"""
Created on 12 May 2017

@author: leem
Based on the ABDB.AbPDB.Chemical_components module by dunbar.

Analyse the chemical component dictionary
http://www.wwpdb.org/ccd.html

These are the types of chemical components in the pdb: ( grep "_chem_comp.type" components.cif | cut -c 50-120 | sort | uniq )

We will bin them into:

peptide:
    "D-beta-peptide, C-gamma linking" 
    "D-gamma-peptide, C-delta linking" 
    "D-peptide linking" 
    "D-PEPTIDE LINKING" 
    "D-peptide NH3 amino terminus" 
    "D-PEPTIDE NH3 AMINO TERMINUS" 
    "L-beta-peptide, C-gamma linking" 
    "L-gamma-peptide, C-delta linking" 
    "L-peptide COOH carboxy terminus" 
    "L-PEPTIDE COOH CARBOXY TERMINUS" 
    "L-peptide linking" 
    "L-PEPTIDE LINKING" 
    "L-peptide NH3 amino terminus" 
    peptide-like 
    PEPTIDE-LIKE 
    "peptide linking" 
    "PEPTIDE LINKING" 

nucleic-acid:
    "DNA linking" 
    "DNA LINKING" 
    "DNA OH 3 prime terminus" 
    "DNA OH 3 PRIME TERMINUS" 
    "L-DNA LINKING"
    "L-RNA LINKING"
    "RNA linking" 
    "RNA LINKING" 
    "RNA OH 3 prime terminus" 

saccharide:
    D-saccharide 
    D-SACCHARIDE 
    "D-saccharide 1,4 and 1,4 linking" 
    "D-SACCHARIDE 1,4 AND 1,4 LINKING" 
    L-saccharide 
    L-SACCHARIDE 
    "L-SACCHARIDE 1,4 AND 1,4 LINKING"
    saccharide 
    SACCHARIDE

non-polymer:
    non-polymer 
    NON-POLYMER 

This has been done in resname_to_type


Common buffers/molecules in the PDB which are unlikely to be hapten antigens 

Method 1:
    This list was taken from the supplementary material (Table 2) of:
    Visualizing ligand molecules in twilight electron density. C. X. Weichenberger, E. Pozharski and B. Rupp. Acta Cryst. (2013). F69, 195-200.
    
    Acknowledgement -  Anthony's JC 03/04/13

Method 2:

    The list of chemical component code to pdb code was taken from:
    http://ligand-expo.rcsb.org/ld-download.html
    Saved as: ./Antibody/AbPDB/dat/Resources/cc-to-pdb.tdd.txt
    
    We look at the number of structures with each code.
    Distribution of number of pdb codes with the chemical component found at:
    ./Antibody/AbPDB/dat/Resources/Frequency_of_cc_in_pdb.pdf
    
    We use the cut-off of 15. (and manually examine those which are over but under 50 and have mainly known Antibodies - just in case it's a pet hapten antigen)
    
    This is a harsh cut off! This is fine for analysing antibodies (as you are unlikely to have more than 15 bound to the same ag). However, change the cutoff for other purposes (suggest at least 200 for "common")
    
    There is still a problem if there is a rarely used code or a newly introduced code for buffer.
    
    fix 120613 will chemical component dict api - runs to online database if cannot find it - obviously requires web access. 
    
    Acknowledgement - JP for suggesting the method.
    The following have been removed from the list as they are either common sugar or peptides
        BGC saccharide  
        NAG saccharide
        XYP saccharide
        XYS saccharide
        MAL saccharide
        MAN saccharide
        GLA saccharide
        GLC saccharide
        A2G saccharide
        LMT saccharide
        PE1 peptide
        F6P saccharide
        DPN peptide
        GAL saccharide
        BOG saccharide
        NGA saccharide
        FUC saccharide
        BMA saccharide
        SUC saccharide
        FUL saccharide
        NDG saccharide

I have updated method 1 list with method 2 list

It is left as a dictionary of dictionaries if we decide to add annotations
    
    
Functions provided are:

is_aa 
is_common_buffer
is_carbohydrate
is_nucleic_acid
is_polymer
get_type
get_chemical_name

each take either a three letter code or a residue object as argument
"""

import sys
from Bio.PDB.Polypeptide import is_aa as bio_is_aa
import re
import tempfile
import urllib.request, urllib.error, urllib.parse

# expourl = "http://ligand-expo.rcsb.org/files/%s/%s/icif/"
expourl = "http://ligand-expo.rcsb.org/reports/%s/%s/%s.cif"
chem_comp_id = re.compile("_chem_comp\.id[ ]+([A-Za-z0-9]+)")
chem_comp_name = re.compile("_chem_comp\.name[ ]+(.+)")
chem_comp_type = re.compile("_chem_comp\.type[ ]+(.+)")

# Try to read the latest chemical components dictionary. If not found then an
# from ABDB import database # does not work as it imports this.
resname_to_name_and_type = {}  # database.get_chemical_components()

# method 1
common_buffers = {
    "PO4": {},
    "SO4": {},
    "GOL": {},
    "EDO": {},
    "BME": {},
    "MPD": {},
    "PEG": {},
    "ACY": {},
    "TRS": {},
    "FMT": {},
    "MES": {},
    "CIT": {},
    "PG4": {},
    "DMS": {},
    "IMD": {},
    "EPE": {},
    "IPA": {},
    "PGE": {},
    "MRD": {},
    "1PE": {},
    "FLC": {},
    "TLA": {},
    "EOH": {},
    "P6G": {},
    "BEN": {},
    "DTT": {},
    "BEZ": {},
    "NHE": {},
    "MLA": {},
    "DOD": {},
    "3PG": {},
    "HTO": {},
    "2PE": {},
    "1PG": {},
    "PGO": {},
    "HEZ": {},
    "PE4": {},
    "MOH": {},
    "15P": {},
    "P33": {},
    "TBU": {},
    "13P": {},
    "ACN": {},
    "12P": {},
    "PG6": {},
    "PGR": {},
    "7PE": {},
    "NH3": {},
    "ETX": {},
    "P4C": {},
    "PG5": {},
    "PE8": {},
    "PG0": {},
    "PE5": {},
    "PDO": {},
    "P22": {},
    "PE3": {},
    "CE1": {},
    "3PO": {},
    "TME": {},
    "9PE": {},
    "8PE": {},
    "PE9": {},
    "PE7": {},
    "PE6": {},
    "PE2": {},
}

# method 2
common_buffers.update(
    {
        "SPM": {},
        "BEN": {},
        "BEF": {},
        "YB": {},
        "EDO": {},
        "MRD": {},
        "APR": {},
        "APC": {},
        "NO2": {},
        "NO3": {},
        "DMS": {},
        "BCL": {},
        "BCT": {},
        "CXS": {},
        "MPD": {},
        "B12": {},
        "MG": {},
        "MN": {},
        "STE": {},
        "NCO": {},
        "OXY": {},
        "ATP": {},
        "MES": {},
        "DIO": {},
        "NAP": {},
        "PER": {},
        "PEG": {},
        "NDP": {},
        "NI": {},
        "NO": {},
        "AZI": {},
        "IPH": {},
        "BME": {},
        "P6G": {},
        "IUM": {},
        "2PE": {},
        "OH": {},
        "NHE": {},
        "PO4": {},
        "C5P": {},
        "ACE": {},
        "ACO": {},
        "ACM": {},
        "ACP": {},
        "ACT": {},
        "ACY": {},
        "TRS": {},
        "TYD": {},
        "DGT": {},
        "UNX": {},
        "PT": {},
        "PB": {},
        "VO4": {},
        "CIT": {},
        "CO": {},
        "CL": {},
        "CS": {},
        "AGS": {},
        "CU": {},
        "CTP": {},
        "DCP": {},
        "BU3": {},
        "BEZ": {},
        "ETX": {},
        "PE4": {},
        "EOH": {},
        "CMO": {},
        "CMP": {},
        "AKG": {},
        "RB": {},
        "SR": {},
        "SF4": {},
        "LDA": {},
        "PEP": {},
        "COA": {},
        "CO3": {},
        "GNP": {},
        "EPE": {},
        "XYL": {},
        "RET": {},
        "CAC": {},
        "PGE": {},
        "PGA": {},
        "PGO": {},
        "PG4": {},
        "PG0": {},
        "SO4": {},
        "FAD": {},
        "AMP": {},
        "TL": {},
        "NH4": {},
        "NH2": {},
        "OCT": {},
        "TLA": {},
        "C8E": {},
        "GSP": {},
        "BA": {},
        "BR": {},
        "GSH": {},
        "PMP": {},
        "FE2": {},
        "FES": {},
        "0QE": {},
        "POP": {},
        "CYN": {},
        "NAD": {},
        "UPG": {},
        "CA": {},
        "CD": {},
        "NAI": {},
        "H4B": {},
        "SCN": {},
        "F09": {},
        "CHD": {},
        "O": {},
        "SIA": {},
        "TAR": {},
        "PLP": {},
        "HC4": {},
        "FMT": {},
        "FMN": {},
        "DTP": {},
        "HEZ": {},
        "HED": {},
        "HEC": {},
        "HEM": {},
        "XE": {},
        "MYR": {},
        "FLC": {},
        "FE": {},
        "SO3": {},
        "CU1": {},
        "CUA": {},
        "TTP": {},
        "SIN": {},
        "ZN": {},
        "TAM": {},
        "SAM": {},
        "U10": {},
        "BTB": {},
        "BTN": {},
        "K": {},
        "F3S": {},
        "IOD": {},
        "HG": {},
        "ADP": {},
        "ADE": {},
        "ADN": {},
        "1PE": {},
        "5GP": {},
        "THP": {},
        "IMP": {},
        "BPH": {},
        "GOL": {},
        "MLA": {},
        "MLI": {},
        "MLT": {},
        "IPA": {},
        "PYR": {},
        "TMP": {},
        "3PG": {},
        "U5P": {},
        "YT3": {},
        "PEO": {},
        "PHQ": {},
        "ANP": {},
        "144": {},
        "NA": {},
        "DTT": {},
        "ALF": {},
        "CDL": {},
        "PLM": {},
        "TPP": {},
        "IMD": {},
    }
)

if not resname_to_name_and_type:
    resname_to_name_and_type = {
        "JN3": ("CHENODEOXYCHOLIC ACID", "non-polymer"),
        "CUY": ("TETRACOSYL OCTADECANOATE", "non-polymer"),  # Added: 2 Aug 2023
        "N36": ("ETHYL VANILLIN", "non-polymer"),  # Added: 3 Aug 2023
        "QOD": (
            "(3R)-N-[(2S,3R)-1-(alpha-D-galactopyranosyloxy)-3-hydroxy-15-methylhexadecan-2-yl]-3-hydroxyheptadecanamide",
            "non-polymer",
        ),  # Added: 3 Aug 2023
        "JN7": (
            "3-CYCLOPENTYL 6-ETHENYL 2-[(THIOPHEN-2-YLACETYL)AMINO]-4,7-DIHYDROTHIENO[2,3-C]PYRIDINE-3,6(5H)-DICARBOXYLATE",
            "non-polymer",
        ),
        "JN5": (
            "(2S)-1,3-BENZOTHIAZOL-2-YL{2-[(2-PYRIDIN-3-YLETHYL)AMINO]PYRIMIDIN-4-YL}ETHANENITRILE",
            "non-polymer",
        ),
        "JN4": (
            "ETHENYL 6-(ETHENYLCARBAMOYL)-2-[(PHENYLACETYL)AMINO]-4,5,6,7-TETRAHYDROTHIENO[2,3-C]PYRIDINE-3-CARBOXYLATE",
            "non-polymer",
        ),
        "JN8": (
            "CYCLOPENTYL 6-(ETHYLCARBAMOYL)-2-[(THIOPHEN-2-YLACETYL)AMINO]-4,5,6,7-TETRAHYDROTHIENO[2,3-C]PYRIDINE-3-CARBOXYLATE",
            "non-polymer",
        ),
        "RMB": ("N1-(5'-PHOSPHO-ALPHA-RIBOSYL)-5-METHYLBENZIMIDAZOLE", "non-polymer"),
        "YM5": (
            "(3-{4-[(2S)-2-(AMINOMETHYL)MORPHOLIN-4-YL]-7H-PYRROLO[2,3-D]PYRIMIDIN-5-YL}PHENYL)METHANOL",
            "non-polymer",
        ),
        "MZ6": (
            "N-[(1S,2R)-3-{(1,3-BENZOTHIAZOL-6-YLSULFONYL)[(2S)-2-METHYLBUTYL]AMINO}-1-BENZYL-2-HYDROXYPROPYL]-3-HYDROXYBENZAMIDE",
            "non-polymer",
        ),
        "PQ1": (
            "PHOSPHORIC ACID MONO-[5-(2-AMINO-5-AMINOMETHYL-4-OXO-3,5-DIHYDRO-4H-PYRIDO[2,3-D]PYRIMIDIN-8-YL)-3,4-DIHYDROXY-TETRAHYDRO-FURAN-2-YLMETHYL] ESTER",
            "nucleic-acid",
        ),
        "NOG": (
            "13-BETA-ETHYL-17-ALPHA-ETHYNYL-17-BETA-HYDROXYGON-4-EN-3-ONE",
            "non-polymer",
        ),
        "KD5": ("4,7-ANHYDRO-3-DEOXY-D-MANNO-OCT-2-ULOSONIC ACID", "non-polymer"),
        "PQ9": (
            "5-[(2E,6E,10E,14E,18E,22E)-3,7,11,15,19,23,27-HEPTAMETHYLOCTACOSA-2,6,10,14,18,22,26-HEPTAENYL]-2,3-DIMETHYLBENZO-1,4-QUINONE",
            "non-polymer",
        ),
        "NDS": ("ETHYL DIMETHYL AMMONIO PROPANE SULFONATE", "non-polymer"),
        "SPX": ("SPHINXOLIDE B", "non-polymer"),
        "SPY": ("PYRIDINE-4-SULFONIC ACID", "non-polymer"),
        "SPZ": (
            "(3S,3'S)-N~1~,N~1~'-BUTANE-1,4-DIYLDIBUTANE-1,3-DIAMINE",
            "non-polymer",
        ),
        "NDR": ("(14BETA,17ALPHA)-17-ETHYNYL-17-HYDROXYESTR-4-EN-3-ONE", "non-polymer"),
        "SPP": (
            "(1-(5-METHANSULPHONAMIDO-1H-INDOL-2-YL-CARBONYL)4-[METHYLAMINO)PYRIDINYL]PIPERAZINE",
            "non-polymer",
        ),
        "KD0": ("3-DEOXY-8-O-PHOSPHONO-D-MANNO-OCT-2-ULOSONIC ACID", "non-polymer"),
        "SPR": ("SPIRAMYCIN I", "non-polymer"),
        "SPS": ("SPARSOMYCIN", "non-polymer"),
        "SPT": ("?", "nucleic-acid"),
        "SPU": (
            "2-{[(R)-{[(2S,3R,4E)-2-AMINO-3-HYDROXYOCTADEC-4-EN-1-YL]OXY}(HYDROXY)PHOSPHORYL]OXY}-N,N,N-TRIMETHYLETHANAMINIUM",
            "non-polymer",
        ),
        "SPV": ("SULFOPYRUVATE", "non-polymer"),
        "NDT": (
            "2-(TOLUENE-4-SULFONYL)-2H-BENZO[D][1,2,3]DIAZABORININ-1-OL",
            "non-polymer",
        ),
        "SPH": ("SPHINGOSINE", "non-polymer"),
        "SPI": (
            "N-HYDROXY-1-(4-METHOXYPHENYL)SULFONYL-4-BENZYLOXYCARBONYL-PIPERAZINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "SPJ": (
            "(3R,3'R)-N~1~,N~1~'-BUTANE-1,4-DIYLDIBUTANE-1,3-DIAMINE",
            "non-polymer",
        ),
        "SPK": ("SPERMINE (FULLY PROTONATED FORM)", "non-polymer"),
        "SPL": (
            "OCTANOIC ACID (2-HYDROXY-1-HYDROXYMETHYL-HEPTADEC-3-ENYL)-AMIDE",
            "non-polymer",
        ),
        "SPM": ("SPERMINE", "non-polymer"),
        "SPN": ("SPEROIDENONE", "non-polymer"),
        "SPO": ("SPHEROIDENE", "non-polymer"),
        "SPA": ("THIOPHENEACETIC ACID", "non-polymer"),
        "SPB": ("4-(4-STYRYL-PHENYLCARBAMOYL)-BUTYRIC ACID", "non-polymer"),
        "IX1": (
            "5-(3-{3-[3-HYDROXY-2-(METHOXYCARBONYL)PHENOXY]PROPENYL}PHENYL)-4-(HYDROXYMETHYL)ISOXAZOLE-3-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "IX6": (
            "(11S)-10-[(2,5-DIMETHYL-1,3-OXAZOL-4-YL)CARBONYL]-11-{2-FLUORO-4-[(2-METHYLPROP-2-EN-1-YL)OXY]PHENYL}-3,3-DIMETHYL-2,3,4,5,10,11-HEXAHYDROTHIOPYRANO[3,2-B][1,5]BENZODIAZEPIN-6-OL 1,1-DIOXIDE",
            "non-polymer",
        ),
        "SPE": ("THERMINE", "non-polymer"),
        "SPF": ("SINAPOYL COENZYME A", "non-polymer"),
        "2FF": (
            "(5-(AMINOMETHYL)-2H-SPIRO[BENZOFURAN-3,4'-PIPERIDINE]-1'-YL)(5-(PHENYLETHYNYL)FURAN-2-YL)METHANONE",
            "non-polymer",
        ),
        "SP8": (
            "N-{(1S)-3-[(4-AMINOBUTYL)AMINO]-1-METHYLPROPYL}BENZAMIDE",
            "non-polymer",
        ),
        "SP9": (
            "N-{(1R)-3-[(4-AMINOBUTYL)AMINO]-1-METHYLPROPYL}BENZAMIDE",
            "non-polymer",
        ),
        "IXH": (
            "8-(2,3-DIHYDRO-1H-INDEN-5-YL)-2-({4-[(3R,5S)-3,5-DIMETHYLPIPERAZIN-1-YL]PHENYL}AMINO)-5-OXO-5,8-DIHYDROPYRIDO[2,3-D]PYRIMIDINE-6-CARBOXAMIDE",
            "non-polymer",
        ),
        "XPH": ("[(1S)-1-AMINO-2-PHENYLETHYL]PHOSPHONIC ACID", "peptide"),
        "IXN": (
            "2-(4-DODECYL-1H-1,2,3-TRIAZOL-1-YL)-5'-O-{[(2-HYDROXYPHENYL)CARBONYL]SULFAMOYL}ADENOSINE",
            "non-polymer",
        ),
        "XPM": ("TETRADECYLPROPANEDIOIC ACID", "non-polymer"),
        "IXM": ("(Z)-1H,1'H-[2,3']BIINDOLYLIDENE-3,2'-DIONE-3-OXIME", "non-polymer"),
        "SP0": (
            "3-FLUORO-N-[1-(4-FLUOROPHENYL)-3-(2-THIENYL)-1H-PYRAZOL-5-YL]BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "SP1": (
            "6-(6-AMINO-PURIN-9-YL)-2-THIOXO-TETRAHYDRO-2-FURO[3,2-D][1,3,2]DIOXAPHOSPHININE-2,7-DIOL",
            "non-polymer",
        ),
        "SP2": ("3,4-DIHYDROSPHEROIDENE", "non-polymer"),
        "SP3": (
            "N-[1-(4-FLUOROPHENYL)-3-(2-THIENYL)-1H-PYRAZOL-5-YL]-3,5-BIS(TRIFLUOROMETHYL)BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "SP5": (
            "N-[3-({4-[(3-AMINOPROPYL)AMINO]BUTYL}AMINO)PROPYL]ACETAMIDE",
            "non-polymer",
        ),
        "IXD": (
            "4-DEOXY-2-O-SULFO-BETA-D-ERYTHRO-HEX-4-ENOPYRANURONIC ACID",
            "non-polymer",
        ),
        "SP7": (
            "N-{(1S)-3-[(4-AMINOBUTYL)AMINO]-1-METHYLPROPYL}ACETAMIDE",
            "non-polymer",
        ),
        "IXX": (
            "3-(5H-DIBENZO[B,F]AZEPIN-5-YL)-N,N-DIMETHYLPROPAN-1-AMINE",
            "non-polymer",
        ),
        "XPX": (
            "(2R)-3-[(HYDROXY{[(2R,3R,5S,6R)-3,4,5-TRIHYDROXY-2,6-BIS(ALPHA-D-MANNOPYRANOSYLOXY)CYCLOHEXYL]OXY}PHOSPHORYL)OXY]PROPANE-1,2-DIYL DIHEXADECANOATE",
            "non-polymer",
        ),
        "W35": (
            "5-(5-(4-(4,5-DIHYDRO-2-OXAZOLY)PHENOXY)PENTYL)-3-METHYL ISOXAZOLE",
            "non-polymer",
        ),
        "Y15": (
            "N-[(1S)-2-{[(1R)-2-(BENZYLOXY)-1-CYANO-1-METHYLETHYL]AMINO}-1-(CYCLOHEXYLMETHYL)-2-OXOETHYL]MORPHOLINE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "IXS": (
            "N-[(1S)-1-BENZYL-2-{[(1S)-2-(ISOBUTYLAMINO)-1-METHYL-2-OXOETHYL]AMINO}ETHYL]-N'-[(1R)-1-(4-FLUOROPHENYL)ETHYL]-5-[METHYL(METHYLSULFONYL)AMINO]ISOPHTHALAMIDE",
            "non-polymer",
        ),
        "XPP": ("HYDROXYPYRROLE-IMIDAZOLE-PYRROLE POLYAMIDE", "non-polymer"),
        "Y11": (
            "N-[(1S)-2-[(4-CYANO-1-METHYLPIPERIDIN-4-YL)AMINO]-1-(CYCLOHEXYLMETHYL)-2-OXOETHYL]MORPHOLINE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "IXW": ("ETHYL (2S)-OXIRAN-2-YLACETATE", "non-polymer"),
        "Y13": (
            "(2E)-3-(3,4-DIHYDROXYPHENYL)-N-[2-(4-HYDROXYPHENYL)ETHYL]ACRYLAMIDE",
            "non-polymer",
        ),
        "XPT": ("1-[CHLORO(DIPROPYL)-LAMBDA~4~-SULFANYL]PROPANE", "non-polymer"),
        "PQ0": (
            "2-AMINO-4-OXO-4,7-DIHYDRO-3H-PYRROLO[2,3-D]PYRIMIDINE-5-CARBONITRILE",
            "non-polymer",
        ),
        "PQB": (
            "[5-AMINO-1-(4-FLUOROPHENYL)-1H-PYRAZOL-4-YL](3-{[(2R)-2,3-DIHYDROXYPROPYL]OXY}PHENYL)METHANONE",
            "non-polymer",
        ),
        "PQC": (
            "3-[(8-PHENYLTHIENO[2,3-H]QUINAZOLIN-2-YL)AMINO]BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "PQE": ("3,6,12,15,18,21,24-HEPTAOXAHEXATRIACONTAN-1-OL", "non-polymer"),
        "JNZ": ("1H-INDAZOL-3-AMINE", "non-polymer"),
        "Z69": ("?", "non-polymer"),
        "PQN": ("PHYLLOQUINONE", "non-polymer"),
        "PQQ": ("PYRROLOQUINOLINE QUINONE", "non-polymer"),
        "PQR": ("4-[METHYL(NITROSO)AMINO]BENZENE-1,2-DIOL", "non-polymer"),
        "Z68": (
            "[4-AMINO-2-(CYCLOHEXYLAMINO)-1,3-THIAZOL-5-YL](3-NITROPHENYL)METHANONE",
            "non-polymer",
        ),
        "JNF": (
            "N-CYCLOHEXYL-4-IMIDAZO[1,2-A]PYRIDIN-3-YL-N-METHYLPYRIMIDIN-2-AMINE",
            "non-polymer",
        ),
        "JNK": (
            "N-{2'-[(4-FLUOROPHENYL)AMINO]-4,4'-BIPYRIDIN-2-YL}-4-METHOXYCYCLOHEXANECARBOXAMIDE",
            "non-polymer",
        ),
        "JNJ": (
            "1-(3-AMINO-1,2-BENZISOXAZOL-5-YL)-6-(4-{2-[(DIMETHYLAMINO)METHYL]-1H-IMIDAZOL-1-YL}-2-FLUOROPHENYL)-7-FLUORO-1H-INDAZOLE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "JNH": (
            "1-[2-(S)-AMINO-3-BIPHENYL-4-YL-PROPIONYL]-PYRROLIDINE-2-(S)-CARBONITRILE",
            "non-polymer",
        ),
        "JNO": (
            "2-{4-[(4-IMIDAZO[1,2-A]PYRIDIN-3-YLPYRIMIDIN-2-YL)AMINO]PIPERIDIN-1-YL}-N-METHYLACETAMIDE",
            "non-polymer",
        ),
        "BEJ": (
            "N,N-[2,5-O-[DI-4-PYRIDIN-3-YL-BENZYL]-GLUCARYL]-DI-[VALYL-AMIDO-METHANE]",
            "non-polymer",
        ),
        "BEK": ("2,3,5,6-TETRAFLUORO-4-METHOXY-BENZAMIDE", "non-polymer"),
        "BEH": (
            "2,5-DIBENZYLOXY-3,4-DIHYDROXY-HEXANEDIOIC ACID BENZYLAMIDE (2-HYDROXY-INDAN-1-YL)-AMIDE",
            "non-polymer",
        ),
        "BEI": (
            "N,N-[2,5-O-[DIBENZYL]-GLUCARYL]-DI-[ISOLEUCYL-AMIDO-METHANE]",
            "non-polymer",
        ),
        "BEN": ("BENZAMIDINE", "non-polymer"),
        "BEO": ("BUTENOIC ACID", "non-polymer"),
        "BEL": (
            "2,4-DINITRO,5-[BIS(2-BROMOETHYL)AMINO]-N-(2',3'-DIOXOPROPYL)BENZAMIDE",
            "non-polymer",
        ),
        "BEM": ("BETA-D-MANNURONIC ACID", "saccharide"),
        "BEB": ("N,N-[2,5-O-DIBENZYL-GLUCARYL]-DI-[1-AMINO-INDAN-2-OL]", "non-polymer"),
        "BEC": (
            "[5-(2-HYDROXY-INDAN-1-YLCARBAMOYL)-3,4-DIHYDROXY-2,5-[DIBENZYL-OXY]-PENTANOYL]-VALINYL-AMIDO-METHANE",
            "non-polymer",
        ),
        "BEA": ("5-METHYL-1,2,4-TRIAZOLO[3,4-B]BENZOTHIAZOLE", "non-polymer"),
        "BEF": ("BERYLLIUM TRIFLUORIDE ION", "non-polymer"),
        "BEG": (
            "2,5-DIBENZYLOXY-3-HYDROXY-HEXANEDIOIC ACID BIS-[(2-HYDROXY-INDAN-1-YL)-AMIDE]",
            "non-polymer",
        ),
        "BED": (
            "N,N-[2,5-O-DI-2-FLUORO-BENZYL-GLUCARYL]-DI-[1-AMINO-INDAN-2-OL]",
            "non-polymer",
        ),
        "BEE": (
            "N,N-[2,5-O-[DI-4-THIOPHEN-3-YL-BENZYL]-GLUCARYL]-DI-[VALYL-AMIDO-METHANE]",
            "non-polymer",
        ),
        "270": ("2H-BENZOIMIDAZOL-2-YLAMINE", "non-polymer"),
        "271": ("N-METHYL-1-PHENYLMETHANAMINE", "non-polymer"),
        "272": ("4-(6-HYDROXY-BENZO[D]ISOXAZOL-3-YL)BENZENE-1,3-DIOL", "non-polymer"),
        "273": ("CYCLOPENTANECARBOXIMIDAMIDE", "non-polymer"),
        "274": (
            "(3Z)-N-(3-CHLOROPHENYL)-3-({3,5-DIMETHYL-4-[(4-METHYLPIPERAZIN-1-YL)CARBONYL]-1H-PYRROL-2-YL}METHYLENE)-N-METHYL-2-OXOINDOLINE-5-SULFONAMIDE",
            "non-polymer",
        ),
        "275": ("5-AMINO-1,2-DIMETHYLPYRIDINIUM", "non-polymer"),
        "276": (
            "4-[[2-[[4-CHLORO-3-(TRIFLUOROMETHYL)PHENYL]AMINO]-3H-BENZIMIDAZOL-5-YL]OXY]-N-METHYL-PYRIDINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "277": (
            "N-(TRANS-4-{(1S,2S)-2-AMINO-3-[(3S)-3-FLUOROPYRROLIDIN-1-YL]-1-METHYL-3-OXOPROPYL}CYCLOHEXYL)-N-METHYLACETAMIDE",
            "non-polymer",
        ),
        "278": ("1-(1-METHYL-1H-PYRROL-2-YL)METHANAMINE", "non-polymer"),
        "279": (
            "N-ETHYL-4-{[5-(METHOXYCARBAMOYL)-2-METHYLPHENYL]AMINO}-5-METHYLPYRROLO[2,1-F][1,2,4]TRIAZINE-6-CARBOXAMIDE",
            "non-polymer",
        ),
        "BEP": ("1-ISOBUTOXY-2-PYRROLIDINO-3[N-BENZYLANILINO] PROPANE", "non-polymer"),
        "BEQ": (
            "N-(CARBOXYMETHYL)-N,N-DIMETHYL-3-[(1-OXODODECYL)AMINO]-1-PROPANAMINIUM INNER SALT",
            "non-polymer",
        ),
        "BEV": ("N-[(2Z)-1,3-THIAZOLIDIN-2-YLIDENE]SULFAMIDE", "non-polymer"),
        "BEW": ("1-HYDROXY-2-SULFANYLPYRIDINIUM", "non-polymer"),
        "BET": ("TRIMETHYL GLYCINE", "non-polymer"),
        "BEU": (
            "N-[(1S,2R,5R,6R)-2-AMINO-5,6-DIHYDROXY-4-(HYDROXYMETHYL)CYCLOHEX-3-EN-1-YL]ACETAMIDE",
            "non-polymer",
        ),
        "GR6": (
            "(3R,3AS,6AR)-HEXAHYDROFURO[2,3-B]FURAN-3-YL {(1S,2R)-1-BENZYL-2-HYDROXY-3-[(7E)-13-METHOXY-1,1-DIOXIDO-3,4,5,6,9,10-HEXAHYDRO-2H-11,1,2-BENZOXATHIAZACYCLOTRIDECIN-2-YL]PROPYL}CARBAMATE",
            "non-polymer",
        ),
        "GR4": ("?", "non-polymer"),
        "Z60": (
            "[4-AMINO-2-(PROP-2-EN-1-YLAMINO)-1,3-THIAZOL-5-YL](5-CHLOROTHIOPHEN-2-YL)METHANONE",
            "non-polymer",
        ),
        "GR3": (
            "3-ACETOXY-17-(1-FORMYL-5-METHYL-3-OXO-HEX-4-ENYL)-16-HYDROXY-4,10,13,14-TETRAMETHYL-2,3,4,5,6,9,10,11,12,13,14,15,16,17-TETRADECAHYDRO-1H-CYCLOPENTA[A]PHENANTHRENE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "GR0": ("GERANYLGERANYL MONOPHOSPHATE", "non-polymer"),
        "GR1": (
            "6-CARBAMIMIDOYL-2-[5-(3-DIETHYLCARBAMOYL-PHENYL)-2-HYDROXY-INDAN-1-YL]-HEXANOIC ACID",
            "non-polymer",
        ),
        "RU9": (
            "2-(4-(3-HYDROXYPHENOXY)PHENYL)-1H-BENZO[D]IMIDAZOLE-5-CARBOXAMIDE",
            "non-polymer",
        ),
        "H1S": ("HEPARIN DISACCHARIDE I-S", "non-polymer"),
        "Y46": (
            "N-(6-AMINO-1-BENZYL-2,4-DIOXO-1,2,3,4-TETRAHYDROPYRIMIDIN-5-YL)BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "GR9": (
            "3-AMINO-6-{4-[(4-METHYLPIPERAZIN-1-YL)SULFONYL]PHENYL}-N-PYRIDIN-3-YLPYRAZINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "P1S": (
            "(6AR,12AR)-3-(HYDROXYMETHYL)-6H-[1,3]DIOXOLO[5,6][1]BENZOFURO[3,2-C]CHROMEN-6A(12AH)-OL",
            "non-polymer",
        ),
        "H1K": ("1-(5-CHLORO-2,4-DIMETHOXYPHENYL)-3-PYRAZIN-2-YLUREA", "non-polymer"),
        "H1L": (
            "(2R)-2-(4-{[3-CHLORO-5-(TRIFLUOROMETHYL)PYRIDIN-2-YL]OXY}PHENOXY)PROPANOIC ACID",
            "non-polymer",
        ),
        "H1M": (
            "METHYL 2-DEOXY-2-(2-HYDROXYETHYL)-ALPHA-D-MANNOPYRANOSIDE",
            "non-polymer",
        ),
        "H1N": (
            "N-[2-OXO-3-((E)-PHENYL{[4-(PIPERIDIN-1-YLMETHYL)PHENYL]IMINO}METHYL)-2,6-DIHYDRO-1H-INDOL-5-YL]ETHANESULFONAMIDE",
            "non-polymer",
        ),
        "H1D": (
            "(2S)-2-AMINO-4-[(2R,3S)-2,3-DIHYDROXY-3-N-HYDROXYCARBAMOYL-PROPYLMERCAPTO]BUTYRIC ACID",
            "non-polymer",
        ),
        "ZPY": (
            "ETHYL 1-{(2S,3S)-3-[(3-{[(1R)-1-(4-FLUOROPHENYL)ETHYL]CARBAMOYL}-5-[METHYL(METHYLSULFONYL)AMINO]BENZOYL)AMINO]-2-HYDROXY-4-PHENYLBUTYL}-1H-PYRAZOLE-4-CARBOXYLATE",
            "non-polymer",
        ),
        "H18": (
            "6-CHLORO-4-(CYCLOHEXYLSULFINYL)-3-PROPYLQUINOLIN-2(1H)-ONE",
            "non-polymer",
        ),
        "GRR": ("(2S)-2-(4-ETHYLPHENOXY)-3-PHENYLPROPANOIC ACID", "non-polymer"),
        "H10": (
            "1-HYDROXY-1-[(2R)-4-{3-[(3S,5S,7S)-TRICYCLO[3.3.1.1~3,7~]DEC-1-YLOXY]PHENYL}BUT-3-YN-2-YL]UREA",
            "non-polymer",
        ),
        "H11": (
            "(5S)-2-(CYCLOOCTYLAMINO)-5-METHYL-5-PROPYL-1,3-THIAZOL-4(5H)-ONE",
            "non-polymer",
        ),
        "H12": ("6-CHLORO-4-(CYCLOHEXYLOXY)-3-PROPYLQUINOLIN-2(1H)-ONE", "non-polymer"),
        "H13": (
            "N-[(1R,3R)-3-HYDROXY-1-(HYDROXYMETHYL)-3-PHENYLPROPYL]TRIDECANAMIDE",
            "non-polymer",
        ),
        "H14": ("(2S,3R)-BETA-HYDROXY-PHENYLALANINE", "peptide"),
        "H15": (
            "N-[(1R,3R)-3-HYDROXY-1-(HYDROXYMETHYL)-3-PHENYLPROPYL]PENTADECANAMIDE",
            "non-polymer",
        ),
        "H16": (
            "6-CHLORO-4-(CYCLOHEXYLSULFANYL)-3-PROPYLQUINOLIN-2(1H)-ONE",
            "non-polymer",
        ),
        "GRF": ("5-O-PHOSPHONO-BETA-D-RIBOFURANOSYLAMINE", "non-polymer"),
        "GRG": ("GERANYLGERANYL DIPHOSPHATE", "non-polymer"),
        "GRD": (
            "BENZYL (S)-1-((2S,3S)-1-(3-(6-AMINO-9H-PURIN-9-YL)PROPYLAMINO)-2-HYDROXY-1-OXOPENTAN-3-YLAMINO)-4-METHYL-1-OXOPENTAN-2-YLCARBAMATE",
            "non-polymer",
        ),
        "GRE": ("2,6-DIHYDROXYBENZOIC ACID", "non-polymer"),
        "GRB": ("?", "nucleic-acid"),
        "GRC": (
            "5'-O-[(R)-{[(R)-[(R)-CHLORO(PHOSPHONO)METHYL](HYDROXY)PHOSPHORYL]OXY}(HYDROXY)PHOSPHORYL]-2'-DEOXYGUANOSINE",
            "non-polymer",
        ),
        "TWT": ("DOCOSANE", "non-polymer"),
        "GRA": ("GLUTARYL-COENZYME A", "non-polymer"),
        "GRO": ("R-2-PHENYL-PROPRIONIC ACID", "non-polymer"),
        "GRL": (
            "(3AS,5R,6AR)-HEXAHYDRO-2H-CYCLOPENTA[B]FURAN-5-YL (2S,3S)-3-HYDROXY-4-(4-(HYDROXYMETHYL)-N-ISOBUTYLPHENYLSULFONAMIDO)-1-PHENYLBUTAN-2-YLCARBAMATE",
            "non-polymer",
        ),
        "GRM": (
            "N-(1,3-BENZODIOXOL-5-YLMETHYL)-1-[(1R)-1-NAPHTHALEN-1-YLETHYL]PIPERIDINE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "GRI": ("6-FORMYL-9-METHOXYPHENAZINE-1-CARBOXYLIC ACID", "non-polymer"),
        "A5A": ("'5'-O-(N-(L-ALANYL)-SULFAMOYL)ADENOSINE", "non-polymer"),
        "27A": (
            "(Z)-2-[2-(4-METHYLPIPERAZIN-1-YL)BENZYL]DIAZENECARBOTHIOAMIDE",
            "non-polymer",
        ),
        "A5C": (
            "(2R,3R,4S,5R)-5-[(1S)-1,2-DIHYDROXYETHYL]OXOLANE-2,3,4-TRIOL",
            "non-polymer",
        ),
        "27C": (
            "5-O-[(S)-BUTOXY(HYDROXY)PHOSPHORYL]-2-DEOXY-2-FLUORO-ALPHA-D-ARABINOFURANOSE",
            "saccharide",
        ),
        "A5D": (
            "(2R,3R,4S,5S)-2-(6-AMINO-9H-PURIN-9-YL)-5-[(BENZYLDISULFANYL)METHYL]TETRAHYDROFURAN-3,4-DIOL",
            "non-polymer",
        ),
        "NOJ": ("1-DEOXYNOJIRIMYCIN", "non-polymer"),
        "NOK": ("2-ACETAMIDO-1,2-DIDEOXYNOJIRMYCIN", "non-polymer"),
        "NOD": ("N-ETHYLHYDROXY-DOXORUBICIN", "non-polymer"),
        "NOE": ("NITROSOETHANE", "non-polymer"),
        "HLU": ("BETA-HYDROXYLEUCINE", "peptide"),
        "HLT": ("2-BROMO-2-CHLORO-1,1,1-TRIFLUOROETHANE", "non-polymer"),
        "A5M": ("2'-AMINE-CYTIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "A5L": (
            "9-(2-DEOXY-2-FLUORO-5-O-PHOSPHONO-BETA-D-ARABINOFURANOSYL)-9H-PURIN-6-AMINE",
            "nucleic-acid",
        ),
        "A5O": ("?", "nucleic-acid"),
        "A5N": ("O-CARBOXY-4-IMINO-L-HOMOSERINE", "peptide"),
        "27P": ("1-[2-(4-BENZYLPHENOXY)ETHYL]PYRROLIDINE", "non-polymer"),
        "A5P": ("ARABINOSE-5-PHOSPHATE", "non-polymer"),
        "BE8": ("?", "non-polymer"),
        "A5R": (
            "AMINO{[(4S)-4-AMINO-9-CARBOXY-5-OXONONYL]AMINO}METHANIMINIUM",
            "peptide",
        ),
        "NOX": ("N-(PARA-GLUTARAMIDOPHENYL-ETHYL)-PIPERIDINIUM-N-OXIDE", "non-polymer"),
        "27U": (
            "N-(4-CARBAMIMIDOYLBENZYL)-1-(3-PHENYLPROPANOYL)-L-PROLINAMIDE",
            "non-polymer",
        ),
        "HLI": (
            "(4-{7-[2-(4-CHLOROPHENOXY)ETHYL]-2-(METHYLAMINO)-6-OXO-6,7-DIHYDRO-1H-PURIN-8-YL}PHENYL)PHOSPHONIC ACID",
            "non-polymer",
        ),
        "SPQ": (
            "(3R,4S,5R)-5-{[(1R)-1-CARBOXY-2-FLUORO-1-(PHOSPHONOOXY)ETHYL]OXY}-4-HYDROXY-3-(PHOSPHONOOXY)CYCLOHEX-1-ENE-1-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "27X": ("AMMOSAMIDE 272", "non-polymer"),
        "BE3": (
            "N,N-[2,5-O-DI-3-FLUORO-BENZYL-GLUCARYL]-DI-[1-AMINO-INDAN-2-OL]",
            "non-polymer",
        ),
        "27Z": (
            "4-AMINO-2-[(4-SULFAMOYLPHENYL)AMINO]-1,3-THIAZOLE-5-CARBOXAMIDE",
            "non-polymer",
        ),
        "BE1": ("N-BROMOACETYL-AMINOETHYL PHOSPHATE", "non-polymer"),
        "BE6": (
            "(2R,3R,4R,5R)-2,5-BIS[(2,5-DIFLUOROBENZYL)OXY]-3,4-DIHYDROXY-N,N'-BIS[(1S,2R)-2-HYDROXY-2,3-DIHYDRO-1H-INDEN-1-YL]HEXANEDIAMIDE",
            "non-polymer",
        ),
        "BE7": ("(4-CARBOXYPHENYL)(CHLORO)MERCURY", "non-polymer"),
        "BE4": (
            "(2R,3R,4R,5R)-2,5-BIS[(2,4-DIFLUOROBENZYL)OXY]-3,4-DIHYDROXY-N,N'-BIS[(1R,2S)-2-HYDROXY-2,3-DIHYDRO-1H-INDEN-1-YL]HEXANEDIAMIDE",
            "non-polymer",
        ),
        "BE5": (
            "(2R,3R,4R,5R)-2,5-BIS[(2,3-DIFLUOROBENZYL)OXY]-3,4-DIHYDROXY-N,N'-BIS[(1S,2R)-2-HYDROXY-2,3-DIHYDRO-1H-INDEN-1-YL]HEXANEDIAMIDE",
            "non-polymer",
        ),
        "FJA": ("FUSICOCCIN J AGLYCONE", "non-polymer"),
        "O1Z": (
            "[4-AMINO-2-(CYCLOHEXYLAMINO)-1,3-THIAZOL-5-YL](NAPHTHALEN-2-YL)METHANONE",
            "non-polymer",
        ),
        "FJD": (
            "3-[(3AS,4R,5S,8AS,8BR)-4-[5-(5-CHLOROTHIOPHEN-2-YL)-1,3-OXAZOL-2-YL]-1,3-DIOXO-4,6,7,8,8A,8B-HEXAHYDRO-3AH-PYRROLO[3,4-A]PYRROLIZIN-2-YL]PROPYL-TRIMETHYL-AZANIUM",
            "non-polymer",
        ),
        "ED2": (
            "N-BENZYL-N-(2-{(4-CYANOPHENYL)[(1-METHYL-1H-IMIDAZOL-5-YL)METHYL]AMINO}ETHYL)-1-METHYL-1H-IMIDAZOLE-4-SULFONAMIDE",
            "non-polymer",
        ),
        "F12": ("2-(ACETYLAMINO)THIOPHENE-3-CARBOXYLIC ACID", "non-polymer"),
        "FJY": (
            "4-[3,3-DIMETHYL-6-(MORPHOLIN-4-YL)-2,3-DIHYDRO-1H-INDOL-1-YL]-7-FLUORO-3-METHYL-2-(PYRIDIN-3-YL)QUINOLINE",
            "non-polymer",
        ),
        "ED1": ("?", "non-polymer"),
        "058": (
            "1-BENZYL-5-CHLORO-3-(2-FLUOROPHENYL)-1H-INDOLE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "059": (
            "(2S)-2-[3-(AMINOMETHYL)PHENYL]-3-{(S)-HYDROXY[(1R)-2-METHYL-1-{[(2-PHENYLETHYL)SULFONYL]AMINO}PROPYL]PHOSPHORYL}PROPANOIC ACID",
            "non-polymer",
        ),
        "ED4": (
            "TERT-BUTYL 4-({(2-{(4-CYANOPHENYL)[(1-METHYL-1H-IMIDAZOL-5-YL)METHYL]AMINO}ETHYL)[(1-METHYL-1H-IMIDAZOL-4-YL)SULFONYL]AMINO}METHYL)PIPERIDINE-1-CARBOXYLATE",
            "non-polymer",
        ),
        "ED5": (
            "TERT-BUTYL 4-({(2-{(4-CYANOPHENYL)[(1-METHYL-1H-IMIDAZOL-5-YL)METHYL]AMINO}ETHYL)[(2-METHYLPHENYL)SULFONYL]AMINO}METHYL)PIPERIDINE-1-CARBOXYLATE",
            "non-polymer",
        ),
        "054": (
            "1-[(2-AMINOPYRIDIN-4-YL)METHYL]-3-(2,4-DIOXO-1,2-DIHYDROTHIENO[3,4-D]PYRIMIDIN-3(4H)-YL)-5-(TRIFLUOROMETHYL)-1H-INDOLE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "055": (
            "(2S)-{[(3-{[(2-CHLORO-6-METHYLPHENYL)CARBAMOYL]AMINO}NAPHTHALEN-2-YL)CARBONYL]AMINO}(PHENYL)ETHANOIC ACID",
            "non-polymer",
        ),
        "056": (
            "3-[(2,4-DICHLOROPHENYL)METHYLSULFANYLMETHYL]BENZOIC ACID",
            "non-polymer",
        ),
        "057": (
            "N-(2-HYDROXY-1,1-DIMETHYLETHYL)-1-METHYL-3-(1H-PYRROLO[2,3-B]PYRIDIN-2-YL)-1H-INDOLE-5-CARBOXAMIDE",
            "non-polymer",
        ),
        "YB": ("YTTERBIUM (III) ION", "non-polymer"),
        "051": (
            "(1R,3R,4S)-4-[({6-[(4-CYANO-2-FLUOROBENZYL)OXY]NAPHTHALEN-2-YL}SULFONYL)AMINO]CYCLOHEXANE-1,3-DICARBOXYLIC ACID",
            "non-polymer",
        ),
        "053": (
            "1-[(2-AMINOPYRIDIN-4-YL)METHYL]-3-(2-OXO-1,2-DIHYDROPYRIDIN-3-YL)-5-(TRIFLUOROMETHYL)-1H-INDOLE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "XP9": ("O-PHOSPHONO-N-(9-SULFANYLNONANOYL)-L-THREONINE", "non-polymer"),
        "T4K": ("?", "non-polymer"),
        "RFN": (
            "(2R)-N-HYDROXY-2-METHYL-2-(METHYLSULFONYL)-4-(2-OXO-4-PHENYLPYRIDIN-1(2H)-YL)BUTANAMIDE",
            "non-polymer",
        ),
        "PLW": (
            "(2S)-2-[5-(1S,2S)-(1-AMINO-2-METHYL-BUTYL)-[1,2,3]TRIAZOL-1-YL]-PROPIONIC ACID",
            "peptide",
        ),
        "XP8": ("O-PHOSPHONO-N-(8-SULFANYLOCTANOYL)-L-THREONINE", "non-polymer"),
        "WSH": (
            "BENZYL [(4R)-1-METHYL-6-PHENYL-4H-[1,2,4]TRIAZOLO[4,3-A][1,4]BENZODIAZEPIN-4-YL]CARBAMATE",
            "non-polymer",
        ),
        "PCW": ("1,2-DIOLEOYL-SN-GLYCERO-3-PHOSPHOCHOLINE", "non-polymer"),
        "MDL": (
            "[DEHYDROXY-N-METHYL-TYROSYL-PROLINYL]-[4,4,5,5,5-PENTAFLUORO-3-OXY-1-[3-INDOLYL]-PENT-2-YL]AMINE",
            "peptide",
        ),
        "LHY": ("L-[(N-HYDROXYAMINO)CARBONYL]PHENYLALANINE", "non-polymer"),
        "PLV": (
            "N-({3-HYDROXY-2-METHYL-5-[(PHOSPHONOOXY)METHYL]PYRIDIN-4-YL}METHYL)-O-METHYL-L-SERINE",
            "non-polymer",
        ),
        "XP1": ("4-(DIMETHYLAMINO)BENZOIC ACID", "non-polymer"),
        "Q27": (
            "7-(2-ETHOXYNAPHTHALEN-1-YL)-6-METHYLQUINAZOLINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "05L": ("HEPTYL 1-SELENO-BETA-D-GLUCOPYRANOSIDE", "non-polymer"),
        "EDC": ("N3,N4-ETHENO-2'-DEOXYCYTIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "05N": ("(3R,4R)-4-HYDROXY-3-METHYL-L-PROLINE", "peptide"),
        "EDA": (
            "3-[2-DEOXY-RIBOFURANOSYL]-3H-1,3,4,5A,8-PENTAAZA-AS-INDACENE-5'-MONOPHOSPHATE",
            "nucleic-acid",
        ),
        "O13": (
            "(E)-4-((1-METHYLPIPERIDIN-3-YLOXYIMINO)METHYL)BENZIMIDAMIDE",
            "non-polymer",
        ),
        "EDG": ("1,4-DIDEOXY-1,4-IMINO-L-ARABINITOL", "non-polymer"),
        "EDD": (
            "2-[3-(3-TERT-BUTYL-5-{[(2,3-DICHLOROPHENYL)CARBAMOYL]IMINO}-2,5-DIHYDRO-1H-PYRAZOL-1-YL)PHENYL]ACETAMIDE",
            "non-polymer",
        ),
        "EDE": ("EDEINE B", "non-polymer"),
        "LMQ": ("(3S)-3-METHYL-L-GLUTAMINE", "peptide"),
        "LMR": ("(2S)-2-HYDROXYBUTANEDIOIC ACID", "non-polymer"),
        "LMS": (
            "[(2R,3S,4R,5R)-5-(6-AMINO-9H-PURIN-9-YL)-3,4-DIHYDROXYTETRAHYDRO-2-FURANYL]METHYL SULFAMATE",
            "nucleic-acid",
        ),
        "EDN": ("ETHANE-1,2-DIAMINE", "non-polymer"),
        "EDO": ("1,2-ETHANEDIOL", "non-polymer"),
        "05B": (
            "2'-[2-(1,3-BENZODIOXOL-5-YL)PYRIMIDIN-4-YL]-5',6'-DIHYDROSPIRO[PIPERIDINE-4,7'-PYRROLO[3,2-C]PYRIDIN]-4'(1'H)-ONE",
            "non-polymer",
        ),
        "EDM": (
            "6-ETHENYL-N,N-DIMETHYL-2-(METHYLSULFONYL)PYRIMIDIN-4-AMINE",
            "non-polymer",
        ),
        "EDR": ("EDROPHONIUM ION", "non-polymer"),
        "X42": (
            "4-{[4-AMINO-5-(3-FLUOROBENZOYL)-1,3-THIAZOL-2-YL]AMINO}BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "EDP": (
            "5-ETHYL-5-[(2R)-PENTAN-2-YL]-2-THIOXODIHYDROPYRIMIDINE-4,6(1H,5H)-DIONE",
            "non-polymer",
        ),
        "LML": ("ISOBUTYL MALONIC ACID", "non-polymer"),
        "EDT": (
            "{[-(BIS-CARBOXYMETHYL-AMINO)-ETHYL]-CARBOXYMETHYL-AMINO}-ACETIC ACID",
            "non-polymer",
        ),
        "05T": ("1-[4-(AMINOMETHYL)CYCLOHEXYL]GUANIDINE", "non-polymer"),
        "LMA": ("LANKAMYCIN", "non-polymer"),
        "CSL": ("(D)-2'-METHYLSELENYL-2'-DEOXYCYTIDINE-5'-PHOSPHATE", "nucleic-acid"),
        "LMC": (
            "(2E,4E,6E)-3-METHYL-6-(1,3,3-TRIMETHYL-1,3-DIHYDRO-2H-INDOL-2-YLIDENE)HEXA-2,4-DIENAL",
            "non-polymer",
        ),
        "LMD": (
            "TETRADECYL 4-O-ALPHA-D-GLUCOPYRANOSYL-BETA-D-GLUCOPYRANOSIDE",
            "non-polymer",
        ),
        "LME": ("(2S,3R)-2-AZANYL-3-METHYL-PENTANEDIOIC ACID", "peptide"),
        "05R": ("2-(3,4-DICHLOROPHENYL)ETHANOIC ACID", "non-polymer"),
        "05S": (
            "(6S)-6,15,15,18-TETRAMETHYL-17-OXO-2,3,4,5,6,7,14,15,16,17-DECAHYDRO-1H-8,12-(METHENO)[1,4,9]TRIAZACYCLOTETRADECINO[9,8-A]INDOLE-9-CARBOXAMIDE",
            "non-polymer",
        ),
        "DPW": ("DECYL 2-TRIMETHYLAZANIUMYLETHYL PHOSPHATE", "non-polymer"),
        "DPV": ("DODECYL 2-(TRIMETHYLAMMONIO)ETHYL PHOSPHATE", "non-polymer"),
        "DPU": (
            "MIXED CARBAMIC PHOSPHORIC ACID ANHYDRIDE OF 7,8-DIAMINONONANIC ACID",
            "non-polymer",
        ),
        "DPT": ("4,7-DIMETHYL-[1,10]PHENANTHROLINE", "non-polymer"),
        "DPS": (
            "3-(1H-INDOL-3-YL)-2-[4-(4-PHENYL-PIPERIDIN-1-YL)-BENZENESULFONYLAMINO]-PROPIONIC ACID",
            "non-polymer",
        ),
        "DPR": ("D-PROLINE", "peptide"),
        "DPQ": (
            "(S)-2-AMINO-3-(4-HYDROXY-3-OXOCYCLOHEXA-1,4-DIENYL)PROPANOIC ACID",
            "peptide",
        ),
        "DPP": ("DIAMINOPROPANOIC ACID", "peptide"),
        "PN1": (
            "(2S,6R)-6-{[(2R)-2-AMINO-2-PHENYLETHANOYL]AMINO}-3,3-DIMETHYL-7-OXO-4-THIA-1-AZABICYCLO[3.2.0]HEPTANE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "PN0": ("PRINOMASTAT", "non-polymer"),
        "PN3": (
            "TERT-BUTYL {(2S)-1-[FORMYL(HYDROXY)AMINO]-3-PHENYLPROPAN-2-YL}CARBAMATE",
            "non-polymer",
        ),
        "PN2": (
            "2-AMINO-3-(HYDROXY-(3-HYDROXY-3-[2-(MERCAPTO-ETHYLCARBAMOYL]-2,2-DIMETHYL-PROPOXY-PHOSPHORYLOXY)-PROPIONIC ACID",
            "non-polymer",
        ),
        "PN5": (
            "(3R)-3-HYDROXY-2,2-DIMETHYL-4-OXO-4-({3-OXO-3-[(2-SULFANYLETHYL)AMINO]PROPYL}AMINO)BUTYL 2,2-DIMETHYLPROPANOATE",
            "non-polymer",
        ),
        "DPZ": ("3,5-DIAMINOPHTHALHYDRAZIDE", "non-polymer"),
        "DPY": ("?", "nucleic-acid"),
        "DPX": (
            "MONO-{4-[(4-AMINO-2-METHYL-PYRIMIDIN-5-YLMETHYL)-AMINO]-2-HYDROXY-3-MERCAPTO-PENT-3-ENYL-PHOSPHONO} ESTER",
            "non-polymer",
        ),
        "DPG": (
            "PHOSPHORIC ACID 2,3-BIS-(3,7,11,15-TETRAMETHYL-HEXADECYLOXY)-PROPYL ESTER 2-HYDROXO-3-PHOSPHONOOXY-PROPYL ESTER",
            "non-polymer",
        ),
        "DPF": ("DIETHYL HYDROGEN PHOSPHATE", "non-polymer"),
        "DPE": ("N-UNDECANYLPHOSPHONATE METHYL ESTER GROUP", "non-polymer"),
        "DPD": ("?", "non-polymer"),
        "DPC": (
            "5-ACETYLAMINO-4-AMINO-6-(PHENETHYL-PROPYL-CARBAMOYL)-5,6-DIHYDRO-4H-PYRAN-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "DPB": (
            "(S)-1-[2'-DEOXY-3',5'-O-(1-PHOSPHONO)BENZYLIDENE-B-D-THREO-PENTOFURANOSYL]THYMINE",
            "nucleic-acid",
        ),
        "DPA": ("?", "non-polymer"),
        "DPO": ("DIPHOSPHATE", "non-polymer"),
        "DPN": ("D-PHENYLALANINE", "peptide"),
        "CF4": (
            "[4-(7,8-DIMETHYL-2,4-DIOXO-3,4-DIHYDROBENZO[G]PTERIDIN-10(2H)-YL)BUTYL]CARBAMIC ACID",
            "non-polymer",
        ),
        "CF5": ("?", "non-polymer"),
        "CF2": (
            "2'-DEOXY-5'-O-{[2-(7,8-DIMETHYL-2,4-DIOXO-3,4-DIHYDROBENZO[G]PTERIDIN-10(2H)-YL)ETHYL]CARBAMOYL}GUANOSINE",
            "non-polymer",
        ),
        "CF3": ("9,9,9-TRIFLUORO-8-OXO-N-PHENYLNONANAMIDE", "non-polymer"),
        "CF0": ("FLUOROMETHANE", "non-polymer"),
        "CF1": (
            "2-{[2-CHLORO-5-(TRIFLUOROMETHYL)PHENYL]AMINO}-5-METHOXYBENZOIC ACID",
            "non-polymer",
        ),
        "LG": ("L-GUANOSINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "LC": ("L-CYTIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "LA": ("LANTHANUM (III) ION", "non-polymer"),
        "LI": ("LITHIUM ION", "non-polymer"),
        "XPC": ("(3S,4R)-4-AMINOPYRROLIDINE-3-CARBOXYLIC ACID", "peptide"),
        "LU": ("LUTETIUM (III) ION", "non-polymer"),
        "XPA": ("1,2-DIDECANOYL-SN-GLYCERO-3-PHOSPHATE", "non-polymer"),
        "A": ("ADENOSINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "BYS": ("2-BENZO[1,3]DIOXOL-5-YLMETHYL-3-BENZYL-SUCCINIC ACID", "non-polymer"),
        "PCT": ("PHOSPHONOACETAMIDE", "non-polymer"),
        "SP6": (
            "BENZYL 6-BENZYL-5,7-DIOXO-6,7-DIHYDRO-5H-[1,3]THIAZOLO[3,2-C]PYRIMIDINE-2-CARBOXYLATE",
            "non-polymer",
        ),
        "MOY": (
            "[(4R)-4-(3-HYDROXYPHENYL)-1,6-DIMETHYL-2-THIOXO-1,2,3,4-TETRAHYDROPYRIMIDIN-5-YL](PHENYL)METHANONE",
            "non-polymer",
        ),
        "PY4": ("2-[O-PHOSPHONOPYRIDOXYL]-AMINO- BUTYRIC ACID", "non-polymer"),
        "PB8": (
            "(2S)-2-[(3R)-3-(ACETYLAMINO)-3-(2-METHYLPROPYL)-2-OXOPYRROLIDIN-1-YL]-N-{(1R,2S)-3-(3,5-DIFLUOROPHENYL)-1-HYDROXY-1-[(3R)-1,2,3,4-TETRAHYDROISOQUINOLIN-3-YL]PROPAN-2-YL}-4-PHENYLBUTANAMIDE",
            "non-polymer",
        ),
        "CFN": ("FE(7)-MO-S(9)-N CLUSTER", "non-polymer"),
        "CFO": ("CHLORO DIIRON-OXO MOIETY", "non-polymer"),
        "CFL": (
            "4-AMINO-1-(2-DEOXY-2-FLUORO-5-O-PHOSPHONO-BETA-D-ARABINOFURANOSYL)PYRIMIDIN-2(1H)-ONE",
            "nucleic-acid",
        ),
        "CFM": ("FE-MO-S CLUSTER", "non-polymer"),
        "DP3": (
            "N-{(4S)-4-AMINO-5-[(2-AMINOETHYL)AMINO]PENTYL}-N'-NITROGUANIDINE",
            "non-polymer",
        ),
        "CFK": ("3-(CARBAMOYLAMINO)-5-PHENYLTHIOPHENE-2-CARBOXAMIDE", "non-polymer"),
        "CFH": ("1,1,1,3,3,3-HEXAFLUOROPROPAN-2-OL", "non-polymer"),
        "CFI": ("1-HYDROXY-2-OXO-1-PHENYLHYDRAZINE", "non-polymer"),
        "CFF": ("CAFFEINE", "non-polymer"),
        "PNP": ("METHYL-PHOSPHONIC ACID MONO-(4-NITRO-PHENYL) ESTER", "non-polymer"),
        "CFD": ("(2R,3E,5R)-5-AMINO-2,4-DIMETHYLHEX-3-ENAL", "non-polymer"),
        "CFE": (
            "(8R)-3-BETA-D-RIBOFURANOSYL-3,6,7,8-TETRAHYDROIMIDAZO[4,5-D][1,3]DIAZEPIN-8-OL",
            "non-polymer",
        ),
        "CFB": (
            "2-CHLORO-9-(2-DEOXY-2-FLUORO-B -D-ARABINOFURANOSYL)-9H-PURIN-6-AMINE",
            "non-polymer",
        ),
        "CFC": ("1-CARBOXY-1'-[(DIMETHYLAMINO)-CARBONYL]FERROCENE", "non-polymer"),
        "DP9": ("L-N(OMEGA)-NITROARGININE-(4R)-AMINO-L-PROLINE AMIDE", "non-polymer"),
        "CFA": ("(2,4-DICHLOROPHENOXY)ACETIC ACID", "non-polymer"),
        "PNI": ("TETRA[N-METHYL-PYRIDYL] PORPHYRIN-NICKEL", "non-polymer"),
        "PNH": ("PENTAN-2-ONE", "non-polymer"),
        "PNK": (
            "(2R,4S)-2-{(R)-CARBOXY[(PHENYLACETYL)AMINO]METHYL}-5,5-DIMETHYL-1,3-THIAZOLIDINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "PNJ": ("PNP-BETA-D-GLUCOSAMINE", "non-polymer"),
        "CFZ": ("?", "nucleic-acid"),
        "PNL": ("PROPANAL", "non-polymer"),
        "CFX": ("CEFOXITIN", "non-polymer"),
        "CFY": (
            "[(4Z)-2-{(2R)-2-[(1S)-1-AMINO-2-PHENYLETHYL]-2-HYDROXY-2,5-DIHYDRO-1,3-THIAZOL-4-YL}-4-(4-HYDROXYBENZYLIDENE)-5-OXO-4,5-DIHYDRO-1H-IMIDAZOL-1-YL]ACETIC ACID",
            "peptide",
        ),
        "CFV": (
            "[(2R)-1-(4-AMINO-2-OXO-PYRIMIDIN-1-YL)-3-HYDROXY-PROPAN-2-YL]OXYMETHYLPHOSPHONIC ACID",
            "non-polymer",
        ),
        "Y19": (
            "1-DEOXY-1-{[(5S)-2,6-DIOXO-5-(PROPANOYLAMINO)-1,2,5,6-TETRAHYDROPYRIMIDIN-4-YL]AMINO}-D-RIBITOL",
            "non-polymer",
        ),
        "CFT": ("TRIFLUOROMETHANE", "non-polymer"),
        "PNB": ("4-NITRO-BENZYLPHOSPHONOBUTANOYL-GLYCINE", "non-polymer"),
        "PNE": ("PARA-NITROPHENYL PHOSPHONOBUTANOYL L-ALANINE", "non-polymer"),
        "CFS": (
            "4-AMINO-5-CYCLOHEXYL-2,2-DIFLUORO-3-HYDROXY-PENTANOIC ACID",
            "non-polymer",
        ),
        "CFP": ("6-CHLORO-2-FLUOROPURINE", "non-polymer"),
        "CFQ": ("1-(2-NITROPHENYL)-2,2,2-TRIFLUOROETHYL]-ARSENOCHOLINE", "non-polymer"),
        "AP8": ("APLYRONINE A", "non-polymer"),
        "AP9": (
            "(2S)-2-({6-[(3-AMINO-5-CHLOROPHENYL)AMINO]-9-ISOPROPYL-9H-PURIN-2-YL}AMINO)-3-METHYLBUTAN-1-OL",
            "non-polymer",
        ),
        "SM": ("SAMARIUM (III) ION", "non-polymer"),
        "AP2": ("PHOSPHOMETHYLPHOSPHONIC ACID ADENOSYL ESTER", "non-polymer"),
        "AP3": (
            "2-(1-CARBOXY-2-HYDROXY-2-METHYL-PROPYL)-5,5-DIMETHYL-THIAZOLIDINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "AP0": ("ACETYL PYRIDINE ADENINE DINUCLEOTIDE, REDUCED", "non-polymer"),
        "AP1": (
            "{3-[3-(3,4-DIMETHOXY-PHENYL)-1-(1-{1-[2-(3,4,5-TRIMETHOXY-PHENYL)-BUTYRYL]-PIPERIDIN-2YL}-VINYLOXY)-PROPYL]-PHENOXY}-ACETIC ACID",
            "non-polymer",
        ),
        "AP6": ("2,4-DIAMINO-6-PHENYL-5,6,7,8,-TETRAHYDROPTERIDINE", "non-polymer"),
        "AP7": ("?", "nucleic-acid"),
        "AP4": (
            "7-AMINO-3,3A,4,5-TETRAHYDRO-8H-2-OXA-5,6,8,9B-TETRAAZA-CYCLOPENTA[A]NAPHTHALENE-1,9-DIONE",
            "non-polymer",
        ),
        "AP5": ("BIS(ADENOSINE)-5'-PENTAPHOSPHATE", "non-polymer"),
        "U8U": ("5-METHYLAMINOMETHYL-2-THIOURIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "MDG": ("7N-METHYL-8-HYDROGUANOSINE-5'-DIPHOSPHATE", "non-polymer"),
        "BYG": (
            "L-GAMMA-GLUTAMYL-S-{(4R)-4-[(6-HYDROXYHEXYL)SULFANYL]-7-NITRO-4,5-DIHYDRO-2,1,3-BENZOXADIAZOL-4-YL}-L-CYSTEINYLGLYCINE",
            "non-polymer",
        ),
        "IJ1": ("2-[(3-METHYLBENZYL)AMINO]ETHANESULFONIC ACID", "non-polymer"),
        "3RC": (
            "3-(1-BENZYL-1H-1,2,3-TRIAZOL-4-YL)-1H-PYRROLO[2,3-B]PYRIDINE",
            "non-polymer",
        ),
        "3RA": (
            "5-(1,3-BENZODIOXOL-5-YLMETHYL)-2-(PHENYLAMINO)-4H-IMIDAZOL-4-ONE",
            "non-polymer",
        ),
        "MR8": ("28-O-METHYLRAPAMYCIN", "non-polymer"),
        "3RG": ("2-[(CARBOXYMETHYL)AMINO]BENZOIC ACID", "non-polymer"),
        "3RD": (
            "4-[(E)-{4-FORMYL-5-HYDROXY-6-METHYL-3-[(PHOSPHONOOXY)METHYL]PYRIDIN-2-YL}DIAZENYL]BENZOIC ACID",
            "non-polymer",
        ),
        "3RE": (
            "2-AMINO-8-(TRANS-4-METHOXYCYCLOHEXYL)-4-METHYL-6-(1H-PYRAZOL-3-YL)PYRIDO[2,3-D]PYRIMIDIN-7(8H)-ONE",
            "non-polymer",
        ),
        "MR4": ("4-(1,3-BENZOXAZOL-2-YL)-2,6-DIMETHYLPHENOL", "non-polymer"),
        "3RK": ("(3S,4R,5R,6S)-AZEPANE-3,4,5,6-TETROL", "non-polymer"),
        "3RH": ("(6S)-6-PHENYL-5,6-DIHYDROBENZO[H]QUINAZOLIN-2-AMINE", "non-polymer"),
        "3RI": (
            "(3S,4R,5R,6S)-1-(2-HYDROXYETHYL)AZEPANE-3,4,5,6-TETROL",
            "non-polymer",
        ),
        "3RN": ("2-(HYDROXYMETHYL)CYCLOPENT-2-EN-1-ONE", "non-polymer"),
        "MR1": ("1-(1,3-BENZODIOXOL-5-YL)METHANAMINE", "non-polymer"),
        "MR2": ("3-DIPHENOL-6-NITRO-3H-BENZO[DE]ISOCHROMEN-1-ONE", "non-polymer"),
        "3RM": (
            "5-{4-[2-(4-BROMOPHENOXY)ETHYL]PIPERAZIN-1-YL}-4H-1,2,4-TRIAZOL-3-AMINE",
            "non-polymer",
        ),
        "3RR": (
            "4-(2,4-DICHLORO-5-METHOXYPHENYL)-2,6-DIMETHYL-7H-PYRROLO[2,3-D]PYRIMIDINE-5-CARBONITRILE",
            "non-polymer",
        ),
        "3RS": (
            "(2R)-3-[2-AMINO-6-(2-METHYLPHENYL)QUINOLIN-3-YL]-N-[(4R)-2,2-DIMETHYLTETRAHYDRO-2H-PYRAN-4-YL]-2-METHYLPROPANAMIDE",
            "non-polymer",
        ),
        "3RP": (
            "4-(2,4-DICHLORO-5-METHOXYPHENYL)-6-METHYLPYRIMIDIN-2-AMINE",
            "non-polymer",
        ),
        "3RQ": (
            "4-(2,4-DICHLORO-5-METHOXYPHENYL)-2-METHYL-7H-PYRROLO[2,3-D]PYRIMIDINE-5-CARBONITRILE",
            "non-polymer",
        ),
        "3RW": (
            "[4-AMINO-7-(PROPAN-2-YL)-7H-PYRROLO[2,3-D]PYRIMIDIN-5-YL](6-{[2-(PYRIDIN-3-YL)ETHYL]AMINO}PYRAZIN-2-YL)METHANONE",
            "non-polymer",
        ),
        "3RU": (
            "3-(2-AMINOQUINOLIN-3-YL)-N-(CYCLOHEXYLMETHYL)PROPANAMIDE",
            "non-polymer",
        ),
        "3RZ": (
            "4-AMINO-2-METHYL-N-(1H-PYRAZOL-3-YL)QUINAZOLINE-8-CARBOXAMIDE",
            "non-polymer",
        ),
        "IYG": ("N-ALPHA-ACETYL-3,5-DIIODOTYROSYLGLYCINE", "non-polymer"),
        "3RX": (
            "N-[(2E)-3,7-DIMETHYLOCTA-2,6-DIEN-1-YL]-N'-[(1R,3S,5R,7R)-TRICYCLO[3.3.1.1~3,7~]DEC-2-YL]ETHANE-1,2-DIAMINE",
            "non-polymer",
        ),
        "BYA": ("N-(PHENYLCARBONYL)-BETA-ALANINE", "non-polymer"),
        "T4B": ("?", "non-polymer"),
        "MRX": (
            "5-BROMO-3-(PYRROLIDIN-1-YLSULFONYL)-1H-INDOLE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "MRY": ("MESO-ERYTHRITOL", "non-polymer"),
        "PQA": (
            "[5-AMINO-1-(4-FLUOROPHENYL)-1H-PYRAZOL-4-YL][3-(PIPERIDIN-4-YLOXY)PHENYL]METHANONE",
            "non-polymer",
        ),
        "MRT": (
            "(5R,9S,12S,15S,18S,21S)-21-BENZYL-12,18-BIS(CARBOXYMETHYL)-15-CYCLOHEXYL-1-(9H-FLUOREN-9-YL)-4-METHYL-9-(2-METHYLPROPYL)-3,6,10,13,16,19-HEXAOXO-5-PHENYL-2-OXA-4,8,11,14,17,20-HEXAAZADOCOSAN-22-OIC ACID",
            "non-polymer",
        ),
        "MRU": ("5-S-METHYL-1-O-PHOSPHONO-5-THIO-D-RIBULOSE", "non-polymer"),
        "SF8": (
            "(2R)-2-(2-{[(1R)-1-CARBOXY-4-{[(3S)-3,4-DICARBOXY-3-HYDROXYBUTANOYL]AMINO}BUTYL]AMINO}-2-OXOETHYL)-2-HYDROXYBUTANEDIOIC ACID",
            "non-polymer",
        ),
        "MRP": ("3-O-METHYL-ALPHA-L-RHAMNOPYRANOSIDE", "saccharide"),
        "MRR": ("(R)-2-METHYLMYRISTOYL-COENZYME A", "non-polymer"),
        "M1L": (
            "2-({[(3E)-5-{4-[(DIMETHYLAMINO)(DIHYDROXY)-LAMBDA~4~-SULFANYL]PHENYL}-8-METHYL-2-OXO-6,7,8,9-TETRAHYDRO-1H-PYRROLO[3,2-H]ISOQUINOLIN-3(2H)-YLIDENE]AMINO}OXY)-4-HYDROXYBUTANOIC ACID",
            "non-polymer",
        ),
        "Z78": (
            "3-(2-FLUOROPHENYL)-1-(2-HYDROXYETHYL)-1-[(6-METHYL-2-OXO-1,2-DIHYDROQUINOLIN-3-YL)METHYL]UREA",
            "non-polymer",
        ),
        "MDE": ("5-MERCAPTOETHANOL-2-DECENOYL-COENZYME A", "non-polymer"),
        "MRN": ("2-METHYL-N-[3-(1-METHYLETHOXY)PHENYL]BENZAMIDE", "non-polymer"),
        "3R5": ("5-METHOXY-2-(PYRIDIN-3-YLMETHOXY)BENZALDEHYDE", "non-polymer"),
        "MRD": ("(4R)-2-METHYLPENTANE-2,4-DIOL", "non-polymer"),
        "MRE": ("2-(4-DIMETHYLAMINOPHENYL)DIAZENYLBENZOIC ACID", "non-polymer"),
        "T4A": ("?", "non-polymer"),
        "MRG": (
            "N2-(3-MERCAPTOPROPYL)-2'-DEOXYGUANOSINE-5'-MONOPHOSPHATE",
            "nucleic-acid",
        ),
        "Z74": (
            "N'-{(1S,2S)-1-(3,5-DIFLUOROBENZYL)-2-HYDROXY-2-[(2R)-4-(PHENYLSULFONYL)PIPERAZIN-2-YL]ETHYL}-5-METHYL-N,N-DIPROPYLBENZENE-1,3-DICARBOXAMIDE",
            "non-polymer",
        ),
        "MRA": (
            "2-[(4-ETHYNYL-2-FLUOROPHENYL)AMINO]-3,4-DIFLUORO-N-(2-HYDROXYETHOXY)BENZAMIDE",
            "non-polymer",
        ),
        "Z76": (
            "N-[(1S,2S)-1-(3,5-DIFLUOROBENZYL)-2-HYDROXY-2-{(2R)-4-[(3-METHYLPHENYL)SULFONYL]PIPERAZIN-2-YL}ETHYL]-3-{[(2R)-2-(METHOXYMETHYL)PYRROLIDIN-1-YL]CARBONYL}-5-METHYLBENZAMIDE",
            "non-polymer",
        ),
        "MRC": ("MUPIROCIN", "non-polymer"),
        "APZ": ("4-AMINOPHTHALHYDRAZIDE", "non-polymer"),
        "APX": ("2'-MONOPHOSPHOADENOSINE-5'-DIPHOSPHORIBOSE", "non-polymer"),
        "APY": ("2-AMINOMETHYL-PYRIDINE", "non-polymer"),
        "APR": ("ADENOSINE-5-DIPHOSPHORIBOSE", "non-polymer"),
        "APS": ("9-HYDROXYPROPYLADENINE, S-ISOMER", "non-polymer"),
        "APP": ("1-ACETYL-2-CARBOXYPIPERIDINE", "peptide"),
        "APQ": (
            "2,6-DIAMINO-8-PROPYLSULFANYLMETHYL-3H-QUINAZOLINE-4-ONE",
            "non-polymer",
        ),
        "APV": (
            "6-(5-AMINO-5-CARBOXY-PENTANOYLAMINO)-3-HYDROXYMETHYL-7-OXO-4-THIA-1-AZA-BICYCLO[3.2.0]HEPTANE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "APW": (
            "{5'-O-[(R)-{[(S)-AMINO(HYDROXY-KAPPAO)PHOSPHORYL]OXY}(HYDROXY-KAPPAO)PHOSPHORYL]ADENOSINATO(2-)}MAGNESIUM",
            "non-polymer",
        ),
        "APT": (
            "2-AMINO-6-[(4-CARBOXY-PHENYLAMINO)-METHYL]-4-HYDROXY-PTERIDIN-1-IUM",
            "non-polymer",
        ),
        "APU": ("ADENYLYL-3'-5'-PHOSPHO-URIDINE-3'-MONOPHOSPHATE", "non-polymer"),
        "APJ": (
            "N~2~-1H-BENZIMIDAZOL-5-YL-N~4~-(3-CYCLOPROPYL-1H-PYRAZOL-5-YL)PYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "APK": (
            "5'-O-[(S)-{[(5S)-5-AMINO-6-OXOHEXYL]AMINO}(HYDROXY)PHOSPHORYL]ADENOSINE",
            "peptide",
        ),
        "APH": ("P-AMIDINOPHENYL-3-ALANINE", "peptide"),
        "API": ("2,6-DIAMINOPIMELIC ACID", "peptide"),
        "APN": ("2-AMINOETHYLGLYCINE-CARBONYLMETHYLENE-ADENINE", "peptide"),
        "APO": ("D-2-AMINO-3-PHOSPHONO-PROPIONIC ACID", "peptide"),
        "APL": (
            "N-(1-BENZYL-3,3,3-TRIFLUORO-2,2-DIHYDROXY-PROPYL)-ACETAMIDE",
            "non-polymer",
        ),
        "APM": ("M-AMIDINOPHENYL-3-ALANINE", "peptide"),
        "APB": ("M-AMINOPHENYLBORONIC ACID", "non-polymer"),
        "APC": ("DIPHOSPHOMETHYLPHOSPHONIC ACID ADENOSYL ESTER", "non-polymer"),
        "APA": (
            "(2S)-3-(4-CARBAMIMIDOYLPHENYL)-2-HYDROXYPROPANOIC ACID",
            "non-polymer",
        ),
        "APF": (
            "1,1,1-TRIFLUORO-3-ACETAMIDO-4-PHENYL BUTAN-2-ONE(N-ACETYL-L-PHENYLALANYL TRIFLUOROMETHYL KETONE)",
            "non-polymer",
        ),
        "APG": ("ATROLACTIC ACID (2-PHENYL-LACTIC ACID)", "non-polymer"),
        "APD": ("3-METHYLPHENYLALANINE", "non-polymer"),
        "APE": ("(1-AMINO-2-PHENYL-ETHYL)-CARBAMIC ACID", "peptide"),
        "0RF": (
            "(2S)-2-(4-CHLOROPHENYL)-1-{4-[(5R,7R)-7-HYDROXY-5-METHYL-6,7-DIHYDRO-5H-CYCLOPENTA[D]PYRIMIDIN-4-YL]PIPERAZIN-1-YL}-3-(PROPAN-2-YLAMINO)PROPAN-1-ONE",
            "non-polymer",
        ),
        "0RA": (
            "3-[3-(1H-PYRROLO[2,3-B]PYRIDIN-5-YL)PHENYL]PROPANOIC ACID",
            "non-polymer",
        ),
        "0RB": (
            "2-AMINO-4-OXO-N-(1H-1,2,3-TRIAZOL-5-YLMETHYL)-1,4-DIHYDROPTERIDINE-7-CARBOXAMIDE",
            "non-polymer",
        ),
        "0RM": (
            "(2R,4S)-2-{(R)-CARBOXY[(2,6-DIMETHOXYBENZOYL)AMINO]METHYL}-5,5-DIMETHYL-1,3-THIAZOLIDINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "0RO": (
            "4-[4-(5,5-DIMETHYL-4,5-DIHYDRO-1,3-THIAZOL-2-YL)PIPERAZIN-1-YL]-6-PROPYLTHIENO[2,3-D]PYRIMIDINE",
            "non-polymer",
        ),
        "0RN": ("SULBACTAM", "non-polymer"),
        "0RU": (
            "8-METHYL-2-[(PYRIMIDIN-2-YLSULFANYL)METHYL]QUINAZOLIN-4(1H)-ONE",
            "non-polymer",
        ),
        "0RT": (
            "4-[4-(5,5-DIMETHYL-4,5-DIHYDRO-1,3-THIAZOL-2-YL)PIPERAZIN-1-YL]-6-(2,2,2-TRIFLUOROETHYL)THIENO[2,3-D]PYRIMIDINE",
            "non-polymer",
        ),
        "391": (
            "4-(1,3-BENZODIOXOL-5-YLOXY)-2-[4-(1H-IMIDAZOL-1-YL)PHENOXY]-6-METHYLPYRIMIDINE",
            "non-polymer",
        ),
        "0RV": (
            "(2S)-2-[(1S,2R)-1-CARBOXY-2-HYDROXYPROPYL]-4-{[(3S,5S)-5-(DIMETHYLCARBAMOYL)PYRROLIDIN-3-YL]SULFANYL}-3-METHYL-2H-PYRROLE-5-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "0RQ": ("DETHIAACETYL-COENZYME A", "non-polymer"),
        "0RP": (
            "ETHYL (2R,4S)-4-{[3,5-BIS(TRIFLUOROMETHYL)BENZYL](METHOXYCARBONYL)AMINO}-2-ETHYL-6-(TRIFLUOROMETHYL)-3,4-DIHYDROQUINOLINE-1(2H)-CARBOXYLATE",
            "non-polymer",
        ),
        "0RS": (
            "{5-[4-(DIMETHYLAMINO)PIPERIDIN-1-YL]-1H-IMIDAZO[4,5-B]PYRIDIN-2-YL}[2-(ISOQUINOLIN-4-YL)PYRIDIN-4-YL]METHANONE",
            "non-polymer",
        ),
        "0RR": (
            "3-{(5Z)-5-[5-CHLORO-1-(2,6-DICHLOROBENZYL)-2-OXO-1,2-DIHYDRO-3H-INDOL-3-YLIDENE]-4-OXO-2-THIOXO-1,3-THIAZOLIDIN-3-YL}PROPANOIC ACID",
            "non-polymer",
        ),
        "KX1": (
            "4-CYCLOHEPTYL-6-(3-PIPERIDIN-1-YLPROPYL)PYRIMIDINE-2-CARBONITRILE",
            "non-polymer",
        ),
        "KX2": (
            "(4E,6Z,8S,9S,10E,12S,13R,14S,16R)-13-HYDROXY-8,14-DIMETHOXY-4,10,12,16-TETRAMETHYL-3,20,22-TRIOXO-19-{[2-(PYRROLIDIN-1-YL)ETHYL]AMINO}-2-AZABICYCLO[16.3.1]DOCOSA-1(21),4,6,10,18-PENTAEN-9-YL CARBAMATE",
            "non-polymer",
        ),
        "398": (
            "N,N-DIBUTYL-4-CHLORANYL-1-[2-(3,4-DIHYDRO-1H-ISOQUINOLIN-2-YLCARBONYL)-4-[(7-IODANYLNAPHTHALEN-2-YL)SULFONYLCARBAMOYL]PHENYL]-5-METHYL-PYRAZOLE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "0RY": (
            "(2Z)-4-[(3-CARBAMOYLPHENYL)AMINO]-4-OXOBUT-2-ENOIC ACID",
            "non-polymer",
        ),
        "0RX": (
            "2-(MORPHOLIN-4-YL)-N-[4-(PYRIDIN-4-YLOXY)PHENYL]PYRIDINE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "0RZ": (
            "(2E)-4-[(3-CARBAMOYLPHENYL)AMINO]-4-OXOBUT-2-ENOIC ACID",
            "non-polymer",
        ),
        "B38": (
            "1'-(1H-INDAZOL-7-YLCARBONYL)-6-METHYLSPIRO[CHROMENE-2,4'-PIPERIDIN]-4(3H)-ONE",
            "non-polymer",
        ),
        "B39": (
            "[(1S,2S,4R)-4-{4-[(1S)-2,3-DIHYDRO-1H-INDEN-1-YLAMINO]-7H-PYRROLO[2,3-D]PYRIMIDIN-7-YL}-2-HYDROXYCYCLOPENTYL]METHYL SULFAMATE",
            "non-polymer",
        ),
        "X5S": (
            "BETA-D-XYLOPYRANOSYL-(1->4)-4-THIO-BETA-D-XYLOPYRANOSYL-(1->4)-4-THIO-BETA-D-XYLOPYRANOSYL-(1->4)-4-THIO-BETA-D-XYLOPYRANOSYL-(1->4)-4-THIO-BETA-D-XYLOPYRANOSE",
            "saccharide",
        ),
        "B34": (
            "(5S)-1-BENZYL-3-(1,1-DIOXIDO-1,2-BENZISOTHIAZOL-3-YL)-4-HYDROXY-5-(1-METHYLETHYL)-1,5-DIHYDRO-2H-PYRROL-2-ONE",
            "non-polymer",
        ),
        "B35": (
            "4-[(1S)-1-(3-FLUORO-4-METHOXYPHENYL)-2-(2-METHOXY-5-NITROPHENYL)ETHYL]-1H-IMIDAZOL-2-AMINE",
            "non-polymer",
        ),
        "B36": (
            "6-{[1-(ANTHRACEN-9-YLCARBONYL)PIPERIDIN-4-YL]METHYL}-2-METHYLQUINOLINE",
            "non-polymer",
        ),
        "B37": (
            "4-({4-[(2-METHYLQUINOLIN-6-YL)METHYL]PIPERIDIN-1-YL}CARBONYL)-2-PHENYLQUINOLINE",
            "non-polymer",
        ),
        "B30": (
            "{2,2'-[(2-{[4-(AMINOSULFONYL)BENZOYL]AMINO}ETHYL)IMINO]DIACETATO(2-)-KAPPAO}COPPER",
            "non-polymer",
        ),
        "B31": (
            "(2S)-4-AMINO-N-[(1R,2S,3R,4R,5S)-5-AMINO-4-[(2,6-DIAMINO-2,6-DIDEOXY-ALPHA-D-GLUCOPYRANOSYL)OXY]-2-HYDROXY-3-(BETA-D-XYLOFURANOSYLOXY)CYCLOHEXYL]-2-HYDROXYBUTANAMIDE",
            "non-polymer",
        ),
        "B32": ("5'-S-[2-(DECYLAMINO)ETHYL]-5'-THIOADENOSINE", "non-polymer"),
        "B33": ("N-ETHYL-N-[3-(PROPYLAMINO)PROPYL]PROPANE-1,3-DIAMINE", "non-polymer"),
        "I3K": (
            "N-(4-CHLOROPHENYL)-5-[(6,7-DIMETHOXYQUINOLIN-4-YL)OXY]-1,3-BENZOXAZOL-2-AMINE",
            "non-polymer",
        ),
        "7EP": (
            "(2R,4S)-2-[(1R)-1-{[(2,6-DIMETHOXYPHENYL)CARBONYL]AMINO}-2-OXOETHYL]-5,5-DIMETHYL-1,3-THIAZOLIDINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "I3N": (
            "1-BENZYL-5-METHOXY-2-METHYL-1H-INDOL-3-YL)-ACETIC ACID",
            "non-polymer",
        ),
        "I3M": ("TRI-IODODE ANION", "non-polymer"),
        "PLZ": (
            "4-[({3-HYDROXY-2-METHYL-5-[(PHOSPHONOOXY)METHYL]PYRIDIN-4-YL}METHYL)AMINO]BUTANOIC ACID",
            "non-polymer",
        ),
        "I3C": ("5-AMINO-2,4,6-TRIIODOBENZENE-1,3-DICARBOXYLIC ACID", "non-polymer"),
        "I3A": ("1H-INDOLE-3-CARBALDEHYDE", "non-polymer"),
        "I3E": ("1-(4-ETHYLPHENYL)PROPAN-1-ONE", "non-polymer"),
        "X55": (
            "(2R)-3-BENZYL-2-(2-BROMO-4-HYDROXY-5-METHOXYPHENYL)-1,3-THIAZOLIDIN-4-ONE",
            "non-polymer",
        ),
        "I39": (
            "[3-AMINO-2-(2-METHYLPHENYL)-1-OXIDOPYRIDIN-4-YL](2,4-DIFLUOROPHENYL)METHANONE",
            "non-polymer",
        ),
        "I38": ("N-(3-MERCAPTOPROPANOYL)-D-ALANINE", "non-polymer"),
        "B3X": ("(3S)-3,5-DIAMINO-5-OXOPENTANOIC ACID", "peptide"),
        "B3Y": ("(3S)-3-AMINO-4-(4-HYDROXYPHENYL)BUTANOIC ACID", "peptide"),
        "B3T": ("3-AMINO-2,3,5-TRIDEOXY-D-THREO-PENTONIC ACID", "peptide"),
        "B3U": ("2-AMINO-L-HISTIDINE", "peptide"),
        "I30": (
            "N'-[(1E)-1-(1H-BENZIMIDAZOL-2-YL)ETHYLIDENE]-5-BROMO-2-HYDROXYBENZOHYDRAZIDE",
            "non-polymer",
        ),
        "B3P": (
            "2-[3-(2-HYDROXY-1,1-DIHYDROXYMETHYL-ETHYLAMINO)-PROPYLAMINO]-2-HYDROXYMETHYL-PROPANE-1,3-DIOL",
            "non-polymer",
        ),
        "B3Q": ("(3S)-3,6-DIAMINO-6-OXOHEXANOIC ACID", "peptide"),
        "B3R": ("3-BROMOPHENOL", "non-polymer"),
        "B3S": ("(3R)-3-AMINO-4-HYDROXYBUTANOIC ACID", "peptide"),
        "B3L": ("(3S)-3-AMINO-5-METHYLHEXANOIC ACID", "peptide"),
        "B3M": ("(3R)-3-AMINO-5-(METHYLSULFANYL)PENTANOIC ACID", "peptide"),
        "B3N": (
            "4-(DIMETHYLAMINO)-N-[7-(HYDROXYAMINO)-7-OXOHEPTYL]BENZAMIDE",
            "non-polymer",
        ),
        "B3H": ("(2S)-2-(BUTYRYLOXY)-3-HYDROXYPROPYL NONANOATE", "non-polymer"),
        "B3I": ("2,3,5-TRIIODOBENZOIC ACID", "non-polymer"),
        "B3K": ("(3S)-3,7-DIAMINOHEPTANOIC ACID", "peptide"),
        "B3D": ("3-AMINOPENTANEDIOIC ACID", "peptide"),
        "B3E": ("(3S)-3-AMINOHEXANEDIOIC ACID", "peptide"),
        "7E9": ("1,3-DIHYDROXYPROPAN-2-YL (Z)-TETRADEC-7-ENOATE", "non-polymer"),
        "7E8": ("(2R)-2,3-DIHYDROXYPROPYL (7Z)-TETRADEC-7-ENOATE", "non-polymer"),
        "B3A": ("(3S)-3-AMINOBUTANOIC ACID", "peptide"),
        "B3C": (
            "BENZYL [(1R)-1-({(2S,4R)-2-({(1S)-5-AMINO-1-[(S)-1,3-BENZOXAZOL-2-YL(HYDROXY)METHYL]PENTYL}CARBAMOYL)-4-[(4-METHYLBENZYL)OXY]PYRROLIDIN-1-YL}CARBONYL)-3-PHENYLPROPYL]CARBAMATE",
            "non-polymer",
        ),
        "39B": (
            "(S)-2,4-DIAMINO-5-((7,8-DICARBAUNDECABORAN-7-YL)METHYL)-6-METHYLPYRIMIDINE",
            "non-polymer",
        ),
        "39A": ("6-AMINO-3,9-DIMETHYL-9H-PURIN-3-IUM", "non-polymer"),
        "39E": (
            "(R)-2,4-DIAMINO-5-((7,8-DICARBAUNDECABORAN-7-YL)METHYL)-6-METHYLPYRIMIDINE",
            "non-polymer",
        ),
        "KXO": (
            "6-(3-FLUOROPHENYL)-N-[1-(2,2,2-TRIFLUOROETHYL)PIPERIDIN-4-YL]PYRIDINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "KXA": (
            "3-(AMINOMETHYL)-4-(2,4-DICHLOROPHENYL)-6-(2-METHOXYPHENYL)-2-METHYL-5,6-DIHYDRO-7H-PYRROLO[3,4-B]PYRIDIN-7-ONE",
            "non-polymer",
        ),
        "KXB": (
            "3-(AMINOMETHYL)-4-(2,4-DICHLOROPHENYL)-6-(2-METHOXYETHYL)-2-METHYL-5,6-DIHYDRO-7H-PYRROLO[3,4-B]PYRIDIN-7-ONE",
            "non-polymer",
        ),
        "0DX": (
            "4,9-BIS{[3-(4-METHYLPIPERAZIN-1-YL)PROPYL]AMINO}-2,7-BIS[3-(MORPHOLIN-4-YL)PROPYL]BENZO[LMN][3,8]PHENANTHROLINE-1,3,6,8(2H,7H)-TETRONE",
            "non-polymer",
        ),
        "0R5": ("?", "non-polymer"),
        "0R4": (
            "N-(4-{5-[3-(ACETYLAMINO)PHENYL]-2-(2-AMINOPYRIDIN-3-YL)-3H-IMIDAZO[4,5-B]PYRIDIN-3-YL}BENZYL)-3-FLUOROBENZAMIDE",
            "non-polymer",
        ),
        "0R6": ("?", "non-polymer"),
        "0R1": ("1,2-BENZISOXAZOLE", "non-polymer"),
        "0R0": ("2-HYDROXYBENZONITRILE", "non-polymer"),
        "0R3": (
            "{(3R,5R,6S)-5-(3-CHLOROPHENYL)-6-(4-CHLOROPHENYL)-1-[(2S,3S)-2-HYDROXYPENTAN-3-YL]-3-METHYL-2-OXOPIPERIDIN-3-YL}ACETIC ACID",
            "non-polymer",
        ),
        "0R2": (
            "[(3R,5R,6S)-1-[(2S)-1-TERT-BUTOXY-1-OXOBUTAN-2-YL]-5-(3-CHLOROPHENYL)-6-(4-CHLOROPHENYL)-2-OXOPIPERIDIN-3-YL]ACETIC ACID",
            "non-polymer",
        ),
        "39Z": (
            "5-HYDROXY-3-[(1R)-1-(1H-PYRROL-2-YL)ETHYL]-2H-INDOL-2-ONE",
            "non-polymer",
        ),
        "0R9": (
            "5-{2-(ETHYLCARBAMOYL)-4-[3-(TRIFLUOROMETHYL)-1H-PYRAZOL-1-YL]-1H-PYRROLO[2,3-B]PYRIDIN-5-YL}PYRIDINE-3-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "DHN": ("5-HYDROXY NORVALINE", "peptide"),
        "PD4": (
            "6-(6-METHOXYPYRIDIN-3-YL)-2-[(2-MORPHOLIN-4-YLETHYL)AMINO]-4-(2-PROPOXYETHYL)PYRIDO[2,3-B]PYRAZIN-3(4H)-ONE",
            "non-polymer",
        ),
        "YBY": (
            "N-{[(1S)-1-CARBOXY-2-(4-HYDROXY-3-IODOPHENYL)ETHYL]CARBAMOYL}-L-GLUTAMIC ACID",
            "non-polymer",
        ),
        "0DQ": ("4-PHENYL-1,2,3-THIADIAZOL-5-AMINE", "non-polymer"),
        "179": (
            "N-[(NAPHTHALEN-2-YLAMINO)(OXO)ACETYL]-BETA-D-GLUCOPYRANOSYLAMINE",
            "non-polymer",
        ),
        "178": ("4-(4-HYDROXY-3-ISOPROPYLPHENYLTHIO)-2-ISOPROPYLPHENOL", "non-polymer"),
        "177": (
            "1-[2-AMINO-2-CYCLOHEXYL-ACETYL]-PYRROLIDINE-3-CARBOXYLIC ACID 5-CHLORO-2-(2-ETHYLCARBAMOYL-ETHOXY)-BENZYLAMIDE",
            "peptide",
        ),
        "176": (
            "1-(2-AMINO-3,3-DIPHENYL-PROPIONYL)-PYRROLIDINE-3-CARBOXYLIC ACID 2,5-DICHLORO-BENZYLAMIDE",
            "peptide",
        ),
        "175": ("3,5-DIHYDRO-5-METHYLIDENE-4H-IMIDAZOL-4-ON", "peptide"),
        "174": ("4-CHLORO-BENZOIC ACID", "non-polymer"),
        "173": ("BENZOYL-FORMIC ACID", "non-polymer"),
        "172": ("2-AMINO-5-HYDROXY-BENZIMIDAZOLE", "non-polymer"),
        "171": ("2-PHENYLAMINO-ETHANESULFONIC ACID", "non-polymer"),
        "170": (
            "(2-[6-CHLORO-3-{[2,2-DIFLUORO-2-(1-OXIDOPYRIDIN-2-YL)ETHYL]AMINO}-2-OXOPYRAZIN-1(2H)-YL]-N-[5-CHLORO-2-(1H-TETRAZOL-1-YL)BENZYL]ACETAMIDE",
            "non-polymer",
        ),
        "DHM": ("2,6-DIMETHYL-7-OCTEN-2-OL", "non-polymer"),
        "SEG": ("HYDROXYALANINE", "peptide"),
        "WPF": (
            "3-[(4S)-6-CHLORO-2-METHYL-4-(4-METHYLPHENYL)QUINAZOLIN-3(4H)-YL]-N,N-DIMETHYLPROPAN-1-AMINE",
            "non-polymer",
        ),
        "SEE": ("TRIHYDROXY(L-SERINATO-KAPPAO~3~)BORATE(1-)", "peptide"),
        "SC": ("2-DEOXY-CYTIDINE-5'-THIOPHOSPHORATE", "nucleic-acid"),
        "HL6": ("N-[(3S)-2-OXOTETRAHYDROFURAN-3-YL]HEXANAMIDE", "non-polymer"),
        "IM5": (
            "2-AMINO-7-{[(3R,4R)-3-HYDROXY-4-(HYDROXYMETHYL)PYRROLIDIN-1-YL]METHYL}-3,5-DIHYDRO-4H-PYRROLO[3,2-D]PYRIMIDIN-4-ONE",
            "non-polymer",
        ),
        "IM4": (
            "(2E)-1-[(6-CHLOROPYRIDIN-3-YL)METHYL]-N-NITROIMIDAZOLIDIN-2-IMINE",
            "non-polymer",
        ),
        "HL4": ("N-[(3S)-2-OXOTETRAHYDROFURAN-3-YL]BUTANAMIDE", "non-polymer"),
        "YAP": (
            "5'-O-[(R)-{[(2S)-2-AMINO-3-(4-HYDROXYPHENYL)PROPANOYL]OXY}(HYDROXY)PHOSPHORYL]ADENOSINE",
            "non-polymer",
        ),
        "HL2": ("(2S,3R)-2-AMINO-3-HYDROXY-4-METHYLPENTANOIC ACID", "peptide"),
        "YAM": (
            "N-METHYL-N-{3-[({2-[(2-OXO-2,3-DIHYDRO-1H-INDOL-5-YL)AMINO]-5-(TRIFLUOROMETHYL)PYRIMIDIN-4-YL}AMINO)METHYL]PYRIDIN-2-YL}METHANESULFONAMIDE",
            "non-polymer",
        ),
        "IM9": (
            "2-{4-[4-({4-[2-METHYL-1-(1-METHYLETHYL)-1H-IMIDAZOL-5-YL]PYRIMIDIN-2-YL}AMINO)PHENYL]PIPERAZIN-1-YL}-2-OXOETHANOL",
            "non-polymer",
        ),
        "YAN": ("1,2-DICHLOROBENZENE", "non-polymer"),
        "YAK": (
            "N-{3-[(4AR,7AS)-1-(4-FLUOROBENZYL)-4-HYDROXY-2-OXO-2,4A,5,6,7,7A-HEXAHYDRO-1H-CYCLOPENTA[B]PYRIDIN-3-YL]-1,1-DIOXIDO-2H-1,2,4-BENZOTHIADIAZIN-7-YL}METHANESULFONAMIDE",
            "non-polymer",
        ),
        "0DF": (
            "1-METHYL-6-PHENYL-1,3-DIHYDRO-2H-IMIDAZO[4,5-B]PYRIDIN-2-ONE",
            "non-polymer",
        ),
        "SF1": (
            "1-BENZYL-5-CHLORO-3-(PHENYLTHIO)-1H-INDOLE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "HL0": ("N-[(3S)-2-OXOTETRAHYDROFURAN-3-YL]DECANAMIDE", "non-polymer"),
        "SEM": ("O-BENZYL-L-SERINE", "peptide"),
        "MDR": ("9-(2-DEOXY-BETA-D-RIBOFURANOSYL)-6-METHYLPURINE", "nucleic-acid"),
        "17N": (
            "8-METHOXYPHENANTHRO[3,4-D][1,3]DIOXOLE-5,6-DICARBOXYLIC ACID",
            "non-polymer",
        ),
        "17M": ("17-METHYL-17-ALPHA-DIHYDROEQUILENIN", "non-polymer"),
        "17L": (
            "(9AS)-8-ACETYL-N-[(3-ETHOXYNAPHTHALEN-1-YL)METHYL]-1,7-DIHYDROXY-3-METHOXY-9A-METHYL-9-OXO-9,9A-DIHYDRODIBENZO[B,D]FURAN-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "17J": (
            "4-(3-PHENYL-1,2,4-THIADIAZOL-5-YL)-N-(PYRIDIN-3-YL)PIPERAZINE-1-CARBOXAMIDE",
            "non-polymer",
        ),
        "17H": (
            "17-HYDROXY-18A-HOMO-19-NOR-17ALPHA-PREGNA-4,9,11-TRIEN-3-ONE",
            "non-polymer",
        ),
        "0DC": ("?", "nucleic-acid"),
        "17D": ("2-BENZYLBENZOIC ACID", "non-polymer"),
        "17C": ("?", "non-polymer"),
        "17B": (
            "(2Z,3R,4S,5R,6R)-2-[(4-AMINOBUTYL)IMINO]-6-(HYDROXYMETHYL)PIPERIDINE-3,4,5-TRIOL",
            "non-polymer",
        ),
        "17A": (
            "(2S)-2-[3-(AMINOMETHYL)PHENYL]-3-[(R)-{(1R)-1-[(BIPHENYL-4-YLSULFONYL)AMINO]-2-METHYLPROPYL}(HYDROXY)PHOSPHORYL]PROPANOIC ACID",
            "non-polymer",
        ),
        "BEY": (
            "(2S)-3-[(R)-[(1S)-1-AMINO-3-PHENYLPROPYL](HYDROXY)PHOSPHORYL]-2-BENZYLPROPANOIC ACID",
            "non-polymer",
        ),
        "17Z": ("2-METHYLNAPHTHALENE-1,4-DIOL", "non-polymer"),
        "4K0": (
            "4-[3-(1H-PYRROLO[2,3-B]PYRIDIN-3-YLMETHYL)-[1,2,4]TRIAZOLO[4,3-B][1,2,4]TRIAZIN-6-YL]BENZENECARBONITRILE",
            "non-polymer",
        ),
        "17W": ("2-[(2-PHENOXYETHYL)SULFANYL]-1H-BENZIMIDAZOLE", "non-polymer"),
        "17V": (
            "2-[1-(PROPAN-2-YL)-1H-1,2,4-TRIAZOL-5-YL]-4,5-DIHYDROTHIENO[3,2-D][1]BENZOXEPINE-8-CARBOXAMIDE",
            "non-polymer",
        ),
        "17T": (
            "N-({(2E)-2-[(4-CHLOROPHENYL)METHYLIDENE]HYDRAZINO}CARBONOTHIOYL)-BETA-D-GLUCOPYRANOSYLAMINE",
            "non-polymer",
        ),
        "17S": (
            "N-({(2E)-2-[(4-METHOXYPHENYL)METHYLIDENE]HYDRAZINO}CARBONOTHIOYL)-BETA-D-GLUCOPYRANOSYLAMINE",
            "non-polymer",
        ),
        "17R": (
            "(2R)-4-[4-FLUORO-2-(TRIFLUOROMETHYL)PHENYL]-2-METHYL-1-{[3-(1H-1,2,4-TRIAZOL-1-YL)PHENYL]SULFONYL}PIPERAZINE",
            "non-polymer",
        ),
        "17P": (
            "1-(CYCLOHEXYLAMINO)-7-(1-METHYL-1H-PYRAZOL-4-YL)-5H-PYRIDO[4,3-B]INDOLE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "2MR": ("N3, N4-DIMETHYLARGININE", "peptide"),
        "IMA": (
            "[4-({[5-BENZYLOXY-1-(3-CARBAMIMIDOYL-BENZYL)-1H-INDOLE-2-CARBONYL]-AMINO}-METHYL)-PHENYL]-TRIMETHYL-AMMONIUM",
            "non-polymer",
        ),
        "2MP": ("3,4-DIMETHYLPHENOL", "non-polymer"),
        "2MQ": (
            "6-AMINO-4-(2-AMINOETHYL)-2-(METHYLAMINO)-1,7-DIHYDRO-8H-IMIDAZO[4,5-G]QUINAZOLIN-8-ONE",
            "non-polymer",
        ),
        "2MV": (
            "(1S,2S,3R,5Z,7E,14BETA,17ALPHA)-17-[(2S,4S)-4-(2-HYDROXY-2-METHYLPROPYL)-2-METHYLTETRAHYDROFURAN-2-YL]-2-METHYL-9,10-SECOANDROSTA-5,7,10-TRIENE-1,3-DIOL",
            "non-polymer",
        ),
        "8DA": ("8-OXODEOXYADENOSINE", "non-polymer"),
        "2MU": ("2',5-DIMETHYLURIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "2MZ": ("2-METHYLIMIDAZOLE", "non-polymer"),
        "2MX": ("?", "non-polymer"),
        "2MY": ("2,6-DIMETHYLPHENOL", "non-polymer"),
        "IMC": (
            "N1-[2-DEOXY-RIBOFURANOSYL]-[2-AMINO-5-METHYL-4-OXO-4H-PYRIMIDINE]-5'-MONOPHOSPHATE",
            "nucleic-acid",
        ),
        "N5P": (
            "1-{2-DEOXY-5-O-[(R)-HYDROXY{[(R)-HYDROXY(PHOSPHONOOXY)PHOSPHORYL]OXY}PHOSPHORYL]-BETA-D-ERYTHRO-PENTOFURANOSYL}-5-NITRO-1H-INDOLE",
            "non-polymer",
        ),
        "2MB": ("2-METHYL BUT-2-ENE", "non-polymer"),
        "2MC": ("METHACRYLYL-COENZYME A", "non-polymer"),
        "2MA": ("2-METHYLADENOSINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "2MG": ("2N-METHYLGUANOSINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "2MD": (
            "MOLYBDENUM COFACTOR,BIS (MOLYBDOPTERIN GUANINE DINUCLEOTIDE)",
            "non-polymer",
        ),
        "2ME": ("METHOXYETHANE", "non-polymer"),
        "DHI": ("D-HISTIDINE", "peptide"),
        "IME": ("TETRA(IMIDAZOLE)DIAQUACOPPER (II)", "non-polymer"),
        "2MH": ("CIS-4-METHYLCYCLOHEXANAMINE", "non-polymer"),
        "2MI": ("2-METHYL-1H-INDOLE", "non-polymer"),
        "2MN": ("2-(2-METHYL-5-NITRO-1H-IMIDAZOL-1-YL)ETHANOL", "non-polymer"),
        "2MO": ("MOLYBDENUM (IV)OXIDE", "non-polymer"),
        "2ML": ("2-METHYLLEUCINE", "peptide"),
        "2MM": ("N,N-DIMETHYL-L-METHIONINE", "non-polymer"),
        "IMG": (
            "1,4-DIDEOXY-1,4-IMINO-1-(S)-(9-DEAZAGUANIN-9-YL)-D-RIBITOL",
            "non-polymer",
        ),
        "9RA": (
            "4-[1-(3,5,5,8,8-PENTAMETHYL-5,6,7,8-TETRAHYDRONAPHTHALEN-2-YL)ETHENYL]BENZOIC ACID",
            "non-polymer",
        ),
        "0D8": ("3-AMINOPROPAN-1-OL", "non-polymer"),
        "NO1": (
            "4-(HYDROXYMETHYL)-3-METHYL-1H-INDOLE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "G97": (
            "N-[2-PHENYL-4-(1H-PYRAZOL-3-YLAMINO)QUINAZOLIN-7-YL]PROP-2-ENAMIDE",
            "non-polymer",
        ),
        "G96": (
            "4-[2-(4-AMINO-1,2,5-OXADIAZOL-3-YL)-6-{[(2R)-2-AMINO-3-PHENYLPROPYL]OXY}-1-ETHYL-1H-IMIDAZO[4,5-C]PYRIDIN-4-YL]-2-METHYLBUT-3-YN-2-OL",
            "non-polymer",
        ),
        "G95": (
            "N-[(1S)-2-AMINO-1-PHENYLETHYL]-5-(1H-PYRROLO[2,3-B]PYRIDIN-4-YL)THIOPHENE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "VC3": ("3-(4-NITROPHENYL)-1H-PYRAZOLE", "non-polymer"),
        "G93": (
            "4-{2-(4-AMINO-1,2,5-OXADIAZOL-3-YL)-1-ETHYL-7-[(3S)-PIPERIDIN-3-YLMETHOXY]-1H-IMIDAZO[4,5-C]PYRIDIN-4-YL}-2-METHYLBUT-3-YN-2-OL",
            "non-polymer",
        ),
        "NO2": ("NITRITE ION", "non-polymer"),
        "G90": (
            "2-[(2S)-3-[[1-(1H-INDOL-3-YL)-2-METHYL-PROPAN-2-YL]AMINO]-2-OXIDANYL-PROPOXY]BENZENECARBONITRILE",
            "non-polymer",
        ),
        "X15": ("1,3-BISPHOSPHOGLYCERIC ACID", "non-polymer"),
        "NO3": ("NITRATE ION", "non-polymer"),
        "IMK": ("2-(BETA-D-GLUCOPYRANOSYL)-5-METHYL-1-BENZIMIDAZOLE", "non-polymer"),
        "G98": (
            "4-[2-(4-AMINO-2,5-DIHYDRO-1,2,5-OXADIAZOL-3-YL)-6-{[(1S)-3-AMINO-1-PHENYLPROPYL]OXY}-1-ETHYL-1H-IMIDAZO[4,5-C]PYRIDIN-4-YL]-2-METHYLBUT-3-EN-2-OL",
            "non-polymer",
        ),
        "G9G": (
            "2-(4-METHYLPIPERAZIN-1-YL)BENZO[C][1,5]NAPHTHYRIDIN-6(5H)-ONE",
            "non-polymer",
        ),
        "G9F": (
            "2-{4-[3-(DIMETHYLAMINO)PROPOXY]PHENYL}-5,6-DIHYDROIMIDAZO[4,5,1-JK][1,4]BENZODIAZEPIN-7(4H)-ONE",
            "non-polymer",
        ),
        "G9D": (
            "3-(MORPHOLIN-4-YLMETHYL)-1,5-DIHYDRO-6H-[1,2]DIAZEPINO[4,5,6-CD]INDOL-6-ONE",
            "non-polymer",
        ),
        "VCE": (
            "3-{[(E)-1-CARBOXY-2-CYCLOPROPYLETHENYL]OXY}-2-HYDROXYBENZOIC ACID",
            "non-polymer",
        ),
        "LFA": ("EICOSANE", "non-polymer"),
        "PDS": (
            "3-{1-[3-(DIMETHYLAMINO)PROPYL]-2-METHYL-1H-INDOL-3-YL}-4-(2-METHYL-1H-INDOL-3-YL)-1H-PYRROLE-2,5-DIONE",
            "non-polymer",
        ),
        "G9M": (
            "N-(6-OXO-5,6-DIHYDROBENZO[C][1,5]NAPHTHYRIDIN-2-YL)-2-(4-PYRROLIDIN-1-YLPIPERIDIN-1-YL)ACETAMIDE",
            "non-polymer",
        ),
        "G9L": (
            "8-FLUORO-2-(3-PIPERIDIN-1-YLPROPANOYL)-1,3,4,5-TETRAHYDROBENZO[C][1,6]NAPHTHYRIDIN-6(2H)-ONE",
            "non-polymer",
        ),
        "G9I": (
            "2-{2-[4-(PYRROLIDIN-1-YLMETHYL)PHENYL]ETHYL}-5,6-DIHYDROIMIDAZO[4,5,1-JK][1,4]BENZODIAZEPIN-7(4H)-ONE",
            "non-polymer",
        ),
        "G9H": (
            "2-[(DIMETHYLAMINO)METHYL]-5,6-DIHYDROIMIDAZO[4,5,1-JK][1,4]BENZODIAZEPIN-7(4H)-ONE",
            "non-polymer",
        ),
        "IMO": ("6-O-PHOSPHORYL INOSINE MONOPHOSPHATE", "non-polymer"),
        "G9P": (
            "(11BR)-3-OXO-1,2,3,11B-TETRAHYDROCHROMENO[4,3,2-DE]ISOQUINOLINE-10-SULFONIC ACID",
            "non-polymer",
        ),
        "IMN": ("INDOMETHACIN", "non-polymer"),
        "X9F": (
            "N^4^-1H-INDAZOL-4-YL-N^2^-[3-(METHYLSULFONYL)PHENYL]PYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "2M3": (
            "2-METHYL-3-{(3S)-1-[(1-PYRIDIN-2-YLCYCLOPROPYL)CARBONYL]PYRROLIDIN-3-YL}-1H-PYRROLO[2,3-B]PYRIDINE",
            "non-polymer",
        ),
        "2M6": (
            "7-DEOXY-L-GLYCERO-ALPHA-D-MANNO-HEPTOPYRANOSYL-(1->2)-ALPHA-D-MANNOPYRANOSYL-(1->2)-ALPHA-D-MANNOPYRANOSYL-(1->3)-ALPHA-D-MANNOPYRANOSE",
            "non-polymer",
        ),
        "2M4": ("2-O-ALPHA-D-MANNOPYRANOSYL-ALPHA-D-MANNOPYRANOSE", "saccharide"),
        "2M5": ("METHYL 7-DEOXY-L-GLYCERO-ALPHA-D-MANNO-HEPTOPYRANOSIDE", "saccharide"),
        "2M8": ("2-O-ALPHA-MANNOSYL-D-GLYCERATE", "non-polymer"),
        "N53": (
            "N-[6,6-DIMETHYL-5-[(2S)-4-METHYL-2-(PHENYLMETHYL)PIPERAZIN-1-YL]CARBONYL-2,4-DIHYDROPYRROLO[3,4-C]PYRAZOL-3-YL]-3-PHENOXY-BENZAMIDE",
            "non-polymer",
        ),
        "N50": (
            "2-[(2,6-DIMETHOXY-5-METHYLPYRIMIDIN-4-YL)METHYLIDENE]PROPANE-1,3-DIOL",
            "non-polymer",
        ),
        "VNL": ("4-HYDROXY-3-METHOXYBENZOATE", "non-polymer"),
        "PDL": ("N-(5'-PHOSPHOPYRIDOXYL)-L-ALANINE", "peptide"),
        "IMU": (
            "PHOSPHORIC ACID MONO-[5-(2-AMINO-4-OXO-4,5-DIHYDRO-3H-PYRROLO[3,2-D]PYRIMIDIN-7-YL)-3,4-DIHYDROXY-PYRROLIDIN-2-YLMETHYL] ESTER",
            "non-polymer",
        ),
        "NXX": ("?", "non-polymer"),
        "IMT": ("4-AMINO-(1-METHYLIMIDAZOLE)-2-CARBOXYLIC ACID", "non-polymer"),
        "PZB": ("4-(4-PHENOXY-1H-PYRAZOL-3-YL)BENZENE-1,3-DIOL", "non-polymer"),
        "IMW": (
            "2-METHYL-5-(1-METHYLETHYL)CYCLOHEXA-2,5-DIENE-1,4-DIONE",
            "non-polymer",
        ),
        "0OX": ("4-AMINO-2-(CYCLOPENTYLOXY)PYRIMIDINE-5-CARBOXAMIDE", "non-polymer"),
        "0OY": ("(4-HYDROXYPHENYL)(MORPHOLIN-4-YL)METHANONE", "non-polymer"),
        "0OV": (
            "[4-AMINO-2-{[(1S,2R)-2-METHYLCYCLOHEXYL]OXY}-5-OXOPYRIDO[2,3-D]PYRIMIDIN-8(5H)-YL]ACETONITRILE",
            "non-polymer",
        ),
        "0OW": (
            "4-AMINO-2-(CYCLOPENTYLOXY)-6-{[(1R,2S)-2-HYDROXYCYCLOPENTYL]OXY}PYRIMIDINE-5-CARBOXAMIDE",
            "non-polymer",
        ),
        "0OT": (
            "2-({3-[(3,5-DIBROMO-2-ETHOXYBENZYL)AMINO]PROPYL}AMINO)QUINOLIN-4(1H)-ONE",
            "non-polymer",
        ),
        "0OU": (
            "2-({3-[(3,5-DICHLOROBENZYL)AMINO]PROPYL}AMINO)QUINOLIN-4(1H)-ONE",
            "non-polymer",
        ),
        "0OR": (
            "N~3~-[(2R)-2-HYDROXY-4-{[(S)-HYDROXY(PHOSPHONOOXY)PHOSPHORYL]OXY}-3,3-DIMETHYLBUTANOYL]-BETA-ALANINAMIDE",
            "non-polymer",
        ),
        "0OS": (
            "{3-[(1S)-1-[({(2S)-1-[(3,5-DICHLOROPHENYL)SULFONYL]PIPERIDIN-2-YL}CARBONYL)OXY]-3-(3,4-DIMETHOXYPHENYL)PROPYL]PHENOXY}ACETIC ACID",
            "non-polymer",
        ),
        "0OP": ("[3-(BENZYLOXY)PHENYL]METHANOL", "non-polymer"),
        "0OQ": (
            "(1R,5S)-3-[4-(TRIFLUOROMETHYL)BENZOYL]-1,2,3,4,5,6-HEXAHYDRO-8H-1,5-METHANOPYRIDO[1,2-A][1,5]DIAZOCIN-8-ONE",
            "non-polymer",
        ),
        "0ON": ("N-PHENYLPYRIDINE-3-CARBOXAMIDE", "non-polymer"),
        "0OO": ("(3-PHENOXYPHENYL)METHANOL", "non-polymer"),
        "0OL": ("PHENYL(PIPERIDIN-1-YL)METHANONE", "non-polymer"),
        "0OM": ("BENZYL PYRIDINE-3-CARBOXYLATE", "non-polymer"),
        "0OJ": ("?", "non-polymer"),
        "0OK": ("3-PHENYLQUINAZOLIN-4(3H)-ONE", "non-polymer"),
        "0OH": ("?", "non-polymer"),
        "9EP": (
            "(2R)-5-[(ACETYLOXY)METHYL]-2-{(1R)-2-OXO-1-[(THIOPHEN-2-YLACETYL)AMINO]ETHYL}-3,6-DIHYDRO-2H-1,3-THIAZINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "0OF": (
            "(3S)-1-ETHYL-3-[3-HYDROXY-5-(PYRIDIN-3-YL)PHENYL]-5-PHENYL-7-(TRIFLUOROMETHYL)-1H-1,5-BENZODIAZEPINE-2,4(3H,5H)-DIONE",
            "non-polymer",
        ),
        "0OG": (
            "3-{5-[3-ETHYL-5-(5-METHYLFURAN-2-YL)-1H-PYRAZOL-1-YL]-1-[(6-OXO-1,6-DIHYDROPYRIDIN-3-YL)METHYL]-1H-BENZIMIDAZOL-2-YL}-4-HYDROXYBENZOIC ACID",
            "non-polymer",
        ),
        "0OD": (
            "TRICHLORO{(1,2,3,4,5-ETA)-1,2,3,4-TETRAMETHYL-5-[2-({5-[(3AS,4S,6AR)-2-OXOHEXAHYDRO-1H-THIENO[3,4-D]IMIDAZOL-4-YL]PENTANOYL}AMINO)ETHYL]CYCLOPENTADIENYL}RHODIUM(1+)",
            "non-polymer",
        ),
        "0OE": (
            "4-{2-[5-(3-CHLOROPHENYL)-1H-PYRAZOL-4-YL]-1-[3-(1H-IMIDAZOL-1-YL)PROPYL]-1H-BENZIMIDAZOL-5-YL}BENZOIC ACID",
            "non-polymer",
        ),
        "0OB": ("3-(1-BENZOTHIOPHEN-2-YL)PROPANOIC ACID", "non-polymer"),
        "0OC": ("N,N-DIMETHYLTHIOPHENE-3-SULFONAMIDE", "non-polymer"),
        "0OA": (
            "(1R,2S,3R,4S,6S)-6-(CYCLOHEXYLMETHOXY)-2,3,4-TRIHYDROXYCYCLOHEXYL (2R)-2-METHOXY-3-(OCTADECYLOXY)PROPYL HYDROGEN (S)-PHOSPHATE",
            "non-polymer",
        ),
        "692": ("N-METHYL-1-(2-THIOPHEN-2-YLPHENYL)METHANAMINE", "non-polymer"),
        "693": (
            "2-(5-{5-[AMINO(IMINIO)METHYL]-1H-BENZIMIDAZOL-2-YL}-2'-FLUORO-6-OXIDO-1,1'-BIPHENYL-3-YL)SUCCINATE",
            "non-polymer",
        ),
        "IMZ": (
            "CIS-[4,5-BIS-(4-BROMOPHENYL)-2-(2-ETHOXY-4-METHOXYPHENYL)-4,5-DIHYDROIMIDAZOL-1-YL]-[4-(2-HYDROXYETHYL)PIPERAZIN-1-YL]METHANONE",
            "non-polymer",
        ),
        "696": (
            "3-{5-[AMINO(IMINIO)METHYL]-1H-INDOL-2-YL}-1,1'-BIPHENYL-2-OLATE",
            "non-polymer",
        ),
        "697": (
            "5-HYDROXY-2-(4-HYDROXYPHENYL)-1-BENZOFURAN-7-CARBONITRILE",
            "non-polymer",
        ),
        "694": (
            "4-BROMO-3-(CARBOXYMETHOXY)-5-(4-HYDROXYPHENYL)THIOPHENE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "695": (
            "N-{2-[7-(METHYLSULFAMOYL)NAPHTHALEN-1-YL]ETHYL}ACETAMIDE",
            "non-polymer",
        ),
        "PDC": ("PYRIDINE-2,6-DICARBOXYLIC ACID", "non-polymer"),
        "698": (
            "(2S,4S,5R)-1-(4-TERT-BUTYLBENZOYL)-2-ISOBUTYL-5-(1,3-THIAZOL-2-YL)PYRROLIDINE-2,4-DICARBOXYLIC ACID",
            "non-polymer",
        ),
        "699": (
            "(2S,4S,5R)-2-ISOBUTYL-5-(2-THIENYL)-1-[4-(TRIFLUOROMETHYL)BENZOYL]PYRROLIDINE-2,4-DICARBOXYLIC ACID",
            "non-polymer",
        ),
        "10X": (
            "(9Z)-11-{(2S,3S)-3-[(2Z)-PENT-2-EN-1-YL]OXIRAN-2-YL}UNDEC-9-ENOIC ACID",
            "non-polymer",
        ),
        "PDB": (
            "1-(2-METHYLPHENYL)-6-[(2R)-3,3,3-TRIFLUORO-2-METHYLPROPYL]-1,7-DIHYDRO-4H-PYRAZOLO[3,4-D]PYRIMIDIN-4-ONE",
            "non-polymer",
        ),
        "ZX1": (
            "(3R)-1-[(4R)-AZEPAN-4-YLCARBAMOYL]-3-(SULFOAMINO)-L-PROLINE",
            "non-polymer",
        ),
        "ZX0": (
            "(3R)-1-[(4S)-AZEPAN-4-YLCARBAMOYL]-3-(SULFOAMINO)-L-PROLINE",
            "non-polymer",
        ),
        "FHL": (
            "(E)-N~6~-[3-CARBOXY-1-(HYDROXYMETHYL)PROPYLIDENE]-L-LYSINE",
            "peptide",
        ),
        "ZX5": (
            "3-(1H-INDOL-3-YL)-N-{[(2R,3R,4R,5R,6S)-3,4,5-TRIHYDROXY-6-METHYLPIPERIDIN-2-YL]METHYL}PROPANAMIDE",
            "non-polymer",
        ),
        "PDA": (
            "2-[(3-HYDROXY-2-METHYL-5-PHOSPHONOOXYMETHYL-PYRIDIN-4-YLMETHYL)-AMINO]-PROPIONIC ACID",
            "non-polymer",
        ),
        "ZX7": (
            "N-{[(2R,3R,4R,5R,6S)-3,4,5-TRIHYDROXY-6-METHYLPIPERIDIN-2-YL]METHYL}-1H-INDOLE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "ZX6": (
            "2-(1H-INDOL-3-YL)-N-{[(2R,3R,4R,5R,6S)-3,4,5-TRIHYDROXY-6-METHYLPIPERIDIN-2-YL]METHYL}ACETAMIDE",
            "non-polymer",
        ),
        "ZX9": (
            "(2S)-2-CYCLOPENTYL-2-PHENYL-N-{[(2R,3R,4R,5R,6S)-3,4,5-TRIHYDROXY-6-METHYLPIPERIDIN-2-YL]METHYL}ETHANAMIDE",
            "non-polymer",
        ),
        "ZX8": (
            "N-{[(2R,3R,4R,5R,6S)-3,4,5-TRIHYDROXY-6-METHYLPIPERIDIN-2-YL]METHYL}-1-BENZOFURAN-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "UXH": (
            "6,8-DIAMINO-7-CHLORO-1-METHYL-2-OXO-1,2-DIHYDROPYRROLO[4,3,2-DE]QUINOLINE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "7X4": (
            "3-({4-[(5-CHLORO-1,3-BENZODIOXOL-4-YL)AMINO]PYRIMIDIN-2-YL}AMINO)BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "7X5": (
            "N'-(5-CHLORO-1,3-BENZODIOXOL-4-YL)-N-(3-METHYLSULFONYLPHENYL)PYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "69P": ("6-PIPERAZIN-1-YL-9H-PURINE", "non-polymer"),
        "7X7": (
            "N'-(5-CHLORO-1,3-BENZODIOXOL-4-YL)-N-(3-MORPHOLIN-4-YLPHENYL)PYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "7X1": (
            "N-(5-CHLORO-1,3-BENZODIOXOL-4-YL)-6-METHOXY-7-(3-PIPERIDIN-1-YLPROPOXY)QUINAZOLIN-4-AMINE",
            "non-polymer",
        ),
        "7X2": ("?", "non-polymer"),
        "7X3": (
            "N'-(3-CHLORO-4-METHOXY-PHENYL)-N-(3,4,5-TRIMETHOXYPHENYL)-1,3,5-TRIAZINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "7X8": (
            "3-({4-[(5-CHLORO-1,3-BENZODIOXOL-4-YL)AMINO]PYRIMIDIN-2-YL}AMINO)BENZAMIDE",
            "non-polymer",
        ),
        "UX9": (
            "METHYL [(2S)-1-{2-(2-{(3R,4S)-3-BENZYL-4-HYDROXY-1-[(1S,2R)-2-HYDROXY-2,3-DIHYDRO-1H-INDEN-1-YL]-2-OXOPYRROLIDIN-3-YL}ETHYL)-2-[4-(PYRIDIN-4-YL)BENZYL]HYDRAZINYL}-3,3-DIMETHYL-1-OXOBUTAN-2-YL]CARBAMATE",
            "non-polymer",
        ),
        "ZEA": ("(2E)-2-METHYL-4-(9H-PURIN-6-YLAMINO)BUT-2-EN-1-OL", "non-polymer"),
        "RH9": ("RHIZOPODIN", "non-polymer"),
        "X": ("?", "nucleic-acid"),
        "ZXZ": ("1,4-DIMETHYLQUINOLIN-2(1H)-ONE", "non-polymer"),
        "ZED": (
            "L-PROLINE, 1-[(2S)-3-MERCAPTO-2-METHYL-1-OXOPROPYL]-4-(PHENYLTHIO)-, 4S",
            "non-polymer",
        ),
        "0O8": (
            "(1,1-DIMETHYLPIPERIDIN-1-IUM-4-YL) OCTADECYL HYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "0O9": (
            "(2S)-2-METHOXY-3-(OCTADECYLOXY)PROPYL (1R,2R,3R,4S,6S)-2,3,4-TRIHYDROXY-6-(2-METHYLPROPOXY)CYCLOHEXYL HYDROGEN (S)-PHOSPHATE",
            "non-polymer",
        ),
        "0O6": ("?", "non-polymer"),
        "0O7": (
            "8-(4-ETHYLPHENYL)-5-METHYL-2,5-DIHYDROPYRAZOLO[4,3-C][2,1]BENZOTHIAZINE 4,4-DIOXIDE",
            "non-polymer",
        ),
        "0O4": (
            "4-(ISOQUINOLIN-1-YL)-N-[2-(MORPHOLIN-4-YL)ETHYL]BENZAMIDE",
            "non-polymer",
        ),
        "0O5": (
            "4-(3-METHYLISOQUINOLIN-1-YL)-N-[2-(MORPHOLIN-4-YL)ETHYL]BENZAMIDE",
            "non-polymer",
        ),
        "0O2": ("?", "non-polymer"),
        "0O3": (
            "3-(HYDROXYMETHYL)BUT-3-EN-1-YL TRIHYDROGEN DIPHOSPHATE",
            "non-polymer",
        ),
        "N5A": (
            "(2S,3S,4R,5R)-N-(2-AMINOETHYL)-5-(6-AMINO-9H-PURIN-9-YL)-3,4-DIHYDROXYTETRAHYDROFURAN-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "PLK": (
            "(3R)-3-[({3-HYDROXY-2-METHYL-5-[(PHOSPHONOOXY)METHYL]PYRIDIN-4-YL}METHYL)AMINO]-5-METHYLHEXANOIC ACID",
            "non-polymer",
        ),
        "DMA": ("DIMETHYLALLYL DIPHOSPHATE", "non-polymer"),
        "DMB": ("?", "non-polymer"),
        "DMC": (
            "3-(4-DIETHYLAMINO-2-HYDROXY-PHENYL)-2-METHYL-PROPIONIC ACID",
            "non-polymer",
        ),
        "DMD": ("5,6-DIMETHYLBENZIMIDAZOLE", "non-polymer"),
        "DME": ("DECAMETHONIUM ION", "non-polymer"),
        "DMF": ("DIMETHYLFORMAMIDE", "non-polymer"),
        "DMG": ("N,N-DIMETHYLGLYCINE", "non-polymer"),
        "DMH": ("N4,N4-DIMETHYL-ASPARAGINE", "peptide"),
        "DMI": ("2,3-DIMETHYLIMIDAZOLIUM ION", "non-polymer"),
        "DMJ": ("1-DEOXYMANNOJIRIMYCIN", "non-polymer"),
        "DMK": ("3,3-DIMETHYL ASPARTIC ACID", "peptide"),
        "DML": ("TERMINAL DIMETHYL", "non-polymer"),
        "DMM": ("3'-DESAMINO-3'-(2-METHOXY-4-MORPHOLINYL)-DOXORUBICIN", "non-polymer"),
        "DMN": ("DIMETHYLAMINE", "non-polymer"),
        "DMO": ("ALPHA-DIFLUOROMETHYLORNITHINE", "non-polymer"),
        "DMP": (
            "[4-R-(-4-ALPHA,5-ALPHA,6-BETA,7-BETA)]-HEXAHYDRO-5,6-BIS(HYDROXY)-[1,3-BIS([4-HYDROXYMETHYL-PHENYL]METHYL)-4,7-BIS(PHENYLMETHYL)]-2H-1,3-DIAZEPINONE",
            "non-polymer",
        ),
        "DMQ": (
            "[4-R-(-4-ALPHA,5-ALPHA,6-BETA,7-BETA)]-HEXAHYDRO-5,6-BIS(HYDROXY)-1,3-BIS([(3-AMINO)PHENYL]METHYL)-4,7-BIS(PHENYLMETHYL)-2H-1,3-DIAZEPINONE",
            "non-polymer",
        ),
        "DMR": ("2-HYDROXY-SUCCINIC ACID", "non-polymer"),
        "DMS": ("DIMETHYL SULFOXIDE", "non-polymer"),
        "DMT": ("3-HYDROXY-4,4-DIMETHYL-2-(METHYLAMINO)-6-OCTENOIC ACID", "peptide"),
        "DMU": ("DECYL-BETA-D-MALTOPYRANOSIDE", "saccharide"),
        "DMV": ("2,3-DIHYDROXY-VALERIANIC ACID", "non-polymer"),
        "DMW": ("2,3-DIMETHYL-1,4-NAPHTHOQUINONE", "non-polymer"),
        "DMX": ("3-[BENZYL(DIMETHYL)AMMONIO]PROPANE-1-SULFONATE", "non-polymer"),
        "DMY": ("DISTAMYCIN A", "non-polymer"),
        "DMZ": (
            "4-METHYL-N-METHYL-N-(2-PHENYL-2H-PYRAZOL-3-YL)BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "PS1": ("1,2-DIDECANOYL-SN-GLYCERO-3-[PHOSPHO-L-SERINE]", "non-polymer"),
        "PS6": (
            "O-[(S)-{[(2S)-2-(HEXANOYLOXY)-3-(TETRADECANOYLOXY)PROPYL]OXY}(HYDROXY)PHOSPHORYL]-D-SERINE",
            "non-polymer",
        ),
        "PS7": (
            "{(1R,2R,3R)-2-[(3E)-4,8-DIMETHYLNONA-3,7-DIEN-1-YL]-2-METHYL-3-[(1E,5E)-2,6,10-TRIMETHYLUNDECA-1,5,9-TRIEN-1-YL]CYCLOPROPYL}METHYL TRIHYDROGEN DIPHOSPHATE",
            "non-polymer",
        ),
        "PS4": (
            "N-({(2S)-1-[(3R)-3-AMINO-4-(3-CHLOROPHENYL)BUTANOYL]PYRROLIDIN-2-YL}METHYL)-3-(METHYLSULFONYL)BENZAMIDE",
            "non-polymer",
        ),
        "PS5": ("PENTASULFIDE-SULFUR", "non-polymer"),
        "ZK5": ("9ALPHA-FLUOROCORTISOL", "non-polymer"),
        "J9Z": (
            "(2R)-2-[(1-BENZOTHIOPHEN-2-YLCARBONYL)AMINO]-3-PHENYLPROPYL PHOSPHATE",
            "non-polymer",
        ),
        "M12": ("10-METHYLUNDECANOIC ACID", "non-polymer"),
        "PLI": (
            "(2E)-2-{[(Z)-{3-HYDROXY-2-METHYL-5-[(PHOSPHONOOXY)METHYL]PYRIDIN-4(1H)-YLIDENE}METHYL]IMINO}PROPANOIC ACID",
            "non-polymer",
        ),
        "MF4": ("TETRAFLUOROMAGNESATE(2-)", "non-polymer"),
        "XR2": (
            "1-METHYL-9-[12-(9-METHYLPHENAZIN-10-IUM-1-YL)-12-OXO-2,11-DIAZA-5,8-DIAZONIADODEC-1-ANOYL]PHENAZIN-10-IUM",
            "non-polymer",
        ),
        "P76": (
            "N-[(1S,2R)-2-HYDROXY-1-(HYDROXYCARBAMOYL)PROPYL]-4-[4-(PHENYLETHYNYL)PHENYL]PIPERIDINE-1-CARBOXAMIDE",
            "non-polymer",
        ),
        "J94": (
            "(5S)-6,6-DIMETHYL-5-[(6R)-8-OXO-6,8-DIHYDROFURO[3,4-E][1,3]BENZODIOXOL-6-YL]-5,6,7,8-TETRAHYDRO[1,3]DIOXOLO[4,5-G]ISOQUINOLIN-6-IUM",
            "non-polymer",
        ),
        "J90": (
            "4-{[(5-NITRO-6-OXO-1,6-DIHYDROPYRIMIDIN-4-YL)AMINO]METHYL}BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "P77": (
            "2-CHLORO-10-[3-(4-METHYLPIPERAZIN-1-YL)PROPYL]-10H-PHENOTHIAZINE",
            "non-polymer",
        ),
        "XRG": ("1-METHYL-3-(N-METHYLCARBAMIMIDOYL)UREA", "non-polymer"),
        "P1Z": ("4-BUTYL-1,2-DIPHENYL-PYRAZOLIDINE-3,5-DIONE", "non-polymer"),
        "XRA": (
            "2-[4-(FURAN-2-YLCARBONYL)PIPERAZIN-1-YL]-6,7-DIMETHOXYQUINAZOLIN-4-AMINE",
            "non-polymer",
        ),
        "XRX": ("3-(DIMETHYLAMINO)BUTYL DIMETHYLCARBAMATE", "non-polymer"),
        "XRS": (
            "(2R)-N,N-DIMETHYL-4-(1-METHYLIMIDAZOL-2-YL)OXY-BUTAN-2-AMINE",
            "non-polymer",
        ),
        "PSJ": ("D-PSICOSE", "saccharide"),
        "FX4": (
            "[(5Z)-5-{[3-(CARBOXYMETHOXY)-4-METHOXYPHENYL]METHYLIDENE}-2,4-DIOXO-1,3-THIAZOLIDIN-3-YL]ACETIC ACID",
            "non-polymer",
        ),
        "PSH": ("1-THIOPHOSPHONO-L-HISTIDINE", "peptide"),
        "PSI": (
            "METHYL N-{(4S,5S)-5-[(L-ALANYL-L-ALANYL)AMINO]-4-HYDROXY-6-PHENYLHEXANOYL}-L-VALYL-L-VALINATE",
            "peptide",
        ),
        "PSN": ("BENZYL 2-FORMYLPHENYL HYDROGEN PHOSPHATE", "non-polymer"),
        "PSO": ("4'-HYDROXYMETHYL-4,5',8-TRIMETHYLPSORALEN", "non-polymer"),
        "PSL": ("PYROSULFATE", "non-polymer"),
        "PSM": (
            "N-[DIHYDROXY(PROPYL)-LAMBDA~4~-SULFANYL]THREONYL-N~1~-{4-[AMINO(IMINO)METHYL]BENZYL}METHIONINAMIDE",
            "peptide",
        ),
        "ME7": ("1,1'-ETHANE-1,2-DIYLBIS(1H-PYRROLE-2,5-DIONE)", "non-polymer"),
        "PSC": (
            "(7R,17E,20E)-4-HYDROXY-N,N,N-TRIMETHYL-9-OXO-7-[(PALMITOYLOXY)METHYL]-3,5,8-TRIOXA-4-PHOSPHAHEXACOSA-17,20-DIEN-1-AMINIUM 4-OXIDE",
            "non-polymer",
        ),
        "PSA": ("3-HYDROXY-4-AMINO-5-PHENYLPENTANOIC ACID", "peptide"),
        "PSF": ("1,2-DICAPROYL-SN-PHOSPHATIDYL-L-SERINE", "non-polymer"),
        "108": (
            "2-(5-CARBAMIMIDOYL-2-HYDROXY-BENZYLAMINO)-PROPIONIC ACID",
            "non-polymer",
        ),
        "PSD": ("5'-O-[N-(PROLYL)-SULFAMOYL]ADENOSINE", "non-polymer"),
        "JLZ": ("3-FLUORO-4-METHYLBENZENECARBOXIMIDAMIDE", "non-polymer"),
        "DM0": ("N~2~,N~2~,N~6~,N~6~-TETRAMETHYL-L-LYSINE", "peptide"),
        "DM1": ("DAUNOMYCIN", "non-polymer"),
        "DM2": ("DOXORUBICIN", "non-polymer"),
        "109": (
            "2-(4-CARBAMIMIDOYL-2-HYDROXY-BENZYLAMINO)-PROPIONIC ACID",
            "non-polymer",
        ),
        "DM4": ("1-O-DEMETHYL-6-DEOXYDOXORUBICIN", "non-polymer"),
        "DM5": ("IDARUBICIN", "non-polymer"),
        "DM6": ("4'-EPIDOXORUBICIN", "non-polymer"),
        "DM7": ("4'-DEOXY-4'-IODODOXORUBICIN", "non-polymer"),
        "DM8": ("2'-BROMO-4'-EPIDAUNORUBICIN", "non-polymer"),
        "DM9": ("N-HYDROXYMETHYL-N-(2-METHOXYETHYL)-DAUNOMYCIN", "non-polymer"),
        "PSP": ("1,3-PROPYLENE-BIS-N,N'-SYN-4-PYRIDINIUMALDOXIME", "non-polymer"),
        "JLN": (
            "5-(FORMYLAMINO)-1-(5-O-PHOSPHONO-BETA-D-RIBOFURANOSYL)-1H-IMIDAZOLE-4-CARBOXYLIC ACID",
            "nucleic-acid",
        ),
        "PSV": ("ALPHA-D-PSICOFURANOSE", "saccharide"),
        "MYS": ("PENTADECANE", "non-polymer"),
        "PST": ("THYMIDINE-5'-THIOPHOSPHATE", "nucleic-acid"),
        "PSU": ("PSEUDOURIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "BCL": ("BACTERIOCHLOROPHYLL A", "non-polymer"),
        "BCM": ("BICYCLOMYCIN", "non-polymer"),
        "BCN": ("BICINE", "non-polymer"),
        "BCO": ("BUTYRYL COENZYME A", "non-polymer"),
        "BCH": ("2-(BUTYRYLSULFANYL)-N,N,N-TRIMETHYLETHANAMINIUM", "non-polymer"),
        "HND": ("(4S)-NONANE-1,4-DIOL", "non-polymer"),
        "BCK": ("(1-METHYL-5-PHENYL-1H-PYRAZOL-4-YL)METHANOL", "non-polymer"),
        "BCD": ("BETA-CYCLODEXTRIN", "saccharide"),
        "BCE": ("(2Z)-N-BIPHENYL-4-YL-2-CYANO-3-HYDROXYBUT-2-ENAMIDE", "non-polymer"),
        "BCF": ("1-CHLORO-2,3,4,5,6-PENTAFLUOROBENZENE", "non-polymer"),
        "BCG": (
            "3-(BENZOYLOXY)-8-METHYL-8-AZABICYCLO[3.2.1]OCTANE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "HNM": ("1-(6-HYDROXYPYRIDIN-3-YL)-4-(METHYLAMINO)BUTAN-1-ONE", "non-polymer"),
        "BPC": (
            "(7R,8R,9R)-7,8,9,10-TETRAHYDROBENZO[DEF]CHRYSENE-7,8,9-TRIOL",
            "non-polymer",
        ),
        "BCB": ("BACTERIOCHLOROPHYLL B", "non-polymer"),
        "BCC": (
            "(3S,6S,8AR)-6-AMINO-6-BENZYL-5-OXOOCTAHYDROINDOLIZINE-3-CARBOXYLIC ACID",
            "peptide",
        ),
        "HNQ": ("5-NITROQUINOLIN-8-OL", "non-polymer"),
        "102": ("GAMMA-DEOXY-GAMMA-SULFO-GUANOSINE-5'-TRIPHOSPHATE", "non-polymer"),
        "258": ("(2-CHLOROETHOXY)BENZENE", "non-polymer"),
        "259": ("1-(METHYLSULFANYL)-4-NITROBENZENE", "non-polymer"),
        "BCX": ("(3R)-3-AMINO-4-SULFANYLBUTANOIC ACID", "peptide"),
        "BCY": ("CARBOXYAMINO[4.3.0]BICYCLO1,6-DIAZANONAN-2-ONE", "non-polymer"),
        "BCZ": (
            "3-(1-ACETYLAMINO-2-ETHYL-BUTYL)-4-GUANIDINO-2-HYDROXY-CYCLOPENTANECARBOXYLIC ACID",
            "non-polymer",
        ),
        "103": ("?", "non-polymer"),
        "BCT": ("BICARBONATE ION", "non-polymer"),
        "BCU": ("(3R)-3,4-DIHYDRO-2H-CHROMEN-3-YLMETHANOL", "non-polymer"),
        "BCV": ("D-(L-A-AMINOADIPOYL)-L-CYSTEINYL-D-CYCLOPROPYLGLYCINE", "non-polymer"),
        "BCW": (
            "6-DEOXY-ALPHA-L-GALACTOPYRANOSYL-(1->4)-[6-DEOXY-ALPHA-L-GALACTOPYRANOSYL-(1->2)-BETA-D-GALACTOPYRANOSYL-(1->3)]-2-(ACETYLAMINO)-2-DEOXY-ALPHA-D-GLUCOPYRANOSE",
            "non-polymer",
        ),
        "BCP": ("2-BROMO-6-CHLORO-PURINE", "non-polymer"),
        "BCQ": (
            "N-[(1R)-1-[({[(5S)-3,5-DIMETHYL-2,5-DIHYDROISOXAZOL-4-YL]METHYL}SULFONYL)METHYL]-2-OXO-2-({(1S)-3-PHENYL-1-[2-(PHENYLSULFONYL)ETHYL]PROPYL}AMINO)ETHYL]MORPHOLINE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "BCR": ("BETA-CAROTENE", "non-polymer"),
        "BCS": ("BENZYLCYSTEINE", "peptide"),
        "3XF": (
            "6-({(3S,4S)-4-[2-({2,2-DIFLUORO-2-[(2R)-PIPERIDIN-2-YL]ETHYL}AMINO)ETHOXY]PYRROLIDIN-3-YL}METHYL)-4-METHYLPYRIDIN-2-AMINE",
            "non-polymer",
        ),
        "TCF": ("3-(4-BROMOPHENYL)PROPANAMIDE", "non-polymer"),
        "WP5": (
            "METHYL [(4S)-6-BROMO-2-METHYL-4-PHENYLQUINAZOLIN-3(4H)-YL]ACETATE",
            "non-polymer",
        ),
        "SR2": (
            "(2E)-N-{4-[(3-BROMOPHENYL)AMINO]QUINAZOLIN-6-YL}-4-(DIMETHYLAMINO)BUT-2-ENAMIDE",
            "non-polymer",
        ),
        "3XH": ("3-HYDROXYHIPPURIC ACID", "peptide"),
        "SR1": ("5-S-METHYL-5-THIO-ALPHA-D-RIBOFURANOSE", "non-polymer"),
        "SRB": ("L-1-NAPHTHYL-2-ACETAMIDO-ETHANE BORONIC ACID", "non-polymer"),
        "NOL": (
            "N-[(BENZYLOXY)CARBONYL]-O-(TERT-BUTYL)-L-THREONYL-3-CYCLOHEXYL-N-[(1S)-2-HYDROXY-1-{[(3S)-2-OXOPYRROLIDIN-3-YL]METHYL}ETHYL]-L-ALANINAMIDE",
            "non-polymer",
        ),
        "PYX": ("S-[S-THIOPYRIDOXAMINYL]CYSTEINE", "peptide"),
        "SRA": ("?", "nucleic-acid"),
        "SRG": (
            "(2S)-2-[[(2S)-1-[[(5S,8S,9E)-2,7-DIOXO-5-PROPAN-2-YL-1,6-DIAZACYCLODODECA-3,9-DIEN-8-YL]AMINO]-3-METHYL-1-OXO-BUTAN-2-YL]CARBAMOYLAMINO]-3-METHYL-BUTANOIC ACID",
            "non-polymer",
        ),
        "RH1": (
            "2,5-DIAZIRIDIN-1-YL-3-(HYDROXYMETHYL)-6-METHYLCYCLOHEXA-2,5-DIENE-1,4-DIONE",
            "non-polymer",
        ),
        "NOM": (
            "7-HYDROXY-5-METHYL-3,3A,5,11B-TETRAHYDRO-1,4-DIOXA-CYCLOPENTA[A]ANTHRACENE-2,6,11-TRIONE",
            "non-polymer",
        ),
        "SKE": (
            "4-({5-AMINO-1-[(2,6-DIFLUOROPHENYL)CARBONYL]-1H-1,2,4-TRIAZOL-3-YL}AMINO)BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "SRI": ("(2Z)-3-HYDROXY-2-IMINOPROPANOIC ACID", "non-polymer"),
        "SRN": ("SORANGICIN A", "non-polymer"),
        "NON": ("METHYL NONANOATE (ESTER)", "non-polymer"),
        "SRL": (
            "[2-(3,5-DI-TERT-BUTYL-4-HYDROXY-PHENYL)-1-(DIETHOXY-PHOSPHORYL)-VINYL]-PHOSPHONIC ACID DIETHLYL ESTER",
            "non-polymer",
        ),
        "SRM": ("SIROHEME", "non-polymer"),
        "SRR": ("4,4'-[2-(4-ETHOXYPHENYL)BUT-1-ENE-1,1-DIYL]DIPHENOL", "non-polymer"),
        "SRS": (
            "4-METHYL-3-(9-OXO-1,8-DIAZA-TRICYCLO[10.6.1.0(13,18)]NONADECA-12(19),13(18),15,17-TETRAENE-10-CARBAMOYL)PENTA-METHYLSULFONEDIIMINE",
            "non-polymer",
        ),
        "SRP": ("SERYL ADENYLATE", "non-polymer"),
        "NOO": ("N-(3-OXOCYCLOHEX-1-EN-1-YL)OCTANAMIDE", "non-polymer"),
        "TR2": ("THIENO[3,2-B]PYRIDIN-7-OL", "non-polymer"),
        "SRT": ("S,R MESO-TARTARIC ACID", "non-polymer"),
        "W23": (
            "2-[(TRIFLUOROMETHYL)SULFONYL]-1,2,3,4-TETRAHYDROISOQUINOLIN-6-OL",
            "non-polymer",
        ),
        "ELI": (
            "6-(3-METHYL-1,4-DIOXO-1,4-DIHYDRONAPHTHALEN-2-YL)HEXANOIC ACID",
            "non-polymer",
        ),
        "SRY": ("STREPTOMYCIN", "non-polymer"),
        "HLZ": ("1-HYDRAZINOPHTHALAZINE", "non-polymer"),
        "TDK": (
            "3-[(4-AMINO-2-METHYLPYRIMIDIN-5-YL)METHYL]-2-{(1S)-1-HYDROXY-1-[(R)-HYDROXY(METHOXY)PHOSPHORYL]ETHYL}-5-(2-{[(S)-HYDROXY(PHOSPHONOOXY)PHOSPHORYL]OXY}ETHYL)-4-METHYL-1,3-THIAZOL-3-IUM",
            "non-polymer",
        ),
        "25H": (
            "{2-[2-(2-AMINO-6-OXO-1,6-DIHYDRO-9H-PURIN-9-YL)ETHOXY]ETHYL}PHOSPHONIC ACID",
            "non-polymer",
        ),
        "NMU": ("N-METHYLUREA", "non-polymer"),
        "RH0": (
            "3-AMINO-5-S-BENZYL-2,3-DIDEOXY-5-THIO-D-ERYTHRO-PENTONIC ACID",
            "peptide",
        ),
        "NMS": (
            "1-(O2-(2-METHYLAMINO-2-OXO-ETHYL)-O5-HYDROXYPHOSPHINYL-BETA-D-RIBOFURANOSYL)THYMINE",
            "nucleic-acid",
        ),
        "NMP": (
            "2-[2-(2-CYCLOHEXYL-2-GUANIDINO-ACETYLAMINO)-ACETYLAMINO]-N-(3-MERCAPTO-PROPYL)-PROPIONAMIDE",
            "non-polymer",
        ),
        "NMQ": (
            "N-METHYL O-NITROPHENYL AMINOETHYLDIPHOSPHATE BERYLLIUM TRIFLUORIDE",
            "non-polymer",
        ),
        "25B": (
            "(1R,2R)-2-[(4-CHLOROPHENYL)CARBONYL]-N-{(1S)-1-[2-(PHENYLSULFONYL)ETHYL]PENTYL}CYCLOHEXANECARBOXAMIDE",
            "non-polymer",
        ),
        "TDB": (
            "6-METHYL-2(PROPANE-1-SULFONYL)-2H-THIENO[3,2-D][1,2,3]DIAZABORININ-1-OL",
            "non-polymer",
        ),
        "TDA": ("N-TRIDECANOIC ACID", "non-polymer"),
        "25A": (
            "5'-O-MONOPHOSPHORYLADENYLYL(2'->5')ADENYLYL(2'->5')ADENOSINE",
            "non-polymer",
        ),
        "TDG": ("THIODIGALACTOSIDE", "saccharide"),
        "NR9": (
            "(5Z)-5-(QUINOLIN-6-YLMETHYLIDENE)-2-[(THIOPHEN-2-YLMETHYL)AMINO]-1,3-THIAZOL-4(5H)-ONE",
            "non-polymer",
        ),
        "25D": (
            "2-CYCLOHEXYL-N-[(3-{[(2,4,6-TRIMETHYLPHENYL)CARBAMOYL]AMINO}NAPHTHALEN-2-YL)CARBONYL]-D-ALANINE",
            "non-polymer",
        ),
        "25E": (
            "N-({(2Z)-2-[(4-HYDROXYPHENYL)METHYLIDENE]HYDRAZINO}CARBONOTHIOYL)-BETA-D-GLUCOPYRANOSYLAMINE",
            "non-polymer",
        ),
        "25Z": (
            "[4-AMINO-2-(PHENYLAMINO)-1,3-THIAZOL-5-YL][3-(TRIFLUOROMETHYL)PHENYL]METHANONE",
            "non-polymer",
        ),
        "HN0": ("?", "nucleic-acid"),
        "HN3": (
            "1-(4-METHYLPHENYL)-2-[(1S,2R,5R,8R,8AR)-1,2,8-TRIHYDROXYOCTAHYDROINDOLIZIN-5-YL]ETHANONE",
            "non-polymer",
        ),
        "HN2": (
            "1-PHENYL-2-[(1S,2R,5R,8R,8AR)-1,2,8-TRIHYDROXYOCTAHYDROINDOLIZIN-5-YL]ETHANONE",
            "non-polymer",
        ),
        "HN5": (
            "(1S,2R,5S,8R,8AR)-5-[2-(4-TERT-BUTYLPHENYL)ETHYL]OCTAHYDROINDOLIZINE-1,2,8-TRIOL",
            "non-polymer",
        ),
        "HN4": (
            "1-(4-TERT-BUTYLPHENYL)-2-[(1S,2R,5R,8R,8AR)-1,2,8-TRIHYDROXYOCTAHYDROINDOLIZIN-5-YL]ETHANONE",
            "non-polymer",
        ),
        "HN7": (
            "1-(4-TERT-BUTYLPHENYL)-2-[(1S,2R,5S,8R,8AR)-1,2,8-TRIHYDROXYOCTAHYDROINDOLIZIN-5-YL]ETHANONE",
            "non-polymer",
        ),
        "HN6": (
            "(1S,2R,5R,8R,8AR)-5-[2-(4-TERT-BUTYLPHENYL)ETHYL]OCTAHYDROINDOLIZINE-1,2,8-TRIOL",
            "non-polymer",
        ),
        "25R": ("2,5-DIMETHYLPYRAZINE", "non-polymer"),
        "NMO": ("NITROGEN MONOXIDE", "non-polymer"),
        "25P": (
            "4-[(2S)-2-{[4-(4-CHLOROPHENOXY)PHENOXY]METHYL}PYRROLIDIN-1-YL]BUTANOATE",
            "non-polymer",
        ),
        "NMM": (
            "(2S)-2-AMINO-5-[(N-METHYLCARBAMIMIDOYL)AMINO]PENTANOIC ACID",
            "peptide",
        ),
        "BC1": (
            "3-{[(9-CYANO-9,10-DIHYDRO-10-METHYLACRIDIN-9-YL)CARBONYL]AMINO}PROPANOIC ACID",
            "non-polymer",
        ),
        "25T": ("2-AMINO-5-METHYLTHIAZOLE", "non-polymer"),
        "BC3": (
            "2-AMINO-7-(PYRIDIN-3-YLMETHYL)-3,5-DIHYDRO-4H-PYRROLO[3,2-D]PYRIMIDIN-4-ONE",
            "non-polymer",
        ),
        "076": (
            "(S)-N-((1R,2S)-1-((2R,3R)-3-BENZYLOXIRAN-2-YL)-1-HYDROXY-3-PHENYLPROPAN-2-YL)-3-METHYL-2-(2-PHENOXYACETAMIDO)BUTANAMIDE",
            "peptide",
        ),
        "077": ("(3S)-5-BIPHENYL-4-YL-3-HYDROXYPENTANOIC ACID", "non-polymer"),
        "074": (
            "[PROPYLAMINO-3-HYDROXY-BUTAN-1,4-DIONYL]-ISOLEUCYL-PROLINE",
            "peptide",
        ),
        "075": ("KNI-10075", "peptide"),
        "072": (
            "(+/-)(2S,5S)-3-(4-(4-CARBOXYPHENYL)BUTYL)-2-HEPTYL-4-OXO-5-THIAZOLIDINE",
            "non-polymer",
        ),
        "073": (
            "(4-{(2S,4E)-2-(1H-1,2,3-BENZOTRIAZOL-1-YL)-2-[4-(METHOXYCARBONYL)PHENYL]-5-PHENYLPENT-4-ENYL}PHENYL)(DIFLUORO)METHYLPHOSPHONIC ACID",
            "non-polymer",
        ),
        "070": (
            "(1R,2S)-2-AZANYL-3-PHENYL-1-[(2R,3R)-3-(PHENYLMETHYL)OXIRAN-2-YL]PROPAN-1-OL",
            "non-polymer",
        ),
        "071": (
            "8-{[2-METHOXY-5-(4-METHYLPIPERAZIN-1-YL)PHENYL]AMINO}-1-METHYL-4,5-DIHYDRO-1H-PYRAZOLO[4,3-H]QUINAZOLINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "MFA": ("ALPHA-L-1-METHYL-FUCOSE", "saccharide"),
        "FHG": (
            "2'-DEOXY-5'-O-[(R)-{[(R)-[(S)-FLUORO(PHOSPHONO)METHYL](HYDROXY)PHOSPHORYL]OXY}(HYDROXY)PHOSPHORYL]GUANOSINE",
            "non-polymer",
        ),
        "FHF": (
            "N-((2R,4S)-2-BUTYL-4-(3-(2-FLUOROPHENYL)UREIDO)-5-METHYL-3-OXOHEXYL)-N-HYDROXYFORMAMIDE",
            "non-polymer",
        ),
        "FHA": (
            "2'-DEOXY-5'-O-[(S)-{(S)-FLUORO[(S)-HYDROXY(PHOSPHONOOXY)PHOSPHORYL]METHYL}(HYDROXY)PHOSPHORYL]ADENOSINE",
            "non-polymer",
        ),
        "078": (
            "(2R,3R,4R,5S)-3,5-BIS(AZANYL)-1,6-DIPHENYL-HEXANE-2,4-DIOL",
            "non-polymer",
        ),
        "079": (
            "(S)-N-((2S,3S,4R,5R)-4-AMINO-3,5-DIHYDROXY-1,6-DIPHENYLHEXAN-2-YL)-3-METHYL-2-(2-PHENOXYACETAMIDO)BUTANAMIDE",
            "peptide",
        ),
        "EB4": (
            "N,N',N''-[(3S,7S,11S)-2,6,10-TRIOXO-1,5,9-TRIOXACYCLODODECANE-3,7,11-TRIYL]TRIS(2,3-DIHYDROXYBENZAMIDE)",
            "non-polymer",
        ),
        "LOS": ("?", "non-polymer"),
        "LOP": (
            "(1R)-2-{[(R)-(2-AMINOETHOXY)(HYDROXY)PHOSPHORYL]OXY}-1-[(DODECANOYLOXY)METHYL]ETHYL (9Z)-OCTADEC-9-ENOATE",
            "non-polymer",
        ),
        "ZPZ": (
            "N-[(1R)-1-(4-FLUOROPHENYL)ETHYL]-N'-[(2S,3S)-3-HYDROXY-4-{4-[(1S)-1-HYDROXYETHYL]-1H-1,2,3-TRIAZOL-1-YL}-1-PHENYLBUTAN-2-YL]-5-[METHYL(METHYLSULFONYL)AMINO]BENZENE-1,3-DICARBOXAMIDE",
            "non-polymer",
        ),
        "LOV": ("5-AMINO-4-HYDROXY-2-ISOPROPYL-7-METHYL-OCTANOIC ACID", "peptide"),
        "EB1": ("SEOCALCITOL", "non-polymer"),
        "NR2": (
            "N-({[4-(AMINOSULFONYL)PHENYL]AMINO}CARBONYL)-4-METHYLBENZENESULFONAMIDE",
            "non-polymer",
        ),
        "FHU": (
            "(5S,6R)-5-FLUORO-6-HYDROXY-PSEUDOURIDINE-5'-MONOPHOSPHATE",
            "nucleic-acid",
        ),
        "LOX": ("3,4,5-TRIHYDOXY-PIPERIDINE-2-ONE-OXIME", "saccharide"),
        "NOC": (
            "3-(6-AMINO-PURIN-9-YL)-5-HYDROXYMETHYL-CYCLOPENTANE-1,2-DIOL",
            "non-polymer",
        ),
        "FHP": (
            "1-HYDROXY-3,7,11-TRIMETHYLDODECA-2,6,10-TRIENE PHOSPHONIC ACID",
            "non-polymer",
        ),
        "MMC": ("METHYL MERCURY ION", "non-polymer"),
        "HLO": (
            "1-[({2,4-BIS[(E)-(HYDROXYIMINO)METHYL]PYRIDINIUM-1-YL}METHOXY)METHYL]-4-CARBAMOYLPYRIDINIUM",
            "non-polymer",
        ),
        "UFM": ("URIDINE-5'-DIPHOSPHATE-MANNOSE", "non-polymer"),
        "HLN": ("HEXYL(NAPHTHALEN-2-YLOXY)PHOSPHINIC ACID", "non-polymer"),
        "PS9": ("OCTATHIOCANE", "non-polymer"),
        "H33": ("8-CHLORO-1,3-DIMETHYL-3,7-DIHYDRO-1H-PURINE-2,6-DIONE", "non-polymer"),
        "BE9": ("PHENYL N-SULFAMOYLMORPHOLINE-4-CARBOXIMIDOATE", "non-polymer"),
        "H37": (
            "D-PHENYLALANYL-N-[(3S)-6-CARBAMIMIDAMIDO-1-CHLORO-2-OXOHEXAN-3-YL]-L-PHENYLALANINAMIDE",
            "peptide",
        ),
        "H34": (
            "(1-{4-[(7S,11S)-12-AMINO-3-CHLORO-6,7,10,11-TETRAHYDRO-7,11-METHANOCYCLOOCTA[B]QUINOLIN-9-YL]BUTYL}-1H-1,2,3-TRIAZOL-4-YL)METHANOL",
            "non-polymer",
        ),
        "H35": ("N-(FURAN-2-YLMETHYL)-7H-PURIN-6-AMINE", "non-polymer"),
        "NR4": (
            "(1S,2R,3S,4R)-1-{(1S)-2-[(2R,3S,4S)-3,4-DIHYDROXY-2-(HYDROXYMETHYL)TETRAHYDROTHIOPHENIUM-1-YL]-1-HYDROXYETHYL}-2,3,4,5-TETRAHYDROXYPENTYL SULFATE",
            "non-polymer",
        ),
        "MY3": (
            "[N-(3-BENZYLCARBAMOYL-OXIRANECARBONYL)-HYDRAZINO] - ACETIC ACID",
            "non-polymer",
        ),
        "A5T": (
            "(1R,5S)-7-{4-[3-(2-CHLORO-3,6-DIFLUOROPHENOXY)PROPYL]PHENYL}-N-CYCLOPROPYL-N-(2,3-DICHLOROBENZYL)-3,9-DIAZABICYCLO[3.3.1]NON-6-ENE-6-CARBOXAMIDE",
            "non-polymer",
        ),
        "TFT": ("(L)-ALPHA-THREOFURANOSYL-THYMINE-3'-MONOPHOSPHATE", "nucleic-acid"),
        "NOZ": ("(2S,3S,4R,5R)-5-(HYDROXYMETHYL)PIPERIDINE-2,3,4-TRIOL", "non-polymer"),
        "H3K": (
            "1-{5-BROMO-2-[(3R)-3-HYDROXYPIPERIDIN-1-YL]PHENYL}-3-(5-CYANOPYRAZIN-2-YL)UREA",
            "non-polymer",
        ),
        "BE2": ("2-AMINOBENZOIC ACID", "peptide"),
        "H3P": ("2,2'-METHANEDIYLBIS(3,4,6-TRICHLOROPHENOL)", "non-polymer"),
        "P7Y": ("2-[(E)-2-NITROETHENYL]PHENOL", "non-polymer"),
        "BE0": ("PIPERIDINE-1-SULFONAMIDE", "non-polymer"),
        "GYN": ("GYMNODIMINE A", "non-polymer"),
        "07V": ("(5R)-2-METHYL-5-(PROP-1-EN-2-YL)CYCLOHEX-2-EN-1-ONE", "non-polymer"),
        "07T": (
            "6-(3-AMINOBENZYL)-4-METHYL-2-METHYLSULFINYL-4,6-DIHYDRO-5H-THIENO[2',3':4,5]PYRROLO[2,3-D]PYRIDAZIN-5-ONE",
            "non-polymer",
        ),
        "07U": (
            "2-METHYL-N~1~-[3-(PYRIDIN-4-YL)-2,6-NAPHTHYRIDIN-1-YL]PROPANE-1,2-DIAMINE",
            "non-polymer",
        ),
        "07R": (
            "1-[(3-HYDROXYPHENYL)METHYL]-3-(4-PYRIDIN-4-YL-1,3-THIAZOL-2-YL)UREA",
            "non-polymer",
        ),
        "07S": (
            "1-(MORPHOLIN-4-YL)-2-[4-(2-{[5-(PYRIDIN-3-YL)-1,3-THIAZOL-2-YL]AMINO}PYRIDIN-4-YL)PIPERAZIN-1-YL]ETHANONE",
            "non-polymer",
        ),
        "07P": (
            "N-[(2S,3R)-10-[(1R,2R)-2-DECYLCYCLOPROPYL]-1-(ALPHA-D-GALACTOPYRANOSYLOXY)-3-HYDROXYDECAN-2-YL]HEXACOSANAMIDE",
            "non-polymer",
        ),
        "07Q": (
            "1-[(1R)-1-(3-METHOXYPHENYL)ETHYL]-3-(4-PYRIDIN-4-YL-1,3-THIAZOL-2-YL)UREA",
            "non-polymer",
        ),
        "EBM": ("2-IODO-N-(1-METHYLETHYL)BENZAMIDE", "non-polymer"),
        "EBN": (
            "(S)-6-AMINO-N-((S)-6-AMINO-1-(((1R,4S)-4-GUANIDINOCYCLOHEXYL)METHYLAMINO)-1-OXOHEXAN-2-YL)-2-(2-(3,4-DICHLOROPHENYL)ACETAMIDO)HEXANAMIDE",
            "non-polymer",
        ),
        "07Z": (
            "4-{[4-AMINO-5-(PYRIDIN-3-YLCARBONYL)-1,3-THIAZOL-2-YL]AMINO}BENZOIC ACID",
            "non-polymer",
        ),
        "EBI": (
            "1-{4-[(3-CYCLOPROPYL-1H-PYRAZOL-5-YL)AMINO]PYRROLO[2,1-F][1,2,4]TRIAZIN-2-YL}-N-(6-FLUOROPYRIDIN-3-YL)-2-METHYL-L-PROLINAMIDE",
            "non-polymer",
        ),
        "NOQ": (
            "(1R,2R)-N-(2-AMINOETHYL)-2-{[(4-METHOXYPHENYL)SULFONYL]METHYL}CYCLOHEXANECARBOXAMIDE",
            "non-polymer",
        ),
        "EBT": (
            "N-[(2R,3R)-3-AMINO-2-HYDROXY-4-PHENYLBUTANOYL]-L-LEUCINE",
            "non-polymer",
        ),
        "07G": (
            "2-(TERT-BUTYLAMINO)-1-(2-CARBOXY-6-CHLORO-1H-INDOL-3-YL)-1-[(3,4-DIFLUOROBENZYL)(FORMYL)AMINO]-2-OXOETHYLIUM",
            "non-polymer",
        ),
        "EBW": (
            "4-(5-{4-[DIMETHYL(PROP-2-ENYL)AMMONIO]PHENYL}-3-OXOPENTYL)-N,N-DIMETHYL-N-PROP-2-ENYLBENZENAMINIUM",
            "non-polymer",
        ),
        "07B": (
            "DIMETHYL (1S,2S,4S,5S)-4,5-DIHYDROXYCYCLOHEXANE-1,2-DICARBOXYLATE",
            "non-polymer",
        ),
        "EBQ": ("2-[(2R)-OXIRAN-2-YL]ETHYL ALPHA-D-GLUCOPYRANOSIDE", "saccharide"),
        "RJA": ("RATJADONE A, BOUND FORM", "non-polymer"),
        "EBS": (
            "3-ETHYL-2-[(2Z)-2-(3-ETHYL-6-SULFO-1,3-BENZOTHIAZOL-2(3H)-YLIDENE)HYDRAZINO]-6-SULFO-3H-1,3-BENZOTHIAZOL-1-IUM",
            "non-polymer",
        ),
        "07N": (
            "2-[(3S,4R)-4-{[(3,4-DICHLORO-5-METHYL-1H-PYRROL-2-YL)CARBONYL]AMINO}-3-FLUOROPIPERIDIN-1-YL]-1,3-THIAZOLE-5-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "07O": ("S-[(2S)-2-PHENYLPROPANOYL]-L-CYSTEINE", "peptide"),
        "07L": ("7-HYDROXY-2H-CHROMEN-2-ONE", "non-polymer"),
        "07M": (
            "8-{[(2-CYANOPYRIDIN-3-YL)METHYL]SULFANYL}-6-HYDROXY-3,4-DIHYDRO-1H-PYRANO[3,4-C]PYRIDINE-5-CARBONITRILE",
            "non-polymer",
        ),
        "07J": (
            "3-(2,6-DICHLORO-3,5-DIMETHOXYPHENYL)-1-(6-{[4-(4-ETHYLPIPERAZIN-1-YL)PHENYL]AMINO}PYRIMIDIN-4-YL)-1-METHYLUREA",
            "non-polymer",
        ),
        "07K": ("8-METHYLQUINOLIN-4-AMINE", "non-polymer"),
        "07H": (
            "[4-(2H-CHROMEN-3-YLMETHYL)PIPERAZIN-1-YL]-[3-(1H-PYRAZOL-5-YL)PHENYL]METHANONE",
            "non-polymer",
        ),
        "DRY": (
            "[(1-{3-[(6-BENZOYL-1-PROPYL-2-NAPHTHYL)OXY]PROPYL}-1H-INDOL-5-YL)OXY]ACETIC ACID",
            "non-polymer",
        ),
        "CX1": (
            "N-[(BENZYLOXY)CARBONYL]-L-HISTIDYL-N-METHYL-L-PHENYLALANYL-L-TYROSINE",
            "non-polymer",
        ),
        "CX2": (
            "2'-DEOXY-5'-O-{(R)-HYDROXY[(2-SULFANYLETHYL)AMINO]PHOSPHORYL}CYTIDINE",
            "nucleic-acid",
        ),
        "DRZ": ("3',4'-DIHYDROXY-PENTANAL-5'-PHOSPHATE", "nucleic-acid"),
        "CX4": ("7-CHLORO-N-(CYCLOPROPYLMETHYL)QUINAZOLIN-4-AMINE", "non-polymer"),
        "01W": ("(2S)-2-AMMONIO-4-[(2,4-DINITROPHENYL)AMINO]BUTANOATE", "peptide"),
        "P0D": (
            "(2S)-3-[(S)-[(1R)-1-AMINOETHYL](PHOSPHONOOXY)PHOSPHORYL]-2-BENZYLPROPANOIC ACID",
            "non-polymer",
        ),
        "DRQ": (
            "(9ALPHA,13BETA,17BETA)-2-[(1Z)-BUT-1-EN-1-YL]ESTRA-1,3,5(10)-TRIENE-3,17-DIOL",
            "non-polymer",
        ),
        "CX9": (
            "3,7-DIMETHYL-1-[3-(3-METHYL-2,6-DIOXO-9H-PURIN-1-YL)PROPYL]PURINE-2,6-DIONE",
            "non-polymer",
        ),
        "DRS": (
            "(9S,12S)-9-(1-METHYLETHYL)-7,10-DIOXO-2-OXA-8,11-DIAZABICYCLO[12.2.2]OCTADECA-1(16),14,17-TRIENE-12-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "DRR": ("?", "non-polymer"),
        "DRU": ("?", "non-polymer"),
        "DRT": ("?", "nucleic-acid"),
        "DRW": (
            "(2S,3R,4S)-2-[(2S,3R)-3-HYDROXY-1-OXOBUTAN-2-YL]-3-METHYL-4-({(3S,5S)-5-[(SULFAMOYLAMINO)METHYL]PYRROLIDIN-3-YL}SULFANYL)-3,4-DIHYDRO-2H-PYRROLE-5-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "DRV": (
            "[(2S,3R,4R,5R)-5-(6-AMINO-9H-PURIN-9-YL)-3,4-DIHYDROXYTETRAHYDROFURAN-2-YL]METHYL (AMINOACETYL)SULFAMATE",
            "non-polymer",
        ),
        "DRI": ("4-O-METHYL-2,6-DIDEOXY-BETA-D-GLUCOSE", "saccharide"),
        "DRH": (
            "(2S)-2-(4-{2-[1,3-BENZOXAZOL-2-YL(HEPTYL)AMINO]ETHYL}PHENOXY)-2-METHYLBUTANOIC ACID",
            "non-polymer",
        ),
        "DRK": (
            "3-({[(3S)-3,4-DIHYDROXYBUTYL]OXY}AMINO)-1H,2'H-2,3'-BIINDOL-2'-ONE",
            "non-polymer",
        ),
        "DRJ": (
            "(2R)-2-(4-{2-[1,3-BENZOXAZOL-2-YL(HEPTYL)AMINO]ETHYL}PHENOXY)-2-METHYLBUTANOIC ACID",
            "non-polymer",
        ),
        "DRM": (
            "{[(1R,2S)-2-(2,4-DIOXO-3,4-DIHYDROPYRIMIDIN-1(2H)-YL)CYCLOPENTYL]OXY}METHYLPHOSPHONIC ACID",
            "nucleic-acid",
        ),
        "DRL": (
            "2-METHYL-3,5,7,8-TETRAHYDRO-4H-THIOPYRANO[4,3-D]PYRIMIDIN-4-ONE",
            "non-polymer",
        ),
        "DRO": (
            "1-(2-{[(3S)-3-(AMINOMETHYL)-3,4-DIHYDROISOQUINOLIN-2(1H)-YL]CARBONYL}PHENYL)-4-CHLORO-5-METHYL-N,N-DIPHENYL-1H-PYRAZOLE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "D64": ("6-(4-METHYLPHENYL)QUINAZOLINE-2,4-DIAMINE", "non-polymer"),
        "DRA": ("11-DEOXY-BETA-RHODOMYCIN", "non-polymer"),
        "DRC": (
            "9-AMINO-N-[2-(4-MORPHOLINYL)ETHYL]-4-ACRIDINECARBOXAMIDE",
            "non-polymer",
        ),
        "DRB": ("?", "non-polymer"),
        "DRE": ("N-(6-AMINOHEXYL)HEXANE-1,6-DIAMINE", "non-polymer"),
        "D67": (
            "2-(1,1-DIFLUOROETHYL)-5-METHYL-N-[4-(TRIFLUOROMETHYL)PHENYL][1,2,4]TRIAZOLO[1,5-A]PYRIMIDIN-7-AMINE",
            "non-polymer",
        ),
        "DRG": ("5,6-DIHYDRO-BENZO[H]CINNOLIN-3-YLAMINE", "non-polymer"),
        "DRF": (
            "(2S)-2-ETHOXY-3-{4-[2-(10H-PHENOXAZIN-10-YL)ETHOXY]PHENYL}PROPANOIC ACID",
            "non-polymer",
        ),
        "INJ": (
            "5-(6-D-RIBITYLAMINO-2,4-DIHYDROXYPYRIMIDIN-5-YL)-1-PENTYL-PHOSPHONIC ACID",
            "non-polymer",
        ),
        "D66": (
            "N-[(2S,3R)-3-AMINO-2-HYDROXY-4-(4-METHOXYPHENYL)BUTANOYL]-L-LEUCINE",
            "non-polymer",
        ),
        "D61": (
            "(2R,3R)-2-(CARBOXYMETHOXY)-4-{[5-(NAPHTHALEN-2-YL)PENTYL]AMINO}-3-{[5-(NAPHTHALEN-2-YL)PENTYL]OXY}-4-OXOBUTANOIC ACID",
            "non-polymer",
        ),
        "D63": (
            "4'-(5-CARBAMIMIDOYL-1H-BENZIMIDAZOL-2-YL)BIPHENYL-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "LM2": ("unknown", "saccharide"),
        "INK": (
            "(RP,SP)-O-(2R)-(1-PHENOXYBUT-2-YL)-METHYLPHOSPHONIC ACID CHLORIDE",
            "non-polymer",
        ),
        "U25": ("URIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "SPG": ("?", "non-polymer"),
        "PDN": ("17,21-DIHYDROXYPREGNA-1,4-DIENE-3,11,20-TRIONE", "non-polymer"),
        "CXP": ("CYCLOHEXANE PROPIONIC ACID", "non-polymer"),
        "DR8": ("N,N,N-TRIMETHYLHEPTA-1,3,5-TRIYN-1-AMINIUM", "non-polymer"),
        "CXR": ("CYCLIC ADENOSINE DIPHOSPHATE-RIBOSE", "non-polymer"),
        "CXS": ("3-CYCLOHEXYL-1-PROPYLSULFONIC ACID", "non-polymer"),
        "CXT": ("CARBOXYATRACTYLOSIDE", "non-polymer"),
        "CXU": ("CLOXACILLIN (OPEN FORM)", "non-polymer"),
        "CXV": (
            "(2R,4S)-2-[(1S)-1-({[3-(2-CHLOROPHENYL)-5-METHYL-1,2-OXAZOL-4-YL]CARBONYL}AMINO)-2-OXOETHYL]-5,5-DIMETHYL-1,3-THIAZOLIDINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "CXW": (
            "2-({[2-({(3S,4S)-4-[(6-AMINO-4-METHYLPYRIDIN-2-YL)METHYL]PYRROLIDIN-3-YL}OXY)ETHYL]AMINO}METHYL)PHENOL",
            "non-polymer",
        ),
        "CXX": (
            "3-(3-CHLORO-5H-DIBENZO[B,F]AZEPIN-5-YL)-N,N-DIMETHYLPROPAN-1-AMINE",
            "non-polymer",
        ),
        "CXY": ("CYTIDINE-5'-DIPHOSPHO-BETA-D-XYLOSE", "non-polymer"),
        "CXZ": (
            "4-CHLORO-6-{5-[(2-MORPHOLIN-4-YLETHYL)AMINO]-1,2-BENZISOXAZOL-3-YL}BENZENE-1,3-DIOL",
            "non-polymer",
        ),
        "DR2": (
            "6-DEOXY-ALPHA-L-IDOPYRANOSYL-(1->2)-ALPHA-D-GLUCOPYRANOSYL-(1->4)-2-(ACETYLAMINO)-2-DEOXY-BETA-L-GULOPYRANOSE",
            "saccharide",
        ),
        "DR5": (
            "METHYL 4-O-L-GLUCOPYRANOSYL-ALPHA-L-RIBO-HEXOPYRANOSIDE",
            "non-polymer",
        ),
        "DR4": ("METHYL 9-(BETA-D-TALOPYRANOSYLOXY)NONANOATE", "saccharide"),
        "DR7": (
            "(3S,8S,9S,12S)-3,12-BIS(1,1-DIMETHYLETHYL)-8-HYDROXY-4,11-DIOXO-9-(PHENYLMETHYL)-6-[[4-(2-PYRIDINYL)PHENYL]METHYL]-2,5,6,10,13-PENTAAZATETRADECANEDIOIC ACID DIMETHYL ESTER",
            "non-polymer",
        ),
        "DR6": (
            "ALPHA-[4-(1,1,3,3 - TETRAMETHYLBUTYL)PHENYL]-OMEGA-HYDROXY-POLY(OXY-1,2-ETHANEDIYL)",
            "non-polymer",
        ),
        "P03": ("2-[3-[3-(2-HYDROXYETHOXY)PROPOXY]PROPOXY]ETHANOL", "non-polymer"),
        "CXA": ("PHENYLALANINE-N-SULFONAMIDE", "non-polymer"),
        "CXB": (
            "[(2-AMINO-ALPHA-METHOXYIMINO-4-THIAZOLYLACETYL)AMINO]METHYLBORONIC ACID",
            "non-polymer",
        ),
        "CXD": (
            "4-CHLORANYL-N-[[4-CHLORANYL-3-(3-CHLORANYL-5-CYANO-PHENOXY)-2-FLUORANYL-PHENYL]METHYL]-1H-IMIDAZOLE-5-CARBOXAMIDE",
            "non-polymer",
        ),
        "CXE": ("PENTAETHYLENE GLYCOL MONODECYL ETHER", "non-polymer"),
        "CXF": ("CYCLOHEXYLFORMAMIDE", "non-polymer"),
        "CXG": (
            "METHYL N-[(2S)-1-[2-[(4R)-5-[[(2S)-3,3-DIMETHYL-1-METHYLAMINO-1-OXO-BUTAN-2-YL]AMINO]-4-HYDROXY-5-OXO-4-(PHENYLMETHYL)PENTYL]-2-[(4-PHENYLPHENYL)METHYL]HYDRAZINYL]-3,3-DIMETHYL-1-OXO-BUTAN-2-YL]CARBAMATE",
            "non-polymer",
        ),
        "O1C": (
            "2ALPHA-(3-HYDROXYPROPYL)-1ALPHA,25-DIHYDROXYVITAMIN D3",
            "non-polymer",
        ),
        "P09": (
            "2-{[4-(3,4-DIMETHYLPHENOXY)PHENYL]CARBAMOYL}BENZOIC ACID",
            "non-polymer",
        ),
        "INE": ("3-BROMO-7-NITROINDAZOLE", "non-polymer"),
        "CXL": ("CYCLOHEXANOL", "non-polymer"),
        "CXM": ("N-CARBOXYMETHIONINE", "peptide"),
        "CXN": ("CLOXACILLIN", "non-polymer"),
        "CXO": ("CYCLOHEXANE-1,2-DIONE", "non-polymer"),
        "MLB": ("6-O-ALPHA-D-GALACTOPYRANOSYL-ALPHA-D-GLUCOPYRANOSE", "saccharide"),
        "2VA": ("2'-(L-VALYL)AMINO-2'-DEOXYADENOSINE", "peptide"),
        "AV9": ("TIVOZANIB", "non-polymer"),
        "MLC": ("MALONYL-COENZYME A", "non-polymer"),
        "ODS": ("4-METHYLPIPERAZINE-1-CARBOXYLIC ACID", "peptide"),
        "AV6": ("4-(2,3-DIHYDRO-1H-PERIMIDIN-2-YL)BENZENE-1,2-DIOL", "non-polymer"),
        "AV1": (
            "[(2R,3S,4R,5R)-5-(6-AMINO-8-BROMO-9H-PURIN-9-YL)-3,4-DIHYDROXYTETRAHYDROFURAN-2-YL]METHYL [(2R,3S,4S)-3,4-DIHYDROXYTETRAHYDROFURAN-2-YL]METHYL DIHYDROGEN DIPHOSPHATE",
            "non-polymer",
        ),
        "AV2": ("ADENOSINE-5'-DIPHOSPHATE-2',3'-VANADATE", "non-polymer"),
        "MP6": (
            "4-(2-CYCLOPROPYLETHYLIDENE)-9-(1H-PYRAZOL-4-YL)-6-{[(1R)-1,2,2-TRIMETHYLPROPYL]AMINO}BENZO[C][1,6]NAPHTHYRIDIN-1(4H)-ONE",
            "non-polymer",
        ),
        "MP7": (
            "1-(3,4-DIFLUOROBENZYL)-2-OXO-N-{(1R)-2-[(2-OXO-2,3-DIHYDRO-1H-BENZIMIDAZOL-5-YL)OXY]-1-PHENYLETHYL}-1,2-DIHYDROPYRIDINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "MP4": ("(2S)-AMINO(3,5-DIHYDROXY-4-METHYLPHENYL)ETHANOIC ACID", "peptide"),
        "MP5": (
            "(4-AMINO-2-METHYLPYRIMIDIN-5-YL)METHYL DIHYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "MP2": ("N-[(BENZYLOXY)CARBONYL]-L-CYSTEINYLGLYCINE", "non-polymer"),
        "MP3": ("3-PHOSPHOGLYCERIC ACID", "non-polymer"),
        "MP0": (
            "(2R,4S,5S)-5-AMINO-6-CYCLOHEXYL-4-HYDROXY-N-METHYL-2-(2-METHYLPROPYL)HEXANAMIDE",
            "peptide",
        ),
        "MP1": ("N-METHYLMESOPORPHYRIN CONTAINING COPPER", "non-polymer"),
        "INF": ("D-[(N-HYDROXYAMINO)CARBONYL]PHENYLALANINE", "non-polymer"),
        "ED8": (
            "5-{6-[(1-METHYLPIPERIDIN-4-YL)OXY]-1H-BENZIMIDAZOL-1-YL}-3-{[2-(TRIFLUOROMETHYL)BENZYL]OXY}THIOPHENE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "MP8": ("(4R)-4-METHYL-L-PROLINE", "peptide"),
        "MP9": (
            "2-AMINO-3-[3-HYDROXY-5-(2-BENZYL-2H-5-TETRAZOLYL)-4-ISOXAZOLYL]-PROPIONIC ACID",
            "non-polymer",
        ),
        "3TD": (
            "(1S)-1,4-ANHYDRO-1-(3-METHYL-2,4-DIOXO-1,2,3,4-TETRAHYDROPYRIMIDIN-5-YL)-5-O-PHOSPHONO-D-RIBITOL",
            "nucleic-acid",
        ),
        "3TF": (
            "(2S)-1-(ALPHA-D-GLUCOPYRANOSYLOXY)-3-(HEXADECANOYLOXY)PROPAN-2-YL (11Z)-OCTADEC-11-ENOATE",
            "non-polymer",
        ),
        "3TG": ("1-(1H-IMIDAZOL-1-YL)-4,4-DIPHENYLBUTAN-2-ONE", "non-polymer"),
        "3TA": (
            "9-CHLORO-2-({5-[3-(DIMETHYLAMINO)PROPYL]-2-METHYLPYRIDIN-3-YL}AMINO)-5,7-DIHYDRO-6H-PYRIMIDO[5,4-D][1]BENZAZEPINE-6-THIONE",
            "non-polymer",
        ),
        "3TB": (
            "N-[(1E,2R)-1-[(2R)-2-{[(2S)-1-AMINO-5-CARBAMIMIDAMIDO-1-OXOPENTAN-2-YL]CARBAMOYL}CYCLOPENTYLIDENE]-3-(PHOSPHONOOXY)PROPAN-2-YL]-L-PHENYLALANINAMIDE",
            "non-polymer",
        ),
        "3TC": (
            "4-AMINO-1-[(2R,5S)-2-(HYDROXYMETHYL)-1,3-OXATHIOLAN-5-YL]PYRIMIDIN-2(1H)-ONE",
            "non-polymer",
        ),
        "3TL": (
            "BENZYL [(1S,4S,7S,8R,9R,10S,13S,16S)-7,10-DIBENZYL-8,9-DIHYDROXY-1,16-DIMETHYL-4,13-BIS(1-METHYLETHYL)-2,5,12,15,18-PENTAOXO-20-PHENYL-19-OXA-3,6,11,14,17-PENTAAZAICOS-1-YL]CARBAMATE",
            "peptide",
        ),
        "3TN": (
            "(2E)-3-[4-HYDROXY-3-(3-METHOXY-5,5,8,8-TETRAMETHYL-5,6,7,8-TETRAHYDRONAPHTHALEN-2-YL)PHENYL]ACRYLIC ACID",
            "non-polymer",
        ),
        "3TO": (
            "3-[2-(3-{[(FURAN-2-YLMETHYL)(METHYL)AMINO]METHYL}PHENYL)ETHYL]PYRIDIN-2-AMINE",
            "non-polymer",
        ),
        "3TH": (
            "2-CHLORO-N-[(3R)-2-OXO-1,2,3,4-TETRAHYDROQUINOLIN-3-YL]-6H-THIENO[2,3-B]PYRROLE-5-CARBOXAMIDE",
            "non-polymer",
        ),
        "3TI": ("1-{(E)-[(4-HYDROXYPHENYL)IMINO]METHYL}NAPHTHALEN-2-OL", "non-polymer"),
        "3TJ": (
            "4-ETHYL-7-FLUORO-3,4-DIHYDRO-2H-1,2,4-BENZOTHIADIAZINE 1,1-DIOXIDE",
            "non-polymer",
        ),
        "3TU": (
            "(2S)-2-(5-{[(2-AMINO-4-OXO-3,4-DIHYDRO[1]BENZOTHIENO[2,3-D]PYRIMIDIN-5-YL)METHYL]AMINO}-1-OXO-1,3-DIHYDRO-2H-ISOINDOL-2-YL)PENTANEDIOIC ACID",
            "non-polymer",
        ),
        "Z5A": (
            "P1-(5'-ADENOSYL)P5-(5'-(3'AZIDO-3'-DEOXYTHYMIDYL))PENTAPHOSPHATE",
            "non-polymer",
        ),
        "3TP": (
            "(2S,3S)-3-{3-[4-(METHYLSULFONYL)PHENYL]-1,2,4-OXADIAZOL-5-YL}-1-OXO-1-PYRROLIDIN-1-YLBUTAN-2-AMINE",
            "non-polymer",
        ),
        "3TR": ("3-AMINO-1,2,4-TRIAZOLE", "non-polymer"),
        "ING": ("D-[(AMINO)CARBONYL]PHENYLALANINE", "non-polymer"),
        "RL2": (
            "N-[3-CARBOXY-2-HYDROXY-PROPIONYL]-L-HOMOPHENYLALANYL-AMINO-2-METHYLBUTANE",
            "peptide",
        ),
        "3TY": (
            "3-[(3E)-3-(BENZYLHYDRAZONO)-4-HYDROXY-6-OXOCYCLOHEXA-1,4-DIEN-1-YL]-L-ALANINE",
            "peptide",
        ),
        "3TZ": (
            "N-[(4-{[(2-AMINO-4-OXO-3,4-DIHYDRO[1]BENZOTHIENO[2,3-D]PYRIMIDIN-5-YL)METHYL]AMINO}PHENYL)CARBONYL]-L-GLUTAMIC ACID",
            "non-polymer",
        ),
        "T5S": ("?", "nucleic-acid"),
        "Z57": (
            "(1S)-1,5-ANHYDRO-1-(4-BROMO-2,5-DIHYDROXYPHENYL)-D-GLUCITOL",
            "non-polymer",
        ),
        "MLN": ("(S)-2-(PHOSPHONOXY)CAPROYL-L-LEUCYL-P-NITROANILIDE", "non-polymer"),
        "Z59": (
            "(2R,3R)-2,3-DIHYDROXY-4-OXO-4-[(2R)-2-PHENYLPYRROLIDIN-1-YL]-N-(THIOPHEN-2-YLMETHYL)BUTANAMIDE",
            "non-polymer",
        ),
        "3T4": (
            "{5-CHLORO-2-[(4-CYANO-3-NITROBENZYL)CARBAMOYL]PHENOXY}ACETIC ACID",
            "non-polymer",
        ),
        "3T5": (
            "(11ALPHA,16ALPHA)-9-FLUORO-11,17-DIHYDROXY-16-METHYL-3,20-DIOXOPREGNA-1,4-DIEN-21-YL DIHYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "VG3": (
            "N-[(1S,2R)-1-BENZYL-3-{[(1S)-2-(CYCLOHEXYLAMINO)-1-METHYL-2-OXOETHYL]AMINO}-2-HYDROXYPROPYL]-3-(PENTYLSULFONYL)BENZAMIDE",
            "non-polymer",
        ),
        "3T0": (
            "(5Z)-5-(5-BROMO-2-METHOXYBENZYLIDENE)-3-(4-METHYLBENZYL)-1,3-THIAZOLIDINE-2,4-DIONE",
            "non-polymer",
        ),
        "3T1": (
            "(3S,11S)-11-(3-CHLORO-4-HYDROXY-5-METHOXYPHENYL)-3-PHENYL-2,3,4,5,10,11-HEXAHYDRO-1H-DIBENZO[B,E][1,4]DIAZEPIN-1-ONE",
            "non-polymer",
        ),
        "MLM": ("3-AMINO-3-OXOPROPANOIC ACID", "non-polymer"),
        "3T8": (
            "N~2~-{2-CHLORO-4-[(4-METHYLPIPERAZIN-1-YL)CARBONYL]PHENYL}-N~2~,N~8~-DIMETHYL-4,5-DIHYDROTHIENO[3,2-D][1]BENZOXEPINE-2,8-DICARBOXAMIDE",
            "non-polymer",
        ),
        "3T9": (
            "2-METHOXY-4-(3-PHENYL-2H-PYRAZOLO[3,4-B]PYRIDIN-6-YL)PHENOL",
            "non-polymer",
        ),
        "MLR": ("MALTOTRIOSE", "non-polymer"),
        "MPD": ("(4S)-2-METHYL-2,4-PENTANEDIOL", "non-polymer"),
        "MPE": ("(1H-INDOL-3-YL)-(2-MERCAPTO-ETHOXYIMINO)-ACETIC ACID", "non-polymer"),
        "AVX": (
            "3-(7-BROMO-1,3-BENZODIOXOL-5-YL)-1-METHYL-1H-PYRAZOL-5-AMINE",
            "non-polymer",
        ),
        "MPC": ("(1S)-MENTHYL HEXYL PHOSPHONATE GROUP", "non-polymer"),
        "MPA": ("(1R)-MENTHYL HEXYL PHOSPHONATE GROUP", "non-polymer"),
        "MPN": (
            "PHOSPHORIC ACID MONO-(2-AMINO-4-OXO-5,6-DITHIOXO-1,5,6,7,8A,9,10,10A-OCTAHYDRO-4H-8-OXA-1,3,9,10-TETRAAZA-ANTHRACEN-7-YLMETHYL) ESTER",
            "non-polymer",
        ),
        "AVU": (
            "[(2R,3S,4R,5R)-5-(6-AMINO-9H-PURIN-9-YL)-3,4-DIHYDROXYTETRAHYDROFURAN-2-YL]METHYL [(2R,3R,4R)-4-FLUORO-3-HYDROXYTETRAHYDROFURAN-2-YL]METHYL DIHYDROGEN DIPHOSPHATE",
            "non-polymer",
        ),
        "AVV": (
            "[(2R,3S,4R,5R)-5-(6-AMINO-2-FLUORO-9H-PURIN-9-YL)-3,4-DIHYDROXYTETRAHYDROFURAN-2-YL]METHYL [(2R,3S,4S)-3,4-DIHYDROXYTETRAHYDROFURAN-2-YL]METHYL DIHYDROGEN DIPHOSPHATE",
            "non-polymer",
        ),
        "AVW": (
            "[(2R,3S,4R,5R)-5-(6-AMINO-9H-PURIN-9-YL)-3,4-DIHYDROXYTETRAHYDROFURAN-2-YL]METHYL [(2R,3R,4S)-4-FLUORO-3-HYDROXYTETRAHYDROFURAN-2-YL]METHYL DIHYDROGEN DIPHOSPHATE",
            "non-polymer",
        ),
        "AVP": ("ACYCLOVIR TRIPHOSPHATE", "non-polymer"),
        "AVQ": (
            "2-[(2,5-DICHLOROBENZYL)SULFANYL]-5-METHYL[1,2,4]TRIAZOLO[1,5-A]PYRIMIDIN-7-OL",
            "non-polymer",
        ),
        "AVR": (
            "(2R)-2-AMINO-1-[2-(1-METHYLETHYL)PYRAZOLO[1,5-A]PYRIDIN-3-YL]PROPAN-1-ONE",
            "non-polymer",
        ),
        "MPI": ("IMIDAZO[1,2-A]PYRIDINE", "non-polymer"),
        "MPV": ("1-METHYL-3-(THIOPHEN-2-YL)-1H-PYRAZOL-5-AMINE", "non-polymer"),
        "MPW": (
            "2-[2-(4-METHOXYPYRIDIN-2-YL)ETHYL]-3H-IMIDAZO[4,5-B]PYRIDINE",
            "non-polymer",
        ),
        "AVN": (
            "(2S)-AMINO[(5S)-3-CHLORO-4,5-DIHYDROISOXAZOL-5-YL]ACETIC ACID",
            "peptide",
        ),
        "INA": ("1,5-BIS(N-BENZYLOXYCARBONYL-L-LEUCINYL)CARBOHYDRAZIDE", "non-polymer"),
        "MPR": ("2-MERCAPTO-PROPION ALDEHYDE", "non-polymer"),
        "MPS": ("METHYLPHOSPHINIC ACID", "non-polymer"),
        "MPP": ("3-(3,4-DIMETHOXYPHENYL)PROPIONIC ACID", "non-polymer"),
        "MPQ": ("N-METHYL-ALPHA-PHENYL-GLYCINE", "peptide"),
        "AVD": (
            "1-(2-CARBOXYPHENYL)-7-CHLORO-6-[(2-CHLORO-4,6-DIFLUOROPHENYL)AMINO]-4-OXO-1,4-DIHYDROQUINOLINE-3-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "AVE": (
            "4-[3-(2-CHLORO-4,5-DIFLUORO-BENZOYL)UREIDO]-3-TRIFLUOROMETHOXYBENZOIC ACID",
            "non-polymer",
        ),
        "AVF": (
            "1-{2-[3-(2-CHLORO-4,5-DIFLUORO-BENZOYL)-UREIDO]-4-FLUORO-PHENYL}-PIPERIDINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "AVG": ("2-AMINO-4-(2-AMINO-ETHOXY)-BUTYRIC ACID", "non-polymer"),
        "MPZ": (
            "4-[(4-METHYLPIPERAZIN-1-YL)METHYL]-N-{3-[(4-PYRIDIN-3-YLPYRIMIDIN-2-YL)AMINO]PHENYL}BENZAMIDE",
            "non-polymer",
        ),
        "MPX": (
            "4-({[(4-METHYLPIPERAZIN-1-YL)AMINO]CARBONOTHIOYL}AMINO)BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "AVC": ("ADENOSINE-5'-MONOPHOSPHATE-2',3'-VANADATE", "nucleic-acid"),
        "0TO": ("N-(3-CHLOROPHENYL)-6,7-DIMETHOXYQUINAZOLIN-4-AMINE", "non-polymer"),
        "0TN": ("DELTA-RU(PHEN)2(DPPZ) COMPLEX", "non-polymer"),
        "0TM": (
            "2-{[4-(2-CHLORO-4,5-DIMETHOXYPHENYL)-5-CYANO-7H-PYRROLO[2,3-D]PYRIMIDIN-2-YL]SULFANYL}-N,N-DIMETHYLACETAMIDE",
            "non-polymer",
        ),
        "0TK": ("(2R)-2,5-DIAMINOPENTANOYL DIHYDROGEN PHOSPHATE", "non-polymer"),
        "0TJ": (
            "N-(6-OXOHEXYL)-2-(4,5,6,7-TETRABROMO-1H-BENZIMIDAZOL-1-YL)ACETAMIDE",
            "non-polymer",
        ),
        "0TH": ("N-(PROPYLSULFONYL)-D-THREONINE", "peptide"),
        "0TF": ("N~6~-D-ORNITHYL-L-LYSINE", "non-polymer"),
        "0TE": (
            "CHLORO{METHYL HYDROGENATO(3-)-KAPPA~2~N,S [PYRIDIN-2-YL(PYRIDIN-2(1H)-YLIDENE-KAPPAN)METHYL]CARBONODITHIOHYDRAZONATE}COPPER",
            "non-polymer",
        ),
        "0TD": ("(3S)-3-(METHYLSULFANYL)-L-ASPARTIC ACID", "peptide"),
        "0TB": (
            "2-AMINO-6-(5-FLUORO-6-METHOXYPYRIDIN-3-YL)-4-METHYL-8-(PYRROLIDIN-1-YL)PYRIDO[2,3-D]PYRIMIDIN-7(8H)-ONE",
            "non-polymer",
        ),
        "0TA": (
            "2-AMINO-8-CYCLOPENTYL-4-METHYL-6-(1H-PYRAZOL-4-YL)PYRIDO[2,3-D]PYRIMIDIN-7(8H)-ONE",
            "non-polymer",
        ),
        "B16": ("1,6-DI-O-PHOSPHONO-BETA-D-GLUCOPYRANOSE", "saccharide"),
        "B17": (
            "{1-[4-(AMINOSULFONYL)PHENYL]-11-[(CARBOXY-KAPPAO)METHYL]-1-OXO-5,8-DIOXA-2,11-DIAZATRIDECAN-13-OATO(2-)-KAPPAO~13~}COPPER",
            "non-polymer",
        ),
        "B14": (
            "1-(6-CYANO-3-PYRIDYLCARBONYL)-5',8'-DIFLUOROSPIRO[PIPERIDINE-4,2'(1'H)-QUINAZOLINE]-4'-AMINE",
            "non-polymer",
        ),
        "B15": (
            "[2,2'-({4-[({2-[4-(AMINOSULFONYL)PHENYL]ETHYL}AMINO)CARBONYL]BENZYL}IMINO)DIACETATO(2-)-KAPPAO]COPPER",
            "non-polymer",
        ),
        "B12": ("COBALAMIN", "non-polymer"),
        "0TZ": (
            "3-(4-AMINO-6-METHYL-1,3,5-TRIAZIN-2-YL)-N-(1H-PYRAZOL-5-YL)IMIDAZO[1,2-A]PYRIDIN-2-AMINE",
            "non-polymer",
        ),
        "B10": (
            "N-{5-TERT-BUTYL-2-METHOXY-3-[({4-[6-(MORPHOLIN-4-YLMETHYL)PYRIDIN-3-YL]NAPHTHALEN-1-YL}CARBAMOYL)AMINO]PHENYL}METHANESULFONAMIDE",
            "non-polymer",
        ),
        "0TX": (
            "(4S)-N~4~-(7-CHLOROQUINOLIN-4-YL)-N~1~,N~1~-DIETHYLPENTANE-1,4-DIAMINE",
            "non-polymer",
        ),
        "0TT": (
            "1-[(2R,4S,5R)-5-(HYDROXYMETHYL)-4-OXIDANYL-OXOLAN-2-YL]-5-[[(5R)-1-[(2R,4S,5R)-5-(HYDROXYMETHYL)-4-OXIDANYL-OXOLAN-2-YL]-5-METHYL-2,4-BIS(OXIDANYLIDENE)-1,3-DIAZINAN-5-YL]METHYL]PYRIMIDINE-2,4-DIONE",
            "non-polymer",
        ),
        "0TS": (
            "BETA-D-GLUCOPYRANOSYL-(1->4)-4-THIO-BETA-D-GLUCOPYRANOSYL-(1->4)-BETA-D-GLUCOPYRANOSYL-(1->4)-4-THIO-BETA-D-GLUCOPYRANOSYL-(1->4)-1-THIO-BETA-D-GLUCOPYRANOSE",
            "saccharide",
        ),
        "0TR": ("2-HYDROXYCYCLOHEPTA-2,4,6-TRIEN-1-ONE", "non-polymer"),
        "0TQ": (
            "(3R,3AS,6AR)-HEXAHYDROFURO[2,3-B]FURAN-3-YL {(2S,3R)-3-HYDROXY-4-[({(3Z)-3-[1-(METHYLAMINO)ETHYLIDENE]-2-OXO-2,3-DIHYDRO-1H-INDOL-5-YL}SULFONYL)(2-METHYLPROPYL)AMINO]-1-PHENYLBUTAN-2-YL}CARBAMATE",
            "non-polymer",
        ),
        "0TP": (
            "4-AMINO-N-METHYL-2-[4-(MORPHOLIN-4-YL)PHENYL]THIENO[3,2-C]PYRIDINE-7-CARBOXAMIDE",
            "non-polymer",
        ),
        "1BP": ("1-BROMOPROPANE-2-OL", "non-polymer"),
        "1BQ": (
            "[4-(2-HYDROXYETHYL)PIPERIDIN-1-YL][4-(5-METHYL-4,4-DIOXIDO-1,5-DIHYDROPYRAZOLO[4,3-C][2,1]BENZOTHIAZIN-8-YL)PHENYL]METHANONE",
            "non-polymer",
        ),
        "1BR": (
            "N-(4-TERT-BUTYLBENZYL)-1,5-DIMETHYL-1,5-DIHYDROPYRAZOLO[4,3-C][2,1]BENZOTHIAZIN-8-AMINE 4,4-DIOXIDE",
            "non-polymer",
        ),
        "1BS": (
            "2-{(1S)-1-[(6-CHLORO-3,3-DIMETHYL-3,4-DIHYDROISOQUINOLIN-1-YL)AMINO]-2-PHENYLETHYL}PYRIDO[4,3-D]PYRIMIDIN-4(1H)-ONE",
            "non-polymer",
        ),
        "1BT": (
            "(2E)-3-(4-{[6-(1,3-BENZOTHIAZOL-5-YLAMINO)-9H-PURIN-2-YL]AMINO}-3,5-DIMETHYLPHENYL)PROP-2-ENENITRILE",
            "non-polymer",
        ),
        "1BU": (
            "1-{4-[(6-AMINOQUINAZOLIN-4-YL)AMINO]PHENYL}-3-[3-TERT-BUTYL-1-(3-METHYLPHENYL)-1H-PYRAZOL-5-YL]UREA",
            "non-polymer",
        ),
        "1BV": ("?", "non-polymer"),
        "1BY": (
            "[2-(PYRIDIN-2-YLAMINO)ETHANE-1,1-DIYL]BIS(PHOSPHONIC ACID)",
            "non-polymer",
        ),
        "I1P": ("2-[4-(N-(3-DIMETHYLAMINOPROPYL)SULPHAMOYL)ANILINO]-", "non-polymer"),
        "INC": (
            "2-(ACETYL-HYDROXY-AMINO)-4-METHYL-PENTANOIC ACID METHYL ESTER",
            "non-polymer",
        ),
        "7GU": ("7-DEAZA-2'-DEOXYGUANOSINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "1BA": ("4-HYDROXY-N'-(4-ISOPROPYLBENZYL)BENZOHYDRAZIDE", "non-polymer"),
        "1BB": (
            "2-{(1S)-1-[(6-CHLORO-3,3-DIMETHYL-3,4-DIHYDROISOQUINOLIN-1-YL)AMINO]-2-PHENYLETHYL}-4-OXO-1,4-DIHYDROPYRIMIDINE-5-CARBONITRILE",
            "non-polymer",
        ),
        "1BC": (
            "2-{(1S)-1-{[(1Z)-6-CHLORO-3,3-DIMETHYL-3,4-DIHYDROISOQUINOLIN-1(2H)-YLIDENE]AMINO}-2-[2-PROPYL-4-(1H-PYRAZOL-4-YL)THIOPHEN-3-YL]ETHYL}PYRIMIDIN-4(5H)-ONE",
            "non-polymer",
        ),
        "1BD": (
            "3-[({(1E)-[2-(TRIFLUOROMETHYL)PHENYL]METHYLIDENE}AMINO)OXY]PROPANOIC ACID",
            "non-polymer",
        ),
        "7GP": ("ETHYL-N-(BETA-D-GLUCOPYRANOSYL)OXAMATE", "non-polymer"),
        "1BF": (
            "N-(6-CHLORO-3,3-DIMETHYL-3,4-DIHYDROISOQUINOLIN-1-YL)-3-[4-(1H-PYRAZOL-4-YL)THIOPHEN-3-YL]-L-ALANINE",
            "non-polymer",
        ),
        "1BG": (
            "(3S,8AR)-2-{(2S)-2-CYCLOHEXYL-2-[(N-METHYL-L-ALANYL)AMINO]ACETYL}-N-[(1R)-1,2,3,4-TETRAHYDRONAPHTHALEN-1-YL]OCTAHYDROPYRROLO[1,2-A]PYRAZINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "1BH": (
            "N-BENZYLOXYCARBONYL-ALA-PRO-3-AMINO-4-PHENYL-BUTAN-2-OL",
            "non-polymer",
        ),
        "1BI": (
            "3-CYCLOHEXYL-2-(FURAN-3-YL)-1-[2-(MORPHOLIN-4-YL)-2-OXOETHYL]-N-(PHENYLSULFONYL)-1H-INDOLE-6-CARBOXAMIDE",
            "non-polymer",
        ),
        "X7E": (
            "(4R)-5-[(S)-(3,4-DIFLUOROPHENYL)(HYDROXY)METHYL]-4-(3-HYDROXYPHENYL)-1,6-DIMETHYL-3,4-DIHYDROPYRIMIDINE-2(1H)-THIONE",
            "non-polymer",
        ),
        "1BL": (
            "(2R)-5-{[(2S,3R)-4-{[1-(3-TERT-BUTYLPHENYL)CYCLOHEXYL]AMINO}-1-(3,5-DIFLUOROPHENYL)-3-HYDROXYBUTAN-2-YL]AMINO}-2-HYDROXY-5-OXOPENTANOIC ACID",
            "non-polymer",
        ),
        "1BM": (
            "3-{[2-(1H-BENZIMIDAZOL-1-YL)-6-{[2-(DIETHYLAMINO)ETHYL]AMINO}PYRIMIDIN-4-YL]AMINO}-4-METHYLPHENOL",
            "non-polymer",
        ),
        "1BN": ("1-BENZYL-1H-IMIDAZOLE", "non-polymer"),
        "1BO": ("1-BUTANOL", "non-polymer"),
        "1B0": (
            "N-METHYL-NALPHA-[(2-METHYL-1H-INDOL-3-YL)ACETYL]-N-PHENYL-L-PHENYLALANINAMIDE",
            "non-polymer",
        ),
        "1B1": (
            "(2R)-2-CYCLOPROPYL-5-METHYL-2-[3-(5-PROP-1-YN-1-YLPYRIDIN-3-YL)PHENYL]-2H-IMIDAZOL-4-AMINE",
            "non-polymer",
        ),
        "I19": (
            "N-(4-{[(3S)-3-(DIMETHYLAMINO)PYRROLIDIN-1-YL]CARBONYL}PHENYL)-5-FLUORO-4-[2-METHYL-1-(1-METHYLETHYL)-1H-IMIDAZOL-5-YL]PYRIMIDIN-2-AMINE",
            "non-polymer",
        ),
        "TF2": (
            "(N-{4-[(ETHYLANILINO)SULFONYL]-2-METHYLPHENYL}-3,3,3-TRIFLUORO-2-HYDROXY-2-METHYLPROPANAMIDE",
            "non-polymer",
        ),
        "1B8": (
            "3-[2-BROMO-4-(1H-PYRAZOL-4-YL)THIOPHEN-3-YL]-N-(6-CHLORO-3,3-DIMETHYL-3,4-DIHYDROISOQUINOLIN-1-YL)-L-ALANINE",
            "non-polymer",
        ),
        "1B9": (
            "3-(4-BROMOTHIOPHEN-3-YL)-N-(6-CHLORO-3,3-DIMETHYL-3,4-DIHYDROISOQUINOLIN-1-YL)-L-ALANINE",
            "non-polymer",
        ),
        "I17": (
            "2-[TRANS-(4-AMINOCYCLOHEXYL)AMINO]-6-(BENZYL-AMINO)-9-CYCLOPENTYLPURINE",
            "non-polymer",
        ),
        "X75": (
            "4-CHLORO-2-{[(6-CHLOROPYRIDIN-3-YL)METHYL]AMINO}-5-NITROBENZAMIDE",
            "non-polymer",
        ),
        "I11": (
            "[[CYCLOHEXANESULFONYL-GLYCYL]-3[PYRIDIN-4-YL-AMINOMETHYL]ALANYL]PIPERIDINE",
            "non-polymer",
        ),
        "I10": (
            "1-[2-(3-BIPHENYL)-4-METHYLVALERYL)]AMINO-3-(2-PYRIDYLSULFONYL)AMINO-2-PROPANONE",
            "non-polymer",
        ),
        "I12": (
            "N-[(5-METHYLISOXAZOL-3-YL)CARBONYL]-L-ALANYL-L-VALYL-N~1~-((1S)-4-ETHOXY-4-OXO-1-{[(3S)-2-OXOPYRROLIDIN-3-YL]METHYL}BUT-2-ENYL)-L-LEUCINAMIDE",
            "non-polymer",
        ),
        "MD0": ("5-AMINO-D-ISOLEUCINE", "peptide"),
        "B1F": ("PHENYLALANINE BORONIC ACID", "peptide"),
        "G3H": ("GLYCERALDEHYDE-3-PHOSPHATE", "non-polymer"),
        "B1E": (
            "3-(2-CYANOPROPAN-2-YL)-N-{4-METHYL-3-[(3-METHYL-4-OXO-3,4-DIHYDROQUINAZOLIN-6-YL)AMINO]PHENYL}BENZAMIDE",
            "non-polymer",
        ),
        "B1L": (
            "3-[(4-HYDROXYBENZOYL)AMINO]AZEPAN-4-YL 4-HYDROXYBENZOATE",
            "non-polymer",
        ),
        "B1M": ("CO-5-METHOXYBENZIMIDAZOLYLCOBAMIDE", "non-polymer"),
        "B1V": (
            "2-[(2-OXO-2-PIPERIDIN-1-YLETHYL)SULFANYL]-6-(TRIFLUOROMETHYL)PYRIMIDIN-4-OL",
            "non-polymer",
        ),
        "B1T": ("2,2'-SULFANEDIYLBIS(4,6-DICHLOROPHENOL)", "non-polymer"),
        "B1P": ("2-DEOXY-5-O-PHOSPHONO-BETA-D-ERYTHRO-PENTOFURANOSE", "nucleic-acid"),
        "0T8": (
            "4-(CYCLOPENTYLAMINO)-N-[3-(2-OXOPYRROLIDIN-1-YL)PROPYL]-2-{[2-(PYRIDIN-4-YL)ETHYL]AMINO}PYRIMIDINE-5-CARBOXAMIDE",
            "non-polymer",
        ),
        "0T7": (
            "3,4-DIMETHYL-1-PROPYL-7-(QUINOLIN-2-YLMETHOXY)IMIDAZO[1,5-A]QUINOXALINE",
            "non-polymer",
        ),
        "0T6": (
            "1-(2-CHLOROPHENYL)-6,8-DIMETHOXY-3-METHYLIMIDAZO[5,1-C][1,2,4]BENZOTRIAZINE",
            "non-polymer",
        ),
        "0T5": (
            "3-NITRO-N-[(1R)-1-PHENYLETHYL]-5-(TRIFLUOROMETHYL)BENZAMIDE",
            "non-polymer",
        ),
        "0T4": (
            "3-(HYDROXYAMINO)-N-[(1R)-1-PHENYLETHYL]-5-(TRIFLUOROMETHYL)BENZAMIDE",
            "non-polymer",
        ),
        "0T3": (
            "4-[(5R)-5,6,7,8-TETRAHYDROIMIDAZO[1,5-A]PYRIDIN-5-YL]BENZONITRILE",
            "non-polymer",
        ),
        "KZL": (
            "5-[(2-AMINO-1,1-DIMETHYLETHYL)AMINO]-7-[(3,5-DIMETHOXYPHENYL)AMINO][1,2,4]TRIAZOLO[4,3-C]PYRIMIDINE-8-CARBOXAMIDE",
            "non-polymer",
        ),
        "0T0": (
            "3-(3,4-DIHYDROISOQUINOLIN-2(1H)-YLSULFONYL)-N-METHYLBENZAMIDE",
            "non-polymer",
        ),
        "151": (
            "(2S)-3-METHYL-2-((2R,3S)-3-[(METHYLSULFONYL)AMINO]-1-{[2-(PYRROLIDIN-1-YLMETHYL)-1,3-OXAZOL-4-YL]CARBONYL}PYRROLIDIN-2-YL)BUTANOIC ACID",
            "non-polymer",
        ),
        "150": ("4,5-DIMETHYL-1,2-PHENYLENEDIAMINE", "non-polymer"),
        "153": (
            "(2S)-2-[(2,4-DICHLORO-BENZOYL)-(3-TRIFLUOROMETHYL-BENZYL)-AMINO]-3-PHENYL-PROPIONIC ACID",
            "non-polymer",
        ),
        "152": ("CARNITINE", "non-polymer"),
        "155": (
            "6-[(Z)-AMINO(IMINO)METHYL]-N-(1-ISOPROPYL-3,4-DIHYDROISOQUINOLIN-7-YL)-2-NAPHTHAMIDE",
            "non-polymer",
        ),
        "154": (
            "(2S)-2-[(5-BENZOFURAN-2-YL-THIOPHEN-2-YLMETHYL)-(2,4-DICHLORO-BENZOYL)-AMINO]-3-PHENYL-PROPIONIC ACID",
            "non-polymer",
        ),
        "157": ("6-(2-HYDROXY-CYCLOPENTYL)-7-OXO-HEPTANAMIDINE", "non-polymer"),
        "156": (
            "4-[3-OXO-3-(5,5,8,8-TETRAMETHYL-5,6,7,8-TETRAHYDRO-NAPHTHALEN-2-YL)-PROPENYL]-BENZOIC ACID",
            "non-polymer",
        ),
        "159": ("2-HYDROXY-5-(2-MERCAPTO-ETHYLSULFAMOYL)-BENZOIC ACID", "non-polymer"),
        "158": ("3-(2-MERCAPTO-ACETYLAMINO)-4-OXO-PENTANOIC ACID", "non-polymer"),
        "4EB": (
            "4-({2-[(2,4-DIMETHYLPHENYL)SULFANYL]ETHYL}AMINO)-N-[(4'-FLUORO-1,1'-BIPHENYL-4-YL)CARBONYL]-3-NITROBENZENESULFONAMIDE",
            "non-polymer",
        ),
        "JDT": ("?", "nucleic-acid"),
        "MD6": ("N-[(3-HYDROXYPYRIDIN-2-YL)CARBONYL]GLYCINE", "peptide"),
        "JFQ": ("TRANS-N-(3-AMINOPROPYL)CYCLOHEXANE-1,4-DIAMINE", "non-polymer"),
        "IBK": (
            "N-ISOBUTYL-2-{[5-(THIOPHEN-2-YL)-1,2-OXAZOL-3-YL]METHOXY}ACETAMIDE",
            "non-polymer",
        ),
        "JDR": (
            "4-({[6-CHLORO-5-FORMYL-2-(METHYLSULFANYL)PYRIMIDIN-4-YL]AMINO}METHYL)BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "MFT": ("?", "nucleic-acid"),
        "E": ("?", "nucleic-acid"),
        "XAE": (
            "3-(2-DEOXY-5-O-PHOSPHONO-BETA-D-ERYTHRO-PENTOFURANOSYL)-3H-IMIDAZO[4,5-G]QUINAZOLIN-8-AMINE",
            "nucleic-acid",
        ),
        "SD8": (
            "2-[4-(2H-1,4-BENZOTHIAZINE-3-YL)-PIPERAZINE-1-LY]-1,3-THIAZOLE-4-CARBOXYLIC ACID ETHYLESTER",
            "non-polymer",
        ),
        "15Q": (
            "2-CHLORO-3'-FLUORO-3-[(E)-(HYDROXYIMINO)METHYL]BIPHENYL-4,4'-DIOL",
            "non-polymer",
        ),
        "15P": ("POLYETHYLENE GLYCOL (N=34)", "non-polymer"),
        "15S": (
            "N-{(2S)-1-[4-(4-FLUOROPHENYL)-3,6-DIHYDROPYRIDIN-1(2H)-YL]-1-OXOPROPAN-2-YL}-2-[(9AR)-7-OXO-2,3,7,8,9,9A-HEXAHYDRO-1H-BENZO[DE][1,7]NAPHTHYRIDIN-1-YL]ACETAMIDE",
            "non-polymer",
        ),
        "15R": (
            "(9AR)-1-[(1-{2-FLUORO-5-[(4-OXO-3,4-DIHYDROPHTHALAZIN-1-YL)METHYL]BENZOYL}PIPERIDIN-4-YL)CARBONYL]-1,2,3,8,9,9A-HEXAHYDRO-7H-BENZO[DE][1,7]NAPHTHYRIDIN-7-ONE",
            "non-polymer",
        ),
        "15U": (
            "(2R,4R)-4-METHYL-1-(N~2~-{[(3S)-3-METHYL-1,2,3,4-TETRAHYDROQUINOLIN-8-YL]SULFONYL}-L-ARGINYL)PIPERIDINE-2-CARBOXYLIC ACID",
            "peptide",
        ),
        "15W": ("7-HYDROXY-2-(4-METHOXYPHENYL)-4H-CHROMEN-4-ONE", "non-polymer"),
        "15V": (
            "N-[1-(3-CHLOROPHENYL)-3-METHYL-1H-PYRAZOL-5-YL]PYRAZOLO[1,5-A]PYRIMIDINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "15Y": (
            "(3S)-3-METHYL-1,2,3,4-TETRAHYDROQUINOLINE-8-SULFONIC ACID",
            "non-polymer",
        ),
        "15X": (
            "1-{4-[(2-FLUORO-6-METHOXY-3'-NITROBIPHENYL-3-YL)METHYL]PHENYL}UREA",
            "non-polymer",
        ),
        "15Z": ("2-(3,4-DIHYDROXYPHENYL)-7-HYDROXY-4H-CHROMEN-4-ONE", "non-polymer"),
        "ROI": ("7,8-DIHYDRO-7,7-DIMETHYL-6-HYDROXYPTERIN", "non-polymer"),
        "15A": (
            "(2Z,3R,4S,5R,6R)-2-[(2-AMINOETHYL)IMINO]-6-(HYDROXYMETHYL)PIPERIDINE-3,4,5-TRIOL",
            "non-polymer",
        ),
        "15C": (
            "2-(4-ACETYLPHENYL)-4-METHYL-1H-ISOINDOLE-1,3(2H)-DIONE",
            "non-polymer",
        ),
        "15B": (
            "N-{3-[4-(3-AMINO-PROPYL)-PIPERAZIN-1-YL]-PROPYL}-3-NITRO-5-(GALACTOPYRANOSYL)-BETA-BENZAMIDE",
            "non-polymer",
        ),
        "15E": ("6-BROMO-3-METHYL-3,4-DIHYDROQUINAZOLIN-2(1H)-ONE", "non-polymer"),
        "15G": (
            "2-{2-[(3,4-DIFLUOROPHENOXY)METHYL]-5-METHOXYPYRIDIN-4-YL}-1,5,6,7-TETRAHYDRO-4H-PYRROLO[3,2-C]PYRIDIN-4-ONE",
            "non-polymer",
        ),
        "15F": (
            "1-BENZYL-4-({[(1E)-2,2,2-TRICHLOROETHANIMIDOYL]OXY}METHYL)PYRIDINIUM",
            "non-polymer",
        ),
        "15I": ("6-(2,9-DIAZASPIRO[5.5]UNDECAN-9-YL)-9H-PURINE", "non-polymer"),
        "15H": (
            "(1S)-1-(1-{3-[4-(1,3-BENZOTHIAZOL-2-YLAMINO)PHENOXY]PYRAZIN-2-YL}PIPERIDIN-4-YL)ETHANOL",
            "non-polymer",
        ),
        "15J": (
            "(1-{3-[4-(1H-BENZIMIDAZOL-2-YLAMINO)PHENOXY]PYRIDIN-2-YL}PIPERIDIN-4-YL)METHANOL",
            "non-polymer",
        ),
        "15M": (
            "(5Z)-7-{(1R,2R,3R,5S)-3,5-DIHYDROXY-2-[(1E,3S)-3-HYDROXY-5-PHENYLPENT-1-ENYL]CYCLOPENTYL}-N-ETHYLHEPT-5-ENAMIDE",
            "non-polymer",
        ),
        "15L": ("GALACTARO-1,5-LACTONE", "saccharide"),
        "4E6": ("(4E,6E)-HEXADECA-1,4,6-TRIENE", "non-polymer"),
        "MYD": (
            "{[5-(6-AMINO-PURIN-7-YL)-3,4-DIHYDROXY-TETRAHYDRO-FURAN-2-YLMETHOXY]-HYDROXY-PHOSPHORYLMETHYL}-PHOSPHONIC ACID MONO-[2-(4-HYDROXY-6-METHOXY-7-METHYL-3-OXO-1,3-DIHYDRO-ISOBENZOFURAN-5-YL)-ETHYL] ESTER",
            "non-polymer",
        ),
        "2KT": ("2-KETOBUTYRIC ACID", "non-polymer"),
        "2KU": ("N-(3-METHOXYPROPYL)-9H-PURIN-6-AMINE", "non-polymer"),
        "NZU": (
            "6-(2H-INDAZOL-4-YL)-1-METHYL-N-[3-(METHYLSULFONYL)PROPYL]-1H-PYRAZOLO[3,4-D]PYRIMIDIN-4-AMINE",
            "non-polymer",
        ),
        "8FG": (
            "N-(5'-PHOSPHO-2'-DEOXYGUANOSIN-8-YL)-2-ACETYLAMINOFLUORENE",
            "nucleic-acid",
        ),
        "F19": (
            "2-[(2-NAPHTHYLSULFONYL)AMINO]ETHYL DIHYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "2KL": (
            "2-AMINO-4-[2,4-DICHLORO-5-(2-PYRROLIDIN-1-YLETHOXY)PHENYL]-N-ETHYLTHIENO[2,3-D]PYRIMIDINE-6-CARBOXAMIDE",
            "non-polymer",
        ),
        "TCG": ("TMG-CHITOTRIOMYCIN", "non-polymer"),
        "Z3E": ("O-BENZYL-L-THREONINE", "peptide"),
        "9PR": ("N-METHYL-L-PROLINAMIDE", "peptide"),
        "9PP": ("2,6-DIAMINO-(S)-9-[2-(PHOSPHONOMETHOXY)PROPYL]PURINE", "non-polymer"),
        "9PQ": (
            "4-[2-(HYDROXYMETHYL)NAPHTHALEN-1-YL]-N-[2-(MORPHOLIN-4-YL)ETHYL]BENZAMIDE",
            "non-polymer",
        ),
        "818": ("18-HYDROXYASCOMYCIN", "non-polymer"),
        "819": (
            "ETHYL (4-{3-[2,4-DIHYDROXY-5-(1-METHYLETHYL)PHENYL]-5-SULFANYL-4H-1,2,4-TRIAZOL-4-YL}BENZYL)CARBAMATE",
            "non-polymer",
        ),
        "9PE": (
            "(1R)-2-{[(S)-(2-AMINOETHOXY)(HYDROXY)PHOSPHORYL]OXY}-1-[(HEPTANOYLOXY)METHYL]ETHYL OCTADECANOATE",
            "non-polymer",
        ),
        "VA3": ("TRIVANADATE", "non-polymer"),
        "812": ("N-BENZYL-4-[(2R)-PYRROLIDIN-2-YLMETHOXY]ANILINE", "non-polymer"),
        "NBY": (
            "N-{[(4-NITROPHENYL)CARBONYL]CARBAMOYL}-BETA-D-GLUCOPYRANOSYLAMINE",
            "non-polymer",
        ),
        "814": (
            "[(3R,5S,6R,11S,12Z,14E)-5,11,21-TRIMETHOXY-3,7,9,15-TETRAMETHYL-6-OXIDANYL-16,20,22-TRIS(OXIDANYLIDENE)-19-PHENYL-17-AZABICYCLO[16.3.1]DOCOSA-1(21),8,12,14,18-PENTAEN-10-YL] CARBAMATE",
            "non-polymer",
        ),
        "815": (
            "THIENO[3,2-B]PYRIDINE-2-SULFONIC ACID [1-(1-AMINO-ISOQUINOLIN-7-YLMETHYL)-2-OXO-PYRROLDIN-3-YL]-AMIDE",
            "non-polymer",
        ),
        "9PL": (
            "(3S,4R)-3-ETHYL-4-[(1-METHYL-1H-IMIDAZOL-5-YL)METHYL]DIHYDROFURAN-2(3H)-ONE",
            "non-polymer",
        ),
        "817": (
            "1-{[N-(1-IMINO-GUANIDINO-METHYL)]SULFANYLMETHYL}-3-TRIFLUOROMETHYL-BENZENE",
            "non-polymer",
        ),
        "EDB": (
            "2-{3-[(5E)-5-{[(2,3-DICHLOROPHENYL)CARBAMOYL]IMINO}-3-THIOPHEN-2-YL-2,5-DIHYDRO-1H-PYRAZOL-1-YL]PHENYL}ACETAMIDE",
            "non-polymer",
        ),
        "VAH": ("(3R)-3-HYDROXY-L-NORVALINE", "peptide"),
        "VAI": ("(2S)-2-AMINO-3-METHYLBUTANE-1,1-DIOL", "peptide"),
        "VAL": ("VALINE", "peptide"),
        "O16": (
            "3-(1,3-BENZOTHIAZOL-2-YL)-2-(1,4-DIOXO-1,2,3,4-TETRAHYDROPHTHALAZIN-6-YL)-5-[(E)-2-PHENYLVINYL]-3H-TETRAAZOL-2-IUM",
            "non-polymer",
        ),
        "VAB": (
            "4-OXO-4-{[2-(PROP-2-EN-1-YLSULFANYL)-1,3-BENZOTHIAZOL-6-YL]AMINO}BUTANOIC ACID",
            "non-polymer",
        ),
        "81A": (
            "2-METHANESULFONYL-BENZENESULFONIC ACID 3-METHYL-5-((1-AMIDINOAMINOOXYMETHYL-CYCLOPROPYL)METHYLOXY)-PHENYLESTER",
            "non-polymer",
        ),
        "VAA": ("N-[VALINYL]-N'-[ADENOSYL]-DIAMINOSUFONE", "non-polymer"),
        "VAF": ("METHYLVALINE", "peptide"),
        "LMZ": ("5-NITROSO-6-RIBITYL-AMINO-2,4(1H,3H)-PYRIMIDINEDIONE", "non-polymer"),
        "VAD": ("DEAMINOHYDROXYVALINE", "peptide"),
        "VAE": (
            "3-{[(E)-1-CARBOXY-2-PHENYLETHENYL]OXY}-2-HYDROXYBENZOIC ACID",
            "non-polymer",
        ),
        "VAZ": (
            "N^6^-{(1R)-2-{[(1S,2R)-1-CARBOXY-2-HYDROXY-2-(METHYLSULFANYL)ETHYL]OXY}-1-[(OXIDOSULFANYL)METHYL]-2-OXOETHYL}-6-OXO-L-LYSINE",
            "non-polymer",
        ),
        "O14": (
            "(E)-4-((1-METHYLPIPERIDIN-4-YLOXYIMINO)METHYL)BENZIMIDAMIDE",
            "non-polymer",
        ),
        "VAR": (
            "9-AMINO-5-(2-AMINOPYRIMIDIN-4-YL)PYRIDO[3',2':4,5]PYRROLO[1,2-C]PYRIMIDIN-4-OL",
            "non-polymer",
        ),
        "RLI": ("?", "non-polymer"),
        "VAP": ("6,7,8-TRIMETHOXY-1,4-DIMETHYLQUINOLIN-2(1H)-ONE", "non-polymer"),
        "T6Q": (
            "2-[[3-(4-ETHANOYL-1,4-DIAZEPAN-1-YL)PHENYL]AMINO]-4-[4-METHYL-2-(METHYLAMINO)-1,3-THIAZOL-5-YL]PYRIMIDINE-5-CARBONITRILE",
            "non-polymer",
        ),
        "VAW": (
            "(2S,3R,4S)-METHYL 4-(2-(2-(1H-INDOL-3-YL)ETHYLAMINO)ETHYL)-2-((2S,3R,4S,5S,6R)-3,4,5-TRIHYDROXY-6-(HYDROXYMETHYL)TETRAHYDRO-2H-PYRAN-2-YLOXY)-3-VINYL-3,4-DIHYDRO-2H-PYRAN-5-CARBOXYLATE",
            "non-polymer",
        ),
        "O12": ("N~5~-DODECANOYL-L-ORNITHINE", "peptide"),
        "2K4": ("1H-INDAZOL-6-OL", "non-polymer"),
        "TQU": ("4-(1H-1,2,4-TRIAZOL-1-YL)QUINOLIN-6-AMINE", "non-polymer"),
        "RLJ": (
            "(2S)-3-(4-CYANOPHENOXY)-N-[4-CYANO-3-(TRIFLUOROMETHYL)PHENYL]-2-HYDROXY-2-METHYLPROPANAMIDE",
            "non-polymer",
        ),
        "TQV": ("O-CRESYL-SALIGENIN PHOSPHATE", "non-polymer"),
        "30E": (
            "(3R,4R)-1-[3-AMINO-6-(3,4,5-TRIMETHOXYPHENYL)PYRAZIN-2-YL]-3-ETHYLPIPERIDINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "YOM": (
            "2,2'-[1,2-PHENYLENEBIS(NITRILOMETHYLIDYNE)]BIS[PHENOLATO]](2-)-N,N',O,O']-IRON",
            "non-polymer",
        ),
        "RSS": (
            "5-METHYL-3-(9-OXO-1,8-DIAZA-TRICYCLO[10.6.1.013,18]NONADECA-12(19),13,15,17-TETRAEN-10-YLCARBAMOYL)-HEXANOIC ACID",
            "non-polymer",
        ),
        "WUL": (
            "8-BROMANYL-2-[[(3S)-3-OXIDANYLPYRROLIDIN-1-YL]METHYL]-3H-[1]BENZOFURO[3,2-D]PYRIMIDIN-4-ONE",
            "non-polymer",
        ),
        "308": ("(3S,5S,7S)-TRICYCLO[3.3.1.1~3,7~]DECAN-1-AMINE", "non-polymer"),
        "6PP": ("3-[(2Z)-3-METHYLPENT-2-EN-1-YL]BENZENE-1,2-DIOL", "non-polymer"),
        "6PR": (
            "(S)-4-AMINO-4-OXO-3-(2-PHOSPHONOACETAMIDO)BUTANOIC ACID",
            "non-polymer",
        ),
        "LMT": ("DODECYL-BETA-D-MALTOSIDE", "saccharide"),
        "MMB": (
            "(2R,3E)-2-{4-[(5-METHYL-2-PHENYL-1,3-OXAZOL-4-YL)METHOXY]BENZYL}-3-(PROPOXYIMINO)BUTANOIC ACID",
            "non-polymer",
        ),
        "XPR": (
            "(2S,3AS,6R,7AS)-6-(SULFOOXY)OCTAHYDRO-1H-INDOLE-2-CARBOXYLIC ACID",
            "peptide",
        ),
        "YOL": (
            "[[2,2'-[4-CARBOXY-1,2-PHENYLENEBIS(NITRILOMETHYLIDYNE)]BIS[PHENOLATO]](2-)-N,N',O,O']-IRON",
            "non-polymer",
        ),
        "LMU": ("DODECYL-ALPHA-D-MALTOSIDE", "saccharide"),
        "6PE": ("1,2-DIHEXANOYL-SN-GLYCERO-3-PHOSPHOETHANOLAMINE", "non-polymer"),
        "RW1": ("4-PHENYLPYRIMIDINE", "non-polymer"),
        "6PG": ("6-PHOSPHOGLUCONIC ACID", "saccharide"),
        "RW3": (
            "N-[6-(4-AMINO-1-{[2-(4-METHYLPIPERAZIN-1-YL)QUINOLIN-3-YL]METHYL}-1H-PYRAZOLO[3,4-D]PYRIMIDIN-3-YL)-1,3-BENZOTHIAZOL-2-YL]ACETAMIDE",
            "non-polymer",
        ),
        "RW4": (
            "3-(2-AMINO-1,3-BENZOTHIAZOL-6-YL)-1-{[2-(4-METHYLPIPERAZIN-1-YL)QUINOLIN-3-YL]METHYL}-1H-PYRAZOLO[3,4-D]PYRIMIDIN-4-AMINE",
            "non-polymer",
        ),
        "O19": (
            "(2R)-N-[4-[2-[(4-MORPHOLIN-4-YLPHENYL)AMINO]PYRIMIDIN-4-YL]PHENYL]PYRROLIDINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "6PC": ("PYRIDINE-2-CARBOXYLIC ACID", "non-polymer"),
        "6PB": ("6-(3-BROMOPHENYL)-7H-PURIN-2-AMINE", "non-polymer"),
        "6PL": (
            "(4S,7R)-4-HYDROXY-N,N,N-TRIMETHYL-9-OXO-7-[(PALMITOYLOXY)METHYL]-3,5,8-TRIOXA-4-PHOSPHAHEXACOSAN-1-AMINIUM 4-OXIDE",
            "non-polymer",
        ),
        "6PO": (
            "9-(2-DEOXY-5-O-PHOSPHONO-BETA-D-ERYTHRO-PENTOFURANOSYL)-6-PROPOXY-9H-PURIN-2-AMINE",
            "nucleic-acid",
        ),
        "6PH": (
            "(1R)-2-(PHOSPHONOOXY)-1-[(TRIDECANOYLOXY)METHYL]ETHYL PENTADECANOATE",
            "non-polymer",
        ),
        "UVX": (
            "[(1R,3R,4R,7S)-3-(2,4-DIOXO-3,4-DIHYDROPYRIMIDIN-1(2H)-YL)-7-HYDROXY-5-METHYLIDENE-2-OXABICYCLO[2.2.1]HEPT-1-YL]METHYL DIHYDROGEN PHOSPHATE",
            "nucleic-acid",
        ),
        "YOK": (
            "[[2,2'-[4-CARBOXYETHYL-1,2-PHENYLENEBIS(NITRILOMETHYLIDYNE)]BIS[PHENOLATO]](2-)-N,N',O,O']-IRON",
            "non-polymer",
        ),
        "UVW": ("ACETYLPHOSPHATE", "non-polymer"),
        "300": ("2-FORMYL-6-METHOXYPHENYL DIHYDROGEN PHOSPHATE", "non-polymer"),
        "SW4": (
            "2-[(1R,2S)-2-[5-[[(2S)-AZETIDIN-2-YL]METHOXY]PYRIDIN-3-YL]CYCLOPROPYL]ETHANOL",
            "non-polymer",
        ),
        "301": ("N6-(PENTA-2,3-DIENYL)ADENINE", "non-polymer"),
        "UVC": ("URIDINE-2',3'-VANADATE", "non-polymer"),
        "302": (
            "4-[(1E,7E)-8-(2,6-DIOXO-1,2,3,6-TETRAHYDROPYRIMIDIN-4-YL)-3,6-DIOXA-2,7-DIAZAOCTA-1,7-DIEN-1-YL]BENZOIC ACID",
            "non-polymer",
        ),
        "303": (
            "6-[(Z)-AMINO(IMINO)METHYL]-N-(1-ISOPROPYL-1,2,3,4-TETRAHYDROISOQUINOLIN-7-YL)-2-NAPHTHAMIDE",
            "non-polymer",
        ),
        "LMN": ("LAURYL MALTOSE NEOPENTYL GLYCOL", "non-polymer"),
        "304": (
            "N-[2-METHYL-5-(METHYLCARBAMOYL)PHENYL]-2-{[(1R)-1-METHYLPROPYL]AMINO}-1,3-THIAZOLE-5-CARBOXAMIDE",
            "non-polymer",
        ),
        "7Z5": (
            "5-[METHYL(PYRIDIN-3-YLMETHYL)AMINO]-2-(PROPANOYLAMINO)BENZOIC ACID",
            "non-polymer",
        ),
        "7Z3": ("1-(3-PYRROLIDIN-1-YLPHENYL)METHANAMINE", "non-polymer"),
        "UPC": (
            "2'-DEOXY-5'-O-[(R)-HYDROXY{[(R)-HYDROXY(PHOSPHONOOXY)PHOSPHORYL]METHYL}PHOSPHORYL]URIDINE",
            "non-polymer",
        ),
        "306": (
            "3-[5-(PIPERIDIN-1-YLMETHYL)-1H-INDOL-2-YL]-6-(1H-PYRAZOL-4-YL)QUINOLIN-2(1H)-ONE",
            "non-polymer",
        ),
        "RLT": (
            "N-(4-FLUOROBENZYL)-5-HYDROXY-1-METHYL-2-(1-METHYL-1-{[(5-METHYL-1,3,4-OXADIAZOL-2-YL)CARBONYL]AMINO}ETHYL)-6-OXO-1,6-DIHYDROPYRIMIDINE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "307": (
            "6,7-DIMETHYL-3-[(METHYL{2-[METHYL({1-[3-(TRIFLUOROMETHYL)PHENYL]-1H-INDOL-3-YL}METHYL)AMINO]ETHYL}AMINO)METHYL]-4H-CHROMEN-4-ONE",
            "non-polymer",
        ),
        "LMB": ("LEPTOMYCIN B, BOUND FORM", "non-polymer"),
        "MG": ("MAGNESIUM ION", "non-polymer"),
        "MA": ("METHYL L-ALANINATE", "peptide"),
        "6P3": ("6-PHENYLPYRIDINE-3-CARBOXYLIC ACID", "non-polymer"),
        "G37": (
            "2-(BENZYLCARBAMOYL-PHENYLACETYLAMINO-METHYL)-5,5-DIMETHYL-THIAZOLIDINE-4-CARBOXYLIC ACID 3-[(1H-BENZIMIDAZOL-2-YLMETHYLCARBAMOYL)-1-BENZYL-2-HYDROXYPROPYL]-AMIDE",
            "non-polymer",
        ),
        "MO": ("MOLYBDENUM ATOM", "non-polymer"),
        "MN": ("MANGANESE (II) ION", "non-polymer"),
        "RLP": ("3-(7-HYDROXY-8-RIBITYLLUMAZINE-6-YL) PROPIONIC ACID", "non-polymer"),
        "LMF": ("NZ(N-METHYL THIOFORMAMIDE) LYSINE", "peptide"),
        "RWY": ("2-(4-PHENOXYPHENOXY)ETHYL THIOCYANATE", "non-polymer"),
        "RWZ": (
            "N-[(2Z)-3,7-DIMETHYLOCTA-2,6-DIEN-1-YL]-N'-[(1R,3S,5R,7R)-TRICYCLO[3.3.1.1~3,7~]DEC-2-YL]ETHANE-1,2-DIAMINE",
            "non-polymer",
        ),
        "LMG": ("1,2-DISTEAROYL-MONOGALACTOSYL-DIGLYCERIDE", "non-polymer"),
        "DOH": ("BETA-HYDROXY ASPARTIC ACID", "peptide"),
        "DOI": (
            "9-(2-DEOXY-5-O-PHOSPHONO-BETA-D-ERYTHRO-PENTOFURANOSYL)-6-(PHOSPHONOOXY)-9H-PURINE",
            "non-polymer",
        ),
        "DON": ("6-DIAZENYL-5-OXO-L-NORLEUCINE", "peptide"),
        "UPX": (
            "N-{4-[(1E)-N-1H-IMIDAZOL-2-YLETHANEHYDRAZONOYL]PHENYL}-7-NITRO-1H-INDOLE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "DOL": (
            "5-(2-DIETHYLAMINO-ETHANESULFONYL)-21-HYDROXY-10-ISOPROPYL-11,19-DIMETHYL-9,26-DIOXA-3,15,28-TRIAZA-TRICYCLO[23.2.1.00,255]OCTACOSA-1(27),12,17,19,25(28)-PENTAENE-2,8,14,23-TETRAONE",
            "non-polymer",
        ),
        "DOM": ("2'-DEOXYMALTOSE", "saccharide"),
        "DOB": ("2,4-DIHYDROXYBENZOIC ACID", "non-polymer"),
        "DOC": ("2',3'-DIDEOXYCYTIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "DOA": ("12-AMINO-DODECANOIC ACID", "peptide"),
        "DOF": (
            "(S)-2-(4-NITROBENZYL)-1,4,7,10-TETRAAZACYCLODODECANE-N,N',N'',N'''-TETRAACETATE",
            "non-polymer",
        ),
        "DOG": ("DIGOXIGENIN", "non-polymer"),
        "DOD": ("DEUTERATED WATER", "non-polymer"),
        "DOE": (
            "(S)-2-(4-(2-(2-HYDROXYETHYLTHIO)-ACETAMIDO)-BENZYL)-1,4,7,10-TETRAAZACYCLODODECANE-N,N',N'',N'''-TETRAACETATE",
            "non-polymer",
        ),
        "DOZ": ("(DIMETHYLAMINO)(HYDROXY)ZINC'", "non-polymer"),
        "DOX": ("DIOXANE", "non-polymer"),
        "DOY": (
            "[(2S)-4-OXO-2-PHENYL-3,4-DIHYDRO-1,5-BENZOTHIAZEPIN-5(2H)-YL]ACETIC ACID",
            "non-polymer",
        ),
        "PU8": (
            "9-BUTYL-8-(2-CHLORO-3,4,5-TRIMETHOXY-BENZYL)-9H-PURIN-6-YLAMINE",
            "non-polymer",
        ),
        "PU9": (
            "8-(2-CHLORO-3,4,5-TRIMETHOXY-BENZYL)-9-PENT-4-YLNYL-9H-PURIN-6-YLAMINE",
            "non-polymer",
        ),
        "DOR": ("(4S)-2,6-DIOXOHEXAHYDROPYRIMIDINE-4-CARBOXYLIC ACID", "non-polymer"),
        "DOS": ("?", "non-polymer"),
        "DOP": ("DIOCTYLPHOSPHATE", "non-polymer"),
        "DOQ": (
            "(2S,3S,4S)-2-CARBOXY-4-[(1Z,3E,5R)-5-CARBOXY-1-METHYL-1,3-HEXADIENYL]-3-PYRROLIDINEACETIC ACID",
            "non-polymer",
        ),
        "PU0": (
            "9-BUTYL-8-(2,5-DIMETHOXY-BENZYL)-2-FLUORO-9H-PURIN-6-YLAMINE",
            "non-polymer",
        ),
        "PU1": (
            "8-(2-CHLORO-3,4,5-TRIMETHOXY-BENZYL)-2-FLUORO-9-PENT-4-YLNYL-9H-PURIN-6-YLAMINE",
            "non-polymer",
        ),
        "DOT": ("3'ANTHRANILOYL-2'-DEOXY-ADENOSINE-5'-TRIPHOSPHATE", "non-polymer"),
        "PU3": ("9-BUTYL-8-(3,4,5-TRIMETHOXYBENZYL)-9H-PURIN-6-AMINE", "non-polymer"),
        "J7Z": (
            "(2S)-3-[(6-O-ALPHA-D-GALACTOPYRANOSYL-BETA-D-GALACTOPYRANOSYL)OXY]-2-[(3Z,6Z,9Z,12Z,15Z)-OCTADECA-3,6,9,12,15-PENTAENOYLOXY]PROPYL (5Z,8Z,11Z,14Z,17Z)-ICOSA-5,8,11,14,17-PENTAENOATE",
            "non-polymer",
        ),
        "G39": (
            "(3R,4R,5S)-4-(ACETYLAMINO)-5-AMINO-3-(PENTAN-3-YLOXY)CYCLOHEX-1-ENE-1-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "XT2": ("DINOPHYSISTOXIN-2", "non-polymer"),
        "XT1": ("DINOPHYSISTOXIN-1", "non-polymer"),
        "NBQ": (
            "2-HYDROXY-5-({1-[(2-NAPHTHYLOXY)METHYL]-3-OXOPROP-1-ENYL}AMINO)TYROSINE",
            "peptide",
        ),
        "J7V": (
            "3'-(7-DIETHYLAMINOCOUMARIN-3-CARBONYLAMINO)-3'-DEOXY-ADP",
            "non-polymer",
        ),
        "Y5H": (
            "3-[(3AS,4R,5S,8AS,8BR)-4-[2-(5-CHLOROTHIOPHEN-2-YL)-1,3-OXAZOL-4-YL]-1,3-DIOXO-4,6,7,8,8A,8B-HEXAHYDRO-3AH-PYRROLO[3,4-A]PYRROLIZIN-2-YL]PROPYL-TRIMETHYL-AZANIUM",
            "non-polymer",
        ),
        "T6F": (
            "2-C-{3-[(4-AMINO-2-METHYLPYRIMIDIN-5-YL)METHYL]-5-(2-{[(R)-HYDROXY(PHOSPHONOOXY)PHOSPHORYL]OXY}ETHYL)-4-METHYL-1,3-THIAZOL-3-IUM-2-YL}-6-O-PHOSPHONO-D-GLUCITOL",
            "non-polymer",
        ),
        "J78": (
            "(METHYLPYRIDAZINE PIPERIDINE PROPYLOXYPHENYL)ETHYLACETATE",
            "non-polymer",
        ),
        "YOF": ("3-FLUOROTYROSINE", "peptide"),
        "XTS": (
            "9-[(2R,3R,4S,5R)-3,4-DIHYDROXY-5-(HYDROXYMETHYL)OXOLAN-2-YL]-3H-PURINE-2,6-DIONE",
            "nucleic-acid",
        ),
        "XTR": (
            "[(1R,4S,6S)-6-HYDROXY-4-(5-METHYL-2,4-DIOXO-3,4-DIHYDROPYRIMIDIN-1(2H)-YL)CYCLOHEX-2-EN-1-YL]METHYL DIHYDROGEN PHOSPHATE",
            "nucleic-acid",
        ),
        "R48": (
            "1-{3-[(6-AMINOQUINAZOLIN-4-YL)AMINO]PHENYL}-3-[3-TERT-BUTYL-1-(4-METHYLPHENYL)-1H-PYRAZOL-5-YL]UREA",
            "non-polymer",
        ),
        "J71": (
            "4-{[(6-METHOXY-5-NITROPYRIMIDIN-4-YL)AMINO]METHYL}BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "J72": (
            "(3E)-5-FLUORO-1-[(6-FLUORO-4H-1,3-BENZODIOXIN-8-YL)METHYL]-1H-INDOLE-2,3-DIONE 3-OXIME",
            "non-polymer",
        ),
        "J73": (
            "METHANOL TRIAZOL ETHYL PHENYL 1O-ALPHA-D-MANNOPYRANOSIDE",
            "non-polymer",
        ),
        "J74": (
            "4-{2-[(6-CHLORO-5-NITROPYRIMIDIN-4-YL)AMINO]ETHYL}BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "J75": (
            "4-({[6-(BENZYLAMINO)-5-NITROPYRIMIDIN-4-YL]AMINO}METHYL)BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "XTY": (
            "(1R)-1,4-ANHYDRO-2-DEOXY-1-(6-METHYL-2,4-DIOXO-1,2,3,4-TETRAHYDROQUINAZOLIN-8-YL)-5-O-PHOSPHONO-D-ERYTHRO-PENTITOL",
            "nucleic-acid",
        ),
        "J77": (
            "(METHYLPYRIDAZINE PIPERIDINE ETHYLOXYPHENYL)ETHYLACETATE",
            "non-polymer",
        ),
        "RJ6": ("N-(3,5-DIMETHOXYPHENYL)IMIDODICARBONIMIDIC DIAMIDE", "non-polymer"),
        "XTF": (
            "[(1R,4R,5S,6R)-5-FLUORO-6-HYDROXY-4-(5-METHYL-2,4-DIOXO-3,4-DIHYDROPYRIMIDIN-1(2H)-YL)CYCLOHEX-2-EN-1-YL]METHYL DIHYDROGEN PHOSPHATE",
            "nucleic-acid",
        ),
        "XTC": (
            "(4S,4AS,5AR,12AS)-4-(DIMETHYLAMINO)-3,10,12,12A-TETRAHYDROXY-9-NITRO-1,11-DIOXO-1,4,4A,5,5A,6,11,12A-OCTAHYDROTETRACENE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "VDB": (
            "(1R,2S,3R,5Z,7E)-17-{(1R)-1-[(2-ETHYL-2-HYDROXYBUTYL)SULFANYL]ETHYL}-2-(2-HYDROXYETHOXY)-9,10-SECOESTRA-5,7,16-TRIENE-1,3-DIOL",
            "non-polymer",
        ),
        "HLE": (
            "3-FORMYL-2-HYDROXY-5-METHYL-HEXANOIC ACID HYDROXYAMIDE",
            "non-polymer",
        ),
        "XTL": (
            "[(1S,4R,6R)-6-HYDROXY-4-(THYMIN-9-YL)CYCLOHEX-2-EN-1-YL]METHYL DIHYDROGEN PHOSPHATE",
            "nucleic-acid",
        ),
        "XTH": (
            "1-(2,3-DIDEOXY-6-O-PHOSPHONO-BETA-D-ERYTHRO-HEXOPYRANOSYL)-5-METHYLPYRIMIDINE-2,4(1H,3H)-DIONE",
            "nucleic-acid",
        ),
        "PUL": ("PULVOMYCIN", "non-polymer"),
        "JJZ": (
            "4-[2-(ANTHRACEN-9-YLMETHYLIDENE)HYDRAZINO]-N-(3-CHLOROPHENYL)-4-OXOBUTANAMIDE",
            "non-polymer",
        ),
        "PUK": ("N-[(2S)-2-AMINO-3-PHENYLPROPYL]-L-PHENYLALANINE", "peptide"),
        "PUG": (
            "3-(1,3,7-TRIHYDRO-9-D-RIBITYL-2,6,8-PURINETRIONE-7-YL)",
            "non-polymer",
        ),
        "PUA": (
            "ADENYLATE-3'-PHOSPHATE-[[2'-DEOXY-URIDINE-5'-PHOSPHATE]-3'-PHOSPHATE]",
            "non-polymer",
        ),
        "PUB": ("PHYCOUROBILIN", "non-polymer"),
        "PUC": (
            "(5Z)-7-{(1R,4S,5S,6R)-6-[(1E,3S)-3-HYDROXYOCT-1-EN-1-YL]-2-OXABICYCLO[2.2.1]HEPT-5-YL}HEPT-5-ENOIC ACID",
            "non-polymer",
        ),
        "NZA": (
            "5-CHLORO-1-(4-CHLOROBENZYL)-3-(PHENYLTHIO)-1H-INDOLE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "DO8": ("3-DEOXY-D-MANNO-2-OCTULOSONATE-8-PHOSPHATE", "non-polymer"),
        "JJL": ("S-[(S)-HYDROXY(5-METHOXYPYRIDIN-3-YL)METHYL]-L-CYSTEINE", "peptide"),
        "JJK": ("S-[(S)-HYDROXY(PYRIDIN-3-YL)METHYL]-L-CYSTEINE", "peptide"),
        "JJJ": ("S-(PYRIDIN-3-YLCARBONYL)-L-CYSTEINE", "peptide"),
        "PUZ": (
            "8-(2,5-DIMETHOXY-BENZYL)-2-FLUORO-9-PENT-9H-PURIN-6-YLAMINE",
            "non-polymer",
        ),
        "HLD": ("4-OXO-4H-PYRAN-2,6-DICARBOXYLIC ACID", "non-polymer"),
        "DO2": ("5,5-DIHYDROXY-6-OXO-L-NORLEUCINE", "peptide"),
        "DO3": (
            "10-((2R)-2-HYDROXYPROPYL)-1,4,7,10-TETRAAZACYCLODODECANE 1,4,7-TRIACETIC ACID",
            "non-polymer",
        ),
        "PUW": ("N-(PHOSPHONOACETYL)-PUTRESCINE", "non-polymer"),
        "3MU": ("?", "nucleic-acid"),
        "DO4": ("2,5-DIOXOPENTANOIC ACID", "non-polymer"),
        "BAV": (
            "(3S,14R,16S)-16-[(1R)-1-HYDROXY-2-{[3-(1-METHYLETHYL)BENZYL]AMINO}ETHYL]-3,4,14-TRIMETHYL-1,4-DIAZACYCLOHEXADECANE-2,5-DIONE",
            "non-polymer",
        ),
        "NC9": (
            "(3R)-3-({(4-AMINOBENZYL)[(4-AMINOPHENYL)ACETYL]AMINO}METHYL)-5-(HYDROXYMETHYL)-2,3,4,7-TETRAHYDRO-1H-AZEPINIUM",
            "non-polymer",
        ),
        "BAT": (
            "4-(N-HYDROXYAMINO)-2R-ISOBUTYL-2S-(2-THIENYLTHIOMETHYL)SUCCINYL-L-PHENYLALANINE-N-METHYLAMIDE",
            "non-polymer",
        ),
        "BAU": (
            "1-((2-HYDROXYETHOXY)METHYL)-5-BENZYLPYRIMIDINE-2,4(1H,3H)-DIONE",
            "non-polymer",
        ),
        "BAR": (
            "8-HYDROXY-2-OXA-BICYCLO[3.3.1]NON-6-ENE-3,5-DICARBOXYLIC ACID",
            "non-polymer",
        ),
        "BAS": ("N-(1-PHENYL-PROPYL)-FORMAMIDE", "non-polymer"),
        "BAP": ("1,2,3-TRIHYDROXY-1,2,3,4-TETRAHYDROBENZO[A]PYRENE", "non-polymer"),
        "NC1": ("NITROCEFIN ACYL-SERINE", "peptide"),
        "NC2": ("NITROCEFIN ACYL-SERINE", "non-polymer"),
        "NC3": ("N-[(CYCLOHEXYLAMINO)CARBONYL]GLYCINE", "non-polymer"),
        "BAZ": ("BIS(5-AMIDINO-BENZIMIDAZOLYL)METHANE ZINC", "non-polymer"),
        "NC5": (
            "3-[(2R)-2-ETHYLPIPERIDIN-1-YL]-N-[6-({3-[(2S)-2-ETHYLPIPERIDIN-1-YL]PROPANOYL}AMINO)ACRIDIN-3-YL]PROPANAMIDE",
            "non-polymer",
        ),
        "BAX": (
            "4-{4-[({[4-CHLORO-3-(TRIFLUOROMETHYL)PHENYL]AMINO}CARBONYL)AMINO]PHENOXY}-N-METHYLPYRIDINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "BAY": ("?", "non-polymer"),
        "BAF": ("(TERT-BUTYLOXYCARBONYL)-ALANYL-AMINO ETHYL-FORMAMIDE", "non-polymer"),
        "BAG": (
            "N-[(E)-2-AMINO-1-(3-{[AMINO(IMINO)METHYL]AMINO}PROPYL)-2-HYDROXYVINYL]BENZAMIDE",
            "non-polymer",
        ),
        "BAD": (
            "(1S)-2-AMINO-1-PHENYLETHYL 6-O-BETA-L-GLUCOPYRANOSYL-ALPHA-D-MANNOPYRANOSIDE",
            "non-polymer",
        ),
        "BAB": ("BIS(5-AMIDINO-BENZIMIDAZOLYL)METHANE", "non-polymer"),
        "239": (
            "6-[(Z)-AMINO(IMINO)METHYL]-N-[4-(AMINOMETHYL)PHENYL]-2-NAPHTHAMIDE",
            "non-polymer",
        ),
        "BAA": ("(TERT-BUTYLOXYCARBONYL)-ALANYL-ALANYL-AMINE", "non-polymer"),
        "234": (
            "5-{2-FLUORO-5-[3-(3-HYDROXY-2-METHOXYCARBONYL-PHENOXY)-PROPENYL]-PHENYL}-ISOXAZOLE-3-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "A94": (
            "N-[(2-AMINO-6-METHYLPYRIMIDIN-4-YL)METHYL]-3-{[(E)-(2-OXODIHYDROFURAN-3(2H)-YLIDENE)METHYL]AMINO}BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "BAL": ("BETA-ALANINE", "peptide"),
        "237": (
            "(2R,4R)-N~1~-(4-CHLOROPHENYL)-N~2~-[3-FLUORO-2'-(METHYLSULFONYL)BIPHENYL-4-YL]-4-METHOXYPYRROLIDINE-1,2-DICARBOXAMIDE",
            "non-polymer",
        ),
        "230": (
            "(2R,4R)-N~1~-(4-CHLOROPHENYL)-N~2~-[2-FLUORO-4-(2-OXOPYRIDIN-1(2H)-YL)PHENYL]-4-METHOXYPYRROLIDINE-1,2-DICARBOXAMIDE",
            "non-polymer",
        ),
        "BAK": ("BIS(5-AMIDINO-2-BENZIMIDAZOLYL)METHANE KETONE", "non-polymer"),
        "232": (
            "1,3-DI[[[10-METHOXY-7H-PYRIDO[4,3-C]CARBAZOL-2-IUMYL]-ETHYL]-PIPERIDIN-4-YL]-PROPANE",
            "non-polymer",
        ),
        "233": (
            "[5-(3,4,5-TRIHYDROXY-6-HYDROXYMETHYL-TETRAHYDRO-PYRAN-2-YLCARBAMOYL)-PENTYL]-CARBAMIC ACID METHYL ESTER",
            "non-polymer",
        ),
        "DPL": ("4-OXOPROLINE", "peptide"),
        "DPK": ("DEPRENYL", "non-polymer"),
        "IVV": (
            "N-(3-METHYLBUTANOYL)-L-VALYL-N-{(1R)-1-[(R)-(2-ETHOXY-2-OXOETHYL)(HYDROXY)PHOSPHORYL]-3-METHYLBUTYL}-L-VALINAMIDE",
            "peptide",
        ),
        "DPJ": ("O,O-DIETHYL HYDROGEN THIOPHOSPHATE", "non-polymer"),
        "ST4": ("4-(ACETYLAMINO)-3-GUANIDINOBENZOIC ACID", "non-polymer"),
        "ST5": ("4-(ACETYLAMINO)-3-[(HYDROXYACETYL)AMINO]BENZOIC ACID", "non-polymer"),
        "ST6": ("4-(ACETYLAMINO)-3-[(AMINOACETYL)AMINO]BENZOIC ACID", "non-polymer"),
        "ST1": ("4-(ACETYLAMINO)-3-HYDROXY-5-NITROBENZOIC ACID", "non-polymer"),
        "ST2": ("4-(ACETYLAMINO)-5-AMINO-3-HYDROXYBENZOIC ACID", "non-polymer"),
        "H": ("ETNAM H HYDROGEN ATOM", "non-polymer"),
        "S8Z": (
            "(8R)-8-[4-(DIFLUOROMETHOXY)PHENYL]-3,3-DIFLUORO-8-[3-(3-METHOXYPROP-1-YN-1-YL)PHENYL]-2,3,4,8-TETRAHYDROIMIDAZO[1,5-A]PYRIMIDIN-6-AMINE",
            "non-polymer",
        ),
        "DPH": ("DEAMINO-METHYL-PHENYLALANINE", "non-polymer"),
        "ST8": (
            "4-{[4-AMINO-6-(CYCLOHEXYLMETHOXY)-5-NITROSOPYRIMIDIN-2-YL]AMINO}BENZAMIDE",
            "non-polymer",
        ),
        "ST9": ("STEAROYL-COENZYME A", "non-polymer"),
        "STD": ("STREPTOLYDIGIN", "non-polymer"),
        "STE": ("STEARIC ACID", "non-polymer"),
        "STF": (
            "(5R)-2-SULFANYL-5-[4-(TRIFLUOROMETHYL)BENZYL]-1,3-THIAZOL-4-ONE",
            "non-polymer",
        ),
        "STG": ("ESTRIOL 3-(B-D-GLUCURONIDE)", "non-polymer"),
        "STA": ("STATINE", "peptide"),
        "STB": ("4-SULFONAMIDE-[4-(THIOMETHYLAMINOBUTANE)]BENZAMIDE", "non-polymer"),
        "STC": (
            "3-[(4-CHLOROANILINO)SULFONYL]THIOPHENE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "STL": ("RESVERATROL", "non-polymer"),
        "KNB": ("3-(METHYLSULFONYL)-L-ALANINE", "peptide"),
        "STN": (
            "2-{[FORMYL(HYDROXY)AMINO]METHYL}-4-METHYLPENTANOIC ACID",
            "non-polymer",
        ),
        "STO": ("STAUROSPORINE", "non-polymer"),
        "STH": ("2-AMINO-4H-1,3-BENZOXATHIIN-4-OL", "non-polymer"),
        "STI": (
            "4-(4-METHYL-PIPERAZIN-1-YLMETHYL)-N-[4-METHYL-3-(4-PYRIDIN-3-YL-PYRIMIDIN-2-YLAMINO)-PHENYL]-BENZAMIDE",
            "non-polymer",
        ),
        "STJ": (
            "3-(6-{[4-(TRIFLUOROMETHOXY)PHENYL]AMINO}PYRIMIDIN-4-YL)BENZAMIDE",
            "non-polymer",
        ),
        "IP8": ("ISOPENTENYL PHOSPHATE", "non-polymer"),
        "STT": (
            "2-(5-METHYL-4-OXO-2-THIOXO-2,4-DIHYDRO-3H-1LAMBDA~4~,3-THIAZOL-3-YL)ETHANESULFONIC ACID",
            "non-polymer",
        ),
        "STU": ("STAUROSPORINE", "non-polymer"),
        "STP": ("?", "non-polymer"),
        "PDH": ("(2S)-2-AMINO-3-PHENYL-1-PROPANOL", "non-polymer"),
        "STR": ("PROGESTERONE", "non-polymer"),
        "STS": (
            "2-[(1S)-1-BENZYL-2-SULFANYLETHYL]-1H-IMIDAZO[4,5-C]PYRIDIN-5-IUM",
            "non-polymer",
        ),
        "STX": (
            "1-(5-TERT-BUTYL-1,3,4-OXADIAZOL-2-YL)-2-(METHYLAMINO)ETHANONE",
            "non-polymer",
        ),
        "9NV": ("DANSYL-L-NORVALINE", "peptide"),
        "STZ": ("STREPTOZOTOCIN", "non-polymer"),
        "NCX": (
            "1-(2-DEOXY-5-O-PHOSPHONO-BETA-D-ERYTHRO-PENTOFURANOSYL)-5-NITRO-1H-INDOLE-3-CARBOXAMIDE",
            "nucleic-acid",
        ),
        "23M": ("2-(3-METHYLPHENYL)-1H-INDOLE-5-CARBOXIMIDAMIDE", "non-polymer"),
        "A9O": ("5-ETHYL-3-METHYL-4-PHENYL-1,2-OXAZOLE", "non-polymer"),
        "A9N": (
            "N-CYCLOPROP-2-EN-1-YL-5-(3,5-DIMETHYL-1,2-OXAZOL-4-YL)-2-METHYL-BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "BA2": (
            "5'-DEOXY-5'-(ETHYLAMINO)-8-{[4-({5-[(3AS,4S,6AR)-2-OXOHEXAHYDRO-1H-THIENO[3,4-D]IMIDAZOL-4-YL]PENTANOYL}AMINO)BUTYL]AMINO}ADENOSINE",
            "non-polymer",
        ),
        "23I": (
            "N-[(1S,2S,4R)-2-HYDROXY-1-ISOBUTYL-5-({(1S)-1-[(ISOPROPYLAMINO)CARBONYL]-2-METHYLPROPYL}AMINO)-4-METHYL-5-OXOPENTYL]-5-[METHYL(METHYLSULFONYL)AMINO]-N'-[(1R)-1-PHENYLETHYL]ISOPHTHALAMIDE",
            "non-polymer",
        ),
        "TFK": (
            "3-[[(METHYLAMINO)SULFONYL]AMINO]-2-OXO-6-PHENYL-N-[3,3,3-TRIFLUORO-1-(1-METHYLETHYL)-2-OXOPHENYL]-1(2H)-PYRIDINE ACETAMIDE",
            "non-polymer",
        ),
        "BA1": ("BALANOL", "non-polymer"),
        "23D": (
            "N2-[(1R,2S)-2-AMINOCYCLOHEXYL]-N6-(3-CHLOROPHENYL)-9-ETHYL-9H-PURINE-2,6-DIAMINE",
            "non-polymer",
        ),
        "23E": (
            "(2E)-3-(4-{[(1-{[(13-CYCLOHEXYL-6-OXO-6,7-DIHYDRO-5H-INDOLO[1,2-D][1,4]BENZODIAZEPIN-10-YL)CARBONYL]AMINO}CYCLOPENTYL)CARBONYL]AMINO}PHENYL)PROP-2-ENOIC ACID",
            "non-polymer",
        ),
        "23F": ("(2Z)-2-AMINO-3-PHENYLACRYLIC ACID", "peptide"),
        "23G": ("?", "nucleic-acid"),
        "NCT": ("(S)-3-(1-METHYLPYRROLIDIN-2-YL)PYRIDINE", "non-polymer"),
        "23A": ("2,3-DIHYDROXYBENZALDEHYDE", "non-polymer"),
        "NCV": ("N-METHYL-6-(1,3-DIHYDROXY-ISOBUTYL)THYMINE", "non-polymer"),
        "23C": (
            "2-[(1R)-1-CARBOXY-2-NAPHTHALEN-1-YLETHYL]-1,3-DIOXO-2,3-DIHYDRO-1H-ISOINDOLE-5-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "NCH": ("P-NITROPHENYL-PHOSPHOCHOLINE", "non-polymer"),
        "NCI": (
            "3,6-BIS{3-[(2R)-(2-ETHYLPIPERIDINO)]PROPIONAMIDO}ACRIDINE",
            "non-polymer",
        ),
        "NCJ": (
            "3,6-BIS{3-[(2R)-2-METHYLPIPERIDINO)]PROPIONAMIDO}ACRIDINE",
            "non-polymer",
        ),
        "NCK": ("3,6-BIS[3-(4-METHYLPIPERIDINO)PROPIONAMIDO]ACRIDINE", "non-polymer"),
        "23X": (
            "N~2~-ACETYL-N-[(2S,3R)-4-{(1,3-BENZOTHIAZOL-6-YLSULFONYL)[(2S)-2-METHYLBUTYL]AMINO}-3-HYDROXY-1-PHENYLBUTAN-2-YL]-L-ISOLEUCINAMIDE",
            "non-polymer",
        ),
        "9NR": ("DANSYL-L-ARGININE", "peptide"),
        "23Z": (
            "(2R)-1-{5-ACETYL-3-[3-{[2-OXO-2-(1H-PYRROL-1-YL)ETHYL]SULFANYL}-4-(TRIFLUOROMETHYL)PHENYL]-4,5,6,7-TETRAHYDRO-1H-PYRAZOLO[4,3-C]PYRIDIN-1-YL}-3-[4-(1-HYDROXY-1-METHYLETHYL)PIPERIDIN-1-YL]PROPAN-2-OL",
            "non-polymer",
        ),
        "NCO": ("COBALT HEXAMMINE(III)", "non-polymer"),
        "A9U": (
            "2-{4-[(1-BENZYLPIPERIDIN-4-YL)METHOXY]PHENYL}-1H-BENZIMIDAZOLE-6-CARBOXAMIDE",
            "non-polymer",
        ),
        "A9T": (
            "2-(4-{[1-(4-CHLOROBENZYL)PIPERIDIN-4-YL]METHOXY}PHENYL)-1H-BENZIMIDAZOLE-5-CARBOXAMIDE",
            "non-polymer",
        ),
        "23V": (
            "N-({(2E)-2-[(3-CHLOROPHENYL)METHYLIDENE]HYDRAZINO}CARBONOTHIOYL)-BETA-D-GLUCOPYRANOSYLAMINE",
            "non-polymer",
        ),
        "NCC": ("?", "non-polymer"),
        "NCD": ("N-CARBAMOYL-L-ASPARTATE", "non-polymer"),
        "A9P": ("3,5 DIMETHYL-4-PHENYL-1,2-OXAZOLE", "non-polymer"),
        "23R": (
            "(1S,2S,3R,5Z,7E,14BETA,17ALPHA,23R)-23-(2-HYDROXY-2-METHYLPROPYL)-2-METHYL-20,24-EPOXY-9,10-SECOCHOLA-5,7,10-TRIENE-1,3-DIOL",
            "non-polymer",
        ),
        "23S": (
            "(S)-2-AMINO-3-(6H-SELENOLO[2,3-B]-PYRROL-4-YL)-PROPIONIC ACID",
            "peptide",
        ),
        "098": (
            "(3R)-N~2~-(CYCLOPROPYLMETHYL)-N~1~-HYDROXY-3-(3-HYDROXYBENZYL)-N~4~-[(1S,2R)-2-HYDROXY-2,3-DIHYDRO-1H-INDEN-1-YL]-L-ASPARTAMIDE",
            "non-polymer",
        ),
        "099": (
            "(2R)-N~4~-HYDROXY-2-(3-HYDROXYBENZYL)-N~1~-[(1S,2R)-2-HYDROXY-2,3-DIHYDRO-1H-INDEN-1-YL]BUTANEDIAMIDE",
            "non-polymer",
        ),
        "FNU": ("?", "nucleic-acid"),
        "LIG": ("3-PYRIDIN-4-YL-2,4-DIHYDRO-INDENO[1,2-.C.]PYRAZOLE", "non-polymer"),
        "FNS": ("N-SULFO-FLAVIN MONONUCLEOTIDE", "non-polymer"),
        "FNR": (
            "1-DEOXY-1-(7,8-DIMETHYL-2,4-DIOXO-3,4-DIHYDRO-2H-BENZO[G]PTERIDIN-1-ID-10(5H)-YL)-5-O-PHOSPHONATO-D-RIBITOL",
            "non-polymer",
        ),
        "LIB": (
            "3-(2-CHLOROPHENYL)-1-(2-{[(1S)-2-HYDROXY-1,2-DIMETHYLPROPYL]AMINO}PYRIMIDIN-4-YL)-1-(4-METHOXYPHENYL)UREA",
            "non-polymer",
        ),
        "FNP": (
            "{[7-(DIFLUORO-PHOSPHONO-METHYL)-NAPHTHALEN-2-YL]-DIFLUORO-METHYL}-PHOSPHONIC ACID",
            "non-polymer",
        ),
        "090": (
            "N-(2,3-DIHYDRO-7,8-DIMETHOXYIMIDAZO[1,2-C] QUINAZOLIN-5-YL)NICOTINAMIDE",
            "non-polymer",
        ),
        "091": ("(3S)-SPIRO[INDOLE-3,3'-PYRROLIDIN]-2(1H)-ONE", "non-polymer"),
        "092": ("?", "non-polymer"),
        "093": (
            "N-(5-(4-CHLORO-3-(2-HYDROXY-ETHYLSULFAMOYL)- PHENYLTHIAZOLE-2-YL)-ACETAMIDE",
            "non-polymer",
        ),
        "094": (
            "1-(2,6-DICHLOROPHENYL)-6-[(2,4-DIFLUOROPHENYL)SULFANYL]-7-(1,2,3,6-TETRAHYDRO-4-PYRIDINYL)-3,4-DIHYDROPYRIDO[3,2-D]PYRIMIDIN-2(1H)-ONE",
            "non-polymer",
        ),
        "095": ("?", "non-polymer"),
        "096": (
            "(2R)-2-{4-HYDROXY-5-METHOXY-2-[3-(4-METHYLPIPERAZIN-1-YL)PROPYL]PHENYL}-3-PYRIDIN-3-YL-1,3-THIAZOLIDIN-4-ONE",
            "non-polymer",
        ),
        "097": (
            "(2S,3R)-N~4~-[(1S)-2,2-DIMETHYL-1-(METHYLCARBAMOYL)PROPYL]-N~1~,2-DIHYDROXY-3-(2-METHYLPROPYL)BUTANEDIAMIDE",
            "non-polymer",
        ),
        "FNG": ("5-AMINOCARBONYL-3-NITROPHENYL-ALPHA-D-GALACTOPYRANOSE", "non-polymer"),
        "LIU": (
            "N-[(4-{[1,1-DIMETHYL-2-(PHENYLTHIO)ETHYL]AMINO}-3-NITROPHENYL)SULFONYL]-4-(4,4-DIMETHYLPIPERIDIN-1-YL)BENZAMIDE",
            "non-polymer",
        ),
        "FNE": (
            "(MU-SULPHIDO)-BIS(MU-CYS,S)-[TRICARBONYLIRON-DI-(CYS,S)NICKEL(II)](FE-NI)",
            "non-polymer",
        ),
        "HFZ": (
            "6-(TRIFLUOROMETHYL)-3,4-DIHYDRO-2H-1,2,4-BENZOTHIADIAZINE-7-SULFONAMIDE 1,1-DIOXIDE",
            "non-polymer",
        ),
        "LIP": ("L-MYO-INOSITOL-1-PHOSPHATE", "non-polymer"),
        "LIQ": (
            "N-{(1S)-2-({(1S,2R)-1-(3,5-DIFLUOROBENZYL)-3-[(3-ETHYLBENZYL)AMINO]-2-HYDROXYPROPYL}AMINO)-2-OXO-1-[(PENTYLSULFONYL)METHYL]ETHYL}NICOTINAMIDE",
            "non-polymer",
        ),
        "FNA": ("N-PHENYLNAPHTHALEN-1-AMINE", "non-polymer"),
        "LIS": ("(2S,4S)-2-AMINO-4,5-EPOXIPENTANOIC ACID", "non-polymer"),
        "FNN": ("3-FLUORO-4-NITROPHENOL", "non-polymer"),
        "FNM": (
            "(5S)-5-METHYL-2-(METHYLSULFANYL)-5-PHENYL-3-(PHENYLAMINO)-3,5-DIHYDRO-4H-IMIDAZOL-4-ONE",
            "non-polymer",
        ),
        "FNK": ("?", "non-polymer"),
        "LIY": (
            "6-ETHYL-5-[(2S)-1-(3-METHOXYPROPYL)-2-PHENYL-1,2,3,4-TETRAHYDROQUINOLIN-7-YL]PYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "FNI": (
            "3-[(5-HYDROXYNAPHTHALEN-2-YL)AMINO]-N-(PYRIMIDIN-4-YL)FURO[2,3-C]PYRIDINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "FNH": (
            "(2R)-({N-[(3S)-3-({[(3S,6S)-6-CYCLOHEXYL-5,8-DIOXO-4,7-DIAZABICYCLO[14.3.1]ICOSA-1(20),16,18-TRIEN-3-YL]CARBONYL}AMINO)-2-OXOHEXANOYL]GLYCYL}AMINO)(PHENYL)ACETIC ACID",
            "non-polymer",
        ),
        "NEH": ("ETHANAMINE", "non-polymer"),
        "GD7": ("(1R)-1,2,2-TRIMETHYLPROPYL (R)-METHYLPHOSPHINATE", "non-polymer"),
        "GD8": ("(1R)-1,2,2-TRIMETHYLPROPYL (S)-METHYLPHOSPHINATE", "non-polymer"),
        "939": (
            "1-METHYL-5-(2-{[5-(4-METHYLPIPERAZIN-1-YL)-2-(TRIFLUOROMETHOXY)PHENYL]AMINO}PYRIMIDIN-4-YL)-1H-PYRROLE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "H52": (
            "(S)-2-METHYL-1-[(4-METHYL-5-ISOQUINOLINE)SULFONYL]-HOMOPIPERAZINE",
            "non-polymer",
        ),
        "GD9": (
            "2-(1H-INDAZOL-4-YL)-6-{[4-(METHYLSULFONYL)PIPERAZIN-1-YL]METHYL}-4-MORPHOLIN-4-YL-THIENO[3,2-D]PYRIMIDINE",
            "non-polymer",
        ),
        "933": (
            "2-CYCLOPROPYL-N-[(2S)-3,3-DIMETHYLBUTAN-2-YL]-5H-PYRROLO[2,3-B]PYRAZINE-7-CARBOXAMIDE",
            "non-polymer",
        ),
        "932": (
            "N-[4-METHYL-3-(1H-1,2,4-TRIAZOL-5-YL)THIOPHEN-2-YL]-2-(2-OXO-3,4-DIHYDROQUINOLIN-1(2H)-YL)ACETAMIDE",
            "non-polymer",
        ),
        "931": (
            "6-CHLORO-N-((3S)-2-OXO-1-{4-[(2R)-2--PYRROLIDINYL] PHENYL}-3-PYRROLIDINYL)-2-NAPHTHALENESULFONAMIDE",
            "non-polymer",
        ),
        "930": (
            "6-CHLORO-N-((3S)-2-OXO-1-{4-[(2S)-2-PYRROLIDINYL]PHENYL}-3-PYRROLIDINYL)-2-NAPHTHALENESULFONAMIDE",
            "non-polymer",
        ),
        "937": (
            "1-(2-HYDROXYETHYL)-8-[[5-(4-METHYLPIPERAZIN-1-YL)-2-(TRIFLUOROMETHOXY)PHENYL]AMINO]-4,5-DIHYDROPYRIMIDO[5,4-G]INDAZOLE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "936": (
            "5-METHOXY-1,2-DIMETHYL-3-(4-NITROPHENOXYMETHYL)INDOLE-4,7-DIONE",
            "non-polymer",
        ),
        "935": (
            "(E)-1-(6-{4-[3-(4-METHYLPIPERAZIN-1-YL)PROPOXY]-3-(TRIFLUOROMETHYL)PHENYL}PYRIDIN-2-YL)METHANIMINE",
            "non-polymer",
        ),
        "NEE": (
            "1-[3-(4-CHLORO-3,5-DIMETHYLPHENOXY)BENZYL]-1H-IMIDAZOLE",
            "non-polymer",
        ),
        "QBT": (
            "[(2R,3S,5R)-3-HYDROXY-5-[(5S)-5-METHYL-2,4-DIOXO-1,3-DIAZINAN-1-YL]OXOLAN-2-YL]METHYL DIHYDROGEN PHOSPHATE",
            "nucleic-acid",
        ),
        "93N": (
            "6-{4-[3-(4-METHYLPIPERAZIN-1-YL)PROPOXY]-3-(TRIFLUOROMETHYL)PHENYL}PYRIDINE-2-CARBONITRILE",
            "non-polymer",
        ),
        "MC": ("1,2-CIS-1-HYDROXY-2,7-DIAMINO-MITOSENE", "non-polymer"),
        "H5M": ("TRANS-3-HYDROXY-5-METHYLPROLINE", "peptide"),
        "93A": (
            "[3,4-DIHYDROXY-5R-(2,2,4-TRIOXO-1,2R,3S,4R-TETRAHYDRO-2L6-IMIDAZO[4,5-C][1,2,6]THIADIAZIN-7-YL)TETRAHYDROFURAN-2-YL]METHYL DIHYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "H5K": (
            "6-{[(5-CYANOPYRAZIN-2-YL)CARBAMOYL]AMINO}NAPHTHALENE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "H5U": (
            "N-({3-[(5S)-5-TERT-BUTYL-1-(4-FLUOROBENZYL)-4-HYDROXY-2-OXO-2,5-DIHYDRO-1H-PYRROL-3-YL]-1,1-DIOXIDO-1,2-BENZISOTHIAZOL-7-YL}METHYL)METHANESULFONAMIDE",
            "non-polymer",
        ),
        "H5W": (
            "6-{[(3R,4R)-4-{[(3E)-4-(3-FLUOROPHENYL)BUT-3-EN-1-YL]OXY}PYRROLIDIN-3-YL]METHYL}-4-METHYLPYRIDIN-2-AMINE",
            "non-polymer",
        ),
        "H5P": ("HYDANTOCIDIN-5'-PHOSPHATE", "non-polymer"),
        "H5S": (
            "(5S)-5-TERT-BUTYL-1-(4-FLUORO-3-METHYLBENZYL)-4-HYDROXY-3-[8-(METHYLSULFONYL)-1,1-DIOXIDO-6,7,8,9-TETRAHYDROISOTHIAZOLO[4,5-H]ISOQUINOLIN-3-YL]-1,5-DIHYDRO-2H-PYRROL-2-ONE",
            "non-polymer",
        ),
        "QBH": (
            "(4S)-4-(2-HYDROXY-5-{[(3S,4S,5R)-4-HYDROXY-1,1-DIOXIDO-5-{[3-(PROPAN-2-YL)BENZYL]AMINO}TETRAHYDRO-2H-THIOPYRAN-3-YL]METHYL}BENZYL)-3-PROPYL-1,3-OXAZOLIDIN-2-ONE",
            "non-polymer",
        ),
        "HFV": (
            "DELTA-(L-ALPHA-AMINOADIPOYL)-L-CYSTEINYL-L-3,3,3,3',3',3'-HEXAFLUOROVALINE",
            "non-polymer",
        ),
        "KAB": ("KABIRAMIDE C", "non-polymer"),
        "09Z": (
            "[4-AMINO-2-(PROPAN-2-YLAMINO)-1,3-THIAZOL-5-YL](PYRIDIN-3-YL)METHANONE",
            "non-polymer",
        ),
        "KAG": ("?", "nucleic-acid"),
        "12K": ("(1-METHYLCYCLOHEXYL)CARBAMIC ACID", "non-polymer"),
        "ME4": (
            "3-CHLORO-N-((1R,2S) -2-(4-(2-OXOPYRIDIN-1(2H)-YL)BENZAMIDO)CYCLOHEXYL)-1H-INDOLE-6-CARBOXAMIDE",
            "non-polymer",
        ),
        "09P": ("1-(5-ETHOXYPYRIDIN-3-YL)-1,4-DIAZEPANE", "non-polymer"),
        "09Q": ("1-(5-PHENYLPYRIDIN-3-YL)-1,4-DIAZEPANE", "non-polymer"),
        "09R": ("1-(6-BROMOPYRIDIN-3-YL)-1,4-DIAZEPANE", "non-polymer"),
        "09S": ("1-(6-BROMO-5-ETHOXYPYRIDIN-3-YL)-1,4-DIAZEPANE", "non-polymer"),
        "09T": ("1-(3,4-DICHLOROBENZYL)-5,6-DIMETHYL-1H-BENZIMIDAZOLE", "non-polymer"),
        "09U": ("(METHOXYMETHYL)CYCLOPROPANE", "non-polymer"),
        "09V": ("CYCLOPROPYLMETHANOL", "non-polymer"),
        "09H": (
            "2-(2,3-DIHYDRO-1-BENZOFURAN-5-YL)-N-[2-(PIPERAZIN-1-YL)PHENYL]-1,3-THIAZOLE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "KAR": ("?", "non-polymer"),
        "09J": (
            "3-[3-(2-CHLORO-6-FLUOROPHENYL)-5-ETHYL-6-OXO-5,6-DIHYDRO[1,2,4]TRIAZOLO[4,3-B]PYRIDAZIN-7-YL]-N-CYCLOPROPYL-4-METHYLBENZAMIDE",
            "non-polymer",
        ),
        "KAP": ("7-KETO-8-AMINOPELARGONIC ACID", "non-polymer"),
        "09L": (
            "4-(3-{[4-(CYCLOPROPYLCARBONYL)PIPERAZIN-1-YL]CARBONYL}-4-FLUOROBENZYL)PHTHALAZIN-1(2H)-ONE",
            "non-polymer",
        ),
        "09M": (
            "(5-HYDROXY-4-{[(1-HYDROXY-2-OXO-1,2-DIHYDROQUINOLIN-3-YL)AMINO]METHYL}-6-METHYLPYRIDIN-3-YL)METHYL DIHYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "09N": (
            "(11Z,14Z)-N-[(2S,3S,4R)-1-(ALPHA-D-GALACTOPYRANOSYLOXY)-3,4-DIHYDROXYOCTADECAN-2-YL]ICOSA-11,14-DIENAMIDE",
            "non-polymer",
        ),
        "09O": ("1-(PYRIDIN-3-YL)-1,4-DIAZEPANE", "non-polymer"),
        "RHA": ("6-DEOXY-MANNOSE-1'-PHOSPHATE", "saccharide"),
        "09A": ("?", "non-polymer"),
        "09B": ("?", "non-polymer"),
        "09C": ("(3S,5R)-3,5-BIS(6-BROMO-1H-INDOL-3-YL)PIPERAZIN-2-ONE", "non-polymer"),
        "09D": ("?", "non-polymer"),
        "09E": ("?", "non-polymer"),
        "09F": ("?", "non-polymer"),
        "09G": ("?", "non-polymer"),
        "CZ8": (
            "({[(2R)-2-{[(4-ETHYL-2,3-DIOXOPIPERAZIN-1-YL)CARBONYL]AMINO}-4-(4-HYDROXYPHENYL)BUTANOYL]AMINO}METHYL)BORONIC ACID",
            "non-polymer",
        ),
        "CZ9": (
            "N-[(DIHYDROXYBORANYL)METHYL]-NALPHA-[(4-ETHYL-2,3-DIOXOPIPERAZIN-1-YL)CARBONYL]-D-TYROSINAMIDE",
            "non-polymer",
        ),
        "P2A": ("1,2-DIETHANOYL-SN-GLYCERO-3-PHOSPHATE", "non-polymer"),
        "IHH": (
            "[4-({4-[(5-CYCLOPROPYL-1H-PYRAZOL-3-YL)AMINO]QUINAZOLIN-2-YL}AMINO)PHENYL]ACETONITRILE",
            "non-polymer",
        ),
        "P2C": (
            "2-[(3,5-DICHLORO-4-TRIOXIDANYLPHENYL)AMINO]BENZOIC ACID",
            "non-polymer",
        ),
        "P2B": (
            "3-[2-MORPHOLIN-4-YL-6-(PYRIDIN-3-YLAMINO)PYRIMIDIN-4-YL]PHENOL",
            "non-polymer",
        ),
        "CZ2": ("S-(DIHYDROXYARSINO)CYSTEINE", "peptide"),
        "CZ3": (
            "1,1'-{HEXA-2,4-DIYNE-1,6-DIYLBIS[OXY{(2S,3R)-2-[(N-METHYL-L-ALANYL)AMINO]-1-OXOBUTANE-3,1-DIYL}(2S)PYRROLIDINE-1,2-DIYLMETHANEDIYL]}BIS[5-(PHENYLSULFANYL)-1H-TETRAZOLE]",
            "non-polymer",
        ),
        "CZ0": ("8-HYDROXY-4-METHYL-9-NITRO-2H-BENZO[G]CHROMEN-2-ONE", "non-polymer"),
        "ME2": ("1-ETHOXY-2-(2-METHOXYETHOXY)ETHANE", "non-polymer"),
        "CZ6": (
            "2-({[(1Z)-1-(2-AMINO-3H-1,3-THIAZOL-1-IUM-4-YL)-2-OXO-2-({[(6S)-4,4,6-TRIMETHYL-1,3,2-DIOXABORINAN-2-YL]METHYL}AMINO)ETHYLIDENE]AMINO}OXY)-2-METHYLPROPANOATE",
            "non-polymer",
        ),
        "CZ7": (
            "({[(2R)-2-{[(4-ETHYL-2,3-DIOXO-3,4-DIHYDROPYRAZIN-1(2H)-YL)CARBONYL]AMINO}-2-(4-HYDROXYPHENYL)ACETYL]AMINO}METHYL)BORONIC ACID",
            "non-polymer",
        ),
        "P2K": ("POLYETHYLENE GLYCOL PEG2000", "non-polymer"),
        "LXZ": ("2-(ACETYLAMINO)-2-DEOXY-ALPHA-D-IDOPYRANOSE", "saccharide"),
        "P2U": ("2'-DEOXY-PSEUDOURIDINE-5'MONOPHOSPHATE", "nucleic-acid"),
        "NEV": ("NEVIRAPINE", "non-polymer"),
        "P2Q": (
            "2-HYDROXY-5-{[(1S,2E)-1-FORMYL-4-PHENOXYBUT-2-EN-1YL]AMINO}-L-TYROSINE",
            "peptide",
        ),
        "P2P": ("N7-(5'-PHOSPHO-ALPHA-RIBOSYL)-2-HYDROXYPURINE", "non-polymer"),
        "P2S": (
            "(2S)-2-AMINO-4-[[(2R)-2-CARBOXYBUTYL](PHOSPHONO)SULFONIMIDOYL]BUTANOIC ACID",
            "non-polymer",
        ),
        "NEW": ("3,3',5,5'-TETRACHLOROBIPHENYL-4,4'-DIOL", "non-polymer"),
        "OMK": (
            "(2S)-2-[2-(4-CHLOROPHENYL)-5,6-DIFLUORO-1H-BENZIMIDAZOL-1-YL]-2-CYCLOHEXYL-N-(2-METHYLPHENYL)ETHANAMIDE",
            "non-polymer",
        ),
        "PCG": ("CYCLIC GUANOSINE MONOPHOSPHATE", "non-polymer"),
        "P2Y": ("(2S)-PYRROLIDIN-2-YLMETHYLAMINE", "peptide"),
        "NET": ("TETRAETHYLAMMONIUM ION", "non-polymer"),
        "797": ("3-(6-HYDROXY-NAPHTHALEN-2-YL)-BENZO[D]ISOOXAZOL-6-OL", "non-polymer"),
        "796": (
            "N-[(1S)-2-AMINO-1-(2,4-DICHLOROBENZYL)ETHYL]-5-[2-(METHYLAMINO)PYRIMIDIN-4-YL]THIOPHENE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "795": ("2-(PYRIDIN-3-YL)-1,3-THIAZOLE-4-CARBALDEHYDE", "non-polymer"),
        "794": ("2-[(CARBOXYCARBONYL)(1-NAPHTHYL)AMINO]BENZOIC ACID", "non-polymer"),
        "793": (
            "3-{[(1R)-1-PHENYLETHYL]AMINO}-4-(PYRIDIN-4-YLAMINO)CYCLOBUT-3-ENE-1,2-DIONE",
            "non-polymer",
        ),
        "792": (
            "N-{[4-(BUT-2-YN-1-YLOXY)PHENYL]SULFONYL}-5-METHYL-D-TRYPTOPHAN",
            "non-polymer",
        ),
        "791": ("2-PHENYLMALONIC ACID", "non-polymer"),
        "790": ("1,2,3,4-TETRAHYDROQUINOLIN-8-YL DIHYDROGEN PHOSPHATE", "non-polymer"),
        "OX4": (
            "3-FLUORO-4-HYDROXYBENZALDEHYDE O-(CYCLOHEXYLCARBONYL)OXIME",
            "non-polymer",
        ),
        "NER": ("(15E)-TETRACOS-15-ENOIC ACID", "non-polymer"),
        "OX6": (
            "TERT-BUTYL [(1R)-2-METHYL-1-(1,3,4-OXADIAZOL-2-YL)PROPYL]CARBAMATE",
            "non-polymer",
        ),
        "OX7": (
            "TERT-BUTYL [(1S)-2-METHYL-1-(1,3,4-OXADIAZOL-2-YL)PROPYL]CARBAMATE",
            "non-polymer",
        ),
        "5TN": (
            "(2E)-3-[4-HYDROXY-3-(5,5,8,8-TETRAMETHYL-3-PROPOXY-5,6,7,8-TETRAHYDRONAPHTHALEN-2-YL)PHENYL]ACRYLIC ACID",
            "non-polymer",
        ),
        "OX1": ("?", "non-polymer"),
        "OX2": ("2-(BETA-D-GLUCOPYRANOSYL)-5-METHYL-1,3,4-OXADIAZOLE", "saccharide"),
        "798": (
            "{4-[(2R)-PYRROLIDIN-2-YLMETHOXY]PHENYL}(4-THIOPHEN-3-YLPHENYL)METHANONE",
            "non-polymer",
        ),
        "5TS": ("5-(1,2-OXAZOL-5-YL)THIOPHENE-2-SULFONIC ACID", "non-polymer"),
        "5TP": ("3-[(2R)-PIPERIDIN-2-YL]-1-(1,3-THIAZOL-2-YL)PROPAN-1-ONE", "peptide"),
        "SYP": (
            "N-[2-({6-[(2-SULFANYLETHYL)AMINO]PYRIMIDIN-4-YL}AMINO)ETHYL]PROPANAMIDE",
            "non-polymer",
        ),
        "ZPH": (
            "METHYL (2S)-2-{[(R)-[(1R)-1-AMINO-3-METHYLBUTYL](HYDROXY)PHOSPHORYL]OXY}-3-PHENYLPROPANOATE",
            "peptide",
        ),
        "NEP": ("N1-PHOSPHONOHISTIDINE", "peptide"),
        "79W": (
            "3-[(2S)-1,1-DIOXIDO-4-OXOTETRAHYDROTHIOPHEN-2-YL]BENZALDEHYDE",
            "non-polymer",
        ),
        "OXX": ("OXALYL-ASPARTYL ANHYDRIDE", "peptide"),
        "OXY": ("OXYGEN MOLECULE", "non-polymer"),
        "OXZ": ("TETRAHYDROOXAZINE", "non-polymer"),
        "OXT": ("?", "non-polymer"),
        "OXP": ("2-OXO-3-PENTENOIC ACID", "non-polymer"),
        "79Z": (
            "5-CYCLOPROPYL-2-(4-FLUOROPHENYL)-6-[(2-HYDROXYETHYL)(METHYLSULFONYL)AMINO]-N-METHYL-1-BENZOFURAN-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "OXR": ("4-OXORETINOL", "non-polymer"),
        "79X": (
            "2-HYDROXY-8,9-DIMETHOXY-6H-ISOINDOLO[2,1-A]INDOL-6-ONE",
            "non-polymer",
        ),
        "OXL": ("OXALATE ION", "non-polymer"),
        "OXM": ("OXAMIC ACID", "non-polymer"),
        "OXN": ("OXTOXYNOL-10", "non-polymer"),
        "OXO": ("OXO GROUP", "non-polymer"),
        "OXI": ("OXOLINIC ACID", "non-polymer"),
        "LXN": ("ALLOXAN", "non-polymer"),
        "79O": (
            "8-BROMO-2-(2-CHLOROPHENYL)[1]BENZOFURO[3,2-D]PYRIMIDIN-4(3H)-ONE",
            "non-polymer",
        ),
        "OXE": ("ORTHO-XYLENE", "non-polymer"),
        "OXG": ("8-OXOGUANINE", "non-polymer"),
        "OXA": ("OXYDIMETHYLENE GROUP", "non-polymer"),
        "OXC": ("OXONIC ACID", "non-polymer"),
        "CZZ": ("THIARSAHYDROXY-CYSTEINE", "peptide"),
        "ZMO": (
            "{[(3S)-3-HYDROXY-2,2-DIMETHYL-4-{[3-({2-[(9Z)-OCTADEC-9-ENOYLSULFANYL]ETHYL}AMINO)-3-OXOPROPYL]AMINO}-4-OXOBUTYL]OXY}(OXO)PHOSPHONIUMOLATE",
            "non-polymer",
        ),
        "CZP": (
            "(8R)-8-(CYCLOPENTYLMETHYL)-2-HYDROPEROXY-2-(4-HYDROXYBENZYL)-6-(4-HYDROXYPHENYL)-7,8-DIHYDROIMIDAZO[1,2-A]PYRAZIN-3(2H)-ONE",
            "non-polymer",
        ),
        "Y10": (
            "(E,2R,3R,4S,5R)-N-(2,3-DIHYDRO-1H-INDEN-2-YL)-2-METHOXY-8,8-DIMETHYL-3,4,5-TRIS(OXIDANYL)NON-6-ENAMIDE",
            "non-polymer",
        ),
        "P25": ("PENTYL TRIHYDROGEN DIPHOSPHATE", "non-polymer"),
        "P24": ("1-(AMINOETHYL)AMINO-4-AMINOBUTANE", "non-polymer"),
        "CZH": ("C2-HYDROPEROXY-COELENTERAZINE", "non-polymer"),
        "CZI": (
            "(2R)-8-BENZYL-2-HYDROPEROXY-6-(4-HYDROXYPHENYL)-2-(4-IODOBENZYL)-7,8-DIHYDROIMIDAZO[1,2-A]PYRAZIN-3(2H)-ONE",
            "non-polymer",
        ),
        "CZN": (
            "(2S,8R)-8-BENZYL-2-HYDROPEROXY-6-(4-HYDROXYPHENYL)-2-(2-NAPHTHYLMETHYL)-7,8-DIHYDROIMIDAZO[1,2-A]PYRAZIN-3(2H)-ONE",
            "non-polymer",
        ),
        "CZO": (
            "[(2R,4S)-2-[(1S,2R)-1-AMINO-2-HYDROXYPROPYL]-2-HYDROXY-4-(4-HYDROXYBENZYL)-5-OXOIMIDAZOLIDIN-1-YL]ACETIC ACID",
            "peptide",
        ),
        "CZL": ("L-CLUSTER (FE8S9)", "non-polymer"),
        "CZM": (
            "'N,N'-BIS-(2-HYDROXY-3-METHYL-BENZYLIDENE)-BENZENE-1,2-DIAMINE'",
            "non-polymer",
        ),
        "CZB": (
            "(2S,8R)-8-BENZYL-2-(4-BROMOBENZYL)-2-HYDROPEROXY-6-(4-HYDROXYPHENYL)-7,8-DIHYDROIMIDAZO[1,2-A]PYRAZIN-3(2H)-ONE",
            "non-polymer",
        ),
        "CZA": (
            "(6AR,11AS,11BR)-10-ACETYL-9-HYDROXY-7,7-DIMETHYL-2,6,6A,7,11A,11B-HEXAHYDRO-11H-PYRROLO[1',2':2,3]ISOINDOLO[4,5,6-CD]INDOL-11-ONE",
            "non-polymer",
        ),
        "P29": (
            "4-(2-(1H-IMIDAZOL-4-YL)ETHYLAMINO)-2-(PHENYLAMINO)PYRAZOLO[1,5-A][1,3,5]TRIAZINE-8-CARBONITRILE",
            "non-polymer",
        ),
        "P28": ("3',5'-DINITRO-N-ACETYL-L-THYRONINE", "non-polymer"),
        "ATF": ("PHOSPHODIFLUOROMETHYLPHOSPHONIC ACID-ADENYLATE ESTER", "non-polymer"),
        "ATG": ("PHOSPHOTHIOPHOSPHORIC ACID-ADENYLATE ESTER", "non-polymer"),
        "ATD": ("THYMIDINE-3'-PHOSPHATE", "nucleic-acid"),
        "ATE": ("16,17-ANDROSTENE-3-OL", "non-polymer"),
        "ATC": (
            "9-(N,N-DIMETHYLGLYCYLAMIDO)-6-DEOXY-6-DEMETHYL-TETRACYCLINE",
            "non-polymer",
        ),
        "2TC": (
            "(4S,4AS,6S,8AS)-6-[(1S)-7-CHLORO-4-HYDROXY-1-METHYL-3-OXO-1,3-DIHYDRO-2-BENZOFURAN-1-YL]-4-(DIMETHYLAMINO)-3,8A-DIHYDROXY-1,8-DIOXO-1,4,4A,5,6,7,8,8A-OCTAHYDRONAPHTHALENE-2-CARBONITRILE",
            "non-polymer",
        ),
        "2TB": ("1,3-DIAMINO-4,5,6-TRIHYDROXY-CYCLOHEXANE", "non-polymer"),
        "ATO": ("CHLOROACETONE", "non-polymer"),
        "ATL": (
            "[(1S,3R,4S,7R)-7-HYDROXY-3-(THYMIN-1-YL)-2,5-DIOXABICYCLO[2.2.1]HEPT-1-YL]METHYL DIHYDROGEN PHOSPHATE",
            "nucleic-acid",
        ),
        "ATM": ("3'-AZIDO-3'-DEOXYTHYMIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "ATJ": ("ETHYL HYDROGEN PHOSPHONATE", "non-polymer"),
        "2TH": (
            "2-CHLORO-N-[(1R,2R)-1-HYDROXY-2,3-DIHYDRO-1H-INDEN-2-YL]-6H-THIENO[2,3-B]PYRROLE-5-CARBOXAMIDE",
            "non-polymer",
        ),
        "ATH": ("4-HYDROXY-ACONITATE ION", "non-polymer"),
        "ATI": (
            "N-(3-AMINO-2-HYDROXY-5-METHYLHEXANOYL)VALYLVALYLASPARTIC ACID",
            "non-polymer",
        ),
        "ATV": (
            "2-[(2R)-OXIRAN-2-YLMETHYL]-1H-BENZO[DE]ISOQUINOLINE-1,3(2H)-DIONE",
            "non-polymer",
        ),
        "ATW": (
            "(1-{6-[6-(PROPAN-2-YLAMINO)-1H-INDAZOL-1-YL]PYRAZIN-2-YL}-1H-PYRROL-3-YL)ACETIC ACID",
            "non-polymer",
        ),
        "ATT": (
            "2-[3-(5-MERCAPTO-[1,3,4]THIADIAZOL-2-YL)-UREIDO]-N-METHYL-3-PHENYL-PROPIONAMIDE",
            "non-polymer",
        ),
        "ATU": (
            "9-NITRO-5,12-DIHYDRO-7H-BENZO[2,3]AZEPINO[4,5-B]INDOL-6-ONE",
            "non-polymer",
        ),
        "ATR": ("2'-MONOPHOSPHOADENOSINE-5'-DIPHOSPHATE", "non-polymer"),
        "2TP": (
            "2-(3-{[4-(HYDROXYAMINO)-2-METHYLPYRIMIDIN-5-YL]METHYL}-4-METHYL-2,3-DIHYDRO-1,3-THIAZOL-5-YL)ETHYL TRIHYDROGEN DIPHOSPHATE",
            "non-polymer",
        ),
        "ATP": ("ADENOSINE-5'-TRIPHOSPHATE", "non-polymer"),
        "ATQ": ("2-AMINOTHIAZOLINE", "non-polymer"),
        "RGT": (
            "2-AMINO-9-{5-O-[(R)-HYDROXY{[(R)-HYDROXY(PHOSPHONOOXY)PHOSPHORYL]OXY}PHOSPHORYL]-BETA-D-RIBOFURANOSYL}-9H-PURIN-6-YL 5-O-PHOSPHONO-BETA-D-RIBOFURANOSIDE",
            "non-polymer",
        ),
        "IHM": (
            "N-(3-{1-[(1S)-1-[3-(CYCLOPROPYLMETHOXY)-4-(DIFLUOROMETHOXY)PHENYL]-2-(1-OXIDOPYRIDIN-4-YL)ETHYL]-1H-PYRAZOL-3-YL}PHENYL)ACETAMIDE",
            "non-polymer",
        ),
        "ATZ": (
            "2-CHLORO-4-ISOPROPYLAMINO-6-ETHYLAMINO -1,3,5-TRIAZINE",
            "non-polymer",
        ),
        "ATX": (
            "(2S)-2-AMINO-1-(5-TERT-BUTYL-1,3,4-OXADIAZOL-2-YL)PROPAN-1-ONE",
            "non-polymer",
        ),
        "ATY": ("3'-O-ACETYLTHYMIDINE-5'-DIPHOSPHATE", "non-polymer"),
        "D98": (
            "2-(4'-CARBAMIMIDOYL-2'-HYDROXYBIPHENYL-4-YL)-1H-BENZIMIDAZOLE-5-CARBOXIMIDAMIDE",
            "non-polymer",
        ),
        "D99": (
            "(3R)-1-{4-[{4-CHLORO-2-[(S)-(2-CHLOROPHENYL)(HYDROXY)METHYL]PHENYL}(2,2-DIMETHYLPROPYL)AMINO]-4-OXOBUTANOYL}PIPERIDINE-3-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "Z3R": ("4-(1,3-BENZOTHIAZOL-2-YL)THIOPHENE-2-SULFONAMIDE", "non-polymer"),
        "D94": (
            "3-[CIS-3-(4-METHYLPIPERAZIN-1-YL)CYCLOBUTYL]-1-(2-PHENYLQUINOLIN-7-YL)IMIDAZO[1,5-A]PYRAZIN-8-AMINE",
            "non-polymer",
        ),
        "Y12": (
            "4-{(1E)-3-OXO-3-[(2-PHENYLETHYL)AMINO]PROP-1-EN-1-YL}-1,2-PHENYLENE DIACETATE",
            "non-polymer",
        ),
        "D90": (
            "N-(2-{[(5-CHLORO-1H-INDOL-2-YL)CARBONYL]AMINO}PHENYL)-5-METHYL-4,5,6,7-TETRAHYDRO[1,3]THIAZOLO[5,4-C]PYRIDINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "D91": (
            "5-CHLORO-2-({3-ISOBUTYL-4-[(5-PYRIDIN-4-YLPYRIMIDIN-2-YL)CARBONYL]PIPERAZIN-1-YL}SULFONYL)-1H-INDOLE",
            "non-polymer",
        ),
        "D92": (
            "N-((1R,2S)-2-(5-CHLORO-1H-INDOLE-2-CARBOXAMIDO)CYCLOHEXYL)-5-METHYL-4,5,6,7-TETRAHYDROTHIAZOLO[5,4-C]PYRIDINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "D93": (
            "N-((1R,2R)-2-(5-CHLORO-1H-INDOLE-2-CARBOXAMIDO)CYCLOHEXYL)-5-METHYL-4,5,6,7-TETRAHYDROTHIAZOLO[5,4-C]PYRIDINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "PCD": (
            "(MOLYBDOPTERIN-CYTOSINE DINUCLEOTIDE-S,S)-DIOXO-AQUA-MOLYBDENUM(V)",
            "non-polymer",
        ),
        "MOI": (
            "(7R,7AS,12BS)-3-METHYL-2,3,4,4A,7,7A-HEXAHYDRO-1H-4,12-METHANO[1]BENZOFURO[3,2-E]ISOQUINOLINE-7,9-DIOL",
            "non-polymer",
        ),
        "RGZ": ("(R)-ROSIGLITAZONE", "non-polymer"),
        "Z34": (
            "N-[2-[5-[AMINO(IMINO)METHYL]-2-HYDROXYPHENOXY]-3,5-DIFLUORO-6-[3-(4,5-DIHYDRO-1-METHYL-1H-IMIDAZOL-2-YL)PHENOXY]PYRIDIN-4-YL]-N-METHYLGLYCINE",
            "non-polymer",
        ),
        "DP7": ("AC-(D)PHE-PRO-BOROARG-OH", "non-polymer"),
        "Z30": (
            "(5R)-5-TERT-BUTYL-4,5,6,7-TETRAHYDRO-1H-INDAZOLE-3-CARBOHYDRAZIDE",
            "non-polymer",
        ),
        "Z31": (
            "[4-AMINO-2-(PROP-2-EN-1-YLAMINO)-1,3-THIAZOL-5-YL](NAPHTHALEN-2-YL)METHANONE",
            "non-polymer",
        ),
        "D9Z": ("5-(2-CHLOROPHENYL)-1,3,4-THIADIAZOLE-2-SULFONAMIDE", "non-polymer"),
        "DP6": (
            "(3R)-3-HYDROXY-5-{[(R)-HYDROXY(PHOSPHONOOXY)PHOSPHORYL]OXY}-3-METHYLPENTANOIC ACID",
            "non-polymer",
        ),
        "MF2": ("CIS-2,6-DIMETHYLMORPHOLINOOCTYLCARBAMYLESEROLINE", "non-polymer"),
        "DP5": (
            "1,4-DIMERCAPTO-ALPHA-D-GLUCOPYRANOSYL-(1-4)-S-1,4-DIMERCAPTO-BETA-D-GLUCOPYRANOSYL-(1-4)-S-1,4-DIMERCAPTO-BETA-D-GLUCOPYRANOSE",
            "saccharide",
        ),
        "DL6": (
            "2-AZIDO-N-((2R,3R,4S,5S,6R)-3,4,5-TRIHYDROXY-6-(HYDROXYMETHYL)-TETRAHYDRO-2H-PYRAN-2-YL)ACETAMIDE",
            "non-polymer",
        ),
        "DP4": (
            "(1S,4S)-4-[DIMETHYL(PHENYL)SILYL]-1-METHYLPIPERIDINE 1-OXIDE",
            "non-polymer",
        ),
        "D9H": (
            "4-({4-CHLORO-6-[(2-HYDROXYETHYL)AMINO]-1,3,5-TRIAZIN-2-YL}AMINO)BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "D9G": ("N-DODECYL-N,N-DIMETHYLGLYCINATE", "non-polymer"),
        "D9A": (
            "N-{[(3R,5S)-7-CHLORO-5-(2,3-DIMETHOXYPHENYL)-1-(2,2-DIMETHYLPROPYL)-2-OXO-1,2,3,5-TETRAHYDRO-4,1-BENZOXAZEPIN-3-YL]ACETYL}-L-ASPARTIC ACID",
            "non-polymer",
        ),
        "3V0": (
            "N-{[(2R)-2,3-DIHYDROXYPROPYL]OXY}-3-[(2-FLUORO-4-IODOPHENYL)AMINO]FURO[3,2-C]PYRIDINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "DP2": (
            "L-LYSYL-N~5~-[(Z)-(2,2-DIHYDROXYHYDRAZINO)(IMINO)METHYL]-L-ORNITHINAMIDE",
            "non-polymer",
        ),
        "DP1": ("L-N(OMEGA)-NITROARGININE-2,4-L-DIAMINOBUTYRIC AMIDE", "non-polymer"),
        "MF3": ("2-AMINO-4-TRIFLUOROMETHYLSULFANYL-BUTYRIC ACID", "peptide"),
        "RGA": ("REIDISPONGIOLIDE A", "non-polymer"),
        "U2G": ("URIDYLYL-2'-5'-PHOSPHO-GUANOSINE", "non-polymer"),
        "K0L": ("(4-AMINO-2,6-DIMETHYLPHENOXY)ACETIC ACID", "peptide"),
        "DP0": ("DISOPYRAMIDE", "non-polymer"),
        "SB3": (
            "1,3-DIPHENYL-1-PROPYL-1-(3,3-DIMETHYL-1,2-DIOXYPENTYL)-2-PIPERIDINE CARBOXYLATE",
            "non-polymer",
        ),
        "AT6": (
            "N-[2-(6-AMINO-4-METHYLPYRIDIN-2-YL)ETHYL]-4-CYANOBENZAMIDE",
            "non-polymer",
        ),
        "AT7": (
            "2'-DEOXY-2'-TRIAZA-1,2-DIEN-2-IUM-1-YL-ADENINE-5'-MONOPHOSPHATE",
            "non-polymer",
        ),
        "AT4": (
            "5'-O-[(R)-HYDROXY(THIOPHOSPHONOOXY)PHOSPHORYL]ADENOSINE",
            "non-polymer",
        ),
        "AT5": (
            "3-[(2S,4S,5R)-5,6-DICHLORO-2,4-DIMETHYL-1-OXOHEXYL]-4-HYDROXY-5,6-DIMETHOXY-2(1H)-PYRIDINONE",
            "non-polymer",
        ),
        "2T1": (
            "2-[({4-[2-(TRIFLUOROMETHYL)PHENYL]PIPERIDIN-1-YL}CARBONYL)AMINO]BENZOIC ACID",
            "non-polymer",
        ),
        "2T0": (
            "2-AMINO-4,5-DIHYDRONAPHTHO[1,2-D][1,3]THIAZOL-8-YL DIHYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "2T3": ("(2R,3R)-4-ANILINO-2,3-DIHYDROXY-4-OXOBUTANOIC ACID", "non-polymer"),
        "AT1": (
            "(S)-2-AMINO-3-(5-TERT-BUTYL-3-(PHOSPHONOMETHOXY)-4-ISOXAZOLYL)PROPIONIC ACID",
            "non-polymer",
        ),
        "PNS": ("4'-PHOSPHOPANTETHEINE", "non-polymer"),
        "H86": (
            "3-(DECYLOXY)-5-(3,5-DIFLUOROPHENYL)-1-(2,2-DIPHOSPHONOETHYL)PYRIDINIUM",
            "non-polymer",
        ),
        "0VQ": (
            "(3E,5E)-6-(3-{2-[3,4-BIS(HYDROXYMETHYL)PHENYL]ETHYL}PHENYL)-1,1,1-TRIFLUORO-2-(TRIFLUOROMETHYL)OCTA-3,5-DIEN-2-OL",
            "non-polymer",
        ),
        "0VP": (
            "3-(5'-{[3,4-BIS(HYDROXYMETHYL)BENZYL]OXY}-2'-ETHYL-2-PROPYLBIPHENYL-4-YL)PENTAN-3-OL",
            "non-polymer",
        ),
        "0VS": (
            "L-GAMMA-GLUTAMYL-S-(3-METHYL-1,4-DIOXO-1,4-DIHYDRONAPHTHALEN-2-YL)-L-CYSTEINYLGLYCINE",
            "non-polymer",
        ),
        "0VR": (
            "1-(O-CARBOXY-PHENYLAMINO)-1-DEOXY-D-RIBULOSE-5-PHOSPHATE",
            "non-polymer",
        ),
        "0VU": (
            "4-({4-[3-(PIPERIDIN-1-YLCARBONYL)PHENYL]PYRIMIDIN-2-YL}AMINO)BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "0VT": ("6-METHYLHEPT-5-EN-2-ONE", "non-polymer"),
        "PNU": (
            "6-CHLORO-2-(1-FURO[2,3-C]PYRIDIN-5-YL-ETHYLSULFANYL)-PYRIMIDIN-4-YLAMINE",
            "non-polymer",
        ),
        "NGR": ("3-O-ALPHA-D-GLUCOPYRANOSYL-ALPHA-D-GLUCOPYRANOSE", "non-polymer"),
        "PNT": ("1,5-BIS(4-AMIDINOPHENOXY)PENTANE", "non-polymer"),
        "0VA": (
            "N-[N-(4-AMINO-3,5-DICHLOROBENZYL)CARBAMIMIDOYL]-3-(4-METHOXYPHENYL)-5-METHYL-1,2-THIAZOLE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "0VB": (
            "N-{N-[4-(ACETYLAMINO)-3-CHLORO-5-METHYLBENZYL]CARBAMIMIDOYL}-3-(4-METHOXYPHENYL)-5-METHYL-1,2-THIAZOLE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "0VE": (
            "3-[8-({4-[ETHYL(2-HYDROXYETHYL)AMINO]PHENYL}AMINO)IMIDAZO[1,2-A]PYRAZIN-5-YL]PHENOL",
            "non-polymer",
        ),
        "PNW": ("4-NITROPHENYL BETA-D-GLUCOPYRANOSIDE", "saccharide"),
        "0VG": (
            "N-{6-[(2S)-2-METHYLPYRROLIDIN-1-YL]PYRIDIN-2-YL}-6-PHENYLIMIDAZO[1,2-B]PYRIDAZIN-8-AMINE",
            "non-polymer",
        ),
        "0VF": (
            "4-{[(3S)-1-{7-[(3,4-DIMETHOXYPHENYL)AMINO][1,3]THIAZOLO[5,4-D]PYRIMIDIN-5-YL}PYRROLIDIN-3-YL]CARBAMOYL}BENZOIC ACID",
            "non-polymer",
        ),
        "0VK": (
            "3-(5'-{[3,4-BIS(HYDROXYMETHYL)BENZYL]OXY}-2'-METHYL-2-PROPYLBIPHENYL-4-YL)PENTAN-3-OL",
            "non-polymer",
        ),
        "PNV": (
            "(2S,5R,6R)-3,3-DIMETHYL-7-OXO-6-(2-PHENOXYACETAMIDO)-4-THIA-1- AZABICYCLO(3.2.0)HEPTANE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "46D": (
            "(4AR,6R,7S,8R,8AR)-8-((5R,6R)-3-CARBOXY-TETRAHYDRO-4,5,6-TRIHYDROXY-2H-PYRAN-2-YLOXY)-HEXAHYDRO-6,7-DIHYDROXY-2-METHYLPYRANO[3,2-D][1,3]DIOXINE-2-CARBOXYLIC ACID)",
            "saccharide",
        ),
        "0VO": (
            "(4E,6Z)-7-(3-{[3,4-BIS(HYDROXYMETHYL)BENZYL]OXY}PHENYL)-3-ETHYLNONA-4,6-DIEN-3-OL",
            "non-polymer",
        ),
        "T42": ("MORPHOLINO-DIPHENYLALANINE-METHOXYPROPYLBORONIC ACID", "peptide"),
        "MF5": (
            "1-(1-METHYLPYRIDIN-1-IUM-2-YL)-N-[[2,3,4,5,6-PENTAKIS(FLUORANYL)PHENYL]METHOXY]METHANIMINE",
            "non-polymer",
        ),
        "7AD": (
            "9-AMINO-N-[3-(DIMETHYLAMINO)PROPYL]ACRIDINE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "7AC": (
            "7-(AMINOMETHYL)-6-(2-CHLOROPHENYL)-1-METHYL-1H-BENZIMIDAZOLE-5-CARBONITRILE",
            "non-polymer",
        ),
        "TCN": ("TETRACYANONICKELATE ION", "non-polymer"),
        "T8X": (
            "2-O-SULFO-ALPHA-D-GLUCOPYRANOSYL 2-O-HEXADECANOYL-3-O-[(2E,4S,6S,8S)-2,4,6,8-TETRAMETHYLTETRACOS-2-ENOYL]-ALPHA-D-GLUCOPYRANOSIDE",
            "non-polymer",
        ),
        "PNM": ("OPEN FORM - PENICILLIN G", "non-polymer"),
        "DSG": ("D-ASPARAGINE", "peptide"),
        "T41": ("?", "nucleic-acid"),
        "7AP": ("(1S,2S)-1-(2,4-DIAMINOPTERIDIN-6-YL)PROPANE-1,2-DIOL", "non-polymer"),
        "PNO": ("P-METHYL AMINOPHOSPHINIC ACID", "non-polymer"),
        "SB6": (
            "4-(4-FLUOROPHENYL)-1-CYCLOROPROPYLMETHYL-5-(4-PYRIDYL)-IMIDAZOLE",
            "non-polymer",
        ),
        "OBZ": ("O-BENZYLHYDROXYLAMINE", "non-polymer"),
        "PNN": ("PENICILLIN G", "non-polymer"),
        "IHG": ("N-ISOPROPYL-N'-HYDROXYGUANIDINE", "non-polymer"),
        "PNA": ("4'-NITROPHENYL-ALPHA-D-MANNOPYRANOSIDE", "saccharide"),
        "JKM": (
            "1-[(2S,3AS,4S,6R,6AR)-2-HYDROXY-6-(HYDROXYMETHYL)-2-OXIDOTETRAHYDROFURO[3,4-D][1,3,2]DIOXAPHOSPHOL-4-YL]PYRIMIDINE-2,4(1H,3H)-DIONE",
            "non-polymer",
        ),
        "7A8": ("(1R)-DIMETHYL-(4-METHYL-CYCLOHEX-3-ENYL)-AMINE", "non-polymer"),
        "PNC": ("PARA-NITROBENZYL GLUTARYL GLYCINIC ACID", "non-polymer"),
        "SJJ": (
            "5-(4-METHOXYPHENYL)-N-(PYRIDIN-4-YLMETHYL)-1,3-OXAZOLE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "0V5": ("(2R)-2-(PHOSPHONOOXY)PROPANOIC ACID", "non-polymer"),
        "0V7": ("2-[(3-FLUOROBENZYL)AMINO]ETHANESULFONIC ACID", "non-polymer"),
        "0V6": (
            "(2Z,6S)-6-{3-CHLORO-5-[5-(PROP-1-YN-1-YL)PYRIDIN-3-YL]THIOPHEN-2-YL}-2-IMINO-3,6-DIMETHYLTETRAHYDROPYRIMIDIN-4(1H)-ONE",
            "non-polymer",
        ),
        "0V8": ("2-[(3-BROMOBENZYL)AMINO]ETHANESULFONIC ACID", "non-polymer"),
        "PND": ("P-NITROPHENYLHYDRAZINE", "non-polymer"),
        "PNG": ("4'-NITROPHENYL-ALPHA-D-GLUCOPYRANOSIDE", "saccharide"),
        "PNF": (
            "6-{4-[HYDROXY-(4-NITRO-PHENOXY)-PHOSPHORYL]-BUTYRYLAMINO}-HEXANOIC ACID",
            "non-polymer",
        ),
        "XPO": ("FORMYL PHOSPHATE", "non-polymer"),
        "4GP": ("N-(BETA-D-GLUCOPYRANOSYL)OXAMIC ACID", "non-polymer"),
        "4GW": ("?", "non-polymer"),
        "W01": (
            "2,6-DIMETHYL-1-(3-[3-METHYL-5-ISOXAZOLYL]-PROPANYL)-4-[2N-METHYL-2H-TETRAZOL-5-YL]-PHENOL",
            "non-polymer",
        ),
        "4GC": (
            "BETA-D-GLUCOPYRANOSYL-(1->4)-BETA-D-GLUCOPYRANOSYL-(1->4)-BETA-D-GLUCOPYRANOSYL-(1->4)-ALPHA-D-GLUCOPYRANOSE",
            "saccharide",
        ),
        "XQK": (
            "METHYL 5-ISOXAZOL-5-YL-3-METHYL-1H-PYRAZOLE-4-CARBOXYLATE",
            "non-polymer",
        ),
        "W02": (
            "2,6-DIMETHYL-1-(3-[3-METHYL-5-ISOXAZOLYL]-PROPANYL)-4-[4-METHYL-2H-TETRAZOL-2-YL]-PHENOL",
            "non-polymer",
        ),
        "4GE": (
            "5-[BENZYL(METHYL)CARBAMOYL]-2-(3-CHLOROPHENYL)-1H-IMIDAZOLE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "YE1": ("?", "non-polymer"),
        "4GH": (
            "N-{(2S,3R)-1-[(4-DEOXY-ALPHA-D-XYLO-HEXOPYRANOSYL)OXY]-3-HYDROXYOCTADECAN-2-YL}HEXACOSANAMIDE",
            "non-polymer",
        ),
        "4GI": (
            "HYDROXY(OXO)[3-({(2S)-2-[(3-PHENYLPROPOXY)CARBONYL]PIPERIDIN-1-YL}SULFONYL)PHENYL]AMMONIUM",
            "non-polymer",
        ),
        "YE7": ("IMIDAZO[2,1-A]ISOQUINOLINE-2-CARBOHYDRAZIDE", "non-polymer"),
        "YE6": ("5-(2-CHLOROPHENYL)FURAN-2-CARBOHYDRAZIDE", "non-polymer"),
        "SB5": (
            "4-(FLUOROPHENYL)-1-CYCLOPROPYLMETHYL-5-(2-AMINO-4-PYRIMIDINYL)IMIDAZOLE",
            "non-polymer",
        ),
        "PDD": ("N-(5'-PHOSPHOPYRIDOXYL)-D-ALANINE", "peptide"),
        "I76": (
            "3-{6-[(3-CHLOROBENZYL)OXY]NAPHTHALEN-2-YL}-1-(PROPAN-2-YL)-1H-PYRAZOLO[3,4-D]PYRIMIDIN-4-AMINE",
            "non-polymer",
        ),
        "I75": ("1-(4-HYDROXY-3-METHOXYPHENYL)ETHANONE", "non-polymer"),
        "T45": ("METHYL 2H-1,2,3-TRIAZOLE-4-CARBOXYLATE", "non-polymer"),
        "XI7": ("2-4-DIOXO-4-PHENYLBUTANOIC ACID", "non-polymer"),
        "G7P": ("6,7-DIDEOXY-7-PHOSPHONO-BETA-D-GLUCO-HEPTOPYRANOSE", "non-polymer"),
        "MEE": ("METHANETHIOL", "non-polymer"),
        "TCO": ("TERT-BUTYL(1S)-1-CYCLOHEXYL-2-OXOETHYLCARBAMATE", "non-polymer"),
        "I7C": ("4-AMINO-6-(TRIFLUOROMETHYL)BENZENE-1,3-DISULFONAMIDE", "non-polymer"),
        "I7B": ("4-AMINO-6-CHLOROBENZENE-1,3-DISULFONAMIDE", "non-polymer"),
        "I7A": ("4,5-DICHLOROBENZENE-1,3-DISULFONAMIDE", "non-polymer"),
        "XIG": ("?", "non-polymer"),
        "XIH": (
            "2-({8-[(3R)-3-AMINOPIPERIDIN-1-YL]-1,3-DIMETHYL-2,6-DIOXO-1,2,3,6-TETRAHYDRO-7H-PURIN-7-YL}METHYL)BENZONITRILE",
            "non-polymer",
        ),
        "PDU": ("5(1-PROPYNYL)-2'-DEOXYURIDINE-5-MONOPHOSPHATE", "nucleic-acid"),
        "XIL": (
            "3-HYDROXY-4-(3,4,5-TRIHYDROXY-TETRAHYDRO-PYRAN-2-YLOXY)-PIPERIDIN-2-ONE",
            "non-polymer",
        ),
        "MEG": ("(2S,3S)-2-AZANYL-3-METHYL-PENTANEDIOIC ACID", "peptide"),
        "NDH": ("(1R, 2S)-CIS 1,2 DIHYDROXY-1,2-DIHYDRONAPHTHALENE", "non-polymer"),
        "U49": (
            "(20S)-19,20,21,22-TETRAHYDRO-19-OXO-5H-18,20-ETHANO-12,14-ETHENO-6,10-METHENO-18H-BENZ[D]IMIDAZO[4,3-K][1,6,9,12]OXATRIAZA-CYCLOOCTADECOSINE-9-CARBONITRILE",
            "non-polymer",
        ),
        "I7T": ("7-IODOTETRACYCLINE", "non-polymer"),
        "I7P": (
            "(1R,2R,3S,4S,5R,6S)-2,3,4,5,6-PENTAKIS(PHOSPHONOOXY)CYCLOHEXYL TRIHYDROGEN DIPHOSPHATE",
            "non-polymer",
        ),
        "XIX": ("3,5-DIAMINO-1H-[1,2,4]TRIAZOLE", "non-polymer"),
        "MEA": ("N-METHYLPHENYLALANINE", "peptide"),
        "XIZ": ("(E)-PYRIDIN-4-YL-ACRYLIC ACID", "non-polymer"),
        "4G2": (
            "2-(3-CHLORO-PHENYL)-5-METHYL-1H-IMIDAZOLE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "MEC": ("ETHYL-CARBAMIC ACID METHYL ESTER", "non-polymer"),
        "4G5": (
            "4-[(2-AMINO-2-OXOETHYL)(METHYL)CARBAMOYL]-2-PHENYL-1H-IMIDAZOLE-5-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "4G8": (
            "4-[BENZYL(CARBOXYMETHYL)CARBAMOYL]-2-PHENYL-1H-IMIDAZOLE-5-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "YEG": ("CYCLOPENTANOL", "non-polymer"),
        "HF1": ("4-(2-HYDROXY-4-FLUOROPHENYLTHIO)-BUTYLPHOSPHONIC ACID", "non-polymer"),
        "SB8": ("2-(3-BENZOYLPHENOXY)ETHYL(HYDROXY)FORMAMIDE", "non-polymer"),
        "MEO": ("OXYGENS BOUND TO MET SD", "non-polymer"),
        "RUR": (
            "[3,3'-(7,12-DIETHYL-3,8,13,17-TETRAMETHYLPORPHYRIN-2,18-DIYL-KAPPA~4~N~21~,N~22~,N~23~,N~24~)DIPROPANOATO(2-)]RUTHENIUM",
            "non-polymer",
        ),
        "8XQ": ("8-HYDROXYQUINOLINE-5-CARBOXYLIC ACID", "non-polymer"),
        "MEN": ("N-METHYL ASPARAGINE", "peptide"),
        "PFY": (
            "(3Z)-5-[(2,6-DICHLOROBENZYL)SULFONYL]-3-[(3,5-DIMETHYL-4-{[(2S)-2-(PYRROLIDIN-1-YLMETHYL)PYRROLIDIN-1-YL]CARBONYL}-1H-PYRROL-2-YL)METHYLIDENE]-1,3-DIHYDRO-2H-INDOL-2-ONE",
            "non-polymer",
        ),
        "WW2": ("CYCLOHEXYL (S)-METHYLPHOSPHONOFLUORIDOATE", "non-polymer"),
        "2IW": (
            "4-[(3AR,4S,7R,7AS)-1,3-DIOXO-1,3,3A,4,7,7A-HEXAHYDRO-2H-4,7-METHANOISOINDOL-2-YL]-N-(4-METHYLQUINOLIN-8-YL)BENZAMIDE",
            "non-polymer",
        ),
        "WW8": (
            "N-[(3S)-1-({6-CHLORO-3-[1-(4-CHLOROBENZYL)-4-PHENYL-1H-IMIDAZOL-5-YL]-1H-INDOL-2-YL}CARBONYL)PYRROLIDIN-3-YL]-N,N',N'-TRIMETHYLPROPANE-1,3-DIAMINE",
            "non-polymer",
        ),
        "2IP": ("D-MYO-INOSITOL-1,4-BISPHOSPHATE", "non-polymer"),
        "2IN": (
            "N-(BENZYLSULFONYL)SERYL-N~1~-{4-[AMINO(IMINO)METHYL]BENZYL}GLYCINAMIDE",
            "non-polymer",
        ),
        "2IM": ("PIPERIDIN-2-IMINE", "non-polymer"),
        "2IH": ("1,1'-HEXANE-1,6-DIYLBIS(1H-IMIDAZOLE)", "non-polymer"),
        "SB9": ("HYDROXY[3-(6-METHYLPYRIDIN-2-YL)PROPYL]FORMAMIDE", "non-polymer"),
        "2IG": (
            "6-ETHYL-5-[9-(3-METHOXYPROPYL)-9H-CARBAZOL-2-YL]PYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "2IB": ("2-IODOBENZYL GROUP", "non-polymer"),
        "2IC": (
            "5-[(5S,9R)-9-(4-CYANOPHENYL)-3-(3,5-DICHLOROPHENYL)-1-METHYL-2,4-DIOXO-1,3,7-TRIAZASPIRO [4.4]NON-7-YL]METHYL]-3-THIOPHENECARBOXYLICACID",
            "non-polymer",
        ),
        "2IA": ("?", "nucleic-acid"),
        "833": ("3-(1,3-BENZODIOXOL-5-YL)-1-METHYL-1H-PYRAZOL-5-AMINE", "non-polymer"),
        "830": (
            "DIPOTASSIUM (2-OXO-2-{[3-(3-PHENOXYPHENYL)PROPYL]AMINO}ETHYL)PHOSPHONATE",
            "non-polymer",
        ),
        "831": (
            "N-[3-(3-{4-[(DIMETHYLAMINO)METHYL]PHENYL}PYRAZOLO[1,5-A]PYRIMIDIN-7-YL)PHENYL]-3-(TRIFLUOROMETHYL)BENZAMIDE",
            "non-polymer",
        ),
        "834": (
            "N-[2-(2-CHLOROPHENYL)-4-METHYL-5-(1-METHYLETHYL)-1H-IMIDAZOL-1-YL]-5-METHYL-2,3-DIHYDRO-1,4-BENZODIOXINE-6-CARBOXAMIDE",
            "non-polymer",
        ),
        "GDA": ("4-DEOXY-4-AMINO-BETA-D-GLUCOSE", "non-polymer"),
        "NDM": ("7,9-DIMETHYLGUANINE", "non-polymer"),
        "839": (
            "3-(9-HYDROXY-1,3-DIOXO-4-PHENYL-2,3-DIHYDROPYRROLO[3,4-C]CARBAZOL-6(1H)-YL)PROPANOIC ACID",
            "non-polymer",
        ),
        "GDB": ("1-(S-GLUTATHIONYL)-2,4-DINITROBENZENE", "non-polymer"),
        "2S8": (
            "(2S)-8-[(TERT-BUTOXYCARBONYL)AMINO]-2-(1H-INDOL-3-YL)OCTANOIC ACID",
            "non-polymer",
        ),
        "GDC": ("GUANOSINE-5'-DIPHOSPHATE-BETA-L-GALACTOSE", "non-polymer"),
        "VG8": (
            "4-(1,4'-BIPIPERIDIN-1'-YL)-7-({5-CHLORO-2-[(2-METHOXYPHENYL)AMINO]PYRIMIDIN-4-YL}AMINO)-2-METHYL-2,3-DIHYDRO-1H-ISOINDOL-1-ONE",
            "non-polymer",
        ),
        "VG9": (
            "6-{[(2,5-DIETHOXYPHENYL)AMINO]METHYL}-5-METHYLPYRIDO[2,3-D]PYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "VG4": (
            "N-[(1S,2R)-1-BENZYL-3-{[(1S)-2-(CYCLOHEXYLAMINO)-1-METHYL-2-OXOETHYL]AMINO}-2-HYDROXYPROPYL]-3-(2-OXO-2,3-DIHYDRO-1H-PYRROL-1-YL)-5-PROPOXYBENZAMIDE",
            "non-polymer",
        ),
        "GDD": ("GUANOSINE-5'-DIPHOSPHATE-ALPHA-D-MANNOSE", "non-polymer"),
        "VG6": (
            "N-[(1S,2R)-1-BENZYL-2-HYDROXY-3-{[3-(TRIFLUOROMETHYL)BENZYL]AMINO}PROPYL]-3-(ETHYLAMINO)-5-(2-OXOPYRROLIDIN-1-YL)BENZAMIDE",
            "non-polymer",
        ),
        "VG7": (
            "N-[(1S,2R)-1-BENZYL-3-(CYCLOHEXYLAMINO)-2-HYDROXYPROPYL]-3-(ETHYLAMINO)-5-(2-OXOPYRROLIDIN-1-YL)BENZAMIDE",
            "non-polymer",
        ),
        "VG0": (
            "N-{(1S,2R)-1-BENZYL-2-HYDROXY-3-[(1,1,5-TRIMETHYLHEXYL)AMINO]PROPYL}-3-(ETHYLAMINO)-5-(2-OXOPYRROLIDIN-1-YL)BENZAMIDE",
            "non-polymer",
        ),
        "VG1": ("ALPHA-D-GLUCOSE-1-PHOSPHATE-6-VANADATE", "saccharide"),
        "VG2": (
            "4-(2-AMINOETHOXY)-N-(3-CHLORO-5-PIPERIDIN-1-YLPHENYL)-3,5-DIMETHYLBENZAMIDE",
            "non-polymer",
        ),
        "MEP": (
            "PHOSPHORIC ACID MONO-[4-METHOXY-3-METHYL-5-(5-METHYL-2,4-DIOXO-3,4-DIHYDRO-2H-PYRIMIDIN-1-YL)-TETRAHYDRO-FURAN-2-YLMETHYL] ESTER",
            "nucleic-acid",
        ),
        "VGL": ("PYRAZINE-2-CARBOXYLIC ACID", "non-polymer"),
        "PRE": ("PREPHENIC ACID", "non-polymer"),
        "VGN": ("2-ACETAMIDO-5-(4-PHENYLPHENYL)BENZOIC ACID", "non-polymer"),
        "VGH": (
            "3-[(1R)-1-(2,6-DICHLORO-3-FLUOROPHENYL)ETHOXY]-5-(1-PIPERIDIN-4-YL-1H-PYRAZOL-4-YL)PYRIDIN-2-AMINE",
            "non-polymer",
        ),
        "MES": ("2-(N-MORPHOLINO)-ETHANESULFONIC ACID", "non-polymer"),
        "VGJ": ("(3R,7S,11S)-3,7,11,15-TETRAMETHYLHEXADECANOIC ACID", "non-polymer"),
        "PRD": (
            "N6-(2,5-DIMETHOXY-BENZYL)-N6-METHYL-PYRIDO[2,3-D]PYRIMIDINE-2,4,6-TRIAMINE",
            "non-polymer",
        ),
        "VGD": ("6-CHLORO-1H-BENZIMIDAZOL-2-AMINE", "non-polymer"),
        "VGF": (
            "1-(3,4-DICHLOROBENZYL)-7-PHENYL-1H-BENZIMIDAZOL-2-AMINE",
            "non-polymer",
        ),
        "MER": (
            "(4R,5S)-3-{[(3S,5S)-5-(DIMETHYLCARBAMOYL)PYRROLIDIN-3-YL]SULFANYL}-5-[(2S,3R)-3-HYDROXY-1-OXOBUTAN-2-YL]-4-METHYL-4,5-DIHYDRO-1H-PYRROLE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "TF5": ("2-[(2',3',4'-TRIFLUOROBIPHENYL-2-YL)OXY]ETHANOL", "non-polymer"),
        "VGB": (
            "(5R,6R,7R,8S)-6,7-DIHYDROXY-5-(HYDROXYMETHYL)-2-(2-PHENYLETHYL)-8-(PROPANOYLAMINO)-5,6,7,8-TETRAHYDRO-1H-IMIDAZO[1,2-A]PYRIDIN-4-IUM",
            "non-polymer",
        ),
        "VGC": (
            "6-CYCLOHEXYL-4-METHYL-5-PHENYL-4H-THIENO[3,2-B]PYRROLE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "PRF": ("7-DEAZA-7-AMINOMETHYL-GUANINE", "non-polymer"),
        "83H": (
            "8-ETHYL-3,10,10-TRIMETHYL-4,5,6,8,10,12-HEXAHYDROPYRAZOLO[4',3':6,7]CYCLOHEPTA[1,2-B]PYRROLO[2,3-F]INDOL-9(1H)-ONE",
            "non-polymer",
        ),
        "VGP": (
            "(1R)-1,4-ANHYDRO-6-DEOXY-1-[(6R)-8-ETHENYL-1,6-DIHYDROXY-10,12-DIMETHOXY-6H-DIBENZO[C,H]CHROMEN-4-YL]-D-GALACTITOL",
            "non-polymer",
        ),
        "MR9": (
            "4-METHYL-3-({3-[2-(METHYLAMINO)PYRIMIDIN-4-YL]PYRIDIN-2-YL}OXY)-N-[2-MORPHOLIN-4-YL-5-(TRIFLUOROMETHYL)PHENYL]BENZAMIDE",
            "non-polymer",
        ),
        "WWV": (
            "4-({[2-(1-METHYLETHYL)PHENYL]CARBAMOYL}AMINO)BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "TD4": (
            "4-{HYDROXY[(PHOSPHONOOXY)ACETYL]AMINO}BUTYL DIHYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "MEY": (
            "N-{[1-(5-CHLORO-2,4-DIHYDROXYPHENYL)-2-OXO-2,3-DIHYDRO-1H-BENZIMIDAZOL-5-YL]METHYL}NAPHTHALENE-1-SULFONAMIDE",
            "non-polymer",
        ),
        "NB1": (
            "(3AR,5R,6S,7R,7AR)-5-(HYDROXYMETHYL)-2-PROPYL-5,6,7,7A-TETRAHYDRO-3AH-PYRANO[3,2-D][1,3]THIAZOLE-6,7-DIOL",
            "non-polymer",
        ),
        "MEX": (
            "1-(2,4-DIHYDROXYPHENYL)-1,3-DIHYDRO-2H-BENZIMIDAZOL-2-ONE",
            "non-polymer",
        ),
        "Q20": (
            "3-(2,4-DIAMINO-6-METHYLQUINAZOLIN-7-YL)-4-ETHOXYBENZALDEHYDE",
            "non-polymer",
        ),
        "PRM": (
            "3,8-DIAMINO-5[3-(DIETHYLMETHYLAMMONIO)PROPYL]-6-PHENYLPHENANTHRIDINIUM",
            "non-polymer",
        ),
        "Q22": (
            "(4AS)-5-[(2,4-DIAMINOPTERIDIN-6-YL)METHYL]-4A,5-DIHYDRO-2H-DIBENZO[B,F]AZEPIN-8-OL",
            "non-polymer",
        ),
        "Q23": (
            "(3R)-1,1-DIOXIDO-2,3-DIHYDROTHIOPHEN-3-YL BENZENESULFONATE",
            "non-polymer",
        ),
        "Q24": ("QUINAZOLINE-2,4-DIAMINE", "non-polymer"),
        "Q26": (
            "N-[3'-(2,4-DIAMINOQUINAZOLIN-7-YL)-4'-ETHOXYBIPHENYL-3-YL]METHANESULFONAMIDE",
            "non-polymer",
        ),
        "MR5": ("4-(1,3-BENZOXAZOL-2-YL)-2,6-DIBROMOPHENOL", "non-polymer"),
        "MEZ": ("(2E)-2-METHYLBUT-2-ENEDIOIC ACID", "non-polymer"),
        "JOZ": ("3,4-DIHYDROISOQUINOLIN-1(2H)-ONE", "non-polymer"),
        "FSA": ("1-S-(CARBOXYMETHYL)-1-THIO-BETA-D-FRUCTOPYRANOSE", "non-polymer"),
        "FSB": ("N-(2-FLOURO-BENZYL)-4-SULFAMOYL-BENZAMIDE", "non-polymer"),
        "FSC": ("FUSICOCCIN", "non-polymer"),
        "PS2": (
            "O-[{[(2R)-2-{[(9S,10S)-9,10-DIBROMOOCTADECANOYL]OXY}-3-(PALMITOYLOXY)PROPYL]OXY}(HYDROXY)PHOSPHORYL]-L-SERINE",
            "non-polymer",
        ),
        "FSE": ("3,7,3',4'-TETRAHYDROXYFLAVONE", "non-polymer"),
        "FSF": ("4FE-2S-O HYBRID CLUSTER", "non-polymer"),
        "FSG": (
            "[(1S)-1-FLUORO-2-(HYDROXYAMINO)-2-OXOETHYL]PHOSPHONIC ACID",
            "non-polymer",
        ),
        "FSH": ("?", "non-polymer"),
        "FSI": (
            "5-(ACETYLAMINO)-3,5-DIDEOXY-3-FLUORO-D-ERYTHRO-ALPHA-L-MANNO-NON-2-ULOPYRANOSONIC ACID",
            "saccharide",
        ),
        "FSK": (
            "6-[(1Z)-3-FLUORO-2-(HYDROXYMETHYL)PROP-1-EN-1-YL]-1,5-DIMETHYLPYRIMIDINE-2,4(1H,3H)-DIONE",
            "non-polymer",
        ),
        "FSM": ("FORSMANN ANTIGEN", "non-polymer"),
        "FSN": (
            "(3ASR,4RS,8ASR,8BRS)-4-(2-(4-FLUOROBENZYL)-1,3-DIOXODEACAHYDROPYRROLO[3,4-A] PYRROLIZIN-4-YL)BENZAMIDINE",
            "non-polymer",
        ),
        "FSO": ("IRON/SULFUR/OXYGEN HYBRID CLUSTER", "non-polymer"),
        "FSP": (
            "[1-(4-FLUOROBENZYL)CYCLOBUTYL]METHYL (1S)-1-[OXO(1H-PYRAZOL-5-YLAMINO)ACETYL]PENTYLCARBAMATE",
            "non-polymer",
        ),
        "RU8": (
            "BIS(2,2'-BIPYRIDINE-KAPPA~2~N~1~,N~1'~)[2-IODO-N-(1,10-PHENANTHROLIN-5-YL-KAPPA~2~N~1~,N~10~)ACETAMIDE]RUTHENIUM(2+)",
            "non-polymer",
        ),
        "FSS": (
            "3-(2,6-DICHLOROPHENYL)-7-({4-[(DIETHYLAMINO)METHOXY]PHENYL}AMINO)-1-METHYL-3,4-DIHYDROPYRIMIDO[4,5-D]PYRIMIDIN-2(1H)-ONE",
            "non-polymer",
        ),
        "FST": (
            "O,O-DIETHYL O-{4-[(R)-METHYLSULFINYL]PHENYL} PHOSPHOROTHIOATE",
            "non-polymer",
        ),
        "PRK": ("N~6~-PROPANOYL-L-LYSINE", "peptide"),
        "FSX": (
            "BIS-(MU-2-OXO),[(MU-3--SULFIDO)-BIS(MU-2--SULFIDO)-TRIS(CYS-S)-TRI-IRON] (AQUA)(GLU-O)IRON(II)",
            "non-polymer",
        ),
        "MR3": ("1-METHYL-1H-PYRROLE", "non-polymer"),
        "RU6": ("unknown", "non-polymer"),
        "RU7": ("PARA-CYMENE RUTHENIUM CHLORIDE", "non-polymer"),
        "RU5": (
            "2-[4-(4-CHLOROPHENOXY)PHENYL]-1H-BENZIMIDAZOLE-6-CARBOXAMIDE",
            "non-polymer",
        ),
        "PHX": (
            "4-{[(Z)-(5-OXO-2-PHENYL-1,3-OXAZOL-4(5H)-YLIDENE)METHYL]AMINO}BUTANOIC ACID",
            "non-polymer",
        ),
        "WP7": (
            "(4S)-6-CHLORO-3-{2-[4-(FURAN-2-YLCARBONYL)PIPERAZIN-1-YL]ETHYL}-2-METHYL-4-PHENYL-3,4-DIHYDROQUINAZOLINE",
            "non-polymer",
        ),
        "XPE": ("3,6,9,12,15,18,21,24,27-NONAOXANONACOSANE-1,29-DIOL", "non-polymer"),
        "R2G": ("2,5-ANHYDRO-4-DEOXY-D-ERYTHRO-PENT-4-ENITOL", "non-polymer"),
        "PRV": ("(2R)-AMINO(2-NITROPHENYL)ETHANOIC ACID", "peptide"),
        "WR1": (
            "NALPHA-[(BENZYLOXY)CARBONYL]-N-[(1R)-4-HYDROXY-1-METHYL-2-OXOBUTYL]-L-PHENYLALANINAMIDE",
            "non-polymer",
        ),
        "ZT0": (
            "(4-CARBOXY-1,3-BENZODIOXOL-5-YL)-N-{2-[(4-METHOXYBENZYL)CARBAMOYL]BENZYL}-N-METHYLMETHANAMINIUM",
            "non-polymer",
        ),
        "ZT2": (
            "5-[(E)-(2-OXO-2,3-DIHYDRO-1H-INDEN-1-YLIDENE)METHYL]-1,3-BENZODIOXOLE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "ZTH": ("(S)-1'-(2',3'-DIHYDROXYPROPYL)-THYMINE", "nucleic-acid"),
        "RG1": ("RHODOPIN GLUCOSIDE", "saccharide"),
        "R2C": ("5-METHOXY-2-[(4-PHENOXYPHENYL)AMINO]BENZOIC ACID", "non-polymer"),
        "PRR": ("3-(METHYL-PYRIDINIUM)ALANINE", "peptide"),
        "ZTX": (
            "2-((4-(4-HYDROXYPIPERIDIN-1-YL)PHENYL)AMINO)-5,11-DIMETHYL-5H-BENZO[E]PYRIMIDO [5,4-B][1,4]DIAZEPIN-6(11H)-ONE",
            "non-polymer",
        ),
        "ZTW": ("RALOXIFENE CORE", "non-polymer"),
        "ZTP": (
            "(2Z,6Z,10Z,14Z,18Z,22Z,26Z)-3,7,11,15,19,23,27,31-OCTAMETHYLDOTRIACONTA-2,6,10,14,18,22,26,30-OCTAEN-1-YL TRIHYDROGEN DIPHOSPHATE",
            "non-polymer",
        ),
        "RUN": (
            "2-[(1R,3S,4S)-1-BENZYL-4-[N-(BENZYLOXYCARBONYL)-L-VALYL]AMINO-3-PHENYLPENTYL]-4(5)-(2-METHYLPROPIONYL)IMIDAZOLE",
            "non-polymer",
        ),
        "RUO": ("OXO-CENTERED TRIRUTHENIUM CLUSTER", "non-polymer"),
        "RUM": (
            "2-({2-[(3R)-3-AMINOPIPERIDIN-1-YL]-5-BROMO-6-OXOPYRIMIDIN-1(6H)-YL}METHYL)BENZONITRILE",
            "non-polymer",
        ),
        "RUB": ("RIBULOSE-1,5-DIPHOSPHATE", "saccharide"),
        "RUC": ("(ETA6-BENZENE)RUTHENIUM", "non-polymer"),
        "RUA": ("(4'-METHYL-2,2'BIPYRIDINE)BIS(2,2'-BIPYRIDINE)", "non-polymer"),
        "RUF": (
            "7-(2-CHLOROBENZYL)-1,3-DIMETHYL-8-PIPERAZIN-1-YL-3,7-DIHYDRO-1H-PURINE-2,6-DIONE",
            "non-polymer",
        ),
        "RUG": (
            "1-BETA-D-GLUCOPYRANOSYL-4-(HYDROXYMETHYL)-1H-1,2,3-TRIAZOLE",
            "non-polymer",
        ),
        "FS1": ("IRON/SULFUR CLUSTER", "non-polymer"),
        "FS2": ("FE-S-O HYBRID CLUSTER", "non-polymer"),
        "FS3": ("FE3-S4 CLUSTER", "non-polymer"),
        "FS4": ("IRON/SULFUR CLUSTER", "non-polymer"),
        "PYV": (
            "4-BROMO-N-(PYRIDIN-2-YLMETHYL)NAPHTHALENE-1-SULFONAMIDE",
            "non-polymer",
        ),
        "FS8": (
            "N-{5-[(7-{[(2R)-2,3-DIHYDROXYPROPYL]OXY}-5-OXO-10,11-DIHYDRO-5H-DIBENZO[A,D][7]ANNULEN-2-YL)AMINO]-2-FLUOROPHENYL}BENZAMIDE",
            "non-polymer",
        ),
        "RUS": (
            "1-(5-O-PHOSPHONO-BETA-D-RIBOFURANOSYL)-2-SELANYLPYRIMIDIN-4(1H)-ONE",
            "nucleic-acid",
        ),
        "RUP": (
            "(R)-2-(3-ADAMANTAN-1-YL-UREIDO)-3-(3-CARBAMIMIDOYL-PHENYL)-N-PHENETHYL-PROPIONAMIDE",
            "non-polymer",
        ),
        "PRZ": ("2-ISOBUTYL-3-METHOXYPYRAZINE", "non-polymer"),
        "QWE": (
            "AMINO{[(4S)-4-({[5-(DIMETHYLAMINO)NAPHTHALEN-1-YL]SULFONYL}AMINO)-5-OXO-5-{(2R)-2-[3-OXO-3-(1,3-THIAZOL-2-YL)PROPYL]PIPERIDIN-1-YL}PENTYL]AMINO}METHANIMINIUM",
            "peptide",
        ),
        "RUT": ("RUTIN", "non-polymer"),
        "VHH": (
            "5-[(3AS,4S,6AR)-1,3-DIMETHYL-2-OXOHEXAHYDRO-1H-THIENO[3,4-D]IMIDAZOL-4-YL]PENTANOIC ACID",
            "non-polymer",
        ),
        "DIL": ("D-ISOLEUCINE", "peptide"),
        "DIM": ("DIIMIDAZOLE LEXITROPSIN", "non-polymer"),
        "DIN": ("1,6-DIHYDROXY NAPHTHALENE", "non-polymer"),
        "DIO": ("1,4-DIETHYLENE DIOXIDE", "non-polymer"),
        "DIH": (
            "3-HYDROXY-4-HYDROXYMETHYL-1-(4-OXO-4,4A,5,7A-TETRAHYDRO-3H-PYRROLO[3,2-D]PYRIMIDIN-7-YLMETHYL)-PYRROLIDINIUM",
            "non-polymer",
        ),
        "DII": ("METHYLPHOSPHONIC ACID DIISOPROPYL ESTER", "non-polymer"),
        "JHG": (
            "4-{(3-PHENYLPROPYL)[(3S,4S)-4-{[4-(TRIFLUOROMETHYL)BENZYL]AMINO}PYRROLIDIN-3-YL]SULFAMOYL}BENZAMIDE",
            "non-polymer",
        ),
        "6MO": ("MOLYBDENUM(VI) ION", "non-polymer"),
        "DID": ("4,4'[1,6-HEXANEDIYLBIS(OXY)]BISBENZENECARBOXIMIDAMIDE", "non-polymer"),
        "DIE": ("2,6-DIETHYLPHENOL", "non-polymer"),
        "DIF": ("2-[2,6-DICHLOROPHENYL)AMINO]BENZENEACETIC ACID", "non-polymer"),
        "DIG": ("2,5-DIDEOXY-2,5-IMINO-D-GLUCITOL", "saccharide"),
        "JHM": ("2-DEOXY-6-O-SULFO-ALPHA-D-ARABINO-HEXOPYRANOSE", "non-polymer"),
        "DIA": ("OCTANE 1,8-DIAMINE", "non-polymer"),
        "DIB": ("3-AMINO-(DIMETHYLPROPYLAMINE)", "non-polymer"),
        "DIC": ("3,4-DICHLOROISOCOUMARIN", "non-polymer"),
        "C19": (
            "3-(2-AMINOQUINAZOLIN-6-YL)-1-(3,3-DIMETHYLINDOLIN-6-YL)-4-METHYLPYRIDIN-2(1H)-ONE",
            "non-polymer",
        ),
        "C18": (
            "7-[(3-CHLOROBENZYL)OXY]-4-[(METHYLAMINO)METHYL]-2H-CHROMEN-2-ONE",
            "non-polymer",
        ),
        "DIX": ("METHYL(CYCLOPENTYL-PROPYL)AMINE", "non-polymer"),
        "DIY": ("5-BUTYLPIPERIDINE", "non-polymer"),
        "DIZ": (
            "(4-CHLOROPHENYL)[3-(4-CHLOROPHENYL)-7-IODO-2,5-DIOXO-1,2,3,5-TETRAHYDRO-4H-1,4-BENZODIAZEPIN-4-YL]ACETIC ACID",
            "non-polymer",
        ),
        "C13": ("1-{3-[(3,5-DICHLOROBENZYL)AMINO]PROPYL}-3-PHENYLUREA", "non-polymer"),
        "C12": (
            "2-(1-AMINO-2-HYDROXYPROPYL)-4-(4-HYDROXYBENZYL)-1-(2-OXOETHYL)-1H-IMIDAZOL-5-OLATE",
            "peptide",
        ),
        "C11": ("UNDECYL-PHOSPHINIC ACID BUTYL ESTER", "non-polymer"),
        "C10": ("HEXAETHYLENE GLYCOL MONODECYL ETHER", "non-polymer"),
        "C17": (
            "7-[(3-CHLOROBENZYL)OXY]-2-OXO-2H-CHROMENE-4-CARBALDEHYDE",
            "non-polymer",
        ),
        "C16": (
            "4-(4-METHYL-1,3-DIOXO-1,3-DIHYDRO-2H-ISOINDOL-2-YL)BENZONITRILE",
            "non-polymer",
        ),
        "C15": ("N-DODECYL-N,N-DIMETHYL-3-AMMONIO-1-PROPANESULFONATE", "non-polymer"),
        "C14": ("TETRADECANE", "non-polymer"),
        "ZO3": ("ZINC ION, 3 WATERS COORDINATED", "non-polymer"),
        "XV1": ("7-(2-FLUOROBENZYL)QUINOLIN-8-OL", "non-polymer"),
        "IR3": ("IRIDIUM (III) ION", "non-polymer"),
        "IR4": (
            "1-[7-(MORPHOLIN-4-YL)-1-(PYRIDIN-2-YL)INDOLIZIN-3-YL]ETHANONE",
            "non-polymer",
        ),
        "IR5": (
            "1-{1-[2-(HYDROXYMETHYL)PHENYL]-7-PHENOXYINDOLIZIN-3-YL}ETHANONE",
            "non-polymer",
        ),
        "IR6": (
            "1-{1-[2-(METHYLSULFONYL)PHENYL]-7-PHENOXYINDOLIZIN-3-YL}ETHANONE",
            "non-polymer",
        ),
        "XV6": ("XV638", "non-polymer"),
        "PHS": ("PHOSPHONIC ACID", "non-polymer"),
        "TXN": ("?", "non-polymer"),
        "I": ("INOSINIC ACID", "nucleic-acid"),
        "R69": (
            "N-[4-({4-[5-(DIMETHYLAMINO)-2-HYDROXYBENZOYL]BENZOYL}AMINO)AZEPAN-3-YL]ISONICOTINAMIDE",
            "non-polymer",
        ),
        "IRS": ("2-METHYLPROP-2-EN-1-AMINE", "non-polymer"),
        "6MC": ("CIS-N6-METHYL-DEOXY-ADENOSINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "IRI": ("IRIDIUM HEXAMMINE ION", "non-polymer"),
        "UTP": ("?", "non-polymer"),
        "IRN": ("1-(5-O-PHOSPHONO-BETA-D-RIBOFURANOSYL)-1H-IMIDAZOLE", "nucleic-acid"),
        "PHR": (
            "3-(1,3,7-TRIHYDRO-9-D-RIBITYL-2,6,8-PURINETRIONE-7-YL ) HEXANE 1-PHOSPHATE",
            "non-polymer",
        ),
        "IRA": ("(5-BROMO-2-METHOXYPHENYL)ACETIC ACID", "non-polymer"),
        "IRB": ("unknown", "non-polymer"),
        "6MA": ("N6-METHYL-DEOXY-ADENOSINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "XVE": (
            "PHENYLMETHYL N-[(2S)-4-CHLORO-3-OXO-1-PHENYL-BUTAN-2-YL]CARBAMATE",
            "non-polymer",
        ),
        "IRE": (
            "3-CHLORO-4-FLUORO-N-[(4Z)-7-METHOXY-6-(3-MORPHOLIN-4-YLPROPOXY)QUINAZOLIN-4(1H)-YLIDENE]ANILINE",
            "non-polymer",
        ),
        "IRF": (
            "1-DEOXY-1-(8-IODO-7-METHYL-2,4-DIOXO-3,4-DIHYDROBENZO[G]PTERIDIN-10(2H)-YL)-D-RIBITOL",
            "non-polymer",
        ),
        "IRG": ("N~4~-CYCLOPROPYL-2-PHENYLQUINAZOLINE-4,7-DIAMINE", "non-polymer"),
        "C1H": (
            "4-(NITROOXY)BUTYL 3-[(5-SULFAMOYL-1,3,4-THIADIAZOL-2-YL)SULFAMOYL]BENZOATE",
            "non-polymer",
        ),
        "C1O": ("CU-O LINKAGE", "non-polymer"),
        "C1N": (
            "N-[(BENZYLOXY)CARBONYL]LEUCYL-N~1~-[3-FLUORO-1-(4-HYDROXYBENZYL)-2-OXOPROPYL]LEUCINAMIDE",
            "non-polymer",
        ),
        "C1M": ("1-(4-CHLOROPHENYL)-1H-TETRAZOLE", "non-polymer"),
        "C1L": (
            "6-METHOXY-3,8-DIMETHYL-4-(MORPHOLIN-4-YLMETHYL)-1H-PYRAZOLO[3,4-B]QUINOLINE",
            "non-polymer",
        ),
        "C1C": ("1-(2-CARBOXYNAPHTH-1YL)-2-NAPHTHOIC ACID", "non-polymer"),
        "C1B": ("?", "non-polymer"),
        "C1A": (
            "(2,5-DIMETHYLBENZENE-1,4-DIYL)DIMETHANEDIYL BIS(N-CARBAMIMIDOYLCARBAMIMIDOTHIOATE)",
            "non-polymer",
        ),
        "PWZ": (
            "(5E)-7-{(1R,4S,5S,6R)-5-[(1E,3S)-3-HYDROXYOCT-1-EN-1-YL]-2-OXABICYCLO[2.2.1]HEPT-6-YL}HEPT-5-ENOIC ACID",
            "non-polymer",
        ),
        "C1F": ("3-CARBOXY-4-METHYL-5-PROPYL-2-FURANPROPIONIC", "non-polymer"),
        "C1E": (
            "(2Z,4E)-3-CHLORO-2-HYDROXY-6-OXO-6-PHENYLHEXA-2,4-DIENOIC ACID",
            "non-polymer",
        ),
        "C1D": (
            "1-(2-{[(6-AMINO-2-METHYLPYRIDIN-3-YL)METHYL]AMINO}ETHYL)-6-CHLORO-3-[(2,2-DIFLUORO-2-PYRIDIN-2-YLETHYL)AMINO]-1,4-DIHYDROPYRAZIN-2-OL",
            "non-polymer",
        ),
        "JH0": ("?", "non-polymer"),
        "JH3": (
            "METHYL (2E,6E)-9-[(2R)-3,3-DIMETHYLOXIRAN-2-YL]-3,7-DIMETHYLNONA-2,6-DIENOATE",
            "non-polymer",
        ),
        "C1X": (
            "(Z)-N~6~-[(4R,5S)-5-(2-CARBOXYETHYL)-4-(CARBOXYMETHYL)DIHYDRO-2H-THIOPYRAN-3(4H)-YLIDENE]-L-LYSINE",
            "peptide",
        ),
        "DI9": ("DICYCLOPENTYL PHOSPHORAMIDATE", "non-polymer"),
        "DI4": ("AC-(D)PHE-PRO-BOROHOMOLYS-OH", "non-polymer"),
        "C1R": ("4-PIPERIDINEBUTYRATE", "non-polymer"),
        "C1Q": (
            "N-{(1S,2S,3R)-1-[(ALPHA-D-GALACTOPYRANOSYLOXY)METHYL]-2,3-DIHYDROXYHEPTADECYL}-10-PHENYLDECANAMIDE",
            "non-polymer",
        ),
        "C1P": (
            "N~2~-(MORPHOLIN-4-YLCARBONYL)-N-[(3S)-1-PHENYL-5-(PHENYLSULFONYL)PENTAN-3-YL]-L-LEUCINAMIDE",
            "non-polymer",
        ),
        "ZP4": ("?", "non-polymer"),
        "DI2": ("AC-(D)PHE-PRO-BOROLYS-OH", "non-polymer"),
        "C1T": (
            "S-[(1S,6R)-3,6-DIHYDROXY-4-METHYL-2,5-DIOXOCYCLOHEX-3-EN-1-YL]-L-CYSTEINE",
            "peptide",
        ),
        "216": (
            "[4-R-(4-ALPHA,6-BETA,7-BETA]-HEXAHYDRO-5,6-DI(HYDROXY)-1,3-DI(ALLYL)-4,7-BISPHENYLMETHYL)-2H-1,3-DIAZEPINONE",
            "non-polymer",
        ),
        "217": ("(2S)-HEX-5-ENE-1,2-DIOL", "non-polymer"),
        "214": (
            "6-[4-((2R)-2-(1H-1,2,3-BENZOTRIAZOL-1-YL)-3-{4-[DIFLUORO(PHOSPHONO)METHYL]PHENYL}-2-PHENYLPROPYL)PHENYL]-2-METHYLQUINOLIN-8-YLPHOSPHONIC ACID",
            "non-polymer",
        ),
        "215": (
            "(1Z)-5-(2-{4-[2-(DIMETHYLAMINO)ETHOXY]PHENYL}-5-PYRIDIN-4-YL-1H-IMIDAZOL-4-YL)INDAN-1-ONE OXIME",
            "non-polymer",
        ),
        "HBI": ("7,8-DIHYDROBIOPTERIN", "non-polymer"),
        "213": ("?", "non-polymer"),
        "210": ("PAMIDRONATE", "non-polymer"),
        "211": ("2,2',2''-NITRILOTRIETHANOL", "non-polymer"),
        "BOP": ("1-BROMO-4-METHOXYBENZENE", "non-polymer"),
        "HBD": ("4-HYDROXYBENZAMIDE", "non-polymer"),
        "BOR": (
            "(1R)-1-AMINO-4-{[(E)-AMINO(IMINO)METHYL]AMINO}BUTYLBORONIC ACID",
            "peptide",
        ),
        "BOS": ("N-[4-(AMINOSULFONYL)PHENYL]-2-MERCAPTOBENZAMIDE", "non-polymer"),
        "BOT": ("BENZOTHIAZOLE", "non-polymer"),
        "BOU": (
            "CBAP, OPEN FORM; BENZOYL-6-AMINO PENICILLOIC ACID, OPEN FORM",
            "non-polymer",
        ),
        "218": (
            "1-[(4-AMINO-2-METHYLPYRIMIDIN-5-YL)METHYL]-3-(2-HYDROXYETHYL)-2-METHYLPYRIDINIUM",
            "non-polymer",
        ),
        "219": ("CYCLOHEXYLMETHYL 2-FORMYLPHENYL HYDROGEN PHOSPHATE", "non-polymer"),
        "NAR": ("NARINGENIN", "non-polymer"),
        "NAS": ("2-NAPHTHALENESULFONIC ACID", "non-polymer"),
        "NAP": ("NADP NICOTINAMIDE-ADENINE-DINUCLEOTIDE PHOSPHATE", "non-polymer"),
        "ILE": ("ISOLEUCINE", "peptide"),
        "BOL": ("2-IODO-N-PHENYLBENZAMIDE", "non-polymer"),
        "BOM": ("HEXADECA-10,12-DIEN-1-OL", "non-polymer"),
        "BON": (
            "N-[4-(4,4,5,5-TETRAMETHYL-1,3,2-DIOXABOROLAN-2-YL)PHENYL]SULFAMIDE",
            "non-polymer",
        ),
        "HBZ": (
            "N-{[(4-HYDROXYPHENYL)CARBONYL]CARBAMOYL}-BETA-D-GLUCOPYRANOSYLAMINE",
            "non-polymer",
        ),
        "HBU": ("4-[HYDROXY-[METHYL-PHOSPHINOYL]]-3-OXO-BUTANOIC ACID", "non-polymer"),
        "BOA": ("2-BROMO-6-HYDROXY-PURINE", "non-polymer"),
        "BOB": (
            "(S)-2-AMINO-5-((R)-1-(CARBOXYMETHYLAMINO)-3-((3S,4R)-1,4-DIHYDROXYNONAN-3-YLTHIO)-1-OXOPROPAN-2-YLAMINO)-5-OXOPENTANOIC ACID",
            "non-polymer",
        ),
        "BOC": ("TERT-BUTYL HYDROGEN CARBONATE", "non-polymer"),
        "HBQ": (
            "ISOPROPYL (2S)-2-ETHYL-7-FLUORO-3-OXO-3,4-DIHYDROQUINOXALINE-1(2H)-CARBOXYLATE",
            "non-polymer",
        ),
        "6MT": ("TRANS-N6-METHYL-DEOXY-ADENOSINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "HBS": ("S,3-HYDROXYBUTAN-2-ONE", "non-polymer"),
        "BOG": ("B-OCTYLGLUCOSIDE", "saccharide"),
        "Y3": ("4-ACETYLAMINO-5-HYDROXYNAPHTHALENE-2,7-DISULFONIC ACID", "non-polymer"),
        "Z73": (
            "2-METHOXY-6,7-DIMETHYL-9-PROPYLIMIDAZO[1,5-A]PYRIDO[3,2-E]PYRAZINE",
            "non-polymer",
        ),
        "PEZ": ("2-(PHOSPHONOOXY)BUTANOIC ACID", "non-polymer"),
        "6MR": ("N-[(4'-IODOBIPHENYL-4-YL)SULFONYL]-D-TRYPTOPHAN", "non-polymer"),
        "ILA": ("N-[ISOLEUCINYL]-N'-[ADENOSYL]-DIAMINOSUFONE", "non-polymer"),
        "PR8": (
            "5'-O-[(R)-HYDROXY{[(2S)-PYRROLIDIN-2-YLCARBONYL]OXY}PHOSPHORYL]ADENOSINE",
            "non-polymer",
        ),
        "J53": (
            "(2S)-2-{4-BUTOXY-3-[({4-[(3S,5S,7S)-TRICYCLO[3.3.1.1~3,7~]DEC-1-YL]BENZOYL}AMINO)METHYL]BENZYL}BUTANOIC ACID",
            "non-polymer",
        ),
        "6MP": ("6-METHYLPURINE", "non-polymer"),
        "J54": (
            "(2R)-3-{[(BENZYLAMINO)CARBONYL]AMINO}-2-HYDROXYPROPANOIC ACID",
            "non-polymer",
        ),
        "9DG": ("9-DEAZAGUANINE", "non-polymer"),
        "SV6": ("TELAPREVIR, BOUND FORM", "peptide"),
        "PXO": ("(1Z)-PROPANAL OXIME", "non-polymer"),
        "SV2": (
            "[{2-[(GUANINE-9-YL)METHYL]PROPANE-1,3-DIYL}BIS(OXYMETHYLENE)]BIS(PHOSPHONIC ACID) ",
            "non-polymer",
        ),
        "SV1": (
            "2-PHENYLETHYL 2-(4-AZANYL-2-METHANOYL-PHENYL)ETHANOATE",
            "non-polymer",
        ),
        "PEW": (
            "(1R)-2-{[(2-AMINOETHOXY)(HYDROXY)PHOSPHORYL]OXY}-1-[(PALMITOYLOXY)METHYL]ETHYL (9S,10S)-9,10-DIBROMOOCTADECANOATE",
            "non-polymer",
        ),
        "PFZ": (
            "1-(N-IMIDAZOLYL)-2-HYDROXY-2-(2,3-DICHLOROPHENYL)OCTANE",
            "non-polymer",
        ),
        "SVE": (
            "N-(2,6-DIETHYLPHENYL)-1-METHYL-8-({4-[(1-METHYLPIPERIDIN-4-YL)CARBAMOYL]-2-(TRIFLUOROMETHOXY)PHENYL}AMINO)-4,5-DIHYDRO-1H-PYRAZOLO[4,3-H]QUINAZOLINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "SVA": ("SERINE VANADATE", "peptide"),
        "SVZ": ("O-[(S)-HYDROXY(ISOPROPYLAMINO)PHOSPHORYL]-D-SERINE", "peptide"),
        "ILH": ("?", "non-polymer"),
        "SVX": ("O-[(R)-ETHOXY(METHYL)PHOSPHORYL]-L-SERINE", "peptide"),
        "SVY": ("O-[BIS(1-METHYLETHOXY)PHOSPHORYL]-L-SERINE", "peptide"),
        "J5Z": (
            "5-METHYL-N-[4-(TRIFLUOROMETHYL)PHENYL][1,2,4]TRIAZOLO[1,5-A]PYRIMIDIN-7-AMINE",
            "non-polymer",
        ),
        "SVW": ("O-[(R)-AMINO(HYDROXY)PHOSPHORYL]-L-SERINE", "peptide"),
        "9DA": ("9-DEAZAADENINE", "non-polymer"),
        "SVR": ("SURAMIN", "non-polymer"),
        "SVS": (
            "5'-DEOXY-5'-({[2-(2-HYDROXYPHENYL)ETHYL]SULFONYL}AMINO)ADENOSINE",
            "non-polymer",
        ),
        "BO8": ("(D-ALPHA-AMINOPIMELYLAMINO)-D-1-ETHYLBORONIC ACID", "non-polymer"),
        "BO9": (
            "N1-[(1R)-1-(DIHYDROXYBORYL)ETHYL]-N2-[(TERT-BUTOXYCARBONYL)-D-GAMMA-GLUTAMYL]-N6-[(BENZYLOXY)CARBONYL-L-LYSINAMIDE",
            "non-polymer",
        ),
        "PAW": (
            "N-(2-AMINOETHYL)-N'-{2-[(2-AMINOETHYL)AMINO]ETHYL}ETHANE-1,2-DIAMINE",
            "non-polymer",
        ),
        "21U": ("D-LEUCYL-N-(3-CHLOROBENZYL)-L-PROLINAMIDE", "non-polymer"),
        "21P": (
            "4-[2-(1,1,3,3-TETRAMETHYL-2,3-DIHYDRO-1H-1,3-BENZODISILOL-5-YL)-1,3-DIOXOLAN-2-YL]BENZOIC ACID",
            "non-polymer",
        ),
        "GXL": ("ALPHA-L-GALACTOPYRANOSE", "non-polymer"),
        "BO1": (
            "N-[4-(AMINOSULFONYL)PHENYL]-BETA-D-GLUCOPYRANOSYLAMINE",
            "non-polymer",
        ),
        "BO2": (
            "N-[(1R)-1-(DIHYDROXYBORYL)-3-METHYLBUTYL]-N-(PYRAZIN-2-YLCARBONYL)-L-PHENYLALANINAMIDE",
            "non-polymer",
        ),
        "BO3": ("BORIC ACID", "non-polymer"),
        "BO4": ("BORATE ION", "non-polymer"),
        "ILT": ("MONOIMIDAZOLE LEXITROPSIN", "non-polymer"),
        "9GP": ("N-(HYDROXYACETYL)-BETA-D-GLUCOPYRANOSYLAMINE", "non-polymer"),
        "BO7": ("2,6-DIMETHOXYBENZOIC ACID", "non-polymer"),
        "NA2": ("SODIUM ION, 2 WATER COORDINATED", "non-polymer"),
        "NA3": (
            "(3S)-3-{[(5-BROMOPYRIDIN-3-YL)CARBONYL]AMINO}-4-OXOBUTANOIC ACID",
            "non-polymer",
        ),
        "NA0": ("3-ACETYLPYRIDINE ADENINE DINUCLEOTIDE PHOSPHATE", "non-polymer"),
        "NA1": ("METHYL N-ACETYL ALLOSAMINE", "non-polymer"),
        "NA6": ("SODIUM ION, 6 WATERS COORDINATED", "non-polymer"),
        "NA7": (
            "[(2R,3R,4R,5R)-5-(6-AMINO-9H-PURIN-9-YL)-3-HYDROXY-4-(PHOSPHONOOXY)TETRAHYDROFURAN-2-YL]METHYL [(2R,3S,4S)-3,4-DIHYDROXYTETRAHYDROFURAN-2-YL]METHYL DIHYDROGEN DIPHOSPHATE",
            "non-polymer",
        ),
        "NA4": (
            "4-[5-(2-CARBOXY-1-FORMYL-ETHYLCARBAMOYL)-PYRIDIN-3-YL]-BENZOIC ACID",
            "non-polymer",
        ),
        "NA5": ("SODIUM ION, 5 WATERS COORDINATED", "non-polymer"),
        "21N": (
            "N-{[(2E)-2-{[4-(TRIFLUOROMETHYL)PHENYL]METHYLIDENE}HYDRAZINO]CARBONOTHIOYL}-BETA-D-GLUCOPYRANOSYLAMINE",
            "non-polymer",
        ),
        "21O": (
            "{(1R)-3-OXO-1-PHENYL-3-[4-(TRIFLUOROMETHYL)PHENYL]PROPYL}PROPANEDIOIC ACID",
            "non-polymer",
        ),
        "21L": (
            "METHYL 3-{4-[({[(9AS)-8-ACETYL-1,7-DIHYDROXY-3-METHOXY-9A-METHYL-9-OXO-9,9A-DIHYDRODIBENZO[B,D]FURAN-4-YL]CARBONYL}AMINO)METHYL]NAPHTHALEN-2-YL}PROPANOATE",
            "non-polymer",
        ),
        "NA9": (
            "N-(CARBOXYMETHYL)-3-CYCLOHEXYL-D-ALANYL-N-({6-[AMINO(IMINO)METHYL]PYRIDIN-3-YL}METHYL)-N~2~-METHYL-L-ALANINAMIDE",
            "non-polymer",
        ),
        "DLS": ("DI-ACETYL-LYSINE", "peptide"),
        "PEI": ("PENTANOIC ACID", "non-polymer"),
        "FLY": (
            "6-(2,4-DIFLUOROPHENOXY)-8-METHYL-2-[(1-METHYLETHYL)AMINO]PYRIDO[2,3-D]PYRIMIDIN-7(8H)-ONE",
            "non-polymer",
        ),
        "FLX": (
            "N-[(FURAN-2-YL)CARBONYL]-(S)-LEUCYL-(R)-[1-AMINO-2(1H-INDOL-3-YL)ETHYL]-PHOSPHONIC ACID",
            "non-polymer",
        ),
        "R6A": ("RUTHENIUM WIRE, 6 CARBON LINKER", "non-polymer"),
        "FLZ": (
            "8-METHYL-6-PHENOXY-2-(TETRAHYDRO-2H-PYRAN-4-YLAMINO)PYRIDO[2,3-D]PYRIMIDIN-7(8H)-ONE",
            "non-polymer",
        ),
        "LKJ": ("?", "non-polymer"),
        "9DX": (
            "(2R,3AR,7R,7AR)-2-[(2S)-2-AMINO-3-HYDROXY-3-OXO-PROPYL]-7-HYDROXY-3,3A,5,6,7,7A-HEXAHYDROFURO[4,5-B]PYRAN-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "FLQ": (
            "N-[6-(ACETYLAMINO)HEXYL]-3',6'-DIHYDROXY-3-OXO-3H-SPIRO[2-BENZOFURAN-1,9'-XANTHENE]-6-CARBOXAMIDE",
            "non-polymer",
        ),
        "FLP": ("FLURBIPROFEN", "non-polymer"),
        "FLS": (
            "6-(2,4-DIFLUOROPHENOXY)-8-METHYL-2-{[(1R)-1-METHYL-2-(METHYLSULFONYL)ETHYL]AMINO}PYRIDO[2,3-D]PYRIMIDIN-7(8H)-ONE",
            "non-polymer",
        ),
        "F68": (
            "N-{[(4-METHYLPHENYL)CARBONYL]CARBAMOYL}-BETA-D-GLUCOPYRANOSYLAMINE",
            "non-polymer",
        ),
        "FLU": ("2-(6-HYDROXY-3-OXO-3H-XANTHEN-9-YL)-BENZOIC ACID", "non-polymer"),
        "FLT": ("FLUOROMALONYL TYROSINE", "peptide"),
        "FLW": (
            "6-(2,4-DIFLUOROPHENOXY)-2-{[3-HYDROXY-1-(2-HYDROXYETHYL)PROPYL]AMINO}-8-METHYLPYRIDO[2,3-D]PYRIMIDIN-7(8H)-ONE",
            "non-polymer",
        ),
        "FLV": ("FLAVIOLIN", "non-polymer"),
        "FLI": (
            "(3R,4S,6S,8S,10R,12R,14R,16S,17E,19E,21E,23E,25E,28R)-3-HEXYL-4,6,8,10,12,14,16-HEPTAHYDROXY-17,28-DIMETHYLOXACYCLOOCTACOSA-17,19,21,23,25-PENTAEN-2-ONE",
            "non-polymer",
        ),
        "FLH": ("FORMALDEHYDE", "non-polymer"),
        "KC6": ("?", "non-polymer"),
        "FLM": ("3-FLUORO-2-METHYL-ANILINE", "non-polymer"),
        "FLL": ("OCTAHEDRAL RU-PYRIDOCARBAZOLE", "non-polymer"),
        "FLO": ("FLUORO GROUP", "non-polymer"),
        "FLN": ("2-PHENYL-4H-CHROMEN-4-ONE", "non-polymer"),
        "FLA": ("TRIFLUOROALANINE", "peptide"),
        "244": (
            "[5-HYDROXY-2-(4-HYDROXYPHENYL)-1-BENZOFURAN-7-YL]ACETONITRILE",
            "non-polymer",
        ),
        "FLB": ("1-(2-ETHOXYETHOXY)-3,5-BIS(TRIFLUOROMETHYL)BENZENE", "non-polymer"),
        "FLE": ("FUROYL-LEUCINE", "peptide"),
        "FLD": (
            "BIS-(N-ETHYLPYRIDINIUM-(3-METHOXYCARBAZOLE))HEXANE-1,6-DIAMINE",
            "non-polymer",
        ),
        "FLG": ("FLUORESCEINYLTHIOUREIDO", "non-polymer"),
        "FLF": ("2-[[3-(TRIFLUOROMETHYL)PHENYL]AMINO] BENZOIC ACID", "non-polymer"),
        "PEG": ("DI(HYDROXYETHYL)ETHER", "non-polymer"),
        "F63": ("N-METHYL-1-[5-(PYRIDIN-3-YLOXY)FURAN-2-YL]METHANAMINE", "non-polymer"),
        "9DS": ("DANSYL-L-SARCOSINE", "peptide"),
        "JHZ": ("?", "non-polymer"),
        "919": (
            "4-[4-({[3-TERT-BUTYL-1-(QUINOLIN-6-YL)-1H-PYRAZOL-5-YL]CARBAMOYL}AMINO)-3-FLUOROPHENOXY]-N-METHYLPYRIDINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "ILX": ("4,5-DIHYDROXYISOLEUCINE", "peptide"),
        "915": (
            "(2R)-2-{[(4-FLUORO-3-METHYLPHENYL)SULFONYL]AMINO}-N-HYDROXY-2-TETRAHYDRO-2H-PYRAN-4-YLACETAMIDE",
            "non-polymer",
        ),
        "914": (
            "[(2R,5R)-5-(6-AMINOPURIN-9-YL)-4-FLUORO-2,5-DIHYDROFURAN-2-YL]OXYMETHYL-[HYDROXY(PHOSPHONOOXY)PHOSPHORYL]OXY-PHOSPHINIC ACID",
            "non-polymer",
        ),
        "916": (
            "N-[(2S,3R)-4-{[(4S)-6-(2,2-DIMETHYLPROPYL)-2,2-DIMETHYL-3,4-DIHYDRO-2H-THIENO[2,3-B]PYRAN-4-YL]AMINO}-3-HYDROXY-1-PHENYLBUTAN-2-YL]ACETAMIDE",
            "non-polymer",
        ),
        "911": ("(1-HYDROXYETHANE-1,1-DIYL)BIS(PHOSPHONIC ACID)", "non-polymer"),
        "910": (
            "5-[3-(BENZYLAMINO)PHENYL]-4-BROMO-3-(CARBOXYMETHOXY)THIOPHENE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "913": ("(3-PYRAZOL-1-YLPHENYL)METHANAMINE", "non-polymer"),
        "H71": (
            "8-[(6-IODO-1,3-BENZODIOXOL-5-YL)THIO]-9-[3-(ISOPROPYLAMINO)PROPYL]-9H-PURIN-6-AMINE",
            "non-polymer",
        ),
        "H7J": (
            "N-[(1S)-1-{1-[(1R,3E)-1-ACETYLPENT-3-EN-1-YL]-1H-1,2,3-TRIAZOL-4-YL}-1,2-DIMETHYLPROPYL]BENZAMIDE",
            "non-polymer",
        ),
        "H7K": (
            "METHYL 5-(6-{[(CIS-4-HYDROXYCYCLOHEXYL)AMINO]METHYL}-2,4-DIHYDROINDENO[1,2-C]PYRAZOL-3-YL)FURAN-2-CARBOXYLATE",
            "non-polymer",
        ),
        "91B": ("1-[3-(4-METHYLPIPERAZIN-1-YL)PHENYL]METHANAMINE", "non-polymer"),
        "DLP": ("1,2-DILINOLEOYL-SN-GLYCERO-3-PHOSPHOCHOLINE", "non-polymer"),
        "91U": (
            "N-ACETYL-3-CYCLOHEXYL-D-ALANYL-N-(3-CHLOROBENZYL)-L-PROLINAMIDE",
            "peptide",
        ),
        "FL9": ("?", "non-polymer"),
        "FL8": ("6,4'-DIHYDROXY-3-METHYL-3',5'-DIBROMOFLAVONE", "non-polymer"),
        "FL1": (
            "5-{[({2-[4-(AMINOSULFONYL)PHENYL]ETHYL}AMINO)CARBONOTHIOYL]AMINO}-2-(6-HYDROXY-3-OXO-3H-XANTHEN-9-YL)BENZOIC ACID",
            "non-polymer",
        ),
        "FL2": ("FLURBIPROFEN METHYL ESTER", "non-polymer"),
        "FL4": (
            "6-(2,4-DIFLUOROPHENOXY)-3-(2-METHYLPHENYL)-1H-PYRAZOLO[3,4-D]PYRIMIDINE",
            "non-polymer",
        ),
        "FL7": ("FLURAZEPAM", "non-polymer"),
        "KCT": (
            "(5S)-7-OXO-6,7,12,13-TETRAHYDRO-5H-INDOLO[2,3-A]PYRROLO[3,4-C]CARBAZOLE-5-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "KCQ": ("(3S)-3-[(2S)-2-AMINO-3-OXOBUTYL]PYRROLIDIN-2-ONE", "non-polymer"),
        "KCP": (
            "2-[CARBOXY-(2-THIOPHEN-2-YL-ACETYLAMINO)-METHYL]-5-METHYL-3,6-DIHYDRO-2H-[1,3]THIAZINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "KCS": ("CARBONOTRITHIOIC ACID", "non-polymer"),
        "R64": (
            "(1-METHYL-1H-IMIDAZOL-2-YL)-(3-METHYL-4-{3-[(PYRIDIN-3-YLMETHYL)-AMINO]-PROPOXY}-BENZOFURAN-2-YL)-METHANONE",
            "non-polymer",
        ),
        "LK7": (
            "(1R,4S,7AS)-1-(1-FORMYLPROP-1-EN-1-YL)-4-METHOXY-2,4,5,6,7,7A-HEXAHYDRO-1H-ISOINDOLE-3-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "LK4": (
            "N-({6-[(4-CYANOBENZYL)OXY]NAPHTHALEN-2-YL}SULFONYL)-D-GLUTAMIC ACID",
            "non-polymer",
        ),
        "LK5": (
            "(1S,4R,7AR)-4-BUTOXY-1-[(1R)-1-FORMYLPROPYL]-2,4,5,6,7,7A-HEXAHYDRO-1H-ISOINDOLE-3-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "LK2": ("N-[(6-BUTOXYNAPHTHALEN-2-YL)SULFONYL]-D-GLUTAMIC ACID", "non-polymer"),
        "KCX": ("LYSINE NZ-CARBOXYLIC ACID", "peptide"),
        "LK0": ("?", "peptide"),
        "LK1": ("N-[(6-BUTOXYNAPHTHALEN-2-YL)SULFONYL]-L-GLUTAMIC ACID", "non-polymer"),
        "P4O": (
            "2-(2-QUINOLIN-3-YLPYRIDIN-4-YL)-1,5,6,7-TETRAHYDRO-4H-PYRROLO[3,2-C]PYRIDIN-4-ONE",
            "non-polymer",
        ),
        "P4L": (
            "2-PYRIDIN-2-YL-7,8-DIHYDRO-5H-THIOPYRANO[4,3-D]PYRIMIDIN-4-OL",
            "non-polymer",
        ),
        "P4G": ("1-ETHOXY-2-(2-ETHOXYETHOXY)ETHANE", "non-polymer"),
        "P4F": ("5,5-DIFLUORO-4-OXO-5-PHOSPHONO-D-NORVALINE", "peptide"),
        "P4E": ("(2S,4E)-2-AMINO-5-PHENYLPENT-4-ENOIC ACID", "peptide"),
        "P4D": (
            "(2R)-2-{[(S)-[(1R)-1-AMINO-2-(4-HYDROXYPHENYL)ETHYL](HYDROXY)PHOSPHORYL]METHYL}BUTANEDIOIC ACID",
            "peptide",
        ),
        "P4C": ("O-ACETALDEHYDYL-HEXAETHYLENE GLYCOL", "non-polymer"),
        "P4A": (
            "4-[4-(4-METHOXYPHENYL)-5-METHYL-1H-PYRAZOL-3-YL]BENZENE-1,3-DIOL",
            "non-polymer",
        ),
        "Y01": ("CHOLESTEROL HEMISUCCINATE", "non-polymer"),
        "OZ2": (
            "(2R)-3-{[(R)-{[(2S)-2,3-DIHYDROXYPROPYL]OXY}(HYDROXY)PHOSPHORYL]OXY}-2-[(6Z)-TRIDEC-6-ENOYLOXY]PROPYL (9Z)-OCTADEC-9-ENOATE",
            "non-polymer",
        ),
        "P4Y": (
            "5,6-DIHYDROXY-2-[(2-PHENYL-1H-INDOL-3-YL)METHYL]PYRIMIDINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "P4T": ("2-PHENYLAMINO-4-METHYL-5-ACETYL THIAZOLE", "non-polymer"),
        "NDP": (
            "NADPH DIHYDRO-NICOTINAMIDE-ADENINE-DINUCLEOTIDE PHOSPHATE",
            "non-polymer",
        ),
        "P4P": ("TETRAPHENYLPHOSPHONIUM", "non-polymer"),
        "NH": ("AMINO SUBSTITUTION GROUP", "non-polymer"),
        "NI": ("NICKEL (II) ION", "non-polymer"),
        "5RG": ("2-AMINO-5-FLUOROBENZOIC ACID", "non-polymer"),
        "TF1": (
            "4-({(2R,5S)-2,5-DIMETHYL-4-[(2R)-3,3,3-TRIFLUORO-2-HYDROXY-2-METHYLPROPANOYL]PIPERAZIN-1-YL}CARBONYL)BENZONITRILE",
            "non-polymer",
        ),
        "NO": ("NITRIC OXIDE", "non-polymer"),
        "5RM": (
            "(5R)-5-(4-METHOXY-3-PROPOXYPHENYL)-5-METHYL-1,3-OXAZOLIDIN-2-ONE",
            "non-polymer",
        ),
        "NC": ("METHYLAMINE GROUP", "non-polymer"),
        "NF": ("N,N'-(3,7-DIAZANONYLENE)-BIS-NAPHTHALIMIDE", "non-polymer"),
        "DLU": (
            "(4R,12AS)-N-(2,4-DIFLUOROBENZYL)-7-HYDROXY-4-METHYL-6,8-DIOXO-3,4,6,8,12,12A-HEXAHYDRO-2H-PYRIDO[1',2':4,5]PYRAZINO[2,1-B][1,3]OXAZINE-9-CARBOXAMIDE",
            "non-polymer",
        ),
        "TCH": ("TAUROCHOLIC ACID", "non-polymer"),
        "NLE": ("NORLEUCINE", "peptide"),
        "5RP": ("RIBULOSE-5-PHOSPHATE", "saccharide"),
        "SS4": ("3-PYRIDIN-4-YL-2,4-DIHYDRO-INDENO[1,2-.C.] PYRAZOLE", "non-polymer"),
        "NP": (
            "4-HYDROXY-3-NITROPHENYLACETYL-EPSILON-AMINOCAPROIC ACID",
            "non-polymer",
        ),
        "NQ": ("2-HYDROXYNAPHTHOQUINONE", "non-polymer"),
        "NT": ("NETROPSIN", "non-polymer"),
        "RWF": ("R-WARFARIN", "non-polymer"),
        "63F": (
            "N-CYCLOPROPYL-6-[(3R)-3-{[4-(TRIFLUOROMETHOXY)BENZYL]CARBAMOYL}-4-{[4-(TRIFLUOROMETHOXY)PHENYL]SULFONYL}PIPERAZIN-1-YL]PYRIDAZINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "YIV": ("N-BUTYLFORMAMIDE", "non-polymer"),
        "63G": (
            "(6S)-3-(2-DEOXY-5-O-PHOSPHONO-BETA-D-ERYTHRO-PENTOFURANOSYL)-6-HYDROXY-5,6,7,8-TETRAHYDROPYRIMIDO[1,2-A]PURIN-10(3H)-ONE",
            "nucleic-acid",
        ),
        "RM8": (
            "2-METHYL-3-{[4-(PYRIDIN-2-YL)PIPERAZIN-1-YL]METHYL}-1H-INDOLE",
            "non-polymer",
        ),
        "SS0": (
            "(8R)-8-[(DIMETHYLAMINO)METHYL]-1-[3-(DIMETHYLAMINO)PROPYL]-1,7,8,9-TETRAHYDROCHROMENO[5,6-D]IMIDAZOL-2-AMINE",
            "non-polymer",
        ),
        "63H": (
            "(6R)-3-(2-DEOXY-5-O-PHOSPHONO-BETA-D-ERYTHRO-PENTOFURANOSYL)-6-HYDROXY-5,6,7,8-TETRAHYDROPYRIMIDO[1,2-A]PURIN-10(3H)-ONE",
            "nucleic-acid",
        ),
        "5R1": (
            "4-[3-AMINO-6-(3,4,5-TRIMETHOXYPHENYL)PYRAZIN-2-YL]-2-METHOXYBENZOIC ACID",
            "non-polymer",
        ),
        "393": (
            "(5-CHLORO-2-{[(3-NITROBENZYL)AMINO]CARBONYL}PHENOXY)ACETIC ACID",
            "non-polymer",
        ),
        "392": (
            "1-{3-[(3,5-DICHLOROBENZYL)AMINO]PROPYL}-3-(2-HYDROXYPHENYL)UREA",
            "non-polymer",
        ),
        "SX": ("SULFUR OXIDE", "non-polymer"),
        "NLG": ("N-ACETYL-L-GLUTAMATE", "non-polymer"),
        "KGY": (
            "4-({[4-(BUT-2-YN-1-YLOXY)PHENYL]SULFONYL}METHYL)-1-[(3,5-DIMETHYLISOXAZOL-4-YL)SULFONYL]-N-HYDROXYPIPERIDINE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "OZT": ("(4S,5R)-5-METHYL-2-OXO-1,3-OXAZOLIDINE-4-CARBOXYLIC ACID", "peptide"),
        "OZR": (
            "N-[(E)-3-[(2R,3S,4S)-3,4-DIHYDROXYOXOLAN-2-YL]PROP-2-ENYL]-2,3-DIHYDROXY-5-NITRO-BENZAMIDE",
            "non-polymer",
        ),
        "OZS": (
            "N-[(E)-3-[(2R,3S,4R,5R)-3,4-DIHYDROXY-5-[4-(TRIFLUOROMETHYL)IMIDAZOL-1-YL]OXOLAN-2-YL]PROP-2-ENYL]-2,3-DIHYDROXY-5-NITRO-BENZAMIDE",
            "non-polymer",
        ),
        "P49": (
            "1-METHYL-8-(PHENYLAMINO)-4,5-DIHYDRO-1H-PYRAZOLO[4,3-H]QUINAZOLINE-3-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "P48": (
            "N,1,4,4-TETRAMETHYL-8-{[4-(4-METHYLPIPERAZIN-1-YL)PHENYL]AMINO}-4,5-DIHYDRO-1H-PYRAZOLO[4,3-H]QUINAZOLINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "NDU": ("?", "nucleic-acid"),
        "397": ("2-(4-HYDROXY-PHENYL)BENZOFURAN-5-OL", "non-polymer"),
        "P45": (
            "2-(4-CHLOROBENZYLAMINO)-4-(PHENYLAMINO)PYRAZOLO[1,5-A][1,3,5]TRIAZINE-8-CARBONITRILE",
            "non-polymer",
        ),
        "P44": (
            "2-(CYCLOHEXYLMETHYLAMINO)-4-(PHENYLAMINO)PYRAZOLO[1,5-A][1,3,5]TRIAZINE-8-CARBONITRILE",
            "non-polymer",
        ),
        "OZZ": (
            "N-[(E)-3-[(2R,3S,4R,5R)-3,4-DIHYDROXY-5-(4-OXOPYRIDIN-1-YL)OXOLAN-2-YL]PROP-2-ENYL]-2,3-DIHYDROXY-5-NITRO-BENZAMIDE",
            "non-polymer",
        ),
        "P42": (
            "(2S)-2-{[(1R)-1-HYDROXYHEXADECYL]OXY}-3-{[(1R)-1-HYDROXYOCTADECYL]OXY}PROPYL 2-(TRIMETHYLAMMONIO)ETHYL PHOSPHATE",
            "non-polymer",
        ),
        "P41": (
            "6-[4-(2-FLUOROPHENYL)-1,3-OXAZOL-5-YL]-N-(1-METHYLETHYL)-1,3-BENZOTHIAZOL-2-AMINE",
            "non-polymer",
        ),
        "396": (
            "4-(2-CHLOROPHENYL)-9-HYDROXY-6-METHYL-1,3-DIOXO-N-(2-PYRROLIDIN-1-YLETHYL)PYRROLO[3,4-G]CARBAZOLE-8-CARBOXAMIDE",
            "non-polymer",
        ),
        "AZH": ("AZIDOHOMOALANINE", "peptide"),
        "AZI": ("AZIDE ION", "non-polymer"),
        "AZK": ("(S)-2-AMINO-6-AZIDOHEXANOIC ACID", "peptide"),
        "AZL": ("1-ETHOXYCARBONYL-D-PHE-PRO-2(4-AMINOBUTYL)HYDRAZINE", "peptide"),
        "AZM": ("5-ACETAMIDO-1,3,4-THIADIAZOLE-2-SULFONAMIDE", "non-polymer"),
        "AZN": ("ALIZARIN RED", "non-polymer"),
        "2RH": ("(2R)-2-HYDROXY-4-METHYLPENTANOIC ACID", "non-polymer"),
        "2RG": (
            "(2S,3R,4S)-4-({(3S,5S)-5-[(3-CARBOXYPHENYL)CARBAMOYL]PYRROLIDIN-3-YL}SULFANYL)-2-[(1S,2R)-1-FORMYL-2-HYDROXYPROPYL]-3-METHYL-3,4-DIHYDRO-2H-PYRROLE-5-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "AZA": ("8-AZAXANTHINE", "non-polymer"),
        "AZB": ("4-{(E)-[4-(PROPAN-2-YL)PHENYL]DIAZENYL}PHENOL", "non-polymer"),
        "AZC": ("?", "non-polymer"),
        "AZD": ("3'-AZIDO-3'-DEOXYTHYMIDINE-5'-DIPHOSPHATE", "non-polymer"),
        "AZE": ("ALL-TRANS AXEROPHTHENE", "non-polymer"),
        "2RA": ("3-AMINO-D-ALANINE", "peptide"),
        "AZG": ("5-AMINO-1H-[1,2,3]TRIAZOLO[4,5-D]PYRIMIDIN-7-OL", "non-polymer"),
        "AZX": (
            "4-[(3-CHLORO-4-{[(2R)-3,3,3-TRIFLUORO-2-HYDROXY-2-METHYLPROPANOYL]AMINO}PHENYL)SULFONYL]-N,N-DIMETHYLBENZAMIDE",
            "non-polymer",
        ),
        "AZY": ("3-AZIDO-L-TYROSINE", "peptide"),
        "AZZ": ("3'-AZIDO-3'-DEOXYTHYMIDINE", "non-polymer"),
        "LW4": (
            "3-(1H-INDOL-3-YL)-4-[2-(4-METHYLPIPERAZIN-1-YL)QUINAZOLIN-4-YL]-1H-PYRROLE-2,5-DIONE",
            "non-polymer",
        ),
        "AZP": (
            "(5S,8S,14R)-ETHYL 11-(3-AMINO-3-OXOPROPYL)-8-BENZYL-14-HYDROXY-5-ISOBUTYL-3,6,9,12-TETRAOXO-1-PHENYL-2-OXA-4,7,10,11-TETRAAZAPENTADECAN-15-OATE",
            "non-polymer",
        ),
        "AZQ": ("AZAPROPAZONE", "non-polymer"),
        "AZR": (
            "2-({[(1Z)-1-(2-AMINO-1,3-THIAZOL-4-YL)-2-OXO-2-{[(2S,3S)-1-OXO-3-(SULFOAMINO)BUTAN-2-YL]AMINO}ETHYLIDENE]AMINO}OXY)-2-METHYLPROPANOIC ACID",
            "non-polymer",
        ),
        "AZS": ("O-DIAZOACETYL-L-SERINE", "peptide"),
        "AZT": ("3'-AZIDO-3'-DEOXYTHYMIDINE-5'-TRIPHOSPHATE", "non-polymer"),
        "AZV": (
            "N-(4-{[({(2R)-3-[(3R)-3-AMINO-4-(2,4,5-TRIFLUOROPHENYL)BUTANOYL]-1,3-THIAZOLIDIN-2-YL}CARBONYL)AMINO]METHYL}PHENYL)-D-VALINE",
            "non-polymer",
        ),
        "3HP": ("3-HYDROXYPHENYLACETATE", "non-polymer"),
        "3HR": ("(3R)-3-HYDROXYBUTANOIC ACID", "non-polymer"),
        "3HT": (
            "5-[1-(4-METHOXYPHENYL)-1H-BENZIMIDAZOL-6-YL]-1,3,4-OXADIAZOLE-2(3H)-THIONE",
            "non-polymer",
        ),
        "3HU": (
            "(S)-1-(2'-AMINO-2'-CARBOXYETHYL)-3-[(2-CARBOXYTHIEN-3-YL)METHYL]THIENO[3,4-D]PYRIMIDIN-2,4-DIONE",
            "non-polymer",
        ),
        "3HX": (
            "(4S)-N-{4-AMINO-5-[(2-AMINOETHYL)(HYDROXYAMINO]-PENTYL}-N'-NITROGUANIDINE",
            "non-polymer",
        ),
        "9IN": (
            "N-(3-FUROYL)-D-VALYL-L-VALYL-N~1~-((1R,2Z)-4-ETHOXY-4-OXO-1-{[(3S)-2-OXOPYRROLIDIN-3-YL]METHYL}BUT-2-ENYL)-D-LEUCINAMIDE",
            "non-polymer",
        ),
        "3HA": ("3-HYDROXYANTHRANILIC ACID", "non-polymer"),
        "3HB": ("3-HYDROXYBENZOIC ACID", "non-polymer"),
        "3HC": ("3-HYDROXYBUTANOYL-COENZYME A", "non-polymer"),
        "3HD": ("3-O-METHYL-O-ALPHA-D-MANNOPYRANOSYL", "non-polymer"),
        "9IP": (
            "N~3~-[3-(5-METHOXYPYRIDIN-3-YL)BENZYL]PYRIDINE-2,3-DIAMINE",
            "non-polymer",
        ),
        "3HF": (
            "N~3~-{(1S,2R)-1-BENZYL-2-HYDROXY-3-[(3-METHOXYBENZYL)AMINO]PROPYL}-N~1~,N~1~-DIBUTYL-1H-INDOLE-1,3-DICARBOXAMIDE",
            "non-polymer",
        ),
        "3HG": ("3-HYDROXYPENTANEDIOIC ACID", "non-polymer"),
        "3HH": (
            "N~1~-BUTYL-5-CYANO-N~3~-{(1S,2R)-1-(3,5-DIFLUOROBENZYL)-2-HYDROXY-3-[(3-METHOXYBENZYL)AMINO]PROPYL}-N~1~-METHYL-1H-INDOLE-1,3-DICARBOXAMIDE",
            "non-polymer",
        ),
        "3HI": (
            "(3R,5R)-7-[2-(4-FLUOROPHENYL)-4-[(2-HYDROXYPHENYL)CARBAMOYL]-5-(1-METHYLETHYL)-3-PHENYL-1H-PYRROL-1-YL]-3,5-DIHYDROXYHEPTANOIC ACID",
            "non-polymer",
        ),
        "3HK": (
            "4-{7-METHOXY-6-[3-(MORPHOLIN-4-YL)PROPOXY]-1,4-DIHYDROINDENO[1,2-C]PYRAZOL-3-YL}BENZONITRILE",
            "non-polymer",
        ),
        "3HL": ("(3S)-3-HYDROXYBUTANOIC ACID", "non-polymer"),
        "3HM": ("5-HYDROXY-6-METHYLPYRIDINE-3-CARBOXYLIC ACID", "non-polymer"),
        "3HO": ("(3S)-3-HYDROXYOCTANOIC ACID", "non-polymer"),
        "3H1": (
            "3-CHLORO-4,6-DIHYDROXY-2-METHYL-5-{(2E,4E)-3-METHYL-5-[(1R,2R,6R)-1,2,6-TRIMETHYL-3-OXOCYCLOHEXYL]PENTA-2,4-DIEN-1-YL}BENZALDEHYDE",
            "non-polymer",
        ),
        "YIR": (
            "(E)-N-(5-(2-CARBAMIMIDOYLHYDRAZONO)-5,6,7,8-TETRAHYDRONAPHTHALEN-2-YL)-7-NITRO-1H-INDOLE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "NLP": ("(1-AMINO-PENTYL)-PHOSPHONIC ACID", "peptide"),
        "Z16": (
            "(1S)-1,5-ANHYDRO-1-(3-CHLORO-2,5-DIHYDROXYPHENYL)-D-GLUCITOL",
            "non-polymer",
        ),
        "3H9": ("(S)-3-HYDROXYHEXANOYL-COA", "non-polymer"),
        "Z14": (
            "4-(4-AMINOPIPERIDIN-1-YL)-2-{[(2-AMINOPYRIMIDIN-5-YL)METHYL]AMINO}-5-NITROBENZAMIDE",
            "non-polymer",
        ),
        "Z15": (
            "(1S)-1,5-ANHYDRO-1-(4-CHLORO-2,5-DIHYDROXYPHENYL)-D-GLUCITOL",
            "non-polymer",
        ),
        "Z12": (
            "2,4-DICHLORO-N-[3,5-DICHLORO-4-(QUINOLIN-3-YLOXY)PHENYL]BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "Z10": (
            "[(1S)-1-[(2-FLUORO-6-PHENYL-PHENYL)CARBONYLAMINO]ETHYL]-TRIHYDROXY-BORON",
            "non-polymer",
        ),
        "HMY": (
            "(1AR,8S,13S,14S,15AR)-5,13,14-TRIHYDROXY-3-METHOXY-8-METHYL-8,9,13,14,15,15A-HEXAHYDRO-6H-OXIRENO[K][2]BENZOXACYCLOTETRADECINE-6,12(1AH)-DIONE",
            "non-polymer",
        ),
        "SEB": ("O-BENZYLSULFONYL-SERINE", "peptide"),
        "A4A": ("[1,3]BENZODIOXOLO[6,5,4-CD]BENZO[F]INDOL-5(6H)-ONE", "non-polymer"),
        "HMT": (
            "(3BETA)-O~3~-[(2R)-2,6-DIHYDROXY-2-(2-METHOXY-2-OXOETHYL)-6-METHYLHEPTANOYL]CEPHALOTAXINE",
            "non-polymer",
        ),
        "HMU": ("5-HYDROXYMETHYL URACIL", "non-polymer"),
        "AZ0": (
            "ETHYL (5S,8S,14S)-14-HYDROXY-5,8,11-TRIS(2-METHYLPROPYL)-3,6,9,12-TETRAOXO-1-PHENYL-2-OXA-4,7,10,11-TETRAAZAPENTADECAN-15-OATE",
            "peptide",
        ),
        "AZ1": ("AZELAIC ACID", "non-polymer"),
        "AZ2": (
            "(2S)-2-ETHOXY-3-[4-(2-{4-[(METHYLSULFONYL)OXY]PHENYL}ETHOXY)PHENYL]PROPANOIC ACID",
            "non-polymer",
        ),
        "AZ3": (
            "4,4'-[HEPTANE-1,7-DIYLBIS(OXY)]DIBENZENECARBOXIMIDAMIDE",
            "non-polymer",
        ),
        "A4L": (
            "9-[(5-(ACETYLAMINO)-6-{[(1S,4R)-8-AMINO-4-[((2R)-6-AMINO-2-{2-[(1S)-5-AMINO-1-FORMYLPENTYL]HYDRAZINO}HEXANOYL)AMINO]-1-(4-AMINOBUTYL)-2,3-DIOXOOCTYL]AMINO}-6-OXOHEXYL)AMINO]-6-CHLORO-2-METHOXYACRIDINIUM",
            "non-polymer",
        ),
        "AZ7": (
            "(3S)-3-(4-CHLOROPHENYL)-4-(5,7-DICHLORO-1H-BENZIMIDAZOL-2-YL)BUTANOIC ACID",
            "non-polymer",
        ),
        "0XZ": (
            "4-AMINO-N-[(1S)-1-(4-CHLOROPHENYL)-3-HYDROXYPROPYL]-1-(7H-PYRROLO[2,3-D]PYRIMIDIN-4-YL)PIPERIDINE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "0XY": (
            "N-[(2S,3R,4R,5S,6S)-6-FLUORO-2,4,5-TRIHYDROXY-6-(HYDROXYMETHYL)TETRAHYDRO-2H-PYRAN-3-YL]ACETAMIDE",
            "saccharide",
        ),
        "I3P": ("D-MYO-INOSITOL-1,4,5-TRIPHOSPHATE", "non-polymer"),
        "WAY": (
            "N-HYDROXY-2-[(4-METHOXY-BENZENESULFONYL)-PYRIDIN-3-YLMETHYL-AMINO]-3-METHYL-BENZAMIDE",
            "non-polymer",
        ),
        "0XS": (
            "2-AMINO-6-BROMO-7-(TRIFLUOROMETHYL)-1,8-NAPHTHYRIDINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "0XR": ("ETHYL (2E)-3-(3,4-DIHYDROXYPHENYL)PROP-2-ENOATE", "non-polymer"),
        "0XW": (
            "(2S,4R)-4-HYDROXY-1,1-DIMETHYLPYRROLIDINIUM-2-CARBOXYLATE",
            "non-polymer",
        ),
        "0XV": (
            "(4-HYDROXY-3,5-DIMETHYLPHENYL)(2-METHYL-1-BENZOFURAN-3-YL)METHANONE",
            "non-polymer",
        ),
        "0XU": (
            "(2S,5S,6E)-2,5-DIAMINO-6-[(3S,4R,5R)-5-(6-AMINO-9H-PURIN-9-YL)-3,4-DIHYDROXYDIHYDROFURAN-2(3H)-YLIDENE]HEXANOIC ACID",
            "non-polymer",
        ),
        "0XT": ("7-METHOXY-6-METHYLPTERIDINE-2,4-DIAMINE", "non-polymer"),
        "0XJ": (
            "(6ALPHA,8ALPHA)-6-(PENT-2-YN-1-YLOXY)ANDROSTA-1,4-DIENE-3,17-DIONE",
            "non-polymer",
        ),
        "P8X": ("DESMETHYL-HBA", "non-polymer"),
        "0XB": (
            "N-(3-FLUOROPHENYL)-4-NITRO-2,1,3-BENZOXADIAZOL-5-AMINE",
            "non-polymer",
        ),
        "0XE": (
            "5,6-DIHYDROXY-4-OXO-2-PHENYL-4H-CHROMEN-7-YL BETA-D-GLUCOPYRANOSIDURONIC ACID",
            "non-polymer",
        ),
        "7CP": (
            "(4R)-7,8-DICHLORO-1',9-DIMETHYL-1-OXO-1,2,4,9-TETRAHYDROSPIRO[BETA-CARBOLINE-3,4'-PIPERIDINE]-4-CARBONITRILE",
            "non-polymer",
        ),
        "7CS": (
            "(3-ENDO)-8-METHYL-8-AZABICYCLO[3.2.1]OCT-3-YL 1H-PYRROLO[2,3-B]PYRIDINE-3-CARBOXYLATE",
            "non-polymer",
        ),
        "1FX": (
            "(2S)-3-CYCLOHEXYL-2-(6-FLUORO-2,4-DIOXO-1,4-DIHYDROQUINAZOLIN-3(2H)-YL)-N-(1,3-THIAZOL-2-YL)PROPANAMIDE",
            "non-polymer",
        ),
        "1FY": (
            "(2S)-3-CYCLOHEXYL-2-[4-(METHYLSULFONYL)-2-OXOPIPERAZIN-1-YL]-N-(1,3-THIAZOL-2-YL)PROPANAMIDE",
            "non-polymer",
        ),
        "HMM": (
            "5-HYDROXYMETHYL-2-METHYL-5,8-DIHYDRO-[1,2,4]TRIAZOLO[1,2-A]PYRIDAZINE-1,3-DIONE",
            "non-polymer",
        ),
        "1FV": (
            "N-{3-[(5-CYCLOPROPYL-2-{[3-(MORPHOLIN-4-YLMETHYL)PHENYL]AMINO}PYRIMIDIN-4-YL)AMINO]PROPYL}CYCLOBUTANECARBOXAMIDE",
            "non-polymer",
        ),
        "1FW": (
            "(2S)-3-CYCLOHEXYL-2-(6-FLUORO-4-OXOQUINAZOLIN-3(4H)-YL)-N-(1,3-THIAZOL-2-YL)PROPANAMIDE",
            "non-polymer",
        ),
        "1FP": ("PIPERAZINE-1-CARBALDEHYDE", "non-polymer"),
        "NLC": (
            "2-(ACETYLAMINO)-2-DEOXY-4-O-BETA-D-GALACTOPYRANOSYL-ALPHA-D-GLUCOPYRANOSE",
            "non-polymer",
        ),
        "1FR": (
            "3-[3-CHLORO-5-(5-{[(1S)-1-PHENYLETHYL]AMINO}ISOXAZOLO[5,4-C]PYRIDIN-3-YL)PHENYL]PROPAN-1-OL",
            "non-polymer",
        ),
        "1FL": ("5-(2,4-DIFLUOROPHENYL)-2-HYDROXY-BENZOIC ACID", "non-polymer"),
        "1FM": (
            "(3S,5Z,8S,9S,11E)-8,9,16-TRIHYDROXY-14-METHOXY-3-METHYL-3,4,9,10-TETRAHYDRO-1H-2-BENZOXACYCLOTETRADECINE-1,7(8H)-DIONE",
            "non-polymer",
        ),
        "1FN": (
            "1-(4-FLUOROPHENYL)-N-[3-FLUORO-4-(1H-PYRROLO[2,3-B]PYRIDIN-4-YLOXY)PHENYL]-2-OXO-1,2-DIHYDROPYRIDINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "1FH": ("12-PHENYLHEME", "non-polymer"),
        "1FK": ("1-[7-AMINO-1-(PYRIMIDIN-2-YL)INDOLIZIN-3-YL]ETHANONE", "non-polymer"),
        "1FD": ("1-[4-(TRIFLUOROMETHOXY)PHENYL]METHANAMINE", "non-polymer"),
        "1FE": ("N-(2-OXO-2,3-DIHYDRO-1H-BENZIMIDAZOL-5-YL)GLYCINAMIDE", "non-polymer"),
        "7CK": (
            "7-CARBOXY-5-HYDROXY-12,13-DIHYDRO-6H-INDOLO[2,3-A]PYRROLO[3,4-C]CARBAZOLE",
            "non-polymer",
        ),
        "1FG": (
            "2-({[2-(3,4-DIHYDROQUINOLIN-1(2H)-YL)-2-OXOETHYL](METHYL)AMINO}METHYL)QUINAZOLIN-4(1H)-ONE",
            "non-polymer",
        ),
        "7CL": ("1-CHLOROPENTANE", "non-polymer"),
        "1FC": (
            "4-AMINO-1-(2-DEOXY-2-FLUORO-5-O-PHOSPHONO-BETA-D-ARABINOFURANOSYL)-2-OXO-1,2-DIHYDROPYRIMIDINE-5-CARBOXYLIC ACID",
            "nucleic-acid",
        ),
        "7C1": (
            "4-(4-METHYL-1,3-DIOXO-1,3-DIHYDRO-2H-ISOINDOL-2-YL)BENZENECARBOXIMIDAMIDE",
            "non-polymer",
        ),
        "P8P": (
            "(2R)-3-{[(R)-HYDROXY(PHOSPHONOOXY)PHOSPHORYL]OXY}PROPANE-1,2-DIYL DIOCTANOATE",
            "non-polymer",
        ),
        "1F8": (
            "2-METHYL-N-(2-SULFANYLETHYL)-1-BENZOFURAN-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "A4P": ("6-(ADENOSINE TETRAPHOSPHATE-METHYL)-7,8-DIHYDROPTERIN", "non-polymer"),
        "1F4": ("?", "non-polymer"),
        "1F6": (
            "(3Z)-3-{[(2-METHOXYETHYL)AMINO]METHYLIDENE}-1-METHYL-5-PHENYL-7-(TRIFLUOROMETHYL)-1H-1,5-BENZODIAZEPINE-2,4(3H,5H)-DIONE",
            "non-polymer",
        ),
        "HMK": (
            "(6AR,12AR)-6H-[1,3]DIOXOLO[5,6][1]BENZOFURO[3,2-C]CHROMENE-3,6A(12AH)-DIOL",
            "non-polymer",
        ),
        "1F0": (
            "[(4S,5R)-2-(4-TERT-BUTYL-2-ETHOXYPHENYL)-4,5-BIS(4-CHLOROPHENYL)-4,5-DIMETHYL-4,5-DIHYDRO-1H-IMIDAZOL-1-YL]{4-[3-(METHYLSULFONYL)PROPYL]PIPERAZIN-1-YL}METHANONE",
            "non-polymer",
        ),
        "1F1": ("1H-INDOLE-6-CARBOXYLIC ACID", "non-polymer"),
        "1F2": ("?", "non-polymer"),
        "1F3": ("?", "non-polymer"),
        "NLI": (
            "N-[2-(CARBAMIMIDAMIDOOXY)ETHYL]-2-{3-[(2,2-DIFLUORO-2-PHENYLETHYL)AMINO]-6-METHYL-2-OXOPYRAZIN-1(2H)-YL}ACETAMIDE",
            "non-polymer",
        ),
        "OEX": ("CA-MN4-O5 CLUSTER", "non-polymer"),
        "NLH": ("(2R,5S)-5-AMINO-2-BUTYL-4,4-DIHYDROXYNONANOIC ACID", "non-polymer"),
        "PE8": ("3,6,9,12,15,18,21-HEPTAOXATRICOSANE-1,23-DIOL", "non-polymer"),
        "NLK": ("(2R,5S)-5-AMINO-2-BUTYL-4-OXONONANOIC ACID", "non-polymer"),
        "NLO": ("O-METHYL-L-NORLEUCINE", "peptide"),
        "HMG": ("3-HYDROXY-3-METHYLGLUTARYL-COENZYME A", "non-polymer"),
        "0X3": (
            "N-(3-CHLORO-5-FLUOROPHENYL)-4-NITRO-2,1,3-BENZOXADIAZOL-5-AMINE",
            "non-polymer",
        ),
        "0X1": (
            "(5-HYDROXY-6-METHYL-4-{[(1-OXO-7-PHENOXY-1,2-DIHYDRO[1,2,4]TRIAZOLO[4,3-A]QUINOLIN-4-YL)AMINO]METHYL}PYRIDIN-3-YL)METHYL DIHYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "F6H": (
            "3,7-ANHYDRO-1,5,6-TRIDEOXY-5-FLUORO-6-(5-METHYL-2,4-DIOXO-3,4-DIHYDROPYRIMIDIN-1(2H)-YL)-2-O-PHOSPHONO-D-GLYCERO-L-ALLO-HEPTITOL",
            "nucleic-acid",
        ),
        "HMA": ("HYDROXYAMINOALANINE", "non-polymer"),
        "OES": ("N-OCTYL-2-HYDROXYETHYL SULFOXIDE", "non-polymer"),
        "I1H": (
            "3-[2-(2,4-DICHLOROPHENYL)ETHOXY]-4-METHOXY-N-[(1-PYRIDIN-4-YLPIPERIDIN-4-YL)METHYL]BENZAMIDE",
            "non-polymer",
        ),
        "F6F": (
            "2-{[4-(TRIFLUOROMETHOXY)BENZOYL]AMINO}ETHYL DIHYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "HMC": ("5-HYDROXYMETHYL-CHONDURITOL", "non-polymer"),
        "NLN": ("NORLEUCINE AMIDE", "peptide"),
        "4AT": ("5-OXOHEXANOIC ACID", "non-polymer"),
        "4AU": ("1H-BENZIMIDAZOL-2-YL(1H-PYRROL-2-YL)METHANONE", "non-polymer"),
        "PE7": ("1-DEOXY-1-THIO-HEPTAETHYLENE GLYCOL", "non-polymer"),
        "4AP": ("4-AMINOPYRIDINE", "non-polymer"),
        "198": ("R-BICALUTAMIDE", "non-polymer"),
        "4AR": (
            "AMINO({(4S)-4-AMINO-6-[(2,6-DIMETHYLBENZOYL)OXY]-5-OXOHEXYL}AMINO)METHANIMINIUM",
            "peptide",
        ),
        "195": (
            "4-{2,4-BIS[(3-NITROBENZOYL)AMINO]PHENOXY}PHTHALIC ACID",
            "non-polymer",
        ),
        "194": ("4-{2-[(3-NITROBENZOYL)AMINO]PHENOXY}PHTHALIC ACID", "non-polymer"),
        "197": (
            "4-[2-ACETYLAMINO-2-(1-BIPHENYL-4-YLMETHYL-2-OXO-AZEPAN-3-YLCARBAMOYL)-ETHYL]-2-FORMYL-BENZOIC ACID",
            "non-polymer",
        ),
        "196": (
            "3-(3-FLUORO-4-HYDROXYPHENYL)-7-HYDROXY-1-NAPHTHONITRILE",
            "non-polymer",
        ),
        "4AX": ("(R)-4-AMINO-ISOXAZOLIDIN-3-ONE", "non-polymer"),
        "190": (
            "N-{(1S)-1-(3-BROMOBENZYL)-4-[(4-BROMOPHENYL)SULFONYL]-6-METHYL-2-OXOHEPTYL}-2-(2,6-DIMETHYLPHENOXY)ACETAMIDE",
            "non-polymer",
        ),
        "4AZ": ("4-(AMINOMETHYL)BENZOIC ACID", "non-polymer"),
        "192": ("1-AMINOCYCLOBUTANECARBOXLIC ACID", "peptide"),
        "4AD": ("ASNAMP", "non-polymer"),
        "4AF": ("4-ACETYL-L-PHENYLALANINE", "peptide"),
        "4AG": ("(2R)-3-HYDROXYPROPANE-1,2-DIYL DIHEXADECANOATE", "non-polymer"),
        "ZOL": ("ZOLEDRONIC ACID", "non-polymer"),
        "4AA": ("4-CHLORO-3-HYDROXYANTHRANILIC ACID", "non-polymer"),
        "4AB": (
            "2,4-DIAMINO-6-[2,3-DIHYDROXY-PROP-3-YL]-5,6,7,8-TETRAHYDROPTERIDINE",
            "non-polymer",
        ),
        "4AC": ("N(4)-ACETYLCYTIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "4AL": ("[4-(AMINOMETHYL)PHENYL]METHANOL", "non-polymer"),
        "4AM": ("4-AMINO-2-DEOXY-2,3-DEHYDRO-N-NEURAMINIC ACID", "non-polymer"),
        "4AN": ("6-AMINO-BENZO[DE]ISOQUINOLINE-1,3-DIONE", "non-polymer"),
        "PE3": (
            "3,6,9,12,15,18,21,24,27,30,33,36,39-TRIDECAOXAHENTETRACONTANE-1,41-DIOL",
            "non-polymer",
        ),
        "4AH": ("?", "non-polymer"),
        "DLF": ("2-DEOXY-ALPHA-L-FUCOPYRANOSIDE", "saccharide"),
        "I59": ("HYDROXY(1-NAPHTHYL)METHYLPHOSPHONIC ACID", "non-polymer"),
        "I58": ("4R-FLUORO-N6-ETHANIMIDOYL-L-LYSINE", "peptide"),
        "XK9": (
            "N-{4-[(1E)-N-(N-HYDROXYCARBAMIMIDOYL)ETHANEHYDRAZONOYL]PHENYL}-7-NITRO-1H-INDOLE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "I32": (
            "4-{[(2S)-3-(TERT-BUTYLAMINO)-2-HYDROXYPROPYL]OXY}-3-IODO-1H-INDOLE-2-CARBONITRILE",
            "non-polymer",
        ),
        "I51": (
            "6-[3-HYDROXY-2-(HYDROXYMETHYL)PROP-1-EN-1-YL]-5-METHYLPYRIMIDINE-2,4(1H,3H)-DIONE",
            "non-polymer",
        ),
        "I50": (
            "N-[2-({[AMINO(IMINO)METHYL]AMINO}OXY)ETHYL]-2-{6-CHLORO-3-[(2,2-DIFLUORO-2-PHENYLETHYL)AMINO]-2-FLUOROPHENYL}ACETAMIDE",
            "non-polymer",
        ),
        "I52": (
            "N-{4-[(1-HYDROXYCARBAMOYL-2-METHYL-PROPYL)-(2-MORPHOLIN-4-YL-ETHYL)-SULFAMOYL]-4-PENTYL-BENZAMIDE",
            "non-polymer",
        ),
        "HM6": (
            "4-(4-(HYDROXYMETHYL)-1H-1,2,3-TRIAZOL-1-YL)PYRIDINE-2,6-DICARBOXYLIC ACID",
            "non-polymer",
        ),
        "WRS": (
            "4-HYDROXY-3-[(1S,3S)-3-HYDROXY-1-PHENYLBUTYL]-2H-CHROMEN-2-ONE",
            "non-polymer",
        ),
        "I5I": (
            "2-(5-ISOTHIOCYANATO-1-BENZOFURAN-2-YL)-4,5-DIHYDRO-1H-IMIDAZOLE",
            "non-polymer",
        ),
        "HM7": ("3-METHYLTHIOETHANE-PYRAZINEDIONE", "non-polymer"),
        "IL0": ("(2S,3S)-2-AMINO-3-METHYLPENTANE-1,1-DIOL", "peptide"),
        "I5A": ("5'-DEOXY-5'-IODOCYTIDINE", "non-polymer"),
        "I5C": ("5-IODO-2'-DEOXY-CYTIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "9D9": ("(1,1-DIFLUOROPENTYL)PHOSPHONIC ACID", "non-polymer"),
        "I1E": ("1-PHENYLPROPAN-1-ONE", "non-polymer"),
        "UFP": ("5-FLUORO-2'-DEOXYURIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "I5P": ("INOSITOL-(1,3,4,5,6)-PENTAKISPHOSPHATE", "non-polymer"),
        "I5S": (
            "ISOQUINOLINE-5-SULFONIC ACID (2-(2-(4-CHLOROBENZYLOXY)ETHYLAMINO)ETHYL)AMIDE",
            "non-polymer",
        ),
        "HM3": ("(4-CHLOROPYRIDIN-2-YL)METHANOL", "non-polymer"),
        "A46": (
            "5-CHLORO-N-{4-[(1R)-1,2-DIHYDROXYETHYL]PHENYL}-1H-INDOLE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "TCB": ("THIOCELLOBIOSE", "saccharide"),
        "19Y": (
            "6-ETHYL-4-METHOXY-2-(PYRIDIN-3-YLSULFANYL)-7H-PYRROLO[2,3-D]PYRIMIDINE-5-CARBALDEHYDE",
            "non-polymer",
        ),
        "19X": (
            "6-({4-[(3R)-3-AMINOPYRROLIDIN-1-YL]-5-CHLORO-6-ETHYL-7H-PYRROLO[2,3-D]PYRIMIDIN-2-YL}SULFANYL)-1H-ISOINDOL-1-ONE",
            "non-polymer",
        ),
        "4A2": (
            "(2E)-2-(3-FLUORANYL-4-METHOXY-PHENYL)IMINO-1-[[2-(TRIFLUOROMETHYL)PHENYL]METHYL]-3H-BENZIMIDAZOLE-5-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "4A3": ("4-AMINO-3-HYDROXYBENZOIC ACID", "non-polymer"),
        "19U": ("1-[(2R)-2-AMINOBUTANOYL]-N-(3-CHLOROBENZYL)-L-PROLINAMIDE", "peptide"),
        "UFT": ("?", "nucleic-acid"),
        "19Q": (
            "2-CYCLOPROPYL-N-[(2S)-3-HYDROXY-3-METHYLBUTAN-2-YL]-5H-PYRROLO[2,3-B]PYRAZINE-7-CARBOXAMIDE",
            "non-polymer",
        ),
        "A44": ("?", "nucleic-acid"),
        "19S": (
            "2-CYCLOPROPYL-N-[(2R)-1-OXO-1-(PIPERIDIN-1-YL)PROPAN-2-YL]-5H-PYRROLO[2,3-B]PYRAZINE-7-CARBOXAMIDE",
            "non-polymer",
        ),
        "19R": (
            "2-CYCLOPROPYL-N-[(2R)-3-HYDROXY-3-METHYLBUTAN-2-YL]-5H-PYRROLO[2,3-B]PYRAZINE-7-CARBOXAMIDE",
            "non-polymer",
        ),
        "19M": (
            "(2S)-2-[[(2S)-2-[2-[(2-AZANYL-4-OXIDANYLIDENE-1H-PTERIDIN-7-YL)CARBONYLAMINO]ETHANOYLAMINO]-3-PHENYL-PROPANOYL]AMINO]-3-PHENYL-PROPANOIC ACID",
            "peptide",
        ),
        "19L": (
            "(2S)-2-[[(2S)-2-[(2-AZANYL-4-OXIDANYLIDENE-1H-PTERIDIN-7-YL)CARBONYLAMINO]-3-OXIDANYL-PROPANOYL]AMINO]-3-(1H-INDOL-3-YL)PROPANOIC ACID",
            "peptide",
        ),
        "19O": (
            "NATRIUM-(5,7 ,9 ,11 ,16 ,18 ,20 ,22 )-5,7,9,11,16,18,20,22-OCTAHYDRO-5,22:7,20:9,18:11,16-TETRAMETHANONONACEN-8,19-BISPHOSPHATE",
            "non-polymer",
        ),
        "246": ("1-BENZYL-3-(2-CHLOROPYRIDIN-4-YL)UREA", "non-polymer"),
        "19H": (
            "6-CHLORO-3-[3-(4-CHLORO-3,5-DIMETHYLPHENOXY)PROPYL]-1H-INDOLE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "19K": (
            "N~4~-[4-(2,4-DIMETHYL-1,3-THIAZOL-5-YL)PYRIMIDIN-2-YL]-N~1~,N~1~-DIMETHYL-2-NITROBENZENE-1,4-DIAMINE",
            "non-polymer",
        ),
        "19J": (
            "2-[2-[(2-AZANYL-4-OXIDANYLIDENE-1H-PTERIDIN-7-YL)CARBONYLAMINO]ETHANOYLAMINO]ETHANOIC ACID",
            "peptide",
        ),
        "19E": ("11-CHLORO-8-METHYL-7H-BENZO[E]PYRIDO[4,3-B]INDOL-3-OL", "non-polymer"),
        "A42": (
            "1-(5-CHLORO-2,4-DIMETHOXYPHENYL)-3-(5-CYANOPYRAZIN-2-YL)UREA",
            "non-polymer",
        ),
        "19G": (
            "3-[3-(4-CHLORO-3,5-DIMETHYLPHENOXY)PROPYL]-1-BENZOTHIOPHENE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "19A": (
            "N,N-DIMETHYL-4-(4-PHENYL-1H-PYRAZOL-3-YL)-1H-PYRROLE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "19C": (
            "2-[(3-CHLORANYL-4-FLUORANYL-PHENYL)METHYL]-6,7-BIS(OXIDANYL)ISOINDOL-1-ONE",
            "non-polymer",
        ),
        "240": (
            "(2S)-2-(2-{[1-(4-METHOXYBENZOYL)-2-METHYL-5-(TRIFLUOROMETHOXY)-1H-INDOL-3-YL]METHYL}PHENOXY)PROPANOIC ACID",
            "non-polymer",
        ),
        "SD4": ("N-HYDROXY-L-ASPARAGINE", "peptide"),
        "243": ("(9Z,11E,13S)-13-HYDROXYOCTADECA-9,11-DIENOIC ACID", "non-polymer"),
        "SD5": (
            "5-[2,6-DI(MORPHOLIN-4-YL)PYRIMIDIN-4-YL]-4-(TRIFLUOROMETHYL)PYRIDIN-2-AMINE",
            "non-polymer",
        ),
        "242": (
            "3-(2-AMINOQUINAZOLIN-6-YL)-4-METHYL-N-[3-(TRIFLUOROMETHYL)PHENYL]BENZAMIDE",
            "non-polymer",
        ),
        "NFG": (
            "2,4-DINITROPHENYL 2-DEOXY-2-FLUORO-BETA-D-GLUCOPYRANOSIDE",
            "non-polymer",
        ),
        "MKE": ("(4S)-4-AMINO-5-OXOHEXANOIC ACID", "non-polymer"),
        "TUO": (
            "2-(HYDRAZINOCARBONYL)-3-PHENYL-1H-INDOLE-5-SULFONAMIDE",
            "non-polymer",
        ),
        "NAA": ("N-ACETYL-D-ALLOSAMINE", "saccharide"),
        "VDN": (
            "2-{2-ETHOXY-5-[(4-ETHYLPIPERAZIN-1-YL)SULFONYL]PHENYL}-5-METHYL-7-PROPYLIMIDAZO[5,1-F][1,2,4]TRIAZIN-4(1H)-ONE",
            "non-polymer",
        ),
        "SD1": (
            "2-[(3,4,5-TRIMETHOXYPHENYL)AMINO]-4-(2,6,6-TRIMETHYL-4-OXO-4,5,6,7-TETRAHYDRO-1H-INDOL-1-YL)BENZAMIDE",
            "non-polymer",
        ),
        "854": ("ETHYL (2S)-1-(BENZYLSULFONYL)PIPERIDINE-2-CARBOXYLATE", "non-polymer"),
        "855": (
            "2-(6-METHYLPYRIDIN-2-YL)-N-PYRIDIN-4-YLQUINAZOLIN-4-AMINE",
            "non-polymer",
        ),
        "857": (
            "4-({2-[(4-CHLORO-3-{[(2S)-1-METHYLPYRROLIDIN-2-YL]METHOXY}PHENYL)AMINO]-1,3-BENZOXAZOL-5-YL}OXY)-N-METHYLPYRIDINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "851": (
            "N-{(1S)-1-CYCLOHEXYL-2-OXO-2-[(2S)-2-(4-PHENYL-1,3-BENZOTHIAZOL-2-YL)PYRROLIDIN-1-YL]ETHYL}-N~2~-METHYL-L-ALANINAMIDE",
            "non-polymer",
        ),
        "852": (
            "2-{4-[2-ACETYLAMINO-2-(1-BIPHENYL-4-YLMETHYL-2-OXO-AZEPAN-3-YLCARBAMOYL)-ETHYL]-2-CARBOXY-PHENYL}-MALONIC ACID",
            "non-polymer",
        ),
        "853": (
            "5-[2-ACETYLAMINO-2-(1-BIPHENYL-4-YLMETHYL-2-OXO-AZEPAN-3-YLCARBAMOYL)-ETHYL]-2-CARBOXYMETHYL-BENZOIC ACID",
            "non-polymer",
        ),
        "NFF": (
            "(2R,4S)-2-[(1R)-1-{[(2-ETHOXYNAPHTHALEN-1-YL)CARBONYL]AMINO}-2-OXOETHYL]-5,5-DIMETHYL-1,3-THIAZOLIDINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "SD3": ("N-HYDROXY-D-ASPARAGINE", "non-polymer"),
        "858": ("C32-O-(1-ETHYL-INDOL-5-YL)ASCOMYCIN", "non-polymer"),
        "859": (
            "2-({2-[(3-HYDROXYPHENYL)AMINO]PYRIMIDIN-4-YL}AMINO)BENZAMIDE",
            "non-polymer",
        ),
        "9TN": ("DEBUTANOYL THAPSIGARGIN", "non-polymer"),
        "WR2": (
            "(2R,3R)-N^1^-[(1S)-2,2-DIMETHYL-1-(METHYLCARBAMOYL)PROPYL]-N^4^-HYDROXY-2-(2-METHYLPROPYL)-3-{[(1,3-THIAZOL-2-YLCARBONYL)AMINO]METHYL}BUTANEDIAMIDE",
            "non-polymer",
        ),
        "9TA": (
            "1,4-BIS-[2-(2-HYDROXY-ETHYLAMINO)-ETHYLAMINO]-ANTHRAQUINONE",
            "non-polymer",
        ),
        "VET": (
            "[(1R,2S,4R)-2-HYDROXY-4-[(5R)-5-HYDROXY-5-METHYL-2,4-DIOXO-IMIDAZOLIDIN-1-YL]CYCLOPENTYL]METHYL DIHYDROGEN PHOSPHATE",
            "nucleic-acid",
        ),
        "VER": ("IRON-OCTAETHYLPORPHYRIN", "non-polymer"),
        "NFE": ("NI-FE ACTIVE CENTER", "non-polymer"),
        "MKK": (
            "1-[(4R)-4-[3-(4-ACETYLPIPERAZIN-1-YL)PROPYL]-1-(2-FLUORO-5-METHYLPHENYL)-4-PHENYL-4,5-DIHYDRO-1H-PYRAZOL-3-YL]ETHANONE",
            "non-polymer",
        ),
        "PFL": ("2,6-BIS(1-METHYLETHYL)PHENOL", "non-polymer"),
        "OHX": ("OSMIUM (III) HEXAMMINE", "non-polymer"),
        "85A": (
            "18-CHLORO-2-OXO-17-[(PYRIDIN-4-YLMETHYL)AMINO]-2,3,11,12,13,14-HEXAHYDRO-1H,10H-4,8-(AZENO)-9,15,1,3,6-BENZODIOXATRIAZACYCLOHEPTADECINE-7-CARBONITRILE",
            "non-polymer",
        ),
        "VEA": ("5-OXA-PROTOPORPHYRIN IX CONTAINING FE", "non-polymer"),
        "MKH": (
            "1-[3-({(2R)-2-[HYDROXY(DIPHENYL)METHYL]PYRROLIDIN-1-YL}SULFONYL)PROPYL]PYRIMIDINE-2,4(1H,3H)-DIONE",
            "non-polymer",
        ),
        "VEK": (
            "6-BROMO-2-(1-METHYL-1H-IMIDAZOL-5-YL)-7-{4-[(5-METHYL-1,2-OXAZOL-3-YL)METHYL]PIPERAZIN-1-YL}-1H-IMIDAZO[4,5-B]PYRIDINE",
            "non-polymer",
        ),
        "TMC": (
            "1-[4-HYDROXY-5-(HYDROXYMETHYL)BICYCLO[3.1.0]HEX-2-YL]-5-METHYLPYRIMIDINE-2,4(1H,3H)-DIONE",
            "non-polymer",
        ),
        "GNB": ("S-P-NITROBENZYLOXYCARBONYLGLUTATHIONE", "non-polymer"),
        "KXN": (
            "(2R,3S)-2-(3,4-DIHYDROXYPHENYL)-3,4-DIHYDRO-2H-CHROMENE-3,5,7-TRIOL",
            "non-polymer",
        ),
        "GNC": ("N~2~-METHYL-L-GLUTAMINE", "peptide"),
        "FTP": (
            "3-(4-AMINO-2-TRIFLUOROMETHYL-PYRIMIDIN-5-YLMETHYL)-4-METHYL-5-(2-PHOSPHONATOOXY-ETHYL)-THIAZOL-3-IUM",
            "non-polymer",
        ),
        "WQP": (
            "5-(ACETYLAMINO)-3,5-DIDEOXY-2-O-PHOSPHONO-D-ERYTHRO-L-MANNO-NONONIC ACID",
            "non-polymer",
        ),
        "GNA": ("2,4-DEOXY-4-GUANIDINO-5-N-ACETYL-NEURAMINIC ACID", "non-polymer"),
        "PDZ": (
            "6-(4-CHLORO-2-FLUORO-3-PHENOXYBENZYL)PYRIDAZIN-3(2H)-ONE",
            "non-polymer",
        ),
        "Q00": ("N-[(3S)-3-AMINO-4-PHENYLBUTYL]-L-PHENYLALANINE", "peptide"),
        "GNG": ("2'-DEOXY-GUANOSINE", "non-polymer"),
        "ZPR": ("N-BENZYLOXYCARBONYL-L-PROLYL-L-PROLINAL", "peptide"),
        "SSV": (
            "(3E,5E,7E,9E,11E,13E)-PENTADECA-3,5,7,9,11,13-HEXAEN-2-ONE",
            "non-polymer",
        ),
        "NFC": ("NI-FE ACTIVE CENTER A-FORM", "non-polymer"),
        "GND": ("2-AMINO-5-GUANIDINO-PENTANOIC ACID", "peptide"),
        "WRG": (
            "(2S)-2-[[(2S)-2-ACETAMIDO-5-[[N-(METHYLCARBAMOYL)CARBAMIMIDOYL]AMINO]PENTANOYL]-METHYL-AMINO]-3-PHENYL-PROPANOIC ACID",
            "peptide",
        ),
        "FTS": (
            "(3R)-3-(FLUOROMETHYL)-N-(3,3,3-TRIFLUOROPROPYL)-1,2,3,4-TETRAHYDROISOQUINOLINE-7-SULFONAMIDE",
            "non-polymer",
        ),
        "GNE": ("1,N2-ETHENOGUANINE", "nucleic-acid"),
        "E90": ("4-[(PYRIMIDIN-2-YLSULFANYL)ACETYL]BENZENESULFONAMIDE", "non-polymer"),
        "FQX": (
            "[1,2,5]OXADIAZOLO[3,4-G]QUINOXALINE-6,7(5H,8H)-DIONE 1-OXIDE",
            "non-polymer",
        ),
        "PDY": (
            "N~7~-(4-ETHOXYPHENYL)-6-METHYL-N~5~-[(3S)-PIPERIDIN-3-YL]PYRAZOLO[1,5-A]PYRIMIDINE-5,7-DIAMINE",
            "non-polymer",
        ),
        "E94": (
            "4-(3-CHLOROPHENYL)-5-{2-[(3-HYDROXYPHENYL)AMINO]PYRIMIDIN-4-YL}-2-{[2-(PIPERIDIN-1-YL)ETHYL]AMINO}THIOPHENE-3-CARBONITRILE",
            "non-polymer",
        ),
        "E97": (
            "[3-(4-{2'-[4-(3-DIMETHYLAMINO-PROPOXY)-PHENYL]-3H,3'H-[5,5']BIBENZOIMIDAZOLYL-2-YL}-PHENOXY)-PROPYL]-DIMETHYL-AMINE",
            "non-polymer",
        ),
        "E96": (
            "4-{[4-HYDROXY-PHENYL]-1H-BENZIMIDAZOLE-5-YL}-BENZIMIDAZOLE-2-YL-[4-HYDROXY-BENZENE]",
            "non-polymer",
        ),
        "FQP": (
            "4-AMINO-2-TRIFLUOROMETHYL-5-HYDROXYMETHYLPYRIMIDINE PYROPHOSPHATE",
            "non-polymer",
        ),
        "RS7": ("(2E)-3-(2-OCT-1-YN-1-YLPHENYL)ACRYLIC ACID", "non-polymer"),
        "GNH": ("AMINOPHOSPHONIC ACID-GUANYLATE ESTER", "non-polymer"),
        "RS2": (
            "N-HYDROXY-2-[4-(4-PHENOXY-BENZENESULFONYL)-TETRAHYDRO-PYRAN-4-YL]-ACETAMIDE",
            "non-polymer",
        ),
        "RS3": (
            "1-DEOXY-1-[8-(DIMETHYLAMINO)-7-METHYL-2,4-DIOXO-3,4-DIHYDROBENZO[G]PTERIDIN-10(2H)-YL]-D-RIBITOL",
            "non-polymer",
        ),
        "GNI": (
            "(3-{4-[2-(2,4-DICHLORO-PHENOXY)-ETHYLCARBAMOYL]-5-PHENYL-ISOXAZOL-3-YL}-PHENYL)-ACETIC ACID",
            "non-polymer",
        ),
        "6TG": (
            "(2S)-2-(4-METHOXY-3,5-DIMETHYLPHENYL)-5-METHYL-2-(3-PYRIMIDIN-5-YLPHENYL)-2H-IMIDAZOL-4-AMINE",
            "non-polymer",
        ),
        "TRY": ("(2Z)-2-AMINO-3-(1H-INDOL-3-YL)PROP-2-ENOIC ACID", "peptide"),
        "T6A": (
            "N-[N-(9-B-D-RIBOFURANOSYLPURIN-6-YL)CARBAMOYL]THREONINE-5'-MONOPHOSPHATE",
            "nucleic-acid",
        ),
        "WRA": (
            "6,6-DIMETHYL-1-[3-(2,4,5-TRICHLOROPHENOXY)PROPOXY]-1,6-DIHYDRO-1,3,5-TRIAZINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "SSO": (
            "1,4-DIDEOXY-1,4-[[2S,3S)-2,4-DIHYDROXY-3-(SULFOXY)BUTYL]EPISULFONIUMYLIDENE]-D-ARABINITOL INNER SALT",
            "non-polymer",
        ),
        "6TP": (
            "4-AMINO-2-[4-(TERT-BUTYLSULFAMOYL)PHENYL]-N-METHYLTHIENO[3,2-C]PYRIDINE-7-CARBOXAMIDE",
            "non-polymer",
        ),
        "PDX": ("2,3-DI-O-SULFO-ALPHA-D-GLUCOPYRANOSE", "non-polymer"),
        "39P": (
            "4-{[5-(ISOXAZOL-3-YLCARBAMOYL)-2-METHYLPHENYL]AMINO}-5-METHYL-N-PROPYLPYRROLO[2,1-F][1,2,4]TRIAZINE-6-CARBOXAMIDE",
            "non-polymer",
        ),
        "GNL": (
            "(3AR,5R,6R,7R,7AR)-5-(HYDROXYMETHYL)-2-METHYL-5,6,7,7A-TETRAHYDRO-3AH-PYRANO[3,2-D][1,3]THIAZOLE-6,7-DIOL",
            "non-polymer",
        ),
        "NNO": ("6-HYDROXYISONICOTINIC ACID N-OXIDE", "non-polymer"),
        "V1A": (
            "S-(1-OXYL-2,2,5,5-TETRAMETHYL-2,5-DIHYDRO-1H-IMIDAZOL-4-YL) METHANESULFONOTHIOATE",
            "non-polymer",
        ),
        "PXA": (
            "2-AMINO-1,9-DICARBONYL-4,6-DIMETHYL-10-DEHYDRO-3-OXO(8-AZA)PHENOXAZINE",
            "non-polymer",
        ),
        "PFP": (
            "2-[5,6-BIS-(4-METHOXY-PHENYL)-FURO[2,3-D]PYRIMIDIN-4-YLAMINO]-ETHANOL",
            "non-polymer",
        ),
        "SSJ": ("?", "non-polymer"),
        "HZH": (
            "1,1,1-TRIFLUORO-3-[(2-PHENYLETHYL)SULFANYL]PROPAN-2-ONE",
            "non-polymer",
        ),
        "PDW": ("METHYL D-PROLINATE", "peptide"),
        "ZRC": ("OXO ZIRCONIUM(IV) CLUSTER", "non-polymer"),
        "NNN": (
            "(2R)-2-{[4-(BENZYLAMINO)-8-(1-METHYLETHYL)PYRAZOLO[1,5-A][1,3,5]TRIAZIN-2-YL]AMINO}BUTAN-1-OL",
            "non-polymer",
        ),
        "ZRA": ("BENZOYL-ARGININE-ALANINE-FLUORO-METHYL KETONE", "peptide"),
        "6T6": (
            "(2R)-2-METHYL-5-PHENYL-2-(3-PYRIDIN-3-YLPHENYL)-2,3-DIHYDRO-1H-IMIDAZOL-4-AMINE",
            "non-polymer",
        ),
        "PVB": ("PURVALANOL B", "non-polymer"),
        "ZRL": ("7-BROMO-2-PYRIDIN-4-YL-5H-THIENO[3,2-C]PYRIDIN-4-ONE", "non-polymer"),
        "ZRK": ("2-(4-PYRIDINYL)FURO[3,2-C]PYRIDIN-4(5H)-ONE", "non-polymer"),
        "RSM": (
            "(2S)-2-(ACETYLAMINO)-N-METHYL-4-[(R)-METHYLSULFINYL]BUTANAMIDE",
            "non-polymer",
        ),
        "FTT": ("3-HYDROXY-TETRADECANOIC ACID", "non-polymer"),
        "RSO": ("R-STYRENE OXIDE", "non-polymer"),
        "RSH": ("3-DEOXY-D-LYXO-HEXONIC ACID", "non-polymer"),
        "OHU": ("?", "nucleic-acid"),
        "RSD": ("RESPINOMYCIN D", "non-polymer"),
        "M31": (
            "N-(BENZYLSULFONYL)GLYCYL-N-[2-(AMINOMETHYL)-5-CHLOROBENZYL]-L-PROLINAMIDE",
            "peptide",
        ),
        "RSF": ("(3R,3AS,6AR)-HEXAHYDROFURO[2,3-B]FURAN-3-OL", "non-polymer"),
        "RSG": ("(3S,3AR,6AS)-HEXAHYDROFURO[2,3-B]FURAN-3-OL", "non-polymer"),
        "RSA": ("N-PROPARGYL-1(S)-AMINOINDAN", "non-polymer"),
        "PDV": ("DIVANADATE ION", "non-polymer"),
        "E9P": ("5-ETHYL-2-PHENOXYPHENOL", "non-polymer"),
        "RSX": (
            "(4AS,7AR)-OCTAHYDRO-1H-CYCLOPENTA[B]PYRIDINE-6,6-DIYLBIS(PHOSPHONIC ACID)",
            "non-polymer",
        ),
        "RSY": (
            "N-[(2R)-1-{[(2S,3S)-5-{[(2R)-1-{[(2S)-1-AMINO-4-METHYL-1-OXOPENTAN-2-YL]AMINO}-3-CHLORO-1-OXOPROPAN-2-YL]AMINO}-3-HYDROXY-5-OXO-1-PHENYLPENTAN-2-YL]AMINO}-3-METHYL-1-OXOBUTAN-2-YL]PYRIDINE-2-CARBOXAMIDE",
            "peptide",
        ),
        "RSZ": (
            "N-[(2S)-1-{[(2S,3S)-3-HYDROXY-5-OXO-5-{[(2R)-1-OXO-3-PHENYL-1-(PROP-2-YN-1-YLAMINO)PROPAN-2-YL]AMINO}-1-PHENYLPENTAN-2-YL]AMINO}-3-METHYL-1-OXOBUTAN-2-YL]PYRIDINE-2-CARBOXAMIDE",
            "peptide",
        ),
        "XXZ": (
            "1-[4-(AMINOMETHYL)BENZOYL]-5'-FLUORO-1'H-SPIRO[PIPERIDINE-4,2'-QUINAZOLIN]-4'-AMINE",
            "non-polymer",
        ),
        "RST": ("RISTOSAMINE", "saccharide"),
        "RSV": ("6-(THIOPHEN-3-YL)QUINOLIN-2-AMINE", "non-polymer"),
        "E9Z": (
            "5-CHLORO-N~4~-CYCLOPROPYL-N~2~-[4-(2-METHOXYETHOXY)PHENYL]PYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "RSP": (
            "4-AMINO-1-(5-O-PHOSPHONO-BETA-D-RIBOFURANOSYL)PYRIMIDINE-2(1H)-THIONE",
            "nucleic-acid",
        ),
        "RSQ": ("?", "nucleic-acid"),
        "GP8": ("1-(2-AMIDINOPHENYL)-3-(PHENOXYPHENYL)UREA", "non-polymer"),
        "PY8": (
            "6-BROMO-13-THIA-2,4,8,12,19-PENTAAZATRICYCLO[12.3.1.1~3,7~]NONADECA-1(18),3(19),4,6,14,16-HEXAENE 13,13-DIOXIDE",
            "non-polymer",
        ),
        "PY9": (
            "4-{[(2R)-2-(2-METHYLPHENYL)PYRROLIDIN-1-YL]CARBONYL}BENZENE-1,3-DIOL",
            "non-polymer",
        ),
        "DKT": (
            "4-[2-(3-BENZYLOXYCARBONYLAMINO-4-CYCLOHEXYL-1-HYDROXY-2-OXO-BUTYLAMINO)-5-GUANIDINO-PENTANOYLAMINO]-4-(1-CARBOXY-2-CYCLOHEXYL-ETHYLCARBAMOYL)-BUTYRIC ACID",
            "non-polymer",
        ),
        "PY0": ("(1S,2S)-1-AMINO-1,2-DIHYDROXYPROPAN-1-OLATE", "non-polymer"),
        "PY1": ("4-(3-PYRIDIN-2-YL-1H-PYRAZOL-4-YL)QUINOLINE", "non-polymer"),
        "PY2": ("3-(MERCAPTOMETHYLENE)PYRIDINE", "non-polymer"),
        "PY3": (
            "5-[AMINO(IMINO)METHYL]-2-[({[6-[3-AMINO-5-({[(1R)-1-METHYLPROPYL]AMINO}CARBONYL)PHENYL]-3-(ISOPROPYLAMINO)-2-OXOPYRAZIN-1(2H)-YL]ACETYL}AMINO)METHYL]-N-PYRIDIN-4-YLBENZAMIDE",
            "non-polymer",
        ),
        "DKZ": (
            "4-AMINO-1-(3-DEOXY-3-FLUORO-BETA-D-GLUCOPYRANOSYL)PYRIMIDIN-2(1H)-ONE",
            "non-polymer",
        ),
        "PY5": ("2-[O-PHOSPHONOPYRIDOXYL]-AMINO-PENTANOIC ACID", "non-polymer"),
        "DKX": (
            "1-(3-DEOXY-3-FLUORO-BETA-D-GLUCOPYRANOSYL)PYRIMIDINE-2,4(1H,3H)-DIONE",
            "non-polymer",
        ),
        "DKY": (
            "1-(3-DEOXY-3-FLUORO-BETA-D-GLUCOPYRANOSYL)-4-[(PHENYLCARBONYL)AMINO]PYRIMIDIN-2(1H)-ONE",
            "non-polymer",
        ),
        "IP2": ("D-MYO-INOSITOL-4,5-BISPHOSPHATE", "non-polymer"),
        "IP3": ("1-HYDROXY-3-METHYLBUTANE", "non-polymer"),
        "DKD": ("(2S)-2-(4-PHENETHYLPHENOXY)-3-PHENYL-PROPANOIC ACID", "non-polymer"),
        "IP1": ("ISOPENICILLIN N", "non-polymer"),
        "498": (
            "P-(5-RUTHENOCENYL-1H-1,2,3-TRIAZOL-1-YL)BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "C38": ("5-IODO-2'-DEOXY-CYTIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "IP4": ("1-METHYL-2-OXY-5,5-DIMETHYL PYRROLIDINE", "non-polymer"),
        "DKA": ("DECANOIC ACID", "non-polymer"),
        "C35": (
            "N~2~-{[1-(4-CHLOROPHENYL)-5-METHYL-1H-1,2,4-TRIAZOL-3-YL]CARBONYL}-N~5~-(DIAMINOMETHYLIDENE)-L-ORNITHYL-L-LEUCYL-L-ISOLEUCYL-4-FLUORO-L-PHENYLALANINAMIDE",
            "non-polymer",
        ),
        "C34": ("N4-METHYL-2'-DEOXY-CYTIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "496": (
            "2-({[4-(HYDROXYMETHYL)-3-(TRIFLUOROMETHYL)-1H-PYRAZOL-1-YL]ACETYL}AMINO)-4,5,6,7-TETRAHYDRO-1-BENZOTHIOPHENE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "497": (
            "6-[AMINO(IMINO)METHYL]-N-[(4R)-4-ETHYL-1,2,3,4-TETRAHYDROISOQUINOLIN-6-YL]-2-NAPHTHAMIDE",
            "non-polymer",
        ),
        "C31": ("?", "nucleic-acid"),
        "DKK": (
            "N-[2-(CARBAMIMIDAMIDOOXY)ETHYL]-2-{6-CYANO-3-[(2,2-DIFLUORO-2-PYRIDIN-2-YLETHYL)AMINO]-2-FLUOROPHENYL}ACETAMIDE",
            "non-polymer",
        ),
        "492": ("1-(2,2'-BITHIOPHEN-5-YL)METHANAMINE", "non-polymer"),
        "C32": ("5-BROMO-2'-DEOXY-CYTIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "NSX": (
            "N~2~,N~6~-BIS(BIPHENYL-4-YLACETYL)-L-LYSYL-D-ARGINYL-N-(2-PHENYLETHYL)-L-TYROSINAMIDE",
            "non-polymer",
        ),
        "NSY": (
            "N~2~,N~6~-BIS(BIPHENYL-4-YLACETYL)-L-LYSYL-D-ARGINYL-N-(2-PHENYLETHYL)-L-PHENYLALANINAMIDE",
            "non-polymer",
        ),
        "PDT": ("1,3-PROPANEDITHIOL", "non-polymer"),
        "NSZ": (
            "N-(BIPHENYL-4-YLACETYL)-S-METHYL-L-CYSTEINYL-D-ARGINYL-N-(2-PHENYLETHYL)-L-PHENYLALANINAMIDE",
            "non-polymer",
        ),
        "NNK": ("4-(AMINOMETHYL)-1-METHYLPYRIDINIUM", "non-polymer"),
        "H25": (
            "5-CHLORO-THIOPHENE-2-CARBOXYLIC ACID ((3S,4S)-1-{[2-FLUORO-4-(2-OXO-2H-PYRIDIN-1-YL)-PHENYLCARBAMOYL]-METHYL}-4-HYDROXY-PYRROLIDIN-3-YL)-AMIDE",
            "non-polymer",
        ),
        "UR2": (
            "4-[3-CARBOXYMETHYL-3-(4-PHOSPHONOOXY-BENZYL)-UREIDO]-4-[(3-CYCLOHEXYL-PROPYL)-METHYL-CARBAMOYL]BUTYRIC ACID",
            "non-polymer",
        ),
        "UR3": ("3-METHYLURIDINE-5'-MONOPHOSHATE", "nucleic-acid"),
        "NFM": ("N-[(4-AMINO-2-METHYLPYRIMIDIN-5-YL)METHYL]FORMAMIDE", "non-polymer"),
        "E67": (
            "7-[(5-AMINOPENTYL)OXY]-N~4~-(1-BENZYLPIPERIDIN-4-YL)-N~2~-[3-(DIMETHYLAMINO)PROPYL]-6-METHOXYQUINAZOLINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "NAG": ("N-ACETYL-D-GLUCOSAMINE", "saccharide"),
        "URE": ("UREA", "non-polymer"),
        "URF": ("5-FLUOROURACIL", "non-polymer"),
        "URA": ("URACIL", "non-polymer"),
        "URB": ("5-BROMOPYRIMIDINE-2,4(1H,3H)-DIONE", "non-polymer"),
        "URC": ("URIC ACID", "non-polymer"),
        "URM": ("URIDINE DIPHOSPHO METHYLENE GALACTOPYRANOSE", "non-polymer"),
        "URN": ("5,6-DIAMINOPYRIMIDINE-2,4(1H,3H)-DIONE", "non-polymer"),
        "RHN": (
            "4,5-DIHYDROXY-9,10-DIOXO-9,10-DIHYDROANTHRACENE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "URI": ("URIDINE", "non-polymer"),
        "URT": (
            "({[(2R,5R)-5-(6-AMINO-9H-PURIN-9-YL)-4-FLUORO-2,5-DIHYDROFURAN-2-YL]OXY}METHYL)PHOSPHONIC ACID",
            "nucleic-acid",
        ),
        "JUP": (
            "4-(2-AMINO-5-{4-[(DIMETHYLAMINO)METHYL]THIOPHEN-2-YL}PYRIDIN-3-YL)-2-{(1R)-1-[2-(TRIFLUOROMETHYL)PHENYL]ETHOXY}BENZAMIDE",
            "non-polymer",
        ),
        "URP": ("N-(AMINOCARBONYL)-BETA-ALANINE", "non-polymer"),
        "NFL": ("2-{[3-(TRIFLUOROMETHYL)PHENYL]AMINO}NICOTINIC ACID", "non-polymer"),
        "URS": ("N-PHENYLTHIOUREA", "non-polymer"),
        "URX": (
            "[(1R,3R,4R,5R,7S)-3-(2,4-DIOXO-3,4-DIHYDROPYRIMIDIN-1(2H)-YL)-7-HYDROXY-5-METHYL-2-OXABICYCLO[2.2.1]HEPT-1-YL]METHYL DIHYDROGEN PHOSPHATE",
            "nucleic-acid",
        ),
        "C3M": (
            "N-[(2R,3S)-3-AMINO-2-HYDROXY-4-PHENYLBUTYL]NAPHTHALENE-2-SULFONAMIDE",
            "non-polymer",
        ),
        "IPC": (
            "3-[ISOPROPYL(4-METHYLBENZOYL)AMINO]-5-PHENYLTHIOPHENE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "C3O": ("2ALPHA-PROPOXY-1ALPHA,25-DIHYDROXYVITAMIN D3", "non-polymer"),
        "DK5": (
            "1-(2,3-DIDEOXY-3-FLUORO-BETA-D-ARABINO-HEXOPYRANOSYL)-4-[(PHENYLCARBONYL)AMINO]PYRIMIDIN-2(1H)-ONE",
            "non-polymer",
        ),
        "DK2": (
            "2-[(1R)-1-CARBOXY-2-(4-HYDROXYPHENYL)ETHYL]-1,3-DIOXOISOINDOLINE-5-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "DK3": ("METHYL (3S)-3-HYDROXYDODECANOATE", "non-polymer"),
        "IPD": ("D-MYO-INOSITOL-1-PHOSPHATE", "non-polymer"),
        "DK1": ("5,7-DICHLORO-4-HYDROXYQUINOLINE-2-CARBOXYLIC ACID", "non-polymer"),
        "PYP": ("2'-DEOXYRIBOFURANOSYLPYRENE-5'-MONOPHOSPHATE", "non-polymer"),
        "C3D": (
            "(5R,6S)-6-PHENYL-5-[4-(2-PYRROLIDIN-1-YLETHOXY)PHENYL]-5,6,7,8-TETRAHYDRONAPHTHALEN-2-OL",
            "non-polymer",
        ),
        "IPH": ("PHENOL", "non-polymer"),
        "IPI": ("3-METHYLBUTYLPHOSPHONIC ACID", "non-polymer"),
        "C3A": (
            "ALLYL {(1S)-1-[(5-{4-[(2,3-DIHYDRO-1H-INDEN-2-YLAMINO)CARBONYL]BENZYL}-1,2,4-OXADIAZOL-3-YL)CARBONYL]-3-PYRROLIDIN-3-YLPROPYL}CARBAMATE",
            "non-polymer",
        ),
        "49A": (
            "4,9-AMINO-2,4-DEOXY-2,3-DEHYDRO-N-ACETYL-NEURAMINIC ACID",
            "non-polymer",
        ),
        "IPL": ("INDOLE-3-PROPANOL PHOSPHATE", "non-polymer"),
        "C3B": ("4-AMINO-1-BETA-D-GLUCOPYRANOSYLPYRIMIDIN-2(1H)-ONE", "non-polymer"),
        "IPR": ("ISOPENTYL PYROPHOSPHATE", "non-polymer"),
        "IPS": ("INORGANIC PHOSPHATE GROUP", "non-polymer"),
        "IPP": ("(P-IODOPHENYLACETYLAMINO)METHYLPHOSPHINIC ACID", "non-polymer"),
        "C3Y": ("S-[(1S)-1-HYDROXY-1-(HYDROXYAMINO)ETHYL]-L-CYSTEINE", "peptide"),
        "C3X": ("2,3-EPOXYPROPYL-BETA-D-XYLOSIDE", "saccharide"),
        "IPT": ("ISOPROPYL-1-BETA-D-THIOGALACTOSIDE", "saccharide"),
        "IPU": ("S-ISOPROPYL-ISOTHIOUREA", "non-polymer"),
        "IPZ": ("2-ISOPROPYL-3-METHOXYPYRAZINE", "non-polymer"),
        "49U": (
            "(S)-N-(4-CARBAMIMIDOYLBENZYL)-1-(3-CYCLOPENTYLPROPANOYL)PYRROLIDINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "IPX": (
            "4-[(1R,2R)-2-{[(5-FLUORO-1H-INDOL-2-YL)CARBONYL]AMINO}-2,3-DIHYDRO-1H-INDEN-1-YL]BUTANOIC ACID",
            "non-polymer",
        ),
        "IPY": ("IMIDAZOLE-PYRROLE POLYAMIDE", "non-polymer"),
        "C3Q": (
            "N-[8-({[5-(DIMETHYLAMINO)NAPHTHALEN-1-YL]SULFONYL}AMINO)OCTYL]-3-[(3S,5S,7S)-TRICYCLO[3.3.1.1~3,7~]DEC-1-YL]PROPANAMIDE",
            "non-polymer",
        ),
        "C3P": ("CYTIDINE-3'-MONOPHOSPHATE", "non-polymer"),
        "C3S": ("CHOLEST-5-EN-3-YL HYDROGEN SULFATE", "non-polymer"),
        "PYG": ("BENZENE-1,2,3-TRIOL", "non-polymer"),
        "NGD": ("NICOTINAMIDE GUANINE DINUCLEOTIDE", "non-polymer"),
        "NGE": (
            "3,5-DIDEOXY-5-[(HYDROXYACETYL)AMINO]-D-GLYCERO-BETA-D-GALACTO-NON-2-ULOPYRANOSONIC ACID",
            "saccharide",
        ),
        "HDU": (
            "N-[4-(2-METHYLIMIDAZO[1,2-A]PYRIDIN-3-YL)-2-PYRIMIDINYL]ACETAMIDE",
            "non-polymer",
        ),
        "HDT": (
            "4-[(4-IMIDAZO[1,2-A]PYRIDIN-3-YLPYRIMIDIN-2-YL)AMINO]BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "HDS": ("1-HEXADECANOSULFONIC ACID", "non-polymer"),
        "NGA": ("N-ACETYL-D-GALACTOSAMINE", "saccharide"),
        "HDP": (
            "[(1S,6S)-6-HYDROXY-4-(5-METHYL-2,4-DIOXO-3,4-DIHYDROPYRIMIDIN-1(2H)-YL)CYCLOHEX-2-EN-1-YL]METHYL DIHYDROGEN PHOSPHATE",
            "nucleic-acid",
        ),
        "NGL": ("ACETYLGALACTOSAMINE-4-SULFATE", "saccharide"),
        "NGM": ("NOGALAMYCIN", "non-polymer"),
        "NGN": ("NITROGEN", "non-polymer"),
        "NGO": (
            "2-METHYL-4,5-DIHYDRO-(1,2-DIDEOXY-ALPHA-D-GLUCOPYRANOSO)[2,1-D]-1,3-OXAZOLE",
            "non-polymer",
        ),
        "NGH": (
            "N-ISOBUTYL-N-[4-METHOXYPHENYLSULFONYL]GLYCYL HYDROXAMIC ACID",
            "non-polymer",
        ),
        "HDZ": ("NITROGEN MOLECULE", "non-polymer"),
        "HDY": (
            "1-(DIMETHYLAMINO)-3-(4-{{4-(2-METHYLIMIDAZO[1,2-A]PYRIDIN-3-YL)PYRIMIDIN-2-YL]AMINO}PHENOXY)PROPAN-2-OL",
            "non-polymer",
        ),
        "HDX": (
            "({3-HYDROXY-2-OXO-4-[2-(PHOSPHONOOXY)ETHYL]PYRIDIN-1(2H)-YL}METHYL)PHOSPHONIC ACID",
            "non-polymer",
        ),
        "NGT": (
            "3AR,5R,6S,7R,7AR-5-HYDROXYMETHYL-2-METHYL-5,6,7,7A-TETRAHYDRO-3AH-PYRANO[3,2-D]THIAZOLE-6,7-DIOL",
            "non-polymer",
        ),
        "HDF": (
            "8-HYDROXY-10-(D-RIBO-2,3,4,5-TETRAHYDROXYPENTYL)-5-DEAZAISOALLOXAZINE",
            "non-polymer",
        ),
        "HDE": ("CIS-HEME D HYDROXYCHLORIN GAMMA-SPIROLACTONE 17R, 18S", "non-polymer"),
        "HDD": ("CIS-HEME D HYDROXYCHLORIN GAMMA-SPIROLACTONE", "non-polymer"),
        "HDC": ("3R-HYDROXYDECANOYL-COENZYME A", "non-polymer"),
        "HDB": (
            "(R)-1-(4-(4-(HYDROXYMETHYL)-1,3,2-DIOXABOROLAN-2-YL)BENZYL)GUANIDINE",
            "non-polymer",
        ),
        "HDA": ("HADACIDIN", "non-polymer"),
        "NGS": ("2-(ACETYLAMINO)-2-DEOXY-6-O-SULFO-BETA-D-GLUCOPYRANOSE", "saccharide"),
        "HDN": ("METHYLHYDRAZINE", "non-polymer"),
        "HDM": ("DIMETHYL PROPIONATE ESTER HEME", "non-polymer"),
        "HDL": ("D-ARABINOHYDROXAMIC ACID", "saccharide"),
        "GZ2": ("3-(1H-TETRAZOL-5-YLAMINO)CYCLOHEX-2-EN-1-ONE", "non-polymer"),
        "NGY": (
            "2-(ACETYLAMINO)-2-DEOXY-6-O-SULFO-ALPHA-D-GLUCOPYRANOSE",
            "saccharide",
        ),
        "HDI": (
            "3-(4-HYDROXYPHENYL)-4,5-DIHYDRO-5-ISOXAZOLE-ACETIC ACID METHYL ESTER",
            "non-polymer",
        ),
        "BMB": (
            "(3S,6S,7R,9AS)-7-[2-(BENZYLAMINO)ETHYL]-N-(DIPHENYLMETHYL)-6-{[(2S)-2-(METHYLAMINO)BUTANOYL]AMINO}-5-OXOOCTAHYDRO-1H-PYRROLO[1,2-A]AZEPINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "BMC": (
            "3-(1,3-BENZODIOXOL-5-YL)-1-METHYL-1H-PYRAZOLE-5-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "PDP": ("PYRIDOXAL-5'-DIPHOSPHATE", "non-polymer"),
        "BMA": ("BETA-D-MANNOSE", "saccharide"),
        "SH4": ("(1R)-1-PHENYLETHYL 4-(ACETYLAMINO)BENZYLPHOSPHONATE", "non-polymer"),
        "BMG": (
            "(4R,5S)-3-(6,7-DIHYDRO-5H-PYRAZOLO[1,2-A][1,2,4]TRIAZOL-4-IUM-6-YLSULFANYL)-5-[(1S,2R)-1-FORMYL-2-HYDROXYPROPYL]-4-METHYL-4,5-DIHYDRO-1H-PYRROLE-2-CARBOXYLATE",
            "non-polymer",
        ),
        "BMD": ("BUTYRAMIDE", "non-polymer"),
        "BME": ("BETA-MERCAPTOETHANOL", "non-polymer"),
        "BMH": (
            "5H-PYRAZOLO(1,2-A)(1,2,4)TRIAZOL-4-IUM, 6-((2-CARBOXY-6-(1-HYDROXYETHYL)-4-METHYL-7-OXO-1-AZABICYCLO(3.2.0)HEPT-2-EN-3-YL)THIO)-6,7-DIHYDRO-, HYDROXIDE, INNER SALT, (4R-(4ALPHA,5BETA,6BETA(R*)))-",
            "non-polymer",
        ),
        "BMI": (
            "3-[5-(1H-IMIDAZOL-1-YL)-7-METHYL-1H-BENZIMIDAZOL-2-YL]-4-[(PYRIDIN-2-YLMETHYL)AMINO]PYRIDIN-2(1H)-ONE",
            "non-polymer",
        ),
        "BMN": (
            "(1R)-1,4-ANHYDRO-2-DEOXY-1-(3-METHOXYNAPHTHALEN-2-YL)-5-O-PHOSPHONO-D-ERYTHRO-PENTITOL",
            "nucleic-acid",
        ),
        "BMO": (
            "N-ACRIDIN-9-YL-N'-[3-(ACRIDIN-9-YLAMINO)PROPYL]PROPANE-1,3-DIAMINE",
            "non-polymer",
        ),
        "BML": ("4-BROMOPHENOL", "non-polymer"),
        "BMM": ("BROMOMETHANE", "non-polymer"),
        "BMS": ("?", "non-polymer"),
        "BMP": ("6-HYDROXYURIDINE-5'-PHOSPHATE", "nucleic-acid"),
        "BMQ": ("?", "nucleic-acid"),
        "BMV": (
            "3-BENZYL-1-(1H-IMIDAZOL-4-YLMETHYL)-4-(THIEN-2-YLSULFONYL)-2,3,4,5-TETRAHYDRO-1H-1,4-BENZODIAZEPINE-7-CARBONITRILE",
            "non-polymer",
        ),
        "BMW": (
            "3-[7-(1H-BENZIMIDAZOL-5-YL)-2-(MORPHOLIN-4-YL)-6,7-DIHYDRO-5H-PYRROLO[2,3-D]PYRIMIDIN-4-YL]PHENOL",
            "non-polymer",
        ),
        "BMT": ("4-METHYL-4-[(E)-2-BUTENYL]-4,N-METHYL-THREONINE", "peptide"),
        "BMU": (
            "1-(5-TERT-BUTYL-2-METHYL-2H-PYRAZOL-3-YL)-3-(4-CHLORO-PHENYL)-UREA",
            "non-polymer",
        ),
        "BMZ": (
            "2-(2-HYDROXY-PHENYL)-1H-BENZOIMIDAZOLE-5-CARBOXAMIDINE",
            "non-polymer",
        ),
        "BMX": (
            "2-(ACETYLAMINO)-2-DEOXY-6-O-PHOSPHONO-ALPHA-D-MANNOPYRANOSE",
            "saccharide",
        ),
        "PDO": ("1,3-PROPANDIOL", "non-polymer"),
        "SHR": ("N-(5-AMINO-5-CARBOXYPENTYL)GLUTAMIC ACID", "peptide"),
        "SHT": (
            "O-PHOSPHONO-N-{(2E)-7-[(2-SULFOETHYL)DITHIO]HEPT-2-ENOYL}-L-THREONINE",
            "non-polymer",
        ),
        "SHU": ("4,6-DIOXOHEPTANOIC ACID", "non-polymer"),
        "SHV": ("HEPTANOIC ACID", "non-polymer"),
        "SHW": (
            "S-[2-[3-[[(2S)-2-HYDROXY-3,3-DIMETHYL-4-PHOSPHONOOXY-BUTANOYL]AMINO]PROPANOYLAMINO]ETHYL] (3R)-3-HYDROXYOCTANETHIOATE",
            "non-polymer",
        ),
        "SHY": (
            "4-[(3AS,4R,7R,8AS,8BR)-2-(1,3-BENZODIOXOL-5-YLMETHYL)-7-HYDROXY-1,3-DIOXODECAHYDROPYRROLO[3,4-A]PYRROLIZIN-4-YL]BENZENECARBOXIMIDAMIDE",
            "non-polymer",
        ),
        "J28": (
            "5-[(E)-(2-AMINO-4-HYDROXY-5-METHYLPHENYL)DIAZENYL]-2,4-DIMETHYLBENZENESULFONIC ACID",
            "non-polymer",
        ),
        "BM2": (
            "S-(R*,R*)]-4-[AMINOIMINOMETHYL)AMINO]-N-[[1-[3-HYDROXY -2-[(2-NAPHTHALENYLSULFONYL)AMINO]-1-OXOPROPYL]-2-PYRROLIDINYL] METHYL]BUTANAMIDE",
            "peptide",
        ),
        "BM3": ("2-(ACETYLAMINO)-2-DEOXY-ALPHA-D-MANNOPYRANOSE", "non-polymer"),
        "SHB": ("METHYL BETA-D-GALACTOPYRANURONATE", "saccharide"),
        "BM1": (
            "(R)-(+)9B-(3-METHYL)PHENYL-2,3-DIHYDROTHIAZOLO[2,3-A]ISOINDOL-5(9BH)-ONE",
            "non-polymer",
        ),
        "BM6": (
            "4-[2-(5,5,8,8-TETRAMETHYL-5,6,7,8-TETRAHYDRO-NAPHTHALEN-2-YL)-[1,3]DIOXOLAN-2-YL]-BENZOIC ACID",
            "non-polymer",
        ),
        "BM4": (
            "3-[(3R)-3-ETHYL-1-{9-[(3S)-3-ETHYL-3-(3-HYDROXYPHENYL)AZEPAN-1-YL]NONYL}AZEPAN-3-YL]PHENOL",
            "non-polymer",
        ),
        "BM5": (
            "(R)-(+) 5(9BH)-OXO-9B-PHENYL-2,3-DIHYDROTHIAZOLO[2,3-A]ISOINDOL-3-CARBOXYLIC ACID METHYL ESTER",
            "non-polymer",
        ),
        "SHH": ("OCTANEDIOIC ACID HYDROXYAMIDE PHENYLAMIDE", "non-polymer"),
        "ECC": ("(4S)-4-AMINO-5-HYDROXYPENTANAMIDE", "peptide"),
        "BM9": (
            "[S-(R*,R*)]-1-(AMINOIMINOMETHYL)-N-[[1-[N-[(2-NAPHTHALENYLSULFONYL)-L-SERYL]-3-PYRROLIDINYL]METHYL]-3-PIPERIDENECARBOXAMIDE",
            "peptide",
        ),
        "SHL": ("5-HYDROXY-6-METHYL-4-OXO-4H-PYRAN-2-CARBOXYLIC ACID", "non-polymer"),
        "L1O": ("4-(2-AMINOETHOXY)-3,5-DICHLOROBENZOIC ACID", "non-polymer"),
        "J3Z": ("(9BETA,13ALPHA)-3-HYDROXYESTRA-1,3,5(10)-TRIEN-17-ONE", "non-polymer"),
        "SHO": ("5-HYDROXYLAEVULINIC ACID", "non-polymer"),
        "MGC": ("ALPHA-METHYL-N-ACETYL-D-GALACTOSAMINE", "non-polymer"),
        "MGB": ("METHYLGLYOXAL BIS-(GUANYLHYDRAZONE)", "non-polymer"),
        "GZL": ("BETA-D-GALACTOFURANOSE", "non-polymer"),
        "CP3": ("COPROPORPHYRIN III", "non-polymer"),
        "MGG": ("2-(2-CARBOXY-ACETYLAMINO)-5-GUANIDINO-PENTANOIC ACID", "peptide"),
        "ECA": (
            "N,N',N''-[BENZENE-1,3,5-TRIYLTRIS(METHYLENE)]TRIS(2,3-DIHYDROXYBENZAMIDE)",
            "non-polymer",
        ),
        "HD1": (
            "6-(4-METHYLPIPERAZIN-1-YL)-N-[(1R,3S)-5-OXIDANYL-2-ADAMANTYL]-2-PROPYLSULFANYL-PYRIDINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "MGD": (
            "2-AMINO-5,6-DIMERCAPTO-7-METHYL-3,7,8A,9-TETRAHYDRO-8-OXA-1,3,9,10-TETRAAZA-ANTHRACEN-4-ONE GUANOSINE DINUCLEOTIDE",
            "non-polymer",
        ),
        "MGK": (
            "METHYL N-(CARBOXYMETHYL)-N-(3-PHENYLPROPYL)GLYCYL-L-HISTIDINATE",
            "non-polymer",
        ),
        "TBP": ("2,4,6-TRIBROMOPHENOL", "non-polymer"),
        "MGI": (
            "METHYL 4-(2,3-DIHYDROXY-5-METHYLPHENOXY)-2-HYDROXY-6-METHYLBENZOATE",
            "non-polymer",
        ),
        "PXG": ("3-[O-PHOSPHONOPYRIDOXYL]--AMINO-BENZOIC ACID", "non-polymer"),
        "MGO": (
            "[[(2R,3S,4R,5R)-5-(6-AMINO-3-METHYL-4-OXO-5H-IMIDAZO[4,5-C]PYRIDIN-1-YL)-3,4-DIHYDROXY-OXOLAN-2-YL]METHOXY-HYDROXY-PHOSPHORYL] PHOSPHONO HYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "GZC": (
            "N,N'''-{ETHANE-1,2-DIYLBIS[NITRILO(E)METHYLYLIDENE(4-HYDROXYBENZENE-3,1-DIYL)]}DIGUANIDINE",
            "non-polymer",
        ),
        "IP6": ("N~3~-[3-(1H-INDOL-6-YL)BENZYL]PYRIDINE-2,3-DIAMINE", "non-polymer"),
        "MGL": ("O1-METHYL-GLUCOSE", "saccharide"),
        "MGS": (
            "1,2-O-DIMETHYL-4-[2,4-DIHYDROXY-BUTYRAMIDO]-4,6-DIDEOXY-ALPHA-D-MANNOPYRANOSIDE",
            "non-polymer",
        ),
        "ECO": ("ESTRADIOL-6 CARBOXYL-METHYL-OXIME", "non-polymer"),
        "MGQ": ("7-BENZYL GUANINE MONOPHOSPHATE", "nucleic-acid"),
        "MGP": ("7-METHYL-GUANOSINE-5'-TRIPHOSPHATE", "non-polymer"),
        "GZZ": (
            "N-{8-[(8-{[(E)-AMINO(IMINO)METHYL]AMINO}OCTYL)AMINO]OCTYL}GUANIDINE",
            "non-polymer",
        ),
        "MGV": ("P-FLUORO-7-BENZYL GUANINE MONOPHOSPHATE", "nucleic-acid"),
        "MGU": (
            "2-O-[2-O-METHYL-4-[2,4-DIHYDROXY-BUTYRAMIDO]-4,6-DIDEOXY-]-ALPHA-D-MANNOPYRANOSYL-[1-O-METHYL-4-[2,4-DIHYDROXY-BUTYRAMIDO]-4,6-DIDEOXY]-ALPHA-D-MANNOPYRANOSE",
            "non-polymer",
        ),
        "ECN": (
            "1-[(2S)-2-[(4-CHLOROBENZYL)OXY]-2-(2,4-DICHLOROPHENYL)ETHYL]-1H-IMIDAZOLE",
            "non-polymer",
        ),
        "TBA": ("TETRABUTYLAMMONIUM ION", "non-polymer"),
        "MGY": ("N-METHYLGLYCINE", "peptide"),
        "MGX": ("1-METHYLGUANIDINE", "non-polymer"),
        "PDM": (
            "4-[4-(2,5-DIOXO-PYRROLIDIN-1-YL)-PHENYLAMINO]-4-HYDROXY-BUTYRIC ACID",
            "non-polymer",
        ),
        "TBD": (
            "2*-DEOXY-THYMIDINE-5*-ALPHA BORANO DIPHOSPHATE (ISOMER RP)",
            "non-polymer",
        ),
        "TBG": ("3-METHYL-L-VALINE", "peptide"),
        "RHD": ("RHODIUM HEXAMINE ION", "non-polymer"),
        "D72": (
            "(2R)-6,8-DICHLORO-7-(2-METHYLPROPOXY)-2-(TRIFLUOROMETHYL)-2H-CHROMENE-3-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "LUR": (
            "{2-[(2-CHLORO-6-FLUOROPHENYL)AMINO]-5-METHYLPHENYL}ACETIC ACID",
            "non-polymer",
        ),
        "ECL": (
            "1-[(2R)-2-[(4-CHLOROBENZYL)OXY]-2-(2,4-DICHLOROPHENYL)ETHYL]-1H-IMIDAZOLE",
            "non-polymer",
        ),
        "LUT": (
            "(3R,3'R,6S)-4,5-DIDEHYDRO-5,6-DIHYDRO-BETA,BETA-CAROTENE-3,3'-DIOL",
            "non-polymer",
        ),
        "R4C": (
            "N-{(2S)-3-[(R)-(4-BROMOPHENYL)(HYDROXY)PHOSPHORYL]-2-[(3-PHENYL-1,2-OXAZOL-5-YL)METHYL]PROPANOYL}-L-ALANYL-L-ALANINAMIDE",
            "peptide",
        ),
        "R4B": (
            "N-[(2S)-3-[(R)-(4-BROMOPHENYL)(HYDROXY)PHOSPHORYL]-2-{[3-(3'-CHLOROBIPHENYL-4-YL)-1,2-OXAZOL-5-YL]METHYL}PROPANOYL]-L-ALANYL-L-ALANINAMIDE",
            "peptide",
        ),
        "LUX": (
            "(3R,3'R,6'S,9R,9'R,13R,13'S)-4',5'-DIDEHYDRO-5',6',7',8',9,9',10,10',11,11',12,12',13,13',14,14',15,15'-OCTADECAHYDRO-BETA,BETA-CAROTENE-3,3'-DIOL",
            "non-polymer",
        ),
        "R4L": (
            "11-CYCLOPENTYL-2-[[2-METHOXY-4-[4-(4-METHYLPIPERAZIN-1-YL)PIPERIDIN-1-YL]CARBONYL-PHENYL]AMINO]-5-METHYL-PYRIMIDO[4,5-B][1,4]BENZODIAZEPIN-6-ONE",
            "non-polymer",
        ),
        "LUZ": ("PTERIDINE-2,4(1H,3H)-DIONE", "non-polymer"),
        "IN": ("INDIUM (III) ION", "non-polymer"),
        "R4K": (
            "(2S,3R)-2-AZANYL-3-HYDROXY-3-[(3S)-3-HYDROXY-2-OXO-1H-INDOL-3-YL]PROPANOIC ACID",
            "peptide",
        ),
        "D71": (
            "1-(3-NITROPHENYL)-3-(PYRIDIN-4-YLMETHYL)PYRIDO[2,3-D]PYRIMIDINE-2,4(1H,3H)-DIONE",
            "non-polymer",
        ),
        "FAH": ("FLUOROACETIC ACID", "non-polymer"),
        "LUF": (
            "(5S)-4-[(3E,7E)-4,8-DIMETHYL-10-(2,6,6-TRIMETHYLCYCLOHEX-1-EN-1-YL)DECA-3,7-DIEN-1-YL]-5-HYDROXYFURAN-2(5H)-ONE",
            "non-polymer",
        ),
        "LUG": (
            "N-(5-{3,4-DIFLUORO-2-[(2-FLUORO-4-IODOPHENYL)AMINO]PHENYL}-1,3,4-OXADIAZOL-2-YL)ETHANE-1,2-DIAMINE",
            "non-polymer",
        ),
        "D76": (
            "(2R)-4-[(6-CHLORO-2-NAPHTHYL)SULFONYL]-1-[(5-METHYL-4,5,6,7-TETRAHYDRO[1,3]THIAZOLO[5,4-C]PYRIDIN-2-YL)CARBONYL]PIPERAZINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "KE4": (
            "(4R)-5-AMINO-4-[({3-HYDROXY-2-METHYL-5-[(PHOSPHONOOXY)METHYL]PYRIDIN-4-YL}METHYL)AMINO]PENTANOIC ACID",
            "non-polymer",
        ),
        "LUM": ("LUMICHROME", "non-polymer"),
        "LUN": (
            "(5-FLUORO-1H-INDOL-2-YL)[(3R)-1'-[(3R)-PIPERIDIN-3-YL]SPIRO[INDOLE-3,3'-PYRROLIDIN]-1(2H)-YL]METHANONE",
            "non-polymer",
        ),
        "LUO": (
            "N-(4-METHOXYPHENYL)-N~2~-[1-(2-METHYLPHENYL)-4-OXO-4,5-DIHYDRO-1H-PYRAZOLO[3,4-D]PYRIMIDIN-6-YL]-L-ALANINAMIDE",
            "non-polymer",
        ),
        "977": ("2-AMINO-5-BROMO-6-PHENYLPYRIMIDIN-4-OL", "non-polymer"),
        "976": (
            "(2S)-2-(6-CHLORO-2-METHYL-4-PHENYLQUINOLIN-3-YL)PENTANOIC ACID",
            "non-polymer",
        ),
        "RNR": (
            "3,6-BIS(3-(3'-(R)-FLUOROPYRROLINDINO)PROPIONAMIDO)ACRIDINE",
            "non-polymer",
        ),
        "974": (
            "1-[2-[5-[AMINO(IMINO)METHYL]-2-HYDROXYPHENOXY]-6-[3-(4,5-DIHYDRO-1-METHYL-1H-IMIDAZOL-2-YL)PHENOXY] PYRIDIN-4-YL]PIPERIDINE-3-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "973": (
            "(5Z)-3-(4-CHLOROPHENYL)-4-HYDROXY-5-(1-NAPHTHYLMETHYLENE)FURAN-2(5H)-ONE",
            "non-polymer",
        ),
        "972": (
            "2-{5-[AMINO(IMINIO)METHYL]-6-CHLORO-1H-BENZIMIDAZOL-2-YL}-6-ISOBUTOXYBENZENOLATE",
            "non-polymer",
        ),
        "RMC": (
            "(2R)-2-({6-[BENZYL(METHYL)AMINO]-9-ISOPROPYL-9H-PURIN-2-YL}AMINO)BUTAN-1-OL",
            "non-polymer",
        ),
        "NS4": ("N-[2-(CYCLOHEXYLOXY)-4-NITROPHENYL]METHANESULFONAMIDE", "non-polymer"),
        "LNO": ("L-LEUCYL-HYDROXYLAMINE", "non-polymer"),
        "979": (
            "5,6-DIPHENYL-N-(2-PIPERAZIN-1-YLETHYL)FURO[2,3-D]PYRIMIDIN-4-AMINE",
            "non-polymer",
        ),
        "RMA": (
            "N-[(1S)-2,3-DIHYDRO-1H-INDEN-1-YL]-N-METHYL-N-PROP-2-YNYLAMINE",
            "non-polymer",
        ),
        "IG": ("ISOGUANOSINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "PDK": (
            "(10S,13R)-3-{2-[{2-[BIS(CARBOXYMETHYL)AMINO]ETHYL}(CARBOXYMETHYL)AMINO]ETHYL}-10-HYDROXY-5,16-DIOXO-13-(TETRADECANOYLOXY)-9,11,15-TRIOXA-3,6-DIAZA-10-PHOSPHANONACOSAN-1-OIC ACID 10-OXIDE",
            "non-polymer",
        ),
        "MO4": ("MAGNESIUM ION, 4 WATERS COORDINATED", "non-polymer"),
        "MX4": ("{1-[4-(BENZYLOXY)-4-OXOBUTANOYL]HYDRAZINO}ACETIC ACID", "non-polymer"),
        "RFL": (
            "8-DEMETHYL-8-DIMETHYLAMINO-FLAVIN-ADENINE-DINUCLEOTIDE",
            "non-polymer",
        ),
        "ZPC": (
            "(5R)-6-(5-CHLOROPYRIDIN-2-YL)-7-OXO-6,7-DIHYDRO-5H-PYRROLO[3,4-B]PYRAZIN-5-YL 4-METHYLPIPERAZINE-1-CARBOXYLATE",
            "non-polymer",
        ),
        "ECQ": ("(3S)-3-[(2S)-2-AMINO-3-HYDROXYBUTYL]PYRROLIDIN-2-ONE", "non-polymer"),
        "NFT": (
            "N-(2-AMINOETHYL)-N~2~-{(1S)-1-[4'-(AMINOSULFONYL)BIPHENYL-4-YL]-2,2,2-TRIFLUOROETHYL}-L-LEUCINAMIDE",
            "non-polymer",
        ),
        "BB8": ("(2S,3S)-BETA-HYDROXY-PHENYLALANINE", "peptide"),
        "TB": ("TERBIUM(III) ION", "non-polymer"),
        "LNE": ("N-ETHYL-L-LEUCINAMIDE", "peptide"),
        "97B": (
            "9-CHLORO-7-(2,6-DIFLUOROPHENYL)-N-{4-[(4-METHYLPIPERAZIN-1-YL)CARBONYL]PHENYL}-5H-PYRIMIDO[5,4-D][2]BENZAZEPIN-2-AMINE",
            "non-polymer",
        ),
        "PDJ": (
            "(2R)-3-{[(2R)-2-AMINO-3-HYDROXYPROPYL]THIO}PROPANE-1,2-DIYL DIHEXADECANOATE",
            "non-polymer",
        ),
        "BB6": ("(2Z)-2-AMINO-3-SULFANYLBUT-2-ENOIC ACID", "peptide"),
        "SDP": ("2-AMINO-3-(DIETHOXY-PHOSPHORYLOXY)-PROPIONIC ACID", "peptide"),
        "QFG": (
            "{(4E)-2-[(1E)-4-AMINO-4-OXOBUTANIMIDOYL]-4-BENZYLIDENE-5-OXO-4,5-DIHYDRO-1H-IMIDAZOL-1-YL}ACETIC ACID",
            "peptide",
        ),
        "KEN": ("N,N-DIMETHYLMETHANAMINE", "non-polymer"),
        "KEM": (
            "(3E,5R)-8-[(1S,3AR,4R,7AS)-1-HYDROXY-7A-METHYL-5-OXOOCTAHYDRO-1H-INDEN-4-YL]-5-METHYL-2,6-DIOXOOCT-3-ENOATE",
            "non-polymer",
        ),
        "BB4": ("2-(5-BROMO-1H-INDOL-3-YL)-N-HYDROXYACETAMIDE", "non-polymer"),
        "KEK": (
            "(3E,5R)-8-(2-CHLOROPHENYL)-5-METHYL-2,6-DIOXOOCT-3-ENOATE",
            "non-polymer",
        ),
        "KEG": ("12-TUNGSTOPHOSPHATE", "non-polymer"),
        "KEF": ("CEFUROXIME (OCT-3-ENE FORM)", "non-polymer"),
        "RMO": ("[ARSENOTHIONITO(2-)-KAPPA~2~O,S](OXO)MOLYBDENUM", "non-polymer"),
        "R45": (
            "N-{(2S)-3-[(S)-(4-BROMOPHENYL)(HYDROXY)PHOSPHORYL]-2-[(3-PHENYL-1,2-OXAZOL-5-YL)METHYL]PROPANOYL}-L-ALPHA-GLUTAMYL-L-ALPHA-GLUTAMINE",
            "peptide",
        ),
        "R47": (
            "N-[(2S)-3-[(S)-(4-BROMOPHENYL)(HYDROXY)PHOSPHORYL]-2-{[3-(3'-CHLOROBIPHENYL-4-YL)-1,2-OXAZOL-5-YL]METHYL}PROPANOYL]-L-ALPHA-GLUTAMYL-L-ALPHA-GLUTAMINE",
            "peptide",
        ),
        "R46": (
            "N-{6-[4-(6-BROMO-1,2-BENZISOTHIAZOL-3-YL)PHENOXY]HEXYL}-N-METHYL-2-PROPEN-1-AMINE",
            "non-polymer",
        ),
        "PDI": ("PHOSPHORIC ACID MONO-(3-HYDROXY-PROPYL) ESTER", "non-polymer"),
        "LNC": (
            "(3S)-5-(3-CARBOXY-3-HYDROXYPROPYL) NICOTINAMIDE-ADENINE-DINUCLEOTIDE",
            "non-polymer",
        ),
        "KEU": (
            "N-{4-[(1R)-4-[(2R,4R,5S)-2,4-DIAMINO-6-OXOHEXAHYDROPYRIMIDIN-5-YL]-1-(2,2,2-TRIFLUORO-1,1-DIHYDROXYETHYL)BUTYL]BENZOYL}-D-GLUTAMIC ACID",
            "non-polymer",
        ),
        "ECX": ("S-ETHYL-L-CYSTEINE", "peptide"),
        "R49": (
            "1-{4-[(6-AMINOQUINOLIN-4-YL)AMINO]PHENYL}-3-[3-TERT-BUTYL-1-(3-METHYLPHENYL)-1H-PYRAZOL-5-YL]UREA",
            "non-polymer",
        ),
        "GN8": (
            "(Z)-3-BENZYL-5-(2-HYDROXY-3-NITROBENZYLIDENE)-2-THIOXOTHIAZOLIDIN-4-ONE",
            "non-polymer",
        ),
        "P6P": ("6-O-PHOSPHONO-ALPHA-D-FRUCTOFURANOSE", "saccharide"),
        "NFR": ("NI-FE REDUCED ACTIVE CENTER", "non-polymer"),
        "P6T": ("1,6-DI-O-PHOSPHONO-D-TAGATOSE", "non-polymer"),
        "SDX": ("S-CITRYLDETHIA COENZYME A", "non-polymer"),
        "P6C": ("2-AMINO-4-ISOPROPYL-PTERIDINE-6-CARBOXYLIC ACID", "non-polymer"),
        "BBL": ("N-[(BENZYLOXY)CARBONYL]-L-ALANINE", "non-polymer"),
        "SDZ": (
            "1-[2-HYDROXY-3-(4-CYCLOHEXYL-PHENOXY)-PROPYL]-4-(2-PYRIDYL)-PIPERAZINE",
            "non-polymer",
        ),
        "P6G": ("HEXAETHYLENE GLYCOL", "non-polymer"),
        "P6F": ("1,6-DI-O-PHOSPHONO-D-FRUCTOSE", "non-polymer"),
        "P6L": (
            "(2S)-3-{[{[(2S)-2,3-DIHYDROXYPROPYL]OXY}(HYDROXY)PHOSPHORYL]OXY}-2-[(6E)-HEXADEC-6-ENOYLOXY]PROPYL (8E)-OCTADEC-8-ENOATE",
            "non-polymer",
        ),
        "NFQ": ("N-[(E)-IMINOMETHYL]-L-ASPARTIC ACID", "non-polymer"),
        "5PL": (
            "(1R,4S,6R)-6-({[2-(ACETYLAMINO)-2-DEOXY-ALPHA-D-GLUCOPYRANOSYL]OXY}METHYL)-4-HYDROXY-1-{[(15-METHYLHEXADECANOYL)OXY]METHYL}-4-OXIDO-7-OXO-3,5-DIOXA-8-AZA-4-PHOSPHAHEPTACOS-1-YL 15-METHYLHEXADECANOATE",
            "non-polymer",
        ),
        "5PM": (
            "4,4'-SULFANEDIYLBIS{N-[(1E)-PYRIDIN-2-YLMETHYLIDENE]ANILINE}",
            "non-polymer",
        ),
        "RM4": ("6-DEOXY-BETA-L-MANNOPYRANOSE", "non-polymer"),
        "5PH": (
            "5-[5-(1-CARBOXYMETHYL-2-OXO-PROPYLCARBAMOYL)-5-PHENYL-PENTYLSULFAMOYL]-2-HYDROXY-BENZOIC ACID",
            "non-polymer",
        ),
        "5PI": (
            "N-(ETHYLSULFONYL)-5-PROPOXY-L-TRYPTOPHYL-N~1~-{4-[AMINO(IMINO)METHYL]BENZYL}-L-GLUTAMAMIDE",
            "non-polymer",
        ),
        "5PG": ("(2S)-(4-HYDROXYPHENYL)(METHYLAMINO)ETHANOIC ACID", "peptide"),
        "BBH": (
            "1-BENZYL-3-(4-METHOXY-BENZENESULFONYL)-6-OXO-HEXAHYDRO-PYRIMIDINE-4-CARBOXYLIC ACID HYDROXYAMIDE",
            "non-polymer",
        ),
        "5PB": (
            "4-[4-(BENZYLOXY)PHENYL]-5-{[2-(4-CHLOROPHENYL)-2-OXOETHYL]SULFANYL}-2,4-DIHYDRO-3H-1,2,4-TRIAZOL-3-ONE",
            "non-polymer",
        ),
        "5PC": ("5(1-PROPYNYL)-2'-DEOXYCYTIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "5PA": (
            "N-[3-HYDROXY-2-METHYL-5-PHOSPHONOOXYMETHYL-PYRIDIN-4-Y-LMETHYL]-1-AMINO-CYCLOPROPANECARBOXYLIC ACID",
            "non-polymer",
        ),
        "PDG": (
            "N-({3-HYDROXY-2-METHYL-5-[(PHOSPHONOOXY)METHYL]PYRIDIN-4-YL}METHYL)-D-GLUTAMIC ACID",
            "non-polymer",
        ),
        "LN5": ("N5-(1-IMINOPROPYL)-L-ORNITHINE", "non-polymer"),
        "BBF": ("1-BROMO-2,3,4,5,6-PENTAFLUOROBENZENE", "non-polymer"),
        "5PV": ("5-PHENYLVALERIC ACID", "non-polymer"),
        "6DE": (
            "1-(2-CHLOROPHENYL)-3,5-DIMETHYL-1H-PYRAZOLE-4-CARBOXYLIC ACID ETHYL ESTER",
            "non-polymer",
        ),
        "5PR": (
            "5-HYDROXY-4-(HYDROXYMETHYL)-6-METHYLPYRIDINE-3-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "LN7": ("N~5~-[(1S)-1-AMINOPENTYL]-L-ORNITHINE", "non-polymer"),
        "5PP": ("5-PENTYL-2-PHENOXYPHENOL", "non-polymer"),
        "SDA": ("4-AMINO-3,5-DICHLOROBENZENESULFONAMIDE", "non-polymer"),
        "IP9": (
            "(2R)-3-{[(R)-{[(1S,2S,3R,4S,5S,6S)-2,6-DIHYDROXY-3,4,5-TRIS(PHOSPHONOOXY)CYCLOHEXYL]OXY}(HYDROXY)PHOSPHORYL]OXY}PROPANE-1,2-DIYL DIOCTANOATE",
            "non-polymer",
        ),
        "LN6": ("N~5~-[(1E)-PENTANIMIDOYL]-L-ORNITHINE", "non-polymer"),
        "SDB": (
            "(4S)-2-[(1S)-1-AMINO-2-HYDROXYETHYL]-3,4,5,6-TETRAHYDROPYRIMIDINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "LN1": (
            "(3R)-4-{[(3,4-DIHYDROXYPHENYL)ACETYL]OXY}-N-(2-FORMYLINDOLIZIN-3-YL)-3-SULFINO-D-VALINE",
            "non-polymer",
        ),
        "I3S": ("(1S,3S,4S)-1,3,4-TRIPHOSPHO-MYO-INOSITOL", "non-polymer"),
        "LN0": (
            "METHYL 6,8-DIDEOXY-6-{[(4R)-1-METHYL-4-PROPYL-L-PROLYL]AMINO}-1-THIO-D-GLYCERO-ALPHA-D-GALACTO-OCTOPYRANOSID-7-ULOSE",
            "non-polymer",
        ),
        "SDF": (
            "[(R)-(2,4-DICHLOROPHENYL)(SULFANYL)METHYL]PHOSPHONIC ACID",
            "non-polymer",
        ),
        "IUM": ("URANYL (VI) ION", "non-polymer"),
        "5P5": (
            "(2R)-3-{[(S)-HYDROXY{[(1R,2R,3R,4R,5S,6R)-2,3,4,6-TETRAHYDROXY-5-(PHOSPHONOOXY)CYCLOHEXYL]OXY}PHOSPHORYL]OXY}PROPANE-1,2-DIYL DIBUTANOATE",
            "non-polymer",
        ),
        "NHP": ("4-(2-AMINOPHENYLTHIO)-BUTYLPHOSPHONIC ACID", "non-polymer"),
        "5P3": ("5-CHLORO-4-HYDROXYBIPHENYL-3-CARBOXYLIC ACID", "non-polymer"),
        "OTH": ("N,O-DIMETHYL-L-THREONINE", "peptide"),
        "P63": (
            "N-(3-(8-CYANO-4-(PHENYLAMINO)PYRAZOLO[1,5-A][1,3,5]TRIAZIN-2-YLAMINO)PHENYL)ACETAMIDE",
            "non-polymer",
        ),
        "OTK": (
            "(3S,3AR,4S,6S,6AS,8R,9R,9AR,9BS)-6-(ACETYLOXY)-4-(BUTANOYLOXY)-3,3A-DIHYDROXY-3,6,9-TRIMETHYL-2-OXODODECAHYDROAZULENO[4,5-B]FURAN-8-YL (2Z)-2-METHYLBUT-2-ENOATE",
            "non-polymer",
        ),
        "P65": (
            "2,4-DIAMINO-6-METHYL-5-[3-(2,4,5-TRICHLOROPHENOXY)PROPYLOXY]PYRIMIDINE",
            "non-polymer",
        ),
        "P66": (
            "4-CHLORO-N-CYCLOPROPYL-3-{[1-(2,6-DIFLUOROPHENYL)-1H-PYRAZOLO[3,4-D]PYRIDAZIN-4-YL]AMINO}BENZAMIDE",
            "non-polymer",
        ),
        "OTA": (
            "2-(OXALYL-AMINO)-4,5,6,7-TETRAHYDRO-THIENO[2,3-C]PYRIDINE-3-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "OTB": ("TERTIARY-BUTOXYFORMIC ACID", "peptide"),
        "FT6": ("6-FLUORO-L-TRYPTOPHAN", "peptide"),
        "OTD": (
            "2-OXO-1,2,3,6-TETRAHYDROPYRIMIDINE-4,6-DICARBOXYLIC ACID",
            "non-polymer",
        ),
        "OTE": ("2-{2-[2-(2-OCTYLOXY-ETHOXY)-ETHOXYL]-ETHOXY}ETHANOL", "non-polymer"),
        "OTG": ("ORTHO-TOLUOYLGLUCOSAMINE", "non-polymer"),
        "OTY": (
            "2-AMINO-3-(4-HYDROXY-6-OXOCYCLOHEXA-1,4-DIENYL)PROPANOIC ACID",
            "peptide",
        ),
        "LI5": ("2-HYDROXY-4,5-DIOXOHEPTYL HYDROGEN PHOSPHONATE", "non-polymer"),
        "OTP": (
            "(2E,6E,10E,14E,18E,22E,26E)-3,7,11,15,19,23,27,31-OCTAMETHYLDOTRIACONTA-2,6,10,14,18,22,26,30-OCTAENYL TRIHYDROGEN DIPHOSPHATE",
            "non-polymer",
        ),
        "OFE": ("IRON(II) OXIDE", "non-polymer"),
        "K2N": (
            "ETHYL [(2S)-5-AMINO-2-METHYL-3-PHENYL-1,2-DIHYDROPYRIDO[3,4-B]PYRAZIN-7-YL]CARBAMATE",
            "non-polymer",
        ),
        "OTT": ("(2E,4E,6E)-OCTA-2,4,6-TRIENOIC ACID", "non-polymer"),
        "OTU": ("3-O-ALPHA-D-GLUCOPYRANOSYL-D-FRUCTOSE", "non-polymer"),
        "EC2": (
            "(2R,3R)-8-CYCLOPROPYL-7-(NAPHTHALEN-1-YLMETHYL)-5-OXO-2-PHENYL-2,3-DIHYDRO-5H-[1,3]THIAZOLO[3,2-A]PYRIDINE-3-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "2PQ": (
            "2-[(1-{3-[4-(BIPHENYL-4-YLCARBONYL)-2-PROPYLPHENOXY]PROPYL}-1,2,3,4-TETRAHYDROQUINOLIN-5-YL)OXY]-2-METHYLPROPANOIC ACID",
            "non-polymer",
        ),
        "2PP": ("2-PROPYL-PENTANOIC ACID", "non-polymer"),
        "AXP": (
            "4-ACETAMIDO-2,4-DIDEXOY-D-GLYCERO-BETA-D-GALACTO-OCTOPYRANOSYLPHOSPHONIC ACID (AN AXIAL PHOSPHONATE)",
            "saccharide",
        ),
        "2PR": (
            "2-AMINO-9-[2-DEOXYRIBOFURANOSYL]-9H-PURINE-5'-MONOPHOSPHATE",
            "nucleic-acid",
        ),
        "2PU": (
            "1-(5-OXO-2,3,5,9B-TETRAHYDRO-1H-PYRROLO[2,1-A]ISOINDOL-9-YL)-3-(5-PYRROLIDIN-2-YL-1H-PYRAZOL-3-YL)-UREA",
            "non-polymer",
        ),
        "2PT": (
            "1-[2-(ACRIDIN-9-YLAMINO)ETHYL]-1,3-DIMETHYLTHIOUREA-PLATINUM(II)-ETHANE-1,2-DIAMINE",
            "non-polymer",
        ),
        "AXT": ("ASTAXANTHIN", "non-polymer"),
        "MZ3": (
            "N-[(1S,2R)-1-BENZYL-2-HYDROXY-3-{[(3-METHOXYPHENYL)SULFONYL](2-THIENYLMETHYL)AMINO}PROPYL]-3-FLUORO-2-METHYLBENZAMIDE",
            "non-polymer",
        ),
        "2PY": (
            "(2S)-1-{[5-(1H-INDAZOL-5-YL)PYRIDIN-3-YL]OXY}-3-[(7AS)-7AH-INDOL-3-YL]PROPAN-2-AMINE",
            "non-polymer",
        ),
        "AXX": (
            "2-[(CYCLOPROPYLCARBONYL)AMINO]-4,5,6,7-TETRAHYDRO-1-BENZOTHIOPHENE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "MZ8": (
            "N-{(1S,2R)-3-[(1,3-BENZOTHIAZOL-6-YLSULFONYL)(ISOBUTYL)AMINO]-1-BENZYL-2-HYDROXYPROPYL}-3-HYDROXYBENZAMIDE",
            "non-polymer",
        ),
        "MZ9": (
            "N-[(1S,2R)-1-BENZYL-2-HYDROXY-3-{[(4-METHOXYPHENYL)SULFONYL][(2S)-2-METHYLBUTYL]AMINO}PROPYL]-4-OXOHEXANAMIDE",
            "non-polymer",
        ),
        "2PA": ("DIAMIDOPHOSPHATE", "non-polymer"),
        "AXC": (
            "5-(1H-BENZIMIDAZOL-2-YLMETHYLSULFANYL)-2-METHYL-[1,2,4]TRIAZOLO[1,5-C]QUINAZOLINE",
            "non-polymer",
        ),
        "2PC": ("3,4-DIHYDRO-2H-PYRROLIUM-5-CARBOXYLATE", "non-polymer"),
        "2PB": ("2-[(DIOXIDOPHOSPHINO)OXY]BENZOATE", "non-polymer"),
        "2PE": ("NONAETHYLENE GLYCOL", "non-polymer"),
        "2PD": (
            "5-HYDROXY-4-(7-METHOXY-1,1-DIOXIDO-2H-1,2,4-BENZOTHIADIAZIN-3-YL)-2-(3-METHYLBUTYL)-6-PHENYLPYRIDAZIN-3(2H)-ONE",
            "non-polymer",
        ),
        "2PG": ("2-PHOSPHOGLYCERIC ACID", "non-polymer"),
        "2PF": ("?", "non-polymer"),
        "2PI": ("2-AMINO-PENTANOIC ACID", "peptide"),
        "2PH": ("[7,8-DIHYDRO-PTERIN-6-YL METHANYL]-PHOSPHONOPHOSPHATE", "non-polymer"),
        "AXI": ("AXITINIB", "non-polymer"),
        "2PM": ("N-[2-(BENZHYDRYLOXY)ETHYL]-N,N-DIMETHYLAMINE", "non-polymer"),
        "2PL": ("PHOSPHOGLYCOLIC ACID", "non-polymer"),
        "AXL": (
            "2-{1-[2-AMINO-2-(4-HYDROXY-PHENYL)-ACETYLAMINO]-2-OXO-ETHYL}-5,5-DIMETHYL-THIAZOLIDINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "AXM": (
            "N-HYDROXY-2-[(3S,5S,7S)-TRICYCLO[3.3.1.1~3,7~]DEC-1-YL]ACETAMIDE",
            "non-polymer",
        ),
        "3JZ": (
            "N-METHYL-N-{2-[({2-[(2-OXO-2,3-DIHYDRO-1H-INDOL-5-YL)AMINO]-5-(TRIFLUOROMETHYL)PYRIMIDIN-4-YL}AMINO)METHYL]PHENYL}METHANESULFONAMIDE",
            "non-polymer",
        ),
        "9OH": ("(TRANS-12,13-EPOXY)-9-HYDROXY-10(E)-OCTADECENOIC ACID", "non-polymer"),
        "9OD": ("(2Z)-9-OXODEC-2-ENOIC ACID", "non-polymer"),
        "UGY": ("(2S)-AMINO(CARBAMOYLAMINO)ETHANOIC ACID", "peptide"),
        "3J8": ("5-METHYLPHENAZIN-1(5H)-ONE", "non-polymer"),
        "NNY": ("1-(4-CHLOROPYRIDIN-2-YL)METHANAMINE", "non-polymer"),
        "44C": ("?", "non-polymer"),
        "44B": (
            "1,1,1,3,3,3-HEXAFLUORO-2-{4-[(2,2,2-TRIFLUOROETHYL)AMINO]PHENYL}PROPAN-2-OL",
            "non-polymer",
        ),
        "AX2": ("1,3,5-TRIAZINE-2,4,6-TRIAMINE", "non-polymer"),
        "2P0": (
            "1-({[(S)-HYDROXY(PHOSPHONOOXY)PHOSPHORYL]OXY}ACETYL)-L-PROLINE",
            "peptide",
        ),
        "AX1": (
            "2-AMINO-6-(1,3-BENZODIOXOL-5-YL)-4-OXO-4,7-DIHYDRO-3H-PYRROLO[2,3-D]PYRIMIDINE-5-CARBONITRILE",
            "non-polymer",
        ),
        "AX6": ("6-[(4-METHOXYBENZYL)SULFANYL]PYRIMIDINE-2,4-DIAMINE", "non-polymer"),
        "2P4": (
            "(5R)-N-BENZYL-5-({(2S)-2-[(1R)-1,2-DIHYDROXYETHYL]PYRROLIDIN-1-YL}CARBONYL)CYCLOPENT-1-ENE-1-CARBOXAMIDE",
            "non-polymer",
        ),
        "AX4": ("6-[(4-METHYLPHENYL)SULFANYL]PYRIMIDINE-2,4-DIAMINE", "non-polymer"),
        "AX5": ("6-(BENZYLSULFANYL)PYRIMIDINE-2,4-DIAMINE", "non-polymer"),
        "2P8": (
            "(2S)-N-BENZYL-2-({(2S)-2-[(1R)-1,2-DIHYDROXYETHYL]PYRROLIDIN-1-YL}CARBONYL)PYRROLIDINE-1-CARBOXAMIDE",
            "non-polymer",
        ),
        "AX8": ("1-(3,4-DICHLOROBENZYL)-1H-BENZIMIDAZOL-2-AMINE", "non-polymer"),
        "D5C": (
            "N-[(3R)-3-AMINO-4-(2,4,5-TRIFLUOROPHENYL)BUTYL]-3'-(TRIFLUOROMETHYL)BIPHENYL-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "MZX": ("5,8-DIMETHOXY-4-METHYLQUINOLIN-2(1H)-ONE", "non-polymer"),
        "MZZ": ("(4-AMINOPHENYL)-IMIDAZOL-1-YL-METHANONE", "non-polymer"),
        "D5G": (
            "(2S)-4-METHYL-2-(3-PHENYLTHIOUREIDO)-N-((3S)-TETRAHYDRO-2-HYDROXY-3-FURANYL)PENTANAMIDE",
            "non-polymer",
        ),
        "EPM": ("N-PALMITOYL-L-METHIONINE", "non-polymer"),
        "MZA": (
            "6-CHLORO-N-[(3S)-1-(5-FLUORO-1,2,3,4-TETRAHYDROISOQUINOLIN-6-YL)-2-OXO-PYRROLIDIN-3-YL]NAPHTHALENE-2-SULFONAMIDE",
            "non-polymer",
        ),
        "MZB": (
            "(1R,2R,3R,4S,5R)-4-AMINO-5-[(R)-METHYLSULFINYL]CYCLOPENTANE-1,2,3-TRIOL",
            "non-polymer",
        ),
        "MZC": (
            "METHYL {3-[2-(ACETYLAMINO)ETHYL]-1H-INDOL-5-YL}CARBAMATE",
            "non-polymer",
        ),
        "MZL": (
            "N-[4-(5-ETHYL-3-PYRIDIN-3-YL-1H-PYRAZOL-1-YL)PHENYL]PYRIDINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "MZM": ("5-ACETAMIDO-4-METHYL-1,3,4-THIADIAZOLE-2-SULFONAMIDE", "non-polymer"),
        "MZO": ("N-(3-METHOXYPHENYL)-9H-PURIN-6-AMINE", "non-polymer"),
        "MZH": ("1,1'-(OXYDIMETHANEDIYL)DIPYRROLIDINE-2,5-DIONE", "non-polymer"),
        "319": (
            "N-({4-[(2-AMINOPYRIDIN-4-YL)OXY]-3-FLUOROPHENYL}CARBAMOYL)-2-(4-FLUOROPHENYL)ACETAMIDE",
            "non-polymer",
        ),
        "318": (
            "N'-[(1S,2S)-2-[(4S)-1-BENZYL-5-OXOIMIDAZOLIDIN-4-YL]-1-(3,5-DIFLUOROBENZYL)-2-HYDROXYETHYL]-5-METHYL-N,N-DIPROPYLBENZENE-1,3-DICARBOXAMIDE",
            "non-polymer",
        ),
        "0ZY": (
            "N-(TERT-BUTOXYCARBONYL)-L-ALANYL-N-{(1R)-5-AMMONIO-1-[HYDROXY(METHOXY)BORANYL]PENTYL}-L-VALINAMIDE",
            "peptide",
        ),
        "0ZX": (
            "N-(TERT-BUTOXYCARBONYL)-L-ALANYL-N-[(1S)-5-AMINO-1-(DIETHOXYBORANYL)PENTYL]-L-VALINAMIDE",
            "peptide",
        ),
        "0ZZ": (
            "5-S-BENZYL-3-({N-[(5-BROMO-2-METHOXYPHENYL)ACETYL]-L-VALYL}AMINO)-2,3-DIDEOXY-5-THIO-D-ERYTHRO-PENTONIC ACID",
            "peptide",
        ),
        "313": ("(DIMETHYL-LAMBDA~4~-SULFANYL)ACETIC ACID", "non-polymer"),
        "0ZT": (
            "N-{(2S,3S)-3-[(TERT-BUTOXYCARBONYL)AMINO]-2-HYDROXY-4-PHENYLBUTYL}-L-PHENYLALANYL-L-ALPHA-GLUTAMYL-L-PHENYLALANINAMIDE",
            "peptide",
        ),
        "0ZW": (
            "N-(TERT-BUTOXYCARBONYL)-L-ALANYL-N-{(1S)-5-AMMONIO-1-[HYDROXY(3-HYDROXYPROPOXY)BORANYL]PENTYL}-L-VALINAMIDE",
            "peptide",
        ),
        "310": (
            "N-CARBAMIMIDOYL-2-[2-(2-CHLOROPHENYL)-5-[4-(4-ETHANOYLPHENOXY)PHENYL]PYRROL-1-YL]ETHANAMIDE",
            "non-polymer",
        ),
        "0ZQ": (
            "N-{(3S)-3-[(TERT-BUTOXYCARBONYL)AMINO]-4-PHENYLBUTYL}-L-PHENYLALANYL-L-ALPHA-GLUTAMYL-L-PHENYLALANINAMIDE",
            "peptide",
        ),
        "0ZP": (
            "N-[(1R,2R,4R)-1-(CYCLOHEXYLMETHYL)-2-HYDROXY-6-METHYL-4-{[(2R)-2-METHYLBUTYL]CARBAMOYL}HEPTYL]-3-(1H-IMIDAZOL-3-IUM-4-YL)-N~2~-[3-NAPHTHALEN-1-YL-2-(NAPHTHALEN-1-YLMETHYL)PROPANOYL]-L-ALANINAMIDE",
            "peptide",
        ),
        "0ZS": (
            "N-{(2R,3S)-3-[(TERT-BUTOXYCARBONYL)AMINO]-2-HYDROXY-4-PHENYLBUTYL}-L-PHENYLALANYL-L-ALPHA-GLUTAMYL-L-PHENYLALANINAMIDE",
            "peptide",
        ),
        "0ZR": (
            "N-{(2R,3S)-3-[(TERT-BUTOXYCARBONYL)AMINO]-2-HYDROXY-4-PHENYLBUTYL}-L-PHENYLALANYL-L-GLUTAMINYL-L-PHENYLALANINAMIDE",
            "peptide",
        ),
        "ZBZ": ("S-(PHENYLCARBONYL)-L-CYSTEINE", "peptide"),
        "0ZL": (
            "N-(ETHOXYCARBONYL)-L-LEUCYL-N-[(1R,2S,3S)-1-(CYCLOHEXYLMETHYL)-2,3-DIHYDROXY-5-METHYLHEXYL]-L-LEUCINAMIDE",
            "peptide",
        ),
        "0ZO": (
            "N-(3-CARBOXYPROPANOYL)-L-ALANYL-L-ALANYL-N-[(2S,3S)-4-CHLORO-3-HYDROXY-1-PHENYLBUTAN-2-YL]-L-PROLINAMIDE",
            "peptide",
        ),
        "0ZN": ("N-[(1R)-1-CARBOXY-3-PHENYLPROPYL]-L-LEUCYL-L-TRYPTOPHAN", "peptide"),
        "0ZI": (
            "AMINO{[(4S)-4-({[5-(DIMETHYLAMINO)NAPHTHALEN-1-YL]SULFONYL}AMINO)-5-(4-ETHYLPIPERIDIN-1-YL)-5-OXOPENTYL]AMINO}METHANIMINIUM",
            "peptide",
        ),
        "0ZJ": (
            "N-(SULFANYLACETYL)-D-PHENYLALANYL-N-[(2S,3S)-6-{[AMINO(IMINIO)METHYL]AMINO}-1-CHLORO-2-HYDROXYHEXAN-3-YL]-L-PROLINAMIDE",
            "peptide",
        ),
        "0ZE": (
            "[[[(4S,5S)-4-[[(3S,6S,8AR)-6-AZANYL-5-OXO-6-(PHENYLMETHYL)-1,2,3,7,8,8A-HEXAHYDROINDOLIZIN-3-YL]CARBONYLAMINO]-5-(1,3-BENZOTHIAZOL-2-YL)-5-HYDROXY-PENTYL]AMINO]-AZANYL-METHYLIDENE]AZANIUM",
            "peptide",
        ),
        "0ZG": (
            "3-[(2S)-2-{[(4-METHYLPHENYL)SULFONYL]AMINO}-3-OXO-3-PIPERIDIN-1-YLPROPYL]BENZENECARBOXIMIDAMIDE",
            "peptide",
        ),
        "0ZA": (
            "3-{(1S)-1-[(6-CHLORO-3,3-DIMETHYL-3,4-DIHYDROISOQUINOLIN-1-YL)AMINO]-2-PHENYLETHYL}-1,2,4-OXADIAZOL-5(2H)-ONE",
            "non-polymer",
        ),
        "0ZC": (
            "(3R)-2-[N-(FURAN-2-YLCARBONYL)-L-LEUCYL]-2,3,4,9-TETRAHYDRO-1H-BETA-CARBOLINE-3-CARBOXYLIC ACID",
            "peptide",
        ),
        "0ZB": (
            "N-BENZYL-N~2~-[(2R)-2-(HYDROXYCARBAMOYL)-4-METHYLPENTANOYL]-L-ALPHA-ASPARAGINE",
            "peptide",
        ),
        "7MR": (
            "(2R)-2-AMINO-3,3,3-TRIFLUORO-N-HYDROXY-2-{[(4-PHENOXYPHENYL)SULFONYL]METHYL}PROPANAMIDE",
            "non-polymer",
        ),
        "7MQ": ("MENAQUINONE-7 (ALTERED)", "non-polymer"),
        "7MP": (
            "7-AMINO-1-METHYL-3-(2-METHYL-5-{[3-(TRIFLUOROMETHYL)BENZOYL]AMINO}PHENYL)-2-OXO-2,3-DIHYDROPYRIMIDO[4,5-D]PYRIMIDIN-1-IUM",
            "non-polymer",
        ),
        "NFX": (
            "7-[(3R)-3-AMINOPYRROLIDIN-1-YL]-8-CHLORO-1-CYCLOPROPYL-6-FLUORO-4-OXO-1,4-DIHYDROQUINOLINE-3-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "ZO0": (
            "4-(4-METHYL-1,4-DIAZEPAN-1-YL)-2-(TRIFLUOROMETHYL)QUINAZOLINE",
            "non-polymer",
        ),
        "FTY": ("DEOXY-DIFLUOROMETHELENE-PHOSPHOTYROSINE", "peptide"),
        "M99": (
            "(R)-3-(2-(3-CYCLOPROPOXY-4-(DIFLUOROMETHOXY)PHENYL)-2-(5-(1,1,1,3,3,3-HEXAFLUORO-2-HYDROXYPROPAN-2-YL)THIAZOL-2-YL)ETHYL)PYRIDINE 1-OXIDE",
            "non-polymer",
        ),
        "YES": (
            "3-(2-HYDROXYETHYL)-2,2-BIS(HYDROXYMETHYL)PENTANE-1,5-DIOL",
            "non-polymer",
        ),
        "7MN": (
            "(Z)-N-[(1E)-1-CARBOXY-2-(2,3-DIHYDRO-1H-INDOL-1-YL)ETHYLIDENE]{3-HYDROXY-2-METHYL-5-[(PHOSPHONOOXY)METHYL]PYRIDIN-4(1H)-YLIDENE}METHANAMINIUM",
            "peptide",
        ),
        "7MC": (
            "5'-O-[(R)-[(N-ACETYL-L-ALPHA-ASPARTYL)AMINO](3-AMINOPROPOXY)PHOSPHORYL]ADENOSINE",
            "non-polymer",
        ),
        "7MG": ("7N-METHYL-8-HYDROGUANOSINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "7ME": (
            "(7S)-7-METHYL-5,6,7,8-TETRAHYDROQUINAZOLINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "7MD": (
            "5'-O-[(R)-(3-AMINOPROPOXY)(L-ALPHA-ASPARTYLAMINO)PHOSPHORYL]ADENOSINE",
            "non-polymer",
        ),
        "SDN": (
            "[(3S)-9-HYDROXY-1-METHYL-10-OXO-4,10-DIHYDRO-3H-BENZO[G]ISOCHROMEN-3-YL]ACETIC ACID",
            "non-polymer",
        ),
        "LYU": ("N~6~-(4-METHYLPENTYL)-L-LYSINE", "peptide"),
        "447": (
            "N-(4-{[6-METHOXY-7-(3-MORPHOLIN-4-YLPROPOXY)QUINAZOLIN-4-YL]AMINO}PHENYL)BENZAMIDE",
            "non-polymer",
        ),
        "OFL": ("O-TRIFLUOROMETHYLPHENYL ANTHRANILIC ACID", "non-polymer"),
        "ZOM": (
            "[5-(4-CHLOROBENZOYL)-1,4-DIMETHYL-1H-PYRROL-2-YL]ACETIC ACID",
            "non-polymer",
        ),
        "ZON": ("1-(1,2-BENZOXAZOL-3-YL)METHANESULFONAMIDE", "non-polymer"),
        "C49": (
            "4-THIO,5-FLUORO,5-METHYL-2'-DEOXY-CYTIDINE-5'-MONOPHOSPHATE",
            "nucleic-acid",
        ),
        "ZOP": (
            "(3Z)-N-ETHYL-N-METHYL-2-OXO-3-(PHENYL{[4-(PIPERIDIN-1-YLMETHYL)PHENYL]AMINO}METHYLIDENE)-2,3-DIHYDRO-1H-INDOLE-6-CARBOXAMIDE",
            "non-polymer",
        ),
        "NIX": ("NALIDIXIC ACID", "non-polymer"),
        "0Z9": ("N-[(BENZYLOXY)CARBONYL]-L-LEUCYL-N-HYDROXY-L-ALANINAMIDE", "peptide"),
        "0Z4": (
            "6-AMMONIO-N-(TRIFLUOROACETYL)-L-NORLEUCYL-N-[4-(1-METHYLETHYL)PHENYL]-L-LEUCINAMIDE",
            "peptide",
        ),
        "0Z6": (
            "D-PHENYLALANYL-N-[(2S,3S)-6-{[AMINO(IMINIO)METHYL]AMINO}-1-CHLORO-2-HYDROXYHEXAN-3-YL]-L-PHENYLALANINAMIDE",
            "peptide",
        ),
        "0Z1": (
            "6-AMMONIO-N-(TRIFLUOROACETYL)-L-NORLEUCYL-N-[4-(1-METHYLETHYL)PHENYL]-L-PROLINAMIDE",
            "peptide",
        ),
        "0Z0": (
            "N-(TRIFLUOROACETYL)-L-PHENYLALANYL-N-[4-(TRIFLUOROMETHYL)PHENYL]-L-ALANINAMIDE",
            "peptide",
        ),
        "0Z3": (
            "6-AMMONIO-N-(TRIFLUOROACETYL)-L-NORLEUCYL-N-[4-(1-METHYLETHYL)PHENYL]-L-PHENYLALANINAMIDE",
            "peptide",
        ),
        "0Z2": (
            "N-(TRIFLUOROACETYL)-L-LEUCYL-N-[4-(TRIFLUOROMETHYL)PHENYL]-L-ALANINAMIDE",
            "peptide",
        ),
        "31I": (
            "(2E)-3-{5-[(2,4-DIAMINOPYRIMIDIN-5-YL)METHYL]-2,3-DIMETHOXYPHENYL}-1-[(1S)-1-(PROPAN-2-YL)PHTHALAZIN-2(1H)-YL]PROP-2-EN-1-ONE",
            "non-polymer",
        ),
        "VVV": ("5,6,8-TRIMETHOXY-4-METHYLQUINOLIN-2(1H)-ONE", "non-polymer"),
        "8DG": ("8-OXO-2'-DEOXYGUANOSINE-5'-TRIPHOSPHATE", "non-polymer"),
        "IK4": (
            "N-({7-[(4-CYANOBENZYL)OXY]NAPHTHALEN-2-YL}SULFONYL)-D-GLUTAMIC ACID",
            "non-polymer",
        ),
        "SXK": ("?", "non-polymer"),
        "4CY": ("NITRILO-L-METHIONINE", "peptide"),
        "MBG": ("METHYL-BETA-GALACTOSE", "saccharide"),
        "4CT": (
            "(3R,4S)-1-[(4-AMINO-5H-PYRROLO[3,2-D]PYRIMIDIN-7-YL)METHYL]-4-{[(4-CHLOROPHENYL)SULFANYL]METHYL}PYRROLIDIN-3-OL",
            "non-polymer",
        ),
        "4CR": (
            "(1R)-4-({[ETHYL(METHYL)AMINO]CARBONYL}OXY)-N-METHYL-N-[(1E)-PROP-2-EN-1-YLIDENE]INDAN-1-AMINIUM",
            "non-polymer",
        ),
        "4CS": (
            "(4S)-2-METHYL-1,4,5,6-TETRAHYDROPYRIMIDINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "4CP": (
            "2-[2-(4-CHLORO-PHENYLSULFANYL)-ACETYLAMINO]-3-(4-GUANIDINO-PHENYL)-PROPIONAMIDE",
            "non-polymer",
        ),
        "4CO": ("4-HYDROXYPHENACYL COENZYME A", "non-polymer"),
        "4CL": ("4-CHLOROBENZENE-1,2-DIOL", "non-polymer"),
        "8DD": ("8-OXO-7,8-DIHYDRO-2'-DEOXY-ADENOSINE-5'-DIPHOSPHATE", "non-polymer"),
        "PYC": ("PYRROLE-2-CARBOXYLATE", "non-polymer"),
        "R8G": (
            "2,7-BIS[3-(4-METHYLPIPERAZIN-1-YL)PROPYL]-4,9-BIS{[3-(4-METHYLPIPERAZIN-1-YL)PROPYL]AMINO}BENZO[LMN][3,8]PHENANTHROLINE-1,3,6,8(2H,7H)-TETRONE",
            "non-polymer",
        ),
        "4CH": ("4-CHLOROPHENOL", "non-polymer"),
        "4CF": ("4-CYANO-L-PHENYLALANINE", "peptide"),
        "4CD": (
            "4-CHLORO-6-{[(2R)-2-(2-METHYLPHENYL)PYRROLIDIN-1-YL]CARBONYL}BENZENE-1,3-DIOL",
            "non-polymer",
        ),
        "4CB": ("4-CARBOXYPHENYLBORONIC ACID", "non-polymer"),
        "4CA": ("4-HYDROXYBENZYL COENZYME A", "non-polymer"),
        "RN8": ("?", "non-polymer"),
        "2MT": ("(4R)-2,2-DIMETHYL-1,3-THIAZOLIDINE-4-CARBOXYLIC ACID", "peptide"),
        "K2D": (
            "(3R,3AS,6AR)-HEXAHYDROFURO[2,3-B]FURAN-3-YL [(1S,2R)-1-BENZYL-3-{(2-ETHYLBUTYL)[(4-METHOXYPHENYL)SULFONYL]AMINO}-2-HYDROXYPROPYL]CARBAMATE",
            "non-polymer",
        ),
        "N5Y": (
            "N-(6-AMINO-1-BENZYL-2,4-DIOXO-1,2,3,4-TETRAHYDROPYRIMIDIN-5-YL)-N-METHYLBENZAMIDE",
            "non-polymer",
        ),
        "ITT": ("INOSINE TRIPHOSPHATE", "non-polymer"),
        "1D4": (
            "CYCLOTRIDECYL 3,4,6-TRIDEOXY-3-(DIMETHYLAMINO)-BETA-D-XYLO-HEXOPYRANOSIDE",
            "non-polymer",
        ),
        "1D2": (
            "CYCLODODECYL 3,4,6-TRIDEOXY-3-(DIMETHYLAMINO)-BETA-D-XYLO-HEXOPYRANOSIDE",
            "non-polymer",
        ),
        "XM5": (
            "6-METHOXY-9-METHYL[1,3]DIOXOLO[4,5-H]QUINOLIN-8(9H)-ONE",
            "non-polymer",
        ),
        "FTN": (
            "N-[3-(1-METHYLETHOXY)PHENYL]-2-(TRIFLUOROMETHYL)BENZAMIDE",
            "non-polymer",
        ),
        "N5T": (
            "(2S)-4-(2,5-DIFLUOROPHENYL)-N,N-DIMETHYL-2-PHENYL-2,5-DIHYDRO-1H-PYRROLE-1-CARBOXAMIDE",
            "non-polymer",
        ),
        "1DF": (
            "(5-CARBOXY-2-HYDROXY-3-METHOXYPHENYL)(HYDROXY)OXOAMMONIUM",
            "non-polymer",
        ),
        "1DG": (
            "(4R)-ISONICOTINIC-ACETYL-NICOTINAMIDE-ADENINE DINUCLEOTIDE",
            "non-polymer",
        ),
        "1DB": ("1-DECANE-SULFONIC-ACID", "non-polymer"),
        "SXV": (
            "N~3~-[(2S)-2-HYDROXY-3,3-DIMETHYL-4-(PHOSPHONOOXY)BUTANOYL]-N-{2-[(3-OXOBUTYL)SULFANYL]ETHYL}-BETA-ALANINAMIDE",
            "non-polymer",
        ),
        "XMJ": (
            "4-[[(1E)-2-(4-CHLOROPHENYL)ETHENYL]SULFONYL]-1-[[1-(4-PYRIDINYL)-4-PIPERIDINYL]METHYL]PIPERAZINONE",
            "non-polymer",
        ),
        "1DA": ("1-DEAZA-ADENOSINE", "non-polymer"),
        "XMD": (
            "(2R)-4-[(6-CHLORO-2-NAPHTHALENYL)SULFONYL]-6-OXO-1-[[1-(4-PYRIDINYL)-4-PIPERIDINYL]METHYL]-2-PIPERAZINECARBOXYLIC ACID ETHYL ESTER",
            "non-polymer",
        ),
        "1DO": ("1-DODECANOL", "non-polymer"),
        "XMF": (
            "4-[(6-CHLORO-2-NAPHTHALENYL)SULFONYL]-1-[[4-HYDROXYMETHYL-1-(4-PYRIDINYL)-4-PIPERIDINYL]METHYL]PIPERAZINONE",
            "non-polymer",
        ),
        "1DM": (
            "N-(2-MORPHOLIN-4-YL-1-MORPHOLIN-4-YLMETHYL-ETHYL)-3-NITRO-5-(3,4,5-TRIHYDROXY-6-HYDROXYMETHYL-TETRAHYDRO-PYRAN-2-YLOXY)-BENZAMIDE",
            "non-polymer",
        ),
        "1DJ": ("(1R)-1-PHENYLPROP-2-YN-1-OL", "non-polymer"),
        "XMA": (
            "(-)-7-[(6-CHLORO-2-NAPHTHALENYL)SULFONYL]TETRAHYDRO-8A-(METHOXYMETHYL)-1'-(4-PYRIDINYL)-SPIRO[5H-OXAZOLO[3,2-A]PYRAZINE-2(3H),4'-PIPERIDIN]-5-ONE",
            "non-polymer",
        ),
        "1DH": ("5-CHLORO-1,3-THIAZOL-2-AMINE", "non-polymer"),
        "1DI": (
            "(2R)-2-AMINO-3-(2,4-DICHLOROPHENYL)-1-(1,3-DIHYDRO-2H-ISOINDOL-2-YL)PROPAN-1-ONE",
            "non-polymer",
        ),
        "1DV": ("5-METHYL-3-(TRIFLUOROMETHYL)-1H-PYRAZOLE", "non-polymer"),
        "1DW": ("2-BROMO-5-HYDROXYBENZALDEHYDE", "non-polymer"),
        "TQ4": ("5-[(4-METHYLPHENYL)SULFANYL]-2,4-QUINAZOLINEDIAMINE", "non-polymer"),
        "1DU": ("2-(1H-PYRAZOL-1-YL)ETHANOL", "non-polymer"),
        "1DR": ("1,10-DIHYDROPYRROLO[2,3-A]CARBAZOLE-3-CARBALDEHYDE", "non-polymer"),
        "1DP": ("N1-DEAZA-ADENOSINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "1DQ": ("5-(TRIFLUOROMETHYL)-1H-IMIDAZOLE", "non-polymer"),
        "MXZ": ("6-DEOXY-2-O-METHYL-ALPHA-L-GALACTOPYRANOSE", "non-polymer"),
        "XMP": ("XANTHOSINE-5'-MONOPHOSPHATE", "non-polymer"),
        "1DY": (
            "N-(2-METHOXYPHENYL)-4-{[3-(4-OXO-3,4-DIHYDROQUINAZOLIN-2-YL)PROPANOYL]AMINO}BENZAMIDE",
            "non-polymer",
        ),
        "N5N": (
            "2-[[(2S)-4-[[(3S)-1-CARBAMIMIDOYLPIPERIDIN-3-YL]METHYLAMINO]-2-(NAPHTHALEN-2-YLSULFONYLAMINO)-4-OXIDANYLIDENE-BUTANOYL]-CYCLOPROPYL-AMINO]ETHANOIC ACID",
            "non-polymer",
        ),
        "MXX": ("5,8-DIMETHOXY-1,4-DIMETHYLQUINOLIN-2(1H)-ONE", "non-polymer"),
        "N5O": (
            "(2S,3S,4R,5R)-5-(6-AMINO-9H-PURIN-9-YL)-3,4-DIHYDROXY-N-(2-HYDROXYETHYL)TETRAHYDROFURAN-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "YYY": ("DEOXYCYTIDINE DIPHOSPHATE", "non-polymer"),
        "4C0": (
            "3-(1H-BENZIMIDAZOL-2-YL)-N-(3-PHENYLPROPANOYL)-D-ALANINE",
            "non-polymer",
        ),
        "S80": (
            "N1-HYDROXY-2-(3-HYDROXY-PROPYL)-3-ISOBUTYL-N4-[1-(2-METHOXY-ETHYL)-2-OXO-AZEPAN-3-YL]-SUCCINAMIDE",
            "non-polymer",
        ),
        "S82": (
            "4-[3-(1-METHYLETHYL)-1,2,4-OXADIAZOL-5-YL]-N-[(1S,2R)-2-PHENYLCYCLOPROPYL]PIPERIDINE-1-CARBOXAMIDE",
            "non-polymer",
        ),
        "N5M": ("?", "nucleic-acid"),
        "YYA": (
            "2,4,5-TRIDEOXY-2-(METHYLAMINO)-4-[(2Z)-PENTA-2,4-DIEN-1-YL]-L-XYLONIC ACID",
            "peptide",
        ),
        "S89": (
            "N-[(2S)-1-HYDROXY-3-PHENYLPROPAN-2-YL]-NALPHA-[(2E)-3-PHENYLPROP-2-ENOYL]-L-PHENYLALANINAMIDE",
            "peptide",
        ),
        "8DR": (
            "(S)-2-(3-(BENZO[C][1,2,5]THIADIAZOLE-4-SULFONAMIDO)THIOPHENE-2-CARBOXAMIDO)-5-GUANIDINOPENTANOIC ACID",
            "non-polymer",
        ),
        "WSA": ("5'-O-[(L-TRYPTOPHYLAMINO)SULFONYL]ADENOSINE", "non-polymer"),
        "HYA": ("2,3,4,N-TETRAHYDROXY-BUTYRIMIDIC ACID", "non-polymer"),
        "HYB": (
            "[2-(2-METHYL-PROPANE-2-SULFONYLMETHYL)-3-NAPHTHALEN-1-YL-PROPIONYL-VALINYL]-PHENYLALANINOL",
            "non-polymer",
        ),
        "HYC": (
            "O5'-[9-(3,17B-DIHYDROXY-1,3,5(10)-ESTRATRIEN-16B-YL)-NONANOYL]ADENOSINE",
            "non-polymer",
        ),
        "HYD": ("HYDROXY GROUP", "non-polymer"),
        "HYE": (
            "(2R,3S,4R)-2-[(S)-(1S)-CYCLOHEX-2-EN-1-YL(HYDROXY)METHYL]-3-HYDROXY-4-(2-HYDROXYETHYL)-3-METHYL-5-OXOPYRROLIDINE-2-CARBALDEHYDE",
            "non-polymer",
        ),
        "HYF": (
            "4-HYDROXY-5-ISOBUTYRYL-6-METHYL-1,3,7-TRIS-(3-METHYL-BUT-2-ENYL)-6-(4-METHYL-PENT-3-ENYL)-BICYCLO[3.3.1]NON-3-ENE-2,9-DIONE",
            "non-polymer",
        ),
        "HYG": ("HYGROMYCIN B", "non-polymer"),
        "HYI": (
            "(2S)-2-AMINO-4-[(2R,3R)-2,3-DIHYDROXY-3-N-HYDROXYCARBAMOYL-PROPYLMERCAPTO]BUTYRIC ACID",
            "non-polymer",
        ),
        "WSK": (
            "(2S)-1-(3,6-DIBROMO-9H-CARBAZOL-9-YL)-3-(DIMETHYLAMINO)PROPAN-2-OL",
            "non-polymer",
        ),
        "N5I": (
            "1-(2-DEOXY-5-O-PHOSPHONO-BETA-D-ERYTHRO-PENTOFURANOSYL)-5-NITRO-1H-INDOLE",
            "nucleic-acid",
        ),
        "HYM": (
            "(4Z)-4-(2-AMINO-5-OXO-3,5-DIHYDRO-4H-IMIDAZOL-4-YLIDENE)-2,3-DICHLORO-4,5,6,7-TETRAHYDROPYRROLO[2,3-C]AZEPIN-8(1H)-ONE",
            "non-polymer",
        ),
        "HYN": ("IMIDAZOLIDINE-2,4-DIONE", "non-polymer"),
        "HYP": ("4-HYDROXYPROLINE", "peptide"),
        "HYQ": (
            "REL-(3AR,4S,7R,7AS)-3A,4,7,7A-TETRAHYDRO-2-(4-NITRO-1-NAPHTHALENYL)-4,7-ETHANO-1H-ISOINDOLE-1,3(2H)-DIONE",
            "non-polymer",
        ),
        "WST": (
            "(3AS,4R,9BR)-4-(4-HYDROXYPHENYL)-1,2,3,3A,4,9B-HEXAHYDROCYCLOPENTA[C]CHROMEN-9-OL",
            "non-polymer",
        ),
        "Z79": ("N-HYDROXY-N~2~-[(4-METHOXYPHENYL)SULFONYL]GLYCINAMIDE", "non-polymer"),
        "N5G": (
            "N-({TRANS-4-[({(2R,4R,4AS,6S,8AS)-2-AMINO-4-HYDROXY-5-[(PHOSPHONOOXY)METHYL]DECAHYDROPTERIDIN-6-YL}METHYL)AMINO]CYCLOHEXYL}CARBONYL)-L-GLUTAMIC ACID",
            "non-polymer",
        ),
        "Q6T": ("1,3-BENZODIOXOLE-4-CARBOXYLIC ACID", "non-polymer"),
        "HYZ": (
            "N-[1-(3-FLUOROBENZYL)-1H-INDAZOL-5-YL]-5-[(PIPERIDIN-1-YLAMINO)METHYL]PYRIMIDINE-4,6-DIAMINE",
            "non-polymer",
        ),
        "Q6W": (
            "3-{5-OXO-3-[1-(PIPERIDIN-4-YL)-1H-PYRAZOL-4-YL]-5H-BENZO[4,5]CYCLOHEPTA[1,2-B]PYRIDIN-7-YL}-N-(PYRIDIN-2-YLMETHYL)PROPANAMIDE",
            "non-polymer",
        ),
        "MXN": ("(2R)-HYDROXY(PHENYL)ETHANENITRILE", "non-polymer"),
        "878": ("5-IODO-2-(OXALYL-AMINO)-BENZOIC ACID", "non-polymer"),
        "879": (
            "6-({2-(2-CHLOROPHENYL)-5-[4-(PYRIMIDIN-5-YLOXY)PHENYL]-1H-PYRROL-1-YL}METHYL)PYRIDIN-2-AMINE",
            "non-polymer",
        ),
        "876": (
            "N-ACETYL-N-[1-(1,1'-BIPHENYL-4-YLMETHYL)-2-OXOAZEPAN-3-YL]-3,4-DIPHOSPHONOPHENYLALANINAMIDE",
            "non-polymer",
        ),
        "877": (
            "N-(4-AMINO-5-CYANO-6-ETHOXYPYRIDIN-2-YL)-2-(4-BROMO-2,5-DIMETHOXYPHENYL)ACETAMIDE",
            "non-polymer",
        ),
        "PYD": ("2,5-DIMETHYL-PYRIMIDIN-4-YLAMINE", "non-polymer"),
        "VK3": ("MENADIONE", "non-polymer"),
        "872": (
            "(3R)-4-[(3R)-3-AMINO-4-(2,4,5-TRIFLUOROPHENYL)BUTANOYL]-3-(2,2,2-TRIFLUOROETHYL)-1,4-DIAZEPAN-2-ONE",
            "non-polymer",
        ),
        "MXL": ("MOXALACTAM", "non-polymer"),
        "870": (
            "N-[7-(3-AMINOPHENYL)-5-METHOXY-1,3-BENZOXAZOL-2-YL]-2,5-DICHLOROBENZENESULFONAMIDE",
            "non-polymer",
        ),
        "N5C": ("?", "non-polymer"),
        "8DX": (
            "(2R,3AR,6S,7AR)-2-[(2S)-2-AMINO-3-HYDROXY-3-OXO-PROPYL]-6-HYDROXY-3,3A,5,6,7,7A-HEXAHYDROFURO[4,5-B]PYRAN-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "PBU": (
            "(2R)-3-{[(R)-HYDROXY{[(1R,2R,3S,4R,5R,6S)-2,3,6-TRIHYDROXY-4,5-BIS(PHOSPHONOOXY)CYCLOHEXYL]OXY}PHOSPHORYL]OXY}PROPANE-1,2-DIYL DIBUTANOATE",
            "non-polymer",
        ),
        "O04": (
            "(E)-4-((TETRAHYDRO-2H-PYRAN-2-YLOXYIMINO)METHYL)BENZIMIDAMIDE",
            "non-polymer",
        ),
        "87Y": (
            "7,8-DIHYDRO-6-HYDROXYMETHYL-7-METHYL-7-[2-PHENYLETHYL]-PTERIN",
            "non-polymer",
        ),
        "MXG": (
            "2-(1-{2-[4-(2-ACETYLAMINO-PROPIONYLAMINO)-4-CARBOXY-BUTYRYLAMINO]-6-AMINO-HEXANOYLAMINO}-2-OXO-ETHYL)-5-METHYLENE-5,6-DIHYDRO-2H-[1,3]THIAZINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "POZ": ("?", "non-polymer"),
        "MXE": ("2-METHOXYETHANOL", "non-polymer"),
        "POX": (
            "4-AMINO-6-{[1-(3-FLUOROBENZYL)-1H-INDAZOL-5-YL]AMINO}PYRIMIDINE-5-CARBALDEHYDE O-(2-METHOXYETHYL)OXIME",
            "non-polymer",
        ),
        "POY": (
            "3-METHYL-1-(5-OXOHEXYL)-7-PROPYL-3,7-DIHYDRO-1H-PURINE-2,6-DIONE",
            "non-polymer",
        ),
        "HY0": ("HYGROMYCIN B VARIANT", "non-polymer"),
        "HY1": ("PHENYLACETALDEHYDE", "non-polymer"),
        "HY2": (
            "2-{2-[4-(3-{[(11AS)-7-METHOXY-5-OXO-2,3,5,10,11,11A-HEXAHYDRO-1H-PYRROLO[2,1-C][1,4]BENZODIAZEPIN-8-YL]OXY}PROPYL)PIPERAZIN-1-YL]ETHYL}-1H-BENZO[DE]ISOQUINOLINE-1,3(2H)-DIONE",
            "non-polymer",
        ),
        "HY3": ("3-HYDROXYPROLINE", "peptide"),
        "HY7": (
            "8-BROMANYL-2-[(4-METHYLPIPERAZIN-1-YL)METHYL]-3H-[1]BENZOFURO[3,2-D]PYRIMIDIN-4-ONE",
            "non-polymer",
        ),
        "POT": (
            "(1S)-1-(PHENOXYMETHYL)PROPYL METHYLPHOSPHONOCHLORIDOATE",
            "non-polymer",
        ),
        "SX5": (
            "4-[(4-BROMO-2-{[(3R,5S)-3,5-DIMETHYLPIPERIDIN-1-YL]CARBONYL}PHENYL)AMINO]-4-OXOBUTANOIC ACID",
            "non-polymer",
        ),
        "RQ3": (
            "2-{4-[(3,5-DIMETHYLANILINO)-CARBONYL-METHYL]-PHENOXY}-2-METHYLPROPIONIC ACID",
            "non-polymer",
        ),
        "FWD": (
            "2-AMINO-3-(5-FLUORO-2,4-DIOXO-3,4-DIHYDRO-2H-PYRIMIDIN-1-YL)-PROPIONIC ACID",
            "non-polymer",
        ),
        "FV0": (
            "(2R,3S)-3-AMINO-1-[(2-METHYLPROPYL)AMINO]-4-PHENYLBUTAN-2-OL",
            "non-polymer",
        ),
        "POS": (
            "1-[[(4-PHENOXYPHENYL)SULFONYL]AMINO]-3-[[N/N-(4-PYRIDINYLCARBONYL)-L-LEUCYL]AMINO]-2-PROPANOL",
            "non-polymer",
        ),
        "E72": (
            "7-[(5-AMINOPENTYL)OXY]-N~4~-[1-(5-AMINOPENTYL)PIPERIDIN-4-YL]-N~2~-[3-(DIMETHYLAMINO)PROPYL]-6-METHOXYQUINAZOLINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "E71": (
            "3-[4-(2,3-DIFLUOROPHENYL)-1,2-OXAZOL-5-YL]-5-(PYRIDIN-4-YL)-1H-PYRROLO[2,3-B]PYRIDINE",
            "non-polymer",
        ),
        "E70": (
            "N-{2-[(4-HYDROXYPHENYL)AMINO]PYRIDIN-3-YL}-4-METHOXYBENZENESULFONAMIDE",
            "non-polymer",
        ),
        "O2": ("OXYGEN MOLECULE", "non-polymer"),
        "E75": (
            "TRANS-4-{[4-(5-METHYL-3-PHENYL-1,2-OXAZOL-4-YL)PYRIMIDIN-2-YL]AMINO}CYCLOHEXANOL",
            "non-polymer",
        ),
        "PON": ("IMIDO DIPHOSPHATE", "non-polymer"),
        "OAE": (
            "(2S,4S)-2-[(R)-{[(2R)-2-AMINO-2-PHENYLACETYL]AMINO}(CARBOXY)METHYL]-5,5-DIMETHYL-1,3-THIAZOLIDINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "FK9": ("PROPYL THIOPROPYL ALPHA-D-MANNOPYRANOSIDE", "non-polymer"),
        "POL": ("N-PROPANOL", "non-polymer"),
        "335": (
            "[4-(2-(1H-1,2,3-BENZOTRIAZOL-1-YL)-3-{4-[DIFLUORO(PHOSPHONO)METHYL]PHENYL}-2-PHENYLPROPYL)PHENYL](DIFLUORO)METHYLPHOSPHONIC ACID",
            "non-polymer",
        ),
        "334": (
            "2-{5-[AMINO(IMINIO)METHYL]-1H-BENZIMIDAZOL-2-YL}-6-BROMO-4-METHYLBENZENOLATE",
            "non-polymer",
        ),
        "337": (
            "N-BENZYL-1-[5-({5-TERT-BUTYL-2-METHOXY-3-[(METHYLSULFONYL)AMINO]PHENYL}CARBAMOYL)-2-METHYLPHENYL]-1H-1,2,3-TRIAZOLE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "PYE": ("TETRAHYDROPYRAN", "non-polymer"),
        "ZPX": (
            "N-[(1R)-1-(4-FLUOROPHENYL)ETHYL]-N'-[(2S,3S)-3-HYDROXY-1-PHENYL-4-(1H-PYRAZOL-1-YL)BUTAN-2-YL]-5-[METHYL(METHYLSULFONYL)AMINO]BENZENE-1,3-DICARBOXAMIDE",
            "non-polymer",
        ),
        "336": (
            "4-{2-[4-(3,10-DIBROMO-8-CHLORO-6,11-DIHYDRO-5H-BENZO[5,6]CYCLOHEPTA[1,2-B]PYRIDIN-11-YL)PIPERIDIN-1-YL]-2-OXOETHYL}PIPERIDINE-1-CARBOXAMIDE",
            "non-polymer",
        ),
        "POH": (
            "(1Z,4Z,9Z,15Z)-5,10,15,20-TETRAKIS(1-METHYLPYRIDIN-1-IUM-4-YL)-21,23-DIHYDROPORPHYRIN",
            "non-polymer",
        ),
        "ZPQ": (
            "N~2~-[(2R,4S,5S)-5-{[N-{[(3,5-DIMETHYL-1H-PYRAZOL-1-YL)METHOXY]CARBONYL}-3-(METHYLSULFONYL)-L-ALANYL]AMINO}-4-HYDROXY-2,7-DIMETHYLOCTANOYL]-N-ISOBUTYL-L-VALINAMIDE",
            "non-polymer",
        ),
        "331": (
            "(2S)-2-(4-[2-(3-[2,4-DIFLUOROPHENYL]-1-HEPTYLUREIDO)ETHYL]PHENOXY)-2-METHYLBUTYRIC ACID",
            "non-polymer",
        ),
        "POI": (
            "N~2~-ACETYL-N~5~-({3-HYDROXY-2-METHYL-5-[(PHOSPHONOOXY)METHYL]PYRIDIN-4-YL}METHYL)-L-ORNITHINE",
            "non-polymer",
        ),
        "LI1": (
            "1-[2,6,10.14-TETRAMETHYL-HEXADECAN-16-YL]-2-[2,10,14-TRIMETHYLHEXADECAN-16-YL]GLYCEROL",
            "non-polymer",
        ),
        "330": (
            "9-HYDROXY-6-(3-HYDROXYPROPYL)-4-(2-METHOXYPHENYL)PYRROLO[3,4-C]CARBAZOLE-1,3(2H,6H)-DIONE",
            "non-polymer",
        ),
        "POF": (
            "(4AS,4BR,10BS,12AS)-12A-METHYL-1,3-DIOXO-2-(PYRIDIN-3-YLMETHYL)-1,2,3,4,4A,4B,5,6,10B,11,12,12A-DODECAHYDRONAPHTHO[2,1-F]ISOQUINOLIN-8-YL SULFAMATE",
            "non-polymer",
        ),
        "SHP": ("(4-HYDROXYMALTOSEPHENYL)GLYCINE", "peptide"),
        "333": (
            "4-(1,3-BENZODIOXOL-5-YLOXY)-2-[4-(1H-IMIDAZOL-1-YL)PHENOXY]PYRIMIDINE",
            "non-polymer",
        ),
        "POG": (
            "(20S)-2,5,8,11,14,17-HEXAMETHYL-3,6,9,12,15,18-HEXAOXAHENICOSANE-1,20-DIOL",
            "non-polymer",
        ),
        "KQ7": (
            "N-(2-METHYL-5-{[(2-MORPHOLIN-4-YLPYRIDIN-4-YL)CARBONYL]AMINO}PHENYL)-6-({2-[(2S)-1-METHYLPYRROLIDIN-2-YL]ETHYL}AMINO)PYRIDINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "NAN": ("5-N-ACETYL-ALPHA-D-NEURAMINIC ACID", "non-polymer"),
        "B91": (
            "3-{(E)-2-[6-(CYCLOPROPYLAMINO)-9H-PURIN-9-YL]ETHENYL}-4-METHYL-N-[3-(4-METHYL-1H-IMIDAZOL-1-YL)-5-(TRIFLUOROMETHYL)PHENYL]BENZAMIDE",
            "non-polymer",
        ),
        "POD": (
            "9-HYDROXY-5-(3,4,5-TRIMETHOXYPHENYL)-5,8,8A,9-TETRAHYDROFURO[3',4':6,7]NAPHTHO[2,3-D][1,3]DIOXOL-6(5AH)-ONE",
            "non-polymer",
        ),
        "ZPG": (
            "(2S,5R,8R,11S,14S,17S,21R)-5,8,11,14,17-PENTAMETHYL-4,7,10,13,16,19-HEXAOXADOCOSANE-2,21-DIOL",
            "non-polymer",
        ),
        "OH": ("HYDROXIDE ION", "non-polymer"),
        "Z70": ("(2S)-2-AMINO-4-(2-AMINO-1H-IMIDAZOL-5-YL)BUTANOIC ACID", "peptide"),
        "E7B": ("?", "non-polymer"),
        "SMM": ("S-ADENOSYLMETHIONINE METHYL ESTER", "non-polymer"),
        "RQX": (
            "2-AMINO-3-METHOXY-6-METHYL-5-[(2E)-3-METHYLHEX-2-EN-1-YL]CYCLOHEXA-2,5-DIENE-1,4-DIONE",
            "non-polymer",
        ),
        "FK5": ("8-DEETHYL-8-[BUT-3-ENYL]-ASCOMYCIN", "non-polymer"),
        "OX": ("BOUND OXYGEN", "non-polymer"),
        "SMK": (
            "(3S,6S,7R,9AS)-6-{[(2S)-2-AMINOBUTANOYL]AMINO}-7-(2-AMINOETHYL)-N-(DIPHENYLMETHYL)-5-OXOOCTAHYDRO-1H-PYRROLO[1,2-A]AZEPINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "POA": ("PHOSPHONOACETALDEHYDE", "non-polymer"),
        "OS": ("OSMIUM ION", "non-polymer"),
        "FW1": ("FUSICOCCIN A AGLYCONE", "non-polymer"),
        "C57": (
            "4-(DIMETHYLAMINO)-N-{[(2S)-1-ETHYLPYRROLIDIN-2-YL]METHYL}-2-METHOXY-5-NITROBENZAMIDE",
            "non-polymer",
        ),
        "C56": (
            "4-(DIMETHYLAMINO)-N-{[(2R)-1-ETHYLPYRROLIDIN-2-YL]METHYL}-2-METHOXY-5-NITROBENZAMIDE",
            "non-polymer",
        ),
        "DEZ": ("4-PHOSPHO-D-ERYTHRONATE", "non-polymer"),
        "C54": ("(1R)-CYCLOPENT-2-ENE-1-CARBOXYLIC ACID", "non-polymer"),
        "C53": (
            "2-IMINO-5-(1-PYRIDIN-2-YL-METH-(E)-YLIDENE)-1,3-THIAZOLIDIN-4-ONE",
            "non-polymer",
        ),
        "C52": (
            "3-(2-AMINOQUINAZOLIN-6-YL)-4-METHYL-1-[3-(TRIFLUOROMETHYL)PHENYL]PYRIDIN-2(1H)-ONE",
            "non-polymer",
        ),
        "Z71": (
            "4-AMINO-N-(3,5-DIFLUOROPHENYL)-2-[(4-SULFAMOYLPHENYL)AMINO]-1,3-THIAZOLE-5-CARBOXAMIDE",
            "non-polymer",
        ),
        "C50": (
            "1-[3-(2,3-DICHLOROPHENOXY)PROPOXY]-6,6-DIMETHYL-1,6-DIHYDRO-1,3,5-TRIAZINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "DEP": ("DIETHYL PHOSPHONATE", "non-polymer"),
        "DEQ": ("DEQUALINIUM", "non-polymer"),
        "DER": ("D-4-PHOSPHOERYTHRONIC ACID", "non-polymer"),
        "DES": ("DIETHYLSTILBESTROL", "non-polymer"),
        "DET": ("UNDECYLAMINE-N,N-DIMETHYL-N-OXIDE", "non-polymer"),
        "DEU": ("CO(III)-(DEUTEROPORPHYRIN IX)", "non-polymer"),
        "JDC": (
            "(3R)-7-HYDROXY-N-{(2S)-1-[(3R,4R)-4-(3-HYDROXYPHENYL)-3,4-DIMETHYLPIPERIDIN-1-YL]-3-METHYLBUTAN-2-YL}-1,2,3,4-TETRAHYDROISOQUINOLINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "C58": (
            "(2S)-3-PHENYL-N~1~-[2-(PYRIDIN-4-YL)-5,6,7,8-TETRAHYDRO[1]BENZOTHIENO[2,3-D]PYRIMIDIN-4-YL]PROPANE-1,2-DIAMINE",
            "non-polymer",
        ),
        "DEH": (
            "2,9-DIHYDROXY-3,10-DIMETHOXY-5,6-DIHYDROISOQUINO[3,2-A]ISOQUINOLINIUM",
            "non-polymer",
        ),
        "DEI": ("(2Z)-2-HYDROXY-3-(METHYLSULFANYL)PROP-2-ENOIC ACID", "non-polymer"),
        "NHE": ("2-[N-CYCLOHEXYLAMINO]ETHANE SULFONIC ACID", "non-polymer"),
        "DEL": ("4-DEOXYLACTOSE", "saccharide"),
        "DEM": ("DECYLOXY-METHANOL", "non-polymer"),
        "DEN": ("INDENE", "non-polymer"),
        "DEO": ("?", "non-polymer"),
        "DEA": ("D-ERITADENINE", "non-polymer"),
        "DEB": ("6-DEOXYERYTHRONOLIDE B", "non-polymer"),
        "DEC": ("SEBACIC ACID", "non-polymer"),
        "DED": ("2-DIMETHYLAMINO-ETHYL-DIPHOSPHATE", "non-polymer"),
        "DEE": (
            "3,5-DIMETHYL-1H-PYRAZOLE-4-CARBOXYLIC ACID ETHYL ESTER",
            "non-polymer",
        ),
        "DEF": ("DEFEROXAMINE MESYLATE FE(III) COMPLEX", "non-polymer"),
        "DEG": ("BUTYL ALPHA-D-MANNOPYRANOSIDE", "non-polymer"),
        "Z77": (
            "1-[(6,7-DIMETHYL-2-OXO-1,2-DIHYDROQUINOLIN-3-YL)METHYL]-1-(2-HYDROXYETHYL)-3-(3-METHOXYPHENYL)THIOUREA",
            "non-polymer",
        ),
        "Z72": (
            "8-(CYCLOPENTYLOXY)-4-[(3,5-DICHLOROPYRIDIN-4-YL)AMINO]-7-METHOXY-2H-CHROMEN-2-ONE",
            "non-polymer",
        ),
        "FKO": ("3-(2-CHLOROPHENYL)-6-(2-FLUOROPHENOXY)-2H-INDAZOLE", "non-polymer"),
        "NHD": ("NICOTINAMIDE PURIN-6-OL-DINUCLEOTIDE", "non-polymer"),
        "UP6": ("6-AZA-UMP", "non-polymer"),
        "UP5": ("P1-(ADENOSINE-5'-P5-(URIDINE-5')PENTAPHOSPHATE", "non-polymer"),
        "UP2": ("UROPORPHYRINOGEN III", "non-polymer"),
        "PO4": ("PHOSPHATE ION", "non-polymer"),
        "UP1": ("URIDINE-5'-DIPHOSPHATE-2-DEOXY-2-FLUOROGALACTOSE", "non-polymer"),
        "UPN": ("?", "non-polymer"),
        "MDJ": (
            "4-AMINO-1-(2-DEOXY-5-O-PHOSPHONO-BETA-D-ERYTHRO-PENTOFURANOSYL)-3-(HYDROXYMETHYL)PYRIDIN-2(1H)-ONE",
            "nucleic-acid",
        ),
        "UPL": ("UNKNOWN BRANCHED FRAGMENT OF PHOSPHOLIPID", "non-polymer"),
        "UPM": ("?", "non-polymer"),
        "MDK": (
            "4-AMINO-1-(2-DEOXY-5-O-PHOSPHONO-BETA-D-ERYTHRO-PENTOFURANOSYL)-3-METHYLPYRIDIN-2(1H)-ONE",
            "nucleic-acid",
        ),
        "UPF": ("URIDINE-5'-DIPHOSPHATE-2-DEOXY-2-FLUOROGALACTOSE", "non-polymer"),
        "PO2": ("HYPOPHOSPHITE", "non-polymer"),
        "MBB": (
            "2'-(3-METHYL-4-DIMETHYLAMINOPHENYL)-5-(4-METHYL-1-PIPERAZINYL)-2,5'-BI-BENZIMIDAZOLE",
            "non-polymer",
        ),
        "MDH": ("N-METHYLDEHYDROBUTYRINE", "peptide"),
        "CAY": ("CARBOXYMETHYLENECYSTEINE", "peptide"),
        "UPA": ("URIDYLYL-2'-5'-PHOSPHO-ADENOSINE", "non-polymer"),
        "PYF": ("3-PYRIDINYLCARBINOL", "non-polymer"),
        "TGC": ("3-THIAGLUTARYL-COA", "non-polymer"),
        "MDN": ("METHYLENEDIPHOSPHONIC ACID", "non-polymer"),
        "HFT": ("HYDROXYFLUTAMIDE", "non-polymer"),
        "UPU": ("?", "non-polymer"),
        "UPR": (
            "2'-DEOXYURIDINE-5'-ALPHA,BETA-AMINOPHOSPHORANE-TRIPHOSPHATE",
            "non-polymer",
        ),
        "UPS": ("2'-DEOXY-5'-O-THIOPHOSPHONOURIDINE", "nucleic-acid"),
        "UPP": ("PHENYL-URIDINE-5'-DIPHOSPHATE", "non-polymer"),
        "MDO": (
            "{2-[(1S)-1-AMINOETHYL]-4-METHYLIDENE-5-OXO-4,5-DIHYDRO-1H-IMIDAZOL-1-YL}ACETIC ACID",
            "peptide",
        ),
        "IVF": ("N-(IODOACETYL)-L-VALYL-L-PHENYLALANINAMIDE", "peptide"),
        "IVA": ("ISOVALERIC ACID", "non-polymer"),
        "C5P": ("CYTIDINE-5'-MONOPHOSPHATE", "non-polymer"),
        "DE1": ("DECAN-1-OL", "non-polymer"),
        "DE3": ("N,N-DIETHYL-3-METHYLBENZAMIDE", "non-polymer"),
        "IVH": ("2-(BUTYLSULFANYL)ADENOSINE", "non-polymer"),
        "NHF": ("1,5-ANHYDRO-D-FRUCTOSE", "saccharide"),
        "C5X": ("4,5-EPOXYPENTYL-BETA-D-XYLOSIDE", "saccharide"),
        "C5G": ("?", "non-polymer"),
        "C5E": (
            "(1R,5S)-1,2,3,4,5,6-HEXAHYDRO-8H-1,5-METHANOPYRIDO[1,2-A][1,5]DIAZOCIN-8-ONE",
            "non-polymer",
        ),
        "C5D": (
            "(1R,3R)-5-[(2E)-3-{(1S,3R)-2,2,3-TRIMETHYL-3-[6,6,6-TRIFLUORO-5-HYDROXY-5-(TRIFLUOROMETHYL)HEX-3-YN-1-YL]CYCLOPENTYL}PROP-2-EN-1-YLIDENE]CYCLOHEXANE-1,3-DIOL",
            "non-polymer",
        ),
        "C5C": ("S-CYCLOPENTYL THIOCYSTEINE", "peptide"),
        "C5B": ("(1S,3R)-1-AMINOCYCLOPENTANE-1,3-DICARBOXYLIC ACID", "non-polymer"),
        "C5A": ("(1S,3S)-1-AMINOCYCLOPENTANE-1,3-DICARBOXYLIC ACID", "non-polymer"),
        "IVS": (
            "3-HYDROXY-6-METHYL-4-(3-METHYL-2-(3-METHYL-2-(3-METHYL-BUTYRYLAMINO)-BUTYRYLAMINO)-BUTYRYLAMINO)-HEPTANOIC ACID ETHYL ESTER",
            "non-polymer",
        ),
        "C5M": (
            "N-{(2R,3S)-3-[(3-CHLOROBENZYL)AMINO]-2-HYDROXY-4-PHENYLBUTYL}-4-METHOXY-2,3,6-TRIMETHYLBENZENESULFONAMIDE",
            "non-polymer",
        ),
        "Z75": (
            "N'-[(1S,2S)-2-[(2R)-4-BENZYLPIPERAZIN-2-YL]-1-(3,5-DIFLUOROBENZYL)-2-HYDROXYETHYL]-5-METHYL-N,N-DIPROPYLBENZENE-1,3-DICARBOXAMIDE",
            "non-polymer",
        ),
        "ACC": (
            "N-[N-[2-AMINO-6-OXO-HEXANOIC ACID-6-YL]CYSTEINYL]-S-METHYLCYSTEINE",
            "non-polymer",
        ),
        "ACB": ("3-METHYL-BETA-D-ASPARTIC ACID", "peptide"),
        "ACA": ("6-AMINOHEXANOIC ACID", "non-polymer"),
        "GD3": ("GADOLINIUM ION", "non-polymer"),
        "ACG": ("MODIFIED ACARBOSE PENTASACCHARIDE", "saccharide"),
        "ACF": (
            "3-{[(2R,5S)-5-{[(2S)-2-(AMINOMETHYL)PYRROLIDIN-1-YL]CARBONYL}PYRROLIDIN-2-YL]METHOXY}-4-CHLOROBENZOIC ACID",
            "non-polymer",
        ),
        "ACE": ("ACETYL GROUP", "non-polymer"),
        "ACD": ("ARACHIDONIC ACID", "non-polymer"),
        "ACK": (
            "(3AR,4R,6R,6AR)-4-(6-AMINO-9H-PURIN-9-YL)-6-(HYDROXYMETHYL)TETRAHYDROFURO[3,4-D][1,3,2]DIOXAPHOSPHOL-2-OL 2-OXIDE",
            "non-polymer",
        ),
        "ACJ": (
            "5-[2-CHLORO-4-(TRIFLUOROMETHYL)PHENOXY]-2-NITROBENZOIC ACID",
            "non-polymer",
        ),
        "ACI": ("6-AMINO-4-HYDROXYMETHYL-CYCLOHEX-4-ENE-1,2,3-TRIOL", "non-polymer"),
        "ACH": ("ACETYLCHOLINE", "non-polymer"),
        "ACO": ("ACETYL COENZYME *A", "non-polymer"),
        "ACN": ("ACETONE", "non-polymer"),
        "ACM": ("ACETAMIDE", "non-polymer"),
        "ACL": ("DEOXY-CHLOROMETHYL-ARGININE", "peptide"),
        "ACS": (
            "1-[(1S)-CARBOXY-2-(METHYLSULFINYL)ETHYL]-(3R)-[(5S)-5-AMINO-5-CARBOXYPENTANAMIDO]-(4R)-SULFANYLAZETIDIN-2-ONE",
            "non-polymer",
        ),
        "ACR": ("ALPHA-ACARBOSE", "saccharide"),
        "ACQ": ("DIPHOSPHOMETHYLPHOSPHONIC ACID ADENYLATE ESTER", "non-polymer"),
        "ACP": ("PHOSPHOMETHYLPHOSPHONIC ACID ADENYLATE ESTER", "non-polymer"),
        "ACW": (
            "D-(L-A-AMINOADIPOYL)-L-CYSTEINYL-B-METHYL-D-CYCLOPROPYLGLYCINE",
            "non-polymer",
        ),
        "ACV": ("L-D-(A-AMINOADIPOYL)-L-CYSTEINYL-D-VALINE", "non-polymer"),
        "ACU": ("ACETALDEHYDE", "non-polymer"),
        "ACT": ("ACETATE ION", "non-polymer"),
        "HFA": ("ALPHA-HYDROXY-BETA-PHENYL-PROPIONIC ACID", "peptide"),
        "ACZ": ("CIS-AMICLENOMYCIN", "non-polymer"),
        "ACY": ("ACETIC ACID", "non-polymer"),
        "ACX": ("ALPHA-CYCLODEXTRIN (CYCLOHEXA-AMYLOSE)", "saccharide"),
        "ME9": ("1,1'-BUTANE-1,4-DIYLBIS(1H-PYRROLE-2,5-DIONE)", "non-polymer"),
        "ME8": (
            "[[(2R,3S,4R,5R)-5-(6-AMINOPURIN-9-YL)-3,4-DIHYDROXY-OXOLAN-2-YL]METHOXY-HYDROXY-PHOSPHORYL] (2S)-2-AZANYL-4-METHYLSULFANYL-BUTANOATE",
            "non-polymer",
        ),
        "HFG": (
            "7-BROMO-6-CHLORO-3-{3-[(2R,3S)-3-HYDROXYPIPERIDIN-2-YL]-2-OXOPROPYL}QUINAZOLIN-4(3H)-ONE",
            "non-polymer",
        ),
        "NEQ": ("N-ETHYLMALEIMIDE", "non-polymer"),
        "BKA": ("3-OXOHEPTANOIC ACID", "non-polymer"),
        "SJ1": (
            "N-(ISOBUTOXYCARBONYL)-D-SERYL-N-((1S)-4-{[AMINO(IMINO)METHYL]AMINO}-1-FORMYLBUTYL)-L-ALANINAMIDE",
            "non-polymer",
        ),
        "GGG": ("GLYCYLGLYCYLGLYCINE", "non-polymer"),
        "J19": (
            "(2Z,5Z)-2-[(2-CHLOROPHENYL)IMINO]-5-(4-HYDROXY-3-METHOXYBENZYLIDENE)-1,3-THIAZOLIDIN-4-ONE",
            "non-polymer",
        ),
        "K0I": (
            "3-[[2-PYRIDIN-2-YL-6-(1,2,4,5-TETRAHYDRO-3-BENZAZEPIN-3-YL)PYRIMIDIN-4-YL]AMINO]PROPANOIC ACID",
            "non-polymer",
        ),
        "J14": (
            "N-{(3S,4R)-4-[(6-AMINO-4-METHYLPYRIDIN-2-YL)METHYL]PYRROLIDIN-3-YL}-N'-[2-(3-FLUOROPHENYL)ETHYL]ETHANE-1,2-DIAMINE",
            "non-polymer",
        ),
        "J15": (
            "N-{3-[4-(3-AMINO-PROPYL)-PIPERAZIN-1-YL]-PROPYL}-3-NITRO-5-(GALACTOPYRANOSYL)-ALPHA-BENZAMIDE",
            "non-polymer",
        ),
        "J12": (
            "N-{3-[4-(3-AMINO-PROPYL)-PIPERAZIN-1-YL]-PROPYL}-3-(2-THIOPHEN-2-YL-ACETYLAMINO)-5-(3,4,5-TRIHYDROXY-6-HYDROXYMETHYL-TETRAHYDRO-PYRAN-2-YLOXY)-BENZAMIDE",
            "non-polymer",
        ),
        "J11": (
            "N-{(3R,4S)-4-[(6-AMINO-4-METHYLPYRIDIN-2-YL)METHYL]PYRROLIDIN-3-YL}-N'-[2-(3-FLUOROPHENYL)ETHYL]ETHANE-1,2-DIAMINE",
            "non-polymer",
        ),
        "K00": (
            "(2R,3R,4R,5R,6S)-3-FLUORO-4,5-DIHYDROXY-6-[(1R,2R)-1,2,3-TRIHYDROXYPROPYL]OXANE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "K01": (
            "N,N-DIETHYL-2-[(2-THIENYLCARBONYL)AMINO]-4,5,6,7-TETRAHYDRO-1-BENZOTHIOPHENE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "K02": (
            "(5R)-N,N-DIETHYL-5-METHYL-2-[(THIOPHEN-2-YLCARBONYL)AMINO]-4,5,6,7-TETRAHYDRO-1-BENZOTHIOPHENE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "K03": (
            "N,N-DIETHYL-5,5-DIMETHYL-2-[(2-THIENYLCARBONYL)AMINO]-4,5,6,7-TETRAHYDRO-1-BENZOTHIOPHENE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "UFO": (
            "1-(1,4-DIMETHYL-1,2,3,4-TETRAHYDROQUINOXALIN-6-YL)METHANAMINE",
            "non-polymer",
        ),
        "K05": (
            "(E)-3,4-DIHYDROXY-N'-[(2-METHOXYNAPHTHALEN-1-YL)METHYLENE]BENZOHYDRAZIDE",
            "non-polymer",
        ),
        "J1D": (
            "5'-{[2-({N-[(2-AMINO-7,7-DIMETHYL-4-OXO-3,4,7,8-TETRAHYDROPTERIDIN-6-YL)CARBONYL]GLYCYL}AMINO)ETHYL]SULFONYL}-5'-DEOXYADENOSINE",
            "non-polymer",
        ),
        "J1B": (
            "5'-S-[1-(2-{[(2-AMINO-7,7-DIMETHYL-4-OXO-3,4,7,8-TETRAHYDROPTERIDIN-6-YL)METHYL]AMINO}ETHYL)PIPERIDIN-4-YL]-5'-THIOADENOSINE",
            "non-polymer",
        ),
        "J1C": (
            "5'-S-[1-(2-{[(2-AMINO-7,7-DIMETHYL-4-OXO-3,4,7,8-TETRAHYDROPTERIDIN-6-YL)CARBONYL]AMINO}ETHYL)PIPERIDIN-4-YL]-5'-THIOADENOSINE",
            "non-polymer",
        ),
        "J1A": (
            "5'-S-[1-(2-{[(2-AMINO-4-OXO-3,4-DIHYDROPTERIDIN-6-YL)METHYL]AMINO}ETHYL)PIPERIDIN-4-YL]-5'-THIOADENOSINE",
            "non-polymer",
        ),
        "BK4": (
            "3-(6-METHOXYNAPHTHALEN-2-YL)-1-(PROPAN-2-YL)-1H-PYRAZOLO[3,4-D]PYRIMIDIN-4-AMINE",
            "non-polymer",
        ),
        "BK5": (
            "3-(6-ETHOXYNAPHTHALEN-2-YL)-1-(PROPAN-2-YL)-1H-PYRAZOLO[3,4-D]PYRIMIDIN-4-AMINE",
            "non-polymer",
        ),
        "BK6": (
            "3-(6-METHOXYNAPHTHALEN-2-YL)-1-(PIPERIDIN-4-YLMETHYL)-1H-PYRAZOLO[3,4-D]PYRIMIDIN-4-AMINE",
            "non-polymer",
        ),
        "BK7": (
            "3-(6-ETHOXYNAPHTHALEN-2-YL)-1-(PIPERIDIN-4-YLMETHYL)-1H-PYRAZOLO[3,4-D]PYRIMIDIN-4-AMINE",
            "non-polymer",
        ),
        "J1Z": ("THIENO[3,2-B]THIOPHENE", "non-polymer"),
        "BK1": (
            "1-(1-METHYLETHYL)-3-(NAPHTHALEN-1-YLMETHYL)-1H-PYRAZOLO[3,4-D]PYRIMIDIN-4-AMINE",
            "non-polymer",
        ),
        "BK2": (
            "1-TERT-BUTYL-3-NAPHTHALEN-2-YL-1H-PYRAZOLO[3,4-D]PYRIMIDIN-4-AMINE",
            "non-polymer",
        ),
        "BK3": (
            "3-(NAPHTHALEN-1-YLMETHYL)-1-(PIPERIDIN-4-YLMETHYL)-1H-PYRAZOLO[3,4-D]PYRIMIDIN-4-AMINE",
            "non-polymer",
        ),
        "MDP": ("N-CARBOXY-N-METHYL-MURAMIC ACID", "saccharide"),
        "GDP": ("GUANOSINE-5'-DIPHOSPHATE", "nucleic-acid"),
        "MED": ("D-METHIONINE", "peptide"),
        "GDR": ("GUANOSINE-5'-DIPHOSPHATE-RHAMNOSE", "nucleic-acid"),
        "GDS": ("OXIDIZED GLUTATHIONE DISULFIDE", "non-polymer"),
        "GDT": (
            "(2E,4E)-N-[(2S,3R)-3-HYDROXY-1-[[(3Z,5S,8S,10S)-10-HYDROXY-5-METHYL-2,7-DIOXO-1,6-DIAZACYCLODODEC-3-EN-8-YL]AMINO]-1-OXOBUTAN-2-YL]DODECA-2,4-DIENAMIDE",
            "non-polymer",
        ),
        "GDU": ("GALACTOSE-URIDINE-5'-DIPHOSPHATE", "non-polymer"),
        "GDV": (
            "(5R,6R,7R,8S)-8-(ACETYLAMINO)-6,7-DIHYDROXY-5-(HYDROXYMETHYL)-N-PHENYL-1,5,6,7,8,8A-HEXAHYDROIMIDAZO[1,2-A]PYRIDINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "MEB": ("2-METHYLPROP-1-ENE", "non-polymer"),
        "GDX": ("?", "non-polymer"),
        "MEL": (
            "[((1R)-2-{(2S)-2-[({4-[AMINO(IMINO)METHYL]BENZYL}AMINO)CARBONYL]AZETIDINYL}-1-CYCLOHEXYL-2-OXOETHYL)AMINO]ACETIC ACID",
            "non-polymer",
        ),
        "HF3": ("SMALLEST HF-OXO-PHOSPHATE CLUSTER HF3", "non-polymer"),
        "HF2": ("(2R)-2-HYDROXY-3-PHENYLPROPANOIC ACID", "peptide"),
        "HF5": ("HF OXO CLUSTER HF5", "non-polymer"),
        "MEK": (
            "3,4-DIFLUORO-2-[(2-FLUORO-4-IODOPHENYL)AMINO]-N-(2-HYDROXYETHOXY)-5-[(2-HYDROXYETHOXY)METHYL]BENZAMIDE",
            "non-polymer",
        ),
        "MDW": (
            "(1R)-N,6-DIHYDROXY-7-METHOXY-2-[(4-METHOXYPHENYL)SULFONYL]-1,2,3,4-TETRAHYDROISOQUINOLINE-1-CARBOXAMIDE",
            "non-polymer",
        ),
        "MEU": ("O-METHYL-GLYCINE", "peptide"),
        "AC2": ("9-HYROXYETHOXYMETHYLGUANINE", "non-polymer"),
        "AC1": (
            "6-METHYL-5-(4,5,6-TRIHYDROXY-3-HYDROXYMETHYL-CYCLOHEX-2-ENYLAMINO)-TETRAHYDRO-PYRAN-2,3,4-TRIOL",
            "non-polymer",
        ),
        "AC0": ("1-PHENYLETHANONE", "non-polymer"),
        "AC7": ("?", "non-polymer"),
        "AC6": ("P-HYDROXYACETOPHENONE", "non-polymer"),
        "AC5": ("1-AMINOCYCLOPENTANECARBOXYLIC ACID", "peptide"),
        "AC4": ("AMINOIMIDAZOLE 4-CARBOXAMIDE RIBONUCLEOTIDE", "non-polymer"),
        "NE6": (
            "METHYL [(1E,5R)-5-{(3S)-3-[(2E,4E)-2,5-DIMETHYLOCTA-2,4-DIENOYL]-2,4-DIOXO-3,4-DIHYDRO-2H-PYRAN-6-YL}HEXYLIDENE]CARBAMATE",
            "non-polymer",
        ),
        "AC9": ("TRISACETYLACETONATOCHROMIUM(III)", "non-polymer"),
        "GDL": ("2-ACETAMIDO-2-DEOXY-D-GLUCONO-1,5-LACTONE", "non-polymer"),
        "GDM": ("GELDANAMYCIN", "non-polymer"),
        "GDN": ("GLUTATHIONE S-(2,4 DINITROBENZENE)", "non-polymer"),
        "GDO": ("?", "nucleic-acid"),
        "SBZ": ("[4-(1,3,2-DIOXABOROLAN-2-YLOXY)METHYL]BENZAMIDINE", "non-polymer"),
        "LWY": ("1-[(R)-HYDROXY(METHYL)PHOSPHORYL]-L-PROLINE", "peptide"),
        "ZMK": (
            "METHYL 7-{[2-({N-[(2S)-2-HYDROXY-3,3-DIMETHYL-4-(PHOSPHONOOXY)BUTANOYL]-BETA-ALANYL}AMINO)ETHYL]SULFANYL}-7-OXOHEPTANOATE",
            "non-polymer",
        ),
        "EZ4": ("(2S,3R,4R)-2,3,4,5-TETRAHYDROXY-N-OXO-PENTANAMIDE", "non-polymer"),
        "KAA": ("5'-O-[(L-LYSYLAMINO)SULFONYL]ADENOSINE", "non-polymer"),
        "R2B": ("1,4-ANHYDRO-D-ERYTHRO-PENT-1-ENITOL", "saccharide"),
        "LWH": (
            "4-(1H-PYRAZOL-5-YL)-2-{4-[(3S)-PYRROLIDIN-3-YLAMINO]QUINAZOLIN-2-YL}PHENOL",
            "non-polymer",
        ),
        "R2Z": (
            "N~2~-({(1R,2Z)-2-[(2R)-2-(FORMYLAMINO)-3-(PHOSPHONOOXY)PROPYLIDENE]CYCLOPENTYL}CARBONYL)-L-ARGININAMIDE",
            "non-polymer",
        ),
        "LWM": (
            "[(1R,2S,4R)-4-{[2-AMINO-5-(FORMYLAMINO)-6-OXO-1,6-DIHYDROPYRIMIDIN-4-YL]AMINO}-2-HYDROXYCYCLOPENTYL]METHYL DIHYDROGEN PHOSPHATE",
            "nucleic-acid",
        ),
        "GG1": (
            "4-{[1-METHYL-2,4-DIOXO-6-(3-PHENYLPROP-1-YN-1-YL)-1,4-DIHYDROQUINAZOLIN-3(2H)-YL]METHYL}BENZOIC ACID",
            "non-polymer",
        ),
        "SN0": ("N-(3-CARBOXYPROPANOYL)-L-NORVALINE", "non-polymer"),
        "Z0M": (
            "N-[(5R,6R,7R,8S)-6,7-DIHYDROXY-5-(HYDROXYMETHYL)-2-(2-PHENYLETHYL)-5,6,7,8-TETRAHYDROIMIDAZO[1,2-A]PYRIDIN-8-YL]-3-SULFANYLPROPANAMIDE",
            "non-polymer",
        ),
        "LWG": ("6,9-DICHLORO[1]BENZOTHIENO[3,2-D]PYRIMIDIN-4(3H)-ONE", "non-polymer"),
        "R2P": (
            "(2R)-2-AMINO-3-(4-HYDROXY-1,2,5-THIADIAZOL-3-YL)PROPANOIC ACID",
            "peptide",
        ),
        "959": ("BENZYL (2-OXOPROPYL)CARBAMATE", "non-polymer"),
        "M98": (
            "(S)-3-(2-(3-CYCLOPROPOXY-4-(DIFLUOROMETHOXY)PHENYL)-2-(5-(1,1,1,3,3,3-HEXAFLUORO-2-HYDROXYPROPAN-2-YL)THIAZOL-2-YL)ETHYL)PYRIDINE 1-OXIDE",
            "non-polymer",
        ),
        "951": (
            "2-[(2E)-2-{[5-(2-CHLOROPHENYL)FURAN-2-YL]METHYLIDENE}HYDRAZINO]BENZOIC ACID",
            "non-polymer",
        ),
        "950": (
            "2-{5-[AMINO(IMINIO)METHYL]-6-FLUORO-1H-BENZIMIDAZOL-2-YL}-6-ISOBUTOXYBENZENOLATE",
            "non-polymer",
        ),
        "GG2": (
            "1-(4-METHOXYPHENYL)-7-OXO-6-[4-(2-OXOPIPERIDIN-1-YL)PHENYL]-4,5,6,7-TETRAHYDRO-1H-PYRAZOLO[3,4-C]PYRIDINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "957": (
            "N-[(1S,2R)-1-(3,5-DIFLUOROBENZYL)-3-({1-[4-(2,2-DIMETHYLPROPYL)THIOPHEN-2-YL]CYCLOPROPYL}AMINO)-2-HYDROXYPROPYL]ACETAMIDE",
            "non-polymer",
        ),
        "956": (
            "1-METHYL-2-OXO-4-[4-(THIOPHEN-2-YLCARBONYL)PIPERAZIN-1-YL]-1,2-DIHYDROQUINOLINE-3-CARBONITRILE",
            "non-polymer",
        ),
        "SN1": ("2,4-DIHYDROXY-TRANS CINNAMIC ACID", "non-polymer"),
        "SBX": (
            "1-CYCLOHEXYL-3-PHENYL-1-PROPYL-1-(3,3-DIMETHYL-1,2-DIOXYPENTYL)-2-PIPERIDINE CARBOXYLATE",
            "non-polymer",
        ),
        "NHO": (
            "NICOTINAMIDE-ADENINE-DINUCLEOTIDE-5-HYDROXY-4-OXONORVALINE",
            "non-polymer",
        ),
        "GG9": (
            "(3R,4S,5R)-5-[(1R)-1-CARBOXY-2,2-DIFLUORO-1-(PHOSPHONOOXY)ETHOXY]-4-HYDROXY-3-(PHOSPHONOOXY)CYCLOHEX-1-ENE-1-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "PHY": (
            "1(S)-AMINOETHYL-(2-CARBOXYPROPYL)PHOSPHORYL-PHOSPHINIC ACID",
            "non-polymer",
        ),
        "SBY": ("3-[DODECYL(DIMETHYL)AMMONIO]PROPANE-1-SULFONATE", "non-polymer"),
        "KGQ": (
            "DIETHYL ({4-[(2S,3R)-2-({[(3R,3AS,6AR)-HEXAHYDROFURO[2,3-B]FURAN-3-YLOXY]CARBONYL}AMINO)-3-HYDROXY-4-{ISOBUTYL[(4-METHOXYPHENYL)SULFONYL]AMINO}BUTYL]PHENOXY}METHYL)PHOSPHONATE",
            "non-polymer",
        ),
        "KGR": ("5-KETO-4-DEOXYGLUCARATE", "non-polymer"),
        "EZT": (
            "(17BETA)-17-{(E)-2-[2-(TRIFLUOROMETHYL)PHENYL]VINYL}ESTRA-1(10),2,4-TRIENE-3,17-DIOL",
            "non-polymer",
        ),
        "LSG": (
            "5-ACETYL-2-[(2-FLUORO-4-IODOPHENYL)AMINO]-N-(2-HYDROXYETHOXY)-1-METHYL-1H-PYRROLE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "EZV": (
            "4-{3-[7-(4-METHYLPIPERAZIN-1-YL)-1H-BENZIMIDAZOL-2-YL]-1H-INDAZOL-6-YL}ANILINE",
            "non-polymer",
        ),
        "T80": ("METHYLPENTA(OXYETHYL) HEPTADECANOATE", "non-polymer"),
        "QDK": ("L-RIBULOSE", "saccharide"),
        "EZR": (
            "3-METHOXY-4-{3-[4-(4-METHYLPIPERAZIN-1-YL)-1H-BENZIMIDAZOL-2-YL]-1H-INDAZOL-6-YL}ANILINE",
            "non-polymer",
        ),
        "EZL": ("6-ETHOXY-1,3-BENZOTHIAZOLE-2-SULFONAMIDE", "non-polymer"),
        "TK1": (
            "1-({5-[4-(METHYLSULFONYL)PHENOXY]-2-PYRIDIN-2-YL-1H-BENZIMIDAZOL-6-YL}METHYL)PYRROLIDINE-2,5-DIONE",
            "non-polymer",
        ),
        "KGC": ("N~6~-[(2R)-2-CARBOXY-5-OXOTETRAHYDROFURAN-2-YL]-L-LYSINE", "peptide"),
        "QDR": (
            "2-{[(3-AMINO-1H-1,2,4-TRIAZOL-5-YL)SULFANYL]METHYL}-8-METHYLQUINAZOLIN-4(3H)-ONE",
            "non-polymer",
        ),
        "QDS": ("N-(QUINOXALIN-2-YLCARBONYL)-D-SERINE", "peptide"),
        "JEF": ("JEFFAMINE", "non-polymer"),
        "R24": (
            "1-[3-TERT-BUTYL-1-(4-METHYLPHENYL)-1H-PYRAZOL-5-YL]-3-{3-[(6-NITROQUINOLIN-4-YL)AMINO]PHENYL}UREA",
            "non-polymer",
        ),
        "R23": (
            "METHYL-[4-(4-PIPERIDINE-1-YLMETHYL-PHENYL)-CYCLOHEXYL]-CARBAMINIC ACID-(4-CHLOROPHENYL)-ESTER",
            "non-polymer",
        ),
        "R22": (
            "4-(3,5-DIMETHYLPHENOXY)-5-(FURAN-2-YLMETHYLSULFANYLMETHYL)-3-IODO-6-METHYLPYRIDIN-2(1H)-ONE",
            "non-polymer",
        ),
        "R21": (
            "2-(3-CHLORO-4-HYDROXYPHENYL)-N-[2-(4-SULFAMOYLPHENYL)ETHYL]ACETAMIDE",
            "non-polymer",
        ),
        "R20": ("4-(3-METHYLPHENYL)-1H-1,2,3-TRIAZOLE", "non-polymer"),
        "CPH": (
            "1,8-DIHYDROXY-7-METHYL-3-(1,3,4-TRIHYDROXY-2-OXO-PENTYL)-3,10-DIHYDRO-2H-ANTHRACEN-9-ONE",
            "non-polymer",
        ),
        "CPI": ("6-CARBOXYPIPERIDINE", "peptide"),
        "CPJ": (
            "BIS-1,2-{[(Z)-2-CARBOXY-2-METHYL-1,3-DIOXANE]-5-YLOXYCARBAMOYL}-ETHANE",
            "non-polymer",
        ),
        "CPK": (
            "BIS-1,2-{[(Z)-2CARBOXY-2-METHYL-1,3-DIOXANE]-5-YLOXYCARBONYL}-PIPERAZINE",
            "non-polymer",
        ),
        "CPL": ("1-PALMITOYL-2-LINOLEOYL-SN-GLYCERO-3-PHOSPHOCHOLINE", "non-polymer"),
        "CPM": ("S-(2-CARBOXY-3-PHENYLPROPYL)THIODIIMINE-S-METHANE", "non-polymer"),
        "CPN": ("2-AMINOETHYLGLYCINE-CARBONYLMETHYLENE-CYTOSINE", "peptide"),
        "CPO": ("COPROGEN", "non-polymer"),
        "P8S": ("HEPTADECAFLUORO-1-OCTANESULFONIC ACID", "non-polymer"),
        "CPA": (
            "2'-DEOXYCYTIDINE-2'-DEOXYADENOSINE-3',5'-MONOPHOSPHATE",
            "non-polymer",
        ),
        "CPB": (
            "2-(2-CHLORO-PHENYL)-5,7-DIHYDROXY-8-(3-HYDROXY-1-METHYL-PIPERIDIN-4-YL)-4H-BENZOPYRAN-4-ONE",
            "non-polymer",
        ),
        "CPC": ("2-METHYL-1-METHYLAMINO-CYCLOPROPANE CARBOXYLIC ACID", "peptide"),
        "CPD": (
            "[1-(3-DIMETHYLAMINO-PROPYL)-3-ETHYL-UREIDO]-[4-(2,2,2-TRIFLUORO-ACETYLAMINO)-BENZYL]PHOSPHINIC ACID-2-(2,2-DIHYDRO-ACETYLAMINO)-3-HYDROXY-1-(4-NITROPHENYL)-PROPYL ESTER",
            "non-polymer",
        ),
        "4XB": (
            "3-METHOXYBENZYL 3-METHYL-4-(PIPERIDIN-4-YLOXY)-1-BENZOFURAN-2-CARBOXYLATE",
            "non-polymer",
        ),
        "CPF": (
            "1-CYCLOPROPYL-6-FLUORO-4-OXO-7-PIPERAZIN-1-YL-1,4-DIHYDROQUINOLINE-3-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "CPG": ("GUANOSINE-5'-PHOSPHATE", "non-polymer"),
        "CPX": (
            "N-[(2R,3S)-1-((2S)-2-{[(CYCLOPENTYLAMINO)CARBONYL]AMINO}-3-METHYLBUTANOYL)-2-(1-FORMYL-1-CYCLOBUTYL)PYRROLIDINYL]CYCLOPROPANECARBOXAMIDE",
            "non-polymer",
        ),
        "CPY": (
            "1-(3-AMINO-4-CYCLOHEXYL-2-HYDROXYBUTYL)PYRIDIN-2(1H)-ONE",
            "non-polymer",
        ),
        "CPZ": ("4-(4-CHLOROPHENYL)IMIDAZOLE", "non-polymer"),
        "P8H": (
            "3-BROMO-5-[(4-PYRROLIDIN-1-YLPIPERIDIN-1-YL)CARBONYL]PYRIDINE",
            "non-polymer",
        ),
        "MD3": ("N-(CARBOXYCARBONYL)-S-(NAPHTHALEN-2-YLMETHYL)-L-CYSTEINE", "peptide"),
        "CPP": ("2-CYCLOPROPYLMETHYLENEPROPANAL", "non-polymer"),
        "CPQ": ("N,N-BIS(3-D-GLUCONAMIDOPROPYL)DEOXYCHOLAMIDE", "non-polymer"),
        "CPR": ("?", "non-polymer"),
        "CPS": (
            "3-[(3-CHOLAMIDOPROPYL)DIMETHYLAMMONIO]-1-PROPANESULFONATE",
            "non-polymer",
        ),
        "CPT": ("CISPLATIN", "non-polymer"),
        "CPU": ("?", "non-polymer"),
        "CPV": (
            "5-AMINO-6-CYCLOHEXYL-4-HYDROXY-2-ISOPROPYL-HEXANOIC ACID",
            "non-polymer",
        ),
        "CPW": (
            "(S)-2-AMINO-3-(1,3,5,7-PENTAHYDRO-2,4-DIOXO-CYCLOPENTA[E]PYRIMIDIN-1-YL) PROIONIC ACID",
            "non-polymer",
        ),
        "TR9": ("(5S)-2,5-DIHYDROXY-3-METHYLCYCLOHEX-2-ENE-1,4-DIONE", "non-polymer"),
        "5NS": ("5-AMINONAPHTHALENE-1-SULFONIC ACID", "non-polymer"),
        "JEN": (
            "3-METHOXY-6-[4-(3-METHYLPHENYL)-1-PIPERAZINYL]PYRIDAZINE",
            "non-polymer",
        ),
        "5NA": ("2-HYDROXY-5-METHYL NAPHTHOIC ACID", "non-polymer"),
        "5NB": ("5'-AMINO-8-BROMO-5'-DEOXYADENOSINE", "non-polymer"),
        "5NC": ("5-AZA-CYTIDINE-5'MONOPHOSPHATE", "nucleic-acid"),
        "5NH": ("?", "non-polymer"),
        "5NI": ("5-NITROINDAZOLE", "non-polymer"),
        "DCN": ("DICLOSAN", "non-polymer"),
        "MD5": ("N-(CARBOXYCARBONYL)-S-(3-NITROBENZYL)-L-CYSTEINE", "peptide"),
        "TQ5": ("5-[4-TERT-BUTYLPHENYLSULFANYL]-2,4-QUINAZOLINEDIAMINE", "non-polymer"),
        "5N5": ("5'-AMINO-5'-DEOXYADENOSINE", "non-polymer"),
        "PZM": ("1-(4-METHOXYPHENYL)METHANAMINE", "non-polymer"),
        "TR0": (
            "(4R)-N-(BIPHENYL-4-YLMETHYL)-4-HYDROXY-1-[(3-METHYLISOXAZOL-5-YL)ACETYL]-L-PROLINAMIDE",
            "non-polymer",
        ),
        "P89": (
            "1-{[(1E)-(3-HYDROXY-2-METHYL-5-{[(TRIHYDROXY-LAMBDA^5^-PHOSPHANYL)OXY]METHYL}PYRIDIN-4-YL)METHYLIDENE]AMINO}UNDECAN-2-ONE",
            "non-polymer",
        ),
        "P88": (
            "7-(4-{[3-(2,6-DIMETHYLPHENYL)-5-(1-METHYLETHYL)ISOXAZOL-4-YL]METHOXY}PHENYL)ISOQUINOLINE-3-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "PZL": (
            "3-[6-BROMO-2-FLUORO-3-(1H-PYRAZOLO[3,4-C]PYRIDAZIN-3-YLMETHYL)PHENOXY]-5-CHLOROBENZONITRILE",
            "non-polymer",
        ),
        "P83": ("1-(9-ETHYL-9H-CARBAZOL-3-YL)-N-METHYLMETHANAMINE", "non-polymer"),
        "P81": ("4-[(E)-2-NITROETHENYL]PHENOL", "non-polymer"),
        "P80": ("3-[(E)-2-NITROETHENYL]PHENOL", "non-polymer"),
        "P86": (
            "2-{[4-(DIETHYLAMINO)PIPERIDIN-1-YL]METHYL}-4,6-DIIODOPHENOL",
            "non-polymer",
        ),
        "P84": (
            "2,4-BIS(IODANYL)-6-[[METHYL-(1-METHYLPIPERIDIN-4-YL)AMINO]METHYL]PHENOL",
            "non-polymer",
        ),
        "CP8": (
            "CP-271485; (6R)-4-BENZYL-6-(1-METHYL-2,2-DIOXIDO-1,3-DIHYDRO-2,1-BENZISOTHIAZOL-5-YL)MORPHOLIN-3-ONE",
            "non-polymer",
        ),
        "CP9": (
            "3-[4-(2-METHYL-IMIDAZO[4,5-C]PYRIDIN-1-YL)BENZYL]-3H-BENZOTHIAZOL-2-ONE",
            "non-polymer",
        ),
        "OVA": (
            "3,4-DIHYDROXY-2-METHOXY-4-METHYL-3-[2-METHYL-3-(3-METHYL-BUT-2-ENYL) -OXIRANYL]-CYCLOHEXANONE",
            "non-polymer",
        ),
        "CP0": ("IRINOTECAN, CPT-11", "non-polymer"),
        "6RG": ("2-PHENYL-N~4~-(2-PHENYLETHYL)QUINAZOLINE-4,7-DIAMINE", "non-polymer"),
        "CP2": ("1,2-DICHLORO-PROPANE", "non-polymer"),
        "4X4": (
            "N-[2-(2-METHOXY-1H-DIPYRIDO[2,3-A:3',2'-E]PYRROLIZIN-11-YL)ETHYL]FURAN-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "CP4": (
            "BIS[5-CHLORO-1H-INDOL-2-YL-CARBONYL-AMINOETHYL]-ETHYLENE GLYCOL",
            "non-polymer",
        ),
        "CP5": ("(2Z)-3-{[OXIDO(OXO)PHOSPHINO]OXY}-2-PHENYLACRYLATE", "non-polymer"),
        "CP6": ("5-(4-CHLORO-PHENYL)-6-ETHYL-PYRIMIDINE-2,4-DIAMINE", "non-polymer"),
        "CP7": ("6-ETHYL-5-PHENYLPYRIMIDINE-2,4-DIAMINE", "non-polymer"),
        "591": (
            "(5S)-2-AMINO-5-(1-ETHYL-1H-PYRAZOL-4-YL)-3-METHYL-5-[3-(PYRIMIDIN-5-YL)PHENYL]-3,5-DIHYDRO-4H-IMIDAZOL-4-ONE",
            "non-polymer",
        ),
        "593": (
            "(5S)-2-AMINO-3-METHYL-5-PHENYL-5-[(3S,5S,7S)-TRICYCLO[3.3.1.1~3,7~]DEC-1-YL]-3,5-DIHYDRO-4H-IMIDAZOL-4-ONE",
            "non-polymer",
        ),
        "592": (
            "N-BENZYL-6-[(3-ENDO)-3-{[(3-METHOXY-2-METHYLPHENYL)CARBONYL]AMINO}-8-AZABICYCLO[3.2.1]OCT-8-YL]PYRIDINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "595": (
            "N-[(1S,5S)-5-{[(4-AMINOPHENYL)SULFONYL](3-METHYLBUTYL)AMINO}-1-METHYL-6-OXOHEXYL]-NALPHA-(METHOXYCARBONYL)-BETA-PHENYL-L-PHENYLALANINAMIDE",
            "non-polymer",
        ),
        "596": (
            "(2Z)-2-(1H-INDAZOL-3-YLMETHYLIDENE)-6-METHOXY-7-(PIPERAZIN-1-YLMETHYL)-1-BENZOFURAN-3(2H)-ONE",
            "non-polymer",
        ),
        "599": (
            "2-[3-{4-CHLORO-3-[(4-CHLOROPHENYL)ETHYNYL]PHENYL}-1-(3-MORPHOLIN-4-YLPROPYL)-1,4,6,7-TETRAHYDRO-5H-PYRAZOLO[4,3-C]PYRIDIN-5-YL]-2-OXOETHANOL",
            "non-polymer",
        ),
        "598": (
            "6-{[1-(BENZYLSULFONYL)PIPERIDIN-4-YL]AMINO}-3-(CARBOXYMETHOXY)THIENO[3,2-B][1]BENZOTHIOPHENE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "D78": (
            "(5S)-N-[(1S,2R)-2-HYDROXY-3-[[(4-METHOXYPHENYL)SULFONYL](2-METHYLPROPYL)AMINO]-1-(PHENYLMETHYL)PROPYL]-2-OXO-3-[3-(TRIFLUOROMETHYL)PHENYL]-5-OXAZOLIDINECARBOXAMIDE",
            "non-polymer",
        ),
        "MX5": (
            "{1-[4-(3,4-DIHYDROQUINOLIN-1(2H)-YL)-4-OXOBUTANOYL]HYDRAZINO}ACETIC ACID",
            "non-polymer",
        ),
        "XBP": ("XYLULOSE-1,5-BISPHOSPHATE", "saccharide"),
        "MX3": (
            "(1-{4-[BENZYL(METHYL)AMINO]-4-OXOBUTANOYL}HYDRAZINO)ACETIC ACID",
            "non-polymer",
        ),
        "MX1": (
            "(2R)-2-((R)-CARBOXY{[CARBOXY(4-HYDROXYPHENYL)ACETYL]AMINO}METHOXYMETHYL)-5-METHYLENE-5,6-DIHYDRO-2H-1,3-OXAZINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "NHT": (
            "(3AR,5R,6S,7R,7AR)-2-(ETHYLAMINO)-5-(HYDROXYMETHYL)-5,6,7,7A-TETRAHYDRO-3AH-PYRANO[3,2-D][1,3]THIAZOLE-6,7-DIOL",
            "non-polymer",
        ),
        "SM1": (
            "N-(BENZYLSULFONYL)-D-SERYL-N-(4-{[AMINO(IMINO)METHYL]AMINO}BENZYL)-L-ALANINAMIDE",
            "non-polymer",
        ),
        "TRL": (
            "{1-[2-(1-FORMYL-PROPYL)-3-METHANESULFONYLAMINO-PYRROLIDINE-1-CARBONYL]-2-METHYL-PROPYL}-CARBAMIC ACID TERT-BUTYL ESTER",
            "non-polymer",
        ),
        "9MQ": (
            "{BENZYL-[(S)-3-[((S)-1-CARBAMIMIDOYL-PIPERIDIN-3-YLMETHYL)-CARBAMOYL]-2-(NAPHTHALENE-2-SULFONYLAMINO)-PROPIONYL]-AMINO}-ACETIC ACID",
            "non-polymer",
        ),
        "9MR": (
            "(3R,4R,5R)-3-HYDROXY-5-(HYDROXYMETHYL)PIPERIDIN-4-YL BETA-D-GLUCOPYRANOSIDE",
            "non-polymer",
        ),
        "9MU": (
            "(R)-N-((S)-1-CARBAMIMIDOYL-PIPERIDIN-3-YLMETHYL)-2-(NAPHTHALENE-2-SULFONYLAMINO)-3-PHENYL-PROPIONAMIDE",
            "non-polymer",
        ),
        "9MT": (
            "3-(3-BUTYL-6-METHOXY-2-METHYL-QUINOLIN-4-YL)SULFANYLPROPANENITRILE",
            "non-polymer",
        ),
        "SN9": (
            "8-METHOXY-1-METHYL-4-(4-(4-(1-METHYLPYRIDINIUM-4-YLAMINO)PHENYLCARBAMOYL)PHENYLAMINO)QUINOLINIUM",
            "non-polymer",
        ),
        "9MX": (
            "(R)-1-[(S)-3-[((S)-1-CARBAMIMIDOYL-PIPERIDIN-3-YLMETHYL)-CARBAMOYL]-2-(NAPHTHALENE-2-SULFONYLAMINO)-PROPIONYL]-4-METHYL-PIPERIDINE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "3LV": (
            "4-[2,2-BIS(HYDROXYMETHYL)-5-OXOPYRROLIDIN-1-YL]-3-[(DIPROPYLAMINO)METHYL]BENZOIC ACID",
            "non-polymer",
        ),
        "9MZ": (
            "2-{(5S)-2-[(2E)-2-(2-HYDROXYBENZYLIDENE)HYDRAZINYL]-4-OXO-4,5-DIHYDRO-1,3-THIAZOL-5-YL}-N-[3-(TRIFLUOROMETHYL)PHENYL]ACETAMIDE",
            "non-polymer",
        ),
        "3LP": (
            "1-(CYCLOHEXYLAMINO)-3-(6-METHYL-3,4-DIHYDRO-1H-CARBAZOL-9(2H)-YL)PROPAN-2-OL",
            "non-polymer",
        ),
        "ZZF": (
            "4-({4-[(2,6-DIMETHYLPYRIDIN-3-YL)OXY]PYRIDIN-2-YL}AMINO)BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "3LR": ("3,6-DIDEOXY-L-ARABINO-HEXONIC ACID", "saccharide"),
        "NHW": ("2-OXOPENTADECYL-COA", "non-polymer"),
        "3LM": (
            "(2E)-2-[({3-HYDROXY-2-METHYL-5-[(PHOSPHONOOXY)METHYL]PYRIDIN-4-YL}METHYL)AMINO]-4-(METHYLSULFANYL)BUT-2-ENOIC ACID",
            "non-polymer",
        ),
        "33N": ("N-(3,3-DIPHENYLPROPYL)PYRIDINE-3-CARBOXAMIDE", "non-polymer"),
        "3LI": (
            "(4R)-3-(4-[4-(2-CHLOROPHENYL)PIPERAZIN-1-YL]-6-{[2-METHYL-6-(METHYLCARBAMOYL)PHENYL]AMINO}-1,3,5-TRIAZIN-2-YL)-N-[2-(HYDROXYAMINO)-2-OXOETHYL]-1,3-THIAZOLIDINE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "9MG": ("9-METHYLGUANINE", "non-polymer"),
        "3LG": (
            "(5-{[(2R)-1-(4-{3-[(2-METHOXYBENZYL)OXY]PROPOXY}PHENYL)-6-OXOPIPERAZIN-2-YL]METHOXY}-1H-INDOL-1-YL)ACETIC ACID",
            "non-polymer",
        ),
        "3LA": ("3-OXODODECANOIC ACID", "non-polymer"),
        "IVK": (
            "(3R,4R)-N-(4-CHLOROPHENYL)-N'-[2-FLUORO-4-(2-OXOPYRIDIN-1(2H)-YL)PHENYL]-1-(2,2,2-TRIFLUOROETHYL)PYRROLIDINE-3,4-DICARBOXAMIDE",
            "non-polymer",
        ),
        "TRB": (
            "4-CHLORO-8-METHYL-7-(3-METHYL-BUT-2-ENYL)-6,7,8,9-TETRAHYDRO-2H-2,7,9A-TRIAZA-BENZO[CD]AZULENE-1-THIONE",
            "non-polymer",
        ),
        "SBA": (
            "CARBOBENZYLOXY-(L)-LEUCINYL-(L)LEUCINYL METHOXYMETHYLKETONE",
            "non-polymer",
        ),
        "3L1": (
            "(2S,3R,5S)-5-(6-AMINO-9H-PURIN-9-YL)-TETRAHYDRO-2-(HYDROXYMETHYL)FURAN-3-OL",
            "non-polymer",
        ),
        "TRF": ("N1-FORMYL-TRYPTOPHAN", "peptide"),
        "MQD": ("2-METHYLPENTANE-1,2,4-TRIOL", "non-polymer"),
        "D7S": (
            "(5BETA,14BETA,17ALPHA,25S)-3-OXOCHOLEST-7-EN-26-OIC ACID",
            "non-polymer",
        ),
        "D7P": (
            "D-PHENYLALANYL-N~5~-[(2,2-DIHYDROXYHYDRAZINO)(IMINO)METHYL]-D-ORNITHINAMIDE",
            "non-polymer",
        ),
        "59R": (
            "6-{[(3R,4R)-4-(2-{[2-(3-FLUOROPHENYL)ETHYL]AMINO}ETHOXY)PYRROLIDIN-3-YL]METHYL}PYRIDIN-2-AMINE",
            "non-polymer",
        ),
        "D7V": (
            "(3Z)-3-(DIBENZO[B,E]OXEPIN-11(6H)-YLIDENE)-N,N-DIMETHYLPROPAN-1-AMINE",
            "non-polymer",
        ),
        "59W": (
            "6-{[(3S,4S)-4-(2-{[2-(3-FLUOROPHENYL)ETHYL]AMINO}ETHOXY)PYRROLIDIN-3-YL]METHYL}PYRIDIN-2-AMINE",
            "non-polymer",
        ),
        "MXY": ("2-O-METHYL FUCOSE", "saccharide"),
        "MXT": (
            "1-(1-HYDROXY-2,2,6,6-TETRAMETHYLPIPERIDIN-4-YL)PYRROLIDINE-2,5-DIONE",
            "non-polymer",
        ),
        "MXS": (
            "(2S)-2-[[2-METHANOYL-7-(METHOXYCARBONYLAMINO)INDOLIZIN-3-YL]AMINO]-3-METHYL-3-SULFINO-BUTANOIC ACID",
            "non-polymer",
        ),
        "MXP": ("MYXOPYRONIN B", "non-polymer"),
        "59A": (
            "N-{[(2S,3S)-3-(ETHOXYCARBONYL)OXIRAN-2-YL]CARBONYL}-L-ISOLEUCINE",
            "non-polymer",
        ),
        "59C": (
            "[(3R,5S,6R,7R,12E)-5,11-DIMETHOXY-3,7,9,15,19-PENTAMETHYL-6-OXIDANYL-16,20,22-TRIS(OXIDANYLIDENE)-21-(PROP-2-ENYLAMINO)-17-AZABICYCLO[16.3.1]DOCOSA-1(21),8,12,14,18-PENTAEN-10-YL] CARBAMATE",
            "non-polymer",
        ),
        "D7A": ("N-[2-(3,4-DIMETHOXYPHENYL)ETHYL]-4-SULFAMOYLBENZAMIDE", "non-polymer"),
        "D7G": (
            "((1S)-1-((((1S)-1-BENZYL-3-(CYCLOPROPYLAMINO)-2,3-DIOXOPROPYL)AMINO)CARBONYL)-3-METHYLBUTYL)CARBAMIC ACID 5-METHOXY-3-OXAPENTYL ESTER",
            "non-polymer",
        ),
        "59G": (
            "6-(4-{[3-(2,6-DICHLOROPHENYL)-5-(PROPAN-2-YL)-1,2-OXAZOL-4-YL]METHOXY}PHENYL)-1-BENZOTHIOPHENE-3-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "MXI": (
            "4-(2-TERT-BUTYL-4-(6-METHOXYNAPHTHALEN-2-YL)-3H-IMIDAZOL-4-YL)PYRIDIN-2-AMINE",
            "non-polymer",
        ),
        "MXF": (
            "(2S,3R)-4-(2-AMINOETHYLCARBAMOYLOXY)-2-[(2-METHANOYLINDOLIZIN-3-YL)AMINO]-3-METHYL-3-SULFINO-BUTANOIC ACID",
            "non-polymer",
        ),
        "D7K": (
            "3-[(4-AMINO-2-METHYLPYRIMIDIN-5-YL)METHYL]-2-{(S)-HYDROXY[(R)-HYDROXY(METHOXY)PHOSPHORYL]PHENYLMETHYL}-5-(2-{[(R)-HYDROXY(PHOSPHONOOXY)PHOSPHORYL]OXY}ETHYL)-4-METHYL-1,3-THIAZOL-3-IUM",
            "non-polymer",
        ),
        "MXD": ("6-PIPERIDIN-1-YLPYRIMIDINE-2,4-DIAMINE 3-OXIDE", "non-polymer"),
        "33B": ("?", "non-polymer"),
        "MXC": (
            "(2S,3R)-2-[(7-AMINOCARBONYL-2-METHANOYL-INDOLIZIN-3-YL)AMINO]-4-AMINOCARBONYLOXY-3-METHYL-3-SULFINO-BUTANOIC ACID",
            "non-polymer",
        ),
        "MXA": (
            "6-(2,5-DIMETHOXY-BENZYL)-5-METHYL-PYRIDO[2,3-D]PYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "339": (
            "(S)-2-(3-((R)-1-(4-BROMOPHENYL)ETHYL)UREIDO)-N-((S)-1-((S)-5-GUANIDINO-1-OXO-1-(THIAZOL-2-YL)PENTAN-2-YLAMINO)-3-METHYL-1-OXOBUTAN-2-YL)-5-UREIDOPENTANAMIDE",
            "non-polymer",
        ),
        "338": ("2-(5-HYDROXY-NAPHTHALEN-1-YL)-1,3-BENZOOXAZOL-6-OL", "non-polymer"),
        "B98": (
            "(3R)-3-(AMINOMETHYL)-9-METHOXY-1,2,3,4-TETRAHYDRO-5H-[1]BENZOTHIENO[3,2-E][1,4]DIAZEPIN-5-ONE",
            "non-polymer",
        ),
        "B99": (
            "(2S)-1-{[6-FURAN-3-YL-5-(3-METHYL-2H-INDAZOL-5-YL)PYRIDIN-3-YL]OXY}-3-(1H-INDOL-3-YL)PROPAN-2-AMINE",
            "non-polymer",
        ),
        "B96": (
            "1-(5-TERT-BUTYL-2-P-TOLYL-2H-PYRAZOL-3-YL)-3-[4-(2-MORPHOLIN-4-YL-ETHOXY)-NAPHTHALEN-1-YL]-UREA",
            "non-polymer",
        ),
        "B97": (
            "(10R)-10-METHYL-3-(6-METHYLPYRIDIN-3-YL)-9,10,11,12-TETRAHYDRO-8H-[1,4]DIAZEPINO[5',6':4,5]THIENO[3,2-F]QUINOLIN-8-ONE",
            "non-polymer",
        ),
        "B94": (
            "(3S)-3-({[(5S,8R)-2-(3-CARBOXYPROPYL)-8-(2-{[(4-CHLOROPHENYL)ACETYL]AMINO}ETHYL)-1,3-DIOXO-2,3,5,8-TETRAHYDRO-1H-[1,2,4]TRIAZOLO[1,2-A]PYRIDAZIN-5-YL]CARBONYL}AMINO)-4-OXOPENTANOIC ACID",
            "non-polymer",
        ),
        "B95": (
            "N-[2-(1H-BENZIMIDAZOL-2-YL)ETHYL]-2-(2-BROMOPHENOXY)ACETAMIDE",
            "non-polymer",
        ),
        "B92": (
            "(3S)-3-({[(5S,10AS)-2-{(2S)-4-CARBOXY-2-[(PHENYLACETYL)AMINO]BUTYL}-1,3-DIOXO-2,3,5,7,8,9,10,10A-OCTAHYDRO-1H-[1,2,4]TRIAZOLO[1,2-A]CINNOLIN-5-YL]CARBONYL}AMINO)-4-OXOPENTANOIC ACID",
            "non-polymer",
        ),
        "B93": (
            "(3S)-3-({[(5S)-2-{2-[(1H-BENZIMIDAZOL-5-YLCARBONYL)AMINO]ETHYL}-7-(CYCLOHEXYLMETHYL)-1,3-DIOXO-2,3,5,8-TETRAHYDRO-1H-[1,2,4]TRIAZOLO[1,2-A]PYRIDAZIN-5-YL]CARBONYL}AMINO)-4-OXOPENTANOIC ACID",
            "non-polymer",
        ),
        "B90": (
            "N-[4-(DIMETHYLPHOSPHORYL)PHENYL]-9-[(E)-2-(5-METHYL-1H-INDAZOL-4-YL)ETHENYL]-9H-PURIN-6-AMINE",
            "non-polymer",
        ),
        "7XY": (
            "{(3-CHLOROBENZYL)[(5-{[(3,3-DIPHENYLPROPYL)AMINO]SULFONYL}-2-THIENYL)METHYL]AMINO}(OXO)ACETIC ACID",
            "non-polymer",
        ),
        "TRQ": (
            "2-AMINO-3-(6,7-DIOXO-6,7-DIHYDRO-1H-INDOL-3-YL)-PROPIONIC ACID",
            "peptide",
        ),
        "TRS": ("2-AMINO-2-HYDROXYMETHYL-PROPANE-1,3-DIOL", "non-polymer"),
        "L4P": ("3-[GLYCEROLYLPHOSPHONYL]-[1,2-DI-PHYTANYL]GLYCEROL", "non-polymer"),
        "L4T": (
            "N-[(CYCLOPENTYLOXY)CARBONYL]-3-METHYL-L-VALYL-(4R)-4-[(8-BROMO-7-METHOXY-2-{2-[(2-METHYLPROPANOYL)AMINO]-1,3-THIAZOL-4-YL}QUINOLIN-4-YL)OXY]-N-[(1R,2S)-1-CARBOXY-2-ETHENYLCYCLOPROPYL]-L-PROLINAMIDE",
            "non-polymer",
        ),
        "TRU": (
            "6-CHLORO-3-(DICHLOROMETHYL)-3,4-DIHYDRO-2H-1,2,4-BENZOTHIADIAZINE-7-SULFONAMIDE 1,1-DIOXIDE",
            "non-polymer",
        ),
        "ITC": ("ISO-7-CHLORTETRACYCLINE", "non-polymer"),
        "TRT": ("FRAGMENT OF TRITON X-100", "non-polymer"),
        "L4C": (
            "(2E)-2-{[(2S)-2-(ACETYLAMINO)-2-CARBOXYETHOXY]IMINO}PENTANEDIOIC ACID",
            "non-polymer",
        ),
        "L4G": (
            "6-(4-METHYLSULFONYL-PHENYL)-5-[4-(2-PIPERIDIN-1-YLETHOXY)PHENOXY]NAPHTHALEN-2-OL",
            "non-polymer",
        ),
        "VD5": (
            "(1R,3R,7E,17E)-17-(5-HYDROXY-1,5-DIMETHYLHEXYLIDENE)-2-METHYLENE-9,10-SECOESTRA-5,7-DIENE-1,3-DIOL",
            "non-polymer",
        ),
        "ZT4": (
            "5-{(E)-[(2R)-2-HYDROXY-2,3-DIHYDRO-1H-INDEN-1-YLIDENE]METHYL}-1,3-BENZODIOXOLE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "ZM3": ("MYCINAMICIN III", "non-polymer"),
        "TB7": (
            "TERT-BUTYL 3-[(4-OXO-3H-PYRIDO[2,3-D]PYRIMIDIN-2-YL)AMINO]AZETIDINE-1-CARBOXYLATE",
            "non-polymer",
        ),
        "KOP": ("METHYL DIHYDROGEN PHOSPHATE", "non-polymer"),
        "L42": (
            "2-(ACETYLAMINO)-2-DEOXY-3-O-(4-DEOXY-ALPHA-L-THREO-HEX-4-ENOPYRANURONOSYL)-4-O-SULFO-BETA-D-GALACTOPYRANOSE",
            "non-polymer",
        ),
        "L41": (
            "{4-[3-(4-ACETYL-3-HYDROXY-2-PROPYLPHENOXY)PROPOXY]PHENOXY}ACETIC ACID",
            "non-polymer",
        ),
        "SBE": ("1,3,2-DIOXABOROLAN-2-OL", "non-polymer"),
        "L47": (
            "3-[(5S)-1-ACETYL-3-(2-CHLOROPHENYL)-4,5-DIHYDRO-1H-PYRAZOL-5-YL]PHENOL",
            "non-polymer",
        ),
        "NHR": (
            "2-{4-[2-(2-AMINO-4-HYDROXY-QUINAZOLIN-6-YL)-1-CARBOXY-ETHYL]-BENZOYLAMINO}-PENTANEDIOIC ACID",
            "non-polymer",
        ),
        "L44": (
            "(2S)-1-HYDROXY-3-(TETRADECANOYLOXY)PROPAN-2-YL DOCOSANOATE",
            "non-polymer",
        ),
        "ZMA": (
            "4-{2-[(7-AMINO-2-FURAN-2-YL[1,2,4]TRIAZOLO[1,5-A][1,3,5]TRIAZIN-5-YL)AMINO]ETHYL}PHENOL",
            "non-polymer",
        ),
        "ZMG": (
            "(5R)-2-[(2-FLUOROPHENYL)AMINO]-5-(1-METHYLETHYL)-1,3-THIAZOL-4(5H)-ONE",
            "non-polymer",
        ),
        "IH5": (
            "{3,5-DICHLORO-4-[4-HYDROXY-3-(PROPAN-2-YL)PHENOXY]PHENYL}ACETIC ACID",
            "non-polymer",
        ),
        "VD4": (
            "(1R,3R,7E,17Z)-17-(5-HYDROXY-1,5-DIMETHYLHEXYLIDENE)-2-METHYLENE-9,10-SECOESTRA-5,7-DIENE-1,3-DIOL",
            "non-polymer",
        ),
        "ZMR": ("ZANAMIVIR", "non-polymer"),
        "ZMP": (
            "({(3S)-3-HYDROXY-2,2-DIMETHYL-4-OXO-4-[(3-OXO-3-{[2-(TETRADECANOYLSULFANYL)ETHYL]AMINO}PROPYL)AMINO]BUTYL}OXY)(OXO)PHOSPHONIUMOLATE",
            "non-polymer",
        ),
        "ZMQ": (
            "{[(3S)-4-{[3-({2-[(9Z)-HEXADEC-9-ENOYLSULFANYL]ETHYL}AMINO)-3-OXOPROPYL]AMINO}-3-HYDROXY-2,2-DIMETHYL-4-OXOBUTYL]OXY}(OXO)PHOSPHONIUMOLATE",
            "non-polymer",
        ),
        "TB6": ("[(2R)-2-HYDROXYPROPYL]PHOSPHONIC ACID", "non-polymer"),
        "33Y": (
            "1-METHYLETHYL 3-[(3,4-DIFLUOROPHENYL)CARBONYL]-1,1-DIMETHYL-1,2,3,6-TETRAHYDROAZEPINO[4,5-B]INDOLE-5-CARBOXYLATE",
            "non-polymer",
        ),
        "33Z": (
            "(2S)-2-(3-{[AMINO(IMINO)METHYL]AMINO}PHENYL)-3-[(R)-[(1R)-1-({N-[(BENZYLOXY)CARBONYL]-L-PHENYLALANYL}AMINO)-2-METHYLPROPYL](HYDROXY)PHOSPHORYL]PROPANOIC ACID",
            "non-polymer",
        ),
        "33U": (
            "BETA-PHENYL-D-PHENYLALANYL-N-(4-CARBAMIMIDOYLBENZYL)-L-PROLINAMIDE",
            "peptide",
        ),
        "33T": (
            "3-CHLORO-4-({(2R)-4-[4-FLUORO-2-(TRIFLUOROMETHYL)PHENYL]-2-METHYLPIPERAZIN-1-YL}SULFONYL)BENZAMIDE",
            "non-polymer",
        ),
        "33P": (
            "{3-[(3-HYDROXY-2-METHYL-5-PHOSPHONOOXYMETHYL-PYRIDIN-4-YLMETHYL)-AMINO]-2-METHYL-PROPYL}-PHOSPHONIC ACID",
            "non-polymer",
        ),
        "RLL": (
            "(2S)-3-[(4-CYANONAPHTHALEN-1-YL)OXY]-N-[4-CYANO-3-(TRIFLUOROMETHYL)PHENYL]-2-HYDROXY-2-METHYLPROPANAMIDE",
            "non-polymer",
        ),
        "B9M": ("(10E)-HEXADEC-10-EN-12-YN-1-OL", "non-polymer"),
        "33D": ("3,3-DIMETHYLBUTAN-1-OL", "non-polymer"),
        "B9D": (
            "(2R,3R,4R,5S,6R)-6-FLUORANYL-6-(HYDROXYMETHYL)OXANE-2,3,4,5-TETROL",
            "saccharide",
        ),
        "33F": (
            "N-[(2S)-BUTAN-2-YL]-6-[(3R)-3-{[4-(TRIFLUOROMETHOXY)BENZYL]CARBAMOYL}-4-{[4-(TRIFLUOROMETHOXY)PHENYL]SULFONYL}PIPERAZIN-1-YL]PYRIDAZINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "33A": (
            "N-BENZYL-4-[4-(3-CHLOROPHENYL)-1H-PYRAZOL-3-YL]-1H-PYRROLE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "B9A": ("MACROCYCLIC-BIS-9-AMINO-ACRIDINE", "non-polymer"),
        "ILC": ("?", "non-polymer"),
        "ZXA": (
            "N-{[(2R,3R,4R,5R,6S)-3,4,5-TRIHYDROXY-6-METHYLPIPERIDIN-2-YL]METHYL}ACETAMIDE",
            "non-polymer",
        ),
        "7X6": (
            "N-[3-[[4-[(5-CHLORO-1,3-BENZODIOXOL-4-YL)AMINO]PYRIMIDIN-2-YL]AMINO]PHENYL]METHANESULFONAMIDE",
            "non-polymer",
        ),
        "PZF": (
            "(3R,4S)-1-[6-(6-METHOXYPYRIDIN-3-YL)PYRIMIDIN-4-YL]-4-(2,4,5-TRIFLUOROPHENYL)PYRROLIDIN-3-AMINE",
            "non-polymer",
        ),
        "ZXB": ("(2S,3R,4S,5R)-2-BENZYLPIPERIDINE-3,4,5-TRIOL", "non-polymer"),
        "SBH": ("(S)-4-BROMO-3-HYDROXY-3-METHYLBUTYL DIPHOSPHATE", "non-polymer"),
        "ZXD": ("(2S,3R,4S,5R)-2-(1-METHYLETHYL)PIPERIDINE-3,4,5-TRIOL", "non-polymer"),
        "ZXG": (
            "2,6-DIBROMO-4-[2-(3,5-DIBROMO-4-HYDROXYPHENYL)PROPAN-2-YL]PHENYL HYDROGEN SULFATE",
            "non-polymer",
        ),
        "PZA": ("PYRAZINE-2-CARBOXAMIDE", "non-polymer"),
        "FP0": (
            "(3ALPHA)-15-HYDROXY-12,13-EPOXYTRICHOTHEC-9-EN-3-YL ACETATE",
            "non-polymer",
        ),
        "SBI": ("SORBINIL", "non-polymer"),
        "1JJ": ("[(1S)-1-HYDROXYPROPYL]PHOSPHONIC ACID", "non-polymer"),
        "XOP": (
            "8-METHYL-8-AZABICYCLO[3.2.1]OCTAN-3-YL PHENYLPHOSPHONATE",
            "non-polymer",
        ),
        "DQX": (
            "8-[3,5-DIFLUORO-4-(MORPHOLIN-4-YLMETHYL)PHENYL]-2-(1-PIPERIDIN-4-YL-1H-PYRAZOL-4-YL)QUINOXALINE",
            "non-polymer",
        ),
        "DQY": ("3-(DIMETHYL-LAMBDA~4~-SULFANYL)PROPANOIC ACID", "non-polymer"),
        "1JX": (
            "(2R)-2-METHYL-1,4-DIHYDROPYRIDO[2,3-B]PYRAZIN-3(2H)-ONE",
            "non-polymer",
        ),
        "1JZ": ("1-[4-(HYDROXYMETHYL)PHENYL]GUANIDINE", "non-polymer"),
        "1JQ": (
            "(2R)-2-METHYL-3-(3-PHENYL-1,2-OXAZOL-5-YL)PROPANOIC ACID",
            "non-polymer",
        ),
        "1JS": (
            "N~1~-HYDROXY-N~5~-(3-HYDROXYPROPYL)-N~2~-[4-(PHENYLETHYNYL)BENZOYL]-L-GLUTAMAMIDE",
            "non-polymer",
        ),
        "SBO": ("TRIHYDROXYANTIMONITE(III)", "non-polymer"),
        "SX3": (
            "4-BROMO-2-{[(3R,5S)-3,5-DIMETHYLPIPERIDIN-1-YL]CARBONYL}ANILINE",
            "non-polymer",
        ),
        "MQ7": ("MENAQUINONE-7", "non-polymer"),
        "TMT": ("2,3,4-TRIMETHYL-1,3-THIAZOLE", "non-polymer"),
        "KFZ": (
            "2-{[(2,5-DICHLOROPHENYL)METHYL]SULFANYL}-5-ETHYL-[1,2,4]TRIAZOLO[1,5-A]PYRIMIDIN-7-OL",
            "non-polymer",
        ),
        "TY8": ("2,4-BIS(HYDROPEROXY)-5-HYDROXY-L-PHENYLALANINE", "peptide"),
        "SBL": ("L-NAPHTHYL-1-ACETAMIDO BORONIC ACID ALANINE", "peptide"),
        "MQ8": ("MENAQUINONE 8", "non-polymer"),
        "TY4": (
            "4-BROMO-2-({[(2-CHLOROPHENYL)CARBONYL]AMINO}METHYL)-6-METHYLPHENYL 2,4-DICHLOROBENZOATE",
            "non-polymer",
        ),
        "TY5": ("O-BENZYL-L-TYROSINE", "peptide"),
        "TY0": ("(3S)-3-AMINO-4-(4-TERT-BUTOXYPHENYL)-1-HYDROXYBUTAN-2-ONE", "peptide"),
        "DQN": ("DUROQUINONE", "non-polymer"),
        "TY2": ("3-AMINO-L-TYROSINE", "peptide"),
        "TY3": ("3-HYDROXY-L-TYROSINE", "peptide"),
        "I5R": (
            "(2R)-2-(1H-INDOL-3-YLMETHYL)-1,4-DIHYDROPYRIDO[2,3-B]PYRAZIN-3(2H)-ONE",
            "non-polymer",
        ),
        "UNB": (
            "N-(4-CARBAMIMIDOYL-BENZYL)-2-[2-HYDROXY-6-METHYL-3-(NAPHTHALENE-1-SULFONYLAMINO)-PHENYL]-ACETAMIDE",
            "non-polymer",
        ),
        "GTP": ("GUANOSINE-5'-TRIPHOSPHATE", "non-polymer"),
        "T5O": ("5-METHOXY-2'-DEOXYURIDINE-5'-PHOSPHATE", "nucleic-acid"),
        "UB1": (
            "3-({3-[(2S)-2-AMINO-2-CARBOXYETHYL]-5-METHYL-2,6-DIOXO-3,6-DIHYDROPYRIMIDIN-1(2H)-YL}METHYL)-4,5-DIBROMOTHIOPHENE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "KFN": (
            "(4S,5R,6R)-4,5-DIHYDROXY-6-[(1R,2R)-1,2,3-TRIHYDROXYPROPYL]-5,6-DIHYDRO-4H-PYRAN-2-CARBOXYLIC ACID",
            "saccharide",
        ),
        "Z96": (
            "TERT-BUTYL 4-{(11S)-8-CHLORO-6-[(R)-HYDROXY(1-METHYL-1H-IMIDAZOL-5-YL)METHYL]-11H-BENZO[5,6]CYCLOHEPTA[1,2-B]PYRIDIN-11-YL}PIPERAZINE-1-CARBOXYLATE",
            "non-polymer",
        ),
        "O90": (
            "{3-CHLORO-4-[(3-{[7-PROPYL-3-(TRIFLUOROMETHYL)-1,2-BENZISOXAZOL-6-YL]OXY}PROPYL)SULFANYL]PHENYL}ACETIC ACID",
            "non-polymer",
        ),
        "TYL": ("N-(4-HYDROXYPHENYL)ACETAMIDE (TYLENOL)", "non-polymer"),
        "TYM": ("TRYPTOPHANYL-5'AMP", "non-polymer"),
        "TYN": ("AMINOBENZOFURAZAN-O-TYROSINE", "peptide"),
        "TYO": (
            "(4Z,6E)-2-AMINO-7-HYDROPEROXY-4-[(E)-2-HYDROXYVINYL]HEPTA-4,6-DIENOIC ACID",
            "peptide",
        ),
        "TYI": ("3,5-DIIODOTYROSINE", "peptide"),
        "TYJ": ("2,5-DIHYDROXY-N-METHYL-L-TYROSINE", "peptide"),
        "TYK": ("TYLOSIN", "non-polymer"),
        "TYD": ("THYMIDINE-5'-DIPHOSPHATE", "non-polymer"),
        "TYE": ("4-[(2S)-2-AMINO-3-HYDROXYPROPYL]PHENOL", "non-polymer"),
        "UB2": (
            "N-[(S)-({[(BENZYLOXY)CARBONYL]AMINO}METHYL)(HYDROXY)PHOSPHORYL]-L-ALANYL-L-LEUCINE",
            "peptide",
        ),
        "VII": ("2,5-DIPHENYLFURAN-3,4-DICARBOXYLIC ACID", "non-polymer"),
        "VIN": (
            "12-(2-HYDROXYETHYL)-2-(1-METHYLETHOXY)-13,14-DIHYDRONAPHTHO[2,1-A]PYRROLO[3,4-C]CARBAZOL-5(12H)-ONE",
            "non-polymer",
        ),
        "TYA": ("?", "non-polymer"),
        "TYB": ("TYROSINAL", "peptide"),
        "TYC": ("L-TYROSINAMIDE", "non-polymer"),
        "VIR": ("VIRGINIAMYCIN M1", "non-polymer"),
        "Z97": (
            "(E)-N~5~-({3-HYDROXY-2-METHYL-5-[(PHOSPHONOOXY)METHYL]PYRIDIN-4-YL}METHYLIDENE)-L-ORNITHINE",
            "non-polymer",
        ),
        "VIQ": (
            "4-METHOXY-2,3,6-TRIMETHYL-N-(1,3,5-TRIMETHYL-1H-PYRAZOL-4-YL)BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "TYX": ("S-(2-ANILINO-2-OXOETHYL)-L-CYSTEINE", "peptide"),
        "TYY": ("3-(4-HYDROXY-3-IMINO-6-OXO-CYCLOHEXA-1,4-DIENYL)-ALANINE", "peptide"),
        "TYZ": ("PARA ACETAMIDO BENZOIC ACID", "non-polymer"),
        "TYT": ("TYROSINE DERIVATIVE", "peptide"),
        "T4G": (
            "6-(2,6-DIMETHYLPYRIDIN-4-YL)-5-PHENYL-1,2,4-TRIAZIN-3-AMINE",
            "non-polymer",
        ),
        "TYV": ("TYVELOSE", "saccharide"),
        "TYW": ("(2S)-2-AMINO-3-(4-HYDROXYPHENYL)PROPANE-1,1-DIOL", "peptide"),
        "TYP": ("CYCLO-(L-TYROSINE-L-PROLINE) INHIBITOR", "non-polymer"),
        "TYQ": ("3-AMINO-6-HYDROXY-TYROSINE", "peptide"),
        "TYR": ("TYROSINE", "peptide"),
        "TYS": ("O-SULFO-L-TYROSINE", "peptide"),
        "FUU": (
            "2-(3-((3-(5-(FURAN-2-YL)-1,3,4-OXADIAZOL-2-YL)UREIDO)METHYL)PHENOXY)ACETIC ACID",
            "non-polymer",
        ),
        "FUR": ("FURAZOLIDONE", "non-polymer"),
        "FUP": ("2',3'-DIDEOXY-3'-FLUORO-URIDIDINE-5'-DIPHOSPHATE", "non-polymer"),
        "FUZ": ("2-[(FURAN-2-YLMETHYL)AMINO]BENZOIC ACID", "non-polymer"),
        "NZQ": ("5,6-DIHYDROXY-NADP", "non-polymer"),
        "FUX": ("5-HYDROXYMETHYL-FURFURAL", "non-polymer"),
        "FUY": ("BETA-L-FUCOSYL-AZIDE", "non-polymer"),
        "FUF": ("2-DEOXY-2-FLUORO-BETA-L-FUCOPYRANOSE", "non-polymer"),
        "FUG": ("FUMAGILLIN", "non-polymer"),
        "FUD": ("D-FRUCTOSE", "saccharide"),
        "E59": (
            "4-{[(4-METHYL-6-OXO-1,6-DIHYDROPYRIMIDIN-2-YL)SULFANYL]ACETYL}BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "FUC": ("ALPHA-L-FUCOSE", "saccharide"),
        "FUA": ("FUSIDIC ACID", "non-polymer"),
        "E55": (
            "3-O-DECYL-2-DEOXY-6-O-{2-DEOXY-3-O-[(3R)-3-METHOXYDECYL]-6-O-METHYL-2-[(11Z)-OCTADEC-11-ENOYLAMINO]-4-O-PHOSPHONO-BETA-D-GLUCOPYRANOSYL}-2-[(3-OXOTETRADECANOYL)AMINO]-1-O-PHOSPHONO-ALPHA-D-GLUCOPYRANOSE",
            "non-polymer",
        ),
        "HGZ": ("6,7,8-TRIMETHOXY-4-METHYLQUINOLIN-2(1H)-ONE", "non-polymer"),
        "E57": (
            "N-[(1S)-1-(3-CHLOROPHENYL)-2-HYDROXYETHYL]-4-(2-{[(2S)-1-HYDROXYBUTAN-2-YL]AMINO}-5-METHYLPYRIMIDIN-4-YL)-1H-PYRROLE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "FUM": ("FUMARIC ACID", "non-polymer"),
        "E50": (
            "2-CHLORO-5-{[(5-ETHYLPYRIMIDIN-2-YL)SULFANYL]ACETYL}BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "E52": ("METHYLATED RUTHENIUM PYRIDOCARBAZOLE", "non-polymer"),
        "HGX": (
            "1-HEPTADECANOYL-2-TRIDECANOYL-3-GLYCEROL-PHOSPHONYL CHOLINE",
            "non-polymer",
        ),
        "QQ2": (
            "[(2-CHLORO-5-METHYLPHENYL){6-[(4-{[(2R)-3-(DIMETHYLAMINO)-2-HYDROXYPROPYL]OXY}PHENYL)AMINO]PYRIMIDIN-4-YL}AMINO]ACETONITRILE",
            "non-polymer",
        ),
        "NZT": (
            "N-(4-{[4-(PYRAZIN-2-YL)PIPERAZIN-1-YL]CARBONYL}PHENYL)QUINOLINE-8-SULFONAMIDE",
            "non-polymer",
        ),
        "NML": ("N-METHYLACETAMIDE", "non-polymer"),
        "6XP": (
            "(2S)-N'-[(3R)-4-CHLORO-7-METHYL-2-OXO-2,3-DIHYDRO-1H-INDOL-3-YL]-2-(4-HYDROXYPHENYL)PROPANEHYDRAZIDE",
            "non-polymer",
        ),
        "HGU": ("N-HYDROXYGUANIDINE", "non-polymer"),
        "Z92": (
            "(4E)-4-[(4-CHLOROPHENYL)HYDRAZONO]-5-(3,4-DIMETHOXYPHENYL)-2,4-DIHYDRO-3H-PYRAZOL-3-ONE",
            "non-polymer",
        ),
        "HGS": ("D-GAMMA-GLUTAMYL-L-CYSTEINYL-BETA-ALANINE", "non-polymer"),
        "HGP": (
            "(1-HEXADECANOYL-2-TETRADECANOYL-GLYCEROL-3-YL) PHOSPHONYL CHOLINE",
            "non-polymer",
        ),
        "OPG": ("OXIRANPSEUDOGLUCOSE", "non-polymer"),
        "6X1": ("2-{[(3,5-DICHLOROPHENYL)CARBAMOYL]AMINO}BENZOIC ACID", "non-polymer"),
        "FU7": (
            "2-AMINO-4-{2,4-DICHLORO-6-[2-(1H-PYRAZOL-1-YL)ETHOXY]PHENYL}-N-(2,2-DIFLUOROPROPYL)-5,7-DIHYDRO-6H-PYRROLO[3,4-D]PYRIMIDINE-6-CARBOXAMIDE",
            "non-polymer",
        ),
        "FU4": ("2,6-ANHYDRO-1-DEOXY-D-GALACTITOL", "saccharide"),
        "FU5": (
            "4-[2-CHLORO-6-(4,4,4-TRIFLUOROBUTOXY)PHENYL]-6-METHYLPYRIMIDIN-2-AMINE",
            "non-polymer",
        ),
        "FU2": ("FURFURAL", "non-polymer"),
        "FU3": (
            "2-AMINO-N-CYCLOBUTYL-4-[2,4-DICHLORO-6-(4,4,4-TRIFLUOROBUTOXY)PHENYL]-5,7-DIHYDRO-6H-PYRROLO[3,4-D]PYRIMIDINE-6-CARBOXAMIDE",
            "non-polymer",
        ),
        "FU1": ("TETRAHYDROFURAN", "non-polymer"),
        "E5G": ("5-HYDROXYPENTYL ALPHA-D-GLUCOPYRANOSIDE", "saccharide"),
        "HGL": (
            "(2S)-2-AMINO-4-(2-AMINO-6-OXO-1,6-DIHYDRO-9H-PURIN-9-YL)BUTANOIC ACID",
            "peptide",
        ),
        "FU9": (
            "3-(CYCLOPROPYLAMINO)-5-{[3-(TRIFLUOROMETHYL)PHENYL]AMINO}PYRIMIDO[4,5-C]QUINOLINE-8-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "ODP": ("4-OXO-NICOTINAMIDE-ADENINE DINUCLEOTIDE PHOSPHATE", "nucleic-acid"),
        "HGM": ("(2R)-2-(4-HYDROXYPHENYL)-2-(METHYLAMINO)ETHANOIC ACID", "peptide"),
        "U04": (
            "({3-[1-(4-HYDROXY-2-OXO-2H-CHROMEN-3-YL)-PROPYL]-PHENYLCARBAMOYL}-METHYL)-CARBAMIC ACID TERT-BUTYL ESTER",
            "non-polymer",
        ),
        "ZTD": (
            "METHYL 4-[({(4R)-4-HYDROXY-1-[(3-METHYLISOXAZOL-5-YL)ACETYL]-L-PROLYL}AMINO)METHYL]BENZOATE",
            "non-polymer",
        ),
        "NZF": (
            "(3S)-N-[3-(TRIFLUOROMETHOXY)BENZYL]-1-{2-[(3,4,5-TRIMETHOXYPHENYL)AMINO]PYRIMIDIN-4-YL}PIPERIDINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "458": (
            "4-[(1S,2S,5S)-5-(HYDROXYMETHYL)-8-METHYL-3-OXABICYCLO[3.3.1]NON-7-EN-2-YL]PHENOL",
            "non-polymer",
        ),
        "459": (
            "4-[(1S,2S,5S,9R)-5-(HYDROXYMETHYL)-8,9-DIMETHYL-3-OXABICYCLO[3.3.1]NON-7-EN-2-YL]PHENOL",
            "non-polymer",
        ),
        "JBT": ("GDP-N-ACETYLPEROSAMINE-COENZYME A", "non-polymer"),
        "450": (
            "{[4-R(-4-ALPHA,5-ALPHA,6-BETA,7-BETA)]-HEXAHYDRO-5,6-BIS(HYDROXY)-1,3-BIS([3-AMINO)PHENYL]METHYL-4,7-BIA(PHENYLMETHYL)-2H-1,3-DIAZEPINONE",
            "non-polymer",
        ),
        "C70": (
            "2-(4-CHLOROPHENYL)-4-[(3S)-PIPERIDIN-3-YLAMINO]THIENO[2,3-D]PYRIDAZINE-7-CARBOXAMIDE",
            "non-polymer",
        ),
        "452": (
            "2-{[N-(2-ACETYL-5-CHLORO-4-FLUOROPHENYL)GLYCYL]AMINO}BENZOIC ACID",
            "non-polymer",
        ),
        "C72": (
            "2-(4-CHLOROPHENYL)-4-[(3S)-PIPERIDIN-3-YLAMINO]THIENO[3,2-C]PYRIDINE-7-CARBOXAMIDE",
            "non-polymer",
        ),
        "C75": (
            "4-{5-[(Z)-(2-IMINO-4-OXO-1,3-THIAZOLIDIN-5-YLIDENE)METHYL]-2-FURYL}-N-METHYLBENZENESULFONAMIDE",
            "non-polymer",
        ),
        "455": (
            "(3R,4R)-1-SULFAMOYL-PYRROLIDINE-3,4-DICARBOXYLIC ACID 3-[(4-CHLORO-PHENYL)-AMIDE] 4-{[2-FLUORO-4-(2-OXO-2H-PYRIDIN-1-YL)-PHENYL]-AMIDE}",
            "non-polymer",
        ),
        "C77": (
            "5'-DEOXY-2',3'-O-(1-METHYLETHYLIDENE)-5'-(4-{5-[(3AS,4S,6AR)-2-OXOHEXAHYDRO-1H-THIENO[3,4-D]IMIDAZOL-4-YL]PENTYL}-1H-1,2,3-TRIAZOL-1-YL)ADENOSINE",
            "non-polymer",
        ),
        "OM3": (
            "(R)-(4-CARBOXY-1,3-BENZODIOXOL-5-YL)METHYL-[[2-(CYCLOHEXYLMETHYLCARBAMOYL)PHENYL]METHYL]-METHYL-AZANIUM",
            "non-polymer",
        ),
        "HGF": (
            "2-[5-(2-METHOXYETHOXY)PYRIDIN-3-YL]-8,9,10,11-TETRAHYDRO-7H-PYRIDO[3',4':4,5]PYRROLO[2,3-F]ISOQUINOLIN-7-ONE",
            "non-polymer",
        ),
        "OM1": (
            "(4-CARBOXY-1,3-BENZODIOXOL-5-YL)METHYL-[[2-[(4-METHOXYPHENYL)METHYLCARBAMOYL]PHENYL]METHYL]AZANIUM",
            "non-polymer",
        ),
        "HGG": ("3-HYDROXYL GLUTARYL COA", "non-polymer"),
        "IT5": ("4-CYANO-4-PHENYLPIPERIDINE-1-CARBODITHIOIC ACID", "non-polymer"),
        "IT2": ("MORPHOLINE-4-CARBODITHIOIC ACID", "non-polymer"),
        "JBI": (
            "9-CYCLOPENTYL-N~8~-(2-FLUOROPHENYL)-N~2~-(4-METHOXYPHENYL)-9H-PURINE-2,8-DIAMINE",
            "non-polymer",
        ),
        "IT1": (
            "(2S)-2-AMINO-6-[[3-HYDROXY-2-METHYL-5-(PHOSPHONOOXYMETHYL)PYRIDIN-4-YL]METHYLIDENEAMINO]HEXANOIC ACID",
            "peptide",
        ),
        "DGB": (
            "N-[4-(1-BENZOYLPIPERIDIN-4-YL)BUTYL]-3-PYRIDIN-3-YLPROPANAMIDE",
            "non-polymer",
        ),
        "DGC": ("4-DEOXY-ALPHA-L-THREO-HEX-4-ENOPYRANURONIC ACID", "saccharide"),
        "385": (
            "(3R,3AS,6AR)-HEXAHYDROFURO[2,3-B]FURAN-3-YL [(1S,2R)-3-[(1,3-BENZODIOXOL-5-YLSULFONYL)(ISOBUTYL)AMINO]-2-HYDROXY-1-{4-[(2-METHYL-1,3-THIAZOL-4-YL)METHOXY]BENZYL}PROPYL]CARBAMATE",
            "non-polymer",
        ),
        "DGA": ("DIACYL GLYCEROL", "non-polymer"),
        "UN4": (
            "5-{[(2-AMINO-9H-PURIN-6-YL)OXY]METHYL}-2-PYRROLIDINONE",
            "non-polymer",
        ),
        "DGG": (
            "1-[GLYCEROLYLPHOSPHONYL]-2-[8-(2-HEXYL-CYCLOPROPYL)-OCTANAL-1-YL]-3-[HEXADECANAL-1-YL]-GLYCEROL",
            "non-polymer",
        ),
        "DGD": ("DIGALACTOSYL DIACYL GLYCEROL (DGDG)", "saccharide"),
        "UN7": (
            "3-[3-(3(S)-METHYLCARBAMOYL-7-SULFOAMINO-3,4,-DIHYDRO-1H-ISOQUINOLIN-2-YL)-3-OXO-PROPYL]-BENZOIC ACID",
            "non-polymer",
        ),
        "DGJ": ("(2R,3S,4R,5S)-2-(HYDROXYMETHYL)PIPERIDINE-3,4,5-TRIOL", "non-polymer"),
        "NZO": ("NOCODAZOLE", "non-polymer"),
        "DGH": ("(2R)-AMINO(4-HYDROXYPHENYL)ACETIC ACID", "peptide"),
        "DGI": ("2'-DEOXYGUANOSINE-5'-DIPHOSPHATE", "nucleic-acid"),
        "DGN": ("D-GLUTAMINE", "peptide"),
        "DGL": ("D-GLUTAMIC ACID", "peptide"),
        "DGM": ("D-GLUCOPYRANOSYLIUM", "saccharide"),
        "DGS": ("3,6-ANHYDRO-D-GALACTOSE-2-SULFATE", "saccharide"),
        "DGP": ("2'-DEOXYGUANOSINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "ODK": ("3-PYRIDIN-2-YL-1H-PYRAZOLE-5-CARBOXYLIC ACID", "non-polymer"),
        "DGT": ("2'-DEOXYGUANOSINE-5'-TRIPHOSPHATE", "non-polymer"),
        "DGU": ("D-GALACTURONIC ACID", "non-polymer"),
        "DGZ": (
            "N-((2R,3S,6S,18S,21S)-2-AMINO-18-(4-BENZOYLBENZYL)-21-CARBAMOYL-3-HYDROXY-6-(NAPHTHALEN-2-YLMETHYL)-4,7,16,19-TETRAOXO-1-PHENYL-11,14-DIOXA-5,8,17,20-TETRAAZAPENTACOSAN-25-YL)HEX-5-YNAMIDE",
            "non-polymer",
        ),
        "DGX": ("DIGOXIN", "non-polymer"),
        "DGY": ("(2R)-2,3-DIHYDROXYPROPANOIC ACID", "non-polymer"),
        "OST": (
            "METHYL (2Z)-3-METHOXY-2-{2-[(E)-2-PHENYLVINYL]PHENYL}ACRYLATE",
            "non-polymer",
        ),
        "UNR": ("3-(2-HYDROXY-2,2-DIPHOSPHONOETHYL)-1-METHYLPYRIDINIUM", "non-polymer"),
        "UNT": (
            "(2S)-3-{4-[3-(5-METHYL-2-PHENYL-1,3-OXAZOL-4-YL)PROPYL]PHENYL}-2-(2H-1,2,3-TRIAZOL-2-YL)PROPANOIC ACID",
            "non-polymer",
        ),
        "UNU": ("BENZAMIDE", "non-polymer"),
        "UNV": (
            "6,7-DIHYDRO-5H-CYCLOPENTA[B]PYRIDINE-6,6-DIYLBIS(PHOSPHONIC ACID)",
            "non-polymer",
        ),
        "UNX": ("UNKNOWN ATOM OR ION", "non-polymer"),
        "DG2": ("2,3-DIPHOSPHOGLYCERIC ACID", "non-polymer"),
        "DG3": ("2'-3'-DIDEOXYGUANOSINE-5'-TRIPHOSPHATE", "non-polymer"),
        "DG0": (
            "4-O-ALPHA-D-GALACTOPYRANURONOSYL-ALPHA-D-GALACTOPYRANURONIC ACID",
            "saccharide",
        ),
        "DG1": (
            "(4S)-ISONICOTINIC-ACETYL-NICOTINAMIDE-ADENINE DINUCLEOTIDE",
            "non-polymer",
        ),
        "DG6": ("2-DEOXY-GLUCITOL-6-PHOSPHATE", "non-polymer"),
        "DG7": (
            "1-[3-TERT-BUTYL-1-(4-METHYLPHENYL)-1H-PYRAZOL-5-YL]-3-{4-[2-(PYRIDIN-3-YLMETHOXY)ETHYL]-1,3-THIAZOL-2-YL}UREA",
            "non-polymer",
        ),
        "UNF": (
            "4-O-(4-DEOXY-BETA-L-THREO-HEX-4-ENOPYRANURONOSYL)-BETA-D-GALACTOPYRANURONIC ACID",
            "non-polymer",
        ),
        "ODA": (
            "9-AMINO-6,10-DIOXO-OCTAHYDRO-PYRIDAZINO[1,2-A][1,2]DIAZEPINE-1-CARBOXYLIC ACID",
            "peptide",
        ),
        "UNH": (
            "({1-[1-CARBAMOYL-PHENYL-METHYL)-CARBAMOYL]-METHYL}-AMINOOXALYL)-BUTYLCARBAMOYL)-3-METHYL-BUTYLCARBAMOYL)-CYCLOHEXYL-METHYL)-CARBAMIC ACID ISOBUTYL ESTER",
            "non-polymer",
        ),
        "UNI": (
            "4-CYANO-N-(3-CYCLOPROPYL(5,6,7,8,9,10-HEXAHYDRO-4-HYDROXY-2-OXO-CYCLOOCTA[B]PYRAN-3-YL)METHYL)PHENYL BENZENSULFONAMIDE",
            "non-polymer",
        ),
        "DG8": ("?", "nucleic-acid"),
        "DG9": (
            "5'-({[(2R,3R)-3-AMINO-2-{[2-({N-[(2R)-2-HYDROXY-3,3-DIMETHYL-4-{[OXIDO(OXO)PHOSPHONIO]OXY}BUTANOYL]-BETA-ALANYL}AMINO)ETHYL]SULFANYL}-4-METHYLPENTYL]SULFONYL}AMINO)-5'-DEOXYADENOSINE",
            "non-polymer",
        ),
        "UNL": ("UNKNOWN LIGAND", "non-polymer"),
        "UNM": (
            "7-[(CIS-4-AMINOCYCLOHEXYL)AMINO]-5-BROMO-1-BENZOFURAN-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "UNN": (
            "(6,7-DIFLUORO-QUINAZOLIN-4-YL)-(1-METHYL-2,2-DIPHENYL-ETHYL)-AMINE",
            "non-polymer",
        ),
        "OMC": ("O2'-METHYLYCYTIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "OMB": ("OXYMETHYLENE BRIDGING GROUP", "non-polymer"),
        "ITL": (
            "(3ALPHA,5ALPHA,8ALPHA)-3-[(N,N-BIS{2-[BIS(CARBOXYMETHYL)AMINO]ETHYL}-L-GAMMA-GLUTAMYL)AMINO]CHOLAN-24-OIC ACID",
            "non-polymer",
        ),
        "ITM": ("ISOCITRIC ACID-MAGNESIUM COMPLEX", "non-polymer"),
        "JB3": ("GDP-N-ACETYLPEROSAMINE", "non-polymer"),
        "JB2": ("GDP-PEROSAMINE", "non-polymer"),
        "JB1": (
            "(1R,3S,5Z)-5-[(2E)-2-[(1R,3AS,7AR)-1-[(2R,3S)-3-(2-HYDROXYETHYL)HEPTAN-2-YL]-7A-METHYL-2,3,3A,5,6,7-HEXAHYDRO-1H-INDEN-4-YLIDENE]ETHYLIDENE]-4-METHYLIDENE-CYCLOHEXANE-1,3-DIOL",
            "non-polymer",
        ),
        "ITI": (
            "N-[3-(5-{2-[(4-MORPHOLIN-4-YLPHENYL)AMINO]PYRIMIDIN-4-YL}IMIDAZO[2,1-B][1,3]THIAZOL-6-YL)PHENYL]-2-PHENYLACETAMIDE",
            "non-polymer",
        ),
        "45P": (
            "3-(5-AMINO-7-HYDROXY-[1,2,3]TRIAZOLO[4,5-D]PYRIMIDIN-2-YL)-N-(3,5-DICHLOROBENZYL)-BENZAMIDE",
            "non-polymer",
        ),
        "C7P": ("CIS-DIAMMINE(PYRIDINE)CHLOROPLATINUM(II)", "non-polymer"),
        "ITD": ("?", "non-polymer"),
        "ITE": (
            "2-(METHYLAMINO)-1,7-DIHYDRO-8H-IMIDAZO[4,5-G]QUINAZOLIN-8-ONE",
            "non-polymer",
        ),
        "ITB": (
            "2-(CARBOXYMETHYL)-1-OXO-1,2-DIHYDRONAPHTHO[1,2-D]ISOTHIAZOLE-4-CARBOXYLIC ACID 3,3-DIOXIDE",
            "non-polymer",
        ),
        "45U": (
            "(S)-N-(4-CARBAMIMIDOYLBENZYL)-1-(2-(CYCLOPENTYLOXY)ETHANOYL)PYRROLIDINE-2-CARBOXAMIDE",
            "peptide",
        ),
        "OMM": (
            "4-({(2S)-2-[2-(4-CHLOROPHENYL)-5,6-DIFLUORO-1H-BENZIMIDAZOL-1-YL]-2-CYCLOHEXYLACETYL}AMINO)-3-FLUOROBENZOIC ACID",
            "non-polymer",
        ),
        "ITA": (
            "{4-[(CARBOXYMETHOXY)CARBONYL]-3,3-DIOXIDO-1-OXONAPHTHO[1,2-D]ISOTHIAZOL-2(1H)-YL}ACETIC ACID",
            "non-polymer",
        ),
        "OMP": ("OROTIDINE-5'-MONOPHOSPHATE", "non-polymer"),
        "C7M": (
            "(2S,3R)-N-[5-CHLORO-2-(2,3-DIHYDRO-1H-TETRAZOL-1-YL)BENZYL]-3-HYDROXY-4-{[(4-METHOXYPHENYL)SULFONYL]AMINO}-1-PHENYLBUTAN-2-AMINIUM",
            "non-polymer",
        ),
        "OMU": ("?", "nucleic-acid"),
        "OMT": ("S-DIOXYMETHIONINE", "peptide"),
        "OMZ": ("(BETAR)-3-CHLORO-BETA-HYDROXY-D-TYROSINE", "peptide"),
        "C7C": (
            "5-[(4-CHLOROPHENYL)SULFANYL]-1,2,3-THIADIAZOLE-4-CARBOXYLATE",
            "non-polymer",
        ),
        "45C": ("4,5-DICHLOROBENZENE-1,2-DIOL", "non-polymer"),
        "ITR": ("IMINO-TRYPTOPHAN", "non-polymer"),
        "ITS": ("INOSITOL 1,3,4,5-TETRAKISPHOSPHATE", "non-polymer"),
        "ITP": (
            "PHOSPHORIC ACID MONO-(2,3,4,6-TETRAHYDROXY-5-PHOSPHONOOXY-CYCLOHEXYL) ESTER",
            "non-polymer",
        ),
        "ITQ": ("1,2,3,4-TETRAHYDROGEN STAUROSPORINE", "non-polymer"),
        "AAE": ("ACETOACETIC ACID", "non-polymer"),
        "AAD": (
            "(2-AMINOOXY-ETHYL)-[5-(6-AMINO-PURIN-9-YL)-3,4-DIHYDROXY-TETRAHYDRO-FURAN-2-YLMETHYL]-METHYL-SULFONIUM",
            "non-polymer",
        ),
        "AAG": ("N-ALPHA-L-ACETYL-ARGININE", "non-polymer"),
        "AAF": (
            "METHYL 4-{[({[(2R,5S)-5-{[(2S)-2-(AMINOMETHYL)PYRROLIDIN-1-YL]CARBONYL}PYRROLIDIN-2-YL]METHYL}AMINO)CARBONYL]AMINO}BENZOATE",
            "non-polymer",
        ),
        "AAA": (
            "(2-ACETYL-5-METHYLANILINO)(2,6-DIBROMOPHENYL)ACETAMIDE",
            "non-polymer",
        ),
        "AAC": ("ACETYLAMINO-ACETIC ACID", "non-polymer"),
        "AAB": ("2-DEOXY-5-PHOSPHORIBOSE GROUP", "saccharide"),
        "AAM": ("ALPHA-ADENOSINE MONOPHOSPHATE", "non-polymer"),
        "AAL": ("3,6-ANHYDRO-L-GALACTOSE", "saccharide"),
        "AAO": ("ACARBOSE DERIVED HEXASACCHARIDE", "saccharide"),
        "AAN": ("2-(4-NITROPHENYL)ACETIC ACID", "non-polymer"),
        "AAI": (
            "TRANS-4-(7-CARBAMOYL-1H-BENZIMIDAZOL-2-YL)-1-PROPYLPIPERIDINIUM",
            "non-polymer",
        ),
        "AAH": ("?", "non-polymer"),
        "GF4": (
            "4-ETHYL-5-METHYL-2-(1H-TETRAZOL-5-YL)-1,2-DIHYDRO-3H-PYRAZOL-3-ONE",
            "non-polymer",
        ),
        "AAU": (
            "(5R,6R)-5-BENZYL-6-HYDROXY-2,4-BIS(4-HYDROXY-3-METHOXYBENZYL)-1-[3-(4-HYDROXYPHENYL)PROPANOYL]-1,2,4-TRIAZEPAN-3-ONE",
            "non-polymer",
        ),
        "AAT": ("S-ADENOSYL-1,8-DIAMINO-3-THIOOCTANE", "non-polymer"),
        "U05": (
            "6,11-DIHYDRO-11-ETHYL-6-METHYL-9-NITRO-5H-PYRIDO[2,3-B][1,5]BENZODIAZEPIN-5-ONE",
            "non-polymer",
        ),
        "AAV": (
            "N-ISOXAZOL-3-YL-4-METHYL-3-[6-(4-METHYLPIPERAZIN-1-YL)-4-OXO-QUINAZOLIN-3-YL]BENZAMIDE",
            "non-polymer",
        ),
        "AAQ": (
            "6-AMINO-4-[2-(BENZYLAMINO)ETHYL]-2-(METHYLAMINO)-1,7-DIHYDRO-8H-IMIDAZO[4,5-G]QUINAZOLIN-8-ONE",
            "non-polymer",
        ),
        "AAP": (
            "ALPHA-(2,6-DICHLOROPHENYL)-ALPHA-(2-ACETYL-5-METHYLANILINO)ACETAMIDE",
            "non-polymer",
        ),
        "AAS": ("3-ACTOXYMERCURI-4-AMINOBENZENESULFONAMIDE", "non-polymer"),
        "AAR": ("ARGININEAMIDE", "peptide"),
        "MC5": (
            "(5S)-5-({6-[(2-FLUOROBENZYL)OXY]NAPHTHALEN-2-YL}METHYL)-1,3-THIAZOLIDINE-2,4-DIONE",
            "non-polymer",
        ),
        "BHF": ("2-PHENYL-4H-BENZO[H]CHROMEN-4-ONE", "non-polymer"),
        "AAY": (
            "8-[2-((2S)-4-HYDROXY-1-{[5-(HYDROXYMETHYL)-6-METHOXY-2-NAPHTHYL]METHYL}-6-OXOPIPERIDIN-2-YL)ETHYL]-3,7-DIMETHYL-1,2,3,7,8,8A-HEXAHYDRONAPHTHALEN-1-YL 2-METHYLBUTANOATE",
            "non-polymer",
        ),
        "AAX": (
            "N-(CYCLOPROPYLMETHYL)-4-(METHYLOXY)-3-({5-[3-(3-PYRIDINYL)PHENYL]-1,3-OXAZOL-2-YL}AMINO)BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "MC1": ("METHICILLIN ACYL-SERINE", "peptide"),
        "AAZ": (
            "N-[5-(ETHYLSULFONYL)-2-METHOXYPHENYL]-5-[3-(2-PYRIDINYL)PHENYL]-1,3-OXAZOL-2-AMINE",
            "non-polymer",
        ),
        "BIN": (
            "2,3-DICARBOXY-4-(2-CHLORO-PHENYL)-1-ETHYL-5-ISOPROPOXYCARBONYL-6-METHYL-PYRIDINIUM",
            "non-polymer",
        ),
        "BIO": ("BIOPTERIN", "non-polymer"),
        "BIL": ("(3R,4S)-3-AMINO-4-METHYLHEXANOIC ACID", "peptide"),
        "BIM": ("BIS-(INDOLE)MALEIMIDE PYRIDINOPHANE", "non-polymer"),
        "BIJ": (
            "7-{[(2Z,5S)-5-CHLORO-2-{[2-METHOXY-4-(4-METHYLPIPERAZIN-1-YL)PHENYL]IMINO}-2,5-DIHYDROPYRIMIDIN-4-YL]AMINO}-2-METHYL-2,3-DIHYDRO-1H-ISOINDOL-1-ONE",
            "non-polymer",
        ),
        "BIK": ("3,7-DIHYDROXY-2-NAPHTHOIC ACID", "non-polymer"),
        "BIH": ("NAPHTHALENE-2,6-DISULFONIC ACID", "non-polymer"),
        "BII": (
            "2-{[5-CHLORO-2-({(1E,4R)-2-METHOXY-4-[(3R)-3-(METHYLAMINO)PYRROLIDIN-1-YL]CYCLOHEXA-2,5-DIEN-1-YLIDENE}AMINO)PYRIMIDIN-4-YL]AMINO}-N-(1-METHYLETHYL)BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "BIF": ("(R)-2-AMINO-3-(4-PHENYLCYCLOHEXYL)PROPANOIC ACID", "peptide"),
        "BIG": (
            "(3R,4S)-1-[(4-AMINO-5H-PYRROLO[3,2-D]PYRIMIDIN-7-YL)METHYL]-4-[(BUTYLSULFANYL)METHYL]PYRROLIDIN-3-OL",
            "non-polymer",
        ),
        "BID": ("BISTRAMIDE A", "non-polymer"),
        "BIE": ("(3,4-DIHYDROXY-2-NITROPHENYL)(PHENYL)METHANONE", "non-polymer"),
        "BIB": ("BETA-AMINO ISOBUTYRATE", "non-polymer"),
        "BIC": (
            "3-AMINO-3-BENZYL-9-CARBOXAMIDE[4.3.0]BICYCLO-1,6-DIAZANONAN-2-ONE",
            "non-polymer",
        ),
        "SL2": ("?", "non-polymer"),
        "BIA": (
            "1-(3,4,DIHYDROXY-5-NITROPHENYL)-3-{4-[3-(TRIFLUOROMETHYL) PHENYL] PIPERAZIN-1-YL}PROPAN-1-ONE",
            "non-polymer",
        ),
        "BIZ": ("BIZELESIN", "non-polymer"),
        "BIX": (
            "(2S)-2-{[(S)-[(3S)-3-AMINO-3-CARBOXYPROPYL](HYDROXY)PHOSPHORYL]METHYL}PENTANEDIOIC ACID",
            "non-polymer",
        ),
        "BIY": (
            "(3S)-2,2-DIMETHYL-3,4-DIHYDRO-2H-1,4-THIAZINE-3,6-DICARBOXYLIC ACID",
            "non-polymer",
        ),
        "BIV": (
            "21-NOR-9,10-SECOCHOLESTA-5,7,10(19)-TRIENE-1,3,25-TRIOL, 20-(4-HYDROXY-4-METHYLPENTYL)-, (1A,3B,5Z,7E)",
            "non-polymer",
        ),
        "BIW": (
            "1-[2-(4-CARBOXYPIPERIDIN-1-YL)-2-OXOETHYL]-3-CYCLOHEXYL-2-FURAN-3-YL-1H-INDOLE-6-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "BIT": (
            "(-)-1-PHENYL-1,2,3,4-TETRAHYDRO-4-HYDROXYPYRROLO[2,3-B]-7-METHYLQUINOLIN-4-ONE",
            "non-polymer",
        ),
        "BIU": ("5-BROMO-L-ISOLEUCINE", "peptide"),
        "BIR": (
            "N-[3-[(1-AMINOETHYL)(HYDROXY)PHOSPHORYL]-2-(1,1'-BIPHENYL-4-YLMETHYL)PROPANOYL]ALANINE",
            "non-polymer",
        ),
        "BIS": (
            "1,1,5,5-TETRAFLUOROPHOSPHOPENTYLPHOSPHONIC ACID ADENYLATE ESTER",
            "non-polymer",
        ),
        "BIP": ("2-BENZYL-3-IODOPROPANOIC ACID", "non-polymer"),
        "BIQ": (
            "BENZYL [12-(2-AMINO-2-OXOETHYL)-4-NITRO-10,13-DIOXO-15-[(PROPYLAMINO)CARBONYL]-2-OXA-11,14-DIAZATRICYCLO[15 .2.2.1~3,7~]DOCOSA-1(19),3(22),4,6,17,20-HEXAEN-9-YL]CARBAMATE",
            "non-polymer",
        ),
        "SLX": (
            "(13AS)-3,10-DIMETHOXY-5,8,13,13A-TETRAHYDRO-6H-ISOQUINO[3,2-A]ISOQUINOLINE-2,9-DIOL",
            "non-polymer",
        ),
        "SLZ": ("L-THIALYSINE", "peptide"),
        "K22": ("4,5,6,7-TETRABROMO-1H,3H-BENZIMIDAZOL-2-THIONE", "non-polymer"),
        "K23": (
            "6-CHLORO-3-[1-(4-CHLOROBENZYL)-4-PHENYL-1H-IMIDAZOL-5-YL]-1H-INDOLE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "K20": (
            "(3R,3AS,6AR)-HEXAHYDROFURO[2,3-B]FURAN-3-YL [(1S,2R)-3-{(1,3-BENZODIOXOL-5-YLSULFONYL)[(2S)-2-METHYLBUTYL]AMINO}-1-BENZYL-2-HYDROXYPROPYL]CARBAMATE",
            "non-polymer",
        ),
        "K21": (
            "4-[3-{1-(4-BENZYL)PIPERODINYL}PROPIONYL]-7-METHOXY-2,3,4,5-TERTRAHYDRO-1,4-BENZOTHIAZEPINE",
            "non-polymer",
        ),
        "K26": (
            "N-ACETYL-L-ILE-L-TYR-(R)-1-AMINO-2-(4-HYDROXYPHENYL)ETHYLPHOSPHONIC ACID",
            "non-polymer",
        ),
        "K27": (
            "4-CARBAMOYL-1-(3-{4-[(E)-(HYDROXYIMINO)METHYL]PYRIDINIUM-1-YL}PROPYL)PYRIDINIUM",
            "non-polymer",
        ),
        "SLR": (
            "(3R,4R)-3-HYDROXY-2-[(1S)-1-HYDROXY-2-METHYLPROPYL]-4-METHYL-5-OXO-D-PROLINE",
            "peptide",
        ),
        "K25": (
            "4,5,6,7-TETRABROMO-N,N-DIMETHYL-1H-BENZIMIDAZOL-2-AMINE",
            "non-polymer",
        ),
        "SLL": (
            "(2S)-2-AZANYL-6-[(4-HYDROXY-4-OXO-BUTANOYL)AMINO]HEXANOIC ACID",
            "peptide",
        ),
        "SLM": ("SIALYLAMIDE", "non-polymer"),
        "SLN": (
            "1-(4-AZIDO-2-NITROPHENYL)AMINO-3-(1-OXYL-2,2,5,5-TETRAMETHYLPYRROLIDINYL-3-CARBAMIDO)-2-PROPYLDISPHOSPHATE",
            "non-polymer",
        ),
        "SLO": (
            "2-[(3Z)-6-FLUORANYL-2-METHYL-3-[(4-METHYLSULFONYLPHENYL)METHYLIDENE]INDEN-1-YL]ETHANOIC ACID",
            "non-polymer",
        ),
        "ELD": ("(9E)-OCTADEC-9-ENAMIDE", "non-polymer"),
        "BI8": (
            "3-[1-(3-AMINOPROPYL)-1H-INDOL-3-YL]-4-(1-METHYL-1H-INDOL-3-YL)-1H-PYRROLE-2,5-DIONE",
            "non-polymer",
        ),
        "BI9": (
            "2-({5-CHLORO-2-[(2-METHOXY-4-MORPHOLIN-4-YLPHENYL)AMINO]PYRIMIDIN-4-YL}AMINO)-N-METHYLBENZAMIDE",
            "non-polymer",
        ),
        "BI6": (
            "(3S,6S,7Z,10AS)-N-(DIPHENYLMETHYL)-6-{[(2S)-2-(METHYLIDENEAMINO)BUTANOYL]AMINO}-5-OXO-1,2,3,5,6,9,10,10A-OCTAHYDROPYRROLO[1,2-A]AZOCINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "BI7": (
            "3-[(3AS,4R,8AS,8BR)-4-[5-(5-CHLORO-2-THIENYL)ISOXAZOL-3-YL]-1,3-DIOXOOCTAHYDROPYRROLO[3,4-A]PYRROLIZIN-2(3H)-YL]-N,N,N-TRIMETHYLPROPAN-1-AMINIUM",
            "non-polymer",
        ),
        "BI4": (
            "3-(1H-INDOL-3-YL)-4-{1-[2-(1-METHYLPYRROLIDIN-2-YL)ETHYL]-1H-INDOL-3-YL}-1H-PYRROLE-2,5-DIONE",
            "non-polymer",
        ),
        "BI5": ("4-PHENOXY-N-(PYRIDIN-2-YLMETHYL)BENZAMIDE", "non-polymer"),
        "BI2": (
            "3-(1H-INDOL-3-YL)-4-(1-{2-[(2S)-1-METHYLPYRROLIDINYL]ETHYL}-1H-INDOL-3-YL)-1H-PYRROLE-2,5-DIONE",
            "non-polymer",
        ),
        "BI3": (
            "3-[1-(3-AMINOPROPYL)-1H-INDOL-3-YL]-4-(1H-INDOL-3-YL)-1H-PYRROLE-2,5-DIONE",
            "non-polymer",
        ),
        "SLB": ("5-N-ACETYL-BETA-D-NEURAMINIC ACID", "non-polymer"),
        "BI1": (
            "3-{1-[3-(DIMETHYLAMINO)PROPYL]-1H-INDOL-3-YL}-4-(1H-INDOL-3-YL)-1H-PYRROLE-2,5-DIONE",
            "non-polymer",
        ),
        "MCO": (
            "1-(3-MERCAPTO-2-METHYL-PROPIONYL)-PYRROLIDINE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "MCN": ("PTERIN CYTOSINE DINUCLEOTIDE", "non-polymer"),
        "MCM": ("7-AMINO-4-METHYL-CHROMEN-2-ONE", "non-polymer"),
        "MCL": ("NZ-(1-CARBOXYETHYL)-LYSINE", "peptide"),
        "MCK": (
            "N-[(2,3-DIHYDROXYPHENYL)CARBONYL]-O-[(2S)-2-{[(2,3-DIHYDROXYPHENYL)CARBONYL]AMINO}-3-(L-SERYLOXY)PROPANOYL]-L-SERINE",
            "non-polymer",
        ),
        "MCJ": (
            "N-[(2,3-DIHYDROXYPHENYL)CARBONYL]-O-[(2S)-2-{[(2,3-DIHYDROXYPHENYL)CARBONYL]AMINO}-3-({N-[(2,3-DIHYDROXYPHENYL)CARBONYL]-L-SERYL}OXY)PROPANOYL]-D-SERINE",
            "non-polymer",
        ),
        "MCI": (
            "(2-MERCAPTOMETHYL-4-PHENYL-BUTYRYLIMINO)-(5-TETRAZOL-1-YLMETHYL-THIOPHEN-2-YL)-ACETIC ACID",
            "non-polymer",
        ),
        "MCH": ("TRICHLOROMETHANE", "non-polymer"),
        "MCG": ("(S)-(ALPHA)-METHYL-4-CARBOXYPHENYLGLYCINE", "peptide"),
        "MCF": (
            "(8R)-3-(5-S-METHYL-5-THIO-BETA-D-RIBOFURANOSYL)-3,6,7,8-TETRAHYDROIMIDAZO[4,5-D][1,3]DIAZEPIN-8-OL",
            "non-polymer",
        ),
        "GFP": ("2-DEOXY-2-FLUORO-ALPHA-D-GLUCOSE-1-PHOSPHATE", "saccharide"),
        "MCD": ("METHYLMALONYL(CARBADETHIA)-COENZYME A", "non-polymer"),
        "MCB": ("METHYLENE CARBONYL GROUP", "non-polymer"),
        "GFT": (
            "(2S)-2-AZANYL-3-[CYCLOHEXYLOXY(METHYL)PHOSPHORYL]OXY-PROPANOIC ACID",
            "peptide",
        ),
        "AA5": (
            "N-[(3-HYDROXY-2-METHYL-5-{[(TRIHYDROXYPHOSPHORANYL)OXY]METHYL}PYRIDIN-4-YL)METHYLENE]METHIONINE",
            "non-polymer",
        ),
        "AA4": ("2-AMINO-5-HYDROXYPENTANOIC ACID", "peptide"),
        "AA7": (
            "N,N'-DI-1,2,3,4-TETRAHYDROACRIDIN-9-YLHEPTANE-1,7-DIAMINE",
            "non-polymer",
        ),
        "AA6": (
            "3-[4,5-DIHYDROXY-2-2-(2-HYDROXYETHYLAMINO)-PHENYL]-2-METHYLAMINO-PROPIONIC ACID",
            "peptide",
        ),
        "AA1": ("[1-(4-ETHOXY-4-OXOBUTANOYL)HYDRAZINO]ACETIC ACID", "non-polymer"),
        "AA0": (
            "N-[4-METHYL-3-[6-(4-METHYLPIPERAZIN-1-YL)-4-OXIDANYLIDENE-QUINAZOLIN-3-YL]PHENYL]-2-MORPHOLIN-4-YL-PYRIDINE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "AA3": ("2-AMINOBUTYRIC ACID", "peptide"),
        "AA2": ("4-ARYL-2-PHENYLAMINO PYRIMIDINE", "non-polymer"),
        "GFB": ("GUANOSINE-5'-DIPHOSPHATE-BETA-L-FUCOPYRANOSE", "non-polymer"),
        "GFC": (
            "5'-O-[(S)-{[(S)-[(S)-CHLORO(FLUORO)PHOSPHONOMETHYL](HYDROXY)PHOSPHORYL]OXY}(HYDROXY)PHOSPHORYL]-2'-DEOXYGUANOSINE",
            "non-polymer",
        ),
        "MCU": ("METHYL 4-O-METHYL-BETA-D-GLUCOPYRANURONATE", "saccharide"),
        "GFA": (
            "2-[4-CHLORO-2-(PHENYLCARBONYL)PHENOXY]-N-PHENYLACETAMIDE",
            "non-polymer",
        ),
        "AA9": (
            "N-((1S,2R)-3-(((1S)-2-(CYCLOHEXYLAMINO)-1--METHYL-2-OXOETHYL)AMINO)-2-HYDROXY-1-(PHENYLMETHYL)PROPYL)-3-(ETHYLAMINO)-5-((METHYLSULFONYL)(PHENYL)AMINO)BENZAMIDE",
            "non-polymer",
        ),
        "AA8": ("ANGUINOMYCIN A, BOUND FORM", "non-polymer"),
        "MCP": ("2-CARBOXY-4-METHYLPIPERIDINE", "non-polymer"),
        "XBB": ("(3S)-2,3,4,5-TETRAHYDROPYRIDIN-3-AMINE", "non-polymer"),
        "KI9": ("2,3,4,6,8-PENTACHLORO-9H-CARBAZOL-1-OL", "non-polymer"),
        "EXB": ("3-(1H-BENZIMIDAZOL-1-YL)PROPANOIC ACID", "non-polymer"),
        "EXC": ("NITROXIDE SPIN-LABELED NUCLEOTIDE C-SPIN", "nucleic-acid"),
        "LQZ": ("2-(DIETHYLAMINO)-N-(2,6-DIMETHYLPHENYL)ETHANAMIDE", "non-polymer"),
        "NFV": ("NI-FE OXIDIZED ACTIVE CENTER", "non-polymer"),
        "EXO": ("(1R,2S,3S,4R)-5-METHYLENECYCLOHEXANE-1,2,3,4-TETRAOL", "non-polymer"),
        "KI1": (
            "1-{4-[4-CHLORO-3-(TRIFLUOROMETHYL)PHENYL]-4-HYDROXYPIPERIDIN-1-YL}-3-SULFANYLPROPAN-1-ONE",
            "non-polymer",
        ),
        "EXM": (
            "(8ALPHA,10ALPHA,13ALPHA)-6-METHYLIDENEANDROSTA-1,4-DIENE-3,17-DIONE",
            "non-polymer",
        ),
        "LQQ": (
            "6-ACETYL-8-CYCLOPENTYL-5-METHYL-2-[(5-PIPERAZIN-1-YLPYRIDIN-2-YL)AMINO]PYRIDO[2,3-D]PYRIMIDIN-7(8H)-ONE",
            "non-polymer",
        ),
        "QJ9": (
            "1-{(3S)-1-[4-(1-BENZOFURAN-2-YL)PYRIMIDIN-2-YL]PIPERIDIN-3-YL}-3-ETHYL-1,3-DIHYDRO-2H-BENZIMIDAZOL-2-ONE",
            "non-polymer",
        ),
        "EXT": (
            "{5-HYDROXY-6-METHYL-4-[(E)-{[2-(PHOSPHONOOXY)ETHYL]IMINO}METHYL]PYRIDIN-3-YL}METHYL DIHYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "Z99": (
            "2-[(1S,2S)-2-CARBOXYCYCLOPROPYL]-3-(9H-XANTHEN-9-YL)-D-ALANINE",
            "non-polymer",
        ),
        "EXY": ("6-[(2R)-OXIRAN-2-YL]-L-NORLEUCINE", "peptide"),
        "RDI": ("RADICICOL", "non-polymer"),
        "NU9": (
            "N-{[4-(4-PHENYL-1,3-THIAZOL-2-YL)TETRAHYDRO-2H-PYRAN-4-YL]METHYL}-3-[5-(TRIFLUOROMETHYL)-1,2,4-OXADIAZOL-3-YL]BENZAMIDE",
            "non-polymer",
        ),
        "B0R": (
            "2,6-DIFLUORO-N-(3-METHOXY-2H-PYRAZOLO[3,4-B]PYRIDIN-5-YL)-3-[(PROPYLSULFONYL)AMINO]BENZAMIDE",
            "non-polymer",
        ),
        "BAO": ("BIS(5-AMIDINO-2-BENZIMIDAZOLYL)METHANONE", "non-polymer"),
        "FUN": (
            "5-(AMINOSULFONYL)-4-CHLORO-2-[(2-FURYLMETHYL)AMINO]BENZOIC ACID",
            "non-polymer",
        ),
        "NTN": ("ISONICOTINAMIDINE", "non-polymer"),
        "LEC": (
            "N-[(2S,3R,4R,5S,6R)-4,5-DIHYDROXY-6-(HYDROXYMETHYL)-2-(4-NITROPHENOXY)OXAN-3-YL]ETHANAMIDE",
            "non-polymer",
        ),
        "LED": ("(4R)-5-OXO-L-LEUCINE", "peptide"),
        "LEE": ("L-GAMMA-GLUTAMYL-S-HEXYL-L-CYSTEINYLGLYCINE", "peptide"),
        "R01": (
            "(4'-{[ALLYL(METHYL)AMINO]METHYL}-1,1'-BIPHENYL-4-YL)(4-BROMOPHENYL)METHANONE",
            "non-polymer",
        ),
        "R00": (
            "N-[(2S,3S)-3-AMINO-2-HYDROXY-4-PHENYLBUTYL]-L-PHENYLALANINE",
            "peptide",
        ),
        "R03": (
            "ALLYL-{6-[3-(4-BROMO-PHENYL)-1-METHYL-1H-INDAZOL-6-YL]OXY}HEXYL)-N-METHYLAMINE",
            "non-polymer",
        ),
        "R02": (
            "{4-[((1S,2S)-2-{[ALLYL(CYCLOPROPYL)AMINO]METHYL}CYCLOPROPYL)METHOXY]PHENYL}(4-BROMOPHENYL)METHANONE",
            "non-polymer",
        ),
        "LEF": ("(4S)-5-FLUORO-L-LEUCINE", "peptide"),
        "KIR": ("KIRROMYCIN", "non-polymer"),
        "KIQ": (
            "(1S,6R)-3-{[3-(TRIFLUOROMETHYL)-5,6-DIHYDRO[1,2,4]TRIAZOLO[4,3-A]PYRAZIN-7(8H)-YL]CARBONYL}-6-(2,4,5-TRIFLUOROPHENYL)CYCLOHEX-3-EN-1-AMINE",
            "non-polymer",
        ),
        "LEG": ("1-(4-HEXYLPHENYL)PROP-2-EN-1-ONE", "non-polymer"),
        "KIV": ("3-METHYL-2-OXOBUTANOIC ACID", "non-polymer"),
        "IPG": ("N-ISOPROPYL GLYCINE", "peptide"),
        "QJZ": (
            "N-{6-[2-(METHYLSULFANYL)PYRIMIDIN-4-YL]-1,3-BENZOTHIAZOL-2-YL}ACETAMIDE",
            "non-polymer",
        ),
        "KIN": (
            "1-[4-(PYRIDIN-4-YLOXY)PHENYL]-3-[3-(TRIFLUOROMETHYL)PHENYL]UREA",
            "non-polymer",
        ),
        "KIM": (
            "N~4~-METHYL-N~4~-(3-METHYL-1H-INDAZOL-6-YL)-N~2~-(3,4,5-TRIMETHOXYPHENYL)PYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "LEI": ("3-SULFANYL-D-VALINE", "peptide"),
        "KIB": ("2,6-DIMETHOXYBENZENE-1,4-DIOL", "non-polymer"),
        "KIA": ("2,6-DIMETHOXYCYCLOHEXA-2,5-DIENE-1,4-DIONE", "non-polymer"),
        "BAM": ("BENZAMIDINE", "non-polymer"),
        "KIF": ("KIFUNENSINE", "non-polymer"),
        "TY9": ("3,4-BIS(HYDROPEROXY)-5-HYDROXY-L-PHENYLALANINE", "peptide"),
        "CRR": ("3,4-DI-1H-INDOL-3-YL-1H-PYRROLE-2,5-DICARBOXYLIC ACID", "non-polymer"),
        "CRS": ("M-CRESOL", "non-polymer"),
        "CRP": (
            "((1RS,3SR)-2,2-DICHLORO-N-[(R)-1-(4-CHLOROPHENYL)ETHYL]-1-ETHYL-3-METHYLCYCLOPROPANECARBOXAMIDE",
            "non-polymer",
        ),
        "CRQ": (
            "[2-(3-CARBAMOYL-1-IMINO-PROPYL)-4-(4-HYDROXY-BENZYLIDENE)-5-OXO-4,5-DIHYDRO-IMIDAZOL-1-YL]-ACETIC ACID",
            "peptide",
        ),
        "CRV": (
            "N-[1-(AMINOMETHYL)CYCLOPROPYL]-3-(BENZYLSULFONYL)-N~2~-[(1S)-2,2,2-TRIFLUORO-1-(4-HYDROXYPHENYL)ETHYL]-L-ALANINAMIDE",
            "non-polymer",
        ),
        "CRW": (
            "[2-(1-AMINOETHYL)-4-METHYLENE-5-OXO-4,5-DIHYDRO-1H-IMIDAZOL-1-YL]ACETIC ACID",
            "peptide",
        ),
        "CRT": ("SPIRILLOXANTHIN", "non-polymer"),
        "CRU": (
            "4-[(4Z)-1-(CARBOXYMETHYL)-4-(4-HYDROXYBENZYLIDENE)-5-OXO-4,5-DIHYDRO-1H-IMIDAZOL-2-YL]-4-IMINOBUTANOIC ACID",
            "peptide",
        ),
        "CRZ": ("4-(9H-CARBAZOL-9-YL)BUTANOIC ACID", "non-polymer"),
        "CRX": (
            "[2-(1-AMINOETHYL)-2-HYDROXY-4-METHYLENE-5-OXOIMIDAZOLIDIN-1-YL]ACETIC ACID",
            "peptide",
        ),
        "CRY": ("PROPANE-1,2,3-TRIOL", "non-polymer"),
        "PVA": ("1-AMINO-2-METHYL-PROPYLPHOSPHONIC ACID", "non-polymer"),
        "CRB": (
            "[1R-(1ALPHA,3BETA,4ALPHA,5BETA)]-5-(PHOSPHONOMETHYL)-1,3,4-TRIHYDROXYCYCLOHEXANE-1-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "CRC": ("CAPRIC ACID", "non-polymer"),
        "PU7": ("9-BUTYL-8-(2,5-DIMETHOXY-BENZYL)-9H-PURIN-6-YLAMINE", "non-polymer"),
        "CRA": (
            "1-DEOXY-1-METHOXYCARBAMIDO-BETA-D-GLUCO-2-HEPTULOPYRANOSONAMIDE",
            "saccharide",
        ),
        "CRF": (
            "[(4Z)-2-[(1R,2R)-1-AMINO-2-HYDROXYPROPYL]-4-(1H-INDOL-3-YLMETHYLIDENE)-5-OXO-4,5-DIHYDRO-1H-IMIDAZOL-1-YL]ACETIC ACID",
            "peptide",
        ),
        "CRG": (
            "[2-(1-AMINO-2-HYDROXY-PROPYL)-4-(3H-IMIDAZOL-4-YLMETHYLENE)-5-OXO-4,5-DIHYDRO-IMIDAZOL-1-YL]-ACETIC ACID",
            "peptide",
        ),
        "CRD": ("(2E)-BUT-2-ENAL", "non-polymer"),
        "CRE": ("GERMACRENE A", "non-polymer"),
        "CRJ": (
            "N-(1-CYANOCYCLOPROPYL)-3-({[(2S)-5-OXOPYRROLIDIN-2-YL]METHYL}SULFONYL)-N~2~-[(1S)-2,2,2-TRIFLUORO-1-(4-FLUOROPHENYL)ETHYL]-L-ALANINAMIDE",
            "non-polymer",
        ),
        "CRK": (
            "4-{(Z)-[2-[3-(METHYLSULFANYL)PROPANOYL]-5-OXO-1-(2-OXOETHYL)-1,5-DIHYDRO-4H-IMIDAZOL-4-YLIDENE]METHYL}BENZENOLATE",
            "peptide",
        ),
        "CRH": (
            "1,2-HYDRO-1-OXY-3,4-HYDRO-3-(1-METHOXY-2-OXY-3,4-DIHYDROXYPENTYL)-8,9-DIHYROXY-7-METHYLANTHRACENE",
            "non-polymer",
        ),
        "CRI": ("?", "non-polymer"),
        "CRN": ("N-[(E)-AMINO(IMINO)METHYL]-N-METHYLGLYCINE", "non-polymer"),
        "CRO": (
            "{2-[(1R,2R)-1-AMINO-2-HYDROXYPROPYL]-4-(4-HYDROXYBENZYLIDENE)-5-OXO-4,5-DIHYDRO-1H-IMIDAZOL-1-YL}ACETIC ACID",
            "peptide",
        ),
        "CRL": (
            "N-{(1R)-2-[(4-CYANO-1,1-DIOXIDOTETRAHYDRO-2H-THIOPYRAN-4-YL)AMINO]-2-OXO-1-[(TRIMETHYLSILYL)METHYL]ETHYL}MORPHOLINE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "CRM": (
            "3-[8-((2S,3S,4R)-2,3,4,5-TETRAHYDROXYPENTYL)-2,4,7-TRIOXO-1,3,8-TRIHYDROPTERIDIN-6-YL]PROPANOIC ACID",
            "non-polymer",
        ),
        "PR": ("PRASEODYMIUM ION", "non-polymer"),
        "LEP": ("N~2~-PHOSPHONO-L-LEUCINAMIDE", "peptide"),
        "13J": (
            "3-{4-AMINO-1-[(3S)-1-PROPANOYLPIPERIDIN-3-YL]-1H-PYRAZOLO[3,4-D]PYRIMIDIN-3-YL}-N-[4-(PROPAN-2-YL)PHENYL]BENZAMIDE",
            "non-polymer",
        ),
        "PT": ("PLATINUM (II) ION", "non-polymer"),
        "PU": ("PUROMYCIN-N-AMINOPHOSPHONIC ACID", "nucleic-acid"),
        "NU5": (
            "3-({2-[(4-{[6-(CYCLOHEXYLMETHOXY)-9H-PURIN-2-YL]AMINO}PHENYL)SULFONYL]ETHYL}AMINO)PROPAN-1-OL",
            "non-polymer",
        ),
        "TY1": ("O-TERT-BUTYL-L-TYROSINE", "peptide"),
        "PB": ("LEAD (II) ION", "non-polymer"),
        "5LK": (
            "4-{(6R,7R)-7-AMINO-3-[3-(4-METHYLPIPERAZIN-1-YL)PHENYL]-6,7-DIHYDROPYRAZOLO[1,5-A]PYRIMIDIN-6-YL}PHENOL",
            "non-polymer",
        ),
        "5LI": (
            "2,3-DICHLORO-N-{2-[FORMYL(HYDROXY)AMINO]ETHYL}BENZAMIDE",
            "non-polymer",
        ),
        "PF": ("PROFLAVINE", "non-polymer"),
        "PD": ("PALLADIUM ION", "non-polymer"),
        "LET": (
            "(Z)-N^6-{3-CARBOXY-1-[(4-CARBOXY-2-OXOBUTOXY)METHYL]PROPYLIDENE}-L-LYSINE",
            "peptide",
        ),
        "PI": ("HYDROGENPHOSPHATE ION", "non-polymer"),
        "LEU": ("LEUCINE", "peptide"),
        "RHC": (
            "5-(3-AMINO-4,4-DIHYROXY-BUTYLSULFANYLMETHYL)-TETRAHYDRO-FURAN-2,3,4-TRIOL",
            "non-polymer",
        ),
        "LEW": (
            "1-BETA-D-GLUCOPYRANOSYL-4-NAPHTHALEN-1-YL-1H-1,2,3-TRIAZOLE",
            "non-polymer",
        ),
        "BAH": ("BIS(5-AMIDINO-2-BENZIMIDAZOLYL)METHANE KETONE HYDRATE", "non-polymer"),
        "OXS": ("2-OXALOSUCCINIC ACID", "non-polymer"),
        "CR2": (
            "{(4Z)-2-(AMINOMETHYL)-4-[(4-HYDROXYPHENYL)METHYLIDENE]-5-OXO-4,5-DIHYDRO-1H-IMIDAZOL-1-YL}ACETIC ACID",
            "peptide",
        ),
        "B01": (
            "D-PHENYLALANYL-N-[(4-CHLOROPYRIDIN-2-YL)METHYL]-L-PROLINAMIDE",
            "peptide",
        ),
        "CR0": (
            "[2-(1-AMINO-2-HYDROXYPROPYL)-2-HYDROXY-4-ISOBUTYL-5-OXO-2,5-DIHYDRO-1H-IMIDAZOL-1-YL]ACETALDEHYDE",
            "peptide",
        ),
        "CR1": ("1-DEOXY-1-METHOXYCARBAMIDO-BETA-D-GLUCOPYRANOSE", "saccharide"),
        "CR6": (
            "1-DEOXY-1-ACETYLAMINO-BETA-D-GLUCO-2-HEPTULOPYRANOSONAMIDE",
            "saccharide",
        ),
        "CR7": (
            "[(4Z)-2-[(1S)-1,5-DIAMINOPENTYL]-4-(4-HYDROXYBENZYLIDENE)-5-OXO-4,5-DIHYDRO-1H-IMIDAZOL-1-YL]ACETIC ACID",
            "peptide",
        ),
        "CR4": (
            "2-{5-[AMINO(IMINIO)METHYL]-1H-BENZIMIDAZOL-2-YL}BENZENOLATE",
            "non-polymer",
        ),
        "B00": (
            "2-AMINO-3-METHYL-6-[(1S,2R)-2-PHENYLCYCLOPROPYL]PYRIMIDIN-4(3H)-ONE",
            "non-polymer",
        ),
        "CR8": (
            "2-[1-AMINO-2-(1H-IMIDAZOL-5-YL)ETHYL]-1-(CARBOXYMETHYL)-4-[(4-OXOCYCLOHEXA-2,5-DIEN-1-YLIDENE)METHYL]-1H-IMIDAZOL-5-OLATE",
            "peptide",
        ),
        "CR9": (
            "2-{5-[AMINO(IMINIO)METHYL]-6-FLUORO-1H-BENZIMIDAZOL-2-YL}-6-[(2-METHYLCYCLOHEXYL)OXY]BENZENOLATE",
            "non-polymer",
        ),
        "B03": ("D-PHENYLALANYL-N-(PYRIDIN-2-YLMETHYL)-L-PROLINAMIDE", "peptide"),
        "OPD": ("9R,13R-12-OXOPHYTODIENOIC ACID", "non-polymer"),
        "OPE": ("PHOSPHORIC ACID MONO-(2-AMINO-ETHYL) ESTER", "non-polymer"),
        "B02": (
            "2-AMINO-3-METHYL-6-[(1S,2R)-2-(3'-METHYLBIPHENYL-4-YL)CYCLOPROPYL]PYRIMIDIN-4(3H)-ONE",
            "non-polymer",
        ),
        "BAI": ("(5-AMIDINO-2-BENZIMIDAZOLYL)(2-BENZIMIDAZOLYL)METHANE", "non-polymer"),
        "OPA": (
            "2-(OXALYL-AMINO)-4,7-DIHYDRO-5H-THIENO[2,3-C]PYRAN-3-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "OPB": (
            "4-BUTYL-1-(4-HYDROXYPHENYL)-2-PHENYLPYRAZOLIDINE-3,5-DIONE",
            "non-polymer",
        ),
        "OPC": (
            "(7R,17E)-4-HYDROXY-N,N,N,7-TETRAMETHYL-7-[(8E)-OCTADEC-8-ENOYLOXY]-10-OXO-3,5,9-TRIOXA-4-PHOSPHAHEPTACOS-17-EN-1-AMINIUM 4-OXIDE",
            "non-polymer",
        ),
        "OPL": ("(2R,3R)-2,4,4-TRIHYDROXY-3-METHYLBUTANAL", "non-polymer"),
        "OPM": ("O1-PENTYL-MANNOSE", "saccharide"),
        "OPN": (
            "{(2-AMINO-ETHYL)-[2-(2-OXO-1,2-DIHYDRO-[1,8]NAPHTHYRIDIN-3-YL)-ACETYL]-AMINO}-ACETIC ACID",
            "non-polymer",
        ),
        "OPH": ("2-HYDROXY-3-PHENYL-PROPIONIC ACID METHYL ESTER", "non-polymer"),
        "OPI": ("PANTOTHENYL-AMINOETHANOL-11-PIVALIC ACID", "non-polymer"),
        "B04": (
            "D-PHENYLALANYL-N-[(1-METHYLPYRIDINIUM-2-YL)METHYL]-L-PROLINAMIDE",
            "peptide",
        ),
        "AXQ": (
            "{(E)-(3R,6S,9R)-3-[(1S,3R)-3-((S)-1 -BUTYLCARBAMOYL-2-METHYL-PROPYLCARB AMOYL)-1-HYDROXY-BUTYL]-6-METHYL-5, 8-DIOXO-1,11-DITHIA-4,7-DIAZA-CYCLO PENTADEC-13-EN-9-YL}-CARBAMIC ACID TERT-BUTYL ESTER",
            "non-polymer",
        ),
        "D15": (
            "N-(5-{[(2S)-4-AMINO-2-(3-CHLOROPHENYL)BUTANOYL]AMINO}-1H-INDAZOL-3-YL)BENZAMIDE",
            "non-polymer",
        ),
        "D16": ("TOMUDEX", "non-polymer"),
        "578": (
            "(5R,6S,8S)-8-[3-(AMINOMETHYL)PHENYL]-6-HYDROXY-5-ISOPROPYL-3-OXO-1-PHENYL-2,7-DIOXA-4-AZA-6-PHOSPHANONAN-9-OIC ACID 6-OXIDE",
            "non-polymer",
        ),
        "D10": ("DECANE", "non-polymer"),
        "D11": ("D-PHOSPHOTHREONINE", "peptide"),
        "D12": ("DODECANE", "non-polymer"),
        "D13": (
            "4-[5-(3-IODO-PHENYL)-2-(4-METHANESULFINYL-PHENYL)-1H-IMIDAZOL-4-YL]-PYRIDINE",
            "non-polymer",
        ),
        "572": (
            "4-[2-(HYDROXYMETHYL)PYRIMIDIN-4-YL]-N,N-DIMETHYLPIPERAZINE-1-SULFONAMIDE",
            "non-polymer",
        ),
        "571": (
            "5-[(2R)-2-HYDROXY-2-PHENYLACETYL]-3-({[4-(4-METHYLPIPERAZIN-1-YL)PHENYL]CARBONYL}AMINO)-1,6-DIHYDROPYRROLO[3,4-C]PYRAZOL-5-IUM ",
            "non-polymer",
        ),
        "570": (
            "2-(2-BENZOYL-PHENYLAMINO)-3-{4-[2-(5-METHYL-2-PHENYL-OXAZOL-4-YL)-ETHOXY]-PHENYL}-PROPIONIC ACID",
            "non-polymer",
        ),
        "577": (
            "4-[5-(4-FLUORO-PHENYL)-2-[4-METHANESULFINYL-PHENYL]-3H-IMIDAZOL-4-YL]-PYRIDINE",
            "non-polymer",
        ),
        "576": ("(3S)-5-(9H-FLUOREN-2-YL)-3-HYDROXYPENTANOIC ACID", "non-polymer"),
        "575": (
            "(4Z)-6-BROMO-4-({[4-(PYRROLIDIN-1-YLMETHYL)PHENYL]AMINO}METHYLIDENE)ISOQUINOLINE-1,3(2H,4H)-DIONE",
            "non-polymer",
        ),
        "B09": ("N-METHOXYBENZENESULFONAMIDE", "non-polymer"),
        "B08": ("unknown", "non-polymer"),
        "TM1": ("?", "non-polymer"),
        "OEM": ("N-METHYL-D-ASPARTIC ACID", "peptide"),
        "V21": (
            "3-METHYLTHIOBENZIMIDAZO[1,2-C][1,2,3]THIADIAZOL-7-SULFONAMIDE",
            "non-polymer",
        ),
        "PXK": (
            "6-(4,4-DIMETHYLPENT-2-YN-1-YL)-1LAMBDA~4~-PYRROLO[2,3-D][1,3]THIAZOLE",
            "non-polymer",
        ),
        "V25": (
            "ETHYL 3-[(E)-2-AMINO-1-CYANOETHENYL]-6,7-DICHLORO-1-METHYL-1H-INDOLE-2-CARBOXYLATE",
            "non-polymer",
        ),
        "9CS": (
            "(1R,2S,3S,4R,6S)-4,6-DIAMINO-3-[(3-AMINO-3-DEOXY-ALPHA-D-GLUCOPYRANOSYL)OXY]-2-HYDROXYCYCLOHEXYL 2,6-DIAMINO-2,6-DIDEOXY-ALPHA-D-GLUCOPYRANOSIDE",
            "non-polymer",
        ),
        "9CR": ("(9CIS)-RETINOIC ACID", "non-polymer"),
        "9CL": ("(CHLOROMETHYL)BENZENE", "non-polymer"),
        "9CA": ("9H-CARBAZOLE", "non-polymer"),
        "TCC": ("2-(2,4-DICHLORO-PHENYLAMINO)-PHENOL", "non-polymer"),
        "R39": (
            "1-[3-TERT-BUTYL-1-(4-METHYLPHENYL)-1H-PYRAZOL-5-YL]-3-[4-(2-MORPHOLIN-4-YLETHOXY)PHENYL]UREA",
            "non-polymer",
        ),
        "PSK": (
            "[(3R,4S)-4-HYDROXY-3-METHYL-2-OXOHEXYL]PHOSPHONIC ACID",
            "non-polymer",
        ),
        "TBT": ("8,9,10,11-TETRAHYDRO-BENZO[A]ANTHRACENE-8,9,10-TRIOL", "non-polymer"),
        "R35": ("(2R)-2-TERT-BUTYLBUTANEDIOIC ACID", "non-polymer"),
        "V2H": (
            "(1S,3R,5Z,7E,22E)-9,10-SECOERGOSTA-5,7,10,22-TETRAENE-1,3-DIOL",
            "non-polymer",
        ),
        "D1T": ("DECANE-1-THIOL", "non-polymer"),
        "D1U": (
            "4-[6,6-DIMETHYL-4-OXO-3-(TRIFLUOROMETHYL)-4,5,6,7-TETRAHYDRO-1H-INDAZOL-1-YL]-2-{[2-(METHYLSULFANYL)ETHYL]AMINO}BENZAMIDE",
            "non-polymer",
        ),
        "D1V": ("2-MERCAPTO(3H)QUINAZOLINONE", "non-polymer"),
        "D1P": ("2'-DEOXY-RIBOFURANOSE-5'-PHOSPHATE", "nucleic-acid"),
        "D1Q": (
            "4-{3-[(4-CHLOROPHENYL)AMINO]-3-OXOPROPYL}-3-({[5-(PROPAN-2-YL)-4,5,6,7-TETRAHYDRO[1,3]THIAZOLO[5,4-C]PYRIDIN-2-YL]CARBONYL}AMINO)BENZOIC ACID",
            "non-polymer",
        ),
        "D1R": (
            "NALPHA-[(4-METHYLPIPERAZIN-1-YL)CARBONYL]-N-{(1S)-3-PHENYL-1-[2-(PHENYLSULFONYL)ETHYL]PROPYL}-L-PHENYLALANINAMIDE",
            "non-polymer",
        ),
        "D1S": (
            "(4E,6Z,8S,9S,10E,12S,13R,14S,16R)-19-{[3-(DIMETHYLAMINO)PROPYL]AMINO}-13-HYDROXY-8,14-DIMETHOXY-4,10,12,16-TETRAMETHYL-3,20,22-TRIOXO-2-AZABICYCLO[16.3.1]DOCOSA-1(21),4,6,10,18-PENTAEN-9-YL CARBAMATE",
            "non-polymer",
        ),
        "KMT": ("4-(METHYLSULFANYL)-2-OXOBUTANOIC ACID", "non-polymer"),
        "D1X": (
            "2,6-DIOXO-1,2,3,6-TETRAHYDROPYRIDINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "D1D": ("(4S,5S)-1,2-DITHIANE-4,5-DIOL", "non-polymer"),
        "PSB": (
            "3-(5-AMINO-7-HYDROXY-[1,2,3]TRIAZOLO[4,5-D]PYRIMIDIN-2-YL)-N-[2-(2-(HYDROXYMETHYL-PHENYLSULFANYL)-BENZYL]-BENZAMIDE",
            "non-polymer",
        ),
        "D1F": (
            "1-(BETA-D-GLUCOPYRANOSYL)-5-(HEPT-1-YN-1-YL)PYRIMIDINE-2,4(1H,3H)-DIONE",
            "non-polymer",
        ),
        "D1G": ("N,N-DIMETHYL(5-(PYRIDIN-3-YL)FURAN-2-YL)METHANAMINE", "non-polymer"),
        "D1A": (
            "N-[4-[(E)-N-CARBAMIMIDAMIDO-C-METHYL-CARBONIMIDOYL]PHENYL]-7-NITRO-1H-INDOLE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "D1B": (
            "2-(5-{4-[AMINO(IMINO)METHYL]PHENYL}-2-THIENYL)-1H-BENZIMIDAZOLE-6- CARBOXIMIDAMIDE DIHYDROCHLORIDE",
            "non-polymer",
        ),
        "D1C": (
            "4-[({[(2R)-2-AMINO-3-(2,4,5-TRIFLUOROPHENYL)PROPYL]SULFAMOYL}AMINO)METHYL]BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "D1L": ("2-[4-(2,4-DICHLOROPHENOXY)PHENOXY]PROPANOIC ACID", "non-polymer"),
        "D1M": (
            "5-ETHYNYL-1-(BETA-D-GLUCOPYRANOSYL)PYRIMIDINE-2,4(1H,3H)-DIONE",
            "non-polymer",
        ),
        "D1N": ("NAPHTHALENE-1,2-DIOL", "non-polymer"),
        "D1H": (
            "1,1'-ETHANE-1,2-DIYLBIS(3,7-DIMETHYL-3,7-DIHYDRO-1H-PURINE-2,6-DIONE)",
            "non-polymer",
        ),
        "D1I": (
            "3-(BETA-D-GLUCOPYRANOSYL)-6-PENTYLFURO[2,3-D]PYRIMIDIN-2(3H)-ONE",
            "non-polymer",
        ),
        "D1J": (
            "1-(BETA-D-GLUCOPYRANOSYL)-5-(PENT-1-YN-1-YL)PYRIMIDINE-2,4(1H,3H)-DIONE",
            "non-polymer",
        ),
        "57D": ("CHRYSIN", "non-polymer"),
        "35G": ("GUANOSINE-3',5'-MONOPHOSPHATE", "non-polymer"),
        "35F": (
            "N-{3-[5-(1H-1,2,4-TRIAZOL-3-YL)-1H-INDAZOL-3-YL]PHENYL}FURAN-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "35C": ("3,5-DICHLOROBENZENE-1,2-DIOL", "non-polymer"),
        "35B": (
            "4-BROMO-3-(CARBOXYMETHOXY)-5-{3-[CYCLOHEXYL(METHYLCARBAMOYL)AMINO]PHENYL}THIOPHENE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "35A": (
            "N-[(5R,14R)-5-AMINO-5,14-DIMETHYL-4-OXO-3-OXA-18-AZATRICYCLO[15.3.1.1~7,11~]DOCOSA-1(21),7(22),8,10,17,19-HEXAEN-19-YL]-N-METHYLMETHANESULFONAMIDE",
            "non-polymer",
        ),
        "35I": (
            "(2E)-3-{5-[(2,4-DIAMINOPYRIMIDIN-5-YL)METHYL]-2,3-DIMETHOXYPHENYL}-1-[(1S)-1-(3,3,3-TRIFLUOROPROPYL)PHTHALAZIN-2(1H)-YL]PROP-2-EN-1-ONE",
            "non-polymer",
        ),
        "PSG": ("PARA-NITROPHENYL 1-THIO-BETA-D-GLUCOPYRANOSIDE", "non-polymer"),
        "PF9": (
            "2-{[4-(1-METHYL-4-PYRIDIN-4-YL-1H-PYRAZOL-3-YL)PHENOXY]METHYL}QUINOLINE",
            "non-polymer",
        ),
        "PF8": (
            "2-{[4-(4-PYRIDIN-4-YL-1H-PYRAZOL-3-YL)PHENOXY]METHYL}QUINOLINE",
            "non-polymer",
        ),
        "PSE": ("O-PHOSPHOETHANOLAMINE", "non-polymer"),
        "35Y": ("3-[(1H-TETRAZOL-5-YLCARBONYL)AMINO]-L-ALANINE", "peptide"),
        "PSZ": (
            "4-[({3-HYDROXY-2-METHYL-5-[(PHOSPHONOOXY)METHYL]PYRIDIN-4-YL}METHYL)AMINO]THIOPHENE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "L6S": ("6-O-SULFO-ALPHA-L-GALACTOSE", "saccharide"),
        "PSX": (
            "[(2S,3R,4S)-2,4-DIHYDROXY-3-METHYLHEXYL]PHOSPHONIC ACID",
            "non-polymer",
        ),
        "3N6": (
            "N-{(1S)-5-AMINO-1-[(4-PYRIDIN-4-YLPIPERAZIN-1-YL)CARBONYL]PENTYL}-3,5-DIBROMO-NALPHA-{[4-(2-OXO-1,4-DIHYDROQUINAZOLIN-3(2H)-YL)PIPERIDIN-1-YL]CARBONYL}-D-TYROSINAMIDE",
            "non-polymer",
        ),
        "3N7": (
            "N~4~-(5-CYCLOPROPYL-1H-PYRAZOL-3-YL)-N~2~-1H-INDAZOL-5-YL-6-METHYLPYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "3N4": ("NAPHTHO[2,1-B]THIOPHEN-1-YLACETIC ACID", "non-polymer"),
        "3N5": (
            "(2R)-1-[(BENZYLOXY)CARBONYL]-2,3-DIHYDRO-1H-INDOLE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "3N2": ("(5-CHLORO-1-BENZOTHIOPHEN-3-YL)ACETIC ACID", "non-polymer"),
        "3N3": (
            "(2S)-1-[(BENZYLOXY)CARBONYL]-2,3-DIHYDRO-1H-INDOLE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "3N1": ("(5-CHLORO-3-METHYL-1-BENZOTHIOPHEN-2-YL)ACETIC ACID", "non-polymer"),
        "PXN": (
            "(2S)-1-[3-{[(2R)-2-HYDROXYPROPYL]OXY}-2,2-BIS({[(2R)-2-HYDROXYPROPYL]OXY}METHYL)PROPOXY]PROPAN-2-OL",
            "non-polymer",
        ),
        "PVC": (
            "(2E,4R,5S)-2,3,4,5-TETRAHYDROXY-6-(PALMITOYLOXY)HEX-2-ENOIC ACID",
            "non-polymer",
        ),
        "ZK1": (
            "{[7-MORPHOLIN-4-YL-2,3-DIOXO-6-(TRIFLUOROMETHYL)-3,4-DIHYDROQUINOXALIN-1(2H)-YL]METHYL}PHOSPHONIC ACID",
            "non-polymer",
        ),
        "3NF": ("N-ACETYL-L-TYROSINE", "peptide"),
        "3NG": (
            "5-[(3-CHLOROPHENYL)AMINO]BENZO[C][2,6]NAPHTHYRIDINE-8-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "3ND": (
            "(3S,4R)-N-(7-CHLORO-1-OXO-1,4-DIHYDROISOQUINOLIN-6-YL)-4-(4-CHLOROPHENYL)PYRROLIDINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "3NE": (
            "4-(TRIFLUOROMETHYL)-3-[3-(TRIFLUOROMETHYL)PHENYL]-1,7-DIHYDRO-6H-PYRAZOLO[3,4-B]PYRIDIN-6-ONE",
            "non-polymer",
        ),
        "3NC": (
            "CIS-4-AMINO-N-(7-CHLORO-1-OXO-1,2-DIHYDROISOQUINOLIN-6-YL)CYCLOHEXANECARBOXAMIDE",
            "non-polymer",
        ),
        "3NA": (
            "{3-[(4,5,7-TRIFLUORO-1,3-BENZOTHIAZOL-2-YL)METHYL]-1H-INDOL-1-YL}ACETIC ACID",
            "non-polymer",
        ),
        "ZKD": (
            "URIDINE-DIPHOSPHATE-1-DEOXY-1-THIO-N-ACETYLGLUCOSAMINE",
            "non-polymer",
        ),
        "NFS": ("FE(4)-NI(1)-S(5) CLUSTER", "non-polymer"),
        "3NM": (
            "4-METHYL-5-[2-(PHOSPHONOOXY)ETHYL]-1,3-THIAZOLE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "3NK": ("1-(6-METHOXYNAPHTHALEN-2-YL)BUTANE-1,3-DIONE", "non-polymer"),
        "3NH": (
            "(3S)-TETRAHYDROFURAN-3-YL (1R,2S)-3-[4-((1R)-2-{[(S)-AMINO(HYDROXY)METHYL]OXY}-2,3-DIHYDRO-1H-INDEN-1-YL)-2-BENZYL-3-OXOPYRROLIDIN-2-YL]-1-BENZYL-2-HYDROXYPROPYLCARBAMATE",
            "non-polymer",
        ),
        "3NI": ("NICKEL (III) ION", "non-polymer"),
        "3NV": (
            "4-[5-CHLORO-3-(1-METHYLETHYL)-1H-PYRAZOL-4-YL]-N-(5-PIPERAZIN-1-YLPYRIDIN-2-YL)PYRIMIDIN-2-AMINE",
            "non-polymer",
        ),
        "3NW": (
            "1-[2-(2-{[2-(DIMETHYLAMINO)ETHYL]AMINO}-6-{2-[(1-METHYLETHYL)AMINO]-1,3-THIAZOL-5-YL}PYRIMIDIN-4-YL)BENZYL]-3-ETHYLUREA",
            "non-polymer",
        ),
        "3NT": ("3-NITROTOLUENE", "non-polymer"),
        "3NU": (
            "4-[3-(1-METHYLETHYL)-1H-PYRAZOL-4-YL]-N-(1-METHYLPIPERIDIN-4-YL)PYRIMIDIN-2-AMINE",
            "non-polymer",
        ),
        "3NS": (
            "(3S)-4-HYDROXY-3-{[(2S)-4-{[2-(2-METHYL-1H-BENZIMIDAZOL-1-YL)ETHYL]AMINO}-2-(1-METHYLETHYL)-4-OXOBUTANOYL]AMINO}BUTANOIC ACID",
            "peptide",
        ),
        "3NP": ("3-NITROPROPANOIC ACID", "non-polymer"),
        "PSS": ("(2S)-2-[(3R)-3-BENZYL-2-OXOPIPERAZIN-1-YL]HEXANOIC ACID", "peptide"),
        "TC1": (
            "3-(5-PHOSPHO-2-DEOXY-BETA-D-RIBOFURANOSYL)-2-OXO-1,3-DIAZA-PHENOTHIAZINE",
            "nucleic-acid",
        ),
        "3NY": ("5-NITRO-1H-BENZOTRIAZOLE", "non-polymer"),
        "PSQ": ("NDELTA-(N'-SULPHODIAMINOPHOSPHINYL)-L-ORNITHINE", "non-polymer"),
        "LPM": ("6,8-DIMERCAPTO-OCTANOIC ACID AMIDE", "non-polymer"),
        "PSW": ("3-(SULFANYLSELANYL)-L-ALANINE", "peptide"),
        "357": (
            "5-ETHYL-3-[(2-METHOXYETHYL)METHYLAMINO]-6-METHYL-4-(3-METHYLBENZYL)PYRIDIN-2(1H)-ONE",
            "non-polymer",
        ),
        "356": (
            "8-[(3R)-3-AMINOPIPERIDIN-1-YL]-7-BUT-2-YN-1-YL-3-METHYL-1-[(4-METHYLQUINAZOLIN-2-YL)METHYL]-3,7-DIHYDRO-1H-PURINE-2,6-DIONE",
            "non-polymer",
        ),
        "GTG": ("7-METHYL-GUANOSINE-5'-TRIPHOSPHATE-5'-GUANOSINE", "non-polymer"),
        "354": (
            "N-[(S)-(4-{[(2-AMINO-4-HYDROXYQUINAZOLIN-6-YL)(DIHYDROXY)-LAMBDA~4~-SULFANYL]AMINO}PHENYL)(HYDROXY)METHYL]-L-GLUTAMIC ACID",
            "non-polymer",
        ),
        "353": (
            "N-{4-[(2-AMINO-3-CHLOROPYRIDIN-4-YL)OXY]-3-FLUOROPHENYL}-4-ETHOXY-1-(4-FLUOROPHENYL)-2-OXO-1,2-DIHYDROPYRIDINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "352": (
            "(5S)-2-{[(1S)-1-(4-FLUOROPHENYL)ETHYL]AMINO}-5-(1-HYDROXY-1-METHYLETHYL)-5-METHYL-1,3-THIAZOL-4(5H)-ONE",
            "non-polymer",
        ),
        "351": (
            "1-(3-{5-[4-(AMINOMETHYL)PHENYL]-1H-PYRROLO[2,3-B]PYRIDIN-3-YL}PHENYL)-3-(2-PHENOXYPHENYL)UREA",
            "non-polymer",
        ),
        "SHI": ("(S)-2-AMINO-3-(3-HYDROXY-ISOXAZOL-4-YL)PROPIONIC ACID", "non-polymer"),
        "359": (
            "3-[[3-[(2R)-3-[[(2S)-5-AMINO-1-[(4-CARBAMIMIDOYLPHENYL)METHYLAMINO]-1,5-DIOXO-PENTAN-2-YL]AMINO]-2-(ETHYLSULFONYLAMINO)-3-OXO-PROPYL]-1H-INDOL-5-YL]OXYMETHYL]BENZOIC ACID",
            "non-polymer",
        ),
        "358": (
            "1-(2,6-DICHLOROPHENYL)-5-(2,4-DIFLUOROPHENYL)-7-PIPERIDIN-4-YL-3,4-DIHYDROQUINOLIN-2(1H)-ONE",
            "non-polymer",
        ),
        "HNA": ("1,8-DI-HYDROXY-4-NITRO-ANTHRAQUINONE", "non-polymer"),
        "NM7": (
            "2-(CARBAMOYLAMINO)-5-(4-FLUOROPHENYL)THIOPHENE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "HNC": ("S-[(1S,2R)-2-HYDROXY-1-(2-OXOETHYL)HEPTYL]-L-CYSTEINE", "peptide"),
        "HNB": ("(2S,5R)-5-PENTYLTETRAHYDROFURAN-2-OL", "non-polymer"),
        "HNE": ("(2E,4R)-4-HYDROXYNON-2-ENAL", "non-polymer"),
        "1H2": (
            "3-(3,5-DIMETHYL-1,2-OXAZOL-4-YL)-5-[(R)-HYDROXY(PHENYL)METHYL]PHENOL",
            "non-polymer",
        ),
        "1H3": (
            "3-(3,5-DIMETHYL-1,2-OXAZOL-4-YL)-5-[(S)-HYDROXY(PHENYL)METHYL]PHENOL",
            "non-polymer",
        ),
        "1H1": ("ETHYL (3R)-4-CYANO-3-HYDROXYBUTANOATE", "non-polymer"),
        "1H4": (
            "N-{3-[(5-CYCLOPROPYL-2-{[3-(2-OXO-2,3-DIHYDRO-1H-PYRROL-1-YL)PHENYL]AMINO}PYRIMIDIN-4-YL)AMINO]PROPYL}CYCLOBUTANECARBOXAMIDE",
            "non-polymer",
        ),
        "XA0": (
            "4-CHLORO-5-NITRO-2-[(3,4,5-TRIFLUOROBENZYL)AMINO]BENZAMIDE",
            "non-polymer",
        ),
        "HNF": ("7-NITRO-9H-FLUOREN-2-OL", "non-polymer"),
        "GTI": ("1-(3-PHENYL-1,2,4-THIADIAZOL-5-YL)PIPERAZINE", "non-polymer"),
        "HNI": ("PROTOPORPHYRIN IX CONTAINING NI(II)", "non-polymer"),
        "7I2": ("7-NITROINDAZOLE-2-CARBOXAMIDINE", "non-polymer"),
        "7I1": (
            "3-METHYL-N-(NAPHTHALEN-1-YLMETHYL)-4-PIPERIDIN-4-YLOXY-1-BENZOFURAN-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "HNH": ("5-(1-METHYL-4,5-DIHYDRO-1H-PYRROL-2-YL)PYRIDIN-2-OL", "non-polymer"),
        "HNK": ("5-[(2R)-1-METHYLPYRROLIDIN-2-YL]PYRIDIN-2-OL", "non-polymer"),
        "LS1": (
            "N-METHYL-4-{[(2-OXO-1,2-DIHYDRO-3H-INDOL-3-YLIDENE)METHYL]AMINO}BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "SGL": ("1-HYDROXY-1-THIO-GLYCEROL", "non-polymer"),
        "7IG": (
            "5-{4-[(3,5-DIFLUOROBENZYL)AMINO]PHENYL}-6-ETHYLPYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "1HS": ("2-{[(2S)-2-HYDROXYPROPYL]SULFANYL}ETHANESULFONIC ACID", "non-polymer"),
        "1HP": (
            "4-(2-AMINO-1-METHYL-1H-IMIDAZO[4,5-B]PYRIDIN-6-YL)PHENOL",
            "non-polymer",
        ),
        "X62": (
            "4-CHLORO-2-{[(2-CHLOROPYRIMIDIN-5-YL)METHYL]AMINO}-5-NITROBENZAMIDE",
            "non-polymer",
        ),
        "1HW": (
            "1-{(3S,3AS)-3-[3-FLUORO-4-(TRIFLUOROMETHOXY)PHENYL]-8-METHOXY-3,3A,4,5-TETRAHYDRO-2H-BENZO[G]INDAZOL-2-YL}-2-HYDROXYETHANONE",
            "non-polymer",
        ),
        "XAP": ("3-AMINOOXY-1-AMINOPROPANE", "non-polymer"),
        "7IN": (
            "N-(BENZYLSULFONYL)SERYL-N~1~-{4-[(Z)-AMINO(IMINO)METHYL]BENZYL}SERINAMIDE",
            "non-polymer",
        ),
        "1HX": (
            "N-[(1S)-1-(2-CHLORO-6-FLUOROPHENYL)ETHYL]-5-CYANO-1-METHYL-1H-PYRROLE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "SGN": ("N,O6-DISULFO-GLUCOSAMINE", "saccharide"),
        "XAT": ("VIOLAXANTHIN", "non-polymer"),
        "XAV": (
            "2-[4-(TRIFLUOROMETHYL)PHENYL]-7,8-DIHYDRO-5H-THIOPYRANO[4,3-D]PYRIMIDIN-4-OL",
            "non-polymer",
        ),
        "BCA": ("4-HYDROXYBENZOYL COENZYME A", "non-polymer"),
        "1HB": ("(4S)-4-AMINO-5-HYDROXY-N,N-DIMETHYLPENTANAMIDE", "peptide"),
        "SGA": ("O3-SULFONYLGALACTOSE", "saccharide"),
        "VIA": (
            "5-{2-ETHOXY-5-[(4-METHYLPIPERAZIN-1-YL)SULFONYL]PHENYL}-1-METHYL-3-PROPYL-1H,6H,7H-PYRAZOLO[4,3-D]PYRIMIDIN-7-ONE",
            "non-polymer",
        ),
        "1HF": (
            "(3E)-3-{(2Z)-[(5Z)-5-(FURAN-2-YLMETHYLIDENE)-4-OXO-1,3-THIAZOLIDIN-2-YLIDENE]HYDRAZINYLIDENE}-2-OXO-2,3-DIHYDRO-1H-INDOLE-5-SULFONIC ACID",
            "non-polymer",
        ),
        "NM8": (
            "6-{5-[(Z)-(2,4-DIOXO-1,3-THIAZOLIDIN-5-YLIDENE)METHYL]FURAN-2-YL}-N-{3-[(4-ETHYLPIPERAZIN-1-YL)METHYL]-5-(TRIFLUOROMETHYL)PHENYL}NAPHTHALENE-1-CARBOXAMIDE",
            "non-polymer",
        ),
        "1HD": ("N-[(2S)-2-AMINO-3-HYDROXYPROPYL]ACETAMIDE", "peptide"),
        "7IP": ("6-[2-(3'-METHOXYBIPHENYL-3-YL)ETHYL]PYRIDIN-2-AMINE", "non-polymer"),
        "1HK": (
            "4-[2-(2,4-DIHYDROINDENO[1,2-C]PYRAZOL-3-YL)ETHYL]-2-METHOXYPHENOL",
            "non-polymer",
        ),
        "XAC": (
            "N-(2-AMINOETHYL)-2-[4-(2,6-DIOXO-1,3-DIPROPYL-2,3,6,7-TETRAHYDRO-1H-PURIN-8-YL)PHENOXY]ACETAMIDE",
            "non-polymer",
        ),
        "1HN": ("1-HYDROXYNAPHTHALENE-2-CARBOXYLIC ACID", "non-polymer"),
        "SGC": ("4-DEOXY-4-THIO-BETA-D-GLUCOPYRANOSE", "saccharide"),
        "VIG": ("4-AMINO HEXANOIC ACID", "non-polymer"),
        "HNP": ("HYDANTOCIDIN-5'-MONOPHOSPHATE", "non-polymer"),
        "HNR": (
            "N-(6-AMINO-1-BUTYL-2,4-DIOXO-1,2,3,4-TETRAHYDROPYRIMIDIN-5-YL)-N-METHYL-BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "HNT": (
            "[(3R)-7-NITRO-1,2,3,4-TETRAHYDROISOQUINOLIN-3-YL]METHANOL",
            "non-polymer",
        ),
        "HNW": ("D-MANNOSE ALPHA1O P-HYDROXYPROPYNYL-PHENYL", "non-polymer"),
        "AXB": ("?", "non-polymer"),
        "ICJ": (
            "(1R,3S,5Z)-5-[(2E)-2-{(1R,3AS,7AR)-1-[(1R,2S,4R)-2-BUTYL-4-HYDROXY-1,5-DIMETHYLHEXYL]-7A-METHYLOCTAHYDRO-4H-INDEN-4-YLIDENE}ETHYLIDENE]-4-METHYLIDENECYCLOHEXANE-1,3-DIOL",
            "non-polymer",
        ),
        "HNV": ("D-MANNOSE ALPHA1O-PROPYNYL P-METHOXY PHENYL", "non-polymer"),
        "ICI": ("ISOCITRIC ACID", "non-polymer"),
        "VIO": ("N5-(1-IMINO-3-BUTENYL)-L-ORNITHINE", "non-polymer"),
        "VO4": ("VANADATE ION", "non-polymer"),
        "252": ("(1-HYDROXYNONANE-1,1-DIYL)BIS(PHOSPHONIC ACID)", "non-polymer"),
        "UD2": ("URIDINE-DIPHOSPHATE-N-ACETYLGALACTOSAMINE", "non-polymer"),
        "253": (
            "10,11-DIMETHOXY-4-METHYLDIBENZO[C,F]-2,7-NAPHTHYRIDINE-3,6-DIAMINE",
            "non-polymer",
        ),
        "ICO": ("1H-INDOLE-3-CARBOXYLIC ACID", "non-polymer"),
        "250": (
            "(2R)-2-HYDROXY-3-[4-(2-HYDROXYETHYL)PIPERAZIN-1-YL]PROPANE-1-SULFONIC ACID",
            "non-polymer",
        ),
        "ICN": ("(S)-2-HYDROXY-2,3-DIMETHYLBUTANENITRILE", "non-polymer"),
        "251": (
            "N-[AMINO(IMINO)METHYL]-2-(2,5-DIPHENYL-1H-PYRROL-1-YL)ACETAMIDE",
            "non-polymer",
        ),
        "TD1": (
            "O-[(2S)-2-AMINO-3-HYDROXYPROPANOYL]-N-(2,3-DIHYDROXYBENZOYL)-L-SERINE",
            "non-polymer",
        ),
        "8PC": ("2-(2,4-DICHLOROPHENOXY)-5-(PYRIDIN-2-YLMETHYL)PHENOL", "non-polymer"),
        "256": ("PHENYL(SULFO)ACETIC ACID", "non-polymer"),
        "8PA": (
            "3-[(4-AMINO-2-METHYLPYRIMIDIN-5-YL)METHYL]-5-(2-{[(S)-HYDROXY(PHOSPHONOOXY)PHOSPHORYL]OXY}ETHYL)-2-[(1S,2E)-1-HYDROXY-3-PYRIDIN-3-YLPROP-2-EN-1-YL]-4-METHYL-1,3-THIAZOL-3-IUM",
            "non-polymer",
        ),
        "ICL": (
            "[1-(3-CHLORO-2-FORMYL-PHENYLCARBAMOYL)-2-METHYL-PROPYL]-CARBAMIC ACID TERT-BUTYL ESTER",
            "non-polymer",
        ),
        "8PG": (
            "(8S,12S)-15S-HYDROXY-9-OXOPROSTA-10Z,13E-DIEN-1-OIC ACID",
            "non-polymer",
        ),
        "8PE": (
            "(2R)-3-{[(S)-(2-AMINOETHOXY)(HYDROXY)PHOSPHORYL]OXY}-2-(TETRADECANOYLOXY)PROPYL OCTADECANOATE",
            "non-polymer",
        ),
        "SG1": ("3-NITRO-4-(2-OXO-PYRROLIDIN-1-YL)-BENZENESULFONAMIDE", "non-polymer"),
        "254": (
            "6-[HYDROXYIMINO-(5,5,8,8-TETRAMETHYL-5,6,7,8-TETRAHYDRO-NAPHTALEN-2-YL)-METHYL]-NAPHTALENE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "8PS": ("5-OCTYL-2-PHENOXYPHENOL", "non-polymer"),
        "8PR": ("PAROXETINE", "non-polymer"),
        "255": (
            "N-(TERT-BUTYL)-4-[5-(PYRIDIN-2-YLAMINO)QUINOLIN-3-YL]BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "SG3": (
            "[(1E,3R,4S)-4-FLUORO-3,5-DIHYDROXY-4-METHYLPENT-1-EN-1-YL]PHOSPHONIC ACID",
            "non-polymer",
        ),
        "RNN": (
            "N,N'-(BENZENE-1,4-DIYLDIETHANE-2,1-DIYL)DIPROPANE-2-SULFONAMIDE",
            "non-polymer",
        ),
        "VIT": ("VITAMIN E", "non-polymer"),
        "UD0": ("?", "non-polymer"),
        "SG4": ("3,4-DI-O-ACETYL-6-O-SULFAMOYL-ALPHA-D-GLUCOPYRANOSE", "saccharide"),
        "SG7": ("3,4-DI-O-ACETYL-6-O-SULFAMOYL-ALPHA-D-MANNOPYRANOSE", "saccharide"),
        "TYU": ("TETRAHYDROURIDINE", "nucleic-acid"),
        "SG6": ("4-O-ACETYL-6-O-SULFAMOYL-ALPHA-D-GALACTOPYRANOSE", "saccharide"),
        "AXA": (
            "(5S)-5-(2-AMINO-2-OXOETHYL)-4-OXO-N-[(3-OXO-3,4-DIHYDRO-2H-1,4-BENZOXAZIN-6-YL)METHYL]-3,4,5,6,7,8-HEXAHYDRO[1]BENZOTHIENO[2,3-D]PYRIMIDINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "VOL": ("L-VALINOL", "peptide"),
        "UD1": ("URIDINE-DIPHOSPHATE-N-ACETYLGLUCOSAMINE", "non-polymer"),
        "ICY": ("S-(2-IODOBENZYL)-L-CYSTEINE", "peptide"),
        "ICX": (
            "METHYL N-[(5Z)-6-({[4-(4-IODOBENZYL)PHENYL]CARBONYL}AMINO)HEX-5-ENOYL]GLYCINATE",
            "non-polymer",
        ),
        "VOR": (
            "(2R,3S)-2-(2,4-DIFLUOROPHENYL)-3-(5-FLUOROPYRIMIDIN-4-YL)-1-(1H-1,2,4-TRIAZOL-1-YL)BUTAN-2-OL",
            "non-polymer",
        ),
        "WBU": ("5-AMINO-1H-PYRIMIDINE-2,4-DIONE", "non-polymer"),
        "AXF": ("?", "non-polymer"),
        "E37": (
            "N~2~-[4-(4-PHENYLTHIOPHEN-2-YL)BENZOYL]-L-ALPHA-GLUTAMINE",
            "non-polymer",
        ),
        "E36": (
            "2-CHLORO-5-[(6,7-DIHYDRO-1H-[1,4]DIOXINO[2,3-F]BENZIMIDAZOL-2-YLSULFANYL)ACETYL]BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "RAJ": (
            "N-{3-[3-(DIMETHYLAMINO)PROPYL]-5-(TRIFLUOROMETHYL)PHENYL}-4-METHYL-3-[(3-PYRIMIDIN-4-YLPYRIDIN-2-YL)AMINO]BENZAMIDE",
            "non-polymer",
        ),
        "SL0": (
            "5,7-DIHYDROXY-2-(4-HYDROXYPHENYL)-4-OXO-4H-CHROMEN-3-YL 3,4-DI-O-ACETYL-6-DEOXY-ALPHA-L-MANNOPYRANOSIDE",
            "non-polymer",
        ),
        "E38": (
            "5-{[(5-BROMO-1H-BENZIMIDAZOL-2-YL)SULFANYL]ACETYL}-2-CHLOROBENZENESULFONAMIDE",
            "non-polymer",
        ),
        "ICS": (
            "IRON-SULFUR-MOLYBDENUM CLUSTER WITH INTERSTITIAL CARBON",
            "non-polymer",
        ),
        "TBI": ("TAZOBACTAM TRANS-ENAMINE INTERMEDIATE", "non-polymer"),
        "ICR": (
            "(4R)-5-IMINO-1-(5-O-PHOSPHONO-BETA-D-RIBOFURANOSYL)-4,5-DIHYDRO-1H-IMIDAZOLE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "ICP": ("2-METHYL-5-METHYLENE-5H-PYRIMIDIN-4-YLIDENEAMINE", "non-polymer"),
        "SR3": (
            "(2S,3R)-4-(2-AMINO-2-OXOETHOXY)-3-(DIHYDROXY-LAMBDA~4~-SULFANYL)-3-METHYL-4-OXO-2-{[(1E)-3-OXOPROP-1-EN-1-YL]AMINO}BUTANOIC ACID",
            "non-polymer",
        ),
        "SL1": (
            "N-((1S)-4-{[AMINO(IMINO)METHYL]AMINO}-1-FORMYLBUTYL)-2-{(3R)-3-[(BENZYLSULFONYL)AMINO]-2-OXO-5-PHENYL-2,3-DIHYDRO-1H-1,4-BENZODIAZEPIN-1-YL}ACETAMIDE",
            "non-polymer",
        ),
        "ICU": ("METHYL(2-ACETOXY-2-(2-CARBOXY-4-AMINO-PHENYL))ACETATE", "non-polymer"),
        "ICT": ("ISOCITRIC ACID", "non-polymer"),
        "UD4": (
            "UDP-2-ACETAMIDO-4-AMINO-2,4,6-TRIDEOXY-ALPHA-D-GLUCOPYRANOSE",
            "non-polymer",
        ),
        "NWL": (
            "6-AMINO-1-BENZYL-5-(ETHYLAMINO)PYRIMIDINE-2,4(1H,3H)-DIONE",
            "non-polymer",
        ),
        "16T": (
            "1-[(3S)-1-{[3-(3-CHLOROPHENOXY)-4-HYDROXYPHENYL]SULFONYL}PIPERIDIN-3-YL]-5-METHYLPYRIMIDINE-2,4(1H,3H)-DIONE",
            "non-polymer",
        ),
        "16U": ("N-(3-CHLOROBENZYL)-1-(4-METHYLPENTANOYL)-L-PROLINAMIDE", "peptide"),
        "16W": (
            "3-[(3-CHLOROPHENOXY)METHYL]-1-(TETRAHYDRO-2H-PYRAN-4-YL)-1H-PYRAZOLO[3,4-D]PYRIMIDIN-4-AMINE",
            "non-polymer",
        ),
        "MAI": ("DEOXO-METHYLARGININE", "peptide"),
        "E3P": (
            "N~2~-[3-(1,1':4',1''-TERPHENYL-4-YL)PROPANOYL]-L-ALPHA-GLUTAMINE",
            "non-polymer",
        ),
        "16P": ("3,6,9,12,15,18-HEXAOXAICOSANE", "non-polymer"),
        "PFF": ("4-FLUORO-L-PHENYLALANINE", "peptide"),
        "E3G": ("ESTRONE BETA-D-GLUCURONIDE", "non-polymer"),
        "PFI": (
            "(6S)-6-CYCLOPENTYL-6-[2-(3-FLUORO-4-ISOPROPOXYPHENYL)ETHYL]-4-HYDROXY-5,6-DIHYDRO-2H-PYRAN-2-ONE",
            "non-polymer",
        ),
        "HGI": ("MERCURY (II) IODIDE", "non-polymer"),
        "E3O": (
            "(16ALPHA,17ALPHA)-ESTRA-1,3,5(10)-TRIENE-3,16,17-TRIOL",
            "non-polymer",
        ),
        "FUB": ("BETA-L-ARABINOFURANOSE", "saccharide"),
        "0S5": (
            "N-BENZOYL-1-[(2R)-3-CYCLOPENTYL-2-{[FORMYL(HYDROXY)AMINO]METHYL}PROPANOYL]-L-PROLINAMIDE",
            "non-polymer",
        ),
        "CIC": ("CITRYL-THIOETHER-COENZYME *A", "non-polymer"),
        "47C": ("?", "nucleic-acid"),
        "CIA": (
            "6-BENZO[1,3]DIOXOL-5-YL-2-METHYL-2,3,6,7,12,12A-HEXAHYDRO-PYRAZINO[1',2':1,6]PYRIDO[3,4-B]INDOLE-1,4-DIONE",
            "non-polymer",
        ),
        "PFJ": (
            "6,7-DIMETHOXY-4-[(3R)-3-(QUINOXALIN-2-YLOXY)PYRROLIDIN-1-YL]QUINAZOLINE",
            "non-polymer",
        ),
        "CIG": ("2-AMINO-6-CHLOROPYRAZINE", "non-polymer"),
        "47D": (
            "{3-[(5-CHLORO-1,3-BENZOTHIAZOL-2-YL)METHYL]-2,4-DIOXO-3,4-DIHYDROPYRIMIDIN-1(2H)-YL}ACETIC ACID",
            "non-polymer",
        ),
        "CID": (
            "6-(4-CHLOROPHENYL)IMIDAZO[2,1-B][1,3]THIAZOLE-5-CARBALDEHYDE O-(3,4-DICHLOROBENZYL)OXIME",
            "non-polymer",
        ),
        "CIJ": (
            "2-(3-CHLORO-4-FLUOROBENZYL)-4,5-DIHYDROXY-1H-ISOINDOLE-1,3(2H)-DIONE",
            "non-polymer",
        ),
        "CII": (
            "N-[(1S)-2-METHYL-1-(PYRIDIN-4-YLCARBAMOYL)PROPYL]CYCLOHEXANECARBOXAMIDE",
            "non-polymer",
        ),
        "CIH": (
            "(2Z)-2-CYANO-N-(2,2'-DICHLOROBIPHENYL-4-YL)-3-HYDROXYBUT-2-ENAMIDE",
            "non-polymer",
        ),
        "47N": ("NEOAGAROBIOSE", "non-polymer"),
        "CIN": ("4-CARBOXYCINNAMIC ACID", "non-polymer"),
        "CIL": ("CILASTATIN", "non-polymer"),
        "CIS": (
            "(15Z)-N-((1S,2R,3E)-2-HYDROXY-1-{[(3-O-SULFO-BETA-D-GALACTOPYRANOSYL)OXY]METHYL}HEPTADEC-3-ENYL)TETRACOS-15-ENAMIDE",
            "non-polymer",
        ),
        "CIR": ("CITRULLINE", "peptide"),
        "CIQ": (
            "2-CYCLOHEXYL-6-METHOXY-N-[1-(1-METHYLETHYL)PIPERIDIN-4-YL]-7-(3-PYRROLIDIN-1-YLPROPOXY)QUINAZOLIN-4-AMINE",
            "non-polymer",
        ),
        "CIP": ("INOSITOL-2-METHYLENE-1,2-CYCLIC-MONOPHOSPHATE", "non-polymer"),
        "CIW": (
            "5-[(5-CHLORO-2-OXO-2,3-DIHYDRO-1H-INDOL-1-YL)METHYL]-1,3-BENZODIOXOLE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "RNG": (
            "(6,10-DIOXO-OCTAHYDRO-PYRIDAZINO[1,2-A][1,2]DIAZEPIN-1-YL)-ACETALDEHYDE FRAGMENT",
            "non-polymer",
        ),
        "CIU": ("N-CYCLOHEXYL-N'-(4-IODOPHENYL)UREA", "non-polymer"),
        "CIT": ("CITRIC ACID", "non-polymer"),
        "CO": ("COBALT (II) ION", "non-polymer"),
        "47X": ("3-(3,4-DIHYDROXYPHENYL)-7-HYDROXY-4H-CHROMEN-4-ONE", "non-polymer"),
        "PFN": ("FENOPROFEN", "non-polymer"),
        "FUL": ("BETA-L-FUCOSE", "saccharide"),
        "DAD": ("2',3'-DIDEOXYADENOSINE-5'-TRIPHOSPHATE", "non-polymer"),
        "DAE": (
            "O,P-DINITROPHENYL AMINOETHYLDIPHOSPHATE-BERYLLIUM TRIFLUORIDE",
            "non-polymer",
        ),
        "DAF": (
            "4,6-DIDEOXY-4-{[(1S,5R,6S)-3-FORMYL-5,6-DIHYDROXY-4-OXOCYCLOHEX-2-EN-1-YL]AMINO}-ALPHA-D-XYLO-HEX-5-ENOPYRANOSE",
            "saccharide",
        ),
        "DAG": ("4,6-DIDEOXY-4-AMINO-BETA-D-GLUCOPYRANOSIDE", "saccharide"),
        "M6T": ("?", "non-polymer"),
        "CM": ("CARBOXYMETHYL GROUP", "non-polymer"),
        "DAB": ("2,4-DIAMINOBUTYRIC ACID", "peptide"),
        "DAC": ("2-DECENOYL N-ACETYL CYSTEAMINE", "non-polymer"),
        "DAL": ("D-ALANINE", "peptide"),
        "DAM": ("N-METHYL-ALPHA-BETA-DEHYDROALANINE", "peptide"),
        "DAN": ("2-DEOXY-2,3-DEHYDRO-N-ACETYL-NEURAMINIC ACID", "non-polymer"),
        "CL": ("CHLORIDE ION", "non-polymer"),
        "DAH": ("3,4-DIHYDROXYPHENYLALANINE", "peptide"),
        "DAI": (
            "(3AS,9AS)-2-PENTYL-4-HYDROXYMETHYL-3A,4,9,9A-TETRAHYDRO-4,9[1',2']-BENZENO-1H-BENZ[F]ISOINDOLE-1,3(2H)-DIONE",
            "non-polymer",
        ),
        "DAJ": ("DECARBAMOYL-2,7-DIAMINOMITOSENE", "non-polymer"),
        "DAK": ("4-(N,N-DIMETHYLAMINO)CINNAMOYL-COA", "non-polymer"),
        "DAT": ("2'-DEOXYADENOSINE-5'-DIPHOSPHATE", "non-polymer"),
        "DAU": ("2'DEOXY-THYMIDINE-5'-DIPHOSPHO-ALPHA-D-GLUCOSE", "non-polymer"),
        "DAV": ("DELTA-AMINO VALERIC ACID", "non-polymer"),
        "DAW": (
            "N-[(S)-(1-{2-OXO-2-[(3-SULFANYLPROPYL)AMINO]ETHYL}-1H-1,2,3-TRIAZOL-5-YL)(PHENYL)METHYL]-4-SULFAMOYLBENZAMIDE",
            "non-polymer",
        ),
        "DAP": ("6-AMIDINE-2-(4-AMIDINO-PHENYL)INDOLE", "non-polymer"),
        "DAQ": (
            "O,P-DINITROPHENYL AMINOPROPYLDIPHOSPHATE BERYLLIUM TRIFLUORIDE",
            "non-polymer",
        ),
        "DAR": ("D-ARGININE", "peptide"),
        "DAS": ("D-ASPARTIC ACID", "peptide"),
        "DAX": (
            "5-BROMO-N[2-(DIMETHYLAMINO)ETHYL]-9-AMINOACRIDINE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "DAY": (
            "1-[(1R,2R,3AS,3BS,10AR,10BS,11S,12AS)-1,11-DIHYDROXY-2,5,10A,12A-TETRAMETHYL-7-PHENYL-1,2,3,3A,3B,7,10,10A,10B,11,12,12A-DODECAHYDROCYCLOPENTA[5,6]NAPHTHO[1,2-F]INDAZOL-1-YL]-2-HYDROXYETHANONE",
            "non-polymer",
        ),
        "DAZ": ("DIAMINOZINC", "non-polymer"),
        "M60": (
            "(4S)-5-[[(2S)-1-[2-(CARBOXYMETHYL)-2-[(3S)-3-HYDROXY-4-OXO-4-(PHENETHYLAMINO)BUTANOYL]HYDRAZINYL]-3-METHYL-1-OXO-BUTAN-2-YL]AMINO]-5-OXO-4-(PHENYLMETHOXYCARBONYLAMINO)PENTANOIC ACID",
            "peptide",
        ),
        "M61": ("?", "non-polymer"),
        "PFW": (
            "7-METHOXY-4-[(3S)-3-PHENYLPIPERIDIN-1-YL]-6-[2-(QUINOLIN-2-YL)ETHOXY]QUINAZOLINE",
            "non-polymer",
        ),
        "M66": ("?", "non-polymer"),
        "M67": (
            "(2S)-N-(4-CARBAMIMIDOYLBENZYL)-1-[(2R)-2-CYCLOHEXYL-2-{[2-OXO-2-(PROPYLAMINO)ETHYL]AMINO}ACETYL]AZETIDINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "CG": ("4-AMIDINOINDAN-1-ONE-2'-AMIDINOHYDRAZONE", "non-polymer"),
        "ULP": ("?", "non-polymer"),
        "SRO": ("SEROTONIN", "non-polymer"),
        "DA4": ("3-ETHYNYLBENZENESULFONAMIDE", "non-polymer"),
        "DA5": (
            "5-BROMO-9-AMINO-N-ETHYL(DIAMINOMETHYL)ACRIDINE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "DA6": (
            "6-BROMO-9-AMINO-N-ETHYL(DIAMINOMETHYL)ACRIDINE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "DA7": (
            "5-FLUORO-9-AMINO-(N-(2-DIMETHYLAMINO)ETHYL)ACRIDINE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "U": ("URIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "DA1": ("PYRIDINE-2,5-DIAMINE", "non-polymer"),
        "DA2": ("NG,NG-DIMETHYL-L-ARGININE", "peptide"),
        "DA3": ("?", "non-polymer"),
        "ULA": ("ULAPUALIDE A", "non-polymer"),
        "DA8": (
            "OCTYL 3-DEOXY-2-O-(6-DEOXY-ALPHA-L-GALACTOPYRANOSYL)-BETA-D-XYLO-HEXOPYRANOSIDE",
            "non-polymer",
        ),
        "DA9": ("2-AZIDO-N-(2-SULFANYLETHYL)ETHANAMIDE", "non-polymer"),
        "OOF": (
            "4-({(2S)-2-[2-(4-CHLOROPHENYL)-5,6-DIFLUORO-1H-BENZIMIDAZOL-1-YL]-2-CYCLOHEXYLACETYL}AMINO)BENZOIC ACID",
            "non-polymer",
        ),
        "OOA": ("3-OXOOCTANOIC ACID", "non-polymer"),
        "IJZ": (
            "5-AMINO-2,6-DIOXO-1,2,3,6-TETRAHYDROPYRIMIDINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "OOK": (
            "4-({(2S)-2-[2-(4-CHLOROPHENYL)-5,6-DIFLUORO-1H-BENZIMIDAZOL-1-YL]-2-CYCLOHEXYLACETYL}AMINO)-3,5-DIFLUOROBENZOIC ACID",
            "non-polymer",
        ),
        "472": (
            "(5S)-2-AMINO-3-METHYL-5-PYRIDIN-4-YL-5-(3-PYRIDIN-3-YLPHENYL)-3,5-DIHYDRO-4H-IMIDAZOL-4-ONE",
            "non-polymer",
        ),
        "CI2": ("(5BETA)-PREGNANE-3,20-DIONE", "non-polymer"),
        "470": (
            "(3AS,6AR)-5-[(4E)-OCT-4-EN-4-YL]-N,4-DIPHENYL-2,3,6,6A-TETRAHYDROPENTALEN-3A(1H)-AMINE",
            "non-polymer",
        ),
        "471": (
            "N-((2S)-2-({(1Z)-1-METHYL-3-OXO-3-[4-(TRIFLUOROMETHYL) PHENYL]PROP-1-ENYL}AMINO)-3-{4-[2-(5-METHYL-2-PHENYL-1,3-OXAZOL-4-YL)ETHOXY]PHENYL}PROPYL)PROPANAMIDE",
            "non-polymer",
        ),
        "477": (
            "3-{2-[5-(DIFLUOROMETHYL)-2H-THIENO[3,2-C]PYRAZOL-3-YL]-1H-INDOL-6-YL}PENTAN-3-OL",
            "non-polymer",
        ),
        "474": (
            "(2S,3S)-3-AMINO-4-(3,3-DIFLUOROPYRROLIDIN-1-YL)-N,N-DIMETHYL-4-OXO-2-(TRANS-4-[1,2,4]TRIAZOLO[1,5-A]PYRIDIN-6-YLCYCLOHEXYL)BUTANAMIDE",
            "non-polymer",
        ),
        "475": (
            "N-[OXO(PYRIDIN-2-YLAMINO)ACETYL]-BETA-D-GLUCOPYRANOSYLAMINE",
            "non-polymer",
        ),
        "478": (
            "{3-[(4-AMINO-BENZENESULFONYL)-ISOBUTYL-AMINO]-1-BENZYL-2-HYDROXY-PROPYL}-CARBAMIC ACID TETRAHYDRO-FURAN-3-YL ESTER",
            "non-polymer",
        ),
        "MAH": ("3-HYDROXY-3-METHYL-GLUTARIC ACID", "non-polymer"),
        "AGO": (
            "2-AMINO-9-(5-O-[(R)-HYDROXY{[(R)-HYDROXY(PHOSPHONOAMINO)PHOSPHORYL]OXY}PHOSPHORYL]-3-O-{[2-(METHYLAMINO)PHENYL]CARBONYL}-BETA-D-ERYTHRO-PENTOFURANOSYL-2-ULOSE)-1,9-DIHYDRO-6H-PURIN-6-ONE",
            "non-polymer",
        ),
        "AGN": ("?", "non-polymer"),
        "AGM": ("5-METHYL-ARGININE", "peptide"),
        "AGL": ("4,6-DIDEOXY-4-AMINO-ALPHA-D-GLUCOSE", "saccharide"),
        "MA5": (
            "2-(6-(2-CYCLOHEXYLETHOXY)-TETRAHYDRO-4,5-DIHYDROXY-2(HYDROXYMETHYL)-2H-PYRAN-3-YLOXY)-TETRAHYDRO-6(HYDROXYMETHYL)-2H-PYRAN-3,4,5-TRIOL",
            "non-polymer",
        ),
        "MA4": ("CYCLOHEXYL-HEXYL-BETA-D-MALTOSIDE", "non-polymer"),
        "AGI": ("5,7-DIHYDROXY-2-(4-HYDROXYPHENYL)-4H-CHROMEN-4-ONE", "non-polymer"),
        "AGH": (
            "N-{(1S,2R,3S)-1-[(ALPHA-D-GALACTOPYRANOSYLOXY)METHYL]-2,3-DIHYDROXYHEPTADECYL}HEXACOSANAMIDE",
            "saccharide",
        ),
        "AGG": ("TIROFIBAN", "non-polymer"),
        "AGF": (
            "O-(((1R)-((N-(PHENYL-METHOXY-CARBONYL)-ALANYL)-AMINO)METHYL)HYDROXYPHOSPHINYL)3-L-PHENYLLACTATE",
            "non-polymer",
        ),
        "AGE": (
            "(2E)-N-HYDROXY-3-[1-METHYL-4-(PHENYLACETYL)-1H-PYRROL-2-YL]PROP-2-ENAMIDE",
            "non-polymer",
        ),
        "AGD": ("3-(2-AMINO-6-OXO-1,6-DIHYDRO-9H-PURIN-9-YL)-D-ALANINE", "peptide"),
        "29B": ("4-[4-(4-FLUOROPHENYL)-1H-PYRAZOL-3-YL]PYRIDINE", "non-polymer"),
        "AGB": ("N-(1-ADAMANTYL)-N'-(4-GUANIDINOBENZYL)UREA", "non-polymer"),
        "AGA": (
            "(1S)-2-{[{[(2S)-2,3-DIHYDROXYPROPYL]OXY}(HYDROXY)PHOSPHORYL]OXY}-1-[(PENTANOYLOXY)METHYL]ETHYL OCTANOATE",
            "non-polymer",
        ),
        "29A": (
            "3-{5-CHLORO-4-[(2,4-DIFLUOROBENZYL)OXY]-6-OXOPYRIMIDIN-1(6H)-YL}-N-(2-HYDROXYETHYL)-4-METHYLBENZAMIDE",
            "non-polymer",
        ),
        "SN6": (
            "1-METHYL-4-[4-[4-(4-(1-METHYLQUINOLINIUM)AMINO)BENZAMIDO]ANILINO]PYRIDINIUM",
            "non-polymer",
        ),
        "CS": ("CESIUM ION", "non-polymer"),
        "NYH": ("1-NITROCYCLOHEXENE", "non-polymer"),
        "SN5": ("2-DEOXY-2-(ETHANETHIOYLAMINO)-BETA-D-GLUCOPYRANOSE", "non-polymer"),
        "SN2": ("5-[BIS-2(CHLORO-ETHYL)-AMINO]-2,4-DINTRO-BENZAMIDE", "non-polymer"),
        "SN3": ("CHLORODYSINOSIN", "non-polymer"),
        "AGY": (
            "2-[(4-TERT-BUTYL-3-NITROPHENYL)CARBONYL]-N-NAPHTHALEN-1-YLHYDRAZINECARBOXAMIDE",
            "non-polymer",
        ),
        "AGX": (
            "(1S)-1-(1H-BENZIMIDAZOL-2-YL)ETHYL (3,4-DICHLOROPHENYL)CARBAMATE",
            "non-polymer",
        ),
        "NYB": ("N-METHYL-S-[(1R)-1-METHYLPROPYL]-L-CYSTEINE", "peptide"),
        "NYC": (
            "[(4E)-2-[(1R,2S)-1-AMINO-2-HYDROXYPROPYL]-4-(1H-INDOL-3-YLMETHYLENE)-5-OXO-4,5-DIHYDRO-1H-IMIDAZOL-1-YL]ACETIC ACID",
            "peptide",
        ),
        "AGU": ("AMINOGUANIDINE", "non-polymer"),
        "29U": (
            "1-[(2R)-2-AMINOBUTANOYL]-N-(4-CARBAMIMIDOYLBENZYL)-L-PROLINAMIDE",
            "non-polymer",
        ),
        "AGS": ("PHOSPHOTHIOPHOSPHORIC ACID-ADENYLATE ESTER", "non-polymer"),
        "166": (
            "6-CARBAMIMIDOYL-2-[2-HYDROXY-6-(4-HYDROXY-PHENYL)-INDAN-1-YL]-HEXANOIC ACID",
            "non-polymer",
        ),
        "AGQ": (
            "3-[(3E)-3-(CARBAMIMIDOYLHYDRAZONO)-4-HYDROXY-6-OXOCYCLOHEXA-1,4-DIEN-1-YL]-L-ALANINE",
            "peptide",
        ),
        "AGP": ("2-DEOXY-2-AMINO GLUCITOL-6-PHOSPHATE", "non-polymer"),
        "BWP": (
            "(2S)-1-{4-[(4-ANILINO-5-BROMOPYRIMIDIN-2-YL)AMINO]PHENOXY}-3-(DIMETHYLAMINO)PROPAN-2-OL",
            "non-polymer",
        ),
        "CP": ("PHOSPHORIC ACID MONO(FORMAMIDE)ESTER", "non-polymer"),
        "BWD": (
            "2-AMINO-3-(5-BROMO-2,4-DIOXO-3,4-DIHYDRO-2H-PYRIMIDIN-1-YL)-PROPIONIC ACID",
            "non-polymer",
        ),
        "CU": ("COPPER (II) ION", "non-polymer"),
        "BW2": (
            "N-(4-{[(2-AMINO-4-OXO-3,4-DIHYDROQUINAZOLIN-6-YL)AMINO]SULFONYL}BENZOYL)GLUTAMIC ACID",
            "non-polymer",
        ),
        "T2T": ("?", "nucleic-acid"),
        "7DA": ("7-DEAZA-2'-DEOXYADENOSINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "K44": (
            "5,6,7,8-TETRABROMO-1-METHYL-2,3-DIHYDRO-1H-IMIDAZO[1,2-A]BENZIMIDAZOLE",
            "non-polymer",
        ),
        "RNY": (
            "N-(3-{[7-METHOXY-6-(2-PYRROLIDIN-1-YLETHOXY)QUINAZOLIN-4-YL]AMINO}-4-METHYLPHENYL)-2-MORPHOLIN-4-YLPYRIDINE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "NMT": (
            "1-(O2-(METHYLCARBAMOYL)-O5-HYDROXYPHOSPHINYL-BETA-D-RIBOFURANOSYL)THYMINE",
            "nucleic-acid",
        ),
        "TDH": ("4,4'-(1H-1,2,3-TRIAZOLE-1,5-DIYL)DIPHENOL", "non-polymer"),
        "CS4": ("S-[3-(3,4-DICHLOROPHENYL)-3-OXOPROPYL]-L-CYSTEINE", "peptide"),
        "TDO": ("DTDP-4-KETO-L-RHAMNOSE", "non-polymer"),
        "D0M": ("PENTAN-3-AMINE", "non-polymer"),
        "MAQ": ("2-AMINO-8-METHYLQUINAZOLIN-4(3H)-ONE", "non-polymer"),
        "MAP": ("MAGNESIUM-5'-ADENYLY-IMIDO-TRIPHOSPHATE", "non-polymer"),
        "MAS": ("DIMETHYLSULFAMIC ACID", "non-polymer"),
        "MAR": ("4'-EPI-4'-(2-DEOXYFUCOSE)DAUNOMYCIN", "non-polymer"),
        "MAU": ("N-METHYL KIRROMYCIN", "non-polymer"),
        "MAT": (
            "2,4-DIDEOXY-4-[2-(PROPYL)AMINO]-3-O-METHYL ALPHA-L-THREO-PENTOPYRANOSIDE",
            "saccharide",
        ),
        "MAW": ("4-DEOXY-ALPHA-L-ERYTHRO-HEX-4-ENOPYRANURONIC ACID", "saccharide"),
        "MAV": ("ALPHA-D-MANNOPYRANURONIC ACID", "saccharide"),
        "MAY": ("METHYL ARACHIDONYL FLUOROPHOSPHONATE", "non-polymer"),
        "MAX": ("MATAIRESINOL", "non-polymer"),
        "SNL": ("SPIRONOLACTONE", "non-polymer"),
        "1CT": (
            "[2-(2-OXO-2-{[2-(TRIFLUOROMETHYL)PHENYL]AMINO}ETHOXY)PHENYL]PHOSPHONIC ACID",
            "non-polymer",
        ),
        "SNH": ("?", "non-polymer"),
        "SNI": (
            "9-(3,4-DIHYDROXY-5-HYDROXYMETHYL-TETRAHYDRO-FURAN-2-YL)-1,9-DIHYDRO-PURINE-6-THIONE",
            "non-polymer",
        ),
        "MAA": ("N-METHYL-L-ALANINE", "peptide"),
        "TDC": ("5A,6-ANHYDROTETRACYCLINE", "non-polymer"),
        "MAC": ("MERCURY ACETATE ION", "non-polymer"),
        "MAB": ("MANNOBIOSE", "saccharide"),
        "MAE": ("MALEIC ACID", "non-polymer"),
        "MAD": ("6-HYDRO-1-METHYLADENOSINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "298": (
            "N-[(1E,2Z)-BUT-2-EN-1-YLIDENE]-N'-[(2E)-BUT-2-EN-1-YLIDENE]BUTANE-1,4-DIAMINE",
            "non-polymer",
        ),
        "299": ("2,6-DIFORMYL-4-METHYLPHENYL DIHYDROGEN PHOSPHATE", "non-polymer"),
        "296": (
            "(3R)-3-AMINO-2,2-DIFLUORO-3-(4-METHOXYPHENYL)PROPANOIC ACID",
            "non-polymer",
        ),
        "297": ("DODECANE-1,12-DIAMINE", "non-polymer"),
        "294": (
            "N-HYDROXY-4-({4-[4-(TRIFLUOROMETHYL)PHENOXY]PHENYL}SULFONYL)TETRAHYDRO-2H-PYRAN-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "295": (
            "(2S,3S)-3-(4-FLUOROPHENYL)-2,3-DIHYDROXYPROPANOIC ACID",
            "non-polymer",
        ),
        "292": (
            "N-(3-CYCLOPROPYL-1H-PYRAZOL-5-YL)-2-(2-NAPHTHYL)ACETAMIDE",
            "non-polymer",
        ),
        "293": (
            "(2S,4R,5S,6R)-6-((S)-1,2-DIHYDROXYETHYL)TETRAHYDRO-2H-PYRAN-2,4,5-TRIOL",
            "saccharide",
        ),
        "290": ("6-METHYL-3,9-DIHYDRO-2H-PURIN-2-ONE", "non-polymer"),
        "291": (
            "PROP-2-EN-1-YL 7-O-CARBAMOYL-L-GLYCERO-ALPHA-D-MANNO-HEPTOPYRANOSIDE",
            "saccharide",
        ),
        "EVH": ("1H-IMIDAZOLE-2-SULFONAMIDE", "non-polymer"),
        "CSM": (
            "1-{2,5-ANHYDRO-6-O-METHYL-4-[(PHOSPHONOOXY)METHYL]-ALPHA-L-MANNOFURANOSYL}PYRIMIDINE-2,4(1H,3H)-DIONE",
            "nucleic-acid",
        ),
        "EVJ": ("THIOPHENE-2-SULFONAMIDE", "non-polymer"),
        "SND": ("THIONICOTINAMIDE-ADENINE-DINUCLEOTIDE", "non-polymer"),
        "4YM": (
            "ETHYL 4-[(2R)-2-(AMINOMETHYL)MORPHOLIN-4-YL]-3-(3-CYANOPHENYL)-1H-PYRAZOLO[3,4-B]PYRIDINE-5-CARBOXYLATE",
            "non-polymer",
        ),
        "EVA": (
            "(2S,3R,4S,5R)-3,4,5-TRIHYDROXY-6-OXOPIPERIDINE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "EVD": ("1-BENZOFURAN-2-SULFONAMIDE", "non-polymer"),
        "D05": (
            "6-(3,4-DIHYDROXYBENZYL)-3-ETHYL-1-(2,4,6-TRICHLOROPHENYL)-1H-PYRAZOLO[3,4-D]PYRIMIDIN-4(5H)-ONE",
            "non-polymer",
        ),
        "EVF": ("1,3-BENZOTHIAZOLE-2-SULFONAMIDE", "non-polymer"),
        "EVG": ("FURAN-2-SULFONAMIDE", "non-polymer"),
        "TPJ": ("(5R)-5-TERT-BUTYL-L-PROLINE", "peptide"),
        "MAK": ("ALPHA-KETOMALONIC ACID", "non-polymer"),
        "CSN": ("N,4-DIHYDROXY-N-OXO-3-(SULFOOXY)BENZENAMINIUM", "non-polymer"),
        "NMX": ("NITROMETHYLDETHIA COENZYME A", "non-polymer"),
        "EVP": ("ETOPOSIDE; VP-16", "non-polymer"),
        "CSI": ("AMINO-(2-IMINO-HEXAHYDRO-PYRIMIDIN-4-YL)-ACETIC ACID", "non-polymer"),
        "EVS": ("16-EPI-VELLOSIMINE", "non-polymer"),
        "NP7": (
            "2-CHLORO-N-[2-(4-METHYLPIPERAZIN-1-YL)-5-NITROPHENYL]BENZAMIDE",
            "non-polymer",
        ),
        "CSH": (
            "[2-(2-HYDROXY-1-METHYL-ETHYL)-4-(1H-IMIDAZOL-4-YLMETHYL)-5-OXO-IMIDAZOLIDIN-1-YL]-ACETIC ACID",
            "peptide",
        ),
        "HN1": (
            "(6S,8R)-3-(2-DEOXY-5-O-PHOSPHONO-BETA-D-ERYTHRO-PENTOFURANOSYL)-8-HYDROXY-6-[(1S)-1-HYDROXYHEXYL]-4,6,7,8-TETRAHYDROPYRIMIDO[1,2-A]PURIN-10(3H)-ONE",
            "nucleic-acid",
        ),
        "LSB": (
            "(4R,5S)-3-[(2R,3S,4S)-2-ETHYL-5-[(3R)-2-ETHYL-3-[2-[(2R,3R)-2-ETHYL-3-METHYL-OXIRAN-2-YL]ETHYL]OXIRAN-2-YL]-3-HYDROXY-4-METHYL-PENTANOYL]-4-METHYL-5-PHENYL-1,3-OXAZOLIDIN-2-ONE",
            "non-polymer",
        ),
        "CSK": ("S-(METHYLSELANYL)-L-CYSTEINE", "peptide"),
        "LSA": ("1,2-BENZISOTHIAZOL-3(2H)-ONE 1,1-DIOXIDE", "non-polymer"),
        "LSO": (
            "(Z)-N~6~-(3-CARBOXY-1-{[(4-CARBOXY-2-OXOBUTYL)SULFONYL]METHYL}PROPYLIDENE)-L-LYSINE",
            "peptide",
        ),
        "CSJ": ("S-[(2-AMINOPHENYL)CARBONYL]-L-CYSTEINE", "peptide"),
        "7DR": (
            "6-ETHYL-5-[3-(5-METHOXYBIPHENYL-3-YL)PROP-1-YN-1-YL]PYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "LSW": (
            "N~2~-[(1S)-1-CARBOXY-3-PHENYLPROPYL]-L-LYSYL-L-TRYPTOPHAN",
            "non-polymer",
        ),
        "NME": ("METHYLAMINE", "non-polymer"),
        "LSR": (
            "1,3,3-TRIMETHYL-2-[(1E,3E)-3-METHYLPENTA-1,3-DIEN-1-YL]CYCLOHEXENE",
            "non-polymer",
        ),
        "LSS": ("5'-O-(L-LEUCYLSULFAMOYL)ADENOSINE", "non-polymer"),
        "LSP": (
            "(7S)-4,7-DIHYDROXY-10-OXO-3,5,9-TRIOXA-4-PHOSPHAUNDECAN-1-AMINIUM 4-OXIDE",
            "non-polymer",
        ),
        "NMB": (
            "2,7-DIBENZYL-1,1-DIOXO-3,6-BIS-PHENOXYMETHYL-[1,2,7]THIADIAZEPANE-4,5-DIOL",
            "non-polymer",
        ),
        "UA4": (
            "(2S)-6-(2,4-DIAMINO-6-ETHYLPYRIMIDIN-5-YL)-2-(3,5-DIFLUOROPHENYL)-4-(3-METHOXYPROPYL)-2-METHYL-2H-1,4-BENZOXAZIN-3(4H)-ONE",
            "non-polymer",
        ),
        "NMC": ("N-CYCLOPROPYLMETHYL GLYCINE", "peptide"),
        "CSF": ("CMP-3FNEUAC", "nucleic-acid"),
        "I2O": (
            "N-CYCLOPROPYL-4-METHYL-3-(2'-OXO-1',2'-DIHYDROSPIRO[CYCLOPENTANE-1,3'-INDOL]-6'-YL)BENZAMIDE",
            "non-polymer",
        ),
        "NMN": ("BETA-NICOTINAMIDE RIBOSE MONOPHOSPHATE", "non-polymer"),
        "LS4": (
            "4-{[(2-OXO-1,2-DIHYDRO-3H-INDOL-3-YLIDENE)METHYL]AMINO}-N-(1,3-THIAZOL-2-YL)BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "LS5": (
            "3-{[4-([AMINO(IMINO)METHYL]AMINOSULFONYL)ANILINO]METHYLENE}-2-OXO-2,3-DIHYDRO-1H-INDOLE",
            "non-polymer",
        ),
        "LS2": (
            "N-METHYL-{4-[2-(7-OXO-6,7-DIHYDRO-8H-[1,3]THIAZOLO[5,4-E]INDOL-8-YLIDENE)HYDRAZINO]PHENYL}METHANESULFONAMIDE",
            "non-polymer",
        ),
        "CSC": (
            "4-(3-ACETOXYMETHYL-2-CARBOXY-8-OXO-5-THIA-1-AZA-BICYCLO[4.2.0]OCT-2-EN-7-YLCARBAMOYL)-1-CARBOXY-BUTYL-AMMONIUM",
            "non-polymer",
        ),
        "I2A": ("HYDROXOCOBALAMIN", "non-polymer"),
        "CSB": ("CYS BOUND TO LEAD ION", "peptide"),
        "BC6": (
            "(4E,8S,9R,10E,12S,13R,14S,16R)-13,20-DIHYDROXY-14-METHOXY-4,8,10,12,16-PENTAMETHYL-3-OXO-2-AZABICYCLO[16.3.1]DOCOSA-1(22),4,10,18,20-PENTAEN-9-YL CARBAMATE",
            "non-polymer",
        ),
        "SNC": ("S-NITROSO-CYSTEINE", "peptide"),
        "HPQ": ("HOMOPHENYLALANINYLMETHANE", "peptide"),
        "I2C": (
            "5'-O-[(S)-HYDROXY{[2-HYDROXY-3,5-DIMETHYL-6-(2-OXOETHYL)PYRIDIN-4-YL]OXY}PHOSPHORYL]GUANOSINE",
            "non-polymer",
        ),
        "KKT": (
            "N-(6-AMINO-1-BENZYL-2,4-DIOXO-1,2,3,4-TETRAHYDROPYRIMIDIN-5-YL)-N-METHYLBUTANE-1-SULFONAMIDE",
            "non-polymer",
        ),
        "I2E": ("3-(1,3-BENZODIOXOL-5-YL)PROPANOIC ACID", "non-polymer"),
        "W32": (
            "4-({3-CHLORO-4-[3-(TRIFLUOROMETHYL)PHENOXY]PHENYL}AMINO)-N-[2-(METHYLSULFONYL)ETHYL]-8,9-DIHYDRO-7H-PYRIMIDO[4,5-B]AZEPINE-6-CARBOXAMIDE",
            "non-polymer",
        ),
        "QHF": (
            "N-[(2S,4S)-1-({4-[2-(3,5-DIMETHYL-1,2-OXAZOL-4-YL)ETHYL]PIPERIDIN-1-YL}SULFONYL)-4-(5-FLUOROPYRIMIDIN-2-YL)-2-METHYLPENTAN-2-YL]-N-HYDROXYFORMAMIDE",
            "non-polymer",
        ),
        "BC2": ("MACBECIN", "non-polymer"),
        "KKK": (
            "1-ACETYL-4-(4-{[(2R,4S)-2-(2,4-DICHLOROPHENYL)-2-(1H-IMIDAZOL-1-YLMETHYL)-1,3-DIOXOLAN-4-YL]METHOXY}PHENYL)PIPERAZINE",
            "non-polymer",
        ),
        "KKJ": ("(2S,3R)-2-HEXYL-3-HYDROXYNONANOIC ACID", "non-polymer"),
        "EV0": ("2-AMINO-6-PROPYLPYRIMIDIN-4(3H)-ONE", "non-polymer"),
        "EV1": ("1-(3,4-DIMETHOXYBENZYL)-6,7-DIMETHOXYISOQUINOLINE", "non-polymer"),
        "EV2": ("3-PYRROLIDIN-1-YLQUINOXALIN-2-AMINE", "non-polymer"),
        "EV3": ("3-(2-AMINO-5-CHLORO-1H-BENZIMIDAZOL-1-YL)PROPAN-1-OL", "non-polymer"),
        "EV4": (
            "4-(2-AMINO-5-CHLORO-1H-BENZIMIDAZOL-1-YL)-N-CYCLOHEXYL-N-METHYLBUTANAMIDE",
            "non-polymer",
        ),
        "EV5": (
            "(3S)-3-(2-AMINO-5-CHLORO-1H-BENZIMIDAZOL-1-YL)-N-(CYCLOHEXYLMETHYL)PENTANAMIDE",
            "non-polymer",
        ),
        "EV6": (
            "(3S)-3-(2-AMINO-5-CHLORO-1H-BENZIMIDAZOL-1-YL)-N-[(1R,3S,5R,7R)-TRICYCLO[3.3.1.1~3,7~]DEC-2-YL]PENTANAMIDE",
            "non-polymer",
        ),
        "P9M": (
            "N-CYCLOPENTYL-5-(3,5-DIMETHYLISOXAZOL-4-YL)-2-METHYLBENZENESULFONAMIDE",
            "non-polymer",
        ),
        "CTT": (
            "BETA-D-GLUCOPYRANOSYL-(1->4)-BETA-D-GLUCOPYRANOSYL-(1->4)-BETA-D-GLUCOPYRANOSYL-(1->4)-BETA-D-GLUCOPYRANOSE",
            "saccharide",
        ),
        "4TZ": (
            "4-{[(4-CYANOPHENYL)(4H-1,2,4-TRIAZOL-4-YL)AMINO]METHYL}PHENYL SULFAMATE",
            "non-polymer",
        ),
        "CTV": ("(2Z)-3,7-DIMETHYLOCTA-2,6-DIENENITRILE", "non-polymer"),
        "CTP": ("CYTIDINE-5'-TRIPHOSPHATE", "non-polymer"),
        "CTR": ("CELLOTRIOSE", "saccharide"),
        "CTS": ("CASTANOSPERMINE", "non-polymer"),
        "4TR": (
            "2-BROMO-4-{[(4-CYANOPHENYL)(4H-1,2,4-TRIAZOL-4-YL)AMINO]METHYL}PHENYL SULFAMATE",
            "non-polymer",
        ),
        "4TP": ("4-HYDROXY-L-THREONINE-5-MONOPHOSPHATE", "non-polymer"),
        "CTX": (
            "(Z)-2-[4-(1,2)-DIPHENYL-1-BUTENYL)-PHENOXY]-N,N-DIMETHYLETHANAMINE",
            "non-polymer",
        ),
        "CTY": ("CLARITHROMYCIN", "non-polymer"),
        "CTZ": ("C2-HYDROXY-COELENTERAZINE", "non-polymer"),
        "4TT": (
            "4,5,6,11-TETRAHYDRO-1H-PYRAZOLO[4',3':6,7]CYCLOHEPTA[1,2-B]INDOLE",
            "non-polymer",
        ),
        "CTD": ("3-DEAZACYTIDINE", "non-polymer"),
        "CTE": ("7-CHLOROTRYPTOPHAN", "peptide"),
        "CTF": (
            "(17BETA)-17-(CYANOMETHYL)-2-METHOXYESTRA-1(10),2,4-TRIEN-3-YL SULFAMATE",
            "non-polymer",
        ),
        "CTG": (
            "(5R,6S)-5,6-DIHYDRO-5,6-DIHYDROXYTHYMIDINE-5'-MONOPHOSPHATE",
            "nucleic-acid",
        ),
        "4TN": (
            "(2E)-3-[3-(3-ETHOXY-5,5,8,8-TETRAMETHYL-5,6,7,8-TETRAHYDRONAPHTHALEN-2-YL)-4-HYDROXYPHENYL]ACRYLIC ACID",
            "non-polymer",
        ),
        "CTB": ("N-2-THIOPHEN-2-YL-ACETAMIDE BORONIC ACID", "non-polymer"),
        "CTC": ("7-CHLOROTETRACYCLINE", "non-polymer"),
        "4TC": ("?", "non-polymer"),
        "4TB": ("4-(2-THIENYL)BUTYRIC ACID", "non-polymer"),
        "4TA": (
            "P1-(5'-ADENOSYL)P4-(5'-(2'-DEOXY-THYMIDYL))TETRAPHOSPHATE",
            "nucleic-acid",
        ),
        "CTO": ("TRIACETYLCHITOTRIOSE", "saccharide"),
        "CTH": ("4-CHLOROTHREONINE", "peptide"),
        "CTI": (
            "1,2-DIMETHOXY-12-METHYL[1,3]BENZODIOXOLO[5,6-C]PHENANTHRIDIN-12-IUM",
            "non-polymer",
        ),
        "CTJ": (
            "1-({(2R)-2-[(1R)-1-{[(2Z)-2-(2-AMINO-1,3-THIAZOL-4-YL)-2-{[(2-CARBOXYPROPAN-2-YL)OXY]IMINO}ACETYL]AMINO}-2-OXOETHYL]-4-CARBOXY-3,6-DIHYDRO-2H-1,3-THIAZIN-5-YL}METHYL)PYRIDINIUM",
            "non-polymer",
        ),
        "LOG": ("LOGNAC", "saccharide"),
        "P9C": (
            "3-({3-[(1S,4S,4AS,6S,7S,9S,9AR)-1,6-DIMETHYL-2-OXO-4-PHENYLDECAHYDRO-6,9-EPOXY-4A,7-METHANOBENZO[7]ANNULEN-1-YL]PROPANOYL}AMINO)-2,4-DIHYDROXYBENZOIC ACID",
            "non-polymer",
        ),
        "RJ1": (
            "N-[2-CHLORO-5-(TRIFLUOROMETHYL)PHENYL]IMIDODICARBONIMIDIC DIAMIDE",
            "non-polymer",
        ),
        "5JZ": (
            "N~2~-{4-[6-(3,4-DIHYDROQUINOLIN-1(2H)-YLCARBONYL)-1H-BENZIMIDAZOL-1-YL]-6-ETHOXY-1,3,5-TRIAZIN-2-YL}-3-(2,2-DIMETHYL-4H-1,3-BENZODIOXIN-6-YL)-N-METHYL-L-ALANINAMIDE",
            "non-polymer",
        ),
        "S4U": ("4-THIOURIDINE-5'-PHOSPHATE", "nucleic-acid"),
        "1SC": ("5'-O-THIOPHOSPHONOCYTIDINE", "nucleic-acid"),
        "1SB": (
            "N'-[(1E)-(4-HYDROXYPHENYL)METHYLIDENE]-2-(3-METHOXYPHENYL)ACETOHYDRAZIDE",
            "non-polymer",
        ),
        "1SA": ("5-AMINO-1,3,4-THIADIAZOLE-2-SULFONAMIDE", "non-polymer"),
        "LOI": (
            "2-(METHYLAMINO)-N-(4-METHYL-1,3-THIAZOL-2-YL)-5-[(4-METHYL-4H-1,2,4-TRIAZOL-3-YL)SULFANYL]BENZAMIDE",
            "non-polymer",
        ),
        "1SD": (
            "(2S,3R,4S,5R,6R)-3,4,5-TRIHYDROXY-6-(HYDROXYMETHYL)OXANE-2-SULFONAMIDE",
            "non-polymer",
        ),
        "1SM": (
            "METHYL 2-[({[(4,6-DIMETHYLPYRIMIDIN-2-YL)AMINO]CARBONYL}AMINO)SULFONYL]BENZOATE",
            "non-polymer",
        ),
        "1SQ": ("ISOQUINOLIN-1-AMINE", "non-polymer"),
        "1SP": ("(2S)-PENTANE-1,2-DIOL", "non-polymer"),
        "FHC": ("2-FLUORO-3-(4-HYDROXYPHENYL)-2E-PROPENEOATE", "non-polymer"),
        "1ST": (
            "(5S,6R,7R,9R)-12-HYDROXY-6-METHOXY-5-METHYL-7-(METHYLAMINO)-6,7,8,9-TETRAHYDRO-5H,14H-5,9-EPOXY-4B,9A,15-TRIAZADIBENZO[B,H]CYCLONONA[1,2,3,4-JKL]CYCLOPENTA[E]-AS-INDACEN-14-ONE",
            "non-polymer",
        ),
        "FHB": ("3-FLUORO-4-HYDROXYBENZOIC ACID", "non-polymer"),
        "ITU": ("ETHYLISOTHIOUREA", "non-polymer"),
        "CT4": (
            "1-[(2-CHLORO-1,3-THIAZOL-5-YL)METHYL]-3-METHYL-2-NITROGUANIDINE",
            "non-polymer",
        ),
        "CT5": (
            "4-[4-(2,3-DIHYDRO-1,4-BENZODIOXIN-6-YL)-3-METHYL-1H-PYRAZOL-5-YL]-6-ETHYLBENZENE-1,3-DIOL",
            "non-polymer",
        ),
        "CT6": (
            "(5Z)-5-(3-BROMOCYCLOHEXA-2,5-DIEN-1-YLIDENE)-N-(PYRIDIN-4-YLMETHYL)-1,5-DIHYDROPYRAZOLO[1,5-A]PYRIMIDIN-7-AMINE",
            "non-polymer",
        ),
        "CT7": (
            "(5-CHLOROPYRAZOLO[1,5-A]PYRIMIDIN-7-YL)-(4-METHANESULFONYLPHENYL)AMINE",
            "non-polymer",
        ),
        "CT0": ("N1-CYCLOPENTYL-N2-(THIAZOL-2-YL)OXALAMIDE", "non-polymer"),
        "CT1": (
            "(1R)-2-METHYL-1-(PHENYLMETHYL)PROPYL[(1S)-1-FORMYLPENTYL]CARBAMATE",
            "non-polymer",
        ),
        "CT2": (
            "1-(PHENYLMETHYL)CYCLOPENTYL[(1S)-1-FORMYLPENTYL]CARBAMATE",
            "non-polymer",
        ),
        "CT3": (
            "BETA-D-GLUCOPYRANOSYL-(1->4)-BETA-D-GLUCOPYRANOSYL-(1->4)-BETA-D-GLUCOPYRANOSE",
            "saccharide",
        ),
        "N30": (
            "(1R,2R,3S,4R,6S)-4,6-DIAMINO-2-[(5-AMINO-5-DEOXY-BETA-D-RIBOFURANOSYL)OXY]-3-HYDROXYCYCLOHEXYL 2-AMINO-2-DEOXY-ALPHA-D-GLUCOPYRANOSIDE",
            "non-polymer",
        ),
        "CT8": (
            "(5Z)-5-(3-BROMOCYCLOHEXA-2,5-DIEN-1-YLIDENE)-N-(PYRIDIN-4-YLMETHYL)-1,5-DIHYDRO[1,2,4]TRIAZOLO[1,5-A]PYRIMIDIN-7-AMINE",
            "non-polymer",
        ),
        "CT9": (
            "4-[5-(TRANS-4-AMINOCYCLOHEXYLAMINO)-3-ISOPROPYLPYRAZOLO[1,5-A]PYRIMIDIN-7-YLAMINO]-N,N-DIMETHYLBENZENESULFONAMIDE",
            "non-polymer",
        ),
        "ORP": ("2-DEOXY-5-PHOSPHONO-RIBOSE", "saccharide"),
        "ORQ": ("N~5~-ACETYL-L-ORNITHINE", "peptide"),
        "ORN": ("L-ORNITHINE", "peptide"),
        "ORO": ("OROTIC ACID", "non-polymer"),
        "Z93": (
            "2-{[(4R)-2,5-DIOXO-4-(4-PYRIDIN-3-YLPHENYL)IMIDAZOLIDIN-4-YL]METHYL}-6-METHOXY-1-OXO-1H-ISOINDOLIUM",
            "non-polymer",
        ),
        "ORH": (
            "3,5-DIOXO-4-(3-PIPERIDIN-1-YLPROPYL)-2-[3-(TRIFLUOROMETHYL)PHENYL]-2,3,4,5-TETRAHYDRO-1,2,4-TRIAZINE-6-CARBONITRILE",
            "non-polymer",
        ),
        "ORG": (
            "4-(CYCLOHEXYLAMINO)-6-PIPERAZIN-1-YL-1,3,5-TRIAZINE-2-CARBONITRILE",
            "non-polymer",
        ),
        "ORD": ("D-ORNITHINE", "peptide"),
        "ORE": (
            "4-(2,7-DIFLUORO-6-HYDROXY-3-OXO-3H-XANTHEN-9-YL)ISOPHTHALIC ACID",
            "non-polymer",
        ),
        "ORB": (
            "N-{(1R)-1-[4-(2-CHLORO-5-FLUOROPYRIDIN-3-YL)PHENYL]ETHYL}-3,5-DIMETHYLISOXAZOLE-4-SULFONAMIDE",
            "non-polymer",
        ),
        "ORC": (
            "(R)-N-[1-[4-(3,5-DIMETHYLISOXAZOL-4-YL)PHENYL]ETHYL]-3,5-DIMETHYLISOXAZOLE-4-SULFONAMIDE",
            "non-polymer",
        ),
        "ORA": (
            "7-HYDROXY-8-[(E)-PHENYLDIAZENYL]NAPHTHALENE-1,3-DISULFONIC ACID",
            "non-polymer",
        ),
        "D38": (
            "N~2~-[(2S)-2-{[1-(4-CARBOXYBENZYL)-1H-1,2,3-TRIAZOL-4-YL]METHYL}-3-METHYLBUTANOYL]-L-LYSINE",
            "non-polymer",
        ),
        "555": ("4-(4-HYDROXYPHENYL)-1-NAPHTHALDEHYDE OXIME", "non-polymer"),
        "D37": (
            "N-{5-[(1-CYCLOHEPTYL-1H-PYRAZOLO[3,4-D]PYRIMIDIN-6-YL)AMINO]PYRIDIN-2-YL}METHANESULFONAMIDE",
            "non-polymer",
        ),
        "557": (
            "N-[(2S)-2-{4-[4-(HYDROXYMETHYL)-3-(TRIFLUOROMETHYL)-1H-PYRAZOL-1-YL]PHENYL}PROPYL]PROPANE-2-SULFONAMIDE",
            "non-polymer",
        ),
        "D35": ("2,5-BIS{[4-(N-CYCLOHEXYLDIAMINOMETHYL)PHENYL]}FURAN", "non-polymer"),
        "D32": (
            "(2,3-DIMETHYL-4-{[2-(PROP-2-YN-1-YLOXY)-4-{[4-(TRIFLUOROMETHYL)PHENOXY]METHYL}PHENYL]SULFANYL}PHENOXY)ACETIC ACID",
            "non-polymer",
        ),
        "550": (
            "METHYL (1R,2S)-2-(HYDROXYCARBAMOYL)-1-{4-[(2-METHYLQUINOLIN-4-YL)METHOXY]BENZYL}CYCLOPROPANECARBOXYLATE",
            "non-polymer",
        ),
        "553": (
            "(7S)-2-(2-AMINOPYRIMIDIN-4-YL)-7-(2-FLUOROETHYL)-1,5,6,7-TETRAHYDRO-4H-PYRROLO[3,2-C]PYRIDIN-4-ONE",
            "non-polymer",
        ),
        "552": (
            "(6S)-1-CHLORO-3-[(4-FLUOROBENZYL)OXY]-6-(PYRROLIDIN-1-YLCARBONYL)PYRROLO[1,2-A]PYRAZIN-4(6H)-ONE",
            "non-polymer",
        ),
        "L64": (
            "1-(4-{4-MORPHOLIN-4-YL-1-[1-(PYRIDIN-3-YLCARBONYL)PIPERIDIN-4-YL]-1H-PYRAZOLO[3,4-D]PYRIMIDIN-6-YL}PHENYL)-3-PYRIDIN-4-YLUREA",
            "non-polymer",
        ),
        "F43": ("FACTOR 430", "non-polymer"),
        "CMD": ("3'-DESAMINO-3'-(3-CYANO-4-MORPHOLINYL)-DOXORUBICIN", "non-polymer"),
        "DKI": (
            "5-AMINO-3-{[4-(AMINOSULFONYL)PHENYL]AMINO}-N-(2,6-DIFLUOROPHENYL)-1H-1,2,4-TRIAZOLE-1-CARBOTHIOAMIDE",
            "non-polymer",
        ),
        "MJA": (
            "(3R,5R)-3,5-DIHYDROXY-7-[(1S,2S,6R,8S,8AR)-8-HYDROXY-2,6-DIMETHYL-1,2,6,7,8,8A-HEXAHYDRONAPHTHALEN-1-YL]HEPTANOIC ACID",
            "non-polymer",
        ),
        "2ZF": (
            "N-(TERT-BUTOXYCARBONYL)-L-ALPHA-GLUTAMYL-N-[(1R)-1-(CARBOXYCARBONYL)-3,3-DIFLUOROPROPYL]-L-LEUCINAMIDE",
            "peptide",
        ),
        "9AT": ("(2S,3R)-2-AMINO-3-HYDROXY-BUTANAMIDE", "peptide"),
        "9AP": ("9-AMINOPHENANTHRENE", "non-polymer"),
        "9AR": ("9-HYDROXY ARISTOLOCHIC ACID", "non-polymer"),
        "9AM": ("9-AMINO-2-DEOXY-2,3-DEHYDRO-N-ACETYL-NEURAMINIC ACID", "non-polymer"),
        "9AL": (
            "(2S,3S)-3-AMINO-2-HYDROXY-N-[(1R)-1-PHENYLPROPYL]BUTANAMIDE",
            "peptide",
        ),
        "2ZS": (
            "N-(MORPHOLIN-4-YLCARBONYL)-L-PHENYLALANYL-N-[(1R,2S)-1-(CYCLOHEXYLMETHYL)-2-HYDROXY-3-(1-METHYLETHOXY)-3-OXOPROPYL]-S-METHYL-L-CYSTEINAMIDE",
            "peptide",
        ),
        "9AD": (
            "9-AMINO-(N-(2-DIMETHYLAMINO)ETHYL)ACRIDINE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "9AC": ("9-ACRIDINECARBONYL", "non-polymer"),
        "SNR": ("NORBIOTIN", "non-polymer"),
        "X1N": (
            "2-METHYL-N-[(1Z)-3-NITROCYCLOHEXA-2,4-DIEN-1-YLIDENE]-4,5-DIHYDRO[1,3]THIAZOLO[4,5-H]QUINAZOLIN-8-AMINE",
            "non-polymer",
        ),
        "L66": (
            "7-(5-HYDROXY-2-METHYLPHENYL)-8-(2-METHOXYPHENYL)-1-METHYL-1H-IMIDAZO[2,1-F]PURINE-2,4(3H,8H)-DIONE",
            "non-polymer",
        ),
        "RNT": ("L-RHAMNITOL", "saccharide"),
        "2Z5": (
            "6-AMMONIO-N-(TRIFLUOROACETYL)-L-NORLEUCYL-N-[4-(TRIFLUOROMETHYL)PHENYL]-L-ALANINAMIDE",
            "peptide",
        ),
        "2Z4": (
            "BENZYL [(1R)-1-({(1S,2S,3S)-1-BENZYL-2-HYDROXY-4-({(1S)-1-[(2-HYDROXY-4-METHOXYBENZYL)CARBAMOYL]-2-METHYLPROPYL}AMINO)-3-[(4-METHOXYBENZYL)AMINO]-4-OXOBUTYL}CARBAMOYL)-2,2-DIMETHYLPROPYL]CARBAMATE",
            "peptide",
        ),
        "2Z3": (
            "N-(MORPHOLIN-4-YLCARBONYL)-L-PHENYLALANYL-N-[(1R)-1-(CYCLOHEXYLMETHYL)-3,3-DIFLUORO-2,2-DIHYDROXY-4-(METHYLAMINO)-4-OXOBUTYL]-L-NORLEUCINAMIDE",
            "peptide",
        ),
        "2Z0": (
            "N-{[5-(DIMETHYLAMINO)NAPHTHALEN-2-YL]SULFONYL}-L-ALPHA-GLUTAMYL-N-[(1S)-4-{[AMINO(IMINIO)METHYL]AMINO}-1-(CHLOROACETYL)BUTYL]GLYCINAMIDE",
            "peptide",
        ),
        "D3Z": (
            "1-(5-DEOXY-BETA-L-XYLOFURANOSYL)PYRIMIDINE-2,4(1H,3H)-DIONE",
            "non-polymer",
        ),
        "D3X": (
            "N-HYDROXY-2-{[(4-METHOXYPHENYL)SULFONYL](2-{[(2R,3R,4S,5S,6R)-3,4,5-TRIHYDROXY-6-(HYDROXYMETHYL)TETRAHYDRO-2H-PYRAN-2-YL]OXY}ETHYL)AMINO}ACETAMIDE",
            "non-polymer",
        ),
        "S45": (
            "(3R)-3-[3-(4-CHLOROPHENYL)-1,2,4-OXADIAZOL-5-YL]PIPERIDINE",
            "non-polymer",
        ),
        "D3T": ("2',3'-DIDEOXY-THYMIDINE-5'-TRIPHOSPHATE", "nucleic-acid"),
        "55V": (
            "6-METHYL-5-[3-METHYL-3-(3,4,5-TRIMETHOXYPHENYL)BUT-1-YN-1-YL]PYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "D3P": ("(2R)-AMINO(3,5-DIHYDROXYPHENYL)ACETIC ACID", "peptide"),
        "D3O": ("TRIDEUTERIOOXIDANIUM", "non-polymer"),
        "D3M": ("3,6-DICHLORO-2-METHOXYBENZOIC ACID", "non-polymer"),
        "55E": (
            "4-(4-HYDROXY-3-METHYLPHENYL)-6-PHENYLPYRIMIDIN-2(5H)-ONE",
            "non-polymer",
        ),
        "D3G": ("(5-(PYRIDIN-3-YL)FURAN-2-YL)METHANAMINE", "non-polymer"),
        "55F": (
            "N-[4-(5-FLUORO-6-METHYLPYRIDIN-2-YL)-5-QUINOXALIN-6-YL-1H-IMIDAZOL-2-YL]ACETAMIDE",
            "non-polymer",
        ),
        "55A": (
            "N~2~-ACETYL-N-[(2S,3R)-4-{(1,3-BENZOTHIAZOL-6-YLSULFONYL)[(2S)-2-METHYLBUTYL]AMINO}-3-HYDROXY-1-PHENYLBUTAN-2-YL]-L-LEUCINAMIDE",
            "non-polymer",
        ),
        "D3A": (
            "(1-{[(4R,6S)-8-CHLORO-6-(2,3-DIMETHOXYPHENYL)-4H,6H-PYRROLO[1,2-A][4,1]BENZOXAZEPIN-4-YL]ACETYL}PIPERIDIN-4-YL)ACETIC ACID",
            "non-polymer",
        ),
        "C78": (
            "[1-[1-(6-CARBAMOYL-CYCLOHEX-2-ENYLCARBAMOYL)-CYCLOHEXYLCARBAMOYL]-2-(4-PHOSPHONOOXY-PHENYL)- ETHYL]-CARBAMIC ACID 3-AMINOBENZYLESTER",
            "non-polymer",
        ),
        "37A": (
            "(1S,5S,7R)-N~7~-(BIPHENYL-4-YLMETHYL)-N~3~-HYDROXY-6,8-DIOXA-3-AZABICYCLO[3.2.1]OCTANE-3,7-DICARBOXAMIDE",
            "non-polymer",
        ),
        "37D": (
            "METHYL 5-FURAN-2-YL-3-METHYL-1H-PYRAZOLE-4-CARBOXYLATE",
            "non-polymer",
        ),
        "37G": (
            "6-(4-{[3-(2,6-DICHLOROPHENYL)-5-(PROPAN-2-YL)-1,2-OXAZOL-4-YL]METHOXY}PHENYL)-1H-INDOLE-3-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "KAW": (
            "N-{(5S)-4,4-DIHYDROXY-6-PHENYL-5-[(PHENYLCARBONYL)AMINO]HEXANOYL}-L-TRYPTOPHAN",
            "non-polymer",
        ),
        "37U": ("D-PHENYLALANYL-N-(3-FLUOROBENZYL)-L-PROLINAMIDE", "peptide"),
        "37T": ("THEOBROMINE", "non-polymer"),
        "6CA": ("?", "non-polymer"),
        "RB0": ("D-RIBITOL", "non-polymer"),
        "6CF": (
            "2'-DEOXY-5'-O-[(S)-{DIFLUORO[(S)-HYDROXY(PHOSPHONOOXY)PHOSPHORYL]METHYL}(HYDROXY)PHOSPHORYL]CYTIDINE",
            "non-polymer",
        ),
        "6CL": ("6-CARBOXYLYSINE", "peptide"),
        "6CM": (
            "N-((E,2S,3R)-1,3-DIHYDROXYOCTADEC-4-EN-2-YL)HEXANAMIDE",
            "non-polymer",
        ),
        "6CN": ("?", "non-polymer"),
        "6CP": ("?", "non-polymer"),
        "C71": (
            "N-[4-(AMINOMETHYL)-1,1-DIOXIDOTETRAHYDRO-2H-THIOPYRAN-4-YL]-3-(1-METHYLCYCLOPENTYL)-N~2~-[(1E)-N-(PHENYLSULFONYL)ETHANIMIDOYL]-L-ALANINAMIDE",
            "non-polymer",
        ),
        "6CR": ("6-CHLORO-9-(BETA-D-RIBOFURANOSYL)-9H-PURIN-2-AMINE", "non-polymer"),
        "6CS": (
            "(4S,5S)-5-HYDROXY-2-METHYL-1,4,5,6-TETRAHYDROPYRIMIDINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "6CT": (
            "PHOSPHORIC ACID MONO-[5-HYDROXYMETHYL-2-METHYL-3-THYMINYL-CYCLOPENTYLMETHYL]ESTER GROUP",
            "nucleic-acid",
        ),
        "L0I": (
            "4-[(2-{4-[(CYCLOPROPYLCARBAMOYL)AMINO]-1H-PYRAZOL-3-YL}-1H-BENZIMIDAZOL-6-YL)METHYL]MORPHOLIN-4-IUM",
            "non-polymer",
        ),
        "6CW": ("6-CHLORO-L-TRYPTOPHAN", "peptide"),
        "L0G": (
            "2-{4-[(CYCLOPROPYLCARBAMOYL)AMINO]-1H-PYRAZOL-3-YL}-6-(MORPHOLIN-4-IUM-4-YLMETHYL)-1H-3,1-BENZIMIDAZOL-3-IUM",
            "non-polymer",
        ),
        "L0F": ("N-[3-(1H-BENZIMIDAZOL-2-YL)-1H-PYRAZOL-4-YL]BENZAMIDE", "non-polymer"),
        "L0E": (
            "4-[(2-{4-[(PHENYLCARBAMOYL)AMINO]-1H-PYRAZOL-3-YL}-1H-BENZIMIDAZOL-5-YL)METHYL]MORPHOLIN-4-IUM",
            "non-polymer",
        ),
        "L0D": ("2-(1H-PYRAZOL-3-YL)-1H-BENZIMIDAZOLE", "non-polymer"),
        "L0C": (
            "4-{[2-(4-{[(4-FLUOROPHENYL)CARBONYL]AMINO}-1H-PYRAZOL-3-YL)-1H-BENZIMIDAZOL-6-YL]METHYL}MORPHOLIN-4-IUM",
            "non-polymer",
        ),
        "C73": (
            "2-(CARBAMOYLAMINO)-5-(4-CHLOROPHENYL)-N-[(3S)-PIPERIDIN-3-YL]THIOPHENE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "NPM": (
            "N-[(1E)-PYRIDIN-2-YLMETHYLENE]-N-[4-(4-{[(1E)-PYRIDIN-2-YLMETHYLENE]AMINO}BENZYL)PHENYL]AMINE",
            "non-polymer",
        ),
        "ZIT": ("AZITHROMYCIN", "non-polymer"),
        "ZIR": ("N-(3-METHYLBUT-2-EN-1-YL)ADENOSINE", "non-polymer"),
        "L09": (
            "N-(3-TERT-BUTYL-1H-PYRAZOL-5-YL)-N'-{4-CHLORO-3-[(PYRIDIN-3-YLOXY)METHYL]PHENYL}UREA",
            "non-polymer",
        ),
        "L08": (
            "1-ACETYL-4-(4-{4-[(2-ETHOXYPHENYL)THIO]-3-NITROPHENYL}PYRIDIN-2-YL)PIPERAZINE",
            "non-polymer",
        ),
        "L07": ("2-PROPYL[1,3]THIAZOLO[4,5-C]QUINOLIN-4-AMINE", "non-polymer"),
        "454": (
            "8,8-DIPHENYL-2,3,4,8-TETRAHYDROIMIDAZO[1,5-A]PYRIMIDIN-6-AMINE",
            "non-polymer",
        ),
        "L05": (
            "1-BENZYL-3-(4-METHOXYPHENYLAMINO)-4-PHENYLPYRROLE-2,5-DIONE",
            "non-polymer",
        ),
        "L04": ("L004", "non-polymer"),
        "L03": ("3-(4-CHLOROPHENYL)-5-(METHYLTHIO)-4H-1,2,4-TRIAZOLE", "non-polymer"),
        "L02": ("4-[3-(4-CHLOROPHENYL)-1H-PYRAZOL-5-YL]PIPERIDINE", "non-polymer"),
        "L01": (
            "3-[({(1S,2R)-1-BENZYL-2-HYDROXY-3-[(3-METHOXYBENZYL)AMINO]PROPYL}AMINO)(HYDROXY)METHYL]-N,N-DIPROPYLBENZAMIDE",
            "non-polymer",
        ),
        "L00": (
            "(2R)-2-(5-{3-CHLORO-6-((2-METHOXYETHYL){[(1S,2S)-2-METHYLCYCLOPROPYL]METHYL}AMINO)-2-[METHYL(METHYLSULFONYL)AMINO]PYRIDIN-4-YL}-1,3,4-OXADIAZOL-2-YL)-1-PHENYLPROPAN-2-AMINE",
            "non-polymer",
        ),
        "ZIG": (
            "2,4-DIFLUORO-N-[2-METHOXY-5-(4-PYRIDAZIN-4-YLQUINOLIN-6-YL)PYRIDIN-3-YL]BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "ZID": ("ISONICOTINIC-ACETYL-NICOTINAMIDE-ADENINE DINUCLEOTIDE", "non-polymer"),
        "6C3": (
            "6-CHLORO-3-(3-METHYLISOXAZOL-5-YL)-4-PHENYLQUINOLIN-2(1H)-ONE",
            "non-polymer",
        ),
        "SLT": (
            "5-(ACETYLAMINO)-3,5-DIDEOXYNON-2-ULOPYRANONOSYL-(2->3)-BETA-D-LYXO-HEXOPYRANOSYL-(1->4)HEXOPYRANOSE",
            "non-polymer",
        ),
        "6C6": ("?", "non-polymer"),
        "ZIN": ("7-(PYRIDIN-2-YLMETHYL)QUINOLIN-8-OL", "non-polymer"),
        "DAA": (
            "8-AMINO-7-CARBOXYAMINO-NONANOIC ACID WITH ALUMINUM FLUORIDE",
            "non-polymer",
        ),
        "3ON": ("(3R)-3-HYDROXY-8'-APOCAROTENOL", "non-polymer"),
        "NPH": ("CYSTEINE-METHYLENE-CARBAMOYL-1,10-PHENANTHROLINE", "peptide"),
        "NPK": (
            "5-[(3AS,4R,6AR)-1-{[(1R)-1-(6-NITRO-1,3-BENZODIOXOL-5-YL)ETHOXY]CARBONYL}-2-OXOHEXAHYDRO-1H-THIENO[3,4-D]IMIDAZOL-4-YL]PENTANOIC ACID",
            "non-polymer",
        ),
        "OM2": (
            "(R)-[2-[[(2S)-BUTAN-2-YL]CARBAMOYL]PHENYL]METHYL-[(4-CARBOXY-1,3-BENZODIOXOL-5-YL)METHYL]-METHYL-AZANIUM",
            "non-polymer",
        ),
        "NPJ": (
            "4-NITROPHENYL 4-O-ALPHA-D-GLUCOPYRANOSYL-ALPHA-D-GALACTOPYRANOSIDE",
            "non-polymer",
        ),
        "SLU": ("?", "non-polymer"),
        "PVE": (
            "(1S)-1-CARBOXY-5-[(3-CARBOXYPROPANOYL)AMINO]-8,9-DIHYDROXY-1,2,3,4-TETRAHYDROPYRIMIDO[1,2-A]QUINOLIN-11-IUM",
            "non-polymer",
        ),
        "371": (
            "2-(3,4-DIHYDROXYPHENYL)-8-(1,1-DIOXIDOISOTHIAZOLIDIN-2-YL)-3-HYDROXY-6-METHYL-4H-CHROMEN-4-ONE",
            "non-polymer",
        ),
        "373": (
            "3-(5-{[4-(AMINOMETHYL)PIPERIDIN-1-YL]METHYL}-1H-INDOL-2-YL)QUINOLIN-2(1H)-ONE",
            "non-polymer",
        ),
        "NPG": ("N-SUCCINYL PHENYLGLYCINE", "non-polymer"),
        "376": (
            "N-[(NAPHTHALEN-1-YLAMINO)(OXO)ACETYL]-BETA-D-GLUCOPYRANOSYLAMINE",
            "non-polymer",
        ),
        "NPF": ("ORTHONITROPHENYL-BETA-D-FUCOPYRANOSIDE", "non-polymer"),
        "NPA": ("2-(4-HYDROXY-3-NITROPHENYL)ACETIC ACID", "non-polymer"),
        "YSA": ("5'-O-[N-(L-TYROSYL)SULFAMOYL]ADENOSINE", "non-polymer"),
        "YSD": (
            "(1R,3R,7E,17BETA)-17-[(2R)-5-ETHYL-5-HYDROXYHEPTAN-2-YL]-2-METHYLIDENE-9,10-SECOESTRA-5,7-DIENE-1,3-DIOL",
            "non-polymer",
        ),
        "XC2": (
            "(3R)-8-CYCLOPROPYL-6-(MORPHOLIN-4-YLMETHYL)-7-(1-NAPHTHYLMETHYL)-5-OXO-2,3-DIHYDRO-5H-[1,3]THIAZOLO[3,2-A]PYRIDINE-3-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "NPC": (
            "4-HYDROXY-3-NITROPHENYLACETYL-EPSILON-AMINOCAPROIC ACID ANION",
            "non-polymer",
        ),
        "NPB": ("3-NITROPHENYLBORONIC ACID", "non-polymer"),
        "1N5": ("(2R,4S)-2,4,7-TRIHYDROXYHEPTANOIC ACID", "non-polymer"),
        "1N1": (
            "N-(2-CHLORO-6-METHYLPHENYL)-2-({6-[4-(2-HYDROXYETHYL)PIPERAZIN-1-YL]-2-METHYLPYRIMIDIN-4-YL}AMINO)-1,3-THIAZOLE-5-CARBOXAMIDE",
            "non-polymer",
        ),
        "1N8": (
            "N-{2-[(N,N-DIETHYLGLYCYL)AMINO]-5-(TRIFLUOROMETHYL)PHENYL}-4-METHYL-3-[2-(METHYLAMINO)QUINAZOLIN-6-YL]BENZAMIDE",
            "non-polymer",
        ),
        "1N9": (
            "2,6-DIMETHYLPHENYL 2-(3,5-DIMETHOXY-4-(3-(4-METHYLPIPERAZIN-1-YL)PROPOXY)PHENYLAMINO)PYRIMIDIN- 4-YL(2,4-DIMETHOXYPHENYL)CARBAMATE",
            "non-polymer",
        ),
        "PVG": ("PHYCOVIOLOBILIN, GREEN LIGHT-ABSORBING FORM", "non-polymer"),
        "VI": ("1,3-DIPHENYL-1H-PYRAZOLE-4,5-DICARBOXYLIC ACID", "non-polymer"),
        "GSL": (
            "(2S,3R)-3-HYDROXY-2-(TETRADECANOYLAMINO)OCTADECYL ALPHA-D-GALACTOPYRANOSIDURONIC ACID",
            "non-polymer",
        ),
        "SNS": (
            "3,6-BIS(3-(3'-(S)-FLUOROPYRROLINDINO)PROPIONAMIDO)ACRIDINE",
            "non-polymer",
        ),
        "1NT": (
            "(2R,4R,5R,6R,7R)-2,4,5,6,7-PENTAHYDROXY-2,8-BIS(PHOSPHONOOXY)OCTANOIC ACID",
            "non-polymer",
        ),
        "1NP": ("1-NAPHTHOL", "non-polymer"),
        "S47": (
            "(3S,4S)-PYRROLIDINE-3,4-DIYL BIS(NAPHTHALEN-1-YLACETATE)",
            "non-polymer",
        ),
        "UN1": ("2-AMINOHEXANEDIOIC ACID", "peptide"),
        "NPZ": (
            "1-METHYL-3-NAPHTHALEN-2-YL-1H-PYRAZOLO[3,4-D]PYRIMIDIN-4-AMINE",
            "non-polymer",
        ),
        "UN2": ("2-AMINO-4,4-DIHYDROXYBUTANOIC ACID", "peptide"),
        "1ND": (
            "1-[4-(1-PYRROLIDINYLCARBONYL)PHENYL]-3-(TRIFLUOROMETHYL)-4,5,6,7-TETRAHYDRO-1H-INDAZOLE",
            "non-polymer",
        ),
        "1NE": (
            "N,N-DIMETHYL-4-[3-(TRIFLUOROMETHYL)-4,5,6,7-TETRAHYDRO-1H-INDAZOL-1-YL]BENZAMIDE",
            "non-polymer",
        ),
        "1NF": (
            "N,N-DIMETHYL-4-[3-(TRIFLUOROMETHYL)-4,5,6,7-TETRAHYDRO-1H-INDAZOL-1-YL]BENZAMIDE",
            "non-polymer",
        ),
        "1NG": (
            "1-{4-[2-OXO-2-(1-PYRROLIDINYL)ETHYL]PHENYL}-3-( TRIFLUOROMETHYL)-4,5,6,7-TETRAHYDRO-1H-INDAZOLE",
            "non-polymer",
        ),
        "1NA": ("N-ACETYL-O-METHYL-D-GLUCOSAMINE", "saccharide"),
        "1NB": ("2-(2-HYDROXY-CYCLOPENTYL)-PENT-4-ENAL", "non-polymer"),
        "1NH": (
            "(2E,6E)-8-[(3-BENZOYLBENZYL)OXY]-3,7-DIMETHYLOCTA-2,6-DIENYL TRIHYDROGEN DIPHOSPHATE",
            "non-polymer",
        ),
        "1NI": (
            "4-[2-(2-ACETYLAMINO-3-NAPHTHALEN-1-YL-PROPIONYLAMINO)-3-METHYL-BUTYRYLAMINO]-3-HYDROXY-6-METHYL-HEPTANOIC ACID [1-(1-CARBAMOYL-2-NAPHTHALEN-1-YL-ETHYLCARBAMOYL)-PROPYL]-AMIDE",
            "non-polymer",
        ),
        "UN5": (
            "{3(S)-METHYLCARBAMOYL-2-[3-(3-SULFOAMINO-PHENYL)-PROPIONYL]-1,2,3,4-TETRAHYDRO-ISOQUINOLIN-7-YL}-SULFAMIC ACID",
            "non-polymer",
        ),
        "XCC": ("FE(4)-NI(1)-S(4) CLUSTER", "non-polymer"),
        "NPV": (
            "4-[8-(3-NITROPHENYL)-1,7-NAPHTHYRIDIN-6-YL]BENZOIC ACID",
            "non-polymer",
        ),
        "XCF": (
            "5-[[(2R)-2-CYCLOPROPYL-7,8-DIMETHOXY-2H-CHROMEN-5-YL]METHYL]PYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "UN6": (
            "(3-{[3-(3-SULFOAMINO-PHENYL)-PROPIONYLAMINO]-METHYL}-PHENYL)-SULFAMIC ACID",
            "non-polymer",
        ),
        "H3S": ("HEPARIN DISACCHARIDE III-S", "non-polymer"),
        "MAO": ("?", "non-polymer"),
        "XCN": ("S-CYANO-L-CYSTEINE", "peptide"),
        "XCL": (
            "[(1S,4R,6R)-6-HYDROXY-4-(CYTOSIN-9-YL)CYCLOHEX-2-EN-1-YL]METHYL DIHYDROGEN PHOSPHATE",
            "nucleic-acid",
        ),
        "UFG": (
            "URIDINE-5'-DIPHOSPHATE-4-DEOXY-4-FLUORO-ALPHA-D-GALACTOSE",
            "non-polymer",
        ),
        "XCR": (
            "[(1R,4S,6S)-4-(4-AMINO-2-OXOPYRIMIDIN-1(2H)-YL)-6-HYDROXYCYCLOHEX-2-EN-1-YL]METHYL DIHYDROGEN PHOSPHATE",
            "nucleic-acid",
        ),
        "UN8": (
            "N-(4-{[3-BUTYL-1-(2-FLUOROBENZYL)-2,6-DIOXO-2,3,6,7-TETRAHYDRO-1H-PURIN-8-YL]METHYL}PHENYL)-1-METHYL-1H-IMIDAZOLE-4-SULFONAMIDE",
            "non-polymer",
        ),
        "XCP": ("(1S,2S)-2-AMINOCYCLOPENTANECARBOXYLIC ACID", "peptide"),
        "XCT": (
            "4-AMINO-1-(2,3-DIDEOXY-6-O-PHOSPHONO-BETA-D-ERYTHRO-HEXOPYRANOSYL)PYRIMIDIN-2(1H)-ONE",
            "nucleic-acid",
        ),
        "UN9": (
            "N-[(1-CHLORO-4-HYDROXYISOQUINOLIN-3-YL)CARBONYL]GLYCINE",
            "non-polymer",
        ),
        "XNB": (
            "N~8~-(CYCLOPROPYLMETHYL)-N~4~-[2-(METHYLSULFANYL)PHENYL]-2-(PIPERAZIN-1-YL)PYRIMIDO[5,4-D]PYRIMIDINE-4,8-DIAMINE",
            "non-polymer",
        ),
        "YS2": (
            "(1R,3R,7E,17BETA)-17-[(2R)-5-HYDROXYPENTAN-2-YL]-2-METHYLIDENE-9,10-SECOESTRA-5,7-DIENE-1,3-DIOL",
            "non-polymer",
        ),
        "XCX": (
            "2-(4-{2-[1,3-BENZOXAZOL-2-YL(HEPTYL)AMINO]ETHYL}PHENOXY)-2-METHYLPROPANOIC ACID",
            "non-polymer",
        ),
        "XCY": (
            "{5-[4-{[4-(AMINOMETHYL)BENZYL]AMINO}-2-OXOPYRIMIDIN-1(2H)- YL]-3-HYDROXYTETRAHYDROFURAN-2-YL}METHYL DIHYDROGEN PHOSPHATE",
            "nucleic-acid",
        ),
        "YS5": (
            "(1R,3R,7E,17BETA)-17-[(2R,3S)-3-BUTYL-5-ETHYL-5-HYDROXYHEPTAN-2-YL]-2-METHYLIDENE-9,10-SECOESTRA-5,7-DIENE-1,3-DIOL",
            "non-polymer",
        ),
        "YS4": (
            "[({4-[4-(PROPAN-2-YLOXY)PHENYL]PYRIDIN-2-YL}AMINO)METHANEDIYL]BIS(PHOSPHONIC ACID)",
            "non-polymer",
        ),
        "TSN": ("TRICHOSTATIN A", "non-polymer"),
        "WYQ": (
            "PROPAN-2-YL {[4-ETHOXY-3-(1-METHYL-7-OXO-3-PROPYL-6,7-DIHYDRO-1H-PYRAZOLO[4,3-D]PYRIMIDIN-5-YL)PHENYL]SULFONYL}CARBAMATE",
            "non-polymer",
        ),
        "TSO": (
            "(1R,3S,5S,8R)-8-HYDROXY-2-OXABICYCLO[3.3.1]NON-6-ENE-3,5-DICARBOXYLIC ACID",
            "non-polymer",
        ),
        "TU0": ("(E)-2-(5-BROMO-2-HYDROXYPHENYL)ETHENESULFONIC ACID", "non-polymer"),
        "PVH": ("HISTIDINE-METHYL-ESTER", "peptide"),
        "WYF": (
            "(2Z)-4,6-DIHYDROXY-2-[(8-METHOXY-1,2,3,4-TETRAHYDROPYRAZINO[1,2-A]INDOL-10-YL)METHYLIDENE]-1-BENZOFURAN-3(2H)-ONE",
            "non-polymer",
        ),
        "WYE": (
            "(2Z)-4,6-DIHYDROXY-2-{[1-METHYL-4-(4-METHYLPIPERAZIN-1-YL)-1H-PYRROLO[2,3-B]PYRIDIN-3-YL]METHYLIDENE}-1-BENZOFURAN-3(2H)-ONE",
            "non-polymer",
        ),
        "R7A": (
            "S-[(4-BROMO-1-OXYL-2,2,5,5-TETRAMETHYL-2,5-DIHYDRO-1H-PYRROL-3-YL)METHYL] METHANESULFONOTHIOATE",
            "non-polymer",
        ),
        "XAR": (
            "[(1R,4S,6S)-4-(6-AMINO-9H-PURIN-9-YL)-6-HYDROXYCYCLOHEX-2-EN-1-YL]METHYL DIHYDROGEN PHOSPHATE",
            "nucleic-acid",
        ),
        "TSK": (
            "3-(5-FLUORO-6-IODO-1-METHYL-1H-INDOL-3-YL)-4-(7-METHOXY-1-BENZOFURAN-3-YL)-1H-PYRROLE-2,5-DIONE",
            "non-polymer",
        ),
        "Q8A": ("QUINOLIN-8-AMINE", "non-polymer"),
        "TSI": ("(2S,3R)-1-AMINO-2-METHYLBUTANE-2,3-DIOL", "non-polymer"),
        "TBN": (
            "'2-(4-AMINO-PYRROLO[2,3-D]PYRIMIDIN-7-YL)-5-HYDROXYMETHYL-TETRAHYDRO-FURAN-3,4-DIOL",
            "non-polymer",
        ),
        "TSD": ("3-TRIMETHYLSILYL-PROPIONATE-2,2,3,3,-D4", "non-polymer"),
        "EBD": (
            "[4-({4-[(5-CYCLOPROPYL-1H-PYRAZOL-3-YL)AMINO]-5-METHYLPYRIMIDIN-2-YL}AMINO)PHENYL]ACETONITRILE",
            "non-polymer",
        ),
        "TSC": ("(1S)-1-AMINO-2-(1H-INDOL-3-YL)ETHANOL", "non-polymer"),
        "Q82": (
            "{[4-R(-4-ALPHA,5-ALPHA,6-BETA,7-BETA)]-HEXAHYDRO-5,6-BIS(HYDROXY)-1,3-BIS(4-HYDROXYMETHYL)METHYL]-4,7-BIS(PHENYLMETHYL)-2H-1,3-DIAZEPIN-2-YLIDENE]CYANAMIDE}",
            "non-polymer",
        ),
        "Q80": (
            "2-({[(2R,4S)-2-[2-(4-CHLOROPHENYL)ETHYL]-2-(1H-IMIDAZOL-1-YLMETHYL)-1,3-DIOXOLAN-4-YL]METHYL}SULFANYL)-5-(TRIFLUOROMETHYL)PYRIDINE",
            "non-polymer",
        ),
        "EBG": (
            "2-HYDROXYMETHYL-6-(2-OXIRANYL-ETHOXY)-TETRAHYDRO-PYRAN-3,4,5-TRIOL",
            "saccharide",
        ),
        "Q86": ("4-PHENYL-1-(1H-1,2,4-TRIAZOL-1-YL)BUTAN-2-ONE", "non-polymer"),
        "Q88": ("9H-PURINE-6,8-DIAMINE", "non-polymer"),
        "LZ6": (
            "L-GAMMA-GLUTAMYL-S-(2-{[4-(3-CARBOXYPROPYL)PHENYL](2-CHLOROETHYL)AMINO}ETHYL)-L-CYSTEINYLGLYCINE",
            "non-polymer",
        ),
        "TUX": (
            "1,5-ANHYDRO-2-C-(CARBOXYMETHYL-N-HYDROXYAMIDE)-2-DEOXY-3-O-MYRISTOYL-D-GLUCITOL",
            "non-polymer",
        ),
        "WY4": (
            "7-(5,6-DIHYDRO-8H-IMIDAZO[2,1-C][1,4]OXAZIN-2-YL)-6-FORMYL-2,7-DIHYDRO- [1,4]THIAZEPINE-3-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "WY2": (
            "(6,7-DIHYDRO-5H-CYCLOPENTA[D]IMIDAZO[2,1-B]THIAZOL-2-YL]-4,7-DIHYDRO[1,4]THIAZEPINE-3,6-DICARBOXYLIC ACID",
            "non-polymer",
        ),
        "TUR": ("3-O-ALPHA-D-GLUCOPYRANOSYL-ALPHA-D-FRUCTOFURANOSE", "non-polymer"),
        "EBC": ("N,N'-ETHANE-1,2-DIYLBIS(2-IODOACETAMIDE)", "non-polymer"),
        "TUH": (
            "5-[2,4-DIHYDROXY-5-(PROPAN-2-YL)PHENYL]-4-(1-METHYL-1H-INDOL-5-YL)-2,4-DIHYDRO-3H-1,2,4-TRIAZOL-3-ONE",
            "non-polymer",
        ),
        "TUI": (
            "2-(4-(2-((3-(5-(PYRIDIN-2-YLTHIO)THIAZOL-2-YL)UREIDO)METHYL)-1H-IMIDAZOL-4-YL)PHENOXY)ACETIC ACID",
            "non-polymer",
        ),
        "TUN": ("ETHYL HYDROGEN DIETHYLAMIDOPHOSPHATE", "non-polymer"),
        "FGL": ("2-AMINOPROPANEDIOIC ACID", "peptide"),
        "TUB": (
            "(1BETA,1'ALPHA)-7',12'-DIHYDROXY-6,6'-DIMETHOXY-2,2',2'-TRIMETHYLTUBOCURARAN-2'-IUM",
            "non-polymer",
        ),
        "MPM": (
            "(1R)-1-[((1E)-{3-HYDROXY-2-METHYL-5-[(PHOSPHONOOXY)METHYL]PYRIDIN-4-YL}METHYLENE)AMINO]-1-METHYLPROPYLPHOSPHONIC ACID",
            "non-polymer",
        ),
        "RO8": (
            "N-{[(2Z)-5-BROMO-1,3-THIAZOL-2(3H)-YLIDENE]CARBAMOYL}-3-CHLOROBENZENESULFONAMIDE",
            "non-polymer",
        ),
        "O33": (
            "(4R)-N-[(1S,2R)-2-HYDROXY-2,3-DIHYDRO-1H-INDEN-1-YL]-3-[(2S,3S)-2-HYDROXY-3-({N-[(ISOQUINOLIN-5-YLOXY)ACETYL]-S-METHYL-L-CYSTEINYL}AMINO)-4-PHENYLBUTANOYL]-5,5-DIMETHYL-1,3-THIAZOLIDINE-4-CARBOXAMIDE",
            "peptide",
        ),
        "E18": (
            "2-(DIMETHYLAMINO)ETHYL (1Z)-2-PHENYL-N-(SULFOOXY)ETHANIMIDOTHIOATE",
            "non-polymer",
        ),
        "UNQ": (
            "5-{6-[(TRANS-4-AMINOCYCLOHEXYL)AMINO]PYRAZIN-2-YL}-1-BENZOFURAN-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "E11": (
            "N~4~-(1-BENZYLPIPERIDIN-4-YL)-N~2~-[3-(DIMETHYLAMINO)PROPYL]-6,7-DIMETHOXYQUINAZOLINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "E10": ("?", "non-polymer"),
        "E13": (
            "N-[(2S,3S)-4-{(4R)-4-[(2,2-DIMETHYLPROPYL)CARBAMOYL]-5,5-DIMETHYL-1,3-THIAZOLIDIN-3-YL}-3-HYDROXY-4-OXO-1-PHENYLBUTAN-2-YL]-N~2~-{(2S)-2-[(METHOXYCARBONYL)AMINO]-2-PHENYLACETYL}-3-METHYL-L-VALINAMIDE",
            "peptide",
        ),
        "E12": ("?", "non-polymer"),
        "E15": (
            "(4R)-3-[(2S,3S)-3-[[(2S)-2-[[(2S)-2-AZANYL-2-PHENYL-ETHANOYL]AMINO]-3,3-DIMETHYL-BUTANOYL]AMINO]-2-HYDROXY-4-PHENYL-BUTANOYL]-5,5-DIMETHYL-N-[(2R)-3-METHYLBUTAN-2-YL]-1,3-THIAZOLIDINE-4-CARBOXAMIDE",
            "peptide",
        ),
        "E14": (
            "(4R)-3-[(2S,3S)-3-[[(2S)-2-[[(2S)-2-AZANYL-2-PHENYL-ETHANOYL]AMINO]-3,3-DIMETHYL-BUTANOYL]AMINO]-2-HYDROXY-4-PHENYL-BUTANOYL]-5,5-DIMETHYL-N-(2-METHYLPROPYL)-1,3-THIAZOLIDINE-4-CARBOXAMIDE",
            "peptide",
        ),
        "E17": ("KNI-10729", "peptide"),
        "E16": (
            "(4R)-3-[(2S,3S)-3-[[(2S)-2-[[(2S)-2-AZANYL-2-PHENYL-ETHANOYL]AMINO]-3,3-DIMETHYL-BUTANOYL]AMINO]-2-HYDROXY-4-PHENYL-BUTANOYL]-N-[(2R)-3,3-DIMETHYLBUTAN-2-YL]-5,5-DIMETHYL-1,3-THIAZOLIDINE-4-CARBOXAMIDE",
            "peptide",
        ),
        "FGM": ("(2R,3R)-2-HEXYL-3-HYDROXYTRIDECANOIC ACID", "non-polymer"),
        "R94": (
            "N-(4-{[4-(2-HYDROXY-5-PIPERIDIN-1-YLBENZOYL)BENZOYL]AMINO}AZEPAN-3-YL)ISONICOTINAMIDE",
            "non-polymer",
        ),
        "LO2": (
            "2-{4-[BUTYL(3-CHLORO-4,5-DIMETHOXYBENZYL)AMINO]PHENYL}-1,1,1,3,3,3-HEXAFLUOROPROPAN-2-OL",
            "non-polymer",
        ),
        "W42": (
            "5-(5-(2,6-DICHLORO-4-(4,5-DIHYDRO-2-OXAZOLYL)PHENOXY)PENTYL)-3-(HYDROXYETHYL OXYMETHYLENEOXYMETHYL) ISOXAZOLE",
            "non-polymer",
        ),
        "FYA": ("ADENOSINE-5'-[PHENYLALANINOL-PHOSPHATE]", "non-polymer"),
        "LO3": ("?", "non-polymer"),
        "FYI": (
            "N-{(2S)-4-(METHYLAMINO)-4-OXO-2-[4-(PHOSPHONOOXY)BENZYL]BUTANOYL}-L-ISOLEUCYL-L-ASPARTAMIDE",
            "peptide",
        ),
        "FYN": ("?", "non-polymer"),
        "FYO": (
            "4-[5-(2,6-DIOXO-1,2,3,6-TETRAHYDROPYRIDIN-4-YL)-1H-1,2,4-TRIAZOL-3-YL]-6-OXO-1,6-DIHYDROPYRIDINE-2-CARBONITRILE",
            "non-polymer",
        ),
        "FYQ": (
            "N~2~-{(2R)-4-(METHYLAMINO)-4-OXO-2-[4-(PHOSPHONOOXY)BENZYL]BUTANOYL}-L-GLUTAMINYL-L-ASPARTAMIDE",
            "peptide",
        ),
        "FYZ": (
            "(2R,3S,4S,5S,6S)-2-(HYDROXYMETHYL)-6-[3-(4-PHENYLPHENYL)PROP-2-YNOXY]OXANE-3,4,5-TRIOL",
            "non-polymer",
        ),
        "EBP": ("DIETHYL 4-METHYLBENZYLPHOSPHONATE", "non-polymer"),
        "FYX": (
            "4-(5-PYRIDIN-4-YL-1H-1,2,4-TRIAZOL-3-YL)PYRIDINE-2-CARBONITRILE",
            "non-polymer",
        ),
        "PXL": (
            "3-HYDROXY-5-(HYDROXYMETHYL)-2-METHYLISONICOTINALDEHYDE",
            "non-polymer",
        ),
        "SNX": (
            "2-[(2-METHOXYETHYL)AMINO]-4-(4-OXO-1,2,3,4-TETRAHYDRO-9H-CARBAZOL-9-YL)BENZAMIDE",
            "non-polymer",
        ),
        "FY0": (
            "(2S)-4-(METHYLAMINO)-4-OXO-2-[4-(PHOSPHONOOXY)BENZYL]BUTANOIC ACID",
            "non-polymer",
        ),
        "FH7": ("N^5^-FORMYL-N^5^-HYDROXY-D-ORNITHINE", "peptide"),
        "QSC": ("(1R)-1-PHENYLETHANAMINE", "non-polymer"),
        "E1X": (
            "PHOSPHORIC ACID MONO-[5-(1-ETHYL-2,6-DIOXO-1,2,3,6-TETRAHYDRO-PURIN-9-YL)-3-HYDROXY-TETRAHYDRO-FURAN-2-YLMETHYL]ESTER",
            "nucleic-acid",
        ),
        "UNA": ("UNDECANAL", "non-polymer"),
        "E1P": ("ETHYL OXO(PIPERIDIN-1-YL)ACETATE", "non-polymer"),
        "E1S": (
            "N~2~-{3-[4-(5-METHYLTHIOPHEN-2-YL)PHENYL]PROPANOYL}-L-ALPHA-GLUTAMINE",
            "non-polymer",
        ),
        "SNY": ("(2E)-3-(4-HYDROXY-3,5-DIMETHOXYPHENYL)PROP-2-ENAL", "non-polymer"),
        "QSO": ("5,7-DIHYDROXY-3-(4-METHOXYPHENYL)-4H-CHROMEN-4-ONE", "non-polymer"),
        "QSN": (
            "N-[6-(1H-IMIDAZOL-1-YL)-7-NITRO-2,4-DIOXO-1,4-DIHYDROQUINAZOLIN-3(2H)-YL]METHANESULFONAMIDE",
            "non-polymer",
        ),
        "G89": (
            "(3R,3AR,5R,6AR)-3-HYDROXYHEXAHYDRO-2H-CYCLOPENTA[B]FURAN-5-YL [(2S,3R)-3-HYDROXY-4-{[(4-METHOXYPHENYL)SULFONYL](2-METHYLPROPYL)AMINO}-1-PHENYLBUTAN-2-YL]CARBAMATE",
            "peptide",
        ),
        "E1H": ("ETHANIMINE", "non-polymer"),
        "UNC": ("5-AMINO-6-NITROPYRIMIDINE-2,4(1H,3H)-DIONE", "non-polymer"),
        "UND": ("UNDECANE", "non-polymer"),
        "E1B": (
            "5-[(3-ETHYNYLPHENYL)AMINO]PYRIMIDO[4,5-C]QUINOLINE-8-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "UNE": ("5-BROMO-7-METHOXY-1-BENZOFURAN-2-CARBOXYLIC ACID", "non-polymer"),
        "CKE": (
            "1-{7-CYCLOHEXYL-6-[4-(4-METHYLPIPERAZIN-1-YL)BENZYL]-7H-PYRROLO[2,3-D]PYRIMIDIN-2-YL}METHANAMINE",
            "non-polymer",
        ),
        "PAQ": ("2-OXY-4-HYDROXY-5-(2-HYDRAZINOPYRIDINE)PHENYLALANINE", "peptide"),
        "CKG": (
            "8-(CYCLOHEXA-2,5-DIEN-1-YLIDENEAMINO)-1-(PIPERIDIN-4-YLMETHYL)-4,5-DIHYDRO-1H-PYRAZOLO[4,3-H]QUINAZOLINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "IH9": ("2-[(3-HYDROXYBENZYL)AMINO]ETHANESULFONIC ACID", "non-polymer"),
        "PAT": ("ALPHA-PHOSPHONO-TRYPTOPHAN", "peptide"),
        "41A": (
            "5,5-DIMETHYL-2-MORPHOLIN-4-YL-5,6-DIHYDRO-1,3-BENZOTHIAZOL-7(4H)-ONE",
            "non-polymer",
        ),
        "PAV": ("(2R,4S)-2-METHYL-2,3,3,4-TETRAHYDROXYTETRAHYDROFURAN", "non-polymer"),
        "CKB": (
            "1-BETA-D-GLUCOPYRANOSYL-5-METHYLPYRIMIDINE-2,4(1H,3H)-DIONE",
            "non-polymer",
        ),
        "41L": ("N-(1H-BENZIMIDAZOL-1-YL)-2-PHENYLACETAMIDE", "non-polymer"),
        "IH3": (
            "2-(2,2-DIPHENYL-ETHYL)-7-METHYL-1,3-DIOXO-2,3,5,8-TETRAHYDRO-1H-[1,2,4] TRIAZOLO[1,2-A]PYRIDAZINE-5-CARBOXYLIC ACID [4-(2-AMINO-3H-IMIDAZOL-4-YL)-CYCLOHEXYL]-AMIDE",
            "non-polymer",
        ),
        "PAZ": (
            "N-[(2R)-2-HYDROXY-3,3-DIMETHYL-4-(PHOSPHONOOXY)BUTANOYL]-BETA-ALANINE",
            "non-polymer",
        ),
        "IH1": (
            "2-[2-(4-BROMO-BENZENESULFONYL)-ETHYL]-1-3-DIOXO-2,3,5,8-TETRAHYDRO-1H-[1,2,4]TRIAZOLO[1,2-A]PYRIDAZINE-5-CARBOXYLIC ACID(4-CARBAMIMIDOYL-CYCLOHEXYLMETHYL)-AMIDE",
            "non-polymer",
        ),
        "CKI": ("N-(2-AMINOETHYL)-5-CHLOROISOQUINOLINE-8-SULFONAMIDE", "non-polymer"),
        "CKH": ("2-FLUORO-N-[2-(2-METHYL-1H-INDOL-3-YL)ETHYL]BENZAMIDE", "non-polymer"),
        "CKK": (
            "N-{[4-({5-[(4-AMINOPIPERIDIN-1-YL)METHYL]PYRROLO[2,1-F][1,2,4]TRIAZIN-4-YL}OXY)-3-FLUOROPHENYL]CARBAMOYL}-2-(4-FLUOROPHENYL)ACETAMIDE",
            "non-polymer",
        ),
        "CKJ": (
            "1-[(TRANS-4-AMINOCYCLOHEXYL)METHYL]-N-BUTYL-3-(4-FLUOROPHENYL)-1H-PYRAZOLO[3,4-D]PYRIMIDIN-6-AMINE",
            "non-polymer",
        ),
        "F23": ("TRICOSANOIC ACID", "non-polymer"),
        "F22": (
            "(4Z,7Z,10Z,13Z,16Z,19Z)-DOCOSA-4,7,10,13,16,19-HEXAENOIC ACID",
            "non-polymer",
        ),
        "F21": (
            "(3R)-3-(FLUOROMETHYL)-7-(THIOMORPHOLIN-4-YLSULFONYL)-1,2,3,4-TETRAHYDROISOQUINOLINE",
            "non-polymer",
        ),
        "F20": (
            "N-BENZOYL-L-PHENYLALANYL-4-[(5S)-1,1-DIOXIDO-3-OXOISOTHIAZOLIDIN-5-YL]-L-PHENYLALANINAMIDE",
            "non-polymer",
        ),
        "F27": (
            "6-[2-(3-BIPHENYL-4-YLPROPANOYL)-1,3-OXAZOL-5-YL]PYRIDINE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "CKP": (
            "(2R,3R,4S,5R)-2,3,4-TRIHYDROXY-5-[(PHOSPHONATOOXY)METHYL]TETRAHYDROFURAN-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "F25": (
            "4-(1R,3AS,4R,8AS,8BR)-[1-DIFLUOROMETHYL-2-(4-FLUOROBENZYL)-3-OXODECAHYDROPYRROLO[3,4-A]PYRROLIZIN-4-YL]BENZAMIDINE",
            "non-polymer",
        ),
        "CKR": (
            "6-AMINO-2-[(2-MORPHOLIN-4-YLETHYL)AMINO]-3,7-DIHYDRO-8H-IMIDAZO[4,5-G]QUINAZOLIN-8-ONE",
            "non-polymer",
        ),
        "OI7": ("1,7-DI-O-PHOSPHONO-BETA-D-ALTRO-HEPT-2-ULOFURANOSE", "non-polymer"),
        "P0G": (
            "8-[(1R)-2-{[1,1-DIMETHYL-2-(2-METHYLPHENYL)ETHYL]AMINO}-1-HYDROXYETHYL]-5-HYDROXY-2H-1,4-BENZOXAZIN-3(4H)-ONE",
            "non-polymer",
        ),
        "F29": (
            "1-ALLYL-5-(2-PHENYLPYRAZOLO[1,5-A]PYRIDIN-3-YL)-1H-PYRAZOLO[3,4-C]PYRIDAZIN-3-AMINE",
            "non-polymer",
        ),
        "PAK": ("9-CYANO PAULLONE", "non-polymer"),
        "PAL": ("N-(PHOSPHONACETYL)-L-ASPARTIC ACID", "non-polymer"),
        "PAM": ("PALMITOLEIC ACID", "non-polymer"),
        "OI1": (
            "3-(4-HYDROXYBENZYL)-2-[1-({[2-(4-HYDROXYPHENYL)ETHYL]AMINO}CARBONYL)BUTYL]-4-OXO-3,6,11,11A-TETRAHYDRO-4H-PYRAZINO[1,2-B]ISOQUINOLIN-2-IUM-1-OLATE",
            "non-polymer",
        ),
        "PAO": ("N-(PHOSPHONOACETYL)-L-ORNITHINE", "non-polymer"),
        "BDS": ("2,3-BIS-BENZO[1,3]DIOXOL-5-YLMETHYL-SUCCINIC ACID", "non-polymer"),
        "DCO": ("3,3-DICHLORO-2-PHOSPHONOMETHYL-ACRYLIC ACID", "non-polymer"),
        "DCL": ("2-AMINO-4-METHYL-PENTAN-1-OL", "peptide"),
        "DCM": ("2'-DEOXYCYTIDINE-5'-MONOPHOSPHATE", "non-polymer"),
        "DCJ": (
            "N-((1-(2,6-DICARBOXYPYRIDIN-4-YL)-1H-1,2,3-TRIAZOL-4-YL)METHYL)-2-HYDROXY-N,N-DIMETHYLETHANAMINIUM",
            "non-polymer",
        ),
        "CX6": (
            "2,3,6A,7,8,9-HEXAHYDRO-11H-[1,4]DIOXINO[2,3-G]PYRROLO[2,1-B][1,3]BENZOXAZIN-11-ONE",
            "non-polymer",
        ),
        "DCH": (
            "3-(7-DIAMINOMETHYL-NAPHTHALEN-2-YL)-PROPIONIC ACID ETHYL ESTER",
            "non-polymer",
        ),
        "DCI": ("2-METHYL-BUTYLAMINE", "non-polymer"),
        "DCF": ("2'-DEOXYCOFORMYCIN", "non-polymer"),
        "DCG": ("2'-DEOXY-GUANOSINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "DCD": ("DIETHYLCARBAMODITHIOIC ACID", "non-polymer"),
        "DCE": ("1,2-DICHLOROETHANE", "non-polymer"),
        "DCB": (
            "2,4-DINITROPHENYL-2-DEOXY-2-FLUORO-BETA-D-CELLOBIOSIDE",
            "non-polymer",
        ),
        "DCC": ("DODECYL-COA", "non-polymer"),
        "DCA": ("DESULFO-COENZYME A", "non-polymer"),
        "DCZ": ("2'-DEOXYCYTIDINE", "non-polymer"),
        "M4O": ("NERYL MOENOMYCIN A", "non-polymer"),
        "DCX": ("DECYL(DIMETHYL)PHOSPHINE OXIDE", "non-polymer"),
        "DCY": ("D-CYSTEINE", "peptide"),
        "DCV": (
            "[(1R)-3-[ACETYL(HYDROXY)AMINO]-1-(3,4-DICHLOROPHENYL)PROPYL]PHOSPHONIC ACID",
            "non-polymer",
        ),
        "DCW": ("DICYCLOHEXYLUREA", "non-polymer"),
        "DCT": ("?", "nucleic-acid"),
        "DCU": ("DECIDIUM", "non-polymer"),
        "DCR": ("ICOSANOIC ACID", "non-polymer"),
        "DCS": (
            "D-[3-HYDROXY-2-METHYL-5-PHOSPHONOOXYMETHYL-PYRIDIN-4-YLMETHYL]-N,O-CYCLOSERYLAMIDE",
            "non-polymer",
        ),
        "DCP": ("2'-DEOXYCYTIDINE-5'-TRIPHOSPHATE", "non-polymer"),
        "DCQ": (
            "2-DECYL-5,6-DIMETHOXY-3-METHYLCYCLOHEXA-2,5-DIENE-1,4-DIONE",
            "non-polymer",
        ),
        "TF4": ("DICHLORO-ACETIC ACID", "non-polymer"),
        "P0H": ("N-[(BENZYLOXY)CARBONYL]GLYCYL-L-PROLINE", "peptide"),
        "TBU": ("TERTIARY-BUTYL ALCOHOL", "non-polymer"),
        "M42": ("6-METHYL[1,3]DIOXOLO[4,5-H]QUINOLIN-8(9H)-ONE", "non-polymer"),
        "M43": ("(17[2]S)-17[2]-METHYLTHIO-COENZYME F43", "non-polymer"),
        "M41": (
            "N-(BENZYLSULFONYL)-D-VALYL-N-[2-(AMINOMETHYL)-5-CHLOROBENZYL]-L-PROLINAMIDE",
            "peptide",
        ),
        "M44": ("?", "non-polymer"),
        "RAS": ("(1R)-N-(PROP-2-EN-1-YL)-2,3-DIHYDRO-1H-INDEN-1-AMINE", "non-polymer"),
        "DC8": (
            "(3AS,4R,9BR)-2,2-DIFLUORO-4-(4-HYDROXYPHENYL)-1,2,3,3A,4,9B-HEXAHYDROCYCLOPENTA[C]CHROMEN-8-OL",
            "non-polymer",
        ),
        "OMG": ("O2'-METHYLGUANOSINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "DC4": ("2-CHLORODIBENZOFURAN", "non-polymer"),
        "DC5": ("2,6-DICHLOROBIPHENYL", "non-polymer"),
        "DC2": ("N-ACETYL-S-[(1S)-1,2-DICHLOROETHYL]-L-CYSTEINE", "peptide"),
        "DC3": (
            "(2R)-4-[4-(3-METHYLPHENYL)-1H-1,2,3-TRIAZOL-1-YL]-1-(2,4,5-TRIFLUOROPHENYL)BUTAN-2-AMINE",
            "non-polymer",
        ),
        "DC0": ("N-[(2S)-2-AMINO-4-METHYLPENTYL]-L-PHENYLALANINE", "peptide"),
        "F2C": (
            "6-[2-(7-PHENYLHEPTANOYL)-1,3-OXAZOL-5-YL]PYRIDINE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "F2B": ("N-(2,3-DIFLUORO-BENZYL)-4-SULFAMOYL-BENZAMIDE", "non-polymer"),
        "F2A": (
            "2'-DEOXY-5'-O-[(S)-HYDROXY{[(S)-HYDROXY(PHOSPHONOOXY)PHOSPHORYL]METHYL}PHOSPHORYL]ADENOSINE",
            "non-polymer",
        ),
        "OIL": ("(2S,3S)-2-HYDROXY-3-METHYLPENTANOIC ACID", "non-polymer"),
        "F2G": (
            "3-(2-{2-[(CYCLOHEXYLMETHOXY-HYDRO-METHYL)-AMINO]-1-HYDROXY-3-METHYL-BUTYLAMINO}-1-HYDROXY-PROPYLAMINO)-PENTANE-1,1,4-TRIOL",
            "non-polymer",
        ),
        "F2F": ("3,4-DIFLUORO-L-PHENYLALANINE", "peptide"),
        "PA6": ("(R)-2-(FORMYLOXY)-3-(PHOSPHONOOXY)PROPYL PENTANOATE", "non-polymer"),
        "PA7": ("N-PROPYL CARBOXYAMIDO ADENOSINE", "non-polymer"),
        "PA8": ("1,2-DIOCTANOYL-SN-GLYCERO-3-PHOSPHATE", "non-polymer"),
        "IHS": ("D-MYO-INOSITOL-HEXASULPHATE", "non-polymer"),
        "F2I": (
            "N'-{(1S,2R)-1-(3,5-DIFLUOROBENZYL)-2-HYDROXY-3-[(3-IODOBENZYL)AMINO]PROPYL}-5-METHYL-N,N-DIPROPYLISOPHTHALAMIDE",
            "non-polymer",
        ),
        "OID": ("10-(1H-IMIDAZOL-1-YL)DECANOIC ACID", "non-polymer"),
        "F2O": ("MU-OXO-DIIRON(III)", "non-polymer"),
        "F2N": (
            "1-HYDRO-5-(3-PHENYLPROPANOYL)-FLAVIN-ADENINE DINUCLEOTIDE",
            "non-polymer",
        ),
        "IHU": (
            "N-(2-CHLORO-4-FLUOROBENZOYL)-N'-(5-HYDROXY-2-METHOXYPHENYL)UREA",
            "non-polymer",
        ),
        "414": (
            "(5R,6R,8S)-8-(3-{[AMINO(IMINO)METHYL]AMINO}PHENYL)-5-CYCLOHEXYL-6-HYDROXY-3-OXO-1-PHENYL-2,7-DIOXA-4-AZA-6-PHOSPHANONAN-9-OIC ACID 6-OXIDE",
            "non-polymer",
        ),
        "415": (
            "1-{3-[(3-CHLORO-5-METHOXYBENZYL)AMINO]PROPYL}-3-PHENYLUREA",
            "non-polymer",
        ),
        "416": (
            "(2S)-7-TERT-BUTYL-6-CHLORO-2-(TRIFLUOROMETHYL)-2H-CHROMENE-3-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "417": (
            "(1S,2R,5S)-5-[3-(TRIFLUOROMETHYL)-5,6-DIHYDRO[1,2,4]TRIAZOLO[4,3-A]PYRAZIN-7(8H)-YL]-2-(2,4,5-TRIFLUOROPHENYL)CYCLOHEXANAMINE",
            "non-polymer",
        ),
        "410": (
            "4-BROMO-3-(CARBOXYMETHOXY)-5-{3-[CYCLOHEXYL(PHENYLCARBONYL)AMINO]PHENYL}THIOPHENE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "411": (
            "3-BROMO-N-[4-[1-(2-CARBAMIMIDAMIDO-2-OXO-ETHYL)-5-PHENYL-PYRROL-2-YL]PHENYL]BENZAMIDE",
            "non-polymer",
        ),
        "412": (
            "(R)-1-(4-(4-(HYDROXYMETHYL)-1,3,2-DIOXABOROLAN-2-YL)PHENETHYL)GUANIDINE",
            "non-polymer",
        ),
        "413": (
            "(S)-[(R)-2-(4-BENZYLOXY-3-METHOXY-PHENYL)-2-(4-CARBAMIMIDOYL-PHENYLAMINO)-ACETYLAMINO]-PHENYL-ACETIC ACID",
            "non-polymer",
        ),
        "IHB": ("3-IODO-4-HYDROXYBENZOIC ACID", "non-polymer"),
        "IHC": (
            "{[2-({[5-(2,6-DICHLOROPHENYL)-1,2,4-TRIAZIN-3-YL]AMINO}METHYL)-1,3-BENZOTHIAZOL-5-YL]OXY}ACETIC ACID",
            "non-polymer",
        ),
        "F2Y": ("3,5-DIFLUORO-L-TYROSINE", "peptide"),
        "IHA": (
            "{[2-({[5-(2,6-DIMETHOXYPHENYL)-1,2,4-TRIAZIN-3-YL]AMINO}METHYL)-1,3-BENZOTHIAZOL-5-YL]OXY}ACETIC ACID",
            "non-polymer",
        ),
        "418": (
            "2-{4-[2-(S)-ALLYLOXYCARBONYLAMINO-3-{4-[(2-CARBOXY-PHENYL)-OXALYL-AMINO]-PHENYL}-PROPIONYLAMINO]-BUTOXY}-6-HYDROXY-BENZOIC ACID METHYL ESTER",
            "non-polymer",
        ),
        "419": (
            "N-[(3AR,6S,6AS)-1-(N-METHYL-L-ALANYL-3-METHYL-L-VALYL)OCTAHYDROCYCLOPENTA[B]PYRROL-6-YL]-2,2-DIPHENYLACETAMIDE",
            "non-polymer",
        ),
        "IHD": ("1-IODOHEXADECANE", "non-polymer"),
        "IHE": (
            "6-(CYCLOHEXYLAMINO)-9-[2-(4-METHYLPIPERAZIN-1-YL)-ETHYL]-9H-PURINE-2-CARBONITRILE",
            "non-polymer",
        ),
        "AEP": (
            "3-[(1-AMINO-2-CARBOXY-ETHYL)-HYDROXY-PHOSPHINOYL]-2-METHYL-PROPIONIC ACID",
            "non-polymer",
        ),
        "AES": ("4-(2-AMINOETHYL)BENZENESULFONYL FLUORIDE", "non-polymer"),
        "AER": ("ABIRATERONE", "non-polymer"),
        "AET": (
            "N-[N-(9-B-D-RIBOFURANOSYLPURIN-6-YL)METHYLCARBAMOYL]THREONINE-5'-MONOPHOSPHATE",
            "nucleic-acid",
        ),
        "AEW": ("1-(3-AMINO-2-METHYLBENZYL)-4-HEXYLPYRIDIN-2(1H)-ONE", "non-polymer"),
        "GBN": ("[1-(AMINOMETHYL)CYCLOHEXYL]ACETIC ACID", "non-polymer"),
        "GBO": (
            "(2E,6E)-3,7-DIMETHYL-8-({5-[(3AS,4S,6AR)-2-OXOHEXAHYDRO-1H-THIENO[3,4-D]IMIDAZOL-4-YL]PENTANOYL}AMINO)OCTA-2,6-DIEN-1-YL TRIHYDROGEN DIPHOSPHATE",
            "non-polymer",
        ),
        "GBL": ("GAMMA-BUTYROLACTONE", "non-polymer"),
        "DRN": ("BISINDOLYLMALEIMIDE IX", "non-polymer"),
        "MO7": ("HEPTAMOLYBDATE", "non-polymer"),
        "MO6": ("MAGNESIUM ION, 6 WATERS COORDINATED", "non-polymer"),
        "GBH": (
            "2-{4-[2-(BETA-D-GLUCOPYRANOSYLOXY)ETHYL]PIPERAZIN-1-YL}ETHANESULFONIC ACID",
            "non-polymer",
        ),
        "GBI": ("S-(3-IODOBENZYL)GLUTATHIONE", "non-polymer"),
        "AEA": ("(2-AMINO-2-CARBAMOYL-ETHYLSULFANYL)-ACETIC ACID", "peptide"),
        "AEE": (
            "6-{4-[(4-ETHYLPIPERAZIN-1-YL)METHYL]PHENYL}-N-[(1R)-1-PHENYLETHYL]-7H-PYRROLO[2,3-D]PYRIMIDIN-4-AMINE",
            "non-polymer",
        ),
        "AED": ("4-(2-AMINOETHYL)-2-CYCLOHEXYLPHENOL", "non-polymer"),
        "AEG": ("4-[(2R)-2-AMINOPROPYL]PHENOL", "non-polymer"),
        "AEF": ("4-(2-AMINOETHYL)PHENOL", "non-polymer"),
        "AEI": ("THREONINE-ASPARTIC ESTER", "peptide"),
        "AEH": ("4-(2-AMINOETHYL)-2-ETHYLPHENOL", "non-polymer"),
        "AEK": (
            "4-{2-[(CYCLOPENTYLMETHYL)AMINO]ETHYL}-2-(METHYLAMINO)-3,7-DIHYDRO-8H-IMIDAZO[4,5-G]QUINAZOLIN-8-ONE",
            "non-polymer",
        ),
        "AEJ": (
            "(1S)-1-{4-[(9AR)-OCTAHYDRO-2H-PYRIDO[1,2-A]PYRAZIN-2-YL]PHENYL}-2-PHENYL-1,2,3,4-TETRAHYDROISOQUINOLIN-6-OL",
            "non-polymer",
        ),
        "AEM": ("2-AMINOETHANIMIDIC ACID", "non-polymer"),
        "P0Y": (
            "(2S,4R)-1-[(2R,3S)-3-AMINO-2-HYDROXY-4-PHENYLBUTYL]-4-[(PYRIDIN-3-YLMETHYL)SULFANYL]PIPERIDINE-2-CARBOXYLIC ACID",
            "peptide",
        ),
        "GBX": (
            "2-AMINO-4-[1-(CARBOXYMETHYL-CARBAMOYL)-2-(9-HYDROXY-7,8-DIOXO-7,8,9,10-TETRAHYDRO-BENZO[DEF]CHRYSEN-10-YLSULFANYL)-ETHYLCARBAMOYL]-BUTYRIC ACID",
            "non-polymer",
        ),
        "AEN": ("5-(1-SULFONAPHTHYL)-ACETYLAMINO-ETHYLAMINE", "non-polymer"),
        "PE2": ("9-(4-HYDROXY-3-(HYDROXYMETHYL)BUT-1-YL)GUANINE", "non-polymer"),
        "BUP": ("?", "non-polymer"),
        "BUQ": ("4-HYDROXY-2-BUTANONE", "non-polymer"),
        "BUT": ("BUTYL GROUP", "non-polymer"),
        "DRD": (
            "2-[(1-{3-[(6-BENZOYL-1-PROPYL-2-NAPHTHYL)OXY]PROPYL}-1H-INDOL-4-YL)OXY]-2-METHYLPROPANOIC ACID",
            "non-polymer",
        ),
        "BUJ": (
            "(3R)-3-(HEXADECANOYLAMINO)-4-(TRIMETHYLAMMONIO)BUTANOATE",
            "non-polymer",
        ),
        "BUK": (
            "N~2~-({[(1R)-1-(4-BROMOPHENYL)ETHYL]AMINO}CARBONYL)ASPARAGINYL-N~1~-{4-{[AMINO(IMINO)METHYL]AMINO}-1-[2,3-DIHYDRO-1,3-THIAZOL-2-YL(HYDROXY)METHYL]BUTYL}VALINAMIDE",
            "non-polymer",
        ),
        "BUI": (
            "(3R)-3-{[(TETRADECYLAMINO)CARBONYL]AMINO}-4-(TRIMETHYLAMMONIO)BUTANOATE",
            "non-polymer",
        ),
        "BUL": ("BULGECIN A", "non-polymer"),
        "BUM": ("2-AMINO-N,3,3-TRIMETHYLBUTANAMIDE", "non-polymer"),
        "BUB": ("1-BUTANE BORONIC ACID", "non-polymer"),
        "BUC": ("S,S-BUTYLTHIOCYSTEINE", "peptide"),
        "RBZ": ("ALPHA-RIBAZOLE-5'-PHOSPHATE", "non-polymer"),
        "BUA": ("BUTANOIC ACID", "non-polymer"),
        "BUG": ("3-METHYL-L-VALINE", "peptide"),
        "BUD": ("(2S,3S)-BUTANE-2,3-DIOL", "non-polymer"),
        "RVB": (
            "3-{[(1Z)-1-CARBOXY-3-METHYLBUT-1-EN-1-YL]OXY}-2-HYDROXYBENZOIC ACID",
            "non-polymer",
        ),
        "BU2": ("1,3-BUTANEDIOL", "non-polymer"),
        "BU3": ("(R,R)-2,3-BUTANEDIOL", "non-polymer"),
        "BU1": ("1,4-BUTANEDIOL", "non-polymer"),
        "BU4": ("(3R)-BUTANE-1,3-DIOL", "non-polymer"),
        "LNK": ("PENTANE", "non-polymer"),
        "BA4": (
            "[[(3R,4R,5S,6R)-3-(BUTANOYLAMINO)-4,5-DIHYDROXY-6-(HYDROXYMETHYL)OXAN-2-YLIDENE]AMINO] N-PHENYLCARBAMATE",
            "non-polymer",
        ),
        "TF3": (
            "N-(2-AMINOETHYL)-2-{3-CHLORO-4-[(4-ISOPROPYLBENZYL)OXY]PHENYL} ACETAMIDE",
            "non-polymer",
        ),
        "RBY": ("METHYLENE ADP-BETA-XYLOSE", "non-polymer"),
        "TBH": (
            "5-(2-{2-[(TERT-BUTOXY-HYDROXY-METHYL)-AMINO]-1-HYDROXY-3-PHENYL-PROPYLAMINO}-3-HYDROXY-3-PENTYLAMINO-PROPYL)-2-CARBOXYMETHOXY-BENZOIC ACID",
            "non-polymer",
        ),
        "45B": (
            "4-{[2-({4-[2-(4-ACETYLPIPERAZIN-1-YL)-2-OXOETHYL]PHENYL}AMINO)-5-FLUOROPYRIMIDIN-4-YL]AMINO}-N-(2-CHLOROPHENYL)BENZAMIDE",
            "non-polymer",
        ),
        "AE1": (
            "2-{[[(1S)-1-AMINO-2-CARBOXYETHYL](DIHYDROXY)PHOSPHORANYL]METHYL}-4-METHYLPENTANOIC ACID",
            "non-polymer",
        ),
        "AE3": ("2-(2-ETHOXYETHOXY)ETHANOL", "non-polymer"),
        "AE2": ("AETIOCHOLANOLONE", "non-polymer"),
        "AE4": ("3,6,9,12,15-PENTAOXAHEPTADECAN-1-OL", "non-polymer"),
        "RAP": ("RAPAMYCIN IMMUNOSUPPRESSANT DRUG", "non-polymer"),
        "M4Z": (
            "(2S)-N-[(4-CARBAMIMIDOYLPHENYL)METHYL]-1-[(2R)-2-CYCLOHEXYL-2-[[2-(ETHYLAMINO)-2-OXIDANYLIDENE-ETHYL]AMINO]ETHANOYL]AZETIDINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "AE9": ("2-AZIDOETHANOL", "non-polymer"),
        "MOR": ("MORPHOLINE-4-CARBOXYLIC ACID", "non-polymer"),
        "BEZ": ("BENZOIC ACID", "non-polymer"),
        "MOP": ("4-METHYLPENTANAL", "non-polymer"),
        "MOW": ("OXO(SULFANYL)MOLYBDENUM(IV) ION", "non-polymer"),
        "MOU": ("MORINIAFUNGIN", "non-polymer"),
        "MOT": (
            "N-[4-[(2,4-DIAMINOFURO[2,3D]PYRIMIDIN-5-YL)METHYL]METHYLAMINO]-BENZOYL]-L-GLUTAMATE",
            "non-polymer",
        ),
        "GB6": (
            "(3R,4R,5R)-3,4-DIHYDROXY-5-({[(1R)-2-HYDROXY-1-PHENYLETHYL]AMINO}METHYL)PYRROLIDIN-2-ONE",
            "non-polymer",
        ),
        "GB7": (
            "(3R,4R,5R)-3,4-DIHYDROXY-5-({[(1R)-2-HYDROXY-1-PHENYLETHYL]AMINO}METHYL)-1-METHYLPYRROLIDIN-2-ONE",
            "non-polymer",
        ),
        "GB4": ("(2E)-3-(2,4-DICHLOROPHENYL)-N-HYDROXYACRYLAMIDE", "non-polymer"),
        "GB5": ("(2E)-3-(4-CHLOROPHENYL)-N-HYDROXYACRYLAMIDE", "non-polymer"),
        "GB2": (
            "(2R,3R,4S)-2-({[(1S)-2-HYDROXY-1-PHENYLETHYL]AMINO}METHYL)PYRROLIDINE-3,4-DIOL",
            "non-polymer",
        ),
        "GB3": (
            "(2R,3R,4S,5R)-2-({[(1R)-2-HYDROXY-1-PHENYLETHYL]AMINO}METHYL)-5-METHYLPYRROLIDINE-3,4-DIOL",
            "non-polymer",
        ),
        "MOM": ("HYDROXY(DIOXO)MOLYBDENUM", "non-polymer"),
        "GB1": (
            "(2R,3R,4S)-2-({[(1R)-2-HYDROXY-1-PHENYLETHYL]AMINO}METHYL)PYRROLIDINE-3,4-DIOL",
            "non-polymer",
        ),
        "MOC": (
            "CARBAMIC ACID 2,6-DIAMINO-5-METHYL-4,7-DIOXO-2,3,4,7-TETRAHYDRO-1H-3A-AZA-CYCLOPENTA[A]INDEN-8-YLMETHYL ESTER",
            "non-polymer",
        ),
        "MOB": ("?", "non-polymer"),
        "MOA": ("MYCOPHENOLIC ACID", "non-polymer"),
        "MOG": ("5-AMINO-2-METHYL-6H-OXAZOLO[5,4-D]PYRIMIDIN-7-ONE", "non-polymer"),
        "MOF": ("MOMETASONE FUROATE", "non-polymer"),
        "MOE": ("METHOXY-ETHOXYL", "non-polymer"),
        "MOD": ("L-METHIONYL ADENYLATE", "peptide"),
        "ETR": ("N-ETHYL RETINAMIDE", "non-polymer"),
        "ETS": (
            "(4S-TRANS)-4-(ETHYLAMINO)-5,6-DIHYDRO-6-METHYL-4H-THIENO(2,3-B)THIOPYRAN-2-SULFONAMIDE-7,7-DIOXIDE",
            "non-polymer",
        ),
        "ETP": (
            "3-(4-BENZENESULFONYL-THIOPHENE-2-SULFONYLAMINO)-PHENYLBORONIC ACID",
            "non-polymer",
        ),
        "ETQ": (
            "3-CHLORO-5-ETHYL-N-{[(2S)-1-ETHYLPYRROLIDIN-2-YL]METHYL}-6-HYDROXY-2-METHOXYBENZAMIDE",
            "non-polymer",
        ),
        "ETV": (
            "4-AMINO-5-FLUORO-1-[(2R,5S)-2-(HYDROXYMETHYL)-1,3-OXATHIOLAN-5-YL]PYRIMIDIN-2(1H)-ONE",
            "non-polymer",
        ),
        "KMB": (
            "2-AMINO-6-AMINOMETHYL-8-PHENYLSULFANYLMETHYL-3H-QUINAZOLIN-4-ONE",
            "non-polymer",
        ),
        "ETT": (
            "5-(ACETYLAMINO)-2,6-ANHYDRO-3,5-DIDEOXY-3-[(2E)-3-(4-METHYLPHENYL)PROP-2-EN-1-YL]-D-GLYCERO-D-GALACTO-NON-2-ENONIC ACID",
            "non-polymer",
        ),
        "QN3": (
            "N,N'-(3S,4S)-PYRROLIDINE-3,4-DIYLBIS(N-BENZYL-2-CHLOROBENZENESULFONAMIDE)",
            "non-polymer",
        ),
        "ETX": ("2-ETHOXYETHANOL", "non-polymer"),
        "ETY": ("4-ETHYLPHENOL", "non-polymer"),
        "KMH": ("(5S)-5-AMINO-3-OXO-HEXANOIC ACID", "non-polymer"),
        "ETB": ("ETHYL COENZYME A", "non-polymer"),
        "ETC": (
            "(R,R)-5,11-CIS-DIETHYL-5,6,11,12-TETRAHYDROCHRYSENE-2,8-DIOL",
            "non-polymer",
        ),
        "1IN": (
            "1-[2-HYDROXY-4-(2-HYDROXY-5-METHYL-CYCLOPENTYLCARBAMOYL)5-PHENYL-PENTYL]-4-(3-PYRIDIN-3-YL-PROPIONYL)-PIPERAZINE-2-CARBOXYLIC ACID TERT-BUTYLAMIDE",
            "non-polymer",
        ),
        "ETF": ("TRIFLUOROETHANOL", "non-polymer"),
        "ETG": (
            "(3S,5S,7S)-N-(2-{2-[2-({[5-(DIMETHYLAMINO)NAPHTHALEN-1-YL]SULFONYL}AMINO)ETHOXY]ETHOXY}ETHYL)TRICYCLO[3.3.1.1~3,7~]DECANE-1-CARBOXAMIDE",
            "non-polymer",
        ),
        "ETD": ("ETHENYLENE GROUP", "non-polymer"),
        "ETE": ("2-{2-[2-2-(METHOXY-ETHOXY)-ETHOXY]-ETHOXY}-ETHANOL", "non-polymer"),
        "ETH": ("ETHYL GROUP", "non-polymer"),
        "ETI": ("IODOETHANE", "non-polymer"),
        "ETN": ("METHYLETHYLAMINE", "non-polymer"),
        "ETO": ("ETHOXYCARBONYL GROUP", "non-polymer"),
        "ETM": ("2-(TRIMETHYLAMMONIUM)ETHYL THIOL", "non-polymer"),
        "BEX": (
            "N-[(R)-[(3,4-DIMETHOXYPHENYL)AMINO](PHENOXY)METHYL]SULFAMIDE",
            "non-polymer",
        ),
        "K66": (
            "[4,5,6,7-TETRABROMO-2-(DIMETHYLAMINO)-1H-BENZIMIDAZOL-1-YL]ACETIC ACID",
            "non-polymer",
        ),
        "GF2": ("?", "nucleic-acid"),
        "K64": (
            "3-[2-HYDROXY-3-(3-HYDROXY-2-METHYL-BENZOYLAMINO)-4-PHENYL-BUTYRYL]-5,5- DIMETHYL-THIAZOLIDINE-4-CARBOXYLIC ACID 2-METHYL-BENZYLAMIDE",
            "non-polymer",
        ),
        "RVP": ("RIBAVIRIN MONOPHOSPHATE", "non-polymer"),
        "K62": (
            "(5S)-3-(3-ACETYLPHENYL)-N-[(1S,2R)-3-[(1,3-BENZODIOXOL-5-YLSULFONYL)(2-METHYLPROPYL)AMINO]-2-HYDROXY-1-(PHENYLMETHYL)PROPYL]-2-OXO-5-OXAZOLIDINECARBOXAMIDE",
            "non-polymer",
        ),
        "TBM": ("4-METHYL-4-[(E)-2-BUTENYL]-4,N-DIMETHYL-THREONINE", "peptide"),
        "K60": (
            "5S)-N-[(1S,2R)-3-[(1,3-BENZODIOXOL-5-YLSULFONYL)(2-METHYLPROPYL)AMINO]-2-HYDROXY-1-(PHENYLMETHYL)PROPYL]-2-OXO-3-[3-(TRIFLUOROMETHYL)PHENYL]-5-OXAZOLIDINECARBOXAMIDE",
            "non-polymer",
        ),
        "1IG": (
            "N-{2-[6-(2,4-DIAMINO-6-ETHYLPYRIMIDIN-5-YL)-2,2-DIMETHYL-3-OXO-2,3-DIHYDRO-4H-1,4-BENZOTHIAZIN-4-YL]ETHYL}ACETAMIDE",
            "non-polymer",
        ),
        "K68": ("(4,5,6,7-TETRABROMO-1H-BENZIMIDAZOL-1-YL)ACETIC ACID", "non-polymer"),
        "GF1": (
            "(2S)-2-[(3AR,4R,7S,7AS)-1,3-DIOXOOCTAHYDRO-2H-4,7-METHANOISOINDOL-2-YL]PROPANOIC ACID",
            "non-polymer",
        ),
        "K6H": (
            "(2S)-2-CHLORANYL-2-[(2R)-5-OXIDANYLIDENE-2H-FURAN-2-YL]ETHANOIC ACID",
            "non-polymer",
        ),
        "GF7": (
            "(3S)-1-(4-ACETYLPHENYL)-5-OXOPYRROLIDINE-3-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "TBL": ("N-[(4-METHOXYPHENYL)SULFONYL]-D-ALANINE", "non-polymer"),
        "0ML": ("(2S,3S)-3-AMINO-2-HYDROXY-N-(4-IODOBENZYL)BUTANAMIDE", "peptide"),
        "K6X": ("3,4,5,6,7-PENTABROMO-1H-INDAZOLE", "non-polymer"),
        "ET2": ("6-FLUORO-1H-BENZIMIDAZOL-2-AMINE", "non-polymer"),
        "ET0": ("3-(5-METHOXY-1H-INDOL-3-YL)PROPANOIC ACID", "non-polymer"),
        "ET1": (
            "3-{5-METHOXY-1-[(4-METHOXYPHENYL)SULFONYL]-1H-INDOL-3-YL}PROPANOIC ACID",
            "non-polymer",
        ),
        "QND": ("QUINALDIC ACID", "non-polymer"),
        "TBO": (
            "5-CHLORO-8-METHYL-7-(3-METHYL-BUT-2-ENYL)-6,7,8,9-TETRAHYDRO-2H-2,7,9A-TRIAZA-BENZO[CD]AZULENE-1-THIONE",
            "non-polymer",
        ),
        "KM1": ("5-CHLORO-8-[(3-CHLOROBENZYL)OXY]QUINOLINE", "non-polymer"),
        "KM0": ("8-(BENZYLOXY)-5-CHLOROQUINOLINE", "non-polymer"),
        "PEK": (
            "(1S)-2-{[(2-AMINOETHOXY)(HYDROXY)PHOSPHORYL]OXY}-1-[(STEAROYLOXY)METHYL]ETHYL (5E,8E,11E,14E)-ICOSA-5,8,11,14-TETRAENOATE",
            "non-polymer",
        ),
        "QNI": ("6-(2,9-DIAZASPIRO[5.5]UNDECAN-2-YL)-9H-PURINE", "non-polymer"),
        "MC9": ("CALCIPOTRIOL", "non-polymer"),
        "CVX": ("O-BUTLYLMETHYLPHOSPHONIC ACID ESTER GROUP", "non-polymer"),
        "MC8": ("MC-1", "non-polymer"),
        "4VQ": (
            "4-[3-AMINO-6-(3,4,5-TRIMETHOXYPHENYL)PYRAZIN-2-YL]-2-METHYLBENZOIC ACID",
            "non-polymer",
        ),
        "LZT": (
            "N~2~-{[(1S)-6-METHOXY-3-OXO-2,3-DIHYDRO-1H-INDEN-1-YL]ACETYL}-N-{(1S)-1-[(4-METHYLBENZYL)CARBAMOYL]-3-PHENYLPROPYL}-L-THREONINAMIDE",
            "non-polymer",
        ),
        "CVM": ("CYMAL-4", "non-polymer"),
        "CVI": ("CRYSTAL VIOLET", "non-polymer"),
        "CVD": (
            "(2S)-1-(8H-CARBAZOL-4-YLOXY)-3-[2-(2-METHOXYPHENOXY)ETHYLAMINO]PROPAN-2-OL",
            "non-polymer",
        ),
        "R9A": ("RUTHENIUM WIRE, 9 CARBON LINKER", "non-polymer"),
        "4VA": (
            "(2S)-4-AMINO-N-[(1R)-1-(4-CHLORO-2-FLUORO-3-PHENOXYPHENYL)PROPYL]-4-OXOBUTAN-2-AMINIUM",
            "non-polymer",
        ),
        "X81": ("1-STEAROYL-2-OLEOYL-SN-GLYCERO-3-PHOSPHATE", "non-polymer"),
        "MC4": ("2-METHYLACETOACETYL COA", "non-polymer"),
        "X87": (
            "4-[(3-AMINOPROPYL)AMINO]-5-NITRO-2-[(PYRIDIN-3-YLMETHYL)AMINO]BENZAMIDE",
            "non-polymer",
        ),
        "X86": (
            "4-[(3-METHOXYPROPYL)AMINO]-5-NITRO-2-[(PYRIDIN-3-YLMETHYL)AMINO]BENZAMIDE",
            "non-polymer",
        ),
        "X85": (
            "4-{[3-(MORPHOLIN-4-YL)PROPYL]AMINO}-5-NITRO-2-[(PYRIDIN-3-YLMETHYL)AMINO]BENZAMIDE",
            "non-polymer",
        ),
        "X84": (
            "4-[(3-HYDROXYPROPYL)AMINO]-5-NITRO-2-[(PYRIDIN-3-YLMETHYL)AMINO]BENZAMIDE",
            "non-polymer",
        ),
        "MC3": ("1,2-DIMYRISTOYL-RAC-GLYCERO-3-PHOSPHOCHOLINE", "non-polymer"),
        "X89": (
            "1-[(2R)-2-[(2,4-DICHLOROBENZYL)OXY]-2-(2,4-DICHLOROPHENYL)ETHYL]-1H-IMIDAZOLE",
            "non-polymer",
        ),
        "X88": ("5-NITRO-2-[(PYRAZIN-2-YLMETHYL)AMINO]BENZAMIDE", "non-polymer"),
        "VWN": (
            "2-[(1-METHYL-1H-BENZIMIDAZOL-2-YL)METHYL]-6-MORPHOLIN-4-YLPYRIMIDIN-4(3H)-ONE",
            "non-polymer",
        ),
        "MC2": ("N2-(L-PHENYLALANYL)-N1-(NAPHTHALENYL)-L-ARIGNINAMIDE", "non-polymer"),
        "TSA": (
            "8-HYDROXY-2-OXA-BICYCLO[3.3.1]NON-6-ENE-3,5-DICARBOXYLIC ACID",
            "non-polymer",
        ),
        "K2Z": (
            "BENZYL [(1S)-1-{[(1S,2S)-1-ETHYL-2-HYDROXY-3-{[3-(4-METHYLPIPERAZIN-1-YL)PROPYL]AMINO}-3-OXOPROPYL]CARBAMOYL}-3-METHYLBUTYL]CARBAMATE",
            "non-polymer",
        ),
        "TS8": ("3,4,5-TRIACETYLOXY-6-(ACETYLOXYMETHYL)OXANE-2-THIOL", "non-polymer"),
        "X8Z": ("L-CAPTOPRIL", "non-polymer"),
        "TS9": ("(2S,3S,4R)-2-AMINO-3,4-DIHYDROXY-3-METHYLPENTANOIC ACID", "peptide"),
        "LO1": ("[[4-(AMINOMETHYL)PHENYL]AMINO]OXO-ACETIC ACID,", "non-polymer"),
        "X8A": (
            "[(5R)-5-CARBOXY-5-(METHYLAMINO)-7-(PIPERIDIN-1-YL)HEPTYL](TRIHYDROXY)BORATE(1-)",
            "non-polymer",
        ),
        "1QQ": ("(3S)-3-METHYL-2-OXOPENTANOIC ACID", "non-polymer"),
        "X8E": (
            "5-METHYL-8-PYRIDIN-4-YL[1,2,4]TRIAZOLO[4,3-A]QUINOLIN-1(2H)-ONE",
            "non-polymer",
        ),
        "X8D": ("5-METHYL[1,2,4]TRIAZOLO[4,3-A]QUINOLIN-1(2H)-ONE", "non-polymer"),
        "Y": ("?", "nucleic-acid"),
        "BER": ("BERBERINE", "non-polymer"),
        "TS4": ("GLUTATHIONYLSPERMIDINE DISULFIDE", "non-polymer"),
        "X8M": ("IPRATROPIUM", "non-polymer"),
        "GU8": ("2,3,6-TRI-O-METHYL-BETA-D-GLUCOPYRANOSE", "saccharide"),
        "RBP": (
            "(3R,4R)-4-(4-BROMOPHENYL)-3-(2-NAPHTHYLMETHOXY)PIPERIDINE",
            "non-polymer",
        ),
        "SAZ": (
            "(1S)-N,4-DIMETHYL-N-(4-METHYLPENT-3-ENYL)CYCLOHEX-3-ENAMINIUM",
            "non-polymer",
        ),
        "4V2": ("3-(5-MERCAPTO-1,3,4-OXADIAZOL-2-YL)-PHENOL", "non-polymer"),
        "PE4": (
            "2-{2-[2-(2-{2-[2-(2-ETHOXY-ETHOXY)-ETHOXY]-ETHOXY}-ETHOXY)-ETHOXY]-ETHOXY}-ETHANOL",
            "non-polymer",
        ),
        "R9P": (
            "8-BROMO-2-{2-CHLORO-4-[(PIPERIDIN-4-YLMETHYL)AMINO]PHENYL}[1]BENZOFURO[3,2-D]PYRIMIDIN-4(3H)-ONE",
            "non-polymer",
        ),
        "BES": (
            "2-(3-AMINO-2-HYDROXY-4-PHENYL-BUTYRYLAMINO)-4-METHYL-PENTANOIC ACID",
            "non-polymer",
        ),
        "TS1": (
            "4-{2,6,8-TRIOXO-9-[(2R,3S,4R)-2,3,4,5-TETRAHYDROXYPENTYL]-1,2,3,6,8,9-HEXAHYDRO-7H-PURIN-7-YL}BUTYL DIHYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "SL6": (
            "(Z)-4-[4-[(4-CHLOROPHENYL)METHYL]-1-(CYCLOHEXYLMETHYL)PIPERIDIN-4-YL]-2-OXIDANYL-4-OXIDANYLIDENE-BUT-2-ENOIC ACID",
            "non-polymer",
        ),
        "REZ": (
            "(2R)-2-AMINO-7-{[(1R)-1-CARBOXYETHYL]AMINO}-7-OXOHEPTANOIC ACID",
            "non-polymer",
        ),
        "53A": (
            "ETHYL 5-{[(1R)-1-(ETHOXYCARBONYL)-2-OXOPROPYL]SULFANYL}-1,2-DIHYDRO[1,2,3]TRIAZOLO[1,5-A]QUINAZOLINE-3-CARBOXYLATE",
            "non-polymer",
        ),
        "53N": ("3-[5-(3-NITROPHENYL)THIOPHEN-2-YL]PROPANOIC ACID", "non-polymer"),
        "53J": (
            "(2E)-3-{5-[(2,4-DIAMINOPYRIMIDIN-5-YL)METHYL]-2,3-DIMETHOXYPHENYL}-1-[(1S)-1-(2-METHYLPROP-1-EN-1-YL)PHTHALAZIN-2(1H)-YL]PROP-2-EN-1-ONE",
            "non-polymer",
        ),
        "53I": (
            "(2E)-3-{5-[(2,4-DIAMINOPYRIMIDIN-5-YL)METHYL]-2,3-DIMETHOXYPHENYL}-1-[(1R)-1-(2-METHYLPROP-1-EN-1-YL)PHTHALAZIN-2(1H)-YL]PROP-2-EN-1-ONE",
            "non-polymer",
        ),
        "53H": ("5'-O-[(3-METHYL-D-VALYL)SULFAMOYL]ADENOSINE", "non-polymer"),
        "53V": (
            "5-[(3R)-3-(5-METHOXY-2',6'-DIMETHYLBIPHENYL-3-YL)BUT-1-YN-1-YL]-6-METHYLPYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "53U": ("D-PHENYLALANYL-N-BENZYL-L-PROLINAMIDE", "peptide"),
        "53T": (
            "5-[(3R)-3-(5-METHOXY-3',5'-DIMETHYLBIPHENYL-3-YL)BUT-1-YN-1-YL]-6-METHYLPYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "53S": (
            "5-[(3R)-3-(5-METHOXY-4'-METHYLBIPHENYL-3-YL)BUT-1-YN-1-YL]-6-METHYLPYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "53R": (
            "5-[(3R)-3-(5-METHOXYBIPHENYL-3-YL)BUT-1-YN-1-YL]-6-METHYLPYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "SL3": ("?", "non-polymer"),
        "N8T": (
            "N-[8-(1,2,3,4-TETRAHYDROACRIDIN-9-YLTHIO)OCTYL]-1,2,3,4-TETRAHYDROACRIDIN-9-AMINE",
            "non-polymer",
        ),
        "2XA": (
            "(2R)-2-AMINO-3-[(4-CHLOROPHENYL)-DIPHENYL-METHYL]SULFANYL-PROPANOIC ACID",
            "peptide",
        ),
        "V63": ("[(1R,5S)-8-AZABICYCLO[3.2.1]OCTAN-3-YL] BENZOATE", "non-polymer"),
        "2XY": (
            "N-[2-NITRO-4-(TRIFLUOROMETHYL)PHENYL]MORPHOLIN-4-AMINE",
            "non-polymer",
        ),
        "VAS": (
            "(2S,4S,5S)-5-AMINO-6-CYCLOHEXYL-4-HYDROXY-2-(1-METHYLETHYL)HEXANOIC ACID",
            "peptide",
        ),
        "N8C": ("1-NITROOCTANE", "non-polymer"),
        "5H1": ("1H-INDOL-5-OL", "non-polymer"),
        "N8M": ("5'-DEOXY-5'-(DIMETHYLAMINO)-8-METHYLADENOSINE", "non-polymer"),
        "2XP": ("(2Z,4Z)-2-IMINO-6-OXOHEX-4-ENOIC ACID", "non-polymer"),
        "REP": (
            "(1,10 PHENANTHROLINE)-(TRI-CARBON MONOXIDE) RHENIUM (I)",
            "non-polymer",
        ),
        "UF0": (
            "O-[(R)-HYDROXY{[(3R)-3-HYDROXY-4-{[3-({2-[(HYDROXYACETYL)AMINO]ETHYL}AMINO)-3-OXOPROPYL]AMINO}-2,2-DIMETHYL-4-OXOBUTYL]OXY}PHOSPHORYL]-L-SERINE",
            "peptide",
        ),
        "5HG": (
            "{[2-(6-AMINO-9H-PURIN-9-YL)ETHOXY]METHYL}PHOSPHONIC ACID",
            "non-polymer",
        ),
        "5HD": ("5-(2-HYDROXYETHYL)NONANE-1,9-DIOL", "non-polymer"),
        "5HE": (
            "(5R,6E,8Z,11Z,14Z,17Z)-5-HYDROXYICOSA-6,8,11,14,17-PENTAENOIC ACID",
            "non-polymer",
        ),
        "5HC": ("?", "nucleic-acid"),
        "5HA": (
            "N-[(1S,2R)-1-BENZYL-3-(CYCLOPROPYLAMINO)-2-HYDROXYPROPYL]-5-[METHYL(METHYLSULFONYL)AMINO]-N'-[(1R)-1-PHENYLETHYL]ISOPHTHALAMIDE",
            "non-polymer",
        ),
        "MDF": ("?", "peptide"),
        "5HK": (
            "2-[3-(3-METHOXY-4-NITROPHENYL)-11-OXO-10,11-DIHYDRO-5H-DIBENZO[B,E][1,4]DIAZEPIN-8-YL]-N,N-DIMETHYLACETAMIDE",
            "non-polymer",
        ),
        "5HI": (
            "(3R,5R)-7-[2-(4-FLUOROPHENYL)-4-{[(1S)-2-HYDROXY-1-PHENYLETHYL]CARBAMOYL}-5-(1-METHYLETHYL)-1H-IMIDAZOL-1-YL]-3,5-DIHYDROXYHEPTANOIC ACID",
            "non-polymer",
        ),
        "K2B": ("(8ALPHA,9BETA)-CHOLEST-4-EN-3-ONE", "non-polymer"),
        "5HT": ("5-HYDROXY-THYMIDINE", "nucleic-acid"),
        "5HU": ("5-HYDROXYMETHYLURIDINE-2'-DEOXY-5'-MONOPHOSPHATE", "nucleic-acid"),
        "5HP": ("PYROGLUTAMIC ACID", "peptide"),
        "K2C": (
            "6,7,12,13-TETRAHYDRO-5H-INDOLO[2,3-A]PYRROLO[3,4-C]CARBAZOL-5-ONE",
            "non-polymer",
        ),
        "V6F": ("(4R)-1-ACETYL-4-HYDROXY-N-METHYL-L-PROLINAMIDE", "non-polymer"),
        "QUM": ("QUINACRINE MUSTARD", "non-polymer"),
        "2X7": ("(3E)-3-IMINOOXEPIN-2(3H)-ONE", "non-polymer"),
        "5HY": ("[(4R)-2,5-DIOXOIMIDAZOLIDIN-4-YL]ACETIC ACID", "non-polymer"),
        "REJ": ("TRICARBONYL (L-SERINE) RHENIUM(I)", "non-polymer"),
        "K2A": (
            "(3R,3AS,6AR)-HEXAHYDROFURO[2,3-B]FURAN-3-YL [(1S,2R)-1-BENZYL-2-HYDROXY-3-({[4-(HYDROXYMETHYL)PHENYL]SULFONYL}[(2S)-2-METHYLBUTYL]AMINO)PROPYL]CARBAMATE",
            "non-polymer",
        ),
        "VAM": (
            "METHYL 4-O-(4-THIO-BETA-D-GLUCOPYRANOSYL)-BETA-D-GLUCOPYRANOSIDE",
            "non-polymer",
        ),
        "537": ("2,6-DIHYDROANTHRA/1,9-CD/PYRAZOL-6-ONE", "non-polymer"),
        "DR1": ("5-METHYL-5H-INDOLO[3,2-B]QUINOLINE", "non-polymer"),
        "535": (
            "ETHYL 3-METHYL-5,6-DIHYDROIMIDAZO[2,1-B][1,3]THIAZOLE-2-CARBOXYLATE",
            "non-polymer",
        ),
        "534": (
            "5-{3-[(4-{3-[4-(1-METHYLETHYL)PHENYL]PYRAZIN-2-YL}PIPERAZIN-1-YL)SULFONYL]PHENYL}PYRIMIDIN-2-AMINE",
            "non-polymer",
        ),
        "533": (
            "6-(2,4-DIFLUOROPHENOXY)-8-METHYL-2-{[(1R)-1-METHYL-2-(2H-TETRAZOL-2-YL)ETHYL]AMINO}PYRIDO[2,3-D]PYRIMIDIN-7(8H)-ONE",
            "non-polymer",
        ),
        "532": (
            "(5S)-2-AMINO-3-METHYL-5-[3-(PYRIDIN-3-YL)PHENYL]-5-(THIOPHEN-3-YL)-3,5-DIHYDRO-4H-IMIDAZOL-4-ONE",
            "non-polymer",
        ),
        "530": (
            "6-(2,4-DIFLUOROPHENOXY)-N-(TETRAHYDRO-2H-PYRAN-4-YL)-1H-PYRAZOLO[3,4-D]PYRIMIDIN-3-AMINE",
            "non-polymer",
        ),
        "W22": (
            "[(2S)-4-METHYL-3-OXO-2,3,4,5-TETRAHYDRO-1H-1,4-BENZODIAZEPIN-2-YL]ACETIC ACID",
            "non-polymer",
        ),
        "UF2": (
            "1-(2-DEOXY-2-FLUORO-5-O-PHOSPHONO-BETA-D-ARABINOFURANOSYL)PYRIMIDINE-2,4(1H,3H)-DIONE",
            "nucleic-acid",
        ),
        "K2E": (
            "(3R,3AS,6AR)-HEXAHYDROFURO[2,3-B]FURAN-3-YL {(1S,2R)-3-[(1,3-BENZODIOXOL-5-YLSULFONYL)(2-ETHYLBUTYL)AMINO]-1-BENZYL-2-HYDROXYPROPYL}CARBAMATE",
            "non-polymer",
        ),
        "DR3": (
            "6-DEOXY-ALPHA-L-GALACTOPYRANOSYL-(1->2)-BETA-D-GALACTOPYRANOSYL-(1->3)-2-(ACETYLAMINO)-2-DEOXY-BETA-D-GLUCOPYRANOSE",
            "saccharide",
        ),
        "538": ("?", "non-polymer"),
        "PTB": (
            "(5-TERT-BUTYL-1,3,4-OXADIAZOL-2-YL)[(2R)-PYRROLIDIN-2-YL]METHANONE",
            "non-polymer",
        ),
        "PE5": ("3,6,9,12,15,18,21,24-OCTAOXAHEXACOSAN-1-OL", "non-polymer"),
        "OX3": ("4-HYDROXYBENZALDEHYDE O-(CYCLOHEXYLCARBONYL)OXIME", "non-polymer"),
        "REC": ("CEPHALOSPORIN", "non-polymer"),
        "0BU": (
            "8-({3-[({3-[(4,6,8-TRISULFONAPHTHALEN-1-YL)CARBAMOYL]PHENYL}CARBAMOYL)AMINO]BENZOYL}AMINO)NAPHTHALENE-1,3,5-TRISULFONIC ACID",
            "non-polymer",
        ),
        "3B3": (
            "(2S)-1-AMINO-3-[(5-NITROQUINOLIN-8-YL)AMINO]PROPAN-2-OL",
            "non-polymer",
        ),
        "0BW": (
            "METHYL 3-[(3-METHOXYBENZOYL)AMINO]-4-(4-METHYLPIPERAZIN-1-YL)BENZOATE",
            "non-polymer",
        ),
        "0BV": ("PHENANTHREN-9-YLACETALDEHYDE", "non-polymer"),
        "3B6": (
            "(2S)-N-(4-CYANO-3-IODOPHENYL)-3-(4-CYANOPHENOXY)-2-HYDROXY-2-METHYLPROPANAMIDE",
            "non-polymer",
        ),
        "3B4": ("2-(CYCLOHEXYLAMINO)BENZOIC ACID", "non-polymer"),
        "0BR": (
            "(4BETA,8ALPHA,9R)-6'-METHOXY-10,11-DIHYDROCINCHONAN-9-OL",
            "non-polymer",
        ),
        "6AB": ("6-AMINOBENZOIC ACID", "non-polymer"),
        "ZG1": (
            "1-ETHYL-5-[3-(2-OXO-2-PYRROLIDIN-1-YLETHYL)-1,2,4-OXADIAZOL-5-YL]-N-(TETRAHYDRO-2H-PYRAN-4-YL)-1H-PYRAZOLO[3,4-B]PYRIDIN-4-AMINE",
            "non-polymer",
        ),
        "3B8": (
            "2-AMINO-7-FLUORO-5-OXO-5H-CHROMENO[2,3-B]PYRIDINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "3B9": (
            "N-[2-(2,4-DIAMINOPYRIDO[2,3-D]PYRIMIDIN-7-YL)-2-METHYLPROPYL]-4-PHENOXYBENZAMIDE",
            "non-polymer",
        ),
        "0BY": (
            "4-({4-[(2-CHLOROPHENYL)AMINO]PYRIMIDIN-2-YL}AMINO)BENZOIC ACID",
            "non-polymer",
        ),
        "0BX": ("4-{[4-(PHENYLAMINO)PYRIMIDIN-2-YL]AMINO}BENZOIC ACID", "non-polymer"),
        "0BZ": (
            "4-({4-[(2-FLUOROPHENYL)AMINO]PYRIMIDIN-2-YL}AMINO)BENZOIC ACID",
            "non-polymer",
        ),
        "6AZ": ("6-AZIDO-UMP", "non-polymer"),
        "0BD": (
            "3-METHYL-1-(2-METHYLPROPYL)BUTYL 4-O-BETA-L-GULOPYRANOSYL-BETA-D-GLUCOPYRANOSIDE",
            "saccharide",
        ),
        "0BB": (
            "4-{3-[(4-METHYL-1-BENZOTHIOPHEN-3-YL)METHYL]-2-OXO-2,3-DIHYDRO-1H-BENZIMIDAZOL-1-YL}BUTANOIC ACID",
            "non-polymer",
        ),
        "0BM": (
            "N-(METHYLSULFONYL)-D-PHENYLALANYL-N-[(1-CARBAMIMIDOYLPIPERIDIN-4-YL)METHYL]-L-PROLINAMIDE",
            "non-polymer",
        ),
        "6AP": ("9H-PURINE-2,6-DIAMINE", "non-polymer"),
        "0BN": ("4-CARBAMIMIDOYL-L-PHENYLALANINE", "peptide"),
        "0BI": (
            "(4S)-4-[(1R)-1-HYDROXY-2-({1-[3-(1-METHYLETHYL)PHENYL]CYCLOPROPYL}AMINO)ETHYL]-19-(METHOXYMETHYL)-11-OXA-3,16-DIAZATRICYCLO[15.3.1.1~6,10~]DOCOSA-1(21),6(22),7,9,17,19-HEXAEN-2-ONE",
            "non-polymer",
        ),
        "6AU": ("?", "non-polymer"),
        "3BR": ("1-BROMOPROPANE", "non-polymer"),
        "3BS": (
            "3-CHLORO-4-{[(1-HYDROXY-2,2,6,6-TETRAMETHYLPIPERIDIN-4-YL)CARBAMOTHIOYL]AMINO}BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "3BP": (
            "2-(3-BIPHENYL-4-YL-2-ETHANESULFONYLAMINO-PROPIONYLAMINO)-PENTANEDIOIC ACID 5-AMIDE 1-(4-CARBAMIMIDOYL-BENZYLAMIDE)",
            "non-polymer",
        ),
        "0B1": (
            "2-AMINO-6-[(1S,2R)-2-(3'-METHOXYBIPHENYL-3-YL)CYCLOPROPYL]-3-METHYLPYRIMIDIN-4(3H)-ONE",
            "non-polymer",
        ),
        "P05": ("D-PHENYLALANYL-N-(PYRIDIN-4-YLMETHYL)-L-PROLINAMIDE", "peptide"),
        "3BT": (
            "2-{2-[3-(2-CARBAMIMIDOYLSULFANYL-ETHYL)-PHENYL]-ETHYL}-ISOTHIOUREA",
            "non-polymer",
        ),
        "3BZ": ("3-CHLOROBENZOATE", "non-polymer"),
        "SLP": ("3-SELANYLPROPANOIC ACID", "non-polymer"),
        "P04": (
            "19-(CYCLOPROPYLAMINO)-4,6,7,15-TETRAHYDRO-5H-16,1-(AZENOMETHENO)-10,14-(METHENO)PYRAZOLO[4,3-O][1,3,9]TRIAZACYCLOHEXADECIN-8(9H)-ONE",
            "non-polymer",
        ),
        "ZGU": ("(S)-1'-(2',3'-DIHYDROXYPROPYL)-GUANINE", "nucleic-acid"),
        "SLQ": (
            "(2S)-2-({9-(1-METHYLETHYL)-6-[(4-PYRIDIN-2-YLBENZYL)AMINO]-9H-PURIN-2-YL}AMINO)BUTAN-1-OL",
            "non-polymer",
        ),
        "3BB": ("3-BROMOBUT-3-EN-1-OL", "non-polymer"),
        "3BC": (
            "N-(R-CARBOXY-ETHYL)-ALPHA-(S)-(2-PHENYLETHYL)GLYCYL-L-ARGININE-N-PHENYLAMIDE",
            "non-polymer",
        ),
        "3BA": (
            "7-AMINO-2-TERT-BUTYL-4-{[2-(1H-IMIDAZOL-4-YL)ETHYL]AMINO}PYRIDO[2,3-D]PYRIMIDINE-6-CARBOXAMIDE",
            "non-polymer",
        ),
        "ZGL": ("D-ALPHA-GLUTAMINE", "peptide"),
        "3BD": (
            "7-AMINO-2-TERT-BUTYL-4-(4-PYRIMIDIN-2-YLPIPERAZIN-1-YL)PYRIDO[2,3-D]PYRIMIDINE-6-CARBOXAMIDE",
            "non-polymer",
        ),
        "3BE": (
            "3-BROMO-N'-[(1E)-(3,5-DIBROMO-2,4-DIHYDROXYPHENYL)METHYLIDENE]BENZOHYDRAZIDE",
            "non-polymer",
        ),
        "3BK": (
            "(2R,3R,4S,5R)-2-[6-AMINO-8-[(3,4-DICHLOROPHENYL)METHYLAMINO]PURIN-9-YL]-5-(HYDROXYMETHYL)OXOLANE-3,4-DIOL",
            "non-polymer",
        ),
        "3BH": (
            "(2R,3R,4S,5R)-2-(6,8-DIAMINOPURIN-9-YL)-5-(HYDROXYMETHYL)OXOLANE-3,4-DIOL",
            "non-polymer",
        ),
        "3BI": (
            "(2S)-2-{[(R)-[(3R)-3-CARBOXY-3-{[(4-{[(2,4-DIAMINOPTERIDIN-6-YL)METHYL](METHYL)AMINO}PHENYL)CARBONYL]AMINO}PROPYL](HYDROXY)PHOSPHORYL]METHYL}PENTANEDIOIC ACID",
            "non-polymer",
        ),
        "3BN": (
            "3-BENZOYL-N-[(1S,2R)-1-BENZYL-3-(CYCLOPROPYLAMINO)-2-HYDROXYPROPYL]-5-[METHYL(METHYLSULFONYL)AMINO]BENZAMIDE",
            "non-polymer",
        ),
        "3BM": (
            "2-[(2-CHLORO-4-IODOPHENYL)AMINO]-N-{[(2R)-2,3-DIHYDROXYPROPYL]OXY}-3,4-DIFLUOROBENZAMIDE",
            "non-polymer",
        ),
        "P08": ("PIK-108", "non-polymer"),
        "LFC": ("6-DEOXY-L-GALACTONIC ACID", "non-polymer"),
        "R4A": (
            "BIS[1H,1'H-2,2'-BIPYRIDINATO(2-)-KAPPA~2~N~1~,N~1'~]{3-[4-(1,10-DIHYDRO-1,10-PHENANTHROLIN-4-YL-KAPPA~2~N~1~,N~10~)BUTOXY]-N,N-DIMETHYLANILINATO(2-)}RUTHENIUM",
            "non-polymer",
        ),
        "SLI": ("2-HYDROXY-N-PHENYLBENZAMIDE", "non-polymer"),
        "XLC": (
            "3-CHLORO-N-[4-CHLORO-2-[[(4-CHLOROPHENYL)AMINO]CARBONYL]PHENYL]-4-[(4-METHYL-1-PIPERAZINYL)METHYL]-2-THIOPHENECARBOXAMIDE",
            "non-polymer",
        ),
        "XSP": ("(2R)-3-OXO-2-(PHOSPHONOOXY)PROPANOIC ACID", "non-polymer"),
        "S0H": ("(S)-2-HYDROXYPROPYLPHOSPHONIC ACID", "non-polymer"),
        "SLD": (
            "(3Z)-N-[(4E)-5-(4-{(5S)-5-[(ACETYLAMINO)METHYL]-2-OXO-1,3-OXAZOLIDIN-3-YL}-2-FLUOROPHENYL)PENT-4-EN-1-YL]-3-(4-METHYL-2,6-DIOXO-1,6-DIHYDROPYRIMIDIN-5(2H)-YLIDENE)PROPANAMIDE",
            "non-polymer",
        ),
        "YQY": (
            "[3-(AMINOMETHYL)-3-OXIDANYL-AZETIDIN-1-YL]-[3,4-BIS(FLUORANYL)-2-[(2-FLUORANYL-4-IODANYL-PHENYL)AMINO]PHENYL]METHANONE",
            "non-polymer",
        ),
        "UBF": (
            "3-({3-[(2S)-2-AMINO-2-CARBOXYETHYL]-5-BROMO-2,6-DIOXO-3,6-DIHYDROPYRIMIDIN-1(2H)-YL}METHYL)THIOPHENE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "S0N": ("O-SUCCINYLBENZOYL-N-COENZYME A", "non-polymer"),
        "SLE": ("2-(THIOMETHYLENE)-4-METHYLPENTANOIC ACID", "non-polymer"),
        "SLF": (
            "SULFATE-3-D-GALACTOSYL-BETA-1-1-N-STEAROYL-D-SPHINGOSINE",
            "non-polymer",
        ),
        "QFI": (
            "({4-[(2S,3R)-2-({[(3R,3AS,6AR)-HEXAHYDROFURO[2,3-B]FURAN-3-YLOXY]CARBONYL}AMINO)-3-HYDROXY-4-{ISOBUTYL[(4-METHOXYPHENYL)SULFONYL]AMINO}BUTYL]PHENOXY}METHYL)PHOSPHONIC ACID",
            "non-polymer",
        ),
        "O8M": (
            "2-[{[(2R,3S,4R,5R)-5-(6-AMINO-8-METHYL-9H-PURIN-9-YL)-3,4-DIHYDROXYTETRAHYDROFURAN-2-YL]METHYL}(METHYL)AMINO]ACETAMIDE",
            "non-polymer",
        ),
        "1L4": (
            "N-[(2R)-3-(3-CYANOPHENYL)-1-OXO-1-(PYRROLIDIN-1-YL)PROPAN-2-YL]-8-FLUORO-1,2,3,4-TETRAHYDROISOQUINOLINE-6-SULFONAMIDE",
            "non-polymer",
        ),
        "1L2": (
            "(2S)-3-(ALPHA-D-GALACTOPYRANOSYLOXY)-2-[(9Z)-OCTADEC-9-ENOYLOXY]PROPYL (9Z,12Z)-OCTADECA-9,12-DIENOATE",
            "non-polymer",
        ),
        "SLA": (
            "(3S,4R)-3-HYDROXY-2-[(1S)-1-HYDROXY-2-METHYLPROPYL]-4-METHYL-5-OXO-D-PROLINE",
            "peptide",
        ),
        "ATA": ("ATRAZINE GLUTATHIONE CONJUGATE", "non-polymer"),
        "L20": (
            "(2S)-1-(1H-INDOL-3-YL)-3-{[5-(3-METHYL-1H-INDAZOL-5-YL)PYRIDIN-3-YL]OXY}PROPAN-2-AMINE",
            "non-polymer",
        ),
        "L23": (
            "4-[1-(2,6-DICHLOROBENZYL)-2-METHYL-1H-IMIDAZOL-4-YL]PYRIMIDIN-2-AMINE",
            "non-polymer",
        ),
        "L22": ("4-(2-AMINO-1,3-THIAZOL-4-YL)PYRIMIDIN-2-AMINE", "non-polymer"),
        "L24": (
            "[[[2-AMINO-5,6,7,8-TETRAHYDRO-4-HYDROXY-PYRIDO[2,3-D]PYRIMIDIN-6-YL]-ETHYL]-PHENYL]-CARBONYL-GLUTAMIC ACID",
            "non-polymer",
        ),
        "L27": (
            "{1-BENZYL-3-[2-BENZYL-3-OXO-4-(1-OXO-1,2,3,4-TETRAHYDRO- ISOQUINOLIN-4-YL)-2,3-DIHYDRO-1H-PYRROL-2-YL]-2- HYDROXY-PROPYL}-CARBAMIC ACID TETRAHYDRO-FURAN-3-YL ESTER",
            "non-polymer",
        ),
        "O8H": (
            "4-IODO-N-[(1-{2-OXO-2-[4-(3-THIOPHEN-2-YL-1,2,4-OXADIAZOL-5-YL)PIPERIDIN-1-YL]ETHYL}-1H-1,2,3-TRIAZOL-4-YL)METHYL]BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "SLC": ("SELENO-CAPTOPRIL", "non-polymer"),
        "L2A": (
            "(2S,5S,8S,11S,15E,20S)-20-AMINO-2-(CARBOXYMETHYL)-11,20-DIMETHYL-5,8-BIS(2-METHYLPROPYL)-3,6,9,21-TETRAOXO-1,4,7,10-TETRAAZACYCLOHENICOS-15-ENE-11-CARBOXYLIC ACID",
            "peptide",
        ),
        "L2C": ("(2S)-3-HYDROXY-2-(NONANOYLOXY)PROPYL LAURATE", "non-polymer"),
        "7UP": (
            "6-(1,2,3,4-TETRAHYDROISOQUINOLIN-6-YLETHYNYL)NAPHTHALENE-2-CARBOXIMIDAMIDE",
            "non-polymer",
        ),
        "L2D": (
            "{2,2'-[(1R,2R)-CYCLOHEXANE-1,2-DIYLBIS{[(PYRIDIN-2-YL-KAPPAN)METHYL]IMINO-KAPPAN}]DIACETATO-KAPPAO(2-)}IRON",
            "non-polymer",
        ),
        "L2G": (
            "N-(4-{1-[4-(4-ACETYLPIPERAZIN-1-YL)-TRANS-CYCLOHEXYL]-4-AMINO-1H-PYRAZOLO[3,4-D]PYRIMIDIN-3-YL}-2-METHOXYPHENYL)-1-METHYL-1H-INDOLE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "1LU": ("(2S)-2-HYDROXY-4-METHYL-PENTANOIC ACID", "non-polymer"),
        "L2M": (
            "{(R)-N-[(1S,2S)-2-{METHYL[(PYRIDIN-2-YL-KAPPAN)METHYL]AMINO-KAPPAN}CYCLOHEXYL]-N-[(PYRIDIN-2-YL-KAPPAN)METHYL]GLYCINATO-KAPPA~2~N,O}IRON(2+)",
            "non-polymer",
        ),
        "1LP": ("TRANYLCYPROMINE", "non-polymer"),
        "L2P": ("2,3-DI-PHYTANYL-GLYCEROL", "non-polymer"),
        "L2S": ("1-{[(4-METHYLPHENYL)SULFANYL]ACETYL}PIPERIDINE", "non-polymer"),
        "L2T": (
            "4-(BENZYLOXY)-N-[(1S,2R)-2-HYDROXY-1-({(1S)-1-[(2-METHYLBENZYL)CARBAMOYL]-3-PHENYLPROPYL}CARBAMOYL)PROPYL]BENZAMIDE",
            "non-polymer",
        ),
        "1LH": (
            "(2S)-4-[(2-AMMONIOETHYL)AMINO]-N-[(1R)-1-(4-CHLORO-2-FLUORO-3-PHENOXYPHENYL)PROPYL]-4-OXOBUTAN-2-AMINIUM",
            "non-polymer",
        ),
        "1LI": (
            "N-[(1S,2R)-3-{[1-(3-TERT-BUTYLPHENYL)CYCLOHEXYL]AMINO}-1-(3,5-DIFLUOROBENZYL)-2-HYDROXYPROPYL]ACETAMIDE",
            "non-polymer",
        ),
        "1LG": (
            "2,4-DICHLORO-N-(3-CYANO-4,5,6,7-TETRAHYDRO-BENZOTHIOPHEN-2YL)-5-(MORPHOLINE-4-SULFONYL)-BENZAMIDE",
            "non-polymer",
        ),
        "XED": ("DEXTROFLOXACINE", "non-polymer"),
        "O8G": (
            "1-(AZIDOACETYL)-4-(3-THIOPHEN-2-YL-1,2,4-OXADIAZOL-5-YL)PIPERIDINE",
            "non-polymer",
        ),
        "S00": (
            "N-(BENZYLSULFONYL)-3-CYCLOHEXYL-D-ALANYL-N-(4-CARBAMIMIDOYLBENZYL)-L-PROLINAMIDE",
            "peptide",
        ),
        "S01": (
            "1-[4-(8-CHLORO-3-METHYL-5,6-DIHYDRO-BENZO[5,6]CYCLOHEPTA[1,2-B]PYRIDIN-11-YLIDENE)-PIPERIDIN-1-YL]-2-PYRIDIN-4-YL-ETHANONE",
            "non-polymer",
        ),
        "S02": (
            "PHOSPHORIC ACID MONO-[2-(4-{2-[4-(2-HYDROXY-ETHOXY)-CYCLOHEXYLIDENE]-ETHYLIDENE}-CYCLOHEXYLOXY) -ETHYL] ESTER",
            "non-polymer",
        ),
        "S03": (
            "1-{4-[4-AMINO-5-(3-METHOXYPHENYL)-7H-PYRROLO[2,3-D]PYRIMIDIN-7-YL]BENZYL}PIPERIDIN-4-OL",
            "non-polymer",
        ),
        "S04": (
            "N-(BENZYLSULFONYL)-3-CYCLOHEXYL-D-ALANYL-N-[2-(AMINOMETHYL)-5-CHLOROBENZYL]-L-PROLINAMIDE",
            "peptide",
        ),
        "S05": (
            "3,4-DIHYDROQUINOLIN-1(2H)-YL[4-(1H-IMIDAZOL-5-YL)PIPERIDIN-1-YL]METHANONE",
            "non-polymer",
        ),
        "S06": (
            "(3S,5R,7R,8S,9S,10R)-7-(HYDROXYMETHYL)-3-(2-NAPHTHYL)-1,6-DIOXA-2-AZASPIRO[4.5]DECANE-8,9,10-TRIOL",
            "non-polymer",
        ),
        "LFX": (
            "(3S)-9-FLUORO-3-METHYL-10-(4-METHYLPIPERAZIN-1-YL)-7-OXO-2,3-DIHYDRO-7H-[1,4]OXAZINO[2,3,4-IJ]QUINOLINE-6-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "O8B": (
            "2-(2-METHYLPHENOXY)-N-[2-(4-METHYLPHENYL)-2H-BENZOTRIAZOL-5-YL]ACETAMIDE",
            "non-polymer",
        ),
        "MDD": ("MALONALDEHYDE", "non-polymer"),
        "SH1": (
            "N-[(2-PHENYLETHYL)SULFONYL]-D-SERYL-N-[(1S)-4-[(DIAMINOMETHYLENE)AMINO]-1-(HYDROXYMETHYL)BUTYL]-L-ALANINAMIDE",
            "non-polymer",
        ),
        "IVM": ("22,23-DIHYDROAVERMECTIN B1A; IVERMECTIN", "non-polymer"),
        "XEZ": (
            "6-(1H-PYRAZOL-3-YL)-3-(1H-PYRAZOL-4-YL)IMIDAZO[1,2-A]PYRAZINE",
            "non-polymer",
        ),
        "HQY": ("QUINOLIN-8-OL", "non-polymer"),
        "VS2": (
            "3-[N-[BENZYLOXYCARBONYL]-PHENYLALANINYL-AMINO]-5-PHENYL-PENTANE-1-SULFONYLMETHYLBENZENE",
            "non-polymer",
        ),
        "VS3": (
            "3-[N-[BENZYLOXYCARBONYL]-PHENYLALANINYL-AMINO]-5-PHENYL-PENTANE-1-SULFONIC ACID 4-NITRO-PHENYL ESTER",
            "non-polymer",
        ),
        "VS4": (
            "3-[[N-[4-METHYL-PIPERAZINYL]CARBONYL]-PHENYLALANINYL-AMINO]-5-PHENYL-PENTANE-1-SULFONIC ACID BENZYLOXY-AMIDE",
            "non-polymer",
        ),
        "HQP": (
            "4-({4-[(5-CYCLOPROPYL-1H-PYRAZOL-3-YL)AMINO]QUINAZOLIN-2-YL}IMINO)CYCLOHEXA-2,5-DIENE-1-CARBONITRILE",
            "non-polymer",
        ),
        "HQQ": ("5-METHYL-5-(4-PHENOXY-PHENYL)-PYRIMIDINE-2,4,6-TRIONE", "non-polymer"),
        "TWN": (
            "(3S)-3-[(2S,3S,4R)-3,4-DIMETHYLTETRAHYDROFURAN-2-YL]BUTYL LAURATE",
            "non-polymer",
        ),
        "HQU": ("3-HYDROXYQUINALDIC ACID", "non-polymer"),
        "HQI": ("N-(8-HYDROXYQUINOLIN-5-YL)ACETAMIDE", "non-polymer"),
        "HQL": (
            "4-(BENZHYDRYLOXY)-1-[3-(1H-TETRAAZOL-5-YL)PROPYL]PIPERIDINE",
            "non-polymer",
        ),
        "MCA": ("METHYLMALONYL-COENZYME A", "non-polymer"),
        "HQN": ("BENZENE-1,2,4-TRIOL", "non-polymer"),
        "HQO": ("2-HEPTYL-4-HYDROXY QUINOLINE N-OXIDE", "non-polymer"),
        "HQA": ("3-(8-HYDROXYQUINOLIN-3-YL)-L-ALANINE", "peptide"),
        "HQC": (
            "3-{6-[(8-HYDROXY-QUINOLINE-2-CARBONYL)-AMINO]-2-THIOPHEN-2-YL-HEXANOYLAMINO}-4-OXO-BUTYRI ACID",
            "non-polymer",
        ),
        "HQD": ("3-HYDROXY-2-METHYLQUINOLIN-4(1H)-ONE", "non-polymer"),
        "HQE": ("BENZENE-1,4-DIOL", "non-polymer"),
        "FAP": (
            "(2S)-1-[4-({6-[(2,6-DIFLUOROPHENYL)AMINO]PYRIMIDIN-4-YL}AMINO)PHENOXY]-3-(DIMETHYLAMINO)PROPAN-2-OL",
            "non-polymer",
        ),
        "GFH": (
            "2'-DEOXY-5'-O-[(R)-{[(R)-[(R)-FLUORO(PHOSPHONO)METHYL](HYDROXY)PHOSPHORYL]OXY}(HYDROXY)PHOSPHORYL]GUANOSINE",
            "non-polymer",
        ),
        "8TX": ("8-THIOXO-3,7,8,9-TETRAHYDRO-1H-PURINE-2,6-DIONE", "non-polymer"),
        "P9L": (
            "1,3-DIMETHYL-6-(MORPHOLIN-4-YLSULFONYL)-3,4-DIHYDROQUINAZOLIN-2(1H)-ONE",
            "non-polymer",
        ),
        "8TO": (
            "(4E,6E,8S,9R,10E,12R,13R,14S,16R)-19-{[2-(DIMETHYLAMINO)ETHYL]AMINO}-13-HYDROXY-8,14-DIMETHOXY-4,10,12,16,21-PENTAMETHYL-3,20,22-TRIOXO-2-AZABICYCLO[16.3.1]DOCOSA-1(21),4,6,10,18-PENTAEN-9-YL CARBAMATE",
            "non-polymer",
        ),
        "DEX": ("DEXAMETHASONE", "non-polymer"),
        "MCZ": (
            "(1S,3R,5Z,7E,14BETA,17ALPHA,20S)-20-(3-HYDROXY-3-METHYLBUTOXY)-9,10-SECOPREGNA-5,7,10-TRIENE-1,3-DIOL",
            "non-polymer",
        ),
        "PTO": ("PSEUDOTROPINE", "non-polymer"),
        "GFL": (
            "2-AMINO-9-(2-DEOXY-2-FLUORO-5-O-PHOSPHONO-BETA-D-ARABINOFURANOSYL)-1,9-DIHYDRO-6H-PURIN-6-ONE",
            "nucleic-acid",
        ),
        "GFM": (
            "2'-DEOXY-5'-O-[(S)-{[(S)-[(1R)-1-FLUORO-1-PHOSPHONOETHYL](HYDROXY)PHOSPHORYL]OXY}(HYDROXY)PHOSPHORYL]GUANOSINE",
            "non-polymer",
        ),
        "Z5P": ("1-[2-(1H-PYRAZOL-1-YL)PHENYL]METHANAMINE", "non-polymer"),
        "MCW": ("2-METHOXYCYCLOHEXA-2,5-DIENE-1,4-DIONE", "non-polymer"),
        "LKM": (
            "N-({6-[(4-CYANO-2-FLUOROBENZYL)OXY]NAPHTHALEN-2-YL}SULFONYL)-D-GLUTAMIC ACID",
            "non-polymer",
        ),
        "M4C": ("?", "nucleic-acid"),
        "MCV": (
            "5-[2-(2,5-DIMETHOXYPHENYL)ETHYL]THIENO[2,3-D]PYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "DEY": (
            "(2R)-2-({3-[5-HYDROXY-2-(3-METHYLBUTYL)-3-OXO-6-THIOPHEN-2-YL-2,3-DIHYDROPYRIDAZIN-4-YL]-1,1-DIOXIDO-2H-1,2,4-BENZOTHIADIAZIN-7-YL}OXY)PROPANAMIDE",
            "non-polymer",
        ),
        "MCE": ("CHLOROETHYL GROUP", "non-polymer"),
        "TBF": ("TERT-BUTYL FORMATE", "non-polymer"),
        "I92": ("4-(2-BROMOPHENYL)-2,4-DIOXOBUTANOIC ACID", "non-polymer"),
        "MCT": ("4-METHYLCATECHOL", "non-polymer"),
        "ATK": (
            "(1-{6-[6-(CYCLOPENTYLAMINO)-1H-INDAZOL-1-YL]PYRAZIN-2-YL}-1H-PYRROL-3-YL)ACETIC ACID",
            "non-polymer",
        ),
        "VSV": (
            "N-({3-[({4-[(Z)-(2,4-DIOXO-1,3-THIAZOLIDIN-5-YLIDENE)METHYL]PHENYL}AMINO)METHYL]PHENYL}CARBONYL)-D-GLUTAMIC ACID",
            "non-polymer",
        ),
        "GFF": (
            "2'-DEOXY-5'-O-[({[DIFLUORO(PHOSPHONO)METHYL](HYDROXY)PHOSPHORYL}OXY)(HYDROXY)PHOSPHORYL]GUANOSINE",
            "non-polymer",
        ),
        "HQ2": (
            "8-HYDROXY-3-(PIPERAZIN-1-YL)QUINOLINE-5-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "HQ6": (
            "N-[(3R,4S,5R,6R,7R)-3,5,6-TRIHYDROXY-7-(HYDROXYMETHYL)AZEPAN-4-YL]ACETAMIDE",
            "non-polymer",
        ),
        "GFG": (
            "ALPHA-D-GLUCOPYRANOSYL-(1->3)-ALPHA-D-FRUCTOFURANOSYL ALPHA-D-GALACTOPYRANOSIDE",
            "saccharide",
        ),
        "VSC": ("WRR-204", "peptide"),
        "PTF": ("[(METHYLSULFANYL)METHYL]BENZENE", "non-polymer"),
        "PTE": ("TUNGSTOPTERIN COFACTOR", "non-polymer"),
        "VSO": ("VINYLSULPHONIC ACID", "non-polymer"),
        "RYN": (
            "N-ETHYL-N'-(1-METHYLETHYL)-6-(METHYLSULFANYL)-1,3,5-TRIAZINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "EOB": (
            "{1,3-PHENYLENEBIS[IMINO(2-OXOETHANE-2,1-DIYL)]}BIS(PHOSPHONIC ACID)",
            "non-polymer",
        ),
        "EOA": ("N-PHENETHYL-FORMAMIDE", "non-polymer"),
        "JMQ": (
            "7-(((2R,3R,4S)-3,4-DIHYDROXY-2-(HYDROXYMETHYL)PYRROLIDIN-1-YL)METHYL)-3H-PYRROLO[3,2-D]PYRIMIDIN-4(5H)-ONE",
            "non-polymer",
        ),
        "RYJ": ("4-SULFAMOYL-N-(2,2,2-TRIFLUOROETHYL)BENZAMIDE", "non-polymer"),
        "YIS": (
            "1-[3-TERT-BUTYL-1-(3-CHLORO-4-HYDROXYPHENYL)-1H-PYRAZOL-5-YL]-3-{2-[(3-{2-[(2-HYDROXYETHYL)SULFANYL]PHENYL}[1,2,4]TRIAZOLO[4,3-A]PYRIDIN-6-YL)SULFANYL]BENZYL}UREA",
            "non-polymer",
        ),
        "EOD": ("11-[(2R,3S)-3-PENTYLOXIRAN-2-YL]UNDECANOIC ACID", "non-polymer"),
        "EOH": ("ETHANOL", "non-polymer"),
        "EOM": ("ETHYLOXYMETHOXYL", "non-polymer"),
        "EOL": ("2-METHOXY-4-(PROP-2-EN-1-YL)PHENOL", "non-polymer"),
        "QQ7": ("CUCURBIT[7]URIL", "non-polymer"),
        "EOP": (
            "{ETHANE-1,2-DIYLBIS[IMINO(2-OXOETHANE-2,1-DIYL)]}BIS(PHOSPHONIC ACID)",
            "non-polymer",
        ),
        "QQ1": (
            "2-[3-(CYCLOHEXYLMETHYL)-5-PHENYL-IMIDAZOL-4-YL]-[1,3]THIAZOLO[4,5-E]PYRIMIDIN-7-AMINE",
            "non-polymer",
        ),
        "PEX": ("1,2-DIDECANOYL-SN-GLYCERO-3-PHOSPHOETHANOLAMINE", "non-polymer"),
        "EOT": (
            "[(1-[(BIS-CARBOXYMETHYL-AMINO)-METHYL]-2-{4-[3-(2-HYDROXY-ETHYL)-THIOUREIDO]-PHENY}-ETHYL)-CARBOXYMETHYL-AMINO]-ACETIC ACID",
            "non-polymer",
        ),
        "RYV": ("N-ETHYL-4-SULFAMOYLBENZAMIDE", "non-polymer"),
        "EOZ": ("3,5-BIS[(PHOSPHONOACETYL)AMINO]BENZOIC ACID", "non-polymer"),
        "EOX": ("ETHYLOXY GROUP", "non-polymer"),
        "GOO": (
            "(1R,2R,3R,4S,5R)-4-AMINO-5-METHOXYCYCLOPENTANE-1,2,3-TRIOL",
            "non-polymer",
        ),
        "S62": ("1,3-BENZOTHIAZOLE-2,5-DISULFONIC ACID", "non-polymer"),
        "YIY": (
            "(1E)-1-[5-TERT-BUTYL-2-(3-FLUOROPHENYL)-1H-PYRAZOL-3-YLIDENE]-3-(4-PYRIDIN-3-YLOXYPHENYL)UREA",
            "non-polymer",
        ),
        "W61": (
            "1,3-THIAZOLIDIN-3-YL[(2S,4S)-4-{4-[2-(TRIFLUOROMETHYL)QUINOLIN-4-YL]PIPERAZIN-1-YL}PYRROLIDIN-2-YL]METHANONE",
            "non-polymer",
        ),
        "RO0": (
            "1-(3,4-DIMETHOXYBENZYL)-6,7-DIMETHOXY-4-{[4-(2-METHOXYPHENYL)PIPERIDIN-1-YL]METHYL}ISOQUINOLINE",
            "non-polymer",
        ),
        "YIX": (
            "1-ETHYL-3-(2-{[3-(1-METHYLETHYL)[1,2,4]TRIAZOLO[4,3-A]PYRIDIN-6-YL]SULFANYL}BENZYL)UREA",
            "non-polymer",
        ),
        "W": ("TUNGSTEN ION", "non-polymer"),
        "KI2": (
            "3-BENZYLOXYCARBONYLAMINO-2-HYDROXY-4-PHENYL-BUTYRIC ACID",
            "non-polymer",
        ),
        "P79": (
            "1-{3-TERT-BUTYL-1-[2-(1,1-DIOXIDOTHIOMORPHOLIN-4-YL)-2-OXOETHYL]-1H-PYRAZOL-5-YL}-3-NAPHTHALEN-2-YLUREA",
            "non-polymer",
        ),
        "TIY": ("2,3,4,6-TETRAHYDROXY-5H-BENZO[7]ANNULEN-5-ONE", "non-polymer"),
        "PTX": (
            "{[(2E)-2-(2-AMINO-1,3-THIAZOL-4-YL)-2-(METHOXYIMINO)ETHANOYL]AMINO}METHYLPHOSPHONIC ACID",
            "non-polymer",
        ),
        "QUS": (
            "(S)-2-AMINO-3-(3,5-DIOXO-[1,2,4]OXADIAZOLIDIN-2-YL)-PROPIONIC ACID",
            "non-polymer",
        ),
        "FAJ": ("FAD-TRANS-2-PHENYLCYCLOPROPYLAMINE ADDUCT", "non-polymer"),
        "LNA": ("(9E,12Z)-10-NITROOCTADECA-9,12-DIENOIC ACID", "non-polymer"),
        "L2O": ("(2S,3R)-3-AMINO-2-HYDROXY-5-METHYLHEXANOIC ACID", "peptide"),
        "4SH": (
            "N-BENZYL-2-[(MORPHOLIN-4-YLACETYL)AMINO]-4,5,6,7-TETRAHYDRO-1-BENZOTHIOPHENE-3-CARBOXAMIDE",
            "peptide",
        ),
        "P9B": (
            "4-{[(PENTAFLUOROPHENYL)CARBAMOYL]AMINO}BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "PTS": (
            "(4S-TRANS)-4-(AMINO)-5,6-DIHYDRO-6-METHYL-4H-THIENO (2,3-B)THIOPYRAN-2-SULFONAMIDE-7,7-DIOXIDE",
            "non-polymer",
        ),
        "TGE": (
            "7-(6-METHOXYPYRIDIN-3-YL)-3-{[2-(MORPHOLIN-4-YL)ETHYL]AMINO}-1-(2-PROPOXYETHYL)PYRIDO[3,4-B]PYRAZIN-2(1H)-ONE",
            "non-polymer",
        ),
        "PTR": ("O-PHOSPHOTYROSINE", "peptide"),
        "QUY": (
            "(2S,3R,4S,5R,6S)-6-[(2S,3S,4R)-2-(HEXACOSANOYLAMINO)-3,4-DIHYDROXY-OCTADECOXY]-3,4,5-TRIHYDROXY-N-(PHENYLMETHYL)OXANE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "QQQ": ("1-(1-METHYLETHYL)-1H-BENZIMIDAZOLE-2-SULFONIC ACID", "non-polymer"),
        "RE3": (
            "3-[(3S)-3-HYDROXY-2-OXO-2,3-DIHYDRO-1H-INDOL-3-YL]-L-ALANINE",
            "peptide",
        ),
        "EO9": (
            "[(1S)-1-[(2,6-DIFLUOROPHENYL)CARBONYLAMINO]ETHYL]-TRIHYDROXY-BORON",
            "non-polymer",
        ),
        "PTP": ("THYMIDINE-3',5'-DIPHOSPHATE", "non-polymer"),
        "SQ": ("3-ETHYLAMINO-4-METHYLAMINO-CYCLOBUTANE-1,2-DIONE", "non-polymer"),
        "CMO": ("CARBON MONOXIDE", "non-polymer"),
        "CMN": (
            "(1R,2R,3S,6R)-6-AMINO-4-(HYDROXYMETHYL)CYCLOHEX-4-ENE-1,2,3-TRIOL",
            "non-polymer",
        ),
        "CMM": (
            "2-[2-(1-CARBAMIMIDOYL-PIPERIDIN-3-YL)-ACETYLAMINO]-3-{4-[2-(3-OXALYL-1H-INDOL-7-YL)ETHYL]-PHENYL}-PROPIONIC ACID METHYL ESTER",
            "non-polymer",
        ),
        "43M": ("4-CHLORO-3-METHYLPHENOL", "non-polymer"),
        "CMK": ("CMP-2-KETO-3-DEOXY-OCTULOSONIC ACID", "non-polymer"),
        "CMJ": ("(20S)-20-METHYLDOTETRACONTANE", "non-polymer"),
        "CMI": (
            "1-(3-CARBAMIMIDOYL-BENZYL)-4-METHYL-1H-INDOLE-2-CARBOXYLIC ACID 3,5-DIMETHYL-BENZYLAMIDE",
            "non-polymer",
        ),
        "CMH": ("S-(METHYLMERCURY)-L-CYSTEINE", "peptide"),
        "43F": (
            "2,4-DICHLORO-1-[4-(CHLOROMETHYL)-2-METHOXYPHENOXY]BENZENE",
            "non-polymer",
        ),
        "CMF": (
            "3-CYCLOHEXYL-1-(2-MORPHOLIN-4-YL-2-OXOETHYL)-2-PHENYL-1H-INDOLE-6-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "CME": ("S,S-(2-HYDROXYETHYL)THIOCYSTEINE", "peptide"),
        "43E": (
            "1-{3-[(3,5-DICHLOROBENZYL)AMINO]PROPYL}-3-THIOPHEN-3-YLUREA",
            "non-polymer",
        ),
        "43B": (
            "3-NITRO-N-{4-[2-(2-PHENYLETHYL)-1,3-BENZOTHIAZOL-5-YL]BENZOYL}-4-{[2-(PHENYLSULFANYL)ETHYL]AMINO}BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "CMB": (
            "1-(3-CARBAMIMIDOYLBENZYL)-N-(3,5-DICHLOROBENZYL)-4-METHYL-1H-INDOLE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "CMA": ("N2-(CARBOXYETHYL)-L-ARGININE", "non-polymer"),
        "43A": (
            "(3Z)-6-(4-HYDROXY-3-METHOXYPHENYL)-3-(1H-PYRROL-2-YLMETHYLENE)-1,3-DIHYDRO-2H-INDOL-2-ONE",
            "non-polymer",
        ),
        "F05": ("2H-ISOINDOL-1-AMINE", "non-polymer"),
        "F04": ("2-ETHYL-4-METHYL-PHTHALAZIN-1-ONE", "non-polymer"),
        "F07": ("THIENO[3,2-C][1,6]NAPHTHYRIDINE", "non-polymer"),
        "F06": ("2-CHLOROBENZYL CARBAMIMIDOTHIOATE", "non-polymer"),
        "F01": (
            "4-AMINO-1-[(2R,3R,4S,5R)-3,4-DIHYDROXY-5-(HYDROXYMETHYL)OXOLAN-2-YL]-5-FLUORO-PYRIMIDIN-2-ONE",
            "non-polymer",
        ),
        "CMZ": (
            "(2S)-1-(2,5-DIMETHYLPHENOXY)-3-MORPHOLIN-4-YLPROPAN-2-OL",
            "non-polymer",
        ),
        "CMY": ("(6-AMINOHEXYL)CARBAMIC ACID", "non-polymer"),
        "CMX": ("CARBOXYMETHYLDETHIA COENZYME *A", "non-polymer"),
        "CMW": (
            "3-{[(4-METHYLPHENYL)SULFONYL]AMINO}PROPYL PYRIDIN-4-YLCARBAMATE",
            "non-polymer",
        ),
        "CMV": (
            "(2R,4S)-2-[(1R)-1-({(2R)-2-[(4-ETHYL-2,3-DIOXOPIPERAZIN-1-YL)AMINO]-2-PHENYLACETYL}AMINO)-2-OXOETHYL]-5,5-DIMETHYL-1,3-THIAZOLIDINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "CMU": ("5-CHLORO-6-(1-(2-IMINOPYRROLIDINYL) METHYL) URACIL", "non-polymer"),
        "CMT": ("O-METHYLCYSTEINE", "peptide"),
        "CMS": ("CARBAMOYL SARCOSINE", "non-polymer"),
        "CMR": ("2'-DEOXY-CYTIDINE-5'-RP-MONOMETHYLPHOSPHONATE", "nucleic-acid"),
        "43P": (
            "4-(1H-IMIDAZOL-4-YL)-3-(5-ETHYL-2,4-DIHYDROXY-PHENYL)-1H-PYRAZOLE",
            "non-polymer",
        ),
        "CMP": ("ADENOSINE-3',5'-CYCLIC-MONOPHOSPHATE", "non-polymer"),
        "M2T": ("5'-DEOXY-5'-(DIMETHYL-LAMBDA~4~-SULFANYL)ADENOSINE", "non-polymer"),
        "M2W": (
            "DELTA-(L-ALPHA-AMINOADIPOYL)-L-CYSTEINYL-O-METHYL-D-THREONINE",
            "non-polymer",
        ),
        "M2P": ("D-MANNITOL-1,6-DIPHOSPHATE", "non-polymer"),
        "M2S": ("3-{[(R)-METHYLSULFINYL]METHYL}-L-VALINE", "peptide"),
        "M2L": (
            "(2R)-2-AMINO-3-(2-DIMETHYLAMINOETHYLSULFANYL)PROPANOIC ACID",
            "peptide",
        ),
        "M2M": ("1-METHOXY-2-(2-METHOXYETHOXY)ETHANE", "non-polymer"),
        "W2R": (
            "1-{3-[2-CHLORO-4-({5-[2-(2-HYDROXYETHOXY)ETHYL]-5H-PYRROLO[3,2-D]PYRIMIDIN-4-YL}AMINO)PHENOXY]PHENYL}-3-CYCLOHEXYLUREA",
            "non-polymer",
        ),
        "M2F": (
            "2,4-DINITROPHENYL 2-DEOXY-2-FLUORO-BETA-D-MANNOPYRANOSIDE",
            "non-polymer",
        ),
        "M2G": ("N2-DIMETHYLGUANOSINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "ILY": ("N~6~-PROPAN-2-YL-L-LYSINE", "non-polymer"),
        "M2C": (
            "(2S)-2-AMINO-4-(METHYLSULFANYL)-1-PYRIDIN-2-YLBUTANE-1,1-DIOL",
            "non-polymer",
        ),
        "M28": ("3-[4-(AMINOSULFONYL)PHENYL]PROPANOIC ACID", "non-polymer"),
        "M29": ("ETHYL 3-[4-(AMINOSULFONYL)PHENYL]PROPANOATE", "non-polymer"),
        "NC7": ("7-{[(CYCLOHEXYLAMINO)CARBONYL]AMINO}HEPTANOIC ACID", "non-polymer"),
        "M24": (
            "5'-O-[(R)-[({3-[5-(2-FLUOROPHENYL)-1,2,4-OXADIAZOL-3-YL]PHENYL}CARBONYL)OXY](HYDROXY)PHOSPHORYL]ADENOSINE",
            "non-polymer",
        ),
        "M25": ("N-{2-[4-(AMINOSULFONYL)PHENYL]ETHYL}ACETAMIDE", "non-polymer"),
        "M21": ("(10E,12Z)-TETRADECA-10,12-DIEN-1-OL", "non-polymer"),
        "M22": (
            "3-CHLORO-5-[2-CHLORO-5-(1H-INDAZOL-3-YLMETHOXY)PHENOXY]BENZONITRILE",
            "non-polymer",
        ),
        "M23": (
            "7-(5-BROMO-2-ETHOXYPHENYL)-6-METHYLQUINAZOLINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "QUI": ("2-CARBOXYQUINOXALINE", "non-polymer"),
        "SB": ("ANTIMONY (III) ION", "non-polymer"),
        "NVG": (
            "4-AMINO-1-(5-{[3-(1H-BENZIMIDAZOL-2-YL)PROPANOYL]AMINO}-5-DEOXY-ALPHA-L-LYXOFURANOSYL)PYRIMIDIN-2(1H)-ONE",
            "non-polymer",
        ),
        "UHF": (
            "N-((2R,4S)-2-BUTYL-5-METHYL-4-(3-(5-METHYLPYRIDIN-2-YL)UREIDO)-3-OXOHEXYL)-N-HYDROXYFORMAMIDE",
            "non-polymer",
        ),
        "PC8": ("1,2-DIOCTANOYL-SN-GLYCERO-3-PHOSPHOCHOLINE", "non-polymer"),
        "PC9": (
            "(7R,14S)-14,15-DIBROMO-4-HYDROXY-N,N,N-TRIMETHYL-9-OXO-7-[(PALMITOYLOXY)METHYL]-3,5,8-TRIOXA-4-PHOSPHAHEXACOSAN-1-AMINIUM 4-OXIDE",
            "non-polymer",
        ),
        "INX": (
            "2-(CARBOXYMETHOXY)-5-[(2S)-2-({(2S)-2-[(3-CARBOXYPROPANOYL)AMINO] -3-PHENYLPROPANOYL}AMINO)-3-OXO-3-(PENTYLAMINO)PROPYL]BENZOIC ACID",
            "non-polymer",
        ),
        "INY": ("CARPROPAMIDE", "non-polymer"),
        "INZ": (
            "2-{4-[(2S)-2-[({[(1S)-1-CARBOXY-2-PHENYLETHYL]AMINO}CARBONYL)AMINO]-3-OXO-3-(PENTYLAMINO)PROPYL]PHENOXY}MALONIC ACID",
            "non-polymer",
        ),
        "INT": (
            "[[PHENYLMETHYLOXY-CARBONYL]-ALANINYL]-VALINYL-[PHENYL-1-HYDROXYPROP-2-YL]-AMINE",
            "non-polymer",
        ),
        "INU": (
            "N-(3-CYCLOPROPYL(5,6,7,8,9,10-HEXAHYDRO-2-OXO-2H-CYCLOOCTA[B]PYRAN-3-YL)METHYL)PHENYLBENZENSULFONAMIDE",
            "non-polymer",
        ),
        "INV": (
            "4-(AMINOSULFONYL)-N-[(4-FLUOROPHENYL)METHYL]-BENZAMIDE",
            "non-polymer",
        ),
        "INW": (
            "4-(AMINOSULFONYL)-N-[(2,4-DIFLUOROPHENYL)METHYL]-BENZAMIDE",
            "non-polymer",
        ),
        "INP": (
            "4-[[N-[(PHENYLMETHOXY)CARBONYL]-/NL/N-LEUCYL]AMINO]-1[(2S)-2-[[[4-(PYRIDINYLMETHOXY)CARBONYL]AMINO]-4-METHYLPENT/NYL]-3-PYRROLIDINONE/N",
            "non-polymer",
        ),
        "INQ": (
            "6-[N-(3-HYDROXY-PHENYL)-3-(MORPHOLIN-4-YLMETHYL)-2H-THIENO[3,2-E]-1,2-THIAZINE-1,1,-DIOXIDE]-SULFONAMIDE",
            "non-polymer",
        ),
        "INR": ("INDIRUBIN-5-SULPHONATE", "non-polymer"),
        "INS": ("1,2,3,4,5,6-HEXAHYDROXY-CYCLOHEXANE", "non-polymer"),
        "INL": (
            "6-[N-(3-METHOXY-PHENYL)-3-(MORPHOLIN-4-YLMETHYL)-2H-THIENO[3,2-E]-1,2-THIAZINE-1,1,-DIOXIDE]-SULFONAMIDE",
            "non-polymer",
        ),
        "INM": (
            "4-AMINO-6-[N-(3-METHOXYLPROPYL)-2H-THIENO[3,2-E][1,2]THIAZINE 1,1-DIOXIDE]-SULFONAMIDE",
            "non-polymer",
        ),
        "INN": (
            "N-{(2R)-2-[2-(HYDROXYAMINO)-2-OXOETHYL]-4-METHYLPENTANOYL}-3-METHYL-L-VALYL-N-(2-AMINOETHYL)-L-ALANINAMIDE",
            "peptide",
        ),
        "INO": ("2-HYDROXYISONICOTINIC ACID N-OXIDE", "non-polymer"),
        "INH": (
            "N-(R-CARBOXY-ETHYL)-ALPHA-(S)-(2-PHENYLETHYL)GLYCYL-L-ARGININE-N-PHENYLAMIDE",
            "non-polymer",
        ),
        "INI": ("5-NITRO-6-RIBITYL-AMINO-2,4(1H,3H)-PYRIMIDINEDIONE", "non-polymer"),
        "438": ("?", "non-polymer"),
        "439": (
            "6-CHLORO-N-{(3S)-1-[(1S)-1-(DIMETHYLAMINO)-2,3-DIHYDRO-1H-INDEN-5-YL]-2-OXO-3-PYRROLIDINYL}-2-NAPHTHALENESULFONAMIDE",
            "non-polymer",
        ),
        "CM7": (
            "N-[(1S,2R)-1-BENZYL-2-HYDROXY-3-{[3-(TRIFLUOROMETHYL)BENZYL]AMINO}PROPYL]-7-ETHYL-1-METHYL-3,4-DIHYDRO-1H-[1,2,5]THIADIAZEPINO[3,4,5-HI]INDOLE-9-CARBOXAMIDE 2,2-DIOXIDE",
            "non-polymer",
        ),
        "437": (
            "1-{3-TERT-BUTYL-1-[4-(HYDROXYMETHYL)PHENYL]-1H-PYRAZOL-5-YL}-3-NAPHTHALEN-1-YLUREA",
            "non-polymer",
        ),
        "434": ("BILH 434", "non-polymer"),
        "CM4": (
            "(2R,3R,4S)-3-(4-HYDROXYPHENYL)-4-METHYL-2-[4-(2-PYRROLIDIN-1-YLETHOXY)PHENYL]CHROMAN-6-OL",
            "non-polymer",
        ),
        "CM3": (
            "(2R,3R,4S)-5-FLUORO-3-(4-HYDROXYPHENYL)-4-METHYL-2-[4-(2-PIPERIDIN-1-YLETHOXY)PHENYL]CHROMAN-6-OL",
            "non-polymer",
        ),
        "433": (
            "2-(4-{3-[1-[2-(2-CHLORO-6-FLUORO-PHENYL)-ETHYL]-3-(2,3-DICHLORO-PHENYL)-UREIDO]-PROPYL}-PHENOXY)-2-METHYL-PROPIONIC ACID",
            "non-polymer",
        ),
        "430": (
            "5-(1H-BENZIMIDAZOL-1-YL)-3-{(1R)-1-[2-(TRIFLUOROMETHYL)PHENYL]ETHOXY}THIOPHENE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "431": ("2-(4-METHYLPHENYL)-4H-CHROMEN-4-ONE", "non-polymer"),
        "GLH": ("N-5-CYCLOHEXYL-N-5-[(CYCLOHEXYLAMINO)CARBONYL]GLUTAMINE", "peptide"),
        "740": (
            "N-[5-(1,1-DIOXIDOISOTHIAZOLIDIN-2-YL)-1H-INDAZOL-3-YL]-2-(4-PIPERIDIN-1-YLPHENYL)ACETAMIDE",
            "non-polymer",
        ),
        "AKY": ("ACLACINOMYCIN Y", "non-polymer"),
        "GLK": ("(4S)-4-AMINO-5,5-DIHYDROXYPENTANOIC ACID", "peptide"),
        "GLL": ("GLYCOLURIL", "non-polymer"),
        "GLM": ("1-AMINO-PROPAN-2-ONE", "peptide"),
        "GLN": ("GLUTAMINE", "peptide"),
        "741": (
            "4-[[3-CHLORO-4-(1-METHYLIMIDAZOL-2-YL)SULFANYL-PHENYL]AMINO]-7-[3-(2-HYDROXYETHYL-METHYL-AMINO)PROPOXY]-6-METHOXY-QUINOLINE-3-CARBONITRILE",
            "non-polymer",
        ),
        "MM5": (
            "1,1'-[1,4-PHENYLENEBIS(METHYLENE)]BIS[1,4,8,11-TETRAAZACYCLOTETRADECANE]NI(II)",
            "non-polymer",
        ),
        "AKR": ("ACRYLIC ACID", "non-polymer"),
        "GLB": ("BETA-D-GALACTOSE", "saccharide"),
        "AKP": (
            "N-(PYRAZIN-2-YLCARBONYL)LEUCYLISOLEUCYL-N~1~-{1-[2-({1-CARBOXY-2-[4-(PHOSPHONOOXY)PHENYL]ETHYL}AMINO)-1,1-DIHYDROXY-2-OXOETHYL]BUT-3-ENYL}-3-CYCLOHEXYLALANINAMIDE",
            "non-polymer",
        ),
        "GLD": ("4,6-DIDEOXYGLUCOSE", "saccharide"),
        "AKV": (
            "{3-[(1R,3S)-1,3-DIHYDROXYPENTYL]-4,5,9,10-TETRAHYDROXY-2-ANTHRYL}ACETATE",
            "non-polymer",
        ),
        "GLF": ("1-FLUORO-ALPHA-1-DEOXY-D-GLUCOSE", "saccharide"),
        "AKT": ("10-DECARBOXYMETHYLACLACINOMYCIN T (DCMAT)", "non-polymer"),
        "AKK": ("(1Z)-PROP-1-EN-1-AMINE", "non-polymer"),
        "GLY": ("GLYCINE", "peptide"),
        "AKI": (
            "1-(4-{2-[(5,6-DIPHENYLFURO[2,3-D]PYRIMIDIN-4-YL)AMINO]ETHYL}PHENYL)-3-PHENYLUREA",
            "non-polymer",
        ),
        "AKH": ("1-BENZOFURAN-7-CARBOXYLIC ACID", "non-polymer"),
        "AKN": (
            "(2R)-4-AMINO-N-{(1R,2S,3S,4R,5S)-5-AMINO-2-[(3-AMINO-3-DEOXY-ALPHA-D-GLUCOPYRANOSYL)OXY]-4-[(6-AMINO-6-DEOXY-ALPHA-D-GLUCOPYRANOSYL)OXY]-3-HYDROXYCYCLOHEXYL}-2-HYDROXYBUTANAMIDE",
            "non-polymer",
        ),
        "AKL": ("3-AMINO-5-CHLORO-4-OXOPENTANOIC ACID", "peptide"),
        "AKC": (
            "2-ETHOXYETHYL (1S,2S)-3-{(2S)-4-[(3AS,8S,8AR)-2-OXO-3,3A,8,8A-TETRAHYDRO-2H-INDENO[1,2-D][1,3]OXAZOL-8-YL]-2-BENZYL-3-OXO-2,3-DIHYDRO-1H-PYRROL-2-YL}-1-BENZYL-2-HYDROXYPROPYLCARBAMATE",
            "non-polymer",
        ),
        "AKB": ("2-AMINO-3-KETOBUTYRIC ACID", "non-polymer"),
        "AKA": ("10-DECARBOXYMETHYLACLACINOMYCIN A (DCMAA)", "non-polymer"),
        "GLS": ("BETA-D-GLUCOPYRANOSE SPIROHYDANTOIN", "saccharide"),
        "AKG": ("2-OXOGLUTARIC ACID", "non-polymer"),
        "GLU": ("GLUTAMIC ACID", "peptide"),
        "GLV": ("GLYOXYLIC ACID", "non-polymer"),
        "AKD": ("5-BROMO-2-[(3S)-PIPERIDIN-3-YLAMINO]BENZOATE", "non-polymer"),
        "TZA": ("TRANS-AMICLENOMYCIN", "non-polymer"),
        "F47": (
            "4-[[6-(4-FLUOROPHENYL)-[1,2,4]TRIAZOLO[4,3-B][1,2,4]TRIAZIN-3-YL]METHYL]PHENOL",
            "non-polymer",
        ),
        "BSZ": ("3-CHLORO-4-METHOXYBENZENESULFONIC ACID", "non-polymer"),
        "BST": (
            "6-(4-AMINO-4-HYDROXY-2-OXO-3,4-DIHYDRO-2H-PYRIMIDIN-1-YL)-3-[3-AMINO-5-(N-METHYL-GUANIDINO)-PENT ANOYLAMINO]-3,6-DIHYDRO-2H-PYRAN-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "BSU": ("1,3-DIPHENYLUREA", "non-polymer"),
        "UD5": ("5-FLUOROURIDINE", "nucleic-acid"),
        "BSP": ("BROMOSULFALEIN", "non-polymer"),
        "PT2": (
            "[ETHANE-1,2-DIAMINATO(2-)-KAPPA~2~N,N'](1,10-PHENANTHROLINE-KAPPA~2~N~1~,N~10~)PLATINUM(2+)",
            "non-polymer",
        ),
        "BSM": (
            "5-(5-CHLORO-2,4-DIHYDROXYPHENYL)-N-ETHYL-4-(4-METHOXYPHENYL)-1H-PYRAZOLE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "BSO": ("BIOTIN-D-SULFOXIDE", "non-polymer"),
        "BSH": (
            "3-[(2R)-2-[(BENZYLSULFONYL)AMINO]-2-(DIHYDROXYBORANYL)ETHYL]BENZOIC ACID",
            "non-polymer",
        ),
        "BSI": (
            "2-(BIPHENYL-4-SULFONYL)-1,2,3,4-TETRAHYDRO-ISOQUINOLINE-3-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "BSJ": ("?", "non-polymer"),
        "CAR": ("?", "nucleic-acid"),
        "BSD": (
            "N-{(1S,2R)-1-BENZYL-2-HYDROXY-3-[(3-METHOXYBENZYL)AMINO]PROPYL}-5-[METHYL(METHYLSULFONYL)AMINO]-N'-[(1R)-1-PHENYLETHYL]BENZENE-1,3-DICARBOXAMIDE",
            "non-polymer",
        ),
        "BSE": ("BETA-3-SERINE", "peptide"),
        "BSF": ("{[(BENZYLSULFONYL)AMINO]METHYL}BORONIC ACID", "non-polymer"),
        "BSG": (
            "4-({[(DIHYDROXYBORANYL)METHYL]SULFAMOYL}METHYL)BENZOIC ACID",
            "non-polymer",
        ),
        "BSA": ("2-(BENZOYLAMINO)ETHANESULFONIC ACID", "non-polymer"),
        "BSB": ("N-BENZYL-4-SULFAMOYL-BENZAMIDE", "non-polymer"),
        "BSC": ("(S)-2-AMINO-4-(AMINO-BUTYLHYDROSULFINYL)BUTANOIC ACID", "non-polymer"),
        "IBC": (
            "3-(3-HYDROXY-7,8-DIHYDRO-6H-CYCLOHEPTA[D]ISOXAZOL-4-YL)-L-ALANINE",
            "non-polymer",
        ),
        "F": ("FLUORIDE ION", "non-polymer"),
        "BS4": ("N-[(4-AMINOPHENYL)SULFONYL]BENZAMIDE", "non-polymer"),
        "BS5": (
            "5'-DEOXY-5'-[({5-[(3AS,4S,6AR)-2-OXOHEXAHYDRO-1H-THIENO[3,4-D]IMIDAZOL-4-YL]PENTANOYL}SULFAMOYL)AMINO]ADENOSINE",
            "non-polymer",
        ),
        "VNP": (
            "3-[(4-AMINO-2-METHYLPYRIMIDIN-5-YL)METHYL]-5-(2-{[HYDROXY(PHOSPHONOAMINO)PHOSPHORYL]OXY}ETHYL)-4-METHYL-1,3-THIAZOL-3-IUM",
            "non-polymer",
        ),
        "BS1": (
            "(3-EXO)-3-(10,11-DIHYDRO-5H-DIBENZO[A,D][7]ANNULEN-5-YLOXY)-8,8-DIMETHYL-8-AZONIABICYCLO[3.2.1]OCTANE",
            "non-polymer",
        ),
        "BS2": (
            "(3-ENDO,8-ANTI)-8-BENZYL-3-(10,11-DIHYDRO-5H-DIBENZO[A,D][7]ANNULEN-5-YLOXY)-8-AZONIABICYCLO[3.2.1]OCTANE",
            "non-polymer",
        ),
        "BS3": ("BISMUTH(III) ION", "non-polymer"),
        "IBF": ("1,2,3,4,5-PENTAFLUORO-6-IODOBENZENE", "non-polymer"),
        "MPB": ("4-HYDROXY-BENZOIC ACID METHYL ESTER", "non-polymer"),
        "IBG": ("GAMMA-GLUTAMYL[S-(2-IODOBENZYL)CYSTEINYL]GLYCINE", "non-polymer"),
        "SBR": (
            "(R)-N-(3-INDOL-1-YL-2-METHYL-PROPYL)-4-SULFAMOYL-BENZAMIDE",
            "non-polymer",
        ),
        "SBS": (
            "(S)-N-(3-INDOL-1-YL-2-METHYL-PROPYL)-4-SULFAMOYL-BENZAMIDE",
            "non-polymer",
        ),
        "SBP": ("[3-(1,3,2-DIOXABOROLAN-2-YLOXY)PROPYL]GUANIDINE", "non-polymer"),
        "AK8": (
            "1-[3-METHYL-4-({3-[2-(METHYLAMINO)PYRIMIDIN-4-YL]PYRIDIN-2-YL}OXY)PHENYL]-3-[3-(TRIFLUOROMETHYL)PHENYL]UREA",
            "non-polymer",
        ),
        "MMY": (
            "5-(2-MORPHOLIN-4-YL-7-PYRIDIN-3-YL-6,7-DIHYDRO-5H-PYRROLO[2,3-D]PYRIMIDIN-4-YL)-PYRIMIDIN-2-YLAMINE",
            "non-polymer",
        ),
        "SBT": ("2-BUTANOL", "non-polymer"),
        "MMZ": ("1-METHYL-1,3-DIHYDRO-2H-IMIDAZOLE-2-THIONE", "non-polymer"),
        "AK3": (
            "1-(5-{2-[(1-METHYL-1H-PYRAZOLO[4,3-D]PYRIMIDIN-7-YL)AMINO]ETHYL}-1,3-THIAZOL-2-YL)-3-[3-(TRIFLUOROMETHYL)PHENYL]UREA",
            "non-polymer",
        ),
        "AK2": (
            "1-(3-CHLOROPHENYL)-3-{5-[2-(THIENO[3,2-D]PYRIMIDIN-4-YLAMINO)ETHYL]-1,3-THIAZOL-2-YL}UREA",
            "non-polymer",
        ),
        "AK1": (
            "1-{5-[2-(THIENO[3,2-D]PYRIMIDIN-4-YLAMINO)ETHYL]-1,3-THIAZOL-2-YL}-3-[3-(TRIFLUOROMETHYL)PHENYL]UREA",
            "non-polymer",
        ),
        "MMV": (
            "3-(2-{3-[(2,4-DIAMINO-6-ETHYLPYRIMIDIN-5-YL)OXY]PROPOXY}PHENYL)PROPANOIC ACID",
            "non-polymer",
        ),
        "AK7": (
            "1-(5-{2-[(6-AMINO-5-BROMOPYRIMIDIN-4-YL)AMINO]ETHYL}-1,3-THIAZOL-2-YL)-3-[3-(TRIFLUOROMETHYL)PHENYL]UREA",
            "non-polymer",
        ),
        "AK6": (
            "4-METHOXY-N-PHENYL-3-({3-[(1H-PYRROLO[2,3-B]PYRIDIN-5-YLMETHYL)SULFANYL]PROPANOYL}AMINO)BENZAMIDE",
            "non-polymer",
        ),
        "AK5": (
            "3-({3-[(6-AMINO-5-BROMOPYRIMIDIN-4-YL)SULFANYL]PROPANOYL}AMINO)-4-METHOXY-N-PHENYLBENZAMIDE",
            "non-polymer",
        ),
        "AK4": (
            "(7-{[2-(2-{[(3-CHLOROPHENYL)CARBAMOYL]AMINO}-1,3-THIAZOL-5-YL)ETHYL]AMINO}-1H-PYRAZOLO[4,3-D]PYRIMIDIN-1-YL)ACETIC ACID",
            "non-polymer",
        ),
        "GL8": ("GALANTHAMINE DERIVATIVE", "non-polymer"),
        "GL9": (
            "N-(8,9,10-TRIHYDROXY-7-HYDROXYMETHYL-2,4-DIOXO-6-OXA-1,3-DIAZA-SPIRO[4.5]DEC-3-YL-ACETAMIDE",
            "saccharide",
        ),
        "MMO": ("N~2~-METHYL-L-ARGININE", "peptide"),
        "MMN": ("5-DEOXY-5-{[(1S)-1-HYDROXYETHYL]AMINO}-D-GLUCITOL", "saccharide"),
        "MMI": (
            "N-[(4S,5S,7R)-8-({(S)-1-[(BENZYLAMINO)OXOMETHYL]-2-METHYLPROPYL}AMINO)-5-HYDROXY-2,7-DIMETHYL-8-OXO-OCT-4-YL]-(4S,7S)-4-ISOPROPYL-2,5,9-TRIOXO-1-OXA-3,6,10-TRIAZACYCLOHEXADECANE-7-CARBOXAMIDE",
            "non-polymer",
        ),
        "MMH": (
            "N-[3-(ACETYLAMINO)PHENYL]-5-{(2E)-2-[(4-METHOXYPHENYL)METHYLIDENE]HYDRAZINO}-3-METHYL-1H-PYRAZOLE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "SBD": ("D-NAPHTHYL-1-ACETAMIDO BORONIC ACID ALANINE", "peptide"),
        "MMJ": (
            "N-METHYL-1-(4-METHYL-2-PHENYL-1,3-THIAZOL-5-YL)METHANAMINE",
            "non-polymer",
        ),
        "GL0": ("BETA-D-GULOPYRANOSE", "saccharide"),
        "GL1": ("1-O-PHOSPHONO-ALPHA-D-GALACTOPYRANOSE", "non-polymer"),
        "GL2": (
            "3-AMINO-8,9,10-TRIHYDROXY-7-HYDROXYMETHYL-6-OXA-1,3-DIAZA-SPIRO[4.5]DECANE-2,4-DIONE",
            "saccharide",
        ),
        "GL3": ("THIOGLYCIN", "peptide"),
        "GL4": (
            "8,9,10-TRIHYDROXY-7-HYDROXYMETHYL-2-THIOXO-6-OXA-1,3-DIAZA-SPIRO[4.5]DECAN-4-ONE",
            "non-polymer",
        ),
        "GL5": (
            "3,8,9,10-TETRAHYDROXY-7-HYDROXYMETHYL-6-OXA-1,3-DIAZA-SPIRO[4.5]DECANE-2,4-DIONE",
            "saccharide",
        ),
        "GL6": (
            "(3,4,5-TRIHYDROXY-6-HYDROXYMETHYL-TETRAHYDRO-PYRAN-2-YL)-PHOSPHORAMIDIC ACID DIMETHYL ESTER",
            "saccharide",
        ),
        "GL7": (
            "8,9,10-TRIHYDROXY-7-HYDROXYMETHYL-3-METHYL-6-OXA-1,3-DIAZA-SPIRO[4.5]DECANE-2,4-DIONE",
            "saccharide",
        ),
        "ERT": (
            "METHYL (4R)-2-ETHYL-2,5,12-TRIHYDROXY-7-METHOXY-6,11-DIOXO-4-{[2,3,6-TRIDEOXY-3-(DIMETHYLAMINO)-BETA-D-RIBO-HEXOPYRANOSYL]OXY}-1H,2H,3H,4H,6H,11H-TETRACENE-1-CARBOXYLATE",
            "non-polymer",
        ),
        "ERU": ("(13Z)-DOCOS-13-ENAMIDE", "non-polymer"),
        "KOJ": ("5-HYDROXY-2-(HYDROXYMETHYL)-4H-PYRAN-4-ONE", "non-polymer"),
        "KOM": ("7,8-DIHYDROXY-4-PHENYL-2H-CHROMEN-2-ONE", "non-polymer"),
        "KOL": (
            "(4R,5S)-3-{[(3S,5S)-5-(DIMETHYLCARBAMOYL)PYRROLIDIN-3-YL]SULFANYL}-5-[(2S,3R)-3-HYDROXY-1-OXOBUTAN-2-YL]-4-METHYL-4,5-DIHYDRO-1H-PYRROLE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "MPJ": ("(1-AMINO-3-METHYLSULFANYL-PROPYL)-PHOSPHINIC ACID", "peptide"),
        "QL2": (
            "6-[(3,5-DIFLUOROPHENYL)AMINO]-9-ETHYL-9H-PURINE-2-CARBONITRILE",
            "non-polymer",
        ),
        "ERY": ("ERYTHROMYCIN A", "non-polymer"),
        "MPK": ("(5-METHYL-3-PHENYL-1,2-OXAZOL-4-YL)METHANOL", "non-polymer"),
        "ERD": (
            "(2S)-2-(3,4-DIHYDROXYPHENYL)-5,7-DIHYDROXY-2,3-DIHYDRO-4H-CHROMEN-4-ONE",
            "non-polymer",
        ),
        "ERE": (
            "(1R,3S,4R,5S)-3-AMINO-2,3,6-TRIDEOXY-3-METHYL-ALPHA-L-ARABINO-HEXOPYRANOSE",
            "saccharide",
        ),
        "ERG": ("ERGOSTEROL", "non-polymer"),
        "ERA": (
            "(7Z,15E,17E)-N-[(2S,3S,4E)-1,3-DIHYDROXYOCTADEC-4-EN-2-YL]TRICOSA-7,15,17-TRIENAMIDE",
            "non-polymer",
        ),
        "RDP": ("ALPHA-D-RIBOSE-1-PHOSPHATE", "saccharide"),
        "ERC": (
            "8-HYDROXY-5-[(1R)-1-HYDROXY-2-({2-[3-METHOXY-4-(3-SULFANYLPROPOXY)PHENYL]ETHYL}AMINO)ETHYL]QUINOLIN-2(1H)-ONE",
            "non-polymer",
        ),
        "ERM": ("ERGOTAMINE", "non-polymer"),
        "ERN": ("ERYTHROMYCYLAMINE", "non-polymer"),
        "KOR": ("L-HOMOCYSTEINE-S-N-S-L-CYSTEINE", "peptide"),
        "KOU": (
            "(E)-N-({3-HYDROXY-2-METHYL-5-[(PHOSPHONOOXY)METHYL]PYRIDIN-4-YL}METHYLIDENE)-L-SERINE",
            "non-polymer",
        ),
        "ERI": (
            "4-O-ACETYL-2,6-DIDEOXY-3-C-METHYL-BETA-L-ARABINO-HEXOPYRANOSE",
            "saccharide",
        ),
        "MPT": ("BETA-MERCAPTOPROPIONIC ACID", "non-polymer"),
        "IBU": ("2-METHYLPROPYL HYDROGEN CARBONATE", "non-polymer"),
        "K88": (
            "(3Z)-N,N-DIMETHYL-2-OXO-3-(4,5,6,7-TETRAHYDRO-1H-INDOL-2-YLMETHYLIDENE)-2,3-DIHYDRO-1H-INDOLE-5-SULFONAMIDE",
            "non-polymer",
        ),
        "MPU": ("D-MONAPTERIN", "non-polymer"),
        "K83": (
            "3-HYDROXY-N-[(2S,3R)-3-HYDROXY-4-{[(4-METHOXYPHENYL)SULFONYL](2-METHYLPROPYL)AMINO}-1-PHENYLBUTAN-2-YL]BENZAMIDE",
            "non-polymer",
        ),
        "K84": (
            "(9Z)-1-(5-PYRIDIN-2-YL-1,3,4-OXADIAZOL-2-YL)OCTADEC-9-EN-1-ONE",
            "non-polymer",
        ),
        "K85": (
            "PROPAN-2-YL 3-{3-[3-METHOXY-5-(TRIFLUOROMETHYL)PHENYL]-1H-1,2,4-TRIAZOL-1-YL}PROPANOATE",
            "non-polymer",
        ),
        "M6P": ("ALPHA-D-MANNOSE-6-PHOSPHATE", "saccharide"),
        "W94": (
            "[(2S,4S)-4-{4-[1-PHENYL-3-(TRIFLUOROMETHYL)-1H-PYRAZOL-5-YL]PIPERIDIN-1-YL}PYRROLIDIN-2-YL](1,3-THIAZOLIDIN-3-YL)METHANONE",
            "non-polymer",
        ),
        "3YH": (
            "3-(3-CARBAMIMIDOYLPHENYL)-N-(2'-SULFAMOYLBIPHENYL-4-YL)-1,2-OXAZOLE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "MDX": ("11-MERCAPTOUNDECANOIC ACID", "non-polymer"),
        "K8X": ("4,5,6,7-TETRAIODO-1H-BENZIMIDAZOLE", "non-polymer"),
        "RDR": (
            "4-[(6-CHLORO-1-BENZOTHIEN-2-YL)SULFONYL]-1-{[1-(2-HYDROXYETHYL)-1H-PYRROLO[3,2-C]PYRIDIN-2-YL]METHYL}PIPERAZIN-2-ONE",
            "non-polymer",
        ),
        "3YP": ("3-CYCLOPENTYL-N-HYDROXYPROPANAMIDE", "non-polymer"),
        "3YY": (
            "(5R)-5-[3-(4-FLUOROPHENYL)-1-PHENYL-1H-PYRAZOL-4-YL]IMIDAZOLIDINE-2,4-DIONE",
            "non-polymer",
        ),
        "PGC": ("O-PHOSPHOGLYCOLOHYDROXAMATE", "non-polymer"),
        "QLZ": (
            "5-[3-(2,5-DIMETHOXYPHENYL)PROP-1-YN-1-YL]-6-PROPYLPYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "ER3": ("ERBIUM (III) ION", "non-polymer"),
        "QLR": (
            "6-METHYL-5-[(3R)-3-(3,4,5-TRIMETHOXYPHENYL)PENT-1-YN-1-YL]PYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "QLP": (
            "N~2~-({3-HYDROXY-2-METHYL-5-[(PHOSPHONOOXY)METHYL]PYRIDIN-4-YL}METHYL)-L-GLUTAMINE",
            "non-polymer",
        ),
        "UDS": ("N-[(3-HYDROXYQUINOLIN-2-YL)CARBONYL]-D-SERINE", "peptide"),
        "GGY": (
            "4-(2-AMINO-5-{4-[(DIMETHYLAMINO)METHYL]THIOPHEN-2-YL}PYRIDIN-3-YL)-2-{[(1R,2Z)-4,4,4-TRIFLUORO-1-METHYLBUT-2-EN-1-YL]OXY}BENZAMIDE",
            "non-polymer",
        ),
        "KO1": ("D-GLYCERO-ALPHA-D-TALO-OCT-2-ULOPYRANOSONIC ACID", "saccharide"),
        "QLG": (
            "[(4Z)-2-[(1Z)-4-AMINO-4-OXOBUTANIMIDOYL]-4-(2-METHYLPROPYLIDENE)-5-OXO-4,5-DIHYDRO-1H-IMIDAZOL-1-YL]ACETIC ACID",
            "peptide",
        ),
        "N4D": (
            "N,4-DIMETHYL-3-[(1-PHENYL-1H-PYRAZOLO[3,4-D]PYRIMIDIN-4-YL)AMINO]BENZAMIDE",
            "non-polymer",
        ),
        "KO2": (
            "PROP-2-EN-1-YL D-GLYCERO-ALPHA-D-TALO-OCT-2-ULOPYRANOSIDONIC ACID",
            "saccharide",
        ),
        "KO4": ("POTASSIUM ION, 4 WATERS COORDINATED", "non-polymer"),
        "M6R": ("5-AMINO-5-DEOXY-1-O-PHOSPHONO-D-MANNITOL", "non-polymer"),
        "PGJ": (
            "2-(ETHOXYMETHYL)-4-(4-FLUOROPHENYL)-3-[2-(2-HYDROXYPHENOXY)PYRIMIDIN-4-YL]ISOXAZOL-5(2H)-ONE",
            "non-polymer",
        ),
        "UDP": ("URIDINE-5'-DIPHOSPHATE", "nucleic-acid"),
        "RHM": (
            "DELTA-ALPHA-RH[2R,9R-DIAMINO-4,7-DIAZADECANE]9,10-PHENANTHRENEQUINONE DIIMINE",
            "non-polymer",
        ),
        "GGX": ("?", "non-polymer"),
        "LNB": ("(9Z,12E)-12-NITROOCTADECA-9,12-DIENOIC ACID", "non-polymer"),
        "PGI": (
            "(5R,6R,7S,8S)-5-(HYDROXYMETHYL)-2-(2-PHENYLETHYL)-1,5,6,7,8,8A-HEXAHYDROIMIDAZO[1,2-A]PYRIDINE-6,7,8-TRIOL",
            "non-polymer",
        ),
        "M6S": (
            "(2S)-N-[(4-CARBAMIMIDOYLPHENYL)METHYL]-1-[(2R)-2-CYCLOHEXYL-2-[[2-(METHYLAMINO)-2-OXIDANYLIDENE-ETHYL]AMINO]ETHANOYL]AZETIDINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "4PG": ("1-[4-(2-OXO-2-PHENYLETHYL)PHENYL]GUANIDINE", "non-polymer"),
        "RU": ("RUTHENIUM ION", "non-polymer"),
        "4PE": (
            "3-(2-DEOXY-5-O-PHOSPHONO-BETA-D-ERYTHRO-PENTOFURANOSYL)-6-METHYL-3H-PYRROLO[2,3-D]PYRIMIDIN-2-OL",
            "nucleic-acid",
        ),
        "4PD": (
            "3-(2-DEOXY-5-O-PHOSPHONO-BETA-D-ERYTHRO-PENTOFURANOSYL)-6-METHYL-1,3-DIHYDRO-2H-PYRROLO[2,3-D]PYRIMIDIN-2-ONE",
            "nucleic-acid",
        ),
        "4PC": (
            "3-(2'-DEOXY-5-O-PHOSPHONO-BETA-D-ERYTHRO-PENTOFURANOSYL)-6-METHYL-3,7-DIHYDRO-2H-PYRROLO[2,3-D]PYRIMIDIN-2-ONE",
            "nucleic-acid",
        ),
        "4PB": ("N-HYDROXY-4-PHOSPHONO-BUTANAMIDE", "non-polymer"),
        "4PA": ("PROPYL ACETATE", "non-polymer"),
        "PGW": (
            "(1R)-2-{[(S)-{[(2S)-2,3-DIHYDROXYPROPYL]OXY}(HYDROXY)PHOSPHORYL]OXY}-1-[(HEXADECANOYLOXY)METHYL]ETHYL (9Z)-OCTADEC-9-ENOATE",
            "non-polymer",
        ),
        "4PO": ("ALPHA-(4-PYRIDYL-1-OXIDE)-N-TERT-BUTYLNITRONE", "non-polymer"),
        "4PN": ("4-PIPERIDINO-PIPERIDINE", "non-polymer"),
        "4PL": ("7-HYDROXY-6-METHYLFURO[3,4-C]PYRIDIN-1(3H)-ONE", "non-polymer"),
        "PGT": (
            "(1S)-2-{[{[(2R)-2,3-DIHYDROXYPROPYL]OXY}(HYDROXY)PHOSPHORYL]OXY}-1-[(PALMITOYLOXY)METHYL]ETHYL STEARATE",
            "non-polymer",
        ),
        "4PI": ("N-(4-METHYLBENZOYL)-4-BENZYLPIPERIDINE", "non-polymer"),
        "4PH": ("4-METHYL-L-PHENYLALANINE", "peptide"),
        "RE": ("RHENIUM", "non-polymer"),
        "4PT": (
            "(2R)-3-{[(S)-{[(2S,3R,5S,6S)-2,6-DIHYDROXY-3,4,5-TRIS(PHOSPHONOOXY)CYCLOHEXYL]OXY}(HYDROXY)PHOSPHORYL]OXY}-2-(1-HYDROXYBUTOXY)PROPYL BUTYRATE",
            "non-polymer",
        ),
        "4PS": (
            "N~3~-[(2R)-2-HYDROXY-4-{[(S)-HYDROXY(PHOSPHONOOXY)PHOSPHORYL]OXY}-3,3-DIMETHYLBUTANOYL]-N-(2-SULFANYLETHYL)-BETA-ALANINAMIDE",
            "non-polymer",
        ),
        "4PR": (
            "(3S)-1-{[(3,5-DIMETHYLISOXAZOL-4-YL)AMINO]CARBONYL}-4,4-DIMETHYLPYRROLIDIN-3-YL{(1S)-1-[1-HYDROXY-2-OXO-2-{[(1R)-1-PHENYLETHYL]AMINO}ETHYL]PENTYL}CARBAMATE",
            "non-polymer",
        ),
        "RB": ("RUBIDIUM ION", "non-polymer"),
        "4PP": ("?", "non-polymer"),
        "PGR": ("R-1,2-PROPANEDIOL", "non-polymer"),
        "RH": ("RHODIUM", "non-polymer"),
        "4PZ": ("4-METHYL-1H-PYRAZOLE", "non-polymer"),
        "4PY": (
            "(2S)-1-(6H-INDOL-3-YL)-3-{[5-(7H-PYRAZOLO[3,4-C]PYRIDIN-5-YL)PYRIDIN-3-YL]OXY}PROPAN-2-AMINE",
            "non-polymer",
        ),
        "KZ9": (
            "(S)-4-(3-HYDROXYPHENYL)-2-THIOXO-1,2,3,4,7,8-HEXAHYDROQUINAZOLIN-5(6H)-ONE",
            "non-polymer",
        ),
        "MIO": ("3-METHYL-1,2-OXAZOL-5-AMINE", "non-polymer"),
        "PTD": ("PENTANEDIAL", "non-polymer"),
        "UNG": (
            "4-O-(4-DEOXY-BETA-L-THREO-HEX-4-ENOPYRANURONOSYL)-ALPHA-D-GALACTOPYRANURONIC ACID",
            "non-polymer",
        ),
        "GGZ": (
            "(2S,3AR,4R,6R,6AR)-4-(6-AMINO-9H-PURIN-9-YL)-6-({[(R)-HYDROXY(SULFOOXY)PHOSPHORYL]OXY}METHYL)TETRAHYDROFURO[3,4-D][1,3,2]DIOXAPHOSPHOL-2-OL 2-OXIDE",
            "non-polymer",
        ),
        "4P5": ("4H-FURO[3,2-B]PYRROLE-5-CARBOXYLIC ACID", "non-polymer"),
        "V4Z": (
            "5-CHLORANYL-N2-[(1S)-1-(5-FLUORANYLPYRIDIN-2-YL)ETHYL]-N4-(3-PROPAN-2-YLOXY-1H-PYRAZOL-5-YL)PYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "GOX": (
            "(2S,3S,4R,5R)-6-(HYDROXYAMINO)-2-(HYDROXYMETHYL)-2,3,4,5-TETRAHYDROPYRIDINE-3,4,5-TRIOL",
            "non-polymer",
        ),
        "R04": (
            "(2E)-N-ALLYL-4-{[3-(4-BROMOPHENYL)-5-FLUORO-1-METHYL-1H-INDAZOL-6-YL]OXY}-N-METHYL-2-BUTEN-1-AMINE",
            "non-polymer",
        ),
        "MIN": ("METHYL-PHE-PRO-AMINO-CYCLOHEXYLGLYCINE", "peptide"),
        "PGZ": (
            "METHYL 2-O-[(S)-(BENZYLOXY)(HYDROXY)PHOSPHORYL]-3-DEOXY-3-{[(4-METHYLPHENYL)CARBONYL]AMINO}-1-THIO-BETA-D-GALACTOPYRANOSIDE",
            "non-polymer",
        ),
        "PGX": (
            "7-[6-(3-HYDROPEROXY-OCT-1-ENYL)-2,3-DIOXA-BICYCLO[2.2.1]HEPT-5-YL]-HEPT-5-ENOIC ACID",
            "non-polymer",
        ),
        "B13": ("5-HYDROXYBENZIMIDAZOLYLCOB(III)AMIDE", "non-polymer"),
        "51K": (
            "2-(2-{3-[3,5-BIS(TRIFLUOROMETHYL)PHENYL]-1H-1,2,4-TRIAZOL-1-YL}ETHYL)-1,3,4-OXADIAZOLE",
            "non-polymer",
        ),
        "B11": (
            "N-[3-(4-FLUOROPHENOXY)PHENYL]-4-[(2-HYDROXYBENZYL)AMINO]PIPERIDINE-1-SULFONAMIDE",
            "non-polymer",
        ),
        "WFE": (
            "N-{[(3S)-3-AMINO-1-(5-ETHYL-7H-PYRROLO[2,3-D]PYRIMIDIN-4-YL)PYRROLIDIN-3-YL]METHYL}-2,4-DIFLUOROBENZAMIDE",
            "non-polymer",
        ),
        "G40": ("?", "non-polymer"),
        "G42": ("8-OXO-2'-DEOXY-GUANOSINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "RNP": (
            "(1E,2R)-1-(ISOPROPYLIMINO)-3-(1-NAPHTHYLOXY)PROPAN-2-OL",
            "non-polymer",
        ),
        "G44": (
            "4-({[4-(3-METHYLBENZOYL)PYRIDIN-2-YL]AMINO}METHYL)BENZENECARBOXIMIDAMIDE",
            "non-polymer",
        ),
        "G46": ("2'-DEOXY-GUANOSINE-5'-MONOTHIOPHOSPHATE", "nucleic-acid"),
        "G47": ("N2-ETHANETHIOL-2'-DEOXY-GUANOSINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "G48": ("2'-O-METHYOXYETHYL-GUANOSINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "51P": (
            "5-[(1Z)-2-(2-METHOXYPHENYL)PROP-1-EN-1-YL]FURO[2,3-D]PYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "WFP": ("3,5-DIFLUORO-L-PHENYLALANINE", "peptide"),
        "51U": ("D-PHENYLALANYL-N-(3-METHYLBENZYL)-L-PROLINAMIDE", "peptide"),
        "2FR": (
            "3-[3-(3-METHYL-6-{[(1S)-1-PHENYLETHYL]AMINO}-1H-PYRAZOLO[4,3-C]PYRIDIN-1-YL)PHENYL]PROPANAMIDE",
            "non-polymer",
        ),
        "PA3": ("PAROMOMYCIN (RING 3)", "non-polymer"),
        "2FP": ("1,6-FRUCTOSE DIPHOSPHATE (LINEAR FORM)", "saccharide"),
        "8ID": ("NICOTINAMIDE-8-IODO-ADENINE-DINUCLEOTIDE", "non-polymer"),
        "2FU": ("BUT-2-ENEDIAL", "non-polymer"),
        "8IG": ("8-IODO-GUANINE", "non-polymer"),
        "2FY": ("1-(2,4-DIHYDROXY-PHENYL)-PROPAN-1-ONE", "non-polymer"),
        "2FX": ("1-BENZOTHIOPHEN-2-YLACETIC ACID", "non-polymer"),
        "8IN": (
            "[3-(1-BENZYL-3-CARBAMOYLMETHYL-2-METHYL-1H-INDOL-5-YLOXY)-PROPYL-]-PHOSPHONIC ACID",
            "non-polymer",
        ),
        "8IP": ("N~3~-(3-PYRIDIN-3-YLBENZYL)PYRIDINE-2,3-DIAMINE", "non-polymer"),
        "2FA": (
            "2-(6-AMINO-2-FLUORO-PURIN-9-YL)-5-HYDROXYMETHYL-TETRAHYDRO-FURAN-3,4-DIOL",
            "non-polymer",
        ),
        "2FG": ("2-FLUORO-2-DEOXY-BETA-D-GALACTOPYRANOSE", "saccharide"),
        "B18": (
            "(4R)-N-[4-({[2-(DIMETHYLAMINO)ETHYL]AMINO}CARBONYL)-1,3-THIAZOL-2-YL]-4-METHYL-1-OXO-2,3,4,9-TETRAHYDRO-1H-BETA-CARBOLINE-6-CARBOXAMIDE",
            "non-polymer",
        ),
        "2FE": ("2'-FLUORO-2'-DEOXY-1,N6-ETHENOADENINE", "nucleic-acid"),
        "2FD": (
            "5-(6-AMINO-2-FLUORO-PURIN-9-YL)-2-HYDROXYMETHYL-TETRAHYDRO-FURAN-3-OL",
            "non-polymer",
        ),
        "5F1": ("5'-FLUORO-2',5'-DIDEOXYADENOSINE", "non-polymer"),
        "2FI": ("2'-FLUORO-2'-DEOXYINOSINE", "nucleic-acid"),
        "2FH": ("2-PHENYLHEME", "non-polymer"),
        "2FM": ("S-(DIFLUOROMETHYL)HOMOCYSTEINE", "peptide"),
        "2FL": ("2-FLUORO-2-DEOXY-LACTOSE", "saccharide"),
        "5FH": ("(5S)-5-BENZYLIMIDAZOLIDINE-2,4-DIONE", "non-polymer"),
        "5FN": (
            "(3AS,5S,6S,7R,7AR)-5-FLUORO-5-(HYDROXYMETHYL)-2-METHYL-5,6,7,7A-TETRAHYDRO-3AH-PYRANO[3,2-D][1,3]OXAZOLE-6,7-DIOL",
            "non-polymer",
        ),
        "5FO": (
            "6-ETHYL-5-FLUORO-2-{5-[(4-METHYLPIPERAZIN-1-YL)SULFONYL]-2-PROPOXYPHENYL}PYRIMIDIN-4(3H)-ONE",
            "non-polymer",
        ),
        "5FA": ("ADENOSINE-5'-PENTAPHOSPHATE", "nucleic-acid"),
        "5FB": (
            "4-(4-{[(5R)-2,4-DIOXO-1,3-THIAZOLIDIN-5-YL]METHYL}-2-METHOXYPHENOXY)-3-(TRIFLUOROMETHYL)BENZONITRILE",
            "non-polymer",
        ),
        "5FC": ("5-FORMYL-2'-DEOXY-CYTIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "5FD": ("5'-FLUORO-5'-DEOXYADENOSINE", "non-polymer"),
        "5FE": (
            "5,5,5-TRIFLUORO-3-HYDROXY-4-[2-(5,5,5-TRIFLUORO-3-HYDROXY-4-{3-METHYL-2-[3-METHYL-2-(3-METHYL-BUTYRYLAMINO)-BUTYRYLAMINO]-BUTYRYLAMINO}-PENTANOYLAMINO)-PROPIONYLAMINO]-PENTANOIC ACID METHYL ESTER",
            "peptide",
        ),
        "RF5": ("2-DEOXY-2-FLUORO-5-O-PHOSPHONO-ALPHA-D-ARABINOFURANOSE", "saccharide"),
        "5FX": ("5-DEOXY-5-FLUORO-D-XYLULOSE", "non-polymer"),
        "8I1": (
            "2-(5-{[(2R)-2-AMINO-3-PHENYLPROPYL]OXY}PYRIDIN-3-YL)-8,9-DIMETHOXYBENZO[C][2,7]NAPHTHYRIDIN-4-AMINE",
            "non-polymer",
        ),
        "5FP": (
            "5,5,5-TRIFLUORO-3-HYDROXY-4-[2-(5,5,5-TRIFLUORO-3-HYDROXY-4-{3-METHYL-2-[3-METHYL-2-(3-METHYL-BUTYRYLAMINO)-BUTYRYLAMINO]-BUTYRYLAMINO}-PENTANOYLAMINO)-PROPIONYLAMINO]-PENTANOIC ACID",
            "peptide",
        ),
        "TEL": ("TELITHROMYCIN", "non-polymer"),
        "5FR": (
            "4-({1-[3-(3-AMINO-3-OXOPROPYL)-5-CHLOROPHENYL]-3-METHYL-1H-PYRAZOLO[4,3-C]PYRIDIN-6-YL}AMINO)-3-METHOXY-N-(1-METHYLPIPERIDIN-4-YL)BENZAMIDE",
            "non-polymer",
        ),
        "5FU": ("5-FLUORO-URIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "5FW": ("ISOPRENALINE", "non-polymer"),
        "G4D": ("4-DEOXY-ALPHA-D-GLUCOSE", "saccharide"),
        "PRL": ("PROFLAVIN", "non-polymer"),
        "G4G": (
            "4,4'-{(3S,4S)-PYRROLIDINE-3,4-DIYLBIS[(BENZYLIMINO)SULFONYL]}DIBENZAMIDE",
            "non-polymer",
        ),
        "G4M": ("?", "non-polymer"),
        "V4O": ("CYCLO-TETRAMETAVANADATE", "non-polymer"),
        "G4P": ("GUANOSINE-5',3'-TETRAPHOSPHATE", "nucleic-acid"),
        "518": (
            "N-[4-(5-{[(1S)-1-CYCLOHEXYL-2-(METHYLAMINO)-2-OXOETHYL]CARBAMOYL}FURAN-2-YL)PHENYL]-1-BENZOFURAN-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "G4S": ("4-O-SULFO-BETA-D-GALACTOPYRANOSE", "saccharide"),
        "TEA": ("TRIETHYLAMMONIUM ION", "non-polymer"),
        "511": (
            "[1-(4-CHLOROBENZOYL)-5-METHOXY-1H-INDOL-3-YL]ACETIC ACID",
            "non-polymer",
        ),
        "WF3": (
            "METHYL (2E)-3-METHOXY-2-[2-({[6-METHYL-3-(TRIFLUOROMETHYL)QUINOXALIN-2-YL]OXY}METHYL)PHENYL]PROP-2-ENOATE",
            "non-polymer",
        ),
        "513": ("1-(5-BROMO-1-BENZOTHIOPHEN-3-YL)METHANAMINE", "non-polymer"),
        "512": ("4-BROMO-3-(CARBOXYMETHOXY)THIOPHENE-2-CARBOXYLIC ACID", "non-polymer"),
        "515": (
            "3-({5-[(N-ACETYL-3-{4-[(CARBOXYCARBONYL)(2-CARBOXYPHENYL)AMINO]-1-NAPHTHYL}-L-ALANYL)AMINO]PENTYL}OXY)-2-NAPHTHOIC ACID",
            "non-polymer",
        ),
        "514": (
            "(3Z)-5-ACETYL-3-(BENZOYLIMINO)-3,6-DIHYDROPYRROLO[3,4-C]PYRAZOL-5-IUM",
            "non-polymer",
        ),
        "517": (
            "1-{2-OXO-3-[(1R)-1-(1H-PYRROL-2-YL)ETHYL]-2H-INDOL-5-YL}UREA",
            "non-polymer",
        ),
        "516": (
            "N-{(1S)-1-CYCLOHEXYL-2-[(2S)-2-(4-NAPHTHALEN-1-YL-1,3-THIAZOL-2-YL)PYRROLIDIN-1-YL]-2-OXOETHYL}-N~2~-METHYL-L-ALANINAMIDE",
            "non-polymer",
        ),
        "623": (
            "3,3'-[3,5-DIFLUORO-4-METHYL-2,6-PYRIDYLENEBIS(OXY)]-BIS(BENZENECARBOXIMIDAMIDE)",
            "non-polymer",
        ),
        "621": (
            "N-ACETYLGLYCYL-N-[(3S)-1-HYDROXY-5-METHYL-2-OXOHEXAN-3-YL]-L-SERINAMIDE",
            "non-polymer",
        ),
        "PRO": ("PROLINE", "peptide"),
        "627": (
            "N-[(3E)-5-[(2R)-2-METHOXY-2-PHENYLACETYL]PYRROLO[3,4-C]PYRAZOL-3(5H)-YLIDENE]-4-(4-METHYLPIPERAZIN-1-YL)BENZAMIDE",
            "non-polymer",
        ),
        "626": (
            "4-(4-METHYLPIPERAZIN-1-YL)-N-[5-(2-THIENYLACETYL)-1,5-DIHYDROPYRROLO[3,4-C]PYRAZOL-3-YL]BENZAMIDE",
            "non-polymer",
        ),
        "625": (
            "4-(4-{1-[(6-AMINOPYRIDIN-2-YL)METHYL]-5-(2-CHLOROPHENYL)-1H-PYRROL-2-YL}PHENOXY)BUTANENITRILE",
            "non-polymer",
        ),
        "624": (
            "2-(2-(3,4-DICHLOROPHENYL)-5-(ISOPROPYLAMINO)-6-OXOPYRIMIDIN-1(6H)-YL)-N-((S)-1-OXO-1-(THIAZOL-2-YL)-5-GUANIDINOPENTAN-2-YL)ACETAMIDE",
            "non-polymer",
        ),
        "629": (
            "[2-AMINO-3-(4-HYDROXY-PHENYL)-PROPIONYLAMINO]-(1,3,4,5-TETRAHYDROXY-4-HYDROXYMETHYL-PIPERIDIN-2-YL)- ACETIC ACID",
            "non-polymer",
        ),
        "628": (
            "4-{[6-(2,6-DICHLOROBENZOYL)IMIDAZO[1,2-A]PYRIDIN-2-YL]AMINO}BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "NBV": (
            "(2R,3R,4R,5S)-1-BUTYL-2-(HYDROXYMETHYL)PIPERIDINE-3,4,5-TRIOL",
            "non-polymer",
        ),
        "U8D": (
            "(3-{[3-(2-AMINO-2-OXOETHYL)-1-BENZYL-2-ETHYL-1H-INDOL-5-YL]OXY}PROPYL)PHOSPHONIC ACID",
            "non-polymer",
        ),
        "NBU": ("N-BUTANE", "non-polymer"),
        "NBT": ("(2S)-2-NITROBUTANE", "non-polymer"),
        "FKL": (
            "N-CYCLOPROPYL-4-METHYL-3-[8-METHYL-7-OXO-2-(TETRAHYDRO-2H-PYRAN-4-YLAMINO)-7,8-DIHYDROPYRIDO[2,3-D]PYRIMIDIN-6-YL]BENZAMIDE",
            "non-polymer",
        ),
        "PRN": ("?", "nucleic-acid"),
        "6GO": ("6-O-METHYLGUANINE", "non-polymer"),
        "3D1": (
            "(2R,3S,5R)-5-(6-AMINO-9H-PURIN-9-YL)-TETRAHYDRO-2-(HYDROXYMETHYL)FURAN-3-OL",
            "non-polymer",
        ),
        "0DY": (
            "N~1~-(3-AMINOBENZYL)-N~2~-[(2R)-2-(HYDROXYCARBAMOYL)-4-METHYLPENTANOYL]-L-ASPARTAMIDE",
            "peptide",
        ),
        "3D3": ("2-[(6-AMINO-7H-PURIN-8-YL)THIO]ACETAMIDE", "non-polymer"),
        "0DV": (
            "(4R,5R,6R,6AS,9S,9AE,10AR)-5-HYDROXY-9-(HYDROXYMETHYL)-6,10A-DIMETHYL-3-(PROPAN-2-YL)-1,2,4,5,6,6A,7,8,9,10A-DECAHYDRODICYCLOPENTA[A,D][8]ANNULEN-4-YL ALPHA-D-GULOPYRANOSIDE",
            "non-polymer",
        ),
        "0DU": (
            "N-(2,6-DICHLOROBENZOYL)-4-[1,6-DIMETHYL-2-OXO-4-(TRIFLUOROMETHYL)-1,2-DIHYDROPYRIDIN-3-YL]-L-PHENYLALANINE",
            "non-polymer",
        ),
        "0DT": ("?", "nucleic-acid"),
        "0DS": (
            "N-{(2R)-2-[2-(HYDROXYAMINO)-2-OXOETHYL]-4-METHYLPENTANOYL}-L-LEUCYL-L-PHENYLALANINAMIDE",
            "peptide",
        ),
        "6GB": (
            "6-O-(6-O-PHOSPHONO-BETA-D-GLUCOPYRANOSYL)-BETA-D-GLUCOPYRANOSE",
            "non-polymer",
        ),
        "0DO": (
            "(2R,4S)-N-TERT-BUTYL-1-[(2S,3S)-3-{[(2,6-DIMETHYLPHENOXY)ACETYL]AMINO}-2-HYDROXY-4-PHENYLBUTYL]-4-(PYRIDIN-4-YLSULFONYL)PIPERIDINE-2-CARBOXAMIDE",
            "peptide",
        ),
        "0DN": ("5'-DEOXY-5'-FLUOROTHYMIDINE", "non-polymer"),
        "I1N": ("(2S,3R,4R,5S,6S)-2,3,4,5,6-PENTAHYDROXYCYCLOHEXANONE", "non-polymer"),
        "0DJ": (
            "3-(1-METHYL-2-OXO-2,3-DIHYDRO-1H-IMIDAZO[4,5-B]PYRIDIN-6-YL)BENZAMIDE",
            "non-polymer",
        ),
        "0DI": (
            "(2R)-2-AMINO-3-(3-CHLOROPHENYL)-1-[4-(2,5-DIFLUOROBENZOYL)PIPERAZIN-1-YL]PROPAN-1-ONE",
            "non-polymer",
        ),
        "0DG": ("?", "nucleic-acid"),
        "6GU": ("6-CHLOROGUANINE", "non-polymer"),
        "RDF": (
            "N-ALPHA-L-RHAMNOPYRANOSYLOXY(HYDROXYPHOSPHINYL)-L-LEUCYL-L-TRYPTOPHAN",
            "peptide",
        ),
        "6GP": ("METHYL-N-(BETA-D-GLUCOPYRANOSYL)OXAMATE", "non-polymer"),
        "0DB": (
            "N-({1-[(2S)-2-CARBOXY-4-PHENYLBUTYL]CYCLOPENTYL}CARBONYL)-L-TRYPTOPHAN",
            "peptide",
        ),
        "1BE": (
            "N-(6-CHLORO-3,3-DIMETHYL-3,4-DIHYDROISOQUINOLIN-1-YL)-3-[2-PROPYL-4-(1H-PYRAZOL-4-YL)THIOPHEN-3-YL]-L-ALANINE",
            "non-polymer",
        ),
        "3DT": ("3-METHYLTHYMIDINE", "non-polymer"),
        "3DP": ("3-(N,N-DIMETHYLOCTYLAMMONIO)PROPANESULFONATE", "non-polymer"),
        "3DQ": (
            "(1R,2S,4S,5R)-2-[(4-METHOXYPHENYL)METHYL]-1,4,5-TRIS(OXIDANYL)-3-OXIDANYLIDENE-CYCLOHEXANE-1-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "3DR": ("1',2'-DIDEOXYRIBOFURANOSE-5'-PHOSPHATE", "nucleic-acid"),
        "3DS": (
            "(4S,5R)-4,5-DIHYDROXY-3-OXOCYCLOHEX-1-ENE-1-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "0D6": (
            "N-{(1R)-3-PHENYL-1-[2-(PHENYLSULFONYL)ETHYL]PROPYL}-N~2~-(PIPERAZIN-1-YLCARBONYL)-L-LEUCINAMIDE",
            "peptide",
        ),
        "HLC": (
            "4-(4-CHLOROPHENOXY)-N-[(3S)-2-OXOTETRAHYDROFURAN-3-YL]BUTANAMIDE",
            "non-polymer",
        ),
        "DN2": (
            "3-({(2S)-2-[({(1S)-1-[({(1S)-1-[(R)-CARBOXY(HYDROXY)METHYL]-3,3-DIFLUOROPROPYL}AMINO)CARBONYL]-3-METHYLBUTYL}AMINO)CARBONYL]-2,3-DIHYDRO-1H-INDOL-2-YL}METHYL)THIOPHENE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "0D3": ("N-[(2S)-2-BENZYL-3-SULFANYLPROPANOYL]-L-ALANYLGLYCINAMIDE", "peptide"),
        "0D2": (
            "(4R,4'S)-4,4'-DIPHENYL-4,4'-BIIMIDAZOLIDINE-2,2',5,5'-TETRONE",
            "non-polymer",
        ),
        "0D1": ("4,4'-(2,2-DICHLOROETHENE-1,1-DIYL)DIPHENOL", "non-polymer"),
        "3DD": (
            "(2R,3R)-4-(4-AMINO-1H-IMIDAZO[4,5-C]PYRIDIN-1-YL)-2,3-DIHYDROXYBUTANOIC ACID",
            "non-polymer",
        ),
        "3DE": (
            "3,5-DIMETHYL-1-PHENYL-1H-PYRAZOLE-4-CARBOXYLIC ACID ETHYL ESTER",
            "non-polymer",
        ),
        "3DG": ("9-(3-DEOXY-BETA-D-RIBOFURANOSYL)GUANINE", "non-polymer"),
        "3DA": ("3'-DEOXYADENOSINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "3DB": (
            "5-[(PHENYLSULFONYL)AMINO]-1,3,4-THIADIAZOLE-2-SULFONAMIDE",
            "non-polymer",
        ),
        "PEE": ("PHOSPHATIDYLETHANOLAMINE", "non-polymer"),
        "RDA": (
            "METHYL 3-CHLORO-2-{3-[(2,5-DIHYDROXY-4-METHOXYPHENYL)AMINO]-3-OXOPROPYL}-4,6-DIHYDROXYBENZOATE",
            "non-polymer",
        ),
        "3DM": ("2,6-DIMETHOXYPHENOL", "non-polymer"),
        "3DH": ("5'-S-ETHYL-5'-THIOADENOSINE", "non-polymer"),
        "NBA": (
            "3-[(3-{[3-(METHYLAMINO)PROPYL]AMINO}PROPYL)AMINO]PROPANE-1,1-DIOL",
            "non-polymer",
        ),
        "ZEB": ("4-HYDROXY-3,4-DIHYDRO-ZEBULARINE", "non-polymer"),
        "ZEC": ("6-HYDROXY-1,3-BENZOTHIAZOLE-2-SULFONAMIDE", "non-polymer"),
        "62P": (
            "6-[(5-CHLORO-3-METHYL-1-BENZOFURAN-2-YL)SULFONYL]PYRIDAZIN-3(2H)-ONE",
            "non-polymer",
        ),
        "62U": (
            "(8S,9R,13R,14S,16R)-21-(FURAN-2-YL)-13-HYDROXY-8,14,19-TRIMETHOXY-16-METHYL-4,10,12-TRIMETHYLIDENE-3,20,22-TRIOXO-2-AZABICYCLO[16.3.1]DOCOSA-1(21),18-DIEN-9-YL CARBAMATE",
            "non-polymer",
        ),
        "NBG": ("1-N-ACETYL-BETA-D-GLUCOSAMINE", "saccharide"),
        "ZEN": ("?", "non-polymer"),
        "NBF": ("[(2-ETHOXY-1-NAPHTHOYL)AMINO]METHYLBORONIC ACID", "non-polymer"),
        "SNP": ("1-(ISOPROPYLAMINO)-3-(1-NAPHTHYLOXY)-2-PROPANOL", "non-polymer"),
        "ZES": (
            "[3-(4-BROMO-2-FLUORO-BENZYL)-7-CHLORO-2,4-DIOXO-3,4-DIHYDRO-2H-QUINAZOLIN-1-YL]-ACETIC ACID",
            "non-polymer",
        ),
        "62A": (
            "2-(5-AMINO-6-OXO-2-M-TOLYL-6H-PYRIMIDIN-1-YL)-N-[4-GUANIDINO-1-(THIAZOLE-2-CARBONYL)-BUTYL]-ACETAMIDE",
            "non-polymer",
        ),
        "62D": (
            "6-(ETHYLAMINO)-2-(METHYLAMINO)-3,7-DIHYDRO-8H-IMIDAZO[4,5-G]QUINAZOLIN-8-ONE",
            "non-polymer",
        ),
        "ZEZ": ("(2Z)-2-METHYL-4-(9H-PURIN-6-YLAMINO)BUT-2-EN-1-OL", "non-polymer"),
        "ZEX": ("ZEAXANTHIN", "non-polymer"),
        "NBD": ("N6-BENZYL-NICOTINAMIDE-ADENINE-DINUCLEOTIDE", "non-polymer"),
        "UDX": ("URIDINE-5'-DIPHOSPHATE-XYLOPYRANOSE", "non-polymer"),
        "W91": (
            "5-(3-(2,6-DICHLORO-4-(4,5-DIHYDRO-2-OXAZOLYL)PHENOXY)PROPYL)-3-METHYL ISOXAZOLE",
            "non-polymer",
        ),
        "NB8": (
            "5'-O-[(R)-{[(2S)-2-AMINO-4-(HYDROXYAMINO)-4-OXOBUTANOYL]OXY}(HYDROXY)PHOSPHORYL]ADENOSINE",
            "peptide",
        ),
        "PRJ": (
            "(2S,3AS,6R,7AS)-6-HYDROXYOCTAHYDRO-1H-INDOLE-2-CARBOXYLIC ACID",
            "peptide",
        ),
        "XD1": (
            "(2R,4S)-2-[(1R)-1-AMINO-2-OXOETHYL]-5,5-DIMETHYL-1,3-THIAZOLIDINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "S2C": ("S-2-(BORONOETHYL)-L-CYSTEINE", "peptide"),
        "S2A": ("SPIROCYCLIC ALKENE", "non-polymer"),
        "S2G": ("(2S)-2-HYDROXYPENTANEDIOIC ACID", "non-polymer"),
        "S2D": ("N-BENZOYL-D-ALANINE", "peptide"),
        "UDC": (
            "2-(ACETYLAMINO)-3-O-[(2S,3S)-2-CARBOXY-3,4-DIHYDROXYBUTAN-2-YL]-2-DEOXY-ALPHA-D-GLUCOPYRANOSE",
            "saccharide",
        ),
        "S2H": ("HYDROGEN DISULFIDE", "non-polymer"),
        "S2N": (
            "S-[2-({N-[(2R)-2-HYDROXY-4-{[(S)-HYDROXY(METHOXY)PHOSPHORYL]OXY}-3,3-DIMETHYLBUTANOYL]-BETA-ALANYL}AMINO)ETHYL] (3R)-3-HYDROXYDODECANETHIOATE",
            "non-polymer",
        ),
        "S2M": (
            "2'-O-[2-(METHOXY)ETHYL]-2-THIOTHYMIDINE-5'-MONOPHOSPHATE",
            "nucleic-acid",
        ),
        "I15": (
            "5-[(5-FLUORO-3-METHYL-1H-INDAZOL-4-YL)OXY]BENZENE-1,3-DICARBONITRILE",
            "non-polymer",
        ),
        "S2P": (
            "(2S)-2-AMINO-3-(4-HYDROXY-1,2,5-THIADIAZOL-3-YL)PROPANOIC ACID",
            "peptide",
        ),
        "UDA": (
            "3'-1-CARBOXY-1-PHOSPHONOOXY-ETHOXY-URIDINE-DIPHOSPHATE-N-ACETYLGLUCOSAMINE",
            "non-polymer",
        ),
        "UDB": (
            "(S)-N-(CYCLOPENTYLMETHYL)-2-(3-(3,5-DIFLUOROPHENYL)UREIDO)-N-(2-(HYDROXYAMINO)-2-OXOETHYL)-3,3-DIMETHYLBUTANAMIDE",
            "non-polymer",
        ),
        "IMB": (
            "[(ISOQUINOLIN-1-YLAMINO)-PHOSPHONO-METHYL]-PHOSPHONIC ACID",
            "non-polymer",
        ),
        "OGK": (
            "(1S,2S)-2-ETHYL-1-({[(3AS,4S,6R,7AS)-6-ETHYL-1-OXOOCTAHYDRO-1H-INDEN-4-YL]CARBONYL}AMINO)CYCLOPROPANECARBOXYLIC ACID",
            "non-polymer",
        ),
        "X72": (
            "4-CHLORO-5-NITRO-2-[(PYRIMIDIN-5-YLMETHYL)AMINO]BENZAMIDE",
            "non-polymer",
        ),
        "X73": (
            "2-{[(6-CHLOROPYRIDIN-3-YL)METHYL]AMINO}-5-NITROBENZAMIDE",
            "non-polymer",
        ),
        "S22": (
            "1-(4-(4-(2-(ISOPROPYLSULFONYL)PHENYLAMINO)-1H-PYRROLO[2,3-B]PYRIDIN-6-YLAMINO)-3-METHOXYPHENYL)PIPERIDIN-4-OL",
            "non-polymer",
        ),
        "S23": (
            "N-{(2S)-3-[(1R)-1-AMINOETHYL](HYDROXY)PHOSPHORYL-2-BENZYLPROPANOYL}-L-PHENYLALANINE",
            "non-polymer",
        ),
        "S21": ("4,4,4-TRIFLUORO-1-(4-NITROPHENYL)BUTANE-1,3-DIONE", "non-polymer"),
        "S27": (
            "N-[[2-METHYL-4-HYDROXYCARBAMOYL]BUT-4-YL-N]-BENZYL-P-[PHENYL]-P-[METHYL]PHOSPHINAMID",
            "non-polymer",
        ),
        "S24": ("S-(2,4-DIHYDROXYPHENYL) HYDROGEN THIOCARBONATE", "non-polymer"),
        "S25": (
            "N-(2-AMINOETHYL)-5-(2-{[4-(MORPHOLIN-4-YL)PYRIDIN-2-YL]AMINO}-1,3-THIAZOL-5-YL)PYRIDINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "S28": (
            "N-(BENZYLSULFONYL)GLYCYL-N-(4-CARBAMIMIDOYLBENZYL)-L-PROLINAMIDE",
            "peptide",
        ),
        "S29": (
            "N-(BENZYLSULFONYL)-D-ALANYL-N-(4-CARBAMIMIDOYLBENZYL)-L-PROLINAMIDE",
            "peptide",
        ),
        "OGG": ("2-(2-CARBAMOYLMETHOXY-ETHOXY)-ACETAMIDE", "non-polymer"),
        "SR": ("STRONTIUM ION", "non-polymer"),
        "XGL": (
            "[(1S,4R,6R)-6-HYDROXY-4-(GUANIN-9-YL)CYCLOHEX-2-EN-1-YL]METHYL DIHYDROGEN PHOSPHATE",
            "nucleic-acid",
        ),
        "OGA": ("N-OXALYLGLYCINE", "non-polymer"),
        "XGA": (
            "6-AMINO-3-(2-DEOXY-5-O-PHOSPHONO-BETA-D-ERYTHRO-PENTOFURANOSYL)-3,7-DIHYDRO-8H-IMIDAZO[4,5-G]QUINAZOLIN-8-ONE",
            "nucleic-acid",
        ),
        "XGU": (
            "2-AMINO-9-(2,3-DIDEOXY-6-O-PHOSPHONO-BETA-D-ERYTHRO-HEXOPYRANOSYL)-1,9-DIHYDRO-6H-PURIN-6-ONE",
            "nucleic-acid",
        ),
        "XGR": (
            "[(1R,4S,6S)-4-(2-AMINO-6-OXO-1,6-DIHYDRO-9H-PURIN-9-YL)-6-HYDROXYCYCLOHEX-2-EN-1-YL]METHYL DIHYDROGEN PHOSPHATE",
            "nucleic-acid",
        ),
        "T8N": ("(2E)-2-(2-METHOXY-2-OXOETHYL)BUT-2-ENEDIOIC ACID", "non-polymer"),
        "XGP": ("1-O-PHOSPHONO-BETA-D-GLUCOPYRANOSE", "non-polymer"),
        "TQD": (
            "(2R,6S)-6-{[METHYL(3,4,5-TRIMETHOXYPHENYL)AMINO]METHYL}-1,2,5,6,7,8-HEXAHYDROQUINAZOLINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "DNS": ("N~6~-{[5-(DIMETHYLAMINO)-1-NAPHTHYL]SULFONYL}-L-LYSINE", "peptide"),
        "DNR": ("?", "nucleic-acid"),
        "TQI": (
            "N-{[4-CHLORO-3-(TRIFLUOROMETHYL)PHENYL]SULFONYL}-L-TRYPTOPHAN",
            "peptide",
        ),
        "TQJ": (
            "N-[(2S)-1-[4-(5-BROMOPYRIDIN-2-YL)PIPERAZIN-1-YL]SULFONYL-5-PYRIMIDIN-2-YL-PENTAN-2-YL]-N-HYDROXY-METHANAMIDE",
            "non-polymer",
        ),
        "TQT": (
            "6-(OCTAHYDRO-1H-INDOL-1-YLMETHYL)DECAHYDROQUINAZOLINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "SY2": (
            "N-{[(1S)-2-METHYL-1-{[(5S,8S)-5-(1-METHYLETHYL)-2,7-DIOXO-1,6-DIAZACYCLODODEC-3-EN-8-YL]CARBAMOYL}PROPYL]CARBAMOYL}-L-VALINE",
            "non-polymer",
        ),
        "SY1": (
            "2-({2-[(3R)-3-AMINOPIPERIDIN-1-YL]-4-OXOQUINAZOLIN-3(4H)-YL}METHYL)BENZONITRILE",
            "non-polymer",
        ),
        "TQP": ("?", "non-polymer"),
        "TQQ": (
            "(S)-2-AMINO-3-(6,7-DIHYDRO-6-IMINO-7-OXO-1H-INDOL-3-YL)PROPANOIC ACID",
            "peptide",
        ),
        "TE1": (
            "(2Z)-3-{2-HYDROXY-5-[(1S)-1-HYDROXY-3-METHYLBUTYL]-4-METHOXYPHENYL}PROP-2-ENOIC ACID",
            "non-polymer",
        ),
        "MIM": (
            "[CYCLOHEXYLETHYL]-[[[[4-[2-METHYL-1-IMIDAZOLYL-BUTYL]PHENYL]ACETYL]-SERYL]-LYSINYL]-AMINE",
            "non-polymer",
        ),
        "SY9": ("STRYCHNINE", "non-polymer"),
        "M6D": ("6-O-PHOSPHONO-BETA-D-MANNOPYRANOSE", "non-polymer"),
        "TQX": (
            "(2S)-TERT-BUTOXY[4-(8-FLUORO-5-METHYL-3,4-DIHYDRO-2H-CHROMEN-6-YL)-2-METHYLQUINOLIN-3-YL]ETHANOIC ACID",
            "non-polymer",
        ),
        "TE2": ("?", "non-polymer"),
        "HSC": ("(S)-3-HYDROXYDECANOYL-COA", "non-polymer"),
        "HSA": (
            "PHOSPHORIC ACID MONO-[2-AMINO-3-(3H-IMIDAZOL-4-YL)-PROPYL]ESTER",
            "non-polymer",
        ),
        "HSF": ("1-HEXADECYLSULFONYL FLUORIDE", "non-polymer"),
        "HSG": ("OCTYL ALPHA-L-ALTROPYRANOSIDE", "saccharide"),
        "HSD": (
            "(1S,2S,3R,6R)-6-AMINO-4-(HYDROXYMETHYL)CYCLOHEX-4-ENE-1,2,3-TRIOL",
            "non-polymer",
        ),
        "HSE": ("L-HOMOSERINE", "peptide"),
        "HSH": ("OCTYL BETA-D-GALACTOPYRANOSIDE", "saccharide"),
        "HSI": (
            "(3R)-3-{4-[(4-CHLOROPHENYL)ETHYNYL]BENZOYL}NONANOIC ACID",
            "non-polymer",
        ),
        "HSO": ("HISTIDINOL", "peptide"),
        "HSL": ("HOMOSERINE LACTONE", "peptide"),
        "DNX": ("3,4-DINITROPHENOL", "non-polymer"),
        "HSR": ("2-(ACETYLAMINO)-2-DEOXY-BETA-D-GLUCOPYRANOSE", "saccharide"),
        "HSS": ("5'-O-[(L-HISTIDYLAMINO)SULFONYL]ADENOSINE", "non-polymer"),
        "HSP": ("4-(2-HYDROXYPHENYLSULFINYL)-BUTYLPHOSPHONIC ACID", "non-polymer"),
        "HSQ": ("2-(ACETYLAMINO)-2-DEOXY-ALPHA-L-IDOPYRANOSE", "saccharide"),
        "HSW": ("N-HYDROXYSULFAMIDE", "non-polymer"),
        "HST": ("4-CARBOXY-5-(1-PENTYL)HEXYLSULFANYL-1,2,3-TRIAZOLE", "non-polymer"),
        "HSU": ("BETA-D-RIBOFURANOSE", "saccharide"),
        "HSZ": ("BETA-D-XYLOPYRANOSE", "saccharide"),
        "HSX": ("5-O-PHOSPHONO-ALPHA-D-RIBOFURANOSE", "saccharide"),
        "HSY": ("ALPHA-L-XYLOPYRANOSE", "saccharide"),
        "SF4": ("IRON/SULFUR CLUSTER", "non-polymer"),
        "PT1": ("PTEROIC ACID", "non-polymer"),
        "0GJ": (
            "L-ALPHA-GLUTAMYL-N-{(1S)-4-{[AMINO(IMINIO)METHYL]AMINO}-1-[(1S)-2-CHLORO-1-HYDROXYETHYL]BUTYL}GLYCINAMIDE",
            "peptide",
        ),
        "SF3": ("FE4-S3 CLUSTER", "non-polymer"),
        "HS2": (
            "2-(ACETYLAMINO)-2-DEOXY-4-O-(4-DEOXY-ALPHA-L-THREO-HEX-4-ENOPYRANURONOSYL)-ALPHA-D-GLUCOPYRANOSE",
            "saccharide",
        ),
        "HS3": (
            "(2R)-3-HYDROXY-2-[(4-METHOXYPHENYL)SULFONYLAMINO]-N-OXO-PROPANAMIDE",
            "non-polymer",
        ),
        "HS1": (
            "2-[[(2S)-2,3-DIHYDROXYPROPYL]-(4-METHOXYPHENYL)SULFONYL-AMINO]-N-OXO-ETHANAMIDE",
            "non-polymer",
        ),
        "HS6": ("2-[(4-FLUOROPHENYL)SULFONYLAMINO]-N-OXO-ETHANAMIDE", "non-polymer"),
        "HS7": ("N-OXO-2-(PHENYLSULFONYLAMINO)ETHANAMIDE", "non-polymer"),
        "HS4": ("N-OXO-2-[(4-PHENYLPHENYL)SULFONYLAMINO]ETHANAMIDE", "non-polymer"),
        "HS5": (
            "2-[(4-FLUOROPHENYL)SULFONYL-(2-HYDROXYETHYL)AMINO]-N-OXO-ETHANAMIDE",
            "non-polymer",
        ),
        "HS8": ("3-(1-SULFO-1H-IMIDAZOL-3-IUM-4-YL)-L-ALANINE", "peptide"),
        "HS9": ("N-ALPHA-METHYL-L-HISTIDINAMIDE", "peptide"),
        "SYC": ("(PYRIDIN-2-YLMETHYL)PHOSPHONIC ACID", "non-polymer"),
        "NVA": ("NORVALINE", "peptide"),
        "SYE": ("(QUINOLIN-2-YLMETHYL)PHOSPHONIC ACID", "non-polymer"),
        "SYD": ("[(5-PHENYLPYRIDIN-2-YL)METHYL]PHOSPHONIC ACID", "non-polymer"),
        "E4N": ("TETRAETHYLAMMONIUM ION", "non-polymer"),
        "SYO": (
            "[(3S)-3-HYDROXY-2,2-DIMETHYL-4-OXO-4-[[3-OXO-3-[2-(3-OXOOCTYLSULFANYL)ETHYLAMINO]PROPYL]AMINO]BUTYL] DIHYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "SYM": ("2S,4R-4-METHYLGLUTAMATE", "non-polymer"),
        "SYS": ("3-[(2-AMINO-2-OXOETHYL)SELANYL]-L-ALANINE", "peptide"),
        "SYR": ("SYRINGATE", "non-polymer"),
        "TQ6": (
            "5-(4-MORPHOLIN-4-YL-PHENYLSULFANYL)-2,4-QUINAZOLINEDIAMINE",
            "non-polymer",
        ),
        "MIZ": (
            "1-[2-(1H-IMIDAZOL-4-YL)ETHYL]-2,4,6-TRIMETHYLPYRIDINIUM",
            "non-polymer",
        ),
        "TQ1": (
            "5-(HYDROXYMETHYL)-8-(1H-PYRROL-2-YL)-2H-[1,2,4]TRIAZOLO[4,3-A]QUINOLIN-1-ONE",
            "non-polymer",
        ),
        "TQ2": (
            "(2S)-TERT-BUTOXY[4-(3,4-DIMETHYLPHENYL)-2-METHYLQUINOLIN-3-YL]ETHANOIC ACID",
            "non-polymer",
        ),
        "TQ3": ("5-PHENYLSULFANYL-2,4-QUINAZOLINEDIAMINE", "non-polymer"),
        "7PH": (
            "(1R)-2-(DODECANOYLOXY)-1-[(PHOSPHONOOXY)METHYL]ETHYL TETRADECANOATE",
            "non-polymer",
        ),
        "SYY": (
            "METHYL 3-[(THIOPHEN-2-YLACETYL)AMINO]THIOPHENE-2-CARBOXYLATE",
            "non-polymer",
        ),
        "TQ8": (
            "N-[(2S)-4-CHLORO-3-OXO-1-PHENYL-BUTAN-2-YL]-4-METHYL-BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "TQ9": ("5-CHLORO-4-OXOPENTANOIC ACID", "non-polymer"),
        "GHA": ("GHAVAMIOL", "non-polymer"),
        "EME": ("N-METHYL-L-GLUTAMIC ACID", "peptide"),
        "EMD": (
            "5-[1-(3,4-DIMETHOXY-BENZOYL)-1,2,3,4-TETRAHYDRO-QUINOLIN-6-YL]-6-METHYL-3,6-DIHYDRO-[1,3,4]THIADIAZIN-2-ONE",
            "non-polymer",
        ),
        "EMF": (
            "ETHYL (1S,2S)-2-(4-HYDROXY-3-METHOXYPHENYL)CYCLOPROPANECARBOXYLATE",
            "non-polymer",
        ),
        "EMA": (
            "(ADENIN-9-YL-ETHOXYMETHYL)-HYDROXYPHOSPHINYL-DIPHOSPHATE",
            "non-polymer",
        ),
        "EMC": ("ETHYL MERCURY ION", "non-polymer"),
        "EMB": ("METHYL-CARBAMIC ACID ETHYL ESTER", "non-polymer"),
        "EMM": ("[METHYL(VINYL)AMINO]METHANOL", "non-polymer"),
        "EML": (
            "2-{ETHYL[(5-{[6-METHYL-3-(1H-PYRAZOL-4-YL)IMIDAZO[1,2-A]PYRAZIN-8-YL]AMINO}ISOTHIAZOL-3-YL)METHYL]AMINO}-2-METHYLPROPAN-1-OL",
            "non-polymer",
        ),
        "EMO": ("3-METHYL-1,6,8-TRIHYDROXYANTHRAQUINONE", "non-polymer"),
        "EMN": (
            "(2S)-1-{[(2S)-3-(2-METHOXYETHOXY)-2-METHYLPROPYL]OXY}PROPAN-2-AMINE",
            "non-polymer",
        ),
        "EMH": (
            "9-ETHYL-6,6-DIMETHYL-8-[4-(MORPHOLIN-4-YL)PIPERIDIN-1-YL]-11-OXO-6,11-DIHYDRO-5H-BENZO[B]CARBAZOLE-3-CARBONITRILE",
            "non-polymer",
        ),
        "EMK": ("?", "non-polymer"),
        "EMJ": ("2,5-DIHYDROXY-3-UNDECYLCYCLOHEXA-2,5-DIENE-1,4-DIONE", "non-polymer"),
        "EMU": ("N-BENZYL-9H-PURIN-6-AMINE", "non-polymer"),
        "EMT": ("2-(ETHYLMERCURI-THIO)-BENZOIC ACID", "non-polymer"),
        "RGC": ("REIDISPONGIOLIDE C", "non-polymer"),
        "EMQ": (
            "1-[2-(DIMETHYLAMINO)ETHYL]-3-{[(2R,4AS,5R,10BS)-5-PHENYL-9-(TRIFLUOROMETHYL)-3,4,4A,5,6,10B-HEXAHYDRO-2H-PYRANO[3,2-C]QUINOLIN-2-YL]METHYL}UREA",
            "non-polymer",
        ),
        "EMP": (
            "2,4-DIDEOXY-4-(ETHYLAMINO)-3-O-METHYL ALPHA-L-THREO-PENTOPYRANOSIDE",
            "non-polymer",
        ),
        "EMR": ("N-AMINOETHYLMORPHOLINE", "non-polymer"),
        "RGI": ("METHYLCARBAMIC ACID", "peptide"),
        "H89": (
            "6-[4-(BENZYLOXY)PHENYL]-1-(PYRIDIN-4-YLMETHYL)-1H-PYRAZOLO[3,4-B]PYRIDINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "RGL": ("ARGINAL", "peptide"),
        "LDC": (
            "4-AMINO-1-(2-DEOXY-BETA-L-ERYTHRO-PENTOFURANOSYL)PYRIMIDIN-2(1H)-ONE",
            "non-polymer",
        ),
        "LDA": ("LAURYL DIMETHYLAMINE-N-OXIDE", "non-polymer"),
        "MIR": ("MONOETHYLPHOSPHORYLSERINE", "peptide"),
        "LDE": (
            "(2R)-2-{[(S)-[(1R)-1-AMINO-3-METHYLBUTYL](HYDROXY)PHOSPHORYL]METHYL}PENTANEDIOIC ACID",
            "peptide",
        ),
        "W05": ("DELTA-(L-ALPHA-AMINOADIPOYL)-L-CYSTEINYL-D-ALANINE", "non-polymer"),
        "LDH": ("N~6~-ETHYL-L-LYSINE", "peptide"),
        "LDO": ("6-HYDROXY-L-NORLEUCINE", "non-polymer"),
        "LDN": (
            "4-[6-(4-PIPERAZIN-1-YLPHENYL)PYRAZOLO[1,5-A]PYRIMIDIN-3-YL]QUINOLINE",
            "non-polymer",
        ),
        "LDM": ("3-LAURYLAMIDO-N,N'-DIMETHYLPROPYLAMINOXYDE", "non-polymer"),
        "W03": (
            "2,6-DIMETHYL-1-(3-[3-METHYL-5-ISOXAZOLYL]-PROPANYL)-4-[2-METHYL-4-ISOXAZOLYL]-PHENOL",
            "non-polymer",
        ),
        "LDR": ("PHENANTHRIDIN-6(5H)-ONE", "non-polymer"),
        "LDP": ("L-DOPAMINE", "non-polymer"),
        "LDT": ("IDD594", "non-polymer"),
        "LDZ": (
            "N-[(BENZYLOXY)CARBONYL]-L-LEUCYL-N-[(2S)-4-METHYL-1-OXOPENTAN-2-YL]-L-LEUCINAMIDE",
            "non-polymer",
        ),
        "LDY": ("ALPHA-D-LYXOPYRANOSE", "saccharide"),
        "7PC": (
            "2-[4-(AMINOMETHYL)-2-CHLOROPHENOXY]-5-PYRIDIN-2-YLPHENOL",
            "non-polymer",
        ),
        "SU1": (
            "3-[(3-(2-CARBOXYETHYL)-4-METHYLPYRROL-2-YL)METHYLENE]-2-INDOLINONE",
            "non-polymer",
        ),
        "B1Q": (
            "(4S)-1-[(2S,3S,11BS)-2-AMINO-9,10-DIMETHOXY-1,3,4,6,7,11B-HEXAHYDRO-2H-PYRIDO[2,1-A]ISOQUINOLIN-3-YL]-4-(FLUOROMETHYL)PYRROLIDIN-2-ONE",
            "non-polymer",
        ),
        "LD2": (
            "6-[(1E)-2-PHENYL-N-(3-SULFANYL-4H-1,2,4-TRIAZOL-4-YL)ETHANIMIDOYL]-2H-1,4-BENZOXAZIN-3(4H)-ONE",
            "non-polymer",
        ),
        "LD1": (
            "6-[(7S)-7-PHENYL-7H-[1,2,4]TRIAZOLO[3,4-B][1,3,4]THIADIAZIN-6-YL]-2H-1,4-BENZOXAZIN-3(4H)-ONE",
            "non-polymer",
        ),
        "POO": (
            "3-CYCLOHEXYL-1-(2-{METHYL[(1-METHYLPIPERIDIN-3-YL)METHYL]AMINO}-2-OXOETHYL)-2-PHENYL-1H-INDOLE-6-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "KZI": (
            "5-{[(1R,2S)-2-AMINOCYCLOHEXYL]AMINO}-3-[(3,5-DIMETHOXYPHENYL)AMINO]PYRAZINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "QWI": (
            "6-(4-PHENYL-1,9-DIAZASPIRO[5.5]UNDECAN-9-YL)-9H-PURINE",
            "non-polymer",
        ),
        "H8K": (
            "TRANS-4-({[3-(FURAN-3-YL)-2,4-DIHYDROINDENO[1,2-C]PYRAZOL-6-YL]METHYL}AMINO)CYCLOHEXANOL",
            "non-polymer",
        ),
        "B1Z": ("ADENOSYLCOBALAMIN", "non-polymer"),
        "P7A": (
            "1-[5-TERT-BUTYL-2-(1,1-DIOXIDOTHIOMORPHOLIN-4-YL)THIOPHEN-3-YL]-3-NAPHTHALEN-1-YLUREA",
            "non-polymer",
        ),
        "H8H": (
            "N-(5-CHLORO-1,3-BENZODIOXOL-4-YL)-7-[2-(4-METHYLPIPERAZIN-1-YL)ETHOXY]-5-(TETRAHYDRO-2H-PYRAN-4-YLOXY)QUINAZOLIN-4-AMINE",
            "non-polymer",
        ),
        "TFN": (
            "N-(BENZYLSULFONYL)-O-TERT-BUTYL-D-SERYL-N-(4-CARBAMIMIDOYLBENZYL)GLYCINAMIDE",
            "peptide",
        ),
        "EM7": (
            "3-AMINO-N-METHYL-6-[3-(1H-TETRAZOL-5-YL)PHENYL]PYRAZINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "EM1": ("CEM-101", "non-polymer"),
        "KZM": (
            "5-{[(1R,2S)-2-AMINOCYCLOHEXYL]AMINO}-7-[(3,5-DIMETHOXYPHENYL)AMINO]-2-(3-HYDROXYPHENYL)[1,2,4]TRIAZOLO[1,5-C]PYRIMIDINE-8-CARBOXAMIDE",
            "non-polymer",
        ),
        "T0M": ("THIOMERSAL", "non-polymer"),
        "COQ": (
            "2,4-DIAMINO-6-[N-(3',5'-DIMETHOXYBENZYL)-N-METHYLAMINO]PYRIDO[2,3-D]PYRIMIDINE",
            "non-polymer",
        ),
        "COP": (
            "N-(4-CARBOXY-4-{4-[(2,4-DIAMINO-PTERIDIN-6-YLMETHYL)-AMINO]-BENZOYLAMINO}-BUTYL)-PHTHALAMIC ACID",
            "non-polymer",
        ),
        "COS": ("COENZYME A PERSULFIDE", "non-polymer"),
        "COR": ("2,4-DIAMINO-1,5-DIPHENYL-3-HYDROXYPENTANE", "non-polymer"),
        "COU": ("COUMARIN", "non-polymer"),
        "COT": ("COA-S-ACETYL TRYPTAMINE", "non-polymer"),
        "COW": ("ANTHRANILOYL-COENZYME A", "non-polymer"),
        "COV": (
            "(1S,3R,5Z,7E,14BETA,17ALPHA)-17-[(2S,4S)-4-(2-HYDROXY-2-METHYLPROPYL)-2-METHYLTETRAHYDROFURAN-2-YL]-9,10-SECOANDROSTA-5,7,10-TRIENE-1,3-DIOL",
            "non-polymer",
        ),
        "COY": ("CO-(ADENIN-9-YL-PENTYL)-COBALAMIN", "non-polymer"),
        "COX": ("4-(5-METHYL-3-PHENYLISOXAZOL-4-YL)BENZENESULFONAMIDE", "non-polymer"),
        "PEV": (
            "(1S)-2-{[(2-AMINOETHOXY)(HYDROXY)PHOSPHORYL]OXY}-1-[(PALMITOYLOXY)METHYL]ETHYL STEARATE",
            "non-polymer",
        ),
        "COZ": ("COENZYME A", "non-polymer"),
        "PEP": ("PHOSPHOENOLPYRUVATE", "non-polymer"),
        "PEQ": ("L-PHOSPHOLACTATE", "non-polymer"),
        "CIY": ("(2E)-3-(4-HYDROXY-3-METHOXYPHENYL)PROP-2-ENAL", "non-polymer"),
        "ILI": (
            "N-[(1S)-1-({[(3S,4S,7R)-3-HYDROXY-7-METHYL-1-(PYRIDIN-2-YLSULFONYL)-2,3,4,7-TETRAHYDRO-1H-AZEPIN-4-YL]AMINO}CARBONYL)-3-METHYLBUTYL]-1-BENZOFURAN-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "COA": ("COENZYME A", "non-polymer"),
        "PEM": (
            "2-[P-[2-P-CHLOROBENZAMIDO)ETHYL]PHENOXY]-2-METHYLPROPIONIC ACID",
            "non-polymer",
        ),
        "COC": ("COCAINE", "non-polymer"),
        "COB": ("CO-METHYLCOBALAMIN", "non-polymer"),
        "COE": ("FURO[2,3D]PYRIMIDINE ANTIFOLATE", "non-polymer"),
        "COD": ("DEPHOSPHO COENZYME A", "non-polymer"),
        "COG": (
            "2,4-DIAMINO-6-[N-(2',5'-DIMETHOXYBENZYL)-N-METHYLAMINO]QUINAZOLINE",
            "non-polymer",
        ),
        "COF": ("TRIFLUOROACETONYL COENZYME A", "non-polymer"),
        "COI": ("2-OXO-4-METHYLPENTANOIC ACID", "non-polymer"),
        "COH": ("PROTOPORPHYRIN IX CONTAINING CO", "non-polymer"),
        "COK": ("S-(THIOETHYLHYDROXY)COENZYME A", "non-polymer"),
        "COJ": ("HYDROGENOBYRINIC ACID", "non-polymer"),
        "COM": ("1-THIOETHANESULFONIC ACID", "non-polymer"),
        "COL": (
            "2-(OXALYL-AMINO)-4,7-DIHYDRO-5H-THIENO[2,3-C]THIOPYRAN-3-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "COO": ("CROTONYL COENZYME A", "non-polymer"),
        "CON": ("COBALT TETRAAMMINE ION", "non-polymer"),
        "LIC": (
            "3-(2-CHLOROBENZYL)-1-(2-{[(1S)-2-HYDROXY-1,2-DIMETHYLPROPYL]AMINO}PYRIMIDIN-4-YL)-1-(4-METHOXYPHENYL)UREA",
            "non-polymer",
        ),
        "TFH": ("NITROGEN OF TRIFLUORO-ETHYLHYDRAZINE", "non-polymer"),
        "4EA": (
            "(1S)-2-(1H-INDOL-3-YL)-1-[({5-[(E)-2-PYRIDIN-4-YLVINYL]PYRIDIN-3-YL}OXY)METHYL]ETHYLAMINE",
            "non-polymer",
        ),
        "NQG": (
            "5-AZANYLIDYNE-N-[(2S)-4-ETHOXY-2-HYDROXY-4-OXOBUTANOYL]-L-NORVALYL-L-ARGINYL-L-TRYPTOPHANAMIDE",
            "non-polymer",
        ),
        "OPT": (
            "S-BENZYL-N-(BIPHENYL-4-YLACETYL)-L-CYSTEINYL-N~5~-(DIAMINOMETHYL)-D-ORNITHYL-N-(2-PHENYLETHYL)-L-TYROSINAMIDE",
            "non-polymer",
        ),
        "TJ2": (
            "3-{(1S)-2-(TERT-BUTYLAMINO)-1-[(4-CHLOROBENZYL)(FORMYL)AMINO]-2-OXOETHYL}-6-CHLORO-1H-INDOLE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "LIL": ("2-TRIDECANOYLOXY-PENTADECANOIC ACID", "non-polymer"),
        "CR3": (
            "2-{5-[AMINO(IMINIO)METHYL]-1H-INDOL-2-YL}-6-(CYCLOPENTYLOXY)BENZENOLATE",
            "non-polymer",
        ),
        "YF1": ("2,5-DIMETHYLPYRIMIDIN-4-AMINE", "non-polymer"),
        "SHC": ("S-HEXYLCYSTEINE", "peptide"),
        "OPP": (
            "1-[PYRROL-1-YL-2,5-DIONE-METHOXYMETHYL]-PYRROLE-2,5-DIONE",
            "non-polymer",
        ),
        "OPQ": (
            "((3S,4S)-4-((2S,3S)-2-(2-(NAPTHALEN-6-YL)ACETAMIDO)-3-METHYLPENTANAMIDO)-N-BUTYL-3-HYDROXY-6-METHYLHEPTANAMIDE)",
            "non-polymer",
        ),
        "CO1": (
            "17-[3-(3,3-DIMETHYL-OXIRANYL)-1-METHYL-PROPYL]-10,13-DIMETHYL-2,3,4,7,8,9,10,11,12,13,14,15,16,17-TETRADECAHYDRO-1H-CYCLOPENTA[A]PHENANTHREN-3-OL",
            "non-polymer",
        ),
        "CO3": ("CARBONATE ION", "non-polymer"),
        "CO2": ("CARBON DIOXIDE", "non-polymer"),
        "CO5": ("COBALT ION,5 WATERS COORDINATED", "non-polymer"),
        "CO4": (
            "2,4-DIAMINO-5-METHYL-6-[(3,4,5-TRIMETHOXY-N-METHYLANILINO)METHYL]PYRIDO[2,3-D]PYRIMIDINE",
            "non-polymer",
        ),
        "CO7": ("?", "non-polymer"),
        "CO6": ("ISOBUTYRYL-COENZYME A", "non-polymer"),
        "CO9": (
            "1-({2-[(1S)-1-AMINOETHYL]-1,3-OXAZOL-4-YL}CARBONYL)-L-PROLYL-L-TRYPTOPHAN",
            "non-polymer",
        ),
        "CO8": ("OCTANOYL-COENZYME A", "non-polymer"),
        "PE6": ("PHOSPHORYL-HEXAETHYLENE GLYCOL", "non-polymer"),
        "CR5": (
            "(2R)-2-(AMINOMETHYL)-2,4-DIHYDROXY-5-OXO-3-(2-OXOETHYL)-2,5-DIHYDRO-1H-IMIDAZOL-3-IUM",
            "peptide",
        ),
        "PE0": ("PTERINE", "non-polymer"),
        "F6B": ("N-(2,6-DIFLOURO-BENZYL)-4-SULFAMOYL-BENZAMIDE", "non-polymer"),
        "F6A": ("N-BIPHENYL-3-YL-2-(TRIFLUOROMETHYL)BENZAMIDE", "non-polymer"),
        "OET": ("ETHYL GROUP", "non-polymer"),
        "IL2": (
            "(5E)-5-[(3AS,4R,5R,6AS)-5-HYDROXY-4-[(1E,3S,4R)-3-HYDROXY-4-METHYLOCT-1-EN-6-YN-1-YL]HEXAHYDROPENTALEN-2(1H)-YLIDENE]PENTANOIC ACID",
            "non-polymer",
        ),
        "F6Z": ("?", "non-polymer"),
        "F6Y": ("?", "non-polymer"),
        "OEC": ("OXYGEN EVOLVING SYSTEM", "non-polymer"),
        "UFR": ("?", "nucleic-acid"),
        "OEG": ("?", "non-polymer"),
        "F6R": ("FRUCTOSE -6-PHOSPHATE", "non-polymer"),
        "F6P": ("FRUCTOSE-6-PHOSPHATE", "saccharide"),
        "GNR": (
            "2-(3,4-DIHYDRO-3-OXO-2H-BENZO[B][1,4]THIAZIN-2-YL)-N-HYDROXYACETAMIDE",
            "non-polymer",
        ),
        "GNS": ("N-SULFO-ALPHA-D-GLUCOSAMINE", "non-polymer"),
        "GNP": ("PHOSPHOAMINOPHOSPHONIC ACID-GUANYLATE ESTER", "non-polymer"),
        "GNQ": (
            "N-[(1R)-1-[(BENZYLSULFONYL)METHYL]-2-{[(1S)-1-METHYL-2-{[4-(TRIFLUOROMETHOXY)PHENYL]AMINO}ETHYL]AMINO}-2-OXOETHYL]MORPHOLINE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "MKC": ("6-BENZYL-1-ETHOXYMETHYL-5-ISOPROPYL URACIL", "non-polymer"),
        "AIX": (
            "(2R,4S)-2-[(1R)-1-{[(2R)-2-AMINO-2-PHENYLACETYL]AMINO}-2-OXOETHYL]-5,5-DIMETHYL-1,3-THIAZOLIDINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "GNT": ("(-)-GALANTHAMINE", "non-polymer"),
        "AIZ": (
            "3-{6-[(2-CHLOROPHENYL)AMINO]-1H-INDAZOL-3-YL}-5-{[4-(DIMETHYLAMINO)BUTANOYL]AMINO}BENZOIC ACID",
            "non-polymer",
        ),
        "AIU": (
            "(2S,3R)-2-(4-{2-[(3S,4S)-3,4-DIMETHYLPYRROLIDIN-1-YL]ETHOXY}PHENYL)-3-(4-HYDROXYPHENYL)-2,3-DIHYDRO-1,4-BENZOXATHIIN-6-OL",
            "non-polymer",
        ),
        "AIT": (
            "(2S,3R)-3-(4-HYDROXYPHENYL)-2-(4-{[(2R)-2-PYRROLIDIN-1-YLPROPYL]OXY}PHENYL)-2,3-DIHYDRO-1,4-BENZOXATHIIN-6-OL",
            "non-polymer",
        ),
        "GNX": ("2-DEOXY-3-O-SULFO-2-(SULFOAMINO)-ALPHA-D-GLUCOPYRANOSE", "saccharide"),
        "AIQ": (
            "2,6-DIAMINO-8-(1H-IMIDAZOL-2-YLSULFANYLMETHYL)-3H-QUINAZOLINE-4-ONE",
            "non-polymer",
        ),
        "T6P": ("TREHALOSE-6-PHOSPHATE", "saccharide"),
        "AIS": ("5-AMINOIMIDAZOLE RIBONUCLEOSIDE", "non-polymer"),
        "AIR": ("5-AMINOIMIDAZOLE RIBONUCLEOTIDE", "non-polymer"),
        "AIM": ("4-(4-AMINO-CYCLOHEXYLIDENE)-4H-IMIDAZOL-2-YLAMINE", "non-polymer"),
        "AIL": (
            "N-{[(3S,5S)-5-(1,3-THIAZOLIDIN-3-YLCARBONYL)PYRROLIDIN-3-YL]METHYL}-1,3-THIAZOLE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "AIO": (
            "[(2R)-1-(L-ALANYL-L-ISOLEUCYL)PYRROLIDIN-2-YL]BORONIC ACID",
            "non-polymer",
        ),
        "AIN": ("2-(ACETYLOXY)BENZOIC ACID", "non-polymer"),
        "GNF": (
            "N~2~-1,3-BENZOXAZOL-2-YL-3-CYCLOHEXYL-N-{2-[(4-METHOXYPHENYL)AMINO]ETHYL}-L-ALANINAMIDE",
            "non-polymer",
        ),
        "AIH": (
            "(2S,3R)-2-(4-{2-[(3R,4R)-3,4-DIMETHYLPYRROLIDIN-1-YL]ETHOXY}PHENYL)-3-(4-HYDROXYPHENYL)-2,3-DIHYDRO-1,4-BENZOXATHIIN-6-OL",
            "non-polymer",
        ),
        "AIK": (
            "N-[3-(DIMETHYLAMINO)PROPYL]-2-({[4-({[4-(FORMYLAMINO)-1-METHYL-1H-PYRROL-2-YL]CARBONYL}AMINO)-1-METHYL-1H-PYRROL-2-YL]CARBONYL}AMINO)-5-ISOPROPYL-1,3-THIAZOLE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "AIJ": (
            "(2S,3R)-3-(4-HYDROXYPHENYL)-2-(4-{[(2S)-2-PYRROLIDIN-1-YLPROPYL]OXY}PHENYL)-2,3-DIHYDRO-1,4-BENZOXATHIIN-6-OL",
            "non-polymer",
        ),
        "T6E": (
            "1-[4-(4-AMINOTHIENO[2,3-D]PYRIMIDIN-5-YL)PHENYL]-3-[2-FLUORO-5-(TRIFLUOROMETHYL)PHENYL]UREA",
            "non-polymer",
        ),
        "AIG": (
            "4-AMINO-2-HEXYLOXY-6-HYDROXYMETHYL-TETRAHYDRO-PYRAN-3,5-DIOL",
            "saccharide",
        ),
        "AIF": (
            "[(2R,3S,4R,5E)-5-[(5-AMINO-2,6-DIOXO-3H-PYRIMIDIN-4-YL)IMINO]-2,3,4-TRIHYDROXY-PENTYL] DIHYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "AIA": (
            "(1S)-2-[(2S,5R)-2-(AMINOMETHYL)-5-ETHYNYLPYRROLIDIN-1-YL]-1-CYCLOPENTYL-2-OXOETHANAMINE",
            "non-polymer",
        ),
        "AIC": (
            "(2S,5R,6R)-6-{[(2R)-2-AMINO-2-PHENYLETHANOYL]AMINO}-3,3-DIMETHYL-7-OXO-4-THIA-1-AZABICYCLO[3.2.0]HEPTANE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "AIB": ("ALPHA-AMINOISOBUTYRIC ACID", "peptide"),
        "NSI": (
            "3-(4-METHOXYPHENYL)-N-(PHENYLSULFONYL)-1-[3-(TRIFLUOROMETHYL)BENZYL]-1H-INDOLE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "NSK": ("N-SUCCINYL LYSINE", "peptide"),
        "NSM": ("NITROSOMETHANE", "non-polymer"),
        "NSN": ("N~2~-(2-AMINO-1-METHYL-2-OXOETHYLIDENE)ASPARAGINATE", "non-polymer"),
        "XAD": (
            "9-(2,3-DIDEOXY-6-O-PHOSPHONO-BETA-D-ERYTHRO-HEXOPYRANOSYL)-9H-PURIN-6-AMINE",
            "nucleic-acid",
        ),
        "NSA": ("2,2-DICHLORO-N-PHENYL-ACETAMIDE", "non-polymer"),
        "NSB": ("N,N,N-TRIMETHYL-3-SULFOPROPAN-1-AMINIUM", "non-polymer"),
        "NSC": ("N,N'-BIS(4-AMINO-2-METHYLQUINOLIN-6-YL)UREA", "non-polymer"),
        "NSD": ("N-(3-AMINOPROPYL)PROPANE-1,3-DIAMINE", "non-polymer"),
        "M06": (
            "3-CHLORO-5-{[4-METHYL-2-OXO-1-(2H-PYRAZOLO[3,4-B]PYRIDIN-3-YLMETHYL)-1,2-DIHYDROPYRIDIN-3-YL]OXY}BENZONITRILE",
            "non-polymer",
        ),
        "M07": (
            "(5R,7R,8S,9S,10R)-7-(HYDROXYMETHYL)-3-(4-METHOXYPHENYL)-1,6-DIOXA-2-AZASPIRO[4.5]DEC-2-ENE-8,9,10-TRIOL",
            "non-polymer",
        ),
        "M04": (
            "1-[4-(4-CHLOROBENZYL)-1-(7H-PYRROLO[2,3-D]PYRIMIDIN-4-YL)PIPERIDIN-4-YL]METHANAMINE",
            "non-polymer",
        ),
        "M05": (
            "4-(4-CHLOROBENZYL)-1-(7H-PYRROLO[2,3-D]PYRIMIDIN-4-YL)PIPERIDIN-4-AMINIUM",
            "non-polymer",
        ),
        "M02": ("1-(9H-PURIN-6-YL)PIPERIDIN-4-AMINE", "non-polymer"),
        "M03": (
            "1-[4-(4-CHLOROPHENYL)-1-(7H-PYRROLO[2,3-D]PYRIMIDIN-4-YL)PIPERIDIN-4-YL]METHANAMINE",
            "non-polymer",
        ),
        "M01": ("1-[1-(9H-PURIN-6-YL)PIPERIDIN-4-YL]METHANAMINE", "non-polymer"),
        "NSP": ("5-(AMINOMETHYL)-2-METHYLPYRIMIDIN-4-AMINE", "non-polymer"),
        "NSS": ("5'-O-[N-(L-ASPARAGINYL)SULFAMOYL]ADENOSINE", "non-polymer"),
        "NST": (
            "3-{[(3-NITROANILINE]SULFONYL}THIOPHENE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "M08": (
            "(5R,7R,8S,9S,10R)-7-(HYDROXYMETHYL)-3-PHENYL-1,6-DIOXA-2-AZASPIRO[4.5]DEC-2-ENE-8,9,10-TRIOL",
            "non-polymer",
        ),
        "M09": (
            "(3S,5R,7R,8S,9S,10R)-7-(HYDROXYMETHYL)-3-(4-NITROPHENYL)-1,6-DIOXA-2-AZASPIRO[4.5]DECANE-8,9,10-TRIOL",
            "non-polymer",
        ),
        "RKT": ("PYRAZINO[2,3-F]QUINOXALINE", "non-polymer"),
        "M0E": ("MOENOMYCIN", "non-polymer"),
        "EAH": (
            "(5S,7E,9E,11Z,14Z)-5-HYDROXYICOSA-7,9,11,14-TETRAENOIC ACID",
            "non-polymer",
        ),
        "M0N": (
            "(1-HYDROXY-2-IMIDAZO[1,2-A]PYRIDIN-3-YLETHANE-1,1-DIYL)BIS(PHOSPHONIC ACID)",
            "non-polymer",
        ),
        "M0H": ("S-(HYDROXYMETHYL)-L-CYSTEINE", "peptide"),
        "NS9": (
            "(2S,3S)-3-{[7-(BENZYLAMINO)-3-(1-METHYLETHYL)PYRAZOLO[1,5-A]PYRIMIDIN-5-YL]AMINO}BUTANE-1,2,4-TRIOL",
            "non-polymer",
        ),
        "M0T": (
            "1-(3,4-DICHLOROBENZYL)-2-METHYL-N-[(1R)-1-PHENYLPROPYL]-1H-BENZIMIDAZOLE-5-CARBOXAMIDE",
            "non-polymer",
        ),
        "RKD": (
            "(2S)-2-(4-CHLOROPHENYL)-2-HYDROXY-2-[4-(1H-PYRAZOL-4-YL)PHENYL]ETHANAMINIUM",
            "non-polymer",
        ),
        "NS1": ("15-TRANS-1,2-DIHYDRONEUROSPORENE", "non-polymer"),
        "NS3": (
            "(3S)-3-CYCLOPENTYL-6-METHYL-7-[(4-METHYLPIPERAZIN-1-YL)SULFONYL]-3,4-DIHYDRO-2H-1,2,4-BENZOTHIADIAZINE 1,1-DIOXIDE",
            "non-polymer",
        ),
        "LII": (
            "(Z)-6-(2-[2,5-DIMETHOXYPHENYL]ETHEN-1-YL)-2,4-DIAMINO-5-METHYLPYRIDO[2,3-D]PYRIMIDINE",
            "non-polymer",
        ),
        "NS5": ("15-CIS-1,2-DIHYDRONEUROSPORENE", "non-polymer"),
        "NS6": (
            "(3R)-3-CYCLOPENTYL-7-[(4-METHYLPIPERAZIN-1-YL)SULFONYL]-3,4-DIHYDRO-2H-1,2-BENZOTHIAZINE 1,1-DIOXIDE",
            "non-polymer",
        ),
        "NS7": (
            "(3R)-3-CYCLOPENTYL-6-METHYL-7-[(4-METHYLPIPERAZIN-1-YL)SULFONYL]-3,4-DIHYDRO-2H-1,2-BENZOTHIAZINE 1,1-DIOXIDE",
            "non-polymer",
        ),
        "SDT": (
            "3,4-DIHYDROXY-9,10-SECOANDROSTA-1(10),2,4-TRIENE-9,17-DIONE",
            "non-polymer",
        ),
        "T07": (
            "5-[(2,4-DICHLOROBENZYL)SULFANYL]-4H-1,2,4-TRIAZOL-3-AMINE",
            "non-polymer",
        ),
        "GN1": (
            "2-(ACETYLAMINO)-2-DEOXY-1-O-PHOSPHONO-ALPHA-D-GLUCOPYRANOSE",
            "saccharide",
        ),
        "AI9": (
            "[(2R,3S,4S,5E)-5-[[5-AZANYL-2,4-BIS(OXIDANYLIDENE)-1H-PYRIMIDIN-6-YL]IMINO]-2,3,4-TRIS(OXIDANYL)PENTYL] DIHYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "GN7": ("N7-2'-DEOXY-GUANOSINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "SDR": ("(2S)-2-[(N-ACETYL-L-ALPHA-ASPARTYL)AMINO]NONANOIC ACID", "peptide"),
        "SDS": ("DODECYL SULFATE", "non-polymer"),
        "AI7": ("3-(HEPTYLOXY)BENZOIC ACID", "non-polymer"),
        "T66": ("2-AMINOETHYLLYSINE-CARBONYLMETHYLENE-THYMINE", "peptide"),
        "AI1": ("N-BENZYL-3-(ALPHA-D-GALACTOS-1-YL)-BENZAMIDE", "non-polymer"),
        "T05": (
            "4-{[(3S)-3-(5-METHYL-2,4-DIOXO-3,4-DIHYDROPYRIMIDIN-1(2H)-YL)PIPERIDIN-1-YL]SULFONYL}-2-(3-METHYLPHENOXY)BENZOIC ACID",
            "non-polymer",
        ),
        "AI3": (
            "2,3-DIMETHOXY-12H-[1,3]DIOXOLO[5,6]INDENO[1,2-C]ISOQUINOLIN-6-IUM",
            "non-polymer",
        ),
        "AI2": (
            "3A-METHYL-5,6-DIHYDRO-FURO[2,3-D][1,3,2]DIOXABOROLE-2,2,6,6A-TETRAOL",
            "non-polymer",
        ),
        "SDD": ("D-SORBOSE", "saccharide"),
        "EAA": ("ETHACRYNIC ACID", "non-polymer"),
        "MK5": (
            "2-[(DICHLOROACETYL)AMINO]ETHYL [(2S,3R)-3-HYDROXY-4-{[(4-METHOXYPHENYL)SULFONYL][(2S)-2-METHYLBUTYL]AMINO}-1-PHENYLBUTAN-2-YL]CARBAMATE",
            "non-polymer",
        ),
        "SDG": (
            "2-AMINO-9-(2-DEOXY-5-O-PHOSPHONO-BETA-D-ERYTHRO-PENTOFURANOSYL)-9H-PURINE-6-SELENOL",
            "nucleic-acid",
        ),
        "MK3": (
            "6-{5-[(2-AMINOPYRIMIDIN-4-YL)AMINO]-2-HYDROXYPHENYL}-N-METHYLIDENE-1-BENZOTHIOPHENE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "MK2": (
            "N~4~-[7-(1-BENZOFURAN-2-YL)-1H-INDAZOL-5-YL]PYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "MK1": (
            "N-[2(R)-HYDROXY-1(S)-INDANYL]-5-[(2(S)-TERTIARY BUTYLAMINOCARBONYL)-4(3-PYRIDYLMETHYL)PIPERAZINO]-4(S)-HYDROXY-2(R)-PHENYLMETHYLPENTANAMIDE",
            "non-polymer",
        ),
        "MK0": (
            "1-O-[(S)-HYDROXY{[(4S,8S,16S,20S)-4,8,12,16,20-PENTAMETHYLHEPTACOSYL]OXY}PHOSPHORYL]-BETA-D-MANNOPYRANOSE",
            "non-polymer",
        ),
        "T03": ("3-[(4-FLUOROBENZYL)SULFANYL]-4H-1,2,4-TRIAZOLE", "non-polymer"),
        "LIJ": (
            "BENZYL [(1S)-2-({(1S,2R)-1-BENZYL-2-HYDROXY-3-[(3-METHOXYBENZYL)AMINO]PROPYL}AMINO)-2-OXO-1-{[(1-PROPYLBUTYL)SULFONYL]METHYL}ETHYL]CARBAMATE",
            "non-polymer",
        ),
        "EAC": ("ETHYL 3-OXOBUTANOATE", "non-polymer"),
        "MK8": ("2-METHYL-L-NORLEUCINE", "peptide"),
        "D14": (
            "7-CHLORO-N-[(1S,2R,4S)-4-(DIMETHYLCARBAMOYL)-2-{[(5-METHYL-5,6-DIHYDRO-4H-PYRROLO[3,4-D][1,3]THIAZOL-2-YL)CARBONYL]AMINO}CYCLOHEXYL]ISOQUINOLINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "EAB": (
            "(3-{(E)-[3-(AMINOMETHYL)PHENYL]DIAZENYL}PHENYL)ACETIC ACID",
            "non-polymer",
        ),
        "EPZ": ("?", "non-polymer"),
        "EPY": (
            "1-HYDROXY-2-S-GLUTATHIONYL-3-PARA-NITROPHENOXY-PROPANE",
            "non-polymer",
        ),
        "EPT": ("HEPTANYL-P-PHENOL", "non-polymer"),
        "EPU": (
            "URIDINE-DIPHOSPHATE-2(N-ACETYLGLUCOSAMINYL) BUTYRIC ACID",
            "non-polymer",
        ),
        "EPS": (
            "5-[(1-CARBOXYVINYL)OXY]-4-HYDROXY-3-(PHOSPHONOOXY)CYCLOHEX-1-ENE-1-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "EPQ": ("(4S)-4-AMINO-6-FLUORO-N,N-DIMETHYL-5-OXOHEXANAMIDE", "non-polymer"),
        "EPN": ("3-(4-NITRO-PHENOXY)-PROPAN-1-OL", "non-polymer"),
        "EPO": ("(2R,3R)-OXIRANE-2,3-DICARBOXYLIC ACID", "non-polymer"),
        "EPL": ("DIETHYL 4-METHOXYPHENYL PHOSPHATE", "non-polymer"),
        "D17": (
            "N-({3-[({4-[(Z)-(4-OXO-2-THIOXO-1,3-THIAZOLIDIN-5-YLIDENE)METHYL]PHENYL}AMINO)METHYL]PHENYL}CARBONYL)-L-GLUTAMIC ACID",
            "non-polymer",
        ),
        "EPJ": ("EPIBATIDINE", "non-polymer"),
        "IMF": ("TETRA(IMIDAZOLE)DIAQUACOPPER (I)", "non-polymer"),
        "EPH": (
            "L-ALPHA-PHOSPHATIDYL-BETA-OLEOYL-GAMMA-PALMITOYL-PHOSPHATIDYLETHANOLAMINE",
            "non-polymer",
        ),
        "EPI": ("4-ETHYLPIPERIDINE", "non-polymer"),
        "EPG": (
            "2-HYDROXYMETHYL-6-OXIRANYLMETHOXY-TETRAHYDRO-PYRAN-3,4,5-TRIOL",
            "saccharide",
        ),
        "EPD": ("EPOTHILONE D", "non-polymer"),
        "EPE": ("4-(2-HYDROXYETHYL)-1-PIPERAZINE ETHANESULFONIC ACID", "non-polymer"),
        "EPB": (
            "7,11-DIHYDROXY-8,8,10,12,16-PENTAMETHYL-3-[1-METHYL-2-(2-METHYL-THIAZOL-4-YL)VINYL]-4,17-DIOXABICYCLO[14.1.0]HEPTADECANE-5,9-DIONE",
            "non-polymer",
        ),
        "EPC": (
            "(1S)-1-[((1E)-{3-HYDROXY-2-METHYL-5-[(PHOSPHONOOXY)METHYL]PYRIDIN-4-YL}METHYLENE)AMINO]ETHYLPHOSPHONIC ACID",
            "non-polymer",
        ),
        "EPA": ("5,8,11,14,17-EICOSAPENTAENOIC ACID", "non-polymer"),
        "LYT": ("BUTYLAMINE", "non-polymer"),
        "CM9": (
            "CIS-4-METHYL-N-[(1S)-3-(METHYLSULFANYL)-1-(PYRIDIN-4-YLCARBAMOYL)PROPYL]CYCLOHEXANECARBOXAMIDE",
            "non-polymer",
        ),
        "LYW": ("1-AMINO-1-CARBONYL PENTANE", "non-polymer"),
        "LYP": ("N~6~-METHYL-N~6~-PROPYL-L-LYSINE", "non-polymer"),
        "LYR": (
            "N~6~-[(2Z,4E,6E,8E)-3,7-DIMETHYL-9-(2,6,6-TRIMETHYLCYCLOHEX-1-EN-1-YL)NONA-2,4,6,8-TETRAENYL]LYSINE",
            "peptide",
        ),
        "LYS": ("LYSINE", "peptide"),
        "POB": (
            "2-[(1-AMINO-ETHYL)-PHOSPHATE-PHOSPHINOYLOXY]-BUTYRIC ACID",
            "non-polymer",
        ),
        "LYX": ("?", "peptide"),
        "R8D": (
            "3-CHLORO-5-[2-CHLORO-5-(1H-PYRAZOLO[3,4-B]PYRIDIN-3-YLMETHOXY)PHENOXY]BENZONITRILE",
            "non-polymer",
        ),
        "LYZ": ("5-HYDROXYLYSINE", "peptide"),
        "LYD": (
            "2-{4-[2-(2-AMINO-4-OXO-4,7-DIHYDRO-3H-PYRROLO[2,3-D]PYRIMIDIN-5-YL)-ETHYL]-BENZOYLAMINO}-3-METHYL-BUTYRIC ACID",
            "non-polymer",
        ),
        "LYF": (
            "(2E,4S)-2-{[(5S)-5-AMINO-5-CARBOXYPENTYL]IMINO}-4-HYDROXYHEPTANEDIOIC ACID",
            "peptide",
        ),
        "LYG": (
            "8-CHLORO-2-{[(3S)-3-HYDROXYPYRROLIDIN-1-YL]METHYL}[1]BENZOTHIENO[3,2-D]PYRIMIDIN-4(3H)-ONE",
            "non-polymer",
        ),
        "LYA": (
            "2-{4-[2-(2-AMINO-4-OXO-4,7-DIHYDRO-3H-PYRROLO[2,3-D]PYRIMIDIN-5-YL)-ETHYL]-BENZOYLAMINO}-PENTANEDIOIC ACID",
            "non-polymer",
        ),
        "LYB": (
            "2-{4-[4-(4-{4-[2-(2-AMINO-4-OXO-4,7-DIHYDRO-3H-PYRROLO[2,3-D]PYRIMIDIN-5-YL)-ETHYL]-BENZOYLAMINO}-4-CARBOXY-BUTYRYLAMINO)-4-CARBOXY-BUTYRYLAMINO}-PENTANEDIOIC ACID",
            "non-polymer",
        ),
        "LYC": ("LYCOPENE", "non-polymer"),
        "LYL": ("2-ALLYL-6-METHYL-PHENOL", "non-polymer"),
        "LYM": ("DEOXY-METHYL-LYSINE", "peptide"),
        "LYN": ("2,6-DIAMINO-HEXANOIC ACID AMIDE", "peptide"),
        "LYO": ("4-HYDROXY-LYSINE", "peptide"),
        "LYH": ("3,4 ENE-LYSINE", "peptide"),
        "LYJ": ("(2S)-2,6-DIAMINOHEXAN-1-OL", "non-polymer"),
        "LYK": ("(2S)-2,6-DIAMINOHEXANE-1,1-DIOL", "peptide"),
        "LY4": (
            "(9R)-9-[(DIMETHYLAMINO)METHYL]-6,7,10,11-TETRAHYDRO-9H,18H-5,21:12,17-DIMETHENODIBENZO[E,K]PYRROLO[3,4-H][1,4,13]OXADIAZACYCLOHEXADECINE-18,20-DIONE",
            "non-polymer",
        ),
        "LY5": (
            "(3S,4AR,6S,8AR)-6-{[(2S)-2-CARBOXY-4,4-DIFLUOROPYRROLIDIN-1-YL]METHYL}DECAHYDROISOQUINOLINE-3-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "LY7": (
            "N-[(2S)-2-(4'-CYANOBIPHENYL-4-YL)PROPYL]PROPANE-2-SULFONAMIDE",
            "non-polymer",
        ),
        "LY0": (
            "(2R)-3-[(R)-[(1R)-1-AMINOETHYL](HYDROXY)PHOSPHORYL]-2-METHYLPROPANOIC ACID",
            "non-polymer",
        ),
        "LY1": ("8,9-DICHLORO-2,3,4,5-TETRAHYDRO-1H-BENZO[C]AZEPINE", "non-polymer"),
        "LY2": ("2-MORPHOLIN-4-YL-7-PHENYL-4H-CHROMEN-4-ONE", "non-polymer"),
        "LY3": (
            "2-{4-[2-(2-AMINO-4-OXO-4,7-DIHYDRO-3H-PYRROLO[2,3-D]PYRIMIDIN-5-YL)-ETHYL]-BENZOYLAMINO}-4-(2H-TETRAZOL-5-YL)-BUTYRIC ACID",
            "non-polymer",
        ),
        "BQN": (
            "4-{(2E)-3-[4-{[2-(PYRIDIN-3-YLMETHOXY)PHENYL]SULFANYL}-2,3-BIS(TRIFLUOROMETHYL)PHENYL]PROP-2-ENOYL}MORPHOLINE",
            "non-polymer",
        ),
        "D18": ("2,5-BIS{[4-(N-ISOPROPYL)DIAMINOMETHYL]PHENYL}FURAN", "non-polymer"),
        "BQM": (
            "3-({4-[(1E)-3-MORPHOLIN-4-YL-3-OXOPROP-1-EN-1-YL]-2,3-BIS(TRIFLUOROMETHYL)PHENYL}SULFANYL)ANILINE",
            "non-polymer",
        ),
        "LY9": ("L-ARABINARIC ACID", "non-polymer"),
        "D19": ("2,5-BIS{[4-(N-CYCLOPROPYLDIAMINOMETHYL)PHENYL]}FURAN", "non-polymer"),
        "R88": (
            "(4-BROMOPHENYL)[4-({(2E)-4-[CYCLOPROPYL(METHYL)AMINO]BUT-2-ENYL}OXY)PHENYL]METHANONE",
            "non-polymer",
        ),
        "SE": ("SELENIUM ATOM", "non-polymer"),
        "EAP": ("2-AMINO-VINYL-PHOSPHATE", "non-polymer"),
        "R83": (
            "N-{2,4-DIMETHOXY-5-[(2-PIPERIDIN-1-YLBENZYL)SULFAMOYL]PHENYL}ACETAMIDE",
            "non-polymer",
        ),
        "R82": (
            "N-(2,4-DIMETHOXY-5-{[(2R)-2-METHYL-2,3-DIHYDRO-1H-INDOL-1-YL]SULFONYL}PHENYL)ACETAMIDE",
            "non-polymer",
        ),
        "R84": (
            "4-{[1-(2-{[({5-[(3-CARBOXYPROPANOYL)AMINO]-2,4-DIMETHOXYPHENYL}SULFONYL)AMINO]METHYL}PHENYL)PIPERIDIN-4-YL]METHOXY}-4-OXOBUTANOIC ACID",
            "non-polymer",
        ),
        "R86": (
            "2-{(2R)-4-[2-(1H-IMIDAZOL-1-YL)-6-METHYLPYRIMIDIN-4-YL]-1-ISOBUTYRYLPIPERAZIN-2-YL}-N-[2-(4-METHOXYPHENYL)ETHYL]ACETAMIDE",
            "non-polymer",
        ),
        "GME": ("5-O-METHYL-GLUTAMIC ACID", "peptide"),
        "EP9": (
            "N-[(BENZYLOXY)CARBONYL]-L-LEUCYL-N-[(2S,3S)-3-HYDROXY-1-(4-HYDROXYPHENYL)-4-OXOBUTAN-2-YL]-L-LEUCINAMIDE",
            "peptide",
        ),
        "EP4": ("5'-DEOXY-5'-(DIMETHYLAMINO)ADENOSINE", "non-polymer"),
        "EP2": (
            "METHYL N-[(2S)-4-{[(1S)-1-{[(2S)-2-CARBOXYPYRROLIDIN-1-YL]CARBONYL}-3-METHYLBUTYL]AMINO}-2-HYDROXY-4-OXOBUTANOYL]-L-LEUCYLGLYCYLGLYCINATE",
            "peptide",
        ),
        "EP0": (
            "N-[(3R)-4-ETHOXY-3-HYDROXY-4-OXOBUTANOYL]-L-ISOLEUCYL-L-PROLINE",
            "non-polymer",
        ),
        "EP1": (
            "3-[4-(2-HYDROXYETHYL)PIPERAZIN-1-YL]PROPANE-1-SULFONIC ACID",
            "non-polymer",
        ),
        "UMF": (
            "1-(2-DEOXY-2-FLUORO-3-O-PHOSPHONO-BETA-L-RIBOFURANOSYL)PYRIMIDINE-2,4(1H,3H)-DIONE",
            "non-polymer",
        ),
        "SHM": ("HOMOBIOTIN", "non-polymer"),
        "YJC": (
            "(4R)-1-{[1-(4-CHLOROPHENYL)CYCLOPROPYL]CARBONYL}-4-[(2-CHLOROPHENYL)SULFONYL]-N-{1-[(E)-IMINOMETHYL]CYCLOPROPYL}-L-PROLINAMIDE",
            "non-polymer",
        ),
        "YJD": ("4-(2-PROPAN-2-YLOXYBENZIMIDAZOL-1-YL)PHENOL", "non-polymer"),
        "HOE": (
            "2'-(4-HYDROXYPHENYL)-5-(4-METHYL-1-PIPERAZINYL)-2,5'-BI-BENZIMIDAZOLE",
            "non-polymer",
        ),
        "YJW": ("4-(5-METHYL-4-PHENYLISOXAZOL-3-YL)BENZENE-1,3-DIOL", "non-polymer"),
        "YJX": (
            "1-(3H-IMIDAZO[4,5-C]PYRIDIN-2-YL)-3,4-DIHYDROPYRIDO[2,1-A]ISOINDOL-6(2H)-ONE",
            "non-polymer",
        ),
        "1U2": (
            "6-[(PHENYLCARBAMOYL)AMINO]NAPHTHALENE-2-CARBOXIMIDAMIDE",
            "non-polymer",
        ),
        "1U4": (
            "N-[(S)-({[(BENZYLOXY)CARBONYL]AMINO}METHYL)(HYDROXY)PHOSPHORYL]-L-PHENYLALANYL-L-LEUCINE",
            "non-polymer",
        ),
        "1U9": (
            "6-{(E)-2-[3-(2-HYDROXYETHYL)PHENYL]ETHENYL}NAPHTHALENE-2-CARBOXIMIDAMIDE",
            "non-polymer",
        ),
        "4RB": (
            "4-(6-{[(1R)-1-(HYDROXYMETHYL)PROPYL]AMINO}IMIDAZO[1,2-B]PYRIDAZIN-3-YL)BENZOIC ACID",
            "non-polymer",
        ),
        "4RG": ("3-(PHENOXYMETHYL)BENZOIC ACID", "non-polymer"),
        "4RR": (
            "(4R)-4-[3-(CYCLOPENTYLOXY)-4-METHOXYPHENYL]PYRROLIDIN-2-ONE",
            "non-polymer",
        ),
        "1UP": (
            "2-[(7-CARBAMIMIDOYL-2-METHOXYNAPHTHALEN-1-YL)OXY]ACETAMIDE",
            "non-polymer",
        ),
        "IKR": (
            "METHYL (2E)-{2-[(4-IODO-2,5-DIMETHYLPHENOXY)METHYL]PHENYL}(METHOXYIMINO)ETHANOATE",
            "non-polymer",
        ),
        "4R8": (
            "(6E,10Z,13Z,16Z,19Z)-4-OXODOCOSA-6,10,13,16,19-PENTAENOIC ACID",
            "non-polymer",
        ),
        "V20": (
            "N6^-[(1R)-2-[(1S)-1-CARBOXY-2-(METHYLSULFANYL)ETHOXY]-2-OXO-1-(SULFANYLMETHYL)ETHYL]-6-OXO-L-LYSINE",
            "non-polymer",
        ),
        "HOH": ("WATER", "non-polymer"),
        "SFG": ("SINEFUNGIN", "non-polymer"),
        "SFD": ("N5-SULFONO FLAVIN-ADENINE DINUCLEOTIDE", "non-polymer"),
        "1UN": (
            "2-[2-HYDROXY-3-(3-HYDROXY-2-METHYL-BENZOYLAMINO)-4-PHENYL SULFANYL-BUTYL]-DECAHYDRO-ISOQUINOLINE-3-CARBOXYLIC ACID TERT-BUTYLAMIDE",
            "non-polymer",
        ),
        "SFE": ("(3S)-3-AMINO-3-PHENYLPROPANOIC ACID", "peptide"),
        "X7O": ("5-AMINO-3-METHYL-PYRROLIDINE-2-CARBOXYLIC ACID", "non-polymer"),
        "SFB": (
            "(2S,3S,4R,5R)-5-(6-AMINOPURIN-9-YL)-N-[[(2S,3S,4R,5R)-5-(2,4-DIOXOPYRIMIDIN-1-YL)-4-HYDROXY-2-(HYDROXYMETHYL)OXOLAN-3-YL]METHYLSULFONYL]-3,4-DIHYDROXY-OXOLANE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "LIR": (
            "2-[(3S)-3-AMINOPIPERIDIN-1-YL]-1-(2-CYANOBENZYL)-5-METHYL-4,6-DIOXO-3,4,5,6-TETRAHYDROPYRROLO[3,4-D]IMIDAZOL-1-IUM",
            "non-polymer",
        ),
        "46Z": (
            "(2R,3R,4S,5S)-4-FLUORO-3,5-DIHYDROXYTETRAHYDROFURAN-2-YL 2-PHENYLETHYL HYDROGEN (S)-PHOSPHATE",
            "saccharide",
        ),
        "SFC": ("(S)-IBUPROFENOYL-COENZYME A", "non-polymer"),
        "SPC": (
            "N-HYDROXY 1N(4-METHOXYPHENYL)SULFONYL-4-(Z,E-N-METHOXYIMINO)PYRROLIDINE-2R-CARBOXAMIDE",
            "non-polymer",
        ),
        "YJ2": (
            "(2S,4R)-1-[1-(4-BROMOPHENYL)CYCLOPROPYL]CARBONYL-4-(2-CHLOROPHENYL)SULFONYL-N-[1-(IMINOMETHYL)CYCLOPROPYL]PYRROLIDINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "YJ8": (
            "(2S,4R)-4-(2-CHLOROPHENYL)SULFONYL-N-[1-(IMINOMETHYL)CYCLOPROPYL]-1-[1-(4-IODOPHENYL)CYCLOPROPYL]CARBONYL-PYRROLIDINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "YJ9": (
            "(2S,4R)-4-(2-CHLOROPHENYL)SULFONYL-N-[1-(IMINOMETHYL)CYCLOPROPYL]-1-[1-[4-(TRIFLUOROMETHYL)PHENYL]CYCLOPROPYL]CARBONYL-PYRROLIDINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "SFA": ("SANGLIFEHRIN A", "non-polymer"),
        "IKX": (
            "4-(4-(2-HYDROXYETHYL)-1H-1,2,3-TRIAZOL-1-YL)PYRIDINE-2,6-DICARBOXYLIC ACID",
            "non-polymer",
        ),
        "TFX": (
            "2-[4-(DIMETHYLAMINO)PHENYL]-3,6-DIMETHYL-1,3-BENZOTHIAZOL-3-IUM",
            "non-polymer",
        ),
        "SPD": ("SPERMIDINE", "non-polymer"),
        "CM1": ("CARBOXYMYCOBACTIN S", "non-polymer"),
        "G69": (
            "N-[(2S)-5-{[4-(HYDROXYMETHYL)-3-(TRIFLUOROMETHYL)-1H-PYRAZOL-1-YL]METHYL}-2,3-DIHYDRO-1H-INDEN-2-YL]PROPANE-2-SULFONAMIDE",
            "non-polymer",
        ),
        "OLN": ("(S)-2-ACETAMIDO-5-UREIDOPENTANOIC ACID", "non-polymer"),
        "OLO": ("OLOMOUCINE", "non-polymer"),
        "2DT": ("3'-DEOXYTHYMIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "2DP": (
            "3-[[3-METHYLPHOSPHONO-GLYCEROLYL]PHOSPHONYL]-[1,2-DI[2,6,10,14-TETRAMETHYL-HEXADECAN-16-YL]GLYCEROL",
            "non-polymer",
        ),
        "2DS": (
            "N-[(2,3-DIHYDROXYPHENYL)CARBONYL]-O-[(2R)-2-{[(2,3-DIHYDROXYPHENYL)CARBONYL]AMINO}-3-HYDROXYPROPANOYL]-L-SERINE",
            "non-polymer",
        ),
        "2DR": ("2-DEOXY-BETA-D-ERYTHRO-PENTOFURANOSE", "non-polymer"),
        "HOZ": ("(4S)-4,5-DIAMINOPENTANOIC ACID", "non-polymer"),
        "2DE": (
            "(2E,6E)-7,11-DIMETHYLDODECA-2,6,10-TRIEN-1-YL TRIHYDROGEN DIPHOSPHATE",
            "non-polymer",
        ),
        "2DD": (
            "3-(5-CHLORO-2,4-DIHYDROXY-PHENYL)-1H-PYRAZOLE-4-CARBOXYLIC ACID 4-SULFAMOYL-BENZYLAMIDE",
            "non-polymer",
        ),
        "2DG": ("2-DEOXY-BETA-D-GALACTOSE", "saccharide"),
        "2DF": (
            "N-(2-DEOXY-BETA-D-ERYTHO-PENTOFURANOSYL-5-PHOSPHATE) FORMAMIDE",
            "nucleic-acid",
        ),
        "2DA": ("2',3'-DIDEOXYADENOSINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "2DC": ("(3ALPHA,8ALPHA,22R)-CHOLEST-5-ENE-3,20,22-TRIOL", "non-polymer"),
        "2DB": ("(S)-(2,3-DIHYDROXYPROPOXY)TRIHYDROXYBORATE", "non-polymer"),
        "2DM": (
            "2-HYDROXY-3-(PYREN-1-YLMETHOXY)PROPYL DIHYDROGEN PHOSPHATE",
            "nucleic-acid",
        ),
        "2DL": ("2-METHYL-4-DIETHYLAMIDE-PHENOL", "non-polymer"),
        "2DO": ("(2S)-2-AMINOHEXANE-1,1-DIOL", "peptide"),
        "2DI": (
            "9-[(2R,5R)-5-(HYDROXYMETHYL)TETRAHYDROFURAN-2-YL]-1,9-DIHYDRO-6H-PURIN-6-ONE",
            "non-polymer",
        ),
        "5DR": (
            "6-ETHYL-5-[3-(4-METHOXYBIPHENYL-3-YL)PROP-1-YN-1-YL]PYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "2D4": (
            "N-[4-HYDROXY-3-(2-HYDROXYNAPHTHALEN-1-YL)PHENYL]THIOPHENE-2-SULFONAMIDE",
            "non-polymer",
        ),
        "2D7": (
            "N-(4-ACETYLPHENYL)-5-(5-CHLORO-2,4-DIHYDROXYPHENYL)-1H-PYRAZOLE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "2D3": (
            "METHYL 3-ISOXAZOL-5-YL-5-METHYL-1H-PYRAZOLE-4-CARBOXYLATE",
            "non-polymer",
        ),
        "TFZ": ("N-PHENYL-2-(TRIFLUOROMETHYL)BENZAMIDE", "non-polymer"),
        "2D9": (
            "4-CHLORO-6-(4-{4-[4-(METHYLSULFONYL)BENZYL]PIPERAZIN-1-YL}-1H-PYRAZOL-5-YL)BENZENE-1,3-DIOL",
            "non-polymer",
        ),
        "5DE": (
            "1-(4-AMINOPHENYL)-3,5-DIMETHYL-1H-PYRAZOLE-4-CARBOXYLIC ACID ETHYL ESTER",
            "non-polymer",
        ),
        "TGF": (
            "2-[4-(4-{4-[(2-AMINO-5-FORMYL-4-OXO-3,4,5,6,7,8-HEXAHYDRO-PTERIDIN-6-YLMETHYL)-AMINO]-BENZOYLAMINO}-4-CARBOXY-BUTYRYLAMINO)-4-CARBOXY-BUTYRYLAMINO]-PENTANEDIOIC ACID",
            "non-polymer",
        ),
        "5DI": ("2-OXO-1,2,DIDEOXY-5F-D-IDOPYRANOSE", "saccharide"),
        "8K6": ("OCTADECANE", "non-polymer"),
        "G6H": (
            "N-{4-METHYL-3-[(3-{4-[(3,4,5-TRIMETHOXYPHENYL)AMINO]-1,3,5-TRIAZIN-2-YL}PYRIDIN-2-YL)AMINO]PHENYL}-3-(TRIFLUOROMETHYL)BENZAMIDE",
            "non-polymer",
        ),
        "G6I": (
            "N-{3-[(3-{4-[(4-METHOXYPHENYL)AMINO]-1,3,5-TRIAZIN-2-YL}PYRIDIN-2-YL)AMINO]-4-METHYLPHENYL}-3-(TRIFLUOROMETHYL)BENZAMIDE",
            "non-polymer",
        ),
        "46A": (
            "N-BUTYL-4,6-DIMETHYL-N-{[2'-(2H-TETRAZOL-5-YL)BIPHENYL-4-YL]METHYL}PYRIMIDIN-2-AMINE",
            "non-polymer",
        ),
        "G6A": (
            "N-CYCLOPROPYL-2',6-DIMETHYL-4'-(5-METHYL-1,3,4-OXADIAZOL-2-YL)BIPHENYL-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "G6G": (
            "N-{3-[(3-{4-[(4-METHOXYPHENYL)AMINO]-1,3,5-TRIAZIN-2-YL}PYRIDIN-2-YL)AMINO]-4-METHYLPHENYL}-4-[(4-METHYLPIPERAZIN-1-YL)METHYL]BENZAMIDE",
            "non-polymer",
        ),
        "G6D": ("6-DEOXY-ALPHA-D-GLUCOSE", "saccharide"),
        "G6X": (
            "2-{6-[(4AS,6R,8AS)-6-HYDROXY-3-METHOXY-5,6,9,10-TETRAHYDRO-4AH-[1]BENZOFURO[3A,3,2-EF][2]BENZAZEPIN-11(12H)-YL]HEXYL}-1,2-BENZISOTHIAZOL-3(2H)-ONE 1,1-DIOXIDE",
            "non-polymer",
        ),
        "SA8": ("S-5'-AZAMETHIONINE-5'-DEOXYADENOSINE", "non-polymer"),
        "IKM": ("IKM-159", "non-polymer"),
        "G6S": ("D-GALACTOSE-6-SULFATE", "saccharide"),
        "G6P": ("ALPHA-D-GLUCOSE-6-PHOSPHATE", "saccharide"),
        "46B": (
            "{4-[6-BROMO-3-(BUTYLAMINO)IMIDAZO[1,2-A]PYRIDIN-2-YL]PHENOXY}ACETIC ACID",
            "non-polymer",
        ),
        "605": (
            "4'-[(1R)-1-AMINO-2-(2,5-DIFLUOROPHENYL)ETHYL]BIPHENYL-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "607": (
            "2-(2'-AMINO-5-{5-[AMINO(IMINIO)METHYL]-1H-BENZIMIDAZOL-2-YL}-6-OXIDO-1,1'-BIPHENYL-3-YL)SUCCINATE",
            "non-polymer",
        ),
        "606": (
            "(2S)-2-(3-{[AMINO(IMINO)METHYL]AMINO}PHENYL)-3-[(S)-HYDROXY(3-PHENYLPROPYL)PHOSPHORYL]PROPANOIC ACID",
            "non-polymer",
        ),
        "600": (
            "6-[4-((2S)-2-(1H-1,2,3-BENZOTRIAZOL-1-YL)-3-{4-[DIFLUORO(PHOSPHONO)METHYL]PHENYL}-2-PHENYLPROPYL)PHENYL]-2-[(1S)-1-METHOXY-3-METHYLBUTYL]QUINOLIN-8-YLPHOSPHONIC ACID",
            "non-polymer",
        ),
        "TH0": ("4-MERCAPTOBENZENE-1,3-DIOL", "non-polymer"),
        "609": (
            "7-OXO-7H-BENZIMIDAZO[2,1-A]BENZ[DE]ISOQUINOLINE-3-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "608": (
            "N-(4-PHENOXYPHENYL)-2-[(PYRIDIN-4-YLMETHYL)AMINO]NICOTINAMIDE",
            "non-polymer",
        ),
        "O75": (
            "N-BENZYL-3-(2-CYANO-6-PROPYLPYRIMIDIN-4-YL)-N-[2-(DIMETHYLAMINO)ETHYL]-5-(TRIFLUOROMETHYL)BENZAMIDE",
            "non-polymer",
        ),
        "PR4": ("1-[(2R)-2-CARBOXY-2-HYDROXYETHYL]-L-PROLINE", "peptide"),
        "468": (
            "(3S)-N-(3-CHLORO-2-METHYLPHENYL)-1-CYCLOHEXYL-5-OXOPYRROLIDINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "O2N": (
            "5-{[(2-{[BIS(4-METHOXYPHENYL)METHYL]CARBAMOYL}BENZYL)(PROP-2-EN-1-YL)AMINO]METHYL}-1,3-BENZODIOXOLE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "3F5": ("5'-O-(2-AMINO-2-OXOETHYL)-8-(METHYLAMINO)ADENOSINE", "non-polymer"),
        "3F1": (
            "N-{[4-(BENZYLOXY)PHENYL](METHYL)-LAMBDA~4~-SULFANYLIDENE}-4-METHYLBENZENESULFONAMIDE",
            "non-polymer",
        ),
        "465": (
            "8-AMINO-7-CHLORO-1-METHYL-6-(METHYLIDENEAMINO)-2-OXO-1,2-DIHYDROPYRROLO[4,3,2-DE]QUINOLINE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "MOK": ("5-METHYL-3-PHENYL-1,2-OXAZOLE-4-CARBOXYLIC ACID", "non-polymer"),
        "6EA": (
            "(1S)-1-(1H-INDOL-3-YLMETHYL)-2-(2-PYRIDIN-4-YL-[1,7]NAPHTYRIDIN-5-YLOXY)-EHYLAMINE",
            "non-polymer",
        ),
        "RFC": ("(R)-IBUPROFENOYL-COENZYME A", "non-polymer"),
        "IK8": (
            "1-(3-AMINO-1,2-BENZISOXAZOL-5-YL)-N-(4-{2-[(DIMETHYLAMINO)METHYL]-1H-IMIDAZOL-1-YL}-2-FLUOROPHENYL)-3-(TRIFLUOROMETHYL)-1H-PYRAZOLE-5-CARBOXAMIDE",
            "non-polymer",
        ),
        "3FZ": ("3-FORMYLBENZENECARBOXIMIDAMIDE", "non-polymer"),
        "3FX": (
            "(2R)-3-(CYCLOHEXYLAMINO)-2-HYDROXYPROPANE-1-SULFONIC ACID",
            "non-polymer",
        ),
        "3FY": (
            "N-(2-{(4-BROMOPHENYL)[(1-METHYL-1H-IMIDAZOL-5-YL)METHYL]AMINO}ETHYL)-1-METHYL-N-(2-METHYLBENZYL)-1H-IMIDAZOLE-4-SULFONAMIDE",
            "non-polymer",
        ),
        "461": (
            "(E)-2-(5-CHLOROTHIOPHEN-2-YL)-N-[(3S)-1-{4-[(1R)-1-(DIMETHYLAMINO)ETHYL]-2-FLUOROPHENYL}-2-OXOPYRROLIDIN-3-YL]ETHENESULFONAMIDE",
            "non-polymer",
        ),
        "3FT": ("?", "non-polymer"),
        "3FR": (
            "3'-CHLORO-5'-(3-METHYL-6-{[(1S)-1-PHENYLETHYL]AMINO}-1H-PYRAZOLO[4,3-C]PYRIDIN-1-YL)BIPHENYL-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "3FP": (
            "(2R)-1-(DIMETHYLAMINO)-3-{4-[(6-{[2-FLUORO-5-(TRIFLUOROMETHYL)PHENYL]AMINO}PYRIMIDIN-4-YL)AMINO]PHENOXY}PROPAN-2-OL",
            "non-polymer",
        ),
        "3FQ": (
            "(2S)-1-(DIMETHYLAMINO)-3-{4-[(6-{[2-FLUORO-5-(TRIFLUOROMETHYL)PHENYL]AMINO}PYRIMIDIN-4-YL)AMINO]PHENOXY}PROPAN-2-OL",
            "non-polymer",
        ),
        "3FN": (
            "6-(2,4-DIFLUOROPHENOXY)-8-METHYL-2-(TETRAHYDRO-2H-PYRAN-4-YLAMINO)PYRIDO[2,3-D]PYRIMIDIN-7(8H)-ONE",
            "non-polymer",
        ),
        "3FL": (
            "3-{[(4-{[(2,6-DIOXO-1,2,3,6-TETRAHYDROPYRIMIDIN-4-YL)METHYL]AMINO}BUTYL)AMINO]METHYL}BENZOIC ACID",
            "non-polymer",
        ),
        "3FM": ("3-O-FORMAMIDO-ALPHA-D-MANNOPYRANOSIDE", "saccharide"),
        "3FI": (
            "3-{(E)-[(3-{[(2,6-DIOXO-1,2,3,6-TETRAHYDROPYRIMIDIN-4-YL)METHYL]AMINO}PROPOXY)IMINO]METHYL}BENZOIC ACID",
            "non-polymer",
        ),
        "3FF": (
            "2-[(2,4-DIFLUOROPHENYL)AMINO]-7-{[(2R)-2,3-DIHYDROXYPROPYL]OXY}-10,11-DIHYDRO-5H-DIBENZO[A,D][7]ANNULEN-5-ONE",
            "non-polymer",
        ),
        "3FG": ("(2S)-AMINO(3,5-DIHYDROXYPHENYL)ETHANOIC ACID", "peptide"),
        "3FD": (
            "4-[[(2R,3S,4R,5R)-5-[6-AMINO-8-[(3,4-DICHLOROPHENYL)METHYLAMINO]PURIN-9-YL]-3,4-DIHYDROXY-OXOLAN-2-YL]METHOXYMETHYL]BENZONITRILE",
            "non-polymer",
        ),
        "462": (
            "N-[AMINO(IMINO)METHYL]-2-[2-(2-CHLOROPHENYL)-4-(4-PROPOXYPHENYL)-3-THIENYL]ACETAMIDE",
            "non-polymer",
        ),
        "3FB": ("(3S)-3-AMINO-4-PHENYLBUTANOIC ACID", "non-polymer"),
        "3FC": (
            "(1R)-2,2-DIMETHYL-1-({5-[4-(TRIFLUOROMETHYL)PHENYL]-1,3,4-OXADIAZOL-2-YL}METHYL)PROPYL (1S)-1-{OXO[(2-OXO-1,3-OXAZOLIDIN-3-YL)AMINO]ACETYL}PENTYLCARBAMATE",
            "non-polymer",
        ),
        "3FA": ("3-FLUOROBENZENE-1,2-DIOL", "non-polymer"),
        "ZCL": ("3,4-DICHLORO-L-PHENYLALANINE", "peptide"),
        "U66": (
            "(20S)-19,20,22,23-TETRAHYDRO-19-OXO-5H,21H-18,20-ETHANO-12,14-ETHENO-6,10-METHENOBENZ[D]IMIDAZO[4,3-L][1,6,9,13]OXATRIAZACYCLONOADECOSINE-9-CARBONITRILE",
            "non-polymer",
        ),
        "PR1": (
            "4-HYDROXY-3-[2-OXO-3-(THIENO[3,2-B]PYRIDINE-2-SULFONYLAMINO)-PYRROLIDIN-1-YLMETHYL]-BENZAMIDINE",
            "non-polymer",
        ),
        "RFA": ("?", "non-polymer"),
        "PMV": (
            "(3R)-3-HYDROXY-3-METHYL-5-(PHOSPHONOOXY)PENTANOIC ACID",
            "non-polymer",
        ),
        "ZCY": ("(S)-1'-(2',3'-DIHYDROXYPROPYL)-CYTOSINE", "nucleic-acid"),
        "FZN": ("?", "peptide"),
        "XY2": (
            "N,N'-DIMETHYL-N-(ACETYL)-N'-(7-NITROBENZ-2-OXA-1,3-DIAZOL-4-YL)ETHYLENEDIAMINE",
            "non-polymer",
        ),
        "JT6": ("3-[(4'-CYANOBIPHENYL-4-YL)OXY]-N-HYDROXYPROPANAMIDE", "non-polymer"),
        "JWS": ("2-(4,6-DIAMINO-1,3,5-TRIAZIN-2-YL)-4-METHOXYPHENOL", "non-polymer"),
        "JWT": (
            "N-(6-AMINO-1-BENZYL-2,4-DIOXO-1,2,3,4-TETRAHYDROPYRIMIDIN-5-YL)-4-FLUORO-N-METHYLBENZENESULFONAMIDE",
            "non-polymer",
        ),
        "PR0": ("3(S)-AMINO-4-PHENYL-BUTAN-2(R)-OL", "peptide"),
        "2O7": (
            "7A-[(4-CYANOPHENYL)METHYL]-6-(3,5-DICHLOROPHENYL)-5-OXO-2,3,5,7A-TETRAHYDRO-1H-PYRROLO[1,2-A]PYRROLE-7-CARBONITRILE",
            "non-polymer",
        ),
        "TFS": ("TRIFLUOROMETHANESULFONIC ACID", "non-polymer"),
        "PPO": ("3H-PYRAZOLO[4,3-D]PYRIMIDIN-7-OL", "non-polymer"),
        "S4G": ("?", "nucleic-acid"),
        "S4A": ("?", "nucleic-acid"),
        "S4C": ("4'-THIO-4'-DEOXY-CYTOSINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "S4M": (
            "5'-[(S)-(3-AMINOPROPYL)(METHYL)-LAMBDA~4~-SULFANYL]-5'-DEOXYADENOSINE",
            "non-polymer",
        ),
        "S4H": ("TRITHIOSULFUROUS ACID", "non-polymer"),
        "0F1": ("3,6-DICHLORO-1-BENZOTHIOPHENE-2-CARBOXYLIC ACID", "non-polymer"),
        "0F0": (
            "3-{[4-(DIMETHYLAMINO)BUTANOYL]AMINO}-N-(4-{[4-(PYRIDIN-3-YL)PYRIMIDIN-2-YL]AMINO}PHENYL)BENZAMIDE",
            "non-polymer",
        ),
        "0F3": ("4-(SULFAMOYLAMINO)BENZENESULFONAMIDE", "non-polymer"),
        "0F2": (
            "3-{4-[2-(MORPHOLIN-4-YL)ETHOXY]-2-(4LAMBDA~4~-THIENO[3,2-C]PYRAZOL-3-YL)-1H-INDOL-6-YL}PENTAN-3-OL",
            "non-polymer",
        ),
        "0F5": (
            "5-(3-{6-[(TRANS-4-AMINOCYCLOHEXYL)AMINO]PYRAZIN-2-YL}BENZYL)-2H-1LAMBDA~4~,3-THIAZOLE-2,4(3H)-DIONE",
            "non-polymer",
        ),
        "0F4": (
            "1-(3-{[(3Z)-2-OXO-3-(1H-PYRROL-2-YLMETHYLIDENE)-2,3-DIHYDRO-1H-INDOL-6-YL]AMINO}PHENYL)-3-[3-(TRIFLUOROMETHYL)PHENYL]UREA",
            "non-polymer",
        ),
        "0F7": (
            "N-[(BENZYLOXY)CARBONYL]-L-ISOLEUCYL-N-[(1R)-1-(CARBOXYCARBONYL)-3,3-DIFLUOROPROPYL]-L-LEUCINAMIDE",
            "peptide",
        ),
        "TFR": (
            "(2R)-2-AMINO-7-OXO-7-{[(2R,3S)-4,4,4-TRIFLUORO-3-HYDROXYBUTAN-2-YL]AMINO}HEPTANOIC ACID",
            "peptide",
        ),
        "0F9": ("4-CHLORO-2-(1H-PYRAZOL-3-YL)PHENOL", "non-polymer"),
        "0FA": ("N~2~-PHOSPHONO-L-LEUCINAMIDE", "peptide"),
        "0FG": ("D-LEUCYL-N-(4-FLUOROBENZYL)-L-PHENYLALANINAMIDE", "peptide"),
        "0FI": ("N-PHENYLTHIOFORMAMIDE", "non-polymer"),
        "0FK": ("8-HYDROXYQUINOLINE-2-CARBOXAMIDE", "non-polymer"),
        "0FL": ("3-AMINO-N-FORMYL-L-ALANINE", "peptide"),
        "0FO": ("6-BROMO-4-HYDROXY-2H-CHROMEN-2-ONE", "non-polymer"),
        "0FN": ("1,3-DIOXO-2,3-DIHYDRO-1H-INDENE-2-CARBONITRILE", "non-polymer"),
        "S44": (
            "(2S)-2-(4-PROPOXY-3-{[({4-[(3S,5S,7S)-TRICYCLO[3.3.1.1~3,7~]DEC-1-YL]PHENYL}CARBONYL)AMINO]METHYL}BENZYL)BUTANOIC ACID",
            "non-polymer",
        ),
        "0FP": (
            "N-(6-AMINOHEXANOYL)-3-METHYL-L-VALYL-3-METHYL-L-VALYL-N~1~-[(2S,3S)-3-HYDROXY-4-OXO-4-{[(1R)-1-PHENYLPROPYL]AMINO}BUTAN-2-YL]-N~4~,N~4~-DIMETHYL-L-ASPARTAMIDE",
            "peptide",
        ),
        "0FS": (
            "(5Z)-5-[3-(TRIFLUOROMETHYL)BENZYLIDENE]-1,3-THIAZOLIDINE-2,4-DIONE",
            "non-polymer",
        ),
        "0FR": ("FURAN-2-YL(1H-INDOL-3-YL)METHANONE", "non-polymer"),
        "0FU": ("2,5,6-TRIAMINOPYRIMIDIN-4-OL", "non-polymer"),
        "0FT": (
            "1-{[1-(ALPHA-L-ARABINOFURANOSYL)-1H-1,2,3-TRIAZOL-4-YL]METHYL}-5-METHYL-2,4-DIOXO-1,2,3,4-TETRAHYDROPYRIMIDINE",
            "non-polymer",
        ),
        "0FW": (
            "4-{3-[(BIPHENYL-4-YLCARBONYL)AMINO]PHENOXY}BENZENE-1,2-DICARBOXYLIC ACID",
            "non-polymer",
        ),
        "0FV": (
            "(2Z)-2-FLUORO-3,7-DIMETHYLOCTA-2,6-DIEN-1-YL TRIHYDROGEN DIPHOSPHATE",
            "non-polymer",
        ),
        "0FY": (
            "3-(1-{2-[(3-FLUOROPYRIDINIUM-4-YL)AMINO]-2-OXOETHYL}-1H-PYRAZOL-4-YL)-6-METHYL-8-[(3-{[(1R,3R)-3-METHYLPIPERIDINIUM-1-YL]METHYL}-1,2-THIAZOL-5-YL)AMINO]IMIDAZO[1,2-A]PYRAZIN-1-IUM",
            "non-polymer",
        ),
        "0FX": ("DTDP-4-AMINO-4,6-DIDEOXYGLUCOSE", "non-polymer"),
        "0FZ": (
            "4-(6-METHOXY-3,4-DIHYDROISOQUINOLIN-1-YL)BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "S48": (
            "METHYL N-{(3S)-1-[(1-METHYL-1H-IMIDAZOL-5-YL)METHYL]-6-PHENYL-1,2,3,4-TETRAHYDROQUINOLIN-3-YL}-N-[(1-METHYL-1H-IMIDAZOL-4-YL)SULFONYL]GLYCINATE",
            "non-polymer",
        ),
        "XYP": ("BETA-D-XYLOPYRANOSE", "saccharide"),
        "XYQ": ("ACETYL-RIBOSYL-ADP", "non-polymer"),
        "XYS": ("XYLOPYRANOSE", "saccharide"),
        "XYZ": ("BETA-D-XYLOFURANOSE", "non-polymer"),
        "UM2": ("(2S)-2-AMINO-N-BUTYL-PROPANAMIDE", "peptide"),
        "JW5": ("?", "nucleic-acid"),
        "XYA": (
            "2-(6-AMINO-OCTAHYDRO-PURIN-9-YL)-5-HYDROXYMETHYL-TETRAHYDRO-FURAN-3,4-DIOL",
            "non-polymer",
        ),
        "PMR": ("PROTOCHLOROPHYLLIDE", "non-polymer"),
        "XYD": ("2,5-DIMETHYLANILINE", "non-polymer"),
        "PPL": ("PIPERIDINE-2-CARBOXYLIC ACID TERT-BUTYLAMIDE", "non-polymer"),
        "XYF": (
            "5(R)-5-FLUORO-BETA-D-XYLOPYRANOSYL-ENZYME INTERMEDIATE",
            "non-polymer",
        ),
        "XYG": (
            "[(4Z)-2-[(1Z)-ETHANIMIDOYL]-4-(4-HYDROXYBENZYLIDENE)-5-OXO-4,5-DIHYDRO-1H-IMIDAZOL-1-YL]ACETIC ACID",
            "peptide",
        ),
        "XYH": ("XYLAROHYDROXAMATE", "non-polymer"),
        "XYL": ("D-XYLITOL", "non-polymer"),
        "XYM": (
            "(2Z)-3-[(CARBOXYMETHYL)SULFANYL]-2-HYDROXYPROP-2-ENOIC ACID",
            "non-polymer",
        ),
        "XYN": ("3-[(CARBOXYMETHYL)SULFANYL]-2-OXOPROPANOIC ACID", "non-polymer"),
        "HGT": (
            "(2-((2R,3S,4R,5S,6S)-3-AMINO-4,5-DIHYDROXY-6-(HYDROXYMETHYL)TETRAHYDRO-2H-PYRAN-2-YLAMINO)ETHOXY)STANNANETRIOL",
            "non-polymer",
        ),
        "GUB": (
            "9-ALPHA-L-LYXOFURANOSYL-N-(2-PHENYLETHYL)-9H-PURIN-6-AMINE",
            "non-polymer",
        ),
        "GUA": ("GLUTARIC ACID", "non-polymer"),
        "TSM": ("2-METHYLPROPANE-2-SULFONIC ACID", "non-polymer"),
        "GUF": ("2-DEOXY-2-FLUORO-BETA-D-GLUCOPYRANURONIC ACID", "saccharide"),
        "TSH": ("2-(1H-INDOL-3-YL)ETHANIMINE", "non-polymer"),
        "GUD": ("GLUCOSE-URIDINE-C1,5'-DIPHOSPHATE", "non-polymer"),
        "TSF": (
            "3-(1,3,7-TRIHYDRO-9-D-RIBITYL-2,6,8-PURINETRIONE-7-YL) 1,1 DIFLUOROPENTANE-1-PHOSPHATE",
            "non-polymer",
        ),
        "GUI": (
            "5-CHLORO-N-[2-METHOXY-4-[4-(4-METHYLPIPERAZIN-1-YL)PIPERIDIN-1-YL]PHENYL]-N'-(2-PROPAN-2-YLSULFONYLPHENYL)PYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "TSE": ("THIOPHOSPHONOACETIC ACID", "non-polymer"),
        "TSB": ("5'-O-(N-(L-THREONYL)-SULFAMOYL)ADENOSINE", "non-polymer"),
        "GUN": ("GUANINE", "non-polymer"),
        "GUM": ("4-METHYL-UMBELLIFERYL-N-ACETYL-CHITOBIOSE", "non-polymer"),
        "GUL": ("5-FLUORO-BETA-L-GULOSYL FLUORIDE", "non-polymer"),
        "GUS": (
            "(E)-N-[(5-CARBAMIMIDAMIDO-2-HYDROXYPHENYL)METHYLIDENE]-L-ALANINE",
            "non-polymer",
        ),
        "GUR": ("GLYCOLURIL", "non-polymer"),
        "GUP": ("ALPHA-L-GULOPYRANOSIDE", "saccharide"),
        "TSZ": ("HYDRAZINECARBOTHIOAMIDE", "non-polymer"),
        "TSX": (
            "N-[4-(1-ALLYL-3-BUTYL-2,6-DIOXO-2,3,6,7-TETRAHYDRO-1H-PURIN-8-YLMETHYL)-PHENYL]-ACETAMIDE",
            "non-polymer",
        ),
        "TSV": ("ITMN-191; DANOPREVIR", "non-polymer"),
        "GUZ": ("2-DEOXY-2-FLUORO-ALPHA-D-GLUCOPYRANURONIC ACID", "saccharide"),
        "TST": (
            "4-METHYL-2-{[4-(TOLUENE-4-SULFONYL)-THIOMORPHOLINE-3-CARBONYL]-AMINO}-PENTANOIC ACID",
            "peptide",
        ),
        "TSU": ("PARA-TOLUENE SULFONATE", "non-polymer"),
        "TSR": ("2-(1H-INDOL-3-YL)ACETAMIDE", "non-polymer"),
        "TSS": ("2-(1H-INDOL-3-YL)ETHANAMINE", "non-polymer"),
        "TSP": ("3'-THIO-THYMIDINE-5'-PHOSPHATE", "non-polymer"),
        "HUD": (
            "TERT-BUTYL {(2S)-1-[(1R,2S,5S)-2-{[(2S,3R)-4-AMINO-1-CYCLOPROPYL-3-HYDROXY-4-OXOBUTAN-2-YL]CARBAMOYL}-6,6-DIMETHYL-3-AZABICYCLO[3.1.0]HEX-3-YL]-3,3-DIMETHYL-1-OXOBUTAN-2-YL}CARBAMATE",
            "peptide",
        ),
        "HUB": ("HUPERZINE B", "non-polymer"),
        "HUP": ("HUPERAINE A", "non-polymer"),
        "XJX": (
            "(2,4-DIHYDROXY-5-PROPAN-2-YL-PHENYL)-[5-[(4-METHYLPIPERAZIN-1-YL)METHYL]ISOINDOL-2-YL]METHANONE",
            "non-polymer",
        ),
        "HUX": (
            "3-CHLORO-9-ETHYL-6,7,8,9,10,11-HEXAHYDRO-7,11-METHANOCYCLOOCTA[B]QUINOLIN-12-AMINE",
            "non-polymer",
        ),
        "HU4": (
            "TERT-BUTYL {(1S)-2-[(1R,2S,5R)-2-({[(1S)-3-AMINO-1-(CYCLOBUTYLMETHYL)-2,3-DIOXOPROPYL]AMINO}CARBONYL)-7,7-DIMETHYL-6-OXA-3-AZABICYCLO[3.2.0]HEPT-3-YL]-1-CYCLOHEXYL-2-OXOETHYL}CARBAMATE",
            "non-polymer",
        ),
        "HU5": (
            "(1R,2S,5S)-N-[(1S)-3-AMINO-1-(CYCLOBUTYLMETHYL)-2,3-DIOXOPROPYL]-3-[(2S)-2-{[(TERT-BUTYLAMINO)CARBONYL]AMINO}-3,3-DIMETHYLBUTANOYL]-6,6-DIMETHYL-3-AZABICYCLO[3.1.0]HEXANE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "HU1": (
            "TERT-BUTYL {(1S)-2-[(1R,2S,5S)-2-({[(1S)-3-AMINO-1-(CYCLOBUTYLMETHYL)-2,3-DIOXOPROPYL]AMINO}CARBONYL)-6,6-DIMETHYL-3-AZABICYCLO[3.1.0]HEX-3-YL]-1-CYCLOHEXYL-2-OXOETHYL}CARBAMATE",
            "non-polymer",
        ),
        "HU2": (
            "(2S)-({N-[(3S)-3-({N-[(2S,4E)-2-ISOPROPYL-7-METHYLOCT-4-ENOYL]-L-LEUCYL}AMINO)-2-OXOHEXANOYL]GLYCYL}AMINO)(PHENYL)ACETIC ACID",
            "non-polymer",
        ),
        "PR9": ("D-PROLINAMIDE", "peptide"),
        "O28": (
            "2-({[3-(TRIFLUOROMETHYL)-4,5,6,7-TETRAHYDRO-1H-INDAZOL-1-YL]ACETYL}AMINO)-4,5,6,7-TETRAHYDRO-1-BENZOTHIOPHENE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "MKR": (
            "1-{(3R,3AR)-3-[3-(4-ACETYLPIPERAZIN-1-YL)PROPYL]-8-FLUORO-3-PHENYL-3A,4-DIHYDRO-3H-PYRAZOLO[5,1-C][1,4]BENZOXAZIN-2-YL}ETHANONE",
            "non-polymer",
        ),
        "A51": (
            "(3E)-3-[(PHENYLAMINO)METHYLIDENE]DIHYDROFURAN-2(3H)-ONE",
            "non-polymer",
        ),
        "VWW": (
            "L-GAMMA-GLUTAMYL-S-BENZYL-N-[(S)-CARBOXY(PHENYL)METHYL]-L-CYSTEINAMIDE",
            "peptide",
        ),
        "VWS": (
            "5-{[(4-CHLOROPHENYL)SULFONYL]AMINO}-2-METHYL-1-BENZOFURAN-3-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "GU3": (
            "METHYL 3-O-METHYL-2,6-DI-O-SULFO-ALPHA-D-GLUCOPYRANOSIDE",
            "saccharide",
        ),
        "GU2": ("2,3-DI-O-METHYL-ALPHA-L-IDOPYRANURONIC ACID", "saccharide"),
        "GU1": ("2,3-DI-O-METHYL-BETA-D-GLUCOPYRANURONIC ACID", "saccharide"),
        "GU0": ("2,3,6-TRI-O-SULFONATO-ALPHA-L-GALACTOPYRANOSE", "saccharide"),
        "GU7": (
            "2-AMINO-7-[2-(2-HYDROXY-1-HYDROXYMETHYL-ETHYLAMINO)-ETHYL]-1,7-DIHYDRO-PURIN-6-ONE",
            "non-polymer",
        ),
        "GU6": ("2,3,6-TRI-O-SULFONATO-ALPHA-D-GLUCOPYRANOSE", "saccharide"),
        "GU5": ("2,3-DI-O-METHYL-6-O-SULFONATO-ALPHA-D-GLUCOPYRANOSE", "saccharide"),
        "GU4": ("2,3,4,6-TETRA-O-SULFONATO-ALPHA-D-GLUCOPYRANOSE", "saccharide"),
        "TS6": ("MONOTHIOPHOSPHATE", "non-polymer"),
        "GU9": ("2,3,6-TRI-O-METHYL-ALPHA-D-GLUCOPYRANOSE", "saccharide"),
        "D1K": (
            "3-(BETA-D-GLUCOPYRANOSYL)-6-PROPYLFURO[2,3-D]PYRIMIDIN-2(3H)-ONE",
            "non-polymer",
        ),
        "TS2": (
            "2-AMINO-4-[4-(4-AMINO-4-CARBOXY-BUTYRYLAMINO)-5,8,19,22-TETRAOXO-1,2-DITHIA-6,9,13,18,21-PENTAAZA-CYCLOTETRACOS-23-YLCARBAMOYL]-BUTYRIC ACID",
            "non-polymer",
        ),
        "TS3": ("BIS[[5-GLUTAMYL-CYSTEINYL-GLYCYL]-AMINOPROPYL]AMINE", "non-polymer"),
        "TS0": (
            "4-{2,6,8-TRIOXO-9-[(2S,3R,4R)-2,3,4,5-TETRAHYDROXYPENTYL]-1,2,3,6,8,9-HEXAHYDRO-7H-PURIN-7-YL}BUTYL DIHYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "JTH": (
            "2,7-DIHYDROXY-4-(PROPAN-2-YL)CYCLOHEPTA-2,4,6-TRIEN-1-ONE",
            "non-polymer",
        ),
        "EKO": (
            "4-[(2R)-2-[(4-CHLOROBENZYL)OXY]-2-(2,4-DICHLOROPHENYL)ETHYL]-1H-IMIDAZOLE",
            "non-polymer",
        ),
        "REX": (
            "GLYCYL-L-ALPHA-AMINO-EPSILON-PIMELYL-D-ALANYL-D-ALANINE",
            "non-polymer",
        ),
        "REY": ("GLYCYL-L-ALPHA-AMINO-EPSILON-PIMELYL-D-ALANINE", "non-polymer"),
        "RER": (
            "(1R,3S,4S,5S)-3-AMINO-2,3,6-TRIDEOXY-3-METHYL-ALPHA-L-ARABINO-HEXOPYRANOSE",
            "saccharide",
        ),
        "RES": ("4-PHOSPHO-D-ERYTHRONOHYDROXAMIC ACID", "non-polymer"),
        "EKE": ("(4S,5R)-4-AMINO-5-HYDROXYHEXANOIC ACID", "non-polymer"),
        "REQ": (
            "(4,7-DIMETHYL-1,10-PHENANTHROLINE-KAPPA~2~N~1~,N~10~)[TRIS(HYDROXYMETHYL)]RHENIUM(+)",
            "non-polymer",
        ),
        "EKB": (
            "6-ETHYL-5-[3-(3,4,5-TRIMETHOXYPHENYL)PROP-1-YN-1-YL]PYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "RET": ("RETINAL", "non-polymer"),
        "PMA": ("PYROMELLITIC ACID", "non-polymer"),
        "REI": ("TRICARBONYL (AQUA) (IMIDAZOLE) RHENIUM(I)", "non-polymer"),
        "REN": ("(S)-RETICULINE", "non-polymer"),
        "REO": ("PERRHENATE", "non-polymer"),
        "REL": ("D-GLUCURONIC ACID", "saccharide"),
        "REM": (
            "NALPHA-[(2S)-2-BENZYL-3-(TERT-BUTYLSULFONYL)PROPANOYL]-N-[(1S,2R,3S)-1-(CYCLOHEXYLMETHYL)-3-CYCLOPROPYL-2,3-DIHYDROXYPROPYL]-L-HISTIDINAMIDE",
            "non-polymer",
        ),
        "REB": (
            "[4-({4-[(5-CYCLOPROPYL-1H-PYRAZOL-3-YL)AMINO]PYRIMIDIN-2-YL}AMINO)PHENYL]ACETONITRILE",
            "non-polymer",
        ),
        "MFB": ("BETA-L-METHYL-FUCOSE", "saccharide"),
        "REA": ("RETINOIC ACID", "non-polymer"),
        "REF": (
            "2,3,7,8-TETRAHYDROXYCHROMENO[5,4,3-CDE]CHROMENE-5,10-DIONE",
            "non-polymer",
        ),
        "RED": ("DIHYDROLIPOIC ACID", "non-polymer"),
        "REE": ("2-(2-CARBOXYETHYL)-1,1,1-TRIMETHYLDIAZANIUM", "non-polymer"),
        "FCP": (
            "(2R)-1-[4-({6-[(2,6-DIFLUOROPHENYL)AMINO]PYRIMIDIN-4-YL}AMINO)PHENOXY]-3-(DIMETHYLAMINO)PROPAN-2-OL",
            "non-polymer",
        ),
        "FCR": ("ALPHA,ALPHA,ALPHA-TRIFLUORO-P-CRESOL", "non-polymer"),
        "FCT": ("DEOXY-2-FLUORO-B-D-CELLOTRIOSIDE", "saccharide"),
        "A53": (
            "5-{5-[(S)-2-AMINO-3-(1H-INDOL-3-YL)-PROPOXYL]-PYRIDIN-3-YL}-3-[1-(1H-PYRROL-2-YL)-METH-(Z)-YLIDENE]-1,3-DIHYDRO-INDOL-2-ONE",
            "non-polymer",
        ),
        "9PC": (
            "3-[(6-CHLORO[1,3]THIAZOLO[5,4-B]PYRIDIN-2-YL)METHOXY]-2,6-DIFLUOROBENZAMIDE",
            "non-polymer",
        ),
        "FCX": ("ALPHA-FLUORO-CARBOXYMETHYLDETHIA COENZYME A COMPLEX", "non-polymer"),
        "FCY": ("FREE CYSTEINE", "non-polymer"),
        "LFN": ("LUMIFLAVIN", "non-polymer"),
        "ZYD": (
            "(1R,3S,5Z)-5-{(2E)-2-[(3AR,7AS)-7A-(7-HYDROXY-7-METHYLOCTYL)OCTAHYDRO-4H-INDEN-4-YLIDENE]ETHYLIDENE}-4-METHYLIDENECYCLOHEXANE-1,3-DIOL",
            "non-polymer",
        ),
        "FCA": ("ALPHA-D-FUCOSE", "saccharide"),
        "FCB": ("BETA-D-FUCOSE", "saccharide"),
        "FCD": ("5-(2-CHLOROPHENYL)FURAN-2-CARBOXYLIC ACID", "non-polymer"),
        "FCF": (
            "3-[(1E,7E)-8-(2,6-DIOXO-1,2,3,6-TETRAHYDROPYRIMIDIN-4-YL)-3,6-DIOXA-2,7-DIAZAOCTA-1,7-DIEN-1-YL]BENZOIC ACID",
            "non-polymer",
        ),
        "FCG": (
            "(N5,C4A)-(ALPHA-HYDROXY-PROPANO)-3,4,4A,5-TETRAHYDRO-FLAVIN-ADENINE DINUCLEOTIDE",
            "non-polymer",
        ),
        "FCH": (
            "(2R,3S)-3-AMINO-1-[(CYCLOHEXYLMETHYL)AMINO]-4-PHENYLBUTAN-2-OL",
            "non-polymer",
        ),
        "FCI": ("FERRICROCIN-IRON", "non-polymer"),
        "FCK": (
            "3-({[3-({[(1E)-(2,6-DIOXO-1,2,3,6-TETRAHYDROPYRIMIDIN-4-YL)METHYLIDENE]AMINO}OXY)PROPYL]AMINO}METHYL)BENZOIC ACID",
            "non-polymer",
        ),
        "FCL": ("3-CHLORO-L-PHENYLALANINE", "peptide"),
        "FCM": (
            "(9AS)-8-ACETYL-N-[(2-ETHYL-4-FLUORONAPHTHALEN-1-YL)METHYL]-1,7-DIHYDROXY-3-METHOXY-9A-METHYL-9-OXO-9,9A-DIHYDRODIBENZO[B,D]FURAN-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "FCN": ("FOSFOMYCIN", "non-polymer"),
        "FCO": ("CARBONMONOXIDE-(DICYANO) IRON", "non-polymer"),
        "FC0": ("N-CARBOXY-L-PHENYLALANINE", "peptide"),
        "FC1": ("THIOCOUMARIN", "non-polymer"),
        "FC2": ("5-(2,5-DICHLOROPHENYL)-2-FUROIC ACID", "non-polymer"),
        "FC3": ("5-[2-(TRIFLUOROMETHYL)PHENYL]-2-FUROIC ACID", "non-polymer"),
        "FC6": ("HEXACYANOFERRATE(3-)", "non-polymer"),
        "FC7": ("FUSICOCCIN A-THF", "non-polymer"),
        "GE2": ("3,5-DIAMINO-CYCLOHEXANOL", "non-polymer"),
        "LOC": (
            "N-[(7S)-1,2,3,10-TETRAMETHOXY-9-OXO-6,7-DIHYDRO-5H-BENZO[D]HEPTALEN-7-YL]ETHANAMIDE",
            "non-polymer",
        ),
        "A55": (
            "N-[2-(4-AMINO-5,8-DIFLUORO-1,2-DIHYDROQUINAZOLIN-2-YL)ETHYL]-3-FURAMIDE",
            "non-polymer",
        ),
        "RE9": (
            "6-(2,3,4,5,6,7-HEXAHYDRO-2,4,4-TRIMETHYL-1-METYLENEINDEN-2-YL)-3-METHYLHEXA-2,4-DIENOIC ACID",
            "non-polymer",
        ),
        "QUV": (
            "N-[(2S,3S,4R)-1-({6-DEOXY-6-[(NAPHTHALEN-1-YLCARBAMOYL)AMINO]-ALPHA-D-GALACTOPYRANOSYL}OXY)-3,4-DIHYDROXYOCTADECAN-2-YL]HEXACOSANAMIDE",
            "non-polymer",
        ),
        "GE1": ("3,4-DIDEOXY-2,6-AMINO-ALPHA-D GALACTOPYRANOSE", "saccharide"),
        "QUX": (
            "N-[(3S,4S,5R)-4,5-DIHYDROXY-1-[(2R,3R,4R,5R,6R)-3,4,5-TRIHYDROXY-6-(HYDROXYMETHYL)OXAN-2-YL]NONADECAN-3-YL]HEXACOSANAMIDE",
            "non-polymer",
        ),
        "RE0": (
            "3-[(3R)-3-HYDROXY-2-OXO-2,3-DIHYDRO-1H-INDOL-3-YL]-L-ALANINE",
            "peptide",
        ),
        "QUZ": ("QUINAZOLINE", "non-polymer"),
        "9PO": ("PYRIDINE-N-OXIDE", "non-polymer"),
        "QUA": (
            "8-HYDROXY-4-(1-HYDROXYETHYL)QUINOLINE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "QUE": ("3,5,7,3',4'-PENTAHYDROXYFLAVONE", "non-polymer"),
        "QUD": (
            "2-AMINO-3-{(1R)-1-CYCLOHEXYL-2-[(CYCLOHEXYLCARBONYL)AMINO]ETHYL}-6-PHENOXYQUINAZOLIN-3-IUM",
            "non-polymer",
        ),
        "EK9": (
            "N-[(1S)-2-HYDROXY-1-PHENYLETHYL]-4-(5-METHYL-2-PHENYLPYRIMIDIN-4-YL)-1H-PYRROLE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "EK7": (
            "9-(DIMETHYLAMINO)-2-[(3-HYDROXYPHENYL)AMINO]-5,6-DIHYDROTHIENO[3,4-H]QUINAZOLINE-7-CARBONITRILE",
            "non-polymer",
        ),
        "EK6": (
            "ETHYL N-{2-CHLORO-4-[5-(5-{[(1S)-1-(3-CHLORO-4-FLUOROPHENYL)-2-HYDROXYETHYL]CARBAMOYL}-1H-PYRROL-3-YL)-1H-PYRAZOL-4-YL]BENZYL}GLYCINATE",
            "non-polymer",
        ),
        "EK5": (
            "4-[4-(3-CHLOROPHENYL)-1H-PYRAZOL-5-YL]-N-(2,3-DIHYDRO-1-BENZOFURAN-5-YLMETHYL)-1H-PYRROLE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "EK4": (
            "N-[(1S)-2-HYDROXY-1-PHENYLETHYL]-4-{4-[3-(TRIFLUOROMETHYL)PHENYL]-1H-PYRAZOL-5-YL}-1H-PYRROLE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "EK3": (
            "N-CYCLOHEXYL-4-[3-(4-FLUOROPHENYL)-1H-PYRAZOL-4-YL]PYRIDIN-2-AMINE",
            "non-polymer",
        ),
        "EK2": (
            "{4-[4-(3,5-DICHLOROPHENYL)-1H-PYRAZOL-5-YL]-1H-PYRROL-2-YL}(MORPHOLIN-4-YL)METHANONE",
            "non-polymer",
        ),
        "QUO": (
            "2-AMINO-7-DEAZA-(2'',3''-DIHYDROXY-CYCLOPENTYLAMINO)-GUANOSINE-5'-MONOPHOSPHATE",
            "nucleic-acid",
        ),
        "EK0": ("3-(4-CHLOROPHENYL)-4,5,6,7-TETRAHYDRO-1H-INDAZOLE", "non-polymer"),
        "A54": (
            "2-[(1R)-3-AMINO-1-PHENYL-PROPOXY]-4-CHLORO-BENZONITRILE",
            "non-polymer",
        ),
        "CAZ": ("ACYLATED CEFTAZIDIME", "non-polymer"),
        "742": ("(2,2-DIPHOSPHONOETHYL)(DODECYL)DIMETHYLPHOSPHONIUM", "non-polymer"),
        "CAX": (
            "(2S,4S)-4-(2,2-DIHYDROXYETHYL)-2,3,3-TRIMETHYLCYCLOPENTANONE",
            "non-polymer",
        ),
        "744": (
            "(3S)-N-(5-CHLORO-2-METHYLPHENYL)-1-CYCLOHEXYL-5-OXOPYRROLIDINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "745": (
            "TRANS-6-(2-PHENYLCYCLOPROPYL)-NAPHTHALENE-2-CARBOXAMIDINE",
            "non-polymer",
        ),
        "746": (
            "4-TERT-BUTYL-N-[2-METHYL-3-(4-METHYL-6-{[4-(MORPHOLIN-4-YLCARBONYL)PHENYL]AMINO}-5-OXO-4,5-DIHYDROPYRAZIN-2-YL)PHENYL]BENZAMIDE",
            "non-polymer",
        ),
        "IBO": ("2-METHYLPROPANAMIDE", "non-polymer"),
        "CAS": ("S-(DIMETHYLARSENIC)CYSTEINE", "peptide"),
        "749": (
            "[(6E,11E)-2,6,12,16-TETRAMETHYLHEPTADECA-2,6,11,15-TETRAENE-9,9-DIYL]BIS(PHOSPHONIC ACID)",
            "non-polymer",
        ),
        "CAQ": ("CATECHOL", "non-polymer"),
        "CAP": ("2-CARBOXYARABINITOL-1,5-DIPHOSPHATE", "saccharide"),
        "IBD": (
            "2-AMINO-4-{2,4-DICHLORO-5-[2-(DIETHYLAMINO)ETHOXY]PHENYL}-N-ETHYLTHIENO[2,3-D]PYRIMIDINE-6-CARBOXAMIDE",
            "non-polymer",
        ),
        "CAV": (
            "5-AMINO-6-CYCLOHEXYL-3,4-DIHYDROXY-2-ISOPROPYL-HEXANOIC ACID",
            "peptide",
        ),
        "CAU": (
            "(2S)-1-(9H-CARBAZOL-4-YLOXY)-3-(ISOPROPYLAMINO)PROPAN-2-OL",
            "non-polymer",
        ),
        "CAT": ("DODECANE-TRIMETHYLAMINE", "non-polymer"),
        "CAK": (
            "[(1R,2R,3S,4S,5R)-2,3,4,5-TETRAHYDROXYCYCLOHEXYL]METHYL DIHYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "CAJ": ("?", "non-polymer"),
        "CAI": ("CARBOMYCIN A", "non-polymer"),
        "CAH": ("5-EXO-HYDROXYCAMPHOR", "non-polymer"),
        "CAO": ("OXIDIZED COENZYME A", "non-polymer"),
        "CAN": ("CANALINE", "non-polymer"),
        "CAM": ("CAMPHOR", "non-polymer"),
        "CAL": ("5-AMINO-6-CYCLOHEXYL-4-HYDROXY-2-ISOBUTYL-HEXANOIC ACID", "peptide"),
        "CAC": ("CACODYLATE ION", "non-polymer"),
        "CAB": ("4,4-DIHYDROXY-5-OXO-L-NORVALINE", "peptide"),
        "CAA": ("ACETOACETYL-COENZYME A", "non-polymer"),
        "IBS": (
            "L-ALPHA-GLYCEROPHOSPHO-D-MYO-INOSITOL-4,5-BIS-PHOSPHATE",
            "non-polymer",
        ),
        "CAG": ("?", "non-polymer"),
        "CAF": ("S-DIMETHYLARSINOYL-CYSTEINE", "peptide"),
        "CAE": ("CAMPHANE", "non-polymer"),
        "CAD": ("CACODYLIC ACID", "non-polymer"),
        "PGF": ("2,5,8,11-TETRAOXATRIDECANE", "non-polymer"),
        "PGG": ("PARA-NITROPHENYLPHOSPHONOBUTANOYL-GLYCINE", "non-polymer"),
        "PGD": (
            "2-AMINO-5,6-DIMERCAPTO-7-METHYL-3,7,8A,9-TETRAHYDRO-8-OXA-1,3,9,10-TETRAAZA-ANTHRACEN-4-ONE GUANOSINE DINUCLEOTIDE",
            "non-polymer",
        ),
        "PGE": ("TRIETHYLENE GLYCOL", "non-polymer"),
        "DYY": (
            "1-[(2,6-DIFLUOROPHENYL)SULFONYL]-4-(2,3-DIHYDRO-1,4-BENZODIOXIN-6-YLSULFONYL)PIPERAZINE",
            "non-polymer",
        ),
        "PGA": ("2-PHOSPHOGLYCOLIC ACID", "non-polymer"),
        "PGN": ("2'-DEOXYGUANOSINE-3',5'-DIPHOSPHATE", "nucleic-acid"),
        "PGO": ("S-1,2-PROPANEDIOL", "non-polymer"),
        "LOF": ("3-PHENYL-LACTIC ACID", "non-polymer"),
        "PGM": (
            "1-MYRISTOYL-2-HYDROXY-SN-GLYCERO-3-[PHOSPHO-RAC-(1-GLYCEROL)]",
            "non-polymer",
        ),
        "DYP": (
            "1-[1-(4-{[(2,4-DIAMINOPTERIDIN-6-YL)METHYL]AMINO}BENZYL)PIPERIDIN-4-YL]PROPAN-1-ONE",
            "non-polymer",
        ),
        "PGK": (
            "(1R)-2-{[{[(2R)-2,3-DIHYDROXYPROPYL]OXY}(HYDROXY)PHOSPHORYL]OXY}-1-[(PALMITOYLOXY)METHYL]ETHYL (9S,10S)-9,10-DIBROMOOCTADECANOATE",
            "non-polymer",
        ),
        "PGH": ("PHOSPHOGLYCOLOHYDROXAMIC ACID", "non-polymer"),
        "DYS": ("S-[5-(2-AMINOETHYL)-2,3-DIHYDROXYPHENYL]-L-CYSTEINE", "peptide"),
        "PGV": (
            "(1R)-2-{[{[(2S)-2,3-DIHYDROXYPROPYL]OXY}(HYDROXY)PHOSPHORYL]OXY}-1-[(PALMITOYLOXY)METHYL]ETHYL (11E)-OCTADEC-11-ENOATE",
            "non-polymer",
        ),
        "DYM": (
            "6-(2,3-DIHYDROXYPROPOXY)-5-ACETAMIDO-5,6-DIHYDRO-4-HYDROXY-4H-PYRAN-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "DYN": ("?", "non-polymer"),
        "PGU": (
            "N-({3-HYDROXY-2-METHYL-5-[(PHOSPHONOOXY)METHYL]PYRIDIN-4-YL}METHYL)-L-GLUTAMIC ACID",
            "non-polymer",
        ),
        "DYH": (
            "(2R,3AR,6S,7R,7AR)-2-[(2S)-2-AMINO-2-CARBOXYETHYL]-6-HYDROXY-7-(METHYLAMINO)HEXAHYDRO-2H-FURO[3,2-B]PYRAN-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "PGS": ("?", "non-polymer"),
        "PGP": ("GUANOSINE-3',5'-DIPHOSPHATE", "nucleic-acid"),
        "PGQ": ("S-1,2-PROPANEDIOL", "non-polymer"),
        "DYG": (
            "(3S)-3-AMINO-3-[(4Z)-1-(CARBOXYMETHYL)-4-[(4-HYDROXYPHENYL)METHYLIDENE]-5-OXO-IMIDAZOL-2-YL]PROPANOIC ACID",
            "peptide",
        ),
        "A56": ("6-(3-BROMO-2-NAPHTHYL)-1,3,5-TRIAZINE-2,4-DIAMINE", "non-polymer"),
        "DYB": ("4-(1,3,2-DIOXABOROLAN-2-YLOXY)BUTAN-1-AMINIUM", "non-polymer"),
        "PGY": ("PHENYLGLYCINE", "peptide"),
        "DY6": (
            "2-({[3-(3,4-DIHYDROISOQUINOLIN-2(1H)-YLSULFONYL)PHENYL]CARBONYL}AMINO)BENZOIC ACID",
            "non-polymer",
        ),
        "SOC": ("DIOXYSELENOCYSTEINE", "peptide"),
        "XAN": ("XANTHINE", "non-polymer"),
        "PG6": (
            "1-(2-METHOXY-ETHOXY)-2-{2-[2-(2-METHOXY-ETHOXY]-ETHOXY}-ETHANE",
            "non-polymer",
        ),
        "PG7": (
            "((2R,3R,5R)-5-(2-AMINO-6-HYDROXY-9H-PURIN-9-YL)-3-HYDROXY-TETRAHYDROFURAN-2-YL)METHYL DIHYDROGEN PHOSPHATE",
            "nucleic-acid",
        ),
        "PG4": ("TETRAETHYLENE GLYCOL", "non-polymer"),
        "PG5": ("1-METHOXY-2-[2-(2-METHOXY-ETHOXY]-ETHANE", "non-polymer"),
        "PG2": ("PROSTAGLANDIN D2", "non-polymer"),
        "PG3": ("GUANIDINE-3-PROPANOL", "non-polymer"),
        "PG0": ("2-(2-METHOXYETHOXY)ETHANOL", "non-polymer"),
        "PG1": ("PENICILLIN G ACYL-SERINE", "peptide"),
        "SOE": ("ALPHA-L-SORBOPYRANOSE", "saccharide"),
        "PG8": ("1,2-DIOCTANOYL-SN-GLYCERO-3-[PHOSPHO-RAC-(1-GLYCEROL)", "non-polymer"),
        "PG9": ("D-PHENYLGLYCINE", "peptide"),
        "UDR": (
            "(E)-S-METHYL 5-(1-(3,7-DIMETHYL-2-OXO-2,3-DIHYDROBENZO[D]OXAZOL-5-YL)-5-(5-METHYL-1,3,4-OXADIAZOL-2-YL)PENT-1-ENYL)-2-METHOXY-3-METHYLBENZOTHIOATE",
            "non-polymer",
        ),
        "F4P": (
            "1-[BIS(4-FLUOROPHENYL)METHYL]-4-(1H-1,2,4-TRIAZOL-1-YLCARBONYL)PIPERAZINE",
            "non-polymer",
        ),
        "CA9": (
            "(1R,2S,3R)-5-[2-[(1R,3AS,7AR)-1-[(2R)-6-HYDROXY-6-METHYL-HEPTAN-2-YL]-7A-METHYL-1,2,3,3A,6,7-HEXAHYDROINDEN-4-YL]ETHYNYL]-2-METHYL-CYCLOHEX-4-ENE-1,3-DIOL",
            "non-polymer",
        ),
        "CA8": ("?", "non-polymer"),
        "74T": (
            "(3R,3AS,6AR)-HEXAHYDROFURO[2,3-B]FURAN-3-YL {(2S,3R)-4-[({2-[(1-CYCLOPENTYLPIPERIDIN-4-YL)AMINO]-1,3-BENZOTHIAZOL-6-YL}SULFONYL)(2-METHYLPROPYL)AMINO]-3-HYDROXY-1-PHENYLBUTAN-2-YL}CARBAMATE",
            "non-polymer",
        ),
        "OGX": (
            "2'-DEOXY-5'-O-{(S)-HYDROXY[(2-SULFANYLETHYL)AMINO]PHOSPHORYL}-8-OXOGUANOSINE",
            "nucleic-acid",
        ),
        "UDT": ("O-DECYL HYDROGEN THIOCARBONATE", "non-polymer"),
        "CA3": ("COA-S-TRIMETHYLENE-ACETYL-TRYPTAMINE", "non-polymer"),
        "CA2": (
            "(1S,3R,4R,5S)-1,3,4-TRIHYDROXY-5-(3-PHENOXYPROPYL)CYCLOHEXANECARBOXYLIC ACID",
            "non-polymer",
        ),
        "CA1": ("ETHYL PROPIONATE", "non-polymer"),
        "CA0": ("5'-O-[(S)-(CARBAMOYLOXY)(HYDROXY)PHOSPHORYL]ADENOSINE", "non-polymer"),
        "CA7": (
            "7-CYCLOHEXYLHEPTYL 4-O-ALPHA-D-GLUCOPYRANOSYL-BETA-D-GLUCOPYRANOSIDE",
            "non-polymer",
        ),
        "CA6": ("?", "non-polymer"),
        "CA5": ("COA-S-ACETYL 5-BROMOTRYPTAMINE", "non-polymer"),
        "CA4": ("CYPROTERONE ACETATE", "non-polymer"),
        "IB8": (
            "(7R)-14-CYCLOHEXYL-7-{[2-(DIMETHYLAMINO)ETHYL](METHYL)AMINO}-7,8-DIHYDRO-6H-INDOLO[1,2-E][1,5]BENZOXAZOCINE-11-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "74A": (
            "N'-{(1S,2S)-1-(3,5-DIFLUOROBENZYL)-2-HYDROXY-2-[(2R)-4-(PHENYLCARBONYL)PIPERAZIN-2-YL]ETHYL}-5-METHYL-N,N-DIPROPYLBENZENE-1,3-DICARBOXAMIDE",
            "non-polymer",
        ),
        "F4C": (
            "2-(4-FLUOROPHENYL)-3-OXO-6-PYRIDIN-4-YL-N-[2-(TRIFLUOROMETHYL)BENZYL]-2,3-DIHYDROPYRIDAZINE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "OGN": (
            "[(Z)-[(3R,4R,5R,6R)-3-ACETAMIDO-6-(HYDROXYMETHYL)-4,5-BIS(OXIDANYL)OXAN-2-YLIDENE]AMINO] N-PHENYLCARBAMATE",
            "non-polymer",
        ),
        "UDG": ("?", "non-polymer"),
        "MKY": (
            "ETHYL [(2Z)-2-(CARBAMIMIDOYLIMINO)-6-HYDROXY-1,3-BENZOTHIAZOL-3(2H)-YL]ACETATE",
            "non-polymer",
        ),
        "SOH": ("HYDROGEN SULFATE", "non-polymer"),
        "A58": (
            "4-(6-{[(4-METHYLCYCLOHEXYL)AMINO]METHYL}-1,4-DIHYDROINDENO[1,2-C]PYRAZOL-3-YL)BENZOIC ACID",
            "non-polymer",
        ),
        "F4H": (
            "1,5-ANHYDRO-2,3-DIDEOXY-3-FLUORO-2-(5-METHYL-2,4-DIOXO-3,4-DIHYDROPYRIMIDIN-1(2H)-YL)-6-O-PHOSPHONO-D-MANNITOL",
            "nucleic-acid",
        ),
        "IB2": ("P1-P2-METHYLENE-P3-THIO-DIADENOSINE TRIPHOSPHATE", "non-polymer"),
        "F4M": ("N5-FORMYL-TETRAHYDROMETHANOPTERIN", "non-polymer"),
        "74M": (
            "METHYL N-({(2S,3S)-3-[(PROPYLAMINO)CARBONYL]OXIRAN-2-YL}CARBONYL)-L-ISOLEUCYL-L-PROLINATE",
            "non-polymer",
        ),
        "UDM": ("URIDINE-DIPHOSPHATE-METHYLENE-N-ACETYL-GLUCOSAMINE", "non-polymer"),
        "MII": ("METHICILLIN", "non-polymer"),
        "MIH": (
            "8-(2,6-DICHLOROPHENYL)-4-(2,4-DIFLUOROPHENYL)-2-PIPERIDIN-4-YL-1,7-NAPHTHYRIDINE 7-OXIDE",
            "non-polymer",
        ),
        "MIK": ("METHYL ISONICOTINIMIDATE LYSINE", "non-polymer"),
        "GHW": (
            "N-({4-[(2-AMINO-6-METHYL-4-OXO-3,4-DIHYDROTHIENO[2,3-D]PYRIMIDIN-5-YL)SULFANYL]PHENYL}CARBONYL)-L-GLUTAMIC ACID",
            "peptide",
        ),
        "GHP": ("(2R)-AMINO(4-HYDROXYPHENYL)ETHANOIC ACID", "peptide"),
        "MIL": ("MILRINONE", "non-polymer"),
        "GHR": (
            "[[(3S,4S,5R,6R)-3,4,5-TRIHYDROXY-6-(HYDROXYMETHYL)PIPERIDIN-2-YLIDENE]AMINO] N-(4-CHLOROPHENYL)CARBAMATE",
            "non-polymer",
        ),
        "GHS": ("6-CARBAMIMIDAMIDOHEXANOIC ACID", "non-polymer"),
        "MIA": (
            "2-METHYLTHIO-N6-ISOPENTENYL-ADENOSINE-5'-MONOPHOSPHATE",
            "nucleic-acid",
        ),
        "SOM": ("METHYL PHOSPHINIC ACID", "non-polymer"),
        "MIC": ("ALPHA-METHYLISOCITRIC ACID", "non-polymer"),
        "MIB": ("(3E)-4-(1-METHYL-1H-INDOL-3-YL)BUT-3-EN-2-ONE", "non-polymer"),
        "SF2": (
            "5-CHLORO-1-(3-METHOXYBENZYL)-3-(PHENYLTHIO)-1H-INDOLE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "MID": (
            "1-[N-(NAPHTHALEN-2-YLSULFONYL)GLYCYL-4-CARBAMIMIDOYL-D-PHENYLALANYL]PIPERIDINE",
            "peptide",
        ),
        "MIG": (
            "(2R,3R,4R,5S)-1-(2-HYDROXYETHYL)-2-(HYDROXYMETHYL)PIPERIDINE-3,4,5-TRIOL",
            "non-polymer",
        ),
        "MBA": ("2-METHYLBUTYLAMINE", "non-polymer"),
        "MIY": (
            "(4S,4AS,5AR,12AS)-4,7-BIS(DIMETHYLAMINO)-3,10,12,12A-TETRAHYDROXY-1,11-DIOXO-1,4,4A,5,5A,6,11,12A-OCTAHYDROTETRACENE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "MIX": (
            "1,4-DIHYDROXY-5,8-BIS({2-[(2-HYDROXYETHYL)AMINO]ETHYL}AMINO)-9,10-ANTHRACENEDIONE",
            "non-polymer",
        ),
        "GHG": (
            "(2S,4S)-2,5-DIAMINO-4-HYDROXY-5-OXOPENTANOIC ACID (NON-PREFERRED NAME)",
            "peptide",
        ),
        "NYS": (
            "S-{5-[(1R)-2-AMINO-1-HYDROXYETHYL]-2,3-DIHYDROXYPHENYL}-L-CYSTEINE",
            "peptide",
        ),
        "GAQ": ("5-HYDROXYPENTANAL", "non-polymer"),
        "GHC": (
            "N-({4-[(2-AMINO-6-ETHYL-4-OXO-3,4-DIHYDROTHIENO[2,3-D]PYRIMIDIN-5-YL)SULFANYL]PHENYL}CARBONYL)-L-GLUTAMIC ACID",
            "peptide",
        ),
        "MIP": ("MONOISOPROPYL ESTER PHOSPHONIC ACID", "non-polymer"),
        "MIS": ("MONOISOPROPYLPHOSPHORYLSERINE", "peptide"),
        "MBC": (
            "2-(5-{4-[AMINO(IMINO)METHYL]PHENYL}-2-FURYL)-1H-BENZIMIDAZOLE-5-CARBOXIMIDAMIDE",
            "non-polymer",
        ),
        "MIU": (
            "N-{(1S)-1-{[4-(3-AMINOPROPYL)PIPERAZIN-1-YL]CARBONYL}-4-[(DIAMINOMETHYLENE)AMINO]BUTYL}-3-(TRIFLUOROMETHYL)BENZENESULFONAMIDE",
            "peptide",
        ),
        "MIT": (
            "AMINO{[(4S)-5-[(2R,4R)-2-CARBOXY-4-METHYLPIPERIDIN-1-YL]-4-({[(3R)-3-METHYL-1,2,3,4-TETRAHYDROQUINOLIN-8-YL]SULFONYL}AMINO)-5-OXOPENTYL]AMINO}METHANIMINIUM",
            "peptide",
        ),
        "MIV": ("MYCINAMICIN IV", "non-polymer"),
        "AOG": (
            "4-AMINO-2-OCTYLOXY-6-HYDROXYMETHYL-TETRAHYDRO-PYRAN-3,5-DIOL",
            "saccharide",
        ),
        "AOF": (
            "N-(5-AMINO-2,6-DIOXO-1,2,3,6-TETRAHYDROPYRIMIDIN-4-YL)-5-O-PHOSPHONO-BETA-D-RIBOFURANOSYLAMINE",
            "non-polymer",
        ),
        "AOE": (
            "N-BUTYL-11-[(7R,8R,9S,13S,14S,17S)-3,17-DIHYDROXY-13-METHYL-7,8,9,11,12,13,14,15,16,17-DECAHYDRO-6H-CYCLOPENTA[A]PHENANTHREN-7-YL]-N-METHYLUNDECANAMIDE",
            "non-polymer",
        ),
        "AOD": (
            "5-BROMO-2-{[(1S,3R)-3-CARBOXYCYCLOHEXYL]AMINO}BENZOIC ACID",
            "non-polymer",
        ),
        "AOC": ("5'-O-PROP-2-YN-1-YLADENOSINE", "non-polymer"),
        "TDM": (
            "2-[(2E)-3-[(4-AMINO-2-METHYLPYRIMIDIN-5-YL)METHYL]-2-(1-HYDROXYETHYLIDENE)-4-METHYL-2,3-DIHYDRO-1,3-THIAZOL-5-YL]ETHYL TRIHYDROGEN DIPHOSPHATE",
            "non-polymer",
        ),
        "AOA": ("(AMINOOXY)ACETIC ACID", "non-polymer"),
        "SOP": ("S-(2-OXOPROPYL)-COENZYME A", "non-polymer"),
        "AOO": (
            "N-ETHYL-6-METHOXY-N'-(1-METHYLETHYL)-1,3,5-TRIAZINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "AON": ("5-ALPHA-ANDROSTANE-3-BETA,17-ALPHA-DIOL", "non-polymer"),
        "AOM": ("5-ALPHA-ANDROSTANE-3-BETA,17BETA-DIOL", "non-polymer"),
        "AOL": ("(1R,2R,3S,4S,5R)-5-AMINOCYCLOPENTANE-1,2,3,4-TETROL", "non-polymer"),
        "SOS": (
            "[(1R,2S,4R)-4-({2-AMINO-5-[BENZYL(FORMYL)AMINO]-6-OXO-1,6-DIHYDROPYRIMIDIN-4-YL}AMINO)-2-HYDROXYCYCLOPENTYL]METHYL DIHYDROGEN PHOSPHATE",
            "nucleic-acid",
        ),
        "T0I": (
            "(2S)-2-AMINO-3-(2,4-DIHYDROXY-5-IMIDAZOL-1-YL-PHENYL)PROPANOIC ACID",
            "peptide",
        ),
        "O2C": ("3'-DEOXY-CYTIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "NQB": (
            "N-CYCLOPROPYL-4-METHYL-3-[6-(4-METHYLPIPERAZIN-1-YL)-4-OXIDANYLIDENE-QUINAZOLIN-3-YL]BENZAMIDE",
            "non-polymer",
        ),
        "AOV": ("ADP ORTHOVANADATE", "non-polymer"),
        "SOR": ("D-SORBITOL", "non-polymer"),
        "AOS": ("D-ALLOSE", "saccharide"),
        "AOR": ("N~2~-ACETYL-L-ORNITHINE", "non-polymer"),
        "AOP": ("PENTYLOXYAMINO-ACETALDEHYDE", "non-polymer"),
        "NQI": (
            "N~2~-[(2S)-2-{[(2R)-4-ETHOXY-2-HYDROXY-4-OXOBUTANOYL]AMINO}PENT-4-ENOYL]-L-ARGINYL-L-TRYPTOPHANAMIDE",
            "non-polymer",
        ),
        "AOX": ("(3BETA,5ALPHA)-3-HYDROXYANDROSTAN-17-ONE", "non-polymer"),
        "TDL": (
            "3-[(4-AMINO-2-METHYLPYRIMIDIN-5-YL)METHYL]-2-(1-CARBOXY-1-HYDROXYETHYL)-5-(2-{[HYDROXY(PHOSPHONOOXY)PHOSPHORYL]OXY}ETHYL)-4-METHYL-1,3-THIAZOL-3-IUM",
            "non-polymer",
        ),
        "MBZ": (
            "1-[2-DEOXYRIBOFURANOSYL]-4-METHYL-BENZOIMIDAZOLE-5'-MONOPHOSPHATE",
            "nucleic-acid",
        ),
        "PUX": (
            "8-BENZO[1,3]DIOXOL-,5-YLMETHYL-9-BUTYL-2-FLUORO-9H-PURIN-6-YLAMINE",
            "non-polymer",
        ),
        "MBT": ("3,7-BIS(DIMETHYLAMINO)PHENOTHIAZIN-5-IUM", "non-polymer"),
        "AO5": ("?", "non-polymer"),
        "T04": (
            "2-[[2-FLUORO-5-[[[4-[(Z)-(4-OXO-2-SULFANYLIDENE-1,3-THIAZOLIDIN-5-YLIDENE)METHYL]PHENYL]AMINO]METHYL]PHENYL]CARBONYLAMINO]PENTANEDIOIC ACID",
            "non-polymer",
        ),
        "AO3": ("ALLOSAMIDIN", "non-polymer"),
        "AO2": (
            "N'-(2S,3R)-3-AMINO-4-CYCLOHEXYL-2-HYDROXY-BUTANO-N-(4-METHYLPHENYL)HYDRAZIDE",
            "non-polymer",
        ),
        "AO1": (
            "(2S,3R)-3-AMINO-2-HYDROXY-5-(ETHYLSULFANYL)PENTANOYL-((S)-(-)-(1-NAPHTHYL)ETHYL)AMIDE",
            "non-polymer",
        ),
        "SOX": (
            "N-[(2S,4S,6R)-2-(DIHYDROXYMETHYL)-4-HYDROXY-3,3-DIMETHYL-7-OXO-4LAMBDA~4~-THIA-1-AZABICYCLO[3.2.0]HEPT-6-YL]-2-PHENYLACETAMIDE",
            "non-polymer",
        ),
        "MBV": ("MESOBILIVERDIN IV ALPHA", "non-polymer"),
        "AO9": (
            "5'-DEOXY-5'-[(IMIDAZO[2,1-B][1,3]THIAZOL-5-YLCARBONYL)AMINO]CYTIDINE",
            "non-polymer",
        ),
        "SFX": (
            "(3S)-N-METHYL-3-PHENYL-3-[4-(TRIFLUOROMETHYL)PHENOXY]PROPAN-1-AMINE",
            "non-polymer",
        ),
        "GH3": ("3'-DEOXY-GUANOSINE-5'-TRIPHOSPHATE", "nucleic-acid"),
        "MBP": (
            "1-[(2-AMINO-6,9-DIHYDRO-1H-PURIN-6-YL)OXY]-3-METHYL-2-BUTANOL",
            "non-polymer",
        ),
        "SFT": ("SULFOGALACTOCERAMIDE", "non-polymer"),
        "SFU": ("METHYL 6-DEOXY-1-SELENO-ALPHA-L-GALACTOPYRANOSIDE", "non-polymer"),
        "SFR": (
            "(2R)-2-[(1S,2R)-1-CARBOXY-2-HYDROXYPROPYL]-5-[(2R)-TETRAHYDROFURAN-2-YL]-2,3-DIHYDRO-1,3-THIAZOLE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "SFP": (
            "5,10,15,20-TETRAKIS(4-SULPFONATOPHENYL)-21H,23H-PORPHINE",
            "non-polymer",
        ),
        "GAB": ("3-AMINOBENZOIC ACID", "non-polymer"),
        "SFN": ("SULFONATE GROUP", "non-polymer"),
        "SFO": ("SULFONIC ACID", "non-polymer"),
        "SFM": ("ANALOGUE OF MACROLIDE FROM DEGRADED SANGLIFEHRIN A", "non-polymer"),
        "MBR": ("TRIBROMOMETHANE", "non-polymer"),
        "SFI": (
            "2-[(3Z)-6-FLUORANYL-2-METHYL-3-[(4-METHYLSULFANYLPHENYL)METHYLIDENE]INDEN-1-YL]ETHANOIC ACID",
            "non-polymer",
        ),
        "MI1": (
            "3-{(3R,4R)-4-METHYL-3-[METHYL(7H-PYRROLO[2,3-D]PYRIMIDIN-4-YL)AMINO]PIPERIDIN-1-YL}-3-OXOPROPANENITRILE",
            "non-polymer",
        ),
        "MI0": (
            "(3R,4S)-4-AZANYL-5-CYCLOHEXYL-2,2-DIFLUORO-3-HYDROXY-N-METHYL-PENTANAMIDE",
            "peptide",
        ),
        "MI3": ("?", "non-polymer"),
        "MI2": ("2-(2-METHYLPHENYL)-1H-INDOLE-5-CARBOXIMIDAMIDE", "non-polymer"),
        "MI5": (
            "N-(3-METHYLISOTHIAZOL-5-YL)-3-(1-METHYL-1H-PYRAZOL-4-YL)-5-[(3R)-PIPERIDIN-3-YL]PYRAZOLO[1,5-A]PYRIMIDIN-7-AMINE",
            "non-polymer",
        ),
        "MI4": (
            "(1R,3R,7E,17BETA)-17-{(1S,2E,5R)-5-HYDROXY-1-METHYL-6-[(3S,5S,7S)-TRICYCLO[3.3.1.1~3,7~]DEC-1-YL]HEX-2-EN-1-YL}-2-METHYLIDENE-9,10-SECOESTRA-5,7-DIENE-1,3-DIOL",
            "non-polymer",
        ),
        "MI6": (
            "(2'R,3R,4'R,5'R)-6-CHLORO-4'-(3-CHLORO-2-FLUOROPHENYL)-2'-(2,2-DIMETHYLPROPYL)-N-(2-MORPHOLIN-4-YLETHYL)-2-OXO-1,2-DIHYDROSPIRO[INDOLE-3,3'-PYRROLIDINE]-5'-CARBOXAMIDE",
            "non-polymer",
        ),
        "KST": ("N~6~-(5-CARBOXY-3-THIENYL)-L-LYSINE", "peptide"),
        "NPI": ("(2S)-2-AMINOHEPTANEDIOIC ACID", "peptide"),
        "KSS": (
            "7-(3,4-DIMETHOXYPHENYL)-5-(ETHYLSULFANYL)IMIDAZO[1,2-C]PYRIMIDINE",
            "non-polymer",
        ),
        "KSR": (
            "6-(5-METHOXY-1-METHYL-1H-INDOL-3-YL)-5H-PYRROLO[2,3-B]PYRAZINE",
            "non-polymer",
        ),
        "KSE": (
            "7-[(2-AMINO-2-METHYLPROPYL)AMINO]-2-[(3,5-DIMETHOXYPHENYL)AMINO]-5-METHYLPYRAZOLO[1,5-A]PYRIMIDINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "KSG": (
            "(1S,2R,3S,4R,5S,6S)-2,3,4,5,6-PENTAHYDROXYCYCLOHEXYL 2-AMINO-4-{[CARBOXY(IMINO)METHYL]AMINO}-2,3,4,6-TETRADEOXY-ALPHA-D-ARABINO-HEXOPYRANOSIDE",
            "non-polymer",
        ),
        "KSF": (
            "N-(2-CHLOROPHENYL)-5-PHENYLIMIDAZO[1,5-A]PYRAZIN-8-AMINE",
            "non-polymer",
        ),
        "KSA": ("K-252A", "non-polymer"),
        "KSC": (
            "7-[(CYCLOPROPYLMETHYL)AMINO]-2-[(4-METHOXYPHENYL)AMINO]-5-METHYLPYRAZOLO[1,5-A]PYRIMIDINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "KSM": (
            "N-(2-CHLORO-6-METHYLPHENYL)-8-[(3S)-3-METHYLPIPERAZIN-1-YL]IMIDAZO[1,5-A]QUINOXALIN-4-AMINE",
            "non-polymer",
        ),
        "KSL": (
            "N-(2,6-DIMETHYLPHENYL)-5-PHENYLIMIDAZO[1,5-A]PYRAZIN-8-AMINE",
            "non-polymer",
        ),
        "KSH": (
            "2-[(3,5-DIMETHOXYPHENYL)AMINO]-5-ETHYL-7-[(2R)-2-(HYDROXYMETHYL)PYRROLIDIN-1-YL]PYRAZOLO[1,5-A]PYRIMIDINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "KSK": (
            "7-[(2-AMINO-2-METHYLPROPYL)AMINO]-5-CYCLOPROPYL-2-[(3,5-DIMETHOXYPHENYL)AMINO]PYRAZOLO[1,5-A]PYRIMIDINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "MOJ": (
            "(5E,7S)-2-AMINO-7-(4-FLUORO-2-PYRIDIN-3-YLPHENYL)-4-METHYL-7,8-DIHYDROQUINAZOLIN-5(6H)-ONE OXIME",
            "non-polymer",
        ),
        "YLQ": (
            "3-[1-[2-(4-METHYLPHENOXY)ETHYL]BENZIMIDAZOL-2-YL]SULFANYLPROPANOIC ACID",
            "non-polymer",
        ),
        "RX3": (
            "N-({(1S,2R)-2-[(S)-[(1R)-1-{[(BENZYLOXY)CARBONYL]AMINO}-2-PHENYLETHYL](HYDROXY)PHOSPHORYL]CYCLOPENTYL}CARBONYL)-L-TRYPTOPHAN",
            "non-polymer",
        ),
        "TL3": (
            "2,5-BIS-O-{4-[AMINO(IMINO)METHYL]PHENYL}-1,4:3,6-DIANHYDRO-D-GLUCITOL",
            "non-polymer",
        ),
        "TL2": (
            "2-O-{4-[AMINO(IMINO)METHYL]PHENYL}-5-O-{3-[AMINO(IMINO)METHYL]PHENYL}-1,4:3,6-DIANHYDRO-D-GLUCITOL",
            "non-polymer",
        ),
        "GA3": ("GIBBERELLIN A3", "non-polymer"),
        "TL1": (
            "2,5-BIS-O-{3-[AMINO(IMINO)METHYL]PHENYL}-1,4:3,6-DIANHYDRO-D-GLUCITOL",
            "non-polymer",
        ),
        "XUA": ("?", "nucleic-acid"),
        "LOL": ("(2S)-2-AMINO-4-METHYLPENTANE-1,1-DIOL", "peptide"),
        "QTR": ("OXYGEN OF QUINONE", "non-polymer"),
        "GA0": ("GLYCYLALANYL-N-2-NAPHTHYL-L-PROLINEAMIDE", "non-polymer"),
        "KS5": (
            "1-(1-METHYLETHYL)-3-QUINOLIN-6-YL-1H-PYRAZOLO[3,4-D]PYRIMIDIN-4-AMINE",
            "non-polymer",
        ),
        "KS4": (
            "1-CYCLOBUTYL-3-(3,4-DIMETHOXYPHENYL)-1H-PYRAZOLO[3,4-D]PYRIMIDIN-4-AMINE",
            "non-polymer",
        ),
        "KS1": (
            "1-CYCLOPENTYL-3-(1H-PYRROLO[2,3-B]PYRIDIN-5-YL)-1H-PYRAZOLO[3,4-D]PYRIMIDIN-4-AMINE",
            "non-polymer",
        ),
        "QTP": (
            "3-{2-DEOXY-5-O-[(R)-HYDROXY{[(R)-HYDROXY(PHOSPHONOOXY)PHOSPHORYL]OXY}PHOSPHORYL]-BETA-D-ERYTHRO-PENTOFURANOSYL}-7-METHYL-3H-IMIDAZO[4,5-B]PYRIDINE",
            "non-polymer",
        ),
        "PMO": ("N1-(5'-PHOSPHO-ALPHA-RIBOSYL)-5-METHOXYBENZIMIDAZOLE", "non-polymer"),
        "TL4": (
            "2-O-{3-[AMINO(IMINO)METHYL]PHENYL}-5-O-{4-[AMINO(IMINO)METHYL]PHENYL}-1,4:3,6-DIANHYDRO-D-GLUCITOL",
            "non-polymer",
        ),
        "SO2": ("SULFUR DIOXIDE", "non-polymer"),
        "LOM": ("THIAZOLE", "non-polymer"),
        "VCA": ("VACCENIC ACID", "non-polymer"),
        "TDF": ("4-(TRIFLUOROMETHYL)-D-PHENYLALANINE", "peptide"),
        "SO4": ("SULFATE ION", "non-polymer"),
        "X6D": (
            "6-BROMO-7-[4-(4-CHLOROBENZYL)PIPERAZIN-1-YL]-2-[4-(MORPHOLIN-4-YLMETHYL)PHENYL]-3H-IMIDAZO[4,5-B]PYRIDINE",
            "non-polymer",
        ),
        "YHO": ("(1S,2S,3R,4R)-4-AMINOCYCLOPENTANE-1,2,3-TRIOL", "non-polymer"),
        "LOR": ("LORACABEF (OPEN FORM)", "non-polymer"),
        "TDE": (
            "(E)-(2R,3R,4S,5R)-3,4,5-TRIHYDROXY-2-METHOXY-8,8-DIMETHYL-NON-6-ENOIC ACID ((3S,6R)-6-HYDROXY-2-OXO-AZEPAN-3-YL)-AMIDE",
            "non-polymer",
        ),
        "GT1": (
            "[5-HYDROXY-4-(METHOXYMETHYL)-6-METHYLPYRIDIN-3-YL]METHYL DIHYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "TDD": ("(2R)-2-AMINO-3,3-DIMETHYL-BUTANOIC ACID", "peptide"),
        "4LG": (
            "METHYL (6-{[(2R)-1-(4-{3-[(2-METHOXYBENZYL)OXY]PROPOXY}PHENYL)-6-OXOPIPERAZIN-2-YL]METHOXY}-1H-INDOL-1-YL)ACETATE",
            "non-polymer",
        ),
        "4LI": (
            "(4R)-3-(4-[4-(2-CHLOROPHENYL)PIPERAZIN-1-YL]-6-{[2-METHYL-6-(METHYLCARBAMOYL)PHENYL]AMINO}-1,3,5-TRIAZIN-2-YL)-N-METHYL-1,3-THIAZOLIDINE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "4LM": (
            "(2E)-2-{[(1E)-{3-HYDROXY-2-METHYL-5-[(PHOSPHONOOXY)METHYL]PYRIDIN-4-YL}METHYLIDENE]AMINO}BUT-2-ENOIC ACID",
            "non-polymer",
        ),
        "4LL": ("2-(BUTYLAMINO)QUINOLIN-8-OL", "non-polymer"),
        "M7R": (
            "(2R)-2-BENZYL-3-(4-PROPOXY-3-{[({4-[(3S,5S,7S)-TRICYCLO[3.3.1.1~3,7~]DEC-1-YL]PHENYL}CARBONYL)AMINO]METHYL}PHENYL)PROPANOIC ACID",
            "non-polymer",
        ),
        "M7P": ("D-GLYCERO-D-MANNOPYRANOSE-7-PHOSPHATE", "non-polymer"),
        "GCN": ("3-DEOXY-D-GLUCOSAMINE", "saccharide"),
        "TDZ": (
            "(5R)-5-(4-{[(2R)-6-HYDROXY-2,5,7,8-TETRAMETHYL-3,4-DIHYDRO-2H-CHROMEN-2-YL]METHOXY}BENZYL)-1,3-THIAZOLIDINE-2,4-DIONE",
            "non-polymer",
        ),
        "P36": ("1-BENZYL-4-(BENZYLOXY)-3-BROMOPYRIDIN-2(1H)-ONE", "non-polymer"),
        "882": (
            "(3R,5R)-7-[5-(ANILINOCARBONYL)-3,4-BIS(4-FLUOROPHENYL)-1-ISOPROPYL-1H-PYRROL-2-YL]-3,5-DIHYDROXYHEPTANOIC ACID",
            "non-polymer",
        ),
        "881": (
            "N',N''-{FURAN-2,5-DIYLBIS[3-(PIPERIDIN-4-YLOXY)BENZENE-4,1-DIYL]}DIPYRIDINE-2-CARBOXIMIDAMIDE",
            "non-polymer",
        ),
        "880": (
            "CYCLOHEXYL-{4-[5-(3,4-DICHLOROPHENYL)-2-PIPERIDIN-4-YL-3-PROPYL-3H-IMIDAZOL-4-YL]-PYRIMIDIN-2-YL}AMINE",
            "non-polymer",
        ),
        "887": (
            "N-CYCLOPROPYL-6-[(6,7-DIMETHOXYQUINOLIN-4-YL)OXY]NAPHTHALENE-1-CARBOXAMIDE",
            "non-polymer",
        ),
        "886": (
            "N-({4'-[(4-ISOBUTYRYLPHENOXY)METHYL]BIPHENYL-4-YL}SULFONYL)-D-VALINE",
            "non-polymer",
        ),
        "885": (
            "1-[(2-AMINO-4-CHLORO-5-METHYLPHENYL)SULFONYL]-L-PROLINE",
            "non-polymer",
        ),
        "889": (
            "N-{6,6-DIMETHYL-5-[(1-METHYLPIPERIDIN-4-YL)CARBONYL]-1,4,5,6-TETRAHYDROPYRROLO[3,4-C]PYRAZOL-3-YL}-3-METHYLBUTANAMIDE",
            "non-polymer",
        ),
        "888": (
            "3-(1,1-DIOXIDO-4H-1,2,4-BENZOTHIADIAZIN-3-YL)-4-HYDROXY-1-(3-METHYLBUTYL)QUINOLIN-2(1H)-ONE",
            "non-polymer",
        ),
        "TDY": ("?", "nucleic-acid"),
        "GTC": (
            "(1R,2S)-2-(5-THIOXO-4,5-DIHYDRO-1H-1,2,4-TRIAZOL-3-YL)CYCLOHEXANECARBOXYLIC ACID",
            "non-polymer",
        ),
        "G04": (
            "(3R,3AS,4R,6AR)-4-METHOXYHEXAHYDROFURO[2,3-B]FURAN-3-YL [(2S,3R)-3-HYDROXY-4-{[(4-METHOXYPHENYL)SULFONYL](2-METHYLPROPYL)AMINO}-1-PHENYLBUTAN-2-YL]CARBAMATE",
            "non-polymer",
        ),
        "G05": (
            "(4AR,6R,7AS)-HEXAHYDRO-4AH-CYCLOPENTA[B][1,4]DIOXIN-6-YL [(1S,2R)-1-BENZYL-2-HYDROXY-3-{[(4-METHOXYPHENYL)SULFONYL](2-METHYLPROPYL)AMINO}PROPYL]CARBAMATE",
            "non-polymer",
        ),
        "G00": ("4-BUTOXY-3-CHLOROBENZYL IMIDOTHIOCARBAMATE", "non-polymer"),
        "G01": ("N-[(R)-HYDROXY(METHYL)PHOSPHORYL]-D-GLUTAMIC ACID", "peptide"),
        "8MO": ("METHOXSALEN", "non-polymer"),
        "2BZ": (
            "5-(2-CHLOROPHENYL)-3-METHYL-7-NITROPYRAZOLO[3,4-B][1,4]BENZODIAZEPINE",
            "non-polymer",
        ),
        "5B2": ("5-PHENYL-1H-INDAZOL-3-AMINE", "non-polymer"),
        "5B3": (
            "4-(3-AMINO-1H-INDAZOL-5-YL)-N-TERT-BUTYLBENZENESULFONAMIDE",
            "non-polymer",
        ),
        "2BU": ("?", "nucleic-acid"),
        "2BT": ("2'-O-BUTYL-THYMIDINE", "nucleic-acid"),
        "2BS": ("ETHYL 2-AMINO-1,3-BENZOTHIAZOLE-6-CARBOXYLATE", "non-polymer"),
        "2BR": ("2-BROMOPHENOL", "non-polymer"),
        "2BP": ("9H-PURIN-2-AMINE", "non-polymer"),
        "2BN": ("(1R,4S)-2-AZABORNANE", "non-polymer"),
        "2BM": ("DIBROMOMETHANE", "non-polymer"),
        "2BL": (
            "(3R)-3-ETHYL-N-[(4-METHYLPHENYL)SULFONYL]-L-ASPARTIC ACID",
            "non-polymer",
        ),
        "2BJ": (
            "2-{[3-(3,4-DIMETHYLPHENOXY)PHENYL]CARBAMOYL}-4-NITROBENZOIC ACID",
            "non-polymer",
        ),
        "2BH": (
            "[(1E,5S)-5-AMINO-5-CARBOXYPENT-1-ENYL](TRIHYDROXY)BORATE(1-)",
            "non-polymer",
        ),
        "2BF": ("ALPHA-BENZYL-AMINOBENZYL-PHOSPHONIC ACID", "non-polymer"),
        "2BE": (
            "4-CHLORO-N'-[(1E)-(3,5-DIBROMO-2,4-DIHYDROXYPHENYL)METHYLIDENE]BENZOHYDRAZIDE",
            "non-polymer",
        ),
        "2BD": ("?", "nucleic-acid"),
        "2BC": (
            "N'-[(1E)-(3,5-DIBROMO-2,4-DIHYDROXYPHENYL)METHYLIDENE]NAPHTHALENE-2-CARBOHYDRAZIDE",
            "non-polymer",
        ),
        "2BB": (
            "(2S,3R)-N~4~-[(1S)-1-(DIMETHYLCARBAMOYL)-2,2-DIMETHYLPROPYL]-N~1~,2-DIHYDROXY-3-(2-METHYLPROPYL)BUTANEDIAMIDE",
            "non-polymer",
        ),
        "2BA": ("BIS-(3',5')-CYCLIC-DIMERIC-ADENOSINE-MONOPHOSPHATE", "non-polymer"),
        "5BT": ("5-BROMO-2'-DEOXYCYTIDINE", "non-polymer"),
        "5BU": ("5-BROMO-URIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "TLX": (
            "N1-(2-AMINO-4-METHYLPENTYL)OCTAHYDRO-PYRROLO[1,2-A] PYRIMIDINE",
            "non-polymer",
        ),
        "5BP": (
            "N-CYCLOPROPYL-4-PYRAZOLO[1,5-B]PYRIDAZIN-3-YLPYRIMIDIN-2-AMINE",
            "non-polymer",
        ),
        "ST3": ("4-(ACETYLAMINO)-3-AMINO BENZOIC ACID", "non-polymer"),
        "5BR": ("1-BROMOPENTANE", "non-polymer"),
        "UHD": ("6-HYDROXY-5-UNDECYL-1,3-BENZOTHIAZOLE-4,7-DIONE", "non-polymer"),
        "2B5": (
            "2-(2-((BENZOFURAN-2-CARBOXAMIDO)METHYL)-5-METHOXY-1H-INDOL-1-YL)ACETIC ACID",
            "non-polymer",
        ),
        "2B4": ("2-(4-BROMANYLPHENOXY)PROPANEDIOIC ACID", "non-polymer"),
        "CK8": (
            "N-[4-(2,4-DIMETHYL-THIAZOL-5-YL)-PYRIMIDIN-2-YL]-N',N'-DIMETHYL-BENZENE-1,4-DIAMINE",
            "non-polymer",
        ),
        "5BD": (
            "N,N'-(9,10-DIOXO-9,10-DIHYDROANTHRACENE-2,7-DIYL)BIS[2-(DIMETHYLAMINO)ACETAMIDE]",
            "non-polymer",
        ),
        "5BE": (
            "3-METHYL-5-[5-(1-METHYLETHYL)-1H-BENZIMIDAZOL-2-YL]-N-(1-METHYLPIPERIDIN-4-YL)-1H-PYRAZOLE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "5BC": (
            "(5-{3-[4-(BIPHENYL-4-YLCARBONYL)-2-PROPYLPHENOXY]PROPOXY}-1H-INDOL-1-YL)ACETIC ACID",
            "non-polymer",
        ),
        "5BM": (
            "(2Z)-BIS{AMINO[(2-AMINOPHENYL)SULFANYL]METHYLIDENE}BUTANEDINITRILE",
            "non-polymer",
        ),
        "5BN": (
            "5-[(2-AMINOETHYL)AMINO]-6-FLUORO-3-(1H-PYRROL-2-YL)BENZO[CD]INDOL-2(1H)-ONE",
            "non-polymer",
        ),
        "5BO": (
            "5-BROMO-6-ETHYL-2-{5-[(4-METHYLPIPERAZIN-1-YL)SULFONYL]-2-PROPOXYPHENYL}PYRIMIDIN-4(3H)-ONE",
            "non-polymer",
        ),
        "8M0": ("MO(8)-O(28) CLUSTER", "non-polymer"),
        "88S": (
            "N-(2-METHYL-1,3-BENZOTHIAZOL-6-YL)-3-UREIDO-PROPANAMIDE",
            "non-polymer",
        ),
        "88R": ("2-((3,4-DIMETHOXYPHENYL)METHYL))PROPANEDIOIC ACID", "non-polymer"),
        "88Q": (
            "1,5:6,10-DIANHYDRO-3,4,7,8-TETRADEOXY-2,9-BIS-C-(HYDROXYMETHYL)-L-MANNO-DECITOL",
            "non-polymer",
        ),
        "88W": (
            "N~3~-(ETHYLCARBAMOYL)-N-(2-METHYL-1,3-BENZOTHIAZOL-6-YL)-BETA-ALANINAMIDE",
            "non-polymer",
        ),
        "88V": (
            "(4-{2-[({3-[(2-METHYL-1,3-BENZOTHIAZOL-6-YL)AMINO]-3-OXOPROPYL}CARBAMOYL)AMINO]ETHYL}BENZYL)PROPANEDIOIC ACID",
            "non-polymer",
        ),
        "GTQ": ("2,5-DIHYDROXYBENZOIC ACID", "non-polymer"),
        "88Z": (
            "N-(3-FLUORO-4-{[6-METHOXY-7-(3-MORPHOLIN-4-YLPROPOXY)QUINOLIN-4-YL]OXY}PHENYL)-N'-(4-FLUOROPHENYL)CYCLOPROPANE-1,1-DICARBOXAMIDE",
            "non-polymer",
        ),
        "88X": (
            "5-ETHYL-2-(PHENYLCARBONYL)IMIDAZO[1,2-A]PYRIMIDIN-7(1H)-ONE",
            "non-polymer",
        ),
        "GTR": ("GALACTURONIC ACID", "saccharide"),
        "G0B": ("?", "non-polymer"),
        "88A": (
            "8-(4-METHYLSULFONYLPHENYL)-N-(4-MORPHOLIN-4-YLPHENYL)-[1,2,4]TRIAZOLO[1,5-A]PYRIDIN-2-AMINE",
            "non-polymer",
        ),
        "GTT": ("GLUTATHIONE", "non-polymer"),
        "88N": (
            "{4-[4-({3-[(2-METHYL-1,3-BENZOTHIAZOL-6-YL)AMINO]-3-OXOPROPYL}AMINO)-4-OXOBUTYL]BENZYL}PROPANEDIOIC ACID",
            "non-polymer",
        ),
        "G0S": ("3-(BETA-D-GALACTOPYRANOSYLTHIO)PROPANOIC ACID", "non-polymer"),
        "669": (
            "1-(5-CARBOXYPENTYL)-5-(2,6-DICHLOROBENZYLOXY)-1H-INDOLE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "667": (
            "6-OXO-8,9,10,11-TETRAHYDRO-7H-CYCLOHEPTA[C][1]BENZOPYRAN-3-O-SULFAMATE",
            "non-polymer",
        ),
        "666": (
            "6-(4-{[2-(3-IODOBENZYL)-3-OXOCYCLOHEX-1-EN-1-YL]AMINO}PHENYL)-5-METHYL-4,5-DIHYDROPYRIDAZIN-3(2H)-ONE",
            "non-polymer",
        ),
        "665": (
            "(3S)-N-(3-BROMOPHENYL)-1-CYCLOHEXYL-5-OXOPYRROLIDINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "664": (
            "1-METHYL-2-(2-{[5-(4-METHYLPIPERAZIN-1-YL)-2-(TRIFLUOROMETHOXY)PHENYL]AMINO}PYRIMIDIN-4-YL)-1,5,6,7-TETRAHYDRO-4H-PYRROLO[3,2-C]PYRIDIN-4-ONE",
            "non-polymer",
        ),
        "663": (
            "3-CHLORO-N-(4-CHLORO-2-{[(5-CHLOROPYRIDIN-2-YL)AMINO]CARBONYL}-6-METHOXYPHENYL)-4-[(1-METHYL-1H-IMIDAZOL-2-YL)METHYL]THIOPHENE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "662": (
            "5-(4-FLUOROPHENYL)-2,3-DIHYDROXY-N-[(E)-3-[(2R,3R,4R,5R)-4-HYDROXY-3-METHYL-5-[6-(PROPYLAMINO)PURIN-9-YL]OXOLAN-2-YL]PROP-2-ENYL]BENZAMIDE",
            "non-polymer",
        ),
        "ZA4": (
            "[(2-FLUOROPHENYL)CARBONYLAMINO]METHYL-TRIHYDROXY-BORON",
            "non-polymer",
        ),
        "ZA5": (
            "[(2,6-DIFLUOROPHENYL)CARBONYLAMINO]METHYL-TRIHYDROXY-BORON",
            "non-polymer",
        ),
        "ZA2": (
            "[(2-CHLOROPHENYL)CARBONYLAMINO]METHYL-TRIHYDROXY-BORON",
            "non-polymer",
        ),
        "ZA3": (
            "TRIHYDROXY-[[(2-NITROPHENYL)CARBONYLAMINO]METHYL]BORON",
            "non-polymer",
        ),
        "XL3": ("PROPANE-1-THIOL", "non-polymer"),
        "9Y9": (
            "TERT-BUTYL {(2S,3R)-4-[(4S)-7-FLUORO-4-METHYL-1,1-DIOXIDO-4,5-DIHYDRO-1,2-BENZOTHIAZEPIN-2(3H)-YL]-3-HYDROXY-1-PHENYLBUTAN-2-YL}CARBAMATE",
            "non-polymer",
        ),
        "X69": (
            "4-METHOXY-5-NITRO-2-[(PYRIDIN-3-YLMETHYL)AMINO]BENZAMIDE",
            "non-polymer",
        ),
        "TLN": (
            "[(1R,3R,4R,7S)-7-HYDROXY-3-(THYMIN-1-YL)-2,5-DIOXABICYCLO[2.2.1]HEPT-1-YL]METHYL DIHYDROGEN PHOSPHATE",
            "nucleic-acid",
        ),
        "NYA": (
            "(1R,3R,7E,17BETA)-17-{(1R,2E,4R)-4-HYDROXY-1-METHYL-4-[(3S,5S,7S)-TRICYCLO[3.3.1.1~3,7~]DEC-1-YL]BUT-2-EN-1-YL}-2-METHYLIDENE-9,10-SECOESTRA-5,7-DIENE-1,3-DIOL",
            "non-polymer",
        ),
        "U4S": (
            "1-(5-DEOXY-5-PYRROLIDIN-1-YL-ALPHA-L-ARABINOFURANOSYL)PYRIMIDINE-2,4(1H,3H)-DIONE",
            "non-polymer",
        ),
        "TLM": ("THIOLACTOMYCIN", "non-polymer"),
        "N23": (
            "N-[2-(2-METHYL-1H-INDOL-3-YL)ETHYL]THIOPHENE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "N22": (
            "5-[3-(2,5-DIMETHOXYPHENYL)PROP-1-YN-1-YL]-6-ETHYLPYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "N21": (
            "(2R)-2-[[4-[[4-[(Z)-(2,4-DIOXO-1,3-THIAZOLIDIN-5-YLIDENE)METHYL]PHENOXY]METHYL]PHENYL]SULFONYLAMINO]PENTANEDIOIC ACID",
            "non-polymer",
        ),
        "N20": ("6-CYCLOHEXYLMETHYLOXY-2-(4'-HYDROXYANILINO)PURINE", "non-polymer"),
        "N25": (
            "1,3-BIS-(4-METHOXY-BENZENESULFONYL)-5,5-DIMETHYL-HEXAHYDRO-PYRIMIDINE-2-CARBOXYLIC ACID HYDROXYAMIDE",
            "non-polymer",
        ),
        "N24": (
            "(2S)-2-(3-BROMOPHENYL)-3-(5-CHLORO-2-HYDROXYPHENYL)-1,3-THIAZOLIDIN-4-ONE",
            "non-polymer",
        ),
        "TDS": (
            "8-HYDROXY-5,7-DIMETHOXY-3-METHYL-2-TRIDECYL-4H-CHROMEN-4-ONE",
            "non-polymer",
        ),
        "K54": (
            "(5S)-N-{(1S,2R)-3-[(1,3-BENZOTHIAZOL-6-YLSULFONYL)(2-METHYLPROPYL)AMINO]-1-BENZYL-2-HYDROXYPROPYL}-2-OXO-3-[2-(TRIFLUOROMETHYL)PHENYL]-1,3-OXAZOLIDINE-5-CARBOXAMIDE",
            "non-polymer",
        ),
        "N2O": ("NITROUS OXIDE", "non-polymer"),
        "N2M": ("5-AMINO-NAPHTALENE-2-MONOSULFONATE", "non-polymer"),
        "N2C": ("N,S-DIMETHYLCYSTEINE", "peptide"),
        "N2B": ("ETHYLBENZENE", "non-polymer"),
        "N2A": (
            "N-[2-CHLORO-5-(1-{3-[4-(6-CHLORO-3-METHYL-2-OXO-2,3-DIHYDRO-1H-BENZIMIDAZOL-1-YL)PIPERIDIN-1-YL]PROPYL}-6-OXO-1,6-DIHYDROPYRIMIDIN-5-YL)BENZYL]-4-FLUOROBENZAMIDE",
            "non-polymer",
        ),
        "N2G": ("?", "non-polymer"),
        "N2D": (
            "N-[2-CHLORO-5-(1-{3-[4-(6-CHLORO-3-METHYL-2-OXO-2,3-DIHYDRO-1H-BENZIMIDAZOL-1-YL)PIPERIDIN-1-YL]PROPYL}-6-OXO-1,6-DIHYDROPYRIDAZIN-3-YL)BENZYL]BENZAMIDE",
            "non-polymer",
        ),
        "P18": ("3-HYDROXY-1-PHENYLPROPAN-1-ONE", "non-polymer"),
        "U46": (
            "(5E)-7-{6-[(1E)-3-HYDROXYOCT-1-ENYL]-2-OXABICYCLO[2.2.1]HEPT-5-YL}HEPT-5-ENOIC ACID",
            "non-polymer",
        ),
        "N2P": ("PENTANE-1,5-DIAMINE", "non-polymer"),
        "MM4": ("MOLYBDENUM (IV) OXIDE", "non-polymer"),
        "N2T": (
            "(2S)-4-(2,5-DIFLUOROPHENYL)-N-METHYL-2-PHENYL-N-PIPERIDIN-4-YL-2,5-DIHYDRO-1H-PYRROLE-1-CARBOXAMIDE",
            "non-polymer",
        ),
        "ZAN": (
            "5'-O-[(S)-HYDROXY{[(S)-HYDROXY(PHOSPHONOOXY)PHOSPHORYL]AMINO}PHOSPHORYL]ADENOSINE",
            "non-polymer",
        ),
        "ZAL": ("3-CYCLOHEXYL-D-ALANINE", "peptide"),
        "ZAM": (
            "3-[(ACETYL-METHYL-AMINO)-METHYL]-4-AMINO-N-METHYL-N-(1-METHYL-1H-INDOL-2-YLMETHYL)-BENZAMIDE",
            "non-polymer",
        ),
        "ZAH": ("N-[(BENZYLOXY)CARBONYL]-L-ALANYL-L-PROLINE", "non-polymer"),
        "ZAF": (
            "O-(((1R)-((N-PHENYLMETHOXYCARBONYL-L-ALANYL)AMINO)ETHYL)HYDROXYPHOSPHONO)-L-BENZYLACETIC ACID",
            "non-polymer",
        ),
        "ZAD": ("(S)-1'-(2',3'-DIHYDROXYPROPYL)-ADENINE", "nucleic-acid"),
        "ZAE": ("N-METHYL-D-PHENYLALANINE", "peptide"),
        "NYG": (
            "[(4Z)-2-[(1S)-1,3-DIAMINO-3-OXOPROPYL]-4-(4-HYDROXYBENZYLIDENE)-5-OXO-4,5-DIHYDRO-1H-IMIDAZOL-1-YL]ACETIC ACID",
            "peptide",
        ),
        "ZAA": (
            "(5-{3-[2-(5-METHYL-2-PHENYL-1,3-OXAZOL-4-YL)ETHOXY]BENZYL}-2-PHENYL-2H-1,2,3-TRIAZOL-4-YL)ACETIC ACID",
            "non-polymer",
        ),
        "PP2": (
            "1-TERT-BUTYL-3-(4-CHLORO-PHENYL)-1H-PYRAZOLO[3,4-D]PYRIMIDIN-4-YLAMINE",
            "non-polymer",
        ),
        "66G": (
            "(4Z)-4-[HYDROXY(SULFANYL)METHYLIDENE]-2-[(2S)-PYRROLIDIN-2-YL]-1,3-OXAZOL-5(4H)-ONE",
            "non-polymer",
        ),
        "MM7": (
            "(1R)-1,4-ANHYDRO-2-DEOXY-1-(2-METHOXY-4-METHYLPHENYL)-5-O-PHOSPHONO-D-ERYTHRO-PENTITOL",
            "nucleic-acid",
        ),
        "ZAR": (
            "6-(4-DIFLUOROMETHOXY-3-METHOXY-PHENYL)-2H-PYRIDAZIN-3-ONE",
            "non-polymer",
        ),
        "ZAS": ("5'-AZIDO-5'-DEOXYADENOSINE", "non-polymer"),
        "ZAP": (
            "[N-(BENZYLOXYCARBONYL)AMINO](4-AMIDINOPHENYL)METHANE-PHOSPHONATE",
            "non-polymer",
        ),
        "SN8": (
            "1,6-DIMETHYL-4-(4-(4-(1-METHYLPYRIDINIUM-4-YLAMINO)PHENYLCARBAMOYL)PHENYLAMINO)QUINOLINIUM",
            "non-polymer",
        ),
        "MM6": ("NICKEL(II)(1,4,8,11-TETRAAZACYCLOTETRADECANE)", "non-polymer"),
        "P1D": (
            "((2S,3AS,4R,6S)-4-(HYDROXYMETHYL)-6-(4-OXO-4,5-DIHYDRO-3H-PYRROLO[3,2-D]PYRIMIDIN-7-YL)-TETRAHYDROFURO[3,4-D][1,3]DIOXOL-2-YL)METHYLPHOSPHONIC ACID",
            "non-polymer",
        ),
        "TDP": ("THIAMIN DIPHOSPHATE", "non-polymer"),
        "L87": (
            "8-BUTYL-1-METHYL-7-(2-METHYLPHENYL)-1H-IMIDAZO[2,1-F]PURINE-2,4(3H,8H)-DIONE",
            "non-polymer",
        ),
        "L86": (
            "(11S)-11-BENZYL-6-CHLORO-1,2,10,11,12,13,14,15,16,17,18,19-DODECAHYDRO-5,9-METHANO-2,5,8,10,13,17-BENZOHEXAAZACYCLOHENICOSINE-3,24-DIONE",
            "non-polymer",
        ),
        "L85": ("4-[(E)-(4-HYDROXYPHENYL)DIAZENYL]BENZENESULFONIC ACID", "non-polymer"),
        "JUG": ("5-HYDROXYNAPHTHALENE-1,4-DIONE", "non-polymer"),
        "L81": (
            "1,3-DIHYDROISOINDOL-2-YL-(6-HYDROXY-3,3-DIMETHYL-1,2-DIHYDROINDOL-5-YL)METHANONE",
            "non-polymer",
        ),
        "S6I": ("4-[(CYCLOPENTYLCARBAMOYL)AMINO]BENZENESULFONAMIDE", "non-polymer"),
        "S6G": ("6-THIO-2'-DEOXYGUANOSINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "P1G": (
            "((2R,4R,6R,6AS)-4-(2-AMINO-6-OXO-1,6-DIHYDROPURIN-9-YL)-6-(HYDROXYMETHYL)-TETRAHYDROFURO[3,4-D][1,3]DIOXOL-2-YL)METHYLPHOSPHONIC ACID",
            "non-polymer",
        ),
        "P1H": (
            "2-PROPYL-ISONICOTINIC-ACYL-NICOTINAMIDE-ADENINE DINUCLEOTIDE",
            "non-polymer",
        ),
        "0H9": ("2-[(4-METHOXYBENZYL)AMINO]ETHANESULFONIC ACID", "non-polymer"),
        "0H8": (
            "N-(3-CARBOXYPROPANOYL)-L-VALYL-N-[(1R)-5-AMINO-1-PHOSPHONOPENTYL]-L-PROLINAMIDE",
            "peptide",
        ),
        "TH6": ("4-HYDROXY-L-THREONINE", "peptide"),
        "MM1": ("1,4,8,11-TETRAAZA-CYCLOTETRADECANE CU(II)", "non-polymer"),
        "0H3": (
            "7-(6-METHOXYPYRIDIN-3-YL)-4-{[2-(PROPAN-2-YLOXY)ETHYL]AMINO}-1-(2-PROPOXYETHYL)PYRIDO[4,3-D]PYRIMIDIN-2(1H)-ONE",
            "non-polymer",
        ),
        "0H2": (
            "[1,3-DIOXO-6-(PYRIDIN-2-YL-KAPPAN)-2,3-DIHYDRO-1H-ISOINDOL-5-YL-KAPPAC~5~][(THIOXOMETHYLIDENE)AZANIDO-KAPPAN](1,4,7-TRITHIONANE-KAPPA~3~S~1~,S~4~,S~7~)RUTHENIUM",
            "non-polymer",
        ),
        "0H6": (
            "6-({(2S)-3-CYCLOPENTYL-2-[4-(TRIFLUOROMETHYL)-1H-IMIDAZOL-1-YL]PROPANOYL}AMINO)PYRIDINE-3-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "0H5": (
            "6-METHOXY-N-(1-METHYL-1H-PYRAZOL-3-YL)QUINAZOLIN-4-AMINE",
            "non-polymer",
        ),
        "0H4": (
            "(2S)-3-CYCLOPENTYL-N-(5-METHYLPYRIDIN-2-YL)-2-[2-OXO-4-(TRIFLUOROMETHYL)PYRIDIN-1(2H)-YL]PROPANAMIDE",
            "non-polymer",
        ),
        "0HK": (
            "(1R,2R,4S,5S,7S)-7-{[HYDROXY(DITHIOPHEN-2-YL)ACETYL]OXY}-9,9-DIMETHYL-3-OXA-9-AZONIATRICYCLO[3.3.1.0~2,4~]NONANE",
            "non-polymer",
        ),
        "L8V": (
            "2-({4-[(2-AMINOPYRIDIN-4-YL)OXY]-3-FLUOROPHENYL}AMINO)-N-(2,4-DIFLUOROPHENYL)PYRIDINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "0HH": ("L-GAMMA-GLUTAMYL-S-NONYL-L-CYSTEINYLGLYCINE", "peptide"),
        "0HO": ("2,1,3-BENZOTHIADIAZOLE-5-CARBOXYLIC ACID", "non-polymer"),
        "0HN": ("1,3-BENZODIOXOLE-5-CARBOXYLIC ACID", "non-polymer"),
        "0HM": (
            "2-{[4-(AMINOMETHYL)PHENYL]CARBAMOYL}-1-[(1-BENZYL-1H-IMIDAZOL-2-YL)METHYL]-3-HYDROXYPYRIDINIUM",
            "non-polymer",
        ),
        "0HL": (
            "4-{[(3R)-3-{[(7-METHOXYNAPHTHALEN-2-YL)SULFONYL](THIOPHEN-3-YLMETHYL)AMINO}-2-OXOPYRROLIDIN-1-YL]METHYL}THIOPHENE-2-CARBOXIMIDAMIDE",
            "non-polymer",
        ),
        "P1L": ("S-PALMITOYL-L-CYSTEINE", "peptide"),
        "0HG": (
            "N-[(4S)-4-AMMONIO-4-CARBOXYBUTANOYL]-S-(4-BROMOBENZYL)-L-CYSTEINYLGLYCINE",
            "peptide",
        ),
        "0HD": ("1-(1H-INDAZOL-5-YL)-3-(2-PHENYLETHYL)UREA", "non-polymer"),
        "JU2": (
            "2-AMINO-6-BROMO-4-OXO-4,7-DIHYDRO-3H-PYRROLO[2,3-D]PYRIMIDINE-5-CARBONITRILE",
            "non-polymer",
        ),
        "0HZ": (
            "AMINO({3-[(3S,8AS)-1,4-DIOXOOCTAHYDROPYRROLO[1,2-A]PYRAZIN-3-YL]PROPYL}AMINO)METHANIMINIUM",
            "peptide",
        ),
        "0HY": (
            "METHYL (3R)-3-(7-AMINO-4,5-DIOXO-1,4,5,6-TETRAHYDROPYRIMIDO[4,5-C]PYRIDAZIN-3-YL)BUTANOATE",
            "non-polymer",
        ),
        "0HX": (
            "PENTAN-3-YL 2-(ACETYLAMINO)-2,4-DIDEOXY-ALPHA-L-THREO-HEX-4-ENOPYRANOSIDURONIC ACID",
            "non-polymer",
        ),
        "MM0": (
            "N,N'-(DIBENZO[B,J][1,7]PHENANTHROLINE-2,10-DIYLDIMETHANEDIYL)DIPROPAN-1-AMINE",
            "non-polymer",
        ),
        "L8B": (
            "(4R)-4-HYDROXY-1-[(3-METHYLISOXAZOL-5-YL)ACETYL]-N-[4-(1,3-OXAZOL-5-YL)BENZYL]-L-PROLINAMIDE",
            "non-polymer",
        ),
        "S69": (
            "N'-[(E)-(2,4-DIHYDROXY-6-METHYLPHENYL)METHYLIDENE]-2-(3-METHOXYPHENYL)ACETOHYDRAZIDE",
            "non-polymer",
        ),
        "DIQ": ("2-METHYL-DECAHYDRO-ISOQUINOLINE-3-CARBOXYLIC ACID", "non-polymer"),
        "0HQ": ("DIAZOMETHANE", "non-polymer"),
        "0HW": (
            "N-[(3S)-4-ETHOXY-3-HYDROXY-4-OXOBUTANOYL]-L-ISOLEUCYL-L-PROLINE",
            "peptide",
        ),
        "0HV": ("3-[(4-NITRONAPHTHALEN-1-YL)AMINO]BENZOIC ACID", "non-polymer"),
        "L8I": (
            "2'-(2-FLUOROPHENYL)-1-METHYL-6',8',9',11'-TETRAHYDROSPIRO[AZETIDINE-3,10'-PYRIDO[3',4':4,5]PYRROLO[2,3-F]ISOQUINOLIN]-7'(5'H)-ONE",
            "non-polymer",
        ),
        "0HT": (
            "METHYL N-[(2S,3S)-3-AMINO-2-HYDROXY-5-METHYLHEXYL]-L-VALYL-L-ISOLEUCYL-L-PHENYLALANINATE",
            "peptide",
        ),
        "DSN": ("D-SERINE", "peptide"),
        "TEI": (
            "2-(3-CYANO-4-ISOBUTOXY-PHENYL)-4-METHYL-5-THIAZOLE-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "NX2": (
            "3-[(3,4-DICHLOROBENZYL)SULFANYL]-N-(METHYLSULFONYL)THIOPHENE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "P1R": ("PYRIMIDINE", "non-polymer"),
        "7SP": (
            "(2R)-2-AMINO-N'-[(E)-(4-HYDROXYNAPHTHALEN-1-YL)METHYLIDENE]-2-PHENYLETHANEHYDRAZIDE",
            "non-polymer",
        ),
        "AR0": (
            "4-AMINO-1-{[AMINO(IMINIO)METHYL]AMINO}-6-[4-(CARBOXYMETHYL)PYRIDINIUM-1-YL]-1,2,3,4,6-PENTADEOXY-D-ERYTHRO-HEXITOL",
            "peptide",
        ),
        "DSM": (
            "3-(10,11-DIHYDRO-5H-DIBENZO[B,F]AZEPIN-5-YL)-N-METHYLPROPAN-1-AMINE",
            "non-polymer",
        ),
        "TDU": (
            "1-(1,2,3-BENZOTHIADIAZOL-6-YL)-3-[2-(CYCLOHEX-1-EN-1-YL)ETHYL]UREA",
            "non-polymer",
        ),
        "A2P": ("ADENOSINE-2'-5'-DIPHOSPHATE", "non-polymer"),
        "GWD": (
            "(3E)-3-[(3,5-DIBROMO-4-HYDROXYPHENYL)METHYLIDENE]-5-IODO-1,3-DIHYDRO-2H-INDOL-2-ONE",
            "non-polymer",
        ),
        "A2R": (
            "[(2R,3R,4R,5R)-5-(6-AMINO-9H-PURIN-9-YL)-3-HYDROXY-4-(PHOSPHONOOXY)TETRAHYDROFURAN-2-YL]METHYL [(2R,3S,4R,5R)-3,4,5-TRIHYDROXYTETRAHYDROFURAN-2-YL]METHYL DIHYDROGEN DIPHOSPHATE",
            "non-polymer",
        ),
        "TMS": (
            "N-PROPYL-4-[(E)-2-(3,4,5-TRIMETHOXY-PHENYL)-VINYL]-BENZAMIDE",
            "non-polymer",
        ),
        "A2T": (
            "N-[1-(3-METHYLBUTYL)PIPERIDIN-4-YL]-N-{4-[METHYL(PYRIDIN-4-YL)AMINO]BENZYL}-4-PENTYLBENZAMIDE",
            "non-polymer",
        ),
        "TMU": ("N-(4-METHOXYBENZYL)-N'-(5-NITRO-1,3-THIAZOL-2-YL)UREA", "non-polymer"),
        "GWB": (
            "4-[(CYCLOPROPYLETHYNYL)OXY]-6-FLUORO-3-ISOPROPYLQUINOLIN-2(1H)-ONE",
            "non-polymer",
        ),
        "TMX": ("2-DEOXY-2-(TRIMETHYLAMMONIO)-BETA-D-GLUCOPYRANOSE", "saccharide"),
        "A2Y": (
            "(3S,5S,7S)-N-{[5-(THIOPHEN-2-YL)-1,2-OXAZOL-3-YL]METHYL}TRICYCLO[3.3.1.1~3,7~]DECAN-1-AMINIUM",
            "non-polymer",
        ),
        "A2Z": ("6,9-DIMETHYL[1,3]DIOXOLO[4,5-H]QUINOLIN-8(9H)-ONE", "non-polymer"),
        "GWI": (
            "N-({4-[({4-CHLORO-2-[(3-CHLORO-5-CYANOPHENYL)CARBONYL]PHENOXY}ACETYL)AMINO]-3-METHYLPHENYL}SULFONYL)PROPANAMIDE",
            "non-polymer",
        ),
        "WBT": (
            "3-FLUORO-5-MORPHOLIN-4-YL-N-[1-(2-PYRIDIN-4-YLETHYL)-1H-INDOL-6-YL]BENZAMIDE",
            "non-polymer",
        ),
        "GWJ": (
            "2-{4-CHLORO-2-[(3-CHLORO-5-CYANOPHENYL)CARBONYL]PHENOXY}-N-(2-METHYL-4-SULFAMOYLPHENYL)ACETAMIDE",
            "non-polymer",
        ),
        "AR1": (
            "(2-{[4-({4-[(4-FORMYLAMINO-1-METHYL-1H-IMIDAZOLE-2-CARBONYL)-AMINO]-1-METHYL-1H-IMIDAZOLE-2-CARBONYL}-AMINO)-1-METHYL-1H-IMIDAZOLE-2-CARBONYL]-AMINO}-ETHYL)-DIMETHYL-AMMONIUM",
            "non-polymer",
        ),
        "TMA": ("TETRAMETHYLAMMONIUM ION", "non-polymer"),
        "TMB": ("N-METHYL-4-[(E)-2-BUTENYL]-4,N-DIMETHYL-THREONINE", "peptide"),
        "A2C": (
            "3,3'-[ETHANE-1,2-DIYLBIS(NITRILOMETHYLYLIDENE)]BIS(4-HYDROXYBENZENECARBOXIMIDAMIDE)",
            "non-polymer",
        ),
        "A2D": (
            "BIS{[(2R,3S,4R,5R)-5-(6-AMINO-9H-PURIN-9-YL)-3,4-DIHYDROXYTETRAHYDROFURAN-2-YL]METHYL} DIHYDROGEN DIPHOSPHATE",
            "non-polymer",
        ),
        "A2E": (
            "(5S)-5-{[10-(1,2,3,4-TETRAHYDROACRIDIN-9-YLAMINO)DECYL]AMINO}-5,6,7,8-TETRAHYDROQUINOLIN-2(1H)-ONE",
            "non-polymer",
        ),
        "A2F": ("2-FLUOROADENINE", "non-polymer"),
        "A2G": ("N-ACETYL-2-DEOXY-2-AMINO-GALACTOSE", "saccharide"),
        "TMH": ("(+)-3,6,6-TRIMETHYLBICYCLO[3.1.1]HEPT-2-ENE", "non-polymer"),
        "TMI": ("1-[PHENYL-(4-PHENYLPHENYL)-METHYL]IMIDAZOLE", "non-polymer"),
        "TMJ": (
            "N-(4-CHLOROPHENYL)-1,2,3,4-TETRAHYDROISOQUINOLINE-7-SULFONAMIDE",
            "non-polymer",
        ),
        "A2K": (
            "(8S,11R,13S,14S,16S,17S)-17-CYCLOPROPYLCARBONYL-16-ETHENYL-13-METHYL-11-(4-PYRIDIN-3-YLPHENYL)-2,6,7,8,11,12,14,15,16,17-DECAHYDRO-1H-CYCLOPENTA[A]PHENANTHREN-3-ONE",
            "non-polymer",
        ),
        "A2L": ("?", "nucleic-acid"),
        "A2M": ("?", "nucleic-acid"),
        "A2N": (
            "[(2R,3S,4R,5R)-4-(ACETYLAMINO)-3,5-DIHYDROXYTETRAHYDROFURAN-2-YL]METHYL [(2R,3S,4R,5R)-5-(6-AMINO-9H-PURIN-9-YL)-3,4-DIHYDROXYTETRAHYDROFURAN-2-YL]METHYL DIHYDROGEN DIPHOSPHATE",
            "non-polymer",
        ),
        "TMO": ("TRIMETHYLAMINE OXIDE", "non-polymer"),
        "DSB": ("4,7-DIOXOSEBACIC ACID", "non-polymer"),
        "HWD": (
            "2-AMINO-3-(2,4-DIOXO-3,4-DIHYDRO-2H-PYRIMIDIN-1-YL)-PROPIONIC ACID",
            "non-polymer",
        ),
        "TH2": ("TESTOSTERONE HEMISUCCINATE", "non-polymer"),
        "NJQ": (
            "3-BENZYL-4-({[6-(CYCLOHEXYLMETHYL)PYRIDIN-2-YL]CARBONYL}AMINO)BENZOIC ACID",
            "non-polymer",
        ),
        "NJP": ("5-HYDROXY-NICOTINAMIDE-ADENINE-DINUCLEOTIDE PHOSPHATE", "non-polymer"),
        "P3A": ("PHOSPHATIDYLGLYCEROL-PHOSPHOGLYCEROL", "non-polymer"),
        "AR3": ("CYTARABINE", "non-polymer"),
        "U3S": (
            "1-(5-DEOXY-5-PIPERIDIN-1-YL-ALPHA-L-ARABINOFURANOSYL)PYRIMIDINE-2,4(1H,3H)-DIONE",
            "non-polymer",
        ),
        "N": ("1-DEOXY-RIBOFURANOSE-5'-PHOSPHATE", "nucleic-acid"),
        "HW8": (
            "4-METHYL-6-{[(3R,4R)-4-{[5-(PYRIDIN-2-YL)PENTYL]OXY}PYRROLIDIN-3-YL]METHYL}PYRIDIN-2-AMINE",
            "non-polymer",
        ),
        "HW9": (
            "4-METHYL-6-{[(3R,4R)-4-{[5-(4-METHYLPYRIDIN-2-YL)PENTYL]OXY}PYRROLIDIN-3-YL]METHYL}PYRIDIN-2-AMINE",
            "non-polymer",
        ),
        "HW6": (
            "6-{[(3R,4R)-4-{[5-(3-CHLORO-5-FLUOROPHENYL)PENTYL]OXY}PYRROLIDIN-3-YL]METHYL}-4-METHYLPYRIDIN-2-AMINE",
            "non-polymer",
        ),
        "HW7": (
            "6-({(3R,4R)-4-[4-(3-CHLORO-5-FLUOROPHENOXY)BUTOXY]PYRROLIDIN-3-YL}METHYL)-4-METHYLPYRIDIN-2-AMINE",
            "non-polymer",
        ),
        "HW4": (
            "6-{[(3S,4S)-4-{[5-(3-FLUOROPHENYL)PENTYL]OXY}PYRROLIDIN-3-YL]METHYL}-4-METHYLPYRIDIN-2-AMINE",
            "non-polymer",
        ),
        "HW5": (
            "6-{[(3S,4S)-4-{[(4E)-5-(3-FLUOROPHENYL)PENT-4-EN-1-YL]OXY}PYRROLIDIN-3-YL]METHYL}-4-METHYLPYRIDIN-2-AMINE",
            "non-polymer",
        ),
        "HW0": (
            "6-{[(3R,4R)-4-{[5-(6-AMINOPYRIDIN-2-YL)PENTYL]OXY}PYRROLIDIN-3-YL]METHYL}-4-METHYLPYRIDIN-2-AMINE",
            "non-polymer",
        ),
        "HW1": (
            "6-[5-({(3R,4R)-4-[(6-AMINO-4-METHYLPYRIDIN-2-YL)METHYL]PYRROLIDIN-3-YL}OXY)PENTYL]-4-METHYLPYRIDIN-2-AMINE",
            "non-polymer",
        ),
        "A20": ("3-AMINOPROPANE-1-SULFONIC ACID", "non-polymer"),
        "A21": (
            "(5S)-2-{[(1S)-1-(2-FLUOROPHENYL)ETHYL]AMINO}-5-METHYL-5-(TRIFLUOROMETHYL)-1,3-THIAZOL-4(5H)-ONE",
            "non-polymer",
        ),
        "A22": (
            "[(2R,3S,4R,5R)-5-(6-AMINO-9H-PURIN-9-YL)-3,4-DIHYDROXYTETRAHYDROFURAN-2-YL]METHYL [(2R,3R,4R,5R)-5-(6-AMINO-9H-PURIN-9-YL)-3-HYDROXY-4-(PHOSPHONOOXY)TETRAHYDROFURAN-2-YL]METHYL DIHYDROGEN DIPHOSPHATE",
            "non-polymer",
        ),
        "A23": ("?", "nucleic-acid"),
        "A24": (
            "(3-NITRO-5-(2-MORPHOLIN-4-YL-ETHYLAMINOCARBONYL)PHENYL)-GALACTOPYRANOSIDE",
            "non-polymer",
        ),
        "A25": (
            "18-CHLORO-11,12,13,14-TETRAHYDRO-1H,10H-8,4-(AZENO)-9,15,1,3,6-BENZODIOXATRIAZACYCLOHEPTADECIN-2-ONE",
            "non-polymer",
        ),
        "A26": (
            "(2Z)-2-CYANO-3-HYDROXY-N-[4-(TRIFLUOROMETHYL)PHENYL]BUT-2-ENAMIDE",
            "non-polymer",
        ),
        "A27": (
            "N-[6-(4-HYDROXYPHENYL)-5-PHENYL-1H-INDAZOL-3-YL]BUTANAMIDE",
            "non-polymer",
        ),
        "A28": ("N-(6,7-DIFLUORO-5-PHENYL-1H-INDAZOL-3-YL)BUTANAMIDE", "non-polymer"),
        "A29": ("2-(3-HYDROXYPHENYL)-2-OXO-ETHANOIC ACID", "non-polymer"),
        "9DN": ("DANSYL-L-ASPARAGINE", "peptide"),
        "GW5": (
            "(2E)-3-{4-[(1E)-1,2-DIPHENYLBUT-1-ENYL]PHENYL}ACRYLIC ACID",
            "non-polymer",
        ),
        "GW4": (
            "N-[(2S)-2-[(2-BENZOYLPHENYL)AMINO]-3-{4-[2-(5-METHYL-2-PHENYL-1,3-OXAZOL-4-YL)ETHOXY]PHENYL}PROPYL]ACETAMIDE",
            "non-polymer",
        ),
        "GW7": (
            "N-{3-CHLORO-4-[(3-FLUOROBENZYL)OXY]PHENYL}-6-ETHYLTHIENO[3,2-D]PYRIMIDIN-4-AMINE",
            "non-polymer",
        ),
        "GW6": (
            "(6ALPHA,11ALPHA,14BETA,16ALPHA,17ALPHA)-6,9-DIFLUORO-17-{[(FLUOROMETHYL)SULFANYL]CARBONYL}-11-HYDROXY-16-METHYL-3-OXOANDROSTA-1,4-DIEN-17-YL FURAN-2-CARBOXYLATE",
            "non-polymer",
        ),
        "VU2": ("(2S)-2-[(4-CHLOROBENZYL)OXY]-2-PHENYLETHANAMINE", "non-polymer"),
        "GW0": (
            "{4-[({2-[3-FLUORO-4-(TRIFLUOROMETHYL)PHENYL]-4-METHYL-1,3-THIAZOL-5-YL}METHYL)SULFANYL]-2-METHYLPHENOXY}ACETIC ACID",
            "non-polymer",
        ),
        "GW3": (
            "7-(1-ETHYL-PROPYL)-7H-PYRROLO-[3,2-F]QUINAZOLINE-1,3-DIAMINE",
            "non-polymer",
        ),
        "GW9": ("2-CHLORO-5-NITRO-N-PHENYLBENZAMIDE", "non-polymer"),
        "GW8": (
            "1-(4-{METHYL[2-({4-[(METHYLSULFONYL)METHYL]PHENYL}AMINO)PYRIMIDIN-4-YL]AMINO}PHENYL)-3-{3-[(4-METHYLPIPERAZIN-1-YL)CARBONYL]PHENYL}UREA",
            "non-polymer",
        ),
        "EIP": ("4-HYDROXY-3-METHYL BUTYL DIPHOSPHATE", "non-polymer"),
        "EIT": (
            "((3R,4R,5R)-4-(2-(1H-IMIDAZOL-1-YL)ETHOXY)-3-HYDROXY-5-(5-METHYL-2,4-DIOXO-3,4-DIHYDROPYRIMIDIN-1(2H)-YL)-TETRAHYDROFURAN-2-YL)METHYL DIHYDROGEN PHOSPHATE",
            "nucleic-acid",
        ),
        "RCP": (
            "(3R)-1'-(9-ANTHRYLCARBONYL)-3-(MORPHOLIN-4-YLCARBONYL)-1,4'-BIPIPERIDINE",
            "non-polymer",
        ),
        "XAZ": ("5,7-DIMETHYLPYRAZOLO[1,5-A]PYRIMIDIN-2-AMINE", "non-polymer"),
        "RCS": (
            "TRISCARBONYL-CYCLOPENTADIENYL-CARBOXY-4-AMINOMETHYLBENZENE-SULFONAMIDE RHENIUM(I)",
            "non-polymer",
        ),
        "RCL": ("RICINOLEIC ACID", "non-polymer"),
        "RCM": (
            "(5R)-5-[(7-CHLORO-1H-INDOL-3-YL)METHYL]-3-METHYLIMIDAZOLIDINE-2,4-DIONE",
            "non-polymer",
        ),
        "EIC": ("LINOLEIC ACID", "non-polymer"),
        "RCO": ("RESORCINOL", "non-polymer"),
        "X65": ("5-NITRO-2-[(4-SULFAMOYLBENZYL)AMINO]BENZAMIDE", "non-polymer"),
        "RCE": (
            "1-{2,5-ANHYDRO-6-DEOXY-4-[(PHOSPHONOOXY)METHYL]-BETA-D-GULOFURANOSYL}PYRIMIDINE-2,4(1H,3H)-DIONE",
            "nucleic-acid",
        ),
        "EIM": (
            "(1R,3S,5Z)-5-[(2E)-2-[(1R,3S,3AS,7AR)-1-[(2R)-6-HYDROXY-6-METHYL-HEPTAN-2-YL]-3-METHOXY-7A-METHYL-2,3,3A,5,6,7-HEXAHYDRO-1H-INDEN-4-YLIDENE]ETHYLIDENE]-4-METHYLIDENE-CYCLOHEXANE-1,3-DIOL",
            "non-polymer",
        ),
        "RCA": (
            "6-{[4-(HYDROXYMETHYL)-5-METHYL-2,6-DIOXOHEXAHYDROPYRIMIDIN-5-YL]METHYL}-5-METHYLPYRIMIDINE-2,4(1H,3H)-DIONE",
            "non-polymer",
        ),
        "PP8": (
            "2-[(1R)-1-(N-(3-METHYLBUTANOYL)-L-VALYL-L-ASPARAGINYL)-AMINO)-3-METHYLBUTYL]HYDROXYPHOSPHINYLOXY]-3-PHENYLPROPANOIC ACID METHYLESTER",
            "non-polymer",
        ),
        "EIN": ("?", "non-polymer"),
        "00C": ("3-SULFO-D-ALANINE", "peptide"),
        "00B": ("(4R)-5,5-DIMETHYL-1,3-THIAZOLIDINE-4-CARBOXYLIC ACID", "peptide"),
        "00A": (
            "5'-O-[(S)-{[(4-CHLOROPHENYL)CARBONYL]OXY}(HYDROXY)PHOSPHORYL]ADENOSINE",
            "non-polymer",
        ),
        "FAY": ("8-FORMYL-FLAVIN-ADENINE DINUCLEOTIDE", "non-polymer"),
        "00G": ("5-[2-(1H-PYRROL-1-YL)ETHOXY]-1H-INDOLE", "non-polymer"),
        "00F": (
            "(2S)-3-[(R)-[(1R)-1-AMINO-2-PHENYLETHYL](HYDROXY)PHOSPHORYL]-2-METHYLPROPANOIC ACID",
            "peptide",
        ),
        "00E": ("MORPHOLIN-4-YLACETIC ACID", "peptide"),
        "00D": ("(2R)-3,3-DIMETHYLBUTAN-2-AMINE", "non-polymer"),
        "00K": (
            "(1S,7S)-7-AMINO-N-[(2R,3S)-7-AMINO-1-(CYCLOHEXYLAMINO)-2-HYDROXY-1-OXOHEPTAN-3-YL]-7-BENZYL-8-OXOHEXAHYDRO-1H-PYRAZOLO[1,2-A]PYRIDAZINE-1-CARBOXAMIDE",
            "peptide",
        ),
        "00J": (
            "N,2-DIMETHYL-6-{[7-(2-MORPHOLIN-4-YLETHOXY)QUINOLIN-4-YL]OXY}-1-BENZOFURAN-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "00I": (
            "N-[(1R)-1-(4-CARBAMIMIDOYLBENZYL)-2-OXO-2-PIPERIDIN-1-YLETHYL]-N~2~-(NAPHTHALEN-2-YLSULFONYL)-L-ALPHA-GLUTAMINE",
            "peptide",
        ),
        "00H": (
            "3-[(R)-[(1R)-1-AMINO-2-PHENYLETHYL](HYDROXY)PHOSPHORYL]PROPANOIC ACID",
            "peptide",
        ),
        "00O": ("BETA-PHENYL-D-PHENYLALANINE", "peptide"),
        "00N": (
            "(1S,7S)-7-AMINO-7-BENZYL-N-{(1S)-4-CARBAMIMIDAMIDO-1-[(S)-HYDROXY(1,3-THIAZOL-2-YL)METHYL]BUTYL}-8-OXOHEXAHYDRO-1H-PYRAZOLO[1,2-A]PYRIDAZINE-1-CARBOXAMIDE",
            "peptide",
        ),
        "00M": ("(2R,3S)-3-AMINO-2-HYDROXY-4-PHENYLBUTANOIC ACID", "peptide"),
        "00L": (
            "(1S,7S)-7-AMINO-7-BENZYL-N-[(1S)-4-CARBAMIMIDAMIDO-1-{(1S)-1-HYDROXY-2-OXO-2-[(2-PHENYLETHYL)AMINO]ETHYL}BUTYL]-8-OXOHEXAHYDRO-1H-PYRAZOLO[1,2-A]PYRIDAZINE-1-CARBOXAMIDE",
            "peptide",
        ),
        "00S": ("4-(AMINOMETHYL)BENZENECARBOXIMIDAMIDE", "non-polymer"),
        "00R": (
            "(5S)-N-[TRANS-4-(2-AMINO-1H-IMIDAZOL-5-YL)CYCLOHEXYL]-1,3-DIOXO-2-[2-(PHENYLSULFONYL)ETHYL]-2,3,5,8-TETRAHYDRO-1H-[1,2,4]TRIAZOLO[1,2-A]PYRIDAZINE-5-CARBOXAMIDE",
            "peptide",
        ),
        "00Q": (
            "D-PHENYLALANYL-N-{(1S)-1-[(S)-1,3-BENZOTHIAZOL-2-YL(HYDROXY)METHYL]-4-CARBAMIMIDAMIDOBUTYL}-L-PROLINAMIDE",
            "peptide",
        ),
        "00P": (
            "(5S)-N-[(TRANS-4-AMINOCYCLOHEXYL)METHYL]-1,3-DIOXO-2-[2-(PHENYLSULFONYL)ETHYL]-2,3,5,8-TETRAHYDRO-1H-[1,2,4]TRIAZOLO[1,2-A]PYRIDAZINE-5-CARBOXAMIDE",
            "peptide",
        ),
        "00W": (
            "3-AMINO-6-CARBAMIMIDAMIDO-3,4,5,6-TETRADEOXY-L-ERYTHRO-HEXONIC ACID",
            "peptide",
        ),
        "00V": ("(2S,3S)-3,7-DIAMINO-2-HYDROXYHEPTANOIC ACID", "peptide"),
        "00U": (
            "(1S,7S)-7-AMINO-7-BENZYL-8-OXOHEXAHYDRO-1H-PYRAZOLO[1,2-A]PYRIDAZINE-1-CARBOXYLIC ACID",
            "peptide",
        ),
        "00T": ("(4-CHLOROBENZENE-1,2-DIYL)DIMETHANAMINE", "non-polymer"),
        "FAB": ("FLAVIN-ADENINE DINUCLEOTIDE-N5-ISOBUTYL KETONE", "non-polymer"),
        "FAC": ("1,1,1,3,3,3-HEXAFLUOROPROPANEDIOL", "non-polymer"),
        "00Y": (
            "1-[(4S,5S)-4-AMINO-5-HYDROXY-5-(1,3-THIAZOL-2-YL)PENTYL]GUANIDINE",
            "peptide",
        ),
        "00X": ("(1S,2R)-1-AMINO-2,3-DIHYDRO-1H-INDEN-2-OL", "non-polymer"),
        "FAF": (
            "2-ACETYLAMINO-4-METHYL-PENTANOIC ACID (1-FORMYL-2-PHENYL-ETHYL)-AMIDE",
            "non-polymer",
        ),
        "FAG": ("?", "nucleic-acid"),
        "FAD": ("FLAVIN-ADENINE DINUCLEOTIDE", "non-polymer"),
        "FAE": ("FLAVIN-N7 PROTONATED-ADENINE DINUCLEOTIDE", "non-polymer"),
        "FA8": ("?", "non-polymer"),
        "FA9": ("?", "non-polymer"),
        "FA2": (
            "5-(6-AMINO-9H-PURIN-9-YL)-4-HYDROXYTETRAHYDROFURAN-3-YL DIHYDROGEN PHOSPHATE",
            "nucleic-acid",
        ),
        "FA3": ("2-ANHYDRO-3-FLUORO-QUINIC ACID", "non-polymer"),
        "FA1": ("2,3 -ANHYDRO-QUINIC ACID", "non-polymer"),
        "FA6": ("3-HYDROXYIMINO QUINIC ACID", "non-polymer"),
        "FA7": ("(2Z,4E)-7-METHYLOCTA-2,4-DIENOIC ACID", "non-polymer"),
        "FA4": ("SM-25453", "non-polymer"),
        "FA5": ("ADENOSINE-5'-[PHENYLALANINYL-PHOSPHATE]", "nucleic-acid"),
        "003": (
            "5-METHYL-7-(2-METHYLPROPYL)-2-(NAPHTHALEN-1-YLMETHYL)-3-PYRIDIN-4-YL-2H-PYRAZOLO[3,4-D]PYRIMIDINE-4,6(5H,7H)-DIONE",
            "non-polymer",
        ),
        "002": (
            "N-[(2R)-2-BENZYL-4-(HYDROXYAMINO)-4-OXOBUTANOYL]-L-ISOLEUCYL-L-LEUCINE",
            "non-polymer",
        ),
        "001": (
            "1-[2,2-DIFLUORO-2-(3,4,5-TRIMETHOXY-PHENYL)-ACETYL]-PIPERIDINE-2-CARBOXYLIC ACID 4-PHENYL-1-(3-PYRIDIN-3-YL-PROPYL)-BUTYL ESTER",
            "non-polymer",
        ),
        "000": ("METHYL HYDROGEN CARBONATE", "non-polymer"),
        "007": ("1-METHYLAMINE-1-BENZYL-CYCLOPENTANE", "non-polymer"),
        "006": (
            "(4R)-3-[(2S,3S)-3-{[(2,6-DIMETHYLPHENOXY)ACETYL]AMINO}-2-HYDROXY-4-PHENYLBUTANOYL]-N-[(1S,2R)-2-HYDROXY-2,3-DIHYDRO-1H-INDEN-1-YL]-5,5-DIMETHYL-1,3-THIAZOLIDINE-4-CARBOXAMIDE",
            "peptide",
        ),
        "005": ("(2S,3S)-3-AMINO-2-HYDROXY-4-PHENYLBUTANOIC ACID", "peptide"),
        "004": ("(2S)-AMINO(PHENYL)ETHANOIC ACID", "peptide"),
        "009": (
            "(4S)-1,4-DIBENZYL-N-[(1S,2R)-1-BENZYL-3-{[3-(DIMETHYLAMINO)BENZYL]AMINO}-2-HYDROXYPROPYL]-2-OXOIMIDAZOLIDINE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "008": (
            "(S)-2-[(R)-3-AMINO-4-(2-FLUORO-PHENYL)-BUTYRYL]-1,2,3,4-TETRAHYDRO-ISOQUINOLINE-3-CARBOXYLIC ACID AMIDE",
            "non-polymer",
        ),
        "5B1": ("5-BROMO-1H-INDAZOL-3-AMINE", "non-polymer"),
        "EI1": ("3-ETHYL-2-(4-HYDROXYPHENYL)-2H-INDAZOL-5-OL", "non-polymer"),
        "XAH": (
            "5'-O-{(R)-HYDROXY[(L-LYSYLAMINO)OXY]PHOSPHORYL}ADENOSINE",
            "non-polymer",
        ),
        "RC8": (
            "(2R)-2-({9-(1-METHYLETHYL)-6-[(4-PYRIDIN-2-YLBENZYL)AMINO]-9H-PURIN-2-YL}AMINO)BUTAN-1-OL",
            "non-polymer",
        ),
        "DRX": (
            "[(1S)-1-(5-CHLORO-1-BENZOTHIEN-3-YL)-2-(2-NAPHTHYLAMINO)-2-OXOETHYL]PHOSPHONIC ACID",
            "non-polymer",
        ),
        "RC7": (
            "{(2R)-4-(4-HYDROXYBENZYL)-2-[2-(1H-IMIDAZOL-4-YL)ETHYL]-5-OXO-2,5-DIHYDRO-1H-IMIDAZOL-1-YL}ACETALDEHYDE",
            "peptide",
        ),
        "RC1": (
            "[3R-[3A,4A,5B(R*)]]-5-(1-CARBOXY-1-PHOSPHONOETHOXY)-4-HYDROXY-3-(PHOSPHONOOXY)-1-CYCLOHEXENE-1-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "RC2": (
            "N-{3-[5-OXO-7-(PIPERAZIN-1-YL)-5H-[1,3,4]THIADIAZOLO[3,2-A]PYRIMIDIN-2-YL]PHENYL}GLYCINAMIDE",
            "non-polymer",
        ),
        "RB6": ("BAL 9141, BOUND FORM; CEFTOBIPROLE, BOUND FORM", "non-polymer"),
        "XAL": (
            "[(1S,4R,6R)-6-HYDROXY-4-(ADENIN-9-YL)CYCLOHEX-2-EN-1-YL]METHYL DIHYDROGEN PHOSPHATE",
            "nucleic-acid",
        ),
        "RB5": ("D-RIBOSE", "saccharide"),
        "OAG": (
            "6-{[(2,5-DICHLOROPHENYL)AMINO]METHYL}PYRIDO[2,3-D]PYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "O8K": (
            "(4Z,6R,7S,8S,9E)-7-HYDROXY-8-METHOXY-4,6-DIMETHYLCYCLOTETRADECA-4,9-DIEN-1-ONE",
            "non-polymer",
        ),
        "768": (
            "(2R)-1-(2,4-DICHLOROPHENOXY)-3-[(2E)-2-IMINO-3-(2-PIPERIDIN-1-YLETHYL)-2,3-DIHYDRO-1H-BENZIMIDAZOL-1-YL]PROPAN-2-OL",
            "non-polymer",
        ),
        "8MG": ("8-METHYL-2'-DEOXYGUANOSINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "CCY": (
            "2-(1-AMINO-2-MERCAPTO-ETHYL)-5-(4-HYDROXY-BENZYL)-3-(ETHANOYL)-3,5-DIHYDRO-IMIDAZOL-4-ONE",
            "peptide",
        ),
        "CCX": (
            "3-CYANO-N-{1-[4-(5-CYANO-1H-INDOL-3-YL)BUTYL]PIPERIDIN-4-YL}-1H-INDOLE-7-CARBOXAMIDE",
            "non-polymer",
        ),
        "OAA": ("OXALOACETATE ION", "non-polymer"),
        "CCZ": ("PHENYL (4-CARBAMIMIDOYLBENZYL)PHOSPHONATE", "non-polymer"),
        "762": (
            "3-{5-[AMINO(IMINIO)METHYL]-6-CHLORO-1H-BENZIMIDAZOL-2-YL}-1,1'-BIPHENYL-2-OLATE",
            "non-polymer",
        ),
        "CCT": (
            "5-(4-CYANOPHENYL)-3-{[(2-METHYLPHENYL)SULFONYL]AMINO}THIOPHENE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "760": ("9-(3-PHENYLMETHYLAMINO)-1,2,3,4-TETRAHYDROACRIDINE", "non-polymer"),
        "761": ("3-(OXALYL-AMINO)-NAPHTHALENE-2-CARBOXYLIC ACID", "non-polymer"),
        "CCQ": ("L-CARNITINYL-COA INNER SALT", "non-polymer"),
        "CCP": ("BUTYLPHOSPHONATE", "non-polymer"),
        "CCS": ("CARBOXYMETHYLATED CYSTEINE", "peptide"),
        "CCR": (
            "[N-[N-(4-METHOXY-2,3,6-TRIMETHYLPHENYLSULFONYL)-L-ASPARTYL]-D-(4-AMIDINO-PHENYLALANYL)]-PIPERIDINE",
            "peptide",
        ),
        "CCM": ("1-METHYL-1-CARBOXY-CYCLOPENTANE", "non-polymer"),
        "CCL": ("N~6~-[(CYCLOPENTYLOXY)CARBONYL]-D-LYSINE", "peptide"),
        "CCO": ("CARBOXYMETHYLTHIO-3-(3-CHLOROPHENYL)-1,2,4-OXADIAZOL", "non-polymer"),
        "CCN": ("ACETONITRILE", "non-polymer"),
        "CCI": ("CALICHEAMICIN GAMMA-1-OLIGOSACCHARIDE", "non-polymer"),
        "CCH": (
            "[7-ETHENYL-12-FORMYL-3,8,13,17-TERTRAMETHYL-21H,23H-PORPHINE-2,18-DIPROPANOATO(2)-N21,N22,N23,N24]IRON",
            "non-polymer",
        ),
        "CCK": (
            "[1-(6-{6-[(1-METHYLETHYL)AMINO]-1H-INDAZOL-1-YL}PYRAZIN-2-YL)-1H-PYRROL-3-YL]ACETIC ACID",
            "non-polymer",
        ),
        "NXL": (
            "(2S,5R)-1-FORMYL-5-[(SULFOOXY)AMINO]PIPERIDINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "CCE": ("2-[(AMINOCARBONYL)OXY]-N,N,N-TRIMETHYLETHANAMINIUM", "non-polymer"),
        "CCD": ("(4R)-4-HYDROXY-N,N,N-TRIMETHYLPENTAN-1-AMINIUM", "non-polymer"),
        "77Z": (
            "2-({(3R)-3-[(3S)-1-(3-METHYLBUTYL)-2,4-DIOXO-1,2,3,4-TETRAHYDROQUINOLIN-3-YL]-1,1-DIOXIDO-3,4-DIHYDRO-2H-1,2,4-BENZOTHIADIAZIN-7-YL}OXY)ACETAMIDE",
            "non-polymer",
        ),
        "CCA": (
            "2-BENZOYLOXY-8-METHYL-8-AZA-BICYCLO[3.2.1]OCTANE-3-CARBOXYLIC ACID METHYL ESTER",
            "non-polymer",
        ),
        "CCC": ("?", "nucleic-acid"),
        "CCB": ("CYCLOBUTYL(CYCLOPENTYL)FORMAMIDE", "non-polymer"),
        "PIH": ("IODOPHENYL", "non-polymer"),
        "PII": (
            "2-[(HYDROXY{[(2R,3R,5S,6R)-2,3,4,5,6-PENTAHYDROXYCYCLOHEXYL]OXY}PHOSPHORYL)OXY]-1-[(PALMITOYLOXY)METHYL]ETHYL HEPTADECANOATE",
            "non-polymer",
        ),
        "LK3": (
            "N-{[6-(PENTYLOXY)NAPHTHALEN-2-YL]SULFONYL}-D-GLUTAMIC ACID",
            "non-polymer",
        ),
        "PIL": (
            "3-(CYCLOPENTYLOXY)-N-(3,5-DICHLOROPYRIDIN-4-YL)-4-METHOXYBENZAMIDE",
            "non-polymer",
        ),
        "PIM": ("4-PHENYL-1H-IMIDAZOLE", "non-polymer"),
        "PIN": ("PIPES; 1,4-PIPERAZINEDIETHANESULFONIC ACID", "non-polymer"),
        "PIO": (
            "[(2R)-2-OCTANOYLOXY-3-[OXIDANYL-[(1R,2R,3S,4R,5R,6S)-2,3,6-TRIS(OXIDANYL)-4,5-DIPHOSPHONOOXY-CYCLOHEXYL]OXY-PHOSPHORYL]OXY-PROPYL] OCTANOATE",
            "non-polymer",
        ),
        "PIA": (
            "[(4Z)-2-[(1S)-1-AMINOETHYL]-4-(4-HYDROXYBENZYLIDENE)-5-OXO-4,5-DIHYDRO-1H-IMIDAZOL-1-YL]ACETIC ACID",
            "peptide",
        ),
        "PIB": (
            "2-(BUTANOYLOXY)-1-{[(HYDROXY{[2,3,4,6-TETRAHYDROXY-5-(PHOSPHONOOXY)CYCLOHEXYL]OXY}PHOSPHORYL)OXY]METHYL}ETHYL BUTANOATE",
            "non-polymer",
        ),
        "PIC": (
            "6-(DIFLUORO-PHOSPHONO-METHYL)-NAPHTHALENE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "PID": ("PERIDININ", "non-polymer"),
        "PIE": ("1,2-DIACYL-SN-GLYCERO-3-PHOSPHOINOSITOL", "non-polymer"),
        "PIF": (
            "(2R)-3-{[(S)-HYDROXY{[(1R,2R,3R,4R,5S,6R)-2,3,5,6-TETRAHYDROXY-4-(PHOSPHONOOXY)CYCLOHEXYL]OXY}PHOSPHORYL]OXY}PROPANE-1,2-DIYL DIOCTANOATE",
            "non-polymer",
        ),
        "PIG": ("2-[2-(2-HYDROXY-ETHOXY)-ETHOXY]-ETHANOL", "non-polymer"),
        "PIX": (
            "4-(3-{[5-(TRIFLUOROMETHYL)PYRIDIN-2-YL]OXY}BENZYL)PIPERIDINE-1-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "PIY": ("2-PHENYL-1H-IMIDAZOLE", "non-polymer"),
        "UCD": (
            "(4S,5R,6R)-6-((2R,3R,4R,5R,6R)-3-ACETAMIDO-2,5-DIHYDROXY-6-(HYDROXYMETHYL)-TETRAHYDRO-2H-PYRAN-4-YLOXY)-4,5-DIHYDROXY-5,6-DIHYDRO-4H-PYRAN-2-CARBOXYLIC ACID",
            "saccharide",
        ),
        "6FA": ("6-HYDROXY-FLAVIN-ADENINE DINUCLEOTIDE", "non-polymer"),
        "PIP": ("PIPERIDINE", "non-polymer"),
        "F5B": ("1,2,3,4,5-PENTAFLUOROBENZENE", "non-polymer"),
        "PIR": (
            "2-(4-AMINO-PHENYL)-5-HYDROXYMETHYL-PYRROLIDINE-3,4-DIOL",
            "non-polymer",
        ),
        "PIS": ("TRIHYDROGEN THIODIPHOSPHATE", "non-polymer"),
        "PIT": ("PICEATANNOL", "non-polymer"),
        "PIU": (
            "1-{2-[4-(AMINOSULFONYL)PHENYL]ETHYL}-2,4,6-TRIMETHYLPYRIDINIUM",
            "non-polymer",
        ),
        "N88": (
            "(1R,4R,5R)-3-(TERT-BUTYLCARBAMOYL)-1,4,5-TRIHYDROXYCYCLOHEX-2-ENE-1-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "PIW": ("1-PHENYL-1H-IMIDAZOLE", "non-polymer"),
        "HMP": ("1-HYDROXYAMINE-2-ISOBUTYLMALONIC ACID", "non-polymer"),
        "JJ3": (
            "(3AS,4R,9BR)-4-(4-HYDROXYPHENYL)-6-(METHOXYMETHYL)-1,2,3,3A,4,9B-HEXAHYDROCYCLOPENTA[C]CHROMEN-8-OL",
            "non-polymer",
        ),
        "77B": (
            "BENZYL N-({(2S,3S)-3-[(BENZYLAMINO)CARBONYL]OXIRAN-2-YL}CARBONYL)-L-ISOLEUCYL-L-PROLINATE",
            "non-polymer",
        ),
        "RH3": ("RHODIUM(III) ION", "non-polymer"),
        "PI8": (
            "N-13-[(10S,13S)-9,12-DIOXO-10-(2-BUTYL)-2-OXA-8,11-DIAZABICYCLO [13.2.2] NONADECA-15,17,18-TRIENE] (2R)-BENZYL-(4S)-HYDROXY-5-AMINOPENTANOIC (1R)-HYDROXY-(2S)-INDANEAMIDE",
            "non-polymer",
        ),
        "F5N": ("2-(IMIDAZO[1,2-A]PYRIDIN-2-YL)-N-PHENYLACETAMIDE", "non-polymer"),
        "UCN": ("7-HYDROXYSTAUROSPORINE", "non-polymer"),
        "PI0": (
            "(2S)-N-[(4-CARBAMIMIDOYLPHENYL)METHYL]-2-[[(2R,3R)-3-METHYL-2-(PHENYLMETHYLSULFONYLAMINO)PENTANOYL]AMINO]PENTANEDIAMIDE",
            "non-polymer",
        ),
        "PI1": ("MACROCYCLIC PEPTIDOMIMETIC INHIBITOR 1", "non-polymer"),
        "PI2": (
            "2-{11-[1-HYDROXY-2-(3-METHYL-BUTYLAMINO)-ETHYL] -6,9-DIOXO-2-OXA-7,10-DIAZA-BICYCLO[11.2.2]HEPTADECA- 1(16),13(17),14-TRIEN-8-YL}-ACETAMIDE",
            "non-polymer",
        ),
        "PI3": (
            "11-[1-HYDROXY-2-(3-METHYL-BUTYLAMINO)-ETHYL]-8-ISOPROPYL-2-OXA-7,10- DIAZA-BICYCLO[11.2.2]HEPTADECA-1(16),13(17),14-TRIENE-6,9-DIONE",
            "non-polymer",
        ),
        "PI4": (
            "1-[2-(8-CARBAMOYLMETHYL-6,9-DIOXO-2-OXA-7,10-DIAZA-BICYCLO[11.2.2]HEPTADECA- 1(16),13(17),14-TRIEN-11-YL)-2-HYDROXY-ETHYL]-PIPERIDINE-2-CARBOXYLIC ACID TERT-BUTYLAMIDE",
            "non-polymer",
        ),
        "PI5": (
            "N-[3-(8-SEC-BUTYL-7,10-DIOXO-2-OXA-6,9-DIAZA-BICYCLO[11.2.2]HEPTADECA-1(16),13(17),14- TRIEN-11-YLAMINO)-2-HYDROXY-1-(4-HYDROXY-BENZYL)-PROPYL]-3-METHYL-2- (2-OXO-PYRROLIDIN-1-YL)-BUTYRAMIDE",
            "non-polymer",
        ),
        "PI6": (
            "[1-BENZYL-3-(8-SEC-BUTYL-7,10-DIOXO-2-OXA-6,9-DIAZA-BICYCLO[11.2.2] HEPTADECA-1(16),13(17),14-TRIEN-11-YLAMINO)-2-HYDROXY-PROPYL]-CARBAMIC ACID TERT-BUTYL ESTER",
            "peptide",
        ),
        "PI7": (
            "N-[3-(8-SEC-BUTYL-7,10-DIOXO-2-OXA-6,9-DIAZA-BICYCLO[11.2.2] HEPTADECA-1(16),13(17),14-TRIEN-11-YAMINO)-2-HYDROXY-1-(4-HYDROXY-BENZYL) -PROPYL]-3-METHYL-2-PROPIONYLAMINO-BUTYRAMIDE",
            "non-polymer",
        ),
        "UBT": (
            "N-[(R)-({[(BENZYLOXY)CARBONYL]AMINO}METHYL)(HYDROXY)PHOSPHORYL]-L-LEUCYLGLYCINE",
            "peptide",
        ),
        "UBU": (
            "N-[(R)-({[(BENZYLOXY)CARBONYL]AMINO}METHYL)(HYDROXY)PHOSPHORYL]-N-ETHYL-L-LEUCINAMIDE",
            "peptide",
        ),
        "UBV": (
            "N-[(S)-({[(BENZYLOXY)CARBONYL]AMINO}METHYL)(HYDROXY)PHOSPHORYL]-L-LEUCYL-L-ISOLEUCINE",
            "peptide",
        ),
        "UBW": (
            "N-[(R)-({[(BENZYLOXY)CARBONYL]AMINO}METHYL)(HYDROXY)PHOSPHORYL]-L-LEUCYL-L-NORVALINE",
            "peptide",
        ),
        "UBP": (
            "3-[3-(2-CARBOXYETHYL)-2,4-DIOXO-3,4-DIHYDROPYRIMIDIN-1(2H)-YL]-L-ALANINE",
            "non-polymer",
        ),
        "CC8": ("(2S)-2-METHYL-3-SULFANYLPROPANOIC ACID", "non-polymer"),
        "UBR": ("?", "nucleic-acid"),
        "UBS": (
            "N-[(S)-({[(BENZYLOXY)CARBONYL]AMINO}METHYL)(HYDROXY)PHOSPHORYL]-L-LEUCYL-L-VALINE",
            "peptide",
        ),
        "CC5": ("BETA-D-ERYTHROFURANOSYL-ADENOSINE", "non-polymer"),
        "CC4": ("2-(SELANYLMETHYL)PROP-2-ENOIC ACID", "non-polymer"),
        "CC7": (
            "4-AMINO-1-[(2S,4AR,6R,7R,7AS)-2,7-DIHYDROXY-2-OXIDOTETRAHYDRO-4H-FURO[3,2-D][1,3,2]DIOXAPHOSPHININ-6-YL]PYRIMIDIN-2(1H)-ONE",
            "non-polymer",
        ),
        "76V": (
            "N-{[(2S,3S)-3-(ETHOXYCARBONYL)OXIRAN-2-YL]CARBONYL}-L-ISOLEUCYL-L-ISOLEUCINE",
            "non-polymer",
        ),
        "CC0": (
            "(4-{2-ACETYLAMINO-2-[1-(3-CARBAMOYL-4-CYCLOHEXYLMETHOXY-PHENYL)-ETHYLCARBAMOYL}-ETHYL}-2-PHOSPHONO-PHENOXY)-ACETIC ACID",
            "non-polymer",
        ),
        "CC3": (
            "N-{3-[(4-{[3-(TRIFLUOROMETHYL)PHENYL]AMINO}PYRIMIDIN-2-YL)AMINO]PHENYL}CYCLOPROPANECARBOXAMIDE",
            "non-polymer",
        ),
        "CC2": (
            "4-HYDROXY-6,7-DIMETHYL-3-(NAPHTHALEN-1-YLMETHYL)-2H-CHROMEN-2-ONE",
            "non-polymer",
        ),
        "UBD": ("?", "nucleic-acid"),
        "UBE": (
            "3-({3-[(2S)-2-AMINO-2-CARBOXYETHYL]-5-METHYL-2,6-DIOXO-3,6-DIHYDROPYRIMIDIN-1(2H)-YL}METHYL)-5-PHENYLTHIOPHENE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "OA5": ("6-(2-PHENOXYETHOXY)-1,3,5-TRIAZINE-2,4-DIAMINE", "non-polymer"),
        "OA4": ("5-METHYL-6-PHENYLQUINAZOLINE-2,4-DIAMINE", "non-polymer"),
        "OA3": ("4-AMINO-7,7-DIMETHYL-7,8-DIHYDROQUINAZOLIN-5(6H)-ONE", "non-polymer"),
        "8MR": (
            "(3R)-4,4-DIFLUORO-3-[(4-METHOXYPHENYL)SULFONYL]BUTANOIC ACID",
            "non-polymer",
        ),
        "OA1": ("(2-AMINO-1,3-OXAZOL-5-YL)-(3-BROMOPHENYL)METHANONE", "non-polymer"),
        "UBC": (
            "(S)-1-(2-AMINO-2-CARBOXYETHYL)-3-(2-CARBOXYBENZYL)PYRIMIDINE-2,4-DIONE",
            "non-polymer",
        ),
        "76A": (
            "3-(4'-HYDROXYBIPHENYL-4-YL)-N-(4-HYDROXYCYCLOHEXYL)-1,4-DIHYDROINDENO[1,2-C]PYRAZOLE-6-CARBOXAMIDE",
            "non-polymer",
        ),
        "UBI": ("?", "nucleic-acid"),
        "F83": (
            "(3R)-N-(4-CHLOROPHENYL)-3-(HYDROXYMETHYL)-1,2,3,4-TETRAHYDROISOQUINOLINE-7-SULFONAMIDE",
            "non-polymer",
        ),
        "TLC": (
            "2-O,3-ETHDIYL-ARABINOFURANOSYL-THYMINE-5'-MONOPHOSPHATE",
            "nucleic-acid",
        ),
        "MWP": ("1-BENZOTHIOPHEN-6-AMINE 1,1-DIOXIDE", "non-polymer"),
        "GJZ": ("(1R,2S)-2-PHENYLCYCLOPROPANAMINE", "non-polymer"),
        "RZ0": ("N-BUTYL-4-SULFAMOYLBENZAMIDE", "non-polymer"),
        "GJW": (
            "7-(PYRROLIDIN-1-YLMETHYL)PYRROLO[1,2-A]QUINOXALIN-4(5H)-ONE",
            "non-polymer",
        ),
        "PU4": ("9-BUTYL-8-(4-METHOXYBENZYL)-9H-PURIN-6-AMINE", "non-polymer"),
        "UC4": (
            "1-METHYL ETHYL 1-CHLORO-5-[[(5,6DIHYDRO-2-METHYL-1,4-OXATHIIN-3-YL)CARBONYL]AMINO]BENZOATE",
            "non-polymer",
        ),
        "HC9": ("(3ALPHA,8ALPHA,22R)-CHOLEST-5-ENE-3,22-DIOL", "non-polymer"),
        "PU5": ("9-BUTYL-8-(3-METHOXYBENZYL)-9H-PURIN-6-AMINE", "non-polymer"),
        "UC3": (
            "1-METHYL ETHYL 2-CHLORO-5-[[[(1-METHYLETHOXY)THIOOXO]METHYL]AMINO]-BENZOATE",
            "non-polymer",
        ),
        "GJB": ("1-(2-PHENYLETHYL)PYRIMIDINE-2,4,6(1H,3H,5H)-TRIONE", "non-polymer"),
        "PU6": ("8-BENZO[1,3]DIOXOL-,5-YLMETHYL-9-BUTYL-9H-", "non-polymer"),
        "F53": (
            "(3R,3AS,6AR)-HEXAHYDROFURO[2,3-B]FURAN-3-YL {(1S,2R)-1-BENZYL-3-[(2-ETHYLBUTYL){[4-(HYDROXYMETHYL)PHENYL]SULFONYL}AMINO]-2-HYDROXYPROPYL}CARBAMATE",
            "non-polymer",
        ),
        "AMI": ("ALLOSAMIZOLINE", "non-polymer"),
        "AMH": ("TRANS-4-AMINOMETHYLCYCLOHEXANE-1-CARBOXYLIC ACID", "non-polymer"),
        "AMJ": ("N~2~-ACETYL-N-BENZYL-O-METHYL-L-SERINAMIDE", "non-polymer"),
        "AMM": (
            "(2S)-2-{[(AMINOMETHYL)(DIHYDROXY)SILYL]METHYL}-4-METHYLPENTANAL",
            "non-polymer",
        ),
        "AML": ("AMYLAMINE", "non-polymer"),
        "AMO": ("ASPARTYL-ADENOSINE-5'-MONOPHOSPHATE", "non-polymer"),
        "AMN": (
            "9-DEOXY-9-AMINO-2-O-METHYL-5-N-ACETYL-ALPHA-D-NEURAMINIC ACID",
            "non-polymer",
        ),
        "AMA": (
            "2-[(3-HYDROXY-2-METHYL-5-PHOSPHONOOXYMETHYL-PYRIDIN-4-YLMETHYL)-AMINO]-2-METHYL-SUCCINIC ACID",
            "non-polymer",
        ),
        "AMC": ("AMINOMETHYLCYCLOHEXANE", "non-polymer"),
        "AMB": ("L-2-AMINO-4-METHOXY-CIS-BUT-3-ENOIC ACID", "non-polymer"),
        "AME": ("N-ACETYLMETHIONINE", "non-polymer"),
        "AMD": ("ASPARTYL-2'-DEOXY-ADENOSINE-5'-MONOPHOSPHATE", "non-polymer"),
        "AMG": ("ALPHA-METHYL-D-GALACTOSIDE", "non-polymer"),
        "AMF": (
            "(3Z,5S,6R,7S,8R,8AS)-3-(OCTYLIMINO)HEXAHYDRO[1,3]THIAZOLO[3,4-A]PYRIDINE-5,6,7,8-TETROL",
            "non-polymer",
        ),
        "AMY": ("ANTIMYCIN", "non-polymer"),
        "AMX": ("AMIDOCARBOXYMETHYLDETHIA COENZYME *A", "non-polymer"),
        "AMZ": ("AMINOIMIDAZOLE 4-CARBOXAMIDE RIBONUCLEOTIDE", "non-polymer"),
        "NWA": ("4,4-DIHYDROXY-N,N,N-TRIMETHYLPENTAN-1-AMINIUM", "non-polymer"),
        "NWB": ("N,N,N-TRIMETHYL-4-OXOPENTAN-1-AMINIUM", "non-polymer"),
        "RBV": (
            "1-(BETA-D-RIBOFURANOSYL)-1H-1,2,4-TRIAZOLE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "AMQ": (
            "(S)-ALPHA-AMINO-3-HYDROXY-5-METHYL-4-ISOXAZOLEPROPIONIC ACID",
            "non-polymer",
        ),
        "AMP": ("ADENOSINE MONOPHOSPHATE", "non-polymer"),
        "AMS": ("3-MERCURI-4-AMINOBENZENESULFONAMIDE", "non-polymer"),
        "AMR": (
            "3,5-DIAMINO-N-(AMINOIMINOMETHYL)-6-CHLOROPYRAZINECARBOXAMIDE",
            "non-polymer",
        ),
        "AMU": ("BETA-N-ACETYLMURAMIC ACID", "saccharide"),
        "AMT": ("2-AMINOTHIAZOLE", "non-polymer"),
        "AMW": ("ADENOSINE MONOTUNGSTATE", "non-polymer"),
        "AMV": (
            "METHYL 2-(ACETYLAMINO)-3-O-[(1R)-1-CARBOXYETHYL]-2-DEOXY-BETA-D-GLUCOPYRANOSIDE",
            "non-polymer",
        ),
        "NW1": (
            "6-CYCLOHEXYLMETHYLOXY-5-NITROSO-PYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "RPL": (
            "[8,9-DIHYDROXY-3-(4-CARBOXY-HYDROXY-HYDROXYMETHYL-AMINO-BUTYL)-2,4-DIOXO-6-OXA-1,3-DIAZA-SPIRO[4.4]NON-7-YLMETHYL] PHOSPHATE",
            "non-polymer",
        ),
        "SBC": (
            "1-[4-(AMINOSULFONYL)PHENYL]-1,6-DIHYDROPYRAZOLO[3,4-E]INDAZOLE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "AM9": (
            "6-(2,6-DIMETHYLPHENYL)-2-{[4-(4-METHYLPIPERAZIN-1-YL)PHENYL]AMINO}PYRIMIDO[5',4':5,6]PYRIMIDO[1,2-A]BENZIMIDAZOL-5(6H)-ONE",
            "non-polymer",
        ),
        "AM8": (
            "N-(3-CHLOROPHENYL)-N-METHYL-2-OXO-3-[(3,4,5-TRIMETHYL-1H-PYRROL-2-YL)METHYL]-2H-INDOLE-5-SULFONAMIDE",
            "non-polymer",
        ),
        "QIL": ("N-FORMYL-L-ISOLEUCINE", "peptide"),
        "ZD6": (
            "4-BROMO-2-FLUORO-N-[(4E)-6-METHOXY-7-[(1-METHYLPIPERIDIN-4-YL)METHOXY]QUINAZOLIN-4(1H)-YLIDENE]ANILINE",
            "non-polymer",
        ),
        "AM1": (
            "(S)-2-AMINO-3-(3-CARBOXY-5-METHYLISOXAZOL-4-YL)PROPIONIC ACID",
            "non-polymer",
        ),
        "AM0": (
            "N-PHENYL-1-{4-[(3,4,5-TRIMETHOXYPHENYL)AMINO]-1,3,5-TRIAZIN-2-YL}-1H-BENZIMIDAZOL-2-AMINE",
            "non-polymer",
        ),
        "AM3": (
            "(3Z,5S,6R,7S,8S,8AR)-3-(OCTYLIMINO)HEXAHYDRO[1,3]OXAZOLO[3,4-A]PYRIDINE-5,6,7,8-TETROL",
            "non-polymer",
        ),
        "AM2": ("APRAMYCIN", "non-polymer"),
        "AM5": (
            "4-METHYL-N~3~-(2-{[4-(4-METHYLPIPERAZIN-1-YL)PHENYL]AMINO}PYRIMIDIN-5-YL)-N~1~-[3-(TRIFLUOROMETHYL)PHENYL]BENZENE-1,3-DICARBOXAMIDE",
            "non-polymer",
        ),
        "AM4": ("4-(METHYLSULFONYL)BENZENECARBOXIMIDAMIDE", "non-polymer"),
        "AM7": (
            "2-BENZYL-5-(3-FLUORO-4-{[6-METHOXY-7-(3-MORPHOLIN-4-YLPROPOXY)QUINOLIN-4-YL]OXY}PHENYL)-3-METHYLPYRIMIDIN-4(3H)-ONE",
            "non-polymer",
        ),
        "AM6": (
            "2-METHYL-N-{4-METHYL-3-[(2-{[4-(4-METHYLPIPERAZIN-1-YL)PHENYL]AMINO}PYRIMIDIN-5-YL)CARBAMOYL]PHENYL}-3-(TRIFLUOROMETHYL)BENZAMIDE",
            "non-polymer",
        ),
        "MW3": ("MANGANESE ION, 3 WATERS COORDINATED", "non-polymer"),
        "MW2": ("MANGANESE DIHYDRATE ION", "non-polymer"),
        "MW1": ("MANGANESE ION, 1 WATER COORDINATED", "non-polymer"),
        "MW5": (
            "6-BENZYL-5-CHLORO[1,2,4]TRIAZOLO[1,5-A]PYRIMIDIN-7(3H)-ONE",
            "non-polymer",
        ),
        "ZHH": ("2-(4-METHOXYPHENYL)ETHANAMINE", "non-polymer"),
        "Z90": (
            "4-({(4-CARBOXYBUTYL)[2-(2-{[4-(2-PHENYLETHYL)BENZYL]OXY}PHENYL)ETHYL]AMINO}METHYL)BENZOIC ACID",
            "non-polymer",
        ),
        "B6S": ("1,3-BENZOTHIAZOLE-6-SULFONIC ACID", "non-polymer"),
        "B6P": ("?", "non-polymer"),
        "TFM": ("S-ETHYL-N-[4-(TRIFLUOROMETHYL)PHENYL]ISOTHIOUREA", "non-polymer"),
        "QIN": (
            "N-[(1S,2S)-2-AMINO-1-(3-THIENYLMETHYL)HEXYL]-2-({[(1S,2S)-2-METHYLCYCLOPROPYL]METHYL}AMINO)-6-[METHYL(METHYLSULFONYL)AMINO]ISONICOTINAMIDE",
            "non-polymer",
        ),
        "B6A": (
            "[2-AMINO-1-(PIPERIDIN-4-YLMETHYL)-1H-BENZIMIDAZOL-6-YL](PHENYL)METHANONE",
            "non-polymer",
        ),
        "B6D": (
            "2,4-BIS(ACETYLAMINO)-2,4,6-TRIDEOXY-BETA-D-GLUCOPYRANOSE",
            "saccharide",
        ),
        "TJF": (
            "4-[4-METHYL-2-(METHYLAMINO)-1,3-THIAZOL-5-YL]-2-{[3-(MORPHOLIN-4-YLCARBONYL)PHENYL]AMINO}PYRIMIDINE-5-CARBONITRILE",
            "non-polymer",
        ),
        "CI3": (
            "(2Z)-4-[(3S)-1-BENZYL-3-(4-CHLOROBENZYL)PIPERIDIN-3-YL]-2-HYDROXY-4-OXOBUT-2-ENOIC ACID",
            "non-polymer",
        ),
        "HMH": ("4-AMINO-5-HYDROXYMETHYL-2-METHYLPYRIMIDINE", "non-polymer"),
        "QO9": (
            "1-[(5R,6R,8R,9R)-4-AMINO-9-{[TERT-BUTYL(DIMETHYL)SILYL]OXY}-6-({[TERT-BUTYL(DIMETHYL)SILYL]OXY}METHYL)-2,2-DIOXIDO-1,7-DIOXA-2-THIASPIRO[4.4]NON-3-EN-8-YL]-5-METHYLPYRIMIDINE-2,4(1H,3H)-DIONE",
            "non-polymer",
        ),
        "HMI": ("2-HYDROXYCARBAMOYL-4-METHYL-PENTANOIC ACID", "non-polymer"),
        "Y28": ("O-BENZYL-N-(CARBOXYCARBONYL)-D-TYROSINE", "peptide"),
        "B63": (
            "4-{[2-(2-AMINO-4-OXO-3,4,7,8-TETRAHYDROPTERIDIN-6-YL)ETHYL]AMINO}BENZOIC ACID",
            "non-polymer",
        ),
        "B62": (
            "(6R)-2-AMINO-6-METHYL-5,6,7,8-TETRAHYDROPTERIDIN-4(3H)-ONE",
            "non-polymer",
        ),
        "B61": (
            "2-AMINO-6-(2-HYDROXY-2-METHYLPROPYL)-8-METHYL-7,8-DIHYDROPTERIDIN-4(3H)-ONE",
            "non-polymer",
        ),
        "B60": ("2-AMINO-8-METHYL-7,8-DIHYDROPTERIDIN-4(3H)-ONE", "non-polymer"),
        "B67": (
            "(2S)-2-HYDROXY-2-METHYL-N-[4-NITRO-3-(TRIFLUOROMETHYL)PHENYL]-3-(PENTAFLUOROPHENOXY)PROPANAMIDE",
            "non-polymer",
        ),
        "B66": (
            "4-{[(1R,2S)-1,2-DIHYDROXY-2-METHYL-3-(4-NITROPHENOXY)PROPYL]AMINO}-2-(TRIFLUOROMETHYL)BENZONITRILE",
            "non-polymer",
        ),
        "B65": (
            "(1R)-4-(3-PHENOXYPHENYL)-1-PHOSPHONOBUTANE-1-SULFONIC ACID",
            "non-polymer",
        ),
        "B64": (
            "4-{[2-(METHOXYCARBONYL)-5-(2-THIENYL)-3-THIENYL]AMINO}-4-OXO-2-BUTENOIC ACID ",
            "non-polymer",
        ),
        "B69": ("?", "non-polymer"),
        "B68": (
            "(2S)-3-[4-(ACETYLAMINO)PHENOXY]-2-HYDROXY-2-METHYL-N-[4-NITRO-3-(TRIFLUOROMETHYL)PHENYL]PROPANAMIDE",
            "non-polymer",
        ),
        "YNM": ("N-METHYL-L-TYROSINE", "peptide"),
        "G0G": (
            "N,N'-(3S,4S)-PYRROLIDINE-3,4-DIYLBIS(N-BENZYLBENZENESULFONAMIDE)",
            "non-polymer",
        ),
        "VO3": ("TETRAMETAVANADATE", "non-polymer"),
        "X09": ("AMINO-AUSTRALINE", "non-polymer"),
        "X03": (
            "(5R)-5-(2-METHYLBUTAN-2-YL)-4,5,6,7-TETRAHYDRO-1H-INDAZOLE-3-CARBOHYDRAZIDE",
            "non-polymer",
        ),
        "X02": (
            "[4-AMINO-2-(PROP-2-EN-1-YLAMINO)-1,3-THIAZOL-5-YL](PHENYL)METHANONE",
            "non-polymer",
        ),
        "X01": ("3-(4-AMINO-1,3,5-TRIAZIN-2-YL)-4-HYDROXYBENZONITRILE", "non-polymer"),
        "RP5": ("5-O-PHOSPHONO-BETA-D-RIBOFURANOSE", "non-polymer"),
        "X06": ("4-CHLORO-2-(4,6-DIAMINO-1,3,5-TRIAZIN-2-YL)PHENOL", "non-polymer"),
        "I84": (
            "[2,6-DIMETHYL-4-(2-O-TOLYL-ACETYLAMINO)-BENZENESULFONYL]-GLYCINE",
            "non-polymer",
        ),
        "I85": (
            "N-{4-[(3S)-4'-OXO-1',4',5',6'-TETRAHYDROSPIRO[PIPERIDINE-3,7'-PYRROLO[3,2-C]PYRIDIN]-2'-YL]PYRIDIN-2-YL}-3-(TRIFLUOROMETHYL)BENZAMIDE",
            "non-polymer",
        ),
        "TFL": (
            "2-[4-(DIMETHYLAMINO)PHENYL]-6-HYDROXY-3-METHYL-1,3-BENZOTHIAZOL-3-IUM",
            "non-polymer",
        ),
        "4NP": ("4-NITROPHENYL PHOSPHATE", "non-polymer"),
        "4NS": ("4-NITROPHENYL SULFATE", "non-polymer"),
        "TT": ("?", "nucleic-acid"),
        "IOY": ("P-IODO-D-PHENYLALANINE", "peptide"),
        "TS": ("THYMIDINE-5'-THIOPHOSPHATE", "nucleic-acid"),
        "4NZ": ("4-NITROBENZENESULFONAMIDE", "non-polymer"),
        "4ND": ("N4-(N,N-DIPHENYLCARBAMOYL)-AMINOGUANIDINE", "non-polymer"),
        "TL": ("THALLIUM (I) ION", "non-polymer"),
        "4NF": ("N-(4-NITROPHENYL)THIOFORMAMIDE", "non-polymer"),
        "4NA": ("1-CHLORO-6-(4-HYDROXYPHENYL)-2-NAPHTHOL", "non-polymer"),
        "4NC": ("4-NITROCATECHOL", "non-polymer"),
        "4NB": ("4-NITROBENZOIC ACID", "non-polymer"),
        "4NM": ("4-NITROPHENYL METHANETHIOL", "non-polymer"),
        "4NL": ("4-AMINOPHENOL", "non-polymer"),
        "4NN": (
            "N-[(5S,6R)-5-HYDROXY-6-(HYDROXYMETHYL)-2-OXO-5,6-DIHYDRO-2H-PYRAN-3-YL]ACETAMIDE",
            "saccharide",
        ),
        "4NI": ("4-NITROBUTANOIC ACID", "non-polymer"),
        "4NH": (
            "4-({4-[(4-AMINOBUT-2-YNYL)OXY]PHENYL}SULFONYL)-N-HYDROXY-2,2-DIMETHYLTHIOMORPHOLINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "AA": ("9-AMINOACRIDINE", "non-polymer"),
        "AG": ("SILVER ION", "non-polymer"),
        "AF": ("2-AMINOFLUORENE", "non-polymer"),
        "AI": (
            "(2S,2'S)-N,N'-[(2S,3S,4S,5S)-1-CYCLOHEXYL-3,4-DIHYDROXY-6-PHENYLHEXANE-2,5-DIYL]BIS[3-METHYL-2-({[METHYL(PYRIDIN-2-YLMETHYL)AMINO]CARBONYL}AMINO)BUTANAMIDE]",
            "non-polymer",
        ),
        "AL": ("ALUMINUM ION", "non-polymer"),
        "XTZ": (
            "4-{[(2-AMINO-4-OXO-3,4,7,8-TETRAHYDROPTERIDIN-6-YL)METHYL]AMINO}-N-(1,3-THIAZOL-2-YL)BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "AP": ("N-1-AMINOPYRENE", "non-polymer"),
        "AS": ("?", "nucleic-acid"),
        "AR": ("ARGON", "non-polymer"),
        "AU": ("GOLD ION", "non-polymer"),
        "X0J": (
            "4-[4-(BIPHENYL-3-YLMETHYL)PIPERAZIN-1-YL]-N-{[4-({(1R)-3-(DIMETHYLAMINO)-1-[(PHENYLSULFANYL)METHYL]PROPYL}AMINO)-3-NITROPHENYL]SULFONYL}BENZAMIDE",
            "non-polymer",
        ),
        "X0A": (
            "2-[4-AMINO-6-(PHENYLAMINO)-1,3,5-TRIAZIN-2-YL]-4-CHLOROPHENOL",
            "non-polymer",
        ),
        "UCL": ("?", "nucleic-acid"),
        "I8P": (
            "(1R,3S,4R,5S,6R)-2,4,5,6-TETRAKIS(PHOSPHONOOXY)CYCLOHEXANE-1,3-DIYL BIS[TRIHYDROGEN (DIPHOSPHATE)]",
            "non-polymer",
        ),
        "LGU": ("ALPHA-L-GULURONATE", "saccharide"),
        "X0W": ("5,6-DIMETHOXY-2-METHYL-1,3-BENZOTHIAZOLE", "non-polymer"),
        "X0V": ("4-(TRIFLUOROMETHYL)BENZENE-1,2-DIAMINE", "non-polymer"),
        "X0U": (
            "6,7-DIHYDRO[1,4]DIOXINO[2,3-F][1,3]BENZOTHIAZOL-2-AMINE",
            "non-polymer",
        ),
        "X0T": (
            "7-[(2S)-2,3-DIHYDROXYPROPYL]-1,3-DIMETHYL-3,7-DIHYDRO-1H-PURINE-2,6-DIONE",
            "non-polymer",
        ),
        "G26": (
            "2-(BENZYLCARBAMOYL-PHENYLACETYLAMINO-METHYL)-5,5-DIMETHYL-THIAZOLIDINE-4-CARBOXYLIC ACID (HYDROXYMETHYL-2-PHENYLETHYL)AMIDE",
            "non-polymer",
        ),
        "G27": (
            "(3R,4R,5R)-5-(HYDROXYMETHYL)-1-(3-PHENYLPROPYL)PIPERIDINE-3,4-DIOL",
            "non-polymer",
        ),
        "G24": (
            "[4-(3-BENZYL-4-HYDROXYBENZYL)-3,5-DIMETHYLPHENOXY]ACETIC ACID",
            "non-polymer",
        ),
        "G25": ("GUANOSINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "G23": (
            "(2R,4S)-2-[(R)-BENZYLCARBAMOYL-PHENYLACETYL-METHYL]-5,5-DIMETHYL-THIAZOLIDINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "G20": (
            "4-ACETYL-4-GUANIDINO-6-METHYL(PROPYL)CARBOXAMIDE-4,5-DIHYDRO-2H-PYRAN-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "G21": (
            "5-ACETYLAMINO-4-AMINO-6-(PHENETHYLPROPYLCARBAMOYL)5,6-DIHYDRO-4H-PYRAN-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "IMY": (
            "1-{[4,5-BIS(4-CHLOROPHENYL)-2-(2-ISOPROPOXY-4-METHOXYPHENYL)-4,5-DIHYDRO-1H-IMIDAZOL-1-YL]CARBONYL}PIPERAZINE",
            "non-polymer",
        ),
        "G28": (
            "5-N-ACETYL-4-AMINO-6-DIETHYLCARBOXAMIDE-4,5-DIHYDRO-2H-PYRAN-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "G29": (
            "(6ALPHA,8ALPHA)-6-(BUT-2-YN-1-YLOXY)ANDROSTA-1,4-DIENE-3,17-DIONE",
            "non-polymer",
        ),
        "XJ1": (
            "(2E)-3-(3-{6-[(TRANS-4-AMINOCYCLOHEXYL)AMINO]PYRAZIN-2-YL}PHENYL)PROP-2-ENOIC ACID",
            "non-polymer",
        ),
        "WLM": (
            "5-CHLORO-2-METHOXY-N-[2-[4-[(5-PROPAN-2-YL-1,3,4-THIADIAZOL-2-YL)SULFAMOYL]PHENYL]ETHYL]BENZAMIDE",
            "non-polymer",
        ),
        "XTN": ("2-AMINO-1,5-DIHYDROPTERIDINE-4,6-DIONE", "non-polymer"),
        "R19": (
            "ALLYL-{6-[3-(4-BROMO-PHENYL)-BENZOFURAN-6-YLOXY]-HEXYL-}-METHYL-AMIN",
            "non-polymer",
        ),
        "THY": ("C2-1-HYDROXY-2-METHYL-BUTYL-THIAMIN DIPHOSPHATE", "non-polymer"),
        "DZD": (
            "5'-O-[(R)-{[(R)-{[(3R,4R)-1-(3-CARBAMOYLBENZYL)-4-HYDROXYPYRROLIDIN-3-YL]METHOXY}(HYDROXY)PHOSPHORYL]METHYL}(HYDROXY)PHOSPHORYL]ADENOSINE",
            "non-polymer",
        ),
        "8OP": (
            "[(2R,3S,4R,5R)-5-(6-AZANYL-8-OXO-7H-PURIN-9-YL)-3,4-DIHYDROXY-OXOLAN-2-YL]METHYL DIHYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "TG1": ("THAPSIGARGIN", "non-polymer"),
        "8OX": (
            "6-AZANYL-9-[(2R,3R,4S,5R)-3,4-DIHYDROXY-5-(HYDROXYMETHYL)OXOLAN-2-YL]-7H-PURIN-8-ONE",
            "non-polymer",
        ),
        "8OG": ("8-OXO-2'-DEOXY-GUANOSINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "8OD": (
            "[(2R,3S,4R,5R)-5-(6-AZANYL-8-OXO-7H-PURIN-9-YL)-3,4-DIHYDROXY-OXOLAN-2-YL]METHYL PHOSPHONO HYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "TFO": (
            "[2-(6-AMINO-9H-PURIN-9-YL)-1-METHYLETHOXY]METHYLPHOSPHONIC ACID",
            "nucleic-acid",
        ),
        "PDR": (
            "3-(4-CHLORO-1H-PYRROLO[2,3-B]PYRIDIN-2-YL)-5,6-DIMETHOXY-1-METHYL-1H-PYRROLO[3,2-B]PYRIDINE",
            "non-polymer",
        ),
        "V04": (
            "4-[(2-FLUORO-4-IODOPHENYL)AMINO]-N-(2-HYDROXYETHOXY)-1H-INDAZOLE-5-CARBOXAMIDE",
            "non-polymer",
        ),
        "CI9": (
            "N-({(1S)-5-[4-(13-{[2,4-BIS(DIHYDROXYAMINO)PHENYL]AMINO}-2,5,8,11-TETRAOXATRIDEC-1-YL)-1H-1,2,3-TRIAZOL-1-YL]-1-CARBOXYPENTYL}CARBAMOYL)-L-GLUTAMIC ACID",
            "non-polymer",
        ),
        "Z2T": (
            "(1S)-1,5-ANHYDRO-1-(3-CHLORO-2-HYDROXY-5-METHOXYPHENYL)-D-GLUCITOL",
            "non-polymer",
        ),
        "PC1": ("1,2-DIACYL-SN-GLYCERO-3-PHOSPHOCHOLINE", "non-polymer"),
        "T38": ("?", "nucleic-acid"),
        "8PP": (
            "2-[(2E,6E,10E,14E,18E,22E,26E)-3,7,11,15,19,23,27,31-OCTAMETHYLDOTRIACONTA-2,6,10,14,18,22,26,30-OCTAENYL]PHENOL",
            "non-polymer",
        ),
        "G2R": (
            "[(2R,3R,4R,5R)-5-(2-AMINO-6-OXO-1,6-DIHYDRO-9H-PURIN-9-YL)-3,4-DIHYDROXYTETRAHYDROFURAN-2-YL]METHYL [(2R,3S,4S)-3,4-DIHYDROXYTETRAHYDROFURAN-2-YL]METHYL DIHYDROGEN DIPHOSPHATE",
            "non-polymer",
        ),
        "G2S": ("GUANOSINE-5'-DITHIOPHOSPHORATE", "nucleic-acid"),
        "G2P": ("PHOSPHOMETHYLPHOSPHONIC ACID GUANYLATE ESTER", "non-polymer"),
        "DZP": (
            "7-CHLORO-1-METHYL-5-PHENYL-1,3-DIHYDRO-2H-1,4-BENZODIAZEPIN-2-ONE",
            "non-polymer",
        ),
        "G2F": ("2-DEOXY-2-FLUORO-ALPHA-D-GLUCOPYRANOSE", "saccharide"),
        "G2G": (
            "5-TERT-BUTYL-2-(4-METHYLPHENYL)-3-{[({4-[6-(MORPHOLIN-4-YLMETHYL)PYRIDIN-3-YL]-1-NAPHTHYL}AMINO)CARBONYL]AMINO}-1H-PYRAZOL-2-IUM",
            "non-polymer",
        ),
        "G2C": (
            "2'-DEOXY-5'-O-[(S)-{[(R)-[DICHLORO(PHOSPHONO)METHYL](HYDROXY)PHOSPHORYL]OXY}(HYDROXY)PHOSPHORYL]GUANOSINE",
            "non-polymer",
        ),
        "G2N": (
            "ETHYL [(2R)-5-AMINO-2-METHYL-3-PHENYL-1,2-DIHYDROPYRIDO[3,4-B]PYRAZIN-7-YL]CARBAMATE",
            "non-polymer",
        ),
        "G2L": ("3'-O-METHYOXYETHYL-GUANOSINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "G2M": (
            "2'-DEOXY-5'-O-[(S)-HYDROXY{[(S)-HYDROXY(1-METHYL-1-PHOSPHONOETHYL)PHOSPHORYL]OXY}PHOSPHORYL]GUANOSINE",
            "non-polymer",
        ),
        "G2H": ("2-HYDROXY-1-(HYDROXYMETHYL)ETHYL DIHYDROGEN PHOSPHATE", "non-polymer"),
        "G2I": (
            "(3R,4R,5R)-3-HYDROXY-5-(HYDROXYMETHYL)PIPERIDIN-4-YL 4-O-BETA-D-GLUCOPYRANOSYL-BETA-D-GLUCOPYRANOSIDE",
            "non-polymer",
        ),
        "T39": ("?", "nucleic-acid"),
        "MO8": (
            "4-METHYL-7,8-DIHYDRO-5H-THIOPYRANO[4,3-D]PYRIMIDIN-2-AMINE",
            "non-polymer",
        ),
        "64M": (
            "5-[(5-METHOXY-1H-PYRROLO[2,3-B]PYRIDIN-3-YL)METHYL]-N-[4-(TRIFLUOROMETHYL)BENZYL]PYRIDIN-2-AMINE",
            "non-polymer",
        ),
        "64P": ("?", "non-polymer"),
        "64U": ("3-CYCLOHEXYL-D-ALANYL-N-(3-CHLOROBENZYL)-L-PROLINAMIDE", "peptide"),
        "64T": ("5-HYDROXY-THYMIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "T36": ("?", "nucleic-acid"),
        "IMX": (
            "[(1R,2R)-2-{[(5-FLUORO-1H-INDOL-2-YL)CARBONYL]AMINO}-2,3-DIHYDRO-1H-INDEN-1-YL]ACETIC ACID",
            "non-polymer",
        ),
        "U2F": (
            "URIDINE-5'-DIPHOSPHATE-2-DEOXY-2-FLUORO-ALPHA-D-GLUCOSE",
            "non-polymer",
        ),
        "PH0": (
            "N-{(2S)-3-[(R)-[(1R)-1-AMINO-2-PHENYLETHYL](HYDROXY)PHOSPHORYL]-2-BENZYLPROPANOYL}-L-PHENYLALANINE",
            "non-polymer",
        ),
        "U2L": ("?", "nucleic-acid"),
        "U2N": ("?", "nucleic-acid"),
        "THT": ("TRANS-2-HEXADECENOYL-(N-ACETYL-CYSTEAMINE)-THIOESTER", "non-polymer"),
        "U2P": (
            "PHOSPHORIC ACID MONO-[2-(2,4-DIOXO-3,4-DIHYDRO-2H-PYRIMIDIN-1-YL)-4-HYDROXY-5-HYDROXYMETHYL-TETRAHYDRO-FURAN-3-YL] ESTER",
            "nucleic-acid",
        ),
        "U2S": (
            "1-(5-DEOXY-5-MORPHOLIN-4-YL-ALPHA-L-ARABINOFURANOSYL)PYRIMIDINE-2,4(1H,3H)-DIONE",
            "non-polymer",
        ),
        "N04": (
            "(2R)-2-[[3-[[3-[(Z)-(2,4-DIOXO-1,3-THIAZOLIDIN-5-YLIDENE)METHYL]PHENOXY]METHYL]PHENYL]CARBONYLAMINO]PENTANEDIOIC ACID",
            "non-polymer",
        ),
        "PH4": (
            "3-{HYDROXY[(PHOSPHONOOXY)ACETYL]AMINO}PROPYL DIHYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "T37": ("?", "nucleic-acid"),
        "PUY": ("PUROMYCIN", "nucleic-acid"),
        "PUT": ("1,4-DIAMINOBUTANE", "non-polymer"),
        "U20": (
            "URIDINE-5'-DIPHOSPHATE-3-O-(R-3-HYDROXYMYRISTOYL)-N-ACETYL-D-GLUCOSAMINE",
            "non-polymer",
        ),
        "U21": (
            "URIDINE-5'-DIPHOSPHATE-3-O-(R-3-HYDROXYDECANOYL)-N-ACETYL-D-GLUCOSAMINE",
            "non-polymer",
        ),
        "U22": (
            "URIDINE-5'-DIPHOSPHATE-3-N-(R-3-HYDROXYLAUROYL)-N-ACETYL-D-GLUCOSAMINE",
            "non-polymer",
        ),
        "THR": ("THREONINE", "peptide"),
        "641": (
            "(3S)-1-CYCLOHEXYL-N-(3,5-DICHLOROPHENYL)-5-OXOPYRROLIDINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "643": (
            "6-(4-{[3-(3,5-DICHLOROPYRIDIN-4-YL)-5-(1-METHYLETHYL)ISOXAZOL-4-YL]METHOXY}-2-METHYLPHENYL)-1-METHYL-1H-INDOLE-3-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "642": (
            "(1S,3R,6S)-4-OXO-6-{4-[(2-PHENYLQUINOLIN-4-YL)METHOXY]PHENYL}-5-AZASPIRO[2.4]HEPTANE-1-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "645": (
            "[1-TERT-BUTYL-3-(2,4-DICHLOROPHENYL)-5-HYDROXY-1H-PYRAZOL-4-YL][2-CHLORO-4-(METHYLSULFONYL)PHENYL]METHANONE",
            "non-polymer",
        ),
        "PUR": ("PURINE RIBOSIDE", "non-polymer"),
        "647": (
            "5-(1H-PYRROLO[2,3-B]PYRIDIN-3-YLMETHYL)-N-[4-(TRIFLUOROMETHYL)BENZYL]PYRIDIN-2-AMINE",
            "non-polymer",
        ),
        "646": (
            "2,6-DICHLORO-4-(2-PIPERAZIN-1-YLPYRIDIN-4-YL)-N-(1,3,5-TRIMETHYL-1H-PYRAZOL-4-YL)BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "TN1": ("5,6,7,8-TETRAHYDRONAPHTHALEN-1-OL", "non-polymer"),
        "RGS": ("4N-MALONYL-CYSTEINYL-2,4-DIAMINOBUTYRATE DISULFIDE", "non-polymer"),
        "NC8": (
            "(3R)-3-({(4-AMINOBENZYL)[(4-AMINOPHENYL)ACETYL]AMINO}METHYL)-5-{[(4-BROMOBENZOYL)OXY]METHYL}-2,3,4,7-TETRAHYDRO-1H-AZEPINIUM",
            "non-polymer",
        ),
        "R1A": (
            "3-{[(2,2,5,5-TETRAMETHYL-1-OXO-2,5-DIHYDRO-1H-PYRROLIUM-3-YL)METHYL]DISULFANYL}-D-ALANINE",
            "peptide",
        ),
        "THQ": (
            "PHOSPHORIC ACID MONO-[2-(4-METHYL-THIAZOL-5-YL)-ETHYL] ESTER",
            "non-polymer",
        ),
        "HM8": (
            "2-[(1S,2R)-1-AMINO-2-(SULFOOXY)PROPYL]-5-HYDROXY-1,3-OXAZOLE-4-CARBOTHIOIC O-ACID",
            "peptide",
        ),
        "MO3": ("MAGNESIUM ION, 3 WATERS COORDINATED", "non-polymer"),
        "I0G": (
            "(3AS,4R,9BR)-4-(4-HYDROXYPHENYL)-1,2,3,3A,4,9B-HEXAHYDROCYCLOPENTA[C]CHROMEN-8-OL",
            "non-polymer",
        ),
        "OF3": ("FERRIC ION, 1 WATER COORDINATED", "non-polymer"),
        "JST": (
            "BUT-3-ENYL-[5-(4-CHLORO-PHENYL)-3,6-DIHYDRO-[1,3,4]THIADIAZIN-2-YLIDENE]-AMINE",
            "non-polymer",
        ),
        "HM9": (
            "6-(3-CARBAMIMIDAMIDOPROPYL)-3,5-DIHYDROXYPYRAZINE-2-CARBOTHIOIC S-ACID",
            "peptide",
        ),
        "0J9": (
            "1-TERT-BUTYL-3-(NAPHTHALEN-1-YL)-1H-PYRAZOLO[3,4-D]PYRIMIDIN-4-AMINE",
            "non-polymer",
        ),
        "0J8": (
            "N-(3-NITROBENZYL)-6-[1-(PIPERIDIN-4-YL)-1H-PYRAZOL-4-YL]-2-(TRIFLUOROMETHYL)PYRIDO[2,3-D]PYRIMIDIN-4-AMINE",
            "non-polymer",
        ),
        "JSR": (
            "6-{[(3R,4S)-4-(2-{[2-(3-FLUOROPHENYL)ETHYL]AMINO}ETHOXY)PYRROLIDIN-3-YL]METHYL}-4-METHYLPYRIDIN-2-AMINE",
            "non-polymer",
        ),
        "JSS": (
            "6-{[(3S,4S)-4-(2-{[2-(3-FLUOROPHENYL)ETHYL]AMINO}ETHOXY)PYRROLIDIN-3-YL]METHYL}-4-METHYLPYRIDIN-2-AMINE",
            "non-polymer",
        ),
        "0J5": (
            "(7-AMINO-4,5-DIOXO-1,4,5,6-TETRAHYDROPYRIMIDO[4,5-C]PYRIDAZIN-3-YL)ACETIC ACID",
            "non-polymer",
        ),
        "0J4": (
            "(2R)-2-(7-AMINO-4,5-DIOXO-1,4,5,6-TETRAHYDROPYRIMIDO[4,5-C]PYRIDAZIN-3-YL)PROPANOIC ACID",
            "non-polymer",
        ),
        "0J7": (
            "3-(PYRIMIDIN-2-YL)-N-[3-(1H-TETRAZOL-5-YL)PHENYL]BENZAMIDE",
            "non-polymer",
        ),
        "0J6": (
            "N-[3-(1H-TETRAZOL-5-YL)PHENYL]-2H-INDAZOLE-5-CARBOXAMIDE",
            "non-polymer",
        ),
        "0J0": (
            "(3-EXO)-8,8-DIMETHYL-3-(4-{[(1-METHYL-2-OXO-1,2-DIHYDROQUINOLIN-4-YL)OXY]METHYL}-1H-1,2,3-TRIAZOL-1-YL)-8-AZONIABICYCLO[3.2.1]OCTANE",
            "non-polymer",
        ),
        "0J3": (
            "3-(4-METHYLPIPERAZIN-1-YL)-N-(3-NITROBENZYL)-7-(TRIFLUOROMETHYL)QUINOLIN-5-AMINE",
            "non-polymer",
        ),
        "0J2": (
            "(3R)-3-(7-AMINO-4,5-DIOXO-1,4,5,6-TETRAHYDROPYRIMIDO[4,5-C]PYRIDAZIN-3-YL)BUTANOIC ACID",
            "non-polymer",
        ),
        "0JM": (
            "(2S,5R,6R,7R)-6-{[(2R)-2-CARBOXY-2-PHENYLACETYL]AMINO}-7-HYDROXY-3,3-DIMETHYL-4-THIA-1-AZABICYCLO[3.2.0]HEPTANE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "0JL": (
            "3-{(1S)-1-[3-(2-METHOXYETHOXY)QUINOLIN-6-YL]ETHYL}-5-(3-METHYL-1,2-THIAZOL-5-YL)-3,5-DIHYDRO-4H-[1,2,3]TRIAZOLO[4,5-C]PYRIDIN-4-ONE",
            "non-polymer",
        ),
        "0JO": (
            "2-{[(E)-{3-HYDROXY-2-METHYL-5-[(PHOSPHONOOXY)METHYL]PYRIDIN-4-YL}METHYLIDENE]AMINO}PROP-2-ENOIC ACID",
            "non-polymer",
        ),
        "6IA": ("N6-ISOPENTENYL-ADENOSINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "6IG": (
            "6-ETHYL-5-[1-(3-METHOXYPROPYL)-1,2,3,4-TETRAHYDROQUINOLIN-7-YL]-N~4~-(2-PHENYLETHYL)PYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "0JK": (
            "5-PHENYL-3-(QUINOLIN-6-YLMETHYL)-3,5,6,7-TETRAHYDRO-4H-[1,2,3]TRIAZOLO[4,5-C]PYRIDIN-4-ONE",
            "non-polymer",
        ),
        "0JJ": (
            "7-METHOXY-N-{[6-(3-METHYL-1,2-THIAZOL-5-YL)[1,2,4]TRIAZOLO[4,3-B]PYRIDAZIN-3-YL]METHYL}-1,5-NAPHTHYRIDIN-4-AMINE",
            "non-polymer",
        ),
        "0JE": (
            "3-[5-(5-ETHOXY-6-FLUORO-1H-BENZIMIDAZOL-2-YL)-1H-PYRAZOL-4-YL]-1,1-DIETHYLUREA",
            "non-polymer",
        ),
        "0JD": ("(2S)-2,3-DIHYDRO-1,4-BENZODIOXINE-2-CARBOXYLIC ACID", "non-polymer"),
        "0JG": (
            "METHYL (2S)-3-{4-AMINO-7-[(1E)-3-HYDROXYPROP-1-EN-1-YL]-5-(4-METHYLPHENYL)-7H-PYRROLO[2,3-D]PYRIMIDIN-6-YL}-2-CYANOPROPANOATE",
            "non-polymer",
        ),
        "0JF": (
            "(3S)-N-(4-METHYLBENZYL)-1-{2-[(3,4,5-TRIMETHOXYPHENYL)AMINO]PYRIMIDIN-4-YL}PIPERIDINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "6IN": (
            "4-(1-BENZYL-3-CARBAMOYLMETHYL-2-METHYL-1H-INDOL-5-YLOXY)-BUTYRIC ACID",
            "non-polymer",
        ),
        "0JC": ("DI-MU-IODOBIS(ETHYLENEDIAMINE)DIPLATINUM(II)", "non-polymer"),
        "0JB": (
            "N-[3-(1H-TETRAZOL-5-YL)PHENYL]-1H-BENZIMIDAZOLE-7-CARBOXAMIDE",
            "non-polymer",
        ),
        "JS4": ("unknown", "non-polymer"),
        "JS5": ("unknown", "non-polymer"),
        "6IP": ("6-[2-(1H-INDOL-6-YL)ETHYL]PYRIDIN-2-AMINE", "non-polymer"),
        "JS7": (
            "4-{2-[(6-METHOXY-5-NITROPYRIMIDIN-4-YL)AMINO]ETHYL}BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "0JY": ("4-METHYL-L-LEUCINE", "non-polymer"),
        "0JX": ("(3E)-4-HYDROXYBUT-3-EN-1-YL TRIHYDROGEN DIPHOSPHATE", "non-polymer"),
        "0JZ": (
            "(4R,5R,11R,12R,18R,19S)-4,11,18-TRIS(1-METHYLETHYL)-6,13,20-TRISELENA-3,10,17,22,23,24-HEXAAZATETRACYCLO[17.2.1.1~5,8~.1~12,15~]TETRACOSA-1(21),7,14-TRIENE-2,9,16-TRIONE",
            "non-polymer",
        ),
        "0JU": ("(4S,5Z)-4-AMINO-5-IMINOPENTANAMIDE", "non-polymer"),
        "0JT": ("(4R)-4-AMINO-5-METHYLHEXANOIC ACID", "peptide"),
        "0JW": (
            "(2E)-2-[(R)-AMINO(HYDROXY)METHYL]-3-(3,4-DIHYDROXYPHENYL)PROP-2-ENENITRILE",
            "non-polymer",
        ),
        "0JV": (
            "METHYL N-[(3S,3AR,5R,6AR)-5-[[(2S,3R)-4-[(4-METHOXYPHENYL)SULFONYL-(2-METHYLPROPYL)AMINO]-3-OXIDANYL-1-PHENYL-BUTAN-2-YL]CARBAMOYLOXY]-3,3A,4,5,6,6A-HEXAHYDRO-2H-CYCLOPENTA[B]FURAN-3-YL]CARBAMATE",
            "non-polymer",
        ),
        "0JQ": (
            "2-{1-[5-(6,7-DIMETHOXYCINNOLIN-4-YL)-3-METHYLPYRIDIN-2-YL]PIPERIDIN-4-YL}PROPAN-2-OL",
            "non-polymer",
        ),
        "0JP": (
            "8,9-DIMETHOXY-1-(1,3-THIAZOL-5-YL)-5,6-DIHYDROIMIDAZO[5,1-A]ISOQUINOLINE",
            "non-polymer",
        ),
        "POC": ("PHOSPHOCHOLINE", "non-polymer"),
        "TFI": (
            "2-[5-METHANESULFONYLAMINO-2-(4-AMINOPHENYL)-6-OXO-1,6-DIHYDRO-1-PYRIMIDINYL]-N-(3,3,3-TRIFLUORO-1-ISOPROPYL-2-OXOPROPYL)ACETAMIDE",
            "non-polymer",
        ),
        "VU3": ("2-(2-(4-PHENYLPIPERIDIN-1-YL)ETHYL)-1H-INDOLE", "non-polymer"),
        "XF5": ("4-(PIPERAZIN-1-YL)-1H-INDOLE", "non-polymer"),
        "T31": ("?", "nucleic-acid"),
        "SPW": ("N,N'-BIS(3-AMMONIOPROPYL)BUTANE-1,4-DIAMINIUM", "non-polymer"),
        "A99": (
            "5'-O-[(S)-HYDROXY{[(R)-HYDROXY({(S)-HYDROXY[(1S)-1-(2-NITROPHENYL)ETHOXY]PHOSPHORYL}OXY)PHOSPHORYL]OXY}PHOSPHORYL]ADENOSINE",
            "non-polymer",
        ),
        "IA2": ("2,4-DICHLORO-6-(3,4,5-TRICHLORO-1H-PYRROL-2YL)PHENOL", "non-polymer"),
        "BAC": ("N-(4-IODO-BENZYL)-FORMAMIDE", "non-polymer"),
        "TOX": ("2-AMINO-3-(1-HYDROPEROXY-1H-INDOL-3-YL)PROPAN-1-OL", "peptide"),
        "GQL": (
            "(E)-[4-(3,5-DIFLUOROPHENYL)-3H-PYRROLO[2,3-B]PYRIDIN-3-YLIDENE](3-METHOXYPHENYL)METHANOL",
            "non-polymer",
        ),
        "TOR": ("TOPIRAMATE", "non-polymer"),
        "TOS": ("P-SULFINOTOLUENE", "non-polymer"),
        "TOP": ("TRIMETHOPRIM", "non-polymer"),
        "TOQ": ("6,7-DIHYDROXY-L-TRYPTOPHAN", "peptide"),
        "TOT": (
            "1,1-(4,4,8,8-TETRAMETHYL-4,8-DIAZAUNDECAMETHYLENE)-BIS-4-3-METHYL-2,3-DIHYDRO-(BENZO-1,3-THIAZOLE)-2-METHYLIDENE)-QUINOLINIUM",
            "non-polymer",
        ),
        "TOU": ("THIOUREA", "non-polymer"),
        "BAN": ("HONH-BENZYLMALONYL-L-ALANYLGLYCINE-P-NITROANILIDE", "non-polymer"),
        "A0H": (
            "N-(3-{3-CHLORO-8-[(4-MORPHOLIN-4-YLPHENYL)AMINO]IMIDAZO[1,2-A]PYRAZIN-6-YL}BENZYL)METHANESULFONAMIDE",
            "non-polymer",
        ),
        "TON": (
            "2-{2-[4-(1,1,3,3-TETRAMETHYLBUTYL)PHENOXY]ETHOXY}ETHANOL",
            "non-polymer",
        ),
        "TOL": ("TOLRESTAT", "non-polymer"),
        "235": (
            "(2S)-2-[3-(AMINOMETHYL)PHENYL]-3-[(R)-HYDROXY{(1R)-2-METHYL-1-[(PHENYLSULFONYL)AMINO]PROPYL}PHOSPHORYL]PROPANOIC ACID",
            "non-polymer",
        ),
        "TOB": ("1,3-DIAMINO-5,6-DIHYDROXYCYCLOHEXANE", "non-polymer"),
        "TOC": ("2,3,6-TRIDEOXY-2,6-DIAMINO GLUCOSE", "saccharide"),
        "A0A": ("(2S)-2-AMINO-4-(FORMYLOXY)-4-OXOBUTANOIC ACID", "non-polymer"),
        "TOF": (
            "1,6-DIMETHYLPYRIMIDO[5,4-E][1,2,4]TRIAZINE-5,7(1H,6H)-DIONE",
            "non-polymer",
        ),
        "TF7": (
            "(3S)-3-(DIHYDROXY-LAMBDA~4~-SULFANYL)-4-(1H-1,2,3-TRIAZOL-1-YL)-D-VALINE",
            "non-polymer",
        ),
        "A0D": ("N-ADAMANTAN-2-YL-1-ETHYL-D-PROLINAMIDE", "non-polymer"),
        "TOE": ("2-[2-(2-METHOXY-ETHOXY)-ETHOXY]-ETHOXYL", "non-polymer"),
        "HIP": ("ND1-PHOSPHONOHISTIDINE", "peptide"),
        "HIQ": ("1-[1,2-DIHYDROXY-1-(HYDROXYMETHYL)ETHYL]-L-HISTIDINE", "peptide"),
        "NHG": ("BIS-NORESEROLINE", "non-polymer"),
        "A96": (
            "6-CHLORO-N-PYRIMIDIN-5-YL-3-{[3-(TRIFLUOROMETHYL)PHENYL]AMINO}-1,2-BENZISOXAZOLE-7-CARBOXAMIDE",
            "non-polymer",
        ),
        "HIT": ("2-(ALPHA-HYDROXYISOVALERYL) PROPIONIC ACID", "non-polymer"),
        "HIU": ("(2R)-3-HYDROXY-2-METHYLPROPANOIC ACID", "non-polymer"),
        "NHC": (
            "(1R,2S,3R,4S)-2,3-DIMETHYL-7-OXABICYCLO[2.2.1]HEPTANE-2,3-DICARBOXYLIC ACID",
            "non-polymer",
        ),
        "NHB": (
            "N-HYDROXY-4-(METHYL{[5-(2-PYRIDINYL)-2-THIENYL]SULFONYL}AMINO)BENZAMIDE",
            "non-polymer",
        ),
        "NHM": ("S-(2-OXO)PENTADECYLCOA", "non-polymer"),
        "A91": (
            "3-({2-[(2-AMINO-6-METHYLPYRIMIDIN-4-YL)ETHYNYL]BENZYL}AMINO)-1,3-OXAZOL-2(3H)-ONE",
            "non-polymer",
        ),
        "HIZ": (
            "1-{4-[(1S)-1-AMINO-2-(BENZYLOXY)ETHYL]-1,3-THIAZOL-2-YL}-3-(3-CHLORO-4-FLUOROPHENYL)UREA",
            "non-polymer",
        ),
        "NHN": ("?", "non-polymer"),
        "NHI": ("?", "non-polymer"),
        "NHK": (
            "N-HYDROXY-N~2~-(2-HYDROXYETHYL)-N~2~-[(4-METHOXYPHENYL)SULFONYL]GLYCINAMIDE",
            "non-polymer",
        ),
        "NHJ": (
            "4-[(4-{[3-(TRIFLUOROMETHYL)PHENYL]AMINO}PYRIMIDIN-2-YL)AMINO]BENZAMIDE",
            "non-polymer",
        ),
        "NHU": (
            "2-({2-[(4-CARBAMOYLPHENYL)AMINO]PYRIMIDIN-4-YL}AMINO)BENZAMIDE",
            "non-polymer",
        ),
        "HIA": ("L-HISTIDINE AMIDE", "peptide"),
        "HIB": ("4-HYDROXY-5-IODO-2,3-DIMETHOXY-6-METHYLBENZOIC ACID", "non-polymer"),
        "HIC": ("4-METHYL-HISTIDINE", "peptide"),
        "HID": ("(5-HYDROXY-1H-INDOL-3-YL)ACETIC ACID", "non-polymer"),
        "A93": (
            "4-FLUOROPHENYL 3-[(E)-2-(4-HYDROXY-3,5-DIMETHYLPHENYL)ETHENYL]BENZOATE",
            "non-polymer",
        ),
        "HIF": ("FE(III)-(4-MESOPORPHYRINONE)", "non-polymer"),
        "HIG": (
            "2-ETHYL-5,7-DIMETHYL-3-{(1S)-5-[2-(1H-TETRAZOL-5-YL)PHENYL]-2,3-DIHYDRO-1H-INDEN-1-YL}-3H-IMIDAZO[4,5-B]PYRIDINE",
            "non-polymer",
        ),
        "HII": ("2-METHYL-3-(2-AMINOTHIAZOLO)PROPANAL", "non-polymer"),
        "A92": (
            "2-{2-FLUORO-4-[(2S)-PIPERIDIN-2-YL]PHENYL}-1H-BENZIMIDAZOLE-7-CARBOXAMIDE",
            "non-polymer",
        ),
        "NHY": ("N-HYDROXYUREA", "non-polymer"),
        "NHX": (
            "N-{(2R)-2-[2-(HYDROXYAMINO)-2-OXOETHYL]-4-METHYLPENTANOYL}-3-NAPHTHALEN-2-YL-L-ALANYL-L-ALANINAMIDE",
            "non-polymer",
        ),
        "HIN": (
            "(2S) N-ACETYL-L-ALANYL-ALPHAL-PHENYLALANYL-CHLOROETHYLKETONE",
            "non-polymer",
        ),
        "HIO": ("N-HYDROXY-N-ISOPROPYLOXAMIC ACID", "non-polymer"),
        "HI0": (
            "(R)-N-(7-(4-((4'-CHLOROBIPHENYL-2-YL)METHYL)PIPERAZIN-1-YL)QUINAZOLIN-4-YL)-4-(4-(DIMETHYLAMINO)-1-(PHENYLTHIO)BUTAN-2-YLAMINO)-3-NITROBENZENESULFONAMIDE",
            "non-polymer",
        ),
        "HI1": (
            "METHYL [(1S)-1-({2-[(4R)-4-BENZYL-4-HYDROXY-5-{[(1S,2R)-2-HYDROXY-2,3-DIHYDRO-1H-INDEN-1-YL]AMINO}-5-OXOPENTYL]-2-(4-BROMOBENZYL)HYDRAZINO}CARBONYL)-2,2-DIMETHYLPROPYL]CARBAMATE ",
            "non-polymer",
        ),
        "HI3": (
            "N-(CYCLOPROPYLSULFONYL)-1-(2,5-DIFLUOROBENZYL)-6-FLUORO-5-METHYL-3-(2-OXO-1,2-DIHYDROPYRIDIN-3-YL)-1H-INDOLE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "HI4": (
            "5-ETHYL-1-(2-FLUORO-5-NITROBENZYL)-3-(2-OXO-1,2-DIHYDROPYRIDIN-3-YL)-1H-INDOLE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "HI5": (
            "2-(BUTYRYLOXY)-1-{[(TETRAHYDROXYPHOSPHORANYL)OXY]METHYL}ETHYL BUTYRATE",
            "non-polymer",
        ),
        "HI6": (
            "4-(AMINOCARBONYL)-1-[({2-[(E)-(HYDROXYIMINO)METHYL]PYRIDINIUM-1-YL}METHOXY)METHYL]PYRIDINIUM",
            "non-polymer",
        ),
        "YFE": ("AMORFRUTIN 1", "non-polymer"),
        "YFF": (
            "2-{[(2S)-1-AMINO-5-{[(1Z)-2-FLUOROETHANIMIDOYL]AMINO}-1-OXOPENTAN-2-YL]CARBAMOYL}BENZOIC ACID",
            "non-polymer",
        ),
        "OFO": ("HYDROXY DIIRON-OXO MOIETY", "non-polymer"),
        "NH5": (
            "N-CYCLOHEXYL-N~3~-[2-(3,4-DICHLOROPHENYL)ETHYL]-N-(2-{[2-(5-HYDROXY-3-OXO-3,4-DIHYDRO-2H-1,4-BENZOXAZIN-8-YL)ETHYL]AMINO}ETHYL)-BETA-ALANINAMIDE",
            "non-polymer",
        ),
        "NH4": ("AMMONIUM ION", "non-polymer"),
        "NH7": (
            "2-[(S)-(4-CHLOROPHENYL)(HYDROXY)(1-METHYL-1H-IMIDAZOL-5-YL)METHYL]-N-MORPHOLIN-4-YL-7-PHENYL-1-BENZOFURAN-5-CARBOXAMIDE",
            "non-polymer",
        ),
        "LI7": ("(3E)-3-[(4-HYDROXYPHENYL)IMINO]-1H-INDOL-2(3H)-ONE", "non-polymer"),
        "NH1": (
            "3-(4-AMINO-2-TERT-BUTYL-5-METHYL-PHENYLSULFANYL)-6-CYCLOPENTYL-4-HYDROXY-6-[2-(4-HYDROXY-PHENYL)-ETHYL]-5,6-DIHYDRO-PYRAN-2-ONE",
            "non-polymer",
        ),
        "GYP": ("METHYL-ALPHA-D-GLUCOPYRANOSIDE", "saccharide"),
        "NH3": ("AMMONIA", "non-polymer"),
        "NH2": ("AMINO GROUP", "non-polymer"),
        "R17": (
            "ALLYL-{4-[3-(4-BROMO-PHENYL)-BENZOFURAN-6-YLOXY]-BUT-2-ENYL}-METHYL-AMINE",
            "non-polymer",
        ),
        "NH8": (
            "3-{2-[(S)-(4-CYANOPHENYL)(HYDROXY)(1-METHYL-1H-IMIDAZOL-5-YL)METHYL]-5-NITRO-1-BENZOFURAN-7-YL}BENZONITRILE",
            "non-polymer",
        ),
        "B19": (
            "N-{[(3AS,5AR,8AR,8BS)-2,2,7,7-TETRAMETHYLTETRAHYDRO-3AH-BIS[1,3]DIOXOLO[4,5-B:4',5'-D]PYRAN-3A-YL]METHYL}SULFAMIDE",
            "non-polymer",
        ),
        "A09": ("N-HYDROXYBENZENESULFONAMIDE", "non-polymer"),
        "YFB": ("AMORFRUTIN B", "non-polymer"),
        "A02": ("ARC-1012", "peptide"),
        "A03": (
            "(10R,20R,23R)-1-[(2S,3S,4R,5R)-5-(6-AMINO-9H-PURIN-9-YL)-3,4-DIHYDROXYTETRAHYDROFURAN-2-YL]-20,23-BIS(3-CARBAMIMIDAMIDOPROPYL)-10-METHYL-1,8,11,18,21-PENTAOXO-2,9,12,19,22-PENTAAZATETRACOSAN-24-AMIDE",
            "peptide",
        ),
        "A00": (
            "N-[(5S)-5-{[(4-AMINOPHENYL)SULFONYL](ISOBUTYL)AMINO}-6-HYDROXYHEXYL]-NALPHA-(METHOXYCARBONYL)-BETA-PHENYL-L-PHENYLALANINAMIDE",
            "non-polymer",
        ),
        "A01": (
            "[(2,6-DIMETHOXYPHENYL)CARBONYLAMINO]METHYL-TRIHYDROXY-BORON",
            "non-polymer",
        ),
        "A06": (
            "(3S)-4-(5-CHLORO-1H-BENZIMIDAZOL-2-YL)-3-(4-CHLOROPHENYL)BUTANOIC ACID",
            "non-polymer",
        ),
        "A07": ("N-(6-CHLORO-5-PHENYL-1H-INDAZOL-3-YL)BUTANAMIDE", "non-polymer"),
        "A04": ("5-[2-(TRIFLUOROMETHOXY)PHENYL]-2-FUROIC ACID", "non-polymer"),
        "A05": ("5-(2-CHLORO-4-NITROPHENYL)-2-FUROIC ACID", "non-polymer"),
        "LI0": (
            "4-(4-BENZYL-4-METHOXYPIPERIDIN-1-YL)-N-[(4-{[1,1-DIMETHYL-2-(PHENYLTHIO)ETHYL]AMINO}-3-NITROPHENYL)SULFONYL]BENZAMIDE",
            "non-polymer",
        ),
        "EGT": ("2-(3,4,5-TRIHYDROXY-PHENYL)-CHROMAN-3,5,7-TRIOL", "non-polymer"),
        "EGM": ("4-ETHYLBENZOIC ACID", "non-polymer"),
        "EGL": ("ETHYLENE GLYCOL", "non-polymer"),
        "EGC": (
            "2-(2-{2-[2-(2-{2-[2-(2-{2-[4-(1,1,3,3-TETRAMETHYL-BUTYL)-PHENOXY]-ETHOXY}-ETHOXY)-ETHOXY]-ETHOXY}-ETHOXY)-ETHOXY]-ETHOXY}-ETHOXY)-ETHANOL",
            "non-polymer",
        ),
        "EGB": (
            "(4S)-4-(3-HYDROXYPHENYL)-7,7-DIMETHYL-2-THIOXO-2,3,4,6,7,8-HEXAHYDROQUINAZOLIN-5(1H)-ONE",
            "non-polymer",
        ),
        "EGG": (
            "(2R,3R)-5,7-DIHYDROXY-2-(3,4,5-TRIHYDROXYPHENYL)-3,4-DIHYDRO-2H-CHROMEN-3-YL 3,4,5-TRIHYDROXYBENZOATE",
            "non-polymer",
        ),
        "EGD": ("N-ETHYLGUANIDINE", "non-polymer"),
        "02E": (
            "(E,1S)-1-[DIHYDROXY(DIPHENOXY)-$L^{5}-PHOSPHANYL]-4-METHOXY-BUT-3-EN-1-AMINE",
            "non-polymer",
        ),
        "02D": ("5-(AMINOMETHYL)THIOPHENE-2-CARBOXIMIDAMIDE", "non-polymer"),
        "02G": ("(3S,4E)-3-HYDROXY-7-SULFANYLHEPT-4-ENOIC ACID", "non-polymer"),
        "02F": ("[(1S)-1-AMINO-4-METHOXYBUTYL]PHOSPHONIC ACID", "non-polymer"),
        "02A": ("(2S)-AZETIDINE-2-CARBOXYLIC ACID", "peptide"),
        "LBT": ("ALPHA-LACTOSE", "saccharide"),
        "02C": ("2-(AMINOMETHYL)-1,3-THIAZOLE-4-CARBOXIMIDAMIDE", "non-polymer"),
        "02B": ("4-METHOXY-2,3,6-TRIMETHYLBENZENESULFONIC ACID", "non-polymer"),
        "LBY": ("N~6~-(TERT-BUTOXYCARBONYL)-L-LYSINE", "peptide"),
        "02L": ("N-(2,2-DIMETHYLPROPYL)-L-ASPARAGINE", "peptide"),
        "02O": ("3-(3-TERT-BUTYL-1,2,4-OXADIAZOL-5-YL)-L-ALANINE", "peptide"),
        "02N": ("1-(4-METHYLPHENYL)METHANAMINE", "non-polymer"),
        "02I": (
            "(6S,7S,8S,10R)-4-AMINO-8-HYDROXY-7,8,9,10-TETRAHYDRO-6H-7,10-EPOXYAZEPINO[1,2-E]PURIN-6-YL DIHYDROGEN PHOSPHATE",
            "nucleic-acid",
        ),
        "02H": (
            "(2S,3S,5S)-2,5-DIAMINO-6-PHENYL-1-[4-(PYRIDIN-3-YL)PHENYL]HEXAN-3-OL",
            "non-polymer",
        ),
        "02K": ("1-AMINOCYCLOHEXANECARBOXYLIC ACID", "peptide"),
        "02J": ("5-METHYL-1,2-OXAZOLE-3-CARBOXYLIC ACID", "peptide"),
        "02U": ("14-METHYLHEXADEC-9-ENOIC ACID", "non-polymer"),
        "BF2": ("BERYLLIUM DIFLUORIDE", "non-polymer"),
        "02W": ("2-(2-METHYL-1H-BENZIMIDAZOL-1-YL)ETHANAMINE", "non-polymer"),
        "BF0": (
            "METHYL 4-{[(5BETA,6ALPHA)-17-(CYCLOPROPYLMETHYL)-3,14-DIHYDROXY-4,5-EPOXYMORPHINAN-6-YL]AMINO}-4-OXOBUTANOATE",
            "non-polymer",
        ),
        "02Q": ("3-(2-METHYLPHENYL)PROPANOIC ACID", "non-polymer"),
        "02P": (
            "D-PHENYLALANYL-N-[(4-CHLORO-1-METHYLPYRIDINIUM-2-YL)METHYL]-L-PROLINAMIDE",
            "peptide",
        ),
        "BF5": (
            "(3S)-10-(4-AMINOPIPERAZIN-1-YL)-9-FLUORO-7-HYDROXY-3-METHYL-2,3-DIHYDRO-8H-[1,4]OXAZINO[2,3,4-IJ]QUINOLINE-6-CARBOXYLATE",
            "non-polymer",
        ),
        "02R": ("1-(2-CHLOROPHENYL)METHANAMINE", "non-polymer"),
        "14B": ("2-AMINO-3,5-DIMETHYLBENZOIC ACID", "non-polymer"),
        "FGG": (
            "(2Z,6E,10E)-2-FLUORO-3,7,11,15-TETRAMETHYLHEXADECA-2,6,10,14-TETRAEN-1-YL TRIHYDROGEN DIPHOSPHATE",
            "non-polymer",
        ),
        "02Y": ("6-DIAZONIO-5-OXO-L-NORLEUCINE", "peptide"),
        "02X": ("(2S)-2-(PROPAN-2-YL)BUTANEDIOIC ACID", "non-polymer"),
        "14C": (
            "2-(2-CHLOROPYRIDIN-4-YL)-4-METHYL-1H-ISOINDOLE-1,3(2H)-DIONE",
            "non-polymer",
        ),
        "BFC": ("(R)-3-HYDROXYTETRADECANAL", "non-polymer"),
        "BFB": (
            "N-[(1S)-1-(AMINOCARBONYL)-4-(ETHANIMIDOYLAMINO)BUTYL]BENZAMIDE",
            "non-polymer",
        ),
        "BFA": (
            "9-BROMO-PHENAZINE-1-CARBOXYLIC ACID (2-DIMETHYLAMINO-ETHYL)-AMIDE",
            "non-polymer",
        ),
        "LB2": ("3-O-BETA-D-GLUCOPYRANOSYL-BETA-D-GLUCOPYRANOSE", "saccharide"),
        "BFG": ("3,3-DIPHENYL-N-(2-SULFANYLETHYL)PROPANAMIDE", "non-polymer"),
        "BFF": (
            "3-[(4-BROMO-2,6-DIFLUOROBENZYL)OXY]-5-{[(4-PYRROLIDIN-1-YLBUTYL)CARBAMOYL]AMINO}ISOTHIAZOLE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "BFE": ("N-(2-SULFANYLETHYL)BENZAMIDE", "non-polymer"),
        "BFD": ("ASPARTATE BERYLLIUM TRIFLUORIDE", "peptide"),
        "FG4": (
            "{2-[(1-BENZOFURAN-2-YLMETHOXY)CARBONYL]-5-METHOXY-1H-INDOL-1-YL}ACETIC ACID",
            "non-polymer",
        ),
        "FG5": (
            "(2-{[(1-BENZOFURAN-2-YLSULFONYL)AMINO]METHYL}-5-METHOXY-1H-INDOL-1-YL)ACETIC ACID",
            "non-polymer",
        ),
        "BFI": (
            "2[4-BROMO-2-FLUOROPHENYL)METHYL]-6-FLUOROSPIRO[ISOQUINOLINE-4-(1H),3'-PYRROLIDINE]-1,2',3,5'(2H)-TETRONE",
            "non-polymer",
        ),
        "BFH": ("1-(CARBOXYMETHYL)-1H-BENZO[G]INDOLE-2-CARBOXYLIC ACID", "non-polymer"),
        "BFO": (
            "BENZYL N-[(NAPHTHALEN-2-YLMETHOXY)CARBONYL]-L-ALANYL-N~5~-[(2R,3S,4S)-3-FORMYL-2-HYDROXY-4-METHYLHEXANOYL]-L-ORNITHINATE",
            "non-polymer",
        ),
        "BFN": (
            "5-(ACETYLAMINO)-9-(BENZOYLAMINO)-3,5,9-TRIDEOXY-3-FLUORO-D-ERYTHRO-ALPHA-L-MANNO-NON-2-ULOPYRANOSONIC ACID",
            "non-polymer",
        ),
        "FG2": (
            "5-METHOXY-N-[(5-METHYLPYRIDIN-2-YL)SULFONYL]-1H-INDOLE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "BFL": ("?", "non-polymer"),
        "BFS": ("N-[1-(4-BROMOPHENYL)ETHYL]-5-FLUORO SALICYLAMIDE", "non-polymer"),
        "024": (
            "4-BROMO-3-(CARBOXYMETHOXY)-5-[3-(CYCLOHEXYLAMINO)PHENYL]THIOPHENE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "027": (
            "7-(4-METHYLPIPERAZIN-1-YL)-4-[(5-METHYL-1H-PYRAZOL-3-YL)AMINO]-2-(PROPAN-2-YL)PHTHALAZIN-1(2H)-ONE",
            "non-polymer",
        ),
        "BFP": ("BETA FRUCTOSE 1,6-DIPHOSPHATE", "saccharide"),
        "021": ("N-CYCLOPENTYLGLYCINE", "peptide"),
        "020": (
            "N-(FURAN-2-YLMETHYL)-2-NITRO-4-(TRIFLUOROMETHYL)ANILINE",
            "non-polymer",
        ),
        "BFU": (
            "1-(5-BROMO-PYRIDIN-2-YL)-3-[2-(6-FLUORO-2-HYDROXY-3-PROPIONYL-PHENYL)-CYCLOPROPYL]-UREA",
            "non-polymer",
        ),
        "14O": ("5-FLUORO-1H-INDOLE", "non-polymer"),
        "BFZ": (
            "7-BROMO-1-[2-(DIETHYLAMINO)ETHYL]-5-(2-FLUOROPHENYL)-1,3-DIHYDRO-2H-1,4-BENZODIAZEPIN-2-ONE",
            "non-polymer",
        ),
        "BFX": (
            "4-{4-[3-(2-BROMO-5-FLUOROPHENOXY)PROPYL]PHENYL}-N-(2-CHLOROBENZYL)-N-CYCLOPROPYL-1,2,5,6-TETRAHYDROPYRIDINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "029": ("CIS-4-(CARBAMIMIDAMIDOMETHYL)CYCLOHEXANECARBOXYLIC ACID", "peptide"),
        "14L": (
            "(11S,12S,13S)-11,12,13,14-TETRAHYDRONAPHTHO[1,2,3,4-PQR]TETRAPHENE-11,12,13-TRIOL",
            "non-polymer",
        ),
        "RA8": (
            "N-(BENZYLSULFONYL)-L-LEUCYL-N-(4-{[AMINO(IMINO)METHYL]AMINO}BUTYL)-L-PROLINAMIDE",
            "peptide",
        ),
        "VS1": (
            "3-[[N-[MORPHOLIN-N-YL]-CARBONYL]-PHENYLALANINYL-AMINO]-5- PHENYL-PENTANE-1-SULFONYLBENZENE",
            "non-polymer",
        ),
        "14M": (
            "{(3S)-1-[5-(CYCLOHEXYLCARBAMOYL)-6-(PROPYLSULFANYL)PYRIDIN-2-YL]PIPERIDIN-3-YL}ACETIC ACID",
            "non-polymer",
        ),
        "SRD": ("D-1-NAPHTHYL-2-ACETAMIDO-ETHANE BORONIC ACID", "non-polymer"),
        "EG3": (
            "PHENYLALANYLAMINODI(ETHYLOXY)ETHYL BENZENESULFONAMIDEAMINOCARBONYLBENZENESULFONAMIDE",
            "non-polymer",
        ),
        "EG2": ("AMINODI(ETHYLOXY)ETHYLAMINOCARBONYLBENZENESULFONAMIDE", "non-polymer"),
        "EG1": (
            "AMINOMETHYLENECARBONYLAMINODI(ETHYLOXY)ETHYLAMINOCARBONYLBENZENESULFONAMIDE",
            "non-polymer",
        ),
        "0U3": (
            "N-HYDROXY-2-{4-[3-(4-{[(2S)-2-HYDROXY-3,3-DIMETHYLBUTYL]OXY}-3-METHYLPHENYL)PENTAN-3-YL]-2-METHYLPHENOXY}ACETAMIDE",
            "non-polymer",
        ),
        "OCH": ("QUINOLIN-2(1H)-ONE", "non-polymer"),
        "SRE": (
            "(1S,4S)-4-(3,4-DICHLOROPHENYL)-N-METHYL-1,2,3,4-TETRAHYDRONAPHTHALEN-1-AMINE",
            "non-polymer",
        ),
        "708": (
            "3-[(E)-2-(2-CHLORO-4-{[3-{[(R)-(2,6-DICHLOROPHENYL)(HYDROXY)-LAMBDA~4~-SULFANYL]METHYL}-5-(1-METHYLETHYL)ISOXAZOL-4-YL]METHOXY}PHENYL)ETHENYL]BENZOIC ACID",
            "non-polymer",
        ),
        "709": (
            "N-(7-CARBAMIMIDOYL-NAPHTHALEN-1-YL)-3-HYDROXY-2-METHYL-BENZAMIDE",
            "non-polymer",
        ),
        "OCO": ("COBALT ION,6 WATERS COORDINATED", "non-polymer"),
        "147": ("1-O-[P-NITROPHENYL]-BETA-D-GALACTOPYRANOSE", "saccharide"),
        "OCA": ("OCTANOIC ACID (CAPRYLIC ACID)", "non-polymer"),
        "705": (
            "N-[(E)-3-[(2R,3S,4R,5R)-3,4-DIHYDROXY-5-(6-PROPYLAMINOPURIN-9-YL)OXOLAN-2-YL]PROP-2-ENYL]-5-(4-FLUOROPHENYL)-2,3-DIHYDROXY-BENZAMIDE",
            "non-polymer",
        ),
        "706": (
            "N-(3,3-DIMETHYL-2,3-DIHYDRO-1H-INDOL-6-YL)-2-[(PYRIDIN-4-YLMETHYL)AMINO]PYRIDINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "OCB": ("OCTANOYLCARNITINE", "non-polymer"),
        "700": (
            "[5-CHLORO-1H-INDOL-2-CARBONYL-PHENYLALANINYL]-AZETIDINE-3-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "701": (
            "(2R)-2-(5-CHLORO-2-THIENYL)-N-{(3S)-1-[(1S)-1-METHYL-2-MORPHOLIN-4-YL-2-OXOETHYL]-2-OXOPYRROLIDIN-3-YL}PROPENE-1-SULFONAMIDE",
            "non-polymer",
        ),
        "702": (
            "(1R)-4-[3-(2-BENZYLPHENOXY)PHENYL]-1-PHOSPHONOBUTANE-1-SULFONIC ACID",
            "non-polymer",
        ),
        "703": (
            "5-(3-MORPHOLIN-4-YL-PROPYL)-2-(3-NITRO-PHENYL)-4-THIOXO-4,5-DIHYDRO-1-THIA-3B,5-DIAZA-CYCLOPENTA[A]PENTALEN-6-ONE",
            "non-polymer",
        ),
        "OCY": ("HYDROXYETHYLCYSTEINE", "peptide"),
        "OCX": ("(5E,11E,14E)-8-OXOICOSA-5,9,11,14-TETRAENOIC ACID", "non-polymer"),
        "THD": (
            "2-[3-[(4-AMINO-2-METHYL-5-PYRIMIDINYL)METHYL]-2-(1,2-DIHYDROXYETHYL)-4-METHYL-1,3-THIAZOL-3-IUM-5-YL]ETHYL TRIHYDROGEN DIPHOSPHATE",
            "non-polymer",
        ),
        "145": ("1-O-[O-NITROPHENYL]-BETA-D-GALACTOPYRANOSE", "non-polymer"),
        "IFA": ("?", "non-polymer"),
        "DRP": ("2-DEOXYRIBOFURANOSYL-PYRIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "IFC": (
            "N-[2-METHYL-5-({[3-(4-METHYL-1H-IMIDAZOL-1-YL)-5-(TRIFLUOROMETHYL)PHENYL]CARBONYL}AMINO)PHENYL]ISOXAZOLE-5-CARBOXAMIDE",
            "non-polymer",
        ),
        "IFL": (
            "(3S,4R,5R)-3,4-DIHYDROXY-5-(HYDROXYMETHYL)PIPERIDIN-2-ONE",
            "non-polymer",
        ),
        "142": ("CARBIDOPA", "non-polymer"),
        "OCS": ("CYSTEINESULFONIC ACID", "peptide"),
        "OCR": ("(5E,8E,11Z,13E)-15-OXOICOSA-5,8,11,13-TETRAENOIC ACID", "non-polymer"),
        "OCT": ("N-OCTANE", "non-polymer"),
        "OCV": (
            "N6-[(1R)-2-{[(1R)-1-CARBOXY-2-METHYLPROPYL]OXY}-1-(MERCAPTOMETHYL)-2-OXOETHYL]-6-OXO-D-LYSINE",
            "non-polymer",
        ),
        "CEG": ("4,6-O-(1-CARBOXYETHYLIDENE)-BETA-D-GLUCOSE", "saccharide"),
        "CEF": ("CEFOTAXIME GROUP", "non-polymer"),
        "CEE": ("(2-CHLOROETHYL)BENZENE", "non-polymer"),
        "CED": (
            "5-METHYL-2-[2-OXO-1-(2-THIOPHEN-2-YL-ACETYLAMINO)-ETHYL]-3,6-DIHYDRO-2H-[1,3]THIAZINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "CEC": ("CHLOROETHYLCARBAMOYL GROUP", "non-polymer"),
        "140": ("N-PALMITOYLGLYCINE", "non-polymer"),
        "CEA": ("S-HYDROXY-CYSTEINE", "peptide"),
        "CEO": (
            "5-METHYLENE-2-[2-OXO-1-(2-THIOPHEN-2-YL-ACETYLAMINO)-ETHYL]-5,6-DIHYDRO-2H-[1,3]THIAZINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "CEN": ("7BETA-(4CARBOXYBUTANAMIDO) CEPHALOSPORANIC ACID", "non-polymer"),
        "CEM": (
            "N-(1-CARBOXY-2-HYDROXY-4-OXO-BUTYL)-N-(3-OXO-CISPROPENYL)AMINE",
            "non-polymer",
        ),
        "141": ("PYRAZOLO[3,4-D]PYRIMIDINE-4,6-DIONE", "non-polymer"),
        "CEK": (
            "(9AS)-8-ACETYL-1,7-DIHYDROXY-3-METHOXY-9A-METHYL-N-(1-NAPHTHYLMETHYL)-9-OXO-9,9A-DIHYDRODIBENZO[B,D]FURAN-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "CEJ": ("CYCLOPENTANE-1,3-DIONE", "non-polymer"),
        "CEI": (
            "N-[3-BENZYL-5-(4-HYDROXYPHENYL)PYRAZIN-2-YL]-2-(4-HYDROXYPHENYL)ACETAMIDE",
            "non-polymer",
        ),
        "CEH": (
            "5-{3-(S)-(4-(R)-ACETYLAMINO-4-CARBOXY-BUTYRYLAMINO)-3-[1-(R)-(1-(R)-CARBOXY-ETHYLCARBAMOYL)-ETHYLCARBAMOYL]-PROPYL}-2-(CARBOXY-PHENYLACETYLAMINO-METHYL)-3,6-DIHYDRO-2H-[1,3]THIAZINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "CEW": (
            "(2R)-2-[(1R)-1-({[(2R)-2-AMINO-2,3-DIHYDRO-1,3-THIAZOL-4-YL](METHOXYIMINO)ACETYL}AMINO)-2-OXOETHYL]-5-METHYL-3,6-DIHYDRO-2H-1,3-THIAZINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "CEV": (
            "ETHYL (4R)-4-AMINO-5-[(3S)-2-OXOPYRROLIDIN-3-YL]PENTANOATE",
            "non-polymer",
        ),
        "CET": (
            "2-CHLORO-4-ETHYLAMINO-6-(R(+)-2'-CYANO-4-BUTYLAMINO)-1,3,5-TRIAZINE",
            "non-polymer",
        ),
        "CES": (
            "2-[CARBOXY-(2-FURAN-2-YL-2-METHOXYIMINO-ACETYLAMINO)-METHYL]-5-METHYL-3,6-DIHYDRO-2H-[1,3]THIAZINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "CER": (
            "(2S, 3R)-3-HYDROXY-4-OXO-7,10-TRANS,TRANS-DODECADIENAMIDE",
            "non-polymer",
        ),
        "CEQ": ("ETHYL-TRIMETHYL-SILANE", "non-polymer"),
        "CEP": ("CEPHALOTHIN GROUP", "non-polymer"),
        "O40": (
            "4-{[METHYL(3-{[7-PROPYL-3-(TRIFLUOROMETHYL)-1,2-BENZISOXAZOL-6-YL]OXY}PROPYL)CARBAMOYL]AMINO}BENZOIC ACID",
            "non-polymer",
        ),
        "CEY": (
            "ALPHA-D-GLUCOPYRANOSYL-(1->4)-ALPHA-D-GLUCOPYRANOSYL-(1->4)-ALPHA-D-GLUCOPYRANOSYL-(1->4)-ALPHA-D-GLUCOPYRANOSYL-(1->4)-ALPHA-D-GLUCOPYRANOSYL-(1->4)-ALPHA-D-GLUCOPYRANOSE",
            "non-polymer",
        ),
        "CEX": (
            "ALPHA-D-GLUCOPYRANOSYL-(1->4)-ALPHA-D-GLUCOPYRANOSYL-(1->4)-ALPHA-D-GLUCOPYRANOSYL-(1->4)-ALPHA-D-GLUCOPYRANOSYL-(1->4)-ALPHA-D-GLUCOPYRANOSE",
            "non-polymer",
        ),
        "PK2": (
            "6-AMINO-4-{2-[(CYCLOHEXYLMETHYL)AMINO]ETHYL}-2-(METHYLAMINO)-1,7-DIHYDRO-8H-IMIDAZO[4,5-G]QUINAZOLIN-8-ONE",
            "non-polymer",
        ),
        "PK3": (
            "6-AMINO-4-(2-HYDROXYETHYL)-2-(METHYLAMINO)-3,7-DIHYDRO-8H-IMIDAZO[4,5-G]QUINAZOLIN-8-ONE",
            "non-polymer",
        ),
        "PK0": (
            "4-[(4R)-4-(4-HYDROXYPHENYL)-3,4-DIMETHYL-2,5-DIOXOIMIDAZOLIDIN-1-YL]-2-(TRIFLUOROMETHYL)BENZONITRILE",
            "non-polymer",
        ),
        "PK1": (
            "(5R)-3-(3,4-DICHLOROPHENYL)-5-(4-HYDROXYPHENYL)-1,5-DIMETHYL-2-THIOXOIMIDAZOLIDIN-4-ONE",
            "non-polymer",
        ),
        "O4M": ("MANGANESE TETRAHYDRATE ION", "non-polymer"),
        "O4N": (
            "5-[[[2-[[(1S)-1-(4-METHOXYPHENYL)BUTYL]CARBAMOYL]PHENYL]METHYL-PROP-2-ENYL-AMINO]METHYL]-1,3-BENZODIOXOLE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "THB": ("TETRAHYDROBIOPTERIN", "non-polymer"),
        "CE6": ("CELLOHEXAOSE", "non-polymer"),
        "CE5": ("CELLOPENTAOSE", "non-polymer"),
        "CE3": (
            "(6S,7S)-3-[(ACETYLOXY)METHYL]-7-{[(2E)-2-(2-AMINO-1,3-THIAZOL-4-YL)-2-(METHOXYIMINO)ETHANOYL]AMINO}-8-OXO-5-THIA-1-AZABICYCLO[4.2.0]OCT-2-ENE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "148": (
            "2-[BIS-(2-HYDROXY-ETHYL)-AMINO]-2-HYDROXYMETHYL-PROPANE-1,3-DIOL",
            "non-polymer",
        ),
        "CE1": ("O-DODECANYL OCTAETHYLENE GLYCOL", "non-polymer"),
        "CE0": (
            "(2R)-2-[(1R)-2-[(2S)-2-AMINO-2-CARBOXYETHOXY]-1-{[(2R)-2-[(4S)-2-AMINO-4,5-DIHYDRO-1,3-THIAZOL-4-YL]-2-(METHOXYAMINO)ACETYL]AMINO}-2-OXOETHYL]-5-METHYL-3,6-DIHYDRO-2H-1,3-THIAZINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "0UC": (
            "4-BROMO-3-[5-(3,5-DICHLORO-4-HYDROXYPHENYL)-1,3,4-OXADIAZOL-2-YL]BENZENESULFONYL FLUORIDE",
            "non-polymer",
        ),
        "NEO": ("NEOPTERIN", "non-polymer"),
        "ZAT": ("4,4'-DIACETYLDIPHENYLUREA-BIS(GUANYLHYDRAZONE)", "non-polymer"),
        "CE9": ("DODECYL NONA ETHYLENE GLYCOL ETHER", "non-polymer"),
        "CE8": ("CELLOOCTAOSE", "non-polymer"),
        "70U": (
            "5-(O-METHYLACETO)-2-THIO-2-DEOXY-URIDINE-5'-MONOPHOSPHATE",
            "nucleic-acid",
        ),
        "OC9": ("OCTAN-1-OL", "non-polymer"),
        "OC8": ("CALCIUM ION, 8 WATERS COORDINATED", "non-polymer"),
        "OC1": ("CALCIUM ION, 1 WATER COORDINATED", "non-polymer"),
        "UN3": ("(2-METHYL-5-PHENYL-2H-PYRAZOL-3-YL)-SULFAMIC ACID", "non-polymer"),
        "OC3": ("CALCIUM ION, 3 WATERS COORDINATED", "non-polymer"),
        "OC2": ("CALCIUM ION, 2 WATERS COORDINATED", "non-polymer"),
        "F8A": (
            "9-[2-(TRIFLUOROMETHYL)BENZYL]-2,3,4,9-TETRAHYDRO-1H-CARBAZOLE-8-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "OC4": ("CALCIUM ION, 4 WATERS COORDINATED", "non-polymer"),
        "OC7": ("CALCIUM ION, 7 WATERS COORDINATED", "non-polymer"),
        "OC6": ("CALCIUM ION, 6 WATERS COORDINATED", "non-polymer"),
        "MUU": (
            "N-[(1S,2R)-1-BENZYL-3-{(CYCLOPROPYLMETHYL)[(3-METHOXYPHENYL)SULFONYL]AMINO}-2-HYDROXYPROPYL]-N'-METHYLSUCCINAMIDE",
            "non-polymer",
        ),
        "MUT": (
            "(5S)-3-(3-ACETYLPHENYL)-N-[(1S,2R)-1-BENZYL-2-HYDROXY-3-{ISOBUTYL[(4-METHOXYPHENYL)SULFONYL]AMINO}PROPYL]-2-OXO-1,3-OXAZOLIDINE-5-CARBOXAMIDE",
            "non-polymer",
        ),
        "MUW": (
            "N-{(5S)-5-[{[4-(AMINOMETHYL)PHENYL]SULFONYL}(ISOBUTYL)AMINO]-6-HYDROXYHEXYL}-NALPHA-(METHOXYCARBONYL)-BETA-PHENYL-L-PHENYLALANINAMIDE",
            "non-polymer",
        ),
        "MUV": (
            "N-{(1S,2R)-1-BENZYL-3-[(CYCLOPROPYLMETHYL)(2-FURYLSULFONYL)AMINO]-2-HYDROXYPROPYL}-N'-METHYLSUCCINAMIDE",
            "non-polymer",
        ),
        "MUP": ("METHOXYUNDECYLPHOSPHINIC ACID", "non-polymer"),
        "1X5": (
            "2-[(CYCLOPROPYLCARBONYL)AMINO]-5-[METHYL(PYRIDIN-3-YLMETHYL)AMINO]BENZOIC ACID",
            "non-polymer",
        ),
        "MUR": ("MURAMIC ACID", "non-polymer"),
        "MUD": (
            "N-{[(1R)-1-CARBOXY-2-(METHYLSULFANYL)ETHYL]CARBAMOYL}-L-GLUTAMIC ACID",
            "non-polymer",
        ),
        "MUG": ("4-METHYLUMBELLIFERYL-ALPHA-D-GLUCOSE", "saccharide"),
        "MUF": (
            "(8ALPHA,10ALPHA,13ALPHA,17BETA)-17-[(4-HYDROXYPHENYL)CARBONYL]ANDROSTA-3,5-DIENE-3-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "MUA": ("9-METHYL URIC ACID", "non-polymer"),
        "MUC": ("[(2S)-5-OXO-2,5-DIHYDROFURAN-2-YL]ACETIC ACID", "non-polymer"),
        "MUB": ("N-ACETYLMURAMIC ACID", "non-polymer"),
        "NEM": ("NE2-METHYLATED HISTIDINE", "peptide"),
        "MUL": ("TIAMULIN", "non-polymer"),
        "MUI": (
            "(5S)-3-(4-ACETYLPHENYL)-N-[(1S,2R)-1-BENZYL-2-HYDROXY-3-{ISOBUTYL[(4-METHOXYPHENYL)SULFONYL]AMINO}PROPYL]-2-OXO-1,3-OXAZOLIDINE-5-CARBOXAMIDE",
            "non-polymer",
        ),
        "MUH": (
            "N-{4-METHYL-3-[(3-PYRIMIDIN-4-YLPYRIDIN-2-YL)AMINO]PHENYL}-3-(TRIFLUOROMETHYL)BENZAMIDE",
            "non-polymer",
        ),
        "MUJ": (
            "5-{3-METHOXY-4-[(4-METHOXYBENZYL)OXY]BENZYL}PYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "ASR": ("4-AMINOPHENYLARSONIC ACID", "non-polymer"),
        "ASQ": ("PHOSPHOASPARTATE", "peptide"),
        "ASP": ("ASPARTIC ACID", "peptide"),
        "ASV": (
            "DELTA-(L-ALPHA-AMINOADIPOYL)-L-CYSTEINYL-D-VINYLGLYCINE",
            "non-polymer",
        ),
        "ASU": ("4'-THIO-2'4'-DIDEOXYRIBOFURANOSE-5'-PHOSPHATE", "nucleic-acid"),
        "AST": ("ARSENITE", "non-polymer"),
        "ASZ": ("2-(DIMETHYLAMINO)ETHANESULFONIC ACID", "non-polymer"),
        "ASY": ("(12E)-10-OXOOCTADEC-12-ENOIC ACID", "non-polymer"),
        "ASX": ("ASP/ASN AMBIGUOUS", "peptide"),
        "NUP": ("?", "non-polymer"),
        "STY": ("TYROSINE-O-SULPHONIC ACID", "peptide"),
        "ASC": ("ASCORBIC ACID", "non-polymer"),
        "ASB": ("ASPARTIC ACID-4-CARBOXYMETHYL ESTER", "peptide"),
        "ASA": ("ASPARTIC ALDEHYDE", "peptide"),
        "ASG": ("2-DEOXY-2-ACETAMIDO-BETA-D-GALACTOSE-4-SULFATE", "saccharide"),
        "ASF": ("3-(BUTYLSULPHONYL)-PROPANOIC ACID", "non-polymer"),
        "ASE": ("N-ACETYL SEROTONIN", "non-polymer"),
        "ASD": ("4-ANDROSTENE-3-17-DIONE", "non-polymer"),
        "ASK": ("DEHYDROXYMETHYLASPARTIC ACID", "peptide"),
        "ASJ": ("(3S)-3-AMINO-4-HYDROXYBUTANOIC ACID", "peptide"),
        "ASI": ("L-ISO-ASPARTATE", "peptide"),
        "ASH": (
            "3-CHLORO-N-(4-MORPHOLIN-4-YLPHENYL)-6-PYRIDIN-3-YLIMIDAZO[1,2-A]PYRAZIN-8-AMINE",
            "non-polymer",
        ),
        "ASO": ("1,5-ANHYDROSORBITOL", "saccharide"),
        "ASN": ("ASPARAGINE", "peptide"),
        "ASM": ("2-AMINO-4-OXO-4(1H-PYRROL-1-YL)BUTANOIC ACID", "peptide"),
        "ASL": ("ASPARTIC ACID-4-CARBOXYETHYL ESTER", "peptide"),
        "AS3": ("4-METHYL-6-PROPYLPYRIDIN-2-AMINE", "non-polymer"),
        "AS2": ("(2R)-2-AMINO-4-OXOBUTANOIC ACID", "peptide"),
        "AS1": ("ARGININOSUCCINATE", "non-polymer"),
        "AS0": (
            "4-[(11BETA,17BETA)-17-METHOXY-17-(METHOXYMETHYL)-3-OXOESTRA-4,9-DIEN-11-YL]BENZALDEHYDE OXIME",
            "non-polymer",
        ),
        "AS7": ("N4-(4-CYANOBENZYL)-L-ASPARAGINE", "peptide"),
        "AS6": (
            "{4-METHYL-3-[(1-METHYLETHYL)(2-{[3-(METHYLSULFONYL)-5-MORPHOLIN-4-YLPHENYL]AMINO}PYRIMIDIN-4-YL)AMINO]PHENYL}METHANOL",
            "non-polymer",
        ),
        "AS5": (
            "2-(ACETYLAMINO)-4-O-[2-(ACETYLAMINO)-2-DEOXY-ALPHA-L-GULOPYRANOSYL]-1,5-ANHYDRO-2-DEOXY-D-MANNITOL",
            "non-polymer",
        ),
        "AS4": ("ALDOSTERONE", "non-polymer"),
        "DUA": ("2',5'-DIDEOXY-5'-(TRITYLAMINO)URIDINE", "non-polymer"),
        "AS9": ("N-[HYDROXY(METHYL)PHOSPHORYL]-L-ASPARTIC ACID", "peptide"),
        "AS8": ("ACETYLSULFATE", "non-polymer"),
        "DUD": ("DEOXYURIDINE-5'-DIPHOSPHATE", "non-polymer"),
        "23N": ("(2R)-2-BENZYL-3-NITROPROPANOIC ACID", "non-polymer"),
        "NU1": ("8-HYDROXY-2-METHYL-3-HYDRO-QUINAZOLIN-4-ONE", "non-polymer"),
        "DUX": ("2,3-DEOXY-3-FLUORO-5-O-TRITYLURIDINE", "non-polymer"),
        "DUZ": ("?", "nucleic-acid"),
        "R": ("?", "nucleic-acid"),
        "DUP": ("?", "non-polymer"),
        "DUQ": (
            "(2S)-2-[(2,4-DIOXOPYRIMIDIN-1-YL)METHYL]-N-(2-HYDROXYETHYL)-4-TRITYLOXY-BUTANAMIDE",
            "non-polymer",
        ),
        "DUR": ("2'-DEOXYURIDINE", "non-polymer"),
        "DUT": ("DEOXYURIDINE-5'-TRIPHOSPHATE", "non-polymer"),
        "BA3": ("BIS(ADENOSINE)-5'-TRIPHOSPHATE", "non-polymer"),
        "MU1": (
            "N~2~-ACETYL-N-[(1S,3S,4S)-4-({[(5S)-3-(3-ACETYLPHENYL)-2-OXO-1,3-OXAZOLIDIN-5-YL]CARBONYL}AMINO)-1-BENZYL-3-HYDROXY-5-PHENYLPENTYL]-L-VALINAMIDE",
            "non-polymer",
        ),
        "MU0": (
            "(5S)-N-[(1S,2S,4S)-1-BENZYL-2-HYDROXY-4-{[(2S)-3-METHYL-2-(2-OXOTETRAHYDROPYRIMIDIN-1(2H)-YL)BUTANOYL]AMINO}-5-PHENYLPENTYL]-2-OXO-3-PHENYL-1,3-OXAZOLIDINE-5-CARBOXAMIDE",
            "non-polymer",
        ),
        "MU2": ("N-ACETYLMURAMYL-L-ALANYL-D-ISOGLUTAMINE", "non-polymer"),
        "T16": ("BOROLOG2", "non-polymer"),
        "NCP": ("CIS-DIAMINODICHLOROPLATINUM", "non-polymer"),
        "A9D": (
            "(2S)-2-AMINO-2-CARBOXYETHYL 1-CARBOXY-L-PROLYL-L-PROLINATE",
            "peptide",
        ),
        "T10": (
            "[1-(1-BENZYL-3-HYDROXY-2-OXO-PROPYLCARBAMOYL)-2-PHENYL-ETHYL]-CARBAMIC ACID BENZYL ESTER",
            "non-polymer",
        ),
        "NCR": ("2-NITRO-P-CRESOL", "non-polymer"),
        "T11": ("4-[3-(TRIFLUOROMETHYL)DIAZIRIDIN-3-YL]-L-PHENYLALANINE", "peptide"),
        "B4W": (
            "2-{4-[(3S)-PYRROLIDIN-3-YLAMINO]QUINAZOLIN-2-YL}PHENOL",
            "non-polymer",
        ),
        "NCS": ("?", "non-polymer"),
        "B4P": ("BIS(ADENOSINE)-5'-TETRAPHOSPHATE", "non-polymer"),
        "D28": (
            "4-{[4-(4-FLUORO-3-METHYLPHENYL)-1,3-THIAZOL-2-YL]AMINO}-2-HYDROXYBENZOIC ACID",
            "non-polymer",
        ),
        "TFA": ("TRIFLUOROACETIC ACID", "non-polymer"),
        "T13": (
            "2,3,4,5,6-PENTAFLUORO-N-(3-FLUORO-4-METHOXYPHENYL)BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "NEI": (
            "(2S,3S,4R,5R)-5-(6-AMINO-2-IODO-9H-PURIN-9-YL)-N-ETHYL-3,4-DIHYDROXYTETRAHYDROFURAN-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "B4D": (
            "(3S)-6-CHLORO-3-[(PROP-2-EN-1-YLSULFANYL)METHYL]-3,4-DIHYDRO-2H-1,2,4-BENZOTHIADIAZINE-7-SULFONAMIDE 1,1-DIOXIDE",
            "non-polymer",
        ),
        "B4G": ("GALACTOTRIOSE", "saccharide"),
        "Z82": ("4-BROMOBENZOIC ACID", "non-polymer"),
        "Z85": (
            "1-{4-[2-(BENZYLOXY)ETHYL]-1,3-THIAZOL-2-YL}-3-[3-TERT-BUTYL-1-(4-METHYLPHENYL)-1H-PYRAZOL-5-YL]UREA",
            "non-polymer",
        ),
        "TFC": ("1,1,1-TRIFLUORO-3-(OCTYLTHIO)ACETONE", "non-polymer"),
        "B4C": (
            "BENZYL [(1S)-5-AMINO-1-{[(1S)-1-({(1S)-5-AMINO-1-[(S)-1,3-BENZOXAZOL-2-YL(HYDROXY)METHYL]PENTYL}CARBAMOYL)-3-PHENYLPROPYL]CARBAMOYL}PENTYL]CARBAMATE",
            "non-polymer",
        ),
        "Z86": (
            "(4-{5-[({4-[2-(BENZYLOXY)ETHYL]-1,3-THIAZOL-2-YL}CARBAMOYL)AMINO]-3-TERT-BUTYL-1H-PYRAZOL-1-YL}PHENYL)ACETIC ACID",
            "non-polymer",
        ),
        "B4M": ("TRIS(BROMANYL)METHANE", "non-polymer"),
        "Z88": ("N-[(4-BROMOPHENYL)SULFONYL]ACETAMIDE", "non-polymer"),
        "B4N": ("BIS(4-NITROPHENYL) HYDROGEN PHOSPHATE", "non-polymer"),
        "I02": (
            "3-{5-[(1R)-1-AMINO-1-METHYL-2-PHENYLETHYL]-1,3,4-OXADIAZOL-2-YL}-N-[(1R)-1-(4-FLUOROPHENYL)ETHYL]-5-[METHYL(METHYLSULFONYL)AMINO]BENZAMIDE",
            "non-polymer",
        ),
        "RRR": (
            "4-{[(E)-2-(5-CHLOROTHIEN-2-YL)VINYL]SULFONYL}-1-(1H-PYRROLO[3,2-C]PYRIDIN-2-YLMETHYL)PIPERAZIN-2-ONE",
            "non-polymer",
        ),
        "I03": (
            "(2S)-2-AMINO-2-BENZYL-3-HYDROXYPROPYL 3-({[(1R)-1-(4-FLUOROPHENYL)ETHYL]AMINO}CARBONYL)-5-[METHYL(METHYLSULFONYL)AMINO]BENZOATE",
            "non-polymer",
        ),
        "3Q3": (
            "(2S)-4-[2-AMINO-6-(3-AMINO-2H-INDAZOL-6-YL)PYRIMIDIN-4-YL]-N-PHENYLMORPHOLINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "3Q2": (
            "6-{2-AMINO-6-[(3R)-3-METHYLMORPHOLIN-4-YL]PYRIMIDIN-4-YL}-2H-INDAZOL-3-AMINE",
            "non-polymer",
        ),
        "3Q1": (
            "6-[2-AMINO-6-(MORPHOLIN-4-YL)PYRIMIDIN-4-YL]-2H-INDAZOL-3-AMINE",
            "non-polymer",
        ),
        "3Q0": (
            "6-(3-AMINO-2H-INDAZOL-6-YL)-N~4~-ETHYLPYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "3Q6": (
            "TERT-BUTYL {(3R,5R)-1-[2-AMINO-6-(3-AMINO-2H-INDAZOL-6-YL)PYRIMIDIN-4-YL]-5-METHYLPIPERIDIN-3-YL}CARBAMATE",
            "non-polymer",
        ),
        "3Q5": (
            "TERT-BUTYL {(3R,6S)-1-[2-AMINO-6-(3-AMINO-2H-INDAZOL-6-YL)PYRIMIDIN-4-YL]-6-METHYLPIPERIDIN-3-YL}CARBAMATE",
            "non-polymer",
        ),
        "3Q4": (
            "(3S,6R)-1-[2-AMINO-6-(3-AMINO-2H-INDAZOL-6-YL)PYRIMIDIN-4-YL]-6-METHYL-N-PHENYLPIPERIDINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "KWS": (
            "{2-[(1R,2R)-1-AMINO-2-HYDROXYPROPYL]-4-METHYLIDENE-5-OXO-4,5-DIHYDRO-1H-IMIDAZOL-1-YL}ACETIC ACID",
            "peptide",
        ),
        "I01": (
            "(4R)-4-(2-FLUORO-6-HYDROXY-3-METHOXY-BENZOYL)-BENZOIC ACID (3R)-3-[(PYRIDINE-4-CARBONYL)AMINO]-AZEPAN-4-YL ESTER",
            "non-polymer",
        ),
        "KWT": (
            "(1S,6BR,9AS,11R,11BR)-9A,11B-DIMETHYL-1-[(METHYLOXY)METHYL]-3,6,9-TRIOXO-1,6,6B,7,8,9,9A,10,11,11B-DECAHYDRO-3H-FURO[4,3,2-DE]INDENO[4,5-H][2]BENZOPYRAN-11-YL ACETATE",
            "non-polymer",
        ),
        "D20": ("N~5~-{IMINO[(2-METHOXYETHYL)AMINO]METHYL}-L-ORNITHINE", "non-polymer"),
        "I06": ("2-PHENETHYL-2,3-DIHYDRO-PHTHALAZINE-1,4-DIONE", "non-polymer"),
        "P1Y": (
            "2-(3-BROMO-4-METHOXYBENZYL)-5,6-DIHYDROXYPYRIMIDINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "KWZ": (
            "4-(3-PIPERIDIN-1-YLPROPYL)-6-[3-(TRIFLUOROMETHYL)PHENYL]PYRIMIDINE-2-CARBONITRILE",
            "non-polymer",
        ),
        "TFE": ("?", "nucleic-acid"),
        "NCM": ("NORCAMPHOR", "non-polymer"),
        "3QZ": ("(9BETA)-17-HYDROXYPREGN-4-ENE-3,20-DIONE", "non-polymer"),
        "I04": (
            "N-{(3R,4R)-4-[4-(2-FLUORO-6-HYDROXY-3-METHOXY-BENZOYL)-BENZOYLAMINO]-AZEPAN-3-YL}ISONICOTINAMIDE",
            "non-polymer",
        ),
        "D25": ("3-PHENYL-5-(1H-PYRAZOL-3-YL)ISOXAZOLE", "non-polymer"),
        "3QP": (
            "(E)-N-{2-HYDROXY-3-METHYL-6-[(PHOSPHONOOXY)METHYL]BENZYLIDENE}-L-ASPARTIC ACID",
            "peptide",
        ),
        "NEG": ("NEGAMYCIN", "non-polymer"),
        "3QV": ("7-HYDROXY-2-OXO-2H-CHROMENE-3-CARBONITRILE", "non-polymer"),
        "3QU": ("4-(4-NITROBENZYL)PYRIDINE", "non-polymer"),
        "3QT": (
            "6-{[6-(1-METHYL-1H-PYRAZOL-4-YL)IMIDAZO[1,2-B]PYRIDAZIN-3-YL]METHYL}QUINOLINE",
            "non-polymer",
        ),
        "23T": ("2',3'-DIDEOXY-THYMIDINE-5'-TRIPHOSPHATE", "non-polymer"),
        "3QI": (
            "5,6-DIMETHYL-3-(4-METHYL-2-PYRIDINYL)-2-THIOXO-2,3-DIHYDROTHIENO[2,3- D]PYRIMIDIN-4(1H)-ONE",
            "non-polymer",
        ),
        "3QO": ("4-BENZYLPYRIDINE", "non-polymer"),
        "3QN": (
            "(E)-N~6~-{2-HYDROXY-3-METHYL-6-[(PHOSPHONOOXY)METHYL]BENZYLIDENE}-L-LYSINE",
            "peptide",
        ),
        "23U": (
            "BETA-PHENYL-D-PHENYLALANYL-N-(3-CHLOROBENZYL)-L-PROLINAMIDE",
            "non-polymer",
        ),
        "3QC": (
            "(4R)-4-(3-HYDROXYPHENYL)-N,N,7,8-TETRAMETHYL-3,4-DIHYDROISOQUINOLINE-2(1H)-CARBOXAMIDE",
            "non-polymer",
        ),
        "D26": ("PHENYL-5-(1H-PYRAZOL-3-YL)-1,3-THIAZOLE", "non-polymer"),
        "I08": (
            "(3R,4S)-N-(4-{TRANS-2-[4-(2-FLUORO-6-HYDROXY-3-METHOXY-BENZOYL)-PHENYL]-VINYL}-AZEPAN-3-YL)-ISONICOTINAMIDE",
            "non-polymer",
        ),
        "B45": (
            "5-(2-CHLORO-4-FLUOROPHENYL)-1-(2,6-DICHLOROPHENYL)-7-[1-(1-METHYLETHYL)PIPERIDIN-4-YL]-3,4-DIHYDROQUINAZOLIN-2(1H)-ONE",
            "non-polymer",
        ),
        "B44": ("?", "non-polymer"),
        "B47": (
            "METHYL ({5-[4-(4-HYDROXYPIPERIDIN-1-YL)-2-PHENYL-1,3-THIAZOL-5-YL]-1H-PYRAZOL-3-YL}METHYL)CARBAMATE",
            "non-polymer",
        ),
        "B46": (
            "PROP-2-YN-1-YL {[5-(4-PIPERIDIN-1-YL-2-PYRIDIN-3-YL-1,3-THIAZOL-5-YL)-1H-PYRAZOL-3-YL]METHYL}CARBAMATE",
            "non-polymer",
        ),
        "B41": ("(2S)-1-(1,3-BENZODIOXOL-5-YL)-N-METHYLPROPAN-2-AMINE", "non-polymer"),
        "B40": ("(2S)-N-METHYL-1-PHENYLPROPAN-2-AMINE", "non-polymer"),
        "B43": (
            "4-AMINO-5-(4-PHENOXYPHENYL)-7H-PYRROLO[2,3-D]PYRIMIDIN-7-YL-CYCLOPENTANE",
            "non-polymer",
        ),
        "B42": (
            "N-{3-[(5R)-1-CYCLOPENTYL-4-HYDROXY-5-METHYL-5-(3-METHYLBUTYL)-2-OXO-1,2,5,6-TETRAHYDROPYRIDIN-3-YL]-1,1-DIOXIDO-4H-1,2,4-BENZOTHIADIAZIN-7-YL}METHANESULFONAMIDE",
            "non-polymer",
        ),
        "TFQ": ("4-(2,2,2-TRIFLUOROETHYL)-L-PHENYLALANINE", "peptide"),
        "NED": ("2,6-DIAMINO-2,6-DIDEOXY-ALPHA-D-GLUCOPYRANOSE", "saccharide"),
        "B49": (
            "N-[2-(DIETHYLAMINO)ETHYL]-5-[(Z)-(5-FLUORO-2-OXO-1,2-DIHYDRO-3H-INDOL-3-YLIDENE)METHYL]-2,4-DIMETHYL-1H-PYRROLE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "B48": (
            "4-METHYL-5-[3-(METHYLSULFANYL)-1H-PYRAZOL-5-YL]-2-THIOPHEN-2-YL-1,3-THIAZOLE",
            "non-polymer",
        ),
        "23Q": (
            "(2S,3S,11BS)-3-[3-(FLUOROMETHYL)PHENYL]-9,10-DIMETHOXY-1,3,4,6,7,11B-HEXAHYDRO-2H-PYRIDO[2,1-A]ISOQUINOLIN-2-AMINE",
            "non-polymer",
        ),
        "NCF": (
            "(2R)-2-{(1R)-2-OXO-1-[(2-THIENYLACETYL)AMINO]ETHYL}-5,6-DIHYDRO-2H-1,3-THIAZINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "NCG": ("NCSI-GLU, NEOCARZINOSTATIN-GLUTATHIONE CHROMOPHORE", "non-polymer"),
        "LID": (
            "8-(2-CHLOROPHENYLAMINO)-2-(2,6-DIFLUOROPHENYLAMINO)-9-ETHYL-9H-PURINE-1,7-DIIUM",
            "non-polymer",
        ),
        "LIE": (
            "2-(2,6-DIFLUOROPHENOXY)-N-(2-FLUOROPHENYL)-9-ISOPROPYL-9H-PURIN-8-AMINE",
            "non-polymer",
        ),
        "LIF": (
            "N-{4-[4-AMINO-6-(4-METHOXYPHENYL)FURO[2,3-D]PYRIMIDIN-5-YL]PHENYL}-N'-[2-FLUORO-5-(TRIFLUOROMETHYL)PHENYL]UREA",
            "non-polymer",
        ),
        "YLY": (
            "(2R)-2-AMINO-6-({[(2S,3R)-3-METHYLPYRROLIDIN-2-YL]CARBONYL}AMINO)HEXANOYL [(2S,3R,4R,5R)-5-(6-AMINO-9H-PURIN-9-YL)-3,4-DIHYDROXYTETRAHYDROFURAN-2-YL]METHYL HYDROGEN (R)-PHOSPHATE",
            "non-polymer",
        ),
        "X25": (
            "4-{[2-(3-BENZYL-1,2,4-OXADIAZOL-5-YL)-3-METHYL-1-BENZOFURAN-4-YL]OXY}PIPERIDINE",
            "non-polymer",
        ),
        "X24": (
            "TERT-BUTYL [(TRANS-4-{[({2-[4-(AMINOMETHYL)PHENYL]QUINOLIN-4-YL}CARBONYL)AMINO]METHYL}CYCLOHEXYL)METHYL]CARBAMATE",
            "non-polymer",
        ),
        "YLL": (
            "(1R,2S,3S,4S,5R,6R)-6-(HYDROXYMETHYL)CYCLOHEXANE-1,2,3,4,5-PENTOL",
            "non-polymer",
        ),
        "X23": (
            "(3S,6S,7R,9AS)-6-{[(2S)-2-AMINOBUTANOYL]AMINO}-7-(AMINOMETHYL)-N-(DIPHENYLMETHYL)-5-OXOOCTAHYDRO-1H-PYRROLO[1,2-A]AZEPINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "LIA": (
            "(4-AMINO-2-{[1-(METHYLSULFONYL)PIPERIDIN-4-YL]AMINO}PYRIMIDIN-5-YL)(2,3-DIFLUORO-6-METHOXYPHENYL)METHANONE",
            "non-polymer",
        ),
        "12A": (
            "2-METHYLTHIO-N6-(AMINOCARBONYL-L-THREONYL)-ADENOSINE-5'-MONOPHOSPHATE",
            "nucleic-acid",
        ),
        "12B": ("BENZO[CD]INDOL-2(1H)-ONE", "non-polymer"),
        "12C": (
            "4-[(3S)-1-AZABICYCLO[2.2.2]OCT-3-YLAMINO]-3-(1H-BENZIMIDAZOL-2-YL)-6-CHLOROQUINOLIN-2(1H)-ONE",
            "non-polymer",
        ),
        "12D": ("?", "non-polymer"),
        "12E": (
            "N-{[(2E)-2-(PYRIDIN-4-YLMETHYLIDENE)HYDRAZINO]CARBONOTHIOYL}-BETA-D-GLUCOPYRANOSYLAMINE",
            "non-polymer",
        ),
        "4HY": (
            "[4-(4-HYDROXY-3-IODO-PHENOXY)-3,5-DIIODO-PHENYL]-ACETIC ACID",
            "non-polymer",
        ),
        "4HX": (
            "(4R)-4-(HYDROXY{N~5~-[IMINO(NITROAMINO)METHYL]-L-ORNITHYL}AMINO)-L-PROLINAMIDE",
            "non-polymer",
        ),
        "12H": ("12-HYDROXYDODECANOIC ACID", "non-polymer"),
        "4HV": (
            "N-[(2R)-1-(3-CYANOPHENYL)-3-HYDROXYPROPAN-2-YL]-5-(3,4-DIMETHOXYPHENYL)FURAN-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "12J": ("2-IODO-N-[3-(1-METHYLETHOXY)PHENYL]BENZAMIDE", "non-polymer"),
        "4HT": ("4-HYDROXYTRYPTOPHAN", "peptide"),
        "12L": (
            "(1R,2S,5S)-6,6-DIMETHYL-3-AZABICYCLO[3.1.0]HEXANE-2-CARBOXYLIC ACID",
            "peptide",
        ),
        "12M": ("(2-ETHYLPHENYL)METHANOL", "non-polymer"),
        "12N": ("(2R,3S)-3-AMINO-4-CYCLOBUTYL-2-HYDROXYBUTANOIC ACID", "non-polymer"),
        "4HP": ("4-HYDROXYPHENYLACETATE", "non-polymer"),
        "12P": ("DODECAETHYLENE GLYCOL", "non-polymer"),
        "4HN": (
            "2-(2'-ETHOXYBIPHENYL-4-YL)-4-HYDROXY-1,6-NAPHTHYRIDINE-3-CARBONITRILE",
            "non-polymer",
        ),
        "4HM": ("4-HYDROPEROXY-2-METHOXY-PHENOL", "non-polymer"),
        "4HL": ("4-(HYDRAZINOMETHYL)PHENOL", "non-polymer"),
        "4HK": (
            "5-{7-ETHYL-6-[(3S)-TETRAHYDROFURAN-3-YLOXY]-2,4-DIHYDROINDENO[1,2-C]PYRAZOL-3-YL}PYRIDINE-2-CARBONITRILE",
            "non-polymer",
        ),
        "12U": (
            "N-CYCLOHEPTYLGLYCYL-N-(4-CARBAMIMIDOYLBENZYL)-L-PROLINAMIDE",
            "non-polymer",
        ),
        "4HI": (
            "(3R,5R)-7-[4-(BENZYLCARBAMOYL)-2-(4-FLUOROPHENYL)-5-(1-METHYLETHYL)-1H-IMIDAZOL-1-YL]-3,5-DIHYDROXYHEPTANOIC ACID",
            "non-polymer",
        ),
        "12W": ("CYCLOPENTYL HYDROGEN CARBONATE", "non-polymer"),
        "4HG": ("N-[(4-HYDROXY-8-IODOISOQUINOLIN-3-YL)CARBONYL]GLYCINE", "non-polymer"),
        "4HF": (
            "N-[(5-{2-[(6R)-2-AMINO-4-OXO-3,4,5,6,7,8-HEXAHYDROPYRIDO[2,3-D]PYRIMIDIN-6-YL]ETHYL}-2-THIENYL)CARBONYL]-L-GLUTAMIC ACID",
            "non-polymer",
        ),
        "12Z": (
            "[4-AMINO-2-(CYCLOHEXYLAMINO)-1,3-THIAZOL-5-YL](PYRIDIN-3-YL)METHANONE",
            "non-polymer",
        ),
        "4HD": (
            "(4S,5E,7Z,10Z,13Z,16Z,19Z)-4-HYDROXYDOCOSA-5,7,10,13,16,19-HEXAENOIC ACID",
            "non-polymer",
        ),
        "4HC": ("4-HYDROXY-2H-CHROMEN-2-ONE", "non-polymer"),
        "4HB": ("4,4'-DIHYDROXYBIPHENYL", "non-polymer"),
        "4HA": ("4-HYDROXYBUTAN-1-AMINIUM", "non-polymer"),
        "V30": (
            "2-[(5,6-DIMETHYL-4-OXO-3,4-DIHYDROTHIENO[2,3-D]PYRIMIDIN-2-YL)SULFANYL]-N-(2-HYDROXYETHYL)ACETAMIDE",
            "non-polymer",
        ),
        "LIO": (
            "[1-PENTADECANOYL-2-DECANOYL-GLYCEROL-3-YL]PHOSPHONYL CHOLINE",
            "non-polymer",
        ),
        "NEC": ("?", "non-polymer"),
        "XP4": ("1,2-DIMYRISTOYL-SN-GLYCERO-3-PHOSPHATE", "non-polymer"),
        "LIH": (
            "6-([5-QUINOLYLAMINO]METHYL)-2,4-DIAMINO-5-METHYLPYRIDO[2,3-D]PYRIMIDINE",
            "non-polymer",
        ),
        "4H0": (
            "(3S)-3-AMINO-4-OXO-5-(2,3,5,6-TETRAFLUOROPHENOXY)PENTANOIC ACID",
            "peptide",
        ),
        "120": (
            "2-(2-OXO-1,2-DIHYDRO-PYRIDIN-3-YL)-1H-BENZOIMIDAZOLE-5-CARBOXAMIDINE",
            "non-polymer",
        ),
        "121": (
            "2-(3-HYDROXY-PYRIDIN-2-YL)-1H-BENZOIMIDAZOLE-5-CARBOXAMIDINE",
            "non-polymer",
        ),
        "122": (
            "2-(2-HYDROXY-PHENYL)-3H-BENZOIMIDAZOLE-5-CARBOXAMIDINE",
            "non-polymer",
        ),
        "123": (
            "2-(2-HYDROXY-5-METHOXY-PHENYL)-1H-BENZOIMIDAZOLE-5-CARBOXAMIDINE",
            "non-polymer",
        ),
        "124": ("2-(2-HYDROXY-PHENYL)-1H-INDOLE-5-CARBOXAMIDINE", "non-polymer"),
        "125": ("2'-O-FLUOROETHYL-5-METHYL-URIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "126": (
            "2'-O-METHYL-[TRI(OXYETHYL)]-5-METHYL-URIDINE-5'-MONOPHOSPHATE",
            "nucleic-acid",
        ),
        "127": (
            "2'-O-AMINOOXY-ETHYL-5-METHYL-URIDINE-5'-MONOPHOSPHATE",
            "nucleic-acid",
        ),
        "128": (
            "SPIRO(2,4,6-TRINITROBENZENE[1,2A]-2O',3O'-METHYLENE-ADENINE-TRIPHOSPHATE",
            "nucleic-acid",
        ),
        "129": ("[2(FORMYL-HYDROXY-AMINO)-ETHYL]-PHOSPHONIC ACID", "non-polymer"),
        "LIK": ("3,6-BIS(METHYLENE)DECANOIC ACID", "non-polymer"),
        "TLA": ("L(+)-TARTARIC ACID", "non-polymer"),
        "PNX": (
            "3,7-DIMETHYL-1-(5-OXOHEXYL)-3,7-DIHYDRO-1H-PURINE-2,6-DIONE",
            "non-polymer",
        ),
        "LIT": (
            "{[5-(5-NITRO-2-FURYL)-1,3,4-OXADIAZOL-2-YL]THIO}ACETIC ACID",
            "non-polymer",
        ),
        "TNT": ("1,3-BIS(AMIDINOPHENOXY)PROPANE", "non-polymer"),
        "X2N": ("POSACONAZOLE", "non-polymer"),
        "X2K": (
            "(3Z)-5-AMINO-3-[(3,5-DIMETHYL-1H-PYRROL-2-YL)METHYLIDENE]-1,3-DIHYDRO-2H-INDOL-2-ONE",
            "non-polymer",
        ),
        "X2E": (
            "6-{[(3R,4R)-4-(2-{[(2R)-1-(3-FLUOROPHENYL)PROPAN-2-YL]AMINO}ETHOXY)PYRROLIDIN-3-YL]METHYL}-4-METHYLPYRIDIN-2-AMINE",
            "non-polymer",
        ),
        "X2D": (
            "6-{[(3R,4R)-4-(2-{[(1R,2S)-2-(3-CHLOROPHENYL)CYCLOPROPYL]AMINO}ETHOXY)PYRROLIDIN-3-YL]METHYL}-4-METHYLPYRIDIN-2-AMINE",
            "non-polymer",
        ),
        "X2F": ("2-DEOXY-2-FLUORO XYLOPYRANOSE", "saccharide"),
        "LIV": (
            "LIVIDOMYCIN A; O-2-AMINO-2,3-DIDEOXY-ALPHA-D-GLUCOPYRANOSYL-(1,4)-O-[BETA-D-MANNOPYRANOSYL-(1,4)-O-2,6-DIAMINO-2,6-DIDEOXY-BETA-L-IDOPYRANOSYL-(1,3)-BETA-D-RIBOFURANOSYL-(1,5)-2-DEOXY-D-STREPTAMINE",
            "non-polymer",
        ),
        "X2C": (
            "4-METHYL-6-{[(3R,4R)-4-(2-{[(1R,2S)-2-(3-METHYLPHENYL)CYCLOPROPYL]AMINO}ETHOXY)PYRROLIDIN-3-YL]METHYL}PYRIDIN-2-AMINE",
            "non-polymer",
        ),
        "LIW": (
            "4-[4-(BIPHENYL-2-YLMETHYL)PIPERAZIN-1-YL]-N-[(4-{[1,1-DIMETHYL-2-(PHENYLTHIO)ETHYL]AMINO}-3-NITROPHENYL)SULFONYL]BENZAMIDE",
            "non-polymer",
        ),
        "S6P": ("D-SORBITOL-6-PHOSPHATE", "saccharide"),
        "X2W": ("N-ACETYL-L-GLUTAMYL 5-PHOSPHATE", "peptide"),
        "NEA": ("5'-DEOXY-5'-[2-(AMINO)ETHYLTHIO]ADENOSINE", "non-polymer"),
        "KA5": (
            "5-CHLORO-N~6~-(2,5-DIMETHOXYBENZYL)QUINAZOLINE-2,4,6-TRIAMINE",
            "non-polymer",
        ),
        "8AX": (
            "6-{[(3R,4R)-4-(2-{[2-(3-CHLOROPHENYL)-2,2-DIFLUOROETHYL]AMINO}ETHOXY)PYRROLIDIN-3-YL]METHYL}-4-METHYLPYRIDIN-2-AMINE",
            "non-polymer",
        ),
        "8AP": ("N~3~-BENZYLPYRIDINE-2,3-DIAMINE", "non-polymer"),
        "LIX": (
            "N-[2-({2-AMINO-6-ETHYL-5-[4-(3-METHOXYPROPYL)-2,2-DIMETHYL-3-OXO-3,4-DIHYDRO-2H-1,4-BENZOXAZIN-6-YL]PYRIMIDIN-4-YL}AMINO)ETHYL]NAPHTHALENE-2-SULFONAMIDE",
            "non-polymer",
        ),
        "8AR": ("HYDROGEN ARSENATE", "non-polymer"),
        "L8P": (
            "({[(2S)-1-(4-AMINO-2-OXOPYRIMIDIN-1(2H)-YL)-3-HYDROXYPROPAN-2-YL]OXY}METHYL)PHOSPHONIC ACID",
            "non-polymer",
        ),
        "LIZ": (
            "N-{(1S,2S)-2-[BIS(CARBOXYMETHYL)AMINO]CYCLOHEXYL}-N-{(2R)-2-[BIS(CARBOXYMETHYL)AMINO]-3-[4-({[2-HYDROXY-1,1-BIS(HYDROXYMETHYL)ETHYL]CARBAMOTHIOYL}AMINO)PHENYL]PROPYL}GLYCINE",
            "non-polymer",
        ),
        "8AN": ("?", "nucleic-acid"),
        "8AC": ("8-AMINOOCTANOIC ACID", "non-polymer"),
        "8AD": (
            "9-AMINO-(N-(2-DIMETHYLAMINO)BUTYL)ACRIDINE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "8AG": ("?", "nucleic-acid"),
        "GKR": ("D-GLUCARATE", "non-polymer"),
        "TNP": ("3,5,6,8-TETRAMETHYL-N-METHYL PHENANTHROLINIUM", "non-polymer"),
        "TFG": (
            "2,2,2-TRIFLUORO-1-{5-[(3-PHENYL-5,6-DIHYDROIMIDAZO[1,2-A]PYRAZIN-7(8H)-YL)CARBONYL]THIOPHEN-2-YL}ETHANE-1,1-DIOL",
            "non-polymer",
        ),
        "BPD": (
            "N-(M-TRIFLUOROMETHYLPHENYL) PHENOXAZINE-4,6-DICARBOXYLIC ACID",
            "non-polymer",
        ),
        "PIV": ("PIVALIC ACID", "peptide"),
        "TS5": ("GLUTATHIONYLSPERMIDINE", "non-polymer"),
        "XLM": (
            "METHYL 3-O-BETA-D-XYLOPYRANOSYL-ALPHA-D-MANNOPYRANOSIDE",
            "non-polymer",
        ),
        "CIB": (
            "2-ACETYLAMINO-4-METHYL-PENTANOIC ACID [1-(1-FORMYL-PENTYLCARBAMOYL)-3-METHYL-BUTYL]-AMIDE",
            "non-polymer",
        ),
        "KET": (
            "2-[(3-HYDROXY-2-METHYL-5-PHOSPHONOOXYMETHYL-PYRIDIN-4-YLMETHYLENE)-AMINO]-SUCCINIC ACID",
            "non-polymer",
        ),
        "TKW": ("?", "non-polymer"),
        "IJ2": ("2-[(FURAN-2-YLMETHYL)AMINO]ETHANESULFONIC ACID", "non-polymer"),
        "NEZ": (
            "6-AMINO-4-[2-(4-METHYLPHENYL)ETHYL]-1,7-DIHYDRO-8H-IMIDAZO[4,5-G]QUINAZOLIN-8-ONE",
            "non-polymer",
        ),
        "MHW": ("3-HYDROXYPICOLINIC ACID", "peptide"),
        "IJ4": ("2-[(4-FLUORO-3-METHYLBENZYL)AMINO]ETHANESULFONIC ACID", "non-polymer"),
        "LL8": ("1-(6-METHOXYNAPHTHALEN-2-YL)BUTANE-1,3-DIONE", "non-polymer"),
        "MA0": (
            "(1R)-6-HYDROXY-N-METHYL-N-[(1Z)-PROP-2-EN-1-YLIDENE]INDAN-1-AMINIUM",
            "non-polymer",
        ),
        "CIE": (
            "2-[[[[(4-CHLORO-6-METHOXY-2-PYRIMIDINYL)AMINO]CARBONYL]AMINO]SULFONYL]BENZOIC ACID ETHYL ESTER",
            "non-polymer",
        ),
        "N61": (
            "1-(3-TERT-BUTYL-1-METHYL-1H-PYRAZOL-5-YL)-3-[4-(PYRIDIN-4-YLOXY)PHENYL]UREA",
            "non-polymer",
        ),
        "MHR": (
            "4-(4-PROPOXY-1H-PYRROLO[2,3-B]PYRIDIN-3-YL)PYRIMIDIN-2-AMINE",
            "non-polymer",
        ),
        "U0E": (
            "N-[[1-[N-ACETAMIDYL]-[1-CYCLOHEXYLMETHYL-2-HYDROXY-4-ISOPROPYL]-BUT-4-YL]-CARBONYL]-GLUTAMINYL-ARGINYL-AMIDE",
            "non-polymer",
        ),
        "N69": ("6-(CYCLOHEXYLMETHOXY)-8-ISOPROPYL-9H-PURIN-2-AMINE", "non-polymer"),
        "MHS": ("N1-METHYLATED HISTIDINE", "peptide"),
        "S60": (
            "6-AMINO-2-METHYL-1,7-DIHYDRO-8H-IMIDAZO[4,5-G]QUINAZOLIN-8-ONE",
            "non-polymer",
        ),
        "L8H": ("4-METHOXYNAPHTHALEN-2-AMINE", "non-polymer"),
        "2N6": (
            "4-(2,4-DIMETHYLPHENYL)-2-(METHYLSULFANYL)-7H-PYRROLO[2,3-D]PYRIMIDINE-5-CARBONITRILE",
            "non-polymer",
        ),
        "T12": (
            "1-(3-HYDROXYPROPYL)-2-[(3-NITROBENZOYL)AMINO]-1H-BENZIMIDAZOL-5-YL PIVALATE",
            "non-polymer",
        ),
        "T1D": (
            "5-(4-METHOXYBIPHENYL-3-YL)-1,2,5-THIADIAZOLIDIN-3-ONE 1,1-DIOXIDE",
            "non-polymer",
        ),
        "2NI": (
            "N-{2-[(4'-CYANO-1,1'-BIPHENYL-4-YL)OXY]ETHYL}-N'-HYDROXY-N-METHYLUREA",
            "non-polymer",
        ),
        "2NH": (
            "N-ACETYL-S-[(2E,6E)-3,7,11-TRIMETHYLDODECA-2,6,10-TRIENYL]-L-CYSTEINYL-D-VALYL-L-ISOLEUCYL-L-METHIONINE",
            "non-polymer",
        ),
        "2NO": ("NITROGEN DIOXIDE", "non-polymer"),
        "N6R": (
            "(2R,4AS,6R,7R,7AR)-6-(6-{[(1R)-1-METHYL-2-PHENYLETHYL]AMINO}-9H-PURIN-9-YL)TETRAHYDRO-4H-FURO[3,2-D][1,3,2]DIOXAPHOSPHININE-2,7-DIOL 2-OXIDE",
            "non-polymer",
        ),
        "2NC": (
            "N-{(2S)-2-[(N-ACETYL-L-THREONYL-L-ISOLEUCYL)AMINO]HEXYL}-L-NORLEUCYL-L-GLUTAMINYL-N~5~-[AMINO(IMINIO)METHYL]-L-ORNITHINAMIDE",
            "peptide",
        ),
        "CIO": ("CILOMILAST", "non-polymer"),
        "2NA": ("NAPHTHALEN-2-YLMETHANOL", "non-polymer"),
        "H53": (
            "2-(2,5-DIHYDROXYPHENYL)-6-(HYDROXYMETHYL)OXANE-3,4,5-TRIOL",
            "non-polymer",
        ),
        "2NF": (
            "[[(2R,3S,4R,5R)-5-(6-AMINOPURIN-9-YL)-3,4-DIHYDROXYOXOLAN-2-YL]METHOXY-HYDROXYPHOSPHORYL] [(2R,3S,4R,5R)-5-(3-CARBAMOYLPYRIDIN-1-IUM-1YL)- 3-FLUORO-,4- HYDROXYOXOLAN-2-YL]METHYL PHOSPHATE",
            "non-polymer",
        ),
        "2ND": ("2'-DEOXYINOSINE", "non-polymer"),
        "N6G": (
            "((2R,3S,4R,5S)-5-(2,6-DIAMINO-9H-PURIN-9-YL)-3,4-DIHYDROXY-TETRAHYDROFURAN-2-YL)METHYL DIHYDROGEN PHOSPHATE",
            "nucleic-acid",
        ),
        "AL0": ("3-[HYDROXY(NITROSO)AMINO]-L-ALANINE", "non-polymer"),
        "T1G": ("(2S,3S)-3-HYDROXY-2-METHYLBUTANOYL-COA", "non-polymer"),
        "N6C": ("1-NITROHEXANE", "non-polymer"),
        "N6A": (
            "N-(6-AMINO-1-BENZYL-2,4-DIOXO-1,2,3,4-TETRAHYDROPYRIMIDIN-5-YL)-N-METHYLBENZENESULFONAMIDE",
            "non-polymer",
        ),
        "2NQ": (
            "N-{2-CHLORO-5-[4-(MORPHOLIN-4-YL)QUINOLIN-6-YL]PYRIDIN-3-YL}-4-FLUOROBENZENESULFONAMIDE",
            "non-polymer",
        ),
        "2NP": ("L-2-AMINO-6-METHYLENE-PIMELIC ACID", "non-polymer"),
        "U02": (
            "[2-(3-{[6-(1-BENZYL-PROPYL)-4-HYDROXY-2-OXO-2H-PYRAN-3-YL]-CYCLOPROPYL-METHYL}-PHENYLCARBAMOYL)-ETHYL]-CARBAMIC ACID TERT-BUTYL ESTER",
            "non-polymer",
        ),
        "U03": ("4-HYDROXY-7-METHOXY-3-(1-PHENYL-PROPYL)-CHROMEN-2-ONE", "non-polymer"),
        "2NT": ("?", "nucleic-acid"),
        "MA2": ("4-S-METHYL-4-THIO-ALPHA-D-GLUCOPYRANOSE", "saccharide"),
        "NVP": (
            "11-CYCLOPROPYL-5,11-DIHYDRO-4-METHYL-6H-DIPYRIDO[3,2-B:2',3'-E][1,4]DIAZEPIN-6-ONE",
            "non-polymer",
        ),
        "H59": (
            "N-{3-[(5S)-5-(1,1-DIMETHYLPROPYL)-1-(4-FLUORO-3-METHYLBENZYL)-4-HYDROXY-2-OXO-2,5-DIHYDRO-1H-PYRROL-3-YL]-1,1-DIOXIDO-4H-1,4-BENZOTHIAZIN-7-YL}METHANESULFONAMIDE",
            "non-polymer",
        ),
        "T1N": (
            "(2S)-2-[6-(METHYLSULFANYL)NAPHTHALEN-2-YL]PROPANOIC ACID",
            "non-polymer",
        ),
        "TFF": (
            "5'-O-[(R)-{[(R)-[DIFLUORO(PHOSPHONO)METHYL](HYDROXY)PHOSPHORYL](DIFLUORO)METHYL}(HYDROXY)PHOSPHORYL]THYMIDINE",
            "nucleic-acid",
        ),
        "C8P": (
            "N-{(1S,2S,3R)-1-[(ALPHA-D-GALACTOPYRANOSYLOXY)METHYL]-2,3-DIHYDROXYHEPTADECYL}-8-PHENYLOCTANAMIDE",
            "non-polymer",
        ),
        "C8R": ("?", "non-polymer"),
        "NVZ": (
            "5-[(3AS,4S,6AR)-1-ACETYL-2-OXOHEXAHYDRO-1H-THIENO[3,4-D]IMIDAZOL-4-YL]PENTANOIC ACID",
            "non-polymer",
        ),
        "IYT": ("N-ALPHA-ACETYL-3,5-DIIODOTYROSYL-D-THREONINE", "peptide"),
        "C8C": ("2-AMINO-6-[2-(1H-INDOL-6-YL)ETHYL]PYRIMIDIN-4(3H)-ONE", "non-polymer"),
        "C8D": (
            "N-[(CYCLOPENTYLOXY)CARBONYL]-3-METHYL-L-VALYL-(4R)-N-{(1R)-3-HYDROXY-1-[HYDROXY(OXIDO)BORANYL]PROPYL}-4-(ISOQUINOLIN-1-YLOXY)-L-PROLINAMIDE",
            "non-polymer",
        ),
        "C8E": ("(HYDROXYETHYLOXY)TRI(ETHYLOXY)OCTANE", "non-polymer"),
        "C8F": (
            "8-(4-FLUOROPHENYL)-N-{(1S,2S,3R)-1-[(ALPHA-D-GALACTOPYRANOSYLOXY)METHYL]-2,3-DIHYDROXYHEPTADECYL}OCTANAMIDE",
            "non-polymer",
        ),
        "IYR": ("3-IODO-TYROSINE", "peptide"),
        "WT4": (
            "N-[(5S)-5-AMINO-5-CARBOXYPENTANOYL]-L-HOMOCYSTEYL-S-METHYL-D-CYSTEINE",
            "non-polymer",
        ),
        "C8M": (
            "3-[{[(2R,3S,4R,5R)-5-(6-AMINO-8-METHYL-9H-PURIN-9-YL)-3,4-DIHYDROXYTETRAHYDROFURAN-2-YL]METHYL}(METHYL)AMINO]PROPANAMIDE",
            "non-polymer",
        ),
        "IYZ": (
            "1-(3-{6-[(CYCLOPROPYLMETHYL)AMINO]IMIDAZO[1,2-B]PYRIDAZIN-3-YL}PHENYL)ETHANONE",
            "non-polymer",
        ),
        "0L7": (
            "2-AMINO-5-(5-AMINOPENT-1-YN-1-YL)-7-{2-DEOXY-5-O-[(S)-HYDROXY{[(S)-HYDROXY(PHOSPHONOOXY)PHOSPHORYL]OXY}PHOSPHORYL]-BETA-D-ERYTHRO-PENTOFURANOSYL}-3,7-DIHYDRO-4H-PYRROLO[2,3-D]PYRIMIDIN-4-ONE",
            "non-polymer",
        ),
        "0L6": ("?", "non-polymer"),
        "0L5": ("?", "non-polymer"),
        "0L4": (
            "5-(5-AMINOPENT-1-YN-1-YL)-7-{2-DEOXY-5-O-[(S)-HYDROXY{[(S)-HYDROXY(PHOSPHONOOXY)PHOSPHORYL]OXY}PHOSPHORYL]-BETA-D-ERYTHRO-PENTOFURANOSYL}-7H-PYRROLO[2,3-D]PYRIMIDIN-4-AMINE",
            "non-polymer",
        ),
        "0L3": (
            "7-{2-DEOXY-5-O-[(S)-HYDROXY{[(S)-HYDROXY(PHOSPHONOOXY)PHOSPHORYL]OXY}PHOSPHORYL]-BETA-D-ERYTHRO-PENTOFURANOSYL}-5-{5-[(10-HYDROXYDECANOYL)AMINO]PENT-1-YN-1-YL}-7H-PYRROLO[2,3-D]PYRIMIDIN-4-AMINE",
            "non-polymer",
        ),
        "PPF": ("PHOSPHONOFORMIC ACID", "non-polymer"),
        "0L1": ("HEXANEDIOIC ACID", "non-polymer"),
        "0L0": (
            "(4-{[(6-BENZYL-1-HYDROXY-7-METHOXY-2-OXO-1,2-DIHYDROQUINOLIN-3-YL)AMINO]METHYL}-5-HYDROXY-6-METHYLPYRIDIN-3-YL)METHYL DIHYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "PPK": ("(DIPHOSPHONO)AMINOPHOSPHONIC ACID", "non-polymer"),
        "PPJ": (
            "N-(HYDROXY{3-HYDROXY-2-METHYL-5-[(PHOSPHONOOXY)METHYL]PYRIDIN-4-YL}METHYL)NORVALINE",
            "non-polymer",
        ),
        "PPI": ("PROPANOIC ACID", "non-polymer"),
        "PPH": ("[(1R)-1-AMINO-2-PHENYLETHYL]PHOSPHONIC ACID", "peptide"),
        "OCN": ("COBALT ION, 2 WATERS COORDINATED", "non-polymer"),
        "PPN": ("PARA-NITROPHENYLALANINE", "peptide"),
        "0L9": (
            "(2S)-[6-BROMO-4-(4-CHLOROPHENYL)-2-METHYLQUINOLIN-3-YL](METHOXY)ETHANOIC ACID",
            "non-polymer",
        ),
        "0L8": (
            "2'-BROMO-6'-(FURAN-3-YL)-4'-(HYDROXYMETHYL)BIPHENYL-4-OL",
            "non-polymer",
        ),
        "PPS": ("?", "non-polymer"),
        "PPR": ("PHOSPHONOPYRUVATE", "non-polymer"),
        "PPQ": ("PHOSPHINOTHRICIN", "non-polymer"),
        "PPP": ("3-PHENYL-1,2-PROPANDIOL", "non-polymer"),
        "PPW": ("7-DEAZA-8-AZA-2'-DEOXYGUANOSINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "PPV": ("PYROPHOSPHATE", "non-polymer"),
        "PPU": ("PUROMYCIN-5'-MONOPHOSPHATE", "nucleic-acid"),
        "PPT": ("3-(P-TOLYL)PROPIONIC ACID", "non-polymer"),
        "NVC": (
            "N-[(2R)-2-{[(2S)-2-(1,3-BENZOXAZOL-2-YL)PYRROLIDIN-1-YL]CARBONYL}HEXYL]-N-HYDROXYFORMAMIDE",
            "non-polymer",
        ),
        "MA7": ("1N-METHYLADENOSINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "PPX": (
            "[PHENYLALANINYL-PROLINYL]-[2-(PYRIDIN-4-YLAMINO)-ETHYL]-AMINE",
            "peptide",
        ),
        "NVB": (
            "N-METHYL-4-[3-(3,4,5-TRIMETHOXYPHENYL)QUINOXALIN-5-YL]BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "0LW": (
            "(2R)-1-(PYRROLIDIN-1-YL)-3-(9H-THIOXANTHEN-9-YL)PROPAN-2-OL",
            "non-polymer",
        ),
        "0LV": ("(2,2-DIMETHYL-2,3-DIHYDRO-1-BENZOFURAN-7-YL)METHANOL", "non-polymer"),
        "0LU": (
            "(2S)-1-(PYRROLIDIN-1-YL)-3-(9H-THIOXANTHEN-9-YL)PROPAN-2-OL",
            "non-polymer",
        ),
        "6OG": ("?", "nucleic-acid"),
        "0LS": (
            "N-{[(3S,4S)-4-BENZYLPYRROLIDIN-3-YL]METHYL}-4-CHLORO-N-PHENYLANILINE",
            "non-polymer",
        ),
        "0LR": ("(3R,4R)-3-(NAPHTHALEN-2-YLMETHOXY)-4-PHENYLPIPERIDINE", "non-polymer"),
        "6OB": (
            "(8E,10S,12Z)-10-HYDROXY-6-OXOOCTADECA-8,12-DIENOIC ACID",
            "non-polymer",
        ),
        "6OC": ("(8R,9Z,12Z)-8-HYDROXY-6-OXOOCTADECA-9,12-DIENOIC ACID", "non-polymer"),
        "0LZ": (
            "N-(4-CHLORO-3-FLUOROPHENYL)-N'-[(1S)-1,2,3,4-TETRAHYDROISOQUINOLIN-1-YLMETHYL]ETHANEDIAMIDE",
            "non-polymer",
        ),
        "6OJ": (
            "7,8-BIS(CHLORANYL)-9-METHYL-3,4-DIHYDRO-2H-PYRIDO[3,4-B]INDOL-1-ONE",
            "non-polymer",
        ),
        "0LX": (
            "(2S)-6-[(4-FLUOROBENZYL)AMINO]-2-[(2R)-2-(4-FLUOROPHENYL)-2-METHOXYETHYL]-N-HYDROXYHEXANAMIDE",
            "non-polymer",
        ),
        "0LG": (
            "N-[(4S,8E,11S)-4-[(1R)-1-HYDROXY-2-{[3-(PROPAN-2-YL)BENZYL]AMINO}ETHYL]-2,13-DIOXO-11-PHENYL-6-OXA-3,12-DIAZABICYCLO[12.3.1]OCTADECA-1(18),8,14,16-TETRAEN-16-YL]-N-METHYLMETHANESULFONAMIDE",
            "non-polymer",
        ),
        "0LF": ("(4R)-4-(4-PHENYLBUTOXY)-L-PROLINE", "peptide"),
        "PP1": (
            "1-TER-BUTYL-3-P-TOLYL-1H-PYRAZOLO[3,4-D]PYRIMIDIN-4-YLAMINE",
            "non-polymer",
        ),
        "0LD": (
            "(5-HYDROXY-4-{[(7-HYDROXY-6-OXO-2-PHENYL-6,7-DIHYDRO-2H-PYRAZOLO[3,4-B]PYRIDIN-5-YL)AMINO]METHYL}-6-METHYLPYRIDIN-3-YL)METHYL DIHYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "0LC": (
            "(1R,2R)-N-(1-CYANOCYCLOPROPYL)-2-[(8-FLUORO-1,3,4,5-TETRAHYDRO-2H-PYRIDO[4,3-B]INDOL-2-YL)CARBONYL]CYCLOHEXANECARBOXAMIDE",
            "non-polymer",
        ),
        "0LB": (
            "(1R,2R)-N-(1-CYANOCYCLOPROPYL)-2-{[4-(4-FLUOROPHENYL)PIPERAZIN-1-YL]CARBONYL}CYCLOHEXANECARBOXAMIDE",
            "non-polymer",
        ),
        "0LA": ("(2S)-2-(6-CHLORO-9H-CARBAZOL-2-YL)PROPANOIC ACID", "non-polymer"),
        "PP4": (
            "METHYL (2S)-[1-((N-FORMYL)-L-VALYL)AMINO-2-(2-NAPHTHYL)ETHYL)HYDROXYPHOSPHINYLOXY]-3-PHENYL PROPANOATE",
            "non-polymer",
        ),
        "0LO": ("5-PHENYLPYRIDINE-3-CARBOXYLIC ACID", "non-polymer"),
        "0LN": (
            "13-CYCLOHEXYL-3-METHOXY-17,22-DIMETHYL-7H-10,6-(METHANOIMINOTHIOIMINOBUTANOIMINOMETHANO)INDOLO[2,1-A][2]BENZAZEPINE-14,23-DIONE 16,16-DIOXIDE",
            "non-polymer",
        ),
        "0LM": (
            "N-(4-CHLORO-3-FLUOROPHENYL)-N'-(1,2,2,6,6-PENTAMETHYLPIPERIDIN-4-YL)ETHANEDIAMIDE",
            "non-polymer",
        ),
        "0LL": (
            "N-[(1S,2S)-2-AMINO-2,3-DIHYDRO-1H-INDEN-1-YL]-N'-(4-CHLORO-3-FLUOROPHENYL)ETHANEDIAMIDE",
            "non-polymer",
        ),
        "0LK": (
            "N-[(1S,2S)-2-CARBAMIMIDAMIDO-2,3-DIHYDRO-1H-INDEN-1-YL]-N'-(4-CHLORO-3-FLUOROPHENYL)ETHANEDIAMIDE",
            "non-polymer",
        ),
        "0LJ": (
            "N-[(1R,2R)-2-CARBAMIMIDAMIDO-2,3-DIHYDRO-1H-INDEN-1-YL]-N'-(4-CHLORO-3-FLUOROPHENYL)ETHANEDIAMIDE",
            "non-polymer",
        ),
        "0LI": (
            "3-(IMIDAZO[1,2-B]PYRIDAZIN-3-YLETHYNYL)-4-METHYL-N-{4-[(4-METHYLPIPERAZIN-1-YL)METHYL]-3-(TRIFLUOROMETHYL)PHENYL}BENZAMIDE",
            "non-polymer",
        ),
        "0LH": ("2-(1H-INDOL-1-YL)ETHANAMINE", "non-polymer"),
        "C80": (
            "(2S,4S,5R,7R)-4-AMINO-8-(BUTYLAMINO)-5-HYDROXY-2,7-DIMETHYL-8-OXOOCTYL 1-BENZYL-1H-INDOLE-3-CARBOXYLATE",
            "non-polymer",
        ),
        "C85": (
            "4-{5-[(Z)-(2-IMINO-4-OXO-1,3-THIAZOLIDIN-5-YLIDENE)METHYL]FURAN-2-YL}BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "C88": (
            "3-(4-METHOXY-3-METHYLPHENYL)-1-(PIPERIDIN-4-YLMETHYL)-1H-PYRAZOLO[3,4-D]PYRIMIDIN-4-AMINE",
            "non-polymer",
        ),
        "T18": (
            "N-{3-[6-FLUORO-1-(4-FLUOROBENZYL)-4-HYDROXY-2-OXO-1,2-DIHYDROQUINOLIN-3-YL]-1,1-DIOXIDO-4H-1,4-BENZOTHIAZIN-7-YL}METHANESULFONAMIDE",
            "non-polymer",
        ),
        "NVI": ("1-VINYLIMIDAZOLE", "non-polymer"),
        "GSQ": (
            "5-CHLORO-N-{(3S)-1-[(1S)-1-METHYL-2-MORPHOLIN-4-YL-2-5-CHLORO-N-{(3S)-1-[(1S)-1-METHYL-2-MORPHOLIN-4-YL-2-SULFONAMIDE",
            "non-polymer",
        ),
        "GSP": ("5'-GUANOSINE-DIPHOSPHATE-MONOTHIOPHOSPHATE", "non-polymer"),
        "GSS": ("?", "nucleic-acid"),
        "DAO": ("LAURIC ACID", "non-polymer"),
        "GSU": ("O5'-(L-GLUTAMYL-SULFAMOYL)-ADENOSINE", "peptide"),
        "BA": ("BARIUM ION", "non-polymer"),
        "GSW": ("D-GAMMA-GLUTAMYL-L-CYSTEINYLGLYCINE", "non-polymer"),
        "BC": ("BENZO[G]CHRYSENE", "non-polymer"),
        "A6T": (
            "(1S,5R)-7-{4-[3-(2-CHLORO-3,6-DIFLUOROPHENOXY)PROPYL]PHENYL}-N-CYCLOPROPYL-N-(2,3-DICHLOROBENZYL)-3,9-DIAZABICYCLO[3.3.1]NON-6-ENE-6-CARBOXAMIDE",
            "non-polymer",
        ),
        "A6U": (
            "1,5-ANHYDRO-2-DEOXY-2-(2,4-DIOXO-3,4-DIHYDROPYRIMIDIN-1(2H)-YL)-6-O-PHOSPHONO-D-ALTRITOL",
            "nucleic-acid",
        ),
        "GSZ": (
            "N-[(5R,6R,7R,8S)-6,7-DIHYDROXY-5-(HYDROXYMETHYL)-2-(2-PHENYLETHYL)-1,5,6,7,8,8A-HEXAHYDROIMIDAZO[1,2-A]PYRIDIN-8-YL]-2-METHYLPROPANAMIDE",
            "non-polymer",
        ),
        "A6P": ("6-O-PHOSPHONO-ALPHA-D-ALLOPYRANOSE", "non-polymer"),
        "TMR": ("2,6-DIDEOXY-4-THIOMETHYL-BETA-D-RIBOHEXOPYRANOSIDE", "saccharide"),
        "TIR": ("TIRANDAMYCIN E", "non-polymer"),
        "TIS": ("O-(1,1-DIHYDROXYETHYL)-L-SERINE", "peptide"),
        "GSA": ("O4-SULFONYLGALACTOSE", "saccharide"),
        "TIM": (
            "(2S)-1-(TERT-BUTYLAMINO)-3-[(4-MORPHOLIN-4-YL-1,2,5-THIADIAZOL-3-YL)OXY]PROPAN-2-OL",
            "non-polymer",
        ),
        "GSC": ("2-ETHYLTHIO GLYCINE", "peptide"),
        "GSB": ("S-BENZYL-GLUTATHIONE", "non-polymer"),
        "BP": (
            "(-)-7S,8R,9R-TRIHYDROXY-7,8,9,10-TETRAHYDRO BENZO[A]PYRENE",
            "non-polymer",
        ),
        "GSD": ("4-THIO-D-GLUCOSE", "saccharide"),
        "BR": ("BROMIDE ION", "non-polymer"),
        "GSF": ("L-GAMMA-GLUTAMYL-3-SULFINO-L-ALANYLGLYCINE", "non-polymer"),
        "A6D": (
            "(2R,3R,4S,5S)-2-(6-AMINO-9H-PURIN-9-YL)-5-{[(3-NITROBENZYL)DISULFANYL]METHYL}TETRAHYDROFURAN-3,4-DIOL",
            "non-polymer",
        ),
        "GSH": ("GLUTATHIONE", "non-polymer"),
        "GSK": (
            "6-CHLORO-N-{(3S)-1-[(1S)-1-METHYL-2-(4-MORPHOLINYL)-2-OXO ETHYL]-2-OXO-3-PYRROLIDINYL}-2-NAPHTHALENESULFONAMIDE",
            "non-polymer",
        ),
        "A6G": (
            "2-(2-AMINO-6-OXO-1,6-DIHYDRO-9H-PURIN-9-YL)-1,5-ANHYDRO-2-DEOXY-6-O-PHOSPHONO-D-ALTRITOL",
            "nucleic-acid",
        ),
        "GSM": ("L-GAMMA-GLUTAMYL-S-METHYLCYSTEINYLGLYCINE", "non-polymer"),
        "A6A": (
            "2-(6-AMINO-9H-PURIN-9-YL)-1,5-ANHYDRO-2-DEOXY-6-O-PHOSPHONO-D-ALTRITOL",
            "nucleic-acid",
        ),
        "A6B": (
            "(3S,5S,7S)-N-[6-({5-[(3AS,4S,6AR)-2-OXOHEXAHYDRO-1H-THIENO[3,4-D]IMIDAZOL-4-YL]PENTANOYL}AMINO)HEXYL]TRICYCLO[3.3.1.1~3,7~]DECANE-1-CARBOXAMIDE",
            "non-polymer",
        ),
        "A6C": (
            "2-(4-AMINO-2-OXOPYRIMIDIN-1(2H)-YL)-1,5-ANHYDRO-2-DEOXY-6-O-PHOSPHONO-D-ALTRITOL",
            "nucleic-acid",
        ),
        "26C": ("?", "non-polymer"),
        "26B": (
            "O-TERT-BUTYL-N-[(3-{[(2,4,6-TRIMETHYLPHENYL)CARBAMOYL]AMINO}NAPHTHALEN-2-YL)CARBONYL]-L-THREONINE",
            "peptide",
        ),
        "26A": ("6N-DIMETHYLADENOSINE", "non-polymer"),
        "NNL": (
            "N-[2-(4-CHLOROPHENYL)ETHYL]-6-METHYL[1,2,4]TRIAZOLO[4,3-B]PYRIDAZIN-8-AMINE",
            "non-polymer",
        ),
        "NES": (
            "2-(2-HYDROXY-1,1-DIHYDROXYMETHYL-ETHYLAMINO)-ETHANESULFONIC ACID",
            "non-polymer",
        ),
        "26F": (
            "(2R)-4-(2,6-DIMETHOXYPYRIMIDIN-4-YL)-N-(4-METHOXYBENZYL)-1-{[4-(TRIFLUOROMETHOXY)PHENYL]SULFONYL}PIPERAZINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "NNI": (
            "O-[2-(1,3-DIOXO-1,3-DIHYDRO-2H-ISOINDOL-2-YL)ETHYL] (4-IODOPHENYL)THIOCARBAMATE",
            "non-polymer",
        ),
        "26D": ("PYRIDINE-2,6-DIAMINE", "non-polymer"),
        "NNG": (
            "2-DEOXY-2-{[(S)-HYDROXY(METHYL)PHOSPHORYL]AMINO}-6-O-PHOSPHONO-ALPHA-D-GLUCOPYRANOSE",
            "non-polymer",
        ),
        "NNF": (
            "2-[4-(4-FLUOROPHENYL)PIPERAZIN-1-YL]-6-METHYLPYRIMIDIN-4(3H)-ONE",
            "non-polymer",
        ),
        "NND": (
            "(2R,3R,4R,5S)-2-(HYDROXYMETHYL)-1-NONYLPIPERIDINE-3,4,5-TRIOL",
            "non-polymer",
        ),
        "26O": (
            "N-({(2E)-2-[(2-HYDROXYPHENYL)METHYLIDENE]HYDRAZINO}CARBONOTHIOYL)-BETA-D-GLUCOPYRANOSYLAMINE",
            "non-polymer",
        ),
        "NNB": (
            "O-[2-(1,3-DIOXO-1,3-DIHYDRO-2H-ISOINDOL-2-YL)ETHYL] (4-BROMOPHENYL)THIOCARBAMATE",
            "non-polymer",
        ),
        "NNA": ("NARLAPREVIR", "non-polymer"),
        "26S": (
            "(2R)-4-(2,6-DIMETHOXYPYRIMIDIN-4-YL)-1-[(4-ETHYLPHENYL)SULFONYL]-N-(4-METHOXYBENZYL)PIPERAZINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "26P": ("2-AMINO-6-OXOPIMELIC ACID", "non-polymer"),
        "NNZ": ("2-(AMINOMETHYL)-1-METHYLPYRIDINIUM", "non-polymer"),
        "26U": (
            "N-(4-CARBAMIMIDOYLBENZYL)-1-(4-METHYLPENTANOYL)-L-PROLINAMIDE",
            "non-polymer",
        ),
        "NNX": ("2-(AMINOMETHYL)-4-CHLORO-1-METHYLPYRIDINIUM", "non-polymer"),
        "NNW": ("1-(PYRIDIN-4-YL)METHANAMINE", "non-polymer"),
        "26Z": (
            "4-{[4-AMINO-5-(3-AMINOBENZOYL)-1,3-THIAZOL-2-YL]AMINO}BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "NNU": ("1-(4-CHLOROPYRIDIN-3-YL)METHANAMINE", "non-polymer"),
        "HKA": ("3-METHOXY-4-PHENOXYBENZOIC ACID", "non-polymer"),
        "NNS": ("4-ACETYL-AMINOPHENOL", "non-polymer"),
        "NNR": ("NICOTINAMIDE RIBOSIDE", "non-polymer"),
        "HK8": (
            "(3-CHLORO-11-OXO-10,11-DIHYDRO-5H-DIBENZO[B,E][1,4]DIAZEPIN-8-YL)ACETIC ACID",
            "non-polymer",
        ),
        "HK9": (
            "2-METHOXY-4-(11-OXO-10,11-DIHYDRO-5H-DIBENZO[B,E][1,4]DIAZEPIN-3-YL)BENZOIC ACID",
            "non-polymer",
        ),
        "RHX": ("[2,6-BIS(4-PHENYL)-1,3-OXAZOLIN-2-YL]RHODIUM(III)", "non-polymer"),
        "HK3": (
            "4'-(6,7-DIMETHOXYINDENO[1,2-C]PYRAZOL-3-YL)BIPHENYL-4-OL",
            "non-polymer",
        ),
        "HK0": (
            "3-(3-METHOXY-4-NITROPHENYL)-6-[2-(MORPHOLIN-4-YL)ETHOXY]-5,10-DIHYDRO-11H-DIBENZO[B,E][1,4]DIAZEPIN-11-ONE",
            "non-polymer",
        ),
        "HK1": (
            "4-(6,7-DIMETHOXY-2,4-DIHYDROINDENO[1,2-C]PYRAZOL-3-YL)PHENOL",
            "non-polymer",
        ),
        "HK6": (
            "8-CHLORO-5,10-DIHYDRO-11H-DIBENZO[B,E][1,4]DIAZEPIN-11-ONE",
            "non-polymer",
        ),
        "FN5": ("CMP-3F(A)-NEU5AC", "non-polymer"),
        "HK4": (
            "4-[(6,7-DIMETHOXY-2,4-DIHYDROINDENO[1,2-C]PYRAZOL-3-YL)ETHYNYL]-2-METHOXYPHENOL",
            "non-polymer",
        ),
        "HK5": (
            "4-[6-(1H-IMIDAZOL-1-YLMETHYL)-7-METHOXY-2,4-DIHYDROINDENO[1,2-C]PYRAZOL-3-YL]BENZONITRILE",
            "non-polymer",
        ),
        "263": ("1-(AZIDOMETHYL)-3-METHYLBENZENE", "non-polymer"),
        "262": ("2-FORMYLPHENYL DIHYDROGEN PHOSPHATE", "non-polymer"),
        "261": ("2-ETHOXYPHENOL", "non-polymer"),
        "260": ("2-(BROMOMETHYL)-1,3-DIFLUOROBENZENE", "non-polymer"),
        "267": ("1-(AZIDOMETHYL)-3-METHYLBENZENE", "non-polymer"),
        "266": ("2-(PROPYLSULFANYL)ETHANOL", "non-polymer"),
        "265": ("2-NITROTHIOPHENE", "non-polymer"),
        "264": ("(PHENYLAMINO)ACETONITRILE", "non-polymer"),
        "269": ("(1R)-3-CHLORO-1-PHENYLPROPAN-1-OL", "non-polymer"),
        "268": ("2-PHENOXYETHANOL", "non-polymer"),
        "NN3": (
            "3-{ISOPROPYL[(TRANS-4-METHYLCYCLOHEXYL)CARBONYL]AMINO}-5-PHENYLTHIOPHENE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "NN2": (
            "1-(2-CYCLOPROPYLETHYL)-3-(1,1-DIOXIDO-2H-1,2,4-BENZOTHIADIAZIN-3-YL)-6-FLUORO-4-HYDROXYQUINOLIN-2(1H)-ONE",
            "non-polymer",
        ),
        "NN1": (
            "2-(2-CHLORO-4-FLUOROPHENOXY)-2-METHYL-N-[(1R,2S,3S,5S,7S)-5-(METHYLSULFONYL)-2-ADAMANTYL]PROPANAMIDE",
            "non-polymer",
        ),
        "KAF": (
            "N-{(5S)-4,4-DIHYDROXY-6-PHENYL-5-[(PHENYLCARBONYL)AMINO]HEXANOYL}-L-PHENYLALANINE",
            "non-polymer",
        ),
        "GS1": ("1-THIO-BETA-D-GLUCOPYRANOSE", "saccharide"),
        "GS3": (
            "2-FLUORO-6-{[2-({2-METHOXY-4-[4-(1-METHYLETHYL)PIPERAZIN-1-YL]PHENYL}AMINO)-7H-PYRROLO[2,3-D]PYRIMIDIN-4-YL]AMINO}BENZAMIDE",
            "non-polymer",
        ),
        "GS2": (
            "2-[(2-{[1-(N,N-DIMETHYLGLYCYL)-5-METHOXY-1H-INDOL-6-YL]AMINO}-7H-PYRROLO[2,3-D]PYRIMIDIN-4-YL)AMINO]-6-FLUORO-N-METHYLBENZAMIDE",
            "non-polymer",
        ),
        "A68": (
            "N~2~-ACETYL-N-[(2S,3R)-3-HYDROXY-4-{[(4-METHOXYPHENYL)SULFONYL](2-METHYLPROPYL)AMINO}-1-PHENYLBUTAN-2-YL]-L-ISOLEUCINAMIDE",
            "non-polymer",
        ),
        "A69": (
            "N-{(2S,3R)-4-[(1,3-BENZOTHIAZOL-6-YLSULFONYL)(3-PHENYLPROPYL)AMINO]-3-HYDROXY-1-PHENYLBUTAN-2-YL}-3-HYDROXYBENZAMIDE",
            "non-polymer",
        ),
        "GS7": (
            "2-FLUORO-6-{[2-({2-METHOXY-4-[(METHYLSULFONYL)METHYL]PHENYL}AMINO)-7H-PYRROLO[2,3-D]PYRIMIDIN-4-YL]AMINO}BENZAMIDE",
            "non-polymer",
        ),
        "GS6": (
            "6-CHLORO-N-{(3S)-1-[(1S)-1-METHYL-2-MORPHOLIN-4-YL-2-OXOETHYL]-2-OXOPYRROLIDIN-3-YL}-1-BENZOTHIOPHENE-2-SULFONAMIDE",
            "non-polymer",
        ),
        "GS9": ("NAPHTHALEN-2-YL 1-THIO-BETA-D-GALACTOPYRANOSIDE", "non-polymer"),
        "A66": ("2-AMINOETHYLLYSINE-CARBONYLMETHYLENE-ADENINE", "peptide"),
        "A60": (
            "(3R,3AS,6AR)-HEXAHYDROFURO[2,3-B]FURAN-3-YL [(1S,2R)-3-{[(4-AMINOPHENYL)SULFONYL](2-ETHYLBUTYL)AMINO}-1-BENZYL-2-HYDROXYPROPYL]CARBAMATE",
            "non-polymer",
        ),
        "A61": (
            "(3R,3AS,6AR)-HEXAHYDROFURO[2,3-B]FURAN-3-YL {(1S,2R)-3-[(1,3-BENZOTHIAZOL-6-YLSULFONYL)(2-ETHYLBUTYL)AMINO]-1-BENZYL-2-HYDROXYPROPYL}CARBAMATE",
            "non-polymer",
        ),
        "SQU": ("2,10,23-TRIMETHYL-TETRACOSANE", "non-polymer"),
        "TI3": (
            "[(2S)-2-SULFANYL-3-PHENYLPROPANOYL]-GLY-(5-PHENYLPROLINE)",
            "non-polymer",
        ),
        "X4Z": (
            "(3E)-N-[(1S)-2-(DIMETHYLAMINO)-1-PHENYLETHYL]-6,6-DIMETHYL-3-[(2-METHYLTHIENO[3,2-D]PYRIMIDIN-4-YL)IMINO]-4,6-DIHYDROPYRROLO[3,4-C]PYRAZOLE-5(3H)-CARBOXAMIDE",
            "non-polymer",
        ),
        "RHQ": ("RHODAMINE 6G", "non-polymer"),
        "HCA": ("3-HYDROXY-3-CARBOXY-ADIPIC ACID", "non-polymer"),
        "SQO": (
            "2-{[2-(4-MORPHOLINYL)ETHYL]AMINO}-1,7-DIHYDRO-8H-IMIDAZO[4,5-G]QUINAZOLIN-8-ONE",
            "non-polymer",
        ),
        "B2": (
            "(2R)-5-FLUORO-2-(2-THIENYL)-1,2-DIHYDROQUINAZOLIN-4-AMINE",
            "non-polymer",
        ),
        "RHP": ("(3R)-3-(PROP-2-YNYLAMINO)INDAN-5-OL", "non-polymer"),
        "TMD": ("(6,7-DIHYDRO)4-[(E)-BUTENYL]-4,N-DIMETHYL-THREONINE", "peptide"),
        "SQA": ("2-AZASQUALENE", "non-polymer"),
        "KAI": ("3-(CARBOXYMETHYL)-4-ISOPROPENYLPROLINE", "non-polymer"),
        "OCG": (
            "5-AMINO-N-[5-(ISOXAZOL-3-YLCARBAMOYL)-2-METHYLPHENYL]-1-PHENYL-1H-PYRAZOLE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "SQD": (
            "1,2-DI-O-ACYL-3-O-[6-DEOXY-6-SULFO-ALPHA-D-GLUCOPYRANOSYL]-SN-GLYCEROL",
            "non-polymer",
        ),
        "W84": (
            "5-(7-(5-HYDRO-4-METHYL-2-OXAZOLYL)PHENOXY)HEPTYL)-3-METHYL ISOXAZOLE",
            "non-polymer",
        ),
        "988": ("8-(3-NITROPHENYL)-6-(PYRIDIN-4-YLMETHYL)QUINOLINE", "non-polymer"),
        "989": (
            "2-{[4-(2-ACETYLAMINO-2-PENTYLCARBAMOYL-ETHYL)-NAPHTHALEN-1-YL]-OXALYL-AMINO}-BENZOIC ACID",
            "non-polymer",
        ),
        "TMF": ("5,10-METHYLENE-6-HYDROFOLIC ACID", "non-polymer"),
        "982": (
            "3-(CARBOXYMETHOXY)THIENO[2,3-B]PYRIDINE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "983": (
            "(5S,9R,10R,12S)-12-[3-(AMINOMETHYL)PHENYL]-5-BENZYL-10-HYDROXY-9-ISOPROPYL-3-OXO-1-PHENYL-2-OXA-7-THIA-4,8-DIAZA-10-PHOSPHATRIDECAN-13-OIC ACID 7,7,10-TRIOXIDE",
            "non-polymer",
        ),
        "980": (
            "(2S)-1-(4-{[2-(2-AMINOPYRIMIDIN-5-YL)-7-METHYL-4-(MORPHOLIN-4-YL)THIENO[3,2-D]PYRIMIDIN-6-YL]METHYL}PIPERAZIN-1-YL)-2-HYDROXYPROPAN-1-ONE",
            "non-polymer",
        ),
        "TMG": ("2-(1,3-THIAZOL-4-YL)-1H-BENZIMIDAZOLE", "non-polymer"),
        "U1": ("URANIUM ATOM", "non-polymer"),
        "H06": ("(E)-4-((2-NICOTINOYLHYDRAZONO)METHYL)BENZIMIDAMIDE", "non-polymer"),
        "984": (
            "CYCLOPROPYL-{4-[5-(3,4-DICHLOROPHENYL)-2-[(1-METHYL)-PIPERIDIN]-4-YL-3-PROPYL-3H-IMIDAZOL-4-YL]-PYRIMIDIN-2-YL}AMINE",
            "non-polymer",
        ),
        "985": (
            "(2E)-3-{3-[6-(4-METHYL-1,4-DIAZEPAN-1-YL)PYRAZIN-2-YL]PHENYL}PROP-2-ENOIC ACID",
            "non-polymer",
        ),
        "EEM": (
            "[(3S)-3-AMINO-4-HYDROXY-4-OXO-BUTYL]-[[(2S,3S,4R,5R)-5-(6-AMINOPURIN-9-YL)-3,4-DIHYDROXY-OXOLAN-2-YL]METHYL]-METHYL-SELANIUM",
            "non-polymer",
        ),
        "F41": ("N-(1,3-BENZODIOXOL-5-YL)-2-(PIPERIDIN-1-YL)ACETAMIDE", "non-polymer"),
        "KAM": (
            "N-[7-KETO-8-AMINOPELARGONIC ACID]-[3-HYDROXY-2-METHYL-5-PHOSPHONOOXYMETHYL-PYRIDIN-4-YL-METHANE]",
            "non-polymer",
        ),
        "EEE": ("ETHYL ACETATE", "non-polymer"),
        "EED": (
            "(9BETA,11ALPHA,13ALPHA,14BETA,17ALPHA)-11-(METHOXYMETHYL)ESTRA-1(10),2,4-TRIENE-3,17-DIOL",
            "non-polymer",
        ),
        "EEG": (
            "N-{3-[4-(4-PHENYLTHIOPHEN-2-YL)PHENYL]PROPANOYL}-L-ALPHA-GLUTAMYL-L-ALPHA-GLUTAMYL-AMIDE",
            "non-polymer",
        ),
        "EEF": (
            "N-(3-BIPHENYL-4-YLPROPANOYL)-L-ALPHA-GLUTAMYL-L-ALPHA-GLUTAMYL-AMIDE",
            "non-polymer",
        ),
        "EEA": (
            "N-{3-[3-(3'-CHLOROBIPHENYL-4-YL)ISOXAZOL-5-YL]PROPANOYL}-L-ALPHA-GLUTAMYL-L-ALPHA-GLUTAMYL-AMIDE",
            "non-polymer",
        ),
        "EEC": (
            "N-[3-(3-PHENYLISOXAZOL-5-YL)PROPANOYL]-L-ALPHA-GLUTAMYL-L-ALPHA-GLUTAMYL-AMIDE",
            "non-polymer",
        ),
        "EEB": (
            "URIDINE-DIPHOSPHATE-3(N-ACETYLGLUCOSAMINYL)BUTYRIC ACID",
            "non-polymer",
        ),
        "LI4": ("3-(1-NAPHTHYLMETHOXY)PYRIDIN-2-AMINE", "non-polymer"),
        "LLY": ("NZ-(DICARBOXYMETHYL)LYSINE", "peptide"),
        "04L": (
            "2-METHYL-N-(2-PHENYL-3H-IMIDAZO[4,5-B]PYRIDIN-6-YL)-5-{[(2E)-3-PHENYLPROP-2-ENOYL]AMINO}BENZAMIDE",
            "non-polymer",
        ),
        "04K": (
            "2-({4-[2-(DIETHYLAMINO)ETHOXY]PHENYL}AMINO)-6-(4-FLUOROPHENOXY)-8-METHYLPYRIDO[2,3-D]PYRIMIDIN-7(8H)-ONE",
            "non-polymer",
        ),
        "ROM": ("C-1027 AROMATIZED CHROMOPHORE", "non-polymer"),
        "RON": ("NORVALINE", "peptide"),
        "04H": ("(3S)-3-(AMINOMETHYL)PIPERIDINE-1-CARBOXIMIDAMIDE", "non-polymer"),
        "04G": ("7-HYDROXY-3H-PHENOXAZIN-3-ONE", "non-polymer"),
        "04F": (
            "(1S,2S)-2-AMINO-1-HYDROXY-3-[(3S)-2-OXOPYRROLIDIN-3-YL]PROPANE-1-SULFONIC ACID",
            "non-polymer",
        ),
        "04E": ("IODOACETIC ACID", "non-polymer"),
        "04D": ("(2R,3R,4S)-4-AMINO-2,6-DIMETHYLHEPTANE-1,3-DIOL", "non-polymer"),
        "04C": (
            "1,2,4-TRIDEOXY-4-METHYL-2-{[N-(MORPHOLIN-4-YLACETYL)-L-ALANYL-O-METHYL-L-TYROSYL]AMINO}-1-PHENYL-D-XYLITOL",
            "peptide",
        ),
        "04B": ("2-AMINO-1,2,4-TRIDEOXY-4-METHYL-1-PHENYL-D-XYLITOL", "non-polymer"),
        "04A": (
            "N,N'-[SULFANEDIYLBIS(ETHANE-2,1-DIYL-1,3,4-THIADIAZOLE-5,2-DIYL)]BIS(2-PHENYLACETAMIDE)",
            "non-polymer",
        ),
        "LLT": ("L-DEOXYTHYMIDINE", "non-polymer"),
        "BD5": (
            "N-[(BENZYLOXY)CARBONYL]-D-TYROSYL-L-PHENYLALANYL-L-TYROSINE",
            "non-polymer",
        ),
        "BD4": (
            "6-(3-AMINOPROPYL)-4,9-DIMETHYLPYRROLO[3,4-C]CARBAZOLE-1,3(2H,6H)-DIONE",
            "non-polymer",
        ),
        "BD7": (
            "N-[(BENZYLOXY)CARBONYL]-L-HISTIDYL-N-METHYL-L-PHENYLALANYL-N-HYDROXY-L-TYROSINAMIDE",
            "non-polymer",
        ),
        "BD6": (
            "N-[(BENZYLOXY)CARBONYL]-L-HISTIDYL-N-METHYL-L-PHENYLALANYL-N-HYDROXY-L-TRYPTOPHANAMIDE",
            "non-polymer",
        ),
        "BD1": ("3A,4,7,7A-TETRAHYDRO-BENZO [1,3] DIOXOL-2-ONE", "non-polymer"),
        "BD0": (
            "(1S,2R)-1-[(5-CHLORO-2,4-DIHYDROXYPHENYL)CARBONYL]-2-{4-[(3,3-DIFLUOROPYRROLIDIN-1-YL)METHYL]PHENYL}PYRROLIDINIUM",
            "non-polymer",
        ),
        "BD3": (
            "NALPHA-[(3-TERT-BUTYL-1-METHYL-1H-PYRAZOL-5-YL)CARBONYL]-N-[(2E)-2-IMINOETHYL]-3-{5-[(Z)-IMINOMETHYL]-1,3,4-OXADIAZOL-2-YL}-L-PHENYLALANINAMIDE",
            "non-polymer",
        ),
        "BD2": (
            "3-[(4-HYDROXYBENZOYL)AMINO]AZEPAN-4-YL 4-(2-HYDROXYBENZOYL)BENZOATE",
            "non-polymer",
        ),
        "LLC": (
            "[6-HYDROXY-2-(4-HYDROXYPHENYL)-1-BENZOTHIEN-3-YL][4-(2-PYRROLIDIN-1-YLETHOXY)PHENYL]METHANONE",
            "non-polymer",
        ),
        "04V": ("(2S)-PYRROLIDINE-2-CARBOXIMIDIC ACID", "peptide"),
        "04U": (
            "AZANYLIDENE-[(3S,5S)-5-CARBOXYPYRROLIDIN-3-YL]IMINO-AZANIUM",
            "peptide",
        ),
        "ROS": ("N,N'-TETRAMETHYL-ROSAMINE", "non-polymer"),
        "BD9": (
            "N-UNDECANOYL-L-HISTIDYL-L-HISTIDYL-N-METHYL-N-(2-PYRIDIN-2-YLETHYL)-L-TYROSINAMIDE",
            "non-polymer",
        ),
        "BD8": (
            "N-UNDECANOYL-L-HISTIDYL-L-PHENYLALANYL-N-METHYL-N-(2-PYRIDIN-2-YLETHYL)-L-TYROSINAMIDE",
            "non-polymer",
        ),
        "?": ("DISORDERED SOLVENT", "non-polymer"),
        "04P": ("(3R)-3-(HYDROXYMETHYL)PIPERIDINE-1-CARBOXIMIDAMIDE", "non-polymer"),
        "BDE": (
            "N'-[(1E)-(3,5-DIBROMO-2,4-DIHYDROXYPHENYL)METHYLENE]NICOTINOHYDRAZIDE",
            "non-polymer",
        ),
        "BDD": ("BROMO-DODECANOL", "non-polymer"),
        "BDG": ("O-2,6-DIAMINO-2,6-DIDEOXY-ALPHA-D-GLUCOPYRANOSE", "saccharide"),
        "BDF": ("BETA-D-FRUCTOPYRANOSE", "non-polymer"),
        "BDA": ("4-METHYLBENZYL-N-BIS[DAUNOMYCIN]", "non-polymer"),
        "LI2": (
            "3-FLUORO-5-MORPHOLIN-4-YL-N-[3-(2-PYRIDIN-4-YLETHYL)-1H-INDOL-5-YL]BENZAMIDE",
            "non-polymer",
        ),
        "BDC": (
            "BIS-(9-OCTYLAMINO(2-DIMETHYLAMINOETHYL)ACRIDINE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "BDB": ("?", "non-polymer"),
        "BDM": ("BRODIMOPRIM-4,6-DICARBOXYLATE", "non-polymer"),
        "BDL": ("N-(BIPHENYL-4-YLSULFONYL)-D-LEUCINE", "non-polymer"),
        "BDO": (
            "(2Z,5R)-3-(3-CHLOROBENZYL)-2-IMINO-5-METHYL-5-(2-METHYLPROPYL)IMIDAZOLIDIN-4-ONE",
            "non-polymer",
        ),
        "BDN": ("BENZAMIDINE", "non-polymer"),
        "BDI": (
            "2-BUTYL-5,6-DIHYDRO-1H-IMIDAZO[4,5-D]PYRIDAZINE-4,7-DIONE",
            "non-polymer",
        ),
        "BDH": ("L-BETA-ASPARTYLHISTIDINE", "non-polymer"),
        "BDK": (
            "2-[5-AMINO-6-OXO-2-(2-THIENYL)-1,6-DIHYDROPYRIMIDIN-1-YL)-N-[3,3-DIFLUORO -1-ISOPROPYL-2-OXO-3-(N-(2-MORPHOLINO ETHYL)CARBAMOYL]PROPYL]ACETAMIDE",
            "non-polymer",
        ),
        "BDJ": (
            "(2Z)-3-(3-CHLOROBENZYL)-2-IMINO-5,5-DIMETHYLIMIDAZOLIDIN-4-ONE",
            "non-polymer",
        ),
        "BDU": ("ME-A-9-N-(BIPHENYL-4-CARBONYL)-AMINO-9-DEOXY-NEU5AC", "non-polymer"),
        "BDT": ("5-BETA-DIHYDROTESTOSTERONE", "non-polymer"),
        "BDW": (
            "(4S)-1-(4-{[(2Z,4R)-4-(2-CYCLOHEXYLETHYL)-4-(CYCLOHEXYLMETHYL)-2-IMINO-5-OXOIMIDAZOLIDIN-1-YL]METHYL}BENZYL)-4-PROPYLIMIDAZOLIDIN-2-ONE",
            "non-polymer",
        ),
        "BDV": (
            "1-BUTYL-3-(4-{[(2Z,4R)-2-IMINO-4-METHYL-4-(2-METHYLPROPYL)-5-OXOIMIDAZOLIDIN-1-YL]METHYL}BENZYL)UREA",
            "non-polymer",
        ),
        "BDQ": (
            "1-(4-CYANOPHENYL)-3-(4-{[(2Z,4R)-2-IMINO-4-METHYL-4-(2-METHYLPROPYL)-5-OXOIMIDAZOLIDIN-1-YL]METHYL}BENZYL)UREA",
            "non-polymer",
        ),
        "BDP": ("BETA-D-GLUCOPYRANURONIC ACID", "saccharide"),
        "049": (
            "2-(ACETYLAMINO)-4,5-ANHYDRO-1,2-DIDEOXY-4-METHYL-1-PHENYL-D-XYLITOL",
            "non-polymer",
        ),
        "BDR": ("BETA-D-RIBOFURANOSYL", "saccharide"),
        "047": (
            "1-CYCLOHEXYL-N-{[1-(4-METHYLPHENYL)-1H-INDOL-3-YL]METHYL}METHANAMINE",
            "non-polymer",
        ),
        "046": (
            "5-{[6-(ACETYLAMINO)PYRIMIDIN-4-YL]OXY}-N-{4-[(4-METHYLPIPERAZIN-1-YL)METHYL]-3-(TRIFLUOROMETHYL)PHENYL}-2,3-DIHYDRO-1H-INDOLE-1-CARBOXAMIDE",
            "non-polymer",
        ),
        "045": (
            "BETA-D-FRUCTOFURANOSYL-(2->6)-BETA-D-FRUCTOFURANOSYL ALPHA-D-GLUCOPYRANOSIDE",
            "saccharide",
        ),
        "044": (
            "N-{3-FLUORO-4-[(7-METHOXYQUINOLIN-4-YL)OXY]PHENYL}-1-[(2R)-2-HYDROXYPROPYL]-5-METHYL-3-OXO-2-PHENYL-2,3-DIHYDRO-1H-PYRAZOLE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "BDY": (
            "N~3~-(2,6-DICHLOROBENZYL)-5-(4-{[(2R)-2-(PYRROLIDIN-1-YLMETHYL)PYRROLIDIN-1-YL]CARBONYL}PHENYL)PYRAZINE-2,3-DIAMINE",
            "non-polymer",
        ),
        "BDX": (
            "(2E,5R)-5-(2-CYCLOHEXYLETHYL)-5-(CYCLOHEXYLMETHYL)-2-IMINO-3-METHYLIMIDAZOLIDIN-4-ONE",
            "non-polymer",
        ),
        "041": (
            "2-(3-FLUORO-4-HYDROXYPHENYL)-7-VINYL-1,3-BENZOXAZOL-5-OL",
            "non-polymer",
        ),
        "BDZ": (
            "6-DEOXY-ALPHA-L-GALACTOPYRANOSYL-(1->3)-[6-DEOXY-ALPHA-L-GALACTOPYRANOSYL-(1->2)-BETA-D-GALACTOPYRANOSYL-(1->4)]-2-(ACETYLAMINO)-2-DEOXY-BETA-D-GLUCOPYRANOSE",
            "non-polymer",
        ),
        "RHB": (
            "N-[9-(2-CARBOXYPHENYL)-6-(DIETHYLAMINO)-3H-XANTHEN-3-YLIDENE]-N-ETHYLETHANAMINIUM",
            "non-polymer",
        ),
        "W8R": (
            "5-(7-(5-HYDRO-4-METHYL-2-OXAZOLYL)PHENOXY)HEPTYL)-3-METHYL ISOXAZOLE",
            "non-polymer",
        ),
        "LI8": (
            "N-(3-{[5-BROMO-2-({3-[(PYRROLIDIN-1-YLCARBONYL)AMINO]PHENYL}AMINO)PYRIMIDIN-4-YL]AMINO}PROPYL)-2,2-DIMETHYLMALONAMIDE",
            "non-polymer",
        ),
        "98P": (
            "D-PHENYLALANYL-N-[(4-CHLORO-1-METHYLPYRIDINIUM-3-YL)METHYL]-L-PROLINAMIDE",
            "peptide",
        ),
        "F42": ("COENZYME F420", "non-polymer"),
        "HIS": ("HISTIDINE", "peptide"),
        "LI9": (
            "4-(4-FLUOROPHENYL)-1-METHYL-5-(2-{[(1S)-1-PHENYLETHYL]AMINO}PYRIMIDIN-4-YL)-2-PIPERIDIN-4-YL-1,2-DIHYDRO-3H-PYRAZOL-3-ONE",
            "non-polymer",
        ),
        "98J": ("5-HYDROXY-3-OXOPENTANOIC ACID", "non-polymer"),
        "ULI": ("TETRACONTANE", "non-polymer"),
        "HAA": ("2-(3,4-DIHYDROXYPHENYL)ACETIC ACID", "non-polymer"),
        "HWG": (
            "N-(TERT-BUTYL)-3,5-DIMETHYL-N'-[(5-METHYL-2,3-DIHYDRO-1,4-BENZODIOXIN-6-YL)CARBONYL]BENZOHYDRAZIDE",
            "non-polymer",
        ),
        "H0E": (
            "(10S)-3,3-DIMETHYL-8-{[(2S)-2-(PHENOXYMETHYL)PYRROLIDIN-1-YL]SULFONYL}-2,3,4,10-TETRAHYDROPYRIMIDO[1,2-A]INDOL-10-OL",
            "non-polymer",
        ),
        "IDZ": ("3-(1H-BENZIMIDAZOL-2-YL)-1H-INDAZOLE", "non-polymer"),
        "IDX": ("2-O-SULFO-ALPHA-L-GULOPYRANURONIC ACID", "saccharide"),
        "IDY": ("METHYL 2-O-SULFO-ALPHA-L-IDOPYRANOSIDURONIC ACID", "saccharide"),
        "IDT": ("4,5-DEHYDRO-L-IDURONIC ACID", "saccharide"),
        "IDU": ("2-O-SULFO-BETA-L-ALTROPYRANURONIC ACID", "saccharide"),
        "IDR": ("L-IDURONIC ACID", "saccharide"),
        "IDS": ("2-O-SULFO-ALPHA-L-IDOPYRANURONIC ACID", "saccharide"),
        "IDP": ("INOSINE-5'-DIPHOSPHATE", "non-polymer"),
        "IDN": (
            "(E)-N-METHYL-N-(1-METHYL-1H-INDOL-3-YLMETHYL)-3-(7-OXO-5,6,7,8-TETRAHYDRO-[1,8]NAPHTHYRIDIN-3-YL)-ACRYLAMIDE",
            "non-polymer",
        ),
        "IDO": ("IODO GROUP", "non-polymer"),
        "IDL": ("IMIDAZO[1,2-A]PYRIDIN-3-YLACETIC ACID", "non-polymer"),
        "IDM": ("INDOLINE", "non-polymer"),
        "IDH": ("(5S)-5-IODODIHYDRO-2,4(1H,3H)-PYRIMIDINEDIONE", "non-polymer"),
        "IDI": ("7-IODO-1,2,3,4-TETRAHYDRO-ISOQUINOLINE", "non-polymer"),
        "IDG": ("O-2,6-DIAMINO-2,6-DIDEOXY-BETA-L-IDOPYRANOSE", "saccharide"),
        "IDD": (
            "(5R,6R,7S,8S)-5-(HYDROXYMETHYL)-2-PHENYL-5,6,7,8-TETRAHYDROIMIDAZO[1,2-A]PYRIDINE-6,7,8-TRIOL",
            "non-polymer",
        ),
        "IDE": (
            "(5R,6R,7S,8S)-3-(ANILINOMETHYL)-5,6,7,8-TETRAHYDRO-5-(HYDROXYMETHYL)-IMIDAZO[1,2-A]PYRIDINE-6,7,8-TRIOL",
            "non-polymer",
        ),
        "IDB": (
            "3-[5-[(3-CARBOXY-2,4,6-TRIIODO-PHENYL)CARBAMOYL]PENTANOYLAMINO]-2,4,6-TRIIODO-BENZOIC ACID",
            "non-polymer",
        ),
        "IDC": (
            "5-HYDROXYMETHYL-5,6,7,8-TETRAHYDRO-IMIDAZO[1,2-A]PYRIDIN-6YL-7,8-DIOL-GLUCOPYRANOSIDE",
            "non-polymer",
        ),
        "72X": (
            "(3R)-3-[(1S)-1-(3-CHLOROPHENYL)-1-HYDROXY-5-METHOXYPENTYL]-N-{(1S)-2-CYCLOHEXYL-1-[(METHYLAMINO)METHYL]ETHYL}PIPERIDINE-1-CARBOXAMIDE",
            "non-polymer",
        ),
        "IDA": (
            "(2-CARBAMOYLMETHYL-5-PROPYL-OCTAHYDRO-INDOL-7-YL)ACETIC ACID",
            "non-polymer",
        ),
        "O62": (
            "6-(4-{[3-(2,6-DICHLOROPHENYL)-5-(1-METHYLETHYL)ISOXAZOL-4-YL]METHOXY}PHENYL)NAPHTHALENE-1-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "CGH": ("N-CYCLOHEPTYLGLYCINE", "peptide"),
        "CGK": ("?", "non-polymer"),
        "O61": (
            "N-[(2S)-1-({2-[5-CHLORO-2-(1H-TETRAZOL-1-YL)PHENYL]ETHYL}AMINO)-1-OXO-3-PHENYLPROPAN-2-YL]-3-OXO-3,4-DIHYDRO-2H-1,4-BENZOTHIAZINE-7-CARBOXAMIDE",
            "peptide",
        ),
        "PMP": ("4'-DEOXY-4'-AMINOPYRIDOXAL-5'-PHOSPHATE", "non-polymer"),
        "CGL": ("CYSTINE-GLUTATHIONE", "non-polymer"),
        "CGO": ("CONGO RED", "non-polymer"),
        "CGN": ("5-OXO-PYRROLIDINE-2-CARBALDEHYDE", "non-polymer"),
        "CGA": ("CARBOXYMETHYLATED GLUTAMIC ACID", "peptide"),
        "TVN": (
            "N-{(2R)-4-(METHYLAMINO)-4-OXO-2-[4-(PHOSPHONOOXY)BENZYL]BUTANOYL}-L-VALYL-L-ASPARTAMIDE",
            "non-polymer",
        ),
        "CGB": ("CALYSTEGINE B2", "non-polymer"),
        "CGE": ("CLOPIDOGREL", "non-polymer"),
        "PMY": ("AGLYCON OF PEPLOMYCIN", "non-polymer"),
        "PMZ": (
            "1-[10-(3-DIMETHYLAMINO-PROPYL)-10H-PHENOTHIAZIN-2-YL]-ETHANONE",
            "non-polymer",
        ),
        "CGF": ("C-(1-AZIDO-ALPHA-D-GLUCOPYRANOSYL) FORMAMIDE", "saccharide"),
        "PMD": (
            "[N-(2,4-DIAMINOPTERIDIN-6-YL)-METHYL]-DIBENZ[B,F]AZEPINE",
            "non-polymer",
        ),
        "P2M": ("N-(PYRIDIN-2-YLMETHYL)NAPHTHALENE-1-SULFONAMIDE", "non-polymer"),
        "HAC": ("BETA-CYCLOHEXYL-ALANINE", "peptide"),
        "PMG": (
            "N-({3-HYDROXY-2-METHYL-5-[(PHOSPHONOOXY)METHYL]PYRIDIN-4-YL}METHYL)-2-METHYL-L-GLUTAMIC ACID",
            "non-polymer",
        ),
        "FE2": ("FE (II) ION", "non-polymer"),
        "FE3": ("4-(2-AMINOETHYL)-3-CHLOROBENZOIC ACID", "non-polymer"),
        "PMB": ("PARA-MERCURY-BENZENESULFONIC ACID", "non-polymer"),
        "FE1": (
            "METHYL 1-(4-{[(2,4-DIAMINOPTERIDIN-6-YL)METHYL](METHYL)AMINO}BENZOYL)PIPERIDINE-4-CARBOXYLATE",
            "non-polymer",
        ),
        "CGQ": (
            "3-[C-[N'-(3-CARBAMIMIDOYL-BENZYLIDENIUM)-HYDRAZINO]-[[AMINOMETHYLIDENE]AMINIUM]-IMINOMETHYL]-BENZAMIDINIUM",
            "non-polymer",
        ),
        "CGP": (
            "2'-DEOXYCYTIDINE-2'-DEOXYGUANOSINE-3',5'-MONOPHOSPHATE",
            "non-polymer",
        ),
        "CGS": (
            "N-HYDROXY-2(R)-[[(4-METHOXYPHENYL)SULFONYL](3-PICOLYL)AMINO]-3-METHYLBUTANAMIDE HYDROCHLORIDE",
            "non-polymer",
        ),
        "CGR": ("CYCLIC GUANOSINE DIPHOSPHATE-RIBOSE", "non-polymer"),
        "CGU": ("GAMMA-CARBOXY-GLUTAMIC ACID", "peptide"),
        "CGT": ("CARBA-GLUCOTROPAEOLIN", "non-polymer"),
        "PMJ": (
            "N~2~-(BENZYLSULFONYL)-D-ARGINYL-N-(4-CARBAMIMIDOYLBENZYL)GLYCINAMIDE",
            "peptide",
        ),
        "FEF": ("?", "non-polymer"),
        "FEG": (
            "5'-O-[(S)-{[2-(CARBOXYMETHYL)-6-HYDROXY-3,5-DIMETHYLPYRIDIN-4-YL]OXY}(HYDROXY)PHOSPHORYL]GUANOSINE",
            "non-polymer",
        ),
        "PM6": ("9H-PURINE-6-THIOL", "non-polymer"),
        "FEE": (
            "N-{(1S,2S,3R)-1-[(ALPHA-D-GALACTOPYRANOSYLOXY)METHYL]-2,3-DIHYDROXYOCTYL}TETRACOSANAMIDE",
            "non-polymer",
        ),
        "FEB": (
            "N~2~-[(3R)-3-HYDROXYDODECANOYL]-L-ASPARAGINYL-N~1~-[(1S)-1-(HYDROXYMETHYL)-3-METHYLBUTYL]-L-GLUTAMAMIDE",
            "non-polymer",
        ),
        "FEC": (
            "1,3,5,8-TETRAMETHYL-PORPHINE-2,4,6,7-TETRAPROPIONIC ACID FERROUS COMPLEX",
            "non-polymer",
        ),
        "PM2": (
            "1-[3-(1-{[5-(2-PHENYLETHYL)PYRIDIN-3-YL]CARBONYL}PIPERIDIN-4-YL)PHENYL]METHANAMINE",
            "non-polymer",
        ),
        "FEA": ("MONOAZIDO-MU-OXO-DIIRON", "non-polymer"),
        "FEN": ("N-(4-HYDROXYPHENYL)ALL-TRANS RETINAMIDE", "non-polymer"),
        "FEO": ("MU-OXO-DIIRON", "non-polymer"),
        "FEL": ("HYDRATED FE", "non-polymer"),
        "FEM": ("N-(2-FERROCENYLETHYL)MALEIMIDE", "non-polymer"),
        "PM8": (
            "S-(2-{[N-(2-HYDROXY-4-{[HYDROXY(OXIDO)PHOSPHINO]OXY}-3,3-DIMETHYLBUTANOYL)-BETA-ALANYL]AMINO}ETHYL) DECANETHIOATE",
            "non-polymer",
        ),
        "PM9": (
            "(2E)-2-{[(Z)-{3-HYDROXY-2-METHYL-5-[(PHOSPHONOOXY)METHYL]PYRIDIN-4(1H)-YLIDENE}METHYL]IMINO}-4-(METHYLSULFANYL)BUTANOIC ACID",
            "non-polymer",
        ),
        "P2H": ("1-(2-HYDROXY-2,2-DIPHOSPHONOETHYL)-3-PHENYLPYRIDINIUM", "non-polymer"),
        "FER": ("3-(4-HYDROXY-3-METHOXYPHENYL)-2-PROPENOIC ACID", "non-polymer"),
        "FES": ("FE2/S2 (INORGANIC) CLUSTER", "non-polymer"),
        "FEP": (
            "[(4-{4-[4-(DIFLUORO-PHOSPHONO-METHYL)-PHENYL]-BUTYL}-PHENYL)-DIFLUORO-METHYL]-PHOSPHONIC ACID",
            "non-polymer",
        ),
        "L3G": (
            "N-(4-{4-AMINO-1-[1-(TETRAHYDRO-2H-PYRAN-4-YL)PIPERIDIN-4-YL]-1H-PYRAZOLO[3,4-D]PYRIMIDIN-3-YL}-2-METHOXYPHENYL)-1-METHYL-1H-INDOLE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "CG1": ("5'-O-[(R)-HYDROXY(METHOXY)PHOSPHORYL]GUANOSINE", "nucleic-acid"),
        "CG3": ("HYDROXY(2-HYDROXYPHENYL)OXOAMMONIUM", "non-polymer"),
        "CG2": ("CYTIDYL-2'-5'-PHOSPHO-GUANOSINE", "non-polymer"),
        "FEX": ("FEXARAMINE", "non-polymer"),
        "ID8": ("2-[(2,3-DIMETHYLPHENYL)AMINO]BENZOIC ACID", "non-polymer"),
        "P2T": ("?", "nucleic-acid"),
        "ID4": (
            "4-[4-(CYCLOHEXYLMETHYL)-1H-1,2,3-TRIAZOL-1-YL]-2,3,5,6-TETRAFLUOROBENZENESULFONAMIDE",
            "non-polymer",
        ),
        "ID5": (
            "[5-FLUORO-2-({[(4,5,7-TRIFLUORO-1,3-BENZOTHIAZOL-2-YL)METHYL]AMINO}CARBONYL)PHENOXY]ACETIC ACID",
            "non-polymer",
        ),
        "ID2": ("5-IODODEOXYURIDINE", "non-polymer"),
        "ID3": (
            "(3E)-4-(2,6,6-TRIMETHYLCYCLOHEX-1-EN-1-YL)BUT-3-EN-2-ONE",
            "non-polymer",
        ),
        "SEP": ("PHOSPHOSERINE", "peptide"),
        "727": (
            "(5R)-5-[(5-METHOXY-3-OXO-1,3-DIHYDRO-2H-INDAZOL-2-YL)METHYL]-5-PHENYLIMIDAZOLIDINE-2,4-DIONE",
            "non-polymer",
        ),
        "724": (
            "(5R)-5-[(5-METHOXY-3-OXO-1,3-DIHYDRO-2H-INDAZOL-2-YL)METHYL]-5-METHYLIMIDAZOLIDINE-2,4-DIONE",
            "non-polymer",
        ),
        "722": ("1-(2,2-DIPHOSPHONOETHYL)-3-(HEPTYLOXY)PYRIDINIUM", "non-polymer"),
        "723": (
            "(6-CHLORO-2-OXO-4-PHENYL-1,2-DIHYDROQUINOLIN-3-YL)ACETIC ACID",
            "non-polymer",
        ),
        "720": (
            "(2S)-5-AMINO-2-[(1-PROPYL-1H-IMIDAZOL-4-YL)METHYL]PENTANOIC ACID",
            "non-polymer",
        ),
        "721": ("3-BUTOXY-1-(2,2-DIPHOSPHONOETHYL)PYRIDINIUM", "non-polymer"),
        "SUY": (
            "(1R,5S,6S)-3-[5-CHLORO-6-ETHYL-2-(PYRIMIDIN-5-YLSULFANYL)-7H-PYRROLO[2,3-D]PYRIMIDIN-4-YL]-3-AZABICYCLO[3.1.0]HEXAN-6-AMINE",
            "non-polymer",
        ),
        "PGL": ("(AMINOMETHYL)PHOSPHONIC ACID", "peptide"),
        "MMU": ("1,3-DIMETHYLUREA", "non-polymer"),
        "MSY": (
            "3-(3-METHYL-5-OXO-4,5-DIHYDRO-1H-PYRAZOL-1-YL)BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "MSU": ("SUCCINIC ACID MONOMETHYL ESTER", "non-polymer"),
        "MST": ("2-T-BUTYLAMINO-4-ETHYLAMINO-6-METHYLTHIO-S-TRIAZINE", "non-polymer"),
        "MSS": ("(MOLYBDOPTERIN-S,S)-OXO-MOLYBDENUM", "non-polymer"),
        "MSR": ("4-(1H-IMIDAZOL-1-YL)PHENOL", "non-polymer"),
        "MSQ": ("4-[3-METHYLSULFANYLANILINO]-6,7-DIMETHOXYQUINAZOLINE", "non-polymer"),
        "MSP": ("5'-O-[(L-METHIONYL)-SULPHAMOYL]ADENOSINE", "peptide"),
        "MSO": ("SELENOMETHIONINE SELENOXIDE", "peptide"),
        "MSN": (
            "(1R,2R,3R,4S,5R)-4-AMINO-5-(METHYLTHIO)CYCLOPENTANE-1,2,3-TRIOL",
            "non-polymer",
        ),
        "MSM": ("(METHYLSULFANYL)METHANE", "non-polymer"),
        "MSL": ("(2S)-2-AMINO-4-(METHYLSULFONIMIDOYL)BUTANOIC ACID", "peptide"),
        "MSJ": (
            "N-{5-[(2,4-DIOXO-3,4-DIHYDROPYRIMIDIN-1(2H)-YL)METHOXY]-2-METHYLPENTAN-2-YL}BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "MSI": (
            "1-METHYL-5-(2-PHENOXYMETHYL-PYRROLIDINE-1-SULFONYL)-1H-INDOLE-2,3-DIONE",
            "non-polymer",
        ),
        "MSH": ("3-METHANESULFONYL-HEXANE", "non-polymer"),
        "MSG": ("7-METHYL-6-THIO-GUANOSINE", "non-polymer"),
        "MSF": ("[METHYLSELENO]ACETATE", "non-polymer"),
        "MSE": ("SELENOMETHIONINE", "peptide"),
        "MSD": (
            "1-[2-(3-ACETYL-2-HYDROXY-6-METHOXY-PHENYL)-CYCLOPROPYL]-3-(5-CYANO-PYRIDIN-2-YL)-THIOUREA",
            "non-polymer",
        ),
        "MSC": (
            "2,5-DIBENZYLOXY-3,4-DIHYDROXY-HEXANEDIOIC ACID 2-CHLORO-6-FLUORO-BENZYLAMIDE (2-HYDROXY-INDAN-1- YL)-AMIDE",
            "non-polymer",
        ),
        "MSB": ("1-METHYLOXY-4-SULFONE-BENZENE", "non-polymer"),
        "MSA": ("(2-S-METHYL) SARCOSINE", "peptide"),
        "HCG": ("DELTA-(L-ALPHA-AMINOADIPOYL)-L-CYSTEINYL-GLYCINE", "non-polymer"),
        "AQU": (
            "1-[1-(3-AMINOPHENYL)-3-TERT-BUTYL-1H-PYRAZOL-5-YL]-3-[4-(QUINAZOLIN-4-YLAMINO)PHENYL]UREA",
            "non-polymer",
        ),
        "5Y0": (
            "N-[(2S)-2,3-BIS(OXIDANYL)PROPOXY]-3,4-BIS(FLUORANYL)-2-[(2-FLUORANYL-4-IODANYL-PHENYL)AMINO]BENZAMIDE",
            "non-polymer",
        ),
        "DW0": (
            "1,1'-PROPANE-1,3-DIYLBIS(3,7-DIMETHYL-3,7-DIHYDRO-1H-PURINE-2,6-DIONE)",
            "non-polymer",
        ),
        "DW1": ("RUTHENIUM PYRIDOCARBAZOLE", "non-polymer"),
        "AQP": ("ADENOSINE-5'-TETRAPHOSPHATE", "non-polymer"),
        "AQS": ("N,N-BIS(3-AMINOPROPYL)-2-ANTHRAQUINONESULFONAMIDE", "non-polymer"),
        "M86": (
            "2-(2-OXOIMIDAZOLIDIN-1-YL)ETHYL [(2S,3R)-3-HYDROXY-4-{[(4-METHOXYPHENYL)SULFONYL][(2S)-2-METHYLBUTYL]AMINO}-1-PHENYLBUTAN-2-YL]CARBAMATE",
            "non-polymer",
        ),
        "M84": ("3-{4-[(PHENYLCARBONYL)AMINO]PHENYL}PROPANOIC ACID", "non-polymer"),
        "CI1": (
            "(2Z)-4-[(3S)-1-(BENZYLSULFONYL)-3-(4-CHLOROBENZYL)PIPERIDIN-3-YL]-2-HYDROXY-4-OXOBUT-2-ENOIC ACID",
            "non-polymer",
        ),
        "M80": (
            "3-[4-({N-[(BENZYLOXY)CARBONYL]-L-PHENYLALANYL}AMINO)PHENYL]PROPANOIC ACID",
            "non-polymer",
        ),
        "AQZ": (
            "N-(3-{[7-METHOXY-6-(2-PYRROLIDIN-1-YLETHOXY)QUINAZOLIN-4-YL]AMINO}-4-METHYLPHENYL)-2-MORPHOLIN-4-YLISONICOTINAMIDE",
            "non-polymer",
        ),
        "AQA": ("4-DEOXY-BETA-L-THREO-HEX-4-ENOPYRANURONIC ACID", "non-polymer"),
        "MLK": ("METHYLLYCACONITINE", "non-polymer"),
        "AQC": ("9,10-DIOXO-9,10-DIHYDROANTHRACENE-2-CARBOXAMIDE", "non-polymer"),
        "AQB": (
            "N-(4-{[4-({[1-(3-AMINOPHENYL)-3-TERT-BUTYL-1H-PYRAZOL-5-YL]CARBAMOYL}AMINO)PHENYL]AMINO}QUINAZOLIN-6-YL)-4-(DIMETHYLAMINO)BUTANAMIDE",
            "non-polymer",
        ),
        "AQM": (
            "N-(4-{[4-({[1-(3-AMINOPHENYL)-3-TERT-BUTYL-1H-PYRAZOL-5-YL]CARBAMOYL}AMINO)PHENYL]AMINO}QUINAZOLIN-6-YL)-3-(4-METHYLPIPERAZIN-1-YL)PROPANAMIDE",
            "non-polymer",
        ),
        "AQO": ("2-AMINOQUINAZOLIN-4(3H)-ONE", "non-polymer"),
        "AQN": ("9,10-DIOXO-9,10-DIHYDROANTHRACENE-2-SULFONIC ACID", "non-polymer"),
        "OOP": (
            "1-DEOXY-1-[(4AR)-4A-[(2R)-1-HYDROXY-5-{[(S)-HYDROXY(PHOSPHONOOXY)PHOSPHORYL]OXY}-3-METHYLIDENEPENTAN-2-YL]-7,8-DIMETHYL-2,4-DIOXO-3,4,4A,5-TETRAHYDROBENZO[G]PTERIDIN-10(2H)-YL]-5-O-PHOSPHONO-D-RIBITOL",
            "non-polymer",
        ),
        "AQ4": (
            "[6,7-BIS(2-METHOXY-ETHOXY)QUINAZOLINE-4-YL]-(3-ETHYNYLPHENYL)AMINE",
            "non-polymer",
        ),
        "DWZ": (
            "(2S,3R,4S)-4-{[(3S,5S)-5-(DIMETHYLCARBAMOYL)PYRROLIDIN-3-YL]SULFANYL}-2-[(1S,2R)-1-FORMYL-2-HYDROXYPROPYL]-3-METHYL-3,4-DIHYDRO-2H-PYRROLE-5-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "CI4": (
            "2-(3-CHLORO-2-FLUOROBENZYL)-6,7-DIHYDROXY-2,3-DIHYDRO-1H-ISOINDOL-1-ONE",
            "non-polymer",
        ),
        "DWB": (
            "(4S)-2'-(5-CHLORO-2-FLUOROPHENYL)-7'-METHOXYSPIRO[1,3-OXAZOLE-4,9'-XANTHEN]-2-AMINE",
            "non-polymer",
        ),
        "DWC": ("PYRIDOCARBAZOLE CYCLOPENTADIENYL OS(CO) COMPLEX", "non-polymer"),
        "DWA": (
            "(4R)-4-[3-(2-FLUOROPYRIDIN-3-YL)PHENYL]-4-(4-METHOXYPHENYL)-4,5-DIHYDRO-1,3-OXAZOL-2-AMINE",
            "non-polymer",
        ),
        "DWD": (
            "(4S)-2'-(2-METHYLPROPOXY)-7'-(PYRIMIDIN-5-YL)SPIRO[1,3-OXAZOLE-4,9'-XANTHEN]-2-AMINE",
            "non-polymer",
        ),
        "M8F": (
            "N-[(5S)-5-AMINO-5-CARBOXYPENTANOYL]-L-CYSTEINYL-D-METHIONINE",
            "non-polymer",
        ),
        "DWH": (
            "METHYL N-{4-CHLORO-6-[(4-SULFAMOYLPHENYL)AMINO]-1,3,5-TRIAZIN-2-YL}GLYCINATE",
            "non-polymer",
        ),
        "M8E": (
            "5'-{[4-(AMINOOXY)BUTYL](METHYL)AMINO}-5'-DEOXY-8-ETHENYLADENOSINE",
            "non-polymer",
        ),
        "DWN": (
            "(2R,4S,6R)-4-AMINO-6-METHYL-5-OXOTETRAHYDRO-2H-PYRAN-2-YL [(2R,3S,5R)-3-HYDROXY-5-(5-METHYL-2,4-DIOXO-3,4-DIHYDROPYRIMIDIN-1(2H)-YL)TETRAHYDROFURAN-2-YL]METHYL DIHYDROGEN DIPHOSPHATE",
            "non-polymer",
        ),
        "5YL": (
            "5-(6-D-RIBITYLAMINO-2,4(1H,3H)PYRIMIDINEDIONE-5-YL) PENTYL-1-PHOSPHONIC ACID",
            "non-polymer",
        ),
        "MS9": ("METHYL 2-AMINO-4-CHLOROBENZOATE", "non-polymer"),
        "MS8": (
            "(2R,3AR,7AR)-2-[(2S)-2-AMINO-3-HYDROXY-3-OXO-PROPYL]-3,3A,5,6,7,7A-HEXAHYDROFURO[4,5-B]PYRAN-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "MS7": ("O-BENZYL-N-METHYL-L-TYROSINAMIDE", "non-polymer"),
        "MS5": (
            "7-METHOXY-2-(3-METHOXYBENZYL)-1,2,3,4-TETRAHYDROISOQUINOLIN-6-YL SULFAMATE",
            "non-polymer",
        ),
        "MS4": ("unknown", "non-polymer"),
        "MS3": (
            "N,N-[2,5-O-DIBENZYL-GLUCARYL]-DI-[VALINYL-AMINOMETHANYL-PYRIDINE]",
            "non-polymer",
        ),
        "MS2": (
            "2,2-DICHLORO-1-METHANESULFINYL-3-METHYL-CYCLOPROPANECARBOXYLIC ACID [1-(4-BROMO-PHENYL)-ETHYL]-AMIDE",
            "non-polymer",
        ),
        "MS1": ("BETA-DADF, MSA, MULTISUBSTRATE ADDUCT INHIBITOR", "non-polymer"),
        "MS0": ("(6-METHOXY-1-BENZOFURAN-3-YL)ACETIC ACID", "non-polymer"),
        "KRI": (
            "(S)-2-AMINO-4-[(2S,3R)-2,3,5-TRIHYDROXY-4-OXO-PENTYL]MERCAPTO-BUTYRIC ACID",
            "non-polymer",
        ),
        "OX5": ("4-HYDROXYBENZALDEHYDE O-(3,3-DIMETHYLBUTANOYL)OXIME", "non-polymer"),
        "MMQ": ("MERCAPTOMETHYL PHOSPHONATE", "non-polymer"),
        "T19": (
            "PHENYLMETHYLENECARBOXY-(METHYLENEAMINO-FORMYL-DIPHENYLMETHYL)METHY-PRO-BOROVAL",
            "peptide",
        ),
        "MA1": ("1,4-DITHIO-ALPHA-D-GLUCOPYRANOSE", "saccharide"),
        "HAZ": (
            "4-{4-[2-(1A,7A-DIMETHYL-4-OXY-OCTAHYDRO-1-OXA-4-AZA-CYCLOPROPA[A]NAPHTHALEN-4-YL) -ACETYLAMINO]-PHENYLCARBAMOYL}-BUTYRIC ACID",
            "non-polymer",
        ),
        "KRL": (
            "5-CYANO-N-[4-(4-METHYLPIPERAZIN-1-YL)-2-PIPERIDIN-1-YLPHENYL]FURAN-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "B2Y": ("1-BIPHENYL-2-YLMETHANAMINE", "non-polymer"),
        "B2X": (
            "4-(5-CHLORO-1H,3H-BENZO[DE]ISOCHROMEN-6-YL)-6-(METHYLSULFANYL)-1,3,5-TRIAZIN-2-AMINE",
            "non-polymer",
        ),
        "B2V": ("VALINE BORONIC ACID", "peptide"),
        "B2T": (
            "4-(2-CHLOROPHENYL)-6-(METHYLSULFANYL)-1,3,5-TRIAZIN-2-AMINE",
            "non-polymer",
        ),
        "B2S": (
            "(3ALPHA,7ALPHA)-3,7,15-TRIHYDROXY-12,13-EPOXYTRICHOTHEC-9-EN-8-ONE",
            "non-polymer",
        ),
        "MA3": ("O1-METHYL-4-DEOXY-4-THIO-ALPHA-D-GLUCOSE", "saccharide"),
        "B2Q": (
            "(2S,3R,11BR)-3-BUTYL-9,10-DIMETHOXY-1,3,4,6,7,11B-HEXAHYDRO-2H-PYRIDO[2,1-A]ISOQUINOLIN-2-AMINE",
            "non-polymer",
        ),
        "B2M": (
            "6-BROMO-N'-[(1Z)-(3,5-DICHLORO-2-HYDROXYPHENYL)METHYLIDENE]-2-METHYLQUINOLINE-4-CARBOHYDRAZIDE",
            "non-polymer",
        ),
        "B2L": (
            "4-(1H,3H-BENZO[DE]ISOCHROMEN-6-YL)-6-METHYLPYRIMIDIN-2-AMINE",
            "non-polymer",
        ),
        "B2K": ("4-METHYL-6-(TOLUENE-4-SULFONYL)-PYRIMIDIN-2-YLAMINE", "non-polymer"),
        "B2J": ("4-(ETHYLSULFANYL)-6-METHYL-1,3,5-TRIAZIN-2-AMINE", "non-polymer"),
        "B2I": ("ISOLEUCINE BORONIC ACID", "peptide"),
        "MD2": ("N,N'-BIS(2,3-BUTADIENYL)-1,4-BUTANE-DIAMINE", "non-polymer"),
        "B2G": ("GALACTOBIOSE", "saccharide"),
        "B2F": ("PHENYLALANINE BORONIC ACID", "peptide"),
        "B2D": (
            "7-METHOXY-N-[(6-PHENYL[1,2,4]TRIAZOLO[4,3-B]PYRIDAZIN-3-YL)METHYL]-1,5-NAPHTHYRIDIN-4-AMINE",
            "non-polymer",
        ),
        "Z67": (
            "4-AMINO-N-(2,6-DIFLUOROPHENYL)-2-[(4-SULFAMOYLPHENYL)AMINO]-1,3-THIAZOLE-5-CARBOXAMIDE",
            "non-polymer",
        ),
        "B2B": (
            "(2Z)-2-[(1R)-3-{[(1R,2S,3R,6S,7S,10S,12S,15E,17E)-18-CARBOXY-16-ETHYL-3,7-DIHYDROXY-1,2,6,10,12-PENTAMETHYL-5-OXOOCTADECA-15,17-DIEN-1-YL]OXY}-1-HYDROXY-3-OXOPROPYL]-3-METHYLBUT-2-ENEDIOIC ACID",
            "non-polymer",
        ),
        "B2A": ("ALANINE BORONIC ACID", "peptide"),
        "Z64": (
            "N-[4-(5-FURAN-2-YL-3-METHYL-1H-PYRAZOL-4-YL)BUTYL]-N-METHYL-7H-PURIN-6-AMINE",
            "non-polymer",
        ),
        "NJ8": ("3-CHLORO-4-(2-HYDROXY-4-PROPYLPHENOXY)BENZONITRILE", "non-polymer"),
        "KYN": ("(2S)-2-AMINO-4-(2-AMINOPHENYL)-4-OXOBUTANOIC ACID", "peptide"),
        "BY6": ("BLEOMYCIN A6", "non-polymer"),
        "KYA": ("4-HYDROXYQUINOLINE-2-CARBOXYLIC ACID", "non-polymer"),
        "MA6": ("6N-DIMETHYLADENOSINE-5'-MONOPHOSHATE", "nucleic-acid"),
        "3S1": ("6-[2-(HYDROXYMETHYL)PHENYL]ISOQUINOLIN-1(2H)-ONE", "non-polymer"),
        "ZDM": (
            "NONYL 4-O-ALPHA-D-GLUCOPYRANOSYL-BETA-D-GLUCOPYRANOSIDE",
            "saccharide",
        ),
        "3S2": (
            "[(2S,3R)-2-FORMYL-1-{[4-(METHYLAMINO)BUTYL]CARBAMOYL}PYRROLIDIN-3-YL]SULFAMIC ACID",
            "non-polymer",
        ),
        "MA9": (
            "(3R,4R)-1-{6-[3-(METHYLSULFONYL)PHENYL]PYRIMIDIN-4-YL}-4-(2,4,5-TRIFLUOROPHENYL)PIPERIDIN-3-AMINE",
            "non-polymer",
        ),
        "MA8": (
            "(2S,3R,5S,6S)-2,3,4,5,6-PENTAHYDROXYCYCLOHEXYL 2-(L-CYSTEINYLAMINO)-2-DEOXY-ALPHA-L-GLUCOPYRANOSIDE",
            "non-polymer",
        ),
        "KYS": (
            "N-(4-{[(2-AMINOETHYL)AMINO]SULFONYL}PHENYL)-5-[(3AS,4S,6AR)-2-OXOHEXAHYDRO-1H-THIENO[3,4-D]IMIDAZOL-4-YL]PENTANAMIDE-(1,2,3,4,5,6-ETA)-BENZENE-CHLORO-RUTHENIUM(III)",
            "non-polymer",
        ),
        "KYQ": ("(E)-N~6~-(1-CARBOXY-2-HYDROXYETHYLIDENE)-L-LYSINE", "peptide"),
        "NYP": (
            "N-[(E)-METHYL](PHENYL)-N-[(E)-2-PROPENYLIDENE]METHANAMINIUM",
            "non-polymer",
        ),
        "KYT": (
            "(3AS,4S,6AR)-4-(5-((3R,4R)-3,4-DIAMINOPYRROLIDIN-1-YL)-5-OXOPENTYL)TETRAHYDRO-1H-THIENO[3,4-D]IMIDAZOL-2(3H)-ONE-P-CYMENE-CHLORO-RUTHENIUM(III)",
            "non-polymer",
        ),
        "RIL": (
            "[(4R)-2,2-DIMETHYL-1,3-DIOXOLAN-4-YL]METHYL HYDROGEN HEX-5-ENYLPHOSPHONATE",
            "non-polymer",
        ),
        "3SZ": (
            "3-{[4-(2,3-DIHYDRO-1,4-BENZODIOXIN-6-YLSULFONYL)-1,4-DIAZEPAN-1-YL]SULFONYL}ANILINE",
            "non-polymer",
        ),
        "AGC": ("ALPHA-D-GLUCOSE", "saccharide"),
        "3SP": (
            "N-(CARBOXYMETHYL)-3-CYCLOHEXYL-D-ALANYL-N-({5-[(E)-AMINO(IMINO)METHYL]THIEN-2-YL}METHYL)-L-PROLINAMIDE",
            "peptide",
        ),
        "BYP": (
            "(2R)-1-{4-[(4-ANILINO-5-BROMOPYRIMIDIN-2-YL)AMINO]PHENOXY}-3-(DIMETHYLAMINO)PROPAN-2-OL",
            "non-polymer",
        ),
        "T8B": ("THERMORUBIN", "non-polymer"),
        "3SL": ("(2R)-3-SULFOLACTIC ACID", "non-polymer"),
        "BYM": (
            "(5E)-5-[(2,2-DIFLUORO-1,3-BENZODIOXOL-5-YL)METHYLENE]-1,3-THIAZOLIDINE-2,4-DIONE",
            "non-polymer",
        ),
        "NYT": (
            "BETA-D-FRUCTOFURANOSYL-(2->1)-BETA-D-FRUCTOFURANOSYL-(2->1)-BETA-D-FRUCTOFURANOSYL ALPHA-D-GLUCOPYRANOSIDE",
            "saccharide",
        ),
        "BYH": (
            "1-[2-(1H-BENZIMIDAZOL-2-YLSULFANYL)ETHYL]-3-METHYL-1,3-DIHYDRO-2H-BENZIMIDAZOLE-2-THIONE",
            "non-polymer",
        ),
        "3SE": (
            "4,4,4-TRIFLUORO-1-{4-[3-(1,3-THIAZOL-2-YL)-1,2,4-OXADIAZOL-5-YL]PIPERIDIN-1-YL}BUTAN-1-ONE",
            "non-polymer",
        ),
        "3SD": (
            "N~4~-(2,2-DIMETHYLPROPYL)-N~1~-{(2S)-1-[(4-METHYLBENZYL)AMINO]-1-OXO-4-PHENYLBUTAN-2-YL}-N~2~-[(5-METHYL-1,2-OXAZOL-3-YL)CARBONYL]-L-ASPARTAMIDE",
            "peptide",
        ),
        "KY1": ("4-(2-AMINOPHENYL)-4-OXOBUTANOIC ACID", "non-polymer"),
        "3SF": (
            "5,5,5-TRIFLUORO-1-{4-[3-(1,3-THIAZOL-2-YL)-1,2,4-OXADIAZOL-5-YL]PIPERIDIN-1-YL}PENTAN-1-ONE",
            "non-polymer",
        ),
        "3SA": ("ACARBOSE DERIVED TRISACCHARIDE", "saccharide"),
        "BYC": ("BENZOYL COENZYME A", "non-polymer"),
        "3SC": (
            "5-(2,3-DICHLOROPHENYL)-N-(PYRIDIN-4-YLMETHYL)-3-THIOCYANATOPYRAZOLO[1,5-A]PYRIMIDIN-7-AMINE",
            "non-polymer",
        ),
        "3SB": (
            "(2S)-2-AMINO-N'-[(1E)-(2,4-DIHYDROXY-6-METHYLPHENYL)METHYLIDENE]-2-PHENYLETHANEHYDRAZIDE",
            "non-polymer",
        ),
        "TM6": ("THIO-MALTOHEXAOSE", "saccharide"),
        "B29": (
            "[2-(3-DIBENZOFURAN-4-YL-PHENYL)-1-HYDROXY-1-PHOSPHONO-ETHYL]-PHOSPHONIC ACID",
            "non-polymer",
        ),
        "B28": ("unknown", "non-polymer"),
        "B27": ("(2R,3S) 3-AMINO-4-MERCAPTO-2-BUTANOL", "peptide"),
        "B26": (
            "N-(4-BROMO-3-METHYLBENZOYL)-4-[DIFLUORO(PHOSPHONO)METHYL]-L-PHENYLALANYL-N~5~-(3-IODOBENZOYL)-L-ORNITHYL-3-{[(4-HYDROXY-3-METHOXYPHENYL)ACETYL]AMINO}-D-ALANINAMIDE",
            "peptide",
        ),
        "B25": ("5-NITRO-1H-INDOLE-2-CARBOXIMIDAMIDE", "non-polymer"),
        "B24": ("1-ETHYL-1,2-DIHYDRO-1,2-AZABORININE", "non-polymer"),
        "B23": ("5-(2-NITROPHENYL)-2-FUROIC ACID", "non-polymer"),
        "B22": (
            "[2,2'-{[2-({3-[({2-[4-(AMINOSULFONYL)PHENYL]ETHYL}AMINO)CARBONYL]PHENYL}AMINO)-2-OXOETHYL]IMINO}DIACETATO(2-)-KAPPAO]COPPER",
            "non-polymer",
        ),
        "B21": ("5-(2-METHOXYPHENYL)-2-FUROIC ACID", "non-polymer"),
        "B20": ("1,2-DIHYDRO-1,2-AZABORININE", "non-polymer"),
        "VYF": (
            "(3S,5R)-5-[4-(2-CHLOROPHENYL)-2,2-DIMETHYL-5-OXOPIPERAZIN-1-YL]-N-(2,6-DIMETHYLHEPTAN-4-YL)PIPERIDINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "MM2": (
            "1,1'-[1,4-PHENYLENEBIS(METHYLENE)]BIS[1,4,8,11-TETRAAZA-CYCLOTETRADECANE]CU(II)2",
            "non-polymer",
        ),
        "PEU": (
            "2,5,8,11,14,17,20,23,26,29,32,35,38,41,44,47,50,53,56,59,62,65,68,71,74,77,80-HEPTACOSAOXADOOCTACONTAN-82-OL",
            "non-polymer",
        ),
        "X46": (
            "[4-AMINO-2-(PHENYLAMINO)-1,3-THIAZOL-5-YL](3-METHOXYPHENYL)METHANONE",
            "non-polymer",
        ),
        "X45": (
            "(5S)-2-AMINO-5-(4-METHOXY-3-METHYLPHENYL)-3-METHYL-5-(3-PYRIDIN-3-YLPHENYL)-3,5-DIHYDRO-4H-IMIDAZOL-4-ONE",
            "non-polymer",
        ),
        "X44": (
            "4-{[4-AMINO-5-(4-SULFAMOYLBENZOYL)-1,3-THIAZOL-2-YL]AMINO}BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "X43": (
            "4-{[4-AMINO-5-(3-NITROBENZOYL)-1,3-THIAZOL-2-YL]AMINO}BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "NYL": ("N-ALLYL-ANILINE", "non-polymer"),
        "X41": ("1-MYRISTOYL-2-MYRISTOLEOYL-SN-GLYCERO-3-PHOSPHATE", "non-polymer"),
        "X40": (
            "4-{[4-AMINO-5-(PYRIDIN-3-YLCARBONYL)-1,3-THIAZOL-2-YL]AMINO}BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "NYM": (
            "PHOSPHORIC ACID MONO-[3-AMINO-5-(5-METHYL-2,4-DIOXO-3,4-DIHYDRO-2H-PYRIMIDIN-1-YL)-TETRAHYDRO-FURAN-2-YLMETHYL] ESTER",
            "nucleic-acid",
        ),
        "YBT": (
            "BIS-(2-HYDROXYETHYL)AMINO-TRIS(HYDROXYMETHYL)METHANE YTTRIUM",
            "non-polymer",
        ),
        "MML": ("1-METHYL-4-(1-METHYLETHYL)BENZENE", "non-polymer"),
        "YBH": (
            "4-OXO-3-{6-[4-(QUINOXALIN-2-YLOXY)-BENZOYLAMINO]-2-THIOPHEN-2-YL-HEXANOYLAMINO}-BUTYRIC ACID",
            "non-polymer",
        ),
        "YBN": (
            "3-METHOXYBENZYL 3-(PIPERIDIN-4-YLOXY)-1-BENZOTHIOPHENE-2-CARBOXYLATE",
            "non-polymer",
        ),
        "VNR": (
            "1-DEOXY-1-[(4AR)-4A-[(2Z)-2-(2-{[(S)-HYDROXY(PHOSPHONOOXY)PHOSPHORYL]OXY}ETHYL)BUT-2-EN-1-YL]-7,8-DIMETHYL-2,4-DIOXO-3,4,4A,5-TETRAHYDROBENZO[G]PTERIDIN-10(2H)-YL]-5-O-PHOSPHONO-D-RIBITOL",
            "non-polymer",
        ),
        "10J": (
            "3-{5-[(2E,4AR,7AR)-6-BENZOYL-2-IMINO-3-METHYL-4-OXOOCTAHYDRO-7AH-PYRROLO[3,4-D]PYRIMIDIN-7A-YL]THIOPHEN-3-YL}BENZONITRILE",
            "non-polymer",
        ),
        "10H": (
            "1-(4-{[(2R)-2-METHYLPIPERIDIN-1-YL]SULFONYL}PHENYL)-1,3-DIHYDRO-2H-PYRROL-2-ONE",
            "non-polymer",
        ),
        "AGT": ("S-{(S)-AMINO[(4-AMINOBUTYL)AMINO]METHYL}-L-CYSTEINE", "peptide"),
        "10O": (
            "3-{5-[(2E,4AR,7AR)-2-IMINO-6-(6-METHOXYPYRIDIN-2-YL)-3-METHYL-4-OXOOCTAHYDRO-7AH-PYRROLO[3,4-D]PYRIMIDIN-7A-YL]THIOPHEN-3-YL}BENZONITRILE",
            "non-polymer",
        ),
        "10L": ("1H-INDAZOL-7-AMINE", "non-polymer"),
        "10M": (
            "DECYL 4-O-ALPHA-D-GLUCOPYRANOSYL-1-THIO-BETA-D-GLUCOPYRANOSIDE",
            "saccharide",
        ),
        "10C": (
            "4-AMINO-1-{2,5-ANHYDRO-4-[(PHOSPHONOOXY)METHYL]-ALPHA-L-LYXOFURANOSYL}PYRIMIDIN-2(1H)-ONE",
            "nucleic-acid",
        ),
        "10A": ("DIDECYL-DIMETHYL-AMMONIUM", "non-polymer"),
        "10F": (
            "N-({4-[{[(2R,4S,4AR,6S,8AS)-2-AMINO-4-HYDROXYDECAHYDROPTERIDIN-6-YL]METHYL}(FORMYL)AMINO]PHENYL}CARBONYL)-D-GLUTAMIC ACID",
            "non-polymer",
        ),
        "10G": (
            "(2E)-3-METHYL-4-SULFANYLBUT-2-EN-1-YL TRIHYDROGEN DIPHOSPHATE",
            "non-polymer",
        ),
        "10D": (
            "(2E)-4-FLUORO-3-METHYLBUT-2-EN-1-YL TRIHYDROGEN DIPHOSPHATE",
            "non-polymer",
        ),
        "10E": (
            "(2E)-4-AMINO-3-METHYLBUT-2-EN-1-YL TRIHYDROGEN DIPHOSPHATE",
            "non-polymer",
        ),
        "10Z": (
            "N-(4-{[(2-METHOXYETHYL)AMINO]METHYL}PHENYL)-6-(PYRIDIN-4-YL)QUINAZOLIN-2-AMINE",
            "non-polymer",
        ),
        "0Q1": (
            "(1R)-6-(3,5-DIMETHYL-1,2-OXAZOL-4-YL)-1-PHENYL-2,3-DIHYDRO-1H-INDEN-1-OL",
            "non-polymer",
        ),
        "0Q2": (
            "(1R,3R)-3-(2-METHYLIMIDAZO[4,5-D]PYRROLO[2,3-B]PYRIDIN-1(8H)-YL)CYCLOHEXANOL",
            "non-polymer",
        ),
        "10Y": (
            "(9Z)-11-{(2R,3S)-3-[(2Z)-PENT-2-EN-1-YL]OXIRAN-2-YL}UNDEC-9-ENOIC ACID",
            "non-polymer",
        ),
        "0Q4": (
            "N-[(2R)-2-({N~5~-[AMINO(IMINIO)METHYL]-L-ORNITHYL-L-VALYL}AMINO)-4-METHYLPENTYL]-L-PHENYLALANYL-L-ALPHA-GLUTAMYL-L-ALANYL-L-NORLEUCINAMIDE",
            "peptide",
        ),
        "0Q5": (
            "4-[5-BUTYL-3-(HYDROXYMETHYL)-1-PHENYL-1H-PYRAZOL-4-YL]-3-(3,4-DIHYDROISOQUINOLIN-2(1H)-YLCARBONYL)-N-{[2-(TRIMETHYLSILYL)ETHYL]SULFONYL}BENZAMIDE",
            "non-polymer",
        ),
        "10R": (
            "1-(2-[(R)-2,4-DIHYDROXYBUTOXY]ETHYL)-12-(5-ETHYL-5-HYDROXYHEPTYL)-1,12-DICARBA-CLOSO-DODECABORANE",
            "non-polymer",
        ),
        "10S": (
            "1-(2-[(S)-2,4-DIHYDROXYBUTOXY]ETHYL)-12-(5-ETHYL-5-HYDROXYHEPTYL)-1,12-DICARBA-CLOSO-DODECABORANE",
            "non-polymer",
        ),
        "10P": (
            "D-PHENYLALANYL-N-[(1-METHYLPYRIDINIUM-3-YL)METHYL]-L-PROLINAMIDE",
            "peptide",
        ),
        "10Q": (
            "2-{(2E,4AR,7AR)-7A-[4-(3-CYANOPHENYL)THIOPHEN-2-YL]-2-IMINO-3-METHYL-4-OXOOCTAHYDRO-6H-PYRROLO[3,4-D]PYRIMIDIN-6-YL}PYRIDINE-3-CARBONITRILE",
            "non-polymer",
        ),
        "10V": (
            "3-{5-[(2E,4AR,7AR)-2-IMINO-6-(3-METHOXYPYRIDIN-2-YL)-3-METHYL-4-OXOOCTAHYDRO-7AH-PYRROLO[3,4-D]PYRIMIDIN-7A-YL]THIOPHEN-3-YL}BENZONITRILE",
            "non-polymer",
        ),
        "10W": (
            "2-FLUORO-5-{5-[(2E,4AR,7AR)-2-IMINO-3-METHYL-4-OXO-6-PHENYLOCTAHYDRO-7AH-PYRROLO[3,4-D]PYRIMIDIN-7A-YL]THIOPHEN-2-YL}BENZONITRILE",
            "non-polymer",
        ),
        "10T": (
            "5-(AMINOMETHYL)-2-METHYL-4-(4-METHYLPHENYL)-6-(2-METHYLPROPYL)PYRIDINE-3-CARBOXIC ACID",
            "non-polymer",
        ),
        "10U": (
            "(S)-N-(4-CARBAMIMIDOYLBENZYL)-1-(2-(CYCLOPENTYLAMINO)ETHANOYL)PYRROLIDINE-2-CARBOXAMIDE",
            "peptide",
        ),
        "0QA": ("4-[METHYL(NITROSO)AMINO]-1-(PYRIDIN-3-YL)BUTAN-1-ONE", "non-polymer"),
        "0QB": ("P2-P3 BUTANEDIAMIDE RENIN INHIBITOR (3)", "non-polymer"),
        "0QE": ("CHLOROMETHANE", "non-polymer"),
        "0QF": (
            "N-[(1R)-2-{[(1R,2R,3S)-1-(CYCLOHEXYLMETHYL)-2,3-DIHYDROXY-5-METHYLHEXYL]AMINO}-1-(ETHYLSULFANYL)-2-OXOETHYL]-NALPHA-(MORPHOLIN-4-YLSULFONYL)-D-PHENYLALANINAMIDE",
            "peptide",
        ),
        "0QG": ("3-METHYL-L-VALYL-L-PROLYL-L-ISOLEUCINE", "peptide"),
        "0QH": (
            "(1R)-1-BENZYL-1-METHYL-1-(2-{[4-(1-METHYLETHYL)PHENYL]AMINO}-2-OXOETHYL)-2-{(2S)-4-METHYL-2-[(TRIFLUOROACETYL)AMINO]PENTANOYL}DIAZANIUM",
            "peptide",
        ),
        "0QI": (
            "O-METHYL-N-[(2S)-4-METHYL-2-(SULFANYLMETHYL)PENTANOYL]-L-TYROSINE",
            "peptide",
        ),
        "0QJ": (
            "5'-[(3-{[(4-TERT-BUTYLPHENYL)CARBAMOYL]AMINO}PROPYL)(METHYL)AMINO]-5'-DEOXYADENOSINE",
            "non-polymer",
        ),
        "0QK": (
            "7-{5-[(3-{[(4-TERT-BUTYLPHENYL)CARBAMOYL]AMINO}PROPYL)(PROPAN-2-YL)AMINO]-5-DEOXY-BETA-D-RIBOFURANOSYL}-7H-PYRROLO[2,3-D]PYRIMIDIN-4-AMINE",
            "non-polymer",
        ),
        "0QL": ("3-[(2-AMINOETHYL)DISULFANYL]-L-ALANINE", "peptide"),
        "0QN": (
            "N-(TRIFLUOROACETYL)-L-VALYL-N-[4-(TRIFLUOROMETHYL)PHENYL]-L-ALANINAMIDE",
            "peptide",
        ),
        "OXQ": ("4-HYDROXY-1,2,5-OXADIAZOLE-3-CARBOXYLIC ACID", "non-polymer"),
        "0QR": ("N-(6-AMINOPYRIDIN-2-YL)-4-FLUOROBENZENESULFONAMIDE", "non-polymer"),
        "0QS": (
            "N~2~-[(2R)-2-BENZYL-3-(TERT-BUTYLSULFONYL)PROPANOYL]-N-{(1R)-1-(CYCLOHEXYLMETHYL)-3,3-DIFLUORO-2,2-DIHYDROXY-4-[(2-MORPHOLIN-4-YLETHYL)AMINO]-4-OXOBUTYL}-3-(1H-IMIDAZOL-3-IUM-4-YL)-L-ALANINAMIDE",
            "peptide",
        ),
        "0QV": ("(4-HYDROXYPIPERIDIN-1-YL)(1H-INDOL-3-YL)METHANETHIONE", "non-polymer"),
        "0QW": ("(2-HYDROXYPHENYL)(PYRROLIDIN-1-YL)METHANETHIONE", "non-polymer"),
        "0QX": ("2-(1H-INDOL-3-YLMETHYL)-1H-IMIDAZO[4,5-C]PYRIDINE", "non-polymer"),
        "0QY": (
            "N-[2-(1H-INDOL-3-YLMETHYL)-1H-BENZIMIDAZOL-5-YL]-L-PROLINAMIDE",
            "non-polymer",
        ),
        "100": (
            "1-(5-CHLOROINDOL-3-YL)-3-HYDROXY-3-(2H-TETRAZOL-5-YL)-PROPENONE",
            "non-polymer",
        ),
        "101": ("?", "non-polymer"),
        "106": ("4-(5-BROMO-2-OXO-2H-INDOL-3-YLAZO)-BENZENESULFONAMIDE", "non-polymer"),
        "107": (
            "4-[(7-OXO-7H-THIAZOLO[5,4-E]INDOL-8-YLMETHYL)-AMINO]-N-PYRIDIN-2-YL-BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "104": ("N,N'-BIS(2-AMINOETHYL)-1,2-ETHANEDIAMINE", "non-polymer"),
        "105": (
            "N-[5-METHYL-3-O-TOLYL-ISOXAZOLE-4-CARBOXYLIC ACID AMIDE] BORONIC ACID",
            "non-polymer",
        ),
        "HO6": (
            "5-[(4-METHYLPHENYL)SULFANYL]-1,2,3-THIADIAZOLE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "X4S": (
            "BETA-D-XYLOPYRANOSYL-(1->4)-4-THIO-BETA-D-XYLOPYRANOSYL-(1->4)-4-THIO-BETA-D-XYLOPYRANOSYL-(1->4)-4-THIO-BETA-D-XYLOPYRANOSE",
            "saccharide",
        ),
        "YB2": ("YTTERBIUM (II) ION", "non-polymer"),
        "RT7": (
            "S-{2-[(2-CHLORO-4-SULFAMOYLPHENYL)AMINO]-2-OXOETHYL} 6,8-DICHLORO-3,4-DIHYDROQUINOLINE-1(2H)-CARBOTHIOATE",
            "non-polymer",
        ),
        "XMH": (
            "4-[(6-CHLORO-2-NAPHTHALENYL)SULFONYL]-1-[[1-(4-PYRIDINYL)-4-PIPERIDINYL]METHYL]-2-PIPERAZINECARBOXYLIC ACID ETHYL ESTER",
            "non-polymer",
        ),
        "X4B": (
            "4-CHLORO-5-NITRO-2-[(PYRIDIN-3-YLMETHYL)AMINO]BENZAMIDE",
            "non-polymer",
        ),
        "X4A": (
            "[(2R,3S,5S)-2,3,5-TRIHYDROXYTETRAHYDROFURAN-2-YL]METHYL DIHYDROGEN PHOSPHATE",
            "nucleic-acid",
        ),
        "RIF": ("RIFAMYCIN CGP 4832", "non-polymer"),
        "RT1": (
            "(2S)-3-[(S)-[(1R)-1-AMINO-2-PHENYLETHYL](HYDROXY)PHOSPHORYL]-2-METHYLPROPANOIC ACID",
            "peptide",
        ),
        "QED": (
            "N,N'-{[(2-{[(2,3-DIHYDROXYPHENYL)CARBONYL]AMINO}ETHYL)IMINO]DIETHANE-2,1-DIYL}BIS(1-HYDROXY-6-OXO-1,6-DIHYDROPYRIDINE-2-CARBOXAMIDE)",
            "non-polymer",
        ),
        "OXK": ("OXALYL-COENZYME A", "non-polymer"),
        "VBN": ("VIBRIOBACTIN", "non-polymer"),
        "OXD": ("OXALIC ACID", "non-polymer"),
        "PER": ("PEROXIDE ION", "non-polymer"),
        "VBP": (
            "4-({3-[(2R)-2-AMINO-2-CARBOXYETHYL]-2,6-DIOXO-3,6-DIHYDROPYRIMIDIN-1(2H)-YL}METHYL)BENZOIC ACID",
            "non-polymer",
        ),
        "VBZ": (
            "(1R,4R,5R,7R,8R)-2-BENZYL-5-HYDROXYMETHYL-2-AZA-BICYCLO[2.2.2]OCTANE-4,7,8-TRIOL",
            "non-polymer",
        ),
        "8GT": ("8-OXO-GUANOSINE-5'-TRIPHOSPHATE", "non-polymer"),
        "RCY": (
            "(3'R)-1'-OXYL-2',2',5',5'-TETRAMETHYL-1,3'-BIPYRROLIDINE-2,5-DIONE",
            "non-polymer",
        ),
        "TLS": ("TELMISARTAN", "non-polymer"),
        "WHF": (
            "3-(5-AMINO-1,3,4-THIADIAZOL-2-YL)-1-THIOPHEN-2-YLPROPAN-1-ONE",
            "non-polymer",
        ),
        "WHJ": ("CARMOTEROL", "non-polymer"),
        "RT8": (
            "NALPHA-[(TRANS-4-METHYLCYCLOHEXYL)CARBONYL]-N-PYRIDIN-4-YL-D-TRYPTOPHANAMIDE",
            "non-polymer",
        ),
        "WHU": (
            "2-AZANYL-N-[[4-BROMANYL-3-(3-CHLORANYL-5-CYANO-PHENOXY)-2-FLUORANYL-PHENYL]METHYL]-4-CHLORANYL-1H-IMIDAZOLE-5-CARBOXAMIDE",
            "non-polymer",
        ),
        "2HG": ("(2R)-2-HYDROXYPENTANEDIOIC ACID", "non-polymer"),
        "PPD": (
            "2-[(3-HYDROXY-2-METHYL-5-PHOSPHONOOXYMETHYL-PYRIDIN-4-YLMETHYLENE)-AMINO]-SUCCINIC ACID",
            "non-polymer",
        ),
        "PEL": ("2-PHENYL-ETHANOL", "non-polymer"),
        "WH6": (
            "2-[(1R)-2-CARBOXY-1-(NAPHTHALEN-1-YLMETHYL)ETHYL]-1,3-DIOXO-2,3-DIHYDRO-1H-ISOINDOLE-5-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "V": ("VANADIUM ION", "non-polymer"),
        "4JZ": (
            "1-[5-TERT-BUTYL-2-(4-METHYLPHENYL)-1,2-DIHYDRO-3H-PYRAZOL-3-YLIDENE]-3-{3-[(PYRIDIN-3-YLOXY)METHYL]-1H-PYRAZOL-5-YL}UREA",
            "non-polymer",
        ),
        "NGK": (
            "2-(ACETYLAMINO)-2-DEOXY-4-O-SULFO-ALPHA-D-GALACTOPYRANOSE",
            "non-polymer",
        ),
        "S5P": ("6-SULFANYLURIDINE-5'-PHOSPHATE", "non-polymer"),
        "HAV": ("HYDROXYAMINOVALINE", "non-polymer"),
        "VB1": (
            "N^6^-[(1R)-2-{[(1S)-1-CARBOXYPROPYL]AMINO}-2-OXO-1-(SULFANYLMETHYL)ETHYL]-6-OXO-L-LYSINE",
            "peptide",
        ),
        "TLP": (
            "3-HYDROXY-2-[(3-HYDROXY-2-METHYL-5-PHOSPHONOOXYMETHYL-PYRIDIN-4-YLMETHYL)-AMINO]-BUTYRIC ACID",
            "non-polymer",
        ),
        "QEH": (
            "(7R)-4,7-DIHYDROXY-N,N,N-TRIMETHYL-10-OXO-3,5,9-TRIOXA-4-PHOSPHAHEPTACOSAN-1-AMINIUM 4-OXIDE",
            "non-polymer",
        ),
        "ZG2": (
            "5-[5-BENZYL-4-(2-OXO-2-PYRROLIDIN-1-YLETHYL)-1,3-OXAZOL-2-YL]-1-ETHYL-N-(TETRAHYDRO-2H-PYRAN-4-YL)-1H-PYRAZOLO[3,4-B]PYRIDIN-4-AMINE",
            "non-polymer",
        ),
        "IKT": ("3-(1-AMINOETHYL)NONANEDIOIC ACID", "non-polymer"),
        "68H": ("SALBUTAMOL", "non-polymer"),
        "UNK": ("UNKNOWN", "peptide"),
        "X6K": (
            "3-(4-MORPHOLIN-4-YLPYRIDO[3',2':4,5]FURO[3,2-D]PYRIMIDIN-2-YL)PHENOL",
            "non-polymer",
        ),
        "GSE": ("L-ALPHA-GLYCEROPHOSPHORYLSERINE", "non-polymer"),
        "NGU": ("U-58872, HYDROXY DERIVATIVE OF NOGALAMYCIN", "non-polymer"),
        "VYD": (
            "(3S,5R)-5-{[4-(2-CHLOROPHENYL)-2,2-DIMETHYL-5-OXOPIPERAZIN-1-YL]METHYL}-N-(3-METHYLBUTYL)PIPERIDINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "N41": ("4-(6-CYCLOHEXYLMETHOXY-9H-PURIN-2-YLAMINO)--BENZAMIDE", "non-polymer"),
        "N44": (
            "5-({[2-(BENZYLCARBAMOYL)BENZYL](PROP-2-EN-1-YL)AMINO}METHYL)-1,3-BENZODIOXOLE-4-CARBOXYLATE",
            "non-polymer",
        ),
        "P26": (
            "4-[2-(1,1,3,3-TETRAMETHYL-2,3-DIHYDRO-1H-INDEN-5-YL)-1,3-DIOXOLAN-2-YL]BENZOIC ACID",
            "non-polymer",
        ),
        "2L8": ("?", "nucleic-acid"),
        "NGV": (
            "METHYL 5,7-DIHYDROXY-2-METHYL-4,6,11-TRIOXO-3,4,6,11-TETRAHYDROTETRACENE-1-CARBOXYLATE",
            "non-polymer",
        ),
        "8C5": (
            "8-CYCLOHEXYL-N-METHOXY-5-OXO-2-{[4-(2-PYRROLIDIN-1-YLETHYL)PHENYL]AMINO}-5,8-DIHYDROPYRIDO[2,3-D]PYRIMIDINE-6-CARBOXAMIDE",
            "non-polymer",
        ),
        "P21": (
            "4-CHLORO-N-(3-METHOXYPROPYL)-N-[(3S)-1-(2-PHENYLETHYL)PIPERIDIN-3-YL]BENZAMIDE",
            "non-polymer",
        ),
        "RCC": ("RED CHLOROPHYLL CATABOLITE", "non-polymer"),
        "SN7": (
            "4-[4-[2-AMINO-4-[4,6-(N-METHYLQUINOLINIUM)AMINO]BENZAMIDO]ANILINO]-N-METHYLPYRIDINIUM MESYLATE",
            "non-polymer",
        ),
        "P20": (
            "2-(5-{[AMINO(IMINO)METHYL]AMINO}-2-CHLOROPHENYL)-3-SULFANYLPROPANOIC ACID",
            "non-polymer",
        ),
        "2LM": (
            "(2E)-2-[({3-HYDROXY-2-METHYL-5-[(PHOSPHONOOXY)METHYL]PYRIDIN-4-YL}METHYL)IMINO]-4-(METHYLSULFANYL)BUTANOIC ACID",
            "non-polymer",
        ),
        "FAZ": (
            "8-(ACETYLAMINO)-N-(2-CHLOROPHENYL)-N-METHYL-4,5-DIHYDROTHIENO[3,2-D][1]BENZOXEPINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "8CA": (
            "9-BENZYL-2,3,4,9-TETRAHYDRO-1H-CARBAZOLE-8-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "2LI": (
            "N-[(1S,2R)-3-{[(5S)-5-(3-TERT-BUTYLPHENYL)-4,5,6,7-TETRAHYDRO-1H-INDAZOL-5-YL]AMINO}-1-(3,5-DIFLUOROBENZYL)-2-HYDROXYPROPYL]ACETAMIDE",
            "non-polymer",
        ),
        "P23": ("PROPYL TRIHYDROGEN DIPHOSPHATE", "non-polymer"),
        "PEH": ("DI-STEAROYL-3-SN-PHOSPHATIDYLETHANOLAMINE", "non-polymer"),
        "2LE": ("(2R,5S)-5-PENTYLTETRAHYDROFURAN-2-OL", "non-polymer"),
        "2LG": (
            "2-CHLORO-N-(3-CYANO-5,6-DIHYDRO-4H-CYCLOPENTA[B]THIOPHEN-2-YL)-5-DIETHYLSULFAMOYL-BENZAMIDE",
            "non-polymer",
        ),
        "2LF": (
            "(6S,7S,8S,10R)-2-AMINO-8-HYDROXY-4-OXO-3,6,7,8,9,10-HEXAHYDRO-4H-7,10-EPOXYAZEPINO[1,2-E]PURIN-6-YL DIHYDROGEN PHOSPHATE",
            "nucleic-acid",
        ),
        "2LA": (
            "2-AMINO-5-(AMINOMETHYL)-7-(2-DEOXY-5-O-PHOSPHONO-BETA-D-ERYTHRO-PENTOFURANOSYL)-3,7-DIHYDRO-4H-PYRROLO[2,3-D]PYRIMIDIN-4-ONE",
            "nucleic-acid",
        ),
        "N4T": (
            "(1S)-1-CYCLOPROPYL-2-[(2S)-4-(2,5-DIFLUOROPHENYL)-2-PHENYL-2,5-DIHYDRO-1H-PYRROL-1-YL]-2-OXOETHANAMINE",
            "non-polymer",
        ),
        "8CL": ("CHLOROBENZENE", "non-polymer"),
        "8CM": ("8-HYDROXYCOUMARIN", "non-polymer"),
        "N4I": (
            "4-({(3S,4S)-4-[(3-HYDROXYBENZYL)AMINO]PYRROLIDIN-3-YL}[4-(TRIFLUOROMETHYL)BENZYL]SULFAMOYL)BENZAMIDE",
            "non-polymer",
        ),
        "8CS": ("?", "non-polymer"),
        "N4M": ("TETRAHYDROMETHANOPTERIN", "non-polymer"),
        "2LU": ("2-AMINO-4-METHYL-PENTANYL GROUP", "peptide"),
        "MMA": ("O1-METHYL-MANNOSE", "saccharide"),
        "8CX": (
            "6-{[(3R,4R)-4-(2-{[(2R)-2-FLUORO-2-(3-FLUOROPHENYL)ETHYL]AMINO}ETHOXY)PYRROLIDIN-3-YL]METHYL}-4-METHYLPYRIDIN-2-AMINE",
            "non-polymer",
        ),
        "N4B": ("N-BUTYLBENZENE", "non-polymer"),
        "N4E": ("N-(4-ETHOXYPHENYL)ACETAMIDE", "non-polymer"),
        "2LP": ("2-ALLYLPHENOL", "non-polymer"),
        "ZW3": (
            "N-{6-[(Z)-(2,4-DIOXO-1,3-THIAZOLIDIN-5-YLIDENE)METHYL]IMIDAZO[1,2-A]PYRIDIN-2-YL}ACETAMIDE",
            "non-polymer",
        ),
        "FAR": ("FARNESYL", "non-polymer"),
        "NGP": ("NOGALAMYCIN (PROTONATED FORM)", "non-polymer"),
        "689": (
            "4-[(1S,2S,5S)-5-(HYDROXYMETHYL)-6,8,9-TRIMETHYL-3-OXABICYCLO[3.3.1]NON-7-EN-2-YL]PHENOL",
            "non-polymer",
        ),
        "688": (
            "2-[AMINO(IMINO)METHYL]-2-HYDROXYPHENOXY]-6-[3-(4,5-DIHYDRO-1H-IMIDAZOL-2-YL)PHENOXY]PYRIDINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "FAS": ("ARABINO-FLAVIN-ADENINE DINUCLEOTIDE", "non-polymer"),
        "685": (
            "2-{2-[(3,5-DIMETHYLPHENYL)AMINO]PYRIMIDIN-4-YL}-N-[(1S)-2-HYDROXY-1-METHYLETHYL]-4-METHYL-1,3-THIAZOLE-5-CARBOXAMIDE",
            "non-polymer",
        ),
        "684": (
            "4-({5-[(2,4-DIAMINOPTERIDIN-6-YL)METHYL]-5H-DIBENZO[B,F]AZEPIN-2-YL}OXY)BUTANOIC ACID",
            "non-polymer",
        ),
        "687": (
            "N-ACETYL-N-[1-(1,1'-BIPHENYL-4-YLMETHYL)-2-OXOAZEPAN-3-YL]-3-FORMYL-O-PHOSPHONOTYROSINAMIDE",
            "non-polymer",
        ),
        "681": (
            "2-(5-AMINO-6-OXO-2-PHENYL-6H-PYRIMIDIN-1-YL)-N-[2-(5-TERT-BUTYL-1,3,4-OXADIAZOL-2-YL)-1-(METHYLETHYL)-2-HYDROXYETHYL]ACETAMIDE",
            "non-polymer",
        ),
        "680": ("6-METHYLAMINO-5-NITROISOCYTOSINE", "non-polymer"),
        "682": (
            "N~2~-(AMINOCARBONYL)-N~1~-{4-{[AMINO(IMINO)METHYL]AMINO}-1-[HYDROXY(1,3-THIAZOL-2-YL)METHYL]BUTYL}VALINAMIDE",
            "non-polymer",
        ),
        "XQI": ("METHYL 2-(3,5-DIHYDROXYPHENYL)ETHANOATE", "non-polymer"),
        "PEJ": ("PERIODATE", "non-polymer"),
        "FAQ": ("PHENYLACETYL COENZYME A", "non-polymer"),
        "GA": ("GALLIUM (III) ION", "non-polymer"),
        "2TD": ("TRIDEC-12-EN-2-ONE", "non-polymer"),
        "FAV": (
            "N-(2-CHLOROPHENYL)-N-METHYL-4H-THIENO[3,2-C]CHROMENE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "XQG": ("6-METHYLSULFANYL-9H-PURINE", "non-polymer"),
        "NGQ": ("N-CARBAMIMIDOYL-L-GLUTAMIC ACID", "non-polymer"),
        "FAT": ("1-HEXYLDECANOIC ACID", "non-polymer"),
        "WCR": (
            "[(2R,4Z)-2-(AMINOMETHYL)-2-HYDROXY-4-(4-HYDROXYBENZYLIDENE)-5-OXOIMIDAZOLIDIN-1-YL]ACETIC ACID",
            "peptide",
        ),
        "0N9": (
            "5-HYDROXY-2-(1-METHYL-1H-IMIDAZOL-4-YL)-6-OXO-1,6-DIHYDROPYRIMIDINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "0N8": (
            "(2Z)-4-[1-BENZYL-4-(4-CHLOROBENZYL)PIPERIDIN-4-YL]-2-HYDROXY-4-OXOBUT-2-ENOIC ACID",
            "non-polymer",
        ),
        "JOS": ("JOSAMYCIN", "non-polymer"),
        "PRA": ("3-PHENYLPROPYLAMINE", "non-polymer"),
        "PRC": (
            "N-[4-METHYL-3-[[4-(3-PYRIDINYL)-2-PYRIMIDINYL]AMINO]PHENYL]-3-PYRIDINECARBOXAMIDE",
            "non-polymer",
        ),
        "PRB": ("13-ACETYLPHORBOL", "non-polymer"),
        "0N1": (
            "(4S,8E,11R)-4-[(1R)-1-HYDROXY-2-{[3-(PROPAN-2-YL)BENZYL]AMINO}ETHYL]-16-METHYL-11-PHENYL-6-OXA-3,12-DIAZABICYCLO[12.3.1]OCTADECA-1(18),8,14,16-TETRAENE-2,13-DIONE",
            "non-polymer",
        ),
        "0N0": (
            "(3S,5R)-N-{[9-(4-METHOXYBUTYL)-9H-XANTHEN-9-YL]METHYL}-5-{[(4-METHYLPHENYL)SULFONYL]AMINO}PIPERIDINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "0N3": (
            "3-({[(DIHYDROXYBORANYL)METHYL]SULFAMOYL}METHYL)BENZOIC ACID",
            "non-polymer",
        ),
        "0N2": (
            "(2R,4S,5R,6R)-4-AMINO-5-HYDROXY-6-METHYLTETRAHYDRO-2H-PYRAN-2-YL [(2R,3S,5R)-3-HYDROXY-5-(5-METHYL-2,4-DIOXO-3,4-DIHYDROPYRIMIDIN-1(2H)-YL)TETRAHYDROFURAN-2-YL]METHYL DIHYDROGEN DIPHOSPHATE",
            "non-polymer",
        ),
        "PRI": ("PYRROLIDINE-2-CARBALDEHYDE", "non-polymer"),
        "0N4": (
            "[({[4-(1H-TETRAZOL-5-YL)PHENYL]SULFONYL}AMINO)METHYL]BORONIC ACID",
            "non-polymer",
        ),
        "0N7": ("2-HYDROXYISOQUINOLINE-1,3(2H,4H)-DIONE", "non-polymer"),
        "0N6": (
            "(R)-(3-(3-CYANOPHENYL)-1-OXO-1-(PYRROLIDIN-1-YL)PROPAN-2-YL)-1,2,3,4-TETRAHYDROISOQUINOLINE-6-SULFONAMIDE",
            "non-polymer",
        ),
        "PRT": ("PHOSPHORIBOSYL ATP", "non-polymer"),
        "PRW": ("(2R,3R)-4-AMINO-2,3-DIHYDROXYBUTANOIC ACID", "non-polymer"),
        "FAI": (
            "5-(FORMYLAMINO)-1-(5-O-PHOSPHONO-BETA-D-RIBOFURANOSYL)-1H-IMIDAZOLE-4-CARBOXAMIDE",
            "nucleic-acid",
        ),
        "PRQ": ("(3S)-3-AMINO-3-(2-NITROPHENYL)PROPANOIC ACID", "peptide"),
        "PRP": ("ALPHA-PHOSPHORIBOSYLPYROPHOSPHORIC ACID", "saccharide"),
        "PRS": ("THIOPROLINE", "peptide"),
        "2TL": ("D-ALLOTHREONINE", "peptide"),
        "FAN": ("FORMANILIDE", "non-polymer"),
        "PRY": ("2-PROPYL-ANILINE", "non-polymer"),
        "PRX": ("ADENOSINE-5'-PROPYLPHOSPHATE", "non-polymer"),
        "6M0": ("MO(6)-O(26) CLUSTER", "non-polymer"),
        "FAO": ("?", "non-polymer"),
        "6MN": ("2-AMINO-2-DEOXY-6-O-PHOSPHONO-ALPHA-D-MANNOPYRANOSE", "non-polymer"),
        "0NX": ("(5-PHENYL-1,2-OXAZOL-3-YL)METHANOL", "non-polymer"),
        "2TN": (
            "2-(4-(2-HYDROXY-3-(ISOPROPYLAMINO)PROPOXY)PHENYL)ETHANAMIDE",
            "non-polymer",
        ),
        "FAL": (
            "(2R)-1-[4-({4-[(2,5-DICHLOROPHENYL)AMINO]PYRIMIDIN-2-YL}AMINO)PHENOXY]-3-(DIMETHYLAMINO)PROPAN-2-OL",
            "non-polymer",
        ),
        "6MI": ("6-METHYL-8-(2-DEOXY-RIBOFURANOSYL)ISOXANTHOPTERIDINE", "nucleic-acid"),
        "0NQ": (
            "N-{(1S)-3-[(5S)-5-TERT-BUTYL-1-(3,3-DIMETHYLBUTYL)-4-HYDROXY-2-OXO-2,5-DIHYDRO-1H-PYRROL-3-YL]-1-ETHOXY-1-OXIDO-1,4-DIHYDRO-2,4,1-BENZODIAZAPHOSPHININ-7-YL}METHANESULFONAMIDE",
            "non-polymer",
        ),
        "0NP": (
            "2',3'-DIDEOXY-2'-FLUORO-3'-TRIAZA-1,2-DIEN-2-IUM-1-YLURIDINE",
            "non-polymer",
        ),
        "0NS": (
            "2-METHOXY-N-(3-METHYL-2-OXO-1,2,3,4-TETRAHYDROQUINAZOLIN-6-YL)BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "6ME": (
            "(6S)-6-METHYL-5,6,7,8-TETRAHYDROQUINAZOLINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "0NU": (
            "3-[(3R)-3-(IMIDAZO[4,5-D]PYRROLO[2,3-B]PYRIDIN-1(6H)-YL)PIPERIDIN-1-YL]-3-OXOPROPANENITRILE",
            "non-polymer",
        ),
        "0NT": (
            "[4-(IMIDAZO[4,5-D]PYRROLO[2,3-B]PYRIDIN-1(6H)-YL)PIPERIDIN-1-YL][(2S)-1-(PROPAN-2-YL)PYRROLIDIN-2-YL]METHANONE",
            "non-polymer",
        ),
        "0NW": (
            "3-[(2S)-3-[4-(2-AMINOETHYL)PIPERIDIN-1-YL]-2-{[(2',4'-DICHLOROBIPHENYL-3-YL)SULFONYL]AMINO}-3-OXOPROPYL]BENZENECARBOXIMIDAMIDE",
            "non-polymer",
        ),
        "0NV": (
            "1-(1-BENZYLPIPERIDIN-4-YL)-1,6-DIHYDROIMIDAZO[4,5-D]PYRROLO[2,3-B]PYRIDINE",
            "non-polymer",
        ),
        "PR5": ("?", "nucleic-acid"),
        "0NH": (
            "1-[4-METHYL-1-(METHYLSULFONYL)PIPERIDIN-4-YL]-1,6-DIHYDROIMIDAZO[4,5-D]PYRROLO[2,3-B]PYRIDINE",
            "non-polymer",
        ),
        "PR7": ("1-[(2S)-2-CARBOXY-2-HYDROXYETHYL]-L-PROLINE", "peptide"),
        "0NJ": ("2-(2-FLUOROPHENYL)-6-METHOXY-1,3-BENZOTHIAZOLE", "non-polymer"),
        "0NM": ("CYANIC ACID", "non-polymer"),
        "0NL": (
            "TERT-BUTYL [(1R,3R)-3-(IMIDAZO[4,5-D]PYRROLO[2,3-B]PYRIDIN-1(6H)-YL)CYCLOPENTYL]CARBAMATE",
            "non-polymer",
        ),
        "PR3": ("S,S-PROPYLTHIOCYSTEINE", "peptide"),
        "0NN": (
            "1-BENZYL-N-[3-(1'H,3H-SPIRO[2-BENZOFURAN-1,4'-PIPERIDIN]-1'-YL)PROPYL]-D-PROLINAMIDE",
            "non-polymer",
        ),
        "0NA": (
            "[({[6-(1H-TETRAZOL-5-YL)PYRIDIN-3-YL]SULFONYL}AMINO)METHYL]BORONIC ACID",
            "non-polymer",
        ),
        "SNB": (
            "1-(3-BROMOPHENYL)-7-CHLORO-6-METHOXY-3,4-DIHYDROISOQUINOLINE",
            "non-polymer",
        ),
        "0NC": ("N-METHYL-L-ALANINAMIDE", "peptide"),
        "0NB": (
            "[({[3-CHLORO-4-(1H-TETRAZOL-5-YL)PHENYL]SULFONYL}AMINO)METHYL]BORONIC ACID",
            "non-polymer",
        ),
        "0NE": (
            "[({[4-(1H-TETRAZOL-5-YL)-2-(TRIFLUOROMETHYL)PHENYL]SULFONYL}AMINO)METHYL]BORONIC ACID",
            "non-polymer",
        ),
        "0ND": (
            "[({[2-CHLORO-4-(1H-TETRAZOL-5-YL)PHENYL]SULFONYL}AMINO)METHYL]BORONIC ACID",
            "non-polymer",
        ),
        "0NG": ("({[(CHLOROMETHYL)SULFONYL]AMINO}METHYL)BORONIC ACID", "non-polymer"),
        "0NF": (
            "(2S)-N-[3-(2-AMINOPROPAN-2-YL)-5-(TRIFLUOROMETHYL)PHENYL]-7-[(7-OXO-5,6,7,8-TETRAHYDRO-1,8-NAPHTHYRIDIN-4-YL)OXY]-1,2,3,4-TETRAHYDRONAPHTHALENE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "2TU": (
            "1-(BETA-D-RIBOFURANOSYL)-2-THIOXO-2,3-DIHYDROPYRIMIDIN-4(1H)-ONE",
            "non-polymer",
        ),
        "FAA": ("N5-(4-HYDROXYBENZYL)FLAVIN-ADENINE DINUCLEOTIDE", "non-polymer"),
        "PSY": ("PHENYL ETHENESULFONATE", "non-polymer"),
        "SNA": (
            "N-{1-[5-(1-CARBAMOYL-2-MERCAPTO-ETHYLCARBAMOYL)-PENTYLCARBAMOYL]-2-[4-(DIFLUORO-PHOSPHONO-METHYL)-PHENYL]-ETHYL}-3-{2-[4-(DIFLUORO-PHOSPHONO-METHYL)-PHENYL]-ACETYLAMINO}-SUCCINAMIC ACID",
            "non-polymer",
        ),
        "7YG": (
            "3-AMINO-6-(4-{[2-(DIMETHYLAMINO)ETHYL]SULFAMOYL}PHENYL)-N-PYRIDIN-3-YLPYRAZINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "SNN": ("L-3-AMINOSUCCINIMIDE", "peptide"),
        "XUG": ("(D)-2'-METHYLSELENYL-2'-DEOXYGUANOSINE-5'-PHOSPHATE", "nucleic-acid"),
        "SNO": ("(S)-PARA-NITROSTYRENE OXIDE", "non-polymer"),
        "XQ0": ("3-METHYL-1H-INDAZOL-6-OL", "non-polymer"),
        "PEF": ("DI-PALMITOYL-3-SN-PHOSPHATIDYLETHANOLAMINE", "non-polymer"),
        "QK0": (
            "N-[6-(6-CHLORO-5-{[(4-FLUOROPHENYL)SULFONYL]AMINO}PYRIDIN-3-YL)-1,3-BENZOTHIAZOL-2-YL]ACETAMIDE",
            "non-polymer",
        ),
        "ATS": ("?", "non-polymer"),
        "Y02": (
            "(2R,3R,4S,5R,6E)-3,4,5-TRIHYDROXY-2-METHOXY-8,8-DIMETHYL-N-[2-(2,4,6-TRIMETHYLPHENOXY)ETHYL]NON-6-ENAMIDE (NON-PREFERRED NAME)",
            "non-polymer",
        ),
        "J80": (
            "(METHYLPYRIDAZINE PIPERIDINE BUTYLOXYPHENYL)ETHYLACETATE",
            "non-polymer",
        ),
        "Y00": ("DOBUTAMINE", "non-polymer"),
        "MAZ": ("FORMIC ACID 3-AMINO-BENZYL ESTER", "non-polymer"),
        "2TS": (
            "(2S)-N-[[2-(AMINOMETHYL)-5-CHLORO-PHENYL]METHYL]-1-[(2R)-5-CARBAMIMIDAMIDO-2-(PHENYLMETHYLSULFONYLAMINO)PENTANOYL]PYRROLIDINE-2-CARBOXAMIDE",
            "peptide",
        ),
        "G30": (
            "(1S,2R)-2-[(2,5-DIFLUOROPHENYL)CARBAMOYL]CYCLOPROPANECARBOXYLIC ACID",
            "non-polymer",
        ),
        "SS5": (
            "(2S)-1-[3-(CYCLOHEXYLMETHOXY)PHENYL]-3-{[5-(3-METHYL-1H-INDAZOL-5-YL)PYRIDIN-3-YL]OXY}PROPAN-2-AMINE",
            "non-polymer",
        ),
        "J88": (
            "(3Z)-1-[(6-FLUORO-4H-1,3-BENZODIOXIN-8-YL)METHYL]-4-PHENYL-1H-INDOLE-2,3-DIONE 3-OXIME",
            "non-polymer",
        ),
        "Y08": (
            "(E,2R,3R,4S,5R)-N-[[(3S)-1-CYCLOPROPYLCARBONYLPIPERIDIN-3-YL]METHYL]-2-METHOXY-8,8-DIMETHYL-3,4,5-TRIS(OXIDANYL)NON-6-ENAMIDE",
            "non-polymer",
        ),
        "SS6": (
            "1-{3-[(6-AMINOQUINAZOLIN-4-YL)AMINO]PHENYL}-3-[3-TERT-BUTYL-1-(3-METHYLPHENYL)-1H-PYRAZOL-5-YL]UREA",
            "non-polymer",
        ),
        "SS1": ("1-PHENYLETHANOL", "non-polymer"),
        "G33": ("?", "nucleic-acid"),
        "SS3": (
            "(2S)-1-{[5-(3-METHYL-1H-INDAZOL-5-YL)PYRIDIN-3-YL]OXY}-3-PHENYLPROPAN-2-AMINE",
            "non-polymer",
        ),
        "SS2": ("(1R)-1-PHENYLETHANOL", "non-polymer"),
        "XUL": ("D-XYLULOSE", "saccharide"),
        "G32": ("O6-METHYL-2'-DEOXY-GUANOSINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "ZOO": (
            "(5R)-2-AMINO-5-[4-(DIFLUOROMETHOXY)PHENYL]-5-[4-FLUORO-3-(5-FLUOROPENT-1-YN-1-YL)PHENYL]-3-METHYL-3,5-DIHYDRO-4H-IMIDAZOL-4-ONE",
            "non-polymer",
        ),
        "G35": (
            "(5R)-5-CARBAMIMIDAMIDO-1-(2-DEOXY-5-O-PHOSPHONO-BETA-D-ERYTHRO-PENTOFURANOSYL)IMIDAZOLIDINE-2,4-DIONE",
            "nucleic-acid",
        ),
        "X6B": (
            "4-AMINO-N-(3-FLUOROPHENYL)-2-[(4-SULFAMOYLPHENYL)AMINO]-1,3-THIAZOLE-5-CARBOXAMIDE",
            "non-polymer",
        ),
        "G34": (
            "(3AS,4R,5S,6S,8R,9R,9AR,10R)-5-HYDROXY-4,6,9,10-TETRAMETHYL-1-OXO-6-VINYLDECAHYDRO-3A,9-PROPANOCYCLOPENTA[8]ANNULEN-8-YL {[(3-EXO)-8-METHYL-8-AZABICYCLO[3.2.1]OCT-3-YL]THIO}ACETATE",
            "non-polymer",
        ),
        "NLQ": ("N~2~-ACETYL-L-GLUTAMINE", "peptide"),
        "24D": (
            "N-(2,4-DICHLOROBENZYL)-4-(PYRIMIDIN-2-YLOXY)PIPERIDINE-1-CARBOXAMIDE",
            "non-polymer",
        ),
        "A4D": ("5'-THIOADENOSINE", "non-polymer"),
        "24F": (
            "(2R)-({[5-(4-ETHOXYPHENYL)THIOPHEN-2-YL]SULFONYL}AMINO){1-[(1-METHYLETHOXY)CARBONYL]PIPERIDIN-4-YL}ETHANOIC ACID",
            "non-polymer",
        ),
        "24A": (
            "{[(4-{[2-(4-AMINOCYCLOHEXYL)-9-ETHYL-9H-PURIN-6-YL]AMINO}PHENYL)(HYDROXY)PHOSPHORYL]METHYL}PHOSPHONIC ACID",
            "non-polymer",
        ),
        "A4C": (
            "9,9'-(HEXANE-1,6-DIYLDIIMINO)BIS{N-[2-(DIMETHYLAMINO)ETHYL]ACRIDINE-4-CARBOXAMIDE}",
            "non-polymer",
        ),
        "HMZ": ("4-(2-HYDROXY-4-METHOXYPHENYL)-4-OXOBUTANOIC ACID", "non-polymer"),
        "24B": ("(2,4-DIFLUOROPHENYL)METHANOL", "non-polymer"),
        "24M": (
            "(1R,2R)-2-[(1S,2S)-2-AMINO-1-HYDROXY-4-METHYLPENTYL]CYCLOPENTANECARBOXYLIC ACID",
            "peptide",
        ),
        "24L": (
            "4-{4-[({[(9AS)-8-ACETYL-1,7-DIHYDROXY-3-METHOXY-9A-METHYL-9-OXO-9,9A-DIHYDRODIBENZO[B,D]FURAN-4-YL]CARBONYL}AMINO)METHYL]NAPHTHALEN-2-YL}BUTANOIC ACID",
            "non-polymer",
        ),
        "24O": (
            "(1R,2R)-2-[(1S,2S)-2-AMINO-1-HYDROXY-4-METHYLPENTYL]-4-OXOCYCLOPENTANECARBOXYLIC ACID",
            "peptide",
        ),
        "G36": ("O6-ETHYL-2'-DEOXYGUANOSINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "24I": (
            "(2S)-2-{[HYDROXY(4-IODOBENZYL)PHOSPHORYL]METHYL}PENTANEDIOIC ACID",
            "non-polymer",
        ),
        "24H": (
            "{[(1S)-2-(2-AMINO-6-OXO-1,6-DIHYDRO-9H-PURIN-9-YL)-1-(HYDROXYMETHYL)ETHOXY]METHYL}PHOSPHONIC ACID",
            "non-polymer",
        ),
        "HMR": ("BETA-HOMOARGININE", "peptide"),
        "HMS": ("5-O-PHOSPHONO-L-RIBULOSE", "saccharide"),
        "24U": ("1-BUTANOYL-N-(4-CARBAMIMIDOYLBENZYL)-L-PROLINAMIDE", "non-polymer"),
        "24T": ("2-AMINO-4-METHYLTHIAZOLE", "non-polymer"),
        "HMN": (
            "2,4,6,7,8,9-HEXAHYDROXY-5-METHYLCARBOXAMIDO NONANOIC ACID",
            "non-polymer",
        ),
        "HMO": ("4'-HYDROXY-7-METHOXYISOFLAVONE", "non-polymer"),
        "DL7": (
            "(5BETA,14BETA,17ALPHA,25R)-3-OXOCHOLEST-7-EN-26-OIC ACID",
            "non-polymer",
        ),
        "24P": ("(2R)-2-[(4-BENZYLPHENOXY)METHYL]PYRROLIDINE", "non-polymer"),
        "24S": (
            "N-({(2Z)-2-[(2-NITROPHENYL)METHYLIDENE]HYDRAZINO}CARBONOTHIOYL)-BETA-D-GLUCOPYRANOSYLAMINE",
            "non-polymer",
        ),
        "DL4": ("(14BETA,17ALPHA,25R)-3-OXOCHOLEST-4-EN-26-OIC ACID", "non-polymer"),
        "HMD": (
            "4-(5-AMINO-4-OXO-4H-PYRAZOL-3-YL)-2-BROMO-4,5,6,7-TETRAHYDRO-3AH-PYRROLO[2,3-C]AZEPIN-8-ONE",
            "non-polymer",
        ),
        "HME": ("PORPHYCENE CONTAINING FE", "non-polymer"),
        "HMF": ("2-AMINO-4-PHENYL-BUTYRIC ACID", "peptide"),
        "DL8": (
            "N-[(4-PHENYL-1H-1,2,3-TRIAZOL-1-YL)ACETYL]-BETA-D-GLUCOPYRANOSYLAMINE",
            "non-polymer",
        ),
        "24X": ("?", "non-polymer"),
        "HMB": ("(4S)-4-HYDROXY-4-METHOXYBUTANOIC ACID", "non-polymer"),
        "24Z": (
            "4-({4-CARBAMOYL-2-NITRO-5-[(PYRIDIN-3-YLMETHYL)AMINO]PHENYL}AMINO)BUTANOIC ACID",
            "non-polymer",
        ),
        "DLA": ("D-LACTIC ACID", "non-polymer"),
        "DLG": ("HEXYL 3-DEOXY-BETA-D-GALACTOPYRANOSE", "saccharide"),
        "AG9": ("AGMATIDINE", "non-polymer"),
        "DLE": ("D-LEUCINE", "peptide"),
        "DLD": (
            "2-AZIDOETHYL 2-(ACETYLAMINO)-4-O-[2-(ACETYLAMINO)-2-DEOXY-BETA-D-GALACTOPYRANOSYL]-2-DEOXY-BETA-D-GLUCOPYRANOSIDE",
            "non-polymer",
        ),
        "HM4": (
            "TERT-BUTYL {2-[(1,3-THIAZOL-2-YLAMINO)CARBONYL]PYRIDIN-3-YL}CARBAMATE",
            "non-polymer",
        ),
        "HM5": (
            "3-[(2,2-DIMETHYLPROPANOYL)AMINO]-N-1,3-THIAZOL-2-YLPYRIDINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "DLI": (
            "(3R,4S)-1-{6-[3-(METHYLSULFONYL)PHENYL]PYRIMIDIN-4-YL}-4-(2,4,5-TRIFLUOROPHENYL)PYRROLIDIN-3-AMINE",
            "non-polymer",
        ),
        "AG8": (
            "S-[(2E,6E,10E)-14-(DIMETHYLAMINO)-3,7,11-TRIMETHYLTETRADECA-2,6,10-TRIEN-1-YL] TRIHYDROGEN THIODIPHOSPHATE",
            "non-polymer",
        ),
        "DLN": (
            "3-{[4-AMINO-1-(1-METHYLETHYL)-1H-PYRAZOLO[3,4-D]PYRIMIDIN-3-YL]ETHYNYL}PHENOL",
            "non-polymer",
        ),
        "HM2": (
            "5-CHLORO-6-METHYL-N-(2-PHENYLETHYL)-2-PYRIDIN-2-YLPYRIMIDIN-4-AMINE",
            "non-polymer",
        ),
        "DLL": ("5'-O-[HYDROXY(PHENYLACETYL)PHOSPHORYL]ADENOSINE", "non-polymer"),
        "245": ("1-(2-CHLOROPYRIDIN-4-YL)-3-PHENYLUREA", "non-polymer"),
        "AG7": (
            "4-{2-(4-FLUORO-BENZYL)-6-METHYL-5-[(5-METHYL-ISOXAZOLE-3-CARBONYL)-AMINO]-4-OXO-HEPTANOYLAMINO}-5-(2-OXO-PYRROLIDIN-3-YL)-PENTANOIC ACID ETHYL ESTER",
            "non-polymer",
        ),
        "247": (
            "(3R)-3-AMINO-2,2-DIFLUORO-3-(4-HYDROXYPHENYL)PROPANOIC ACID",
            "non-polymer",
        ),
        "A45": (
            "3-(5-AMINO-7-HYDROXY-[1,2,3]TRIAZOLO[4,5-D]PYRIMIDIN-2-YL)-BENZOIC ACID",
            "non-polymer",
        ),
        "241": (
            "(2S)-2-(3-{[1-(4-METHOXYBENZOYL)-2-METHYL-5-(TRIFLUOROMETHOXY)-1H-INDOL-3-YL]METHYL}PHENOXY)PROPANOIC ACID",
            "non-polymer",
        ),
        "A43": ("?", "nucleic-acid"),
        "A40": ("?", "nucleic-acid"),
        "AG6": (
            "N-[(BENZYLOXY)CARBONYL]-L-ALPHA-GLUTAMYL-N-[(1S)-4-OXO-4-PHENYL-1-PROPYLBUT-2-EN-1-YL]-L-PHENYLALANINAMIDE",
            "non-polymer",
        ),
        "DLZ": (
            "1-DEOXY-1-(6,7-DIMETHYL-2,4-DIOXO-3,4-DIHYDROPTERIDIN-8(2H)-YL)-D-RIBITOL",
            "non-polymer",
        ),
        "DLY": ("D-LYSINE", "peptide"),
        "AG5": (
            "N-(2-PHENYLETHYL)-2-(PHENYLSULFANYL)-5-SULFAMOYLPYRIDINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "A48": ("4-[(DIMESITYLBORYL)(2,2,2-TRIFLUOROETHYL)AMINO]PHENOL", "non-polymer"),
        "A49": (
            "N-{1-[(1-CARBAMOYLCYCLOPROPYL)METHYL]PIPERIDIN-4-YL}-N-CYCLOPROPYL-4-[(1S)-2,2,2-TRIFLUORO-1-HYDROXY-1-METHYLETHYL]BENZAMIDE",
            "non-polymer",
        ),
        "UP3": ("?", "non-polymer"),
        "AG4": (
            "N-(3-METHOXYPROPYL)-2-(PHENYLSULFANYL)-5-SULFAMOYLPYRIDINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "PVW": (
            "3-(2-METHOXYETHYL)-2-THIOXO-1,2,3,7-TETRAHYDRO-6H-PURIN-6-ONE",
            "non-polymer",
        ),
        "SSU": ("URIDINE-5'-PHOSPHOROTHIOATE", "nucleic-acid"),
        "AG3": ("1-{4-[(3-AMINOPROPYL)AMINO]BUTYL}GUANIDINE", "non-polymer"),
        "Y0X": ("?", "non-polymer"),
        "J8Z": (
            "(2R)-2-[(1-BENZOTHIOPHEN-2-YLCARBONYL)AMINO]-3-(3-METHYLPHENYL)PROPYL PHOSPHATE",
            "non-polymer",
        ),
        "S54": (
            "N-(BENZYLSULFONYL)-D-VALYL-N-(4-CARBAMIMIDOYLBENZYL)-L-PROLINAMIDE",
            "peptide",
        ),
        "SSP": ("DITHIOBIS-(PROPANE)", "non-polymer"),
        "SSS": (
            "N-{[5-AMINO-1-(5-O-PHOSPHONO-BETA-D-ARABINOFURANOSYL)-1H-IMIDAZOL-4-YL]CARBONYL}-L-ASPARTIC ACID",
            "non-polymer",
        ),
        "AG2": ("AGMATINE", "non-polymer"),
        "SSM": (
            "(2S)-2-(ACETYLAMINO)-N-METHYL-4-[(S)-METHYLSULFINYL]BUTANAMIDE",
            "non-polymer",
        ),
        "SSL": (
            "(3-{[2-(DIHYDROXYAMINO)-4-(2LAMBDA~5~-TRIAZA-1,2-DIENYL)PHENYL]AMINO}-9-HYDROXY-8,8,10,10-TETRAMETHYL-1,5-DIOXA-9-AZASPIRO[5.5]UNDEC-3-YL)METHYL TRIHYDROGEN DIPHOSPHATE",
            "non-polymer",
        ),
        "PEB": ("PHYCOERYTHROBILIN", "non-polymer"),
        "SSN": ("4-OXOBUTANOIC ACID", "non-polymer"),
        "SSI": (
            "[(3S)-4-HYDROXY-3-{[(4-OXO-4,5-DIHYDRO-3H-PYRROLO[3,2-D]PYRIMIDIN-7-YL)METHYL]AMINO}BUTYL]PHOSPHONIC ACID",
            "non-polymer",
        ),
        "AG1": ("RUTHENIUM PYRIDOCARBAZOLE", "non-polymer"),
        "CP1": (
            "2-(METHYLAMINO)-ETHYLGLYCINE-CARBONYLMETHYLENE-CYTOSINE",
            "nucleic-acid",
        ),
        "SSE": (
            "1,4-DIDEOXY-1,4-[[2S,3S)-2,4-DIHYDROXY-3-(SULFOXY)BUTYL]EPISELENONIUMYLIDENE]-D-ARABINITOL INNER SALT",
            "non-polymer",
        ),
        "SSD": (
            "1,4-DIDEOXY-1,4-[[2R,3R)-2,4-DIHYDROXY-3-(SULFOXY)BUTYL]EPISULFONIUMYLIDENE]-D-ARABINITOL INNER SALT",
            "non-polymer",
        ),
        "SSG": ("1,4-DEOXY-1,4-DITHIO-BETA-D-GLUCOPYRANOSE", "saccharide"),
        "AG0": (
            "(6R)-6-CYCLOPENTYL-6-[2-(2,6-DIETHYLPYRIDIN-4-YL)ETHYL]-3-[(5,7-DIMETHYL[1,2,4]TRIAZOLO[1,5-A]PYRIMIDIN-2-YL)METHYL]-4-HYDROXY-5,6-DIHYDRO-2H-PYRAN-2-ONE",
            "non-polymer",
        ),
        "SSA": ("5'-O-(N-(L-SERYL)-SULFAMOYL)ADENOSINE", "non-polymer"),
        "SSC": ("(2S,5S)-5-CARBOXYMETHYLPROLINE", "non-polymer"),
        "SSB": ("3-BUTYLTHIOLANE 1-OXIDE", "non-polymer"),
        "OAL": (
            "(1,8-DIHYDROXY-9,10-DIOXO-9,10-DIHYDRO-ANTHRACEN-2-YL)-ACETIC ACID",
            "non-polymer",
        ),
        "EVI": ("1,3-THIAZOLE-2-SULFONAMIDE", "non-polymer"),
        "2RB": (
            "N'-[(1E)-(3,5-DIBROMO-2,4-DIHYDROXYPHENYL)METHYLIDENE]-4-METHOXYBENZOHYDRAZIDE",
            "non-polymer",
        ),
        "H24": (
            "(6S)-2-AMINO-6-(3'-METHOXYBIPHENYL-3-YL)-3,6-DIMETHYL-5,6-DIHYDROPYRIMIDIN-4(3H)-ONE",
            "non-polymer",
        ),
        "H26": (
            "{2-[2-(6-OXO-1,6-DIHYDRO-9H-PURIN-9-YL)ETHOXY]ETHYL}PHOSPHONIC ACID",
            "non-polymer",
        ),
        "H21": (
            "5-CHLORO-N-[1-(2-{[2-FLUORO-4-(2-OXOPYRIDIN-1(2H)-YL)PHENYL]AMINO}-2-OXOETHYL)-1H-1,2,4-TRIAZOL-3-YL]THIOPHENE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "H20": (
            "6-CHLORO-4-(CYCLOHEXYLOXY)-3-ISOPROPYLQUINOLIN-2(1H)-ONE",
            "non-polymer",
        ),
        "H23": ("(1-HYDROXYDODECANE-1,1-DIYL)BIS(PHOSPHONIC ACID)", "non-polymer"),
        "H22": (
            "5-CHLORO-N-[(3R)-1-(2-{[2-FLUORO-4-(2-OXOPYRIDIN-1(2H)-YL)PHENYL]AMINO}-2-OXOETHYL)PYRROLIDIN-3-YL]THIOPHENE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "PEC": ("S,S-PENTYLTHIOCYSTEINE", "peptide"),
        "TLB": (
            "2'-O,3'-C-METHYLENE-ARABINOFURANOSYL-THYMINE-5'-MONOPHOSPHATE",
            "nucleic-acid",
        ),
        "ECG": (
            "3-HYDROXY-8-METHYL-8-AZA-BICYCLO[3.2.1]OCTANE-2-CARBOXYLIC ACID METHYL ESTER",
            "non-polymer",
        ),
        "06P": (
            "D-PHENYLALANYL-N-[(1-METHYLPYRIDINIUM-4-YL)METHYL]-L-PROLINAMIDE",
            "peptide",
        ),
        "RMP": ("2'-DEOXY-ADENOSINE-5'-RP-MONOMETHYLPHOSPHONATE", "nucleic-acid"),
        "RVI": (
            "(2R)-2-{[2-AMINO-6-(2-METHYLPHENYL)QUINOLIN-3-YL]METHYL}-N-(CYCLOHEXYLMETHYL)PENTANAMIDE",
            "non-polymer",
        ),
        "06U": (
            "6-ETHYL-5-{(3R)-3-[3-METHOXY-5-(PYRIDIN-4-YL)PHENYL]BUT-1-YN-1-YL}PYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "06T": (
            "(5R,6S)-3-(L-ALANYL)-5,6,15,15,18-PENTAMETHYL-17-OXO-2,3,4,5,6,7,14,15,16,17-DECAHYDRO-1H-12,8-(METHENO)[1,5,9]TRIAZACYCLOTETRADECINO[1,2-A]INDOLE-9-CARBOXAMIDE",
            "non-polymer",
        ),
        "06W": (
            "6-ETHYL-5-{(3S)-3-[3-METHOXY-5-(PYRIDIN-4-YL)PHENYL]BUT-1-YN-1-YL}PYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "06V": (
            "6-ETHYL-5-{(3R)-3-[3-METHOXY-5-(MORPHOLIN-4-YL)PHENYL]BUT-1-YN-1-YL}PYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "06Y": ("2-PHENOXYETHANOIC ACID", "non-polymer"),
        "06X": ("AMLODIPINE", "non-polymer"),
        "06Z": (
            "{4-AMINO-2-[(3-FLUOROPHENYL)AMINO]-1,3-THIAZOL-5-YL}(PHENYL)METHANONE",
            "non-polymer",
        ),
        "LNQ": (
            "3-AMINO-4-{3-[2-(2-PROPOXY-ETHOXY)-ETHOXY]-PROPYLAMINO}-CYCLOBUT-3-ENE-1,2-DIONE",
            "non-polymer",
        ),
        "LNP": (
            "(3ALPHA,9BETA,10ALPHA,13ALPHA)-30-CYCLOPROPYLIDENELANOST-7-EN-3-OL",
            "non-polymer",
        ),
        "ECH": ("BETA,BETA-CAROTEN-4-ONE", "non-polymer"),
        "LNM": ("N-METHYL-L-LEUCINAMIDE", "peptide"),
        "LNL": ("ALPHA-LINOLENIC ACID", "non-polymer"),
        "06C": ("IODOMETHANE", "non-polymer"),
        "ECT": ("ECTEINASCIDIN 743", "non-polymer"),
        "ECS": ("2,3,17BETA-TRIHYDROXY-1,3,5(10)-ESTRATRIENE", "non-polymer"),
        "LNH": (
            "3-{5-(ACETYLAMINO)-3-[3-CYANO-7-(CYCLOPROPYLAMINO)PYRAZOLO[1,5-A]PYRIMIDIN-5-YL]-1H-INDOL-1-YL}PROPANOIC ACID",
            "non-polymer",
        ),
        "BB9": ("(2Z)-2-AMINO-3-SULFANYLPROP-2-ENOIC ACID", "peptide"),
        "06F": (
            "1-[3-TERT-BUTYL-1-(4-METHYLPHENYL)-1H-PYRAZOL-5-YL]-3-{3-[(5-OXO-6,7,8,9-TETRAHYDRO-5H-BENZO[7]ANNULEN-2-YL)AMINO]PHENYL}UREA",
            "non-polymer",
        ),
        "BB7": ("(2Z)-2-AMINO-4-METHOXY-3-SULFANYLBUT-2-ENOIC ACID", "peptide"),
        "06H": (
            "(6S)-4,6,15,15,18-PENTAMETHYL-5,17-DIOXO-2,3,4,5,6,7,14,15,16,17-DECAHYDRO-1H-12,8-(METHENO)[1,4,9]TRIAZACYCLOTETRADECINO[9,8-A]INDOLE-9-CARBOXAMIDE",
            "non-polymer",
        ),
        "LNG": ("DELTA-3ISOTETRADECENOIC ACID", "non-polymer"),
        "06J": (
            "(3AR)-13,13,16-TRIMETHYL-15-OXO-1,2,3,3A,4,5,12,14,15,17,18,19-DODECAHYDRO-13H-10,6-(METHENO)PYRROLO[2',1':3,4][1,4,9]TRIAZACYCLOTETRADECINO[9,8-A]INDOLE-7-CARBOXAMIDE",
            "non-polymer",
        ),
        "BB3": (
            "3-(3,5-DIBROMO-4-HYDROXY-BENZOYL)-2-ETHYL-BENZOFURAN-6-SULFONIC ACID DIMETHYLAMIDE",
            "non-polymer",
        ),
        "BB2": ("ACTINONIN", "non-polymer"),
        "BB1": (
            "2-[(FORMYL-HYDROXY-AMINO)-METHYL]-HEXANOIC ACID (1-DIMETHYLCARBAMOYL-2,2-DIMETHYL-PROPYL)-AMIDE",
            "non-polymer",
        ),
        "BB0": ("1-[6-(DIMETHYLAMINO)NAPHTHALEN-2-YL]ETHANONE", "non-polymer"),
        "061": (
            "2-BUTYL-6-HYDROXY-3-[2'-(1H-TETRAZOL-5-YL)-BIPHENYL-4-YLMETHYL]-3H-QUINAZOLIN-4-ONE",
            "non-polymer",
        ),
        "060": ("S-METHYL-D-CYSTEINE", "peptide"),
        "BBM": (
            "5-BROMO-N-(2,3-DIHYDROXYPROPOXY)-3,4-DIFLUORO-2-[(2-FLUORO-4-IODOPHENYL)AMINO]BENZAMIDE",
            "non-polymer",
        ),
        "062": (
            "3-AZA-9-HYDROXY-9-FLUORENYLCARBONYL-L-PROLYL-2-AMINOMETHYL-5-CHLOROBENZYLAMIDE, N-OXIDE",
            "non-polymer",
        ),
        "065": (
            "(3R,3AS,6AR)-HEXAHYDROFURO[2,3-B]FURAN-3-YL(2S,3R)-3-HYDROXY-4-(N-ISOBUTYLBENZO[D][1,3]DIOXOLE-5-SULFONAMIDO)-1-PHENYLBUTAN-2-YLCARBAMATE",
            "non-polymer",
        ),
        "064": (
            "3-[(E)-2-(2-CHLORO-4-{[3-(2,6-DICHLOROPHENYL)-5-(1-METHYLETHYL)ISOXAZOL-4-YL]METHOXY}PHENYL)ETHENYL]BENZOIC ACID",
            "non-polymer",
        ),
        "067": (
            "P-(4-RUTHENOCENYL-1H-1,2,3-TRIAZOL-1-YL)BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "066": (
            "6-{[3-(DIMETHYLCARBAMOYL)PHENYL]SULFONYL}-4-[(3-METHOXYPHENYL)AMINO]-8-METHYLQUINOLINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "069": ("N,3-DIHYDROXYBENZAMIDE", "non-polymer"),
        "068": ("?", "non-polymer"),
        "BBE": (
            "N-(6-AMINO-1-BENZYL-2,4-DIOXO-1,2,3,4-TETRAHYDROPYRIMIDIN-5-YL)-N,3-DIMETHYLBENZENESULFONAMIDE",
            "non-polymer",
        ),
        "JKK": ("2-[(3-CHLOROBENZYL)AMINO]ETHANESULFONIC ACID", "non-polymer"),
        "BBC": ("3-[(4-AMINOBUTYL)SULFINYL]-2-IMINOPROPAN-1-OL", "peptide"),
        "BBB": (
            "1-((2-HYDROXYETHOXY)METHYL)-5-(3-(BENZYLOXY)BENZYL)-6-HYDROXYPYRIMIDINE-2,4(1H,3H)-DIONE",
            "non-polymer",
        ),
        "BBA": ("2,7-BIS-(4-AMIDINOBENZYLIDENE)-CYCLOHEPTAN-1-ONE", "non-polymer"),
        "EVE": ("1H-BENZIMIDAZOLE-2-SULFONAMIDE", "non-polymer"),
        "EC5": (
            "(2R)-2-[5-CYCLOPROPYL-6-(HYDROXYSULFANYL)-4-(NAPHTHALEN-1-YLMETHYL)-2-OXOPYRIDIN-1(2H)-YL]-3-PHENYLPROPANOIC ACID",
            "non-polymer",
        ),
        "BBZ": (
            "2'-(4-DIMETHYLAMINOPHENYL)-5-(4-METHYL-1-PIPERAZINYL)-2,5'-BI-BENZIMIDAZOLE",
            "non-polymer",
        ),
        "BBY": (
            "5-(7-BROMO-1,3-BENZODIOXOL-5-YL)-1-METHYL-1H-PYRAZOL-3-AMINE",
            "non-polymer",
        ),
        "BBX": ("1-BROMOETHANE", "non-polymer"),
        "BBV": (
            "BENZYL 2-(ACETYLAMINO)-2-DEOXY-ALPHA-D-GLUCOPYRANOSIDE",
            "non-polymer",
        ),
        "BBU": ("1-BROMOBUTANE", "non-polymer"),
        "BBT": (
            "2-HYDROXY-5-[4-(2-HYDROXY-ETHYL)-PIPERIDIN-1-YL]-5-PHENYL-1H-PYRIMIDINE-4,6-DIONE",
            "non-polymer",
        ),
        "BBS": ("4-TERT-BUTYLBENZENESULFONIC ACID", "non-polymer"),
        "BBR": ("(3-BROMO-2-OXO-PROPOXY)PHOSPHONIC ACID", "non-polymer"),
        "BBQ": (
            "CYCLOPENTYL N-[(2S)-1-[(2S,4R)-2-[[(4R)-8-HYDROXY-1,6,10-TRIOXA-5$L^{4}-BORASPIRO[4.5]DECAN-4-YL]CARBAMOYL]-4-ISOQUINOLIN-1-YLOXY-PYRROLIDIN-1-YL]-3,3-DIMETHYL-1-OXO-BUTAN-2-YL]CARBAMATE",
            "non-polymer",
        ),
        "BBP": (
            "(5R)-3-TERT-BUTYL-1-(6-CHLORO-1,3-BENZOTHIAZOL-2-YL)-4,5-DIHYDRO-1H-PYRAZOL-5-OL",
            "non-polymer",
        ),
        "H2U": ("5,6-DIHYDROURIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "H2V": (
            "S-[2-({N-[(2S)-2-HYDROXY-3,3-DIMETHYL-4-(PHOSPHONOOXY)BUTANOYL]-BETA-ALANYL}AMINO)ETHYL] AMINOETHANETHIOATE",
            "non-polymer",
        ),
        "H2P": ("HEPTULOSE-2-PHOSPHATE", "saccharide"),
        "H2S": ("HYDROSULFURIC ACID", "non-polymer"),
        "H2K": (
            "1-{5-CHLORO-2-[(3R)-PYRROLIDIN-3-YLOXY]PHENYL}-3-(5-CYANOPYRAZIN-2-YL)UREA",
            "non-polymer",
        ),
        "OAH": (
            "9-(2-DEOXY-5-O-PHOSPHONO-ALPHA-D-THREO-PENTOFURANOSYL)-9H-PURIN-6-AMINE",
            "non-polymer",
        ),
        "RKP": ("RU(PHEN)2(DPPZ) COMPLEX", "non-polymer"),
        "H2A": (
            "(6R,7R)-3-[(ACETYLOXY)METHYL]-7-{[(6S)-6-(GLYCYLAMINO)-7-OXIDO-7-OXOHEPTANOYL]AMINO}-8-OXO-5-THIA-1-AZABICYCLO[4.2.0]OCTANE-2-CARBOXYLATE",
            "non-polymer",
        ),
        "H2B": (
            "2-AMINO-6-(1,2-DIHYDROXY-PROPYL)-7,8-DIHYDRO-6H-PTERIDIN-4-ONE",
            "non-polymer",
        ),
        "SEY": ("SELENOUREA", "non-polymer"),
        "T49": ("?", "nucleic-acid"),
        "TLF": ("2-[(3-CHLORO-2-METHYLPHENYL)AMINO]BENZOIC ACID", "non-polymer"),
        "CYS": ("CYSTEINE", "peptide"),
        "CYR": (
            "N~5~-[{[(2R)-2-AMINO-2-CARBOXYETHYL]SULFANYL}(IMINIO)METHYL]-L-ORNITHINATE",
            "peptide",
        ),
        "CYQ": ("2-AMINO-3-PHOSPHONOMETHYLSULFANYL-PROPIONIC ACID", "peptide"),
        "CYP": (
            "9R,10R-9-GLUTATHIONYL-10-HYDROXY-9,10-DIHYDRO PHENANTHRENE",
            "non-polymer",
        ),
        "CYW": ("S-[(2R)-2-HYDROXYPROPANOYL]-L-CYSTEINE", "peptide"),
        "CYV": (
            "ETHYL (4R)-4-{[(2R,5S)-5-{[N-(TERT-BUTOXYCARBONYL)-L-SERYL]AMINO}-6-METHYL-2-(3-METHYLBUT-2-EN-1-YL)-4-OXOHEPTANOYL]AMINO}-5-[(3R)-2-OXOPYRROLIDIN-3-YL]PENTANOATE",
            "non-polymer",
        ),
        "CYU": ("CALYCULIN A", "non-polymer"),
        "CYT": ("6-AMINOPYRIMIDIN-2(1H)-ONE", "non-polymer"),
        "POV": (
            "(2S)-3-(HEXADECANOYLOXY)-2-[(9Z)-OCTADEC-9-ENOYLOXY]PROPYL 2-(TRIMETHYLAMMONIO)ETHYL PHOSPHATE",
            "non-polymer",
        ),
        "CYZ": ("CYCLOTHIAZIDE", "non-polymer"),
        "CYY": ("2-DEOXYSTREPTAMINE", "non-polymer"),
        "CYX": ("(3-FORMYL-BUT-3-ENYL)-PHOSPHONIC ACID", "non-polymer"),
        "POR": ("PORPHYRIN FE(III)", "non-polymer"),
        "O09": ("(E)-2-(4-CARBAMIMIDOYLBENZYLIDENEAMINOOXY)ACETIC ACID", "non-polymer"),
        "POP": ("PYROPHOSPHATE 2-", "non-polymer"),
        "POQ": (
            "BIS(((3S,4S,5R,6R)-5-(ETHYL(PHOSPHORYLOXY))-3,4,6-TRIHYDROXY-TETRAHYDRO-2H-PYRAN-2-YL)METHYL) HYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "CYC": ("PHYCOCYANOBILIN", "non-polymer"),
        "CYB": ("PHYCOUROBILIN", "non-polymer"),
        "CYA": ("TWO OXYGEN ATOMS BOUND TO SG OF CYS", "peptide"),
        "POM": ("CIS-5-METHYL-4-OXOPROLINE", "peptide"),
        "CYG": (
            "2-AMINO-4-(AMINO-3-OXO-PROPYLSULFANYLCARBONYL)-BUTYRIC ACID",
            "peptide",
        ),
        "CYF": (
            "5-[2-(2-AMINO-2-CARBAMOYL-ETHYLSULFANYL)-ACETYLAMINO]-2-(3,6-DIHYDROXY-9,9A-DIHYDRO-3H-XANTHEN-9-YL)-BENZOIC ACID",
            "peptide",
        ),
        "CYE": ("?", "non-polymer"),
        "CYD": ("2-AMINO-6-(CYSTEIN-S-YL)-5-OXO-HEXANOIC ACID", "peptide"),
        "CYK": ("N-HEXANOYL-L-HOMOCYSTEINE", "non-polymer"),
        "CYJ": (
            "(Z)-N~6~-[(4R,5S)-5-(2-CARBOXYETHYL)-4-(CARBOXYMETHYL)PIPERIDIN-3-YLIDENE]-L-LYSINE",
            "peptide",
        ),
        "CYI": ("CYCLOHEXYL ISOCYANIDE", "non-polymer"),
        "CYH": ("CYCLOHEXANONE", "non-polymer"),
        "CYO": ("OXYGENS BOUND TO CYS SG", "non-polymer"),
        "CYN": ("CYANIDE ION", "non-polymer"),
        "CYM": ("S-METHYLCYSTEINE", "peptide"),
        "CYL": ("VALIENAMINE", "non-polymer"),
        "CY3": ("2-AMINO-3-MERCAPTO-PROPIONAMIDE", "peptide"),
        "FKI": ("5,5-DI-FLUORO-2-KETO-3-AMINOPENTANOIC ACID", "non-polymer"),
        "CY1": ("ACETAMIDOMETHYLCYSTEINE", "peptide"),
        "CY0": (
            "S-{3-[(4-ANILINOQUINAZOLIN-6-YL)AMINO]-3-OXOPROPYL}-L-CYSTEINE",
            "peptide",
        ),
        "CY7": (
            "S-{(1S)-3-{[(4Z)-4-{[3-CHLORO-4-(PYRIDIN-2-YLMETHOXY)PHENYL]IMINO}-3-CYANO-7-ETHOXY-1,4-DIHYDROQUINOLIN-6-YL]AMINO}-1-[(DIMETHYLAMINO)METHYL]-3-OXOPROPYL}-L-CYSTEINE",
            "non-polymer",
        ),
        "CY6": (
            "N-((3S,6R)-6-((S,E)-4-ETHOXYCARBONYL-1-((S)-2-OXOPYRROLIDIN-3-YL)BUT-3-EN-2-YLCARBAMOYL)-2,9-DIMETHYL-4-OXODEC-8-EN-3-YL)-5-METHYLISOXAZOLE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "FKN": (
            "3-[2-CHLORO-5-(METHYLSULFONYL)PHENYL]-6-(2,4-DIFLUOROPHENOXY)-1H-PYRAZOLO[3,4-D]PYRIMIDINE",
            "non-polymer",
        ),
        "CY4": ("S-BUTYRYL-CYSTEIN", "peptide"),
        "FKA": (
            "BENZYL-CARBAMIC ACID [8-DEETHYL-ASCOMYCIN-8-YL]ETHYL ESTER",
            "non-polymer",
        ),
        "CY9": (
            "(3S,3AR,4S,6S,6AR,7E,10S,12R,13E,15R,15AR)-3-BENZYL-6,12-DIHYDROXY-4,10,12-TRIMETHYL-5-METHYLIDENE-1,11-DIOXO-2,3,3A,4,5,6,6A,9,10,11,12,15-DODECAHYDRO-1H-CYCLOUNDECA[D]ISOINDOL-15-YL ACETATE",
            "non-polymer",
        ),
        "PO5": (
            "(2S)-1-(DIMETHYLAMINO)-3-(4-{[4-(6-MORPHOLIN-4-YLPYRAZOLO[1,5-B]PYRIDAZIN-3-YL)PYRIMIDIN-2-YL]AMINO}PHENOXY)PROPAN-2-OL",
            "non-polymer",
        ),
        "FKD": (
            "3-DEOXY-3-FLUORO-D-ERYTHRO-ALPHA-L-MANNO-NON-2-ULOPYRANOSONIC ACID",
            "non-polymer",
        ),
        "PO3": ("PHOSPHITE ION", "non-polymer"),
        "PO0": ("1-BENZYL-(R)-PROPYLAMINE", "non-polymer"),
        "PO1": (
            "(9BETA,13ALPHA,14BETA,17ALPHA)-2-METHOXYESTRA-1,3,5(10)-TRIENE-3,17-DIYL DISULFAMATE",
            "non-polymer",
        ),
        "TLD": ("4-METHYLBENZENE-1,2-DITHIOL", "non-polymer"),
        "JK2": (
            "3-{5-[(2-FLUOROPHENYL)AMINO]-1H-INDAZOL-1-YL}-N-(3,4,5-TRIMETHOXYPHENYL)BENZAMIDE",
            "non-polymer",
        ),
        "FKP": ("METHYLPIPERAZINOFORSKOLIN", "non-polymer"),
        "JK3": (
            "5-BROMO-N-(3-CHLORO-2-(4-(PROP-2-YNYL)PIPERAZIN-1-YL)PHENYL)FURAN-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "LSC": (
            "(4R,7R,18E)-4,7-DIHYDROXY-N,N,N-TRIMETHYL-10-OXO-3,5,9-TRIOXA-4-PHOSPHAHEPTACOS-18-EN-1-AMINIUM 4-OXIDE",
            "non-polymer",
        ),
        "NAD": ("NICOTINAMIDE-ADENINE-DINUCLEOTIDE", "non-polymer"),
        "QI2": (
            "(2S,4R)-2-(2-{[3-(4-FLUORO-3-METHYLPHENYL)PROPYL](METHYL)AMINO}ETHYL)-4-(4-FLUOROPHENYL)-N-HYDROXY-4-METHOXYBUTANAMIDE",
            "non-polymer",
        ),
        "TM5": ("THIO-MALTOPENTAOSE", "saccharide"),
        "LSM": (
            "DODECYL 4-O-ALPHA-D-GLUCOPYRANOSYL-1-SELENO-BETA-D-GLUCOPYRANOSIDE",
            "saccharide",
        ),
        "L3O": ("(2S,3S)-3-AMINO-2-HYDROXY-5-METHYLHEXANOIC ACID", "peptide"),
        "ZBT": ("2-[(1S)-1-METHYLPROPYL]-4,5-DIHYDRO-1,3-THIAZOLE", "non-polymer"),
        "FRT": (
            "N-(2-METHOXYETHYL)-4-({4-[2-METHYL-1-(1-METHYLETHYL)-1H-IMIDAZOL-5-YL]PYRIMIDIN-2-YL}AMINO)BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "IM8": (
            "2-[1-(4-CHLOROBENZOYL)-5-METHOXY-2-METHYL-1H-INDOL-3-YL]-N-[(1R)-1-(HYDROXYMETHYL)PROPYL]ACETAMIDE",
            "non-polymer",
        ),
        "OAD": ("?", "nucleic-acid"),
        "OAC": ("TRANS-O-HYDROXY-ALPHA-METHYL CINNAMATE", "non-polymer"),
        "TTT": (
            "5-AMINO-2-METHYL-N-[(1R)-1-NAPHTHALEN-1-YLETHYL]BENZAMIDE",
            "non-polymer",
        ),
        "DPM": (
            "3-{2-[4-(2-CARBOXY-ETHYL)-3-CARBOXYMETHYL-5H-PYRROL-2-YLMETHYL]-4-CARBOXYMETHYL-5-METHYL-2H-PYRROL-3-YL}-PROPIONIC ACID",
            "non-polymer",
        ),
        "N3M": ("N-(3-AMINOPHENYL)ACETAMIDE", "non-polymer"),
        "MQA": ("MITOMYCIN A", "non-polymer"),
        "DQ6": (
            "(2S)-2-AMINO-5-(3-METHYLPHENYL)-5,5-DIPHENYLPROPANOIC ACID",
            "non-polymer",
        ),
        "DQ7": (
            "(2R)-2-AMINO-5-(3-METHYLPHENYL)-5,5-DIPHENYLPROPANOIC ACID",
            "non-polymer",
        ),
        "DQ1": (
            "ETHYL 4-(5-{[(2,4-DIAMINOQUINAZOLIN-6-YL)METHYL]AMINO}-2-METHOXYPHENOXY)BUTANOATE",
            "non-polymer",
        ),
        "MQI": ("8-HYDROSULFONYL-2-METHYLQUINOLINE", "non-polymer"),
        "UB3": (
            "N-[(S)-({[(BENZYLOXY)CARBONYL]AMINO}METHYL)(HYDROXY)PHOSPHORYL]-L-VALYL-L-LEUCINE",
            "non-polymer",
        ),
        "DQ8": (
            "1-(3-{[(2-AMINOETHYL)SULFANYL](DIPHENYL)METHYL}PHENYL)ETHANONE",
            "non-polymer",
        ),
        "MQN": ("2-METHYLQUINOLIN-6-AMINE", "non-polymer"),
        "MQQ": (
            "N-[(1S)-1-[({(1S)-3-PHENYL-1-[2-(PHENYLSULFONYL)ETHYL]PROPYL}AMINO)CARBONYL]-3-(PHENYLSULFONYL)PROPYL]MORPHOLINE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "CCU": ("(2Z,4Z)-HEXA-2,4-DIENEDIOIC ACID", "non-polymer"),
        "MQU": (
            "7-[2-METHOXY-1-(METHOXYMETHYL)ETHYL]-7H-PYRROLO[3,2-F] QUINAZOLINE-1,3-DIAMINE",
            "non-polymer",
        ),
        "MQT": (
            "METHYL 2-O-ACETYL-3-O-(4-METHYLBENZOYL)-BETA-D-TALOPYRANOSIDE",
            "non-polymer",
        ),
        "N3A": (
            "2,4-DIHYDROXY-3-({3-[(2R,4AR,8S,8AR,9R)-9-HYDROXY-8-METHYL-3-METHYLIDENE-7-OXO-1,3,4,7,8,8A-HEXAHYDRO-2H-2,4A-ETHANONAPHTHALEN-8-YL]PROPANOYL}AMINO)BENZOIC ACID",
            "non-polymer",
        ),
        "MQY": (
            "(5S)-5-METHYL-1-(QUINOLIN-4-YLMETHYL)-3-{4-[(TRIFLUOROMETHYL)SULFONYL]PHENYL}IMIDAZOLIDINE-2,4-DIONE",
            "non-polymer",
        ),
        "OAM": (
            "4-{4-[3-(1H-IMIDAZOL-1-YL)PROPYL]-5-METHYL-4H-1,2,4-TRIAZOL-3-YL}-1-METHYL-1H-PYRAZOL-5-AMINE",
            "non-polymer",
        ),
        "DQT": (
            "METHYL 3-DEOXY-2-O-(4-METHYLBENZOYL)-3-[(4-METHYLBENZOYL)AMINO]-BETA-D-TALOPYRANOSIDE",
            "non-polymer",
        ),
        "DQU": ("2,6-DIAMINO-3H-QUINAZOLIN-4-ONE", "non-polymer"),
        "CCV": (
            "6-[3-HYDROXY-2-(HYDROXYMETHYL)PROPYL]-5-METHYL-2,4(1H,3H)-PYRIMIDINEDIONE",
            "non-polymer",
        ),
        "DQQ": ("2,5-DIDEOXY-2,5-IMINO-D-MANNITOL", "saccharide"),
        "DQR": (
            "BETA-D-FRUCTOFURANOSYL-(2->1)-BETA-D-FRUCTOFURANOSYL ALPHA-D-GLUCOPYRANOSIDE",
            "saccharide",
        ),
        "OAK": ("N-(PHENYLCARBONYL)-BETA-D-GLUCOPYRANOSYLAMINE", "non-polymer"),
        "UPG": ("URIDINE-5'-DIPHOSPHATE-GLUCOSE", "non-polymer"),
        "AW3": (
            "5-BROMO-7-{5-[(3-{[(4-TERT-BUTYLPHENYL)CARBAMOYL]AMINO}PROPYL)AMINO]-5-DEOXY-BETA-D-RIBOFURANOSYL}-7H-PYRROLO[2,3-D]PYRIMIDIN-4-AMINE",
            "non-polymer",
        ),
        "AW2": (
            "5-BROMO-7-{5-[(3-{[(4-TERT-BUTYLPHENYL)CARBAMOYL]AMINO}PROPYL)(PROPAN-2-YL)AMINO]-5-DEOXY-BETA-D-RIBOFURANOSYL}-7H-PYRROLO[2,3-D]PYRIMIDIN-4-AMINE",
            "non-polymer",
        ),
        "AW1": (
            "5'-[(3-{[(4-TERT-BUTYLPHENYL)CARBAMOYL]AMINO}PROPYL)(PROPAN-2-YL)AMINO]-5'-DEOXYADENOSINE",
            "non-polymer",
        ),
        "AW0": (
            "5'-DEOXY-5'-[(3-{[(4-METHYLPHENYL)CARBAMOYL]AMINO}PROPYL)(PROPAN-2-YL)AMINO]ADENOSINE",
            "non-polymer",
        ),
        "MQ1": ("5-(4-METHOXYPHENOXY)-2,4-QUINAZOLINEDIAMINE", "non-polymer"),
        "MQ0": (
            "(3-AMINOMETHYL-CINNOLIN-4-YL)-(3,3-DIPHENYL-ALLYLIDENE)-AMINE",
            "non-polymer",
        ),
        "DQA": ("1,3,4-TRIHYDROXY-5-OXO-CYCLOHEXANECARBOXYLIC ACID", "non-polymer"),
        "DQB": (
            "4-[(4-{[(2-AMINO-4-OXO-3,4-DIHYDROQUINAZOLIN-6-YL)METHYL]AMINO}BENZOYL)AMINO]BUTANOIC ACID",
            "non-polymer",
        ),
        "MQ9": ("MENAQUINONE-9", "non-polymer"),
        "5WB": (
            "5-[(3S)-3-METHOXY-3-(3,4,5-TRIMETHOXYPHENYL)PROP-1-YN-1-YL]-6-METHYLPYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "5WA": (
            "(1R)-3-(2,4-DIAMINO-6-METHYLPYRIMIDIN-5-YL)-1-(3,4,5-TRIMETHOXYPHENYL)PROP-2-YN-1-OL",
            "non-polymer",
        ),
        "DQO": (
            "1-(2,6-DICHLOROPHENYL)-5-(2,4-DIFLUOROPHENYL)-7-PIPERAZIN-1-YL-3,4-DIHYDROQUINAZOLIN-2(1H)-ONE",
            "non-polymer",
        ),
        "DQH": (
            "(2R,3R)-2-(3,4-DIHYDROXYPHENYL)-3,5,7-TRIHYDROXY-2,3-DIHYDRO-4H-CHROMEN-4-ONE",
            "non-polymer",
        ),
        "OAW": (
            "2-AMINO-4-METHYL-8-(1-METHYLETHYL)-6-(1H-PYRAZOL-4-YL)PTERIDIN-7(8H)-ONE",
            "non-polymer",
        ),
        "XJH": (
            "6-{[(3R,4R)-4-(3-PHENOXYPHENOXY)PYRROLIDIN-3-YL]METHYL}PYRIDIN-2-AMINE",
            "non-polymer",
        ),
        "VLT": ("(2E,4S)-4-AMINO-5-(4-HYDROXYPHENYL)PENT-2-ENOIC ACID", "peptide"),
        "NAB": ("?", "non-polymer"),
        "T3S": ("5'-DEOXY-5'-PIPERIDIN-1-YLTHYMIDINE", "non-polymer"),
        "OAS": ("O-ACETYLSERINE", "peptide"),
        "OAR": ("N-(4-AMINO-5-HYDROXY-PENTYL)-GUANIDINE", "non-polymer"),
        "Z17": (
            "(3R)-3-(7-AMINO-1-METHYL-4,5-DIOXO-1,4,5,6-TETRAHYDROPYRIMIDO[4,5-C]PYRIDAZIN-3-YL)BUTANOIC ACID",
            "non-polymer",
        ),
        "UPE": ("?", "nucleic-acid"),
        "OAP": (
            "4-(S)-[(1-OXO-7-PHENYLHEPTYL)AMINO]-5-[4-(PHENYLMETHYL)PHENYLTHIO]PENTANOIC ACID",
            "non-polymer",
        ),
        "332": (
            "(3R)-3-[(1,2,3,4-TETRAHYDROISOQUINOLIN-7-YLOXY)METHYL]-2,3-DIHYDROTHIENO[2,3-F][1,4]OXAZEPIN-5-AMINE",
            "non-polymer",
        ),
        "G86": (
            "N-(TERT-BUTOXYCARBONYL)-O-TERT-BUTYL-L-THREONYL-N-{(2R)-5-ETHOXY-5-OXO-1-[(3S)-2-OXOPYRROLIDIN-3-YL]PENTAN-2-YL}-L-PHENYLALANINAMIDE",
            "non-polymer",
        ),
        "HAI": ("CYCLOHEXYLAMMONIUM ION", "non-polymer"),
        "TT8": (
            "(2S)-2-AZANYL-4-[[(2S,3S,4R,5R)-5-[6-(METHYLAMINO)PURIN-9-YL]-3,4-BIS(OXIDANYL)OXOLAN-2-YL]METHYLSULFANYL]BUTANOIC ACID",
            "non-polymer",
        ),
        "NAC": ("3-ACETYL PYRIDINE ADENINE DINUCLEOTIDE", "non-polymer"),
        "T3P": ("THYMIDINE-3'-PHOSPHATE", "nucleic-acid"),
        "Z46": (
            "(5R)-5-PROPYL-4,5,6,7-TETRAHYDRO-1H-INDAZOLE-3-CARBOHYDRAZIDE",
            "non-polymer",
        ),
        "2T5": (
            "[(8H-INDENO[1,2-D][1,3]THIAZOL-4-YLOXY)METHYL]PHOSPHONIC ACID",
            "non-polymer",
        ),
        "PCY": ("PACTAMYCIN", "non-polymer"),
        "B0D": (
            "(2R,3S,4R,5R,6R)-2,6-DIFLUORO-2-(HYDROXYMETHYL)TETRAHYDRO-2H-PYRAN-3,4,5-TRIOL",
            "saccharide",
        ),
        "2T4": (
            "{[(2-AMINO-8H-INDENO[1,2-D][1,3]THIAZOL-4-YL)OXY]METHYL}PHOSPHONIC ACID",
            "non-polymer",
        ),
        "38M": (
            "3-(3-CHLOROPHENYL)-2-({(1S)-1-[(6S)-2,8-DIAZASPIRO[5.5]UNDEC-2-YLCARBONYL]PENTYL}SULFANYL)QUINAZOLIN-4(3H)-ONE",
            "non-polymer",
        ),
        "38P": (
            "4-{[5-(CYCLOPROPYLCARBAMOYL)-2-METHYLPHENYL]AMINO}-5-METHYL-N-PROPYLPYRROLO[2,1-F][1,2,4]TRIAZINE-6-CARBOXAMIDE",
            "non-polymer",
        ),
        "38R": (
            "[4-({4-[(3-CYCLOPENTYL-1H-PYRAZOL-5-YL)AMINO]PYRIMIDIN-2-YL}AMINO)PHENYL]ACETONITRILE",
            "non-polymer",
        ),
        "2T6": (
            "{[(7-CARBAMOYL-8H-INDENO[1,2-D][1,3]THIAZOL-4-YL)OXY]METHYL}PHOSPHONIC ACID",
            "non-polymer",
        ),
        "AT2": (
            "ETHYL 4-[(4-METHYLPYRIDIN-2-YL)AMINO]PIPERIDINE-1-CARBOXYLATE",
            "non-polymer",
        ),
        "X1H": (
            "[6-HYDROXY-2-(4-HYDROXYPHENYL)-1-BENZOTHIOPHEN-3-YL](4-METHOXYPHENYL)METHANONE",
            "non-polymer",
        ),
        "AT3": ("ACETYLTHIOCHOLINE", "non-polymer"),
        "C": ("CYTIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "GSI": (
            "2-(5-CHLOROTHIOPHEN-2-YL)-N-[(3S)-1-(4-{2-[(DIMETHYLAMINO)METHYL]-1H-IMIDAZOL-1-YL}-2-FLUOROPHENYL)-2-OXOPYRROLIDIN-3-YL]ETHANESULFONAMIDE",
            "non-polymer",
        ),
        "0U": ("L-URIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "380": (
            "(2R)-2-({4-[AMINO(IMINO)METHYL]PHENYL}AMINO)-N-BENZYL-2-(3,4-DIMETHOXYPHENYL)ACETAMIDE",
            "non-polymer",
        ),
        "0C": ("L-CYTIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "382": (
            "(R)-2-(2-(1H-1,2,4-TRIAZOL-1-YL)BENZYL)-N-(2,2-DIFLUORO-2-(PIPERIDIN-2-YL)ETHYL)OXAZOLO[4,5-C]PYRIDIN-4-AMINE",
            "non-polymer",
        ),
        "383": (
            "[2-AMINO-3-(4-HYDROXY-PHENYL)-PROPIONYLAMINO]- (2,4,5,8-TETRAHYDROXY-7-OXA-2-AZA-BICYCLO[3.2.1]OCT-3-YL)- ACETIC ACID",
            "non-polymer",
        ),
        "0G": ("L-GUANOSINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "B07": ("2,6 DIMETHOXYBENZAMIDOBORONIC ACID", "non-polymer"),
        "387": (
            "1-{3-[(3,5-DICHLOROBENZYL)AMINO]PROPYL}-3-(3-HYDROXYPHENYL)UREA",
            "non-polymer",
        ),
        "388": (
            "(2-{[(4-BROMO-2-FLUOROBENZYL)AMINO]CARBONYL}-5-CHLOROPHENOXY)ACETIC ACID",
            "non-polymer",
        ),
        "389": ("L-ALANYL-L-VALYL-N-(2,2-DIPHENYLETHYL)-L-PROLINAMIDE", "non-polymer"),
        "SXX": ("SINAPINATE", "non-polymer"),
        "KAH": ("5-AMINO-4-HYDROXYHEXANOIC ACID", "non-polymer"),
        "3UB": (
            "PROPAN-2-YL HYDROGEN (S)-[(1R)-1-{[(BENZYLOXY)CARBONYL]AMINO}ETHYL]PHOSPHONATE",
            "non-polymer",
        ),
        "3UN": (
            "(4-{4-[(TERT-BUTOXYCARBONYL)AMINO]-2,2-BIS(ETHOXYCARBONYL)BUTYL}PHENYL)SULFAMIC ACID",
            "non-polymer",
        ),
        "S1Y": (
            "[(2S,3R)-1-{[(4S)-1-(2-AMINOETHYL)AZEPAN-4-YL]CARBAMOYL}-2-FORMYLPYRROLIDIN-3-YL]SULFAMIC ACID",
            "non-polymer",
        ),
        "SDC": ("S-[2-(AMINOSULFONYL)ETHYL]-D-CYSTEINE", "non-polymer"),
        "LS3": (
            "3-{[(2,2-DIOXIDO-1,3-DIHYDRO-2-BENZOTHIEN-5-YL)AMINO]METHYLENE}-5-(1,3-OXAZOL-5-YL)-1,3-DIHYDRO-2H-INDOL-2-ONE",
            "non-polymer",
        ),
        "1C3": ("1-(3,4,5-TRIMETHOXYBENZYL)-L-PROLINE", "peptide"),
        "1C2": (
            "3(S)-METHYLCARBAMOYL-7-SULFOAMINO-3,4-DIHYDRO-1H-ISOQUINOLINE-2-CARBOXYLIC ACID TERT-BUTYL ESTER",
            "non-polymer",
        ),
        "1C5": (
            "[[O-PHOSPHONO-N-ACETYL-TYROSINYL]-GLUTAMYL-3[CYCLOHEXYLMETHYL]ALANINYL]-AMINE",
            "non-polymer",
        ),
        "1C4": ("?", "non-polymer"),
        "I20": (
            "(2S)-2-({(2S)-3-[(R)-[(1R)-1-({(4S)-4-AMINO-5-[(2S)-2-CYANOPYRROLIDIN-1-YL]-5-OXOPENTANOYL}AMINO)ETHYL](HYDROXY)PHOSPHORYL]-2-BENZYLPROPANOYL}AMINO)PROPANOIC ACID",
            "non-polymer",
        ),
        "I21": (
            "2-[(5R)-5-AMINO-5-METHYL-4,16-DIOXO-14-PHENYL-3-OXA-15-AZATRICYCLO[15.3.1.1~7,11~]DOCOSA-1(21),7(22),8,10,12,14,17,19-OCTAEN-19-YL]BENZONITRILE",
            "non-polymer",
        ),
        "1C9": (
            "(2E)-N-{4-[(3-CHLORO-4-FLUOROPHENYL)AMINO]-7-METHOXYQUINAZOLIN-6-YL}-4-(PIPERIDIN-1-YL)BUT-2-ENAMIDE",
            "non-polymer",
        ),
        "I23": (
            "N-CYCLOHEPTYL-1-[(2,4,6-TRIMETHYLPHENYL)SULFONYL]PIPERIDINE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "I24": (
            "2-BUTOXY-9-(2,6-DIFLUOROBENZYL)-N-(2-MORPHOLIN-4-YLETHYL)-9H-PURIN-6-AMINE",
            "non-polymer",
        ),
        "I25": (
            "(2R)-({4-[AMINO(IMINO)METHYL]PHENYL}AMINO){3-[3-(DIMETHYLAMINO)-2,2-DIMETHYLPROPOXY]-5-ETHYLPHENYL}ACETIC ACID",
            "non-polymer",
        ),
        "I26": (
            "(2R)-[(4-CARBAMIMIDOYLPHENYL)AMINO]{3-[3-(DIMETHYLAMINO)-2,2-DIMETHYLPROPOXY]-5-ETHYL-2-FLUOROPHENYL}ETHANOIC ACID",
            "non-polymer",
        ),
        "X66": ("2-(4,6-DIAMINO-1,3,5-TRIAZIN-2-YL)BENZENE-1,4-DIOL", "non-polymer"),
        "KAN": ("KANAMYCIN A", "non-polymer"),
        "16O": (
            "N-({(2E)-2-[(4-BROMOPHENYL)METHYLIDENE]HYDRAZINO}CARBONOTHIOYL)-BETA-D-GLUCOPYRANOSYLAMINE",
            "non-polymer",
        ),
        "16H": (
            "N-[(1R,3R)-3-HYDROXY-1-(HYDROXYMETHYL)-3-PHENYLPROPYL]HEXADECANAMIDE",
            "non-polymer",
        ),
        "16D": ("HEXANE-1,6-DIAMINE", "non-polymer"),
        "16E": ("N-(16-CARBOXYHEXADECANOYL)-L-GLUTAMIC ACID", "non-polymer"),
        "16F": (
            "N-({(2Z)-2-[(3-HYDROXYPHENYL)METHYLIDENE]HYDRAZINO}CARBONOTHIOYL)-BETA-D-GLUCOPYRANOSYLAMINE",
            "non-polymer",
        ),
        "16G": ("N-ACETYL-D-GLUCOSAMINE-6-PHOSPHATE", "saccharide"),
        "16A": ("CETYL-TRIMETHYL-AMMONIUM", "non-polymer"),
        "16B": ("?", "non-polymer"),
        "16C": (
            "N-((E,2S,3R)-1,3-DIHYDROXYOCTADEC-4-EN-2-YL)PALMITAMIDE",
            "non-polymer",
        ),
        "SWA": ("1S-8AB-OCTAHYDRO-INDOLIZIDINE-1A,2A,8B-TRIOL", "non-polymer"),
        "0S8": (
            "5-BROMO-N~4~-[3-(3-METHOXYPROPYL)-1H-PYRAZOL-5-YL]-N~2~-[(3-METHYL-1,2-OXAZOL-5-YL)METHYL]PYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "0S9": (
            "5-BROMO-N~2~-[(3-METHYL-1,2-OXAZOL-5-YL)METHYL]-N~4~-[3-(2-PHENYLETHYL)-1H-PYRAZOL-5-YL]PYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "16X": (
            "4-(2-AMINO-4-METHYL-1,3-THIAZOL-5-YL)-N-(3-DIOXAZIRIDIN-3-YLPHENYL)PYRIMIDIN-2-AMINE",
            "non-polymer",
        ),
        "16Y": ("16-HYDROXYHEXADECANOIC ACID", "non-polymer"),
        "0S2": (
            "5-(3,3-DIMETHYLBUT-1-YN-1-YL)-3-{[(TRANS-4-METHYLCYCLOHEXYL)CARBONYL](PROPAN-2-YL)AMINO}THIOPHENE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "0S3": (
            "5-(3,3-DIMETHYLBUT-1-YN-1-YL)-3-{2,2-DIMETHYL-1-[(TRANS-4-METHYLCYCLOHEXYL)CARBONYL]HYDRAZINYL}THIOPHENE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "0S0": (
            "4-[(E)-(6-HYDROXY-2-OXO-1,2-DIHYDROPYRIDIN-3-YL)DIAZENYL]BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "0S1": (
            "N-((1R,2S)-2-ALLYL-4-OXOCYCLOBUTYL)-4-METHYLBENZENESULFONAMIDE, BOUND FORM",
            "non-polymer",
        ),
        "0S6": (
            "METHYL [(6S)-4-(4-CHLOROPHENYL)-2,3,9-TRIMETHYL-6H-THIENO[3,2-F][1,2,4]TRIAZOLO[4,3-A][1,4]DIAZEPIN-6-YL]ACETATE",
            "non-polymer",
        ),
        "0S7": (
            "5-BROMO-N~4~-(3-METHYL-1H-PYRAZOL-5-YL)-N~2~-[2-(PYRIDIN-3-YL)ETHYL]PYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "0S4": (
            "N-HYDROXY-2-{4-[3-(4-{[(2R)-2-HYDROXY-3,3-DIMETHYLBUTYL]OXY}-3-METHYLPHENYL)PENTAN-3-YL]-2-METHYLPHENOXY}ACETAMIDE",
            "non-polymer",
        ),
        "16S": ("2-(3,4-DIHYDROXYPHENYL)-4H-CHROMEN-4-ONE", "non-polymer"),
        "0SJ": (
            "(1R,5S)-3-(4-CHLOROBENZOYL)-1,2,3,4,5,6-HEXAHYDRO-8H-1,5-METHANOPYRIDO[1,2-A][1,5]DIAZOCIN-8-ONE",
            "non-polymer",
        ),
        "0SK": (
            "8-(HYDROXYAMINO)-2-[(2S)-2-METHYL-1,4-DIOXA-8-AZASPIRO[4.5]DEC-8-YL]-6-(TRIFLUOROMETHYL)-4H-1,3-BENZOTHIAZIN-4-ONE",
            "non-polymer",
        ),
        "0SH": (
            "N-[(2S,3S,4R)-1-(ALPHA-D-GALACTOPYRANOSYLOXY)-3,4-DIHYDROXYOCTADECAN-2-YL]HEXADECANAMIDE",
            "non-polymer",
        ),
        "CH": ("?", "nucleic-acid"),
        "0SN": (
            "{4-[4-({3-[(2-METHYL-1,3-BENZOTHIAZOL-6-YL)AMINO]-3-OXOPROPYL}AMINO)-4-OXOBUTYL]BENZYL}PROPANEDIOIC ACID",
            "non-polymer",
        ),
        "CN": ("CYANIDE", "non-polymer"),
        "0SL": (
            "(3R)-1-(NAPHTHALEN-2-YLSULFONYL)PIPERIDINE-3-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "0SM": (
            "TRIMETHYL-[2-[[(2S,3S)-2-(OCTADECANOYLAMINO)-3-OXIDANYL-BUTOXY]-OXIDANYL-PHOSPHORYL]OXYETHYL]AZANIUM",
            "non-polymer",
        ),
        "0SB": (
            "N-{6-[3-(PIPERAZIN-1-YL)PHENYL]PYRIDIN-2-YL}-4-(TRIFLUOROMETHYL)PYRIDIN-2-AMINE",
            "non-polymer",
        ),
        "CA": ("CALCIUM ION", "non-polymer"),
        "0SA": (
            "UNDECYL 4-O-ALPHA-D-GLUCOPYRANOSYL-1-THIO-BETA-D-GLUCOPYRANOSIDE",
            "non-polymer",
        ),
        "0SF": (
            "(2R)-3-{[4-(4-CHLORO-3-ETHYLPHENOXY)PYRIMIDIN-2-YL][3-(1,1,2,2-TETRAFLUOROETHOXY)BENZYL]AMINO}-1,1,1-TRIFLUOROPROPAN-2-OL",
            "non-polymer",
        ),
        "0SG": (
            "N-{(2S,3R,4E)-1-[(3,6-DI-O-SULFO-BETA-D-GALACTOPYRANOSYL)OXY]-3-HYDROXYOCTADEC-4-EN-2-YL}DODECANAMIDE",
            "non-polymer",
        ),
        "CE": ("CERIUM (III) ION", "non-polymer"),
        "CD": ("CADMIUM ION", "non-polymer"),
        "0SZ": (
            "3-(3,4-DIHYDROISOQUINOLIN-2(1H)-YLSULFONYL)BENZOIC ACID",
            "non-polymer",
        ),
        "0SX": (
            "8-CHLORO-2-[(2S)-PYRROLIDIN-2-YL][1]BENZOFURO[3,2-D]PYRIMIDIN-4(3H)-ONE",
            "non-polymer",
        ),
        "0SY": (
            "2-(PYRIDIN-4-YL)-1,5,6,7-TETRAHYDRO-4H-PYRROLO[3,2-C]PYRIDIN-4-ONE",
            "non-polymer",
        ),
        "168": ("PANTOTHENYL-AMINOETHANOL-ACETATE PIVALIC ACID", "non-polymer"),
        "169": (
            "1-(4-TERT-BUTYLCARBAMOYL-PIPERAZINE-1-CARBONYL)-3-(3-GUANIDINO-PROPYL)-4-OXO-AZETIDINE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "PIQ": ("2-AMINO-1-METHYL-6-PHENYLIMIDAZO[4,5-B]PYRIDINE", "non-polymer"),
        "164": (
            "2-(3-CARBOXYPROPIONYL)-6-HYDROXY-CYCLOHEXA-2,4-DIENE CARBOXYLIC ACID",
            "non-polymer",
        ),
        "165": (
            "N-(BENZYLSULFONYL)-3-CYCLOHEXYLALANYL-N-(2-AMINO-1,3-BENZOTHIAZOL-6-YL)PROLINAMIDE",
            "peptide",
        ),
        "0SP": ("?", "nucleic-acid"),
        "167": (
            "6-CARBAMIMIDOYL-2-[2-HYDROXY-5-(3-METHOXY-PHENYL)-INDAN-1-YL]-HEXANOIC ACID",
            "non-polymer",
        ),
        "160": (
            "3-(3-{2-[(5-METHANESULFONYL-THIOPHENE-2-CARBONYL)-AMINO]-ETHYLDISULFANYLMETHYL}- BENZENESULFONYLAMINO)-4-OXO-PENTANOIC ACID",
            "non-polymer",
        ),
        "161": (
            "5-[4-(1-CARBOXYMETHYL-2-OXO-PROPYLCARBAMOYL)-BENZYLSULFAMOYL]-2-HYDROXY-BENZOIC ACID",
            "non-polymer",
        ),
        "162": (
            "(2-{2-[(5-CARBAMIMIDOYL-1-METHYL-1H-PYRROL-3-YLMETHYL)-CARBAMOYL]-PYRROL-1-YL} -1-CYCLOHEXYLMETHYL-2-OXO-ETHYLAMINO)-ACETIC ACID",
            "peptide",
        ),
        "163": (
            "(2-{2-[(5-CARBAMIMIDOYL-1-METHYL-1H-PYRROL-2-YLMETHYL)-CARBAMOYL]-PYRROL-1-YL}- 1-CYCLOHEXYLMETHYL-2-OXO-ETHYLAMINO)-ACETIC ACID",
            "peptide",
        ),
        "1CS": (
            "1-(2-CHLOROPHENYLSULFONYL)-3-(4-METHOXY-6-METHYL-L,3,5-TRIAZIN-2-YL)UREA",
            "non-polymer",
        ),
        "1CR": (
            "POPO-2-[4-(2-(4-(METHOXY)-1H-1,2,3-TRIAZOL-1-YL)ETHYL)BENZENESULFONAMIDE]-7,12-BIS-[3-(4-(METHOXY)-1H-1,2,3-TRIAZOL-1-YL)PROPANOIC ACID]-CRYPTOPHANE-A",
            "non-polymer",
        ),
        "U17": (
            "METHYL N-[(2S,3R)-3-AMINO-2-HYDROXYHEPTANOYL]-L-SERYL-L-LEUCINATE",
            "non-polymer",
        ),
        "1CP": ("COPROPORPHYRIN I", "non-polymer"),
        "7DE": (
            "3,5-DIMETHYL-1-(3-NITROPHENYL)-1H-PYRAZOLE-4-CARBOXYLIC ACID ETHYL ESTER",
            "non-polymer",
        ),
        "1CU": ("COPPER ION, 1 WATER COORDINATED", "non-polymer"),
        "7DG": ("7-DEAZAGUANINE", "non-polymer"),
        "I2P": ("D-MYO-INOSITOL-2,4,5-TRIPHOSPHATE", "non-polymer"),
        "1CY": (
            "1-(4-CHLOROPHENYL)-6,6-DIMETHYL-1,6-DIHYDRO-1,3,5-TRIAZINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "1CX": (
            "(5S)-4,5-DIFLUORO-6-[(2-FLUORO-4-IODOPHENYL)IMINO]-N-(2-HYDROXYETHOXY)CYCLOHEXA-1,3-DIENE-1-CARBOXAMIDE",
            "non-polymer",
        ),
        "1CC": ("?", "nucleic-acid"),
        "1CB": ("CYCLOPENTANAMINE", "non-polymer"),
        "1CA": ("DESOXYCORTICOSTERONE", "non-polymer"),
        "1CG": (
            "[(4-{(2S)-2-(1,3-BENZOXAZOL-2-YL)-2-[(4-FLUOROPHENYL)SULFAMOYL]ETHYL}PHENYL)AMINO](OXO)ACETIC ACID",
            "non-polymer",
        ),
        "I2M": ("3-METHYL-L-ALLOISOLEUCINE", "peptide"),
        "1CE": (
            "3-(1H-TETRAZOL-5-YLMETHYL)-5,6,7,8-TETRAHYDRO[1]BENZOTHIENO[2,3-D]PYRIMIDIN-4(3H)-ONE",
            "non-polymer",
        ),
        "1CD": (
            "(13R,15S)-13-METHYL-16-OXA-8,9,12,22,24-PENTAAZAHEXACYCLO[15.6.2.16,9.1,12,15.0,2,7.0,21,25]HEPTACOSA-1(24),2,4,6,17(25),18,20-HEPTAENE-23,26-DIONE",
            "non-polymer",
        ),
        "X6A": (
            "2-{[(2-AMINOPYRIMIDIN-5-YL)METHYL]AMINO}-4-CHLORO-5-NITROBENZAMIDE",
            "non-polymer",
        ),
        "1CJ": ("(R)-ETHOXY(PYRAZOLO[1,5-A]PYRIDIN-3-YL)METHANOL", "non-polymer"),
        "1CI": ("1-(4-CHLOROPHENYL)-1H-IMIDAZOLE", "non-polymer"),
        "1CH": (
            "N-(3,3-DIMETHYL-3,4-DIHYDROISOQUINOLIN-1-YL)-L-PHENYLALANINE",
            "non-polymer",
        ),
        "1CN": (
            "5-{[(4-AMINO-3-CHLORO-5-FLUOROPHENYL)SULFONYL]AMINO}-1,3,4-THIADIAZOLE-2-SULFONAMIDE",
            "non-polymer",
        ),
        "1CM": ("(2R)-2-PHENYL-N-PYRIDIN-4-YLBUTANAMIDE", "non-polymer"),
        "1CL": ("FE(8)-S(7) CLUSTER, OXIDIZED", "non-polymer"),
        "R3W": ("CHLOROACETIC ACID", "non-polymer"),
        "NAK": ("AMINO-ACRYLATE", "non-polymer"),
        "VSP": (
            "N-(3-{[4-AMINO-1-(PROPAN-2-YL)-1H-PYRAZOLO[3,4-D]PYRIMIDIN-3-YL]METHYL}PHENYL)ETHANESULFONAMIDE",
            "non-polymer",
        ),
        "G8M": (
            "(1S,2R)-2-[(S)-AMINO(CARBOXY)METHYL]CYCLOBUTANECARBOXYLIC ACID",
            "peptide",
        ),
        "80A": (
            "(R)-PYRIDIN-4-YL[4-(2-PYRROLIDIN-1-YLETHOXY)PHENYL]METHANOL",
            "non-polymer",
        ),
        "80G": ("1-(4-THIOPHEN-2-YLPHENYL)METHANAMINE", "non-polymer"),
        "PCQ": ("3,5,3',5'-TETRACHLORO-BIPHENYL-4,4'-DIOL", "non-polymer"),
        "KKR": (
            "1-[4-[[2-(1H-INDAZOL-4-YL)-4-MORPHOLIN-4-YL-THIENO[3,2-D]PYRIMIDIN-6-YL]METHYL]PIPERAZIN-1-YL]-6-METHYL-HEPT-5-ENE-1,4-DIONE",
            "non-polymer",
        ),
        "NAH": ("NICOTINAMIDE-ADENINE-DINUCLEOTIDE", "non-polymer"),
        "MOH": ("METHANOL", "non-polymer"),
        "OCM": ("COBALT ION, 3 WATERS COORDINATED", "non-polymer"),
        "T7Z": (
            "4-(4-METHYL-2-METHYLIMINO-3H-1,3-THIAZOL-5-YL)-2-[(4-METHYL-3-MORPHOLIN-4-YLSULFONYL-PHENYL)AMINO]PYRIMIDINE-5-CARBONITRILE",
            "non-polymer",
        ),
        "T1C": ("TIGECYCLINE", "non-polymer"),
        "4D9": (
            "3-(1H-BENZIMIDAZOL-2-YL)-N-(1-BENZOTHIOPHEN-2-YLCARBONYL)-D-ALANINE",
            "non-polymer",
        ),
        "SER": ("SERINE", "peptide"),
        "4D7": (
            "3-(1H-BENZIMIDAZOL-2-YL)-N-[(2-METHYLFURAN-3-YL)CARBONYL]-D-ALANINE",
            "non-polymer",
        ),
        "UD6": ("UDP-2-KETOGLC", "non-polymer"),
        "NAI": ("1,4-DIHYDRONICOTINAMIDE ADENINE DINUCLEOTIDE", "non-polymer"),
        "VA": ("(Z)-OCTADEC-11-ENYL ACETATE", "non-polymer"),
        "4DI": ("4-(DIMETHYLAMINO)BUTYL IMIDOTHIOCARBAMATE", "non-polymer"),
        "4DH": (
            "3-(1H-BENZIMIDAZOL-2-YL)-N-[(1-METHYL-3-PHENYL-1H-PYRAZOL-5-YL)CARBONYL]-D-ALANINE",
            "non-polymer",
        ),
        "MOO": ("MOLYBDATE ION", "non-polymer"),
        "4DM": (
            "(4AS,12AS)-3,10,11,12A-TETRAHYDROXY-6-METHYL-1,12-DIOXO-1,4,4A,5,12,12A-HEXAHYDROTETRACENE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "4DC": (
            "(2R)-3-CYCLOPENTYL-2-[4-(METHYLSULFONYL)PHENYL]-N-(1,3-THIAZOL-2-YL)PROPANAMIDE",
            "non-polymer",
        ),
        "4DB": (
            "(2S)-2-AMINO-4-[5-(DIMETHYLAMINO)-1,3-DIOXO-1,3-DIHYDRO-2H-ISOINDOL-2-YL]BUTANOIC ACID",
            "peptide",
        ),
        "4DG": (
            "2-[(2-AMINO-6-OXO-1,6-DIHYDRO-9H-PURIN-9-YL)METHOXY]ETHYL DIHYDROGEN PHOSPHATE",
            "nucleic-acid",
        ),
        "20S": (
            "BETA-D-FRUCTOFURANOSYL 6-O-OCTANOYL-ALPHA-D-GLUCOPYRANOSIDE",
            "non-polymer",
        ),
        "4DE": (
            "1-(4-METHOXYPHENYL)-3,5-DIMETHYL-1H-PYRAZOLE-4-CARBOXYLIC ACID ETHYL ESTER",
            "non-polymer",
        ),
        "IMI": ("2-IMINOBIOTIN", "non-polymer"),
        "VR": ("2-METHYLPROPYL HYDROGEN (R)-METHYLPHOSPHONATE", "non-polymer"),
        "4DX": ("(1S,2S)-2-METHYLCYCLOHEXANOL", "non-polymer"),
        "VX": ("O-ETHYLMETHYLPHOSPHONIC ACID ESTER GROUP", "non-polymer"),
        "4DP": (
            "3-[5-(DIMETHYLAMINO)-1,3-DIOXO-1,3-DIHYDRO-2H-ISOINDOL-2-YL]-L-ALANINE",
            "peptide",
        ),
        "PFR": (
            "7-METHOXY-4-[(3S)-3-PHENYLPIPERIDIN-1-YL]-6-[2-(PYRIDIN-2-YL)ETHOXY]QUINAZOLINE",
            "non-polymer",
        ),
        "TD3": (
            "2-[HYDROXY(4-HYDROXYBUTYL)AMINO]-2-OXOETHYL DIHYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "4DU": (
            "1-(2-DEOXY-5-O-PHOSPHONO-BETA-D-ERYTHRO-PENTOFURANOSYL)-1H-IMIDAZO[4,5-C]PYRIDIN-4-AMINE",
            "nucleic-acid",
        ),
        "WJ1": ("(S)-2-CHLORO-3-PHENYLPROPANOIC ACID", "non-polymer"),
        "MON": (
            "4-(2-{[4-{[3-(4-CHLOROPHENYL)PROPYL]SULFANYL}-6-(1-PIPERAZINYL)-1,3,5-TRIAZIN-2-YL]AMINO}ETHYL)PHENOL",
            "non-polymer",
        ),
        "PI9": ("?", "non-polymer"),
        "IMH": (
            "1,4-DIDEOXY-4-AZA-1-(S)-(9-DEAZAHYPOXANTHIN-9-YL)-D-RIBITOL",
            "non-polymer",
        ),
        "G84": (
            "O-TERT-BUTYL-N-[(9H-FLUOREN-9-YLMETHOXY)CARBONYL]-L-THREONYL-N-{(2R)-5-ETHOXY-5-OXO-1-[(3S)-2-OXOPYRROLIDIN-3-YL]PENTAN-2-YL}-L-PHENYLALANINAMIDE",
            "non-polymer",
        ),
        "G85": (
            "N-[(BENZYLOXY)CARBONYL]-O-TERT-BUTYL-L-SERYL-N-{(2R)-5-ETHOXY-5-OXO-1-[(3S)-2-OXOPYRROLIDIN-3-YL]PENTAN-2-YL}-L-PHENYLALANINAMIDE",
            "peptide",
        ),
        "809": (
            "4-(2-CHLOROPHENYL)-8-(2-HYDROXYETHYL)-6-METHYLPYRROLO[3,4-E]INDOLE-1,3(2H,6H)-DIONE",
            "non-polymer",
        ),
        "808": (
            "N-[(2S)-5-{[5-(TRIFLUOROMETHYL)FURAN-2-YL]METHYL}-2,3-DIHYDRO-1H-INDEN-2-YL]PROPANE-2-SULFONAMIDE",
            "non-polymer",
        ),
        "G80": (
            "(3AS,4R,5S,6S,8R,9R,9AR,10R)-5-HYDROXY-4,6,9,10-TETRAMETHYL-1-OXO-6-VINYLDECAHYDRO-3A,9-PROPANOCYCLOPENTA[8]ANNULEN-8-YL (PIPERIDIN-4-YLTHIO)ACETATE",
            "non-polymer",
        ),
        "G81": (
            "ETHYL (4R)-4-{[N-(TERT-BUTOXYCARBONYL)-L-PHENYLALANYL]AMINO}-5-[(3S)-2-OXOPYRROLIDIN-3-YL]PENTANOATE",
            "non-polymer",
        ),
        "G82": (
            "ETHYL (5S,8S,11R)-8-BENZYL-5-(3-TERT-BUTOXY-3-OXOPROPYL)-3,6,9-TRIOXO-11-{[(3S)-2-OXOPYRROLIDIN-3-YL]METHYL}-1-PHENYL-2-OXA-4,7,10-TRIAZATETRADECAN-14-OATE",
            "non-polymer",
        ),
        "G83": (
            "ETHYL (5S,8S,11R)-8-BENZYL-5-(2-TERT-BUTOXY-2-OXOETHYL)-3,6,9-TRIOXO-11-{[(3S)-2-OXOPYRROLIDIN-3-YL]METHYL}-1-PHENYL-2-OXA-4,7,10-TRIAZATETRADECAN-14-OATE",
            "non-polymer",
        ),
        "803": ("LOVASTATIN", "non-polymer"),
        "802": (
            "2-{5-[AMINO(IMINIO)METHYL]-1H-BENZIMIDAZOL-2-YL}-4-FLUOROBENZENOLATE",
            "non-polymer",
        ),
        "801": (
            "2-{5-[AMINO(IMINIO)METHYL]-1H-BENZIMIDAZOL-2-YL}PYRIDIN-3-OLATE",
            "non-polymer",
        ),
        "G88": ("(2S)-2-(PHOSPHONOMETHYL)PENTANEDIOIC ACID", "non-polymer"),
        "806": (
            "7-[[2-[[1-(1-IMINOETHYL)PIPERIDIN-4-YL]OXY]-9H-CARBOZOL-9-YL]METHYL]NAPHTHALENE-2-CARBOXIMIDAMID",
            "non-polymer",
        ),
        "804": (
            "6-(1,4,5,6-TETRAHYDROPYRIMIDIN-2-YL)-2-[4'-(1,4,5,6-TETRAHYDROPYRIMIDIN-2-YL)BIPHENYL-4-YL]-1H-INDOLE",
            "non-polymer",
        ),
        "ZYF": ("?", "non-polymer"),
        "ZYG": ("?", "non-polymer"),
        "F69": ("(2-THIOPHEN-2-YLPHENYL)METHANOL", "non-polymer"),
        "ZYE": (
            "N-{(1S,2R)-1-BENZYL-2-HYDROXY-3-[(3-METHOXYBENZYL)AMINO]PROPYL}-1-ETHYL-4-(2-OXOPYRROLIDIN-1-YL)-1H-INDOLE-6-CARBOXAMIDE",
            "non-polymer",
        ),
        "ZYC": ("ETHYL (2E)-3-(4-HYDROXY-3-METHOXYPHENYL)PROP-2-ENOATE", "non-polymer"),
        "ZYA": ("BENZOYL-TYROSINE-ALANINE-FLUORO-METHYL KETONE", "peptide"),
        "ZYN": (
            "6-(3-CHLORO-4-FLUOROBENZYL)-4-HYDROXY-N,N-DIMETHYL-2-(1-METHYLETHYL)-3,5-DIOXO-2,3,5,6,7,8-HEXAHYDRO-2,6-NAPHTHYRIDINE-1-CARBOXAMIDE",
            "non-polymer",
        ),
        "ZYO": (
            "(6S)-2-(3-CHLORO-4-FLUOROBENZYL)-8-ETHYL-10-HYDROXY-6-METHYL-4-(5-METHYL-1,1-DIOXIDO-1,2,5-THIADIAZOLIDIN-2-YL)-7,8-DIHYDROPYRAZINO[1',2':1,5]PYRROLO[2,3-D]PYRIDAZINE-1,9(2H,6H)-DIONE",
            "non-polymer",
        ),
        "ZYL": (
            "1-[3-(HEXYLOXY)BENZYL]-4-HYDROXY-2-OXO-1,2-DIHYDROPYRIDINE-3-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "ZYM": (
            "2,4-DIOXO-4-{[3-(3-PHENOXYPHENYL)PROPYL]AMINO}BUTANOIC ACID",
            "non-polymer",
        ),
        "PCU": ("CU(II)MESO(4-N-TETRAMETHYLPYRIDYL)PORPHYRIN", "non-polymer"),
        "ZYK": (
            "(4R)-4-({[(3,4-DIMETHOXYPHENYL)CARBONYL]AMINO}OXY)-L-PROLINE",
            "peptide",
        ),
        "ZYH": (
            "(2-CYCLOHEXYL-1,3-BENZOXAZOL-6-YL){3-[4-(PYRIMIDIN-2-YL)PIPERAZIN-1-YL]AZETIDIN-1-YL}METHANONE",
            "non-polymer",
        ),
        "ZYI": (
            "N-(5-CHLORO-1,3-BENZOXAZOL-2-YL)-2-CYCLOPENTYLACETAMIDE",
            "non-polymer",
        ),
        "ZYV": ("1-[(2S)-4-(7H-PURIN-6-YL)MORPHOLIN-2-YL]METHANAMINE", "non-polymer"),
        "ZYW": (
            "1-[(2S)-4-(5-BROMO-1H-PYRAZOLO[3,4-B]PYRIDIN-4-YL)MORPHOLIN-2-YL]METHANAMINE",
            "non-polymer",
        ),
        "ZYT": (
            "2-(METHYLSULFANYL)-5-(THIOPHEN-2-YLMETHYL)-1H-IMIDAZOL-4-OL",
            "non-polymer",
        ),
        "ZYU": ("6-MORPHOLIN-4-YL-9H-PURINE", "non-polymer"),
        "ZYR": (
            "5,6,7,8-TETRAHYDRO[1]BENZOTHIENO[2,3-D]PYRIMIDIN-4(3H)-ONE",
            "non-polymer",
        ),
        "ZYS": (
            "[4-AMINO-2-(TERT-BUTYLAMINO)-1,3-THIAZOL-5-YL](PHENYL)METHANONE",
            "non-polymer",
        ),
        "ZYP": (
            "5-(1,1-DIOXIDO-1,2-THIAZINAN-2-YL)-N-(4-FLUOROBENZYL)-8-HYDROXY-1,6-NAPHTHYRIDINE-7-CARBOXAMIDE",
            "non-polymer",
        ),
        "ZYQ": (
            "N-(4-OXO-5,6,7,8-TETRAHYDRO-4H-[1,3]THIAZOLO[5,4-C]AZEPIN-2-YL)ACETAMIDE",
            "non-polymer",
        ),
        "NAL": ("BETA-(2-NAPHTHYL)-ALANINE", "peptide"),
        "ZYZ": (
            "(4-{(2S)-2-[(TERT-BUTOXYCARBONYL)AMINO]-3-METHOXY-3-OXOPROPYL}PHENYL)METHANESELENINIC ACID",
            "non-polymer",
        ),
        "ZYX": ("4-(2-AMINOETHYL)BENZENESULFONAMIDE", "non-polymer"),
        "MOL": ("3-AMINO-3-BENZYL-[4.3.0]BICYCLO-1,6-DIAZANONAN-2-ONE", "non-polymer"),
        "RVC": (
            "3-{[(Z)-1-CARBOXY-2-CYCLOPROPYLETHENYL]OXY}-2-HYDROXYBENZOIC ACID",
            "non-polymer",
        ),
        "20A": (
            "1-ETHYL-N-(PHENYLMETHYL)-4-(TETRAHYDRO-2H-PYRAN-4-YLAMINO)-1H-PYRAZOLO[3,4-B]PYRIDINE-5-CARBOXAMIDE",
            "non-polymer",
        ),
        "RVA": (
            "3-{[(Z)-1-CARBOXY-2-PHENYLETHENYL]OXY}-2-HYDROXYBENZOIC ACID",
            "non-polymer",
        ),
        "OR8": (
            "2-CHLORO-N-[[4-(3,5-DIMETHYLISOXAZOL-4-YL)PHENYL]METHYL]-1,4-DIMETHYL-1H-PYRAZOLE-4-SULFONAMIDE",
            "non-polymer",
        ),
        "RVE": ("3-[(1-CARBOXYETHENYL)OXY]-2-HYDROXYBENZOIC ACID", "non-polymer"),
        "RVD": (
            "3-{[(1Z)-1-CARBOXYBUT-1-EN-1-YL]OXY}-2-HYDROXYBENZOIC ACID",
            "non-polymer",
        ),
        "IMJ": (
            "(2E)-N-[(1,2-DIMETHYL-1H-INDOL-3-YL)METHYL]-N-METHYL-3-(7-OXO-5,6,7,8-TETRAHYDRO-1,8-NAPHTHYRIDIN-3-YL)PROP-2-ENAMIDE",
            "non-polymer",
        ),
        "GYU": (
            "2-(ACETYLAMINO)-2-DEOXY-1-O-(PROPYLCARBAMOYL)-ALPHA-D-GLUCOPYRANOSE",
            "non-polymer",
        ),
        "GYT": (
            "BIS-(2-ACETAMIDO-2-DEOXY-ALPHA-D-GLUCOPYRANOSYLOXYCARBONYL)-4,7,10-TRIOXA-1,13-TRIDECANEDIAMINE",
            "non-polymer",
        ),
        "NAM": ("NAM NAPTHYLAMINOALANINE", "peptide"),
        "GYS": (
            "[(4Z)-2-(1-AMINO-2-HYDROXYETHYL)-4-(4-HYDROXYBENZYLIDENE)-5-OXO-4,5-DIHYDRO-1H-IMIDAZOL-1-YL]ACETIC ACID",
            "peptide",
        ),
        "RVX": ("O-[METHYL(2-METHYLPROPOXY)PHOSPHORYL]-L-SERINE", "peptide"),
        "TGG": ("GAMMA-GLUTAMYL-S-(1,2-DICARBOXYETHYL)CYSTEINYLGLYCINE", "non-polymer"),
        "OR5": ("?", "non-polymer"),
        "DH9": (
            "(2S,3S,5S)-5-[(N-FORMYL-L-LEUCYL)OXY]-2-HEXYL-3-HYDROXYHEXADECANOIC ACID",
            "non-polymer",
        ),
        "DH8": (
            "(2BETA,7BETA,16S,17R,19E,21BETA)-21-(BETA-D-GLUCOPYRANOSYLOXY)-2,7-DIHYDRO-7,17-CYCLOSARPAGAN-17-YL ACETATE",
            "non-polymer",
        ),
        "2JZ": (
            "N~2~-(3,4-DIMETHOXYPHENYL)-N~4~-[2-(2-FLUOROPHENYL)ETHYL]-N~6~-QUINOLIN-6-YL-1,3,5-TRIAZINE-2,4,6-TRIAMINE",
            "non-polymer",
        ),
        "8EP": (
            "(2R,3AR,6R,7S,7AR)-2-[(2S)-2-AMINO-3-HYDROXY-3-OXO-PROPYL]-6,7-DIHYDROXY-3,3A,5,6,7,7A-HEXAHYDROFURO[4,5-B]PYRAN-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "IML": ("N-METHYL-ISOLEUCINE", "peptide"),
        "DH3": ("?", "non-polymer"),
        "DH2": (
            "(2S,3S)-2-(3,4-DIHYDROXYPHENYL)-3,5,7-TRIHYDROXY-2,3-DIHYDRO-4H-CHROMEN-4-ONE",
            "non-polymer",
        ),
        "ZY6": (
            "1-[(2S)-4-(5-PHENYL-1H-PYRAZOLO[3,4-B]PYRIDIN-4-YL)MORPHOLIN-2-YL]METHANAMINE",
            "non-polymer",
        ),
        "ZY7": (
            "(3E)-3-[(2,4-DIMETHOXYPHENYL)METHYLIDENE]-3,4,5,6-TETRAHYDRO-2,3'-BIPYRIDINE",
            "non-polymer",
        ),
        "ZY4": (
            "N-[(1S)-1-BENZYL-2,2-DIHYDROXY-3-(TETRAHYDRO-2H-PYRAN-4-YLAMINO)PROPYL]-6-ETHYL-1-METHYL-1,3,4,6-TETRAHYDRO[1,2]THIAZEPINO[5,4,3-CD]INDOLE-8-CARBOXAMIDE 2,2-DIOXIDE",
            "non-polymer",
        ),
        "ZY5": (
            "4-[(E)-5,6-DIHYDRO-2,3'-BIPYRIDIN-3(4H)-YLIDENEMETHYL]-3-METHOXYPHENOL",
            "non-polymer",
        ),
        "ZY2": (
            "N-{(1S,2R)-1-BENZYL-2-HYDROXY-3-[(3-METHOXYBENZYL)AMINO]PROPYL}-8-ETHYL-1-METHYL-3,4,7,8-TETRAHYDRO-1H,6H-[1,2,5]THIADIAZEPINO[5,4,3-DE]QUINOXALINE-10-CARBOXAMIDE 2,2-DIOXIDE",
            "non-polymer",
        ),
        "TGL": ("TRISTEAROYLGLYCEROL", "non-polymer"),
        "ZY0": (
            "N-[(1S,2R)-1-BENZYL-2-HYDROXY-3-{[3-(TRIFLUOROMETHYL)BENZYL]AMINO}PROPYL]-4-ETHYL-8-(2-OXOPYRROLIDIN-1-YL)QUINOLINE-6-CARBOXAMIDE",
            "non-polymer",
        ),
        "ZY1": (
            "N-{(1S,2R)-1-BENZYL-2-HYDROXY-3-[(3-METHOXYBENZYL)AMINO]PROPYL}-7-ETHYL-1-METHYL-3,4-DIHYDRO-1H-[1,2,5]THIADIAZEPINO[3,4,5-HI]INDOLE-9-CARBOXAMIDE 2,2-DIOXIDE",
            "non-polymer",
        ),
        "CTA": ("CYCLOTHEONAMIDE A", "non-polymer"),
        "DHO": ("DEOXYCHOLIC ACID", "non-polymer"),
        "UW9": (
            "1-[(2S)-2-(DIMETHYLAMINO)-3-METHYLBUTYL]-3-(6-ETHOXYNAPHTHALEN-2-YL)-1H-PYRAZOLO[3,4-D]PYRIMIDIN-4-AMINE",
            "non-polymer",
        ),
        "T1A": ("TETRAETHYLARSONIUM ION", "non-polymer"),
        "200": ("4-CHLORO-L-PHENYLALANINE", "peptide"),
        "CC1": (
            "{4-[2-ACETYLAMINO-2-(3-CARBAMOYL-2-CYCLOHEXYLMETHOXY-6,7,8,9-TETRAHYDRO-5H-BENZOCYCLOHEPTEN-5YLCARBAMOYL)-ETHYL]-2-PHOSPHONO-PHENYL}-PHOSPHONIC ACID",
            "non-polymer",
        ),
        "203": (
            "1,5-DIHYDROIMIDAZO[4,5-C][1,2,6]THIADIAZIN-4(3H)-ONE 2,2-DIOXIDE",
            "non-polymer",
        ),
        "UBY": (
            "N-[(S)-({[(BENZYLOXY)CARBONYL]AMINO}METHYL)(HYDROXY)PHOSPHORYL]-L-LEUCYL-L-ALANINE",
            "peptide",
        ),
        "CTL": ("CIS-(1R,2S)-2-AMINO-1,2,3,4-TETRAHYDRONAPHTHALEN-1-OL", "non-polymer"),
        "DHL": ("2-AMINO-ETHANETHIOL", "peptide"),
        "UBZ": (
            "(2S)-2-({N-[(R)-({[(BENZYLOXY)CARBONYL]AMINO}METHYL)(HYDROXY)PHOSPHORYL]-L-LEUCYL}AMINO)BUTANOIC ACID",
            "peptide",
        ),
        "CTM": (
            "(2S)-3-(4-{[2-(4-CHLOROPHENYL)-1,3-THIAZOL-4-YL]METHOXY}-2-METHYLPHENYL)-2-ETHOXYPROPANOIC ACID",
            "non-polymer",
        ),
        "DHK": ("3-DEHYDROSHIKIMATE", "non-polymer"),
        "XSN": ("L-ALPHA-ASPARAGINE", "peptide"),
        "CTN": ("4-AMINO-1-BETA-D-RIBOFURANOSYL-2(1H)-PYRIMIDINONE", "non-polymer"),
        "DHJ": (
            "4-(2-CARBOXYETHYL)(HYDROXY)PHOSPHORYL]-3-OXOBUTANOIC ACID",
            "non-polymer",
        ),
        "JMZ": ("6-(2-AMINOPYRIMIDIN-4-YL)-1H-INDAZOL-3-AMINE", "non-polymer"),
        "PTN": ("PLATINUM TRIAMINE ION", "non-polymer"),
        "PTM": ("ALPHA-METHYL-O-PHOSPHOTYROSINE", "peptide"),
        "PTL": ("PENTANAL", "non-polymer"),
        "PTK": ("PYRENE-1,3,6,8-TETRASULFONIC ACID", "non-polymer"),
        "207": ("1-AMINO-6-CYCLOHEX-3-ENYLMETHYLOXYPURINE", "non-polymer"),
        "PTI": (
            "2-PHENYL-1-[4-(2-PIPERIDIN-1-YL-ETHOXY)-PHENYL]-1,2,3,4-TETRAHYDRO-ISOQUINOLIN-6-OL",
            "non-polymer",
        ),
        "PTH": ("CE1-METHYLENE-HYDROXY-PHOSPHOTYROSINE", "peptide"),
        "PTG": ("(5E,14E)-11-OXOPROSTA-5,9,12,14-TETRAEN-1-OIC ACID", "non-polymer"),
        "JMS": ("2-[(2,6-DICHLORO-3-METHYL-PHENYL)AMINO]BENZOIC ACID", "non-polymer"),
        "JMP": ("2-{[3-(1-METHYLETHOXY)PHENYL]CARBAMOYL}BENZOIC ACID", "non-polymer"),
        "DHH": ("(S)-2-AMINO-7,7-DIHYDROXYHEPTANOIC ACID", "non-polymer"),
        "PTC": (
            "2-METHYLCARBAMOYL-3-(4-PHOSPHONOOXY-PHENYL)-CYCLOPROPANECARBOXYLIC ACID",
            "non-polymer",
        ),
        "8NC": (
            "(3R,4R)-1-METHANESULFONYL-PYRROLIDINE-3,4--DICARBOXYLIC ACID 3-[(3-FLUORO-4-METHOXY-PHENYL)-AMIDE] 4-{[2-FLUORO-4-(2-OXO-2H-PYRIDIN-1-YL)-PHENYL]-AMIDE}",
            "non-polymer",
        ),
        "PTA": (
            "[(1-AMINO-3-METHYL-BUTYL)-HYDROXY-PHOSPHINOYL]-ACETIC ACID",
            "peptide",
        ),
        "DHG": (
            "PHOSPHONIC ACID 2-DODECANOYLAMINO-HEXYL ESTER PROPYL ESTER",
            "non-polymer",
        ),
        "PTY": ("PHOSPHATIDYLETHANOLAMINE", "non-polymer"),
        "DHF": ("DIHYDROFOLIC ACID", "non-polymer"),
        "PTW": ("1,3,5-TRIAZA-7-PHOSPHATRICYCLO[3.3.1.1~3,7~]DECANE", "non-polymer"),
        "PTU": ("2-ETHYL-1-PHENYL-ISOTHIOUREA", "non-polymer"),
        "PTT": ("TUNGSTOPTERIN", "non-polymer"),
        "LBF": ("LEPTOMYCIN B", "non-polymer"),
        "DHE": ("HEME D", "non-polymer"),
        "PTQ": ("2-PHENYLETHYL 1-THIO-BETA-D-GALACTOPYRANOSIDE", "saccharide"),
        "OA2": (
            "2-AMINO-N,N-BIS(PHENYLMETHYL)-1,3-OXAZOLE-5-CARBOXAMIDE",
            "non-polymer",
        ),
        "6SP": ("(3S)-PYRROLIDIN-3-OL", "non-polymer"),
        "JM8": ("N~5~-[(3-(ETHYLSULFANYL)PROPANIMIDOYL]-L-ORNITHINE", "non-polymer"),
        "DHD": ("2,4-DIOXO-PENTANEDIOIC ACID", "non-polymer"),
        "UBB": ("2',3'-DIDEOXY-URIDINE-5'-MONOPHOSPHATE", "non-polymer"),
        "JM2": ("N~5~-[2-(METHYLSULFANYL)ETHANIMIDOYL]-L-ORNITHINE", "non-polymer"),
        "JM1": (
            "FLOURINATED PYRIDOCARBAZOLE CYCLOPENTADIENYL RU(CO) COMPLEX",
            "non-polymer",
        ),
        "JM6": ("N~5~-[4-(METHYLSULFANYL)BUTANIMIDOYL]-L-ORNITHINE", "non-polymer"),
        "JM7": ("N~5~-[2-(ETHYLSULFANYL)ETHANIMIDOYL]-L-ORNITHINE", "non-polymer"),
        "JM4": (
            "N~5~-{3-[(1-METHYLETHYL)SULFANYL]PROPANIMIDOYL}-L-ORNITHINE",
            "non-polymer",
        ),
        "JM5": ("N~5~-[2-(PROPYLSULFANYL)ETHANIMIDOYL]-L-ORNITHINE", "non-polymer"),
        "6SA": ("ACARBOSE DERIVED HEXASACCHARIDE", "saccharide"),
        "6SC": (
            "3-BROMO-5-PHENYL-N-(PYRIDIN-4-YLMETHYL)PYRAZOLO[1,5-A]PYRIMIDIN-7-AMINE",
            "non-polymer",
        ),
        "6SE": ("TRISELANE", "non-polymer"),
        "2CE": (
            "N-[2-(2-CARBAMOYLMETHOXY-ETHOXY)-ETHYL]-2-[2-(4-CHLORO-PHENYLSULFANYL)-ACETYLAMINO]-3-(4-GUANIDINO-PHENYL)-PROPIONAMIDE",
            "non-polymer",
        ),
        "T30": (
            "3-[1-(4-FLUOROPHENYL)CYCLOPROPYL]-4-(1-METHYLETHYL)-5-[4-(TRIFLUOROMETHOXY)PHENYL]-4H-1,2,4-TRIAZOLE",
            "non-polymer",
        ),
        "PT5": (
            "(1S)-2-{[(R)-HYDROXY{[(1R,2R,3S,4R,5R,6S)-2,3,6-TRIHYDROXY-4,5-BIS(PHOSPHONOOXY)CYCLOHEXYL]OXY}PHOSPHORYL]OXY}-1-[(OCTADECANOYLOXY)METHYL]ETHYL (8E,11E)-ICOSA-5,8,11,14-TETRAENOATE",
            "non-polymer",
        ),
        "PT4": ("PLATINUM (IV) ION", "non-polymer"),
        "PT3": ("N-PROPYL-TARTRAMIC ACID", "non-polymer"),
        "N32": (
            "2,4-DIHYDROXY-3-({3-[(2S,4AS,8S,8AR)-8-METHYL-3-METHYLIDENE-7-OXO-1,3,4,7,8,8A-HEXAHYDRO-2H-2,4A-ETHANONAPHTHALEN-8-YL]PROPANOYL}AMINO)BENZOIC ACID",
            "non-polymer",
        ),
        "PCL": ("PLATINUM(II) DI-CHLORIDE", "non-polymer"),
        "PT0": (
            "(R)-[(1R)-1-AMINO-3-METHYLBUTYL](2-ETHOXY-2-OXOETHYL)PHOSPHINIC ACID",
            "peptide",
        ),
        "XS2": ("METHYL4,4II,4III,4IV-TETRATHIO-BETA-D-XYLOPENTOSIDE", "non-polymer"),
        "N33": ("?", "non-polymer"),
        "NAW": ("SODIUM ION, 3 WATERS COORDINATED", "non-polymer"),
        "UWU": ("6-CHLORANYL-7-OXIDANYL-4-PHENYL-CHROMEN-2-ONE", "non-polymer"),
        "N35": (
            "N-{3-[5-HYDROXY-2-(3-METHYLBUTYL)-3-OXO-6-PYRROLIDIN-1-YL-2,3-DIHYDROPYRIDAZIN-4-YL]-1,1-DIOXIDO-2H-1,2,4-BENZOTHIADIAZIN-7-YL}METHANESULFONAMIDE",
            "non-polymer",
        ),
        "UWN": (
            "[2-(PHENYLAMINO)BENZENE-1,4-DIYL]BIS{[4-(PYRROLIDIN-1-YL)PIPERIDIN-1-YL]METHANONE}",
            "non-polymer",
        ),
        "VX3": ("2,3-DIPHENYL-1H-INDOLE-7-CARBOXYLIC ACID", "non-polymer"),
        "PCM": (
            "1-[N[(PHENYLMETHOXY)CARBONYL]-L-LEUCYL-4-[[N/N-[(PHENYLMETHOXY)CARBONYL]-/NL-LEUCYL]AMINO]-3-PYRROLIDINONE/N",
            "non-polymer",
        ),
        "SUG": ("N~2~-(3-CARBOXYPROPANOYL)-L-ARGININE", "non-polymer"),
        "SUF": (
            "[2-(DIMETHYL-LAMBDA~4~-SULFANYL)-1-HYDROXYETHANE-1,1-DIYL]BIS(PHOSPHONIC ACID)",
            "non-polymer",
        ),
        "SUE": ("MK-5172", "non-polymer"),
        "SUD": (
            "4-DIPHOSPHOCYTIDYL-2-C-METHYL-D-ERYTHRITOL 2-PHOSPHATE",
            "non-polymer",
        ),
        "SUC": ("SUCROSE", "saccharide"),
        "SUB": (
            "3-PHENYL-2-{[4-(TOLUENE-4-SULFONYL)-THIOMORPHOLINE-3-CARBONYL]-AMINO}-PROPIONIC ACID ETHYL ESTER",
            "peptide",
        ),
        "SUA": (
            "(4-SULFAMOYL-PHENYL)-THIOCARBAMIC ACID O-(2-THIOPHEN-3-YL-ETHYL) ESTER",
            "non-polymer",
        ),
        "TH9": (
            "(2Z)-4-({3-[3-(HEXYLOXY)PHENYL]PROPYL}AMINO)-2-HYDROXY-4-OXOBUT-2-ENOIC ACID",
            "non-polymer",
        ),
        "SUO": ("N~2~-(3-CARBOXYPROPANOYL)-L-ORNITHINE", "non-polymer"),
        "SUN": ("O-[(R)-(DIMETHYLAMINO)(ETHOXY)PHOSPHORYL]-L-SERINE", "peptide"),
        "SUM": (
            "2-[3,4-DIHYDROXY-2-HYDROXYMETHYL-5-(2-HYDROXY-NONYL)-TETRAHYDRO-FURAN-2-YLOXY]-6-HYDROXYMETHYL-TETRA HYDRO-PYRAN-3,4,5-TRIOL",
            "non-polymer",
        ),
        "J60": (
            "5-[(E)-(5-CHLORO-2-OXO-1,2-DIHYDRO-3H-INDOL-3-YLIDENE)METHYL]-N-[2-(DIETHYLAMINO)ETHYL]-2,4-DIMETHYL-1H-PYRROLE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "J67": (
            "(3Z)-1-[(6-FLUORO-4H-1,3-BENZODIOXIN-8-YL)METHYL]-4-[(E)-2-PHENYLETHENYL]-1H-INDOLE-2,3-DIONE 3-OXIME",
            "non-polymer",
        ),
        "SUJ": (
            "(2R,3R)-2-[(3S,6R)-3-AMINO-6-HYDROXY-2-OXOPIPERIDINYL]-3-HYDROXYBUTANOIC ACID",
            "non-polymer",
        ),
        "SUI": ("(3-AMINO-2,5-DIOXO-1-PYRROLIDINYL)ACETIC ACID", "peptide"),
        "SUW": ("ME-A-9-N-(NAPHTHYL-2-CARBONYL)-AMINO-9-DEOXY-NEU5AC", "non-polymer"),
        "SUS": (
            "2-DEOXY-3,6-DI-O-SULFO-2-(SULFOAMINO)-ALPHA-D-GLUCOPYRANOSE",
            "saccharide",
        ),
        "SUR": ("1-(BETA-D-RIBOFURANOSYL)-2-THIO-URACIL-5'-PHOSPHATE", "nucleic-acid"),
        "SUP": ("SUCROSE-6-PHOSHPATE", "saccharide"),
        "SUZ": (
            "[(1Z)-5-FLUORO-2-METHYL-1-{4-[METHYLSULFINYL]BENZYLIDENE}-1H-INDEN-3-YL]ACETIC ACID",
            "non-polymer",
        ),
        "5DS": (
            "(R)-1-AMINO-1-[5-(DIMETHYLAMINO)-1,3,4-THIADIAZOL-2-YL]METHANESULFONAMIDE",
            "non-polymer",
        ),
        "22O": (
            "N-({(2E)-2-[(3-BROMOPHENYL)METHYLIDENE]HYDRAZINO}CARBONOTHIOYL)-BETA-D-GLUCOPYRANOSYLAMINE",
            "non-polymer",
        ),
        "NBZ": ("NITROBENZENE", "non-polymer"),
        "22M": ("2-(2-METHYLPHENYL)-1H-INDOLE-6-CARBOXIMIDAMIDE", "non-polymer"),
        "NBX": (
            "N-{[(4-AMINOPHENYL)CARBONYL]CARBAMOYL}-BETA-D-GLUCOPYRANOSYLAMINE",
            "non-polymer",
        ),
        "22K": (
            "5-[(1R,3S)-3-AMINOCYCLOHEXYL]-6-BROMO-3-(1-METHYL-1H-PYRAZOL-4-YL)PYRAZOLO[1,5-A]PYRIMIDIN-7-AMINE",
            "non-polymer",
        ),
        "L9R": (
            "(2S)-3-(OCTADECANOYLOXY)-2-[(9Z)-OCTADEC-9-ENOYLOXY]PROPYL 2-(TRIMETHYLAMMONIO)ETHYL PHOSPHATE",
            "non-polymer",
        ),
        "TEN": ("TRIETHYL PHOSPHATE", "non-polymer"),
        "TEO": ("MALATE LIKE INTERMEDIATE", "non-polymer"),
        "NBS": ("?", "non-polymer"),
        "22F": ("(4-FLUOROPHENYL)(PYRIDIN-4-YL)METHANONE", "non-polymer"),
        "22E": (
            "5,11-DIMETHYL-1-OXO-2,6-DIHYDRO-1H-PYRIDO[4,3-B]CARBAZOL-9-YL BENZOATE",
            "non-polymer",
        ),
        "NBP": ("NICOTINAMIDE 8-BROMO-ADENINE DINUCLEOTIDE PHOSPHATE", "non-polymer"),
        "TED": ("?", "nucleic-acid"),
        "22B": ("BACTERIORUBERIN", "non-polymer"),
        "22A": (
            "7-({[2-HYDROXY-1-(HYDROXYMETHYL)ETHYL]AMINO}METHYL)-3,5-DIHYDRO-4H-PYRROLO[3,2-D]PYRIMIDIN-4-ONE",
            "non-polymer",
        ),
        "IUG": ("1-(2-NITROPHENYL)ETHYL UDP-GALACTOSE", "non-polymer"),
        "DN4": (
            "[[(2R,3R,4R,5R)-5-(6-AMINOPURIN-9-YL)-3-OXIDANYL-4-PHOSPHONOOXY-OXOLAN-2-YL]METHOXY-OXIDANYL-PHOSPHORYL] [(2R,3S,4R,5R)-5-(3-CARBOXYPYRIDIN-1-IUM-1-YL)-3,4-BIS(OXIDANYL)OXOLAN-2-YL]METHYL PHOSPHATE",
            "non-polymer",
        ),
        "NBI": ("2-NITRO-N-PHENYLBENZAMIDE", "non-polymer"),
        "DN6": (
            "N-[3-(1H-TETRAZOL-5-YL)PHENYL]-3-(TRIFLUOROMETHYL)BENZAMIDE",
            "non-polymer",
        ),
        "DN1": (
            "3-({(2S)-2-[({(1R)-1-[({(1R)-1-[(R)-CARBOXY(HYDROXY)METHYL]-3,3-DIFLUOROPROPYL}AMINO)CARBONYL]-3-METHYLBUTYL}AMINO)CARBONYL]-2,3-DIHYDRO-1H-INDOL-2-YL}METHYL)THIOPHENE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "22Z": (
            "4-(HEXYLAMINO)-5-NITRO-2-[(PYRIDIN-3-YLMETHYL)AMINO]BENZAMIDE",
            "non-polymer",
        ),
        "DN3": (
            "3-[(DIMETHYLAMINO)METHYL]-N-[3-(1H-TETRAZOL-5-YL)PHENYL]BENZAMIDE",
            "non-polymer",
        ),
        "22X": (
            "(3R)-3-[(1S)-4-(ACETYLAMINO)-1-(3-CHLOROPHENYL)-1-HYDROXYBUTYL]-N-{(1S)-2-CYCLOHEXYL-1-[(METHYLAMINO)METHYL]ETHYL}PIPERIDINE-1-CARBOXAMIDE",
            "non-polymer",
        ),
        "NBC": ("S-4-NITROBUTYRYL-COA", "non-polymer"),
        "NBB": ("N-BUTYL-BENZENESULFONAMIDE", "non-polymer"),
        "22U": ("D-PHENYLALANYL-N-(3-CHLOROBENZYL)-L-PROLINAMIDE", "peptide"),
        "TES": ("TESTOSTERONE", "non-polymer"),
        "22S": (
            "N-({(2E)-2-[(2-CHLOROPHENYL)METHYLIDENE]HYDRAZINO}CARBONOTHIOYL)-BETA-D-GLUCOPYRANOSYLAMINE",
            "non-polymer",
        ),
        "DN8": ("3-BROMO-N-[3-(1H-TETRAZOL-5-YL)PHENYL]BENZAMIDE", "non-polymer"),
        "NBE": ("NITROSOBENZENE", "non-polymer"),
        "22P": ("(2R)-2-(AMINOMETHYL)-2-HYDROXY-4-PHENYLBUTANOIC ACID", "non-polymer"),
        "DNE": ("D-NORLEUCINE", "peptide"),
        "DND": ("NICOTINIC ACID ADENINE DINUCLEOTIDE", "non-polymer"),
        "DNG": ("N-FORMYL-D-NORLEUCINE", "peptide"),
        "DNF": ("2,4-DINITROPHENOL", "non-polymer"),
        "DNA": ("1,4-DIHYDROXY-2-NAPHTHOIC ACID", "non-polymer"),
        "DNC": ("3,5-DINITROCATECHOL", "non-polymer"),
        "DNB": (
            "(2S,3S,4R,5R)-2-(3,4-DIAMINOPHENYL)-5-(HYDROXYMETHYL)PYRROLIDINE-3,4-DIOL",
            "non-polymer",
        ),
        "DNM": ("N-METHYL-D-NORLEUCINE", "non-polymer"),
        "DNL": ("6-AMINO-HEXANAL", "peptide"),
        "DNO": ("D-MANNOSE", "saccharide"),
        "DNN": ("7,8-DIAMINO-NONANOIC ACID", "non-polymer"),
        "DNH": ("1R,2R-DIAMINOCYCLOHEXANE", "non-polymer"),
        "DNJ": ("1-DEOXY-NOJIRIMYCIN", "non-polymer"),
        "DNT": ("2-[1-METHYLHEXYL]-4,6-DINITROPHENOL", "non-polymer"),
        "DNQ": ("6,7-DINITROQUINOXALINE-2,3-DIONE", "non-polymer"),
        "DNP": ("3-AMINO-ALANINE", "peptide"),
        "229": (
            "7-({[(1R,2S)-2,3-DIHYDROXY-1-(HYDROXYMETHYL)PROPYL]AMINO}METHYL)-3,5-DIHYDRO-4H-PYRROLO[3,2-D]PYRIMIDIN-4-ONE",
            "non-polymer",
        ),
        "228": (
            "N-[2-(1,3-BENZODIOXOL-5-YL)ETHYL]-1-[2-(1H-IMIDAZOL-1-YL)-6-METHYLPYRIMIDIN-4-YL]-D-PROLINAMIDE",
            "non-polymer",
        ),
        "227": (
            "6,7-DIMETHOXY-4-{8-[(4-METHYLPIPERAZIN-1-YL)SULFONYL]-3,4-DIHYDROISOQUINOLIN-2(1H)-YL}QUINAZOLINE",
            "non-polymer",
        ),
        "226": ("NAPHTHALENE-1,2,4,5,7-PENTOL", "non-polymer"),
        "225": ("FELODIPINE", "non-polymer"),
        "224": (
            "1-({2-[2-(4-CHLOROPHENYL)ETHYL]-1,3-DIOXOLAN-2-YL}METHYL)-1H-IMIDAZOLE",
            "non-polymer",
        ),
        "223": (
            "(3R,4R)-1-[(4-AMINO-5H-PYRROLO[3,2-D]PYRIMIDIN-7-YL)METHYL]-4-(HYDROXYMETHYL)PYRROLIDIN-3-OL",
            "non-polymer",
        ),
        "222": (
            "2-(3-{(2-CHLORO-4-FLUOROPHENYL)[1-(2-CHLOROPHENYL)-6-OXO-1,6-DIHYDROPYRIDAZIN-3-YL]AMINO}PROPYL)-1H-ISOINDOLE-1,3(2H)-DIONE",
            "non-polymer",
        ),
        "221": (
            "(2R,3R)-3-{[3,5-BIS(TRIFLUOROMETHYL)PHENYL]AMINO}-2-CYANO-3-THIOXOPROPANAMIDE",
            "non-polymer",
        ),
        "220": ("UNDECA-3,7-DIENE-1,3,7,11-TETRACARBALDEHYDE", "non-polymer"),
        "J6Z": ("(2R)-BUTANE-2-SULFONATE", "non-polymer"),
        "U36": ("2'-O-METHYOXYETHYL-URIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "SU4": (
            "N-CYCLOPROPYL-4-OXIDANYL-N-[(2R)-2-OXIDANYL-2-PHENYL-PROPYL]BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "SU3": ("VANIPREVIR, MK-7009", "non-polymer"),
        "SU2": (
            "3-[4-(1-FORMYLPIPERAZIN-4-YL)-BENZYLIDENYL]-2-INDOLINONE",
            "non-polymer",
        ),
        "D50": (
            "N-{(2S,3R)-3-AMINO-4-[4-(BENZYLOXY)PHENYL]-2-HYDROXYBUTANOYL}-L-LEUCINE",
            "non-polymer",
        ),
        "SEC": ("2-AMINO-3-SELENINO-PROPIONIC ACID", "peptide"),
        "DR9": (
            "1-CIS-9-OCTADECANOYL-2-CIS-9-HEXADECANOYL PHOSPHATIDYL GLYCEROL",
            "non-polymer",
        ),
        "X2M": (
            "(3Z)-3-[(3,5-DIMETHYL-1H-PYRROL-2-YL)METHYLIDENE]-1,3-DIHYDRO-2H-INDOL-2-ONE",
            "non-polymer",
        ),
        "SU9": (
            "(3Z)-3-(1H-IMIDAZOL-5-YLMETHYLENE)-5-METHOXY-1H-INDOL-2(3H)-ONE",
            "non-polymer",
        ),
        "U32": (
            "4-[(5-ISOPROPYL-1,3-THIAZOL-2-YL)AMINO]BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "H4B": ("5,6,7,8-TETRAHYDROBIOPTERIN", "non-polymer"),
        "QCT": (
            "2-(3,4-DIHYDROXYPHENYL)-5,7-DIHYDROXY-4-OXO-4H-CHROMEN-3-YL 6-DEOXY-ALPHA-L-MANNOPYRANOSIDE",
            "non-polymer",
        ),
        "H4M": ("5,10-DIMETHYLENE TETRAHYDROMETHANOPTERIN", "non-polymer"),
        "H4K": ("1-(5-CYANOPYRAZIN-2-YL)-3-ISOQUINOLIN-3-YLUREA", "non-polymer"),
        "VX2": (
            "N-CYCLOHEXYL-3-[3-(TRIFLUOROMETHYL)PHENYL][1,2,4]TRIAZOLO[4,3-B]PYRIDAZIN-6-AMINE",
            "non-polymer",
        ),
        "H4W": (
            "6-{[(3R,4R)-4-{[5-(3-FLUOROPHENYL)PENTYL]OXY}PYRROLIDIN-3-YL]METHYL}-4-METHYLPYRIDIN-2-AMINE",
            "non-polymer",
        ),
        "PCF": ("1,2-DIACYL-SN-GLYCERO-3-PHOSHOCHOLINE", "non-polymer"),
        "H4S": (
            "N-[(2S,3S,4R)-3,4-DIHYDROXY-1-{[(1S,2S,3R,4R,5S)-2,3,4,5-TETRAHYDROXYCYCLOHEXYL]AMINO}OCTADECAN-2-YL]HEXACOSANAMIDE",
            "non-polymer",
        ),
        "LCZ": ("(2S)-2-AMINO-4-METHYLPENTANAL", "peptide"),
        "H4P": (
            "1-DEOXY-6-O-PHOSPHONO-1-[(PHOSPHONOMETHYL)AMINO]-L-THREO-HEXITOL",
            "non-polymer",
        ),
        "H4Z": (
            "5-(4-{[1-(2-AMINO-5-FORMYL-7-METHYL-4-OXO-3,4,5,6,7,8-HEXAHYDROPTERIDIN-6-YL)ETHYL]AMINO}PHENYL)-5-DEOXY-1-O-{5-O-[(1,3-DICARBOXYPROPOXY)(HYDROXY)PHOSPHORYL]PENTOFURANOSYL}PENTITOL",
            "non-polymer",
        ),
        "NAY": (
            "6,7,8,9-TETRAHYDROXY-5-METHYLCARBOXAMIDO-2-OXONONANOIC ACID",
            "non-polymer",
        ),
        "08Z": (
            "4-{[4-AMINO-5-(PYRIDIN-3-YLCARBONYL)-1,3-THIAZOL-2-YL]AMINO}BENZAMIDE",
            "non-polymer",
        ),
        "08Y": ("BROMOERGOCRYPTINE", "non-polymer"),
        "08X": (
            "(2S,5S,8S,13S,16Z)-5-(CYCLOHEXYLMETHYL)-3,6,9,15,18-PENTAOXO-2-(3-PHENYLPROP-2-EN-1-YL)-8-{3-[(PYRAZIN-2-YLCARBONYL)AMINO]PROPYL}-1,4,7,10,14-PENTAAZACYCLOOCTADEC-16-ENE-13-CARBOXAMIDE",
            "peptide",
        ),
        "EAM": (
            "2-[(4S)-6-(4-CHLOROPHENYL)-8-METHOXY-1-METHYL-4H-[1,2,4]TRIAZOLO[4,3-A][1,4]BENZODIAZEPIN-4-YL]-N-ETHYLACETAMIDE",
            "non-polymer",
        ),
        "EAL": (
            "1-((2S)-2-{[(1S)-1-CARBOXY-3-PHENYLPROPYL]AMINO}PROPANOYL)-L-PROLINE",
            "non-polymer",
        ),
        "RKS": (
            "METHYL N~6~-ACETYL-N~2~-[(2,3-DIHYDROXYPHENYL)CARBONYL]-N~6~-HYDROXY-D-LYSYL-N~6~-ACETYL-N~6~-HYDROXY-L-LYSINATE",
            "non-polymer",
        ),
        "08S": (
            "3-CHLORO-6-FLUORO-N-[2-[4-[(5-PROPAN-2-YL-1,3,4-THIADIAZOL-2-YL)SULFAMOYL]PHENYL]ETHYL]-1-BENZOTHIOPHENE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "08R": (
            "6-{[(3S,4S)-4-{2-[(2-METHOXYBENZYL)AMINO]ETHOXY}PYRROLIDIN-3-YL]METHYL}-4-METHYLPYRIDIN-2-AMINE",
            "non-polymer",
        ),
        "08Q": (
            "5'-O-{(S)-HYDROXY[(2-SULFANYLETHYL)AMINO]PHOSPHORYL}THYMIDINE",
            "non-polymer",
        ),
        "08P": ("N-(CARBOXYCARBONYL)-D-CYSTEINE", "peptide"),
        "08W": (
            "N-METHYL-1-{3-[(METHYLSULFONYL)AMINO]-2,4-DIOXO-7-(TRIFLUOROMETHYL)-1,2,3,4-TETRAHYDROQUINAZOLIN-6-YL}-1H-IMIDAZOLE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "EAD": (
            "2-ETHYL-ISONICOTINIC-ACYL-NICOTINAMIDE-ADENINE DINUCLEOTIDE",
            "non-polymer",
        ),
        "EAG": (
            "2-AMINOETHYL 2-(ACETYLAMINO)-2-DEOXY-BETA-D-GLUCOPYRANOSIDE",
            "saccharide",
        ),
        "08T": (
            "[[[(2R,3S,4R,5R)-5-(6-AMINOPURIN-9-YL)-3,4-BIS(OXIDANYL)OXOLAN-2-YL]METHOXY-OXIDANYL-PHOSPHORYL]OXY-OXIDANYL-PHOSPHORYL]OXY-TRIS(FLUORANYL)BERYLLIUM ",
            "non-polymer",
        ),
        "08K": (
            "8-CHLORO-1,4-DIMETHYL-6-PHENYL-4H-[1,2,4]TRIAZOLO[4,3-A][1,3,4]BENZOTRIAZEPINE",
            "non-polymer",
        ),
        "08J": (
            "8-CHLORO-6-(2-FLUOROPHENYL)-1-METHYL-4H-IMIDAZO[1,5-A][1,4]BENZODIAZEPINE",
            "non-polymer",
        ),
        "08H": (
            "8-CHLORO-1-METHYL-6-PHENYL-4H-[1,2,4]TRIAZOLO[4,3-A][1,4]BENZODIAZEPINE",
            "non-polymer",
        ),
        "HO2": ("(9E,11E,13S)-13-HYDROXYOCTADECA-9,11-DIENOIC ACID", "non-polymer"),
        "HO3": ("HOLMIUM (III) ATOM", "non-polymer"),
        "08M": ("N~5~-(PYRAZIN-2-YLCARBONYL)-L-ORNITHINE", "peptide"),
        "08L": (
            "(2S,5S,8S,13S,16Z)-5-BENZYL-2-(CYCLOHEXYLMETHYL)-3,6,9,15,18-PENTAOXO-8-{3-[(PYRAZIN-2-YLCARBONYL)AMINO]PROPYL}-1,4,7,10,14-PENTAAZACYCLOOCTADEC-16-ENE-13-CARBOXYLIC ACID",
            "peptide",
        ),
        "RKL": ("RU(TAP)2(DPPZ) COMPLEX", "non-polymer"),
        "08B": (
            "4-BROMO-5-METHYL-N-[1-(3-NITROPYRIDIN-2-YL)PIPERIDIN-4-YL]-1H-PYRROLE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "MWL": (
            "3-PHENYL-4-(PYRIDIN-4-YL)-6-[4-(PYRIMIDIN-2-YL)PIPERAZIN-1-YL]PYRIDAZINE",
            "non-polymer",
        ),
        "08F": (
            "1-[(2-AMINOPYRIDIN-4-YL)METHYL]-5-CHLORO-N-({3-[(METHYLSULFONYL)AMINO]PHENYL}SULFONYL)-3-(2-OXO-1,2-DIHYDROPYRIDIN-3-YL)-1H-INDOLE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "08E": (
            "1-[(2-AMINOPYRIDIN-4-YL)METHYL]-5-CHLORO-3-(2-OXO-1,2-DIHYDROPYRIDIN-3-YL)-1H-INDOLE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "08D": ("SULFAMETHOXAZOLE", "non-polymer"),
        "088": (
            "3-[(5-{[3-(2,6-DICHLOROPHENYL)-5-(1-METHYLETHYL)ISOXAZOL-4-YL]METHOXY}-1H-INDOL-1-YL)METHYL]BENZOIC ACID",
            "non-polymer",
        ),
        "HOB": (
            "(3ALPHA,8ALPHA,14BETA,17ALPHA)-CHOLEST-5-EN-3-YL {3-[(2-DEOXY-5-O-PHOSPHONO-BETA-D-ERYTHRO-PENTOFURANOSYL)OXY]PROPYL}CARBAMATE",
            "non-polymer",
        ),
        "HOC": ("(2S)-2-HYDROXYOCTANOIC ACID", "non-polymer"),
        "HOA": ("HYDROXYAMINE", "non-polymer"),
        "082": (
            "(5R)-5-(3-{[3-(6-METHOXY-1,2-BENZOXAZOL-3-YL)-2-OXO-2,3-DIHYDRO-1H-BENZIMIDAZOL-1-YL]METHYL}PHENYL)-5-METHYL-1,3-OXAZOLIDINE-2,4-DIONE",
            "non-polymer",
        ),
        "HOL": (
            "ALPHA-CHOLEST-5-EN-3-YL 3-[(2-DEOXY-5-O-PHOSPHONO-ALPHA-D-GLYCERO-PENTOFURANOSYL)OXY]PROPYLCARBAMATE",
            "nucleic-acid",
        ),
        "HOM": ("7-HYDROXY-4-METHYL-3-(2-HYDROXY-ETHYL)COUMARIN", "non-polymer"),
        "087": (
            "2-CHLORO-6-(4-{[(3,4-DICHLORO-5-METHYL-1H-PYRROL-2-YL)CARBONYL]AMINO}PIPERIDIN-1-YL)PYRIDINE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "085": (
            "4-[5-(1,3-BENZODIOXOL-5-YL)-4-(PYRIDIN-2-YL)-1H-IMIDAZOL-2-YL]BENZAMIDE",
            "non-polymer",
        ),
        "084": (
            "4-[5-[2-(1-PHENYL-ETHYLAMINO)-PYRIMIDIN-4-YL]-1-METHYL-4-(3-TRIFLUOROMETHYLPHENYL)-1H-IMIDAZOL-2-YL]-PIPERIDINE",
            "non-polymer",
        ),
        "HOS": ("PHOMOPSIN A", "non-polymer"),
        "HOP": (
            "(1S,2S,5S)2-(4-GLUTARIDYLBENZYL)-5-PHENYL-1-CYCLOHEXANOL",
            "non-polymer",
        ),
        "HOQ": (
            "(1R)-2-(1-CARBOXY-2-HYDROXY-2-METHYL-PROPYL)-5,5-DIMETHYL-THIAZOLIDINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "EA1": ("3H-IMIDAZO[2,1-I]PURINE", "non-polymer"),
        "T2M": ("TERT-BUTYL 2-CYANO-2-METHYLHYDRAZINECARBOXYLATE", "non-polymer"),
        "EA2": ("AMINOETHANOLPYROPHOSPHATE", "non-polymer"),
        "EA5": (
            "(4S)-4-AMINO-5-[({3-HYDROXY-2-METHYL-5-[(PHOSPHONOOXY)METHYL]PYRIDIN-4-YL}METHYL)AMINO]PENTANOIC ACID",
            "non-polymer",
        ),
        "EA4": ("ETHYL N-[(4-AMINOBENZYL)CARBAMOYL]GLYCINATE", "non-polymer"),
        "HOX": ("4-AMINO-L-PHENYLALANINE", "peptide"),
        "QC0": (
            "N-[4-METHYL-3-[6-(4-METHYLPIPERAZIN-1-YL)-4-OXIDANYLIDENE-QUINAZOLIN-3-YL]PHENYL]FURAN-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "PCE": (
            "2-(3-AMINO-4-CYCLOHEXYL-2-HYDROXY-BUTYL)-PENT-4-YNOIC ACID",
            "peptide",
        ),
        "SEO": ("2-MERCAPTOETHANOL", "non-polymer"),
        "T2A": (
            "5-METHYL-7-PHENYL-N-(3,4,5-TRIMETHOXYPHENYL)IMIDAZO[5,1-F][1,2,4]TRIAZIN-2-AMINE",
            "non-polymer",
        ),
        "TRX": ("6-HYDROXYTRYPTOPHAN", "peptide"),
        "QEL": (
            "4-[(1R,2S)-2-(4-BENZYLPIPERIDIN-1-YL)-1-HYDROXYPROPYL]PHENOL",
            "non-polymer",
        ),
        "TFP": (
            "10-[3-(4-METHYL-PIPERAZIN-1-YL)-PROPYL]-2-TRIFLUOROMETHYL-10H-PHENOTHIAZINE",
            "non-polymer",
        ),
        "OYS": ("2-PHENYL-N-(4-SULFAMOYLPHENYL)ACETAMIDE", "non-polymer"),
        "LCF": (
            "[6-(4-CHLOROPHENYL)-2,2-DIMETHYL-7-PHENYL-2,3-DIHYDRO-1H-PYRROLIZIN-5-YL]ACETIC ACID",
            "non-polymer",
        ),
        "OYQ": ("(4S)-4-METHYL-N-(4-SULFAMOYLPHENYL)HEXANAMIDE", "non-polymer"),
        "OYP": (
            "3-{4-[(TRIDEC-2-YN-1-YLOXY)METHYL]PHENYL}PROPANOIC ACID",
            "non-polymer",
        ),
        "OYJ": (
            "(3R,4R)-1-METHANESULFONYL-PYRROLIDINE-3,4-DICARBOXYLIC ACID 3-[(4-CHLORO-3-FLUORO-PHENYL)-AMIDE] 4-{[2-FLUORO-4-(2-OXO-2H-PYRIDIN-1-YL)-PHENYL]-AMIDE}",
            "non-polymer",
        ),
        "OYA": ("OCTANAL", "non-polymer"),
        "QAP": (
            "2-PHENYLETHYL (2E)-3-(3,4-DIHYDROXYPHENYL)PROP-2-ENOATE",
            "non-polymer",
        ),
        "NWD": (
            "3-(5-NITRO-2,4-DIOXO-3,4-DIHYDROPYRIMIDIN-1(2H)-YL)-L-ALANINE",
            "peptide",
        ),
        "FI0": (
            "(2R,3S)-3-AMINO-1-{[(2S)-2-METHYLBUTYL]AMINO}-4-PHENYLBUTAN-2-OL",
            "non-polymer",
        ),
        "FI1": (
            "6-CHLORO-N-((3S)-2-OXO-1-(2-OXO-2-((5S)-8-OXO-5,6-DIHYDRO-1H-1,5-METHANOPYRIDO[1,2-A][1,5]DIAZOCIN-3(2H,4H,8H)-YL)ETHYL)PIPERIDIN-3-YL)NAPHTHALENE-2-SULFONAMIDE",
            "non-polymer",
        ),
        "FI4": (
            "(2S)-1-{[3-(2-CHLOROPHENYL)-6-(2,4-DIFLUOROPHENOXY)-1H-PYRAZOLO[3,4-D]PYRIMIDIN-4-YL]AMINO}PROPAN-2-OL",
            "non-polymer",
        ),
        "O29": (
            "N-[2-(DIMETHYLAMINO)ETHYL]-2-({[3-(TRIFLUOROMETHYL)-4,5,6,7-TETRAHYDRO-1H-INDAZOL-1-YL]ACETYL}AMINO)-4,5,6,7-TETRAHYDRO-1-BENZOTHIOPHENE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "O26": (
            "2-[({3-TERT-BUTYL-4-[(METHYLAMINO)METHYL]-1H-PYRAZOL-1-YL}ACETYL)AMINO]-4,5,6,7-TETRAHYDRO-1-BENZOTHIOPHENE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "O27": (
            "N-[(3R)-PYRROLIDIN-3-YL]-2-({[3-(TRIFLUOROMETHYL)-4,5,6,7-TETRAHYDRO-1H-INDAZOL-1-YL]ACETYL}AMINO)-4,5,6,7-TETRAHYDRO-1-BENZOTHIOPHENE-3-CARBOXAMIDE",
            "peptide",
        ),
        "D56": (
            "4-(2-AMINOETHOXY)-N-(3-CHLORO-2-ETHOXY-5-PIPERIDIN-1-YLPHENYL)-3,5-DIMETHYLBENZAMIDE",
            "non-polymer",
        ),
        "O25": (
            "2-[({4-[(ETHYLAMINO)METHYL]-3-(TRIFLUOROMETHYL)-1H-PYRAZOL-1-YL}ACETYL)AMINO]-4,5,6,7-TETRAHYDRO-1-BENZOTHIOPHENE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "O22": (
            "4-[(4-AMINO-5-CYANO-6-ETHOXYPYRIDIN-2-YL)AMINO]BENZAMIDE",
            "non-polymer",
        ),
        "FIR": (
            "(2R,4S)-2-AMINOFORMYL-6-FLUORO-SPIRO[CHROMAN-4,4'-IMIDAZOLIDINE]-2',5'-DIONE",
            "non-polymer",
        ),
        "FIS": (
            "(2S,4R)-2-AMINOFORMYL-6-FLUORO-SPIRO[CHROMAN-4,4'-IMIDAZOLIDINE]-2',5'-DIONE",
            "non-polymer",
        ),
        "FIP": ("5-FLUOROINDOLE PROPANOL PHOSPHATE", "non-polymer"),
        "FIV": ("NAPHTHALENE-2-CARBOXYLIC ACID", "non-polymer"),
        "T1P": (
            "3-{2,6,8-TRIOXO-9-[(2R,3S,4R)-2,3,4,5-TETRAHYDROXYPENTYL]-1,2,3,6,8,9-HEXAHYDRO-7H-PURIN-7-YL}PROPYL DIHYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "FIT": (
            "(4AR,4BS,6AS,7S,9AS,9BS,11AR)-N-TERT-BUTYL-4A,6A-DIMETHYL-2-OXO-2,4A,4B,5,6,6A,7,8,9,9A,9B,10,11,11A-TETRADECAHYDRO-1H-INDENO[5,4-F]QUINOLINE-7-CARBOXAMIDE",
            "non-polymer",
        ),
        "LHO": (
            "2-(2-DEOXY-5-O-PHOSPHONO-BETA-D-ERYTHRO-PENTOFURANOSYL)-6-METHYLISOQUINOLINE-1(2H)-THIONE",
            "nucleic-acid",
        ),
        "O2W": (
            "METHYL ({[(3AR,4R,7R,7AR)-2-(4-AMINOPHENYL)-1,3-DIOXOOCTAHYDRO-4H-4,7-ETHANOISOINDOL-4-YL]CARBAMOYL}OXY)ACETATE",
            "non-polymer",
        ),
        "FIX": ("D-FRUCTURONIC ACID", "saccharide"),
        "LHL": (
            "3-(2,6-DICHLOROPHENYL)-7-({4-[2-(DIETHYLAMINO)ETHOXY]PHENYL}AMINO)-1-METHYL-3,4-DIHYDROPYRIMIDO[4,5-D]PYRIMIDIN-2(1H)-ONE",
            "non-polymer",
        ),
        "VD3": (
            "(1S,3Z)-3-[(2E)-2-[(1R,3AR,7AS)-7A-METHYL-1-[(2R)-6-METHYLHEPTAN-2-YL]-2,3,3A,5,6,7-HEXAHYDRO-1H-INDEN-4-YLIDENE]ETHYLIDENE]-4-METHYLIDENE-CYCLOHEXAN-1-OL",
            "non-polymer",
        ),
        "LHJ": (
            "N-CYCLOPROPYL-3-[1-(2,4-DIFLUOROPHENYL)-7-METHYL-6-OXO-6,7-DIHYDRO-1H-PYRAZOLO[3,4-B]PYRIDIN-5-YL]-4-METHYLBENZAMIDE",
            "non-polymer",
        ),
        "LHI": ("?", "non-polymer"),
        "FIB": ("1,2,3,4,5-PENTAFLUORO-6-IODOBENZENE", "non-polymer"),
        "FIC": ("5-FLUOROINDOLE-2-CARBOXYLIC ACID", "non-polymer"),
        "LHU": ("L-URIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "T3Y": (
            "25,26,27,28-TETRAHYDROXYPENTACYCLO[19.3.1.1~3,7~.1~9,13~.1~15,19~]OCTACOSA-1(25),3(28),4,6,9(27),10,12,15(26),16,18,21,23-DODECAENE-5,11,17,23-TETRASULFONIC ACID",
            "non-polymer",
        ),
        "FID": (
            "(2S,4S)-2-AMINOFORMYL-6-FLUORO-SPIRO[CHROMAN-4,4'-IMIDAZOLIDINE]-2',5'-DIONE",
            "non-polymer",
        ),
        "T2P": (
            "3-{2,6,8-TRIOXO-9-[(2S,3R,4R)-2,3,4,5-TETRAHYDROXYPENTYL]-1,2,3,6,8,9-HEXAHYDRO-7H-PURIN-7-YL}PROPYL DIHYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "O2G": (
            "2'-DEOXY-N,N-DIMETHYL-5'-O-[OXIDO(OXO)PHOSPHONIO]GUANOSINE",
            "nucleic-acid",
        ),
        "FIH": (
            "5-(4-FLUOROPHENYL)-3-{[(4-METHYLPHENYL)SULFONYL]AMINO}THIOPHENE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "FII": (
            "[(3,7,11-TRIMETHYL-DODECA-2,6,10-TRIENYLOXYCARBAMOYL)-METHYL]-PHOSPHONIC ACID",
            "non-polymer",
        ),
        "FIN": ("?", "non-polymer"),
        "FIO": ("N~5~-[(1E)-2-FLUOROETHANIMIDOYL]-L-ORNITHINE", "peptide"),
        "FIL": (
            "(1E)-1-[3-(CYCLOPENTYLOXY)-4-METHOXYPHENYL]ETHANONE O-(AMINOCARBONYL)OXIME",
            "non-polymer",
        ),
        "T1Q": (
            "N-(2,4-DIFLUORO-3-{2-[(3-HYDROXYPROPYL)AMINO]-8-METHYL-7-OXO-7,8-DIHYDROPYRIDO[2,3-D]PYRIMIDIN-6-YL}PHENYL)PROPANE-1-SULFONAMIDE",
            "non-polymer",
        ),
        "ORX": (
            "N~2~-({3-HYDROXY-2-METHYL-5-[(PHOSPHONOOXY)METHYL]PYRIDIN-4-YL}METHYL)-D-ORNITHINE",
            "non-polymer",
        ),
        "LCE": (
            "(1S,3S,4S,5S)-7,7-DICHLORO-3-METHOXY-2-THIABICYCLO[3.2.0]HEPTAN-6-ONE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "OY0": ("(2S)-2-TERT-BUTYL-N-(4-SULFAMOYLPHENYL)PENTANAMIDE", "non-polymer"),
        "D8Y": ("3-[2-(3-CHLOROPHENYL)ETHYL]PYRIDIN-2-AMINE", "non-polymer"),
        "DR0": ("N-(HYDROXYMETHYL)-N,N-DIMETHYLHEXAN-1-AMINIUM", "non-polymer"),
        "D8W": (
            "5-[(PHENYLSULFONYL)AMINO]-1,3,4-THIADIAZOLE-2-SULFONAMIDE",
            "non-polymer",
        ),
        "D8U": ("DEUTERIUM(1+)", "non-polymer"),
        "D8T": ("D-ARABINONIC ACID", "non-polymer"),
        "NE8": (
            "6-AMINO-4-(2-PHENYLETHYL)-1,7-DIHYDRO-8H-IMIDAZO[4,5-G]QUINAZOLIN-8-ONE",
            "non-polymer",
        ),
        "D8I": ("2-[(3-METHOXYBENZYL)AMINO]ETHANESULFONIC ACID", "non-polymer"),
        "D8G": (
            "6-(2-FLUOROBENZYL)-2,4-DIMETHYL-4,6-DIHYDRO-5H-THIENO[2',3':4,5]PYRROLO[2,3-D]PYRIDAZIN-5-ONE",
            "non-polymer",
        ),
        "ME5": (
            "5-CHLORO-N-((1R,2S)-2-(4-(2-OXOPYRIDIN-1(2H)-YL)BENZAMIDO) CYCLOPENTYL)THIOPHENE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "SEI": (
            "1-{3-METHYL-2-[4-(MORPHOLINE-4-CARBONYL)-BENZOYLAMINO]-BUTYRYL}-PYRROLIDINE-2-CARBOXYLIC ACID (3,3,4,4,4-PENTAFLUORO-1-ISOPROPYL-2-OXO-BUTYL)-AMIDE",
            "non-polymer",
        ),
        "DM3": ("6-DEOXYDAUNOMYCIN", "non-polymer"),
        "DS6": (
            "(4S)-4-HYDROXY-5-[4-(3-{4-[(3S)-3-HYDROXY-4,4-DIMETHYLPENTYL]-3-METHYLPHENYL}PENTAN-3-YL)-2-METHYLPHENOXY]PENTANOIC ACID",
            "non-polymer",
        ),
        "X2O": (
            "(2R)-2-AMINO-3-[(2R)-2-METHYL-1,1-DIPHENYL-BUTYL]SULFANYL-PROPANOIC ACID",
            "non-polymer",
        ),
        "DS2": (
            "{4-[3-(4-{[(2R)-2-HYDROXY-3,3-DIMETHYLBUTYL]OXY}-3-METHYLPHENYL)PENTAN-3-YL]-2-METHYLPHENOXY}ACETIC ACID",
            "non-polymer",
        ),
        "DS3": (
            "5-{4-[3-(4-{[(2R)-2-HYDROXY-3,3-DIMETHYLBUTYL]OXY}-3-METHYLPHENYL)PENTAN-3-YL]-2-METHYLPHENOXY}PENTANOIC ACID",
            "non-polymer",
        ),
        "P10": (
            "[1-(3-HYDROXY-2-OXO-1-PHENETHYL-PROPYLCARBAMOYL)2-PHENYL-ETHYL]-CARBAMIC ACID PYRIDIN-4-YLMETHYL ESTER",
            "non-polymer",
        ),
        "P11": (
            "IMIDAZOLE-PYRROLE-BETA ALANINE-IMIDAZOLE-BETA ALANINE-IMIDAZOLE-PYRROLE-BETA ALANINE-DIMETHYLAMINO PROPYLAMIDE",
            "non-polymer",
        ),
        "P12": (
            "4-[AMINO(IMINO)METHYL]-1-[2-(3-AMMONIOPROPOXY)-5-METHOXYBENZYL]PIPERAZIN-1-IUM",
            "non-polymer",
        ),
        "P13": (
            "N-[2-(3-AMINOPROPOXY)-5-(1H-INDOL-5-YL)BENZYL]-N-(2-PIPERAZIN-1-YLETHYL)AMINE",
            "non-polymer",
        ),
        "P14": (
            "N-[2-(2-{[(4-{[AMINO(IMINO)METHYL]AMINO}BUTYL)AMINO]METHYL}-4-METHOXYPHENOXY)ETHYL]GUANIDINE",
            "non-polymer",
        ),
        "P15": ("2,5,8,11,14,17-HEXAOXANONADECAN-19-OL", "non-polymer"),
        "P16": (
            "6-(2,6-DICHLOROPHENYL)-2-{[3-(HYDROXYMETHYL)PHENYL]AMINO}-8-METHYLPYRIDO[2,3-D]PYRIMIDIN-7(8H)-ONE",
            "non-polymer",
        ),
        "P17": (
            "6-(2,6-DICHLORO-PHENYL)-8-METHYL-2-(3-METHYLSULFANYL-PHENYLAMINO)-8H-PYRIDO[2,3-D]PYRIMIDIN-7-ONE",
            "non-polymer",
        ),
        "312": (
            "2-(5-{5-[AMINO(IMINIO)METHYL]-1H-BENZIMIDAZOL-2-YL}-2'-METHOXY-6-OXIDO-1,1'-BIPHENYL-3-YL)SUCCINATE",
            "non-polymer",
        ),
        "P19": (
            "N,N'-DIPHENYLPYRAZOLO[1,5-A][1,3,5]TRIAZINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "P1A": ("2,3,14,20,22-PENTAHYDROXYCHOLEST-7-EN-6-ONE", "non-polymer"),
        "P1B": (
            "(5R)-5-{4-[2-(5-ETHYLPYRIDIN-2-YL)ETHOXY]BENZYL}-1,3-THIAZOLIDINE-2,4-DIONE",
            "non-polymer",
        ),
        "P1C": ("DEACETOXYCEPHALOSPORIN-C", "non-polymer"),
        "DSZ": ("5'-O-(L-ALPHA-ASPARTYLSULFAMOYL)ADENOSINE", "non-polymer"),
        "P1E": (
            "4-{[4-{[(1R,2R)-2-(DIMETHYLAMINO)CYCLOPENTYL]AMINO}-5-(TRIFLUOROMETHYL)PYRIMIDIN-2-YL]AMINO}-N-METHYLBENZENESULFONAMIDE",
            "non-polymer",
        ),
        "DSX": ("BATIMASTAT", "non-polymer"),
        "DSY": ("5-(DIMETHYLAMINO)-2-NAPHTHALENESULFONIC ACID", "non-polymer"),
        "DSV": ("N-(DIBENZO[B,D]THIOPHEN-3-YLSULFONYL)-L-VALINE", "non-polymer"),
        "DST": ("DIMETHYLALLYL S-THIOLODIPHOSPHATE", "non-polymer"),
        "DSU": (
            "((2R,3S,4S,5S)-3,4-DIHYDROXY-5-(HYDROXYMETHYL)-5-((2R,3S,4S,5S,6R)-3,4,5-TRIHYDROXY-6-METHOXY-TETRAHYDRO-2H-PYRAN-2-YLOXY)-TETRAHYDROFURAN-2-YL)METHYL NONANOATE",
            "non-polymer",
        ),
        "DSR": ("2,6-DIDEOXY-4-THIO-BETA-D-ALLOSEPYRANOSIDE", "saccharide"),
        "DSS": ("METHYL METHYLSULFINYLMETHYL SULFIDE", "non-polymer"),
        "DSP": ("D-ASPARTIC ACID", "peptide"),
        "P1O": ("1,2-DIDECANOYL-SN-GLYCERO-3-PHOSPHOCHOLINE", "non-polymer"),
        "5UM": (
            "N-(2-FLUORO-4-SULFAMOYLPHENYL)-2-(THIOPHEN-2-YL)ACETAMIDE",
            "non-polymer",
        ),
        "DSO": (
            "ADAMANTANE-1-CARBOXYLIC ACID-5-DIMETHYLAMINO-NAPHTHALENE-1-SULFONYLAMINO-OCTYL-AMIDE",
            "non-polymer",
        ),
        "DSL": ("MONO-TRANS, OCTA-CIS DECAPRENYL-PHOSPHATE", "non-polymer"),
        "5UN": ("N-(4-SULFAMOYLPHENYL)-2-(THIOPHEN-2-YL)ACETAMIDE", "non-polymer"),
        "P1T": (
            "2-[({3-HYDROXY-2-METHYL-5-[(PHOSPHONOOXY)METHYL]PYRIDIN-4-YL}METHYL)AMINO]ACRYLIC ACID",
            "non-polymer",
        ),
        "DSK": (
            "(2R,3S,4S)-1-[(2S,3S,4R,5R,6S)-2,3,4,5,6,7-HEXAHYDROXYHEPTYL]-3,4-DIHYDROXY-2-(HYDROXYMETHYL)TETRAHYDROTHIOPHENIUM (NON-PREFERRED NAME)",
            "non-polymer",
        ),
        "DSH": ("5'-S-(3-AMINOPROPYL)-5'-THIOADENOSINE", "non-polymer"),
        "DSI": (
            "4-HYDROXY-6-(1H-INDOLE-2-CARBONYL)-8-METHYL-3,6,7,8-TETRAHYDRO-3,6-DIAZA-AS-INDACENE-2-CARBOXYLIC ACID METHYL ESTER",
            "non-polymer",
        ),
        "DSF": ("(2S)-2-(DIFLUOROMETHOXY)-1,1,1,2-TETRAFLUOROETHANE", "non-polymer"),
        "5UD": ("5-FLUOROURIDINE", "non-polymer"),
        "DSD": ("7-(CARBOXYAMINO)-8-AMINO-NONANOIC ACID", "non-polymer"),
        "DSE": ("N-METHYL-D-SERINE", "peptide"),
        "5UA": ("5'-O-CARBOXY-2'-DEOXYADENOSINE", "nucleic-acid"),
        "DSC": ("DODECANESULFONATE ION", "non-polymer"),
        "DSA": (
            "4-HYDROXY-8-METHYL-6-(4,5,6-TRIMETHOXY-1H-INDOLE-2-CARBONYL)-3,6,7,8-TETRAHYDRO-3,6-DIAZA-AS-INDACENE-2-CARBOXYLIC ACID METHYL ESTER",
            "non-polymer",
        ),
        "T29": ("TRI166 (BIFUNCTIONAL BORONATE INHIBITOR)", "non-polymer"),
        "TDI": (
            "(3R,4S)-1-[(4-AMINO-5H-PYRROLO[3,2-D]PYRIMIDIN-7-YL)METHYL]-4-[(METHYLSULFANYL)METHYL]PYRROLIDIN-3-OL",
            "non-polymer",
        ),
        "D88": (
            "4-[(3,4-DICHLORO-5-METHOXYPHENYL)AMINO]-6-METHOXY-7-[3-(4-METHYLPIPERAZIN-1-YL)PROPOXY]QUINOLINE-3-CARBONITRILE",
            "non-polymer",
        ),
        "D83": (
            "2-(4'-CARBAMIMIDOYLBIPHENYL-4-YL)-1H-INDOLE-6-CARBOXIMIDAMIDE",
            "non-polymer",
        ),
        "RIV": (
            "5-CHLORO-N-({(5S)-2-OXO-3-[4-(3-OXOMORPHOLIN-4-YL)PHENYL]-1,3-OXAZOLIDIN-5-YL}METHYL)THIOPHENE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "U34": ("?", "nucleic-acid"),
        "ME6": (
            "[(2R,3S,5R)-5-(4-AZANYL-3-METHYL-2-OXO-PYRIMIDIN-3-IUM-1-YL)-3-HYDROXY-OXOLAN-2-YL]METHYL DIHYDROGEN PHOSPHATE",
            "nucleic-acid",
        ),
        "RIT": ("RITONAVIR", "peptide"),
        "Z27": (
            "2-CHLORO-N-{3-CHLORO-4-[(5-CHLORO-1,3-BENZOTHIAZOL-2-YL)SULFANYL]PHENYL}-4-(TRIFLUOROMETHYL)BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "Z25": (
            "3-(7-AMINO-4,5-DIOXO-1,4,5,6-TETRAHYDROPYRIMIDO[4,5-C]PYRIDAZIN-3-YL)PROPANOIC ACID",
            "non-polymer",
        ),
        "Z23": ("DI[BENZAMIDIN-3-YLOXY]-3,5-DIFLUORO-4-METHYL-PYRIDINE", "non-polymer"),
        "Z22": ("S-METHYL METHANESULFONOTHIOATE", "non-polymer"),
        "Z20": (
            "(2Z,5Z)-2-[(2-CHLOROPHENYL)IMINO]-5-(4-HYDROXY-3-NITROBENZYLIDENE)-1,3-THIAZOLIDIN-4-ONE",
            "non-polymer",
        ),
        "TM3": (
            "2-[[6-[[[2-(3-HYDROXYPROPYL)-5-METHYLPHENYL]AMINO]METHYL]-2-[[3-(4-MORPHOLINYL)PROPYL]AMINO]-1H-BENZIMIDAZOL-1-YL]METHYL]-6-METHYL-3-PYRIDINOL",
            "non-polymer",
        ),
        "RIS": (
            "1-HYDROXY-2-(3-PYRIDINYL)ETHYLIDENE BIS-PHOSPHONIC ACID",
            "non-polymer",
        ),
        "3PY": ("3-HYDROXYPYRUVIC ACID", "non-polymer"),
        "RIP": ("RIBOSE(PYRANOSE FORM)", "saccharide"),
        "ME1": (
            "5-CHLORO-N-(2-(4-(2-OXOPYRIDIN-1(2H)-YL)BENZAMIDO)ETHYL)THIOPHENE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "LCO": ("CHLORATE ION", "non-polymer"),
        "3W5": (
            "N-[(3S)-1,1-DIOXIDOTETRAHYDRO-2H-THIOPYRAN-3-YL]-4-(4-{[(3S)-3-HYDROXY-1-AZABICYCLO[2.2.2]OCT-3-YL]ETHYNYL}PHENOXY)BENZAMIDE",
            "non-polymer",
        ),
        "3W4": (
            "N-[(3R)-1,1-DIOXIDOTETRAHYDRO-2H-THIOPYRAN-3-YL]-4-(4-{[(3S)-3-HYDROXY-1-AZABICYCLO[2.2.2]OCT-3-YL]ETHYNYL}PHENOXY)BENZAMIDE",
            "non-polymer",
        ),
        "RIN": ("3-AMINO-AZACYCLOTRIDECAN-2-ONE", "non-polymer"),
        "T22": (
            "2-({6-[(3R)-3-AMINOPIPERIDIN-1-YL]-3-METHYL-2,4-DIOXO-3,4-DIHYDROPYRIMIDIN-1(2H)-YL}METHYL)BENZONITRILE",
            "non-polymer",
        ),
        "TRR": (
            "2,4-DIAMINO-5-(3,4,5-TRIMETHOXY-BENZYL)-PYRIMIDIN-1-IUM",
            "non-polymer",
        ),
        "RMY": (
            "ETHYL 3-O-[5-(ACETYLAMINO)-3,5-DIDEOXY-D-GLYCERO-ALPHA-D-GALACTO-NON-2-ULOPYRANONOSYL]-1-THIO-BETA-D-GALACTOPYRANOSIDE",
            "non-polymer",
        ),
        "OCC": (
            "2ALPHA-(3-HYDROXYPROPOXY)-1ALPHA,25-DIHYDROXYVITAMIN D3",
            "non-polymer",
        ),
        "T24": (
            "(9Z,11E,13S,15Z)-13-HYDROXYOCTADECA-9,11,15-TRIENOIC ACID",
            "non-polymer",
        ),
        "D36": (
            "N-(2-AMINOETHYL)-N-{5-[(1-CYCLOHEPTYL-1H-PYRAZOLO[3,4-D]PYRIMIDIN-6-YL)AMINO]PYRIDIN-2-YL}METHANESULFONAMIDE",
            "non-polymer",
        ),
        "AU3": ("GOLD 3+ ION", "non-polymer"),
        "ME0": ("HYDROXY-L-METHIONINE", "peptide"),
        "AU4": ("4,4'-(AMINOMETHYLENE)BIS(N,N-DIMETHYLANILINE)", "non-polymer"),
        "SE4": ("SELENATE ION", "non-polymer"),
        "T26": (
            "(2R)-2-[[3-[[4-[(Z)-(4-OXO-2-SULFANYLIDENE-1,3-THIAZOLIDIN-5-YLIDENE)METHYL]PHENYL]METHYLAMINO]PHENYL]CARBONYLAMINO]PENTANEDIOIC ACID",
            "non-polymer",
        ),
        "D34": ("2,5-BIS{[4-(N-CYCLOBUTYLDIAMINOMETHYL)PHENYL]}FURAN", "non-polymer"),
        "AUC": ("GOLD (I) CYANIDE ION", "non-polymer"),
        "NE2": ("3,3',4',5-TETRACHLOROBIPHENYL-4-OL", "non-polymer"),
        "AUK": (
            "1-HYDROXY-2-METHYL-3-[(2E,6E)-3,7,11-TRIMETHYLDODECA-2,6,10-TRIEN-1-YL]QUINOLIN-4(1H)-ONE",
            "non-polymer",
        ),
        "D33": (
            "1-(2-DEOXY-5-O-PHOSPHONO-BETA-D-ERYTHRO-PENTOFURANOSYL)-1H-IMIDAZOLE",
            "nucleic-acid",
        ),
        "AUP": (
            "2-(2-PHENYL-3-PYRIDIN-2-YL-4,5,6,7-TETRAHYDRO-2H-ISOPHOSPHINDOL-1-YL)PYRIDINE",
            "non-polymer",
        ),
        "ME3": ("RUTHENIUM-PYRIDOCARBAZOLE-3", "non-polymer"),
        "AUR": ("AUROVERTIN B", "non-polymer"),
        "D30": (
            "[(4-{[2-(PENT-2-YN-1-YLOXY)-4-{[4-(TRIFLUOROMETHYL)PHENOXY]METHYL}PHENYL]SULFANYL}-5,6,7,8-TETRAHYDRONAPHTHALEN-1-YL)OXY]ACETIC ACID",
            "non-polymer",
        ),
        "LCM": ("N,N'-BUTANE-1,4-DIYLBIS(2,3-DIHYDROXYBENZAMIDE)", "non-polymer"),
        "D31": (
            "2-(4-(AMINOMETHYL)PIPERIDIN-1-YL)-N-(3_CYCLOHEXYL-4-OXO-2,4-DIHYDROINDENO[1,2-C]PYRAZOL-5-YL)ACETAMIDE",
            "non-polymer",
        ),
        "RID": (
            "(3R,5R)-7-[2-(4-FLUOROPHENYL)-5-(1-METHYLETHYL)-4-(MORPHOLIN-4-YLSULFONYL)-3-PHENYL-1H-PYRROL-1-YL]-3,5-DIHYDROXYHEPTANOIC ACID",
            "non-polymer",
        ),
        "RIE": (
            "(3R,5R)-7-[1-(4-FLUOROPHENYL)-4-(1-METHYLETHYL)-3-{METHYL[(1R)-1-PHENYLETHYL]CARBAMOYL}-1H-PYRAZOL-5-YL]-3,5-DIHYDROXYHEPTANOIC ACID",
            "non-polymer",
        ),
        "W2P": (
            "N-{2-[4-({3-CHLORO-4-[3-(TRIFLUOROMETHYL)PHENOXY]PHENYL}AMINO)-5H-PYRROLO[3,2-D]PYRIMIDIN-5-YL]ETHYL}-4-(DIMETHYLAMINO)BUTANAMIDE",
            "non-polymer",
        ),
        "3WC": ("4-METHYL-2-(PIPERAZIN-1-YL) QUINOLINE", "non-polymer"),
        "RIC": (
            "PHOSPHORIC ACID MONO-[5-(5,6-DIMETHYL-BENZOIMIDAZOL-1-YL)-3,4-DIHYDROXY-TETRAHYDRO-FURAN-2YLMETHYL] ESTER",
            "non-polymer",
        ),
        "RIA": ("2'-O-[(5'-PHOSPHO)RIBOSYL]ADENOSINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "ZZP": (
            "N-(3-{[(1Z)-3,5-DIMETHOXYCYCLOHEXA-2,4-DIEN-1-YLIDENE]AMINO}QUINOXALIN-2-YL)-4-FLUOROBENZENESULFONAMIDE",
            "non-polymer",
        ),
        "1A5": ("4-(2-OXO-1,3-OXAZOLIDIN-3-YL)BENZAMIDE", "non-polymer"),
        "1A4": (
            "N-[3-(2-OXO-2,3-DIHYDRO-1,3-THIAZOL-4-YL)PHENYL]THIOPHENE-2-SULFONAMIDE",
            "non-polymer",
        ),
        "1A7": ("4-(2-FLUOROPHENYL)-1,3-THIAZOL-2(3H)-ONE", "non-polymer"),
        "1A6": (
            "3-{[(3-METHYL-1,2-OXAZOL-5-YL)METHYL]SULFANYL}[1,2,4]TRIAZOLO[4,3-A]PYRIDINE",
            "non-polymer",
        ),
        "1A1": (
            "N-[7-(1H-IMIDAZOL-1-YL)-2-(PYRIDIN-3-YL)[1,3]THIAZOLO[5,4-D]PYRIMIDIN-5-YL]CYCLOPROPANECARBOXAMIDE",
            "non-polymer",
        ),
        "1A0": (
            "(3R)-1-[5-CHLORO-6-ETHYL-2-(PYRIDO[2,3-B]PYRAZIN-7-YLSULFANYL)-7H-PYRROLO[2,3-D]PYRIMIDIN-4-YL]PYRROLIDIN-3-AMINE",
            "non-polymer",
        ),
        "1A3": (
            "N-[3-(2-OXO-2,3-DIHYDRO-1,3-THIAZOL-4-YL)PHENYL]-1-PHENYLMETHANESULFONAMIDE",
            "non-polymer",
        ),
        "1A2": (
            "5-(4'-AMINO-1'-ETHYL-5',8'-DIFLUORO-1'H-SPIRO[PIPERIDINE-4,2'-QUINAZOLINE]-1-YLCARBONYL)PICOLINONITRILE",
            "non-polymer",
        ),
        "MO9": (
            "(1R)-2-[(CYANOMETHYL)AMINO]-1-({[2-(DIFLUOROMETHOXY)BENZYL]SULFONYL}METHYL)-2-OXOETHYL MORPHOLINE-4-CARBOXYLATE",
            "non-polymer",
        ),
        "1A9": (
            "3-CYCLOHEXYL-N-{3-(2-OXO-2,3-DIHYDRO-1,3-THIAZOL-4-YL)-5-[(THIOPHEN-2-YLSULFONYL)AMINO]PHENYL}PROPANAMIDE",
            "non-polymer",
        ),
        "1A8": (
            "N-{3-(2-OXO-2,3-DIHYDRO-1,3-THIAZOL-4-YL)-5-[(THIOPHEN-2-YLSULFONYL)AMINO]PHENYL}BUTANAMIDE",
            "non-polymer",
        ),
        "I0P": ("D-MYO INOSITOL 1,4,5,6 TETRAKISPHOSPHATE", "non-polymer"),
        "XHP": ("2-AMINO-6-METHYLIDENE-6,7-DIHYDROPTERIDIN-4(3H)-ONE", "non-polymer"),
        "7F3": (
            "(1S)-4-FLUORO-1-(4-FLUORO-3-PYRIMIDIN-5-YLPHENYL)-1-[2-(TRIFLUOROMETHYL)PYRIDIN-4-YL]-1H-ISOINDOL-3-AMINE",
            "non-polymer",
        ),
        "I0X": (
            "NALPHA-[(BENZYLOXY)CARBONYL]-N-[(1S)-1-(4-TERT-BUTOXYBENZYL)-3-DIAZO-2-OXOPROPYL]-L-PHENYLALANINAMIDE",
            "peptide",
        ),
        "I0Y": (
            "NALPHA-[(BENZYLOXY)CARBONYL]-N-[(2S)-1-(4-TERT-BUTOXYPHENYL)-4-HYDROXY-3-OXOBUTAN-2-YL]-L-PHENYLALANINAMIDE",
            "peptide",
        ),
        "YFD": ("AMORFRUTIN 2", "non-polymer"),
        "14W": ("N-(3-(AMINOMETHYL)BENZYL)ACETAMIDINE", "non-polymer"),
        "14T": (
            "3,4-DIFLUOROPHENYL 2-DEOXY-2-[(DIFLUOROACETYL)AMINO]-BETA-D-GLUCOPYRANOSIDE",
            "saccharide",
        ),
        "NE1": ("2',3,3',4',5-PENTACHLOROBIPHENYL-4-OL", "non-polymer"),
        "14S": (
            "4-(CYCLOPENTYLAMINO)-2-[(2-METHOXYBENZYL)AMINO]-N-[3-(2-OXOPYRROLIDIN-1-YL)PROPYL]PYRIMIDINE-5-CARBOXAMIDE",
            "non-polymer",
        ),
        "2TY": ("2-HYDROXY-5-{[(1E)-2-PHENYLETHYLIDENE]AMINO}-L-TYROSINE", "peptide"),
        "14Q": (
            "5-{3-[3-(1,3-BENZODIOXOL-5-YL)-5-METHOXYPHENYL]PROP-1-YN-1-YL}-6-ETHYLPYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "14Z": (
            "N-ETHYL-3-METHYL-2-OXO-1,2,3,4-TETRAHYDROQUINAZOLINE-6-SULFONAMIDE",
            "non-polymer",
        ),
        "14X": (
            "3-METHYL-6-(PYRROLIDIN-1-YLSULFONYL)-3,4-DIHYDROQUINAZOLIN-2(1H)-ONE",
            "non-polymer",
        ),
        "14Y": ("METHYL(E)-OCTADEC-9-ENOATE", "non-polymer"),
        "14F": ("2-AMINO-4,5-DIMETHOXYBENZOIC ACID", "non-polymer"),
        "14D": (
            "5-METHYL-1-[(3S)-1-{3-[3-(TRIFLUOROMETHYL)PHENOXY]BENZYL}PIPERIDIN-3-YL]PYRIMIDINE-2,4(1H,3H)-DIONE",
            "non-polymer",
        ),
        "14E": (
            "(2S)-3-[(S)-(4-BROMOPHENYL)(HYDROXY)PHOSPHORYL]-2-[(3-PHENYL-1,2-OXAZOL-5-YL)METHYL]PROPANOIC ACID",
            "non-polymer",
        ),
        "0U8": (
            "(1R,4R,5S,6S,7R,10R,11S,12S)-1,7-BIS(HYDROXYMETHYL)-2,8,13,14-TETRAOXATRICYCLO[8.2.1.1~4,7~]TETRADECANE-5,6,11,12-TETROL",
            "non-polymer",
        ),
        "0U9": ("S-BICALUTAMIDE", "non-polymer"),
        "14A": (
            "N-ALLYL-5-AMIDINOAMINOOXY-PROPYLOXY-3-CHLORO-N-CYCLOPENTYLBENZAMIDE",
            "non-polymer",
        ),
        "14N": ("4-(4-METHYLPIPERAZIN-1-YL)BENZOIC ACID", "non-polymer"),
        "0U5": (
            "(2E)-3-{5-[(2,4-DIAMINO-6-PROPYLPYRIMIDIN-5-YL)METHYL]-2,3-DIMETHOXYPHENYL}-1-[(1S)-1-(2-METHYLPROP-1-EN-1-YL)PHTHALAZIN-2(1H)-YL]PROP-2-EN-1-ONE",
            "non-polymer",
        ),
        "0U6": (
            "(2E)-3-{5-[(2,4-DIAMINO-6-ETHYLPYRIMIDIN-5-YL)METHYL]-2,3-DIMETHOXYPHENYL}-1-[(1S)-1-(2-METHYLPROP-1-EN-1-YL)PHTHALAZIN-2(1H)-YL]PROP-2-EN-1-ONE",
            "non-polymer",
        ),
        "0U7": (
            "3-[5-(3,5-DICHLOROPHENYL)-1,3,4-OXADIAZOL-2-YL]BENZENESULFONYL FLUORIDE",
            "non-polymer",
        ),
        "14K": (
            "4-(2-[(5-FLUORO-6-METHOXYPYRIDIN-3-YL)AMINO]-5-{(1R)-1-[4-(METHYLSULFONYL)PIPERAZIN-1-YL]ETHYL}PYRIDIN-3-YL)-6-METHYL-1,3,5-TRIAZIN-2-AMINE",
            "non-polymer",
        ),
        "14H": (
            "N-[(1R,3R)-3-HYDROXY-1-(HYDROXYMETHYL)-3-PHENYLPROPYL]TETRADECANAMIDE",
            "non-polymer",
        ),
        "14I": ("?", "non-polymer"),
        "146": (
            "[4R-(4ALPHA,5ALPHA,6ALPHA,7ALPHA)]-3,3'-{{TETRAHYDRO-5,6-DIHYDROXY-2-OXO-4,7-BIS(PHENYLMETHYL)-1H-1,3-DIAZEPINE-1,3(2H)-DIYL]BIS(METHYLENE)]BIS[N-1H-BENZIMIDAZOL-2-YLBENZAMIDE]",
            "non-polymer",
        ),
        "0UM": (
            "(2S,5S)-2-AMINO-6-[(2R,3S,4R,5R)-5-(6-AMINO-9H-PURIN-9-YL)-3,4-DIHYDROXYTETRAHYDROFURAN-2-YL]-5-(PROPYLAMINO)HEXANOIC ACID",
            "non-polymer",
        ),
        "0UN": (
            "N-{3-[(5-CHLORO-2-{[2-METHOXY-4-(4-METHYLPIPERAZIN-1-YL)PHENYL]AMINO}PYRIMIDIN-4-YL)OXY]PHENYL}PROP-2-ENAMIDE",
            "non-polymer",
        ),
        "0UO": ("4-METHOXY-L-TRYPTOPHAN", "peptide"),
        "0UH": ("?", "nucleic-acid"),
        "143": ("S-2,3-DIHYDRO-5-GLYCIN-2-YL-ISOXAZOL-3-YL-CYSTEINE", "peptide"),
        "0UJ": (
            "N-({1-[(1R,2R,4S)-BICYCLO[2.2.1]HEPT-2-YL]-1,6-DIHYDROIMIDAZO[4,5-D]PYRROLO[2,3-B]PYRIDIN-2-YL}METHYL)METHANESULFONAMIDE",
            "non-polymer",
        ),
        "0UK": ("?", "non-polymer"),
        "RI2": ("1,5-DI-O-PHOSPHONO-ALPHA-D-RIBOFURANOSE", "non-polymer"),
        "0UA": (
            "3-[5-(3,5-DICHLORO-4-HYDROXYPHENYL)-1,3,4-OXADIAZOL-2-YL]BENZENESULFONYL FLUORIDE",
            "non-polymer",
        ),
        "0UB": (
            "BETA-D-FRUCTOFURANOSYL-(2->6)-BETA-D-FRUCTOFURANOSYL-(2->6)-BETA-D-FRUCTOFURANOSE",
            "non-polymer",
        ),
        "149": ("D-GALACTONOLACTONE", "saccharide"),
        "YF4": (
            "5-{[ETHYL(METHYL)AMINO]METHYL}-2-METHYL-5,6-DIHYDROPYRIMIDIN-4-AMINE",
            "non-polymer",
        ),
        "NAJ": ("NICOTINAMIDE-ADENINE-DINUCLEOTIDE (ACIDIC FORM)", "non-polymer"),
        "XIN": (
            "METHYL (3Z)-3-{[(4-{METHYL[(4-METHYLPIPERAZIN-1-YL)ACETYL]AMINO}PHENYL)AMINO](PHENYL)METHYLIDENE}-2-OXO-2,3-DIHYDRO-1H-INDOLE-6-CARBOXYLATE",
            "non-polymer",
        ),
        "0UZ": ("(2R)-AMINO(3-CHLORO-4-HYDROXYPHENYL)ETHANOIC ACID", "peptide"),
        "YF3": (
            "2-{[(4-AMINO-2-METHYLPYRIMIDIN-5-YL)METHYL]AMINO}PROPANE-1-THIOL",
            "non-polymer",
        ),
        "0UU": (
            "METHYL CIS-4-[2-(BENZOYLAMINO)-6-(PIPERIDIN-1-YLMETHYL)-1H-BENZIMIDAZOL-1-YL]CYCLOHEXANECARBOXYLATE",
            "non-polymer",
        ),
        "0UV": (
            "4-FLUORO-N-{(2E)-6-{[4-(2-HYDROXYPROPAN-2-YL)PIPERIDIN-1-YL]METHYL}-1-[CIS-4-(PROPAN-2-YLCARBAMOYL)CYCLOHEXYL]-1,3-DIHYDRO-2H-BENZIMIDAZOL-2-YLIDENE}BENZAMIDE",
            "non-polymer",
        ),
        "HAG": ("4-S-GLUTATHIONYL-5-PENTYL-TETRAHYDRO-FURAN-2-OL", "non-polymer"),
        "0UP": (
            "4-{[(5R,7S)-1-(3-FLUOROPHENYL)-3,7-DIMETHYL-2,2-DIOXIDO-2-THIA-1,3,8-TRIAZASPIRO[4.5]DEC-8-YL]METHYL}-2-(PROPAN-2-YLOXY)PHENOL",
            "non-polymer",
        ),
        "0UQ": (
            "(5R,7S)-1-(3-FLUOROPHENYL)-3,7-DIMETHYL-8-[3-(PROPAN-2-YLOXY)BENZYL]-2-THIA-1,3,8-TRIAZASPIRO[4.5]DECANE 2,2-DIOXIDE",
            "non-polymer",
        ),
        "0US": (
            "N-{1-[CIS-4-(HYDROXYMETHYL)CYCLOHEXYL]-5-(PIPERIDIN-1-YLMETHYL)-1H-BENZIMIDAZOL-2-YL}-3-(PROP-2-EN-1-YLSULFAMOYL)BENZAMIDE",
            "non-polymer",
        ),
        "1AU": (
            "1-[1-(3-AMINOPHENYL)-3-TERT-BUTYL-1H-PYRAZOL-5-YL]-3-NAPHTHALEN-1-YLUREA",
            "non-polymer",
        ),
        "1AW": (
            "1-[1-(3-AMINOPHENYL)-3-TERT-BUTYL-1H-PYRAZOL-5-YL]-3-PHENYLUREA",
            "non-polymer",
        ),
        "1AV": (
            "7-(1H-IMIDAZOL-1-YL)-2-(PYRIDIN-3-YL)[1,3]THIAZOLO[5,4-D]PYRIMIDIN-5-AMINE",
            "non-polymer",
        ),
        "1AQ": (
            "(3S,7R,8AR)-2-{(2S)-2-(4,4-DIFLUOROCYCLOHEXYL)-2-[(N-METHYL-L-ALANYL)AMINO]ACETYL}-N-[(4R)-3,4-DIHYDRO-2H-CHROMEN-4-YL]-7-ETHOXYOCTAHYDROPYRROLO[1,2-A]PYRAZINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "1AP": ("2,6-DIAMINOPURINE NUCLEOTIDE", "nucleic-acid"),
        "1AR": (
            "4-O-ACETYL-2,6-DIDEOXY-3-C-METHYL-ALPHA-L-ARABINO-HEXOPYRANOSE",
            "saccharide",
        ),
        "7FA": (
            "METHYL (S)-(6Z,9Z)-OCTADECA-6,9,12-TRIEN-1-YLPHOSPHONOFLUORIDATE",
            "non-polymer",
        ),
        "1AX": ("(2R,2'R)-3,3'-OXYDIPROPANE-1,2-DIOL", "non-polymer"),
        "1AZ": (
            "1-(3,4-DICHLOROBENZYL)-3,7-DIMETHYL-8-MORPHOLIN-4-YL-3,7-DIHYDRO-1H-PURINE-2,6-DIONE",
            "non-polymer",
        ),
        "1AE": (
            "(1Z)-2-HYDROXY-3-OXOHEX-1-EN-1-YL DIHYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "1AD": (
            "(1S)-2-[(2S,5R)-2-(AMINOMETHYL)-5-PROP-1-YN-1-YLPYRROLIDIN-1-YL]-1-CYCLOPENTYL-2-OXOETHANAMINE",
            "non-polymer",
        ),
        "1AG": (
            "(2S)-2,3-DIHYDROXYPROPYL (5Z,8Z,11Z,14Z)-ICOSA-5,8,11,14-TETRAENOATE",
            "non-polymer",
        ),
        "1AF": ("2,4-BIS{[4-(N-CYCLOPENTYLDIAMINOMETHYL)PHENYL]}FURAN", "non-polymer"),
        "1AA": (
            "5'-O-[({[(2-{[(2-AMINOPHENYL)CARBONYL]OXY}ETHYL)OXY]PHOSPHINATO}OXY)PHOSPHINATO]CYTIDINE",
            "non-polymer",
        ),
        "TH4": (
            "{(2Z)-3-[(6-CHLOROPYRIDIN-3-YL)METHYL]-1,3-THIAZOLIDIN-2-YLIDENE}CYANAMIDE",
            "non-polymer",
        ),
        "1AC": ("1-AMINOCYCLOPROPANECARBOXYLIC ACID", "peptide"),
        "1AB": ("1,4-DIDEOXY-1,4-IMINO-D-ARABINITOL", "non-polymer"),
        "1AM": (
            "2-(6-METHOXY-1-OXO-1,3-DIHYDRO-2H-ISOINDOL-2-YL)-N-[4-(PIPERAZIN-1-YL)PYRIDIN-3-YL]-1,3-THIAZOLE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "1AL": ("ALLANTOATE ION", "non-polymer"),
        "1AO": (
            "2-(1H-INDAZOL-1-YL)-N-[2-(PIPERAZIN-1-YL)PHENYL]-1,3-THIAZOLE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "1AN": ("2-FLUOROANILINE", "non-polymer"),
        "SEN": ("O-[N,N-DIMETHYLPHOSPHORAMIDATE]-L-SERINE", "peptide"),
        "1AH": ("3-AMINO-3-BENZYL-[4.3.0]BICYCLO-1,6-DIAZANONAN-2-ONE", "non-polymer"),
        "1AK": (
            "4-{5-[(E)-2-{4-(2-CHLOROPHENYL)-5-[5-(METHYLSULFONYL)PYRIDIN-2-YL]-4H-1,2,4-TRIAZOL-3-YL}ETHENYL]-1,3,4-OXADIAZOL-2-YL}BENZONITRILE",
            "non-polymer",
        ),
        "1AJ": (
            "6,7-DIHYDROTHIENO[3,2-C]PYRIDIN-5(4H)-YL(1H-IMIDAZOL-1-YL)METHANONE",
            "non-polymer",
        ),
        "82A": (
            "(S)-N-(1-(3-CHLORO-4-FLUOROPHENYL)-2-HYDROXYETHYL)-4-(4-(3-CHLOROPHENYL)-1H-PYRAZOL-3-YL)-1H-PYRROLE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "F4S": ("FE4-S3 CLUSTER", "non-polymer"),
        "82X": (
            "3-(6-{[3-(2,6-DICHLOROPHENYL)-5-(1-METHYLETHYL)ISOXAZOL-4-YL]METHOXY}-1-BENZOTHIOPHEN-2-YL)BENZOIC ACID",
            "non-polymer",
        ),
        "PAX": ("?", "non-polymer"),
        "XIY": ("2-HYDROXYMETHYL-BENZOIMIDAZOLE", "non-polymer"),
        "SCL": ("ACETIC ACID SALICYLOYL-AMINO-ESTER", "non-polymer"),
        "HAR": ("N-OMEGA-HYDROXY-L-ARGININE", "peptide"),
        "RZ7": (
            "4-SULFAMOYL-N-(2,2,3,3,4,4,5,5,6,6,6-UNDECAFLUOROHEXYL)BENZAMIDE",
            "non-polymer",
        ),
        "RZ5": ("N-PENTYL-4-SULFAMOYLBENZAMIDE", "non-polymer"),
        "WTC": (
            "6-BENZYL-1-CYCLOPENTYL-1,5-DIHYDRO-4H-PYRAZOLO[3,4-D]PYRIMIDIN-4-ONE",
            "non-polymer",
        ),
        "PAY": ("OCTANE-1,3,5,7-TETRACARBOXYLIC ACID", "non-polymer"),
        "4F8": ("2-PHENYL-1H-IMIDAZOLE-4,5-DICARBOXYLIC ACID", "non-polymer"),
        "SCO": ("SUCCINAMIDE-COA", "non-polymer"),
        "WTI": (
            "4-[2-AMINO-5-(2,3-DIHYDROTHIENO[3,4-B][1,4]DIOXIN-5-YL)PYRIDIN-3-YL]BENZAMIDE",
            "non-polymer",
        ),
        "WTJ": (
            "2-AMINO-5-(2,3-DIHYDROTHIENO[3,4-B][1,4]DIOXIN-5-YL)-N-[2-(DIMETHYLAMINO)ETHYL]PYRIDINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "4F3": (
            "[2-(1-AMINO-2-HYDROXY-PROPYL)-4-(4-FLUORO-1H-INDOL-3-YLMETHYL)-5-HYDROXY-IMIDAZOL-1-YL]-ACETIC ACID",
            "peptide",
        ),
        "R5P": ("RIBOSE-5-PHOSPHATE", "saccharide"),
        "4FI": (
            "(2R)-2-[(2-METHYL-5-PHENYL-PYRAZOL-3-YL)CARBONYLAMINO]-3-NAPHTHALEN-2-YL-PROPANOIC ACID",
            "non-polymer",
        ),
        "RZ1": (
            "N-(2,2,3,3,4,4,5,5,5-NONAFLUOROPENTYL)-4-SULFAMOYLBENZAMIDE",
            "non-polymer",
        ),
        "4FJ": (
            "1'-[7-FLUORO-3-METHYL-2-(PYRIDIN-2-YL)QUINOLIN-4-YL]-6'-(MORPHOLIN-4-YL)-1',2,2',3,5,6-HEXAHYDROSPIRO[PYRAN-4,3'-PYRROLO[3,2-B]PYRIDINE]",
            "non-polymer",
        ),
        "4FE": ("(2E)-3-(3-HYDROXY-4-METHOXYPHENYL)PROP-2-ENOIC ACID", "non-polymer"),
        "4FF": ("5-METHYL-2-PHENYL-1H-IMIDAZOLE-4-CARBOXYLIC ACID", "non-polymer"),
        "4FA": ("4-FLUOROPHENETHYL ALCOHOL", "non-polymer"),
        "R71": (
            "[4-({6-[ALLYL(METHYL)AMINO]HEXYL}OXY)-2-FLUOROPHENYL](4-BROMOPHENYL)METHANONE",
            "non-polymer",
        ),
        "4FC": ("?", "non-polymer"),
        "4FB": ("(4S)-4-FLUORO-L-PROLINE", "peptide"),
        "4FY": (
            "4-(MORPHOLIN-4-YLCARBONYL)-2-PHENYL-1H-IMIDAZOLE-5-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "SCN": ("THIOCYANATE ION", "non-polymer"),
        "4FZ": ("4-FORMYLBENZENECARBOXIMIDAMIDE", "non-polymer"),
        "OCE": ("OCTANEDIOIC ACID", "non-polymer"),
        "4FW": ("4-FLUOROTRYPTOPHANE", "peptide"),
        "4FP": ("4-(4-FLUOROBENZYL)PIPERIDINE", "non-polymer"),
        "4FR": (
            "3'-CHLORO-5'-[6-({2-METHOXY-4-[(1-METHYLPIPERIDIN-4-YL)CARBAMOYL]PHENYL}AMINO)-3-METHYL-1H-PYRAZOLO[4,3-C]PYRIDIN-1-YL]BIPHENYL-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "YEN": (
            "N-{(2S)-4-(METHYLAMINO)-4-OXO-2-[4-(PHOSPHONOOXY)BENZYL]BUTANOYL}-L-ALPHA-GLUTAMYL-L-ASPARTAMIDE",
            "peptide",
        ),
        "SCI": ("L-THIOCITRULLINE", "non-polymer"),
        "829": ("5'-DEOXY-5'-[(PYRIDIN-4-YLCARBONYL)AMINO]CYTIDINE", "non-polymer"),
        "825": (
            "[{2-BROMO-4-[(2R)-3-OXO-2,3-DIPHENYLPROPYL]PHENYL}(DIFLUORO)METHYL]PHOSPHONIC ACID",
            "non-polymer",
        ),
        "824": (
            "9-HYDROXY-4-PHENYLPYRROLO[3,4-C]CARBAZOLE-1,3(2H,6H)-DIONE",
            "non-polymer",
        ),
        "826": (
            "1,3,4,9-TETRAHYDRO-2-(HYDROXYBENZOYL)-9-[(4-HYDROXYPHENYL)METHYL]-6-METHOXY-2H-PYRIDO[3,4-B]INDOLE",
            "non-polymer",
        ),
        "821": ("RU82129", "non-polymer"),
        "822": (
            "N-ACETYL-N-[1-(1,1'-BIPHENYL-4-YLMETHYL)-2-OXOAZEPAN-3-YL]-4-[DIFLUORO(PHOSPHONO)METHYL]PHENYLALANINAMIDE",
            "non-polymer",
        ),
        "1MZ": ("1-METHYLIMIDAZOLE", "non-polymer"),
        "RME": (
            "N-ETHYL-2-(4-METHYLPIPERAZIN-1-YL)PYRIDINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "ZWZ": (
            "(2R,3R,4R,5R,6S)-2-(AMINOMETHYL)-6-METHYLPIPERIDINE-3,4,5-TRIOL",
            "non-polymer",
        ),
        "RZD": (
            "(2R)-2-(ACETYLAMINO)-3-[(2S)-2-{[2-(1H-INDOL-3-YL)ETHYL]CARBAMOYL}PYRROLIDIN-1-YL]PROPYL DIHYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "TP9": (
            "(3Z)-4-{[(4-AMINO-2-METHYLPYRIMIDIN-5-YL)METHYL]AMINO}-3-MERCAPTOPENT-3-EN-1-YL TRIHYDROGEN DIPHOSPHATE",
            "non-polymer",
        ),
        "SE8": (
            "5-[(2-{[(3S)-5-{[(2S)-2-AMINO-2-CARBOXYETHYL]AMINO}-3-CARBOXY-3-HYDROXY-5-OXOPENTANOYL]AMINO}ETHYL)AMINO]-2,5-DIOXOPENTANOIC ACID",
            "non-polymer",
        ),
        "RTA": ("?", "non-polymer"),
        "RTC": ("RHENIUM (I) TRICARBONYL", "non-polymer"),
        "RTB": ("?", "non-polymer"),
        "RTM": (
            "3-(2-AMINOQUINOLIN-3-YL)-N-CYCLOHEXYL-N-METHYLPROPANAMIDE",
            "non-polymer",
        ),
        "RTL": ("RETINOL", "non-polymer"),
        "RTN": (
            "3-[2-AMINO-6-(2-METHYLPHENYL)QUINOLIN-3-YL]-N-CYCLOHEXYLPROPANAMIDE",
            "non-polymer",
        ),
        "RTH": (
            "6-[2-(3,3-DIMETHYLBUT-1-YN-1-YL)PHENYL]QUINOLIN-2-AMINE",
            "non-polymer",
        ),
        "RTP": ("RIBAVIRIN TRIPHOSPHATE", "nucleic-acid"),
        "RTR": (
            "4-({4-[(6-CHLORO-1-BENZOTHIEN-2-YL)SULFONYL]-2-OXOPIPERAZIN-1-YL}METHYL)BENZENECARBOXIMIDAMIDE",
            "non-polymer",
        ),
        "TP8": (
            "5-(2-{[HYDROXY(PHOSPHONOOXY)PHOSPHORYL]OXY}ETHYL)-2-[(1Z)-1-HYDROXY-3-(PHOSPHONOOXY)PROP-1-EN-1-YL]-3-{[(4Z)-4-IMINO-2-METHYL-4,5-DIHYDROPYRIMIDIN-5-YL]METHYL}-4-METHYL-1,3-THIAZOL-3-IUM",
            "non-polymer",
        ),
        "NC4": ("4-{[(CYCLOHEXYLAMINO)CARBONYL]AMINO}BUTANOIC ACID", "non-polymer"),
        "RTY": (
            "4-{4-[(2S,3R)-2-AMINO-4-BROMO-3-HYDROXYBUTYL]PHENOXY}BUTANOIC ACID",
            "non-polymer",
        ),
        "RTZ": (
            "10-{2-[(2R)-1-METHYLPIPERIDIN-2-YL]ETHYL}-2-(METHYLSULFANYL)-10H-PHENOTHIAZINE",
            "non-polymer",
        ),
        "2HY": (
            "2-HYDROXYETHYL (4R)-4-(4-CYANOPHENYL)-6-METHYL-2-OXO-1-[3-(TRIFLUOROMETHYL)PHENYL]-1,2,3,4-TETRAHYDROPYRIMIDINE-5-CARBOXYLATE",
            "non-polymer",
        ),
        "G": ("GUANOSINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "2HP": ("DIHYDROGENPHOSPHATE ION", "non-polymer"),
        "2HS": (
            "HEXYLPHOSPHONIC ACID (S)-2-METHYL-3-PHENYLPROPYL ESTER",
            "non-polymer",
        ),
        "2HR": (
            "HEXYLPHOSPHONIC ACID (R)-2-METHYL-3-PHENYLPROPYL ESTER",
            "non-polymer",
        ),
        "2HT": ("3-METHYLBENZONITRILE", "non-polymer"),
        "2HI": (
            "(2S,3R)-2,7-DIHYDROXY-3-(4-HYDROXYPHENYL)-2,3-DIHYDRO-4H-CHROMEN-4-ONE",
            "non-polymer",
        ),
        "2HK": (
            "4'-{6-METHOXY-7-[2-(PIPERIDIN-1-YL)ETHOXY]-1,2-DIHYDROINDENO[1,2-C]PYRAZOL-3-YL}BIPHENYL-4-OL",
            "non-polymer",
        ),
        "2HM": (
            "(2R)-2-[2-(HYDROXYAMINO)-2-OXOETHYL]-4-METHYLPENTANOIC ACID",
            "peptide",
        ),
        "RT3": (
            "S-{2-[(2-CHLORO-4-SULFAMOYLPHENYL)AMINO]-2-OXOETHYL} 6-METHYL-3,4-DIHYDROQUINOLINE-1(2H)-CARBOTHIOATE",
            "non-polymer",
        ),
        "2HN": (
            "(1R,2S,4S,5R)-1,4,5-TRIS(OXIDANYL)-3-OXIDANYLIDENE-2-[[2,3,4,5,6-PENTAKIS(FLUORANYL)PHENYL]METHYL]CYCLOHEXANE-1-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "2HA": ("DIHYDROXYACETONE", "non-polymer"),
        "2HC": ("(2E)-3-(2-HYDROXYPHENYL)ACRYLIC ACID", "non-polymer"),
        "2HE": ("(2-HYDROXYETHYL)PHOSPHONIC ACID", "non-polymer"),
        "MAG": ("BETA-METHYL-N-ACETYL-D-GLUCOSAMINE", "saccharide"),
        "8GP": ("?", "non-polymer"),
        "2HF": ("2-FLUORO-L-HISTIDINE", "peptide"),
        "RZW": ("5,6,8-TRIMETHOXY-1,4-DIMETHYLQUINOLIN-2(1H)-ONE", "non-polymer"),
        "H64": (
            "8-(6-BROMO-BENZO[1,3]DIOXOL-5-YLSULFANYL)-9-(3-ISOPROPYLAMINO-PROPYL)-ADENINE",
            "non-polymer",
        ),
        "UC2": (
            "N-[4-CLORO-3-(T-BUTYLOXOME)PHENYL-2-METHYL-3-FURAN-CARBOTHIAMIDE",
            "non-polymer",
        ),
        "NC6": ("6-{[(CYCLOHEXYLAMINO)CARBONYL]AMINO}HEXANOIC ACID", "non-polymer"),
        "MAF": ("2-DEOXY-2-FLUORO-ALPHA-D-MANNOSE", "non-polymer"),
        "ISZ": ("4-(DIAZENYLCARBONYL)PYRIDINE", "non-polymer"),
        "ISY": (
            "3-METHYLBUT-3-ENYLSULFANYL(PHOSPHONOOXY)PHOSPHINIC ACID",
            "non-polymer",
        ),
        "ISX": ("GLUCOSE BETA-1,3-ISOFAGAMINE", "non-polymer"),
        "ISS": ("(2E)-3-[(4-HYDROXY-2-OXOBUTYL)AMINO]PROP-2-ENAL", "non-polymer"),
        "ISQ": ("ISOQUINOLINE", "non-polymer"),
        "ISP": ("PHOSPHORYLISOPROPANE", "non-polymer"),
        "WW7": ("N-(6-AMINOHEXYL)-5-CHLORO-1-NAPHTHALENESULFONAMIDE", "non-polymer"),
        "ISU": ("SE-ETHYL-ISOSELENOUREA", "non-polymer"),
        "IST": ("ISOSTATINE", "non-polymer"),
        "ISJ": (
            "(3R,4R)-3-[(1-CARBOXYETHENYL)OXY]-4-HYDROXYCYCLOHEXA-1,5-DIENE-1-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "ISI": (
            "(7S)-7-[(DIMETHYLAMINO)METHYL]-1-[3-(DIMETHYLAMINO)PROPYL]-7,8-DIHYDRO-1H-FURO[3,2-E]BENZIMIDAZOL-2-AMINE",
            "non-polymer",
        ),
        "ISH": (
            "(7R)-7-[(DIMETHYLAMINO)METHYL]-1-[3-(DIMETHYLAMINO)PROPYL]-7,8-DIHYDRO-1H-FURO[3,2-E]BENZIMIDAZOL-2-AMINE",
            "non-polymer",
        ),
        "ISO": ("PARA-ISOPROPYLANILINE", "non-polymer"),
        "ISN": ("ISATIN", "non-polymer"),
        "ISM": ("5-AMINO-2-(2-METHOXY-2-OXOETHYL)BENZOIC ACID", "non-polymer"),
        "ISC": (
            "(5S,6S)-5-[(1-CARBOXYETHENYL)OXY]-6-HYDROXYCYCLOHEXA-1,3-DIENE-1-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "ISB": ("ISOBUTYRIC ACID", "non-polymer"),
        "ISA": ("3-(4-IODO-PHENYL)-2-MERCAPTO-PROPIONIC ACID", "non-polymer"),
        "ISF": ("?", "non-polymer"),
        "ISE": ("(2R,3S,4S,5R,6S)-2,3,4,5,6-PENTAHYDROXYCYCLOHEXANONE", "non-polymer"),
        "ISD": (
            "(5R)-5-[(1R)-1,2-DIHYDROXYETHYL]-3,4-DIHYDROXYFURAN-2(5H)-ONE",
            "non-polymer",
        ),
        "PVS": ("(ETHENYLSULFONYL)BENZENE", "non-polymer"),
        "ON1": (
            "2-(3,4-DICHLOROBENZYL)-5,6-DIHYDROXYPYRIMIDINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "JKZ": (
            "4-(TRIFLUOROMETHYL)-1,5,6,7-TETRAHYDRO-2H-CYCLOPENTA[B]PYRIDIN-2-ONE",
            "non-polymer",
        ),
        "LAM": (
            "4'-NITROPHENYL-S-(BETA-D-GLUCOPYRANOSYL)-(1-3)-(3-THIO-BETA-D-GLUCOPYRANOSYL)-(1-3)-BETA-D-GLUCOPYRANOSIDE",
            "non-polymer",
        ),
        "JKT": (
            "TERT-BUTYL N-[[1-[(3S,5S)-5-[(2S)-2-[AZANYL(OXIDANYL)METHYL]PYRROLIDIN-1-YL]CARBONYL-1-(4-PHENYLBUTANOYL)PYRROLIDIN-3-YL]-1,2,3-TRIAZOL-4-YL]METHYL]CARBAMATE",
            "non-polymer",
        ),
        "PVX": ("?", "nucleic-acid"),
        "T44": ("3,5,3',5'-TETRAIODO-L-THYRONINE", "non-polymer"),
        "JKF": (
            "6-[(R)-(4-CHLOROPHENYL)(METHOXY)(1-METHYL-1H-IMIDAZOL-5-YL)METHYL]-4-(2,6-DIFLUOROPHENYL)-1-METHYLQUINOLIN-2(1H)-ONE",
            "non-polymer",
        ),
        "FR3": (
            "1-((1R)-1-(HYDROXYMETHYL)-3-(1-NAPHTHYL)PROPYL)-1H-IMIDAZOLE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "FR2": (
            "1-((1R)-1-(HYDROXYMETHYL)-3-PHENYLPROPYL)-1H-IMIDAZOLE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "FR1": (
            "2-[4-[[(S)-1-[[(S)-2-[[(RS)-3,3,3-TRIFLUORO-1-ISOPROPYL-2-OXOPROPYL]AMINOCARBONYL]PYRROLIDIN-1-YL-]CARBONYL]-2-METHYLPROPYL]AMINOCARBONYL]BENZOYLAMINO]ACETIC ACID",
            "non-polymer",
        ),
        "FR0": (
            "N''-(4-(5-((1H-BENZIMIDAZOL-2-YLAMINO)METHYL)-2-THIENYL)-1,3-THIAZOL-2-YL)GUANIDINE",
            "non-polymer",
        ),
        "FR7": (
            "1-{(1R,2S)-1-[2-(2,3,-DICHLOROPHENYL)ETHYL]-2-HYDROXYPROPYL}-1H-IMIDAZOLE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "FR6": (
            "1-{(1R,2S)-2-HYDROXY-1-[2-(1-NAPHTHYL)ETHYL]PROPYL}-1H-IMIDAZOLE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "FR5": (
            "1-((1R)-1-(HYDROXYMETHYL)-3-(6-((3-(1-METHYL-1H-BENZIMIDAZOL-2-YL)PROPANOYL)AMINO)-1H-INDOL-1-YL)PROPYL)-1H-IMIDAZOLE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "FR4": (
            "1-((1R)-1-(HYDROXYMETHYL)-3-{6-[(3-PHENYLPROPANOYL)AMINO]-1H-INDOL-1-YL}PROPYL)-1H-IMIDAZOLE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "JKE": ("2-SULFANYLBENZOIC ACID", "non-polymer"),
        "FR9": (
            "1-[(1R)-3-(6-{[(BENZYLAMINO)CARBONYL]AMINO}-1H-INDOL-1-YL)-1-(HYDROXYMETHYL)PROPYL]-1H-IMIDAZOLE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "FR8": (
            "1-((1R)-1-(HYDROXYMETHYL)-3-{6-[(5-PHENYLPENTANOYL)AMINO]-1H-INDOL-1-YL}PROPYL)-1H-IMIDAZOLE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "PVL": ("PYRUVOYL GROUP", "peptide"),
        "PVN": ("PHYCOVIOLOBILIN", "non-polymer"),
        "FRC": (
            "1-{(1R,2S)-2-HYDROXY-1-[2-(2-NAPHTHYLOXY)ETHYL]PROPYL}-1H-IMIDAZONE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "FRB": (
            "2-CYCLOHEXYL-N-(2-{4-[5-(2,3-DICHLORO-PHENYL)-2H-PYRAZOL-3-YL]-PIPERIDIN-1-YL}-2-OXO-ETHYL)-2-GUANIDINO-ACETAMIDE",
            "non-polymer",
        ),
        "FRA": (
            "[4-(4-ACETYLAMINO-PHENYL)-3,5-DIOXO-4-AZA-TRICYCLO[5.2.2.0 2,6]UNDEC-1-YLCARBAMOYLOXY]-ACETIC ACID",
            "non-polymer",
        ),
        "FRG": (
            "2-[3-METHYL-4-(N-METHYL-GUANIDINO)-BUTYRYLAMINO]-3-(4-PHENYLETHYNYL-PHENYL)-PROPIONIC ACID METHYL ESTER",
            "non-polymer",
        ),
        "FRF": ("PHE-REDUCED-PHE", "non-polymer"),
        "FRE": ("FERULOYL COENZYME A", "non-polymer"),
        "FRD": ("1-PHENYL-2-AMINOPROPANE", "peptide"),
        "FRK": (
            "N-[4,5-BIS(4-HYDROXYPHENYL)-1,3-THIAZOL-2-YL]HEXANAMIDE",
            "non-polymer",
        ),
        "FRJ": (
            "3-(3,5-DIBROMO-4-HYDROXY-BENZOYL)-2-ETHYL-BENZOFURAN-6-SULFONIC ACID [4-(THIAZOL-2-YLSULFAMOYL)-PHENYL]-AMIDE",
            "non-polymer",
        ),
        "FRI": (
            "2-GUANIDINO-4-METHYL-PENTANOIC ACID [2-(4-{5-[4-(4-ACETYLAMINO-BENZYLOXY)-2,3-DICHLORO-PHENYL]-2-METHYL-2H-PYRAZOL-3-YL}-PIPERIDIN-1-YL)-2-OXO-ETHYL]-AMIDE",
            "non-polymer",
        ),
        "FRH": (
            "5-[2,3-DICHLORO-4-(5-{1-[2-(2-GUANIDINO-4-METHYL-PENTANOYLAMINO)-ACETYL]-PIPERIDIN-4-YL}-1-METHYL-1H-PYRAZOL-3-YL)-PHENOXYMETHYL]-FURAN-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "FRO": (
            "(5R,6S)-6-[(1R)-1-HYDROXYETHYL]-7-OXO-3-[(2R)-TETRAHYDROFURAN-2-YL]-4-THIA-1-AZABICYCLO[3.2.0]HEPT-2-ENE-2-CARBOXYLATE",
            "non-polymer",
        ),
        "JK1": (
            "3-{4-[(PHENYLCARBAMOYL)AMINO]-1H-PYRAZOL-1-YL}-N-(3,4,5-TRIMETHOXYPHENYL)BENZAMIDE",
            "non-polymer",
        ),
        "FRM": (
            "2-{3-[4-(4-FLUOROPHENYL)-3,6-DIHYDRO-1(2H)-PYRIDINYL]PROPYL}-8-METHYL-4(3H)-QUINAZOLINONE",
            "non-polymer",
        ),
        "FRL": (
            "1-((1R,2S)-1-{2-[2-(4-CHLOROPHENYL)-1,3-BENZOXAZOL-7-YL]ETHYL}-2-HYDROXYPROPYL)-1H-IMIDAZOLE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "FRS": (
            "1-[5-METHYL-2-(TRIFLUOROMETHYL)FURAN-3-YL]-3-[(2Z)-5-(2-{[6-(1H-1,2,4-TRIAZOL-3-YLAMINO)PYRIMIDIN-4-YL]AMINO}ETHYL)-1,3-THIAZOL-2(3H)-YLIDENE]UREA",
            "non-polymer",
        ),
        "FRR": (
            "(3R,5Z,8S,9S,11E)-8,9,16-TRIHYDROXY-14-METHOXY-3-METHYL-3,4,9,10-TETRAHYDRO-1H-2-BENZOXACYCLOTETRADECINE-1,7(8H)-DIONE",
            "non-polymer",
        ),
        "FRQ": (
            "5-FLUORO-1-[4-(4-PHENYL-3,6-DIHYDROPYRIDIN-1(2H)-YL)BUTYL]QUINAZOLINE-2,4(1H,3H)-DIONE",
            "non-polymer",
        ),
        "FRP": (
            "N1-((2S,3S,5R)-3-AMINO-6-(4-FLUOROPHENYLAMINO)-5-METHYL-6-OXO-1-PHENYLHEXAN-2-YL)-N3,N3-DIPROPYLISOPHTHALAMIDE",
            "non-polymer",
        ),
        "FRW": (
            "4-[[(2S)-3-METHYL-1-OXO-1-[(2S)-2-[[(3S)-1,1,1-TRIFLUORO-4-METHYL-2-OXO-PENTAN-3-YL]CARBAMOYL]PYRROLIDIN-1-YL]BUTAN-2-YL]CARBAMOYL]BENZOIC ACID",
            "non-polymer",
        ),
        "FRV": (
            "4-{[4-(1-CYCLOPROPYL-2-METHYL-1H-IMIDAZOL-5-YL)PYRIMIDIN-2-YL]AMINO}-N-METHYLBENZENESULFONAMIDE",
            "non-polymer",
        ),
        "FRU": ("FRUCTOSE", "saccharide"),
        "6QM": (
            "4-[5-(4-ETHOXYPHENYL)-1,2,3-THIADIAZOL-4-YL]-6-ETHYLBENZENE-1,3-DIOL",
            "non-polymer",
        ),
        "FRZ": (
            "5-(2-PHENYLPYRAZOLO[1,5-A]PYRIDIN-3-YL)-1H-PYRAZOLO[3,4-C]PYRIDAZIN-3-AMINE",
            "non-polymer",
        ),
        "FRY": (
            "(2S)-N-[(3S)-1-(2-AMINO-2-OXOETHYL)-2-OXO-1,2,3,4-TETRAHYDROQUINOLIN-3-YL]-2-CHLORO-2H-THIENO[2,3-B]PYRROLE-5-CARBOXAMIDE",
            "non-polymer",
        ),
        "FRX": (
            "(S)-2-CHLORO-N-(1-(2-(2-HYDROXYETHYLAMINO)-2-OXOETHYL)-2-OXO-1,2,3,4-TETRAHYDROQUINOLIN-3-YL)-6H-THIENO[2,3-B]PYRROLE-5-CARBOXAMIDE",
            "non-polymer",
        ),
        "XU5": (
            "(2S,4R)-4-(2-CHLOROPHENYL)SULFONYL-N-[1-(IMINOMETHYL)CYCLOPROPYL]-1-[1-(4-METHYLPHENYL)CYCLOPROPYL]CARBONYL-PYRROLIDINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "RML": (
            "(11,12-DIMETHYLDIPYRIDO[3,2-A:2',3'-C]PHENAZINE-KAPPA~2~N~4~,N~5~)[BIS(PYRAZINO[2,3-F]QUINOXALINE-KAPPA~2~N~1~,N~10~)]RUTHENIUM(2+)",
            "non-polymer",
        ),
        "XU3": (
            "(2S,4R)-4-(2-CHLOROPHENYL)SULFONYL-1-[1-(5-CHLOROTHIOPHEN-2-YL)CYCLOPROPYL]CARBONYL-N-[1-(IMINOMETHYL)CYCLOPROPYL]PYRROLIDINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "IS3": ("ISOPENTENYL DIPHOSPHATE BETA-S", "non-polymer"),
        "IS2": ("[(4-ETHYLPHENYL)AMINO]CARBONYLPHOSPHONIC ACID", "non-polymer"),
        "IS1": (
            "3-[(1E)-3-OXO-3-({4-[1-(PHENYLCARBONYL)PIPERIDIN-4-YL]BUTYL}AMINO)PROP-1-EN-1-YL]-1-BETA-D-RIBOFURANOSYLPYRIDINIUM",
            "non-polymer",
        ),
        "MAJ": ("INDANE-5-SULFONAMIDE", "non-polymer"),
        "PAD": (
            "5-BETA-D-RIBOFURANOSYLPICOLINAMIDE ADENINE-DINUCLEOTIDE",
            "non-polymer",
        ),
        "DN": ("?", "nucleic-acid"),
        "SWH": (
            "{4-[(4-{(Z)-[3-(4-FLUOROBENZYL)-2,4-DIOXO-1,3-THIAZOLIDIN-5-YLIDENE]METHYL}PHENOXY)METHYL]PHENYL}(TRIHYDROXY)BORATE(1-)",
            "non-polymer",
        ),
        "UNJ": ("5-BROMO-1-BENZOFURAN-2-CARBOXYLIC ACID", "non-polymer"),
        "DI": ("2'-DEOXYINOSINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "J45": (
            "4-[(6-CHLORO-5-NITROPYRIMIDIN-4-YL)AMINO]BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "DG": ("2'-DEOXYGUANOSINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "SWE": ("ALPHA-D-FRUCTOFURANOSYL ALPHA-D-GLUCOPYRANOSIDE", "non-polymer"),
        "DC": ("2'-DEOXYCYTIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "J43": (
            "4-{[6-CHLORO-5-FORMYL-2-(METHYLSULFANYL)PYRIMIDIN-4-YL]AMINO}BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "DA": ("2'-DEOXYADENOSINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "MAM": ("1-O-METHYL-ALPHA-D-MANNOSE", "non-polymer"),
        "F09": ("NONAN-1-OL", "non-polymer"),
        "Q2Y": (
            "4-OXO-3-[2-(5-{[4-(QUINOXALIN-2-YLAMINO)-BENZOYLAMINO]-METHYL}-THIOPHEN-2-YL)-ACETYLAMINO]-PENTANOIC ACID",
            "non-polymer",
        ),
        "DT": ("THYMIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "DU": ("2'-DEOXYURIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "SCC": ("THIOETHYL GROUP", "non-polymer"),
        "RM2": ("4-HYDROXY-N-PROPARGYL-1(R)-AMINOINDAN", "non-polymer"),
        "A8S": (
            "(2Z,4E)-5-[(1S)-1-HYDROXY-2,6,6-TRIMETHYL-4-OXOCYCLOHEX-2-EN-1-YL]-3-METHYLPENTA-2,4-DIENOIC ACID",
            "non-polymer",
        ),
        "A8P": (
            "8-N-OCTYLAMINO-ADENOSINE DIPHOSPHATE HYDROXYPYRROLIDINEDIOL",
            "non-polymer",
        ),
        "GYC": (
            "[(4Z)-2-[(1R)-1-AMINO-2-MERCAPTOETHYL]-4-(4-HYDROXYBENZYLIDENE)-5-OXO-4,5-DIHYDRO-1H-IMIDAZOL-1-YL]ACETIC ACID",
            "peptide",
        ),
        "TGT": ("TAGETITOXIN", "non-polymer"),
        "TGU": ("TRIGALACTURONIC ACID", "non-polymer"),
        "20X": (
            "N-({(2E)-2-[(4-FLUOROPHENYL)METHYLIDENE]HYDRAZINO}CARBONOTHIOYL)-BETA-D-GLUCOPYRANOSYLAMINE",
            "non-polymer",
        ),
        "GYM": ("(2R)-2,3-DIHYDROXYPROPYL TETRADECANOATE", "non-polymer"),
        "20Z": (
            "4-{[4-AMINO-5-(NAPHTHALEN-2-YLCARBONYL)-1,3-THIAZOL-2-YL]AMINO}BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "GWE": (
            "N-{4-[AMINO(DIHYDROXY)-LAMBDA~4~-SULFANYL]-2-METHYLPHENYL}-2-(4-CHLORO-2-{[3-FLUORO-5-(TRIFLUOROMETHYL)PHENYL]CARBONYL}PHENOXY)ACETAMIDE",
            "non-polymer",
        ),
        "MAL": ("MALTOSE", "saccharide"),
        "A8B": (
            "N-QUINOLIN-4-YL-N'-(1,2,3,4-TETRAHYDROACRIDIN-9-YL)OCTANE-1,8-DIAMINE",
            "non-polymer",
        ),
        "GYV": (
            "2-(ACETYLAMINO)-1-O-CARBAMOYL-2-DEOXY-ALPHA-D-GLUCOPYRANOSE",
            "saccharide",
        ),
        "20C": (
            "2-HYDROXY-4-(4-METHYL-1,3-DIOXO-1,3-DIHYDRO-2H-ISOINDOL-2-YL)BENZOIC ACID",
            "non-polymer",
        ),
        "20B": ("2-(3-HYDROXYPHENYL)-4H-CHROMEN-4-ONE", "non-polymer"),
        "20E": (
            "(2BETA,3BETA,5BETA,22R)-2,3,14,20,22,25-HEXAHYDROXYCHOLEST-7-EN-6-ONE",
            "non-polymer",
        ),
        "20D": ("2-(4-FLUOROPHENYL)-4H-CHROMEN-4-ONE", "non-polymer"),
        "A8D": ("5-METHOXY-N-(METHYLSULFONYL)-1H-INDOLE-2-CARBOXAMIDE", "non-polymer"),
        "A8E": ("(2S)-2-AMINO-4-BROMOPENT-4-ENOIC ACID", "peptide"),
        "DH7": (
            "(6E,10R,13S,18E)-2,6,10,15,19,23-HEXAMETHYLTETRACOSA-2,6,10,11,13,14,18,22-OCTAENE",
            "non-polymer",
        ),
        "DH4": (
            "2-[(1R)-1-{[(E)-AZEPAN-1-YLMETHYLIDENE]AMINO}-2-OXOETHYL]-5,5-DIMETHYL-1,3-THIAZOLIDINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "A8N": (
            "N-9-(1',2',3',4'-TETRAHYDROACRIDINYL)-1,8-DIAMINOOCTANE",
            "non-polymer",
        ),
        "20L": ("(4-PHENOXYPHENYL)METHYLAZANIUM", "non-polymer"),
        "DH1": (
            "2,4-DIAMINO-5-[2-METHOXY-5-(4-CARBOXYBUTYLOXY)BENZYL]PYRIMIDINE",
            "non-polymer",
        ),
        "A8M": (
            "5'-[(3-AMINOPROPYL)(METHYL)AMINO]-5'-DEOXY-8-METHYLADENOSINE",
            "non-polymer",
        ),
        "201": ("3-AMIDO-5-BIPHENYL-BENZOIC ACID", "non-polymer"),
        "A83": (
            "TERT-BUTYL 4-[({[1-((1S,2R)-1-BENZYL-2-HYDROXY-3-{ISOBUTYL[(4-METHOXYPHENYL)SULFONYL]AMINO}PROPYL)-1H-1,2,3-TRIAZOL-4-YL]METHYL}AMINO)CARBONYL]BENZYLCARBAMATE",
            "non-polymer",
        ),
        "A80": (
            "4-[(5,5,8,8-TETRAMETHYL-5,6,7,8-TETRAHYDRONAPHTHALEN-2-YL)CARBAMOYL]BENZOIC ACID",
            "non-polymer",
        ),
        "202": ("BROMIC ACID", "non-polymer"),
        "205": (
            "2-(2,6-DICHLOROPHENYL)-1,3-BENZOXAZOLE-6-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "204": (
            "2-AMINO-4-HYDROXYPYRIMIDINE-5-CARBOXYLIC ACID ETHYL ESTER",
            "non-polymer",
        ),
        "A84": (
            "2-[({2-[(1Z)-3-(DIMETHYLAMINO)PROP-1-ENYL]-4-FLUOROPHENYL}SULFONYL)AMINO]-5,6,7,8-TETRAHYDRONAPHTHALENE-1-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "A85": (
            "N-{1-BENZYL-2,2-DIFLUORO-3,3-DIHYDROXY-4-[3-METHYL-2-(3-METHYL-3-PYRIDIN-2-YLMETHYL-UREIDO)-BUTYRYLAMINO]-5-PHENYL-PENTYL}-3-METHYL-2-(3-METHYL-3-PYRIDIN-2-YLMETHYL-UREIDO)-BUTYRAMIDE",
            "non-polymer",
        ),
        "209": ("8-AMINO-1,3-DIMETHYL-3,7-DIHYDROPURINE-2,6-DIONE", "non-polymer"),
        "208": (
            "(2S)-3-(1-{[2-(2-CHLOROPHENYL)-5-METHYL-1,3-OXAZOL-4-YL]METHYL}-1H-INDOL-5-YL)-2-ETHOXYPROPANOIC ACID",
            "non-polymer",
        ),
        "A88": (
            "(5R,6R)-2,4-BIS-(4-HYDROXY-3-METHOXYBENZYL)-1,5-DIBENZYL-3-OXO-6-HYDROXY-1,2,4-TRIAZACYCLOHEPTANE",
            "non-polymer",
        ),
        "A89": (
            "N-(4-BROMO-3-METHYLBENZOYL)-4-[DIFLUORO(PHOSPHONO)METHYL]-L-PHENYLALANYL-N~5~-(3-IODOBENZOYL)-L-ORNITHINAMIDE",
            "peptide",
        ),
        "DHC": ("CAFFEIC ACID", "non-polymer"),
        "DHB": ("3,4-DIHYDROXYBENZOIC ACID", "non-polymer"),
        "DHA": ("2-AMINO-ACRYLIC ACID", "peptide"),
        "D3F": (
            "2-[(2,4-DICHLORO-5-METHYLPHENYL)SULFONYL]-1,3-DINITRO-5-(TRIFLUOROMETHYL)BENZENE",
            "non-polymer",
        ),
        "ZN3": ("ZINC ION, 1 WATER COORDINATED", "non-polymer"),
        "GY3": (
            "8-(2-ETHENYL-6-ETHYL-4-METHYLPHENYL)TETRAHYDRO-7H-PYRAZOLO[1,2-D][1,4,5]OXADIAZEPINE-7,9(8H)-DIONE",
            "non-polymer",
        ),
        "DHZ": ("3,4-DIHYDRO-1H-PYRIMIDIN-2-ONE NUCLEOSIDE", "non-polymer"),
        "DHY": ("2-(3,4-DIHYDROXYPHENYL)ACETIC ACID", "non-polymer"),
        "PAG": ("2-PHOSPHO-D-GLYCERIC ACID", "non-polymer"),
        "DHV": ("3-HYDROXY-D-VALINE", "peptide"),
        "DHU": ("5,6-DIHYDROURIDINE-5'-PHOSPHATE", "nucleic-acid"),
        "DHT": ("5-ALPHA-DIHYDROTESTOSTERONE", "non-polymer"),
        "DHS": ("3-AMINO-4,5-DIHYDROXY-CYCLOHEX-1-ENECARBOXYLATE", "non-polymer"),
        "DHR": ("(2S)-HYDROXY(4-HYDROXYPHENYL)ETHANENITRILE", "non-polymer"),
        "DHQ": ("3,4-DIHYDRO-5-METHYL-ISOQUINOLINONE", "non-polymer"),
        "DHP": ("3-DECYL-2,5-DIOXO-4-HYDROXY-3-PYRROLINE", "peptide"),
        "D3E": (
            "N-CYCLOPROPYL-N-(TRANS-4-PYRIDIN-3-YLCYCLOHEXYL)-4-[(1S)-2,2,2-TRIFLUORO-1-HYDROXY-1-METHYLETHYL]BENZAMIDE",
            "non-polymer",
        ),
        "T5X": (
            "2-C-{3-[(4-AMINO-2-METHYLPYRIMIDIN-5-YL)METHYL]-5-(2-{[(R)-HYDROXY(PHOSPHONOOXY)PHOSPHORYL]OXY}ETHYL)-4-METHYL-1,3-THIAZOL-3-IUM-2-YL}-5-O-PHOSPHONO-D-XYLITOL",
            "non-polymer",
        ),
        "J4Z": (
            "N-ANTHRACEN-2-YL-5-METHYL[1,2,4]TRIAZOLO[1,5-A]PYRIMIDIN-7-AMINE",
            "non-polymer",
        ),
        "MAN": ("ALPHA-D-MANNOSE", "saccharide"),
        "SW1": ("1-[2-(THIOPHEN-2-YL)-1,3-THIAZOL-4-YL]METHANAMINE", "non-polymer"),
        "RM1": ("N-METHYL-1(R)-AMINOINDAN", "non-polymer"),
        "SW2": ("1-[3-(1H-PYRROL-1-YL)PHENYL]METHANAMINE", "non-polymer"),
        "D3": (
            "1-(2-DEOXY-BETA-D-RIBOFURANOSYL)-4-(3-BENZAMIDO)PHENYLIMIDAZOLE",
            "nucleic-acid",
        ),
        "SCM": ("SPECTINOMYCIN", "non-polymer"),
        "H6K": (
            "1-(5-CYANOPYRAZIN-2-YL)-3-(5-PHENYL-1H-PYRAZOL-3-YL)UREA",
            "non-polymer",
        ),
        "QAQ": (
            "[(4-{2-[(3-HYDROXYPHENYL)AMINO]-1H-BENZIMIDAZOL-1-YL}-1,3,5-TRIAZIN-2-YL)AMINO]ACETONITRILE",
            "non-polymer",
        ),
        "92M": (
            "9-CYCLOPENTYL-2-(4-(4-HYDROXYPIPERIDIN-1-YL)-2-METHOXYPHENYLAMINO)-5-METHYL-8,9-DIHYDRO-5H-PYRIMIDO[4,5-B][1,4]DIAZEPIN-6(7H)-ONE",
            "non-polymer",
        ),
        "QAR": (
            "1-(4-AMINO-6-METHYL-1,3,5-TRIAZIN-2-YL)-N-(1H-PYRAZOL-3-YL)-1H-BENZIMIDAZOL-2-AMINE",
            "non-polymer",
        ),
        "92G": ("4-(2-AMINO-1,3-THIAZOL-4-YL)PHENOL", "non-polymer"),
        "IJ6": ("2-[(CYCLOHEXYLMETHYL)AMINO]ETHANESULFONIC ACID", "non-polymer"),
        "QAB": (
            "2-AMINO-4,5,6,7-TETRAHYDRO-1-BENZOTHIOPHENE-3-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "H6P": (
            "(2E)-4-HYDROXY-3-METHYLBUT-2-EN-1-YL TRIHYDROGEN DIPHOSPHATE",
            "non-polymer",
        ),
        "BN3": (
            "4-[3-CHLORO-4-({[(2,4-DICHLOROBENZOYL)AMINO]CARBONYL}AMINO)PHENOXY]BUTANOIC ACID",
            "non-polymer",
        ),
        "BN2": (
            "7-[2,6-DICHLORO-4-({[(2-CHLOROBENZOYL)AMINO]CARBONYL}AMINO)PHENOXY]HEPTANOIC ACID",
            "non-polymer",
        ),
        "BN1": (
            "(S)-2-AMINO-3-[3-HYDROXY-5-(2-METHYL-2H-TETRAZOL-5-YL)ISOXAZOL-4-YL]PROPIONIC ACID",
            "non-polymer",
        ),
        "BN6": (
            "(2S,8R,9S,15S)-15-CYCLOHEXYL-9,12-BIS(CYCLOPROPYLMETHYL)-8-HYDROXY-20-METHYL-4,7,11,14,17-PENTAOXO-2-PHENYL-18-OXA-3,6,10,12,13,16-HEXAAZAHENICOSAN-1-OIC ACID",
            "non-polymer",
        ),
        "BN5": (
            "5-[3-({[(2,4-DICHLOROBENZOYL)AMINO]CARBONYL}AMINO)-2-METHYLPHENOXY]PENTANOIC ACID",
            "non-polymer",
        ),
        "BN4": (
            "4-[4-({[(2,4-DICHLOROBENZOYL)AMINO]CARBONYL}AMINO)-2,3-DIMETHYLPHENOXY]BUTANOIC ACID",
            "non-polymer",
        ),
        "HA8": ("N6-(BUTA-2,3-DIENYL)ADENINE", "non-polymer"),
        "RIO": ("RIBOSTAMYCIN", "non-polymer"),
        "PAJ": ("PANTOYL ADENYLATE", "non-polymer"),
        "RIM": ("RIMANTADINE", "non-polymer"),
        "PE1": (
            "N~2~-({3-HYDROXY-2-METHYL-5-[(PHOSPHONOOXY)METHYL]PYRIDIN-4-YL}METHYL)-L-LYSINE",
            "peptide",
        ),
        "HA0": ("HETEROARYLALANINE 5-PHENYL OXAZOLE", "non-polymer"),
        "HA1": ("9-(5,5-DIFLUORO-5-PHOSPHONOPENTYL)GUANINE", "non-polymer"),
        "HA2": (
            "6-[{(2S)-2-AMINO-3-[4-(BENZYLOXY)PHENYL]PROPYL}(HYDROXY)AMINO]-6-OXOHEXANOIC ACID)",
            "non-polymer",
        ),
        "HA3": (
            "N-HYDROXY-5-[(3-PHENYL-5,6-DIHYDROIMIDAZO[1,2-A]PYRAZIN-7(8H)-YL)CARBONYL]THIOPHENE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "RIB": ("RIBOSE", "saccharide"),
        "HA5": (
            "(5R)-3-HYDROXY-4,5,6,7-TETRAHYDROISOXAZOLO[5,4-C]PYRIDINE-5-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "TIL": (
            "[2-[5-CARBOXYETHYL-2-PHENOLATO(NITRILOMETHYLIDYNE)][PHENOLATO]]CHROMIUM(III)",
            "non-polymer",
        ),
        "HA7": (
            "(7S)-3-HYDROXY-4,5,6,7-TETRAHYDROISOXAZOLO[5,4-C]PYRIDINE-7-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "HAH": (
            "(1S,2S,3R,6R)-6-[(2-HYDROXYBENZYL)AMINO]CYCLOHEX-4-ENE-1,2,3-TRIOL",
            "non-polymer",
        ),
        "BNZ": ("BENZENE", "non-polymer"),
        "HAL": (
            "N-{[2-({[1-(4-CARBOXYBUTANOYL)AMINO]-2-PHENYLETHYL}-HYDROXYPHOSPHINYL)OXY]ACETYL}-2-PHENYLETHYLAMINE",
            "non-polymer",
        ),
        "HAM": ("HISTIDYL-ADENOSINE MONOPHOSPHATE", "non-polymer"),
        "HAN": ("2-AMINO-5,6-DIMETHYL-BENZIMIDAZOLE-1-PENTANOIC ACID", "non-polymer"),
        "HAO": (
            "{[3-(HYDRAZINOCARBONYL)-4-METHOXYPHENYL]AMINO}(OXO)ACETIC ACID",
            "peptide",
        ),
        "BNS": ("BENZENESULFONIC ACID", "non-polymer"),
        "BNR": ("BIS-DAUNORUBICIN", "non-polymer"),
        "HAB": ("?", "non-polymer"),
        "BNP": (
            "(3R,4R)-4-{4-[3-(BENZYLOXY)PROPOXY]PHENYL}-3-(2-NAPHTHYLMETHOXY)PIPERIDINE",
            "non-polymer",
        ),
        "HAD": ("(CARBOXYHYDROXYAMINO)ETHANOIC ACID", "non-polymer"),
        "HAE": ("ACETOHYDROXAMIC ACID", "non-polymer"),
        "BNU": (
            "6-BROMO-1-ETHYL-4-OXO-7-(PIPERAZIN-1-YL)-1,4-DIHYDROQUINOLINE-3-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "BNT": ("2,5-DIBROMO-3-ISOPROPYL-6-METHYLBENZO-1,4-QUINONE", "non-polymer"),
        "HAX": ("N-HYDROXYAMIDOCARBOXYMETHYLDETHIA COENZYME *A", "non-polymer"),
        "BNI": (
            "5-(2-OXO-HEXAHYDRO-THIENO[3,4-D]IMIDAZOL-6-YL)-PENTANOIC ACID (4-NITRO-PHENYL)-AMIDE",
            "non-polymer",
        ),
        "BNH": (
            "[(N-{3-[(N-{CYCLOHEXYL[(ISOBUTOXYCARBONYL)AMINO]ACETYL}-3-CYCLOPROPYLALANYL)AMINO]-4-CYCLOPROPYL-2-OXOBUTANOYL}GLYCYL)AMINO](PHENYL)ACETIC ACID",
            "non-polymer",
        ),
        "BNO": ("NORLEUCINE BORONIC ACID", "peptide"),
        "BNN": ("ACETYL-P-AMIDINOPHENYLALANINE", "peptide"),
        "BNL": ("BIPHENYL", "non-polymer"),
        "HAP": (
            "(N-(2-HYDROXAMATEMETHYLENE-4-METHYL-PENTOYL)PHENYLALANYL)METHYL AMINE",
            "non-polymer",
        ),
        "HAQ": (
            "5-AMINO-4-OXO-1,2,4,5,6,7-HEXAHYDRO-AZEPINO[3,2,1-HI]INDOLE-2-CARBOXYLIC ACID",
            "peptide",
        ),
        "BNA": ("BIS-(NAPHTHYL-1-METHYL) ACETIC ACID", "non-polymer"),
        "HAS": ("HEME-AS", "non-polymer"),
        "BNG": ("B-NONYLGLUCOSIDE", "saccharide"),
        "BNF": ("N-BENZYLFORMAMIDE", "non-polymer"),
        "BNE": (
            "2-[3-(3-ETHYL-1-METHYL-2-OXO-AZEPAN-3-YL)-PHENOXY]-4-[1-AMINO-1-(1-METHYL-1H-IMIDIZOL-5-YL)-ETHYL]-BENZONITRILE",
            "non-polymer",
        ),
        "BND": ("ME-A-N-BENZOYL-AMINO-9-DEOXY-NEU5AC", "non-polymer"),
        "ZN2": ("ZINC ION ON 3-FOLD CRYSTAL AXIS", "non-polymer"),
        "RFP": ("RIFAMPICIN", "non-polymer"),
        "KU8": (
            "4-[3-(1,4-DIAZEPAN-1-YLCARBONYL)-4-FLUOROBENZYL]PHTHALAZIN-1(2H)-ONE",
            "non-polymer",
        ),
        "928": ("5-AMINO-1-(4-CHLOROPHENYL)-1H-PYRAZOLE-4-CARBONITRILE", "non-polymer"),
        "929": (
            "N-[4-(1-BENZYLCARBAMOYL-2-METHYL-PROPYLCARBAMOYL)-1-(3,5-DIFLUORO-BENZYLOXYMETHYL)-2-HYDROXY-PENTYL]-5-[METHYL(METHANESULFONYL)AMINO]-N'-(1-PHENYL-ETHYL)-ISOPHTHALAMIDE",
            "non-polymer",
        ),
        "JK5": (
            "6-{[(3R,4R)-4-(2-{[2-(3-CHLORO-5-FLUOROPHENYL)-2,2-DIFLUOROETHYL]AMINO}ETHOXY)PYRROLIDIN-3-YL]METHYL}-4-METHYLPYRIDIN-2-AMINE",
            "non-polymer",
        ),
        "922": (
            "(5R,6S,8S)-8-(3-{[AMINO(IMINO)METHYL]AMINO}PHENYL)-6-HYDROXY-5-ISOPROPYL-3-OXO-1-PHENYL-2,7-DIOXA-4-AZA-6-PHOSPHANONAN-9-OIC ACID 6-OXIDE",
            "non-polymer",
        ),
        "923": (
            "(5Z)-4-HYDROXY-3-[(2R)-2-(2-HYDROXYPROPAN-2-YL)-2,3-DIHYDRO-1-BENZOFURAN-5-YL]-5-{[(2R)-2-(2-HYDROXYPROPAN-2-YL)-2,3-DIHYDRO-1-BENZOFURAN-5-YL]METHYLIDENE}FURAN-2(5H)-ONE",
            "non-polymer",
        ),
        "924": (
            "1-[5-(4-AMINOTHIENO[3,2-C]PYRIDIN-3-YL)-2,3-DIHYDRO-1H-INDOL-1-YL]-2-PHENYLETHANONE",
            "non-polymer",
        ),
        "925": (
            "3-[7-[(3-HYDROXYPHENYL)AMINO]PYRAZOLO[1,5-A]PYRIMIDIN-2-YL]-N-(1-HYDROXY-2,2,6,6-TETRAMETHYL-PIPERIDIN-4-YL)BENZAMIDE",
            "non-polymer",
        ),
        "78P": (
            "(2R)-2-(7-CARBAMOYL-1H-BENZIMIDAZOL-2-YL)-2-METHYLPYRROLIDINIUM",
            "non-polymer",
        ),
        "LJH": (
            "N,N'-(IMINODIETHANE-2,1-DIYL)BIS(4-AMINO-N-BENZYLBENZENESULFONAMIDE)",
            "non-polymer",
        ),
        "T7M": (
            "(2R)-1-(HEPTADECANOYLOXY)-3-{[(R)-HYDROXY{[(1R,2R,3R,4R,5S,6R)-2,3,5,6-TETRAHYDROXY-4-(PHOSPHONOOXY)CYCLOHEXYL]OXY}PHOSPHORYL]OXY}PROPAN-2-YL (5Z,8Z,11Z,14Z)-ICOSA-5,8,11,14-TETRAENOATE",
            "non-polymer",
        ),
        "LAA": ("(3R)-3-HYDROXY-L-ALPHA-ASPARAGINE", "peptide"),
        "78A": (
            "N-({(2S,3S)-3-[(BENZYLAMINO)CARBONYL]OXIRAN-2-YL}CARBONYL)-L-ISOLEUCYL-L-PROLINE",
            "non-polymer",
        ),
        "78M": ("(2S)-2,3-DIHYDROXYPROPYL(7Z)-PENTADEC-7-ENOATE", "non-polymer"),
        "78N": ("(2R)-2,3-DIHYDROXYPROPYL(7Z)-PENTADEC-7-ENOATE", "non-polymer"),
        "78H": ("7,8-DIHYDROPTEROATE", "non-polymer"),
        "VGI": (
            "2-(3-BROMOPHENYL)-6-[(2-HYDROXYETHYL)AMINO]-1H-BENZO[DE]ISOQUINOLINE-1,3(2H)-DIONE",
            "non-polymer",
        ),
        "FO5": (
            "3-OXO-3-{4-[3-(THIOPHEN-2-YL)-1,2,4-OXADIAZOL-5-YL]PIPERIDIN-1-YL}PROPANENITRILE",
            "non-polymer",
        ),
        "NEX": ("unknown", "non-polymer"),
        "FO1": (
            "1-DEOXY-1-(8-HYDROXY-2,4-DIOXO-3,4-DIHYDROPYRIMIDO[4,5-B]QUINOLIN-10(2H)-YL)-D-RIBITOL",
            "non-polymer",
        ),
        "MZ4": (
            "N-[(1S,2R)-1-BENZYL-2-HYDROXY-3-{[(3-METHOXYPHENYL)SULFONYL](2-THIENYLMETHYL)AMINO}PROPYL]-3,4-DIHYDROXYBENZAMIDE",
            "non-polymer",
        ),
        "PAN": ("5-PHOSPHO-D-ARABINOHYDROXAMIC ACID", "non-polymer"),
        "VGK": (
            "4-[2-AMINO-5-(3,4,5-TRIMETHOXYPHENYL)PYRIDIN-3-YL]BENZOIC ACID",
            "non-polymer",
        ),
        "LAB": ("LATRUNCULIN B", "non-polymer"),
        "LJ9": (
            "3-ETHYL-6-[(3AR,6AS)-HEXAHYDROPYRROLO[3,4-C]PYRROL-2(1H)-YL]-1-PHENYL-1H-INDAZOLE",
            "non-polymer",
        ),
        "LJ1": ("2,6-DIMETHYL-4-[(E)-2-PHENYLETHENYL]PHENOL", "non-polymer"),
        "LJ3": ("3,5-DIBROMOBIPHENYL-4-OL", "non-polymer"),
        "LJ2": ("2,6-DIBROMO-4-[(E)-2-PHENYLETHENYL]PHENOL", "non-polymer"),
        "LJ5": ("N-(3,5-DIBROMO-4-HYDROXYPHENYL)BENZAMIDE", "non-polymer"),
        "LJ4": ("2,6-DIBROMO-4-PHENOXYPHENOL", "non-polymer"),
        "FOT": (
            "5-FLUORO-2,6-DIOXO-1,2,3,6-TETRAHYDROPYRIMIDINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "FOU": (
            "(5S,6R)-5-FLUORO-6-HYDROXYDIHYDROPYRIMIDINE-2,4(1H,3H)-DIONE",
            "non-polymer",
        ),
        "VGG": (
            "1-TERT-BUTYL-3-(3-METHYLBENZYL)-1H-PYRAZOLO[3,4-D]PYRIMIDIN-4-AMINE",
            "non-polymer",
        ),
        "FOP": ("2-HYDROXY-PROPYL-AMMONIUM", "non-polymer"),
        "FOQ": (
            "[(1S)-3-[ACETYL(HYDROXY)AMINO]-1-(PYRIDIN-4-YL)PROPYL]PHOSPHONIC ACID",
            "non-polymer",
        ),
        "FOR": ("FORMYL GROUP", "non-polymer"),
        "FOS": ("[[N-(BENZYLOXYCARBONYL)AMINO]METHYL]PHOSPHATE", "non-polymer"),
        "LJE": (
            "N-{3-METHYL-4-[(3-PYRIMIDIN-4-YLPYRIDIN-2-YL)OXY]PHENYL}-3-(TRIFLUOROMETHYL)BENZAMIDE",
            "non-polymer",
        ),
        "LAC": ("LACTIC ACID", "non-polymer"),
        "FOX": ("?", "nucleic-acid"),
        "FOZ": (
            "N-[4-({[(6S)-2-AMINO-5-FORMYL-4-OXO-1,4,5,6,7,8-HEXAHYDROPTERIDIN-6-YL]METHYL}AMINO)BENZOYL]-L-GLUTAMIC ACID",
            "non-polymer",
        ),
        "LJF": (
            "N-[4-({3-[5-FLUORO-2-(METHYLIDENEAMINO)PYRIMIDIN-4-YL]PYRIDIN-2-YL}OXY)PHENYL]-2-(PHENYLAMINO)BENZAMIDE",
            "non-polymer",
        ),
        "JK4": (
            "6-{[(3R,4R)-4-(2-{[2-(2,3-DIFLUOROPHENYL)-2,2-DIFLUOROETHYL]AMINO}ETHOXY)PYRROLIDIN-3-YL]METHYL}-4-METHYLPYRIDIN-2-AMINE",
            "non-polymer",
        ),
        "FOE": (
            "2-(2-AMINO-3-OXO-PROPYLSULFANYL)-N-(4-FLUORO-PHENYL)-N-ISOPROPYL-ACETAMIDE",
            "peptide",
        ),
        "FOG": ("(4S,5S)-5-AMINO-4-HYDROXY-6-PHENYLHEXANOIC ACID", "peptide"),
        "FOA": ("2-FUROIC ACID", "non-polymer"),
        "FOB": (
            "[(1R)-3-[ACETYL(HYDROXY)AMINO]-1-(PYRIDIN-4-YL)PROPYL]PHOSPHONIC ACID",
            "non-polymer",
        ),
        "FOC": ("FUCITOL", "non-polymer"),
        "FOL": ("FOLIC ACID", "non-polymer"),
        "FOM": ("3-[FORMYL(HYDROXY)AMINO]PROPYLPHOSPHONIC ACID", "non-polymer"),
        "FON": (
            "N-{[4-({[(6R)-2-AMINO-5-FORMYL-4-OXO-1,4,5,6,7,8-HEXAHYDROPTERIDIN-6-YL]METHYL}AMINO)PHENYL]CARBONYL}-L-GLUTAMIC ACID",
            "non-polymer",
        ),
        "FOO": (
            "(3E)-4-{3-HYDROXY-2-METHYL-5-[(PHOSPHONOOXY)METHYL]PYRIDIN-4-YL}-2-IMINOBUT-3-ENOIC ACID",
            "non-polymer",
        ),
        "FOH": ("FARNESOL", "non-polymer"),
        "FOI": (
            "METHYL 3-{[(5S)-1-(HYDROXYAMINO)-5H-INDEN-5-YL]AMINO}FURO[2,3-C]PYRIDINE-2-CARBOXYLATE",
            "non-polymer",
        ),
        "FOK": ("FORSKOLIN", "non-polymer"),
        "784": (
            "[AMINO (4-{(3AS,4R,8AS,8BR)-1,3-DIOXO-2- [3-(TRIMETHYLAMMONIO) PROPYL]DECAHYDROPYRROLO[3,4-A] PYRROLIZIN-4-YL}PHENYL) METHYLENE]AMMONIUM",
            "non-polymer",
        ),
        "785": (
            "2-{5-[AMINO(IMINIO)METHYL]-1H-BENZIMIDAZOL-2-YL}-4-(TRIFLUOROMETHOXY)BENZENOLATE",
            "non-polymer",
        ),
        "786": (
            "(2R,3R)-2,3-DIHYDROXY-4-{(2R)-2-[2-(METHYLAMINO)-5-(METHYLSULFONYL)-1,3-THIAZOL-4-YL]PYRROLIDIN-1-YL}-4-OXO-N-{(1R)-1-[4-(1H-PYRAZOL-1-YL)PHENYL]ETHYL}BUTANAMIDE",
            "non-polymer",
        ),
        "787": ("(PHENYL-PHOSPHONO-METHYL)-PHOSPHONIC ACID", "non-polymer"),
        "780": (
            "3-{5-[AMINO(IMINIO)METHYL]-1H-BENZIMIDAZOL-2-YL}-1,1'-BIPHENYL-2-OLATE",
            "non-polymer",
        ),
        "783": (
            "3-{5-[AMINO(IMINIO)METHYL]-1H-INDOL-2-YL}-5-METHOXY-1,1'-BIPHENYL-2-OLATE",
            "non-polymer",
        ),
        "SCT": ("(SOUTH)-METHANOCARBA-THYMIDINE", "non-polymer"),
        "789": ("3-BROMO-6-HYDROXY-2-(4-HYDROXYPHENYL)-1H-INDEN-1-ONE", "non-polymer"),
        "IU6": ("6-ETHYL-ISO-URSODEOXYCHOLIC ACID", "non-polymer"),
        "LJZ": (
            "4,4'-{CYCLOHEXA-2,5-DIENE-1,4-DIYLIDENEBIS[(E)METHYLYLIDENE(E)DIAZENE-2,1-DIYL]}BIS[N-(2-CHLOROPHENYL)-4-OXOBUTANAMIDE]",
            "non-polymer",
        ),
        "NG6": ("N-ACETYL-D-GALACTOSAMINE 6-SULFATE", "saccharide"),
        "MZ7": (
            "N~2~-ACETYL-N-{(1S,2R)-3-[(1,3-BENZOTHIAZOL-6-YLSULFONYL)(PENTYL)AMINO]-1-BENZYL-2-HYDROXYPROPYL}-L-VALINAMIDE",
            "non-polymer",
        ),
        "TPT": ("?", "non-polymer"),
        "ROK": (
            "4-AMINO-N-[(2-SULFANYLETHYL)CARBAMOYL]BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "OKI": (
            "(2S)-2-[2-(4-CHLOROPHENYL)-1H-BENZIMIDAZOL-1-YL]-N,2-DICYCLOHEXYLETHANAMIDE",
            "non-polymer",
        ),
        "SCV": (
            "N6-[(1S)-2-{[(1R)-1-CARBOXY-2-METHYLPROPYL]OXY}-1-(MERCAPTOCARBONYL)-2-OXOETHYL]-6-OXO-L-LYSINE",
            "non-polymer",
        ),
        "QUN": ("QUINACRINE", "non-polymer"),
        "CH1": ("3'-DEOXY-CYTIDINE-5'-TRIPHOSPHATE", "non-polymer"),
        "SAY": (
            "[(2R,3S,5R)-3-HYDROXY-5-(3-HYDROXY-4-METHANOYL-PHENYL)OXOLAN-2-YL]METHYL DIHYDROGEN PHOSPHATE",
            "nucleic-acid",
        ),
        "TPS": ("THIAMIN PHOSPHATE", "non-polymer"),
        "P38": (
            "5-METHYL-4-[(2-METHYL-5-{[(2-MORPHOLIN-4-YLPYRIDIN-4-YL)CARBONYL]AMINO}PHENYL)AMINO]-N-(1-PHENYLETHENYL)PYRROLO[2,1-F][1,2,4]TRIAZINE-6-CARBOXAMIDE",
            "non-polymer",
        ),
        "CH3": ("METHYL GROUP", "non-polymer"),
        "P32": (
            "4-{[(2S)-3-(TERT-BUTYLAMINO)-2-HYDROXYPROPYL]OXY}-3H-INDOLE-2-CARBONITRILE",
            "non-polymer",
        ),
        "CH4": (
            "3-[(2-AMINOETHYL)AMINO]-2-{[(2-AMINOETHYL)AMINO]METHYL}PROPANAL",
            "non-polymer",
        ),
        "MZ1": (
            "N~2~-ACETYL-N-[(1S,2R)-1-BENZYL-2-HYDROXY-3-{(2-THIENYLMETHYL)[(2,4,5-TRIFLUOROPHENYL)SULFONYL]AMINO}PROPYL]-L-ALANINAMIDE",
            "non-polymer",
        ),
        "P37": (
            "N-CYCLOPROPYL-3-{[1-(2,4-DIFLUOROPHENYL)-7-METHYL-6-OXO-6,7-DIHYDRO-1H-PYRAZOLO[3,4-B]PYRIDIN-4-YL]AMINO}-4-METHYLBENZAMIDE",
            "non-polymer",
        ),
        "P34": (
            "N~2~,N~2~-DIMETHYL-N~1~-(6-OXO-5,6-DIHYDROPHENANTHRIDIN-2-YL)GLYCINAMIDE",
            "non-polymer",
        ),
        "P3N": (
            "N-[(2S)-3-(3-TERT-BUTYL-1,2,4-OXADIAZOL-5-YL)-1-({(2S)-1-[(4-METHYLBENZYL)AMINO]-1-OXO-4-PHENYLBUTAN-2-YL}AMINO)-1-OXOPROPAN-2-YL]-5-METHYL-1,2-OXAZOLE-3-CARBOXAMIDE",
            "peptide",
        ),
        "CH6": (
            "{(4Z)-2-[(1S)-1-AMINO-3-(METHYLSULFANYL)PROPYL]-4-[(4-HYDROXYPHENYL)METHYLIDENE]-5-OXO-4,5-DIHYDRO-1H-IMIDAZOL-1-YL}ACETIC ACID",
            "peptide",
        ),
        "P3M": ("1-3 SUGAR RING OF PENTAMANNOSYL 6-PHOSPHATE", "saccharide"),
        "CH7": (
            "[(4Z)-4-(4-HYDROXYBENZYLIDENE)-5-OXO-2-(3,4,5,6-TETRAHYDROPYRIDIN-2-YL)-4,5-DIHYDRO-1H-IMIDAZOL-1-YL]ACETIC ACID",
            "peptide",
        ),
        "P3F": (
            "PHOSPHORIC ACID MONO-(5-HYDROXY-6-METHYL-4-{[2-(2-TRIFLUOROMETHYL-BENZOYLAMINO)-ACETYL]-HYDRAZONOMETHYL}-PYRIDIN-3-YLMETHYL)ESTER",
            "non-polymer",
        ),
        "P3G": ("3,6,9,12,15-PENTAOXAHEPTADECANE", "non-polymer"),
        "P3D": (
            "(4-{(E)-[(5-AMINOPENTYL)IMINO]METHYL}-5-HYDROXY-6-METHYLPYRIDIN-3-YL)METHYL DIHYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "P3Z": (
            "(S)-CYCLOPROPYL(4-METHOXYPHENYL)PYRIMIDIN-5-YLMETHANOL",
            "non-polymer",
        ),
        "SAS": (
            "2-HYDROXY-(5-([4-(2-PYRIDINYLAMINO)SULFONYL]PHENYL)AZO)BENZOIC ACID",
            "non-polymer",
        ),
        "NG1": ("N-ACETYL-ALPHA-D-GALACTOSAMINE 1-PHOSPHATE", "saccharide"),
        "P3S": ("L-METHIONINE-S-SULFOXIMINE PHOSPHATE", "non-polymer"),
        "P3P": (
            "(2S)-2-AMINO-4-[METHYL(PHOSPHONOOXY)PHOSPHORYL]BUTANOIC ACID",
            "non-polymer",
        ),
        "P3Q": (
            "2-HYDROXY-5-{[(1E)-1-(2-OXOETHYLIDENE)-4-PHENYLBUTYL]AMINO}-L-TYROSINE",
            "peptide",
        ),
        "P3T": (
            "(4-{[(4-AMINOBUTYL)AMINO]METHYL}-5-HYDROXY-6-METHYLPYRIDIN-3-YL)METHYL DIHYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "TDX": ("THYMIDINE-5'-DIPHOSPHO-BETA-D-XYLOSE", "non-polymer"),
        "5SE": ("?", "nucleic-acid"),
        "5SD": ("5ALPHA-ANDROSTAN-3,17-DIONE", "non-polymer"),
        "5SC": (
            "3-((3-BROMO-5-O-TOLYLPYRAZOLO[1,5-A]PYRIMIDIN-7-YLAMINO)METHYL)PYRIDINE 1-OXIDE",
            "non-polymer",
        ),
        "5SA": ("ACARBOSE DERIVED PENTASACCHARIDE", "saccharide"),
        "TIN": ("?", "non-polymer"),
        "5SI": (
            "2-{2-DEOXY-5-O-[(R)-HYDROXY{[(R)-HYDROXY(PHOSPHONOOXY)PHOSPHORYL]OXY}PHOSPHORYL]-BETA-D-ERYTHRO-PENTOFURANOSYL}-6-METHYLISOQUINOLINE-1(2H)-THIONE",
            "non-polymer",
        ),
        "L0B": ("ALPHA-LOBELINE", "non-polymer"),
        "5SP": ("5-O-PHOSPHONO-D-XYLULOSE", "saccharide"),
        "SEH": ("S-BENZYL PHENYLACETOTHIOHYDROXIMATE-O-SULFATE", "non-polymer"),
        "Z01": ("3-[(2,2-DIMETHYLPROPANOYL)AMINO]-L-ALANINE", "peptide"),
        "Z00": (
            "4-[2-[[3-CHLORO-5-(TRIFLUOROMETHYL)PYRIDIN-2-YL]AMINO]ETHYL]-N-(5-PROPAN-2-YL-1,3,4-THIADIAZOL-2-YL)BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "Z02": ("5-NITRO-2-[(PYRIDIN-2-YLMETHYL)AMINO]BENZAMIDE", "non-polymer"),
        "Z04": (
            "2-{[(2-AMINOPYRIMIDIN-5-YL)METHYL]AMINO}-4-(MORPHOLIN-4-YL)-5-NITROBENZAMIDE",
            "non-polymer",
        ),
        "Z06": ("(3-ACETAMIDO-5-CARBOXY-PHENYL)-TRIHYDROXY-BORON", "non-polymer"),
        "Y27": (
            "(R)-TRANS-4-(1-AMINOETHYL)-N-(4-PYRIDYL) CYCLOHEXANECARBOXAMIDE",
            "non-polymer",
        ),
        "CHD": ("CHOLIC ACID", "non-polymer"),
        "WWZ": (
            "4-{[(4-FLUOROPHENYL)CARBAMOYL]AMINO}BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "SCR": ("SUCROSE OCTASULFATE", "saccharide"),
        "RPF": (
            "1-{4-[3-(2-METHOXY-BENZYLOXY)-PROPOXY]-PHENYL}-6-(1,2,,3,4-TETRAHYDRO-QUINOLIN-7-YLOXYMETHYL)-PIPERAZIN-2-ONE",
            "non-polymer",
        ),
        "QG8": ("?", "non-polymer"),
        "PLU": ("LEUCINE PHOSPHONIC ACID", "non-polymer"),
        "MO2": ("MAGNESIUM ION, 2 WATERS COORDINATED", "non-polymer"),
        "CHF": ("CYCLOHEXYLFLUOROSTATONE", "peptide"),
        "Q21": (
            "(2S)-2-(NAPHTHALEN-1-YLOXY)-N-[2-(PYRIDIN-4-YL)-1,3-BENZOXAZOL-5-YL]PROPANAMIDE",
            "non-polymer",
        ),
        "X07": ("5-NITRO-2-[(PYRIDIN-3-YLMETHYL)AMINO]BENZAMIDE", "non-polymer"),
        "WWF": ("C2-1-HYDROXY-3-METHYL-BUTYL-THIAMIN", "non-polymer"),
        "X05": ("AMINO-CASTANOSPERMINE", "non-polymer"),
        "MG1": ("?", "nucleic-acid"),
        "L06": (
            "5-{[AMINO(IMINO)METHYL]AMINO}-2-(SULFANYLMETHYL)PENTANOIC ACID",
            "non-polymer",
        ),
        "X04": (
            "(2-AMINO-4-FORMYL-5-HYDROXY-6-METHYLPYRIDIN-3-YL)METHYL DIHYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "VD1": (
            "5-{2-[1-(1-METHYL-PROPYL)-7A-METHYL-OCTAHYDRO-INDEN-4-YLIDENE]-ETHYLIDENE}-2-METHYLENE-CYCLOHEXANE-1,3-DIOL",
            "non-polymer",
        ),
        "TN5": ("2-(4-AMINO-2-CHLOROPHENOXY)-5-CHLOROPHENOL", "non-polymer"),
        "2UN": (
            "{4-[2-BENZYL-3-METHOXY-2-(METHOXYCARBONYL)-3-OXOPROPYL]PHENYL}SULFAMIC ACID",
            "non-polymer",
        ),
        "TN4": (
            "(1R,2S,3S,4R)-4-HYDROXY-2-METHOXY-4-METHYL-3-[(2R,3R)-2-METHYL-3-(3-METHYLBUT-2-EN-1-YL)OXIRAN-2-YL]CYCLOHEXYL (CHLOROACETYL)CARBAMATE",
            "non-polymer",
        ),
        "JAR": (
            "6-[(3S,4S)-1-BENZYL-4-METHYLPYRROLIDIN-3-YL]-1-(1-METHYLETHYL)-1,5-DIHYDRO-4H-PYRAZOLO[3,4-D]PYRIMIDIN-4-ONE",
            "non-polymer",
        ),
        "2UA": (
            "(3S)-3-[(2R)-2-AMINO-5-CYCLOPROPYL-5-OXOPENTYL]PYRROLIDIN-2-ONE",
            "peptide",
        ),
        "TN7": ("ETHYL HYDROGEN ETHYLAMIDOPHOSPHATE", "non-polymer"),
        "JAS": ("JASPISAMIDE A", "non-polymer"),
        "TN6": ("DIETHYLPHOSPHORAMIDIC ACID", "non-polymer"),
        "LAN": ("LANOSTEROL", "non-polymer"),
        "LAT": ("BETA-LACTOSE", "saccharide"),
        "2UP": ("NAPHTHALENE-2-CARBOXIMIDAMIDE", "non-polymer"),
        "XY3": (
            "5-[(5-{[4-{[4-(2-HYDROXYETHYL)PIPERAZIN-1-YL]METHYL}-3-(TRIFLUOROMETHYL)PHENYL]CARBAMOYL}-2-METHYLPHENYL)ETHYNYL]-1-METHYL-1H-IMIDAZOLE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "JAN": (
            "6-[(S)-AMINO(4-CHLOROPHENYL)(1-METHYL-1H-IMIDAZOL-5-YL)METHYL]-4-(3-CHLOROPHENYL)-1-METHYLQUINOLIN-2(1H)-ONE",
            "non-polymer",
        ),
        "3IP": ("3-(BENZYLOXY)PYRIDIN-2-AMINE", "non-polymer"),
        "TN3": (
            "3-CHLORO-4-(4-CHLORO-2-HYDROXYPHENOXY)-N-METHYLBENZAMIDE",
            "non-polymer",
        ),
        "TN2": ("5-CHLORO-2-(2-CHLORO-4-NITROPHENOXY)PHENOL", "non-polymer"),
        "SA9": ("2-(HYDROXYMETHYL)PHENOL", "non-polymer"),
        "3IB": ("3-INDOLEBUTYRIC ACID", "non-polymer"),
        "CHS": ("4-AMINO-5-CYCLOHEXYL-3-HYDROXY-PENTANOIC ACID", "peptide"),
        "3IG": (
            "N-{2-[6-(2,4-DIAMINO-6-ETHYLPYRIMIDIN-5-YL)-2,2-DIMETHYL-3-OXO-2,3-DIHYDRO-4H-1,4-BENZOXAZIN-4-YL]ETHYL}ACETAMIDE",
            "non-polymer",
        ),
        "3ID": ("3H-INDOLE-5,6-DIOL", "non-polymer"),
        "Z0A": ("(3S)-3-[(2R)-2-AMINO-5-OXOHEXYL]PYRROLIDIN-2-ONE", "non-polymer"),
        "9JZ": (
            "(3S)-1-{(2S)-2-CYCLOHEXYL-2-[(N-METHYL-L-ALANYL)AMINO]ACETYL}-3-METHYL-N-(2-PYRIMIDIN-2-YLPHENYL)-L-PROLINAMIDE",
            "non-polymer",
        ),
        "3IH": (
            "N-[(5-METHYLISOXAZOL-3-YL)CARBONYL]ALANYL-L-VALYL-N~1~-((1R,2Z)-4-(BENZYLOXY)-4-OXO-1-{[(3R)-2-OXOPYRROLIDIN-3-YL]METHYL}BUT-2-ENYL)-L-LEUCINAMIDE",
            "non-polymer",
        ),
        "3IO": ("3-(1H-INDOL-3-YL)-2-OXOPROPANOIC ACID", "non-polymer"),
        "3IN": (
            "N-[2(S)-CYCLOPENTYL-1(R)-HYDROXY-3(R)METHYL]-5-[(2(S)-TERTIARY-BUTYLAMINO-CARBONYL)-4-(N1-(2)-(N-METHYLPIPERAZINYL)-3-CHLORO-PYRAZINYL-5-CARBONYL)-PIPERAZINO]-4(S)-HYDROXY-2(R)-PHENYLMETHYL-PENTANAMIDE",
            "non-polymer",
        ),
        "3IL": ("3-(INDOL-3-YL) LACTATE", "non-polymer"),
        "XJG": (
            "1,3-DIHYDROISOINDOL-2-YL-(2-HYDROXY-4-METHOXY-5-PROPAN-2-YL-PHENYL)METHANONE",
            "non-polymer",
        ),
        "ROC": (
            "(2S)-N-[(2S,3R)-4-[(2S,3S,4AS,8AS)-3-(TERT-BUTYLCARBAMOYL)-3,4,4A,5,6,7,8,8A-OCTAHYDRO-1H-ISOQUINOLIN-2-YL]-3-HYDROXY-1-PHENYL-BUTAN-2-YL]-2-(QUINOLIN-2-YLCARBONYLAMINO)BUTANEDIAMIDE",
            "peptide",
        ),
        "1G8": ("4-(AMINOMETHYL)-D-PHENYLALANINE", "peptide"),
        "1G6": (
            "N3PHE-PHE(4-NH2CH2)-LEU-PHE(4-NH2CH2)-METHYL VINYL SULFONE, BOUND FORM",
            "peptide",
        ),
        "1G5": (
            "HMB-VAL-SER-PHE(4-NH2CH2)-METHYL VINYL SULFONE, BOUND FORM",
            "peptide",
        ),
        "JAI": (
            "N-({(1R,2R)-3-OXO-2-[(2Z)-PENT-2-EN-1-YL]CYCLOPENTYL}ACETYL)-L-ISOLEUCINE",
            "non-polymer",
        ),
        "1G3": (
            "(2S)-1-[4-(AMINOMETHYL)PHENYL]-4-(METHYLSULFONYL)BUTAN-2-AMINE",
            "peptide",
        ),
        "1G2": ("(2S)-2-AZIDO-3-PHENYLPROPANOIC ACID", "peptide"),
        "1G1": (
            "N3PHE-LEU-LEU-PHE(4-NH2CH2)-METHYL VINYL SULFONE, BOUND FORM",
            "peptide",
        ),
        "1G0": (
            "CIS-3-[8-AMINO-1-(4-PHENOXYPHENYL)IMIDAZO[1,5-A]PYRAZIN-3-YL]CYCLOBUTANOL",
            "non-polymer",
        ),
        "SA3": (
            "2,10-DIHYDROXY-12-(BETA-D-GLUCOPYRANOSYL)-6,7,12,13-TETRAHYDROINDOLO[2,3-A]PYRROLO[3,4-C]CARBAZOLE-5,7-DIONE",
            "non-polymer",
        ),
        "I6P": ("INOSITOL 1,2,3,4,5,6-HEXAKISPHOSPHATE", "non-polymer"),
        "LAV": ("N-[(2S)-2-AMINO-4-METHYLPENTYL]-L-VALINE", "peptide"),
        "XJS": ("?", "non-polymer"),
        "ZIO": ("OLEANDOMYCIN", "non-polymer"),
        "I6X": (
            "5-{(1R)-3-AMINO-4-FLUORO-1-[3-(5-PROP-1-YN-1-YLPYRIDIN-3-YL)PHENYL]-1H-ISOINDOL-1-YL}-1-ETHYL-3-METHYLPYRIDIN-2(1H)-ONE",
            "non-polymer",
        ),
        "SA1": (
            "(3AR,6R,6AS)-6-((S)-((S)-CYCLOHEX-2-ENYL)(HYDROXY)METHYL)-6A-METHYL-4-OXO-HEXAHYDRO-2H-FURO[3,2-C]PYRROLE-6-CARBALDEHYDE",
            "non-polymer",
        ),
        "0W6": ("(4S)-4-AMINOPENTANOIC ACID", "peptide"),
        "0W7": (
            "2-[4-(MORPHOLIN-4-YL)-6-OXO-1,6-DIHYDROPYRIMIDIN-2-YL]-N-PHENYLACETAMIDE",
            "non-polymer",
        ),
        "0W5": ("(2E,4E)-11-METHYLDODECA-2,4-DIENOIC ACID", "non-polymer"),
        "YDJ": (
            "5-(3-FLUOROPHENYL)-N-[(3S)-3-PIPERIDYL]-3-UREIDO-THIOPHENE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "0W3": ("IONOMYCIN", "non-polymer"),
        "0W0": (
            "(4R,5S,8Z)-8-(2-AMINO-1,3-THIAZOL-4-YL)-1-[3-(1,5-DIHYDROXY-4-OXO-1,4-DIHYDROPYRIDIN-2-YL)-1,2-OXAZOL-5-YL]-5-FORMYL-11,11-DIMETHYL-1,7-DIOXO-4-(SULFOAMINO)-10-OXA-2,6,9-TRIAZADODEC-8-EN-12-OIC ACID",
            "non-polymer",
        ),
        "YDI": (
            "5-[4-(2-DIMETHYLAMINOETHYLOXY)PHENYL]-2-UREIDO-THIOPHENE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "0W8": ("1-ETHYL-1,3-DIHYDRO-2H-BENZIMIDAZOL-2-ONE", "non-polymer"),
        "SA6": (
            "(2R,3S,4R)-2-[(S)-(1S)-CYCLOHEX-2-EN-1-YL(HYDROXY)METHYL]-4-ETHYL-3-HYDROXY-3-METHYL-5-OXOPYRROLIDINE-2-CARBALDEHYDE",
            "non-polymer",
        ),
        "YDR": (
            "6-S-ALPHA-D-MANNOPYRANOSYL-6-THIO-ALPHA-D-MANNOPYRANOSE",
            "non-polymer",
        ),
        "YDP": (
            "(3S,4S),-3,4-BIS-[(4-CARBAMOYL-BENZENSULFONYL)-(3-METHYL-BUT-2-ENYL)-AMINO]-PYRROLIDINE",
            "non-polymer",
        ),
        "0WV": ("2-METHYL-5-(PROP-1-EN-2-YL)PHENOL", "non-polymer"),
        "0WT": (
            "(1R,5S,6S)-3-[5-CHLORO-6-ETHYL-2-(PYRIMIDIN-5-YLOXY)-7H-PYRROLO[2,3-D]PYRIMIDIN-4-YL]-3-AZABICYCLO[3.1.0]HEXAN-6-AMINE",
            "non-polymer",
        ),
        "0WU": ("(5S)-2-METHYL-5-(PROP-1-EN-2-YL)CYCLOHEX-2-EN-1-ONE", "non-polymer"),
        "0WR": (
            "2-[1-({2-[2-(DIMETHYLAMINO)-1H-BENZIMIDAZOL-1-YL]-9-METHYL-6-(MORPHOLIN-4-YL)-9H-PURIN-8-YL}METHYL)PIPERIDIN-4-YL]PROPAN-2-OL",
            "non-polymer",
        ),
        "0WS": ("2-(1H-PYRAZOL-3-YL)PYRIDINE-4-CARBOXYLIC ACID", "non-polymer"),
        "0WP": (
            "3-{[3-(2-CYANOPROPAN-2-YL)BENZOYL]AMINO}-2,6-DIFLUORO-N-(3-METHOXY-2H-PYRAZOLO[3,4-B]PYRIDIN-5-YL)BENZAMIDE",
            "non-polymer",
        ),
        "0WZ": ("3-(1H-PYRAZOL-1-YL)-L-TYROSINE", "peptide"),
        "0WG": (
            "7-AMINO-6-(AMINOMETHYL)-5-(2-BROMOPHENYL)-1,3-DIMETHYLPYRIDO[2,3-D]PYRIMIDINE-2,4(1H,3H)-DIONE",
            "non-polymer",
        ),
        "0WD": ("?", "non-polymer"),
        "0WE": (
            "N-METHYL-N-[(3-METHYL-1-BENZOFURAN-2-YL)METHYL]-3-(7-OXO-5,6,7,8-TETRAHYDRO-1,8-NAPHTHYRIDIN-3-YL)PROPANAMIDE",
            "non-polymer",
        ),
        "0WB": (
            "3-{2-[(5-FLUORO-2-HYDROXYPHENYL)AMINO]-1,3-THIAZOL-4-YL}BENZONITRILE",
            "non-polymer",
        ),
        "0WC": (
            "(2R)-4-[1-(2-AMINO-5-CHLOROPYRIMIDIN-4-YL)-2,3-DIHYDRO-1H-INDOL-6-YL]-2-(1,3-THIAZOL-2-YL)BUT-3-YN-2-OL",
            "non-polymer",
        ),
        "0WA": (
            "4-FLUORO-2-{[4-(PYRIDIN-4-YL)-1,3-THIAZOL-2-YL]AMINO}PHENOL",
            "non-polymer",
        ),
        "0WN": (
            "N-{4-[(3-CHLORO-4-FLUOROPHENYL)AMINO]-7-[(3S)-TETRAHYDROFURAN-3-YLOXY]QUINAZOLIN-6-YL}-4-(DIMETHYLAMINO)BUTANAMIDE",
            "non-polymer",
        ),
        "0WO": (
            "(2R,4S)-2-[(R)-CARBOXY{[(5-METHYL-3-PHENYL-1,2-OXAZOL-4-YL)CARBONYL]AMINO}METHYL]-5,5-DIMETHYL-1,3-THIAZOLIDINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "T5M": (
            "2-TERT-BUTYL-4-(1,3-DIHYDRO-2H-ISOINDOL-2-YLCARBONYL)PHENOL",
            "non-polymer",
        ),
        "0WM": (
            "(2E)-N-{4-[(3-CHLORO-4-FLUOROPHENYL)AMINO]-7-[(3S)-TETRAHYDROFURAN-3-YLOXY]QUINAZOLIN-6-YL}-4-(DIMETHYLAMINO)BUT-2-ENAMIDE",
            "non-polymer",
        ),
        "0WH": (
            "1-[5-(4-AMINO-7-METHYL-7H-PYRROLO[2,3-D]PYRIMIDIN-5-YL)-2,3-DIHYDRO-1H-INDOL-1-YL]-2-[3-(TRIFLUOROMETHYL)PHENYL]ETHANONE",
            "non-polymer",
        ),
        "ROF": (
            "3-(CYCLOPROPYLMETHOXY)-N-(3,5-DICHLOROPYRIDIN-4-YL)-4-(DIFLUOROMETHOXY)BENZAMIDE",
            "non-polymer",
        ),
        "TE": ("TELLURIUM", "non-polymer"),
        "M5S": (
            "ALPHA-D-MANNOPYRANOSYL-(1->3)-[ALPHA-D-MANNOPYRANOSYL-(1->3)-[ALPHA-D-MANNOPYRANOSYL-(1->6)]-ALPHA-D-MANNOPYRANOSYL-(1->6)]-ALPHA-D-MANNOPYRANOSE",
            "non-polymer",
        ),
        "1GW": (
            "5-(2-FORMYLTHIEN-5-YL)-URIDINE-5'-DIPHOSPHATE-ALPHA-D-GALACTOSE",
            "non-polymer",
        ),
        "1GV": ("4,4'-THIENE-2,5-DIYLBIS(3-METHYLPHENOL)", "non-polymer"),
        "1GU": (
            "4-[2-(BUT-3-EN-1-YL)-7-(TRIFLUOROMETHYL)-2H-INDAZOL-3-YL]BENZENE-1,3-DIOL",
            "non-polymer",
        ),
        "1GT": (
            "4-[1-(BUT-3-EN-1-YL)-7-(TRIFLUOROMETHYL)-1H-INDAZOL-3-YL]BENZENE-1,3-DIOL",
            "non-polymer",
        ),
        "1GS": (
            "4-[2-(2-METHYLPROPYL)-7-(TRIFLUOROMETHYL)-2H-INDAZOL-3-YL]BENZENE-1,3-DIOL",
            "non-polymer",
        ),
        "1GR": (
            "4-[1-(2-METHYLPROPYL)-7-(TRIFLUOROMETHYL)-1H-INDAZOL-3-YL]BENZENE-1,3-DIOL",
            "non-polymer",
        ),
        "1GQ": (
            "4-[1-BUTYL-7-(TRIFLUOROMETHYL)-1H-INDAZOL-3-YL]BENZENE-1,3-DIOL",
            "non-polymer",
        ),
        "1GP": ("SN-GLYCEROL-1-PHOSPHATE", "non-polymer"),
        "1GN": ("2-DEOXY-2-AMINOGALACTOSE", "saccharide"),
        "1GM": (
            "4-[2-ETHYL-7-(TRIFLUOROMETHYL)-2H-INDAZOL-3-YL]BENZENE-1,3-DIOL",
            "non-polymer",
        ),
        "1GL": ("4-O-METHYL-2,6-DIDEOXY-ALPHA-D-GALACTO-HEXOPYRANOSE", "saccharide"),
        "1GK": (
            "N-(FURAN-2-YLMETHYL)-4-[(2-METHYLPHENYL)CARBONYL]-1H-PYRROLE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "1GJ": (
            "4-[2-BENZYL-7-(TRIFLUOROMETHYL)-2H-INDAZOL-3-YL]BENZENE-1,3-DIOL",
            "non-polymer",
        ),
        "1GH": (
            "7-(3,5-DIMETHYL-1,2-OXAZOL-4-YL)-8-METHOXY-1-[(1R)-1-(PYRIDIN-2-YL)ETHYL]-1H,2H,3H-IMIDAZO[4,5-C]QUINOLIN-2-ONE",
            "non-polymer",
        ),
        "1GG": (
            "(6S)-N-(4-CARBAMIMIDOYLBENZYL)-1-CHLORO-3-(CYCLOBUTYLAMINO)-8,8-DIETHYL-4-OXO-4,6,7,8-TETRAHYDROPYRROLO[1,2-A]PYRAZINE-6-CARBOXAMIDE",
            "non-polymer",
        ),
        "1GE": ("?", "non-polymer"),
        "1GC": (
            "2'-DEOXY-5'-O-[(R)-HYDROXY{[(S)-HYDROXY(PHOSPHONOOXY)PHOSPHORYL]METHYL}PHOSPHORYL]GUANOSINE",
            "non-polymer",
        ),
        "1GA": ("GERANILINE", "non-polymer"),
        "84T": (
            "[(2S,4S,5R)-5-(6-AMINOPURIN-9-YL)-4-OXIDANYL-OXOLAN-2-YL]METHOXY-N-[(2S,3R)-4-METHYL-2,3-BIS(OXIDANYL)PENTANOYL]PHOSPHONAMIDIC ACID",
            "non-polymer",
        ),
        "R7U": ("RUTHENIUM WIRE, 7 CARBON LINKER", "non-polymer"),
        "GVK": (
            "6-{4-[4-(4-CHLOROPHENYL)PIPERIDIN-4-YL]PHENYL}-9H-PURINE",
            "non-polymer",
        ),
        "VDY": (
            "3-{2-[1-(5-HYDROXY-1,5-DIMETHYL-HEXYL)-7A-METHYL-OCTAHYDRO-INDEN-4-YLIDENE]-ETHYLIDENE}-4-METHYLENE-CYCLOHEXANOL",
            "non-polymer",
        ),
        "VDX": (
            "5-{2-[1-(5-HYDROXY-1,5-DIMETHYL-HEXYL)-7A-METHYL-OCTAHYDRO-INDEN-4-YLIDENE]-ETHYLIDENE}-4-METHYLENE-CYCLOHEXANE-1,3-DIOL",
            "non-polymer",
        ),
        "VDZ": (
            "5-{2-[1-(5-HYDROXY-1,5-DIMETHYL-HEXYL)-7A-METHYL-OCTAHYDRO-INDEN-4-YLIDENE]-ETHYLIDENE}-2-METHYLENE-CYCLOHEXANE-1,3-DIOL",
            "non-polymer",
        ),
        "LAR": ("LATRUNCULIN A", "non-polymer"),
        "VDA": (
            "(1R,2R,3R,5Z)-17-{(1S)-1-[(2-ETHYL-2-HYDROXYBUTYL)SULFANYL]ETHYL}-2-(2-HYDROXYETHOXY)-9,10-SECOESTRA-5,7,16-TRIENE-1,3-DIOL",
            "non-polymer",
        ),
        "TNV": (
            "[2-(6-AMINO-9H-PURIN-9-YL)-1-METHYLETHOXY]METHYL-TRIPHOSPHATE",
            "non-polymer",
        ),
        "84A": (
            "P-(5-FERROCENYL-1H-1,2,3-TRIAZOL-1-YL)BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "GVN": (
            "(2R)-2-(4-CHLOROPHENYL)-2-[4-(1H-PYRAZOL-4-YL)PHENYL]ETHANAMINE",
            "non-polymer",
        ),
        "ROX": ("ROXITHROMYCIN", "non-polymer"),
        "VDM": (
            "(1S,2S,3R,6S)-4-(HYDROXYMETHYL)-6-{[(1S,2S,3S,4R,5R)-2,3,4-TRIHYDROXY-5-(HYDROXYMETHYL)CYCLOHEXYL]AMINO}CYCLOHEX-4-ENE-1,2,3-TRIOL",
            "non-polymer",
        ),
        "VDL": ("(2R,3R)-2,3-DIAMINOBUTANOIC ACID", "peptide"),
        "VDO": (
            "[(1R,2R,3S,4S,5S)-2,3,4-TRIHYDROXY-5-{[(1S,4R,5S,6S)-4,5,6-TRIHYDROXY-3-(HYDROXYMETHYL)CYCLOHEX-2-EN-1-YL]AMINO}CYCLOHEXYL]METHYL DIHYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "GVO": (
            "(2S)-2-(4-CHLOROPHENYL)-2-[4-(1H-PYRAZOL-4-YL)PHENYL]ETHANAMINE",
            "non-polymer",
        ),
        "O": ("OXYGEN ATOM", "non-polymer"),
        "TNS": (
            "5-[{4-[(E)-2-(4-NITROPHENYL)VINYL]PHENYL}(METHYL)AMINO]PENTANOIC ACID",
            "non-polymer",
        ),
        "NM2": ("3-CARBOXY-N,N,N-TRIMETHYLPROPAN-1-AMINIUM", "non-polymer"),
        "GVM": ("(3R)-3-HYDROXY-5,5-DIMETHYLHEXANOIC ACID", "non-polymer"),
        "MG8": ("N-OCTANOYL-N-METHYLGLUCAMINE", "non-polymer"),
        "PXM": (
            "4-(AMINOMETHYL)-5-(HYDROXYMETHYL)-2-METHYLPYRIDIN-3-OL",
            "non-polymer",
        ),
        "TNY": (
            "(2S,3R)-3-{[(2S,3R,4R,5R,6R)-3-(ACETYLAMINO)-4,5-DIHYDROXY-6-(HYDROXYMETHYL)TETRAHYDRO-2H-PYRAN-2-YL]OXY}-2-AMINOBUTANOIC ACID (NON-PREFERRED NAME)",
            "peptide",
        ),
        "WV7": (
            "2-(TRIFLUOROACETYL)-1,2,3,4-TETRAHYDROISOQUINOLIN-6-OL",
            "non-polymer",
        ),
        "GVG": ("3-METHYL-4-PHENYL-1H-PYRAZOLE", "non-polymer"),
        "M5M": ("2'-(N-ACETAMIDE)-CYTIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "T5A": ("P1-(5'-ADENOSYL)P5-(5'-THYMIDYL)PENTAPHOSPHATE", "non-polymer"),
        "847": (
            "2-(3-{5-[AMINO(IMINIO)METHYL]-1H-BENZIMIDAZOL-2-YL}-5-BROMO-4-OXIDOPHENYL)SUCCINATE",
            "non-polymer",
        ),
        "846": (
            "[4R--(1ALPHA,5ALPHA,7BETA)]-3-[(CYCLOPROPHYLMETHYL)HEXAHYDRO-5,6-DIHYDROXY-2-OXO-4,7-BIS(PHENYLMETHYL)-1H-1,3-DIAZEPIN]METHYL-2-THIAZOLYLBENZAMIDE",
            "non-polymer",
        ),
        "845": (
            "N-(3-{5-[5-(1H-INDOL-4-YL)-1,3,4-OXADIAZOL-2-YL]-3-(TRIFLUOROMETHYL)-1H-PYRAZOL-1-YL}BENZYL)-L-ALANINAMIDE",
            "non-polymer",
        ),
        "VD2": (
            "5-{2-[1-(5-HYDROXY-1,5-DIMETHYL-HEXYL)-7A-METHYL-OCTAHYDRO-INDEN-4-YLIDENE]-ETHYLIDENE}-2-METHYL-CYCLOHEXANE-1,3-DIOL",
            "non-polymer",
        ),
        "843": (
            "N-ACETYL-N-[1-(1,1'-BIPHENYL-4-YLMETHYL)-2-OXOAZEPAN-3-YL]-O-PHOSPHONOTYROSINAMIDE",
            "non-polymer",
        ),
        "842": (
            "N-[(1S,2R)-1-(3,5-DIFLUOROBENZYL)-2-HYDROXY-3-({1-[3-(1H-PYRAZOL-1-YL)PHENYL]CYCLOHEXYL}AMINO)PROPYL]ACETAMIDE",
            "non-polymer",
        ),
        "840": ("PLATENSIMYCIN A1", "non-polymer"),
        "HSM": ("HISTAMINE", "non-polymer"),
        "849": (
            "2-{4-[3-FLUORO-2-(2-METHOXYPHENYL)-1H-INDOL-5-YL] PIPERIDIN-1-YL}-N-METHYLETHANAMINE",
            "non-polymer",
        ),
        "848": ("2-(PYRIDIN-3-YLMETHOXY)ANILINE", "non-polymer"),
        "KHA": (
            "1-(4-{[(2-HYDROXYETHYL)AMINO]CARBONYL}BENZYL)-1-METHYLPIPERIDINIUM",
            "non-polymer",
        ),
        "TNF": ("PICRIC ACID", "non-polymer"),
        "TNA": (
            "N-[(3S)-3-{[(3S)-3-AMINO-3-CARBOXYPROPYL]AMINO}-3-CARBOXYPROPYL]-L-GLUTAMIC ACID",
            "non-polymer",
        ),
        "NM1": (
            "3-(7-BENZYL-4,5-DIHYDROXY-1,1-DIOXO-3,6-BIS-PHENOXYMETHYL-1L6-[1,2,7]THIADIAZEPAN-2-YLMETHYL)-N-METHYL-BENZAMIDE",
            "non-polymer",
        ),
        "T5C": (
            "2-{(1E,3Z,5E,7E)-7-[3,3-DIMETHYL-5-SULFO-1-(2-SULFOETHYL)-1,3-DIHYDRO-2H-INDOL-2-YLIDENE]-4-METHYLHEPTA-1,3,5-TRIEN-1-YL}-3,3-DIMETHYL-5-SULFO-1-(2-SULFOETHYL)-3H-INDOLIUM",
            "non-polymer",
        ),
        "TNM": (
            "[(2R)-1-(6-AMINOPURIN-9-YL)PROPAN-2-YL]OXYMETHYL-PHOSPHONOOXY-PHOSPHINIC ACID",
            "non-polymer",
        ),
        "YSH": (
            "1-[3-CYANO-4-(NEOPENTYLOXY)PHENYL]-1H-PYRAZOLE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "TNL": ("2,4,6-TRINITROTOLUENE", "non-polymer"),
        "ZUK": ("5-PYRIMIDIN-2-YL-D-NORVALINE", "peptide"),
        "ZUN": (
            "(4R)-N-[4-(DIHYDROXYAMINO)BENZYL]-4-HYDROXY-1-[(3-METHYLISOXAZOL-5-YL)ACETYL]-L-PROLINAMIDE",
            "non-polymer",
        ),
        "TXP": (
            "1,4,5,6-TETRAHYDRONICOTINAMIDE ADENINE DINUCLEOTIDE PHOSPHATE",
            "non-polymer",
        ),
        "ONO": (
            "2-[2-({[4-(DIAMINOMETHYL)PHENYL]AMINO}CARBONYL)-6-METHOXYPYRIDIN-3-YL]-5-{[(1-FORMYL-2,2-DIMETHYLPROPYL)AMINO]CARBONYL}BENZOIC ACID",
            "non-polymer",
        ),
        "RRH": ("?", "non-polymer"),
        "RRG": ("?", "non-polymer"),
        "RRC": ("R-ROSCOVITINE", "non-polymer"),
        "RRA": ("PROP-2-YN-1-AMINE", "non-polymer"),
        "VZV": (
            "1-{[(5R,6S)-5,6-BIS(4-CHLOROPHENYL)-6-METHYL-3-(PROPAN-2-YL)-5,6-DIHYDROIMIDAZO[2,1-B][1,3]THIAZOL-2-YL]CARBONYL}-N,N-DIMETHYL-L-PROLINAMIDE",
            "non-polymer",
        ),
        "Q17": ("2-{4-[(3-AMINOPROPYL)AMINO]QUINAZOLIN-2-YL}PHENOL", "non-polymer"),
        "P39": (
            "5-METHYL-4-[(2-METHYL-5-{[(3-MORPHOLIN-4-YLPHENYL)CARBONYL]AMINO}PHENYL)AMINO]-N-[(1S)-1-PHENYLETHYL]PYRROLO[2,1-F][1,2,4]TRIAZINE-6-CARBOXAMIDE",
            "non-polymer",
        ),
        "Q11": ("7-(2-METHOXYPHENYL)QUINAZOLINE-2,4-DIAMINE", "non-polymer"),
        "TNK": ("6-BENZYL-1-BENZYLOXYMETHYL-5-ISOPROPYL URACIL", "non-polymer"),
        "Q12": (
            "7-(3,4-DIMETHOXYPHENYL)-6-METHYLQUINAZOLINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "RRT": ("(R)-RETRO-THIORPHAN", "non-polymer"),
        "Q19": (
            "1-[3-(2,4-DIAMINO-6-METHYLQUINAZOLIN-7-YL)PHENYL]ETHANONE",
            "non-polymer",
        ),
        "Q18": (
            "(2Z)-7-(AZEPAN-1-YLMETHYL)-6-HYDROXY-2-(1H-INDOL-3-YLMETHYLIDENE)-1-BENZOFURAN-3(2H)-ONE",
            "non-polymer",
        ),
        "RRP": (
            "3-({4-[(6-CHLORO-1-BENZOTHIEN-2-YL)SULFONYL]-2-OXOPIPERAZIN-1-YL}METHYL)BENZENECARBOXIMIDAMIDE",
            "non-polymer",
        ),
        "K7J": (
            "N-(3-CARBOXYPROPANOYL)-L-ALANYL-L-ALANYL-N-[(2S,3S)-4-CHLORO-3-HYDROXY-1-PHENYLBUTAN-2-YL]-L-PROLINAMIDE",
            "non-polymer",
        ),
        "P33": ("3,6,9,12,15,18-HEXAOXAICOSANE-1,20-DIOL", "non-polymer"),
        "RR6": (
            "3-HYDROXY-7-(4-{1-[2-HYDROXY-3-(2-HYDROXY-5-SULFO-PHENYLAZO)-BENZYL]-2-SULFO-ETHYLAMINO}-[1,2,5]TRIAZIN-2-YLAMINO)-2-(2-HYDROXY-5-SULFO-PHENYLAZO)-NAPTHALENE-1,8-DISULFONIC ACID",
            "non-polymer",
        ),
        "RR1": (
            "5-(4,6-DIAMINO-[1,3,5]TRIAZIN-2-YLAMINO)-4-HYDROXY-3-(2-SULFO-PHENYLAZO)-NAPHTHALENE-2,7-DISULFONIC ACID",
            "non-polymer",
        ),
        "ZU3": (
            "N-[(BENZYLOXY)CARBONYL]-3-[(2,2-DIMETHYLPROPANOYL)AMINO]-L-ALANYL-N-[(1R)-4-OXO-1-{[(3S)-2-OXOPYRROLIDIN-3-YL]METHYL}PENTYL]-L-LEUCINAMIDE",
            "peptide",
        ),
        "ZU0": ("O-TERT-BUTYL-L-THREONINE", "peptide"),
        "ZU4": ("1-(2-CHLOROBENZYL)-4-HEXYLPYRIDIN-2(1H)-ONE", "non-polymer"),
        "ZU5": (
            "N-[(BENZYLOXY)CARBONYL]-O-TERT-BUTYL-L-THREONYL-N-[(1R)-4-CYCLOPROPYL-4-OXO-1-{[(3S)-2-OXOPYRROLIDIN-3-YL]METHYL}BUTYL]-L-LEUCINAMIDE",
            "peptide",
        ),
        "9UP": ("METHYL (7-CARBAMIMIDOYLNAPHTHALEN-1-YL)CARBAMATE", "non-polymer"),
        "SAN": ("SULFANILAMIDE", "non-polymer"),
        "USC": (
            "5-[(METHYLSULFONYL)AMINO]BENZENE-1,3-DICARBOXYLIC ACID",
            "non-polymer",
        ),
        "IQX": (
            "N-[5-(5-{[(2,4-DIMETHYL-1,3-THIAZOL-5-YL)SULFONYL]AMINO}-6-FLUOROPYRIDIN-3-YL)-4-METHYL-1,3-THIAZOL-2-YL]ACETAMIDE",
            "non-polymer",
        ),
        "IQZ": (
            "(12BS)-1,2,3,4,12,12B-HEXAHYDROINDOLO[2,3-A]QUINOLIZIN-7(6H)-ONE",
            "non-polymer",
        ),
        "IQU": ("N-(2-AMINOETHYL)ISOQUINOLINE-5-SULFONAMIDE", "non-polymer"),
        "USM": ("?", "nucleic-acid"),
        "ROP": ("PROPIONAMIDE", "non-polymer"),
        "IQP": ("1-(5-ISOQUINOLINESULFONYL)-2-METHYLPIPERAZINE", "non-polymer"),
        "IQS": ("N-[2-(METHYLAMINO)ETHYL]-5-ISOQUINOLINESULFONAMIDE", "non-polymer"),
        "PCS": ("PHENYLALANYLMETHYLCHLORIDE", "peptide"),
        "IQO": (
            "1-(1-(4-(7-PHENYL-1H-IMIDAZO[4,5-G]QUINOXALIN-6-YL)BENZYL)PIPERIDIN-4-YL)-1H-BENZO[D]IMIDAZOL-2(3H)-ONE",
            "non-polymer",
        ),
        "USQ": ("SULFOQUINOVOSE-URIDINE-C1,5'-DIPHOSPHATE", "non-polymer"),
        "XWG": (
            "(3AS,4R,5S,8AS,8BR)-4-[5-(5-CHLOROTHIOPHEN-2-YL)-1,2-OXAZOL-3-YL]-2-[3-[1-(2-HYDROXYETHYL)PYRROLIDIN-1-IUM-1-YL]PROPYL]-4,6,7,8,8A,8B-HEXAHYDRO-3AH-PYRROLO[3,4-A]PYRROLIZINE-1,3-DIONE",
            "non-polymer",
        ),
        "IQA": (
            "(5-OXO-5,6-DIHYDRO-INDOLO[1,2-A]QUINAZOLIN-7-YL)-ACETIC ACID",
            "non-polymer",
        ),
        "IQB": (
            "N-[2-(4-BROMOCINNAMYLAMINO)ETHYL]-5-ISOQUINOLINE SULFONAMIDE",
            "non-polymer",
        ),
        "PXZ": (
            "2-AMINO-1,9-DICARBONYL-4,6-DIMETHYL-10-DEHYDRO-PHENOXAZIN-3-ONE",
            "non-polymer",
        ),
        "PXY": ("PARA-XYLENE", "non-polymer"),
        "C0K": (
            "{(1S)-1-(3,4-DIFLUOROPHENYL)-4-[HYDROXY(METHYL)AMINO]-4-OXOBUTYL}PHOSPHONIC ACID",
            "non-polymer",
        ),
        "C0M": (
            "N-{3-CYANO-6-[3-(1-PIPERIDINYL)PROPANOYL]-4,5,6,7-TETRAHYDROTHIENO[2,3-C]PYRIDIN-2-YL}1-NAPHTHALENECARBOXAMIDE",
            "non-polymer",
        ),
        "PXS": ("(2S)-PROPANE-1,2-DIYL DIHEXADECANOATE", "non-polymer"),
        "48D": (
            "(4E,6S,7S,8S,9E)-7-HYDROXY-8-METHOXY-4,6-DIMETHYLCYCLOTETRADECA-4,9-DIEN-1-ONE",
            "non-polymer",
        ),
        "48B": (
            "2-CHLORO-N-[4-({5-FLUORO-2-[(4-HYDROXYPHENYL)AMINO]PYRIMIDIN-4-YL}AMINO)PHENYL]BENZAMIDE",
            "non-polymer",
        ),
        "PXU": ("2-HYDROXY-L-PROLINE", "peptide"),
        "PXT": ("PECTENOTOXIN-2", "non-polymer"),
        "FP5": ("PHENYLACETAMIDOMETHYL BORONIC ACID", "non-polymer"),
        "C0Y": (
            "(3-TERT-BUTYL-4-HYDROXYPHENYL)MORPHOLIN-4-YL-METHANONE",
            "non-polymer",
        ),
        "C0Z": (
            "6-CHLORO-N-[(3S)-1-[(1S)-1-DIMETHYLAMINO-2,3-DIHYDRO-1H-INDEN-5-YL]-2-OXO-PYRROLIDIN-3-YL]NAPHTHALENE-2-SULFONAMIDE",
            "non-polymer",
        ),
        "FP6": ("PYRROLIDINE-1-CARBALDEHYDE", "non-polymer"),
        "FP1": (
            "N-HYDROXY-1-(1-METHYLPYRIDIN-2(1H)-YLIDENE)METHANAMINE",
            "non-polymer",
        ),
        "48Z": (
            "2-DEOXY-2-FLUORO-5-O-THIOPHOSPHONO-ALPHA-D-ARABINOFURANOSE",
            "saccharide",
        ),
        "FP3": (
            "2-CHLORO-5-{[2-(PYRIMIDIN-2-YL)FURO[2,3-C]PYRIDIN-3-YL]AMINO}PHENOL",
            "non-polymer",
        ),
        "FP2": ("2-FLUOROPHENOL", "non-polymer"),
        "C0P": (
            "4-CHLORO-6-[2,4-DICHLORO-5-(2-MORPHOLIN-4-YLETHOXY)PHENYL]PYRIMIDIN-2-AMINE",
            "non-polymer",
        ),
        "MCY": ("5-METHYL-2'-DEOXYCYTIDINE", "nucleic-acid"),
        "C0R": ("CORTICOSTERONE", "non-polymer"),
        "48T": (
            "L-GAMMA-GLUTAMYL-S-(2-BIPHENYL-4-YL-2-OXOETHYL)-L-CYSTEINYLGLYCINE",
            "non-polymer",
        ),
        "C0T": ("CALICHEAMICIN T0", "non-polymer"),
        "PXF": (
            "2-AMINO-1,9-DICARBONYL-4,6-DIMETHYL-10-DEHYDRO-8-FLUORO-3-OXOPHENOXAZINE",
            "non-polymer",
        ),
        "PXE": (
            "2-METHYL-5-(1-METHYLETHYL)CYCLOHEXA-2,5-DIENE-1,4-DIONE 1-OXIME",
            "non-polymer",
        ),
        "FPE": ("3-FLUORO-2-(PHOSPHONOOXY)PROPANOIC ACID", "non-polymer"),
        "FPD": ("5A-(3-FORMYLPHENYLSULFANYL)-DIHYDROBICYCLOMYCIN", "non-polymer"),
        "FPG": (
            "(1S)-1-[(1R)-1-FLUOROETHYL]-1,5-DIMETHYLHEXYL TRIHYDROGEN DIPHOSPHATE",
            "non-polymer",
        ),
        "FPF": (
            "(2Z,6E)-2-FLUORO-3,7,11-TRIMETHYLDODECA-2,6,10-TRIEN-1-YL TRIHYDROGEN DIPHOSPHATE",
            "non-polymer",
        ),
        "FPA": ("BROMO(DIFLUORO)ACETIC ACID", "non-polymer"),
        "FPC": ("D-FRUCTOSE-6-PHOSPHATE (OPEN FORM)", "non-polymer"),
        "FPB": (
            "N-({(2S)-1-[(3R)-3-AMINO-4-(2-FLUOROPHENYL)BUTANOYL]PYRROLIDIN-2-YL}METHYL)BENZAMIDE",
            "non-polymer",
        ),
        "FPM": (
            "(5R,6S)-6-(1-HYDROXYETHYL)-7-OXO-3-[(2R)-OXOLAN-2-YL]-4-THIA-1-AZABICYCLO[3.2.0]HEPT-2-ENE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "PX2": ("1,2-DILAUROYL-SN-GLYCERO-3-PHOSPHATE", "non-polymer"),
        "FPO": ("FLUORO-PHOSPHITE ION", "non-polymer"),
        "FPN": ("4-FLUOROPHENOL", "non-polymer"),
        "FPI": ("N-FORMYLPIPERIDINE", "non-polymer"),
        "FPH": (
            "3-(4-FLUOROPHENYL)-2-PYRIDIN-4-YL-1H-PYRROLO[3,2-B]PYRIDIN-1-OL",
            "non-polymer",
        ),
        "PX5": (
            "4-(1,3-BENZOTHIAZOL-2-YL)-4-HYDROXYCYCLOHEXA-2,5-DIEN-1-ONE",
            "non-polymer",
        ),
        "PX4": ("1,2-DIMYRISTOYL-SN-GLYCERO-3-PHOSPHOCHOLINE", "non-polymer"),
        "C08": (
            "DIFLUORO(5-{2-[(5-OCTYL-1H-PYRROL-2-YL-KAPPAN)METHYLIDENE]-2H-PYRROL-5-YL-KAPPAN}PENTANOATO)BORON",
            "non-polymer",
        ),
        "C09": ("2-(ETHOXYMETHYL)-1H-IMIDAZO[4,5-C]QUINOLIN-4-AMINE", "non-polymer"),
        "FPW": (
            "1-BENZYL-N-{5-[(6,7-DIMETHOXYQUINOLIN-4-YL)OXY]PYRIDIN-2-YL}-2-OXO-1,2-DIHYDROPYRIDINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "6WO": ("OXO-TUNGSTEN(VI)", "non-polymer"),
        "FPQ": (
            "(2R)-3-(PHOSPHONOOXY)-2-{[(2Z,6E)-3,7,11-TRIMETHYLDODECA-2,6,10-TRIEN-1-YL]OXY}PROPANOIC ACID",
            "non-polymer",
        ),
        "FPP": ("FARNESYL DIPHOSPHATE", "non-polymer"),
        "FPS": (
            "S-[(2E,6E)-3,7,11-TRIMETHYLDODECA-2,6,10-TRIENYL] TRIHYDROGEN THIODIPHOSPHATE",
            "non-polymer",
        ),
        "488": (
            "(2R,4R)-2-[(4R)-2-(2-HYDROXYPHENYL)-4,5-DIHYDRO-1,3-THIAZOL-4-YL]-3-METHYL-1,3-THIAZOLIDINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "C00": (
            "2-(4-HYDROXYBIPHENYL-3-YL)-4-METHYL-1H-ISOINDOLE-1,3(2H)-DIONE",
            "non-polymer",
        ),
        "486": (
            "11-(4-DIMETHYLAMINO-PHENYL)-17-HYDROXY-13-METHYL-17-PROP-1-YNYL-1,2,6,7,8,11,12,13,14,15,16,17-DODEC AHYDRO-CYCLOPENTA[A]PHENANTHREN-3-ONE",
            "non-polymer",
        ),
        "C02": (
            "4-(5-BENZENESULFONYLAMINO-1-METHYL-1H-BENZOIMIDAZOL-2-YLMETHYL)-BENZAMIDINE",
            "non-polymer",
        ),
        "C03": (
            "(2R)-2-(4-CHLORO-3-{[3-(6-METHOXY-1,2-BENZISOXAZOL-3-YL)-2-METHYL-6-(TRIFLUOROMETHOXY)-1H-INDOL-1-YL]METHYL}PHENOXY)PROPANOIC ACID",
            "non-polymer",
        ),
        "FPY": (
            "(4S)-5-FLUORO-4-HYDROXY-3,4-DIHYDROPYRIMIDIN-2(1H)-ONE",
            "non-polymer",
        ),
        "FPX": (
            "N-{4-METHYL-3-[8-METHYL-7-OXO-2-(PHENYLAMINO)-7,8-DIHYDROPYRIDO[2,3-D]PYRIMIDIN-6-YL]PHENYL}-3-(TRIFLUOROMETHYL)BENZAMIDE",
            "non-polymer",
        ),
        "C07": (
            "1-AMINO-5-(5-HYDROXY-2-METHYLPHENYL)-7,8,9,10-TETRAHYDROPYRIMIDO[4,5-C]ISOQUINOLIN-6(5H)-ONE",
            "non-polymer",
        ),
        "T15": (
            "2-{3-[(BENZYLSULFONYL)AMINO]-6-METHYL-2-OXOPYRIDIN-1(2H)-YL}-N-({1-[2-(TERT-BUTYLAMINO)-2-OXOETHYL]-4-METHYL-1H-IMIDAZOL-5-YL}METHYL)ACETAMIDE",
            "non-polymer",
        ),
        "451": (
            "N-(6,7,9,10,17,18,20,21-OCTAHYDRODIBENZO[B,K][1,4,7,10,13,16]HEXAOXACYCLOOCTADECIN-2-YL)ACETAMIDE",
            "non-polymer",
        ),
        "IQ0": ("CYCLOPENTANECARBOXYLIC ACID", "non-polymer"),
        "XW1": ("(2S)-2-AMINO-7-ETHOXY-7-OXOHEPTANOIC ACID", "peptide"),
        "US5": (
            "1-(5-O-PHOSPHONO-BETA-D-RIBOFURANOSYL)-4-SELANYLPYRIMIDIN-2(1H)-ONE",
            "nucleic-acid",
        ),
        "P3Y": (
            "5-[3-(2-METHOXYPHENYL)-1H-PYRROLO[2,3-B]PYRIDIN-5-YL]-N,N-DIMETHYLPYRIDINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "US3": (
            "1-(2-DEOXY-5-O-PHOSPHONO-BETA-D-ERYTHRO-PENTOFURANOSYL)-5-METHYL-2-SELANYLPYRIMIDIN-4(1H)-ONE",
            "nucleic-acid",
        ),
        "US2": ("5-SME-DEOXYURIDINE-5'-PHOSPHATE", "nucleic-acid"),
        "US1": ("?", "nucleic-acid"),
        "PLQ": ("1,4-BENZOQUINONE", "non-polymer"),
        "SIS": (
            "(1S,2S,3R,4S,6R)-4,6-DIAMINO-3-{[(2S,3R)-3-AMINO-6-(AMINOMETHYL)-3,4-DIHYDRO-2H-PYRAN-2-YL]OXY}-2-HYDROXYCYCLOHEXYL 3-DEOXY-4-C-METHYL-3-(METHYLAMINO)-BETA-L-ARABINOPYRANOSIDE",
            "non-polymer",
        ),
        "SIR": ("COBALT SIROHYDROCHLORIN", "non-polymer"),
        "PDE": ("PARA-NITROPHENYL PHOSPHONOBUTANOYL D-ALANINE", "non-polymer"),
        "J2I": (
            "2-(2,6-DIFLUORO-4-METHOXYPHENYL)-1-(4-{4-[(3-METHYL-1H-PYRAZOL-5-YL)AMINO]PYRROLO[2,1-F][1,2,4]TRIAZIN-2-YL}PIPERAZIN-1-YL)ETHANONE",
            "non-polymer",
        ),
        "SIC": (
            "(2R)-2-[(3S)-3-AMINO-2,5-DIOXOPYRROLIDIN-1-YL]-3-SULFANYLPROPANOIC ACID",
            "peptide",
        ),
        "SIB": (
            "(2S)-2-AMINO-4-({[(2S,3S,4R,5R)-3,4-DIHYDROXY-5-(6-OXO-1,6-DIHYDRO-9H-PURIN-9-YL)TETRAHYDROFURAN-2-YL]METHYL}THIO)BUTANOIC ACID",
            "peptide",
        ),
        "SIA": ("O-SIALIC ACID", "non-polymer"),
        "SIG": ("STIGMATELLIN", "non-polymer"),
        "SIF": ("3-TRIMETHYLSILYLSUCCINIC ACID", "non-polymer"),
        "SIE": (
            "SULFAMIC ACID 1-DECYL-2-(4-SULFAMOYLOXYPHENYL)-1H-INDOL-6-YL ESTER",
            "non-polymer",
        ),
        "SID": (
            "METHYL 9-S-ACETYL-5-(ACETYLAMINO)-3,5-DIDEOXY-9-THIO-D-GLYCERO-ALPHA-D-GALACTO-NON-2-ULOPYRANOSIDONIC ACID",
            "non-polymer",
        ),
        "SIJ": ("6-(1,9-DIAZASPIRO[5.5]UNDECAN-9-YL)-9H-PURINE", "non-polymer"),
        "SII": (
            "N-(4-{[4-(CYCLOHEXYLAMINO)-1-(3-FLUOROPHENYL)-2-OXO-1,3,8-TRIAZASPIRO[4.5]DEC-3-EN-8-YL]METHYL}PHENYL)ACETAMIDE",
            "non-polymer",
        ),
        "SIH": ("PROGESTERONE-11-ALPHA-OL-HEMISUCCINATE", "non-polymer"),
        "SIO": (
            "METHYL 4,9-DI-O-ACETYL-5-(ACETYLAMINO)-3,5-DIDEOXY-D-GLYCERO-ALPHA-D-GALACTO-NON-2-ULOPYRANOSIDONIC ACID",
            "saccharide",
        ),
        "J2Z": (
            "(9BETA,13ALPHA,16BETA)-3,16-DIHYDROXYESTRA-1,3,5(10)-TRIEN-17-ONE",
            "non-polymer",
        ),
        "SIM": ("SIMVASTATIN ACID", "non-polymer"),
        "SIL": (
            "[(4S)-2,2-DIMETHYL-1,3-DIOXOLAN-4-YL]METHYL HYDROGEN HEX-5-ENYLPHOSPHONATE",
            "non-polymer",
        ),
        "TAT": ("ADENOSINE-5'-RP-ALPHA-THIO-TRIPHOSPHATE", "non-polymer"),
        "TAU": ("2-AMINOETHANESULFONIC ACID", "non-polymer"),
        "TAV": (
            "N-METHYL-N-{2-[(2-NAPHTHYLSULFONYL)AMINO]-5-[(2-NAPHTHYLSULFONYL)OXY]BENZOYL}-L-ASPARTIC ACID",
            "peptide",
        ),
        "TAP": ("7-THIONICOTINAMIDE-ADENINE-DINUCLEOTIDE PHOSPHATE", "non-polymer"),
        "TAQ": ("2,4,6-TRIAMINOQUINAZOLINE", "non-polymer"),
        "TAR": ("D(-)-TARTARIC ACID", "non-polymer"),
        "TAS": ("TRIHYDROXYARSENITE(III)", "non-polymer"),
        "TAX": (
            "(Z)-2-[4-(1,2)-DIPHENYL-1-BUTENYL)-PHENOXY]-N,N-DIMETHYLETHANAMINIUM",
            "non-polymer",
        ),
        "TAY": ("2,2-DIMETHYLPROPANAMIDE", "non-polymer"),
        "TAZ": ("TAZOBACTAM", "non-polymer"),
        "TAD": (
            "BETA-METHYLENE-THIAZOLE-4-CARBOXYAMIDE-ADENINE DINUCLEOTIDE",
            "non-polymer",
        ),
        "JI7": (
            "N-{(3R,4S)-4-[(6-AMINO-4-METHYLPYRIDIN-2-YL)METHYL]PYRROLIDIN-3-YL}-N'-(3-CHLOROBENZYL)ETHANE-1,2-DIAMINE",
            "non-polymer",
        ),
        "JI4": (
            "N-{(3S,4S)-4-[(6-AMINO-4-METHYLPYRIDIN-2-YL)METHYL]PYRROLIDIN-3-YL}-N'-[2-(3-FLUOROPHENYL)ETHYL]ETHANE-1,2-DIAMINE",
            "non-polymer",
        ),
        "JI5": (
            "N-{(3R,4R)-4-[(6-AMINO-4-METHYLPYRIDIN-2-YL)METHYL]PYRROLIDIN-3-YL}-N'-[2-(3-FLUOROPHENYL)ETHYL]ETHANE-1,2-DIAMINE",
            "non-polymer",
        ),
        "JI2": (
            "N-{(3S,4S)-4-[(6-AMINOPYRIDIN-2-YL)METHYL]PYRROLIDIN-3-YL}ETHANE-1,2-DIAMINE",
            "non-polymer",
        ),
        "JI3": (
            "N-{(3S,4S)-4-[(6-AMINO-4-METHYLPYRIDIN-2-YL)METHYL]PYRROLIDIN-3-YL}-N'-(4-CHLOROBENZYL)ETHANE-1,2-DIAMINE",
            "non-polymer",
        ),
        "TAB": (
            "ACETIC ACID N-[2-CHLORO-5-[6-ETHYL-2,4-DIAMINO-PYRIMID-5-YL]-PHENYL]-[BENZYL-TRIAZEN-3-YL]ETHYL ESTER",
            "non-polymer",
        ),
        "JI1": (
            "3-({(3S,4S)-4-[(6-AMINOPYRIDIN-2-YL)METHYL]PYRROLIDIN-3-YL}AMINO)PROPAN-1-OL",
            "non-polymer",
        ),
        "TAL": ("9-(6-DEOXY-ALPHA-L-TALOFURANOSYL)-6-METHYLPURINE", "nucleic-acid"),
        "T98": (
            "5-(4,4-DIMETHYL-2-THIOXO-1,4-DIHYDRO-2H-3,1-BENZOXAZIN-6-YL)-1-METHYL-1H-PYRROLE-2-CARBONITRILE",
            "non-polymer",
        ),
        "TAN": ("2,2-DIMETHYLPROPANENITRILE", "non-polymer"),
        "TAO": ("TROLEANDOMYCIN", "non-polymer"),
        "SDK": (
            "1,3-BIS[[N-[(PHENYLMETHOXY)CARBONYL]-L-LEUCYL]AMINO]-2-PROPANONE",
            "non-polymer",
        ),
        "TA4": (
            "(S)-2-[4-(AMINOMETHYL)-1H-1,2,3-TRIAZOL-1-YL]-4-METHYLPENTANOIC ACID",
            "peptide",
        ),
        "TA5": ("2H-1,2,3-TRIAZOL-4-YLMETHANOL", "non-polymer"),
        "TA6": ("6-O-PHOSPHONO-BETA-D-TAGATOFURANOSE", "non-polymer"),
        "DJR": (
            "(3R,3AS,6AR)-HEXAHYDROFURO[2,3-B]FURAN-3-YL [(1S,2R)-1-BENZYL-2-HYDROXY-3-{ISOBUTYL[(4-METHOXYPHENYL)SULFONYL]AMINO}PROPYL]CARBAMATE",
            "non-polymer",
        ),
        "DJT": (
            "(2S,4R)-4-(2-CHLOROPHENYL)SULFONYL-1-[1-(4-FLUOROPHENYL)CYCLOPROPYL]CARBONYL-N-[1-(IMINOMETHYL)CYCLOPROPYL]PYRROLIDINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "TA2": ("(2R,3S)-3-AMINO-3-PHENYLPROPANE-1,2-DIOL", "non-polymer"),
        "TA3": (
            "(4S,5R)-3-(2-DEOXY-5-O-PHOSPHONO-BETA-D-ERYTHRO-PENTOFURANOSYL)-5-METHYL-1,3-DIAZABICYCLO[2.2.0]HEXAN-2-ONE",
            "nucleic-acid",
        ),
        "JIN": (
            "6-(2,6-DICHLOROPHENYL)-2-[(4-FLUORO-3-METHYLPHENYL)AMINO]-8-METHYLPYRIDO[2,3-D]PYRIMIDIN-7(8H)-ONE",
            "non-polymer",
        ),
        "DJZ": (
            "3,5-DIMETHYL-4-{[2-({1-[4-(METHYLSULFONYL)BENZYL]PIPERIDIN-4-YL}AMINO)PYRIMIDIN-4-YL]OXY}BENZONITRILE",
            "non-polymer",
        ),
        "TA8": ("POLYACRYLIC ACID", "non-polymer"),
        "TA9": (
            "9-OXO-N-[[(2R,3R,4R,5R,6S)-3,4,5-TRIHYDROXY-6-METHYL-PIPERIDIN-2-YL]METHYL]FLUORENE-1-CARBOXAMIDE",
            "non-polymer",
        ),
        "SU0": (
            "2-(7-METHOXY-2-OXO-2H-CHROMEN-4-YL)-N-(4-SULFAMOYLPHENYL)ACETAMIDE",
            "non-polymer",
        ),
        "DJK": (
            "N-[4-(3-BROMO-PHENYLAMINO)-QUINAZOLIN-6-YL]-ACRYLAMIDE",
            "non-polymer",
        ),
        "DJL": (
            "(1R)-1-(4-CHLORO-2-FLUORO-3-PHENOXYPHENYL)PROPAN-1-AMINIUM",
            "non-polymer",
        ),
        "DJN": (
            "N-[(3S,4R,5S,6R)-4,5-DIHYDROXY-6-(HYDROXYMETHYL)PIPERIDIN-3-YL]ACETAMIDE",
            "non-polymer",
        ),
        "SI2": ("O-SIALIC ACID (CHAIR CONFORMATION)", "non-polymer"),
        "SI1": (
            "N-(BUTYLSULFONYL)-D-SERYL-N-{4-[AMINO(IMINO)METHYL]BENZYL}-L-ALANINAMIDE",
            "non-polymer",
        ),
        "LAK": ("BETA-D-GALACTOPYRANOSYL-1-6-BETA-D-GLUCOPYRANOSE", "saccharide"),
        "MCS": ("MALONYL CYSTEINE", "peptide"),
        "J25": (
            "DIETHYL 2-[(THIOPHEN-2-YLACETYL)AMINO]-4,7-DIHYDROTHIENO[2,3-C]PYRIDINE-3,6(5H)-DICARBOXYLATE",
            "non-polymer",
        ),
        "LA8": ("L-ADENOSINE-5'-DIPHOSPHATE", "non-polymer"),
        "D5M": ("2'-DEOXYADENOSINE-5'-MONOPHOSPHATE", "non-polymer"),
        "GCA": (
            "6-(3',5'-DIMETHYLBENZYL)-1-ETHOXYMETHYL-5-ISOPROPYLURACIL",
            "non-polymer",
        ),
        "MCR": ("SULFANYLACETIC ACID", "non-polymer"),
        "SC5": (
            "2-[((R)-{[4-(AMINOMETHYL)PHENYL]AMINO}{[(1R)-1-PHENYLETHYL]AMINO}METHYL)AMINO]ETHANE-1,1-DIOL",
            "non-polymer",
        ),
        "T9N": (
            "2-[(3-HYDROXYPHENYL)AMINO]-4-[4-METHYL-2-(METHYLAMINO)-1,3-THIAZOL-5-YL]PYRIMIDINE-5-CARBONITRILE",
            "non-polymer",
        ),
        "SRX": (
            "CHLORO{N-[DI(PYRIDIN-2-YL-KAPPAN)METHYL]-4-SULFAMOYLBENZAMIDE}[(1,2,3,4,5,6-ETA)-(1R,2R,3R,4S,5S,6S)-1,2,3,4,5,6-HEXAMETHYLCYCLOHEXANE-1,2,3,4,5,6-HEXAYL]RUTHENIUM(2+)",
            "non-polymer",
        ),
        "PLP": ("PYRIDOXAL-5'-PHOSPHATE", "non-polymer"),
        "QGG": (
            "1-({4-CHLORO-1-[(DIAMINOMETHYLIDENE)AMINO]ISOQUINOLIN-7-YL}SULFONYL)-D-PROLINE",
            "non-polymer",
        ),
        "LGL": (
            "1-(3-(5-OXO-4,5-DIHYDRO-1H-1,2,4-TRIAZOL-3-YL)PHENYL)-6-(2'-(PYRROLIDIN-1-YLMETHYL)BIPHENYL-4-YL)-3-(TRIFLUOROMETHYL)-5,6-DIHYDRO-1H-PYRAZOLO[3,4-C]PYRIDIN-7(4H)-ONE",
            "non-polymer",
        ),
        "QGA": (
            "1-[(3-CYCLOHEXYLPROPANOYL)(2-HYDROXYETHYL)AMINO]-1-DEOXY-D-ALLITOL",
            "non-polymer",
        ),
        "GCD": ("4,5-DEHYDRO-D-GLUCURONIC ACID", "saccharide"),
        "HC2": ("20-HYDROXYCHOLESTEROL", "non-polymer"),
        "HC3": ("25-HYDROXYCHOLESTEROL", "non-polymer"),
        "HC0": ("2 IRON/2 SULFUR/6 CARBONYL/1 WATER INORGANIC CLUSTER", "non-polymer"),
        "HC1": ("2 IRON/2 SULFUR/5 CARBONYL/2 WATER INORGANIC CLUSTER", "non-polymer"),
        "HC6": (
            "(2S,2'S)-2,2'-BIPHENYL-4,4'-DIYLBIS(2-HYDROXY-4,4-DIMETHYLMORPHOLIN-4-IUM)",
            "non-polymer",
        ),
        "GCG": ("BIS(GAMMA-GLUTAMYL-CYSTEINYL-GLYCINYL)SPERMIDINE", "non-polymer"),
        "HC4": ("PARA-COUMARIC ACID", "non-polymer"),
        "HC5": (
            "(R)-3-CARBOXY-2-(HEXANOYLOXY)-N,N,N-TRIMETHYLPROPAN-1-AMINIUM",
            "non-polymer",
        ),
        "BL5": ("?", "non-polymer"),
        "BL4": (
            "(3AS)-3A-HYDROXY-5-METHYL-1-PHENYL-1,2,3,3A-TETRAHYDRO-4H-PYRROLO[2,3-B]QUINOLIN-4-ONE",
            "non-polymer",
        ),
        "BL7": (
            "(3AS)-3A-HYDROXY-1-PHENYL-1,2,3,3A-TETRAHYDRO-4H-PYRROLO[2,3-B]QUINOLIN-4-ONE",
            "non-polymer",
        ),
        "BL6": (
            "(3AS)-3A-HYDROXY-7-METHYL-1-PHENYL-1,2,3,3A-TETRAHYDRO-4H-PYRROLO[2,3-B]QUINOLIN-4-ONE",
            "non-polymer",
        ),
        "BL1": (
            "4-CHLORO-N-[(2S)-2-METHYL-2,3-DIHYDRO-1H-INDOL-1-YL]-3-SULFAMOYLBENZAMIDE",
            "non-polymer",
        ),
        "BL0": (
            "1-[5-(DIMETHYLAMINO)-1,3,4-THIADIAZOL-2-YL]METHANESULFONAMIDE",
            "non-polymer",
        ),
        "BL3": ("BILIVERDIN XIII ALPHA", "non-polymer"),
        "KBE": ("BETA-LYSINE", "peptide"),
        "I22": ("D-ALTRO-HEPT-2-ULOSE 7-PHOSPHATE", "non-polymer"),
        "NF2": (
            "(1S)-1,4-ANHYDRO-1-(2,4-DIFLUORO-5-METHYLPHENYL)-5-O-PHOSPHONO-D-RIBITOL",
            "nucleic-acid",
        ),
        "GCH": ("GLYCOCHOLIC ACID", "non-polymer"),
        "KBP": ("S-[2-(ACETYLAMINO)ETHYL] (3R)-3-HYDROXYDECANETHIOATE", "non-polymer"),
        "LA3": (
            "(2R)-2-[4-(1,3-BENZODIOXOL-5-YL)BENZYL]-N~4~-HYDROXY-N~1~-[(1S,2R)-2-HYDROXY-2,3-DIHYDRO-1H-INDEN-1-YL]BUTANEDIAMIDE",
            "non-polymer",
        ),
        "GCK": ("?", "nucleic-acid"),
        "KBT": (
            "3-{[3-BUTYL-5-(1-METHYLETHYL)-2,6-DIOXO-1,2,3,6-TETRAHYDROPYRIMIDIN-4-YL]CARBONYL}-5-METHYLBENZONITRILE",
            "non-polymer",
        ),
        "HCR": ("7-HYDROXYCHOLESTEROL", "non-polymer"),
        "HCS": ("2-AMINO-4-MERCAPTO-BUTYRIC ACID", "peptide"),
        "HCP": (
            "1-[((1E)-{3-HYDROXY-2-METHYL-5-[(PHOSPHONOOXY)METHYL]PYRIDIN-4-YL}METHYLENE)AMINO]CYCLOPENTYLPHOSPHONIC ACID",
            "non-polymer",
        ),
        "NFD": ("NICOTINAMIDE 2-FLUORO-ADENINE DINUCLEOTIDE", "non-polymer"),
        "BLY": ("LYSINE BORONIC ACID", "peptide"),
        "HCW": (
            "1,3-BIS{4-[(1E)-N-(4,5-DIHYDRO-1H-IMIDAZOL-2-YL)ETHANEHYDRAZONOYL]PHENYL}UREA",
            "non-polymer",
        ),
        "HCT": ("(2R)-BUTANE-1,2,4-TRICARBOXYLIC ACID", "non-polymer"),
        "BLZ": (
            "5-[5,6-BIS(METHYLOXY)-1H-BENZIMIDAZOL-1-YL]-3-{[1-(2-CHLOROPHENYL)ETHYL]OXY}-2-THIOPHENECARBOXAMIDE",
            "non-polymer",
        ),
        "HCZ": (
            "6-CHLORO-3,4-DIHYDRO-2H-1,2,4-BENZOTHIADIAZINE-7-SULFONAMIDE 1,1-DIOXIDE",
            "non-polymer",
        ),
        "BLT": (
            "1,4-DIDEOXY-1,4-[[2R,3R)-2,4-DIHYDROXY-3-(SULFOXY)BUTYL]EPISELENONIUMYLIDENE]-D-ARABINITOL INNER SALT",
            "non-polymer",
        ),
        "HCX": ("(SOUTH) BICYCLO[3.1.0]HEXANE", "non-polymer"),
        "BLV": ("BILIVERDIN IX GAMMA CHROMOPHORE", "non-polymer"),
        "BLQ": (
            "(3E)-3-({4-[(2-CARBOXYETHYL)CARBAMOYL]PHENYL}HYDRAZONO)-6-OXOCYCLOHEXA-1,4-DIENE-1-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "BLP": (
            "(5-HYDROXY-6-METHYL-4-((2-(2-(2-NITROBENZAMIDO)ACETYL)HYDRAZINYL)METHYL)PYRIDIN-3-YL)METHYL DIHYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "BLS": ("BLASTICIDIN S", "non-polymer"),
        "GCL": ("1-CYCLOHEXYL-2-AMINO-3,4-DIHYDROXY-6-METHYLHEPTANE", "non-polymer"),
        "BLM": ("BLEOMYCIN A2", "non-polymer"),
        "BLL": (
            "(2R,3R,4R,5R)-3,4-DIHYDROXY-N,N'-BIS[(1S,2R)-2-HYDROXY-2,3-DIHYDRO-1H-INDEN-1-YL]-2,5-BIS(2-PHENYLETHYL)HEXANEDIAMIDE",
            "non-polymer",
        ),
        "BLO": (
            "1-(4-{[(3R)-3-AMINO-5-{[(Z)-AMINO(IMINO)METHYL](METHYL)AMINO}PENTANOYL]AMINO}-2,3,4-TRIDEOXY-D-ERYTHRO-HEX-2-ENOPYRANURONOSYL)-4-HYDROXYPYRIMIDIN-2(1H)-ONE",
            "non-polymer",
        ),
        "BLN": (
            "MORPHOLINE-4-CARBOXYLIC ACID [1S-(2-BENZYLOXY-1R-CYANO-ETHYLCARBAMOYL)-3-METHYL-BUTYL]AMIDE",
            "non-polymer",
        ),
        "BLI": (
            "4-OXO-2-PHENYLMETHANESULFONYL-OCTAHYDRO-PYRROLO[1,2-A]PYRAZINE-6-CARBOXYLIC ACID [1-(N-HYDROXYCARBAMIMIDOYL)-PIPERIDIN-4-YLMETHYL]-AMIDE",
            "non-polymer",
        ),
        "GCO": ("GLUCONIC ACID", "saccharide"),
        "BLK": (
            "[[(2R,3S,4R,5R)-5-(6-AMINOPURIN-9-YL)-3,4-BIS(OXIDANYL)OXOLAN-2-YL]METHOXY-OXIDANYL-PHOSPHORYL] (2S)-2-AZANYL-6-[(2-METHYLPROPAN-2-YL)OXYCARBONYLAMINO]HEXANOATE",
            "non-polymer",
        ),
        "NFP": ("(1E,3S)-5-PHENYL-1-(PHENYLSULFONYL)PENT-1-EN-3-AMINE", "non-polymer"),
        "BLE": ("LEUCINE BORONIC ACID", "peptide"),
        "BLD": ("BRASSINOLIDE", "non-polymer"),
        "BLG": (
            "4-O-(4-O-SULFONYL-N-ACETYLGLUCOSAMININYL)-5-METHYLHYDROXY-L-PROLINE-TAURINE",
            "non-polymer",
        ),
        "BLF": (
            "6-CHLORO-3-[1-(4-CHLOROBENZYL)-4-PHENYL-1H-IMIDAZOL-5-YL]-N-[2-(MORPHOLIN-4-YL)ETHYL]-1H-INDOLE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "BLA": ("BILIVERDINE IX ALPHA", "non-polymer"),
        "HCO": ("2-ACETYL-PROTOPORPHYRIN IX", "non-polymer"),
        "HCL": (
            "(2S)-2-AZANYL-2-[2-CHLORANYL-3,5-BIS(OXIDANYL)PHENYL]ETHANOIC ACID",
            "peptide",
        ),
        "BLB": ("BLEOMYCIN B2", "non-polymer"),
        "902": (
            "{[5-(3-{[1-(BENZYLSULFONYL)PIPERIDIN-4-YL]AMINO}PHENYL)-4-BROMO-2-(2H-TETRAZOL-5-YL)-3-THIENYL]OXY}ACETIC ACID",
            "non-polymer",
        ),
        "903": (
            "2-{4-[2-ACETYLAMINO-2-(1-BIPHENYL-4-YLMETHYL-2-OXO-AZEPAN-3-YLCARBAMOYL)-ETHYL]-2-METHOXYCARBONYL-PHENYL}-2-FLUORO-MALONIC ACID",
            "non-polymer",
        ),
        "900": (
            "N-(4-CHLOROPHENYL)-6-[(6,7-DIMETHOXYQUINOLIN-4-YL)OXY]NAPHTHALENE-1-CARBOXAMIDE",
            "non-polymer",
        ),
        "901": (
            "2-[(4-{2-ACETYLAMINO-2-[4-(1-CARBOXY-3-METHYLSULFANYL-PROPYLCARBAMOYL)-BUTYLCARBAMOYL]-ETHYL}-2-ETHYL-PHENYL)-OXALYL-AMINO]-BENZOIC ACID",
            "non-polymer",
        ),
        "907": (
            "2-{5-[AMINO(IMINIO)METHYL]-1H-INDOL-2-YL}-6-BROMO-4-METHYLBENZENOLATE",
            "non-polymer",
        ),
        "QG9": (
            "METHYL ((S)-1-(2-(3-((3S,4S)-3-BENZYL-4-HYDROXY-1-((1S,2R)-2-HYDROXY-2,3-DIHYDRO-1H-INDEN-1-YL)-2-OXOPYRROLIDIN-3-YL)PROPYL)-2-(4-(PYRIDIN-4-YL)BENZYL)HYDRAZINYL)-3,3-DIMETHYL-1-OXOBUTAN-2-YL)CARBAMATE",
            "non-polymer",
        ),
        "905": (
            "(R)-4-[2-(3-AMINO-BENZENESULFONYLAMINO)-1-(3,5-DIETHOXY-2-FLUOROPHENYL)-2-OXO-ETHYLAMINO]-2-HYDROXY-BENZAMIDINE",
            "non-polymer",
        ),
        "TID": ("4-(5-AMINO-1,3,4-THIADIAZOL-2-YL)BENZOIC ACID", "non-polymer"),
        "PC6": (
            "7-[(DODECANOYLOXY)METHYL]-4-HYDROXY-N,N,N-TRIMETHYL-9-OXO-3,5,8-TRIOXA-4-PHOSPHADOTRIACONTAN-1-AMINIUM 4-OXIDE",
            "non-polymer",
        ),
        "GCU": ("D-GLUCURONIC ACID", "saccharide"),
        "GCW": ("4-O-METHYL-BETA-D-GLUCURONIC ACID", "saccharide"),
        "GCY": (
            "N-[(2S,3R)-1-(BETA-D-GALACTOPYRANOSYLOXY)-3-HYDROXYOCTADEC-4-EN-2-YL]TETRACOSANAMIDE",
            "non-polymer",
        ),
        "O8A": (
            "N-CYCLOPROPYL-5-[2-METHYL-5-(TRIFLUOROMETHOXY)-1H-BENZIMIDAZOL-1-YL]THIOPHENE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "LT3": (
            "(1S,4R,9S)-5-(TRIFLUOROMETHYL)-1,2,3,4-TETRAHYDRO-1,4-METHANONAPHTHALEN-9-AMINE",
            "non-polymer",
        ),
        "LT2": ("(3S)-3-HYDROXY-2,3-DIHYDRO-1H-INDOLE-5,6-DIONE", "non-polymer"),
        "LT1": ("(3-CHLOROPROPYL)BENZENE", "non-polymer"),
        "LT6": (
            "2-(2-{[(1S)-1-BENZYL-2-HYDROXYETHYL]AMINO}-2-OXOETHOXY)-N-BUTYL-6-FLUORO-N-METHYLBENZAMIDE",
            "non-polymer",
        ),
        "LT5": ("(1R)-2-AMINO-1-[3-(TRIFLUOROMETHYL)PHENYL]ETHANOL", "non-polymer"),
        "LT4": ("NOREPINEPHRINE", "non-polymer"),
        "FM4": (
            "[(2R,3S)-3-METHYLOXIRAN-2-YL]-PHOSPHONOOXY-PHOSPHINIC ACID",
            "non-polymer",
        ),
        "DIK": (
            "5-{[4-(5-METHYL-3-OXOHEX-4-EN-1-YL)PHENYL]AMINO}-5-OXOPENTANOIC ACID",
            "non-polymer",
        ),
        "FM2": (
            "2-(7-AMINO-6-METHYL-3H-PYRAZOLO[4,3-D]PYRIMIDIN-3-YL)-5-HYDROXYMETHYL-TETRAHYDRO-FURAN-3,4-DIOL",
            "non-polymer",
        ),
        "FM0": (
            "(3R)-3-(FLUOROMETHYL)-3-HYDROXY-5-{[(S)-HYDROXY(PHOSPHONOOXY)PHOSPHORYL]OXY}PENTANOIC ACID",
            "non-polymer",
        ),
        "FM1": (
            "2-HYDROXYMETHYL-5-(7-METHYLAMINO-3H-PYRAZOLO[4,3-D]PYRIMIDIN-3-YL)-TETRAHYDRO-FURAN-3,4-DIOL",
            "non-polymer",
        ),
        "PC4": ("TETRACHLOROPLATINATE(II)", "non-polymer"),
        "R78": (
            "4-{[(7R)-8-CYCLOPENTYL-7-ETHYL-5-METHYL-6-OXO-5,6,7,8-TETRAHYDROPTERIDIN-2-YL]AMINO}-3-METHOXY-N-(1-METHYLPIPERIDIN-4-YL)BENZAMIDE",
            "non-polymer",
        ),
        "R79": (
            "N-(6-{[3-(4-BROMOPHENYL)-1,2-BENZISOTHIAZOL-6-YL]OXY}HEXYL)-N-METHYLPROP-2-EN-1-AMINE",
            "non-polymer",
        ),
        "VLL": ("(2S)-2,3-DIAMINOBUTANOIC ACID", "peptide"),
        "OUT": ("CARBAMIC ACID", "non-polymer"),
        "LTR": ("L-TRYPTOPHAN", "peptide"),
        "FMZ": (
            "(2R,4S)-5,5-DIMETHYL-2-[(1R)-2-OXO-1-({(2R)-2-[(2-OXOIMIDAZOLIDIN-1-YL)AMINO]-2-PHENYLACETYL}AMINO)ETHYL]-1,3-THIAZOLIDINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "FMX": ("FAMOXADONE", "non-polymer"),
        "FMY": (
            "N-[(2Z)-4-(3-FLUORO-5-MORPHOLIN-4-YLPHENYL)PYRIMIDIN-2(1H)-YLIDENE]-4-(3-MORPHOLIN-4-YL-1H-1,2,4-TRIAZOL-1-YL)ANILINE",
            "non-polymer",
        ),
        "FMV": (
            "(1S,3R,5Z,7E,14BETA,17ALPHA,20S)-20-[5-(1-HYDROXY-1-METHYLETHYL)FURAN-2-YL]-9,10-SECOPREGNA-5,7,10-TRIENE-1,3-DIOL",
            "non-polymer",
        ),
        "FMT": ("FORMIC ACID", "non-polymer"),
        "FMU": ("5-FLUORO-5-METHYLURIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "FMR": ("FUMARATE", "non-polymer"),
        "FMS": ("TRIFLUOROMETHANE SULFONAMIDE", "non-polymer"),
        "FMP": ("FORMYCIN-5'-MONOPHOSPHATE", "non-polymer"),
        "FMQ": ("[1-(4-FLUOROPHENYL)-5-METHYL-1H-PYRAZOL-4-YL]METHANOL", "non-polymer"),
        "FMN": ("FLAVIN MONONUCLEOTIDE", "non-polymer"),
        "LTB": ("LEUKOTRIENE B4", "non-polymer"),
        "FML": (
            "6-(2,4-DIFLUOROPHENOXY)-N-[(1S)-1-METHYL-2-(METHYLSULFONYL)ETHYL]-1H-PYRAZOLO[3,4-D]PYRIMIDIN-3-AMINE",
            "non-polymer",
        ),
        "FMM": (
            "N-{3-CHLORO-4-[(3-FLUOROBENZYL)OXY]PHENYL}-6-[5-({[2-(METHYLSULFONYL)ETHYL]AMINO}METHYL)-2-FURYL]-4-QUINAZOLINAMINE",
            "non-polymer",
        ),
        "FMJ": ("4-(5-METHYL-3-PHENYLISOXAZOL-4-YL)PYRIMIDIN-2-AMINE", "non-polymer"),
        "FMK": (
            "6-(2,4-DIFLUOROPHENOXY)-8-METHYL-2-{[(1S)-1-METHYL-2-(2H-TETRAZOL-2-YL)ETHYL]AMINO}PYRIDO[2,3-D]PYRIMIDIN-7(8H)-ONE",
            "non-polymer",
        ),
        "FMI": ("FE-(4-MESOPORPHYRINONE)-R-ISOMER", "non-polymer"),
        "FMF": ("2-DEOXY-2-FLUOROHEXOPYRANOSYL FLUORIDE", "non-polymer"),
        "FMG": (
            "2-AMINO-9-(2-DEOXY-2-FLUORO-5-O-PHOSPHONO-BETA-D-ARABINOFURANOSYL)-7-METHYL-6-OXO-6,9-DIHYDRO-1H-PURIN-7-IUM",
            "nucleic-acid",
        ),
        "FMD": (
            "4-((3R,4S,5R)-4-AMINO-3,5-DIHYDROXY-HEX-1-YNYL)-5-FLUORO-3-[1-(3-METHOXY-1H-PYRROL-2-YL)-METH-(Z)-YLIDENE]-1,3-DIHYDRO-INDOL-2-ONE",
            "non-polymer",
        ),
        "FME": ("N-FORMYLMETHIONINE", "peptide"),
        "FMB": ("FORMYCIN B", "non-polymer"),
        "FMC": (
            "(1S)-1-(7-AMINO-1H-PYRAZOLO[4,3-D]PYRIMIDIN-3-YL)-1,4-ANHYDRO-D-RIBITOL",
            "non-polymer",
        ),
        "LTM": ("N-ACETYL GLUCONOLACTAM", "saccharide"),
        "FMA": ("6-(3-TETRADECANOIC ACID) FLAVINE MONONUCLEOTIDE", "non-polymer"),
        "OCQ": (
            "(3S)-3-[(2S)-2-AMINO-4-HYDROXY-3-OXOBUTYL]PYRROLIDIN-2-ONE",
            "non-polymer",
        ),
        "P1P": ("3-{[OXIDO(OXO)PHOSPHINO]OXY}PROPAN-1-OLATE", "nucleic-acid"),
        "TD7": (
            "(4E)-4-{3-[(4-AMINO-2-METHYLPYRIMIDIN-5-YL)METHYL]-5-(2-{[(S)-HYDROXY(PHOSPHONOOXY)PHOSPHORYL]OXY}ETHYL)-4-METHYL-1,3-THIAZOL-2(3H)-YLIDENE}-4-HYDROXYBUTANOIC ACID",
            "non-polymer",
        ),
        "DIT": ("DITERCALINIUM", "non-polymer"),
        "FTR": ("FLUOROTRYPTOPHANE", "peptide"),
        "VR1": (
            "(2S)-2-({(5Z)-5-[(5-ETHYL-2-FURYL)METHYLENE]-4-OXO-4,5-DIHYDRO-1,3-THIAZOL-2-YL}AMINO)-2-(4-FLUOROPHENYL)-N-[(4-NITROPHENYL)SULFONYL]ACETAMIDE",
            "non-polymer",
        ),
        "DIU": ("2-HYDROXY-3,5-DIIODO-BENZOIC ACID", "non-polymer"),
        "SAX": ("2-(DODECYLOXY)-6-HYDROXYBENZOIC ACID", "non-polymer"),
        "DIV": ("D-ISOVALINE", "peptide"),
        "DIW": (
            "1-(3-SULFOPROPYL)-4-[(1E,3E)-3-(1,3,3-TRIMETHYL-1,3-DIHYDRO-2H-INDOL-2-YLIDENE)PROP-1-EN-1-YL]QUINOLINIUM",
            "non-polymer",
        ),
        "D4N": (
            "1-{[(3R)-3-METHYL-4-({4-[(1S)-2,2,2-TRIFLUORO-1-HYDROXY-1-METHYLETHYL]PHENYL}SULFONYL)PIPERAZIN-1-YL]METHYL}CYCLOPROPANECARBOXAMIDE",
            "non-polymer",
        ),
        "D4M": (
            "[(5R)-5-(5-METHYL-2,4-DIOXO-3,4-DIHYDROPYRIMIDIN-1(2H)-YL)-2,5-DIHYDROFURAN-2-YL]METHYL DIHYDROGEN PHOSPHATE",
            "nucleic-acid",
        ),
        "XCS": (
            "(1R)-1-(4-AMINO-6-METHYL-2-OXO-1,2-DIHYDROQUINAZOLIN-8-YL)-1,4-ANHYDRO-2-DEOXY-5-O-PHOSPHONO-D-ERYTHRO-PENTITOL",
            "nucleic-acid",
        ),
        "DIP": ("DIPENTYLAMINE", "non-polymer"),
        "D4G": ("?", "non-polymer"),
        "D4D": ("?", "non-polymer"),
        "UA1": (
            "N-(TERT-BUTOXYCARBONYL)-L-TYROSYL-N-METHYL-4-(SULFOAMINO)-L-PHENYLALANINAMIDE",
            "non-polymer",
        ),
        "P58": ("BENZYL(METHYL)CARBAMODITHIOIC ACID", "non-polymer"),
        "DIR": ("3-{[(E)-AMINO(HYDROXYIMINO)METHYL]AMINO}ALANINE", "peptide"),
        "P54": (
            "8-[(2,4-DIMETHYLPHENYL)SULFANYL]-3-PENT-4-YN-1-YL-3H-PURIN-6-AMINE",
            "non-polymer",
        ),
        "P55": (
            "2-(4-ETHYLPIPERAZIN-1-YL)-4-(PHENYLAMINO)PYRAZOLO[1,5-A][1,3,5]TRIAZINE-8-CARBONITRILE",
            "non-polymer",
        ),
        "D4P": ("(2S)-AMINO(4-HYDROXYPHENYL)ACETIC ACID", "peptide"),
        "P51": (
            "2-{[4-(DIETHYLAMINO)PIPERIDIN-1-YL]METHYL}-6-IODO-4-(3-PHENOXYPROP-1-YN-1-YL)PHENOL",
            "non-polymer",
        ),
        "D4T": ("?", "non-polymer"),
        "UA2": (
            "3,4-PYRROLIDINEDIOL,2-(4-AMINO-5H-PYRROLO[3,2-D]PYRIMIDIN-7-YL)-5-(HYDROXYMETHYL)-2S,3S,4R,5R",
            "non-polymer",
        ),
        "P5K": (
            "1-{5-TERT-BUTYL-3-[(1,1-DIOXIDOTHIOMORPHOLIN-4-YL)CARBONYL]THIOPHEN-2-YL}-3-NAPHTHALEN-1-YLUREA",
            "non-polymer",
        ),
        "P5D": (
            "(2R)-2-{[(S)-[(1R)-1-AMINO-2-PHENYLETHYL](HYDROXY)PHOSPHORYL]METHYL}BUTANEDIOIC ACID",
            "peptide",
        ),
        "P5A": ("'5'-O-(N-(L-PROLYL)-SULFAMOYL)ADENOSINE", "non-polymer"),
        "P5B": (
            "N-(ETHYLSULFONYL)TRYPTOPHYL-N~1~-{4-[AMINO(IMINO)METHYL]BENZYL}GLUTAMAMIDE",
            "non-polymer",
        ),
        "P5C": (
            "2-{[(1R,2S)-2-AMINOCYCLOHEXYL]AMINO}-4-[(3-METHYLPHENYL)AMINO]PYRIMIDINE-5-CARBOXAMIDE",
            "non-polymer",
        ),
        "YS3": (
            "(1R,3R,7E,17BETA)-17-[(2R)-5-HYDROXY-5-METHYLHEXAN-2-YL]-2-METHYLIDENE-9,10-SECOESTRA-5,7-DIENE-1,3-DIOL",
            "non-polymer",
        ),
        "RAD": ("C49-METHYL RAPAMYCIN", "non-polymer"),
        "D42": (
            "1-(3-(2,4-DIMETHYLTHIAZOL-5-YL)-4-OXO-2,4-DIHYDROINDENO[1,2-C]PYRAZOL-5-YL)-3-(4-METHYLPIPERAZIN-1-YL)UREA",
            "non-polymer",
        ),
        "G2Q": ("GUANOSINE DIPHOSPHATE RIBOSE", "non-polymer"),
        "P5P": ("?", "nucleic-acid"),
        "D46": (
            "N-[2-({[(5-CHLOROTHIOPHEN-2-YL)CARBONYL]AMINO}METHYL)PHENYL]-5-METHYL-4,5,6,7-TETRAHYDRO[1,3]THIAZOLO[5,4-C]PYRIDINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "P": ("?", "nucleic-acid"),
        "5QC": (
            "1-(3-AMINO-1,2-BENZISOXAZOL-5-YL)-6-(2'-{[(3R)-3-HYDROXYPYRROLIDIN-1-YL]METHYL}BIPHENYL-4-YL)-3-(TRIFLUOROMETHYL)-1,4,5,6-TETRAHYDRO-7H-PYRAZOLO[3,4-C]PYRIDIN-7-ONE",
            "non-polymer",
        ),
        "OCP": (
            "OCTYL-PHOSPHINIC ACID 1,2-BIS-OCTYLCARBAMOYLOXY-ETHYL ESTER",
            "non-polymer",
        ),
        "ZNB": ("?", "non-polymer"),
        "ZNE": (
            "(1R,3S,5Z)-5-[(2E)-2-{(1R,3AS,7AR)-1-[(1R,5S)-5-HYDROXY-1-METHYL-5-(1,3-THIAZOL-2-YL)PENTYL]-7A-METHYLOCTAHYDRO-4H-INDEN-4-YLIDENE}ETHYLIDENE]-4-METHYLIDENECYCLOHEXANE-1,3-DIOL",
            "non-polymer",
        ),
        "JH2": (
            "METHYL (2E,6E)-9-[(2R,3S)-3-ETHYL-3-METHYLOXIRAN-2-YL]-3,7-DIMETHYLNONA-2,6-DIENOATE",
            "non-polymer",
        ),
        "ZNH": ("PROTOPORPHYRIN IX CONTAINING ZN", "non-polymer"),
        "ZNO": ("ZINC ION, 2 WATERS COORDINATED", "non-polymer"),
        "ZNP": ("(2R)-2-BENZYL-3-NITROPROPANAMIDE", "non-polymer"),
        "TIF": (
            "N-(BENZYLSULFONYL)-D-LEUCYL-N-(4-CARBAMIMIDOYLBENZYL)-L-PROLINAMIDE",
            "peptide",
        ),
        "AY9": ("ASCOPYRONE M", "non-polymer"),
        "SE0": (
            "(2R)-2-[(1R)-2-[(2S)-2-AMINO-2-CARBOXYETHOXY]-2-OXO-1-{[(2R)-TETRAHYDROTHIOPHEN-2-YLACETYL]AMINO}ETHYL]-5-METHYLIDENE-5,6-DIHYDRO-2H-1,3-THIAZINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "AY1": ("ANTIMYCIN A1", "non-polymer"),
        "AY0": (
            "(1R,2R,3S)-2-(METHYLCARBAMOYL)-3-[4-(PHOSPHONOOXY)PHENYL]CYCLOPROPANECARBOXYLIC ACID",
            "peptide",
        ),
        "2S2": ("(2S)-2-(1H-INDOL-3-YL)HEXANOIC ACID", "non-polymer"),
        "2S3": ("(2S)-2-(1H-INDOL-3-YL)PENTANOIC ACID", "non-polymer"),
        "AYM": (
            "3-(6-AMINOPYRIDIN-3-YL)-N-METHYL-N-[(1-METHYL-1H-INDOL-2-YL)METHYL]ACRYLAMIDE",
            "non-polymer",
        ),
        "AYL": (
            "5'-O-{(S)-HYDROXY[3-(4-HYDROXYPHENYL)PROPOXY]PHOSPHORYL}ADENOSINE",
            "non-polymer",
        ),
        "AYI": ("?", "peptide"),
        "2SI": ("2-O-SULFO-A-L-IDURONIC ACID", "peptide"),
        "2SK": (
            "(1R,2R,3R,4S,5R)-4-(BENZYLAMINO)-5-(METHYLTHIO)CYCLOPENTANE-1,2,3-TRIOL",
            "non-polymer",
        ),
        "AYE": ("PROP-2-EN-1-AMINE", "non-polymer"),
        "AYD": ("?", "non-polymer"),
        "AYG": (
            "[(4E)-2-[(1S)-1-AMINOETHYL]-4-(4-HYDROXYBENZYLIDENE)-5-OXO-4,5-DIHYDRO-1H-IMIDAZOL-1-YL]ACETIC ACID",
            "peptide",
        ),
        "AYA": ("N-ACETYLALANINE", "peptide"),
        "2SA": (
            "2-[9-(3,4-DIHYDROXY-5-PHOSPHONOOXYMETHYL-TETRAHYDRO-FURAN-2-YL)-9H-PURIN-6-YLAMINO]-SUCCINIC ACID",
            "nucleic-acid",
        ),
        "2SB": (
            "(2S)-2-AMINO-N'-[(1E)-(3-BROMO-4-HYDROXYPHENYL)METHYLIDENE]-2-PHENYLETHANEHYDRAZIDE",
            "non-polymer",
        ),
        "2SC": (
            "(5-PHENYL-7-(PYRIDIN-3-YLMETHYLAMINO)PYRAZOLO[1,5-A]PYRIMIDIN-3-YL)METHANOL",
            "non-polymer",
        ),
        "AYX": ("4-{[(3-NITROPHENYL)CARBAMOYL]AMINO}BENZENESULFONAMIDE", "non-polymer"),
        "2ST": ("?", "nucleic-acid"),
        "MEF": (
            "N-({4-[(6AR)-3-AMINO-1-OXO-1,2,5,6,6A,7-HEXAHYDROIMIDAZO[1,5-F]PTERIDIN-8(9H)-YL]PHENYL}CARBONYL)-L-GLUTAMIC ACID",
            "non-polymer",
        ),
        "AYQ": (
            "N~2~-({(1R,2R,3S)-2-(METHYLCARBAMOYL)-3-[4-(PHOSPHONOOXY)PHENYL]CYCLOPROPYL}CARBONYL)-D-GLUTAMINYL-D-ASPARTAMIDE",
            "peptide",
        ),
        "LAI": ("L-ARABINOSE", "saccharide"),
        "9HI": (
            "(3R,5R)-7-{3-[(4-CARBAMOYLPHENYL)SULFAMOYL]-4,5-BIS(4-FLUOROPHENYL)-2-(1-METHYLETHYL)-1H-PYRROL-1-YL}-3,5-DIHYDROXYHEPTANOIC ACID",
            "non-polymer",
        ),
        "9HO": ("(9S,10E,12Z)-9-HYDROXYOCTADECA-10,12-DIENOIC ACID", "non-polymer"),
        "3KR": (
            "1-(2-METHOXYETHYL)-N-OXO-4-({4-[4-(TRIFLUOROMETHOXY)PHENOXY]PHENYL}SULFONYL)PIPERIDINE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "RAA": ("ALPHA-RHAMNOSE", "saccharide"),
        "FFX": ("(2R,3S,5R,6S)-2,6-DIFLUOROOXANE-3,4,5-TRIOL", "non-polymer"),
        "3KE": (
            "4-{[({3-[2-(4-METHOXYBENZYL)-2H-TETRAZOL-5-YL]PHENYL}CARBONYL)AMINO]METHYL}BENZOIC ACID",
            "non-polymer",
        ),
        "LAD": ("ADENOSINE-5'-[LYSYL-PHOSPHATE]", "non-polymer"),
        "9HX": ("9-DEAZAHYPOXANTHINE", "non-polymer"),
        "3KC": (
            "3-CHLORO-4-(4H-3,4,7-TRIAZADIBENZO[CD,F]AZULEN-6-YL)PHENOL",
            "non-polymer",
        ),
        "LAE": (
            "3-OXO-OCTANOIC ACID (2-OXO-TETRAHYDRO-FURAN-3-YL)-AMIDE",
            "non-polymer",
        ),
        "9HP": ("9-(4-HYDROXYPHENYL)-2,7-PHENANTHROLINE", "non-polymer"),
        "LAF": ("5-FLUOROLEVULINIC ACID", "non-polymer"),
        "PCP": (
            "1-ALPHA-PYROPHOSPHORYL-2-ALPHA,3-ALPHA-DIHYDROXY-4-BETA-CYCLOPENTANE-METHANOL-5-PHOSPHATE",
            "non-polymer",
        ),
        "1EC": (
            "[(5R)-5-AMINO-5-CARBOXY-8-HYDROXYOCTYL](TRIHYDROXY)BORATE(1-)",
            "non-polymer",
        ),
        "FFT": (
            "(2Z,6Z)-3,7,11-TRIMETHYLDODECA-2,6,10-TRIEN-1-YL DIHYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "1EE": (
            "[(5R)-5-AMINO-5-CARBOXY-7-(4-HYDROXYPIPERIDIN-1-YL)HEPTYL](TRIHYDROXY)BORATE(1-)",
            "non-polymer",
        ),
        "XLI": (
            "N-ACETYL-L-LEUCYL-N-[(4S,5S,7R)-8-(BUTYLAMINO)-5-HYDROXY-2,7-DIMETHYL-8-OXOOCTAN-4-YL]-L-METHIONINAMIDE",
            "peptide",
        ),
        "XLF": ("BETA-D-XYLULOFURANOSE", "saccharide"),
        "I4D": ("D-MYO-INOSITOL-4-PHOSPHATE", "non-polymer"),
        "XLD": (
            "3-CHLORO-N-[4-CHLORO-2-[[(5-CHLORO-2-PYRIDINYL)AMINO]CARBONYL]-6-METHOXYPHENYL]-4-[[(4,5-DIHYDRO-2-OXAZOLYL)METHYLAMINO]METHYL]-2-THIOPHENECARBOXAMIDE",
            "non-polymer",
        ),
        "1EM": ("(1S)-2-HYDROXY-1-[(NONANOYLOXY)METHYL]ETHYL MYRISTATE", "non-polymer"),
        "1EL": (
            "9-HYDROXY-5,11-DIMETHYL-4,6-DIHYDRO-1H-PYRIDO[4,3-B]CARBAZOL-1-ONE",
            "non-polymer",
        ),
        "I4A": (
            "5-(3-CARBAMOYLBENZYL)-5,6,7,8,9,10-HEXAHYDROCYCLOHEPTA[B]INDOLE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "1EU": (
            "1-(4-ACETYL-6-PYRIDIN-3-YL-1H-BENZIMIDAZOL-2-YL)-3-ETHYLUREA",
            "non-polymer",
        ),
        "FFQ": ("[(1R)-1-HYDROXYPROPYL]PHOSPHONIC ACID", "non-polymer"),
        "XLX": ("4,4'-(ETHYLENEDIIMINO)BIS[4-OXOBUTYRATE]", "non-polymer"),
        "1EY": ("(11Z,13Z)-HEXADECA-11,13-DIENAL", "non-polymer"),
        "1EX": ("(11Z,13Z)-HEXADECA-11,13-DIEN-1-OL", "non-polymer"),
        "1EZ": (
            "N-(4-SULFAMOYLPHENYL)-2-[(3S,5S,7S)-TRICYCLO[3.3.1.1~3,7~]DEC-1-YL]ACETAMIDE",
            "non-polymer",
        ),
        "XLS": ("D-XYLOSE (LINEAR FORM)", "saccharide"),
        "I4P": ("(1S,3R,4R,6S)-1,3,4,6-TETRAPKISPHOSPHATE", "non-polymer"),
        "0Y8": ("(4R)-4-[(7-BROMOISOQUINOLIN-1-YL)OXY]-L-PROLINE", "peptide"),
        "0Y9": ("(1R,2S)-1-AMINO-2-ETHENYLCYCLOPROPANECARBOXYLIC ACID", "peptide"),
        "0Y5": (
            "4-{[(3S)-3-(5-METHYL-2,4-DIOXO-3,4-DIHYDROPYRIMIDIN-1(2H)-YL)PIPERIDIN-1-YL]METHYL}-2-[3-(TRIFLUOROMETHYL)PHENOXY]BENZOIC ACID",
            "non-polymer",
        ),
        "0Y7": (
            "{(3R,5R,6S)-5-(3-CHLOROPHENYL)-6-(4-CHLOROPHENYL)-1-[(2S)-1-HYDROXYBUTAN-2-YL]-2-OXOPIPERIDIN-3-YL}ACETIC ACID",
            "non-polymer",
        ),
        "IRP": (
            "(1S)-1(9-DEAZAHYPOXANTHIN-9YL)1,4-DIDEOXY-1,4-IMINO-D-RIBITOL-5-PHOSPHATE",
            "non-polymer",
        ),
        "O58": (
            "{4-[(N-{3-[5-CHLORO-2-(1H-TETRAZOL-1-YL)PHENYL]PROPANOYL}-L-PHENYLALANYL)AMINO]PHENYL}ACETIC ACID",
            "non-polymer",
        ),
        "YZE": (
            "N~2~-{[(1S)-1-CARBOXYBUT-3-EN-1-YL]CARBAMOYL}-N~6~-[(4-IODOPHENYL)CARBONYL]-L-LYSINE",
            "non-polymer",
        ),
        "LAX": ("EICOSA-8,11,14-TRIENOIC ACID", "non-polymer"),
        "VM1": ("5-(2-THIENYL)NICOTINIC ACID", "non-polymer"),
        "0YX": (
            "(5-BROMO-2-{[3-(OCTYLOXY)BENZYL]OXY}PHENYL)PHOSPHONIC ACID",
            "non-polymer",
        ),
        "0YY": (
            "(2Z)-2-HYDROXY-4-[4-(OCTYLOXY)PHENYL]-4-OXOBUT-2-ENOIC ACID",
            "non-polymer",
        ),
        "0YZ": (
            "(2Z)-4-[3-(DECYLOXY)PHENYL]-2-HYDROXY-4-OXOBUT-2-ENOIC ACID",
            "non-polymer",
        ),
        "LAZ": ("N-(2-AMINOETHYL)-P-CHLOROBENZAMIDE", "non-polymer"),
        "YZ6": (
            "(E,2R,3R,4S,5R)-N-[(3R)-3-(FURAN-2-YL)-3-PHENYL-PROPYL]-2-METHOXY-8,8-DIMETHYL-3,4,5-TRIS(OXIDANYL)NON-6-ENAMIDE",
            "non-polymer",
        ),
        "0YQ": ("?", "non-polymer"),
        "0YR": ("(2R,3R,4R)-N,2,3,4,5-PENTAKIS(OXIDANYL)PENTANAMIDE", "non-polymer"),
        "0YS": (
            "(5R)-2-[(1S,2R)-2-AMINO-2-CARBOXY-1-HYDROXYETHYL]-5-{(2S)-2-CARBOXY-2-[(3,5-DICHLORO-4-HYDROXYBENZOYL)AMINO]ETHYL}-L-PROLINE",
            "non-polymer",
        ),
        "0YT": ("2-(ACETYLAMINO)-2-DEOXY-5-THIO-BETA-D-GLUCOPYRANOSE", "saccharide"),
        "0YU": (
            "N~4~-[4-(4,5-DIHYDRO-1H-IMIDAZOL-2-YL)PHENYL]-N~1~-[4-(1H-IMIDAZOL-2-YL)PHENYL]-2-NITROBENZENE-1,4-DICARBOXAMIDE",
            "non-polymer",
        ),
        "0YV": (
            "2,2'-{BENZENE-1,3-DIYLBIS[ETHYNE-2,1-DIYL(5-BROMOBENZENE-3,1-DIYL)]}DIETHANAMINE",
            "non-polymer",
        ),
        "0YW": ("2-{[3-(DECYLOXY)BENZOYL]AMINO}-5-NITROBENZOIC ACID", "non-polymer"),
        "0YH": (
            "1-[3-TERT-BUTYL-1-(4-METHYLPHENYL)-1H-PYRAZOL-5-YL]-3-[3-(4-METHOXY-2-METHYLPHENYL)-1H-PYRAZOL-5-YL]UREA",
            "non-polymer",
        ),
        "FFG": (
            "(R)-6-(2'-((3-HYDROXYPYRROLIDIN-1-YL)METHYL)BIPHENYL-4-YL)-1-(3-(5-OXO-4,5-DIHYDRO-1H-1,2,4-TRIAZOL-3-YL)PHENYL)-3-(TRIFLUOROMETHYL)-5,6-DIHYDRO-1H-PYRAZOLO[3,4-C]PYRIDIN-7(4H)-ONE",
            "non-polymer",
        ),
        "0YJ": (
            "7-({[3-TERT-BUTYL-1-(4-METHYLPHENYL)-1H-PYRAZOL-5-YL]CARBAMOYL}AMINO)-N-(PROPAN-2-YL)-1H-INDOLE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "0YL": ("?", "non-polymer"),
        "0YO": (
            "2-[5-METHOXY-2-(QUINOLIN-3-YL)PYRIMIDIN-4-YL]-1,5,6,7-TETRAHYDRO-4H-PYRROLO[3,2-C]PYRIDIN-4-ONE",
            "non-polymer",
        ),
        "EU": ("EUROPIUM ION", "non-polymer"),
        "0YA": ("CYCLOPROPANESULFONAMIDE", "non-polymer"),
        "0YB": (
            "2-(3-BROMOPHENOXY)-4-{(1R)-3,3-DIMETHYL-1-[(3S)-3-(5-METHYL-2,4-DIOXO-3,4-DIHYDROPYRIMIDIN-1(2H)-YL)PIPERIDIN-1-YL]BUTYL}BENZOIC ACID",
            "non-polymer",
        ),
        "0YC": ("1-(PYRIDIN-3-YL)METHANAMINE", "non-polymer"),
        "0YD": ("?", "non-polymer"),
        "0YE": ("3-(3-PHENYL-1,2-OXAZOL-5-YL)PROPANOIC ACID", "non-polymer"),
        "0YF": ("3-(BIPHENYL-4-YL)PROPANOIC ACID", "non-polymer"),
        "0YG": ("N-[(2Z)-2-AMINO-3-(4-HYDROXYPHENYL)PROP-2-ENOYL]GLYCINE", "peptide"),
        "LAP": (
            "[2-((1-OXODODECANOXY-(2-HYDROXY-3-PROPANYL))-PHOSPHONATE-OXY)-ETHYL]-TRIMETHYLAMMONIUM",
            "non-polymer",
        ),
        "LAQ": (
            "5'-O-[(R)-({5-[(3R)-1,2-DITHIOLAN-3-YL]PENTANOYL}OXY)(HYDROXY)PHOSPHORYL]ADENOSINE",
            "non-polymer",
        ),
        "1E1": ("(2R,3S)-3-AMINO-4-CYCLOPROPYL-2-HYDROXYBUTANAMIDE", "non-polymer"),
        "MEM": (
            "(4R,5S,6S)-3-{[(3S,5S)-5-(DIMETHYLCARBAMOYL)PYRROLIDIN-3-YL]SULFANYL}-6-[(1R)-1-HYDROXYETHYL]-4-METHYL-7-OXO-1-AZABICYCLO[3.2.0]HEPT-2-ENE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "7BD": ("3-[(9H-FLUOREN-9-YLIDENEAMINO)OXY]PROPANOIC ACID", "non-polymer"),
        "7BB": (
            "TRANS-4-({[(E)-1-AMINO-2-SULFANYLETHENYL](4-CARBOXYBUTYL)CARBAMOYL}OXY)CYCLOHEXANECARBOXYLIC ACID",
            "peptide",
        ),
        "1E4": (
            "1-O-[3-(5-CARBOXYPYRIDIN-2-YL)-5-FLUOROPHENYL]-6-O-[4-({[(5-CARBOXYPYRIDIN-2-YL)SULFANYL]ACETYL}AMINO)-2-CHLORO-5-METHOXYPHENYL]-D-MANNITOL",
            "non-polymer",
        ),
        "1E7": (
            "6-({2-[(5-CHLORO-2-METHOXYPHENYL)AMINO]-2-OXOETHYL}SULFANYL)PYRIDINE-3-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "1E6": (
            "6-({2-[(5-CHLORO-4-{[(2S)-2,3-DIHYDROXYPROPYL]OXY}-2-METHOXYPHENYL)AMINO]-2-OXOETHYL}SULFANYL)PYRIDINE-3-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "I46": (
            "2-FLUORO-4-[4-(4-FLUOROPHENYL)-1H-PYRAZOL-3-YL]PYRIDINE",
            "non-polymer",
        ),
        "I47": (
            "2-{4-[5-(4-CHLOROPHENYL)-4-PYRIMIDIN-4-YL-1H-PYRAZOL-3-YL]PIPERIDIN-1-YL}-2-OXOETHANOL",
            "non-polymer",
        ),
        "I45": (
            "3-{3-BROMO-4-[(2,4-DIFLUOROBENZYL)OXY]-6-METHYL-2-OXOPYRIDIN-1(2H)-YL}-N,4-DIMETHYLBENZAMIDE",
            "non-polymer",
        ),
        "I42": ("HYDROXY(DIOXIDO)OXOVANADIUM", "non-polymer"),
        "I43": (
            "6-[3-HYDROXY-2-(HYDROXYMETHYL)PROP-1-EN-1-YL]-4-METHOXY-1,5-DIMETHYLPYRIMIDIN-2(1H)-ONE",
            "non-polymer",
        ),
        "I40": ("9-(3-IODOBENZYLAMINO)-1,2,3,4-TETRAHYDROACRIDINE", "non-polymer"),
        "I41": (
            "2-[[2-[[(2S)-3-(3H-IMIDAZOL-4-YL)-1-METHOXY-1-OXO-PROPAN-2-YL]AMINO]-2-OXO-ETHYL]-(PHENYLMETHYL)AMINO]ETHANOIC ACID",
            "non-polymer",
        ),
        "SCA": ("SUCCINYL-COENZYME A", "non-polymer"),
        "TCP": ("5'-METHYLTHYMIDINE", "nucleic-acid"),
        "VJJ": ("3-PHENOXYBENZOIC ACID", "non-polymer"),
        "86A": (
            "(2S)-(3-{[AMINO(IMINO)METHYL]AMINO}PHENYL){[(S)-HYDROXY{(1R)-2-METHYL-1-[(3-PHENYLPROPANOYL)AMINO]PROPYL}PHOSPHORYL]OXY}ACETIC ACID",
            "non-polymer",
        ),
        "LAG": (
            "MALTOSYL-ALPHA (1,4)-(Z,3S,4S,5R,6R)-3,4,5-TRIHYDROXY-6-HYDROXYMETHYL-PIPERIDIN-2-ONE OXIME",
            "saccharide",
        ),
        "DTN": ("DITHIONITE", "non-polymer"),
        "IRC": ("(R)-2-HYDROXY-4-METHYLPENTANOYL-COA", "non-polymer"),
        "4B0": (
            "3-(4,5,6,7-TETRABROMO-1H-BENZOTRIAZOL-1-YL)PROPAN-1-OL",
            "non-polymer",
        ),
        "4B3": (
            "4-BROMO-3-(CARBOXYMETHOXY)-5-{3-[(3,3,5,5-TETRAMETHYLCYCLOHEXYL)AMINO]PHENYL}THIOPHENE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "WPP": ("PIPERACILLIN", "non-polymer"),
        "TCQ": (
            "(1S,2R)-N-{(1E)-5-[(2S)-2-AMINO-2-CARBOXYETHYL]-2-HYDROXY-4-OXOCYCLOHEXA-2,5-DIEN-1-YLIDENE}-2-PHENYLCYCLOPROPANAMINIUM",
            "peptide",
        ),
        "182": (
            "1-((2-HYDROXYETHOXY)METHYL)-5-(PHENYLSELANYL)PYRIMIDINE-2,4(1H,3H)-DIONE",
            "non-polymer",
        ),
        "183": (
            "1-((2-HYDROXYETHOXY)METHYL)-5-(3-(BENZYLOXY)BENZYL)PYRIMIDINE-2,4(1H,3H)-DIONE",
            "non-polymer",
        ),
        "180": (
            "2-(S)-[N-(3-PYRIDYLSULFONYL)AMINO]-3-[[2-CARBONYL-5-[2-(PIPERIDIN-4-YL)ETHYL]-THIENO[2,3-B]THIOPHENEYL]AMINO]-PROPIONIC ACID",
            "non-polymer",
        ),
        "181": (
            "1-((2-HYDROXYETHOXY)METHYL)-5-(PHENYLTHIO)PYRIMIDINE-2,4(1H,3H)-DIONE",
            "non-polymer",
        ),
        "186": (
            "[1-(1-METHYL-4,5-DIOXO-PENT-2-ENYLCARBAMOYL)-2-PHENYL-ETHYL]-CARBAMIC ACID BENZYL ESTER",
            "non-polymer",
        ),
        "187": (
            "1-{3-[AMINO(IMINO)METHYL]PHENYL}-N-[4-(1H-BENZIMIDAZOL-1-YL)-2-FLUOROPHENYL]-3-(TRIFLUOROMETHYL)-1H-PYRAZOLE-5-CARBOXAMIDE",
            "non-polymer",
        ),
        "184": (
            "6-[HYDROXY-(5,5,8,8-TETRAMETHYL-5,6,7,8-TETRAHYDRO-NAPHTALEN-2-YL)-METHYL]-NAPHTALENE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "185": (
            "(6-[4-(AMINOMETHYL)-2,6-DIMETHYLPHENOXY]-2-{[4-(AMINOMETHYL)PHENYL]AMINO}-5-BROMOPYRIMIDIN-4-YL)METHANOL",
            "non-polymer",
        ),
        "WPE": (
            "N-{2-[(4S)-6-CHLORO-2-METHYL-4-PHENYLQUINAZOLIN-3(4H)-YL]ETHYL}FURAN-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "188": ("PYOCHELIN FE(III)", "non-polymer"),
        "189": (
            "N-BENZYL-2-(2,6-DIMETHYLPHENOXY)-N-[((3R,4S)-4-{[ISOBUTYL(PHENYLSULFONYL)AMINO]METHYL}PYRROLIDIN-3-YL)METHYL]ACETAMIDE",
            "non-polymer",
        ),
        "18B": ("181,182-DIHYDROBILIVERDIN", "non-polymer"),
        "18C": (
            "N-((E,2S,3R)-1,3-DIHYDROXYOCTADEC-4-EN-2-YL)STEARAMIDE",
            "non-polymer",
        ),
        "4BZ": ("4-(HYDROXYMETHYL)BENZAMIDINE", "non-polymer"),
        "18D": (
            "3,5-DIDEOXY-5-(PROPANOYLAMINO)-D-GLYCERO-ALPHA-D-GALACTO-NON-2-ULOPYRANOSONIC ACID",
            "non-polymer",
        ),
        "18E": (
            "7-CHLORO-10-METHYL-11H-BENZO[G]PYRIDO[4,3-B]INDOL-3-OL",
            "non-polymer",
        ),
        "4BQ": (
            "(2S)-2-AMINO-5-OXO-5-[(4-PHENYLMETHOXYPHENYL)AMINO]PENTANOIC ACID",
            "non-polymer",
        ),
        "4BP": ("4,4'-BIS(BROMOMETHYL)BIPHENYL", "non-polymer"),
        "4BS": ("4-AMINO-N-[4-(BENZYLOXY)PHENYL]BUTANAMIDE", "non-polymer"),
        "4BR": ("2-(4-METHYLPHENOXY)ETHYLPHOSPHINATE", "non-polymer"),
        "18N": ("1H-BENZO[DE]ISOQUINOLINE-1,3(2H)-DIONE", "non-polymer"),
        "18O": (
            "N-({(2E)-2-[(4-NITROPHENYL)METHYLIDENE]HYDRAZINO}CARBONOTHIOYL)-BETA-D-GLUCOPYRANOSYLAMINE",
            "non-polymer",
        ),
        "18L": (
            "(2R)-2-HYDROXY-3-(PHOSPHONOOXY)PROPYL (9Z,12Z)-OCTADECA-9,12-DIENOATE",
            "non-polymer",
        ),
        "18M": ("?", "nucleic-acid"),
        "18R": (
            "3-{1-[(3R)-1-ACRYLOYLPIPERIDIN-3-YL]-4-AMINO-1H-PYRAZOLO[3,4-D]PYRIMIDIN-3-YL}-N-(3-TERT-BUTYLPHENYL)BENZAMIDE",
            "non-polymer",
        ),
        "18S": (
            "N-({(2E)-2-[(4-METHYLPHENYL)METHYLIDENE]HYDRAZINO}CARBONOTHIOYL)-BETA-D-GLUCOPYRANOSYLAMINE",
            "non-polymer",
        ),
        "18P": (
            "N-{3-[(2R)-6-AMINO-2,4-DIMETHYL-3-OXO-2,3,4,5-TETRAHYDROPYRAZIN-2-YL]PHENYL}-5-CHLOROPYRIDINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "18Q": ("?", "nucleic-acid"),
        "4BM": (
            "N-{[(2R)-2,3-DIHYDROXYPROPYL]OXY}-3,4-DIFLUORO-2-[(2-FLUORO-4-IODOPHENYL)AMINO]BENZAMIDE",
            "non-polymer",
        ),
        "18W": (
            "3-[(5Z)-5-{[3-(2-CARBOXYETHYL)-4-(CARBOXYMETHYL)-5-METHYL-1H-PYRROL-2-YL]METHYLIDENE}-4-(CARBOXYMETHYL)-2-OXO-2,5-DIHYDRO-1H-PYRROL-3-YL]PROPANOIC ACID",
            "non-polymer",
        ),
        "4BO": (
            "(3S)-3-AMINO-4-OXO-4-[(4-PHENYLMETHOXYPHENYL)AMINO]BUTANOIC ACID",
            "non-polymer",
        ),
        "18Z": (
            "2-{[(2-AMINOPYRIMIDIN-5-YL)METHYL]AMINO}-5-NITRO-4-{[2-(PIPERAZIN-1-YL)ETHYL]AMINO}BENZAMIDE",
            "non-polymer",
        ),
        "RAI": (
            "1-(4-CARBOXY-2-GUANIDINOPENTYL)-5,5'-DI(HYDROXYMETHYL)PYRROLIDIN-2-ONE",
            "non-polymer",
        ),
        "4BC": (
            "4-(1,3-BENZODIOXOL-5-YL)-5-(5-ETHYL-2,4-DIHYDROXYPHENYL)-2H-PYRAZOLE-3-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "4BB": (
            "4-TERT-BUTYL-N'-[(1E)-(3,5-DIBROMO-2,4-DIHYDROXYPHENYL)METHYLIDENE]BENZOHYDRAZIDE",
            "non-polymer",
        ),
        "4BE": (
            "4-BROMO-N'-[(1E)-(3,5-DIBROMO-2,4-DIHYDROXYPHENYL)METHYLIDENE]BENZOHYDRAZIDE",
            "non-polymer",
        ),
        "4BD": (
            "5'-O-[(S)-{DIFLUORO[(S)-HYDROXY(PHOSPHONOOXY)PHOSPHORYL]METHYL}(HYDROXY)PHOSPHORYL]THYMIDINE",
            "non-polymer",
        ),
        "4BG": ("N-[4-(BENZYLOXY)PHENYL]GLYCINAMIDE", "non-polymer"),
        "4BF": ("4-BROMO-L-PHENYLALANINE", "peptide"),
        "869": (
            "(1-TERT-BUTYL-5-HYDROXY-1H-PYRAZOL-4-YL)[6-(METHYLSULFONYL)-4'-METHOXY-2-METHYL-1,1'-BIPHENYL-3-YL]METHANONE",
            "non-polymer",
        ),
        "DTZ": ("ZINC(II)HYDROGENSULFIDE", "non-polymer"),
        "861": (
            "3-(3,4,5-TRIMETHOXYPHENYL)PROPYL (2S)-1-(BENZYLSULFONYL)PIPERIDINE-2-CARBOXYLATE",
            "non-polymer",
        ),
        "MEI": (
            "(2E,4E)-11-METHOXY-3,7,11-TRIMETHYLDODECA-2,4-DIENOIC ACID",
            "non-polymer",
        ),
        "RO1": (
            "(3S,4AS,8AS)-N-(TERT-BUTYL)-2-[(3S)-3-({3-(METHYLSULFONYL)-N-[(PYRIDIN-3-YLOXY)ACETYL]-L-VALYL}AMINO)-2-OXO-4-PHENYLBUTYL]DECAHYDROISOQUINOLINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "864": (
            "(2S)-2-[3-(AMINOMETHYL)PHENYL]-3-[(R)-[(1R)-1-{[(BENZYLOXY)CARBONYL]AMINO}-2-METHYLPROPYL](HYDROXY)PHOSPHORYL]PROPANOIC ACID",
            "non-polymer",
        ),
        "866": (
            "5-AMINO-N-[(2S)-2-({[(2,6-DICHLOROPHENYL)CARBONYL](ETHYL)AMINO}METHYL)-3,3,3-TRIFLUORO-2-HYDROXYPROPYL]-1-(4-FLUOROPHENYL)-1H-PYRAZOLE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "DTX": ("DIGITOXIGENIN", "non-polymer"),
        "RAV": (
            "5-ETHYL-5-[(2S)-PENTAN-2-YL]PYRIMIDINE-2,4,6(1H,3H,5H)-TRIONE",
            "non-polymer",
        ),
        "ZST": (
            "3,4-DIHYDRO-4-OXO-3-((5-TRIFLUOROMETHYL-2-BENZOTHIAZOLYL)METHYL)-1-PHTHALAZINE ACETIC ACID",
            "non-polymer",
        ),
        "ZSV": (
            "(R)-[(2S)-5-CARBOXY-2-(2-CARBOXYETHYL)-2,3-DIHYDRO-1,4-BENZODIOXIN-6-YL]METHYL-[[2-[(4-METHOXYPHENYL)METHYLCARBAMOYL]PHENYL]METHYL]-PROP-2-ENYL-AZANIUM",
            "non-polymer",
        ),
        "ZSW": (
            "(R)-[2-[[(2R)-BUTAN-2-YL]CARBAMOYL]PHENYL]METHYL-[[(2S)-5-CARBOXY-2-(2-CARBOXYETHYL)-2,3-DIHYDRO-1,4-BENZODIOXIN-6-YL]METHYL]-PROP-2-ENYL-AZANIUM",
            "non-polymer",
        ),
        "VNF": (
            "4'-CHLORO-N-[(1R)-2-(1H-IMIDAZOL-1-YL)-1-PHENYLETHYL]BIPHENYL-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "TCL": ("TRICLOSAN", "non-polymer"),
        "RO2": (
            "3-[5-[5-(4-METHYL-PIPERAZIN-1-YL)-1H-IMIDAZO[4,5-B]PYRIDIN-2-YL]-BENZIMIDAZOL-2-YL]-PHENOL",
            "non-polymer",
        ),
        "RAW": ("BETA,BETA-CAROTEN-4-ONE", "non-polymer"),
        "ZSC": (
            "5-[METHYL(METHYLSULFONYL)AMINO]BENZENE-1,3-DICARBOXYLIC ACID",
            "non-polymer",
        ),
        "DTQ": ("4-[3-HYDROXYANILINO]-6,7-DIMETHOXYQUINAZOLINE", "non-polymer"),
        "ZYJ": (
            "(4R)-4-({[(1E)-(3,4-DIMETHOXYPHENYL)METHYLIDENE]AMINO}OXY)-L-PROLINE",
            "peptide",
        ),
        "Z48": (
            "N-[(1S)-2-HYDROXY-1-PHENYLETHYL]-4-[5-METHYL-2-(PHENYLAMINO)PYRIMIDIN-4-YL]-1H-PYRROLE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "RPP": (
            "2-RIBOFURANOSYL-3-IODO-2,3-DIHYDRO-1H-PYRAZOLO[3,4-D]PYRIMIDIN-4-YLAMINE",
            "non-polymer",
        ),
        "RPS": ("PHTALIMIDE-RUTHENIUM COMPLEX", "non-polymer"),
        "DTP": ("?", "non-polymer"),
        "RPT": ("RIFAPENTINE", "non-polymer"),
        "PZZ": (
            "3-{3-[(DIMETHYLAMINO)METHYL]-1H-INDOL-7-YL}PROPAN-1-OL",
            "non-polymer",
        ),
        "RPX": ("C16-ETHOXY-RAPAMYCIN,IMMUNOSUPPRESSANT DRUG", "non-polymer"),
        "RPA": ("RHODOPINAL GLUCOSIDE", "saccharide"),
        "HX2": ("(2R,5R)-HEXANE-2,5-DIOL", "non-polymer"),
        "E8U": ("(4R)-4-OXIDANYL-2-OXIDANYLIDENE-HEPTANEDIOIC ACID", "non-polymer"),
        "RPD": (
            "[8,9-DIHYDROXY-3-(4-CARBOXY-HYDROXY-HYDROXYMETHYL-AMINO-BUTYL)-2,4-DIOXO-6-OXA-1,3-DIAZA-SPIRO[4.4]NON-7-YLMETHYL] PHOSPHATE",
            "non-polymer",
        ),
        "EUD": ("EUDESMANE CATION", "non-polymer"),
        "RPH": ("6-HYDROXO-OCTAHYDRO-INDOLE-2-CARBALDEHYDE", "non-polymer"),
        "PJ2": (
            "(5Z,12Z,15S)-15-HYDROXY-11-OXOPROSTA-5,9,12-TRIEN-1-OIC ACID",
            "non-polymer",
        ),
        "Q72": (
            "(S)-4-(4-(3-(3-CARBAMIMIDOYLPHENYL)-2-(2,4,6-TRIISOPROPYLPHENYLSULFONAMIDO)PROPANOYL)PIPERAZINE-1-CARBONYL)PIPERIDINE-1-CARBOXIMIDAMIDE",
            "non-polymer",
        ),
        "RPN": ("(R)-1-PARA-NITRO-PHENYL-2-AZIDO-ETHANOL", "non-polymer"),
        "HXC": ("HEXANOYL-COENZYME A", "non-polymer"),
        "HXB": ("?", "non-polymer"),
        "HXA": ("DOCOSA-4,7,10,13,16,19-HEXAENOIC ACID", "non-polymer"),
        "RP2": (
            "(2R,4AR,6R,7R,7AS)-6-(6-AMINO-8-BROMO-9H-PURIN-9-YL)TETRAHYDRO-4H-FURO[3,2-D][1,3,2]DIOXAPHOSPHININE-2,7-DIOL 2-SULFIDE",
            "non-polymer",
        ),
        "RO4": (
            "[[1-[N-HYDROXY-ACETAMIDYL]-3-METHYL-BUTYL]-CARBONYL-LEUCINYL]-ALANINE ETHYL ESTER",
            "non-polymer",
        ),
        "RP4": (
            "(1S,4S,5S)-1,4,5-TRIHYDROXY-3-[3-(PHENYLTHIO)PHENYL]CYCLOHEX-2-ENE-1-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "HXD": ("(3R)-3-HYDROXYDODECANOIC ACID", "non-polymer"),
        "HXH": (
            "(1S,2S,3R,6R)-6-[(4-PHENOXYBENZYL)AMINO]CYCLOHEX-4-ENE-1,2,3-TRIOL",
            "non-polymer",
        ),
        "HXL": (
            "2-(4-(4-CARBAMIMIDOYLPHENOXY)PHENYL)-1H-BENZO[D]IMIDAZOLE-6-CARBOXIMIDAMIDE",
            "non-polymer",
        ),
        "E86": (
            "4-{2-[(2-CHLOROPHENYL)AMINO]-5-METHYLPYRIMIDIN-4-YL}-N-[(1S)-2-HYDROXY-1-PHENYLETHYL]-1H-PYRROLE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "HXP": ("3,6-DIHYDROXY-XANTHENE-9-PROPIONIC ACID", "non-polymer"),
        "DI5": ("AC-(D)PHE-PRO-BOROHOMOORNITHINE-OH", "non-polymer"),
        "HXZ": ("?", "non-polymer"),
        "HXX": ("3,6-DICHLORO-2-HYDROXYBENZOIC ACID", "non-polymer"),
        "3II": (
            "N-[2-(2,4-DICHLOROPHENYL)ETHYL]-2-{8-[(2,4-DIMETHOXYPHENYL)CARBONYL]-4-OXO-1-PHENYL-1,3,8-TRIAZASPIRO[4.5]DEC-3-YL}ACETAMIDE",
            "non-polymer",
        ),
        "E89": (
            "6-AMINO-4-[2-(4-METHOXYPHENYL)ETHYL]-1,7-DIHYDRO-8H-IMIDAZO[4,5-G]QUINAZOLIN-8-ONE",
            "non-polymer",
        ),
        "ZS4": (
            "2-(DIFLUOROMETHYL)-1-(4,6-DIMORPHOLIN-4-YL-1,3,5-TRIAZIN-2-YL)-1H-BENZIMIDAZOLE",
            "non-polymer",
        ),
        "TCK": (
            "N-[(1S)-5-AMINO-1-(CHLOROACETYL)PENTYL]-4-METHYLBENZENESULFONAMIDE",
            "peptide",
        ),
        "RO5": (
            "N-[(5-BROMO-1,3-THIAZOL-2-YL)CARBAMOYL]-3-ETHYLBENZENESULFONAMIDE",
            "non-polymer",
        ),
        "ZS0": (
            "3-(4-FLUOROBENZYL)-2-THIOXO-1,2,3,7-TETRAHYDRO-6H-PURIN-6-ONE",
            "non-polymer",
        ),
        "RAR": (
            "5-(3,4-DIMETHOXY-5-{(1E)-3-OXO-3-[(1S)-1-PROPYLPHTHALAZIN-2(1H)-YL]PROP-1-EN-1-YL}BENZYL)PYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "DI3": ("AC-(D)PHE-PRO-BORO-N-BUTYL-AMIDINO-GLYCINE-OH", "non-polymer"),
        "MET": ("METHIONINE", "peptide"),
        "BOX": ("BENZOIC ACID", "non-polymer"),
        "UZR": (
            "2'-DEOXY-2'-TRIAZA-1,2-DIEN-2-IUM-1-YL-URIDINE-5'-MONOPHOSPHATE",
            "non-polymer",
        ),
        "HBL": ("7,8-DIHYDRO-L-BIOPTERIN", "non-polymer"),
        "BOZ": ("BIS(5-AMIDINO-BENZIMIDAZOLYL)METHANONE ZINC", "non-polymer"),
        "HBN": ("N-(2-NAPHTHYL)HISTIDINAMIDE", "peptide"),
        "5PY": (
            "1-(2'-DEOXY-5'-O-PHOSPHONO-BETA-D-ERYTHRO-PENTOFURANOSYL)-5-METHYLPYRIMIDIN-2(1H)-ONE",
            "nucleic-acid",
        ),
        "NAF": (
            "M-(N,N,N-TRIMETHYLAMMONIO)-2,2,2-TRIFLUORO-1,1-DIHYDROXYETHYLBENZENE",
            "non-polymer",
        ),
        "RO7": (
            "(2S)-2-METHOXY-3-{4-[2-(5-METHYL-2-PHENYL-1,3-OXAZOL-4-YL)ETHOXY]-1-BENZOTHIOPHEN-7-YL}PROPANOIC ACID",
            "non-polymer",
        ),
        "HBH": (
            "2-[(8S,11S)-11-{(1R)-1-HYDROXY-2-[ISOPENTYL(PHENYLSULFONYL)AMINO]ETHYL}-6,9-DIOXO-2-OXA-7,10-DIAZABICYCLO[11.2.2]HEPTADECA-1(15),13,16-TRIEN-8-YL]ACETAMIDE",
            "non-polymer",
        ),
        "C2R": (
            "5-AMINO-1-(5-O-PHOSPHONO-BETA-D-RIBOFURANOSYL)-1H-IMIDAZOLE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "C2S": ("CYTIDINE-5'-DITHIOPHOSPHORATE", "nucleic-acid"),
        "C2P": ("CYTIDINE-2'-MONOPHOSPHATE", "non-polymer"),
        "IWD": (
            "2-AMINO-3-(5-IODO-2,4-DIOXO-3,4-DIHYDRO-2H-PYRIMIDIN-1-YL)-PROPIONIC ACID",
            "non-polymer",
        ),
        "HBK": (
            "(2S)-2,4-DIHYDROXY-7-METHOXY-2H-1,4-BENZOXAZIN-3(4H)-ONE",
            "non-polymer",
        ),
        "C2U": ("3,5-DICHLORO-2-HYDROXYBENZOIC ACID", "non-polymer"),
        "PZU": ("3,6-DIDEOXY-ALPHA-D-RIBO-HEXOPYRANOSE", "non-polymer"),
        "C2X": ("?", "non-polymer"),
        "NAE": ("NICOTINAMIDE ADENINE DINUCLEOTIDE ACETONE ADDUCT", "non-polymer"),
        "PZQ": ("PRAZIQUANTEL", "non-polymer"),
        "PZP": (
            "[5-HYDROXY-4-(IMINOMETHYL)-6-METHYL-PYRIDIN-3-YL]METHYL DIHYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "IWH": (
            "1-(2,4-DIMETHYLBENZYL)-6-OXO-1,6-DIHYDROPYRIDINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "C2B": ("1-(4-CHLOROPHENYL)METHANAMINE", "non-polymer"),
        "C2C": ("CU-CL-CU LINKAGE", "non-polymer"),
        "PZO": ("PYRAZOLE", "non-polymer"),
        "C2A": ("1-(3-CHLOROPHENYL)METHANAMINE", "non-polymer"),
        "C2F": ("5-METHYL-5,6,7,8-TETRAHYDROFOLIC ACID", "non-polymer"),
        "C2G": ("unknown", "non-polymer"),
        "C2D": (
            "6-CHLORO-1-(2-{[(5-CHLORO-1-BENZOTHIEN-3-YL)METHYL]AMINO}ETHYL)-3-[(2-PYRIDIN-2-YLETHYL)AMINO]-1,4-DIHYDROPYRAZIN-2-OL",
            "non-polymer",
        ),
        "C2E": ("C-DI-GMP, CYCLIC DIGUANOSINE MONOPHOSPHATE", "non-polymer"),
        "PZE": ("PIPERAZINE", "non-polymer"),
        "PZD": (
            "(11AS)-7,8-DIMETHOXY-2-NAPHTHALEN-2-YL-1,10,11,11A-TETRAHYDRO-5H-PYRROLO[2,1-C][1,4]BENZODIAZEPIN-5-ONE",
            "non-polymer",
        ),
        "MEV": ("(R)-MEVALONATE", "non-polymer"),
        "C2N": ("3-CHLORO-D-ALANINE", "peptide"),
        "C2O": ("CU-O-CU LINKAGE", "non-polymer"),
        "C2L": ("?", "nucleic-acid"),
        "C2M": ("5-(3,4-DICHLOROPHENYL)-6-METHYLPYRIMIDINE-2,4-DIAMINE", "non-polymer"),
        "C22": ("3-CHLORO-L-ALANINE", "peptide"),
        "C23": (
            "1-[3-METHYL-4-(PIPERIDIN-4-YLOXY)-1-BENZOFURAN-2-YL]-3-PHENYLPROPAN-1-ONE",
            "non-polymer",
        ),
        "C20": (
            "ACETYL-NH-VAL-CYCLOHEXYL-CH2[NCH2CHOH]CH2-BENZYL-VAL-NH-ACETYL",
            "non-polymer",
        ),
        "C21": ("THIOPHENE-2-CARBOXYLIC ACID", "non-polymer"),
        "C26": ("(2R)-2-ETHYL-1-HEXANESULFONIC ACID", "non-polymer"),
        "C27": (
            "(6R)-2-AMINO-6-[2-(3'-METHOXYBIPHENYL-3-YL)ETHYL]-3,6-DIMETHYL-5,6-DIHYDROPYRIMIDIN-4(3H)-ONE",
            "non-polymer",
        ),
        "C24": (
            "3-({2-[(4-CARBAMIMIDOYL-PHENYLAMINO)-METHYL]-3-METHYL-3H-BENZOIMIDAZOLE-5-CARBONYL}-PYRIDIN-2-YL-AMINO)-PROPIONIC ACID ETHYL ESTER",
            "non-polymer",
        ),
        "C25": ("CYTIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "FVF": (
            "O-((((N-PHENYL-METHOXY-CARBONYL)-PHENYL ALANYL-CARBONYL)AMINO)-ISOBUTYL)HYDROXY PHOSPHINYL)-3-PHENYLACETIC ACID",
            "non-polymer",
        ),
        "C28": (
            "N-[1-(AMINOMETHYL)CYCLOPROPYL]-3-(MORPHOLIN-4-YLSULFONYL)-N~2~-[(1S)-2,2,2-TRIFLUORO-1-(4-FLUOROPHENYL)ETHYL]-L-ALANINAMIDE",
            "non-polymer",
        ),
        "PZ1": (
            "(6R)-6-({[1-(3-HYDROXYPROPYL)-1,7-DIHYDROQUINOLIN-7-YL]OXY}METHYL)-1-(4-{3-[(2-METHOXYBENZYL)OXY]PROPOXY}PHENYL)PIPERAZIN-2-ONE",
            "non-polymer",
        ),
        "FVA": ("N-FORMYL-L-VALINE", "peptide"),
        "NAO": ("SODIUM ION, 1 WATER COORDINATED", "non-polymer"),
        "IW6": (
            "2,6-DIBROMO-4-[(E)-2-(2,6-DIMETHOXYPHENYL)ETHENYL]ANILINE",
            "non-polymer",
        ),
        "IW5": ("4-[(E)-2-(4-AMINOPHENYL)ETHENYL]-2,6-DIBROMOANILINE", "non-polymer"),
        "IW4": ("4-[(E)-2-(3-AMINOPHENYL)ETHENYL]-2,6-DIBROMOANILINE", "non-polymer"),
        "IW3": ("4-[(E)-2-(4-AMINOPHENYL)ETHENYL]-2,6-DIBROMOPHENOL", "non-polymer"),
        "HBC": (
            "(2-AMINO-3-PHENYL-BICYCLO[2.2.1]HEPT-2-YL)-PHENYL-METHANONE",
            "non-polymer",
        ),
        "IW1": (
            "2,6-DIBROMO-4-[(E)-2-(2,6-DICHLOROPHENYL)ETHENYL]PHENOL",
            "non-polymer",
        ),
        "FVX": ("FLUVOXAMINE", "non-polymer"),
        "MEQ": ("N5-METHYLGLUTAMINE", "peptide"),
        "BOW": ("OCTYL SULFAMATE", "non-polymer"),
        "FVQ": ("3-PYRIDIN-3-YLPROP-2-YN-1-YL ALPHA-D-MANNOPYRANOSIDE", "non-polymer"),
        "PA0": ("PHENYLARSINE OXIDE", "non-polymer"),
        "6UL": ("TETRACOSYL PALMITATE", "non-polymer"),
        "UQ9": ("UBIQUINONE-9", "non-polymer"),
        "UQ8": ("UBIQUINONE-8", "non-polymer"),
        "NAQ": ("NICOTINAMIDE ADENINE DINUCLEOTIDE 3-PENTANONE ADDUCT", "non-polymer"),
        "6UP": ("8-AMINONAPHTHALENE-2-CARBOXIMIDAMIDE", "non-polymer"),
        "TCD": (
            "3,5-DICHLORO-2-{4-[(3,5-DICHLOROPYRIDIN-2-YL)OXY]PHENOXY}PYRIDINE",
            "non-polymer",
        ),
        "UQ1": ("UBIQUINONE-1", "non-polymer"),
        "HBY": (
            "(S)-4-ISOPROPOXYCARBONYL-6-METHOXY-3-METHYLTHIOMETHYL-3,4-DIHYDROQUINOXALIN-2(1H)-THIONE",
            "non-polymer",
        ),
        "UQ2": ("UBIQUINONE-2", "non-polymer"),
        "UQ5": (
            "2,3-DIMETHOXY-5-METHYL-6-(3,11,15,19-TETRAMETHYL-EICOSA-2,6,10,14,18-PENTAENYL)-[1,4]BENZOQUINONE",
            "non-polymer",
        ),
        "UQ7": ("UBIQUINONE-7", "non-polymer"),
        "GX1": (
            "2'-DEOXY-5'-O-{(S)-HYDROXY[(2-SULFANYLETHYL)AMINO]PHOSPHORYL}GUANOSINE",
            "nucleic-acid",
        ),
        "VME": ("METHYL L-VALINATE", "peptide"),
        "NAT": (
            "ETHYL 4-(3-HYDROXYPHENYL)-6-METHYL-2-THIOXO-1,2,3,4-TETRAHYDROPYRIMIDINE-5-CARBOXYLATE",
            "non-polymer",
        ),
        "PA1": ("2-AMINO-2-DEOXY-ALPHA-D-GLUCOPYRANOSE", "non-polymer"),
        "NAU": (
            "4,4,6,7,8,9-HEXAHYDROXY-5-METHYLCARBOXAMIDONONANOIC ACID",
            "non-polymer",
        ),
        "SKY": (
            "1-BETA-D-GLUCOPYRANOSYL-4-NAPHTHALEN-2-YL-1H-1,2,3-TRIAZOLE",
            "non-polymer",
        ),
        "SKZ": ("SCHIZOKINEN", "non-polymer"),
        "NBL": (
            "N2-[(BENZYLOXY)CARBONYL]-N1-[(3S)-1-CYANOPYRROLIDIN-3-YL]-L-LEUCINAMIDE",
            "non-polymer",
        ),
        "31U": ("D-LEUCYL-N-(4-CARBAMIMIDOYLBENZYL)-L-PROLINAMIDE", "non-polymer"),
        "OAF": ("2,2-DIFLUORO-3,3-DIHYDROXYBUTANEDIOIC ACID", "non-polymer"),
        "J0Z": ("BENZYL ACETATE", "non-polymer"),
        "SKA": ("7,8-DICHLORO-1,2,3,4-TETRAHYDROISOQUINOLINE", "non-polymer"),
        "TCE": ("3-[BIS(2-CARBOXYETHYL)PHOSPHANYL]PROPANOIC ACID", "non-polymer"),
        "HBT": (
            "(6S)-6-METHYL-2-{[(2-NITROPHENYL)CARBONYL]AMINO}-4,5,6,7-TETRAHYDRO-1-BENZOTHIOPHENE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "SKM": (
            "(3R,4S,5R)-3,4,5-TRIHYDROXYCYCLOHEX-1-ENE-1-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "NAX": (
            "BETA-6-HYDROXY-1,4,5,6-TETRHYDRONICOTINAMIDE ADENINE DINUCLEOTIDE",
            "non-polymer",
        ),
        "GER": ("GERAN-8-YL GERAN", "non-polymer"),
        "DD9": ("NONANE", "non-polymer"),
        "GEP": (
            "N-METHYL-N-(PARA-GLUTARAMIDOPHENYL-ETHYL)-PIPERIDINIUM ION",
            "non-polymer",
        ),
        "MFD": (
            "(2S,3S,8S,9S)-3-AMINO-9-METHOXY-2,6,8-TRIMETHYL-10-PHENYLDECA-4,6-DIENOIC ACID",
            "non-polymer",
        ),
        "PA2": ("PAROMOMYCIN (RING 2)", "non-polymer"),
        "TCX": ("ETHYLPHOSPHORAMIDIC ACID", "non-polymer"),
        "GET": ("GENETICIN", "non-polymer"),
        "DD3": (
            "5-[(1S)-1-(3-CHLOROPHENYL)ETHOXY]QUINAZOLINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "DD2": (
            "5-{[1-(2,3-DICHLOROBENZYL)PIPERIDIN-4-YL]METHOXY}QUINAZOLINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "DD1": (
            "5-{[1-(2-FLUOROBENZYL)PIPERIDIN-4-YL]METHOXY}QUINAZOLINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "TCU": ("5-HEXYL-2-(2-METHYLPHENOXY)PHENOL", "non-polymer"),
        "TCR": ("CYCLOMETHYLTRYPTOPHAN", "peptide"),
        "MFM": (
            "(E)-3-(5((5-(4-CHLOROPHENYL)FURAN-2-YL)METHYLENE)-4-OXO-2-THIOXOTHIAZOLIDIN-3-YL)PROPANOIC ACID",
            "non-polymer",
        ),
        "MFN": (
            "N-[4,5,7-TRICARBOXYHEPTANOYL]-L-GAMMA-GLUTAMYL-N-{2-[4-({5-[(FORMYLAMINO)METHYL]-3-FURYL}METHOXY)PHENYL]ETHYL}-D-GLUTAMINE",
            "non-polymer",
        ),
        "BOE": ("?", "nucleic-acid"),
        "MFP": (
            "((2-BROMO-4-METHYLPHENYL){6-[(4-{[(2R)-3-(DIMETHYLAMINO)-2-HYDROXYPROPYL]OXY}PHENYL)AMINO]PYRIMIDIN-4-YL}AMINO)ACETONITRILE",
            "non-polymer",
        ),
        "MFQ": (
            "((2-BROMO-4-METHYLPHENYL){6-[(4-{[(2S)-3-(DIMETHYLAMINO)-2-HYDROXYPROPYL]OXY}PHENYL)AMINO]PYRIMIDIN-4-YL}AMINO)ACETONITRILE",
            "non-polymer",
        ),
        "MFR": (
            "4-(4-METHOXY-1H-PYRROLO[2,3-B]PYRIDIN-3-YL)PYRIMIDIN-2-AMINE",
            "non-polymer",
        ),
        "MFS": (
            "(1S)-2,3,6-TRI-O-ACETYL-1,5-ANHYDRO-1-SULFAMOYL-4-O-(2,3,4,6-TETRA-O-ACETYL-BETA-D-GALACTOPYRANOSYL)-D-GLUCITOL",
            "non-polymer",
        ),
        "GEG": ("(4E)-4-AMINOHEX-4-ENOIC ACID", "non-polymer"),
        "MFU": ("ALPHA-L-METHYL-FUCOSE", "saccharide"),
        "GEE": ("ETHYL GLYCINATE", "peptide"),
        "TCI": (
            "(6AR,10AR)-6,6,9-TRIMETHYL-3-PENTYL-6A,7,8,10A-TETRAHYDRO-6H-BENZO[C]CHROMEN-1-OL",
            "non-polymer",
        ),
        "GEK": (
            "(2S)-4-[{[(2S,3S,4R,5R)-5-(6-AMINO-9H-PURIN-9-YL)-3,4-DIHYDROXYTETRAHYDROFURAN-2-YL]METHYL}(CARBOXYLATOMETHYL)SULFONIO]-2-AMMONIOBUTANOATE (NON-PREFERRED NAME)",
            "non-polymer",
        ),
        "JG1": (
            "7-PHENYL-1-[5-(PYRIDIN-2-YL)-1,3,4-OXADIAZOL-2-YL]HEPTAN-1-ONE",
            "non-polymer",
        ),
        "JG2": (
            "7-PHENYL-1-[5-(PYRIDIN-2-YL)-1,3,4-OXADIAZOL-2-YL]HEPTANE-1,1-DIOL",
            "non-polymer",
        ),
        "HBR": ("R,3-HYDROXYBUTAN-2-ONE", "non-polymer"),
        "GEO": ("GEMCITABINE", "non-polymer"),
        "GEN": ("GENISTEIN", "non-polymer"),
        "GEM": ("(2-GUANIDINOETHYLMERCAPTO)SUCCINIC ACID", "non-polymer"),
        "GEL": (
            "1-O-OCTYL-2-HEPTYLPHOSPHONYL-SN-GLYCERO-3-PHOSPHOETHANOLAMINE",
            "non-polymer",
        ),
        "GE3": ("5-METHYL-4-METHYLAMINO-TETRAHYDRO-PYRAN-2,3,5-TRIOL", "non-polymer"),
        "DDZ": ("3,3-DIHYDROXY L-ALANINE", "peptide"),
        "DDY": ("2',3'-DIDEOXYCYTOSINE-5'-DIPHOSPHATE", "non-polymer"),
        "DDX": ("2',3'-DEHYDRO-2',3'-DIDEOXYRIBOFURANOSE-5'-PHOSPHATE", "nucleic-acid"),
        "VMX": ("5'-{[(3-AMINOPROPYL)SULFONYL]AMINO}-5'-DEOXYADENOSINE", "non-polymer"),
        "TC8": (
            "3-(2,6-DIFLUOROPHENYL)-2-(METHYLTHIO)QUINAZOLIN-4(3H)-ONE",
            "non-polymer",
        ),
        "TC9": ("D-TUBOCURARINE", "non-polymer"),
        "DDS": ("?", "non-polymer"),
        "DDR": ("(2S)-3-HYDROXYPROPANE-1,2-DIYL DIDECANOATE", "non-polymer"),
        "DDQ": ("DECYLAMINE-N,N-DIMETHYL-N-OXIDE", "non-polymer"),
        "DDP": ("2,4-DIAMINO-4,6-DIHYDROXYPYRIMIDINE", "non-polymer"),
        "TC2": (
            "N-{2-[BIS(2-{[(2,3-DIHYDROXYPHENYL)CARBONYL]AMINO}ETHYL)AMINO]ETHYL}-1-HYDROXY-6-OXO-1,6-DIHYDROPYRIDINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "RYX": ("N-(2,2,3,3,3-PENTAFLUOROPROPYL)-4-SULFAMOYLBENZAMIDE", "non-polymer"),
        "DDU": ("2'-5'DIDEOXYURIDINE", "non-polymer"),
        "DDT": ("N,O-DIDANSYL-L-TYROSINE", "non-polymer"),
        "DDJ": ("3,3'-DIMETHOXYBIPHENYL-4,4'-DIAMINE", "non-polymer"),
        "DDI": ("?", "non-polymer"),
        "DDH": (
            "[7,12-DEACETYL-3,8,13,17-TETRAMETHYL-21H,23H-PORPHINE-2,18-DIPROPANOATO(2-)-N21,N22,N23,N24]-IRON",
            "non-polymer",
        ),
        "DDO": ("6-HYDROXY-D-NORLEUCINE", "non-polymer"),
        "DDN": ("3,4-DIHYDRO-2'-DEOXYURIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "DDM": ("2-HYDROXYMETHYL-PIPERIDINE-3,4,5-TRIOL", "non-polymer"),
        "DDL": ("2,6-DIDEOXY-BETA-D-GALACTOSE", "saccharide"),
        "DDC": ("7-HYDROXY-2-PHENYL-CHROMAN-4-ONE", "non-polymer"),
        "DDB": ("3-METHYL-2,6-DIDEOXY-BETA-D-ALLOPYRANOSE", "saccharide"),
        "DDA": ("2,6-DIDEOXY-BETA-D-GLUCOSE", "saccharide"),
        "DDG": ("2',3'-DIDEOXY-GUANOSINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "DDF": ("5,10-DIDEAZATETRAHYDROFOLIC ACID", "non-polymer"),
        "DDE": (
            "{3-[4-(2-AMINO-2-CARBOXY-ETHYL)-1H-IMIDAZOL-2-YL]-1-CARBAMOYL-PROPYL}-TRIMETHYL-AMMONIUM",
            "peptide",
        ),
        "DDD": (
            "(5R,10S)-5-{[(CARBOXYMETHYL)AMINO]CARBONYL}-7-OXO-3-THIA-1,6-DIAZABICYCLO[4.4.1]UNDECANE-10-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "PA4": ("PAROMOMYCIN (RING 4)", "non-polymer"),
        "SK7": (
            "6-[3-HYDROXY-2-(HYDROXYMETHYL)PROP-1-EN-1-YL]-1,5-DIMETHYLPYRIMIDINE-2,4(1H,3H)-DIONE",
            "non-polymer",
        ),
        "TUC": ("(ISOQUINOLIN-5-YLOXY)ACETIC ACID", "non-polymer"),
        "SK1": (
            "N-(BENZYLSULFONYL)-D-SERYL-N-{4-[AMINO(IMINO)METHYL]BENZYL}-L-ALANINAMIDE",
            "non-polymer",
        ),
        "SK3": (
            "(1R,2R,3S,4S,5R)-5-(BENZYLAMINO)CYCLOPENTANE-1,2,3,4-TETROL",
            "non-polymer",
        ),
        "SK2": (
            "(3R)-METHYLCARBAMOYL-7-SULFOAMINO-3,4-DIHYDRO-1H-ISOQUINOLINE-2-CARBOXYLIC ACID BENZYL ESTER",
            "non-polymer",
        ),
        "TUD": ("TAUROCHENODEOXYCHOLIC ACID", "non-polymer"),
        "VMS": ("5'O-[N-(L-VALYL)SULPHAMOYL]ADENOSINE", "peptide"),
        "XG4": (
            "2'-DEOXY-5'-O-[(R)-HYDROXY{[(R)-HYDROXY(PHOSPHONOOXY)PHOSPHORYL]AMINO}PHOSPHORYL]GUANOSINE",
            "non-polymer",
        ),
        "J01": (
            "(2R,3Z,5R)-3-(2-HYDROXYETHYLIDENE)-7-OXO-4-OXA-1-AZABICYCLO[3.2.0]HEPTANE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "J07": (
            "4-{[5-CHLORO-4-(1H-INDOL-3-YL)PYRIMIDIN-2-YL]AMINO}-N-ETHYLPIPERIDINE-1-CARBOXAMIDE",
            "non-polymer",
        ),
        "K13": (
            "(3R,3AS,6AR)-HEXAHYDROFURO[2,3-B]FURAN-3-YL [(1S,2R)-3-{[(4-AMINOPHENYL)SULFONYL][(2S)-2-METHYLBUTYL]AMINO}-1-BENZYL-2-HYDROXYPROPYL]CARBAMATE",
            "non-polymer",
        ),
        "K12": ("12-HYDROXYDODECYL DIHYDROGEN PHOSPHATE", "non-polymer"),
        "K11": (
            "N-[4-({3-[2-(METHYLAMINO)PYRIMIDIN-4-YL]PYRIDIN-2-YL}OXY)NAPHTHALEN-1-YL]-6-(TRIFLUOROMETHYL)-1H-BENZIMIDAZOL-2-AMINE",
            "non-polymer",
        ),
        "K10": (
            "1-TERT-BUTYL-3-(2,5-DIMETHYLBENZYL)-1H-PYRAZOLO[3,4-D]PYRIMIDIN-4-AMINE",
            "non-polymer",
        ),
        "K17": ("4,5,6,7-TETRABROMO-BENZIMIDAZOLE", "non-polymer"),
        "K15": (
            "5'-{[(3S)-3-AMINO-3-CARBOXYPROPYL](HEXYL)AMINO}-5'-DEOXYADENOSINE",
            "non-polymer",
        ),
        "K14": (
            "(3R,3AS,6AR)-HEXAHYDROFURO[2,3-B]FURAN-3-YL [(1S,2R)-1-BENZYL-2-HYDROXY-3-{[(4-METHOXYPHENYL)SULFONYL][(2S)-2-METHYLBUTYL]AMINO}PROPYL]CARBAMATE",
            "non-polymer",
        ),
        "K19": (
            "(3R,3AS,6AR)-HEXAHYDROFURO[2,3-B]FURAN-3-YL [(1S,2R)-3-{(1,3-BENZOTHIAZOL-6-YLSULFONYL)[(2S)-2-METHYLBUTYL]AMINO}-1-BENZYL-2-HYDROXYPROPYL]CARBAMATE",
            "non-polymer",
        ),
        "TCA": ("PHENYLETHYLENECARBOXYLIC ACID", "non-polymer"),
        "BJ1": (
            "(1R,2S)-1-[(1S)-1,2-DIHYDROXYETHYL]-3-[(2R,3S,4S)-3,4-DIHYDROXY-2-(HYDROXYMETHYL)TETRAHYDROTHIOPHENIUM-1-YL]-2-HYDROXYPROPYL SULFATE",
            "non-polymer",
        ),
        "ZEM": ("20-OXO-PROTOPORPHYRIN IX CONTAINING ZN(II)", "non-polymer"),
        "HE4": ("HEPTAN-1-OL", "non-polymer"),
        "HE5": ("ZINC(II)-DEUTEROPORPHYRIN DIMETHYLESTER", "non-polymer"),
        "HE6": ("6,7-DICARBOXYL-1,2,3,4,5,8-HEXAMETHYLHEMIN", "non-polymer"),
        "HE7": (
            "4-{[(14BETA,17ALPHA)-3-HYDROXYESTRA-1,3,5(10)-TRIEN-17-YL]OXY}-4-OXOBUTANOIC ACID",
            "non-polymer",
        ),
        "HE0": (
            "(2R,4S)-2-[(1R)-1-{[(6S)-6-CARBOXY-6-(GLYCYLAMINO)HEXANOYL]AMINO}-2-OXOETHYL]-5,5-DIMETHYL-1,3-THIAZOLIDINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "HE1": ("4-(2-HYDROXYPHENYLTHIO)-1-BUTENYLPHOSPHONIC ACID", "non-polymer"),
        "HE2": ("HEXAN-1-OL", "non-polymer"),
        "HE3": ("2-HYDROXYETHYL DIHYDROTHIACHROME DIPHOSPHATE", "non-polymer"),
        "KDD": ("2,6-ANHYDRO-3,5-DIDEOXY-D-RIBO-OCT-2-ENONIC ACID", "non-polymer"),
        "KDE": (
            "PROP-2-EN-1-YL 3-DEOXY-BETA-L-GULO-OCT-2-ULOPYRANOSIDONIC ACID",
            "non-polymer",
        ),
        "KDF": ("3-DEOXY-ALPHA-D-ERYTHRO-HEX-2-ULOFURANOSONIC ACID", "non-polymer"),
        "KDG": ("2-KETO-3-DEOXYGLUCONATE", "non-polymer"),
        "QEM": (
            "4-[(1R,2S)-3-(4-BENZYLPIPERIDIN-1-YL)-1-HYDROXY-2-METHYLPROPYL]PHENOL",
            "non-polymer",
        ),
        "KDA": ("(3-DEOXY-D-MANNO-OCT-2-ULOSONIC ACID)-2-O-ALLYL", "saccharide"),
        "KDB": (
            "3,4,5-TRIDEOXY-ALPHA-D-ERYTHRO-OCT-3-EN-2-ULOPYRANOSONIC ACID",
            "non-polymer",
        ),
        "SOL": ("L-SORBOSE", "saccharide"),
        "KDP": ("2-KETO-DEOXY-GALACTOSE", "non-polymer"),
        "ND4": ("AMMONIUM CATION WITH D", "non-polymer"),
        "KDR": (
            "PROP-2-EN-1-YL 3-DEOXY-ALPHA-D-MANNO-OCTOS-2-ULOPYRANOSIDE",
            "non-polymer",
        ),
        "HET": (
            "3-[2-(2-CYCLOPENTYL-6-{[4-(DIMETHYLPHOSPHORYL)PHENYL]AMINO}-9H-PURIN-9-YL)ETHYL]PHENOL",
            "non-polymer",
        ),
        "HEU": ("3-(2-HYDROXYETHYL)-2'-DEOXYURIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "HEV": (
            "5,8-DIMETHYL-1,2,3,4-TETRAVINYLPORPHINE-6,7-DIPROPIONIC ACID FERROUS COMPLEX",
            "non-polymer",
        ),
        "HEP": ("PHENYL[1-(N-SUCCINYLAMINO)PENTYL]PHOSPHONATE", "non-polymer"),
        "HEQ": (
            "3-[18-(4-HYDROXY-2,6,6-TRIMETHYL-CYCLOHEX-1-ENYL)-3,7,12,16-TETRAMETHYL-OCTADECA-1,3,5,7,9,11,13,15, 17-NONAENYL]-2,4,4-TRIMETHYL-CYCLOHEX-2-ENONE",
            "non-polymer",
        ),
        "NDO": (
            "NICOTINAMIDE-(6-DEAMINO-6-HYDROXY-ADENINE)-DINUCLEOTIDE PHOSPHATE",
            "non-polymer",
        ),
        "HES": ("ZINC SUBSTITUTED HEME C", "non-polymer"),
        "NDA": ("3-AMINOMETHYL-PYRIDINIUM-ADENINE-DINUCLEOTIDE", "non-polymer"),
        "NDC": ("NICOTINAMIDE ADENINE DINUCLEOTIDE CYCLOHEXANONE", "non-polymer"),
        "NDB": (
            "N-(2-HYDROXYETHYL)-N,N-DIMETHYL-3-SULFOPROPAN-1-AMINIUM",
            "non-polymer",
        ),
        "HEX": ("HEXANE", "non-polymer"),
        "HEY": (
            "(2E)-2-[({3-HYDROXY-2-METHYL-5-[(PHOSPHONOOXY)METHYL]PYRIDIN-4-YL}METHYL)AMINO]-5-PHOSPHONOPENT-2-ENOIC ACID",
            "non-polymer",
        ),
        "HEZ": ("HEXANE-1,6-DIOL", "non-polymer"),
        "NDF": ("N-(CARBOXYCARBONYL)-D-PHENYLALANINE", "peptide"),
        "HED": ("2-HYDROXYETHYL DISULFIDE", "non-polymer"),
        "HEE": ("N-HEXYLPHOSPHONATE ETHYL ESTER", "non-polymer"),
        "HEF": ("1-(2-HYDROXYETHYLOXYMETHYL)-6-PHENYL THIOTHYMINE", "non-polymer"),
        "HEG": ("PROTOPORPHYRIN IX CONTAINING MG", "non-polymer"),
        "HEA": ("HEME-A", "non-polymer"),
        "HEB": ("HEME B/C", "non-polymer"),
        "HEC": ("HEME C", "non-polymer"),
        "HEL": (
            "(2S,5R,6R)-6-{[(6R)-6-(GLYCYLAMINO)-7-OXIDO-7-OXOHEPTANOYL]AMINO}-3,3-DIMETHYL-7-OXO-4-THIA-1-AZABICYCLO[3.2.0]HEPTANE-2-CARBOXYLATE",
            "non-polymer",
        ),
        "HEM": ("PROTOPORPHYRIN IX CONTAINING FE", "non-polymer"),
        "HEN": (
            "2-[(3-HYDROXY-2-METHYL-5-PHOSPHONOOXYMETHYL-PYRIDIN-4-YLMETHYL)-IMINO]-5-PHOSPHONO-PENT-3-ENOIC ACID",
            "non-polymer",
        ),
        "HEO": ("HEME O", "non-polymer"),
        "HEH": ("HEDAMYCIN", "non-polymer"),
        "KD1": (
            "3-DEOXY-ALPHA-D-MANNO-OCT-2-ULOPYRANONOSYL-(2->8)-3-DEOXY-ALPHA-D-MANNO-OCT-2-ULOPYRANONOSYL-(2->4)-3-DEOXY-ALPHA-D-MANNO-OCT-2-ULOPYRANONOSYL-(2->6)-2-AMINO-2-DEOXY-4-O-PHOSPHONO-BETA-D-GLUCOPYRANOSE",
            "non-polymer",
        ),
        "NDV": (
            "(2S,4S)-2-[(1R)-1-{[(2R)-2-AMINO-2-PHENYLACETYL]AMINO}-2-OXOETHYL]-5,5-DIMETHYL-1,3-THIAZOLIDINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "964": (
            "3-(4-{2-[2-(2-BROMO-ACETYLAMINO)-ETHYLDISULFANYL]-ETHYLCARBAMOYL}-CYCLOHEXYLCARBAMOYL)-PYRAZINE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "965": (
            "[3-(3-{[2-CHLORO-3-(TRIFLUOROMETHYL)BENZYL](2,2-DIPHENYLETHYL)AMINO}PROPOXY)PHENYL]ACETIC ACID",
            "non-polymer",
        ),
        "BJC": (
            "(1R,3S)-N-[(1S,2R)-1-BENZYL-2-HYDROXY-3-{[3-(1-METHYLETHYL)BENZYL]AMINO}PROPYL]-3-[1-METHYL-1-(2-OXOPIPERIDIN-1-YL)ETHYL]CYCLOHEXANECARBOXAMIDE",
            "non-polymer",
        ),
        "961": (
            "3-FLUORO-4-[2-HYDROXY-2-(5,5,8,8-TETRAMETHYL-5,6,7,8,-TETRAHYDRO-NAPHTALEN-2-YL)-ACETYLAMINO]-BENZOIC ACID",
            "non-polymer",
        ),
        "BJA": (
            "(2S)-1-[(2R)-2-(BENZYLSULFONYLAMINO)-5-GUANIDINO-PENTANOYL]-N-[(4-CARBAMIMIDOYLPHENYL)METHYL]PYRROLIDINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "BJM": (
            "(1S,3S,5S)-2-{(2S)-2-AMINO-2-[(1R,3S,5R,7S)-3-HYDROXYTRICYCLO[3.3.1.1~3,7~]DEC-1-YL]ACETYL}-2-AZABICYCLO[3.1.0]HEXANE-3-CARBONITRILE",
            "non-polymer",
        ),
        "PJJ": (
            "(2R,4S,5S)-5-AMINO-2-BENZYL-4-HYDROXY-6-PHENYLHEXANOIC ACID",
            "peptide",
        ),
        "968": (
            "2-[(7-HYDROXY-NAPHTHALEN-1-YL)-OXALYL-AMINO]-BENZOIC ACID",
            "non-polymer",
        ),
        "BJI": (
            "1(R)-1-ACETAMIDO-2-(3-CARBOXYPHENYL)ETHYL BORONIC ACID",
            "non-polymer",
        ),
        "BJH": (
            "1(R)-1-ACETAMIDO-2-(3-CARBOXY-2-HYDROXYPHENYL)ETHYL BORONIC ACID",
            "peptide",
        ),
        "SON": ("ADENOSINE PHOSPHONOACETIC ACID", "non-polymer"),
        "BJP": (
            "(1R)-2-PHENYLACETAMIDO-2-(3-CARBOXYPHENYL)ETHYL BORONIC ACID",
            "non-polymer",
        ),
        "V1N": (
            "[(2S,3R,4R,5R)-5-(6-AMINOPURIN-9-YL)-3,4-BIS(OXIDANYL)OXOLAN-2-YL] 2-(1H-INDOL-3-YL)ETHYL HYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "SCH": ("S-METHYL-THIO-CYSTEINE", "peptide"),
        "BJZ": (
            "6-[(5S,9R)-9-(4-CYANOPHENYL)-3-(3,5-DICHLOROPHENYL)-1-METHYL-2,4-DIOXO-1,3,7-TRIAZASPIRO[4.4]NON-7-YL]PYRIDINE-3-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "PJA": (
            "4-FORMYL-3-HYDROXY-8-METHOXY-1,9-DIMETHYL-11-OXO-11H-DIBENZO[B,E][1,4]DIOXEPINE-6-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "PJC": (
            "3-(1H-INDOL-3-YL)-4-[2-(4-OXOPIPERIDIN-1-YL)-5-(TRIFLUOROMETHYL)PYRIMIDIN-4-YL]-1H-PYRROLE-2,5-DIONE",
            "non-polymer",
        ),
        "OWL": (
            "N-1H-IMIDAZOL-2-YL-N'-[4-(1H-IMIDAZOL-2-YLAMINO)PHENYL]BENZENE-1,4-DIAMINE",
            "non-polymer",
        ),
        "NCQ": ("3,6-BIS[3-(AZEPAN-1-YL)PROPIONAMIDO]ACRIDINE", "non-polymer"),
        "DT9": (
            "(1S)-6,7-DIMETHOXY-1-METHYL-3,4-DIHYDROISOQUINOLINE-2(1H)-SULFONAMIDE",
            "non-polymer",
        ),
        "DT8": (
            "N-((1R,2S,5S)-2-(5-CHLORO-1H-INDOLE-2-CARBOXAMIDO)-5-(DIMETHYLCARBAMOYL)CYCLOHEXYL)-5-METHYL-4,5,6,7-TETRAHYDROTHIAZOLO[5,4-C]PYRIDINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "R52": ("5-O-PHOSPHONO-D-RIBOSE", "non-polymer"),
        "R51": ("5-O-PHOSPHONO-D-RIBOSE", "non-polymer"),
        "R56": (
            "5-OXO-6-PHENYLMETHANESULFONYLAMINO-HEXAHYDRO-THIAZOLO[3,2-A]PYRIDINE-3-CARBOXYLIC ACID (3-GUANIDINO-PROPYL)-AMIDE",
            "non-polymer",
        ),
        "R55": (
            "N-[4-({4-[5-(4,4-DIMETHYLPIPERIDIN-1-YL)-2-HYDROXYBENZOYL]BENZOYL}AMINO)AZEPAN-3-YL]ISONICOTINAMIDE",
            "non-polymer",
        ),
        "LVS": (
            "4-IODO-3-NITROPHENYL ACETYL-LEUCINYL-LEUCINYL-LEUCINYL-VINYLSULFONE",
            "non-polymer",
        ),
        "R5B": ("RUTHENIUM WIRE WC5", "non-polymer"),
        "R5A": ("RUTHENIUM WIRE, 5 CARBON LINKER", "non-polymer"),
        "PPM": ("PHOSPHONOMETHYL GROUP", "non-polymer"),
        "DT3": ("(2S,3S)-2,3-DIHYDROXY-4-SULFANYLBUTANE-1-SULFONATE", "non-polymer"),
        "LVG": ("L-VINYLGLYCINE", "peptide"),
        "W43": (
            "5-(7-(6-CHLORO-4-(5-HYDRO-4-METHYL-2-OXAZOLYL)PHENOXY)HEPTYL)-3-METHYL ISOXAZOLE",
            "non-polymer",
        ),
        "LVA": (
            "(3R,5R)-7-((1R,2R,6S,8R,8AS)-2,6-DIMETHYL-8-{[(2R)-2-METHYLBUTANOYL]OXY}-1,2,6,7,8,8A-HEXAHYDRONAPHTHALEN-1-YL)-3,5-DIHYDROXYHEPTANOIC ACID",
            "non-polymer",
        ),
        "YDK": (
            "2-(CARBAMOYLAMINO)-5-PHENYL-N-[(3S)-PIPERIDIN-3-YL]THIOPHENE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "ZYY": (
            "N-[(6S)-2-(4-FLUOROBENZYL)-10-HYDROXY-6-METHYL-8-(1-METHYLETHYL)-1,9-DIOXO-1,2,6,7,8,9-HEXAHYDROPYRAZINO[1',2':1,5]PYRROLO[2,3-D]PYRIDAZIN-4-YL]-N-METHYLMETHANESULFONAMIDE",
            "non-polymer",
        ),
        "LVO": (
            "5-(ACETYLAMINO)-2,6-ANHYDRO-4-CARBAMIMIDAMIDO-3,4,5-TRIDEOXY-7-O-METHYL-9-O-OCTANOYL-D-GLYCERO-D-GALACTO-NON-2-ENONIC ACID",
            "non-polymer",
        ),
        "LVN": ("(3S)-4-OXO-L-VALINE", "peptide"),
        "FNZ": (
            "4-[(1S,2R)-3-(4-FLUOROPHENYL)-2-HYDROXY-1-(1H-1,2,4-TRIAZOL-1-YL)PROPYL]BENZONITRILE",
            "non-polymer",
        ),
        "DT1": (
            "4-{[5-(CYCLOHEXYLMETHOXY)[1,2,4]TRIAZOLO[1,5-A]PYRIMIDIN-7-YL]AMINO}BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "XE": ("XENON", "non-polymer"),
        "CBY": ("COB(II)INAMIDE", "non-polymer"),
        "SVV": ("O-[(S)-AMINO(HYDROXY)PHOSPHORYL]-L-SERINE", "peptide"),
        "MYY": (
            "(2R)-3-(PHOSPHONOOXY)-2-(TETRADECANOYLOXY)PROPYL PALMITATE",
            "non-polymer",
        ),
        "MYX": (
            "(2Z,6E)-7-{2'-[(2E,4E)-1,6-DIMETHYLHEPTA-2,4-DIENYL]-2,4'-BI-1,3-THIAZOL-4-YL}-3,5-DIMETHOXY-4-METHYLHEPTA-2,6-DIENAMIDE",
            "non-polymer",
        ),
        "P74": (
            "TERT-BUTYL [3-(3-{[4-(DIETHYLAMINO)PIPERIDIN-1-YL]METHYL}-4-HYDROXY-5-IODOPHENYL)PROP-2-YN-1-YL]CARBAMATE",
            "non-polymer",
        ),
        "MYQ": ("(S)-(1,2-DICARBOXYETHYLTHIO)GOLD", "non-polymer"),
        "MYP": (
            "4-MORPHOLIN-4-YL-PIPERIDINE-1-CARBOXYLIC ACID [1-(3-BENZENESULFONYL-1-PROPYL-ALLYLCARBAMOYL)-2-PHENYLETHYL]-AMIDE",
            "non-polymer",
        ),
        "D6G": ("2-DEOXY-6-O-PHOSPHONO-ALPHA-D-ARABINO-HEXOPYRANOSE", "saccharide"),
        "MYR": ("MYRISTIC ACID", "non-polymer"),
        "MYU": (
            "3,5,6,7-TETRAHYDROXY-2-(3,4-DIHYDROXYPHENYL)-4H-CHROMEN-4-ONE",
            "non-polymer",
        ),
        "MYT": ("METYRAPONE", "non-polymer"),
        "P78": (
            "1-{5-TERT-BUTYL-3-[(1,1-DIOXIDOTHIOMORPHOLIN-4-YL)CARBONYL]THIOPHEN-2-YL}-3-NAPHTHALEN-2-YLUREA",
            "non-polymer",
        ),
        "MYV": ("MYCINAMICIN V", "non-polymer"),
        "MYI": ("(5-METHOXY-1H-INDOL-3-YL)ACETIC ACID", "non-polymer"),
        "MYK": ("N~6~-TETRADECANOYL-L-LYSINE", "peptide"),
        "MYL": ("MYCALAMIDE A", "non-polymer"),
        "RCB": (
            "4-NITROPHENYL BETA-D-GLUCOPYRANOSYL-(1->4)-BETA-D-GLUCOPYRANOSYL-(1->4)-BETA-D-GLUCOPYRANOSIDE",
            "non-polymer",
        ),
        "MYN": (
            "(2S)-AMINO[(4R)-2-AMINO-1,4,5,6-TETRAHYDROPYRIMIDIN-4-YL]ETHANOIC ACID",
            "peptide",
        ),
        "MYA": ("TETRADECANOYL-COA", "non-polymer"),
        "MYC": (
            "3,5,7-TRIHYDROXY-2-(3,4,5-TRIHYDROXYPHENYL)-4H-CHROMEN-4-ONE",
            "non-polymer",
        ),
        "CQ0": (
            "3-(3-TERT-BUTYL[1,2,4]TRIAZOLO[4,3-A]PYRIDIN-7-YL)-N-CYCLOPROPYL-4-METHYLBENZAMIDE",
            "non-polymer",
        ),
        "MYE": (
            "N2-({[(4-BROMOPHENYL)METHYL]OXY}CARBONYL)-N1-[(1S)-1-FORMYLPENTYL]-L-LEUCINAMIDE",
            "non-polymer",
        ),
        "D6P": ("(3,4,5,7-TETRAHYDROXY-HEPT-1-ENYL)-PHOSPHONIC ACID", "non-polymer"),
        "MYG": ("GLUCOSAMINYL-(ALPHA-6)-D-MYO-INOSITOL", "non-polymer"),
        "MYF": (
            "5,7-DIHYDROXY-2-(3,4,5-TRIHYDROXYPHENYL)-4H-CHROMEN-4-ONE",
            "non-polymer",
        ),
        "MY9": (
            "N-{(1S,2S)-1-BENZYL-2-HYDROXY-2-[(4S)-1,2,2-TRIMETHYL-5-OXOIMIDAZOLIDIN-4-YL]ETHYL}-N'-[(1R)-1-(4-FLUOROPHENYL)ETHYL]-5-[METHYL(METHYLSULFONYL)AMINO]ISOPHTHALAMIDE",
            "non-polymer",
        ),
        "NT2": ("2,3,4,6-TETRA-O-SULFOHEXOPYRANOSE", "non-polymer"),
        "CQL": ("5-CHLORO-7-IODOQUINOLIN-8-OL", "non-polymer"),
        "MY1": (
            "[N-(3-DIBENZYLCARBAMOYL-OXIRANECARBONYL)-HYDRAZINO]-ACETIC ACID",
            "non-polymer",
        ),
        "MY0": (
            "(2S)-4-[1-(CARBOXYMETHYL)HYDRAZINYL]-2-HYDROXY-4-OXOBUTANOIC ACID",
            "peptide",
        ),
        "CQA": (
            "4-[(7-CHLOROQUINOLIN-4-YL)AMINO]-2-[(DIETHYLAMINO)METHYL]PHENOL",
            "non-polymer",
        ),
        "MY2": (
            "{1-[(3S)-4-(BENZYLOXY)-3-HYDROXY-4-OXOBUTANOYL]HYDRAZINO}ACETIC ACID",
            "non-polymer",
        ),
        "MY5": (
            "{1-[(3S)-4-{[(1S)-2-(BENZYLAMINO)-1-METHYL-2-OXOETHYL]AMINO}-3-HYDROXY-4-OXOBUTANOYL]HYDRAZINO}ACETIC ACID",
            "non-polymer",
        ),
        "MY4": (
            "(1-{(3S)-3-HYDROXY-4-OXO-4-[(2-PHENYLETHYL)AMINO]BUTANOYL}HYDRAZINO)ACETIC ACID",
            "non-polymer",
        ),
        "MY6": ("2-AMINO-7-METHYL-1,7-DIHYDRO-6H-PURIN-6-ONE", "non-polymer"),
        "P7B": (
            "1-{5-TERT-BUTYL-3-[(5-OXO-1,4-DIAZEPAN-1-YL)CARBONYL]THIOPHEN-2-YL}-3-NAPHTHALEN-1-YLUREA",
            "non-polymer",
        ),
        "P7C": (
            "1-[5-TERT-BUTYL-3-({4-[2-(DIMETHYLAMINO)ETHYL]-5-OXO-1,4-DIAZEPAN-1-YL}CARBONYL)THIOPHEN-2-YL]-3-(2,3-DICHLOROPHENYL)UREA",
            "non-polymer",
        ),
        "NCW": (
            "(1S,2R,3S,4R,5R)-2,3,4-TRIHYDROXY-N-OCTYL-6-OXA-8-AZABICYCLO[3.2.1]OCTANE-8-CARBOTHIOAMIDE",
            "non-polymer",
        ),
        "CQR": (
            "[(4Z)-2-(AMINOMETHYL)-4-(4-HYDROXYBENZYLIDENE)-5-OXO-4,5-DIHYDRO-1H-IMIDAZOL-1-YL]ACETIC ACID",
            "peptide",
        ),
        "CQQ": (
            "4-{[4-(DIMETHYLAMINO)BUTANOYL]AMINO}-N-(3-{[4-(PYRIDIN-3-YL)PYRIMIDIN-2-YL]AMINO}PHENYL)BENZAMIDE",
            "non-polymer",
        ),
        "CQP": (
            "(4R,2S)-5'-(4-(4-CHLOROBENZYLOXY)PYRROLIDIN-2-YLMETHANESULFONYL)ISOQUINOLINE",
            "non-polymer",
        ),
        "CQW": (
            "5-(5-CHLORO-7H-PYRROLO[2,3-D]PYRIMIDIN-4-YL)-4,5,6,7-TETRAHYDRO-1H-IMIDAZO[4,5-C]PYRIDINE",
            "non-polymer",
        ),
        "TDR": ("THYMINE", "non-polymer"),
        "CQU": (
            "N-[2-(5-METHYL-4H-1,2,4-TRIAZOL-3-YL)PHENYL]-7H-PYRROLO[2,3-D]PYRIMIDIN-4-AMINE",
            "non-polymer",
        ),
        "P7I": ("(2-AMINOETHYL)PHOSPHONIC ACID", "non-polymer"),
        "5OP": (
            "(5E,13E)-11-HYDROXY-9,15-DIOXOPROSTA-5,13-DIEN-1-OIC ACID",
            "non-polymer",
        ),
        "Y2B": (
            "6-{[(3R,4R)-4-(2-{[(1S,2R)-2-(3-FLUOROPHENYL)CYCLOPROPYL]AMINO}ETHOXY)PYRROLIDIN-3-YL]METHYL}-4-METHYLPYRIDIN-2-AMINE",
            "non-polymer",
        ),
        "5OC": ("?", "nucleic-acid"),
        "5OB": ("5-METHOXYBENZIMIDAZOLE", "non-polymer"),
        "5OH": (
            "(2S)-AMINO[(4R,6S)-2-AMINO-6-HYDROXY-3,4,5,6-TETRAHYDROPYRIMIDIN-4-YL]ETHANOIC ACID",
            "peptide",
        ),
        "ZLD": (
            "N-{[(5S)-3-(3-FLUORO-4-MORPHOLIN-4-YLPHENYL)-2-OXO-1,3-OXAZOLIDIN-5-YL]METHYL}ACETAMIDE",
            "non-polymer",
        ),
        "SOY": ("OXACILLIN-ACYLATED SERINE", "peptide"),
        "AL8": (
            "(R)-3,4-DIHYDRO-2-(3-METHOXYPHENYL)-4-METHYLAMINO-2H-THIENO[3,2-E]-1,2-THIAZINE-6-SULFONAMIDE-1,1-DIOXIDE",
            "non-polymer",
        ),
        "21Z": (
            "3-[(4-AMINO-5-BENZOYL-1,3-THIAZOL-2-YL)AMINO]BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "3M4": ("4-[5-(2-ETHOXYPHENYL)-1,2,4-OXADIAZOL-3-YL]PYRIDINE", "non-polymer"),
        "3M3": (
            "4-[3-(2-FLUOROPHENOXY)PROPYL]-3,5-DIMETHYL-1H-PYRAZOLE",
            "non-polymer",
        ),
        "3M2": (
            "(5Z)-2-AMINO-5-(3,5-DIBROMO-4-HYDROXYBENZYLIDENE)-1-METHYL-1,5-DIHYDRO-4H-IMIDAZOL-4-ONE",
            "non-polymer",
        ),
        "3M1": (
            "(3,4-DIHYDROXY-5-NITROPHENYL)(2-FLUOROPHENYL)METHANONE",
            "non-polymer",
        ),
        "AL6": (
            "2-(3-METHOXYPHENYL)-2H-THIENO-[3,2-E]-1,2-THIAZINE-6-SULFINAMIDE-1,1-DIOXIDE",
            "non-polymer",
        ),
        "PR6": ("3-CHLORO-3-OXOPROPANOIC ACID", "non-polymer"),
        "AL7": (
            "(S)-3,4-DIHYDRO-2-(3-METHOXYPHENYL)-4-METHYLAMINO-2H-THIENO[3,2-E]-1,2-THIAZINE-6-SULFONAMIDE-1,1-DIOXIDE",
            "non-polymer",
        ),
        "AL4": (
            "(R)-4-ETHYLAMINO-3,4-DIHYDRO-2-(2-METHOYLETHYL)-2H-THIENO[3,2-E]-1,2-THIAZINE-6-SULFONAMIDE-1,1-DIOXIDE",
            "non-polymer",
        ),
        "30X": (
            "N~2~-[4-CYANO-3-(TRIFLUOROMETHYL)PHENYL]-N,N-DIMETHYL-N~2~-(2,2,2-TRIFLUOROETHYL)-L-ALANINAMIDE",
            "non-polymer",
        ),
        "30Z": (
            "4-[(2R,3S)-3-[(3,4-DIHYDROXYPHENYL)METHYL]-2-METHYLBUTYL]BENZENE-1,2-DIOL",
            "non-polymer",
        ),
        "T32": ("?", "nucleic-acid"),
        "KTH": ("2-KETOTHIAZOLE", "non-polymer"),
        "T33": ("3,3'-DEIODO-THYROXINE", "non-polymer"),
        "30K": (
            "(3Z)-5-[(1-ETHYLPIPERIDIN-4-YL)AMINO]-3-[(5-METHOXY-1H-BENZIMIDAZOL-2-YL)(PHENYL)METHYLIDENE]-1,3-DIHYDRO-2H-INDOL-2-ONE",
            "non-polymer",
        ),
        "B8M": ("(8E,10Z)-HEXADECA-8,10-DIEN-1-OL", "non-polymer"),
        "B8L": (
            "3-[(3-SEC-BUTYL-4-HYDROXYBENZOYL)AMINO]AZEPAN-4-YL 4-(2-HYDROXY-5-METHOXYBENZOYL)BENZOATE",
            "non-polymer",
        ),
        "30A": (
            "6-BENZYL-5-CHLORO-7-HYDROXYPYRAZOLO[1,5-A]PYRIMIDINE-3-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "30B": ("?", "non-polymer"),
        "AL1": (
            "3,,4-DIHYDRO-2-(3-METHOXYPHENYL)-2H-THIENO-[3,2-E]-1,2-THIAZINE-6-SULFONAMIDE-1,1-DIOXIDE",
            "non-polymer",
        ),
        "B8D": ("4-O-METHYL-ALPHA-D-GLUCOPYRANOSE", "saccharide"),
        "XJ0": ("(E)-3-(2-AMINO-PYRIDINE-5YL)-ACRYLIC ACID", "non-polymer"),
        "B89": (
            "(5S)-2-[(1E)-N-{[(2E)-3-CHLOROPROP-2-EN-1-YL]OXY}PROPANIMIDOYL]-3-HYDROXY-5-(TETRAHYDRO-2H-PYRAN-4-YL)CYCLOHEX-2-EN-1-ONE",
            "non-polymer",
        ),
        "B88": (
            "(2S)-2-{[(S)-[(1R)-1-AMINOETHYL](HYDROXY)PHOSPHORYL]METHYL}BUTANEDIOIC ACID",
            "non-polymer",
        ),
        "B81": ("(3ALPHA,8ALPHA,17BETA)-ANDROST-5-ENE-3,17-DIOL", "non-polymer"),
        "B80": (
            "N-{3-[(6S)-6-ETHYL-1-(4-FLUOROBENZYL)-4-HYDROXY-2-OXO-1,2,5,6-TETRAHYDROPYRIDIN-3-YL]-1,1-DIOXIDO-2H-1,2,4-BENZOTHIADIAZIN-7-YL}METHANESULFONAMIDE",
            "non-polymer",
        ),
        "B83": (
            "2-[({4-[(3R,5S)-3,5-DIMETHYLPIPERIDIN-1-YL]-3-PHENOXYPHENYL}CARBONYL)AMINO]BENZOIC ACID",
            "non-polymer",
        ),
        "B82": (
            "2-({[4-BROMO-3-(DIETHYLSULFAMOYL)PHENYL]CARBONYL}AMINO)BENZOIC ACID",
            "non-polymer",
        ),
        "B85": ("BENZYLPHOSPHONIC ACID", "non-polymer"),
        "B84": (
            "(2R)-3-CYCLOPENTYL-N-(5-METHOXY[1,3]THIAZOLO[5,4-B]PYRIDIN-2-YL)-2-{4-[(4-METHYLPIPERAZIN-1-YL)SULFONYL]PHENYL}PROPANAMIDE",
            "non-polymer",
        ),
        "B87": (
            "4-(1-BENZOTHIOPHEN-2-YL)-6-[4-(2-OXO-2-PYRROLIDIN-1-YLETHYL)PIPERAZIN-1-YL]PYRIMIDINE",
            "non-polymer",
        ),
        "B86": ("2'-DEOXY-5-FLUOROCYTIDINE", "non-polymer"),
        "NA8": ("(1S)-1-CARBOXY-2-NAPHTHALEN-2-YLETHANAMINIUM", "peptide"),
        "9NH": (
            "N-[5-({[2-FLUORO-3-(TRIFLUOROMETHYL)PHENYL]AMINO}CARBONYL)-2-METHYLPHENYL]-4-METHOXY-2-[(4-PIPERAZIN-1-YLPHENYL)AMINO]PYRIMIDINE-5-CARBOXAMIDE",
            "non-polymer",
        ),
        "3MY": ("3-CHLORO-D-TYROSINE", "peptide"),
        "9NE": ("DANSYL-L-GLUTAMATE", "peptide"),
        "9NF": ("DANSYL-L-PHENYLALANINE", "peptide"),
        "3MT": ("3-METHYLTHIAZOLIUM ION", "non-polymer"),
        "3MS": (
            "N-{3-[4-HYDROXY-1-(3-METHYLBUTYL)-2-OXO-1,2-DIHYDROPYRROLO[1,2-B]PYRIDAZIN-3-YL]-1,1-DIOXIDO-2H-1,2,4-BENZOTHIADIAZIN-7-YL}METHANESULFONAMIDE",
            "non-polymer",
        ),
        "3MR": (
            "(4S)-4-(2-AMINO-6-PHENOXYQUINAZOLIN-3(4H)-YL)-N,4-DICYCLOHEXYL-N-METHYLBUTANAMIDE",
            "non-polymer",
        ),
        "3MQ": ("(5-THIOPHEN-2-YLISOXAZOL-3-YL)METHANOL", "non-polymer"),
        "3MP": ("3-METHYLPYRIDINE", "non-polymer"),
        "3MO": ("3-METHYL-1,3-OXAZOLIDIN-2-ONE", "non-polymer"),
        "3MN": (
            "3-({3-[3-(1H-IMIDAZOL-1-YL)PROPYL]-5-METHYL-5-(1-NAPHTHYL)-2,4-DIOXOIMIDAZOLIDIN-1-YL}METHYL)BENZONITRILE",
            "non-polymer",
        ),
        "3MM": (
            "(1R)-1-CARBOXY-N,N,N-TRIMETHYL-3-(METHYLSULFANYL)PROPAN-1-AMINIUM",
            "peptide",
        ),
        "3ML": (
            "[(2S)-3-METHYL-5-OXO-2,5-DIHYDROFURAN-2-YL]ACETIC ACID",
            "non-polymer",
        ),
        "3MI": (
            "2-(3,5-DICHLOROPHENYL)-1,3-BENZOXAZOLE-6-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "LIM": ("3-OXO-PENTADECANOIC ACID", "non-polymer"),
        "3MG": ("3-O-METHYL-BETA-D-GLUCOPYRANOSE", "saccharide"),
        "3MF": ("3-O-METHYLFRUCTOSE IN LINEAR FORM", "saccharide"),
        "3ME": ("?", "nucleic-acid"),
        "3MD": ("2S,3S-3-METHYLASPARTIC ACID", "peptide"),
        "3MC": ("3-METHYLCYTOSINE", "non-polymer"),
        "3MB": ("3-METHOXYBENZAMIDE", "non-polymer"),
        "3MA": ("6-AMINO-3-METHYLPURINE", "non-polymer"),
        "R6C": (
            "(2R)-2-{[6-(BENZYLOXY)-9-ISOPROPYL-9H-PURIN-2-YL]AMINO}BUTAN-1-OL",
            "non-polymer",
        ),
        "XNS": ("METHYL ESTER OCTYL ALPHA-1O-D-MANNOPYRANNOSIDE", "non-polymer"),
        "XNW": (
            "(1R,2R,4S,5R)-1,4,5-TRIHYDROXY-2-(4-METHOXYBENZYL)-3-OXOCYCLOHEXANECARBOXYLIC ACID",
            "non-polymer",
        ),
        "XNV": (
            "ETHYL (4R)-4-({(2S)-2-[3-{[(5-METHYL-1,2-OXAZOL-3-YL)CARBONYL]AMINO}-2-OXOPYRIDIN-1(2H)-YL]PENT-4-YNOYL}AMINO)-5-[(3S)-2-OXOPYRROLIDIN-3-YL]PENTANOATE",
            "non-polymer",
        ),
        "7L0": (
            "N-[6-(5-METHYLSULFONYLPYRIDIN-3-YL)-[1,2,4]TRIAZOLO[1,5-A]PYRIDIN-2-YL]ETHANAMIDE",
            "non-polymer",
        ),
        "1KA": ("(2-HYDROXYETHOXY)ACETALDEHYDE", "non-polymer"),
        "XNZ": (
            "(11R)-10-ACETYL-11-[4-(BENZYLOXY)-2-CHLOROPHENYL]-6-HYDROXY-3,3-DIMETHYL-2,3,4,5,10,11-HEXAHYDRO-1H-DIBENZO[B,E][1,4]DIAZEPIN-1-ONE",
            "non-polymer",
        ),
        "7L4": ("2-PHENYLQUINOLIN-4-OL", "non-polymer"),
        "XNA": (
            "N~8~-(CYCLOPROPYLMETHYL)-N~4~-(2-METHYLPHENYL)-2-(PIPERAZIN-1-YL)PYRIMIDO[5,4-D]PYRIMIDINE-4,8-DIAMINE",
            "non-polymer",
        ),
        "XNC": (
            "(11S)-11-[4-(BENZYLOXY)-2-FLUOROPHENYL]-3,3-DIMETHYL-10-[(6-METHYLPYRIDIN-2-YL)CARBONYL]-2,3,4,5,10,11-HEXAHYDROTHIOPYRANO[3,2-B][1,5]BENZODIAZEPIN-6-OL 1,1-DIOXIDE",
            "non-polymer",
        ),
        "1KX": (
            "{(1S,4R)-4-[2-AMINO-6-(CYCLOPROPYLAMINO)-9H-PURIN-9-YL]CYCLOPENT-2-EN-1-YL}METHANOL",
            "non-polymer",
        ),
        "XND": (
            "(11R)-10-ACETYL-11-(2,4-DICHLOROPHENYL)-6-HYDROXY-3,3-DIMETHYL-2,3,4,5,10,11-HEXAHYDRO-1H-DIBENZO[B,E][1,4]DIAZEPIN-1-ONE",
            "non-polymer",
        ),
        "XNI": (
            "(11S)-10-ACETYL-11-[4-(BENZYLOXY)-3-CHLOROPHENYL]-3,3-DIMETHYL-2,3,4,5,10,11-HEXAHYDRO-1H-DIBENZO[B,E][1,4]DIAZEPIN-1-ONE",
            "non-polymer",
        ),
        "R6G": ("RHODAMINE 6G", "non-polymer"),
        "1KP": (
            "(3E)-6,7-DICHLORO-3-(HYDROXYIMINO)-1,3-DIHYDRO-2H-INDOL-2-ONE",
            "non-polymer",
        ),
        "XNM": (
            "(4-{[4-(4-CHLOROPHENYL)PYRIMIDIN-2-YL]AMINO}PHENYL)[4-(2-HYDROXYETHYL)PIPERAZIN-1-YL]METHANONE",
            "non-polymer",
        ),
        "XNN": (
            "N~8~-(CYCLOPROPYLMETHYL)-2-(2,6-DIAZASPIRO[3.3]HEPT-2-YL)-N~4~-[2-(METHYLSULFANYL)PHENYL]PYRIMIDO[5,4-D]PYRIMIDINE-4,8-DIAMINE",
            "non-polymer",
        ),
        "NCN": ("NICOTINATE MONONUCLEOTIDE", "non-polymer"),
        "RX0": (
            "[(1S,3R,4S)-3-AMINO-4-HYDROXYCYCLOPENTYL]{(3R)-3-[(1S)-1-(BIPHENYL-2-YL)-1-HYDROXY-5-METHOXYPENTYL]PIPERIDIN-1-YL}METHANONE",
            "non-polymer",
        ),
        "LKG": (
            "4-[(6-PHENYL[1,2,4]TRIAZOLO[4,3-B]PYRIDAZIN-3-YL)METHYL]PHENOL",
            "non-polymer",
        ),
        "FLR": ("(2R)-2-(3-FLUORO-4-PHENYL-PHENYL)PROPANOIC ACID", "non-polymer"),
        "G49": ("N2-METHYL-2'-DEOXY-GUANOSINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "YX0": (
            "[(3E)-3-(1-HYDROXYETHYLIDENE)-2,3-DIHYDROISOXAZOL-5-YL]METHYL 6-DEOXY-ALPHA-L-GALACTOPYRANOSIDE",
            "non-polymer",
        ),
        "YX1": (
            "2-DEOXY-2-{[(2-HYDROXY-1-METHYLHYDRAZINO)CARBONYL]AMINO}-BETA-D-GLUCOPYRANOSE",
            "non-polymer",
        ),
        "LKC": (
            "4-AMINO-1-[(1S,3R,4R,7S)-7-HYDROXY-1-(HYDROXYMETHYL)-2,5-DIOXABICYCLO[2.2.1]HEPT-3-YL]-5-METHYLPYRIMIDIN-2(1H)-ONE",
            "nucleic-acid",
        ),
        "LKA": ("5-THIO-A/B-D-MANNOPYRANOSYLAMINE", "non-polymer"),
        "XN1": (
            "N-[2-HYDROXY-1-INDANYL]-5-[(2-TERTIARYBUTYLAMINOCARBONYL)-4(3-PYRIDYLMETHYL)PIPERAZINO]-4-HYDROXY-2-(1-PHENYLETHYL)-PENTANAMIDE",
            "non-polymer",
        ),
        "XN3": (
            "N-[2(R)-HYDROXY-1(S)-INDANYL]-5-[(2(S)-TERTIARY BUTYLAMINOCARBONYL)-4(BENZO[1,3]DIOXOL-5-YLMETHYL)-PIPERAZINO]-4(S)-HYDROXY-2(R)-PHENYLMETHYLPENTANAMIDE",
            "non-polymer",
        ),
        "XN2": (
            "N-[2-HYDROXY-1-INDANYL]-5-[(2-TERTIARYBUTYLAMINOCARBONYL)-4(BENZO[1,3]DIOXOL-5-YLMETHYL)-PIPERAZINO]-4-HYDROXY-2-(1-PHENYLETHYL)-PENTANAMIDE",
            "non-polymer",
        ),
        "S99": (
            "7-PHENYL-1-(4-PYRIDIN-2-YL-1,3-OXAZOL-2-YL)HEPTANE-1,1-DIOL",
            "non-polymer",
        ),
        "7LI": (
            "(5Z)-5-{2-[(3R)-3-AMINOPIPERIDIN-1-YL]-3-(PROPAN-2-YLOXY)BENZYLIDENE}-1,3-THIAZOLIDINE-2,4-DIONE",
            "non-polymer",
        ),
        "MGF": ("TRIFLUOROMAGNESATE", "non-polymer"),
        "RA4": (
            "N-(4-NITROBENZOYL)-L-LEUCYL-N-(4-{[AMINO(IMINO)METHYL]AMINO}BUTYL)-L-PROLINAMIDE",
            "peptide",
        ),
        "JSZ": (
            "ETHYL 4-({(2S)-2-HYDROXY-3-[(1-METHYLETHYL)AMINO]PROPYL}OXY)-3-METHYL-1-BENZOFURAN-2-CARBOXYLATE",
            "non-polymer",
        ),
        "QX": ("QUINOXALINE-2-CARBOXYLIC ACID", "non-polymer"),
        "QSI": ("5'-O-[N-(L-GLUTAMINYL)-SULFAMOYL]ADENOSINE", "non-polymer"),
        "NCA": ("NICOTINAMIDE", "non-polymer"),
        "VHW": ("?", "non-polymer"),
        "LIN": ("3-OXO-BUTYRIC ACID", "non-polymer"),
        "0JN": (
            "1-TERT-BUTYL-3-(NAPHTHALEN-1-YLMETHYL)-1H-PYRAZOLO[3,4-D]PYRIMIDIN-4-AMINE",
            "non-polymer",
        ),
        "VHI": (
            "5-{(3AS,4S,6AR)-1-[(BENZYLOXY)CARBONYL]-2-OXOHEXAHYDRO-1H-THIENO[3,4-D]IMIDAZOL-4-YL}PENTANOIC ACID",
            "non-polymer",
        ),
        "238": ("?", "non-polymer"),
        "VHE": (
            "4-AMINO-18,20-DIMETHYL-7-THIA-3,5,11,15-TETRAAZATRICYCLO[15.3.1.1(2,6)]DOCOSA-1(20),2,4,6(22),17(21),18-HEXAENE-10,16-DIONE",
            "non-polymer",
        ),
        "VHD": (
            "4-(1,3-DIHYDRO-2H-ISOINDOL-2-YLCARBONYL)-6-(1-METHYLETHYL)BENZENE-1,3-DIOL",
            "non-polymer",
        ),
        "VHA": (
            "22-METHYL-13,18-DIOXA-7-THIA-3,5-DIAZATETRACYCLO[17.3.1.1~2,6~.1~8,12~]PENTACOSA-1(23),2(25),3,5,8(24),9,11,19,21-NONAEN-4-AMINE",
            "non-polymer",
        ),
        "FLC": ("CITRATE ANION", "non-polymer"),
        "VHC": (
            "4-AMINO-20,22-DIMETHYL-13-OXA-7-THIA-3,5,17-TRIAZATETRACYCLO[17.3.1.1~2,6~.1~8,12~]PENTACOSA-1(23),2(25),3,5,8(24),9,11,19,21-NONAEN-18-ONE",
            "non-polymer",
        ),
        "0JH": (
            "TERT-BUTYL (2S)-3-[4-AMINO-7-(3-HYDROXYPROPYL)-5-(4-METHYLPHENYL)-7H-PYRROLO[2,3-D]PYRIMIDIN-6-YL]-2-CYANOPROPANOATE",
            "non-polymer",
        ),
        "NCB": ("N-CARBAMOYL-ALANINE", "peptide"),
        "TXL": ("TAXOTERE", "non-polymer"),
        "TXI": (
            "(2R)-N-(4-CHLORO-3-OXOBUTYL)-2,4-DIHYDROXY-3,3-DIMETHYLBUTANAMIDE",
            "non-polymer",
        ),
        "TXF": (
            "(Z)-4-(1-{4-[2-(DIMETHYLAMINO)ETHOXY]PHENYL}-5-HYDROXY-2-PHENYLPENT-1-ENYL)PHENOL",
            "non-polymer",
        ),
        "TXD": ("1,4,5,6-TETRAHYDRONICOTINAMIDE ADENINE DINUCLEOTIDE", "non-polymer"),
        "TXC": (
            "2-[(2-AMINO-6-OXO-1,6-DIHYDRO-9H-PURIN-9-YL)METHOXY]ETHYL L-VALINATE",
            "non-polymer",
        ),
        "TXA": ("THREONYLCARBAMOYLADENYLATE", "non-polymer"),
        "LKS": (
            "N-(1-IMINIO-2-PHENYLETHYL)-5-THIOHEXOPYRANOSYLAMINE BROMIDE",
            "non-polymer",
        ),
        "TXZ": ("O-PHOSPHONO-N-(6-SULFANYLHEXANOYL)-L-THREONINE", "non-polymer"),
        "ZY3": (
            "N-{(1S,2R)-1-BENZYL-2-HYDROXY-3-[(3-METHOXYBENZYL)AMINO]PROPYL}-6-(ETHYLAMINO)-1-METHYL-1,3,4,5-TETRAHYDRO-2,1-BENZOTHIAZEPINE-8-CARBOXAMIDE 2,2-DIOXIDE",
            "non-polymer",
        ),
        "TXT": (
            "4,6-DIDEOXY-4-{[(1S,2R,3R,4S,5S)-2,3,4-TRIHYDROXY-5-(HYDROXYMETHYL)CYCLOHEXYL]AMINO}-ALPHA-D-ALLOPYRANOSYL-(1->4)-ALPHA-D-GLUCOPYRANOSYL-(1->4)-ALPHA-D-GLUCOPYRANOSYL-(1->4)-ALPHA-D-GLUCOPYRANOSE",
            "non-polymer",
        ),
        "TXS": (
            "1-(2,5-DIDEOXY-5-PYRROLIDIN-1-YL-BETA-L-ERYTHRO-PENTOFURANOSYL)-5-METHYLPYRIMIDINE-2,4(1H,3H)-DIONE",
            "non-polymer",
        ),
        "TXQ": ("1,2,5,8-TETRAHYDROXYANTHRACENE-9,10-DIONE", "non-polymer"),
        "OER": ("SR-MN4-O5 CLUSTER", "non-polymer"),
        "0JA": (
            "2-CHLORO-3-(1-CYANOCYCLOPROPYL)-N-[5-({2-[(CYCLOPROPYLCARBONYL)AMINO][1,3]THIAZOLO[5,4-B]PYRIDIN-5-YL}OXY)-2-FLUOROPHENYL]BENZAMIDE",
            "non-polymer",
        ),
        "TX5": (
            "(1R,3R)-5-((Z)-2-((1R,7AS)-HEXAHYDRO-1-((S)-6-HYDROXY-6-METHYLHEPT-4-YN-2-YL)-7A-METHYL-1H-INDEN-4(7AH)-YLIDENE)ETHYLIDENE)CYCLOHEXANE-1,3-DIOL",
            "non-polymer",
        ),
        "TX4": ("L-THREONOHYDROXAMATE 4-PHOSPHATE", "non-polymer"),
        "TX3": (
            "2,4-DIBROMO-6-({[(2-CHLOROPHENYL)CARBONYL]AMINO}METHYL)PHENYL 2-METHYLBENZOATE",
            "non-polymer",
        ),
        "TX2": (
            "5'-O-{(R)-HYDROXY[(2-SULFANYLETHYL)AMINO]PHOSPHORYL}THYMIDINE",
            "nucleic-acid",
        ),
        "TX1": (
            "2,4-DIBROMO-6-({[(2-NITROPHENYL)CARBONYL]AMINO}METHYL)PHENYL 2-CHLOROBENZOATE",
            "non-polymer",
        ),
        "X7A": (
            "[(5R)-5-AMINO-5-CARBOXY-7-(PIPERIDIN-1-YL)HEPTYL](TRIHYDROXY)BORATE(1-)",
            "non-polymer",
        ),
        "VH1": (
            "2-(4-AMINOBENZYL)-1,3-DIOXO-2,3-DIHYDRO-1H-BENZO[DE]ISOQUINOLINE-5-SULFONATE",
            "non-polymer",
        ),
        "JS6": ("?", "non-polymer"),
        "VH2": (
            "N-[5-[(4-BROMOPHENYL)METHYL]-4-HYDROXY-1,3-THIAZOL-2-YL]NAPHTHALENE-1-SULFONAMIDE",
            "non-polymer",
        ),
        "SBN": ("2-AMINO-6-(3,5-DIMETHYLPHENYL)SULFONYLBENZONITRILE", "non-polymer"),
        "N5F": (
            "(2S)-2-[({3-HYDROXY-2-METHYL-5-[(PHOSPHONOOXY)METHYL]PYRIDIN-4-YL}METHYL)AMINO]HEXANEDIOIC ACID",
            "non-polymer",
        ),
        "NCE": ("3,6-BIS[(3-MORPHOLINOPROPIONAMIDO)] ACRIDINE", "non-polymer"),
        "PAP": ("3'-PHOSPHATE-ADENOSINE-5'-DIPHOSPHATE", "non-polymer"),
        "PAR": ("PAROMOMYCIN", "non-polymer"),
        "PAS": ("2-AMINO-4-OXO-4-PHOSPHONOOXY-BUTYRIC ACID", "peptide"),
        "V3L": ("?", "nucleic-acid"),
        "PAU": ("PANTOTHENOIC ACID", "peptide"),
        "SBB": ("N-[2-(1H-INDOL-5-YL)-BUTYL]-4-SULFAMOYL-BENZAMIDE", "non-polymer"),
        "Q50": (
            "{(1S)-1-BENZYL-4-[3-CARBAMOYL-1-(1-CARBAMOYL-2-PHENYL-ETHYLCARBAMOYL)-(S)-PROPYLCARBAMOYL]-2-OXO-5-PHENYL-PENTYL}-CARBAMIC ACID TERT-BUTYL ESTER",
            "non-polymer",
        ),
        "E6C": (
            "N-[1-HYDROXYCARBOXYETHYL-CARBONYL]LEUCYLAMINO-2-METHYL-BUTANE",
            "non-polymer",
        ),
        "IQ1": (
            "3-(2-TERT-BUTYL-5-(PYRIDIN-4-YL)-1H-IMIDAZOL-4-YL)QUINOLINE",
            "non-polymer",
        ),
        "IH2": (
            "2-(2,2-DIPHENYL-ETHYL)-7-METHYL-1,3-DIOXO-2,3,5,8-TETRAHYDRO-1H-[1,2,4]TRIAZOLO [1,2-A]PYRIDAZINE-5-CARBOXYLIC ACID(4-CARBAMIMIDOYL-CYCLOHEXYLMETHYL)-AMIDE",
            "non-polymer",
        ),
        "HZ1": ("?", "non-polymer"),
        "HZ3": (
            "DIMETHYL (1R,4S)-5,6-BIS(4-HYDROXYPHENYL)-7-OXABICYCLO[2.2.1]HEPTA-2,5-DIENE-2,3-DICARBOXYLATE",
            "non-polymer",
        ),
        "OEV": (
            "(1S,2S,3R,6R)-4-(HYDROXYMETHYL)-6-(OCTYLAMINO)CYCLOHEX-4-ENE-1,2,3-TRIOL",
            "non-polymer",
        ),
        "NR3": (
            "(1S,2R,3S,4S)-1-{(1S)-2-[(2R,3S,4S)-3,4-DIHYDROXY-2-(HYDROXYMETHYL)TETRAHYDROTHIOPHENIUM-1-YL]-1-HYDROXYETHYL}-2,3,4,5-TETRAHYDROXYPENTYL SULFATE",
            "non-polymer",
        ),
        "NTY": ("BETA NOR-TYROSINE", "peptide"),
        "WRB": (
            "1-[3-(4-BROMO-PHENOXY)-PROPOXY]-6,6-DIMETHYL-1.6-DIHYDRO-[1,3,5]TRIAZINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "T3O": (
            "4-[(1S,2R,5S)-4,4,8-TRIMETHYL-3-OXABICYCLO[3.3.1]NON-7-EN-2-YL]PHENOL",
            "non-polymer",
        ),
        "IH6": (
            "(3S)-4-{[4-(BUT-2-YNYLOXY)PHENYL]SULFONYL}-N-HYDROXY-2,2-DIMETHYLTHIOMORPHOLINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "ALL": ("D-ALLOPYRANOSE", "saccharide"),
        "NTZ": ("NOJIRIMYCINE TETRAZOLE", "non-polymer"),
        "WRR": (
            "4-HYDROXY-3-[(1S,3R)-3-HYDROXY-1-PHENYLBUTYL]-2H-CHROMEN-2-ONE",
            "non-polymer",
        ),
        "HZP": ("(4S)-4-HYDROXY-L-PROLINE", "peptide"),
        "BZ3": ("5-METHOXY-1H-INDOLE", "non-polymer"),
        "E63": (
            "6-({4-[(3-CYCLOPROPYL-1H-PYRAZOL-5-YL)AMINO]-5-(PHENYLAMINO)PYRIMIDIN-2-YL}AMINO)-1,2-DIHYDRO-3H-INDAZOL-3-ONE",
            "non-polymer",
        ),
        "E64": (
            "N-[N-[1-HYDROXYCARBOXYETHYL-CARBONYL]LEUCYLAMINO-BUTYL]-GUANIDINE",
            "non-polymer",
        ),
        "E65": (
            "4-{[(4,6-DIMETHYLPYRIMIDIN-2-YL)SULFANYL]ACETYL}BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "ALK": (
            "(1R)-1-[DIHYDROXY(METHOXY)-LAMBDA^5^-PHOSPHANYL]ETHANOL",
            "non-polymer",
        ),
        "TAF": ("?", "nucleic-acid"),
        "ALH": ("6-PHENYL[5H]PYRROLO[2,3-B]PYRAZINE", "non-polymer"),
        "N5B": ("N-(5-CYCLOPROPYL-1H-PYRAZOL-3-YL)BENZAMIDE", "non-polymer"),
        "I63": (
            "{3-[(1R)-3-(3,4-DIMETHOXYPHENYL)-1-({[(2S)-1-(3,3-DIMETHYL-2-OXOPENTANOYL)PIPERIDIN-2-YL]CARBONYL}OXY)PROPYL]PHENOXY}ACETIC ACID",
            "non-polymer",
        ),
        "T9T": ("TRIPHENYLSTANNANYL", "non-polymer"),
        "PAB": ("4-AMINOBENZOIC ACID", "non-polymer"),
        "T3F": (
            "(3R,4S,5R,6R)-4-AMINO-3,5-DIHYDROXY-6-METHYLOXAN-2-YL][HYDROXY-[[(2R,3S,5R)-3-HYDROXY-5-(5-METHYL-2,4-DIOXOPYRIMIDIN-1-YL)OXOLAN-2-YL]METHOXY]PHOSPHORYL] HYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "PAC": ("2-PHENYLACETIC ACID", "non-polymer"),
        "UA3": (
            "1-(3-O-PHOSPHONO-BETA-L-ARABINOFURANOSYL)PYRIMIDINE-2,4(1H,3H)-DIONE",
            "non-polymer",
        ),
        "H70": ("2,4-DICHLORO-6-(3,4,5-TRIBROMO-1H-PYRROL-2-YL)PHENOL", "non-polymer"),
        "NTP": ("HEPARIN PENTASACCHARIDE", "saccharide"),
        "OTS": ("4-(2S-AMINO-1-HYDROXYETHYL)PHENOL", "non-polymer"),
        "M90": (
            "(2S)-1-{[(2R)-1-{[(2S,3R)-1-CYCLOHEXYL-3-HYDROXY-4-(PYRIDIN-4-YLOXY)BUTAN-2-YL]AMINO}-3-(METHYLSULFANYL)-1-OXOPROPAN-2-YL]AMINO}-1-OXO-3-PHENYLPROPAN-2-YL 4-AMINOPIPERIDINE-1-CARBOXYLATE",
            "peptide",
        ),
        "NTS": ("NAPHTHALENE TRISULFONATE", "non-polymer"),
        "PAE": ("PHOSPHONOACETIC ACID", "non-polymer"),
        "NTR": ("META-HYDROXY BETA NOR-TYROSINE", "peptide"),
        "PAF": ("PANTOATE", "non-polymer"),
        "NTU": ("N1,N14-BIS((S-METHYL)ISOTHIOUREIDO)TETRADECANE", "non-polymer"),
        "F24": (
            "(7R,12R,13R)-13-FORMYL-12,14-DIHYDROXY-3,5,7-TRIMETHYLTETRADECA-2,4-DIENOIC ACID",
            "non-polymer",
        ),
        "UOC": ("UNDECAN-2-ONE", "non-polymer"),
        "PAH": ("PHOSPHONOACETOHYDROXAMIC ACID", "non-polymer"),
        "L0H": ("N-[3-(1H-BENZIMIDAZOL-2-YL)-1H-PYRAZOL-4-YL]BENZAMIDE", "non-polymer"),
        "ZPN": (
            "(2Z,4E)-N-[(S)-[(1S,2E,5S,8E,10Z,17S)-3,11-DIMETHYL-19-METHYLIDENE-7,13-DIOXO-6,21-DIOXABICYCLO[15.3.1]HENICOSA-2,8,10-TRIEN-5-YL](HYDROXY)METHYL]HEXA-2,4-DIENAMIDE",
            "non-polymer",
        ),
        "PAI": (
            "{[(2,2-DIHYDROXY-ETHYL)-(2,3,4,5-TETRAHYDROXY-6-PHOSPHONOOXY-HEXYL)-AMINO]-METHYL}-PHOSPHONIC ACID",
            "non-polymer",
        ),
        "BZ8": (
            "4-CHLORO-6-[5-(4-ETHOXYPHENYL)-1,2,3-THIADIAZOL-4-YL BENZENE-1,3-DIOL",
            "non-polymer",
        ),
        "MZR": (
            "5-HYDROXY-1-(BETA-D-RIBOFURANOSYL)-1H-IMIDAZOLE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "FT1": ("4-(2,4-DICHLOROPHENOXY)-3-HYDROXYBENZALDEHYDE", "non-polymer"),
        "FT0": ("3-HYDROXY-4-PHENOXYBENZALDEHYDE", "non-polymer"),
        "FT3": ("2-(2,4-DICHLOROPHENOXY)-5-(HYDROXYMETHYL)PHENOL", "non-polymer"),
        "FT2": ("5-(CHLOROMETHYL)-2-(2,4-DICHLOROPHENOXY)PHENOL", "non-polymer"),
        "C4P": (
            "N-[(1R)-2-(BENZYLSULFANYL)-1-FORMYLETHYL]-N-(MORPHOLIN-4-YLCARBONYL)-L-PHENYLALANINAMIDE",
            "non-polymer",
        ),
        "C4Q": ("4-(1,4,7,10-TETRAAZACYCLODODECAN-1-YLMETHYL)QUINOLINE", "non-polymer"),
        "C4R": (
            "(R)-2-AMINO-3-(BICYCLO[2.2.1]HEPTA-2,5-DIEN-7-YLTHIO)PROPANOIC ACID RHODIUM",
            "peptide",
        ),
        "C4S": ("?", "nucleic-acid"),
        "44P": ("5'-O-[(S)-HYDROXY(PHOSPHONOMETHYL)PHOSPHORYL]URIDINE", "non-polymer"),
        "C4X": ("3,4-EPOXYBUTYL-BETA-D-XYLOSIDE", "saccharide"),
        "44U": ("BETA-PHENYL-D-PHENYLALANYL-N-PROPYL-L-PROLINAMIDE", "non-polymer"),
        "C4E": ("N-PHENYL-1H-PYRROLO[2,3-B]PYRIDIN-3-AMINE", "non-polymer"),
        "C4F": ("3-(3-METHOXYBENZYL)-1H-PYRROLO[2,3-B]PYRIDINE", "non-polymer"),
        "C4A": (
            "ETHYL {(1S)-5-AMINO-1-[(5-{4-[(2,3-DIHYDRO-1H-INDEN-2-YLAMINO)CARBONYL]BENZYL}-1,2,4-OXADIAZOL-3-YL)CARBONYL]PENTYL}CARBAMATE",
            "non-polymer",
        ),
        "C4B": ("1-BETA-D-GLUCOPYRANOSYL-1,3,5-TRIAZINANE-2,4,6-TRIONE", "non-polymer"),
        "C4C": (
            "N-(4-CHLOROBENZYL)-N',N'-DIMETHYL-N-PYRIDIN-2-YLETHANE-1,2-DIAMINE",
            "non-polymer",
        ),
        "C4L": ("N-BUTANOYL-L-HOMOSERINE", "non-polymer"),
        "C4M": (
            "N-[(2R,3S)-3-AMINO-2-HYDROXY-4-PHENYLBUTYL]-4-METHOXY-2,3,6-TRIMETHYLBENZENESULFONAMIDE",
            "non-polymer",
        ),
        "44A": ("9-BETA-D-RIBOFURANURONOSYL-9H-PURIN-6-AMINE", "non-polymer"),
        "IUR": ("5-IODOURACIL", "non-polymer"),
        "C4H": (
            "N-[4-({[5-(DIMETHYLAMINO)-1-NAPHTHYL]SULFONYL}AMINO)BUTYL]-3-SULFANYLPROPANAMIDE",
            "non-polymer",
        ),
        "44F": (
            "1-{3-[(3-CHLORO-5-METHOXYBENZYL)AMINO]PROPYL}-3-THIOPHEN-3-YLUREA",
            "non-polymer",
        ),
        "44D": (
            "7-[5-(4-AMINO-5-HYDROXY-6-METHYL-TETRAHYDRO-PYRAN-2-YLOXY)-4-HYDROXY-6-METHYL-TETRAHYDRO-PYRAN-2-YLOXY]-6,9,11-TRIHYDROXY-9-(2-HYDROXY-ACETYL)-7,8,9,10-TETRAHYDRO-NAPHTHACENE-5,12-DIONE",
            "non-polymer",
        ),
        "C44": (
            "N-{(1S,2R)-1-BENZYL-2-HYDROXY-3-[(1S)-1,2,3,4-TETRAHYDRONAPHTHALEN-1-YLAMINO]PROPYL}-3-(1,1-DIOXIDO-1,2-THIAZINAN-2-YL)-5-(ETHYLAMINO)BENZAMIDE",
            "non-polymer",
        ),
        "C45": ("N4-METHOXY-2'-DEOXY-CYTIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "C46": (
            "6H,8H-3,4-DIHYDROPYRIMIDO[4,5-C][1,2]OXAZIN-7-0NE(CYTIDINE)-5'-MONOPHOSPHATE",
            "nucleic-acid",
        ),
        "448": (
            "(2R)-4-[(8R)-8-METHYL-2-(TRIFLUOROMETHYL)-5,6-DIHYDRO[1,2,4]TRIAZOLO[1,5-A]PYRAZIN-7(8H)-YL]-4-OXO-1-(2,4,5-TRIFLUOROPHENYL)BUTAN-2-AMINE",
            "non-polymer",
        ),
        "C40": (
            "N-[(2R,4S,5S,7R)-4-AMINO-8-(BUTYLAMINO)-5-HYDROXY-2,7-DIMETHYL-8-OXOOCTYL]-2-(3-METHOXYPROPOXY)BENZAMIDE",
            "non-polymer",
        ),
        "C41": ("ALISKIREN", "non-polymer"),
        "C42": ("3'-AMINO-2'-DEOXY-CYTIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "C43": ("2'-O-METHYOXYETHYL-CYTIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "443": (
            "1-(1-{(2S)-3-[(6-CHLORONAPHTHALEN-2-YL)SULFONYL]-2-HYDROXYPROPANOYL}PIPERIDIN-4-YL)TETRAHYDROPYRIMIDIN-2(1H)-ONE",
            "non-polymer",
        ),
        "442": (
            "2-[3,5-DIBROMO-4-(4-HYDROXY-3-{HYDROXY[(2-PHENYLETHYL)AMINO]METHYL}PHENOXY)PHENYL]ETHANE-1,1-DIOL",
            "non-polymer",
        ),
        "440": (
            "3-{[4-(BUT-2-YN-1-YLOXY)PHENYL]SULFONYL}PROPANE-1-THIOL",
            "non-polymer",
        ),
        "C48": (
            "N1-(1-DIMETHYLCARBAMOYL-2-PHENYL-ETHYL)-2-OXO-N4-(2-PYRIDIN-2-YL-ETHYL)-SUCCINAMIDE",
            "non-polymer",
        ),
        "446": (
            "4-[(4-{[6-BROMO-3-(METHOXYCARBONYL)-1-OXO-4-PHENYLISOQUINOLIN-2(1H)-YL]METHYL}PHENYL)AMINO]-4-OXOBUTANOIC ACID",
            "non-polymer",
        ),
        "445": ("N-[OXO(PHENYLAMINO)ACETYL]-BETA-D-GLUCOPYRANOSYLAMINE", "non-polymer"),
        "444": (
            "N-(2,2,2-TRIFLUOROETHYL)-N-{4-[2,2,2-TRIFLUORO-1-HYDROXY-1-(TRIFLUOROMETHYL)ETHYL]PHENYL}BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "FTA": (
            "3-(4-PHENYLAMINO-PHENYLAMINO)-2-(1H-TETRAZOL-5-YL)-ACRYLONITRILE",
            "non-polymer",
        ),
        "NTA": ("NITRILOTRIACETIC ACID", "non-polymer"),
        "FTC": (
            "N-[[3-FLUORO-4-ETHOXY-PYRID-2-YL]ETHYL]-N'-[5-CHLORO-PYRIDYL]-THIOUREA",
            "non-polymer",
        ),
        "FTB": (
            "N-{4-[1-(2-FLUOROBENZYL)-3-BUTYL-2,6-DIOXO-2,3,6,7-TETRAHYDRO-1H-PURIN-8-YLMETHYL]-PHENYL}-ACETAMIDE",
            "non-polymer",
        ),
        "FTE": ("PHENYLISOTHIOCYANATE", "non-polymer"),
        "Z63": (
            "[4-AMINO-2-(PROP-2-EN-1-YLAMINO)-1,3-THIAZOL-5-YL](PYRIDIN-4-YL)METHANONE",
            "non-polymer",
        ),
        "YR1": (
            "2-HYDROXY-5-[(5S)-3-HYDROXY-5-(4-NITROPHENYL)-2-OXO-4-(PHENYLCARBONYL)-2,5-DIHYDRO-1H-PYRROL-1-YL]BENZOIC ACID",
            "non-polymer",
        ),
        "FTI": (
            "1-{2-[3-(4-CYANO-BENZYL)-3H-IMIDAZOL-4-YL]-ACETYL}-5-NAPHTHALEN-1-YL-1,2,3,6-TETRAHYDRO-PYRIDINE-4-CARBONITRILE",
            "non-polymer",
        ),
        "FTH": (
            "1-[2-(4-CYANO-BENZYLAMINO)-3-(3-METHYL-3H-IMIDAZOL-4-YL)-PROPIONYL]-5-NAPHTHALEN-1-YL-1,2,3,6-TETRAHYDRO-PYRIDINE-4-CARBONITRILE",
            "non-polymer",
        ),
        "IU5": ("ISO-URSODEOXYCHOLIC ACID", "non-polymer"),
        "FTL": (
            "4-[(3-CYANO-BENZYL)-(3-METHYL-3H-IMIDAZOL-4-YLMETHYL)-AMINO]-2-NAPHTHALEN-1-YL-BENZONITRILE",
            "non-polymer",
        ),
        "DCK": (
            "(2S)-3-CARBOXY-2-HYDROXY-N,N,N-TRIMETHYLPROPAN-1-AMINIUM",
            "non-polymer",
        ),
        "NCY": ("N-METHYLCYSTEINE", "peptide"),
        "NTB": ("TERTIARY-BUTYLAMINE", "non-polymer"),
        "NTE": (
            "[3,3'-{7-ETHENYL-3,8,13,17-TETRAMETHYL-12-[(E)-2-NITROETHENYL]PORPHYRIN-2,18-DIYL-KAPPA~4~N~21~,N~22~,N~23~,N~24~}DIPROPANOATO(2-)]IRON",
            "non-polymer",
        ),
        "NTD": (
            "2-{HYDROXY[2-NITRO-4-(TRIFLUOROMETHYL)PHENYL]METHYLENE}CYCLOHEXANE-1,3-DIONE",
            "non-polymer",
        ),
        "Z62": (
            "[4-AMINO-2-(PROP-2-EN-1-YLAMINO)-1,3-THIAZOL-5-YL](3-NITROPHENYL)METHANONE",
            "non-polymer",
        ),
        "RX5": (
            "METHYL (2-{(R)-(3-CHLOROPHENYL)[(3R)-1-({(2S)-2-(METHYLAMINO)-3-[(3R)-TETRAHYDRO-2H-PYRAN-3-YL]PROPYL}CARBAMOYL)PIPERIDIN-3-YL]METHOXY}ETHYL)CARBAMATE",
            "non-polymer",
        ),
        "M4S": ("4-METHOXYBENZENESULFONAMIDE", "non-polymer"),
        "NTF": ("N-TRIFLURO-ACETYL-BETA-D-GLUCOPYRANOSYLAMINE", "saccharide"),
        "SMZ": ("N~5~-[(E)-IMINO(METHYLSULFANYL)METHYL]-L-ORNITHINE", "non-polymer"),
        "SMY": (
            "(2R)-3-(1H-INDOL-3-YL)-1-{4-[(5S)-5-METHYL-5,7-DIHYDROTHIENO[3,4-D]PYRIMIDIN-4-YL]PIPERAZIN-1-YL}-1-OXOPROPAN-2-AMINE",
            "non-polymer",
        ),
        "SMX": (
            "(6R,7R)-7-{[(2R)-2-HYDROXY-2-PHENYLACETYL]AMINO}-3-{[(1-METHYL-1H-TETRAZOL-5-YL)SULFANYL]METHYL}-8-OXO-5-THIA-1-AZABICYCLO[4.2.0]OCT-2-ENE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "SMT": ("2'-[(METHYLTHIO)ETHYLOXY]-THYMIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "SMS": (
            "SULFAMIC ACID 2,3-O-(1-METHYLETHYLIDENE)-4,5-O-SULFONYL-BETA-FRUCTOPYRANOSE ESTER",
            "non-polymer",
        ),
        "SMR": (
            "N-{(2S)-3-[(3S)-8',9'-DIHYDRO-1H,3'H-SPIRO[PIPERIDINE-3,7'-PYRANO[3,2-E]INDAZOL]-1-YL]-2-HYDROXYPROPYL}-N-(2-ETHOXYETHYL)-2,6-DIMETHYLBENZENESULFONAMIDE",
            "non-polymer",
        ),
        "SMQ": ("SIMOCYCLINONE C4", "non-polymer"),
        "SMP": ("2'-DEOXY-ADENOSINE-5'-SP-MONOMETHYLPHOSPHONATE", "nucleic-acid"),
        "SMO": ("DIOXOSULFIDOMOLYBDENUM(VI) ION", "non-polymer"),
        "FA": ("FOLIC ACID", "non-polymer"),
        "FB": ("5A-FORMYLBICYCLOMYCIN", "non-polymer"),
        "SML": ("PHENYL METHYL SULFONE", "non-polymer"),
        "REW": (
            "(2S,3S)-1-(4-METHOXYPHENYL)-3-(3-(2-(5-METHYL-2-PHENYLOXAZOL-4-YL)ETHOXY)BENZYL)-4-OXOAZETIDINE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "FE": ("FE (III) ION", "non-polymer"),
        "SMI": ("SPECTINOMYCIN", "non-polymer"),
        "SMH": (
            "N-(2-ETHOXYETHYL)-N-{(2S)-2-HYDROXY-3-[(2R)-6-HYDROXY-4-OXO-3,4-DIHYDRO-1'H-SPIRO[CHROMENE-2,3'-PIPERIDIN]-1'-YL]PROPYL}-2,6-DIMETHYLBENZENESULFONAMIDE",
            "non-polymer",
        ),
        "SMG": ("N-SUCCINYL METHIONINE", "non-polymer"),
        "SMF": ("4-SULFOMETHYL-L-PHENYLALANINE", "peptide"),
        "SME": ("METHIONINE SULFOXIDE", "peptide"),
        "SMD": (
            "METHYL-2-S-(ALPHA-D-MANNOPYRANOSYL)-2-THIO-ALPHA-D-MANNOPYRANOSIDE",
            "non-polymer",
        ),
        "SMC": ("S-METHYLCYSTEINE", "peptide"),
        "SMB": ("2-METHYLBUTANOIC ACID", "non-polymer"),
        "SMA": ("STIGMATELLIN A", "non-polymer"),
        "JE2": (
            "(4R)-3-{(2S,3S)-2-HYDROXY-3-[(3-HYDROXY-2-METHYLBENZOYL)AMINO]-4-PHENYLBUTANOYL}-5,5-DIMETHYL-N-(2-METHYLBENZYL)-1,3-THIAZOLIDINE-4-CARBOXAMIDE",
            "peptide",
        ),
        "GGT": ("TRIOXACARCIN A", "non-polymer"),
        "GGW": (
            "METHYL [(1S,4S,5S,7S,10S)-4-BENZYL-1,10-DI-TERT-BUTYL-5-HYDROXY-2,9,12-TRIOXO-7-(4-PYRIDIN-2-YLBENZYL)-13-OXA-3,8,11-TRIAZATETRADEC-1-YL]CARBAMATE",
            "non-polymer",
        ),
        "GGV": ("?", "non-polymer"),
        "DF9": (
            "(3R,4S)-1-[(4-AMINO-5H-PYRROLO[3,2-D]PYRIMIDIN-7-YL)METHYL]-4-[(BENZYLSULFANYL)METHYL]PYRROLIDIN-3-OL",
            "non-polymer",
        ),
        "GGP": ("BETA-1,2,3,4,6-PENTA-O-GALLOYL-D-GLUCOPYRANOSE", "non-polymer"),
        "GGS": (
            "PHOSPHONOOXY-[(10E)-3,7,11,15-TETRAMETHYLHEXADECA-2,6,10,14-TETRAENYL]SULFANYL-PHOSPHINIC ACID",
            "non-polymer",
        ),
        "MDM": ("METHYL-O3-(ALPHA-D-MANNOSE)-ALPHA-D-MANNOSE", "saccharide"),
        "MDB": (
            "(S)-N-(CYCLOPENTYLMETHYL)-N-(2-(HYDROXYAMINO)-2-OXOETHYL)-2-(3-(2-METHOXYPHENYL)UREIDO)-3,3-DIMETHYLBUTANAMIDE",
            "non-polymer",
        ),
        "MDC": (
            "N-[2-(1-MALEIMIDYL)ETHYL]-7-DIETHYLAMINOCOUMARIN-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "MDA": ("2,6-DIDEOXY-3 C-METHYL-D-RIBOPYRANOSIDE", "saccharide"),
        "DF1": (
            "(2R)-1-[(5,6-DIPHENYL-7H-PYRROLO[2,3-D]PYRIMIDIN-4-YL)AMINO]PROPAN-2-OL",
            "non-polymer",
        ),
        "DF0": ("(2-FLUOROBIPHENYL-4-YL)ACETIC ACID", "non-polymer"),
        "DF3": (
            "N-[2-(3-{[2-(2,3-DIHYDRO-1,4-BENZODIOXIN-6-YLAMINO)-2-OXOETHYL]SULFANYL}-1H-INDOL-1-YL)ETHYL]-3-(TRIFLUOROMETHYL)BENZAMIDE",
            "non-polymer",
        ),
        "DF2": (
            "(2R)-3-{[(4Z)-5,6-DIPHENYL-6,7-DIHYDRO-4H-PYRROLO[2,3-D]PYRIMIDIN-4-YLIDENE]AMINO}PROPANE-1,2-DIOL",
            "non-polymer",
        ),
        "MDZ": (
            "N~6~-METHYL-6-OXO-L-LYSINE - 2-[(3-MERCAPTOBUTANOYL)OXY]-3-METHYLBUTANOIC ACID",
            "non-polymer",
        ),
        "GGD": (
            "NONADEC-10-ENOIC ACID 2-[3,4-DIHYDROXY-6-HYDROXYMETHYL-5-(3,4,5-TRIHYDROXY-6-HYDROXYMETHYL-TETRAHYDRO-PYRAN-2-YLOXY)-TETRAHYDRO-PYRAN-2-YLOXY]-1-OCTADEC-9-ENOYLOXYMETHYL-ETHYL ESTER",
            "non-polymer",
        ),
        "FHM": (
            "S-3-(4-FLUOROPHENOXY)-2-HYDROXY-2-METHYL-N-[4-NITRO-3-(TRIFLUOROMETHYL)PHENYL]PROPANAMIDE",
            "non-polymer",
        ),
        "MDY": (
            "(2Z)-N-(3-CHLORO-2'-METHOXYBIPHENYL-4-YL)-2-CYANO-3-HYDROXYBUT-2-ENAMIDE",
            "non-polymer",
        ),
        "GGA": (
            "D-GAMMA-GLUTAMYL-N-{[(R)-{4-[(4-AMINOBUTYL)AMINO]BUTYL}(PHOSPHONOOXY)PHOSPHORYL]METHYL}-D-ALANINAMIDE",
            "non-polymer",
        ),
        "GGC": ("1-MENAPHTHYL GLUTATHIONE CONJUGATE", "non-polymer"),
        "GGB": ("L-CANAVANINE", "non-polymer"),
        "GGM": (
            "3'-O-(N-METHYLANTHRANILOYL)-BETA:GAMMA-IMIDOGUANOSINE-5'-TRIPHOSPHATE",
            "non-polymer",
        ),
        "GGL": ("GAMMA-L-GLUTAMIC ACID", "peptide"),
        "GGO": (
            "(4R,5R)-5-AMINO-1-[2-(1,3-BENZODIOXOL-5-YL)ETHYL]-4-(2,4,5-TRIFLUOROPHENYL)PIPERIDIN-2-ONE",
            "non-polymer",
        ),
        "MDQ": ("?", "nucleic-acid"),
        "MDV": (
            "(7S,8S)-3-(2-DEOXY-5-O-PHOSPHONO-BETA-D-ERYTHRO-PENTOFURANOSYL)-7,8-DIHYDRO-3H-IMIDAZO[2,1-I]PURINE-7,8-DIOL",
            "nucleic-acid",
        ),
        "GGH": (
            "2'-DEOXY-5'-O-(HYDROXY{[HYDROXY(PHOSPHONOMETHYL)PHOSPHORYL]OXY}PHOSPHORYL)GUANOSINE",
            "non-polymer",
        ),
        "MDU": ("?", "nucleic-acid"),
        "GG5": ("4-[3-(4-FLUOROPHENYL)-1H-PYRAZOL-4-YL]PYRIDINE", "non-polymer"),
        "GG4": ("?", "non-polymer"),
        "GG7": ("[(1R)-1-AMINOETHYL]PHOSPHONIC ACID", "non-polymer"),
        "GG6": ("[(1S,2S)-1,2-DIHYDROXYPROPYL]PHOSPHONIC ACID", "non-polymer"),
        "DFY": ("(5,6-DIPHENYL-FURO[2,3-D]PYRIMIDIN-4-YLAMINO)-ACETIC", "non-polymer"),
        "DFX": ("1,2-DEOXY-2-FLUORO-XYLOPYRANOSE", "saccharide"),
        "GG3": (
            "{1-HYDROXY-3-[METHYL(4-PHENYLBUTYL)AMINO]PROPANE-1,1-DIYL}BIS(PHOSPHONIC ACID)",
            "non-polymer",
        ),
        "DFZ": ("3-AMINO-3-BENZYL-[4.3.0]BICYCLO-1,6-DIAZANONAN-2-ONE", "non-polymer"),
        "DFU": ("(2S,3R,4S,5R)-2-METHYLPIPERIDINE-3,4,5-TRIOL", "non-polymer"),
        "DFT": (
            "1-[2-DEOXYRIBOFURANOSYL]-2,4-DIFLUORO-5-METHYL-BENZENE-5'MONOPHOSPHATE",
            "nucleic-acid",
        ),
        "DFW": ("N-(5,6-DIPHENYLFURO[2,3-D]PYRIMIDIN-4-YL)GLYCINE", "non-polymer"),
        "DFV": ("7-HYDROXY-2-(4-HYDROXY-PHENYL)-CHROMAN-4-ONE", "non-polymer"),
        "DFQ": (
            "(2E)-3-{6-[(1R)-1-(2,6-DICHLORO-3-FLUOROPHENYL)ETHOXY]QUINOLIN-3-YL}-N-METHYLPROP-2-ENAMIDE",
            "non-polymer",
        ),
        "DFP": ("DIISOPROPYL PHOSPHONATE", "non-polymer"),
        "DFR": ("3-DEOXY-3-METHYL-D-FRUCTOSE", "saccharide"),
        "DFM": (
            "N-ACETYL-L-PHENYLALANYL-4-[DIFLUORO(PHOSPHONO)METHYL]-L-PHENYLALANINAMIDE",
            "non-polymer",
        ),
        "DFL": ("2-(4-HYDROXY-PHENYL)-CHROMAN-4-ONE", "non-polymer"),
        "DFO": ("2,2-DIFLUORO-3-HYDROSTATINE", "peptide"),
        "DFN": (
            "3-[3-(2,3-DIHYDROXY-PROPYLAMINO)-PHENYL]-4-(5-FLUORO-1-METHYL-1H-INDOL-3-YL)-PYRROLE-2,5-DIONE",
            "non-polymer",
        ),
        "DFI": ("2,2-DIFLUOROSTATINE", "peptide"),
        "DFH": ("1,1-DIPHENYLETHANOL", "non-polymer"),
        "DFK": (
            "D-PHENYLALANYL-N-[(1S)-4-{[(Z)-AMINO(IMINO)METHYL]AMINO}-1-(CHLOROACETYL)BUTYL]-L-PROLINAMIDE",
            "non-polymer",
        ),
        "DFJ": (
            "N-({4-[DIFLUORO(PHOSPHONO)METHYL]PHENYL}ACETYL)-L-PHENYLALANYL-4-[DIFLUORO(PHOSPHONO)METHYL]-L-PHENYLALANINAMIDE",
            "non-polymer",
        ),
        "DFE": ("2-(1,1-DIFLUOROETHOXY)-1,1,1-TRIFLUOROETHANE", "non-polymer"),
        "DFD": ("DECYL FORMATE", "non-polymer"),
        "DFG": ("?", "nucleic-acid"),
        "MD1": (
            "PHOSPHORIC ACID 4-(2-AMINO-4-OXO-3,4,5,6,-TETRAHYDRO-PTERIDIN-6-YL)-2-HYDROXY-3,4-DIMERCAPTO-BUT-3-EN-YL ESTER GUANYLATE ESTER",
            "non-polymer",
        ),
        "DFA": ("DIPHENYLACETIC ACID", "non-polymer"),
        "MD7": (
            "(2Z)-N-BIPHENYL-4-YL-2-CYANO-3-CYCLOPROPYL-3-HYDROXYPROP-2-ENAMIDE",
            "non-polymer",
        ),
        "DFC": ("?", "nucleic-acid"),
        "DFB": ("2,3-DIFLUOROBENZYL ALCOHOL", "non-polymer"),
        "M4A": ("2-METHYLQUINOLIN-4-AMINE", "non-polymer"),
        "7JZ": ("2-DEOXY-2,2-DIFLUORO-BETA-D-LYXO-HEXOPYRANOSE", "saccharide"),
        "SM9": (
            "N-(2-ETHOXYETHYL)-N-{(2S)-2-HYDROXY-3-[(5R)-2-(QUINAZOLIN-4-YL)-2,7-DIAZASPIRO[4.5]DEC-7-YL]PROPYL}-2,6-DIMETHYLBENZENESULFONAMIDE",
            "non-polymer",
        ),
        "SM8": ("SIMOCYCLINONE D8", "non-polymer"),
        "SM7": (
            "6-[1-(PIPERIDIN-4-YL)-3-(PYRIDIN-4-YL)-1H-PYRAZOL-4-YL]INDENO[1,2-C]PYRAZOLE",
            "non-polymer",
        ),
        "SM6": (
            "ETHYL 3-{[1-(HYDROXYAMINO)-2H-INDEN-5-YL]AMINO}THIENO[2,3-C]PYRIDINE-2-CARBOXYLATE",
            "non-polymer",
        ),
        "SM5": (
            "(1E)-5-(1-PIPERIDIN-4-YL-3-PYRIDIN-4-YL-1H-PYRAZOL-4-YL)-2,3-DIHYDRO-1H-INDEN-1-ONE OXIME",
            "non-polymer",
        ),
        "SM4": (
            "(1R)-1-(2-THIOPHEN-2-YL-ACETYLAMINO)-1-(3-(2-CARBOXYVINYL)-PHENYL) METHYLBORONIC ACID",
            "non-polymer",
        ),
        "SM3": (
            "(1R)-1-(2-THIENYLACETYLAMINO)-1-PHENYLMETHYLBORONIC ACID",
            "non-polymer",
        ),
        "SM2": (
            "(1R)-1-(2-THIENYLACETYLAMINO)-1-(3-CARBOXYPHENYL)METHYLBORONIC ACID",
            "non-polymer",
        ),
        "FHO": ("N^5^-FORMYL-N^5^-HYDROXY-L-ORNITHINE", "peptide"),
        "NCZ": (
            "2-HYDROXY-7-METHOXY-5-METHYL-NAPHTHALENE-1-CARBOXYLIC ACID MESO-2,5-DIHYDROXY-CYCLOPENT-3-ENYL ESTER",
            "non-polymer",
        ),
        "YH": ("YTTRIUM ION, 1 WATER COORDINATED", "non-polymer"),
        "K37": (
            "4,5,6,7-TETRABROMO-2-(METHYLSULFANYL)-1H-BENZIMIDAZOLE",
            "non-polymer",
        ),
        "K36": (
            "(1S,2S)-2-({N-[(BENZYLOXY)CARBONYL]-L-LEUCYL}AMINO)-1-HYDROXY-3-[(3S)-2-OXOPYRROLIDIN-3-YL]PROPANE-1-SULFONIC ACID",
            "peptide",
        ),
        "K30": (
            "(2S)-4-(2,5-DIFLUOROPHENYL)-N-[(3R,4S)-3-FLUORO-1-METHYLPIPERIDIN-4-YL]-2-(HYDROXYMETHYL)-N-METHYL-2-PHENYL-2,5-DIHYDRO-1H-PYRROLE-1-CARBOXAMIDE",
            "non-polymer",
        ),
        "K32": ("4,5,6,7-TETRABROMO-1H,3H-BENZIMIDAZOL-2-ONE", "non-polymer"),
        "TOM": ("4-(2-METHOXYPHENYL)-2-OXOBUT-3-ENOIC ACID", "non-polymer"),
        "FHN": (
            "(2S,3R,5R,6S)-3,4,5-TRIHYDROXY-2,6-BIS(HYDROXYMETHYL)PIPERIDINIUM",
            "non-polymer",
        ),
        "YG": ("WYBUTOSINE", "nucleic-acid"),
        "MGT": ("7N-METHYL-8-HYDROGUANOSINE-5'-TRIPHOSPHATE", "non-polymer"),
        "94M": (
            "6-CHLORO-9-[(4-METHOXY-3,5-DIMETHYLPYRIDIN-2-YL)METHYL]-9H-PURIN-2-AMINE",
            "non-polymer",
        ),
        "BH1": ("BACTERIOPHEOPHYTIN", "non-polymer"),
        "BH0": (
            "3-({5-BENZYL-6-HYDROXY-2,4-BIS-(4-HYDROXY-BENZYL)-3-OXO-[1,2,4]-TRIAZEPANE-1-SULFONYL)-BENZONITRILE",
            "non-polymer",
        ),
        "BH2": ("(3R)-3-HYDROXY-L-ASPARTIC ACID", "peptide"),
        "BH4": ("?", "non-polymer"),
        "BH7": ("E-AMINO BIOTINYL CAPROIC ACID", "non-polymer"),
        "BH6": ("{[(2-CHLOROBENZOYL)AMINO]METHYL}BORONIC ACID", "non-polymer"),
        "QKH": ("METHYL-CYCLO-HEPTA-AMYLOSE", "non-polymer"),
        "HG9": ("1-DEOXY-1-[(2-HYDROXYETHYL)(NONANOYL)AMINO]HEXITOL", "non-polymer"),
        "HG7": (
            "(2S)-2-AZANYL-N-[3-(DIMETHYLAMINO)PROPYL]-2-(3-HYDROXYPHENYL)ETHANAMIDE",
            "peptide",
        ),
        "FHI": ("IODODIFLUNISAL", "non-polymer"),
        "HG2": ("DIBROMOMERCURY", "non-polymer"),
        "AB0": (
            "3,7:6,10:9,14-TRIANHYDRO-2,5,11,12,13-PENTADEOXY-4-O-(METHOXYMETHYL)-L-ARABINO-L-ALLO-TETRADEC-12-ENONIC ACID",
            "non-polymer",
        ),
        "AB1": (
            "N-{1-BENZYL-4-[2-(2,6-DIMETHYL-PHENOXY)-ACETYLAMINO]-3-HYDROXY-5-PHENYL-PENTYL}-3-METHYL-2-(2-OXO-TETRAHYDRO-PYRIMIDIN-1-YL)-BUTYRAMIDE",
            "non-polymer",
        ),
        "AB2": (
            "[1-((1S,2R)-1-BENZYL-2-HYDROXY-3-{ISOBUTYL[(4-METHOXYPHENYL)SULFONYL]AMINO}PROPYL)-1H-1,2,3-TRIAZOL-4-YL]METHYL (1R,2R)-2-HYDROXY-2,3-DIHYDRO-1H-INDEN-1-YLCARBAMATE",
            "non-polymer",
        ),
        "AB3": (
            "3-(4-AMINO-1-TERT-BUTYL-1H-PYRAZOLO[3,4-D]PYRIMIDIN-3-YL)PHENOL",
            "non-polymer",
        ),
        "AB4": (
            "2,5-DICHLORO-N-[4-HYDROXY-3-(2-HYDROXY-1-NAPHTHYL)PHENYL]BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "AB5": ("2(S)-AMINO-6-BORONOHEXANOIC ACID", "non-polymer"),
        "AB6": (
            "(2R)-4-AMINO-N-((1R,2S,3R,4R,5S)-5-AMINO-4-[(2-AMINO-2-DEOXY-ALPHA-D-GLUCOPYRANOSYL)OXY]-2-{2-[(3-AMINOPROPYL)AMINO]ETHOXY}-3-HYDROXYCYCLOHEXYL)-2-HYDROXYBUTANAMIDE",
            "non-polymer",
        ),
        "AB7": ("ALPHA-AMINOBUTYRIC ACID", "peptide"),
        "AB8": (
            "(1S,3R,8AS)-8-(2-{(4S,6S)-3-(4-HYDROXY-3-METHOXYBENZYL)-4-[2-(METHYLAMINO)-2-OXOETHYL]-2-OXO-1,3-OXAZINAN-6-YL}ETHYL)-3,7-DIMETHYL-1,2,3,7,8,8A-HEXAHYDRONAPHTHALEN-1-YL (2R)-2-METHYLBUTANOATE",
            "non-polymer",
        ),
        "AB9": (
            "(2R)-4-AMINO-N-{(1R,2S,3R,4R,5S)-5-AMINO-2-{2-[(2-AMINOETHYL)AMINO]ETHOXY}-4-[(2,6-DIAMINO-2,6-DIDEOXY-ALPHA-D-GLUCOPYRANOSYL)OXY]-3-HYDROXYCYCLOHEXYL}-2-HYDROXYBUTANAMIDE",
            "non-polymer",
        ),
        "QKV": (
            "(6-BROMO-1'H,4H-SPIRO[1,3-BENZODIOXINE-2,4'-PIPERIDIN]-1'-YL)METHANOL",
            "non-polymer",
        ),
        "NZS": (
            "6-(1,1-DIOXIDOTHIOMORPHOLIN-4-YL)-N-(3-METHOXYPHENYL)-1-METHYL-1H-PYRAZOLO[3,4-D]PYRIMIDIN-4-AMINE",
            "non-polymer",
        ),
        "ABA": ("ALPHA-AMINOBUTYRIC ACID", "peptide"),
        "ABB": (
            "(2S,4R)-1-ACETYL-N-[(1S)-4-[(AMINOIMINOMETHYL)AMINO]-1-(2-BENZOTHIAZOLYLCARBONYL)BUTYL]-4-HYDROXY-2-PYRROLIDINECARBOXAMIDE",
            "non-polymer",
        ),
        "ABC": ("MODIFIED ACARBOSE HEXASACCHARIDE", "saccharide"),
        "ABD": ("ACARBOSE DERIVED HEXASACCHARIDE", "saccharide"),
        "ABE": ("ABEQUOSE", "saccharide"),
        "ABF": ("BETA-D-ARABINOFURANOSE-5'-PHOSPHATE", "saccharide"),
        "ABG": ("?", "non-polymer"),
        "ABH": ("2(S)-AMINO-6-BORONOHEXANOIC ACID", "non-polymer"),
        "ABI": ("5-AMIDINO-BENZIMIDAZOLE", "non-polymer"),
        "ABJ": (
            "3-[4-AMINO-1-(1-METHYLETHYL)-1H-PYRAZOLO[3,4-D]PYRIMIDIN-3-YL]PHENOL",
            "non-polymer",
        ),
        "ABK": ("3-AMINO-5,5-DI-FLUORO-2-KETO-PENTAN-1-OIC ACID", "non-polymer"),
        "ABL": (
            "(2R,3R,4R,5R)-4,5-DIHYDROXY-2-(HYDROXYMETHYL)-6-OXOPIPERIDIN-3-YL BETA-D-GLUCOPYRANOSIDE",
            "saccharide",
        ),
        "ABM": ("METHYL PHOSPHONIC ACID ADENOSINE ESTER", "non-polymer"),
        "ABN": ("BENZYLAMINE", "non-polymer"),
        "ABO": (
            "4-[3-(1H-BENZIMIDAZOL-2-YL)-1H-INDAZOL-6-YL]-2-METHOXYPHENOL",
            "non-polymer",
        ),
        "ABP": ("8-BROMOADENOSINE-5'-DIPHOSPHATE", "non-polymer"),
        "ABQ": (
            "[4-AMINO-7-(PROPAN-2-YL)-7H-PYRROLO[2,3-D]PYRIMIDIN-5-YL](6-{[(3S,4R)-4-(4-FLUOROPHENYL)TETRAHYDROFURAN-3-YL]AMINO}PYRAZIN-2-YL)METHANONE",
            "non-polymer",
        ),
        "ABR": (
            "(R)-(N-PHENYL-2-HYDROXY-ETHYL)-2'-DEOXY-ADENOSINE-5'-MONOPHOSPHATE",
            "nucleic-acid",
        ),
        "ABS": (
            "(S)-(N-PHENYL-2-HYDROXY-ETHYL)-2'-DEOXY-ADENOSINE-5'-MONOPHOSPHATE",
            "nucleic-acid",
        ),
        "ABT": ("?", "nucleic-acid"),
        "ABU": ("GAMMA-AMINO-BUTANOIC ACID", "non-polymer"),
        "ABV": ("1,3-BENZOTHIAZOL-2-AMINE", "non-polymer"),
        "ABW": (
            "5-[1-(ACETYLAMINO)-3-METHYLBUTYL]-2,5-ANHYDRO-3,4-DIDEOXY-4-(METHOXYCARBONYL)PENTONIC ACID",
            "non-polymer",
        ),
        "ABX": (
            "5-[1-(ACETYLAMINO)-3-METHYLBUTYL]-4-(METHOXYCARBONYL)PROLINE",
            "non-polymer",
        ),
        "ABY": (
            "N-(4-AMINOBUTANOYL)-S-(4-METHOXYBENZYL)-L-CYSTEINYLGLYCINE",
            "non-polymer",
        ),
        "ABZ": (
            "4-[4-AMINO-6-(5-CHLORO-1H-INDOL-4-YLMETHYL)-[1,3,5]TRIAZIN-2-YLAMINO]-BENZONITRILE",
            "non-polymer",
        ),
        "NZH": (
            "(2S)-2-AMINO-3-[1-(1H-TETRAAZOL-5-YL)-1H-IMIDAZOL-4-YL]PROPANAL",
            "peptide",
        ),
        "HGB": ("4-(HYDROXYMERCURY)BENZOIC ACID", "non-polymer"),
        "HGC": ("METHYL MERCURY ION", "non-polymer"),
        "QK5": (
            "(3-{(3R)-1-[4-(1-BENZOTHIOPHEN-2-YL)PYRIMIDIN-2-YL]PIPERIDIN-3-YL}-2-METHYL-1H-PYRROLO[2,3-B]PYRIDIN-1-YL)ACETONITRILE",
            "non-polymer",
        ),
        "HGA": ("GLUTAMINE HYDROXAMATE", "non-polymer"),
        "BHI": ("4-BROMO-3-HYDROXY-3-METHYL BUTYL DIPHOSPHATE", "non-polymer"),
        "BHH": ("N-BUTYL-N'-HYDROXYGUANIDINE", "non-polymer"),
        "BHK": ("(2R)-2-BENZYL-5-HYDROXY-4-OXOPENTANOIC ACID", "non-polymer"),
        "R68": (
            "N-{4-[(4-{3-[(2R)-3,3-DIMETHYLPIPERIDIN-2-YL]-2-FLUORO-6-HYDROXYBENZOYL}BENZOYL)AMINO]AZEPAN-3-YL}ISONICOTINAMIDE",
            "non-polymer",
        ),
        "BHM": (
            "(R)-3-BROMO-2-HYDROXY-2-METHYL-N-[4-NITRO-3-(TRIFLUOROMETHYL)PHENYL]PROPANAMIDE",
            "non-polymer",
        ),
        "BHL": ("6-BROMOHEXAN-1-OL", "non-polymer"),
        "BHO": ("BENZHYDROXAMIC ACID", "non-polymer"),
        "BHN": (
            "2-[2-[CARBOXYMETHYL-[(2-HYDROXYPHENYL)METHYL]AMINO]ETHYL-[(2-HYDROXYPHENYL)METHYL]AMINO]ETHANOIC ACID",
            "non-polymer",
        ),
        "BHA": ("2-HYDROXY-4-AMINOBENZOIC ACID", "non-polymer"),
        "BHC": ("BENZENE HEXACARBOXYLIC ACID", "non-polymer"),
        "BHB": (
            "2,6-DIAMINO-8-(2-DIMETHYLAMINOETHYLSULFANYLMETHYL)-3H-QUINAZOLIN-4-ONE",
            "non-polymer",
        ),
        "942": (
            "N-[6-AZANYL-1-BUTYL-2,4-BIS(OXIDANYLIDENE)PYRIMIDIN-5-YL]-N-METHYL-BENZAMIDE",
            "non-polymer",
        ),
        "BHD": ("(3S)-3-HYDROXY-L-ASPARTIC ACID", "peptide"),
        "BHG": (
            "2-HEXYLOXY-6-HYDROXYMETHYL-TETRAHYDRO-PYRAN-3,4,5-TRIOL",
            "saccharide",
        ),
        "941": (
            "2-(4-{2-TERT-BUTOXYCARBONYLAMINO-2-[4-(3-HYDROXY-2-METHOXYCARBONYL-PHENOXY)-BUTYLCARBAMOYL]-ETHYL}-PHENOXY)-MALONIC ACID",
            "non-polymer",
        ),
        "BHY": (
            "N,N'-[BIPHENYL-4,4'-DIYLDI(2R)PROPANE-2,1-DIYL]DIMETHANESULFONAMIDE",
            "non-polymer",
        ),
        "BHX": (
            "(2S)-2-{[(S)-(2-CARBOXYETHYL)(HYDROXY)PHOSPHORYL]METHYL}PENTANEDIOIC ACID",
            "non-polymer",
        ),
        "K3G": ("PHOSPHOMOLYBDATE KEGGIN", "non-polymer"),
        "BHZ": (
            "2-[2-[CARBOXYMETHYL(PHENYLMETHYL)AMINO]ETHYL-[(2-HYDROXYPHENYL)METHYL]AMINO]ETHANOIC ACID",
            "non-polymer",
        ),
        "NHL": ("(4S)-4-(2-NAPHTHYLMETHYL)-D-GLUTAMIC ACID", "peptide"),
        "BHQ": ("2,5-DITERT-BUTYLBENZENE-1,4-DIOL", "non-polymer"),
        "BHP": (
            "(S)-5-(4-BENZYLOXY-PHENYL)-4-(7-PHENYL-HEPTANOYLAMINO)-PENTANOIC ACID",
            "non-polymer",
        ),
        "BHS": ("6S-5,6,7,8-TETRAHYDROBIOPTERIN", "non-polymer"),
        "BHR": (
            "2-[2-[CARBOXYMETHYL-[(2-HYDROXYPHENYL)METHYL]AMINO]ETHYL-[(2,3-DIHYDROXYPHENYL)METHYL]AMINO]ETHANOIC ACID",
            "non-polymer",
        ),
        "BHT": (
            "N-[4-(BIS{4-[(3-SULFOPHENYL)AMINO]PHENYL}METHYLENE)CYCLOHEXA-2,5-DIEN-1-YLIDENE]-4-SULFOBENZENAMINIUM",
            "non-polymer",
        ),
        "ZGC": ("3-(4-METHOXYPHENYL)PROPANOIC ACID", "non-polymer"),
        "SCU": ("N,N,N-TRIMETHYL-2-[(4-OXOBUTANOYL)OXY]ETHANAMINIUM", "non-polymer"),
        "OEF": ("3,5-DIBROMO-4-(3-ISOPROPYL-PHENOXY)BENZOIC ACID", "non-polymer"),
        "KT3": (
            "N-{4-4-(2,4-DIAMINO-6-OXO-1,6-DIHYDRO-PYRIMIDIN-5-YL)-1-(2,2,2-TRIFLUORO-1,1-DIHYDROXY-ETHYL)-BUT-2-YL-BENZOYL}-GAMMA-GLUTAMYL-GAMMA-GLUTAMYL-GLUTAMIC ACID",
            "non-polymer",
        ),
        "GLJ": ("5,5-DIHYDROXY-L-NORVALINE", "peptide"),
        "M91": (
            "(2S)-1-{[(2R)-1-{[(2S,3S)-1-CYCLOHEXYL-3-HYDROXY-4-(2-OXOPYRIDIN-1(2H)-YL)BUTAN-2-YL]AMINO}-3-(METHYLSULFANYL)-1-OXOPROPAN-2-YL]AMINO}-1-OXO-3-PHENYLPROPAN-2-YL 4-AMINOPIPERIDINE-1-CARBOXYLATE",
            "peptide",
        ),
        "LP9": (
            "3-[4-(DIETHYLAMINO)PHENOXY]-6-(ETHOXYCARBONYL)-5,8-DIHYDROXY-7-OXO-7,8-DIHYDRO-1,8-NAPHTHYRIDIN-1-IUM",
            "non-polymer",
        ),
        "LP8": (
            "3-CYCLOPENTYL-1,4-DIHYDROXY-1,8-NAPHTHYRIDIN-2(1H)-ONE",
            "non-polymer",
        ),
        "LP7": (
            "ETHYL 1,4-DIHYDROXY-2-OXO-1,2-DIHYDRO-1,8-NAPHTHYRIDINE-3-CARBOXYLATE",
            "non-polymer",
        ),
        "LP6": ("6-PIPERIDIN-1-YL-L-NORLEUCINE", "peptide"),
        "LP5": (
            "(R)-((2R,3S,4R,5R,6R)-3-HYDROXY-2-(HYDROXYMETHYL)-5-((R)-3-HYDROXYTETRADECANAMIDO)-6-(PHOSPHONOOXY)TETRAHYDRO-2H-PYRAN-4-YL) 3-HYDROXYTETRADECANOATE",
            "non-polymer",
        ),
        "LP4": (
            "2-DEOXY-3-O-[(3R)-3-HYDROXYTETRADECANOYL]-2-{[(3R)-3-HYDROXYTETRADECANOYL]AMINO}-4-O-PHOSPHONO-BETA-D-GLUCOPYRANOSE",
            "non-polymer",
        ),
        "LP3": (
            "(7R)-4,7-DIHYDROXY-N,N,N-TRIMETHYL-10-OXO-3,5,9-TRIOXA-4-PHOSPHAHEPTACOSAN-1-AMINIUM 4-OXIDE",
            "non-polymer",
        ),
        "LP2": (
            "N2-[[[3-NAPHTYL-N-METHYLCARBONYL-ALANINYL-VALINYL]-AMINO-[1-[2-METHYL-PROPANYL]-2-HYDROXY-PROPAN-3-YL]-CARBONYL]-3-METHYL-ALANYL]-3-NAPHTYL-ALANYL-AMINE",
            "non-polymer",
        ),
        "LP1": (
            "4-[2-(2-ACETYLAMINO-3-NAPHTALEN-1-YL-PROPIONYLAMINO)-4-METHYL-PENTANOYLAMINO]-3-HYDROXY-6-METHYL-HEPTANOIC ACID [1-(1-CARBAMOYL-2-NAPHTHALEN-1-YL-ETHYLCARBAMOYL)-PROPYL]-AMIDE",
            "non-polymer",
        ),
        "NHV": (
            "(5R,6R,7S,8R)-6,7,8-TRIHYDROXY-5-(HYDROXYMETHYL)-2-[(PHENYLAMINO)METHYL]-5,6,7,8-TETRAHYDRO-1H-IMIDAZO[1,2-A]PYRIDIN-4-IUM",
            "non-polymer",
        ),
        "EY5": ("ERYTHROMYCIN D", "non-polymer"),
        "OQB": (
            "4-OXO-3-[(6-{[4-(QUINOXALIN-2-YLAMINO)-BENZOYLAMINO]-METHYL}-PYRIDINE-3-CARBONYL)-AMINO]-BUTYRIC ACID",
            "non-polymer",
        ),
        "PJ7": ("4-AMINOFURAN-2-CARBOXYLIC ACID", "non-polymer"),
        "NHQ": ("N1-ACETYLSPERMINE-S-COA", "non-polymer"),
        "R36": (
            "4-AMINO-N-{4-[2-(2,6-DIMETHYL-PHENOXY)-ACETYLAMINO]-3-HYDROXY-1-ISOBUTYL-5-PHENYL-PENTYL}-BENZAMIDE",
            "non-polymer",
        ),
        "R37": (
            "3-AMINO-N-{4-[2-(2,6-DIMETHYL-PHENOXY)-ACETYLAMINO]-3-HYDROXY-1-ISOBUTYL-5-PHENYL-PENTYL}-BENZAMIDE",
            "non-polymer",
        ),
        "R31": (
            "(2S,4S,5S)-5-AMINO-N-(3-AMINO-2,2-DIMETHYL-3-OXOPROPYL)-4-HYDROXY-6-{4-[2-(3-METHOXYPROPOXY)PHENYL]-3-OXOPIPERAZIN-1-YL}-2-(PROPAN-2-YL)HEXANAMIDE",
            "non-polymer",
        ),
        "R32": (
            "(2S,4S,5S)-5-AMINO-N-(3-AMINO-2,2-DIMETHYL-3-OXOPROPYL)-6-[4-(2-CHLOROPHENYL)-2,2-DIMETHYL-5-OXOPIPERAZIN-1-YL]-4-HYDROXY-2-(PROPAN-2-YL)HEXANAMIDE",
            "non-polymer",
        ),
        "HIE": (
            "4-[6,6-DIMETHYL-4-OXO-3-(TRIFLUOROMETHYL)-4,5,6,7-TETRAHYDRO-1H-INDAZOL-1-YL]-2-[(CIS-4-HYDROXYCYCLOHEXYL)AMINO]BENZAMIDE",
            "non-polymer",
        ),
        "SKP": (
            "5-(1-CARBOXY-1-PHOSPHONOOXY-ETHOXYL)-4-HYDROXY-3-PHOSPHONOOXY-CYCLOHEX-1-ENECARBOXYLIC ACID",
            "non-polymer",
        ),
        "IHZ": (
            "5-[(2-METHYL-5-{[3-(TRIFLUOROMETHYL)PHENYL]CARBAMOYL}PHENYL)AMINO]PYRIDINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "NHS": ("10-FORMYL-5,8,10-TRIDEAZAFOLIC ACID", "non-polymer"),
        "EYG": (
            "{(2R,4Z)-2-[(1R)-1-AMINO-3-(METHYLSULFANYL)PROPYL]-2-HYDROXY-4-[(4-HYDROXYPHENYL)METHYLIDENE]-5-OXOIMIDAZOLIDIN-1-YL}ACETIC ACID",
            "peptide",
        ),
        "LPX": (
            "(2S)-3-{[(R)-(2-AMINOETHOXY)(HYDROXY)PHOSPHORYL]OXY}-2-HYDROXYPROPYL HEXADECANOATE",
            "non-polymer",
        ),
        "EYK": ("2-TERT-BUTYLBENZENE-1,4-DIOL", "non-polymer"),
        "LPT": ("CIS-(AMMINE)(CYCLOHEXYLAMINE)PLATINUM(II) COMPLEX", "non-polymer"),
        "LPS": (
            "O-{HYDROXY[((2R)-2-HYDROXY-3-{[(1S)-1-HYDROXYPENTADECYL]OXY}PROPYL)OXY]PHOSPHORYL}-L-SERINE",
            "peptide",
        ),
        "IHX": (
            "3-[3-(4-METHYLPIPERAZIN-1-YL)-7-(TRIFLUOROMETHYL)QUINOXALIN-5-YL]PHENOL",
            "non-polymer",
        ),
        "LPQ": (
            "(3R,4S)-N-{2-CHLORO-5-[(CYCLOPROPYLAMINO)METHYL]BENZYL}-N-CYCLOPROPYL-4-{6-[2-(2,6-DICHLORO-4-METHYLPHENOXY)ETHOXY]PYRIDIN-3-YL}PIPERIDINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "LPP": (
            "2-(HEXADECANOYLOXY)-1-[(PHOSPHONOOXY)METHYL]ETHYL HEXADECANOATE",
            "non-polymer",
        ),
        "LPO": (
            "(3S,4R)-N-[2-CHLORO-5-(2-METHOXYETHYL)BENZYL]-N-CYCLOPROPYL-4-{6-[2-(2,6-DICHLORO-4-METHYLPHENOXY)ETHOXY]PYRIDIN-3-YL}-4-HYDROXYPIPERIDINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "LPN": (
            "(3R,4S)-N-[2-CHLORO-5-(3-METHOXYPROPYL)BENZYL]-N-CYCLOPROPYL-4-{4-[2-(2,6-DICHLORO-4-METHYLPHENOXY)ETHOXY]PHENYL}PIPERIDINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "EYS": ("S-SELANYLCYSTEINAL", "peptide"),
        "LPL": ("LEU-HYDROXYETHYLENE-LEU", "peptide"),
        "LPG": ("L-PROPARGYLGLYCINE", "peptide"),
        "LPF": (
            "1,1,1-TRIFLUORO-3-((N-ACETYL)-L-LEUCYLAMIDO)-4-PHENYL-BUTAN-2-ONE(N-ACETYL-L-LEUCYL-L-PHENYLALANYL TRIFLUOROMETHYL KETONE)",
            "non-polymer",
        ),
        "LPE": ("1-O-OCTADECYL-SN-GLYCERO-3-PHOSPHOCHOLINE", "non-polymer"),
        "LPD": ("L-PROLINAMIDE", "peptide"),
        "LPC": ("[1-MYRISTOYL-GLYCEROL-3-YL]PHOSPHONYLCHOLINE", "non-polymer"),
        "LPB": ("5-[(3S)-1,2-DITHIOLAN-3-YL]PENTANOIC ACID", "non-polymer"),
        "LPA": ("LIPOIC ACID", "non-polymer"),
        "PA5": ("5-PHOSPHOARABINONIC ACID", "saccharide"),
        "TCY": (
            "(2R,3AS,4AR,5AR,5BS)-2-(6-AMINO-9H-PURIN-9-YL)-3A-HYDROXYHEXAHYDROCYCLOPROPA[4,5]CYCLOPENTA[1,2-B]FURAN-5A(4H)-YL DIHYDROGEN PHOSPHATE",
            "nucleic-acid",
        ),
        "PA9": ("N~2~-ACETYL-N~5~-(PHOSPHONOACETYL)-L-ORNITHINE", "non-polymer"),
        "IHP": ("INOSITOL HEXAKISPHOSPHATE", "non-polymer"),
        "ZF0": ("1-(3-METHYLPHENYL)METHANAMINE", "non-polymer"),
        "OIC": ("OCTAHYDROINDOLE-2-CARBOXYLIC ACID", "peptide"),
        "NNC": (
            "O-[2-(1,3-DIOXO-1,3-DIHYDRO-2H-ISOINDOL-2-YL)ETHYL] (4-CHLOROPHENYL)THIOCARBAMATE",
            "non-polymer",
        ),
        "W29": ("4-(3-ETHYLTHIOPHEN-2-YL)BENZENE-1,2-DIOL", "non-polymer"),
        "NFN": (
            "6-[[(2-ETHOXY-1-NAPHTHALENYL)CARBONYL]AMINO]-3,3-DIMETHYL-7-OXO-4-THIA-1-AZABICYCLO[3.2.0]HEPTANE-2-CARBOXYLATE",
            "non-polymer",
        ),
        "CK5": (
            "3-[4-(2,4-DIMETHYL-THIAZOL-5-YL)-PYRIMIDIN-2-YLAMINO]-PHENOL",
            "non-polymer",
        ),
        "L18": (
            "(2S)-1-METHYL-2-[(2S,4R)-2-METHYL-4-PHENYLPENTYL]PIPERIDINE",
            "non-polymer",
        ),
        "F2P": ("1,6-DI-O-PHOSPHONO-D-ALLITOL", "non-polymer"),
        "D0R": (
            "N-[(2S,4S,5S)-5-AMINO-4-HYDROXY-1,6-DIPHENYLHEXAN-2-YL]-N~2~-(METHYL{[2-(PROPAN-2-YL)-1,3-THIAZOL-4-YL]METHYL}CARBAMOYL)-L-VALINAMIDE",
            "peptide",
        ),
        "CK4": (
            "4-(2,4-DIMETHYL-1,3-THIAZOL-5-YL)-N-[4-(TRIFLUOROMETHYL)PHENYL]PYRIMIDIN-2-AMINE",
            "non-polymer",
        ),
        "P90": (
            "{4-[(2S,4E)-2-(1,3-BENZOTHIAZOL-2-YL)-2-(1H-1,2,3-BENZOTRIAZOL-1-YL)-5-PHENYLPENT-4-ENYL]PHENYL}(DIFLUORO)METHYLPHOSPHONIC ACID",
            "non-polymer",
        ),
        "P91": (
            "8-BROMO-4-(2-CHLOROPHENYL)-N-(2-HYDROXYETHYL)-6-METHYL-1,3-DIOXO-1,2,3,6-TETRAHYDROPYRROLO[3,4-E]INDOLE-7-CARBOXAMIDE",
            "non-polymer",
        ),
        "58Z": (
            "1-[2-(1H-INDOL-5-YLOXY)ETHYL]PIPERIDINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "CK7": (
            "[4-(2-AMINO-4-METHYL-THIAZOL-5-YL)-PYRIMIDIN-2-YL]-(3-NITRO-PHENYL)-AMINE",
            "non-polymer",
        ),
        "P96": (
            "2-{[4-(DIETHYLAMINO)PIPERIDIN-1-YL]METHYL}-6-IODO-4-[3-(PHENYLAMINO)PROP-1-YN-1-YL]PHENOL",
            "non-polymer",
        ),
        "P97": (
            "D-PHENYLALANYL-N-[(4-CHLOROPYRIDIN-3-YL)METHYL]-L-PROLINAMIDE",
            "peptide",
        ),
        "CK6": (
            "4-[4-(4-METHYL-2-METHYLAMINO-THIAZOL-5-YL)-PYRIMIDIN-2-YLAMINO]-PHENOL",
            "non-polymer",
        ),
        "CS9": (
            "N'-{(1S,2R)-1-(3,5-DIFLUOROBENZYL)-2-HYDROXY-2-[(2R,4R)-4-PHENOXYPYRROLIDIN-2-YL]ETHYL}-5-METHYL-N,N-DIPROPYLISOPHTHALAMIDE",
            "non-polymer",
        ),
        "CS8": ("3-THIAOCTANOYL-COENZYME A", "nucleic-acid"),
        "CS5": (
            "N'-{(1S,2R)-1-(3,5-DIFLUOROBENZYL)-2-HYDROXY-3-[(3-METHOXYBENZYL)AMINO]PROPYL}-5-METHYL-N,N-DIPROPYLISOPHTHALAMIDE",
            "non-polymer",
        ),
        "CK1": ("4-(2,5-DICHLOROTHIEN-3-YL)PYRIMIDIN-2-AMINE", "non-polymer"),
        "CS7": (
            "N'-[(1S,2R)-2-[(2R,4R)-4-(BENZYLOXY)PYRROLIDIN-2-YL]-1-(3,5-DIFLUOROBENZYL)-2-HYDROXYETHYL]-5-METHYL-N,N-DIPROPYLISOPHTHALAMIDE",
            "non-polymer",
        ),
        "CS1": ("S-(2-ANILINYL-SULFANYL)-CYSTEINE", "peptide"),
        "CS0": ("S-(2-HYDROXYETHYL)-L-CYSTEINE", "peptide"),
        "CS3": ("S-[3-OXO-3-(2-THIENYL)PROPYL]-L-CYSTEINE", "peptide"),
        "CS2": ("D-MANNONIC ACID", "non-polymer"),
        "586": (
            "N-[4-(1-BENZYLCARBAMOYL-2-METHYL-PROPYLCARBAMOYL)-1-(3,5-DIFLUORO-PHENOXYMETHYL)-2-HYDROXY-4-METHOXY-BUTYL]-5-(METHANESULFONYL-METHYL-AMINO)-N'-(1-PHENYLETHYL)-ISOPHTHALAMIDE",
            "non-polymer",
        ),
        "587": ("C32-O-(1-METHYL-INDOL-5-YL) 18-HYDROXY-ASCOMYCIN", "non-polymer"),
        "CSO": ("S-HYDROXYCYSTEINE", "peptide"),
        "585": (
            "6-({5-FLUORO-2-[(3,4,5-TRIMETHOXYPHENYL)AMINO]PYRIMIDIN-4-YL}AMINO)-2,2-DIMETHYL-2H-PYRIDO[3,2-B][1,4]OXAZIN-3(4H)-ONE",
            "non-polymer",
        ),
        "582": (
            "5-ETHYL-3-METHYL-1,5-DIHYDRO-4H-PYRAZOLO[4,3-C]QUINOLIN-4-ONE",
            "non-polymer",
        ),
        "CK3": (
            "N-[4-(2,4-DIMETHYL-1,3-THIAZOL-5-YL)PYRIMIDIN-2-YL]-N'-HYDROXYIMIDOFORMAMIDE",
            "non-polymer",
        ),
        "580": (
            "3-(4-FLUOROPHENYL)-2-(6-METHYLPYRIDIN-2-YL)-5,6-DIHYDRO-4H-PYRROLO[1,2-B]PYRAZOLE",
            "non-polymer",
        ),
        "581": (
            "2-[3-(ACETYLAMINO)PHENYL]-5-HYDROXY-6-OXO-1,6-DIHYDROPYRIMIDINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "CSE": ("SELENOCYSTEINE", "peptide"),
        "CSD": ("3-SULFINOALANINE", "peptide"),
        "CSG": ("2'-AMINO-2'-DEOXYCYTIDINE-5'-TRIPHOSPHATE", "non-polymer"),
        "CK2": ("4-(2,4-DIMETHYL-1,3-THIAZOL-5-YL)PYRIMIDIN-2-AMINE", "non-polymer"),
        "CSA": ("S-ACETONYLCYSTEINE", "peptide"),
        "588": (
            "2-{[2-(2-CARBAMOYL-VINYL)-4-(2-METHANESULFONYLAMINO-2-PENTYLCARBAMOYL-ETHYL)-PHENYL]-OXALYL-AMINO}-BENZOIC ACID",
            "non-polymer",
        ),
        "589": (
            "5-[2-(CYCLOHEXYLAMINO)PYRIDIN-4-YL]-4-NAPHTHALEN-2-YL-2-(TETRAHYDRO-2H-PYRAN-4-YL)-2,4-DIHYDRO-3H-1,2,4-TRIAZOL-3-ONE",
            "non-polymer",
        ),
        "P9I": ("1-[1-(PYRIDIN-2-YL)INDOLIZIN-3-YL]ETHAN-1-ONE", "non-polymer"),
        "P9J": (
            "4-[(5-METHYL-1H-PYRAZOL-3-YL)AMINO]-2-PHENYLPHTHALAZIN-1(2H)-ONE",
            "non-polymer",
        ),
        "NK": ("SALICYLALDEHYDE", "non-polymer"),
        "CSY": (
            "[4-(4-HYDROXY-BENZYL)-2-(2-HYDROXY-1-METHYL-ETHYL)-5-OXO-IMIDAZOLIDIN-1-YL]-ACETIC ACID",
            "peptide",
        ),
        "CSX": ("S-OXY CYSTEINE", "peptide"),
        "CSZ": ("S-SELANYL CYSTEINE", "peptide"),
        "CSU": ("CYSTEINE-S-SULFONIC ACID", "peptide"),
        "P9A": (
            "3-({3-[(1S,4AS,6S,7S,9S,9AR)-1,6-DIMETHYL-2-OXODECAHYDRO-6,9-EPOXY-4A,7-METHANOBENZO[7]ANNULEN-1-YL]PROPANOYL}AMINO)-2,4-DIHYDROXYBENZOIC ACID",
            "non-polymer",
        ),
        "CSW": ("CYSTEINE-S-DIOXIDE", "peptide"),
        "CSV": ("?", "non-polymer"),
        "CSQ": ("?", "non-polymer"),
        "CSP": ("S-PHOSPHOCYSTEINE", "peptide"),
        "CSS": ("S-MERCAPTOCYSTEINE", "peptide"),
        "CSR": ("S-ARSONOCYSTEINE", "peptide"),
        "5MU": ("?", "nucleic-acid"),
        "KTL": (
            "(1S,2R,3R,4S)-1-{(1S)-2-[(2R,3S,4S)-3,4-DIHYDROXY-2-(HYDROXYMETHYL)TETRAHYDROTHIOPHENIUM-1-YL]-1-HYDROXYETHYL}-2,3,4,5-TETRAHYDROXYPENTYL SULFATE",
            "non-polymer",
        ),
        "5MP": ("5-METHYLPYRROLE", "non-polymer"),
        "5MS": (
            "N-{2-METHYL-5-[(6-PHENYLPYRIMIDIN-4-YL)AMINO]PHENYL}METHANESULFONAMIDE",
            "non-polymer",
        ),
        "5MR": (
            "NALPHA-{(2S)-3-[(S)-HYDROXY(PHENYL)PHOSPHORYL]-2-[(3-PHENYLISOXAZOL-5-YL)METHYL]PROPANOYL}-L-TRYPTOPHANAMIDE",
            "non-polymer",
        ),
        "CK9": (
            "2-{[(2-{[(1R)-1-(HYDROXYMETHYL)PROPYL]AMINO}-9-ISOPROPYL-9H-PURIN-6-YL)AMINO]METHYL}PHENOL",
            "non-polymer",
        ),
        "5MY": ("MYO-INOSITOL-(1,3,4,5,6)-PENTAKISPHOSPHATE", "non-polymer"),
        "5MZ": ("3-[5-(CHLOROMETHYL)-1,2,4-OXADIAZOL-3-YL]ANILINE", "non-polymer"),
        "5MD": ("5-METHYL-2'-DEOXYPSEUDOURIDINE", "nucleic-acid"),
        "5MA": (
            "1-(2,5-DICHLORO-3-{[5-CHLORO-1-(2H-PYRAZOLO[3,4-B]PYRIDIN-3-YLMETHYL)-1H-BENZOTRIAZOL-4-YL]OXY}PHENYL)METHANAMINE",
            "non-polymer",
        ),
        "5MC": ("5-METHYLCYTIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "5MB": ("5-METHYLBENZIMIDAZOLE", "non-polymer"),
        "5MM": ("2-O-METHYL-1,6-DI-O-PHOSPHONO-D-MANNITOL", "saccharide"),
        "1X6": ("O-[(2E)-3-AMINOPROP-2-ENOYL]-L-SERINE", "peptide"),
        "OIP": ("?", "nucleic-acid"),
        "5MI": ("5-METHYL INDOLE", "non-polymer"),
        "3OC": ("(1R)-3-OXOCYCLOHEXANECARBOXYLIC ACID", "non-polymer"),
        "3OB": (
            "3-CHLORO-5-({6-[2-(3,4-DIHYDROISOQUINOLIN-2(1H)-YL)-2-OXOETHYL]-3-(DIMETHYLAMINO)-2-OXO-1,2-DIHYDROPYRIDIN-4-YL}OXY)BENZONITRILE",
            "non-polymer",
        ),
        "3OE": (
            "4-(BENZYLOXY)-N-[(2S,3R)-3-HYDROXY-1-{[(2S)-1-{[(3-METHYLTHIOPHEN-2-YL)METHYL]AMINO}-1-OXO-4-PHENYLBUTAN-2-YL]AMINO}-1-OXOBUTAN-2-YL]BENZAMIDE",
            "peptide",
        ),
        "3OD": ("?", "non-polymer"),
        "3OG": (
            "(1S)-1-[(2R)-6-OXOTETRAHYDRO-2H-PYRAN-2-YL]UNDECYL ACETATE",
            "non-polymer",
        ),
        "3OF": ("HYDRATED FE (III) ION, 2 WATERS COORDINATED", "non-polymer"),
        "3OH": ("3-HYDROXY-PROPANOIC ACID", "non-polymer"),
        "3OJ": (
            "4-[(1-HYDROXY-1,3-DIHYDRO-2,1-BENZOXABOROL-5-YL)OXY]BENZENE-1,2-DICARBONITRILE",
            "non-polymer",
        ),
        "3OL": ("1-OCTEN-3-OL", "non-polymer"),
        "9L9": (
            "(2S)-2-[[4-[[2,4-BIS(AZANYL)-6-OXIDANYLIDENE-1H-PYRIMIDIN-5-YL]CARBAMOYLAMINO]PHENYL]CARBONYLAMINO]PENTANEDIOIC ACID",
            "non-polymer",
        ),
        "3OQ": (
            "3-(2-FLUOROETHYL)-4-({4-[(2S)-1,1,1-TRIFLUORO-2-HYDROXYPROPAN-2-YL]PHENYL}SULFONYL)BENZONITRILE",
            "non-polymer",
        ),
        "3OS": (
            "2-[(4-ETHYNYL-2-FLUOROPHENYL)AMINO]-3,4-DIFLUORO-N-(2-HYDROXYETHOXY)-5-{[(2-HYDROXYETHOXY)IMINO]METHYL}BENZAMIDE",
            "non-polymer",
        ),
        "3OR": (
            "3,4-DIFLUORO-2-[(2-FLUORO-4-IODOPHENYL)AMINO]-N-(2-HYDROXYETHOXY)-5-[(3-OXO-1,2-OXAZINAN-2-YL)METHYL]BENZAMIDE",
            "non-polymer",
        ),
        "3OW": ("?", "non-polymer"),
        "3OY": (
            "N-PROPANOYL-D-ALANYL-(4R)-4-[(7-BROMOISOQUINOLIN-1-YL)OXY]-N-{(1R,2S)-1-[(CYCLOPROPYLSULFONYL)CARBAMOYL]-2-ETHENYLCYCLOPROPYL}-L-PROLINAMIDE",
            "non-polymer",
        ),
        "3OX": ("?", "non-polymer"),
        "3OZ": (
            "[4-({(1S)-1-[(2,4-DICHLOROPHENYL)CARBAMOYL]-1,3-DIHYDRO-2H-ISOINDOL-2-YL}METHYL)-2-METHYLPHENOXY]ACETIC ACID",
            "non-polymer",
        ),
        "GBC": ("GABACULINE", "non-polymer"),
        "X13": (
            "N,N,N-TRIMETHYL-13-[(METHYLSULFONYL)SULFANYL]TRIDECAN-1-AMINIUM",
            "non-polymer",
        ),
        "KTA": ("(S)-CYCLOHEXANONE-2-ACETATE", "non-polymer"),
        "L14": ("TRANS-CYCLOHEXANE-1,4-DICARBOXYLIC ACID", "non-polymer"),
        "NMY": ("NEOMYCIN", "non-polymer"),
        "SEL": ("2-AMINO-1,3-PROPANEDIOL", "peptide"),
        "L15": ("2-[4-(3-METHYL-1H-PYRAZOL-4-YL)PHENYL]ETHANAMINE", "non-polymer"),
        "X17": (
            "(5S)-2-AMINO-5-(4-METHOXY-3-METHYLPHENYL)-3-METHYL-5-[(3S,5S,7S)-TRICYCLO[3.3.1.1~3,7~]DEC-1-YL]-3,5-DIHYDRO-4H-IMIDAZOL-4-ONE",
            "non-polymer",
        ),
        "MO1": ("MAGNESIUM ION, 1 WATER COORDINATED", "non-polymer"),
        "L17": ("L17", "non-polymer"),
        "NEU": ("L-NEOPTERIN", "non-polymer"),
        "NMF": (
            "6-[(2R)-2-(FLUOROMETHYL)-3-HYDROXY-PROPYL]-1,5-DIMETHYL-PYRIMIDINE-2,4-DIONE",
            "non-polymer",
        ),
        "LAU": ("DODECANOIC ACID", "non-polymer"),
        "MO5": ("MAGNESIUM ION, 5 WATERS COORDINATED", "non-polymer"),
        "L13": (
            "4-AMINO-5-(2-METHYLPHENYL)-2,4-DIHYDRO-3H-1,2,4-TRIAZOLE-3-THIONE",
            "non-polymer",
        ),
        "MR0": (
            "N~3~-{(1S,2R)-1-(3,5-DIFLUOROBENZYL)-2-HYDROXY-3-[(3-METHOXYBENZYL)AMINO]PROPYL}-N~1~,N~1~-DIPROPYLBENZENE-1,3,5-TRICARBOXAMIDE",
            "non-polymer",
        ),
        "TGP": ("?", "nucleic-acid"),
        "RRS": (
            "N4-HYDROXY-2-ISOBUTYL-N1-(9-OXO-1,8-DIAZA-TRICYCLO[10.6.1.013,18]NONADECA-12(19),13,15,17-TETRAEN-10-YL)-SUCCINAMIDE",
            "non-polymer",
        ),
        "GLI": ("ALA-ALA-PHE-PSI((S)-CHOH-CH2)-GLY-VAL-VAL-OME", "non-polymer"),
        "NMG": ("GUANIDINO ACETATE", "non-polymer"),
        "3O0": (
            "{4-AMINO-2-[(4-CHLOROPHENYL)AMINO]-1,3-THIAZOL-5-YL}(3-NITROPHENYL)METHANONE",
            "non-polymer",
        ),
        "3O3": ("N-[(3S)-3-AMINO-3-CARBOXYPROPYL]-L-GLUTAMIC ACID", "peptide"),
        "3O2": (
            "N-HYDROXY-1-(2-METHOXYETHYL)-4-{[4-(3-{5-[4-(TRIFLUOROMETHOXY)PHENYL]-2H-TETRAZOL-2-YL}PROPOXY)PHENYL]SULFONYL}PIPERIDINE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "3O4": (
            "N-[4-[2-[(4-MORPHOLIN-4-YLPHENYL)AMINO]PYRIMIDIN-4-YL]PHENYL]ETHANAMIDE",
            "non-polymer",
        ),
        "9LO": (
            "4-[[2,4-BIS(AZANYL)-6-OXIDANYLIDENE-1H-PYRIMIDIN-5-YL]CARBAMOYLAMINO]BENZOIC ACID",
            "non-polymer",
        ),
        "9LI": ("2-(4,6-DICHLORO-2-METHYL-1H-INDOL-3-YL)ETHANAMINE", "non-polymer"),
        "L51": (
            "1-{3-[(6-AMINOQUINOLIN-4-YL)AMINO]PHENYL}-3-[3-TERT-BUTYL-1-(4-METHYLPHENYL)-1H-PYRAZOL-5-YL]UREA",
            "non-polymer",
        ),
        "7NA": ("2-HYDROXY-7-METHOXY-5-METHYL NAPHTHOIC ACID", "non-polymer"),
        "GBP": ("S-(N-HYDROXY-N-BROMOPHENYLCARBAMOYL)GLUTATHIONE", "non-polymer"),
        "7NH": (
            "[2'-HYDROXY-3'-(1H-PYRROLO[3,2-C]PYRIDIN-2-YL)-BIPHENYL-3-YLMETHYL]-UREA",
            "non-polymer",
        ),
        "7NI": ("7-NITROINDAZOLE", "non-polymer"),
        "Y14": (
            "N-[(1S)-2-{[(3S)-1-BENZYL-3-CYANOPYRROLIDIN-3-YL]AMINO}-1-(CYCLOHEXYLMETHYL)-2-OXOETHYL]MORPHOLINE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "NMD": (
            "N-NAPHTHALEN-1-YLMETHYL-2'-[3,5-DIMETHOXYBENZAMIDO]-2'-DEOXY-ADENOSINE",
            "non-polymer",
        ),
        "W2X": (
            "N~6~-[(1R)-1-({[(1R,2R)-1-CARBOXY-3-HYDROXY-2-METHYLPROPYL]OXY}CARBONYL)-2-MERCAPTOPROP-2-EN-1-YL]-6-OXO-L-LYSINE",
            "non-polymer",
        ),
        "1IT": ("DIBENZOFURAN", "non-polymer"),
        "1IQ": (
            "2-(4-ISOPROPYL-4-METHYL-5-OXO-4,5-DIHYDRO-1H-IMIDAZOL-2-YL)QUINOLINE-3-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "1IP": ("N~2~-(PHOSPHONOACETYL)-L-ASPARAGINE", "non-polymer"),
        "SKD": (
            "2-ACETYLAMINO-7-(1,2-DIHYDROXY-ETHYL)-3-HYDROXY-6,8-DIOXA-BICYCLO[3.2.1]OCTANE-5-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "322": (
            "[(5R)-5-(2,3-DIBROMO-5-ETHOXY-4-HYDROXYBENZYL)-4-OXO-2-THIOXO-1,3-THIAZOLIDIN-3-YL]ACETIC ACID",
            "non-polymer",
        ),
        "323": (
            "2-[3,6-BIS(DIMETHYLAMINO)XANTHEN-9-YL]-5-METHANOYL-BENZOATE",
            "non-polymer",
        ),
        "320": (
            "2-(4-FLUOROPHENYL)-N-{[3-FLUORO-4-(1H-PYRROLO[2,3-B]PYRIDIN-4-YLOXY)PHENYL]CARBAMOYL}ACETAMIDE",
            "non-polymer",
        ),
        "321": (
            "N-{3-[5-HYDROXY-2-(3-METHYLBUTYL)-3-OXO-6-THIOPHEN-2-YL-2,3-DIHYDROPYRIDAZIN-4-YL]-1,1-DIOXIDO-2H-1,4-BENZOTHIAZIN-7-YL}METHANESULFONAMIDE",
            "non-polymer",
        ),
        "326": (
            "2-{(E)-[5-HYDROXY-3-METHYL-1-(2-METHYL-4-SULFOPHENYL)-1H-PYRAZOL-4-YL]DIAZENYL}-4-SULFOBENZOIC ACID",
            "non-polymer",
        ),
        "327": (
            "(3S,5E)-3-PROPYL-3,4-DIHYDROTHIENO[2,3-F][1,4]OXAZEPIN-5(2H)-IMINE",
            "non-polymer",
        ),
        "324": (
            "N-{3-[(5-CHLORO-1H-PYRROLO[2,3-B]PYRIDIN-3-YL)CARBONYL]-2,4-DIFLUOROPHENYL}PROPANE-1-SULFONAMIDE",
            "non-polymer",
        ),
        "325": (
            "N-{2,4-DIFLUORO-3-[(5-PYRIDIN-3-YL-1H-PYRROLO[2,3-B]PYRIDIN-3-YL)CARBONYL]PHENYL}ETHANESULFONAMIDE",
            "non-polymer",
        ),
        "328": (
            "4-({4-[(4-METHOXYPYRIDIN-2-YL)AMINO]PIPERIDIN-1-YL}CARBONYL)BENZONITRILE",
            "non-polymer",
        ),
        "329": (
            "(2S)-2-METHYL-2,3-DIHYDROTHIENO[2,3-F][1,4]OXAZEPIN-5-AMINE",
            "non-polymer",
        ),
        "32A": (
            "2-[(5-HEX-1-YN-1-YLFURAN-2-YL)CARBONYL]-N-METHYLHYDRAZINECARBOTHIOAMIDE",
            "non-polymer",
        ),
        "32G": (
            "(3AS,4S,6AR)-4-(5-{1-[4-(6-AMINO-9H-PURIN-9-YL)BUTYL]-1H-1,2,3-TRIAZOL-4-YL}PENTYL)TETRAHYDRO-1H-THIENO[3,4-D]IMIDAZOL-2(3H)-ONE",
            "non-polymer",
        ),
        "32D": (
            "(1S)-1-[2-(DIFLUOROMETHYL)PYRIDIN-4-YL]-4-FLUORO-1-(3-PYRIMIDIN-5-YLPHENYL)-1H-ISOINDOL-3-AMINE",
            "non-polymer",
        ),
        "T3": ("3,5,3'TRIIODOTHYRONINE", "non-polymer"),
        "32S": (
            "(S)-2-AMINO-3-(4H-SELENOLO[3,2-B]-PYRROL-6-YL)-PROPIONIC ACID",
            "peptide",
        ),
        "32P": (
            "3-{2-[(5-AMINOPENTYL)AMINO]-2-OXOETHOXY}-5-({[1-(4-FLUOROPHENYL)ETHYL]AMINO}CARBONYL)PHENYL PHENYLMETHANESULFONATE",
            "non-polymer",
        ),
        "32W": (
            "N-(2-METHOXYETHYL)-4-[(6-PYRIDIN-4-YLQUINAZOLIN-2-YL)AMINO]BENZAMIDE",
            "non-polymer",
        ),
        "32T": (
            "(S)-2-AMINO-3-(4H-THIENO[3,2-B]-PYRROL-6-YL)-PROPIONIC ACID",
            "peptide",
        ),
        "32U": (
            "D-PHENYLALANYL-N-{4-[AMINO(IMINIO)METHYL]BENZYL}-L-PROLINAMIDE",
            "peptide",
        ),
        "KRM": ("4-[5-(4-PHENOXYPHENYL)-1H-PYRAZOL-3-YL]MORPHOLINE", "non-polymer"),
        "L5T": (
            "N-[(CYCLOPENTYLOXY)CARBONYL]-3-METHYL-L-VALYL-(4R)-N-[(1R,2S)-1-CARBOXY-2-ETHENYLCYCLOPROPYL]-4-[(7-METHOXY-2-{2-[(2-METHYLPROPANOYL)AMINO]-1,3-THIAZOL-4-YL}QUINOLIN-4-YL)OXY]-L-PROLINAMIDE",
            "non-polymer",
        ),
        "L5G": (
            "7-METHOXY-4-[(6-PHENYL[1,2,4]TRIAZOLO[4,3-B]PYRIDAZIN-3-YL)METHOXY]QUINOLINE",
            "non-polymer",
        ),
        "RR8": (
            "(3R,4R)-1-METHANESULFONYL-PYRROLIDINE-3,4-DICARBOXYLIC ACID 3-[(4-CHLORO-PHENYL)-AMIDE] 4-{[2-FLUORO-4-(2-OXO-2H-PYRIDIN-1-YL)-PHENYL]-AMIDE}",
            "non-polymer",
        ),
        "TZP": ("4-METHYL-5-HYDROXYETHYLTHIAZOLE PHOSPHATE", "non-polymer"),
        "NMA": (
            "N-[3-HYDROXY-2-METHYL-5-PHOSPHONOOXYMETHYL-PYRIDIN-4-YLMETHYL]-2-METHYLALANINE",
            "non-polymer",
        ),
        "TZT": ("SOBLIDOTIN", "non-polymer"),
        "TZY": (
            "6-[4-(4-FLUOROPHENYL)-1,3-OXAZOL-5-YL]-3-ISOPROPYL[1,2,4]TRIAZOLO[4,3-A]PYRIDINE",
            "non-polymer",
        ),
        "TZZ": ("TRIMETHYL PHOSPHATE", "non-polymer"),
        "VNI": (
            "N-[(1R)-1-(2,4-DICHLOROPHENYL)-2-(1H-IMIDAZOL-1-YL)ETHYL]-4-(5-PHENYL-1,3,4-OXADIAZOL-2-YL)BENZAMIDE",
            "non-polymer",
        ),
        "OAI": ("6-(OXALYL-AMINO)-1H-INDOLE-5-CARBOXYLIC ACID", "non-polymer"),
        "TZC": ("1,2,4-TRIAZOLE-CARBOXAMIDINE", "non-polymer"),
        "TZB": (
            "(4S)-2-[(1E)-1-AMINOPROP-1-ENYL]-4,5-DIHYDRO-1,3-THIAZOLE-4-CARBOXYLIC ACID",
            "peptide",
        ),
        "TZE": ("2-(4-METHYL-THIAZOL-5-YL)-ETHANOL", "non-polymer"),
        "TZD": (
            "2-{3-[(4-AMINO-2-METHYLPYRIMIDIN-5-YL)METHYL]-4-METHYL-2-OXO-2,3-DIHYDRO-1,3-THIAZOL-5-YL}ETHYL TRIHYDROGEN DIPHOSPHATE",
            "non-polymer",
        ),
        "TZG": (
            "(S)-2-AMINO-3-(2-(2-CARBOXYETHYL)-5-CHLORO-4-NITROPHENYL)PROPIONIC ACID",
            "non-polymer",
        ),
        "TZF": ("4-(1H-1,2,4-TRIAZOL-1-YL)PHENOL", "non-polymer"),
        "TZM": ("3-(1H-1,2,4-TRIAZOL-1-YLMETHYL)ANILINE", "non-polymer"),
        "TZL": ("2-(SEC-BUTYL)THIAZOLE", "non-polymer"),
        "TZO": ("1,3-THIAZOLE-4-CARBOXYLIC ACID", "peptide"),
        "TZ1": (
            "N-{5-[(7-CHLOROQUINOLIN-4-YL)SULFANYL]-1,3,4-THIADIAZOL-2-YL}THIOPHENE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "TZ5": (
            "3,8-DIAMINO-6-PHENYL-5-[6-[1-[2-[(1,2,3,4-TETRAHYDRO-9-ACRIDINYL)AMINO]ETHYL]-1H-1,2,3-TRIAZOL-5-YL]HEXYL]-PHENANTHRIDINIUM",
            "non-polymer",
        ),
        "TZ4": (
            "3,8-DIAMINO-6-PHENYL-5-[6-[1-[2-[(1,2,3,4-TETRAHYDRO-9-ACRIDINYL)AMINO]ETHYL]-1H-1,2,3-TRIAZOL-4-YL]HEXYL]-PHENANTHRIDINIUM",
            "non-polymer",
        ),
        "XDP": ("D-XYLULOSE-2,2-DIOL-1,5-BISPHOSPHATE", "saccharide"),
        "VN4": ("OXIDO(DIOXO)VANADIUM", "non-polymer"),
        "VN3": ("VANADATE ION", "non-polymer"),
        "TOY": ("TOBRAMYCIN", "non-polymer"),
        "P8D": ("PHOSPHINATE PSEUDODIPEPTIDE L-ALA-D-ASP", "non-polymer"),
        "RGP": ("GAMMA-GLUTAMYL PHOSPHATE", "non-polymer"),
        "Q4A": (
            "N-(1-BENZYLPIPERIDIN-4-YL)-6,7-DIMETHOXY-2-(4-METHYL-1,4-DIAZEPAN-1-YL)QUINAZOLIN-4-AMINE",
            "non-polymer",
        ),
        "BHE": (
            "OCTYL 2-O-(6-DEOXY-ALPHA-L-GALACTOPYRANOSYL)-BETA-D-GALACTOPYRANOSIDE",
            "non-polymer",
        ),
        "T4P": (
            "3-{2,6,8-TRIOXO-9-[(2R,3R,4R)-2,3,4,5-TETRAHYDROXYPENTYL]-1,2,3,6,8,9-HEXAHYDRO-7H-PURIN-7-YL}PROPYL DIHYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "OAN": (
            "O-(2-ACETAMIDO-2-DEOXY D-GLUCOPYRANOSYLIDENE) AMINO-N-PHENYLCARBAMATE",
            "non-polymer",
        ),
        "E4P": ("ERYTHOSE-4-PHOSPHATE", "saccharide"),
        "E4B": ("ETHYL 4-HYDROXYBENZOATE", "non-polymer"),
        "TPW": (
            "2-{4-[(4-AMINO-2-METHYLPYRIMIDIN-5-YL)METHYL]-3-METHYLTHIOPHEN-2-YL}ETHYL TRIHYDROGEN DIPHOSPHATE",
            "non-polymer",
        ),
        "E4D": (
            "(2S,3R)-2-(4-(2-(PIPERIDIN-1-YL)ETHOXY)PHENYL)-2,3-DIHYDRO-3-(4-HYDROXYPHENYL)BENZO[B][1,4]OXATHIIN-6-OL",
            "non-polymer",
        ),
        "E4H": (
            "(3R,4S,5S,7R,9E,11R,12R)-12-ETHYL-4-HYDROXY-3,5,7,11-TETRAMETHYLOXACYCLODODEC-9-ENE-2,8-DIONE",
            "non-polymer",
        ),
        "3XC": (
            "6-{[(3R,4R)-4-(2-{[2,2-DIFLUORO-2-(4-FLUOROPHENYL)ETHYL]AMINO}ETHOXY)PYRROLIDIN-3-YL]METHYL}-4-METHYLPYRIDIN-2-AMINE",
            "non-polymer",
        ),
        "E4M": ("?", "non-polymer"),
        "RAF": ("RAFFINOSE", "saccharide"),
        "E41": (
            "N-HYDROXY-N^2^-(3-METHYLBUTYL)-N^2^-(NAPHTHALEN-2-YLSULFONYL)-D-VALINAMIDE",
            "non-polymer",
        ),
        "E46": (
            "N-(PIPERIDIN-4-YLMETHYL)-3-[3-(TRIFLUOROMETHYLOXY)PHENYL]-[1,2,3]TRIAZOLO[4,5-B]PYRIDIN-5-AMINE",
            "non-polymer",
        ),
        "LBP": (
            "(2S)-2-AMINO-4-(S-BUTYL-N-PHOSPHONOSULFONIMIDOYL)BUTANOIC ACID",
            "non-polymer",
        ),
        "E49": (
            "2-CHLORO-5-{[(4,6-DIMETHYLPYRIMIDIN-2-YL)SULFANYL]ACETYL}BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "LBS": ("6-O-BETA-D-FRUCTOFURANOSYL-BETA-D-FRUCTOFURANOSE", "non-polymer"),
        "P47": ("(2Z)-5-(4-CHLOROPHENYL)-3-PHENYLPENT-2-ENOIC ACID", "non-polymer"),
        "S": ("SULFUR ATOM", "non-polymer"),
        "RAE": ("2-DEOXY-ALPHA-RHAMNOSE", "saccharide"),
        "GLE": (
            "1-DECYL-3-TRIFLUORO ETHYL-SN-GLYCERO-2-PHOSPHOMETHANOL",
            "non-polymer",
        ),
        "UC1": (
            "2-METHYL-FURAN-3-CARBOTHIOIC ACID [4-CHLORO-3-(3-METHYL-BUT-2-ENYLOXY)-PHENYL]-AMIDE",
            "non-polymer",
        ),
        "RAB": (
            "2-(6-AMINO-PURIN-9-YL)-5-HYDROXYMETHYL-TETRAHYDRO-FURAN-3,4-DIOL",
            "non-polymer",
        ),
        "3XD": (
            "6-{[(3R,4R)-4-{2-[(2,2-DIFLUORO-2-PHENYLETHYL)AMINO]ETHOXY}PYRROLIDIN-3-YL]METHYL}-4-METHYLPYRIDIN-2-AMINE",
            "non-polymer",
        ),
        "RAC": (
            "4-(2-DIMETHYLCARBAMOYL-PHENYLCARBAMOYLOXYMETHYL)-BENZOIC ACID",
            "non-polymer",
        ),
        "LBW": ("?", "non-polymer"),
        "P43": (
            "1-{3-[1-({5-[(2-FLUOROPHENYL)ETHYNYL]FURAN-2-YL}CARBONYL)PIPERIDIN-4-YL]PHENYL}METHANAMINE",
            "non-polymer",
        ),
        "LBV": ("2(R),3(E)- PHYTOCHROMOBILIN", "non-polymer"),
        "RAN": (
            "PHOSPHORIC ACID MONO-[3,4,6-TRIHYDROXY-5-(1-HYDROXY-ETHYLAMINO)-TETRAHYDRO-PYRAN-2-YLMETHYL] ESTER",
            "non-polymer",
        ),
        "NMH": ("(R)-N-(1-METHYL-HEXYL)-FORMAMIDE", "non-polymer"),
        "RAO": ("1-O-METHYL-ALPHA-RHAMNOSE", "saccharide"),
        "P40": (
            "N-CYCLOPROPYL-4-METHYL-3-{2-[(2-MORPHOLIN-4-YLETHYL)AMINO]QUINAZOLIN-6-YL}BENZAMIDE",
            "non-polymer",
        ),
        "RAL": ("RALOXIFENE", "non-polymer"),
        "RAM": ("ALPHA-L-RHAMNOSE", "saccharide"),
        "UMQ": ("UNDECYL-MALTOSIDE", "non-polymer"),
        "UMP": ("DUMP", "non-polymer"),
        "UMS": ("2'-METHYLSELENYL-2'-DEOXYURIDINE-5'-PHOSPHATE", "nucleic-acid"),
        "FGP": ("2-AMINO-3-HYDROXY-3-PHOSPHONOOXY-PROPIONIC ACID", "peptide"),
        "UMX": (
            "[(1R,3R,4R,5S,7S)-3-(2,4-DIOXO-3,4-DIHYDROPYRIMIDIN-1(2H)-YL)-7-HYDROXY-5-METHYL-2-OXABICYCLO[2.2.1]HEPT-1-YL]METHYL DIHYDROGEN PHOSPHATE",
            "nucleic-acid",
        ),
        "UMG": ("METHYL-UMBELLIFERTL-N-ACETYL-CHITOTRIOSE", "non-polymer"),
        "2RL": (
            "N-(4-CHLOROPHENYL)-7-[(6,7-DIMETHOXYQUINOLIN-4-YL)OXY]-2,3-DIHYDRO-1,4-BENZOXAZINE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "UMA": ("URIDINE-5'-DIPHOSPHATE-N-ACETYLMURAMOYL-L-ALANINE", "peptide"),
        "FGR": (
            "N-(N-FORMYLGLYCYL)-5-O-PHOSPHONO-BETA-D-RIBOFURANOSYLAMINE",
            "non-polymer",
        ),
        "UMC": ("?", "non-polymer"),
        "FGS": ("5-(ACETYLAMINO)-5-DEOXY-1-O-PHOSPHONO-L-IDITOL", "non-polymer"),
        "BF3": (
            "2-(3-CHLORO-2-FLUOROBENZYL)-4,5-DIHYDROXY-1H-ISOINDOLE-1,3(2H)-DIONE",
            "non-polymer",
        ),
        "OLB": ("(2S)-2,3-DIHYDROXYPROPYL (9Z)-OCTADEC-9-ENOATE", "non-polymer"),
        "OLC": ("(2R)-2,3-DIHYDROXYPROPYL (9Z)-OCTADEC-9-ENOATE", "non-polymer"),
        "OLD": ("3-[2-(4-BROMOPHENYL)-2-OXOETHYL]-L-HISTIDINE", "peptide"),
        "OLE": ("2-HYDROXY-4-METHYL-PENTANOIC ACID", "peptide"),
        "OLF": (
            "4-({(2S)-2-[2-(4-CHLOROPHENYL)-5,6-DIFLUORO-1H-BENZIMIDAZOL-1-YL]-2-CYCLOHEXYLACETYL}AMINO)-3-METHYLBENZOIC ACID",
            "non-polymer",
        ),
        "02T": ("1-(3-METHYLTHIOPHEN-2-YL)METHANAMINE", "non-polymer"),
        "46U": (
            "(S)-N-(4-CARBAMIMIDOYLBENZYL)-1-(2-(CYCLOHEXYLOXY)ETHANOYL)PYRROLIDINE-2-CARBOXAMIDE",
            "peptide",
        ),
        "OLI": ("OLEIC ACID", "non-polymer"),
        "IKY": (
            "N-HYDROXY-2-[2-(TRIFLUOROMETHYL)PHENYL]-1,3-OXAZOLE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "AZO": (
            "METHYL (2Z)-2-(2-{[6-(2-CYANOPHENOXY)PYRIMIDIN-4-YL]OXY}PHENYL)-3-METHOXYACRYLATE",
            "non-polymer",
        ),
        "RAT": ("D-ARABINARIC ACID", "saccharide"),
        "C6P": (
            "N-({3-HYDROXY-2-METHYL-5-[(PHOSPHONOOXY)METHYL]PYRIDIN-4-YL}METHYL)-L-CYSTEINE",
            "non-polymer",
        ),
        "C6Q": (
            "N-{(1S,2S,3R)-1-[(ALPHA-D-GALACTOPYRANOSYLOXY)METHYL]-2,3-DIHYDROXYHEPTADECYL}-6-PHENYLHEXANAMIDE",
            "non-polymer",
        ),
        "46M": (
            "(4AR,6R,7S,8R,8AS)-HEXAHYDRO-6,7,8-TRIHYDROXY-2-METHYLPYRANO[3,2-D][1,3]DIOXINE-2-CARBOXYLIC ACID",
            "saccharide",
        ),
        "C6L": ("N-HEXANOYL-L-HOMOSERINE", "non-polymer"),
        "02V": ("(2S)-(3,4-DIHYDROXYPHENYL)(METHYLAMINO)ETHANOIC ACID", "peptide"),
        "OLT": ("O-METHYL-L-THREONINE", "peptide"),
        "OLU": (
            "2-(6-HYDROXY-1,3-BENZOTHIAZOL-2-YL)-1,3-THIAZOL-4(5H)-ONE",
            "non-polymer",
        ),
        "C6G": (
            "6-(CARBOXYMETHOXY)-9-(2-DEOXY-5-O-PHOSPHONO-BETA-D-ERYTHRO-PENTOFURANOSYL)-9H-PURIN-2-AMINE",
            "nucleic-acid",
        ),
        "OLZ": ("O-(2-AMINOETHYL)-L-SERINE", "peptide"),
        "46F": (
            "(2R)-4-(6-CHLOROPYRIDAZIN-3-YL)-N-(4-METHOXYBENZYL)-1-{[4-(TRIFLUOROMETHOXY)PHENYL]SULFONYL}PIPERAZINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "C6B": (
            "[(2S,3S,4R,5S)-5-(4-AMINO-2-OXO-PYRIMIDIN-1-YL)-4-HYDROXY-2-(HYDROXYMETHYL)OXOLAN-3-YL] DIHYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "C6C": ("S-CYCLOHEXYL THIOCYSTEINE", "peptide"),
        "46C": (
            "4-[(6-CHLORO-1-OXO-4-PHENYL-3-PROPANOYLISOQUINOLIN-2(1H)-YL)METHYL]BENZOIC ACID",
            "non-polymer",
        ),
        "C6A": (
            "N-[(2S,3R)-4-{[(2R,4S)-2-CYCLOPROPYL-6-(2,2-DIMETHYLPROPYL)-3,4-DIHYDRO-2H-CHROMEN-4-YL]AMINO}-1-(3,5-DIFLUOROPHENYL)-3-HYDROXYBUTAN-2-YL]ACETAMIDE",
            "non-polymer",
        ),
        "IK2": ("4'-DEOXY-4'-ACETYLYAMINO-PYRIDOXAL-5'-PHOSPHATE", "non-polymer"),
        "2RD": ("5-(DODECYLTHIO)-1H-1,2,3-TRIAZOLE-4-CARBOXYLIC ACID", "non-polymer"),
        "469": (
            "2-(6-CHLORO-5-{[(2R,5S)-4-(4-FLUOROBENZYL)-2,5-DIMETHYLPIPERAZIN-1-YL]CARBONYL}-1-METHYL-1H-INDOL-3-YL)-N,N-DIMETHYL-2-OXOACETAMIDE",
            "non-polymer",
        ),
        "02S": ("4-(BENZYLOXY)BENZOIC ACID", "non-polymer"),
        "C69": (
            "(2S)-2-(3-BROMOPHENYL)-3-(2,4-DIMETHOXYPHENYL)-1,3-THIAZOLIDIN-4-ONE",
            "non-polymer",
        ),
        "C66": ("2-AMINOETHYLLYSINE-CARBONYLMETHYLENE-CYTOSINE", "peptide"),
        "464": (
            "3-[5-({5-[(AMINOCARBONYL)AMINO]-2-OXO-2H-INDOL-3-YL}METHYL)-1H-PYRROL-3-YL]-N-(2-PIPERIDIN-1-YLETHYL)BENZAMIDE",
            "non-polymer",
        ),
        "C64": (
            "N-(4-BROMOPHENYL)-2-[2-(1,3-THIAZOL-2-YL)-1H-BENZIMIDAZOL-1-YL]ACETAMIDE",
            "non-polymer",
        ),
        "BF4": ("BERYLLIUM TETRAFLUORIDE ION", "non-polymer"),
        "C62": (
            "4-{5-[(Z)-(2,4-DIOXO-1,3-THIAZOLIDIN-5-YLIDENE)METHYL]FURAN-2-YL}BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "460": (
            "2-[5-(6-METHYLPYRIDIN-2-YL)-2,3-DIHYDRO-1H-PYRAZOL-4-YL]-1,5-NAPHTHYRIDINE",
            "non-polymer",
        ),
        "C60": (
            "[[[3-(2-METHYL-PROPANE-2-SULFONYL)-1-BENZENYL]-2-PROPYL]-CARBONYL-HISTIDYL]-AMINO-[CYCLOHEXYLMETHYL]-[2-HYDROXY-4-ISOPROPYL]-PENTAN-5-OIC ACID BUTYLAMIDE",
            "non-polymer",
        ),
        "C61": (
            "(2R,4S,5S,7S)-5-AMINO-N-BUTYL-4-HYDROXY-7-[4-METHOXY-3-(3-METHOXYPROPOXY)BENZYL]-2,8-DIMETHYLNONANAMIDE",
            "non-polymer",
        ),
        "OL1": (
            "7-PHENYL-1-(5-PYRIDIN-2-YL-1,3-OXAZOL-2-YL)HEPTANE-1,1-DIOL",
            "non-polymer",
        ),
        "OL2": (
            "4-[3-AMINO-6-(3,4,5-TRIMETHOXYPHENYL)PYRAZIN-2-YL]-2-ETHOXYBENZOIC ACID",
            "non-polymer",
        ),
        "MOX": (
            "(2R)-2-[(1R)-1-{[(2S)-2-CARBOXY-2-(4-HYDROXYPHENYL)ACETYL]AMINO}-1-METHOXY-2-OXOETHYL]-5-METHYLIDENE-5,6-DIHYDRO-2H-1,3-OXAZINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "AZF": ("AZAFAGOMINE", "non-polymer"),
        "TGA": ("METHANETHIOSULFONYL-GALACTOSIDE", "non-polymer"),
        "UM1": (
            "(2R)-2-[[(2R)-2-(2-AMINOETHANOYLAMINO)PROPANOYL]AMINO]PROPANOIC ACID",
            "peptide",
        ),
        "UM0": (
            "(2S)-2-[[(2R)-2-[(2S,3R,4R,5S,6R)-3-ACETAMIDO-2,5-DIHYDROXY-6-(HYDROXYMETHYL)OXAN-4-YL]OXYPROPANOYL]AMINO]PROPANOIC ACID",
            "non-polymer",
        ),
        "UM3": ("?", "non-polymer"),
        "TB9": (
            "4-CHLORO-8-METHYL-7-(3-METHYL-BUT-2-ENYL)-6,7,8,9-TETRAHYDRO-2H-2,7,9A-TRIAZA-BENZO[CD]AZULENE-1-THIONE",
            "non-polymer",
        ),
        "RAZ": (
            "(1R)-N,4-DIMETHYL-N-(4-METHYLPENT-3-ENYL)CYCLOHEX-3-ENAMINIUM",
            "non-polymer",
        ),
        "OKA": ("OKADAIC ACID", "non-polymer"),
        "FGA": ("GAMMA-D-GLUTAMIC ACID", "peptide"),
        "MOS": ("DIOXOTHIOMOLYBDENUM(VI) ION", "non-polymer"),
        "OLA": ("OLEIC ACID", "non-polymer"),
        "02Z": ("4-AMINO-2-(PHENYLAMINO)-1,3-THIAZOLE-5-CARBOXAMIDE", "non-polymer"),
        "TOK": (
            "(3ALPHA,8ALPHA)-17-(1H-BENZIMIDAZOL-1-YL)ANDROSTA-5,16-DIEN-3-OL",
            "non-polymer",
        ),
        "M73": (
            "METHYL (2S)-3-({[(2S,3R)-3-HYDROXY-4-{[(4-METHOXYPHENYL)SULFONYL][(2S)-2-METHYLBUTYL]AMINO}-1-PHENYLBUTAN-2-YL]CARBAMOYL}OXY)-2-METHYLPROPANOATE",
            "non-polymer",
        ),
        "M77": ("5-(1,4-DIAZEPAN-1-SULFONYL)ISOQUINOLINE", "non-polymer"),
        "M76": (
            "N-(4-FLUOROBENZYL)-2,3-DIHYDROXY-1-OXO-1,2-DIHYDROISOQUINOLINE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "SOA": ("ISATOIC ANHYDRIDE", "non-polymer"),
        "MBM": (
            "N-{N'-(2-METHYL-1-BENZOFURAN-5-YL)-N-[(3S)-2-OXO-1-(2-OXO-2-PYRROLIDIN-1-YLETHYL)AZEPAN-3-YL]CARBAMIMIDOYL}PYRIDINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "MBN": ("TOLUENE", "non-polymer"),
        "MBO": ("MERCURIBENZOIC ACID", "non-polymer"),
        "MBH": ("(1R)-1-BENZYL-1-(CARBOXYMETHYL)-1-METHYLDIAZANIUM", "non-polymer"),
        "JC1": (
            "(1R,3S,5Z)-5-[(2E)-2-[(1R,3AS,7AR)-1-[(2S,3S)-3-(2-HYDROXYETHYL)HEPTAN-2-YL]-7A-METHYL-2,3,3A,5,6,7-HEXAHYDRO-1H-INDEN-4-YLIDENE]ETHYLIDENE]-4-METHYLIDENE-CYCLOHEXANE-1,3-DIOL",
            "non-polymer",
        ),
        "SOG": (
            "2-HYDROXYMETHYL-6-OCTYLSULFANYL-TETRAHYDRO-PYRAN-3,4,5-TRIOL",
            "non-polymer",
        ),
        "GAX": (
            "1-{3-[(4-PYRIDIN-2-YLPIPERAZIN-1-YL)SULFONYL]PHENYL}-3-(1,3-THIAZOL-2-YL)UREA",
            "non-polymer",
        ),
        "MBD": ("3-METHYLCATECHOL", "non-polymer"),
        "GAV": ("GUANOSINE-5'-RP-ALPHA-THIO-TRIPHOSPHATE", "non-polymer"),
        "GAU": ("(4S)-4-AMINO-5-HYDROXYPENTANOIC ACID", "peptide"),
        "GAT": ("4'-AMINOPHENYL-ALPHA-D-GALACTOPYRANOSIDE", "non-polymer"),
        "GAS": ("?", "non-polymer"),
        "GAR": ("GLYCINAMIDE RIBONUCLEOTIDE", "non-polymer"),
        "395": (
            "S-3-FLUORO-4-[2-HYDROXY-2-(5,5,8,8-TETRAMETHYL-5,6,7,8,-TETRAHYDRO-NAPHTALEN-2-YL)-ACETYLAMINO]-BENZOIC ACID",
            "non-polymer",
        ),
        "GAP": ("GLYCYL-ADENOSINE-5'-PHOSPHATE", "non-polymer"),
        "GAO": ("?", "nucleic-acid"),
        "GAN": (
            "2-[3-BENZYL-5-(1-ALANYL-AMINOETHYL)-2,3,6,7-TETRAHYDRO-1H-AZEPIN-1-YL]-1-OXOPROPYL-VALINYL-VALINE-METHYLESTER",
            "non-polymer",
        ),
        "GAM": ("ALPHA-METHYL-L-GLUTAMIC ACID", "non-polymer"),
        "GAL": ("BETA-D-GALACTOSE", "saccharide"),
        "GAJ": ("N-TETRAZOL-5-YL 9-OXO-9H-XANTHENE-2 SULPHONAMIDE", "non-polymer"),
        "GAI": ("GUANIDINE", "non-polymer"),
        "GAH": (
            "N-{[(2S)-1-(N-{[4-({[AMINO(IMINO)METHYL]AMINO}METHYL)CYCLOHEXYL]CARBONYL}-3-CYCLOHEXYL-L-ALANYL)AZETIDIN-2-YL]CARBONYL}-L-TYROSYL-N~6~-[AMINO(IMINO)METHYL]-L-LYSINAMIDE",
            "peptide",
        ),
        "GAG": ("3-HYDROXYISOXAZOLE-4-CARBOXYLIC ACID", "non-polymer"),
        "GAF": ("2-DEOXY-2-FLUORO-ALPHA-D-GALACTOPYRANOSE", "non-polymer"),
        "GAE": ("D-GALACTARIC ACID", "non-polymer"),
        "GAD": ("2,6-ANHYDRO-3-DEOXY-D-ERYTHRO-HEX-2-ENONIC ACID", "saccharide"),
        "GAC": ("DIHYDRO-ACARBOSE", "saccharide"),
        "FG6": (
            "{2-[(1-BENZOFURAN-2-YLSULFONYL)CARBAMOYL]-5-METHOXY-1H-INDOL-1-YL}ACETIC ACID",
            "non-polymer",
        ),
        "GAA": ("METANITROPHENYL-ALPHA-D-GALACTOSIDE", "non-polymer"),
        "MBS": ("?", "non-polymer"),
        "FG7": ("ETHYL 3-AMINOBENZOATE", "non-polymer"),
        "GA9": (
            "3,3-BIS(3-BROMO-4-HYDROXYPHENYL)-7-CHLORO-1H,3H-BENZO[DE]ISOCHROMEN-1-ONE",
            "non-polymer",
        ),
        "GA4": ("GIBBERELLIN A4", "non-polymer"),
        "394": (
            "R-3-FLUORO-4-[2-HYDROXY-2-(5,5,8,8-TETRAMETHYL-5,6,7,8,-TETRAHYDRO-NAPHTALEN-2-YL)-ACETYLAMINO]-BENZOIC ACID",
            "non-polymer",
        ),
        "GA2": ("9-(1,3-DIHYDROXY-PROPOXYMETHANE)GUANINE", "non-polymer"),
        "FG1": (
            "N-[4-(2-{2-[3-(2-BROMO-ACETYLAMINO)-PROPIONYLAMINO]-3-HYDROXY-PROPIONYLAMINO}-ETHYL)-PHENYL]-OXALAMIC ACID",
            "non-polymer",
        ),
        "SO1": ("SORDARIN", "non-polymer"),
        "L1D": (
            "(2R,4S)-N^1^-(4-CHLOROPHENYL)-N^2^-[2-FLUORO-4-(2-OXOPYRIDIN-1(2H)-YL)PHENYL]-4-HYDROXY-4-PHENYLPYRROLIDINE-1,2-DICARBOXAMIDE",
            "non-polymer",
        ),
        "SO3": ("SULFITE ION", "non-polymer"),
        "PXX": ("N-1,10-PHENANTHROLIN-5-YLACETAMIDE", "non-polymer"),
        "MB9": (
            "(4R)-7-CHLORO-9-METHYL-1-OXO-1,2,4,9-TETRAHYDROSPIRO[BETA-CARBOLINE-3,4'-PIPERIDINE]-4-CARBONITRILE",
            "non-polymer",
        ),
        "JCB": ("N-(P-TOLUENESULFONYL)-INDOLE", "non-polymer"),
        "L1E": (
            "N~1~-(4-CHLOROPHENYL)-6-METHYL-N~5~-[3-(7H-PURIN-6-YL)PYRIDIN-2-YL]ISOQUINOLINE-1,5-DIAMINE",
            "non-polymer",
        ),
        "JCL": (
            "4-(6-CHLORO-2,4-DIOXO-1,2,3,4-TETRAHYDROPYRIMIDIN-5-YL) BUTYL PHOSPHATE",
            "non-polymer",
        ),
        "FG3": (
            "(5-METHOXY-2-{[(5-METHYLPYRIDIN-2-YL)SULFONYL]CARBAMOYL}-1H-INDOL-1-YL)ACETIC ACID",
            "non-polymer",
        ),
        "MB0": (
            "(4R)-7,8-DICHLORO-1',9-DIMETHYL-1-OXO-1,2,4,9-TETRAHYDROSPIRO[BETA-CARBOLINE-3,4'-PIPERIDINE]-4-CARBONITRILE",
            "non-polymer",
        ),
        "MB1": ("2-CHLORO-5-NITROBENZENESULFONAMIDE", "non-polymer"),
        "MB3": ("1-METHYLPYRROLIDIN-2-ONE", "non-polymer"),
        "M7S": (
            "(2S)-2-BENZYL-3-(4-PROPOXY-3-{[({4-[(3S,5S,7S)-TRICYCLO[3.3.1.1~3,7~]DEC-1-YL]PHENYL}CARBONYL)AMINO]METHYL}PHENYL)PROPANOIC ACID",
            "non-polymer",
        ),
        "025": ("3-(TRIFLUOROMETHYL)BENZENESULFONIC ACID", "non-polymer"),
        "L1G": (
            "N-(4-{4-AMINO-1-[4-(4-METHYLPIPERAZIN-1-YL)-TRANS-CYCLOHEXYL]-1H-PYRAZOLO[3,4-D]PYRIMIDIN-3-YL}-2-METHOXYPHENYL)-1-METHYL-1H-INDOLE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "G6Q": ("GLUCOSE-6-PHOSPHATE", "non-polymer"),
        "BFQ": ("IBANDRONATE", "non-polymer"),
        "L1A": (
            "N-(MORPHOLIN-4-YLSULFONYL)-L-PHENYLALANYL-3-(2-AMINO-1,3-THIAZOL-4-YL)-N-{(1R,2R,3S)-1-[(1R)-CYCLOHEX-3-EN-1-YLMETHYL]-2,3-DIHYDROXY-5-METHYLHEXYL}-L-ALANINAMIDE",
            "non-polymer",
        ),
        "M7B": ("7-O-PHOSPHONO-D-GLYCERO-BETA-D-MANNO-HEPTOPYRANOSE", "non-polymer"),
        "026": ("3-(PIPERAZIN-1-YL)PROPAN-1-AMINE", "non-polymer"),
        "M7G": ("7N-METHYL-8-HYDROGUANOSINE-5'-DIPHOSPHATE", "non-polymer"),
        "RA2": (
            "1-[4-CARBOXY-2-(3-PENTYLAMINO)PHENYL]-5,5'-DI(HYDROXYMETHYL)PYRROLIDIN-2-ONE",
            "non-polymer",
        ),
        "L1C": (
            "(2R,4S)-N^1^-(4-CHLOROPHENYL)-4-(2,4-DIFLUOROPHENYL)-4-HYDROXY-N^2^-(2-OXO-2H-1,3'-BIPYRIDIN-6'-YL)PYRROLIDINE-1,2-DICARBOXAMIDE",
            "non-polymer",
        ),
        "M7M": ("?", "non-polymer"),
        "BV3": (
            "1,3-BIS-([[3-(4-{3-[3-NITRO-5-(GALACTOPYRANOSYLOXY)-BENZOYLAMINO]-PROPYL}-PIPERAZIN-1-YL)-PROPYLAMINO-3,4-DIOXO-CYCLOBUTENYL]-AMINO-ETHYL]-AMINO-CARBONYLOXY)-2-AMINO-PROPANE",
            "non-polymer",
        ),
        "BV2": (
            "1,3-BIS-([3-(4-{3-[3-NITRO-5-(GALACTOPYRANOSYLOXY)-BENZOYLAMINO]-PROPYL}-PIPERAZIN-1-YL)-PROPYL-AMINO]-CARBONYLOXY)-2-AMINO-PROPANE",
            "non-polymer",
        ),
        "BV1": ("BV1", "non-polymer"),
        "PXP": ("PYRIDOXINE-5'-PHOSPHATE", "non-polymer"),
        "023": (
            "N^2^-[(2R)-2-{(1S)-1-[FORMYL(HYDROXY)AMINO]ETHYL}-5-PHENYLPENTANOYL]-N,3-DIMETHYL-L-VALINAMIDE",
            "non-polymer",
        ),
        "BV4": (
            "1,3-BIS-([3-[3-[3-(4-{3-[3-NITRO-5-(GALACTOPYRANOSYLOXY)-BENZOYLAMINO]-PROPYL}-PIPERAZIN-1-YL)-PROPYLAMINO-3,4-DIOXO-CYCLOBUTENYL]-AMINO-PROPOXY-ETHOXY-ETHOXY]-PROPYL-]AMINO-CARBONYLOXY)-2-AMINO-PROPANE",
            "non-polymer",
        ),
        "K57": (
            "(4R)-N-TERT-BUTYL-3-{(2S,3S)-2-HYDROXY-3-[(3-HYDROXY-2-METHYLBENZOYL)AMINO]-4-PHENYLBUTANOYL}-5,5-DIMETHYL-1,3-THIAZOLIDINE-4-CARBOXAMIDE",
            "peptide",
        ),
        "K55": (
            "(2S)-2-{4-BUTOXY-3-[({[2-FLUORO-4-(TRIFLUOROMETHYL)PHENYL]CARBONYL}AMINO)METHYL]BENZYL}BUTANOIC ACID",
            "non-polymer",
        ),
        "022": (
            "3-{1-(4-CARBAMOYL-2-METHYLPHENYL)-5-[4-(1H-IMIDAZOL-1-YL)PHENYL]-1H-PYRROL-2-YL}PROPANOIC ACID",
            "non-polymer",
        ),
        "K53": (
            "(5S)-N-{(1S,2R)-3-[(1,3-BENZODIOXOL-5-YLSULFONYL)(2-METHYLPROPYL)AMINO]-1-BENZYL-2-HYDROXYPROPYL}-3-(2-HYDROXYPHENYL)-2-OXO-1,3-OXAZOLIDINE-5-CARBOXAMIDE",
            "non-polymer",
        ),
        "KME": (
            "(1E)-PROP-1-EN-1-YL 3-DEOXY-7-O-METHYL-ALPHA-D-MANNO-OCT-2-ULOPYRANOSIDONIC ACID",
            "non-polymer",
        ),
        "K51": (
            "4-(6-{[(1S)-1-(HYDROXYMETHYL)-2-METHYLPROPYL]AMINO}IMIDAZO[1,2-B]PYRIDAZIN-3-YL)BENZONITRILE",
            "non-polymer",
        ),
        "C0E": (
            "3-FLUORO-6-(4-FLUOROPHENYL)-2-HYDROXY-6-OXOHEXA-2,4-DIENOIC ACID",
            "non-polymer",
        ),
        "QN7": (
            "(8S)-3,3-DIFLUORO-8-(2'-FLUORO-3'-METHOXYBIPHENYL-3-YL)-8-PYRIDIN-4-YL-2,3,4,8-TETRAHYDROIMIDAZO[1,5-A]PYRIMIDIN-6-AMINE",
            "non-polymer",
        ),
        "VG5": (
            " N-[(1S,2R)-1-BENZYL-3-{[(1S)-2-(CYCLOHEXYLAMINO)-1-METHYL-2-OXOETHYL]AMINO}-2-HYDROXYPROPYL]-3-(ETHYLAMINO)-5-(2-OXOPYRROLIDIN-1-YL)BENZAMIDE",
            "non-polymer",
        ),
        "PE9": ("PENTAN-1-OL", "non-polymer"),
        "QN1": (
            "4,4'-[(3S,4S)-PYRROLIDINE-3,4-DIYLBIS({[4-(TRIFLUOROMETHYL)BENZYL]IMINO}SULFONYL)]DIBENZAMIDE",
            "non-polymer",
        ),
        "NX5": ("2-[(3,4-DICHLOROBENZYL)SULFANYL]BENZOIC ACID", "non-polymer"),
        "NX4": (
            "5-[({3-[(3,4-DICHLOROBENZYL)SULFANYL]THIOPHEN-2-YL}CARBONYL)SULFAMOYL]-2-METHOXYBENZOIC ACID",
            "non-polymer",
        ),
        "NX6": ("N-[(BENZYLOXY)CARBONYL]-L-ASPARTIC ACID", "non-polymer"),
        "NX1": (
            "3-[(3-BROMO-4-CHLOROBENZYL)SULFANYL]THIOPHENE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "QN2": (
            "N,N'-(3S,4S)-PYRROLIDINE-3,4-DIYLBIS[N-(4-IODOBENZYL)BENZENESULFONAMIDE]",
            "non-polymer",
        ),
        "NX3": (
            "N-[(3-AMINOPHENYL)SULFONYL]-3-[(3,4-DICHLOROBENZYL)SULFANYL]THIOPHENE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "FP4": (
            "3-(4-{[2-(PYRIMIDIN-2-YL)FURO[2,3-C]PYRIDIN-3-YL]AMINO}-1H-INDAZOL-3-YL)PROPAN-1-OL",
            "non-polymer",
        ),
        "KHP": (
            "2-HYDROXYMETHYL-5-(4-NITRO-PHENOXY)-TETRAHYDRO-FURAN-3,4-DIOL",
            "non-polymer",
        ),
        "QID": (
            "3-HYDROXY-6-(PHENYLSULFONYL)QUINAZOLINE-2,4(1H,3H)-DIONE",
            "non-polymer",
        ),
        "QIG": (
            "N-1H-INDAZOL-5-YL-2-(6-METHYLPYRIDIN-2-YL)QUINAZOLIN-4-AMINE",
            "non-polymer",
        ),
        "PXI": (
            "4-{[4-(DIMETHYLAMINO)-3-HYDROXY-6-METHYLTETRAHYDRO-2H-PYRAN-2-YL]OXY}-12-ETHYL-3,5,7,11-TETRAMETHYLOXACYCLODODEC-9-ENE-2,8-DIONE",
            "non-polymer",
        ),
        "QIC": (
            "(1S,3R,4S,5R)-1,3,4,5-TETRAHYDROXYCYCLOHEXANECARBOXYLIC ACID",
            "non-polymer",
        ),
        "028": ("(1-HYDROXYHEPTANE-1,1-DIYL)BIS(PHOSPHONIC ACID)", "non-polymer"),
        "289": ("D-GLYCERO-ALPHA-D-MANNO-HEPTOPYRANOSE", "saccharide"),
        "288": (
            "5-CHLORO-1H-INDOLE-2-CARBOXYLIC ACID{[CYCLOPENTYL-(2-HYDROXY-ETHYL)-CARBAMOYL]-METHYL}-AMIDE",
            "non-polymer",
        ),
        "QIX": (
            "N~2~-[(2R)-4-(HYDROXYAMINO)-2-(2-NAPHTHYLMETHYL)-4-OXOBUTANOYL]-L-ARGINYLGLYCYLGLYCINAMIDE",
            "non-polymer",
        ),
        "KHO": (
            "4-(3,4-DIHYDROXY-5-HYDROXYMETHYL-TETRAHYDRO-FURAN-2-YLOXY)-TETRAHYDRO-PYRAN-2,3,5-TRIOL",
            "non-polymer",
        ),
        "281": (
            "(2S)-2-[3-(AMINOMETHYL)PHENYL]-3-[(S)-{(1R)-1-[(2,1,3-BENZOTHIADIAZOL-4-YLSULFONYL)AMINO]-2-METHYLPROPYL}(HYDROXY)PHOSPHORYL]PROPANOIC ACID",
            "non-polymer",
        ),
        "280": ("1-METHYL-1H-IMIDAZOLE-5-CARBALDEHYDE", "non-polymer"),
        "283": (
            "(3S)-1-{[4-(BUT-2-YN-1-YLOXY)PHENYL]SULFONYL}PYRROLIDINE-3-THIOL",
            "non-polymer",
        ),
        "282": ("3-METHOXYPYRIDINE", "non-polymer"),
        "285": ("(2R,4R,5R)-2,4,5-TRIMETHYL-1,3-OXAZOLIDINE", "non-polymer"),
        "284": ("(4S)-4-METHYL-1,4,5,6-TETRAHYDROPYRIDIN-2-AMINE", "non-polymer"),
        "287": (
            "4-{[5-(METHOXYCARBAMOYL)-2-METHYLPHENYL]AMINO}-5-METHYL-N-[(1S)-1-PHENYLETHYL]PYRROLO[2,1-F][1,2,4]TRIAZINE-6-CARBOXAMIDE",
            "non-polymer",
        ),
        "286": ("2-ETHENYL-1-METHYLPYRIDINIUM", "non-polymer"),
        "NXW": (
            "2-[(3,4-DICHLOROBENZYL)SULFANYL]-4-{[3-({N-[2-(METHYLAMINO)ETHYL]GLYCYL}AMINO)PHENYL]ETHYNYL}BENZOIC ACID",
            "non-polymer",
        ),
        "NXV": (
            "2-[(3,4-DICHLOROBENZYL)SULFANYL]-4-(4-HYDROXYBUT-1-YN-1-YL)BENZOIC ACID",
            "non-polymer",
        ),
        "NXP": (
            "N-{4-[(1E)-N-CARBAMIMIDOYLETHANEHYDRAZONOYL]PHENYL}-1H-INDOLE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "28N": (
            "8-BROMO-4H-[1,2,4]OXADIAZOLO[3,4-C][1,4]BENZOXAZIN-1-ONE",
            "non-polymer",
        ),
        "KH1": (
            "5-(2-{1-[1-(4-ETHYL-4-HYDROXY-HEXYLOXY)-ETHYL]-7A-METHYL-OCTAHYDRO-INDEN-4-YLIDENE}-ETHYLIDENE)-4-METHYLENE-CYCLOHEXANE-1,3-DIOL",
            "non-polymer",
        ),
        "KH3": (
            "5'-{[(3S)-3-AMINO-3-CARBOXYPROPYL][2-(HEXYLAMINO)ETHYL]AMINO}-5'-DEOXYADENOSINE",
            "non-polymer",
        ),
        "NXY": (
            "3-[(3,4-DICHLOROBENZYL)SULFANYL]THIOPHENE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "4OX": ("4-OXODECANEDIOIC ACID", "non-polymer"),
        "NXZ": (
            "1-[(3S,4S)-4-AMINO-1-(6-PHENYLPYRIMIDIN-4-YL)PYRROLIDIN-3-YL]PIPERIDIN-2-ONE",
            "non-polymer",
        ),
        "NXD": (
            "METHYL 5-(ACETYLAMINO)-9-{[AMINO(OXO)ACETYL]AMINO}-3,5,9-TRIDEOXY-D-GLYCERO-ALPHA-D-GLUCO-NON-2-ULOPYRANOSIDONIC ACID",
            "saccharide",
        ),
        "NXG": (
            "2-{[4-(DIETHYLAMINO)PIPERIDIN-1-YL]METHYL}-4-(3-HYDROXYPROP-1-YN-1-YL)-6-IODOPHENOL",
            "non-polymer",
        ),
        "NXA": ("N-CARBOXYALANINE", "non-polymer"),
        "NXC": (
            "(2,6-BIS[(DIMETHYLAMINO-KAPPAN)METHYL]-4-{3-[(S)-ETHOXY(4-NITROPHENOXY)PHOSPHORYL]PROPYL}PHENYL-KAPPAC~1~)(CHLORO)PLATINUM(2+)",
            "non-polymer",
        ),
        "PXB": ("PARECOXIB", "non-polymer"),
        "PY7": ("PYRIDIN-4-YLMETHANOL", "non-polymer"),
        "28P": ("N-[3-(4-BENZYLPHENOXY)PROPYL]-N-METHYL-BETA-ALANINE", "non-polymer"),
        "28S": ("6-BROMO-N-(PROP-2-EN-1-YL)QUINAZOLIN-4-AMINE", "non-polymer"),
        "NXN": (
            "2-(2,4-DICHLORO-PHENOXY)-N-(2-MERCAPTO-ETHYL)-ACETAMIDE",
            "non-polymer",
        ),
        "QI1": ("4-[BIS(4-CHLOROBENZYL)AMINO]-N-HYDROXYBUTANAMIDE", "non-polymer"),
        "LI6": ("3,4-DIHYDROXY-1-METHYLQUINOLIN-2(1H)-ONE", "non-polymer"),
        "QI3": (
            "(4R)-4-(4-CHLOROPHENOXY)-1-[(4-CHLOROPHENYL)SULFONYL]-N-HYDROXY-L-PROLINAMIDE",
            "non-polymer",
        ),
        "13F": (
            "3-METHYL-2-OXO-N-PHENYL-1,2,3,4-TETRAHYDROQUINAZOLINE-6-SULFONAMIDE",
            "non-polymer",
        ),
        "BVP": (
            "(E)-5-(2-BROMOVINYL)-2'-DEOXYURIDINE-5'-MONOPHOSPHATE",
            "nucleic-acid",
        ),
        "ZIP": ("N-(3-METHYLBUT-2-EN-1-YL)-9H-PURIN-6-AMINE", "non-polymer"),
        "FP9": ("(4R)-4-FLUORO-L-PROLINE", "peptide"),
        "TGN": (
            "3-(HYDROXY-PHENYL-PHOSPHINOYLOXY)-8-METHYL-8-AZA-BICYCLO[3.2.1]OCTANE-2-CARBOXYLIC ACID METHYL ESTER",
            "non-polymer",
        ),
        "BVC": ("1-THIEN-3-YLMETHANAMINE", "non-polymer"),
        "BVA": ("TRIHYDROXY[(N-HYDROXYBENZAMIDATO)OXO]VANADATE", "non-polymer"),
        "ETA": ("ETHANOLAMINE", "peptide"),
        "BVG": ("1H-IMIDAZOL-2-YLMETHANOL", "non-polymer"),
        "BVF": ("4-METHYLPYRIDIN-2-AMINE", "non-polymer"),
        "BVD": ("5-BROMOVINYLDEOXYURIDINE", "non-polymer"),
        "K5E": (
            "(4S,5R)-5-HYDROXY-4-(3-HYDROXYPHENYL)-3,4,5,6,7,8-HEXAHYDROQUINAZOLINE-2(1H)-THIONE",
            "non-polymer",
        ),
        "OKH": (
            "2-(4-CHLOROPHENYL)-1-[(1S)-1-CYCLOHEXYL-2-(CYCLOHEXYLAMINO)-2-OXOETHYL]-1H-BENZIMIDAZOLE-6-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "4OP": ("(2S,4S)-2-CARBOXY-4-HYDROXY-1,1-DIMETHYLPYRROLIDINIUM", "peptide"),
        "BVL": (
            "(2E)-1-[2-HYDROXY-4-METHOXY-5-(3-METHYLBUT-2-EN-1-YL)PHENYL]-3-(4-HYDROXYPHENYL)PROP-2-EN-1-ONE",
            "non-polymer",
        ),
        "13H": ("(9Z,11E,13S)-13-HYDROXYOCTADECA-9,11-DIENOIC ACID", "non-polymer"),
        "KMP": ("3,5,7-TRIHYDROXY-2-(4-HYDROXYPHENYL)-4H-CHROMEN-4-ONE", "non-polymer"),
        "PX8": ("1,2-DISTEAROYL-SN-GLYCERO-3-PHOSPHATE", "non-polymer"),
        "R16": ("HEXADECANE", "non-polymer"),
        "4OJ": ("(2-METHYLPHENYL) DIHYDROGEN PHOSPHATE", "non-polymer"),
        "R14": (
            "N,N'-[ACRIDINE-3,6-DIYLBIS(1H-1,2,3-TRIAZOLE-1,4-DIYLBENZENE-3,1-DIYL)]BIS[3-(DIETHYLAMINO)PROPANAMIDE]",
            "non-polymer",
        ),
        "R15": (
            "N,N'-(HEPTANE-1,7-DIYLDICARBAMOYL)BIS(3-CHLOROBENZENESULFONAMIDE)",
            "non-polymer",
        ),
        "R12": (
            "(2E,4E,6E,8E)-9-(4-HYDROXY-2,3,6-TRIMETHYLPHENYL)-3,7-DIMETHYLNONA-2,4,6,8-TETRAENOIC ACID",
            "non-polymer",
        ),
        "R13": (
            "3-METHYL-7-(5,5,8,8-TETRAMETHYL-5,6,7,8-TETRAHYDRO-NAPHTHALEN-2-YL) -OCTA-2,4,6-TRIENOIC ACID",
            "non-polymer",
        ),
        "R10": ("5-O-PHOSPHONO-D-RIBONIC ACID", "non-polymer"),
        "R11": (
            "4-{[1-METHYL-5-(2-METHYL-BENZOIMIDAZOL-1-YLMETHYL)-1H-BENZOIMIDAZOL-2-YLMETHYL]-AMINO}-BENZAMIDINE",
            "non-polymer",
        ),
        "OSP": ("SULTHIAME", "non-polymer"),
        "OSS": (
            "6-(HYDROXYETHYLDITHIO)-8-(AMINOMETHYLTHIO)OCTANOIC ACID",
            "non-polymer",
        ),
        "Z19": ("1H-INDAZOLE-3-CARBOHYDRAZIDE", "non-polymer"),
        "OSU": ("N-OCTANOYLSUCROSE", "non-polymer"),
        "13Q": (
            "[(3R,5R,6S)-5-(3-CHLOROPHENYL)-6-(4-CHLOROPHENYL)-1-(CYCLOPROPYLMETHYL)-2-OXOPIPERIDIN-3-YL]ACETIC ACID",
            "non-polymer",
        ),
        "R18": ("(17BETA)-17-HYDROXY-17-METHYLESTRA-4,9,11-TRIEN-3-ONE", "non-polymer"),
        "OSV": ("RUTHENIUM OCTASPORINE 4", "non-polymer"),
        "PX3": (
            "5-[[[2-[(4-METHOXYPHENYL)METHYLCARBAMOYL]PHENYL]METHYL-PROP-2-ENYL-AMINO]METHYL]-1,3-BENZODIOXOLE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "OSM": ("1-(OXIDOSULFANYL)METHANAMINE", "non-polymer"),
        "OSL": ("(2R)-2-HYDROXY-3-(SULFOOXY)PROPANOIC ACID", "non-polymer"),
        "OKJ": ("(2S,3S)-3-AMINO-2-HYDROXY-4-(4-HYDROXYPHENYL)BUTANAL", "non-polymer"),
        "OSA": (
            "7-AMINO-4-HYDROXY-3-[(E)-(5-HYDROXY-7-SULFONAPHTHALEN-2-YL)DIAZENYL]NAPHTHALENE-2-SULFONIC ACID",
            "non-polymer",
        ),
        "OSC": (
            "(2R,3AS,6R,7AS)-N-(2-{1-[AMINO(IMINO)METHYL]-2,5-DIHYDRO-1H-PYRROL-3-YL}ETHYL)-6-HYDROXY-1-{N-[(2S)-2-HYDROXY-3-PHENYLPROPANOYL]PHENYLALANYL}OCTAHYDRO-1H-INDOLE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "OSB": ("2-SUCCINYLBENZOATE", "non-polymer"),
        "OSE": ("O-SULFO-L-SERINE", "peptide"),
        "PX1": (
            "(1Z)-7-AMINO-1-(HYDROXYMETHYLENE)-2-IMINO-4,6-DIMETHYL-3-OXO-2,3-DIHYDRO-1H-PHENOXAZINE-9-CARBALDEHYDE",
            "non-polymer",
        ),
        "OSF": ("OCTYL SULFATE", "non-polymer"),
        "R1T": (
            "2-{4-[(4-AMINO-2-METHYLPYRIMIDIN-5-YL)METHYL]-5-[(1R)-1-HYDROXYETHYL]-3-METHYL-2-THIENYL}ETHYL TRIHYDROGEN DIPHOSPHATE",
            "non-polymer",
        ),
        "R1P": ("RIBOSE-1-PHOSPHATE", "saccharide"),
        "OS1": ("RUTHENIUM OCTASPORINE", "non-polymer"),
        "EWB": (
            "[(1S)-1-{[(2-BENZYLPHENYL)CARBONYL]AMINO}ETHYL](TRIHYDROXY)BORATE(1-)",
            "non-polymer",
        ),
        "EWA": (
            "[(1S)-1-{[(2,6-DIMETHOXYPHENYL)CARBONYL]AMINO}ETHYL]BORONIC ACID",
            "non-polymer",
        ),
        "R1Z": ("DELTA-[RU(BPY)2DPPZ]2+", "non-polymer"),
        "OS4": ("OSMIUM 4+ ION", "non-polymer"),
        "R1X": ("1,4-ANHYDRO-D-RIBITOL", "saccharide"),
        "PX6": ("1,2-DIPALMITOYL-SN-GLYCERO-3-PHOSPHATE", "non-polymer"),
        "R1F": (
            "S-[(1-OXYL-2,2,5,5-TETRAMETHYL-4-PHENYL-2,5-DIHYDRO-1H-PYRROL-3-YL)METHYL] METHANESULFONOTHIOATE",
            "non-polymer",
        ),
        "R1B": (
            "S-[(1-OXYL-2,2,4,5,5-PENTAMETHYL-2,5-DIHYDRO-1H-PYRROL-3-YL)METHYL] METHANESULFONOTHIOATE",
            "non-polymer",
        ),
        "R1C": (
            "BIS(2,2'-BIPYRIDINE-KAPPA~2~N~1~,N~1'~)[CHRYSENE-5,6-DIIMINATO(2-)-KAPPA~2~N,N']RHODIUM(4+)",
            "non-polymer",
        ),
        "13Z": (
            "2-[(2-{(2S,4S)-2-[(R)-(5-TERT-BUTYL-1,3,4-OXADIAZOL-2-YL)(HYDROXY)METHYL]-4-FLUOROPYRROLIDIN-1-YL}-2-OXOETHYL)AMINO]-2-METHYLPROPAN-1-OL",
            "non-polymer",
        ),
        "EWQ": ("6,8-DIMETHOXY-1,4-DIMETHYLQUINOLIN-2(1H)-ONE", "non-polymer"),
        "FPU": (
            "1-{5-[(6,7-DIMETHOXYQUINOLIN-4-YL)OXY]PYRIDIN-2-YL}-3-[(1R,2S)-2-PHENYLCYCLOPROPYL]UREA",
            "non-polymer",
        ),
        "13X": ("BENZENE-1,3,5-TRIOL", "non-polymer"),
        "LRG": ("(2S)-2-(BIPHENYL-4-YLOXY)-3-PHENYLPROPANOIC ACID", "non-polymer"),
        "FPT": (
            "N-[[3-FLUORO-4-ETHOXY-PYRID-2-YL]ETHYL]-N'-[5-NITRILOMETHYL-PYRIDYL]-THIOUREA",
            "non-polymer",
        ),
        "LRH": ("6-DEOXY-BETA-L-FRUCTOFURANOSE", "non-polymer"),
        "743": (
            "(1R,5R)-5-{(1Z)-N-[(4'-FLUOROBIPHENYL-4-YL)METHOXY]BUTANIMIDOYL}-2,2-DIMETHYL-4,6-DIOXOCYCLOHEXANECARBONITRILE",
            "non-polymer",
        ),
        "LRU": ("?", "non-polymer"),
        "S12": (
            "O-[(S)-HYDROXY{[(2S)-2-HYDROXY-3-(OCTADEC-9-ENOYLOXY)PROPYL]OXY}PHOSPHORYL]-L-SERINE",
            "peptide",
        ),
        "VRX": (
            "(2S)-({(5Z)-5-[(5-ETHYL-2-FURYL)METHYLENE]-4-OXO-4,5-DIHYDRO-1,3-THIAZOL-2-YL}AMINO)(4-FLUOROPHENYL)ACETIC ACID",
            "non-polymer",
        ),
        "VGM": (
            "4-[2-AMINO-5-(4-HYDROXY-3-METHOXYPHENYL)PYRIDIN-3-YL]BENZAMIDE",
            "non-polymer",
        ),
        "FPR": ("(3S)-3-AMINO-4-OXO-7-PHENYLHEPTANOIC ACID", "peptide"),
        "C01": (
            "(2S)-(4-ISOPROPYLPHENYL)[(2-METHYL-3-OXO-5,7-DIPROPYL-2,3-DIHYDRO-1,2-BENZISOXAZOL-6-YL)OXY]ACETATE",
            "non-polymer",
        ),
        "F89": (
            "S)-2-(5(((1,2-DIHYDRO-3-METHYL-1-OXOBENZO(F)QUINAZOLIN-9-YL)METHYL)AMINO)1-OXO-2-ISOINDOLINYL)GLUTARIC ACID",
            "non-polymer",
        ),
        "S14": (
            "6-(4-{(1S,2S)-2-AMINO-1-[(DIMETHYLAMINO)CARBONYL]-3-[(3S)-3-FLUOROPYRROLIDIN-1-YL]-3-OXOPROPYL}PHENYL)-1H-[1,2,4]TRIAZOLO[1,5-A]PYRIDIN-4-IUM",
            "non-polymer",
        ),
        "RGG": ("(2R)-2,3-DIHYDROXYPROPYL BETA-D-GALACTOPYRANOSIDE", "saccharide"),
        "485": (
            "[2-AMINO-3-(4-HYDROXY-PHENYL)-PROPIONYLAMINO]- (3,4,5-TRIHYDROXY-6-METHYL-TETRAHYDRO-PYRAN-2-YL)- ACETIC ACID",
            "non-polymer",
        ),
        "IB": (
            "2'-(3-IODO-4-METHOXYPHENYL)-5-(4-METHYL-1-PIPERAZINYL)-2,5'-BI-BENZIMIDAZOLE",
            "non-polymer",
        ),
        "OCL": ("COBALT ION, 1 WATER COORDINATED", "non-polymer"),
        "TG0": (
            "(2S,4R)-1-(3-AMINO-2-METHYLBENZOYL)-4-HYDROXY-N-(4-(4-METHYLTHIAZOL-5-YL)BENZYL)PYRROLIDINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "D2X": (
            "3-[(4-HYDROXY-2-METHYLPYRIMIDIN-5-YL)METHYL]-5-(2-{[(R)-HYDROXY(PHOSPHONOOXY)PHOSPHORYL]OXY}ETHYL)-4-METHYL-1,3-THIAZOL-3-IUM",
            "non-polymer",
        ),
        "56Z": (
            "4-{[4-AMINO-5-(4-METHOXYBENZOYL)-1,3-THIAZOL-2-YL]AMINO}BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "484": (
            "3-(5'-{2-[3,4-BIS(HYDROXYMETHYL)PHENYL]ETHYL}-2'-METHYL-2-PROPYLBIPHENYL-4-YL)PENTAN-3-OL",
            "non-polymer",
        ),
        "CU3": ("COPPER (III) ION", "non-polymer"),
        "CU1": ("COPPER (I) ION", "non-polymer"),
        "D2Q": (
            "ETHYL 5-{2-[(2,4-DIAMINO-5-METHYLPYRIDO[2,3-D]PYRIMIDIN-6-YL)METHYL]-4-METHOXYPHENOXY}PENTANOATE",
            "non-polymer",
        ),
        "D2P": ("(2S)-2,3-DICHLOROPROPAN-1-OL", "non-polymer"),
        "D2S": ("4,4'-DISULFANEDIYLDIBENZENE-1,3-DIOL", "non-polymer"),
        "D2R": (
            "6-{2-[(2,4-DIAMINO-5-METHYLPYRIDO[2,3-D]PYRIMIDIN-6-YL)METHYL]-4-METHOXYPHENYL}HEX-5-YNOIC ACID",
            "non-polymer",
        ),
        "CU9": (
            "6,6',7',12'-TETRAMETHOXY-2,2,2',2'-TETRAMETHYLTUBOCURARAN-2,2'-DIIUM",
            "non-polymer",
        ),
        "D2V": (
            "(3S,5Z,7E,22E)-9,10-SECOERGOSTA-5,7,10,22-TETRAEN-3-OL",
            "non-polymer",
        ),
        "D2H": (
            "5-[(1E,3R)-2-(2-METHOXYPHENYL)-3-METHYLPENT-1-EN-1-YL]FURO[2,3-D]PYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "D2K": (
            "ETHYL 4-{3-[(2,4-DIAMINO-5-METHYLPYRIDO[2,3-D]PYRIMIDIN-6-YL)METHYL]-4-METHOXYPHENOXY}BUTANOATE",
            "non-polymer",
        ),
        "D2J": (
            "6-{3-[(2,4-DIAMINO-5-METHYLPYRIDO[2,3-D]PYRIMIDIN-6-YL)METHYL]-4-METHOXYPHENOXY}HEXANOIC ACID",
            "non-polymer",
        ),
        "D2M": ("4-(2,6-DIMETHOXYPHENOXY)-2,6-DIMETHOXYPHENOL", "non-polymer"),
        "D2O": (
            "6-[2-METHOXY-5-(2-PHENYLETHOXY)BENZYL]-5-METHYLPYRIDO[2,3-D]PYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "D2N": (
            "ETHYL 6-{3-[(2,4-DIAMINO-5-METHYLPYRIDO[2,3-D]PYRIMIDIN-6-YL)METHYL]-4-METHOXYPHENOXY}HEXANOATE",
            "non-polymer",
        ),
        "D2A": (
            "2-(4'-AMIDINOBIPHENYL-4-YL)-1H-BENZIMIDAZOLE-5-AMIDINE",
            "non-polymer",
        ),
        "56A": ("3-(8-PHENYLOCTYL)-L-HISTIDINE", "peptide"),
        "D2C": (
            "(2S,4S,4AR,5AS,6S,11R,11AS,12R,12AR)-7-CHLORO-4-(DIMETHYLAMINO)-6,10,11,12-TETRAHYDROXY-1,3-DIOXO-1,2,3,4,4A,5,5A,6,11,11A,12,12A-DODECAHYDROTETRACENE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "D2B": (
            "5-[(1Z)-2-(2-METHOXYPHENYL)-3-METHYLBUT-1-EN-1-YL]FURO[2,3-D]PYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "D2E": (
            "5-[(E)-2-CYCLOPROPYL-2-(2-METHOXYPHENYL)ETHENYL]FURO[2,3-D]PYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "D2D": (
            "5-[(1E)-2-(2-METHOXYPHENYL)-4-METHYLPENT-1-EN-1-YL]FURO[2,3-D]PYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "D2G": ("N-METHYL(5-(PYRIDIN-3-YL)FURAN-2-YL)METHANAMINE", "non-polymer"),
        "D2F": (
            "5-[(1E)-2-(2-METHOXYPHENYL)HEX-1-EN-1-YL]FURO[2,3-D]PYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "568": ("N,N-[2,5-O-DIBENZYL-GLUCARYL]-DI-[1-AMINO-INDAN-2-OL]", "non-polymer"),
        "569": (
            "(5R)-2-AMINO-5-(4-FLUORO-3-PYRIMIDIN-5-YLPHENYL)-3-METHYL-5-[4-(TRIFLUOROMETHOXY)PHENYL]-3,5-DIHYDRO-4H-IMIDAZOL-4-ONE",
            "non-polymer",
        ),
        "CUS": ("(N-SALICYLIDEN-L-LEUCINATO)-COPPER(II)", "non-polymer"),
        "CUP": ("(N-SALICYLIDEN-L-PHENYLALANATO)-COPPER(II)", "non-polymer"),
        "561": (
            "6-HYDROXY-3-{(4R)-1-[4-(1-NAPHTHYLAMINO)-4-OXOBUTYL]-1,2,3-TRIAZOLIDIN-4-YL}-1-BENZOFURAN-5-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "D23": (
            "6-(3-AMINOPHENYL)-N-(TERT-BUTYL)-2-(TRIFLUOROMETHYL)QUINAZOLIN-4-AMINE",
            "non-polymer",
        ),
        "D22": ("N-OCTYLOCTAN-1-AMINE", "non-polymer"),
        "564": (
            "6-(5,5,8,8-TETRAMETHYL-5,6,7,8-TETRAHYDRO-NAPHTALENE-2-CARBONYL)-NAPHTALENE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "565": (
            "(3R)-4-[(3R)-3-AMINO-4-(2,4,5-TRIFLUOROPHENYL)BUTANOYL]-3-METHYL-1,4-DIAZEPAN-2-ONE",
            "non-polymer",
        ),
        "566": (
            "(3S)-1-CYCLOHEXYL-5-OXO-N-PHENYLPYRROLIDINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "567": (
            "2-[[(2R)-1-[[(2S)-5-AMINO-1-[(4-CARBAMIMIDOYLPHENYL)METHYLAMINO]-1,5-DIOXO-PENTAN-2-YL]AMINO]-3-(1H-INDOL-3-YL)-1-OXO-PROPAN-2-YL]SULFAMOYL]ETHANOIC ACID",
            "non-polymer",
        ),
        "CUF": (
            "N,N-BIS[4-[[1-(2-ETHYL)PIPERIDINE]OXY]SALICYLIDENE]-4,5-DIFLUORO-1,2-PHENYLENEDIAMINE-COPPER (II)",
            "non-polymer",
        ),
        "CUE": ("3,9-DIHYDROXY-[1]BENZOFURO[3,2-C]CHROMEN-6-ONE", "non-polymer"),
        "CUD": ("3-(4-AMINO-2-OXOPYRIMIDIN-1(2H)-YL)-D-ALANINE", "peptide"),
        "CUC": ("CYCLOHEXYLGLYCYL ", "peptide"),
        "CUB": ("CU(I)-S-MO(IV)(=O)O-NBIC CLUSTER", "non-polymer"),
        "CUA": ("DINUCLEAR COPPER ION", "non-polymer"),
        "JZ6": ("4-CHLORO-1H-PYRAZOLE", "non-polymer"),
        "CUO": ("CU2-O2 CLUSTER", "non-polymer"),
        "CUN": ("CU(I)-S-MO(IV)(=O)OH CLUSTER", "non-polymer"),
        "CUM": ("CU(I)-S-MO(VI)(=O)OH CLUSTER", "non-polymer"),
        "CUL": ("COPPER (II) CHLORIDE", "non-polymer"),
        "CUK": ("[4CU:2S] CLUSTER", "non-polymer"),
        "L79": (
            "(S)-(2E)-3[4-(5,5,8,8-TETRAMETHYL-5,6,7,8-TETRAHYDRO-2-NAPHTHALENYL)TETRAHYDRO-1-BENZOFURAN-2-YL]-2-PROPENOIC ACID",
            "non-polymer",
        ),
        "YR3": (
            "(2S)-3-{4-[1-ETHYL-1-(4-{[(2R)-2-HYDROXY-3,3-DIMETHYLBUTYL]OXY}-3-METHYLPHENYL)PROPYL]-2-METHYLPHENOXY}PROPANE-1,2-DIOL",
            "non-polymer",
        ),
        "V35": ("D-1-(4-CHLOROPHENYL)-2-(ACETAMIDO)ETHANE BORONIC ACID", "non-polymer"),
        "V36": ("L-1-(4-CHLOROPHENYL)-2-(ACETAMIDO)ETHANE BORONIC ACID", "non-polymer"),
        "V37": (
            "[(1R,5S)-8-PHENETHYL-8-AZABICYCLO[3.2.1]OCTAN-3-YL] BENZOATE",
            "non-polymer",
        ),
        "V38": (
            "[(1R,5S)-8-[(2R)-2-HYDROXY-2-PHENYL-ETHYL]-8-METHYL-8-AZONIABICYCLO[3.2.1]OCTAN-3-YL] BENZOATE",
            "non-polymer",
        ),
        "YR4": (
            "(2R)-2-{4-[3-(4-{[(2R)-2-HYDROXY-3,3-DIMETHYLBUTYL]OXY}-3-METHYLPHENYL)PENTAN-3-YL]-2-METHYLPHENOXY}BUTANE-1,4-DIOL",
            "non-polymer",
        ),
        "IFG": (
            "(2R,3R,4S,5R)-2-ACETAMIDO-3,4-DIHYDROXY-5-HYDROXYMETHYL-PIPERIDINE",
            "non-polymer",
        ),
        "ZHP": ("(S)-1'-(2',3'-DIHYDROXYPROPYL)-HYDROXYPYRIDONE", "nucleic-acid"),
        "3AI": (
            "N-[(2S)-2-AMINO-3-PHENYLPROPYL]-D-METHIONYL-L-ALANYL-L-ISOLEUCINE",
            "peptide",
        ),
        "3AH": ("[HISTIDIN-1-YL-4H-[1,2,4]TRIAZOL-5-YL]-AMINE", "peptide"),
        "3AO": ("1-(3-AMINOPHENYL)ETHANONE", "non-polymer"),
        "3AN": ("?", "non-polymer"),
        "3AM": (
            "[(2R,3S,4R,5R)-5-(6-AMINOPURIN-9-YL)-4-HYDROXY-2-(HYDROXYMETHYL)OXOLAN-3-YL] DIHYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "3AL": ("1-[(4S)-2,5-DIOXOIMIDAZOLIDIN-4-YL]UREA", "non-polymer"),
        "3AC": (
            "(2E)-3-{3-[(5-ETHYL-3-IODO-6-METHYL-2-OXO-1,2-DIHYDROPYRIDIN-4-YL)OXY]PHENYL}ACRYLONITRILE",
            "non-polymer",
        ),
        "3AB": ("3-AMINOBENZAMIDE", "non-polymer"),
        "3AA": ("3-AMINOPYRIDINE-ADENINE DINUCLEOTIDE PHOSPHATE", "non-polymer"),
        "ZHZ": ("2-(4-METHOXYPHENYL)ACETAMIDE", "non-polymer"),
        "3AG": (
            "2-[METHYL-(4-METHYL-PENT-3-ENYL)-AMINO]-ETHYL-DIPHOSPHATE",
            "non-polymer",
        ),
        "3AF": ("2,5-BIS{[4-(2-PENTYL)DIAMINOMETHYL]PHENYL}FURAN", "non-polymer"),
        "3AD": ("3'-DEOXYADENOSINE", "non-polymer"),
        "3AZ": ("3-(AMINOMETHYL)BENZOIC ACID", "peptide"),
        "3AY": ("PYRIMIDINE-2,4,6-TRIAMINE", "non-polymer"),
        "VR0": ("N~5~-[N-(METHYLCARBAMOYL)CARBAMIMIDOYL]-L-ORNITHINE", "peptide"),
        "IFM": ("5-HYDROXYMETHYL-3,4-DIHYDROXYPIPERIDINE", "non-polymer"),
        "3AS": (
            "(3AS,4R,9BR)-2,2-DIFLUORO-4-(4-HYDROXYPHENYL)-6-(METHOXYMETHYL)-1,2,3,3A,4,9B-HEXAHYDROCYCLOPENTA[C]CHROMEN-8-OL",
            "non-polymer",
        ),
        "3AR": ("N-OMEGA-PROPYL-L-ARGININE", "peptide"),
        "3AP": ("3-AMINOPYRIDINE", "non-polymer"),
        "3AW": ("6-METHYL-1,3,5-TRIAZINE-2,4-DIAMINE", "non-polymer"),
        "3AU": ("NONE", "nucleic-acid"),
        "3AT": ("3'-DEOXYADENOSINE-5'-TRIPHOSPHATE", "non-polymer"),
        "S1S": (
            "2-(4-((9S,10S,14S,Z)-18-(2-AMINO-2-OXOETHYL)-9-(CARBOXYMETHYL)-14-(NAPHTHALEN-1-YLMETHYL)-8,17,20-TRIOXO-7,16,19-TRIAZASPIRO[5.14]ICOS-11-EN-10-YL)PHENYL)MALONIC ACID",
            "non-polymer",
        ),
        "US4": (
            "[(2R,3S,5R)-3-({[(2R,3S,5R)-5-(6-AMINO-9H-PURIN-9-YL)-3-HYDROXYTETRAHYDROFURAN-2-YL]METHOXY}METHOXY)-5-(5-METHYL-2,4-DIOXO-3,4-DIHYDROPYRIMIDIN-1(2H)-YL)TETRAHYDROFURAN-2-YL]METHYL DIHYDROGEN PHOSPHATE",
            "nucleic-acid",
        ),
        "TV4": (
            "N-(6-AMINO-5-BROMOPYRIDIN-3-YL)-2,6-DIFLUORO-3-[(PROPYLSULFONYL)AMINO]BENZAMIDE",
            "non-polymer",
        ),
        "PKR": ("(2S)-1-METHYLPYRROLIDINE-2-CARBALDEHYDE", "peptide"),
        "S1P": (
            "(2S,3R,4E)-2-AMINO-3-HYDROXYOCTADEC-4-EN-1-YL DIHYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "NOA": ("NAPHTHYLOXYACETIC ACID", "non-polymer"),
        "YRG": ("(2S)-2-(4-CHLOROPHENOXY)-3-PHENYLPROPANOIC ACID", "non-polymer"),
        "CEB": (
            "2-CHLORO-4-ETHYLAMINO-6-(S(-)-2'-CYANO-4-BUTYLAMINO)-1,3,5-TRIAZINE",
            "non-polymer",
        ),
        "ZH1": ("?", "non-polymer"),
        "ZH3": ("ZINC TRIHYDROXIDE", "non-polymer"),
        "ZH2": (
            "4-[4-(4-AMINOPHENYL)BUTA-1,3-DIYN-1-YL]-N-[(2S,3R)-3-HYDROXY-1-NITROSO-1-OXOBUTAN-2-YL]BENZAMIDE",
            "non-polymer",
        ),
        "ZH4": (
            "4-[4-(3-AMINOPHENYL)BUTA-1,3-DIYN-1-YL]-N-[(2S,3R)-3-HYDROXY-1-NITROSO-1-OXOBUTAN-2-YL]BENZAMIDE",
            "non-polymer",
        ),
        "AZ5": (
            "5-CHLORO-N2-[(1S)-1-(5-FLUOROPYRIMIDIN-2-YL)ETHYL]-N4-(5-METHYL-1H-PYRAZOL-3-YL)PYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "6MZ": ("N6-METHYLADENOSINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "NOB": ("(DIHYDROXYBORANYLOXY-HYDROXY-BORANYL)OXYLITHIUM", "non-polymer"),
        "3A3": ("?", "non-polymer"),
        "3A2": (
            "1-{3-[(R)-{(1R)-1-[(GLYCYL-L-PROLYL)AMINO]-2-PHENYLETHYL}(HYDROXY)PHOSPHORYL]PROPANOYL}-L-PROLYL-D-NORLEUCINE",
            "peptide",
        ),
        "9BB": (
            "TRANS-4-({[(E)-1-AMINO-2-SULFANYLETHENYL]CARBAMOYL}OXY)CYCLOHEXANECARBOXYLIC ACID",
            "peptide",
        ),
        "9BD": (
            "2-(1H-IMIDAZOL-1-YL)-9-METHOXY-8-(2-METHOXYETHOXY)BENZO[C][2,7]NAPHTHYRIDIN-4-AMINE",
            "non-polymer",
        ),
        "3A6": (
            "3-AMINO-6-FLUORO-2-[4-(2-METHYLPYRIDIN-4-YL)PHENYL]-N-(METHYLSULFONYL)QUINOLINE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "3A5": (
            "(2Z,4S)-3-AZA-5-CARBOXYL-2-METHYL-4(METHYLCARBOXY)PENT-2-ENOYL",
            "peptide",
        ),
        "7HE": ("7-METHYL-7-DEPROPIONATEHEMIN", "non-polymer"),
        "CEL": (
            "4-[5-(4-METHYLPHENYL)-3-(TRIFLUOROMETHYL)-1H-PYRAZOL-1-YL]BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "Z": (
            "1-(2-DEOXY-5-O-PHOSPHONO-BETA-D-ERYTHRO-PENTOFURANOSYL)PYRIMIDIN-2(1H)-ONE",
            "nucleic-acid",
        ),
        "7HA": (
            "N-(3-{[2-(4-CHLOROPHENYL)-5-METHYL-1,3-OXAZOL-4-YL]METHOXY}BENZYL)-N-(METHOXYCARBONYL)GLYCINE",
            "peptide",
        ),
        "XBT": ("2-[(1R)-1-METHYLPROPYL]-4,5-DIHYDRO-1,3-THIAZOLE", "non-polymer"),
        "XBV": (
            "(3R,4R)-1-(2,2-DIFLUORO-ETHYL)-PYRROLIDINE-3,4-DICARBOXYLIC ACID 3-[(5-CHLORO-PYRIDIN-2-YL)-AMIDE]-4-{[2-FLUORO-4-(2-OXO-2H-PYRIDIN-1-YL)-PHENYL]-AMIDE}",
            "non-polymer",
        ),
        "FJZ": (
            "N-CYCLOHEXYL-4-(1H-IMIDAZOL-5-YL)PIPERIDINE-1-CARBOTHIOAMIDE",
            "non-polymer",
        ),
        "7HI": (
            "(3R,5R)-7-[3-(BIPHENYL-4-YLCARBAMOYL)-2-ETHYL-5,6,7,8-TETRAHYDROCYCLOHEPTA[B]PYRROL-1(4H)-YL]-3,5-DIHYDROXYHEPTANOIC ACID",
            "non-polymer",
        ),
        "7HK": (
            "METHYL [11-OXO-3-(PYRIDIN-4-YLAMINO)-10,11-DIHYDRO-5H-DIBENZO[B,E][1,4]DIAZEPIN-8-YL]ACETATE",
            "non-polymer",
        ),
        "7HV": ("2-[(3,4-DICHLOROBENZYL)AMINO]ETHANESULFONIC ACID", "non-polymer"),
        "7HP": ("7-HYDROXY-PYRAZOLO[4,3-D]PYRIMIDINE", "non-polymer"),
        "XBK": ("4,5-DE-EPOXYPIMARICIN", "non-polymer"),
        "XBJ": (
            "4-FLUORO-2-(4-{[(3S,4R)-4-(1-HYDROXY-1-METHYLETHYL)PYRROLIDIN-3-YL]AMINO}-6,7-DIMETHOXYQUINAZOLIN-2-YL)PHENOL",
            "non-polymer",
        ),
        "XBG": (
            "(2S,3AS,6AS)-OCTAHYDROCYCLOPENTA[B]PYRROLE-2-CARBOXYLIC-ACID",
            "non-polymer",
        ),
        "1OL": ("(2R,4S,5S)-5-AMINO-4-HYDROXY-2,7-DIMETHYLOCTANOIC ACID", "peptide"),
        "S1A": ("SORAPHEN A", "non-polymer"),
        "1OH": ("4-(1-METHYL-1-PHENYLETHYL)PHENOL", "non-polymer"),
        "O48": ("DECYL SULFAMATE", "non-polymer"),
        "PKC": ("(2R,3S)-1,3-DIAMINO-4-PHENYLBUTAN-2-OL", "peptide"),
        "344": ("6-AMINO-3,7-DIHYDRO-IMIDAZO[4,5-G]QUINAZOLIN-8-ONE", "non-polymer"),
        "345": (
            "4-[4-(4-CHLORO-PHENOXY)-BENZENESULFONYL]-4-METHYL-TETRAHYDRO-PYRAN;COMPOUND WITH N-HYDROXY-ACETAMIDE",
            "non-polymer",
        ),
        "346": (
            "(R)-(4-CARBAMIMIDOYL-PHENYLAMINO)-[5-ETHOXY-2-FLUORO-3-[(R)-TETRAHYDRO-FURAN-3-YLOXY]-PHENYL]-ACETIC ACID",
            "non-polymer",
        ),
        "347": (
            "TERT-BUTYL 4-({[4-(BUT-2-YN-1-YLAMINO)PHENYL]SULFONYL}METHYL)-4-[(HYDROXYAMINO)CARBONYL]PIPERIDINE-1-CARBOXYLATE",
            "non-polymer",
        ),
        "340": (
            "5-METHOXY-1,2-DIMETHYL-3-(PHENOXYMETHYL)INDOLE-4,7-DIONE",
            "non-polymer",
        ),
        "341": ("(3,5-DIFLUOROPHENYL)METHANOL", "non-polymer"),
        "342": (
            "(3S)-1-(1,3-BENZODIOXOL-5-YLMETHYL)-3-[4-(1H-IMIDAZOL-1-YL)PHENOXY]PIPERIDINE",
            "non-polymer",
        ),
        "343": (
            "(2S)-(3-{[AMINO(IMINO)METHYL]AMINO}PHENYL){[(S)-[(1R)-1-({N-[(BENZYLOXY)CARBONYL]-L-TYROSYL}AMINO)-2-METHYLPROPYL](HYDROXY)PHOSPHORYL]OXY}ACETIC ACID",
            "non-polymer",
        ),
        "PKF": ("CARBOBENZOXY-PRO-LYS-PHE-Y(PO2)-ALA-PRO-OME", "non-polymer"),
        "TEJ": (
            "(1S,3R,5Z,7E,20S,23S)-1,3-DIHYDROXY-23,26-EPOXY-9,10-SECOCHOLESTA-5,7,10,25(27)-TETRAEN-26-ONE",
            "non-polymer",
        ),
        "348": (
            "4-[(1R,3AS,4R,8AS,8BR)-1-ISOPROPYL-2-(4-METHOXYBENZYL)-3-OXODECAHYDROPYRROLO[3,4-A]PYRROLIZIN-4-YL]BENZENECARBOXIMIDAMIDE",
            "non-polymer",
        ),
        "349": (
            "5-{[4-{[2-(PYRROLIDIN-1-YLSULFONYL)BENZYL]AMINO}-5-(TRIFLUOROMETHYL)PYRIMIDIN-2-YL]AMINO}-1,3-DIHYDRO-2H-INDOL-2-ONE",
            "non-polymer",
        ),
        "34D": ("3,5-DIHYDROXYBENZOATE", "non-polymer"),
        "34F": (
            "(1S)-1-(3,4-DICHLOROPHENYL)-3-[FORMYL(HYDROXY)AMINO]PROPYL}PHOSPHONIC ACID",
            "non-polymer",
        ),
        "GS": ("GUANOSINE-5'-THIO-MONOPHOSPHATE", "nucleic-acid"),
        "34A": ("3,4-DIMETHYLANILINE", "non-polymer"),
        "34B": (
            "2,4-DIAMINO-5-(1-O-CARBORANYLMETHYL)-6-METHYLPYRIMIDINE",
            "non-polymer",
        ),
        "34C": ("(3-CHLORO-4-PROPOXY-PHENYL)-ACETIC ACID", "non-polymer"),
        "34M": (
            "3-(4-METHOXYPHENYL)-5-({[4-(4-METHOXYPHENYL)-5-METHYL-4H-1,2,4-TRIAZOL-3-YL]SULFANYL}METHYL)-1,2,4-OXADIAZOLE",
            "non-polymer",
        ),
        "34N": (
            "4-CYANO-N-{(3S)-3-(4-FLUOROPHENYL)-3-[4-(METHYLSULFONYL)PHENYL]PROPYL}BENZAMIDE",
            "non-polymer",
        ),
        "34O": (
            "2-(1,3-BENZODIOXOL-5-YL)-5-[(3-FLUORO-4-METHOXYBENZYL)SULFANYL]-1,3,4-OXADIAZOLE",
            "non-polymer",
        ),
        "34H": ("(2R)-2-HYDROXY-3-(4-HYDROXYPHENYL)PROPANOIC ACID", "non-polymer"),
        "34T": (
            "(3R)-8-(DIOXIDOSULFANYL)-3-METHYL-1,2,3,4-TETRAHYDROQUINOLINE",
            "non-polymer",
        ),
        "GD": ("GADOLINIUM ATOM", "non-polymer"),
        "34P": (
            "N-(CARBOXYMETHYL)-3-CYCLOHEXYL-D-ALANYL-N-({4-[(E)-AMINO(IMINO)METHYL]-1,3-THIAZOL-2-YL}METHYL)-L-PROLINAMIDE",
            "peptide",
        ),
        "34Q": (
            "(3R,4R)-4-(PYRROLIDIN-1-YLCARBONYL)-1-(QUINOXALIN-2-YLCARBONYL)PYRROLIDIN-3-AMINE",
            "non-polymer",
        ),
        "34R": (
            "(2E)-3-{5-[(2,4-DIAMINOPYRIMIDIN-5-YL)METHYL]-2,3-DIMETHOXYPHENYL}-1-[(1R)-1-PHENYLPHTHALAZIN-2(1H)-YL]PROP-2-EN-1-ONE",
            "non-polymer",
        ),
        "34S": (
            "(2E)-3-{5-[(2,4-DIAMINOPYRIMIDIN-5-YL)METHYL]-2,3-DIMETHOXYPHENYL}-1-[(1S)-1-PHENYLPHTHALAZIN-2(1H)-YL]PROP-2-EN-1-ONE",
            "non-polymer",
        ),
        "34Z": ("3,4-DICHLOROBENZOATE", "non-polymer"),
        "6BC": (
            "5-CHLORO-7-HYDROXY-6-(PHENYLMETHYL)PYRAZOLO[1,5-A]PYRIMIDINE-3-CARBONITRILE",
            "non-polymer",
        ),
        "6BG": ("6-BROMO-6-DEOXY-BETA-D-GLUCOPYRANOSE", "non-polymer"),
        "1O2": (
            "(2S)-3-(ALPHA-D-GALACTOPYRANOSYLOXY)-2-(HEXADECANOYLOXY)PROPYL (9Z)-OCTADEC-9-ENOATE",
            "non-polymer",
        ),
        "GBD": ("4-HYDROXY-1,2,5-THIADIAZOLE-3-CARBOXYLIC ACID", "non-polymer"),
        "6BD": (
            "(2S)-3-[(9H-FLUOREN-9-YLIDENEAMINO)OXY]-2-METHYLPROPANOIC ACID",
            "non-polymer",
        ),
        "OIO": ("8-(1H-IMIDAZOL-1-YL)OCTANOIC ACID", "non-polymer"),
        "VGA": (
            "1-[(3S)-5-PHENYL-3-THIOPHEN-2-YL-3H-1,4-BENZODIAZEPIN-2-YL]AZETIDIN-3-OL",
            "non-polymer",
        ),
        "6BP": ("HEXABROMOPLATINATE(IV)", "non-polymer"),
        "L7O": (
            "N-{5-[2-(METHYLAMINO)PYRIMIDIN-4-YL]-2-OXO-1,2-DIHYDROPYRIDIN-3-YL}-4-(PIPERIDIN-1-YL)BENZAMIDE",
            "non-polymer",
        ),
        "L7A": ("N-(6-OXO-1,6-DIHYDRO-3,4'-BIPYRIDIN-5-YL)BENZAMIDE", "non-polymer"),
        "GBE": (
            "3-{4-[2-(BETA-D-GLUCOPYRANOSYLOXY)ETHYL]PIPERAZIN-1-YL}PROPANE-1-SULFONIC ACID",
            "non-polymer",
        ),
        "OIN": (
            "(1R,5S)-8-METHYL-8-AZABICYCLO[3.2.1]OCT-3-YL (2R)-3-HYDROXY-2-PHENYLPROPANOATE",
            "non-polymer",
        ),
        "TEM": (
            "N-(2-HYDROXY-4-OXO-BUTYL)-N-(3-OXO-TRANSPROPENYL)AMINE",
            "non-polymer",
        ),
        "TTZ": (
            "N-[(3AS,4R,5S,6S,6AS)-4,5,6-TRIHYDROXY-4-(HYDROXYMETHYL)-4,5,6,6A-TETRAHYDRO-3AH-CYCLOPENTA[D][1,3]THIAZOL-2-YL]-ALPHA-D-GLUCOPYRANOSYLAMINE",
            "non-polymer",
        ),
        "TTY": (
            "4-OXO-4,5,6,7-TETRAHYDRO-1-BENZOFURAN-3-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "TTX": ("TENTOXIN", "non-polymer"),
        "VLM": ("VALINYLAMINE", "peptide"),
        "UEP": ("6-ETHYL-UMP", "non-polymer"),
        "TTS": (
            "3-((3E)-4-HYDROXY-3-{[2-(4-HYDROXYPHENYL)ETHYL]IMINO}-6-OXOCYCLOHEXA-1,4-DIEN-1-YL)ALANINE",
            "peptide",
        ),
        "TTR": ("9-ACETYL-2,3,4,9-TETRAHYDRO-1H-CARBAZOL-1-ONE", "non-polymer"),
        "TTQ": ("6-AMINO-7-HYDROXY-L-TRYPTOPHAN", "peptide"),
        "TTP": ("THYMIDINE-5'-TRIPHOSPHATE", "non-polymer"),
        "SIN": ("SUCCINIC ACID", "non-polymer"),
        "TTI": ("5-TELLURIUM-DEOXYURIDINE-5'-PHOSPHATE", "nucleic-acid"),
        "TTH": ("2,6,10,14-TETRAMETHYL-HEXADECA-2,6,10,14-TETRAENE", "non-polymer"),
        "TTO": ("(3,4-DIHYDROXY-PHENYL)-TRIPHENYL-ARSONIUM", "non-polymer"),
        "TTN": ("TARTRONATE", "non-polymer"),
        "TTM": ("N3-ETHYL-THYMIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "TTL": (
            "TRANS-(1S,2S)-2-AMINO-1,2,3,4-TETRAHYDRONAPHTHALEN-1-OL",
            "non-polymer",
        ),
        "TTC": (
            "(S)-10-[(DIMETHYLAMINO)METHYL]-4-ETHYL-4,9-DIHYDROXY-1H-PYRANO[3',4':6,7]INOLIZINO[1,2-B]-QUINOLINE-3,14(4H,12H)-DIONE",
            "non-polymer",
        ),
        "TTB": (
            "4-[(1E)-2-(5,5,8,8-TETRAMETHYL-5,6,7,8-TETRAHYDRONAPHTHALEN-2-YL)PROP-1-ENYL]BENZOIC ACID",
            "non-polymer",
        ),
        "TTA": ("TETRAPHENYL-ARSONIUM", "non-polymer"),
        "TTG": (
            "2-(1-DIMETHYLAMINOMETHYL-2-HYDROXY-8-HYDROXYMETHYL-9-OXO-9,11-DIHYDRO-INDOLIZINO[1,2-B]QUINOLIN-7-YL)-2-HYDROXY-BUTYRIC ACID",
            "non-polymer",
        ),
        "TTF": ("4,4,4-TRIFLUORO-1-THIEN-2-YLBUTANE-1,3-DIONE", "non-polymer"),
        "TTE": ("PHOSPHOMETHYL PHOSPHONIC ACID DEOXYTHYMIDYLATE ESTER", "non-polymer"),
        "TTD": ("CIS-SYN CYCLOBUTANE THYMINE DIMER", "nucleic-acid"),
        "ZN": ("ZINC ION", "non-polymer"),
        "NOY": ("(2R,3S,4R,5R)-5-(HYDROXYMETHYL)PIPERIDINE-2,3,4-TRIOL", "non-polymer"),
        "QNB": (
            "(3R)-1-AZABICYCLO[2.2.2]OCT-3-YL HYDROXY(DIPHENYL)ACETATE",
            "non-polymer",
        ),
        "L75": (
            "N-[2(R)-HYDROXY-1(S)-INDANYL]-2(R)-PHENYLMETHYL-4(S)-HYDROXY-5-[4-[2-BENZOFURANYLMETHYL]-2(S)-[TERT-BUTYLAMINOCARBONYL]-PIPERAZINYL]-PENTANEAMIDE",
            "non-polymer",
        ),
        "TT2": (
            "(1R,3R,7E,17BETA)-17-{(1S,2E,5R)-5-HYDROXY-1-METHYL-5-[(3S,5S,7S)-TRICYCLO[3.3.1.1~3,7~]DEC-1-YL]PENT-2-EN-1-YL}-2-METHYLIDENE-9,10-SECOESTRA-5,7-DIENE-1,3-DIOL",
            "non-polymer",
        ),
        "TT1": (
            "4-BENZYL-3-[(2-CHLOROBENZYL)SULFANYL]-5-THIOPHEN-2-YL-4H-1,2,4-TRIAZOLE",
            "non-polymer",
        ),
        "QNC": ("QUINOLINE-2-CARBOXYLIC ACID", "non-polymer"),
        "F59": (
            "N-[(BIPHENYL-4-YLCARBONYL)CARBAMOYL]-BETA-D-GLUCOPYRANOSYLAMINE",
            "non-polymer",
        ),
        "QNO": ("2-NONYL-4-HYDROXYQUINOLINE N-OXIDE", "non-polymer"),
        "CE2": ("3-(5-TERT-BUTYL-3-OXIDOISOXAZOL-4-YL)-L-ALANINATE", "non-polymer"),
        "NOT": ("6-AZIDO-L-NORLEUCINE", "peptide"),
        "7CA": (
            "5,7-DIHYDROXY-2-(4-METHOXYPHENYL)-8-(3-METHYLBUTYL)-4-OXO-4H-CHROMEN-3-YL 6-DEOXY-ALPHA-L-MANNOPYRANOSIDE",
            "non-polymer",
        ),
        "E2P": ("PROSTAGLANDIN B2", "non-polymer"),
        "TAE": ("N1-[(E)-4-DIHYDROXYPHOSPHONYL-BUT-2-ENYL]-THYMINE", "non-polymer"),
        "Q97": (
            "4,4'-[(1Z)-1-(4-ETHOXYPHENYL)BUT-1-ENE-1,2-DIYL]DIPHENOL",
            "non-polymer",
        ),
        "TEB": (
            "(4R,5S)-3-(1-(4,5-DIHYDROTHIAZOL-2-YL)AZETIDIN-3-YLTHIO)-5-((2S,3R)-3-HYDROXY-1-OXOBUTAN-2-YL)-4-METHYL-4,5- DIHYDRO-1H-PYRROLE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "QQ3": (
            "(3R)-3-(4-METHYL-1,3-DIOXO-1,3-DIHYDRO-2H-PYRROLO[3,4-C]QUINOLIN-2-YL)HEXANOIC ACID",
            "non-polymer",
        ),
        "TAG": ("D-TAGATOSE", "saccharide"),
        "E2B": (
            "3-{[(9BETA,14BETA,16ALPHA,17ALPHA)-3,17-DIHYDROXYESTRA-1,3,5(10)-TRIEN-16-YL]METHYL}BENZAMIDE",
            "non-polymer",
        ),
        "FJP": (
            "(2E,6Z)-3,7,11-TRIMETHYLDODECA-2,6,10-TRIEN-1-YL DIHYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "E2M": (
            "CIS-4-{[2-({4-[(1E)-3-MORPHOLIN-4-YL-3-OXOPROP-1-EN-1-YL]-2,3-BIS(TRIFLUOROMETHYL)PHENYL}SULFANYL)PHENOXY]METHYL}CYCLOHEXANECARBOXYLIC ACID",
            "non-polymer",
        ),
        "E2I": (
            "4-{[(5-BUTYLPYRIMIDIN-2-YL)SULFANYL]ACETYL}BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "TAA": (
            "[4-(2,2,2-TRIFLUORO-ACETYLAMINO)-BENZYL]-PHOSPHONIC ACID MONO-[2-(2,2-DICHLORO-1-HYDROXY-ETHYLAMINO)-3-HYDROXY-1-(4-NITRO-PHENYL)-PROPYL] ESTER",
            "non-polymer",
        ),
        "E27": (
            "2-CHLORO-5-[(1H-IMIDAZO[4,5-C]QUINOLIN-2-YLSULFANYL)ACETYL]BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "E20": (
            "1-BENZYL-4-[(5,6-DIMETHOXY-1-INDANON-2-YL)METHYL]PIPERIDINE",
            "non-polymer",
        ),
        "NBO": ("NABUMETONE", "non-polymer"),
        "CVR": ("?", "non-polymer"),
        "TAC": ("TETRACYCLINE", "non-polymer"),
        "E28": (
            "4-{4-[4-(AMINOMETHYL)-3-(TRIFLUOROMETHYL)PHENYL]-1H-PYRAZOL-5-YL}-N-(2,3-DIHYDRO-1-BENZOFURAN-5-YLMETHYL)-1H-PYRROLE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "I2H": (
            "5-[[(2R)-2-CYCLOPROPYL-7,8-DIMETHOXY-CHROMAN-5-YL]METHYL]PYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "TAM": ("TRIS(HYDROXYETHYL)AMINOMETHANE", "non-polymer"),
        "NOV": ("NOVOBIOCIN", "non-polymer"),
        "Q9G": (
            "5-(5-FLUORO-2-METHYLPHENYL)-6,7,8,9-TETRAHYDRO-3H-PYRAZOLO[3,4-C]ISOQUINOLIN-1-AMINE",
            "non-polymer",
        ),
        "1FF": ("1-METHYL-5-PHENYL-1H-PYRAZOLE-4-CARBOXYLIC ACID", "non-polymer"),
        "Q9C": ("N-{4-[(3-AMINOPROPYL)AMINO]BUTYL}ACETAMIDE", "non-polymer"),
        "W56": (
            "5-(5-(4-(5-HYDRO-4-METHYL-2-OXAZOLYL)PHENOXY)PENTYL)-3-METHYL ISOXAZOLE",
            "non-polymer",
        ),
        "W54": (
            "5-(5-(2,6-DICHLORO-4-(4,5-DIHYDRO-2-OXAZOLY)PHENOXY)PENTYL)-3-METHYL ISOXAZOLE",
            "non-polymer",
        ),
        "W59": (
            "5-(7-(5-HYDRO-4-ETHYL-2-OXAZOLYL)PHENOXY)HEPTYL)-3-METHYL ISOXAZOLE",
            "non-polymer",
        ),
        "1FA": (
            "[(7-{[2-(3-MORPHOLIN-4-YLPROP-1-YN-1-YL)-6-{[4-(TRIFLUOROMETHYL)PHENYL]ETHYNYL}PYRIDIN-4-YL]THIO}-2,3-DIHYDRO-1H-INDEN-4-YL)OXY]ACETIC ACID",
            "non-polymer",
        ),
        "LGM": (
            "3-METHYL-1-(3-(5-OXO-4,5-DIHYDRO-1H-1,2,4-TRIAZOL-3-YL)PHENYL)-6-(2'-(PYRROLIDIN-1-YLMETHYL)BIPHENYL-4-YL)-5,6-DIHYDRO-1H-PYRAZOLO[3,4-C]PYRIDIN-7(4H)-ONE",
            "non-polymer",
        ),
        "NOW": (
            "NALPHA-[(3-TERT-BUTYL-1-METHYL-1H-PYRAZOL-5-YL)CARBONYL]-N-[(2Z)-2-IMINOETHYL]-3-METHYL-L-PHENYLALANINAMIDE",
            "non-polymer",
        ),
        "TAK": (
            "6-[4-(2-PIPERIDIN-1-YLETHOXY)PHENYL]-3-PYRIDIN-4-YLPYRAZOLO[1,5-A]PYRIMIDINE",
            "non-polymer",
        ),
        "TEE": ("2-AMINO-ETHENETHIOL", "non-polymer"),
        "UKW": (
            "4-ETHYNYL-N-[(1S,2R)-2-HYDROXY-1-(OXOCARBAMOYL)PROPYL]BENZAMIDE",
            "non-polymer",
        ),
        "UKP": (
            "[2,4,6-TRIISOPROPYL-PHENYLSULFONYL-L-[3-AMIDINO-PHENYLALANINYL]]-N'-BETA-ALANINYL-PIPERAZINE",
            "non-polymer",
        ),
        "TA7": ("4,4,4-TRIFLUORO-1-(4-METHOXYPHENYL)BUTANE-1,3-DIONE", "non-polymer"),
        "NOP": ("PYRIDOXAL-5'-PHOSPHATE-N-OXIDE", "non-polymer"),
        "UKC": (
            "8-ETHYL-6-METHOXY-3-METHYL-8H-1,3A,7,8,9-PENTAAZA-DIBENZO[E,H]AZULENE",
            "non-polymer",
        ),
        "UKA": (
            "6-(2,2-DIPHENYLETHYLAMINO)-9-[(2R,3R,4S,5S)-5-(ETHYLCARBAMOYL)-3,4-DIHYDROXY-OXOLAN-2-YL]-N-[2-[(1-PYRIDIN-2-YLPIPERIDIN-4-YL)CARBAMOYLAMINO]ETHYL]PURINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "TA1": ("TAXOL", "non-polymer"),
        "ONH": ("N~5~-HYDROXY-L-ORNITHINE", "peptide"),
        "TEF": ("4-(2H-TETRAZOL-2-YL)-L-PHENYLALANINE", "peptide"),
        "IIP": ("IMIDAZOLE-PYRROLE POLYAMIDE", "non-polymer"),
        "ONL": ("5-OXO-L-NORLEUCINE", "peptide"),
        "ONM": ("3'-O-(N-METHYLANTHRANILOYL)-GUANOSINE-5'-TRIPHOSPHATE", "non-polymer"),
        "ONB": ("5H-DIBENZO[B,F]AZEPINE", "non-polymer"),
        "CVB": ("4-(CARBOXYVIN-2-YL)PHENYLBORONIC ACID", "non-polymer"),
        "ONA": ("unknown", "non-polymer"),
        "FX1": ("BETA-D-THREO-HEXO-2,4-DIULO-2,5-FURANOSE", "non-polymer"),
        "FX3": (
            "5-O-[(2E)-3-(4-HYDROXY-3-METHOXYPHENYL)PROP-2-ENOYL]-ALPHA-L-RIBOFURANOSE",
            "non-polymer",
        ),
        "ONE": ("1-(BETA-D-RIBOFURANOSYL)-PYRIDIN-4-ONE-5'-PHOSPHATE", "nucleic-acid"),
        "IIE": (
            "1-{2-[(4-CHLOROPHENYL)AMINO]-2-OXOETHYL}-N-(1-ISOPROPYLPIPERIDIN-4-YL)-1H-INDOLE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "IID": (
            "N-(1-ISOPROPYLPIPERIDIN-4-YL)-1-(3-METHOXYBENZYL)-1H-INDOLE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "IIG": (
            "N-CYCLOPROPYL-N-(CIS-4-CYCLOPROPYL-4-HYDROXYCYCLOHEXYL)-4-[(1S)-2,2,2-TRIFLUORO-1-HYDROXY-1-METHYLETHYL]BENZAMIDE",
            "non-polymer",
        ),
        "RZ4": ("O-TERT-BUTYL-L-SERINE", "peptide"),
        "IIA": (
            "1-{[5-(5-CHLORO-2-THIENYL)ISOXAZOL-3-YL]METHYL}-N-(1-ISOPROPYLPIPERIDIN-4-YL)-1H-INDOLE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "IIC": ("4-IMIDAZOLMETHYLENE-5-IMIDAZOLONE CHROMOPHORE", "peptide"),
        "IIB": (
            "1-{[5-(5-CHLORO-2-THIENYL)ISOXAZOL-3-YL]METHYL}-3-CYANO-N-(1-ISOPROPYLPIPERIDIN-4-YL)-7-METHYL-1H-INDOLE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "PW1": ("[2,4-BIS(FLUORANYL)-3-PHENOXY-PHENYL]METHYLAZANIUM", "non-polymer"),
        "IIL": ("ISO-ISOLEUCINE", "peptide"),
        "ONP": (
            "O-NITROPHENYL AMINOETHYLDIPHOSPHATE BERYLLIUM TRIFLUORIDE",
            "non-polymer",
        ),
        "IIN": ("2-METHYL-3-AMINO-3-PHENYLISOINDOL-1-ONE", "non-polymer"),
        "RZ8": ("N-HEXYL-4-SULFAMOYLBENZAMIDE", "non-polymer"),
        "II4": ("1-(3-METHYLPHENYL)-1H-BENZIMIDAZOL-5-AMINE", "non-polymer"),
        "JIM": ("(2R)-2,3-DIHYDROXYPROPYL ACETATE", "non-polymer"),
        "FXY": ("1-METHYLHEPTYLFORMAMIDE", "non-polymer"),
        "II2": (
            "5-AMINO-2-{4-[(4-AMINOPHENYL)SULFANYL]PHENYL}-1H-ISOINDOLE-1,3(2H)-DIONE",
            "non-polymer",
        ),
        "FXV": (
            "METHYL-3-(4'-N-OXOPYRIDYLPHENOYL)-3-METHYL-2-(M-AMIDINOBENZYL)-PROPIONATE",
            "non-polymer",
        ),
        "X63": (
            "2-{[(2-AMINOPYRIMIDIN-5-YL)METHYL]AMINO}-4-[(2-HYDROXYETHYL)AMINO]-5-NITROBENZAMIDE",
            "non-polymer",
        ),
        "FXP": (
            "2-DEOXY-2-FLUORO-4-O-BETA-D-XYLOPYRANOSYL-ALPHA-D-XYLOPYRANOSE",
            "saccharide",
        ),
        "FXM": (
            "2,2-DIMETHYLPROPANOYLOXYMETHYL (6R,7R)-7-[(Z)-2-(2-AMINOTHIAZOL-4-YL)PENT-2-ENYLAMINO]-3-CARBAMOYLOXYMETHYL-8-OXO-5-THIA-1-AZABICYCLO[4.2.0]OCT-2-ENE-2-CARBOXYLATE",
            "non-polymer",
        ),
        "NOR": ("CYCLOHEXYL-NORSTATINE", "peptide"),
        "FXN": (
            "5-FLUORO-1H-INDOLE-2-CARBOXYLIC ACID-(2-MERCAPTO-ETHYL)-AMIDE",
            "non-polymer",
        ),
        "FXI": (
            "1-(BUTYL{[(4-CHLOROPHENYL)AMINO]CARBONYL}AMINO)-N-[3-FLUORO-2'-(METHYLSULFONYL)BIPHENYL-4-YL]CYCLOPROPANECARBOXAMIDE",
            "non-polymer",
        ),
        "FXG": (
            "N-BUTYL-3-{[6-(9H-PURIN-6-YLAMINO)HEXANOYL]AMINO}BENZAMIDE",
            "non-polymer",
        ),
        "NBN": ("N-BUTYL ISOCYANIDE", "non-polymer"),
        "FXA": (
            "N-(2-(((5-CHLORO-2-PYRIDINYL)AMINO)SULFONYL)PHENYL)-4-(2-OXO-1(2H)-PYRIDINYL)BENZAMIDE",
            "non-polymer",
        ),
        "RZX": (
            "2-[(R)-{[4-(3-METHOXYPROPOXY)-3-METHYLPYRIDIN-2-YL]METHYL}SULFINYL]-1H-BENZIMIDAZOLE",
            "non-polymer",
        ),
        "OC5": ("CALCIUM ION, 5 WATERS COORDINATED", "non-polymer"),
        "NOS": ("INOSINE", "non-polymer"),
        "T3M": (
            "4-[3-AMINO-6-(3,4,5-TRIMETHOXYPHENYL)PYRAZIN-2-YL]BENZOIC ACID",
            "non-polymer",
        ),
        "M54": (
            "6-(3-AMINOPROPYL)-4-(3-HYDROXYPHENYL)-9-(1H-PYRAZOL-4-YL)BENZO[H]ISOQUINOLIN-1(2H)-ONE",
            "non-polymer",
        ),
        "M51": (
            "{(2S,4S)-4-[4-(3-METHYL-1-PHENYL-1H-PYRAZOL-5-YL)PIPERAZIN-1-YL]PYRROLIDIN-2-YL}(1,3-THIAZOLIDIN-3-YL)METHANONE",
            "non-polymer",
        ),
        "DB9": (
            "2-{4'-[AMINO(IMINO)METHYL]BIPHENYL-3-YL}-1H-BENZIMIDAZOLE-6-CARBOXIMIDAMIDE",
            "non-polymer",
        ),
        "DB8": (
            "4-[(2,4-DICHLORO-5-METHOXYPHENYL)AMINO]-6-METHOXY-7-[3-(4-METHYLPIPERAZIN-1-YL)PROPOXY]QUINOLINE-3-CARBONITRILE",
            "non-polymer",
        ),
        "TB1": ("(3S)-3-(BENZYLOXY)-L-ASPARTIC ACID", "non-polymer"),
        "DB1": ("2,3-DIHYDROXYBENZAMIDE", "non-polymer"),
        "DB0": (
            "N-(DIBENZO[B,F]OXEPIN-10-YLMETHYL)-N-METHYL-N-PROP-2-YNYLAMINE",
            "non-polymer",
        ),
        "DB3": (
            "(11Z,14E)-N-[(2S,3S,4R)-1-(ALPHA-D-GALACTOPYRANOSYLOXY)-3,4-DIHYDROXYOCTADECAN-2-YL]ICOSA-11,14-DIENAMIDE",
            "non-polymer",
        ),
        "DB7": ("DIBUTYL BENZENE-1,2-DICARBOXYLATE", "non-polymer"),
        "DB6": (
            "(11E,14E)-N-[(2S,3S,4R)-1-(ALPHA-D-GLUCOPYRANOSYLOXY)-3,4-DIHYDROXYOCTADECAN-2-YL]ICOSA-11,14-DIENAMIDE",
            "non-polymer",
        ),
        "SAK": (
            "(2S)-5-HYDROXY-2-(4-HYDROXYPHENYL)-7-METHOXY-2,3-DIHYDRO-4H-CHROMEN-4-ONE",
            "non-polymer",
        ),
        "SAI": ("S-ADENOSYL-L-HOMOSELENOCYSTEINE", "non-polymer"),
        "SAH": ("S-ADENOSYL-L-HOMOCYSTEINE", "peptide"),
        "0KT": (
            "N-METHYL-N-(4-{(1S)-2,2,2-TRIFLUORO-1-HYDROXY-1-[1-(2-METHOXYETHYL)-1H-PYRROL-2-YL]ETHYL}PHENYL)BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "SAM": ("S-ADENOSYLMETHIONINE", "non-polymer"),
        "SAL": ("2-HYDROXYBENZOIC ACID", "non-polymer"),
        "SAC": ("N-ACETYL-SERINE", "peptide"),
        "SAB": ("4-SULFONAMIDE-[1-(4-AMINOBUTANE)]BENZAMIDE", "non-polymer"),
        "SAA": ("(S)-ATROLACTIC ACID", "non-polymer"),
        "SAG": (
            "(S)-(+)-2-[4-(FLUOROBENZYLOXY-BENZYLAMINO)PROPIONAMIDE]",
            "non-polymer",
        ),
        "SAF": ("3-[(1S)-1-(DIMETHYLAMINO)ETHYL]PHENOL", "non-polymer"),
        "SAE": ("SELENAZOLE-4-CARBOXYAMIDE-ADENINE DINUCLEOTIDE", "non-polymer"),
        "SAD": (
            "BETA-METHYLENE-SELENAZOLE-4-CARBOXYAMIDE-ADENINE DINUCLEOTIDE",
            "non-polymer",
        ),
        "CH0": (
            "2-{[(3ALPHA,5ALPHA,7ALPHA,8ALPHA,10ALPHA,12ALPHA,17ALPHA)-3,12-BIS{2-[(4-O-ALPHA-D-GLUCOPYRANOSYL-BETA-D-GLUCOPYRANOSYL)OXY]ETHOXY}CHOLAN-7-YL]OXY}ETHYL 4-O-ALPHA-D-GLUCOPYRANOSYL-BETA-D-GLUCOPYRANOSIDE",
            "non-polymer",
        ),
        "406": (
            "N-[3-(4,5'-BIPYRIMIDIN-2-YLAMINO)-4-METHYLPHENYL]-4-{[(3S)-3-(DIMETHYLAMINO)PYRROLIDIN-1-YL]METHYL}-3-(TRIFLUOROMETHYL)BENZAMIDE",
            "non-polymer",
        ),
        "CH2": ("METHYLENE GROUP", "non-polymer"),
        "404": (
            "(2R)-2-{[7-(BENZYLAMINO)-3-(PROPAN-2-YL)-1H-PYRAZOLO[4,3-D]PYRIMIDIN-5-YL]AMINO}BUTAN-1-OL",
            "non-polymer",
        ),
        "403": (
            "(2R,3R)-4-[4-(2-CHLOROPHENYL)PIPERAZIN-1-YL]-2,3-DIHYDROXY-4-OXO-N-(2-THIOPHEN-2-YLETHYL)BUTANAMIDE",
            "non-polymer",
        ),
        "CH5": (
            "2-(((R)-2,3-DIHYDROXYPROPYL)PHOSPHORYLOXY)-N,N,N-TRIMETHYLETHANAMINIUM",
            "non-polymer",
        ),
        "401": (
            "(2S)-2-{3-[({[2-FLUORO-4-(TRIFLUOROMETHYL)PHENYL]CARBONYL}AMINO)METHYL]-4-METHOXYBENZYL}BUTANOIC ACID",
            "non-polymer",
        ),
        "400": (
            "3-({[4-(4-METHYLPIPERAZIN-1-YL)PHENYL]CARBONYL}AMINO)-N-[(1R)-1-PHENYLPROPYL]-1H-THIENO[3,2-C]PYRAZOLE-5-CARBOXAMIDE",
            "non-polymer",
        ),
        "CH8": ("5-CHLORO-2-PHENOXYPHENOL", "non-polymer"),
        "SAR": ("SARCOSINE", "peptide"),
        "SAQ": (
            "2-[(THIOPHEN-2-YLMETHYL)AMINO]-1,7-DIHYDRO-8H-IMIDAZO[4,5-G]QUINAZOLIN-8-ONE",
            "non-polymer",
        ),
        "SAP": ("?", "non-polymer"),
        "SAW": (
            "(1R,3S,5R,7S)-N-[8-({[5-(DIMETHYLAMINO)NAPHTHALEN-1-YL]SULFONYL}AMINO)OCTYL]-3-HYDROXYTRICYCLO[3.3.1.1~3,7~]DECANE-1-CARBOXAMIDE",
            "non-polymer",
        ),
        "SAV": (
            "N~4~-(3-METHYL-1H-INDAZOL-6-YL)-N~2~-(3,4,5-TRIMETHOXYPHENYL)PYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "SAU": (
            "13-METHYL[1,3]BENZODIOXOLO[5,6-C][1,3]DIOXOLO[4,5-I]PHENANTHRIDIN-13-IUM",
            "non-polymer",
        ),
        "SAT": ("SULFOACETIC ACID", "non-polymer"),
        "40G": (
            "N-[(2-AMINO-6-OXO-1,6-DIHYDRO-9H-PURIN-9-YL)ACETYL]-N-[(2S)-2-AMINOPROPYL]GLYCINE",
            "peptide",
        ),
        "CHA": ("1-HYDROXY-2-AMINO-3-CYCLOHEXYLPROPANE", "non-polymer"),
        "CHB": ("3-CHLORO-4-HYDROXYBENZOIC ACID", "non-polymer"),
        "CHC": ("6-ETHYL-CHENODEOXYCHOLIC ACID", "non-polymer"),
        "40C": (
            "N-[(4-AMINO-2-OXOPYRIMIDIN-1(2H)-YL)ACETYL]-N-[(2S)-2-AMINOPROPYL]GLYCINE",
            "peptide",
        ),
        "40A": (
            "N-[(2S)-2-AMINOPROPYL]-N-[(6-AMINO-9H-PURIN-9-YL)ACETYL]GLYCINE",
            "peptide",
        ),
        "CHG": ("CYCLOHEXYL-GLYCINE", "peptide"),
        "CHH": ("N,N,N-TRIMETHYL-4-OXOPENTAN-1-AMINIUM", "non-polymer"),
        "CHI": (
            "5-CHLORO-1H-INDOLE-2-CARBOXYLIC ACID [1-(4-FLUOROBENZYL)-2-(4-HYDROXYPIPERIDIN-1YL)-2-OXOETHYL]AMIDE",
            "non-polymer",
        ),
        "CHK": (
            "6-[(CYCLOHEXYLACETYL)(2-HYDROXYETHYL)AMINO]-6-DEOXY-D-XYLO-HEXITOL",
            "non-polymer",
        ),
        "CHL": ("CHLOROPHYLL B", "non-polymer"),
        "CHM": ("1,3-DICHLORO-PROPAN-2-ONE", "non-polymer"),
        "CHN": ("4-AMINOMETHYL-CYCLOHEXYLAMINE", "non-polymer"),
        "CHO": ("GLYCOCHENODEOXYCHOLIC ACID", "non-polymer"),
        "CHP": ("3-CHLORO-4-HYDROXYPHENYLGLYCINE", "peptide"),
        "CHQ": ("CYCLO-(L-HISTIDINE-L-PROLINE) INHIBITOR", "non-polymer"),
        "CHR": ("NEOCARZINOSTATIN-CHROMOPHORE", "non-polymer"),
        "40T": (
            "N-[(2S)-2-AMINOPROPYL]-N-[(5-METHYL-2,4-DIOXO-3,4-DIHYDROPYRIMIDIN-1(2H)-YL)ACETYL]GLYCINE",
            "peptide",
        ),
        "CHT": ("CHOLINE ION", "non-polymer"),
        "JAK": (
            "2-METHYL-1-(PIPERIDIN-4-YL)-1,6-DIHYDROIMIDAZO[4,5-D]PYRROLO[2,3-B]PYRIDINE",
            "non-polymer",
        ),
        "CHW": (
            "3-CHLORO-4,6-DIHYDROXY-5-[(2E,6E,8S)-8-HYDROXY-3,7-DIMETHYLNONA-2,6-DIEN-1-YL]-2-METHYLBENZALDEHYDE",
            "non-polymer",
        ),
        "CHX": ("CYCLOHEXANE", "non-polymer"),
        "SA2": (
            "(3R)-4-[(4-CARBOXYBUTANOYL)OXY]-N-[(1E)-3-OXOPROP-1-EN-1-YL]-3-SULFINO-D-VALINE",
            "non-polymer",
        ),
        "CHZ": ("1,8-DIHYDROXYANTHRACENE-9,10-DIONE", "non-polymer"),
        "SA0": ("2-(HYDROXYMETHYL)PHENYL BETA-D-GLUCOPYRANOSIDE", "saccharide"),
        "LOB": ("LOBELINE", "non-polymer"),
        "JAC": (
            "1-({4-[3-(TRIFLUOROMETHYL)-6,7-DIHYDROPYRANO[4,3-C]PYRAZOL-1(4H)-YL]PHENYL}METHYL)-2-PYRROLIDINONE",
            "non-polymer",
        ),
        "DBI": (
            "PHOSPHORIC ACID MONO-[5-(5,6-DIMETHYL-BENZOIMIDAZOL-1-YL)-4-HYDROXY-2-HYDROXYMETHYL-TETRAHYDRO-FURAN-3-YL] ESTER",
            "non-polymer",
        ),
        "DBH": ("2,3-DIHYDROXY-BENZOIC ACID", "non-polymer"),
        "DBK": (
            "(3S,3AR,4S,6S,6AS,8R,9BS)-6-(ACETYLOXY)-3,3A-DIHYDROXY-3,6,9-TRIMETHYL-8-{[(2Z)-2-METHYLBUT-2-ENOYL]OXY}-2-OXO-2,3,3A,4,5,6,6A,7,8,9B-DECAHYDROAZULENO[4,5-B]FURAN-4-YL DODECANOATE",
            "non-polymer",
        ),
        "DBJ": ("2,3-DIHYDRO-1,4-BENZODIOXIN-5-YLMETHANOL", "non-polymer"),
        "DBM": ("9-(6-DEOXY-BETA-D-ALLOFURANOSYL)-6-METHYLPURINE", "nucleic-acid"),
        "M5P": ("(S)-2-(MERCAPTOMETHYL)-5-PHENYLPENTANOIC ACID", "non-polymer"),
        "DBO": (
            "N-{(1S,2R)-1-BENZYL-2-HYDROXY-3-[(3-METHYLBENZYL)AMINO]PROPYL}DIBENZO[B,F]OXEPINE-10-CARBOXAMIDE",
            "non-polymer",
        ),
        "DBN": (
            "6-(4,5-DIHYDRO-1H-IMIDAZOL-2-YL)-2-{5-[4-(4,5-DIHYDRO-1H-IMIDAZOL-2-YL)PHENYL]THIEN-2-YL}-1H-BENZIMIDAZOLE",
            "non-polymer",
        ),
        "DBA": ("(2,6-DIMETHYL-PHENOXY)-ACETIC ACID", "non-polymer"),
        "DBC": ("(4S,5S)-4,5-BIS(MERCAPTOMETHYL)-1,3-DIOXOLAN-2-OL", "non-polymer"),
        "DBB": ("D-ALPHA-AMINOBUTYRIC ACID", "peptide"),
        "DBE": ("BIS(4-HYDROXYPHENYL)METHANONE", "non-polymer"),
        "DBD": (
            "7-(1,1-DIOXO-1H-BENZO[D]ISOTHIAZOL-3-YLOXYMETHYL)-2-(OXALYL-AMINO)-4,7-DIHYDRO-5H-THIENO[2,3-C]PYRAN-3-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "DBG": (
            "3-AMINOBENZOPHENONE-4-YL-AMINOHYDROXYPHOSPHINYLAMINOPHOSPHONIC ACID-GUANYLATE ESTER",
            "non-polymer",
        ),
        "DBF": ("DIBENZOFURAN-4,6-DICARBOXYLIC ACID", "non-polymer"),
        "DBY": ("3,5 DIBROMOTYROSINE", "peptide"),
        "DBX": ("2,5-DIHYDROXYBENZENESULFONIC ACID", "non-polymer"),
        "M5G": ("?", "non-polymer"),
        "DBZ": ("3-(BENZOYLAMINO)-L-ALANINE", "peptide"),
        "DBQ": ("DEBROMOHYMENIALDISINE", "non-polymer"),
        "DBP": ("1,3-DIAMINOBENZYL PHENYLALANINE", "non-polymer"),
        "DBS": ("2-(2,3-DIHYDROXY-BENZOYLAMINO)-3-HYDROXY-PROPIONIC ACID", "peptide"),
        "DBR": (
            "ADAMANTANE-1-CARBOXYLIC ACID-5-DIMETHYLAMINO-NAPHTHALENE-1-SULFONYLAMINO-BUTYL-AMIDE",
            "non-polymer",
        ),
        "DBU": ("(2Z)-2-AMINOBUT-2-ENOIC ACID", "peptide"),
        "DBT": ("5-HEPTYL-6-HYDROXY-1,3-BENZOTHIAZOLE-4,7-DIONE", "non-polymer"),
        "DBW": (
            "4,14-DIHYDRO-8H-[1,3]DIOXOLO[4,5-G]ISOQUINO[3,2-A]ISOQUINOLINE-9,10-DIOL",
            "non-polymer",
        ),
        "DBV": ("15,16-DIHYDROBILIVERDIN", "non-polymer"),
        "BT5": ("BIOTINYL-5-AMP", "nucleic-acid"),
        "GBR": (
            "5'-O-[(R)-{[(S)-[(R)-BROMO(PHOSPHONO)METHYL](HYDROXY)PHOSPHORYL]OXY}(HYDROXY)PHOSPHORYL]-2'-DEOXYGUANOSINE",
            "non-polymer",
        ),
        "BT6": ("BENZENETHIOL", "non-polymer"),
        "BT1": (
            "{2-[4-(2-PYRROLIDIN-1-YL-ETHOXY)-PHENYL]-BENZO[B]THIOPHEN-3-YL}-[4-(2-PYRROLIDIN-1-YL-ETHOXY)-PHENYL]-METHANONE",
            "non-polymer",
        ),
        "BT3": (
            "3-[4-(2-PYRROLIDIN-1-YL-ETHOXY)-BENZYL]-2-4-(2-PYRROLIDIN-1-YL-ETHOXY)-PHENYL] -BENZO[B]THIOPHEN-6-OL",
            "non-polymer",
        ),
        "BT2": (
            "3-[4-(2-PYRROLIDIN-1-YL-ETHOXY)-BENZYL]-2-4-(2-PYRROLIDIN-1-YL-ETHOXY)-PHENYL] -BENZO[B]THIOPHENE",
            "non-polymer",
        ),
        "K7I": (
            "L-ALANYL-N-[(1S,2R)-1-BENZYL-2-HYDROXYPROPYL]-L-ALANINAMIDE",
            "non-polymer",
        ),
        "0KE": (
            "(5-HYDROXY-4-{[(1-HYDROXY-7-METHOXY-2-OXO-1,2-DIHYDROQUINOLIN-3-YL)AMINO]METHYL}-6-METHYLPYRIDIN-3-YL)METHYL DIHYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "BT9": ("3-{[4-(TRIFLUOROMETHYL)PHENYL]AMINO}BENZOIC ACID", "non-polymer"),
        "BT8": (
            "3-ETHOXY-6-{2-[1-(6-METHYLPYRIDAZIN-3-YL)PIPERIDIN-4-YL]ETHOXY}-1,2-BENZOXAZOLE",
            "non-polymer",
        ),
        "L37": ("LY374571", "non-polymer"),
        "U11": (
            "METHYL N-{(2S,3R)-3-AMINO-2-HYDROXY-3-[4-(TRIFLUOROMETHYL)PHENYL]PROPANOYL}ALANYLGLYCINATE",
            "non-polymer",
        ),
        "U10": ("UBIQUINONE-10", "non-polymer"),
        "R8E": (
            "3-{5-[(6-AMINO-1H-PYRAZOLO[3,4-B]PYRIDIN-3-YL)METHOXY]-2-CHLOROPHENOXY}-5-CHLOROBENZONITRILE",
            "non-polymer",
        ),
        "GBS": ("4-GUANIDINOBENZOIC ACID", "non-polymer"),
        "U15": (
            "METHYL N-[(2S,3R)-3-AMINO-2-HYDROXY-3-(4-METHYLPHENYL)PROPANOYL]-D-ALANYL-D-LEUCINATE",
            "non-polymer",
        ),
        "N3E": (
            "1-[3-(4-CARBOXYPIPERIDIN-1-YL)-3-DEOXY-BETA-D-ARABINOFURANOSYL]PYRIMIDINE-2,4(1H,3H)-DIONE",
            "non-polymer",
        ),
        "U14": ("3-(5-AMINO-3-IMINO-3H-PYRAZOL-4-YLAZO)-BENZOIC ACID", "non-polymer"),
        "GC1": ("2,6-ANHYDRO-L-GULONIC ACID", "saccharide"),
        "HPS": ("PHENYL DIHYDROGEN PHOSPHATE", "non-polymer"),
        "GC3": (
            "(3AR,5R,6S,7R,7AR)-2,5-BIS(HYDROXYMETHYL)-5,6,7,7A-TETRAHYDRO-3AH-PYRANO[3,2-D][1,3]OXAZOLE-6,7-DIOL",
            "non-polymer",
        ),
        "GC2": ("6-ACETAMIDO-6-DEOXY-CASTANOSPERMINE", "non-polymer"),
        "AF8": (
            "3-HYDROXY-N-{(2S,3R)-3-HYDROXY-4-[(2-METHYLPROPYL){[5-(1,2-OXAZOL-5-YL)THIOPHEN-2-YL]SULFONYL}AMINO]-1-PHENYLBUTAN-2-YL}BENZAMIDE",
            "non-polymer",
        ),
        "GC4": ("4-DEOXY-D-GLUCURONIC ACID", "saccharide"),
        "GC7": ("1-GUANIDINIUM-7-AMINOHEPTANE", "non-polymer"),
        "HPR": ("6-HYDROXY-7,8-DIHYDRO PURINE NUCLEOSIDE", "non-polymer"),
        "AF4": ("2,4-BIS{[4-(N-ISOPORPYL)DIAMINOMETHYL]PHENYL}FURAN", "non-polymer"),
        "T95": (
            "N-(2,6-DIMETHYLPHENYL)-4-(2-ETHOXYPHENOXY)-2-({4-[4-(2-HYDROXYETHYL)PIPERAZIN-1-YL]PHENYL}AMINO)PYRIMIDINE-5-CARBOXAMIDE",
            "non-polymer",
        ),
        "U16": (
            "METHYL N-[(2S,3R)-3-AMINO-2-HYDROXY-3-(4-ISOPROPYLPHENYL)PROPANOYL]-D-ALANYL-D-LEUCINATE",
            "non-polymer",
        ),
        "AF1": (
            "4,6-DIDEOXY-4-{[(1S,4S,5S,6S)-4,5,6-TRIHYDROXY-3-(HYDROXYMETHYL)CYCLOHEX-2-EN-1-YL]AMINO}-BETA-D-GLUCOPYRANOSE",
            "saccharide",
        ),
        "AF2": ("?", "nucleic-acid"),
        "AF3": ("ALUMINUM FLUORIDE", "non-polymer"),
        "AFL": ("ALPHA-L-FUCOSE", "saccharide"),
        "AFN": ("8,9-DIHYDRO-9-HYDROXY-AFLATOXIN B1", "non-polymer"),
        "AFO": ("OCTYL ALPHA-D-ARABINOFURANOSIDE", "saccharide"),
        "AFH": (
            "[(2R,3S,4R,5R)-5-(6-AMINO-9H-PURIN-9-YL)-3,4-DIHYDROXYTETRAHYDROFURAN-2-YL]METHYL (2S,3R,4S,5S,6R)-6-[(1S)-1,2-DIHYDROXYETHYL]-3-FLUORO-4,5-DIHYDROXYTETRAHYDRO-2H-PYRAN-2-YL DIHYDROGEN DIPHOSPHATE",
            "non-polymer",
        ),
        "AFI": (
            "2-[4-(4-CHLOROPHENYL)CYCLOHEXYLIDENE]-3,4-DIHYDROXY-1(2H)-NAPHTHALENONE",
            "non-polymer",
        ),
        "AFJ": (
            "(2R,4S,5S)-N-BUTYL-4-HYDROXY-2,7-DIMETHYL-5-{[N-(4-METHYLPENTANOYL)-L-METHIONYL]AMINO}OCTANAMIDE",
            "non-polymer",
        ),
        "AFD": ("ALPHA-D-ALLOPYRANOSE", "saccharide"),
        "AFE": (
            "5,7-DIHYDROXY-2-(4-HYDROXYPHENYL)-4-OXO-4H-CHROMEN-3-YL 6-DEOXY-ALPHA-L-MANNOPYRANOSIDE",
            "non-polymer",
        ),
        "AFF": ("2-ACETYLAMINOFLUORENE-3-YL", "nucleic-acid"),
        "AFG": ("N-(5'-PHOSPHO-2'-DEOXYGUANOSIN-8-YL)-2-AMINOFLUORENE", "nucleic-acid"),
        "GCM": ("GLYCYLMETHYLENE GROUP", "peptide"),
        "AFA": ("N-[7-METHYL-OCT-2,4-DIENOYL]ASPARAGINE", "peptide"),
        "AFB": (
            "1,6,7,8,9,11A,12,13,14,14A-DECAHYDRO-1,13-DIHYDROXY-6-METHYL-4H-CYCLOPENT[F]OXACYCLOTRIDECIN-4-ONE",
            "non-polymer",
        ),
        "AFC": ("(3R,12R)-3-AMINO-12-METHYLTETRADECANOIC ACID", "peptide"),
        "GCQ": ("GEMCITABINE DIPHOSPHATE", "non-polymer"),
        "GCP": ("PHOSPHOMETHYLPHOSPHONIC ACID GUANYLATE ESTER", "non-polymer"),
        "GCS": ("D-GLUCOSAMINE", "saccharide"),
        "GCR": ("GALLICHROME", "non-polymer"),
        "AFX": (
            "(4S,5S)-4-(6-AMINO-9H-PURIN-9-YL)-3-FLUORO-5-HYDROXY-2-(HYDROXYMETHYL)CYCLOPENT-2-EN-1-ONE",
            "non-polymer",
        ),
        "GCT": ("4,5-DIHYDROXY-TETRAHYDRO-PYRAN-2-CARBOXYLIC ACID", "saccharide"),
        "AFZ": ("2,4-BIS(4-AMIDINOPHENYL)FURAN", "non-polymer"),
        "GCV": ("4-O-METHYL-ALPHA-D-GLUCURONIC ACID", "saccharide"),
        "AFT": ("AFLATOXIN B1", "non-polymer"),
        "AFP": ("ALPHA FRUCTOSE 1,6-DIPHOSPHATE", "saccharide"),
        "AFQ": (
            "6-AMINO-4-{2-[(CYCLOPENTYLMETHYL)AMINO]ETHYL}-2-(METHYLAMINO)-1,7-DIHYDRO-8H-IMIDAZO[4,5-G]QUINAZOLIN-8-ONE",
            "non-polymer",
        ),
        "AFR": ("2-OXO-1,2,DIDEOXY-5F-D-GLUCOPYRANOSE", "saccharide"),
        "BTU": ("3-O-ALPHA-D-GLUCOPYRANOSYL-BETA-D-FRUCTOFURANOSE", "non-polymer"),
        "BTT": ("BENZENE-1,2,4,5-TETROL", "non-polymer"),
        "BTW": ("(2S,3R)-2-BENZYL-3-SULFANYLBUTANOIC ACID", "non-polymer"),
        "BTQ": ("EPI-BIOTIN", "non-polymer"),
        "BTP": ("2-THIOMETHYL-3-PHENYLPROPANOIC ACID", "non-polymer"),
        "BTS": ("3-(2-BENZOTHIAZOLYLTHIO)-1-PROPANESULFONIC ACID", "non-polymer"),
        "BTR": ("6-BROMO-TRYPTOPHAN", "peptide"),
        "IBA": (
            "4-[(2R)-2-(AMINOMETHYL)-2-(HYDROXYMETHYL)-5-OXOPYRROLIDIN-1-YL]-3-[(1-ETHYLPROPYL)AMINO]BENZOIC ACID",
            "non-polymer",
        ),
        "BTY": (
            "4-AMINO-7-METHYLPYRAZOLO[1,5-A][1,3,5]TRIAZIN-2(1H)-ONE",
            "non-polymer",
        ),
        "BTX": (
            "((2R,3S,4R,5R)-5-(6-AMINO-9H-PURIN-9-YL)-3,4-DIHYDROXY-TETRAHYDROFURAN-2-YL)METHYL 5-((3AS,4S,6AR)-2-OXO-HEXAHYDRO-1H-THIENO[3,4-D]IMIDAZOL-4-YL)PENTYL HYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "BTZ": (
            "N-(4-{[AMINO(IMINO)METHYL]AMINO}BUTYL)-2,4'-BI-1,3-THIAZOLE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "BTE": ("2,1,3-BENZOTHIADIAZOL-4-AMINE", "non-polymer"),
        "BTD": ("5-BROMOTHIENYLDEOXYURIDINE", "non-polymer"),
        "BTF": ("?", "non-polymer"),
        "BTA": ("4-DEMETHYL-LEUCINE", "peptide"),
        "K74": (
            "N-METHYL-2-[(4,5,6,7-TETRABROMO-1-METHYL-1H-BENZIMIDAZOL-2-YL)SULFANYL]ACETAMIDE",
            "non-polymer",
        ),
        "BTC": ("CYSTEINE", "peptide"),
        "BTB": (
            "2-[BIS-(2-HYDROXY-ETHYL)-AMINO]-2-HYDROXYMETHYL-PROPANE-1,3-DIOL",
            "non-polymer",
        ),
        "BTM": ("N-BENZYL-N,N-DIETHYLETHANAMINIUM", "non-polymer"),
        "BTL": ("BETAINE ALDEHYDE", "non-polymer"),
        "BTO": (
            "4-[3-(3-NITROPHENYL)-1,2,4-OXADIAZOL-5-YL]BUTANOIC ACID",
            "non-polymer",
        ),
        "BTN": ("BIOTIN", "non-polymer"),
        "BTI": (
            "5-(HEXAHYDRO-2-OXO-1H-THIENO[3,4-D]IMIDAZOL-6-YL)PENTANAL",
            "non-polymer",
        ),
        "BTH": (
            "4-OXO-3-{6-[4-(QUINOXALIN-2-YLAMINO)-BENZOYLAMINO]-2-THIOPHEN-2-YL-HEXANOYLAMINO}-PENTANOIC ACID",
            "non-polymer",
        ),
        "BTK": ("N~6~-BUTANOYL-L-LYSINE", "peptide"),
        "BTJ": (
            "N-[(2-{2-[(N-{(2S,3R)-3-AMINO-4-[4-(BENZYLOXY)PHENYL]-2-HYDROXYBUTANOYL}-L-ALANYL)AMINO]ETHOXY}ETHOXY)ACETYL]-4-BENZOYL-L-PHENYLALANYL-N~6~-HEX-5-YNOYLLYSINAMIDE",
            "non-polymer",
        ),
        "1RB": ("1-ALPHA-D-RIBOFURANOSYL-BENZIMIAZOLE-5'-PHOSPHATE", "non-polymer"),
        "1RC": (
            "(1R)-2-[(5-CHLORO-2,4-DIHYDROXYPHENYL)CARBONYL]-N-ETHYL-2,3-DIHYDRO-1H-ISOINDOLE-1-CARBOXAMIDE",
            "non-polymer",
        ),
        "1RG": (
            "(4R,5S)-3-({(3S,5S)-5-[(3-CARBOXYPHENYL)CARBAMOYL]PYRROLIDIN-3-YL}SULFANYL)-5-[(1S,2R)-1-FORMYL-2-HYDROXYPROPYL]-4-METHYL-4,5-DIHYDRO-1H-PYRROLE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "TER": (
            "N-(3-AMINO-PROPYL)-N-(5-AMINOPROPYL)-1,4-DIAMINOBUTANE",
            "non-polymer",
        ),
        "1RO": (
            "(5E)-5-(QUINOLIN-6-YLMETHYLIDENE)-2-[(THIOPHEN-2-YLMETHYL)AMINO]-1,3-THIAZOL-4(5H)-ONE",
            "non-polymer",
        ),
        "199": (
            "N-{5-[4-(4-METHYLPIPERAZIN-1-YL)PHENYL]-1H-PYRROLO[2,3-B]PYRIDIN-3-YL}NICOTINAMIDE",
            "non-polymer",
        ),
        "WZ1": (
            "METHYL 3-S-ALPHA-D-MANNOPYRANOSYL-3-THIO-ALPHA-D-MANNOPYRANOSIDE",
            "non-polymer",
        ),
        "X8I": (
            "7-(3-HYDROXYPHENYL)-5-METHYL[1,2,4]TRIAZOLO[4,3-A]QUINOLIN-1(2H)-ONE",
            "non-polymer",
        ),
        "KJ2": (
            "4-{4-[4-(3-AMINOPROPOXY)PHENYL]-1H-PYRAZOL-5-YL}-6-CHLOROBENZENE-1,3-DIOL",
            "non-polymer",
        ),
        "VA1": (
            "{5-[(DIFORMYLAMINO)METHYL]-1-BENZOTHIEN-2-YL}BORONIC ACID",
            "non-polymer",
        ),
        "EU3": ("EUROPIUM (III) ION", "non-polymer"),
        "N3F": ("{4-[(2-AMINOPHENYL)AMINO]PHENYL}(PHENYL)METHANONE", "non-polymer"),
        "EUI": (
            "[3,4-BIS(FLUORANYL)-2-[(2-FLUORANYL-4-IODANYL-PHENYL)AMINO]PHENYL]-[3-OXIDANYL-3-[(2S)-PIPERIDIN-2-YL]AZETIDIN-1-YL]METHANONE",
            "non-polymer",
        ),
        "TET": (
            "URIDINE-DIPHOSPHATE-2(N-ACETYLGLUCOSAMINYL-3-FLUORO-2-PHOSPHONOOXY)PROPIONIC ACID",
            "non-polymer",
        ),
        "191": ("PROPIONYL COENZYME A", "non-polymer"),
        "EUG": ("2-METHOXY-4-[(1E)-PROP-1-EN-1-YL]PHENOL", "non-polymer"),
        "RNS": ("L-RHAMNOSE", "saccharide"),
        "NUF": (
            "N,N-BIS[4-[[1-(2-ETHYL)PIPERIDINE]OXY]SALICYLIDENE]-4,5-DIFLUORO-1,2-PHENYLENEDIAMINE-NICKEL (II)",
            "non-polymer",
        ),
        "193": ("(2S)-4-(BETA-ALANYLAMINO)-2-AMINOBUTANOIC ACID", "peptide"),
        "KJZ": (
            "[(2E)-1-{4-[(2-CHLOROPHENYL)AMINO]-4-OXOBUTANOYL}-2-(NAPHTHALEN-1-YLMETHYLIDENE)HYDRAZINO]ACETIC ACID",
            "non-polymer",
        ),
        "TOA": ("3-DEOXY-3-AMINO GLUCOSE", "saccharide"),
        "DU2": (
            "2',5'-DIDEOXY-5'-{[(R)-(1-METHYL-1H-IMIDAZOL-2-YL)(PHENYL)METHYL]AMINO}URIDINE",
            "non-polymer",
        ),
        "EUR": (
            "4-CHLORANYL-N-[[4-CHLORANYL-3-(3-CHLORANYL-5-CYANO-PHENOXY)-2-FLUORANYL-PHENYL]METHYL]-2-(HYDROXYMETHYL)-1H-IMIDAZOLE-5-CARBOXAMIDE",
            "non-polymer",
        ),
        "MBF": ("2-DEOXY-2-FLUORO-BETA-D-MANNOSE", "non-polymer"),
        "SWI": ("SWINHOLIDE A", "non-polymer"),
        "DU3": ("2',5'-DIDEOXY-5'-[(DIPHENYLMETHYL)AMINO]URIDINE", "non-polymer"),
        "DU4": (
            "2',5'-DIDEOXY-5'-[(DIPHENYLMETHYL)(METHYL)AMINO]URIDINE",
            "non-polymer",
        ),
        "XIF": ("PIPERIDINE-3,4-DIOL", "non-polymer"),
        "5IC": ("5-IODO-CYTIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "5ID": (
            "(2R,3R,4S,5R)-2-(4-AMINO-5-IODO-7H-PYRROLO[2,3-D]PYRIMIDIN-7-YL)-5-(HYDROXYMETHYL)TETRAHYDROFURAN-3,4-DIOL",
            "non-polymer",
        ),
        "5IG": (
            "6-(2,4-DIAMINO-6-ETHYLPYRIMIDIN-5-YL)-4-(3-METHOXYPROPYL)-2,2-DIMETHYL-2H-1,4-BENZOXAZIN-3(4H)-ONE",
            "non-polymer",
        ),
        "5IO": (
            "6-ETHYL-5-IODO-2-{5-[(4-METHYLPIPERAZIN-1-YL)SULFONYL]-2-PROPOXYPHENYL}PYRIMIDIN-4(3H)-ONE",
            "non-polymer",
        ),
        "5IN": (
            "N-(BENZYLSULFONYL)-L-SERYL-N~1~-{4-[AMINO(IMINO)METHYL]BENZYL}-O-BENZYL-L-SERINAMIDE",
            "non-polymer",
        ),
        "5IQ": ("ISOQUINOLIN-5-AMINE", "non-polymer"),
        "5IP": (
            "(1R,2R,3R,4R,5S,6S)-6-HYDROXYCYCLOHEXANE-1,2,3,4,5-PENTAYL PENTAKIS[DIHYDROGEN (PHOSPHATE)]",
            "non-polymer",
        ),
        "5IU": ("5-IODO-2'-DEOXYURIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "5IT": ("5-IODO-THYMIDINE-5'-PHOSPHATE", "nucleic-acid"),
        "QGF": (
            "N-[(2S,4S)-1-({4-[(2,4-DICHLOROBENZYL)OXY]PIPERIDIN-1-YL}SULFONYL)-4-(5-FLUOROPYRIMIDIN-2-YL)-2-METHYLPENTAN-2-YL]-N-HYDROXYFORMAMIDE",
            "non-polymer",
        ),
        "CV1": ("8-BROMO-CYCLIC-ADP-RIBOSE", "non-polymer"),
        "RYZ": (
            "N-(2,2,3,3,4,4,4-HEPTAFLUOROBUTYL)-4-SULFAMOYLBENZAMIDE",
            "non-polymer",
        ),
        "DUM": ("DUMMY ATOMS", "non-polymer"),
        "DUN": (
            "2,4(1H,3H)-PYRIMIDINEDIONE, 1-[2-DEOXY-5-O-[HYDROXY(PHOSPHONOAMINO)PHOSPHINYL]-BETA-D-ERYTHRO-PENTOFURANOSYL]-",
            "non-polymer",
        ),
        "CW1": ("COTYLENIN A", "non-polymer"),
        "DUO": (
            "4-HYDROXY-2,8-DIMETHYL-1-OXO-6-(4,5,6-TRIMETHOXY-1H-INDOLE-2-CARBONYL)-1,2,3,6,7,8-HEXAHYDRO-3,6-DIAZA-AS-INDACENE-2-CARBOXYLIC ACID METHYL ESTER",
            "non-polymer",
        ),
        "CW7": ("COTYLENIN A", "non-polymer"),
        "4U7": (
            "[(5S)-5-AMINO-5-CARBOXY-6,6-DIFLUOROHEXYL](TRIHYDROXY)BORATE(1-)",
            "peptide",
        ),
        "OIS": (
            "(S)-N-((2S,3R)-3-HYDROXY-4-(4-((E)-(HYDROXYIMINO)METHYL)-N-ISOBUTYLPHENYLSULFONAMIDO)-1-PHENYLBUTAN-2-YL)-3-METHYL-2-(3-((2-METHYLTHIAZOL-4-YL)METHYL)-2-OXOIMIDAZOLIDIN-1-YL)BUTANAMIDE",
            "non-polymer",
        ),
        "NU7": (
            "N-{[1-METHYL-4-(4-PHENYL-1,3-THIAZOL-2-YL)PIPERIDIN-4-YL]METHYL}-3-[5-(TRIFLUOROMETHYL)-1,2,4-OXADIAZOL-3-YL]BENZAMIDE",
            "non-polymer",
        ),
        "RE2": ("5-[2-(4-HYDROXYPHENYL)ETHYL]BENZENE-1,3-DIOL", "non-polymer"),
        "DUC": ("DIHYDROPYRIMIDINE-2,4(1H,3H)-DIONE", "non-polymer"),
        "4UP": (
            "6-[(2S,3S)-3-PHENYLOXIRAN-2-YL]NAPHTHALENE-2-CARBOXIMIDAMIDE",
            "non-polymer",
        ),
        "CWR": ("(4-METHYL-5-OXO-2,5-DIHYDRO-1H-IMIDAZOL-1-YL)ACETIC ACID", "peptide"),
        "HFP": ("ALPHA-HYDROXYFARNESYLPHOSPHONIC ACID", "non-polymer"),
        "CWT": (
            "(5Z)-5-(1,3-BENZODIOXOL-5-YLMETHYLIDENE)-3-METHYL-2-(PROPAN-2-YLAMINO)IMIDAZOL-4-ONE",
            "non-polymer",
        ),
        "OIR": ("N-(3-PHENYL-2-SULFANYLPROPANOYL)PHENYLALANYLALANINE", "non-polymer"),
        "KBI": (
            "TRANS-4-({9-[(3S)-TETRAHYDROFURAN-3-YL]-8-[(2,4,6-TRIFLUOROPHENYL)AMINO]-9H-PURIN-2-YL}AMINO)CYCLOHEXANOL",
            "non-polymer",
        ),
        "4UN": (
            "{4-[2,2-BIS(5-METHYL-1,2,4-OXADIAZOL-3-YL)-3-PHENYLPROPYL]PHENYL}SULFAMIC ACID",
            "non-polymer",
        ),
        "TB2": ("(4-TERT-BUTYLPHENYL)ACETALDEHYDE", "non-polymer"),
        "4UB": ("3-PHENYL-6-(1H-PYRAZOL-4-YL)IMIDAZO[1,2-A]PYRAZINE", "non-polymer"),
        "CWB": (
            "1-METHYL-N-[(1R,5S)-9-METHYL-9-AZABICYCLO[3.3.1]NONAN-3-YL]INDAZOLE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "CWD": (
            "3-(5-CHLORO-2,4-DIOXO-3,4-DIHYDROPYRIMIDIN-1(2H)-YL)-L-ALANINE",
            "peptide",
        ),
        "HC7": (
            "(2S)-2-[4'-({DIMETHYL[2-(PHOSPHONOOXY)ETHYL]AMMONIO}ACETYL)BIPHENYL-4-YL]-2-HYDROXY-4,4-DIMETHYLMORPHOLIN-4-IUM",
            "non-polymer",
        ),
        "O3N": (
            "(2S)-2-(2-CARBOXYETHYL)-6-{[{2-[(CYCLOHEXYLMETHYL)CARBAMOYL]BENZYL}(PROP-2-EN-1-YL)AMINO]METHYL}-2,3-DIHYDRO-1,4-BENZODIOXINE-5-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "VAK": (
            "METHYL (1R,2R,4S)-2-ETHYL-2,4,5,7-TETRAHYDROXY-6,11-DIOXO-1,2,3,4,6,11-HEXAHYDROTETRACENE-1-CARBOXYLATE",
            "non-polymer",
        ),
        "V12": (
            "5'-[(1H-BENZIMIDAZOL-2-YLACETYL)AMINO]-5'-DEOXYCYTIDINE",
            "non-polymer",
        ),
        "HFS": ("1-(1-HYDROXY-5-ISOQUINOLINESULFONYL)HOMOPIPERAZINE", "non-polymer"),
        "KBM": (
            "(4-CHLOROPHENYL)(5-METHOXY-1-BENZOFURAN-2-YL)METHANONE",
            "non-polymer",
        ),
        "V15": ("2-AMINO-6-(2-PHENYLETHYL)PYRIMIDIN-4(3H)-ONE", "non-polymer"),
        "5I5": (
            "7-(5-DEOXY-BETA-D-RIBOFURANOSYL)-5-IODO-7H-PYRROLO[2,3-D]PYRIMIDIN-4-AMINE",
            "non-polymer",
        ),
        "RYY": ("N-PROPYL-4-SULFAMOYLBENZAMIDE", "non-polymer"),
        "HC8": (
            "1-(3-ACETYL-4,5-DIHYDROXY-7-METHOXYNAPHTHALEN-2-YL)PROPAN-2-ONE",
            "non-polymer",
        ),
        "3CM": (
            "3-CYCLOHEXYLPROPYL 4-O-ALPHA-D-GLUCOPYRANOSYL-BETA-D-GLUCOPYRANOSIDE",
            "saccharide",
        ),
        "3CL": ("3-CHLOROPROPANOL", "non-polymer"),
        "3CO": ("COBALT (III) ION", "non-polymer"),
        "3CN": ("3-AMINOPROPANE", "non-polymer"),
        "NCL": (
            "3,6-BIS{3-(3-[(3R)-METHYLPIPERIDINO)]PROPIONAMIDO}ACRIDINE",
            "non-polymer",
        ),
        "3CH": ("3-CHLOROPHENOL", "non-polymer"),
        "3CE": ("3-CHLOROBENZENE-1,2-DIOL", "non-polymer"),
        "3CD": ("3-(CHLOROACETYL) PYRIDINE ADENINE DINUCLEOTIDE", "non-polymer"),
        "3CF": ("3-CYANO-L-PHENYLALANINE", "peptide"),
        "3CA": ("?", "non-polymer"),
        "3CC": (
            "N-[(2R)-5-(AMINOSULFONYL)-2,3-DIHYDRO-1H-INDEN-2-YL]-2-PROPYLPENTANAMIDE",
            "non-polymer",
        ),
        "3CB": (
            "N-[(3-CARBOXYBENZYL)SULFONYL]ISOLEUCYL-N~1~-{4-[AMINO(IMINO)METHYL]BENZYL}-5-IMINOORNITHINAMIDE",
            "non-polymer",
        ),
        "3CY": (
            "(3S)-3-[({(2S)-5-[(N-ACETYL-L-ALPHA-ASPARTYL)AMINO]-4-OXO-1,2,4,5,6,7-HEXAHYDROAZEPINO[3,2,1-HI]INDOL-2-YL}CARBONYL)AMINO]-5-(BENZYLSULFANYL)-4-OXOPENTANOIC ACID",
            "non-polymer",
        ),
        "3CX": (
            "(2S)-3-(CYCLOHEXYLAMINO)-2-HYDROXYPROPANE-1-SULFONIC ACID",
            "non-polymer",
        ),
        "3CZ": (
            "(2R)-1-[(4-TERT-BUTYLPHENYL)SULFONYL]-2-METHYL-4-(4-NITROPHENYL)PIPERAZINE",
            "non-polymer",
        ),
        "3CU": ("CASUARINE", "non-polymer"),
        "ZFB": (
            "(3R)-3-{[(BENZYLOXY)CARBONYL]AMINO}-2-OXO-4-PHENYLBUTANE-1-DIAZONIUM",
            "peptide",
        ),
        "RE1": ("GLYCYL-L-A-AMINOPIMELYL-E-(D-2-AMINOETHYL)PHOSPHONATE", "non-polymer"),
        "3CP": ("3-CARBOXYPROPYL-COENZYME A", "non-polymer"),
        "3CS": (
            "3-[3-(3,3-DIMETHYLBUTANOYL)-1-(4-IODOBENZYL)-5-(QUINOLIN-2-YLMETHOXY)-1H-INDOL-2-YL]-2,2-DIMETHYLPROPANOIC ACID",
            "non-polymer",
        ),
        "542": (
            "N-[(E)-3-[(2R,3S,4R,5R)-3,4-DIHYDROXY-5-(6-METHYLAMINOPURIN-9-YL)OXOLAN-2-YL]PROP-2-ENYL]-5-(4-FLUOROPHENYL)-2,3-DIHYDROXY-BENZAMIDE",
            "non-polymer",
        ),
        "543": ("CALCIUM ION, 6 WATERS PLUS ETHANOL COORDINATED", "non-polymer"),
        "540": (
            "8-FLUORO-6-METHOXY-3,4-DIMETHYL-1-(3-METHYLPYRIDIN-4-YL)IMIDAZO[1,5-A]QUINOXALINE",
            "non-polymer",
        ),
        "541": (
            "(2R)-N-HYDROXY-2-[(3S)-3-METHYL-3-{4-[(2-METHYLQUINOLIN-4-YL)METHOXY]PHENYL}-2-OXOPYRROLIDIN-1-YL]PROPANAMIDE",
            "non-polymer",
        ),
        "546": (
            "2-METHOXY-6,7-DIMETHYL-9-(4-METHYLPYRIDIN-3-YL)IMIDAZO[1,5-A]PYRIDO[3,2-E]PYRAZINE",
            "non-polymer",
        ),
        "547": (
            "2,3-DIPHENYL-N-(2-PIPERAZIN-1-YLETHYL)FURO[2,3-B]PYRIDIN-4-AMINE",
            "non-polymer",
        ),
        "544": (
            "2-(1-METHYL-3-OXO-3-PHENYL-PROPYLAMINO)-3-{4-[2-(5-METHYL-2-PHENYL-OXAZOL-4-YL)-ETHOXY]-PHENYL}-PROPIONIC ACID",
            "non-polymer",
        ),
        "545": (
            "[2-AMINO-3-(4-HYDROXY-PHENYL)-PROPIONYLAMINO]-(1,3,4,5-TETRAHYDROXY-4-HYDROXYMETHYL-PIPERIDIN-2-YL)- ACETIC ACID BUTYL ESTER",
            "non-polymer",
        ),
        "548": (
            "6-CHLORO-3,4-DIMETHYL-1-(3-METHYLPYRIDIN-4-YL)-8-(TRIFLUOROMETHYL)IMIDAZO[1,5-A]QUINOXALINE",
            "non-polymer",
        ),
        "L53": (
            "4-[4-(4-AMINOPHENYL)BUTA-1,3-DIYN-1-YL]-N-[(2S,3S)-3-HYDROXY-1-NITROSO-1-OXOBUTAN-2-YL]BENZAMIDE",
            "non-polymer",
        ),
        "RNB": (
            "3-CHLORO-5-[(2E)-3,7-DIMETHYLOCTA-2,6-DIEN-1-YL]-4,6-DIHYDROXY-2-METHYLBENZALDEHYDE",
            "non-polymer",
        ),
        "54F": ("3-(PYRIDIN-2-YLOXY)ANILINE", "non-polymer"),
        "54D": ("METHYL THIOPHENE-2-CARBOXYLATE", "non-polymer"),
        "UE1": ("MC-1 (OPEN FORM)", "non-polymer"),
        "54H": ("5'-O-(D-VALYLSULFAMOYL)ADENOSINE", "non-polymer"),
        "U1K": (
            "(S)-2-AMINO-3-(3-HYDROXY-5-TERT-BUTYLISOTHIAZOL-4-YL) PROPRIONIC ACID",
            "non-polymer",
        ),
        "XK2": (
            "[4R-(4ALPHA,5ALPHA,6BETA,7BETA)]-HEXAHYDRO-5,6-DIHYDROXY-1,3-BIS[2-NAPHTHYL-METHYL]-4,7-BIS(PHENYLMETHYL)-2H-1,3-DIAZEPIN-2-ONE",
            "non-polymer",
        ),
        "XK3": (
            "4-[3-AMINO-6-(3-ETHYLTHIOPHEN-2-YL)PYRAZIN-2-YL]CYCLOHEXANE-1-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "U1N": (
            "4-[(3R)-3-{[2-(4-FLUOROPHENYL)-2-OXOETHYL]AMINO}BUTYL]BENZAMIDE",
            "non-polymer",
        ),
        "WDR": (
            "(4S)-6-(3,5-DIMETHYL-1,2-OXAZOL-4-YL)-4-(2-HYDROXYETHOXY)-3-METHYL-3,4-DIHYDROQUINAZOLIN-2(1H)-ONE",
            "non-polymer",
        ),
        "WAI": (
            "4-{2-[4-(2-AMINOETHYL)PIPERAZIN-1-YL]PYRIDIN-4-YL}-N-(3-CHLORO-4-METHYLPHENYL)PYRIMIDIN-2-AMINE",
            "non-polymer",
        ),
        "PCV": (
            "5-AMINO-3-HYDROXY-2-(2-OXO-AZETIDIN-1-YL)-PENTANOIC ACID",
            "non-polymer",
        ),
        "3C4": ("(3-CHLORO-4-HYDROXYPHENYL)ACETIC ACID", "non-polymer"),
        "3C3": (
            "4,4'-(1-PROPYL-1H-1,2,4-TRIAZOLE-3,5-DIYL)BIS(2,5-DIHYDRO-1,2,5-OXADIAZOL-3-AMINE)",
            "non-polymer",
        ),
        "0A0": ("2-METHYL-L-ASPARTIC ACID", "peptide"),
        "0A1": ("O-METHYL-L-TYROSINE", "peptide"),
        "0A2": (
            "[(1R)-1,5-DIAMINOPENTYL][BIS(ETHANOLATO)]HYDROXYBORATE(1-)",
            "peptide",
        ),
        "0A3": (
            "(2R)-2-[(R)-{[(2Z)-2-(2-AMINO-1,3-THIAZOL-4-YL)-2-(METHOXYIMINO)ACETYL]AMINO}(CARBOXY)METHYL]-5-METHYL-3,6-DIHYDRO-2H-1,3-THIAZINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "0A4": (
            "(2R)-2-{(R)-CARBOXY[(THIOPHEN-2-YLACETYL)AMINO]METHYL}-5-METHYLIDENE-5,6-DIHYDRO-2H-1,3-THIAZINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "0A5": ("4-BROMOBUTANOIC ACID", "non-polymer"),
        "0A6": ("(3R)-4-ETHOXY-3-HYDROXY-4-OXOBUTANOIC ACID", "non-polymer"),
        "0A7": ("2-(CARBOXYMETHYL)BENZOIC ACID", "non-polymer"),
        "0A8": ("S-[(2-CHLOROETHYL)CARBAMOYL]-L-CYSTEINE", "peptide"),
        "0A9": ("METHYL L-PHENYLALANINATE", "peptide"),
        "1MS": (
            "N-[(4-METHYLPYRIMIDIN-2-YL)CARBAMOYL]-2-NITROBENZENESULFONAMIDE",
            "non-polymer",
        ),
        "1MR": ("N-METHYLANILINE", "non-polymer"),
        "L10": (
            "N-[(3Z)-5-TERT-BUTYL-2-PHENYL-1,2-DIHYDRO-3H-PYRAZOL-3-YLIDENE]-N'-(4-CHLOROPHENYL)UREA",
            "non-polymer",
        ),
        "L11": (
            "N-[4-CHLORO-3-(PYRIDIN-3-YLOXYMETHYL)-PHENYL]-3-FLUORO-",
            "non-polymer",
        ),
        "L12": ("3-(2-PYRIDIN-4-YLETHYL)-1H-INDOLE", "non-polymer"),
        "7JA": (
            "N-({(1R,2S)-3-OXO-2-[(2Z)-PENT-2-EN-1-YL]CYCLOPENTYL}ACETYL)-L-ISOLEUCINE",
            "peptide",
        ),
        "D02": (
            "METHYL 1-(2,3,5,6-TETRAFLUORO-4-SULFAMOYLPHENYL)-1H-1,2,3-TRIAZOLE-4-CARBOXYLATE",
            "non-polymer",
        ),
        "1MM": (
            "METHYL 2-[({[(4-METHOXY-6-METHYL-1,3,5-TRIAZIN-2-YL)AMINO]CARBONYL}AMINO)SULFONYL]BENZOATE",
            "non-polymer",
        ),
        "OEU": (
            "N-{(2S)-1-[(2-CHLOROBENZYL)AMINO]-1-OXO-4-PHENYLBUTAN-2-YL}-N~2~-[3-(2-METHYLPHENYL)PROPANOYL]-L-THREONINAMIDE",
            "peptide",
        ),
        "1MA": ("6-HYDRO-1-METHYLADENOSINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "HCV": (
            "N-[(5S)-5-AMINO-5-CARBOXYPENTANOYL]-L-HOMOCYSTEYL-D-VALINE",
            "non-polymer",
        ),
        "1MC": ("1-METHYLCYTOSINE", "non-polymer"),
        "1ME": ("(2S)-2-AMINO-3-METHYLSULFANYL-BUTANOIC ACID", "peptide"),
        "1MG": ("1N-METHYLGUANOSINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "NFB": ("NI-FE ACTIVE CENTER B-FORM", "non-polymer"),
        "XDE": (
            "4-[(1S)-1-HYDROXY-2-(METHYLAMINO)ETHYL]BENZENE-1,2-DIOL",
            "non-polymer",
        ),
        "NFA": ("PHENYLALANINE AMIDE", "peptide"),
        "XDN": ("PIPERIDINE-3,4,5-TRIOL", "non-polymer"),
        "XDL": ("XYLOSE-DERIVED ISOFAGOMINE LACTAM", "non-polymer"),
        "XDK": ("2-AMINO-4-PYRIDYL-PYRIMIDINE", "non-polymer"),
        "XDI": ("4,4'-PROPANE-2,2-DIYLBIS(2,6-DIBROMOPHENOL)", "non-polymer"),
        "XDH": ("4,4'-PROPANE-2,2-DIYLBIS(2,6-DICHLOROPHENOL)", "non-polymer"),
        "NFO": ("NI-FE OXIDIZED ACTIVE CENTER", "non-polymer"),
        "XDR": (
            "HEXYL (5S,6R,8R)-6-HYDROXY-5-METHYL-13-OXO-5,6,7,8-TETRAHYDRO-13H-5,8-EPOXY-4B,8A,14-TRIAZADIBENZO[B,H]CYCLOOCTA[1,2,3,4-JKL]CYCLOPENTA[E]-AS-INDACENE-6-CARBOXYLATE",
            "non-polymer",
        ),
        "368": (
            "(1R)-2-{[AMINO(IMINO)METHYL]AMINO}-1-{4-[(4R)-4-(HYDROXYMETHYL)-1,3,2-DIOXABOROLAN-2-YL]PHENYL}ETHYL NICOTINATE",
            "non-polymer",
        ),
        "369": (
            "5-(4-HYDROXYPHENOXY)-6-(3-HYDROXYPHENYL)-7-METHYLNAPHTHALEN-2-OL",
            "non-polymer",
        ),
        "367": (
            "3-HYDROXYPROPYL 3-[({7-[AMINO(IMINO)METHYL]-1-NAPHTHYL}AMINO)CARBONYL]BENZENESULFONATE",
            "non-polymer",
        ),
        "364": (
            "3-BIPHENYL-3-YL-1-(2-HYDROXY-2,2-DIPHOSPHONOETHYL)PYRIDINIUM",
            "non-polymer",
        ),
        "K": ("POTASSIUM ION", "non-polymer"),
        "361": (
            "(7R,8R)-8-(2,4,5-TRIFLUOROPHENYL)-6,7,8,9-TETRAHYDROIMIDAZO[1,2-A:4,5-C']DIPYRIDIN-7-AMINE",
            "non-polymer",
        ),
        "HCY": (
            "(11ALPHA,14BETA)-11,17,21-TRIHYDROXYPREGN-4-ENE-3,20-DIONE",
            "non-polymer",
        ),
        "OHM": ("3-OXO-N-[(3S)-2-OXOPYRROLIDIN-3-YL]DODECANAMIDE", "non-polymer"),
        "36F": (
            "5-METHYL-3-[4-(4-{5-[(3AS,4S,6AR)-2-OXOHEXAHYDRO-1H-THIENO[3,4-D]IMIDAZOL-4-YL]PENTYL}-1H-1,2,3-TRIAZOL-1-YL)BUTYL]-1,3-BENZOXAZOL-2(3H)-ONE",
            "non-polymer",
        ),
        "TMQ": ("TRIMETREXATE", "non-polymer"),
        "OHN": ("N-3-OXO-DODECANOYL-L-HOMOSERINE LACTONE", "non-polymer"),
        "36A": (
            "(S)-2-AMINO-4-(((2S,3S,4R,5R)-5-(6-(3-CHLOROBENZYLAMINO)-9H-PURIN-9-YL)-3,4-DIHYDROXYTETRAHYDROFURAN-2-YL)METHYLTHIO)BUTANOIC ACID",
            "non-polymer",
        ),
        "XD6": (
            "(5Z)-13-CHLORO-14,16-DIHYDROXY-1,11-DIOXO-3,4,7,8,9,10,11,12-OCTAHYDRO-1H-2-BENZOXACYCLOTETRADECINE-6-CARBALDEHYDE",
            "non-polymer",
        ),
        "UIB": (
            "(1R,3AS,4R,8AS,8BR)-4-{5-(PHENYL[1,3]DIOXOL-5-YLMETHYL)-4-ETHYL-2,3,3-TRIMETHYL-6-OXO-OCTAHYDRO-PYRROLO[3,4-C]PYRROL-1-YL}-BENZAMIDINE",
            "non-polymer",
        ),
        "XD2": (
            "(2R)-2-[(1R)-1-{[(2R)-2-HYDROXY-2-PHENYLACETYL]AMINO}-2-OXOETHYL]-5-{[(1-METHYL-1H-TETRAZOL-5-YL)SULFANYL]METHYL}-3,6-DIHYDRO-2H-1,3-THIAZINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "VAC": (
            "N,N-BIS(2-HYDROXY-1-INDANYL)-2,6- DIPHENYLMETHYL-4-HYDROXY-1,7-HEPTANDIAMIDE",
            "non-polymer",
        ),
        "NFH": (
            "N~2~-[(2R)-2-{[FORMYL(HYDROXY)AMINO]METHYL}-4-METHYLPENTANOYL]-N,3-DIMETHYL-L-VALINAMIDE",
            "non-polymer",
        ),
        "XD9": (
            "(4R,6R,7S)-4,6,7-TRIHYDROXY-2-[(1E)-PROP-1-EN-1-YL]-4,5,6,7-TETRAHYDRO-1-BENZOTHIOPHENE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "0AP": ("?", "nucleic-acid"),
        "HCB": ("5-HYDROXYBENZIMIDAZOLYLCOBAMIDE", "non-polymer"),
        "0AR": (
            "N-[N-[(4S)-4-AZANYL-5-HYDROXY-5-OXO-PENTYL]CARBAMIMIDOYL]ETHANAMIDE",
            "peptide",
        ),
        "0AS": ("ASIATIC ACID", "non-polymer"),
        "0AT": (
            "2-(ACETYLAMINO)-2-DEOXY-6-O-PHOSPHONO-ALPHA-D-GLUCOPYRANOSE",
            "saccharide",
        ),
        "0AU": ("?", "nucleic-acid"),
        "0AV": ("?", "nucleic-acid"),
        "HCC": ("2',4,4'-TRIHYDROXYCHALCONE", "non-polymer"),
        "0AY": ("DIETHYL [(1R)-1,5-DIAMINOPENTYL]BORONATE", "peptide"),
        "0AZ": ("(4R)-4-HYDROXY-D-PROLINE", "peptide"),
        "L1P": ("3-PHOSPHORYL-[1,2-DI-PHYTANYL]GLYCEROL", "non-polymer"),
        "NFU": ("FORMYL[BIS(HYDROCYANATO-1KAPPAC)]IRONNICKEL(FE-NI)", "non-polymer"),
        "L1R": (
            "4-(2-AMINOETHOXY)-3,5-DICHLORO-N-[3-(1-METHYLETHOXY)PHENYL]BENZAMIDE",
            "non-polymer",
        ),
        "0AA": ("METHYL L-VALINATE", "peptide"),
        "0AB": ("(3S,4S)-3-AMINO-4-METHYL-3,4-DIHYDRO-2H-PYRAN-2-ONE", "peptide"),
        "0AC": ("(4S,5S)-5-AMINO-4-HYDROXY-6-PHENYLHEXANOIC ACID", "peptide"),
        "0AD": ("?", "nucleic-acid"),
        "0AF": ("7-HYDROXY-L-TRYPTOPHAN", "peptide"),
        "0AG": ("N-(ETHOXYCARBONYL)-L-LEUCINE", "peptide"),
        "0AH": ("O-(BROMOACETYL)-L-SERINE", "peptide"),
        "0AI": (
            "METHYL 2-(ACETYLAMINO)-2-DEOXY-ALPHA-D-GALACTOPYRANOSIDURONIC ACID",
            "non-polymer",
        ),
        "0AJ": ("(1R)-1-(ETHYLSULFANYL)-1-(METHYLSULFANYL)ETHANE", "non-polymer"),
        "0AK": ("(2S)-2-AMINO-4-(2-CHLOROETHOXY)-4-OXOBUTANOIC ACID", "peptide"),
        "0AL": ("4-O-PHOSPHONO-D-GLYCERO-BETA-D-MANNO-HEPTOPYRANOSE", "non-polymer"),
        "0AM": ("?", "nucleic-acid"),
        "L1B": (
            "7-(2,4-DIAMINO-6-ETHYLPYRIMIDIN-5-YL)-1-(3-METHOXYPROPYL)QUINOLINIUM",
            "non-polymer",
        ),
        "0AO": ("(2S)-2-AMINO-3-NAPHTHALEN-1-YLPROPANAL", "non-polymer"),
        "N3C": (
            "4-{4-[(4'-CHLOROBIPHENYL-2-YL)METHYL]PIPERAZIN-1-YL}-N-{[4-({(1R)-3-(DIMETHYLAMINO)-1-[(PHENYLTHIO)METHYL]PROPYL}AMINO)-3-NITROPHENYL]SULFONYL}BENZAMIDE",
            "non-polymer",
        ),
        "MR6": ("2-(3,5-DIMETHYLPHENYL)-1,3-BENZOXAZOLE", "non-polymer"),
        "HCD": ("(3ALPHA,8ALPHA)-CHOLEST-5-ENE-3,20-DIOL", "non-polymer"),
        "SWG": (
            "2-[(4Z)-2-[(1R)-1-AMINO-2-HYDROXY-ETHYL]-4-(1H-INDOL-3-YLMETHYLIDENE)-5-OXO-IMIDAZOL-1-YL]ETHANOIC ACID",
            "peptide",
        ),
        "SK4": (
            "4-[(3-METHOXYPHENYL)AMINO]-6-(METHYLSULFONYL)QUINOLINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "OHA": ("10-OXOHEXADECANOIC ACID", "non-polymer"),
        "KB2": (
            "(3S)-3-(4-{(1S)-1,2-DIMETHYL-1-[(QUINOLIN-6-YLMETHYL)AMINO]PROPYL}-1H-1,2,3-TRIAZOL-1-YL)HEPTAN-2-ONE",
            "non-polymer",
        ),
        "ZK9": (
            "[3-AMINOMETHYLPHENYL-OXY-]-[BENZAMIDIN-3-YLOXY]-3,5-DIFLUORO-4-METHYL-PYRIDINE",
            "non-polymer",
        ),
        "HFM": ("2-FORMYL-PROTOPORPHRYN IX", "non-polymer"),
        "UIN": (
            "5,6,7,8,9,10-HEXAHYDRO-4-HYDROXY-3-(1-PHENYLPROPYL)CYCLOOCTA[B]PYRAN-2-ONE",
            "non-polymer",
        ),
        "N8E": ("3,6,9,12,15-PENTAOXATRICOSAN-1-OL", "non-polymer"),
        "S19": (
            "2-{[(3R,4R)-3-AMINOTETRAHYDRO-2H-PYRAN-4-YL]AMINO}-4-[(4-METHYLPHENYL)AMINO]PYRIMIDINE-5-CARBOXAMIDE",
            "non-polymer",
        ),
        "HCI": ("HYDROCINNAMIC ACID", "non-polymer"),
        "SWF": ("S-WARFARIN", "non-polymer"),
        "VRV": (
            "6-(5-BROMO-2-HYDROXYPHENYL)-2-OXO-4-PHENYL-1,2-DIHYDROPYRIDINE-3-CARBONITRILE",
            "non-polymer",
        ),
        "RX8": (
            "1-[4-AMINO-2-(ETHOXYMETHYL)-1H-IMIDAZO[4,5-C]QUINOLIN-1-YL]-2-METHYLPROPAN-2-OL",
            "non-polymer",
        ),
        "S13": (
            "(3S,5R,7R,8S,9S,10R)-7-(HYDROXYMETHYL)-3-(4-METHYLPHENYL)-1,6-DIOXA-2-AZASPIRO[4.5]DECANE-8,9,10-TRIOL",
            "non-polymer",
        ),
        "HCN": (
            "2 IRON/2 SULFUR/3 CARBONYL/2 CYANIDE/WATER/METHYLETHER CLUSTER",
            "non-polymer",
        ),
        "S11": (
            "5-(3,5-DICHLOROPHENYL)THIO-4-ISOPROPYL-1-(PYRIDIN-4-YL-METHYL)-1H-IMIDAZOL-2-YL-METHYL CARBAMATE",
            "non-polymer",
        ),
        "S10": ("DIETHYL PROPANE-1,3-DIYLBISCARBAMATE", "non-polymer"),
        "S17": (
            "METHYL 4-BROMO-N-[8-(HYDROXYAMINO)-8-OXOOCTANOYL]-L-PHENYLALANINATE",
            "non-polymer",
        ),
        "TVP": ("(2S)-1-ACETYL-2-METHYL-1,2,3,4-TETRAHYDROQUINOLINE", "non-polymer"),
        "NFZ": ("NITROFURAZONE", "non-polymer"),
        "VRB": ("PHYCOVIOLOBILIN, DOUBLE BOUND FORM", "non-polymer"),
        "VRA": (
            "N-{3,4-DIFLUORO-2-[(2-FLUORO-4-IODOPHENYL)AMINO]-6-METHOXYPHENYL}-1-[(2S)-2,3-DIHYDROXYPROPYL]CYCLOPROPANESULFONAMIDE",
            "non-polymer",
        ),
        "IOL": ("4-IODOPHENOL", "non-polymer"),
        "HFL": ("5-(3,3-DIHYDROXYPROPENY)-3-METHOXY-BENZENE-1,2-DIOL", "non-polymer"),
        "JZ9": ("4,5,6,7-TETRAHYDRO-1H-INDOLE", "non-polymer"),
        "JZ8": (
            "5-METHYL-7-(NAPHTHALEN-2-YLAMINO)-1H-[1,2,4]TRIAZOLO[1,5-A]PYRIMIDINE-3,8-DIIUM",
            "non-polymer",
        ),
        "JZ7": ("(2S)-2-ETHOXY-3,4-DIHYDRO-2H-PYRAN", "non-polymer"),
        "IOK": (
            "N-[(1R)-3-(4-HYDROXYPHENYL)-1-METHYLPROPYL]-2-(2-PHENYL-1H-INDOL-3-YL)ACETAMIDE",
            "non-polymer",
        ),
        "JZ5": ("(NZ)-N-(THIOPHEN-2-YLMETHYLIDENE)HYDROXYLAMINE", "non-polymer"),
        "JZ4": ("2-PROPYLPHENOL", "non-polymer"),
        "JZ3": ("GUAIACOL", "non-polymer"),
        "JZ2": ("5-CHLORO-2-METHYLPHENOL", "non-polymer"),
        "JZ1": ("2-ETHYLPHENOL", "non-polymer"),
        "JZ0": ("O-CRESOL", "non-polymer"),
        "JZO": (
            "4-[3-(METHOXYMETHYL)PHENYL]-1,2-DIMETHYL-5-QUINOXALIN-6-YL-1,2-DIHYDRO-3H-PYRAZOL-3-ONE",
            "non-polymer",
        ),
        "JZN": (
            "1-{[3-(4-{[(2R)-4-(5-FLUORO-2-METHOXYPHENYL)-2-HYDROXY-4-METHYL-2-(TRIFLUOROMETHYL)PENTYL]AMINO}-6-METHYL-1H-INDAZOL-1-YL)PHENYL]CARBONYL}-D-PROLINAMIDE",
            "non-polymer",
        ),
        "JZM": ("5-[(4-CHLOROPHENYL)SULFANYL]QUINAZOLINE-2,4-DIAMINE", "non-polymer"),
        "JZL": (
            "7-AMINO-2-[(2-CHLOROBENZYL)AMINO]-1-{[(1S,2S)-2-HYDROXYCYCLOHEPTYL]METHYL}-1H-BENZIMIDAZOLE-5-CARBOXAMIDE",
            "non-polymer",
        ),
        "JZK": (
            "2-[(2-CHLOROBENZYL)AMINO]-1-(CYCLOHEXYLMETHYL)-1H-BENZIMIDAZOLE-5-CARBOXAMIDE",
            "non-polymer",
        ),
        "JZJ": (
            "(3S)-3-[4-(4-BROMOPHENYL)-1H-IMIDAZOL-2-YL]-1,2,3,4-TETRAHYDROISOQUINOLINE",
            "non-polymer",
        ),
        "JZI": (
            "(2R,4E)-2-[(NAPHTHALEN-2-YLCARBONYL)AMINO]-5-PHENYLPENT-4-ENOIC ACID",
            "non-polymer",
        ),
        "JZH": (
            "(3S)-3-(4-HYDROXYPHENYL)-1,5-DIHYDRO-1,5,12-TRIAZABENZO[4,5]CYCLOOCTA[1,2,3-CD]INDEN-4(3H)-ONE",
            "non-polymer",
        ),
        "JZG": (
            "3-{1-[3-(BIPHENYL-4-YLAMINO)-3-OXOPROPYL]-1H-1,2,3-TRIAZOL-4-YL}-6-HYDROXY-1-METHYL-2-PHENYL-1H-INDOLE-5-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "JZF": (
            "ETHYL (4-{(2R)-2-[(1S,3S,5S)-3,5-DIMETHYL-2-OXOCYCLOHEXYL]-2-HYDROXYETHYL}-2,6-DIOXOPIPERIDIN-1-YL)ACETATE",
            "non-polymer",
        ),
        "JZE": ("?", "non-polymer"),
        "JZD": ("?", "non-polymer"),
        "JZC": (
            "(5E,9E,11E)-14,16-DIHYDROXY-3,4,7,8-TETRAHYDRO-1H-2-BENZOXACYCLOTETRADECINE-1,11(12H)-DIONE 11-[O-(2-OXO-2-PIPERIDIN-1-YLETHYL)OXIME]",
            "non-polymer",
        ),
        "JZB": (
            "(5E,9E,11E)-13-CHLORO-14,16-DIHYDROXY-3,4,7,8-TETRAHYDRO-1H-2-BENZOXACYCLOTETRADECINE-1,11(12H)-DIONE 11-[O-(2-OXO-2-PIPERIDIN-1-YLETHYL)OXIME]",
            "non-polymer",
        ),
        "JZA": (
            "4-[(1,1-DIOXIDO-1,2-THIAZETIDIN-2-YL)CARBONYL]MORPHOLINE",
            "non-polymer",
        ),
        "S1T": (
            "2-{4-[(4-AMINO-2-METHYLPYRIMIDIN-5-YL)METHYL]-5-[(1S)-1-HYDROXYETHYL]-3-METHYL-2-THIENYL}ETHYL TRIHYDROGEN DIPHOSPHATE",
            "non-polymer",
        ),
        "OHT": ("4-HYDROXYTAMOXIFEN", "non-polymer"),
        "IFP": (
            "2-TRIFLUOROMETHYL-5-METHYLENE-5H-PYRIMIDIN-4-YLIDENEAMINE",
            "non-polymer",
        ),
        "S1H": ("1-HEXADECANOSULFONYL-O-L-SERINE", "peptide"),
        "JZZ": (
            "4-[7-(3,3-DIMETHYLBUT-1-YN-1-YL)NAPHTHALEN-1-YL]-5-METHOXY-2-METHYL-2,4-DIHYDRO-3H-1,2,4-TRIAZOL-3-ONE",
            "non-polymer",
        ),
        "JZY": (
            "4-METHYL-5-(6-{[4-(METHYLSULFONYL)PIPERAZIN-1-YL]METHYL}-4-MORPHOLIN-4-YLTHIENO[3,2-D]PYRIMIDIN-2-YL)PYRIMIDIN-2-AMINE",
            "non-polymer",
        ),
        "JZX": (
            "5-(6-{[4-(METHYLSULFONYL)PIPERAZIN-1-YL]METHYL}-4-MORPHOLIN-4-YLTHIENO[3,2-D]PYRIMIDIN-2-YL)PYRIDIN-2-AMINE",
            "non-polymer",
        ),
        "JZW": (
            "[3-(6-{[4-(METHYLSULFONYL)PIPERAZIN-1-YL]METHYL}-4-MORPHOLIN-4-YLTHIENO[3,2-D]PYRIMIDIN-2-YL)PHENYL]METHANOL",
            "non-polymer",
        ),
        "JZV": (
            "METHYL (2E)-(METHOXYIMINO)(2-{[({(1Z)-1-[3-(TRIFLUOROMETHYL)PHENYL]ETHYLIDENE}AMINO)OXY]METHYL}PHENYL)ETHANOATE",
            "non-polymer",
        ),
        "JZU": ("5'-DEOXY-5'-(SULFAMOYLAMINO)ADENOSINE", "non-polymer"),
        "JZT": ("?", "non-polymer"),
        "JZS": (
            "N-[(1R)-2-AMINO-1-METHYL-2-OXOETHYL]-3-(6-METHYL-4-{[3,3,3-TRIFLUORO-2-HYDROXY-2-(TRIFLUOROMETHYL)PROPYL]AMINO}-1H-INDAZOL-1-YL)BENZAMIDE",
            "non-polymer",
        ),
        "JZR": ("HEXYL BETA-D-GLUCOPYRANOSIDE", "non-polymer"),
        "JZQ": (
            "(4R)-3-[(2S,3S)-3-{[(2,6-DIFLUOROPHENOXY)ACETYL]AMINO}-2-HYDROXY-4-PHENYLBUTANOYL]-N-[(1S,2R)-2-HYDROXY-2,3-DIHYDRO-1H-INDEN-1-YL]-5,5-DIMETHYL-1,3-THIAZOLIDINE-4-CARBOXAMIDE",
            "peptide",
        ),
        "JZP": (
            "(4R)-3-[(2S,3S)-3-{[(2,6-DICHLOROPHENOXY)ACETYL]AMINO}-2-HYDROXY-4-PHENYLBUTANOYL]-N-[(1S,2R)-2-HYDROXY-2,3-DIHYDRO-1H-INDEN-1-YL]-5,5-DIMETHYL-1,3-THIAZOLIDINE-4-CARBOXAMIDE",
            "peptide",
        ),
        "Z81": (
            "(5S)-2-AMINO-5-(2',5'-DIFLUOROBIPHENYL-3-YL)-3-METHYL-5-PYRIDIN-4-YL-3,5-DIHYDRO-4H-IMIDAZOL-4-ONE",
            "non-polymer",
        ),
        "UIZ": (
            "(1R,3AS,4R,8AS,8BR)-4-(2-BENZO[1,3]DIOXOL-5-YLMETHYL)-1-BENZYL-3-OXO-DECAHYDRO-PYRROLO[3,4-A]PYRROZILIN-4-YL-BENZAMIDINE",
            "non-polymer",
        ),
        "Z80": (
            "3-(2-CHLORO-10H-PHENOTHIAZIN-10-YL)-N,N-DIMETHYLPROPAN-1-AMINE",
            "non-polymer",
        ),
        "OHP": ("(2-HYDROXYPHENYL)ACETIC ACID", "non-polymer"),
        "19Z": ("2-[(4-HYDROXYBENZYL)AMINO]-5-NITROBENZAMIDE", "non-polymer"),
        "Z83": (
            "(4-{3-TERT-BUTYL-5-[(1,3-THIAZOL-2-YLCARBAMOYL)AMINO]-1H-PYRAZOL-1-YL}PHENYL)ACETIC ACID",
            "non-polymer",
        ),
        "RX4": (
            "N~2~-ACETYL-N-{(1R)-1-[(S)-[(2S)-3-{[(2S)-1-AMINO-1-OXOPROPAN-2-YL]AMINO}-2-METHYL-3-OXOPROPYL](HYDROXY)PHOSPHORYL]-2-PHENYLETHYL}-L-ALPHA-ASPARAGINE",
            "peptide",
        ),
        "RX6": (
            "METHYL (2-{(R)-(3-CHLOROPHENYL)[(3R)-1-({(2S)-1-(METHYLAMINO)-3-[(3R)-TETRAHYDRO-2H-PYRAN-3-YL]PROPAN-2-YL}CARBAMOYL)PIPERIDIN-3-YL]METHOXY}ETHYL)CARBAMATE",
            "non-polymer",
        ),
        "Z84": (
            "1-[3-TERT-BUTYL-1-(4-METHYLPHENYL)-1H-PYRAZOL-5-YL]-3-{4-[2-(PYRIDIN-4-YLMETHOXY)ETHYL]-1,3-THIAZOL-2-YL}UREA",
            "non-polymer",
        ),
        "Z87": (
            "1-[3-TERT-BUTYL-1-(4-METHYLPHENYL)-1H-PYRAZOL-5-YL]-3-(1,3-THIAZOL-2-YL)UREA",
            "non-polymer",
        ),
        "N87": (
            "(1R,4R,5R)-1,4,5-TRIHYDROXY-3-[3-(PHENYLCARBONYL)PHENYL]CYCLOHEX-2-ENE-1-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "W7E": (
            "2-[[4-[2-[[3-[(2-METHYL-1,3-BENZOTHIAZOL-6-YL)AMINO]-3-OXIDANYLIDENE-PROPYL]CARBAMOYLAMINO]ETHOXY]PHENYL]METHYL]PROPANEDIOIC ACID",
            "non-polymer",
        ),
        "Z89": (
            "(2R)-2-[(5Z)-5-(5-BROMO-2-OXO-1,2-DIHYDRO-3H-INDOL-3-YLIDENE)-4-OXO-2-THIOXO-1,3-THIAZOLIDIN-3-YL]BUTANEDIOIC ACID",
            "non-polymer",
        ),
        "HP8": (
            "5'-O-{[(2,3-DIHYDROXYPHENYL)CARBONYL]SULFAMOYL}ADENOSINE",
            "non-polymer",
        ),
        "QRG": ("N-ACETYL-L-ALPHA-ASPARTYL-L-METHIONINE", "peptide"),
        "HP3": ("2,3-PROPANDIOL", "non-polymer"),
        "HP2": ("HYDROXYPYRROLE-IMIDAZOLE-PYRROLE POLYAMIDE", "non-polymer"),
        "HP1": ("IMIDAZOLE-PYRROLE-HYDROXYPYRROLE POLYAMIDE", "non-polymer"),
        "HP0": (
            "(3S,4AS,8AS)-2-[(2R,3S)-3-AMINO-2-HYDROXY-4-PHENYLBUTYL]DECAHYDROISOQUINOLINE-3-CARBOXYLIC ACID",
            "peptide",
        ),
        "HP7": (
            "(2S,3S,4R,5R,6R)-5-ACETAMIDO-6-[[[(2R,3S,4R,5R)-5-(2,4-DIOXOPYRIMIDIN-1-YL)-3,4-DIHYDROXY-OXOLAN-2-YL]METHOXY-HYDROXY-PHOSPHORYL]OXY-HYDROXY-PHOSPHORYL]OXY-3,4-DIHYDROXY-OXANE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "HP6": ("HEPTANE", "non-polymer"),
        "QRP": (
            "(3S,8AS)-3-(1H-INDOL-3-YLMETHYL)HEXAHYDROPYRROLO[1,2-A]PYRAZINE-1,4-DIONE",
            "non-polymer",
        ),
        "QRW": (
            "3-({[4-METHYL-5-(PYRIDIN-4-YL)-4H-1,2,4-TRIAZOL-3-YL]METHYL}AMINO)-N-[2-(TRIFLUOROMETHYL)BENZYL]BENZAMIDE",
            "non-polymer",
        ),
        "QRX": (
            "N-(2,6-DIFLUOROBENZYL)-3-({[4-PROPYL-5-(PYRIMIDIN-4-YL)-4H-1,2,4-TRIAZOL-3-YL]METHYL}AMINO)BENZAMIDE",
            "non-polymer",
        ),
        "HPZ": ("(2E,4E)-2-HYDROXY-6-OXO-6-PHENYLHEXA-2,4-DIENOIC ACID", "non-polymer"),
        "HPY": ("4-HYDROXY-3,4-DIHYDRO-1H-PYRIMIDIN-2-ONE", "non-polymer"),
        "HPX": ("(2Z,4E)-2-HYDROXY-6-OXO-6-PHENYLHEXA-2,4-DIENOIC ACID", "non-polymer"),
        "E08": (
            "TRIHYDROXY-[(1S)-1-[(2-PROPAN-2-YLPHENYL)CARBONYLAMINO]ETHYL]BORON",
            "non-polymer",
        ),
        "E09": (
            "3-HYDROXYMETHYL-5-AZIRIDINYL-1METHYL-2-[1H-INDOLE-4,7-DIONE]-PROPANOL",
            "non-polymer",
        ),
        "E06": (
            "[(1R)-1-[(2,6-DIFLUOROPHENYL)CARBONYLAMINO]ETHYL]-TRIHYDROXY-BORON",
            "non-polymer",
        ),
        "E07": (
            "TRIHYDROXY-[(1S)-1-[[2-(PHENYLMETHYL)PHENYL]CARBONYLAMINO]ETHYL]BORON",
            "non-polymer",
        ),
        "E04": (
            "3-CARBOXAMIDO-1,3,5(10)-ESTRATRIEN-17(R)-SPIRO-2'(5',5'-DIMETHYL-6'OXO)TETRAHYDROPYRAN",
            "non-polymer",
        ),
        "HPP": ("HYDROXYPHENYL PROPIONIC ACID", "non-polymer"),
        "E02": (
            "5-(1H-BENZIMIDAZOL-1-YLACETYL)-2-CHLOROBENZENESULFONAMIDE",
            "non-polymer",
        ),
        "HPV": ("2-OXO-3-(PHOSPHONOOXY)PROPANOIC ACID", "non-polymer"),
        "HPT": ("6-HYDROXYPROPYLTHYMINE", "non-polymer"),
        "HPK": ("(3E)-2,6-DIOXO-6-PHENYLHEX-3-ENOATE", "non-polymer"),
        "HPJ": (
            "1-HYDROXY-6-[1-(3-METHOXYPROPYL)-1H-PYRROLO[2,3-B]PYRIDIN-5-YL]-4-PHENYLPYRIDIN-2(1H)-ONE",
            "non-polymer",
        ),
        "HPI": (
            "N-(1-CARBOXY-3-PHENYLPROPYL)PHENYLALANYL-ALPHA-ASPARAGINE",
            "non-polymer",
        ),
        "HPH": ("(2S)-2-AMINO-3-PHENYLPROPANE-1,1-DIOL", "peptide"),
        "HPO": ("2-OXOHEPTYLPHOSPHONIC ACID", "non-polymer"),
        "HPN": ("HEPTANAMIDE", "non-polymer"),
        "HPM": (
            "N-{5-[(7-{[(2S)-2-HYDROXY-3-PIPERIDIN-1-YLPROPYL]OXY}-6-METHOXYQUINAZOLIN-4-YL)AMINO]PYRIMIDIN-2-YL}BENZAMIDE",
            "non-polymer",
        ),
        "HPL": ("HEPTYLFORMAMIDE", "non-polymer"),
        "HPC": ("3-AMINO-4-PHENYL-BUTAN-2-ONE", "peptide"),
        "HPB": ("2-HYDROXY-3-AMINO-4-PHENYL BUTANE", "non-polymer"),
        "HPA": ("HYPOXANTHINE", "non-polymer"),
        "HPG": ("HYDROXYPROPYLOXY GROUP", "non-polymer"),
        "HPF": ("1-[(2-HYDROXYLPHENYL)AMINO]3-GLYCEROLPHOSPHATE", "non-polymer"),
        "HPE": ("HOMOPHENYLALANINE", "peptide"),
        "HPD": ("1-HYDROXY-PENTANE-3,4-DIOL-5-PHOSPHATE", "non-polymer"),
        "8SP": (
            "O-[(R)-{[(2R)-2,3-BIS(OCTANOYLOXY)PROPYL]OXY}(HYDROXY)PHOSPHORYL]-L-SERINE",
            "peptide",
        ),
        "W71": (
            "5-(7-(4-(4,5-DIHYDRO-2-OXAZOLYL)PHENOXY)HEPTYL)-3-METHYL ISOXAZOLE",
            "non-polymer",
        ),
        "8ST": (
            "N-(4-CHLOROPHENYL)-2-[(PYRIDIN-4-YLMETHYL)AMINO]BENZAMIDE",
            "non-polymer",
        ),
        "W72": (
            "6-DEOXY-6-[(2R,3R,4R)-3,4-DIHYDROXY-2-(HYDROXYMETHYL)PYRROLIDIN-1-YL]-L-GULONIC ACID",
            "non-polymer",
        ),
        "HFB": ("HEXAFLUOROBENZENE", "non-polymer"),
        "UI3": (
            "7-METHOXY-8-[1-(METHYLSULFONYL)-1H-PYRAZOL-4-YL]NAPHTHALENE-2-CARBOXIMIDAMIDE",
            "non-polymer",
        ),
        "19B": (
            "4-METHYL-N-(3-MORPHOLIN-4-YLPHENYL)-3-(3-PIPERIDIN-4-YL-1,2-BENZISOXAZOL-6-YL)BENZAMIDE",
            "non-polymer",
        ),
        "5X": (
            "5R-(2E-METHYL-3-PHENYL-ALLYL)-3-(BENZENESULFONYLAMINO)-4-OXO-2-THIONOTHIAZOLIDINE",
            "non-polymer",
        ),
        "TKL": (
            "(S)-2-(4-((S)-1,5-DIAMINOPENTYL)-1H-1,2,3-TRIAZOL-1-YL)-4-METHYLPENTANOIC ACID",
            "non-polymer",
        ),
        "5H": (
            "5R-(4-BROMOPHENYLMETHYL)-3-(BENZENESULFONYLAMINO)-4-OXO-2-THIONOTHIAZOLIDINE",
            "non-polymer",
        ),
        "OHL": (
            "N-HYDROXY-N~2~-[(4-METHOXYPHENYL)SULFONYL]-N~2~-(PYRIDIN-4-YLMETHYL)GLYCINAMIDE",
            "non-polymer",
        ),
        "F3A": (
            "2'-DEOXY-5'-O-[(S)-{DIFLUORO[(S)-HYDROXY(PHOSPHONOOXY)PHOSPHORYL]METHYL}(HYDROXY)PHOSPHORYL]ADENOSINE",
            "non-polymer",
        ),
        "F3B": ("1,3,5-TRICHLORO-2,4,6-TRIFLUOROBENZENE", "non-polymer"),
        "OHO": ("CYCLOHEXANE AMINOCARBOXYLIC ACID", "non-polymer"),
        "OHH": (
            "2-[3-({METHYL[1-(2-NAPHTHOYL)PIPERIDIN-4-YL]AMINO}CARBONYL)-2-NAPHTHYL]-1-(1-NAPHTHYL)-2-OXOETHYLPHOSPHONIC ACID",
            "non-polymer",
        ),
        "OHI": ("3-(2-OXO-2H-IMIDAZOL-4-YL)-L-ALANINE", "peptide"),
        "F3F": (
            "S-[5-(TRIFLUOROMETHYL)-4H-1,2,4-TRIAZOL-3-YL] 5-(PHENYLETHYNYL)FURAN-2-CARBOTHIOATE",
            "non-polymer",
        ),
        "IOX": ("PHENYLMETHANESULFONIC ACID", "non-polymer"),
        "F3H": (
            "1,5-ANHYDRO-2,3-DIDEOXY-3-FLUORO-2-(5-METHYL-2,4-DIOXO-3,4-DIHYDROPYRIMIDIN-1(2H)-YL)-6-O-PHOSPHONO-D-ALTRITOL",
            "nucleic-acid",
        ),
        "OHE": ("HYDROXYETHYL GROUP", "non-polymer"),
        "IOT": ("?", "non-polymer"),
        "IOS": ("3-SULFOOXY-1H-INDOLE", "non-polymer"),
        "F3M": (
            "METHYL 4-AMINO-2,4,5-TRIDEOXY-5,5,5-TRIFLUORO-L-THREO-PENTONATE",
            "peptide",
        ),
        "IOP": ("INDOLYLPROPIONIC ACID", "non-polymer"),
        "F3P": (
            "(1S)-1-[(1S)-1-FLUOROETHYL]-1,5-DIMETHYLHEXYL TRIHYDROGEN DIPHOSPHATE",
            "non-polymer",
        ),
        "ION": ("ARISTOLOCHENE", "non-polymer"),
        "IOM": ("(DIAMINOMETHYL-METHYL-AMINO)-ACETIC ACID", "non-polymer"),
        "F3S": ("FE3-S4 CLUSTER", "non-polymer"),
        "F3T": (
            "4-AMINO-2,4,5-TRIDEOXY-5,5,5-TRIFLUORO-L-THREO-PENTONIC ACID",
            "peptide",
        ),
        "IOH": ("2-PROPANOL, ISOPROPANOL", "non-polymer"),
        "IOG": (
            "N-[(1R)-3-(4-HYDROXYPHENYL)-1-METHYLPROPYL]-2-[2-PHENYL-6-(2-PIPERIDIN-1-YLETHOXY)-1H-INDOL-3-YL]ACETAMIDE",
            "non-polymer",
        ),
        "IOF": (
            "4-(AMINOSULFONYL)-N-[(3,4,5-TRIFLUOROPHENYL)METHYL]-BENZAMIDE",
            "non-polymer",
        ),
        "IOE": (
            "4-(AMINOSULFONYL)-N-[(2,4,6-TRIFLUOROPHENYL)METHYL]-BENZAMIDE",
            "non-polymer",
        ),
        "IOD": ("IODIDE ION", "non-polymer"),
        "IOC": (
            "4-(AMINOSULFONYL)-N-[(2,3,4-TRIFLUOROPHENYL)METHYL]-BENZAMIDE",
            "non-polymer",
        ),
        "IOB": ("3-IODO-BENZYL ALCOHOL", "non-polymer"),
        "IOA": (
            "4-(AMINOSULFONYL)-N-[(2,5-DIFLUOROPHENYL)METHYL]-BENZAMIDE",
            "non-polymer",
        ),
        "OHS": ("O-(CARBOXYSULFANYL)-4-OXO-L-HOMOSERINE", "peptide"),
        "OLP": (
            "2-(2-HYDROXYETHYLAMINO)-6-(3-CHLOROANILINO)-9-ISOPROPYLPURINE",
            "non-polymer",
        ),
        "RXA": ("ISOQUINOLINE-1,3,4(2H)-TRIONE", "non-polymer"),
        "RXC": (
            "(1S)-1-(3-CHLOROPHENYL)-2-OXO-2-[(1,3,4-TRIOXO-1,2,3,4-TETRAHYDROISOQUINOLIN-5-YL)AMINO]ETHYL ACETATE",
            "non-polymer",
        ),
        "RXB": (
            "(1S)-2-OXO-1-PHENYL-2-[(1,3,4-TRIOXO-1,2,3,4-TETRAHYDROISOQUINOLIN-5-YL)AMINO]ETHYL ACETATE",
            "non-polymer",
        ),
        "IO3": ("IOPHENOXIC ACID", "non-polymer"),
        "RXD": (
            "N-[3-(2-FLUOROETHOXY)PHENYL]-N'-(1,3,4-TRIOXO-1,2,3,4-TETRAHYDROISOQUINOLIN-6-YL)BUTANEDIAMIDE",
            "non-polymer",
        ),
        "UI1": (
            "6-[(Z)-AMINO(IMINO)METHYL]-N-[4-(AMINOMETHYL)PHENYL]-4-(PYRIMIDIN-2-YLAMINO)-2-NAPHTHAMIDE",
            "non-polymer",
        ),
        "F32": (
            "(5S)-5-{4-[(2S)-2-(1H-BENZIMIDAZOL-2-YL)-2-(1,3-BENZOTHIAZOL-2-YLAMINO)ETHYL]PHENYL}ISOTHIAZOLIDIN-3-ONE 1,1-DIOXIDE",
            "non-polymer",
        ),
        "UI2": ("8-(PYRIMIDIN-2-YLAMINO)NAPHTHALENE-2-CARBOXIMIDAMIDE", "non-polymer"),
        "F34": (
            "N-[(3S,4R,5R,6R)-4,5,6-TRIHYDROXYAZEPAN-3-YL]ACETAMIDE",
            "non-polymer",
        ),
        "RXP": (
            "1-BENZYLOXYCARBONYLAMINO-2-PHENYL-ETHYL)-{2-[1-CARBAMOYL-2-(1H-INDOL-3-YL)-ETHYLCARBAMOYL]-5-PHENYL-PENTYL}-PHOSPHINIC ACID",
            "non-polymer",
        ),
        "RXR": (
            "[2,2,5,5-TETRAMETHYL-3,4-BIS(SULFANYLMETHYL)-2,5-DIHYDRO-1H-PYRROL-1-YL]OXIDANYL RADICAL",
            "non-polymer",
        ),
        "RXV": (
            "6-METHOXY-4-(2-{4-[([1,3]OXATHIOLO[5,4-C]PYRIDIN-6-YLMETHYL)AMINO]PIPERIDIN-1-YL}ETHYL)QUINOLINE-3-CARBONITRILE",
            "non-polymer",
        ),
        "QYT": (
            "(5E)-5-(QUINOXALIN-6-YLMETHYLENE)-1,3-THIAZOLIDINE-2,4-DIONE",
            "non-polymer",
        ),
        "Z8B": ("5'-AZIDO-8-BROMO-5'-DEOXYADENOSINE", "non-polymer"),
        "M38": (
            "4-(5,11-DIOXO-5H-INDENO[1,2-C]ISOQUINOLIN-6(11H)-YL)BUTANOATE",
            "non-polymer",
        ),
        "M37": (
            "1-[2-(AMINOMETHYL)PHENYL]-N-(3-FLUORO-2'-SULFAMOYLBIPHENYL-4-YL)-3-(TRIFLUOROMETHYL)-1H-PYRAZOLE-5-CARBOXAMIDE",
            "non-polymer",
        ),
        "M35": (
            "1-[3-(AMINOMETHYL)PHENYL]-N-[3-FLUORO-2'-(METHYLSULFONYL)BIPHENYL-4-YL]-3-(TRIFLUOROMETHYL)-1H-PYRAZOLE-5-CARBOXAMIDE",
            "non-polymer",
        ),
        "HQ": ("2-HEPTYL-1,4-DIHYDROXY-QUINOLINIUM", "non-polymer"),
        "M33": (
            "5'-O-[(S)-HYDROXY{[(S)-HYDROXY(METHYL)PHOSPHORYL]OXY}PHOSPHORYL]ADENOSINE",
            "non-polymer",
        ),
        "M32": (
            "N-(BENZYLSULFONYL)-D-ALANYL-N-[2-(AMINOMETHYL)-5-CHLOROBENZYL]-L-PROLINAMIDE",
            "peptide",
        ),
        "HT": (
            "2'-(4-HYDROXYPHENYL)-5-(4-METHYL-1-PIPERAZINYL)-2,5'-BI-BENZIMIDAZOLE",
            "non-polymer",
        ),
        "M30": ("N-CYCLOHEXYLGLYCINE", "peptide"),
        "IZC": ("2H-IMIDAZOL-4-YLACETIC ACID", "non-polymer"),
        "HO": ("HOLMIUM ATOM", "non-polymer"),
        "MPG": ("1-MONOOLEOYL-RAC-GLYCEROL", "non-polymer"),
        "HG": ("MERCURY (II) ION", "non-polymer"),
        "MNX": ("1,8-DI-HYDROXY-4-NITRO-XANTHEN-9-ONE", "non-polymer"),
        "MNY": ("5,8-DI-AMINO-1,4-DIHYDROXY-ANTHRAQUINONE", "non-polymer"),
        "PB7": (
            "(2S)-2-{(3S)-3-(ACETYLAMINO)-3-[(2S)-BUTAN-2-YL]-2-OXOPYRROLIDIN-1-YL}-N-{(2S,3R)-3-HYDROXY-4-[(3-METHOXYBENZYL)AMINO]-1-PHENYLBUTAN-2-YL}-4-PHENYLBUTANAMIDE",
            "non-polymer",
        ),
        "PB6": (
            "3-[(1E)-BUT-1-EN-1-YL]-1-(2,2-DIPHOSPHONOETHYL)PYRIDINIUM",
            "non-polymer",
        ),
        "PB1": ("3-(2-AMINOETHYL)-4-(AMINOMETHYL)HEPTANEDIOIC ACID", "non-polymer"),
        "PB0": (
            "(2S)-2-[(3R)-3-(ACETYLAMINO)-3-(2-METHYLPROPYL)-2-OXOPYRROLIDIN-1-YL]-N-{(1R,2S)-3-(3,5-DIFLUOROPHENYL)-1-HYDROXY-1-[(2R,4R)-4-PROPOXYPYRROLIDIN-2-YL]PROPAN-2-YL}-4-PHENYLBUTANAMIDE",
            "non-polymer",
        ),
        "PB3": ("N,N'-BIS(3-METHYLBENZYL)PYRIMIDINE-4,6-DICARBOXAMIDE", "non-polymer"),
        "PB2": ("1-(BIPHENYL-4-YLMETHYL)-1H-IMIDAZOLE", "non-polymer"),
        "MNP": ("2-(3-NITROPHENYL)ACETIC ACID", "non-polymer"),
        "MNQ": (
            "M-NITROPHENYL AMINOETHYLDIPHOSPHATE BERYLLIUM TRIFLUORIDE",
            "non-polymer",
        ),
        "MNR": ("PROTOPORPHYRIN IX CONTAINING MN", "non-polymer"),
        "MNS": ("5-(DIMETHYLAMINO)-1-NAPHTHALENESULFONAMIDE", "non-polymer"),
        "MNT": (
            "2'(3')-O-N-METHYLANTHRANILOYL-ADENOSINE-5'-DIPHOSPHATE",
            "non-polymer",
        ),
        "MNU": (
            "(2R,4S)-1-[(4R)-3,4-DIHYDROXYTETRAHYDROFURAN-2-YL]-5-[(METHYLAMINO)METHYL]-1,2,3,4-TETRAHYDROPYRIMIDINE-2,4-DIOL-5'-MONOPHOSPHATE",
            "nucleic-acid",
        ),
        "MNV": ("N-METHYL-C-AMINO VALINE", "peptide"),
        "SCB": (
            "2-CHLORO-5-(5-{(E)-[(2Z)-3-(2-METHOXYETHYL)-4-OXO-2-(PHENYLIMINO)-1,3-THIAZOLIDIN-5-YLIDENE]METHYL}-2-FURYL)BENZOIC ACID",
            "non-polymer",
        ),
        "429": (
            "2-{4-[2-ACETYLAMINO-3-(4-CARBOXYMETHOXY-3-HYDROXY-PHENYL)-PROPIONYLAMINO]-BUTOXY}-6-HYDROXY-BENZOIC ACID METHYL ESTER",
            "non-polymer",
        ),
        "428": ("2-(1-METHYLETHYL)IMIDAZO[1,2-B]PYRIDAZIN-6-OL", "non-polymer"),
        "MNK": (
            "(2S)-5,7-DIHYDROXY-9-METHYL-2-(PROP-1-EN-2-YL)-1,2,3,4-TETRAHYDRO-6H-BENZO[7]ANNULEN-6-ONE",
            "non-polymer",
        ),
        "MNL": ("4,N-DIMETHYLNORLEUCINE", "peptide"),
        "MNM": (
            "(2S,3S,4R,5R)-2,3,4-TRIHYDROXY-5-HYDROXYMETHYL-PIPERIDINE",
            "non-polymer",
        ),
        "MNN": ("(S)-MANDELIC ACID NITRILE", "non-polymer"),
        "MNO": (
            "(3S)-N-METHANESULFONYL-3-({1-[N-(2-NAPHTOYL)-L-VALYL]-L-PROLYL}AMINO)-4-OXOBUTANAMIDE",
            "non-polymer",
        ),
        "421": (
            "6-CARBAMIMIDOYL-4-(3-HYDROXY-2-METHYL-BENZOYLAMINO)-NAPHTHALENE-2-CARBOXYLIC ACID METHYL ESTER",
            "non-polymer",
        ),
        "420": (
            "N-[2-({[5-(DIMETHYLAMINO)NAPHTHALEN-1-YL]SULFONYL}AMINO)ETHYL]-2-IODOACETAMIDE",
            "non-polymer",
        ),
        "MNB": ("5-MERCAPTO-2-NITRO-BENZOIC ACID", "non-polymer"),
        "422": (
            "3-(5-{[4-(AMINOMETHYL)PIPERIDIN-1-YL]METHYL}-1H-INDOL-2-YL)-1H-INDAZOLE-6-CARBONITRILE",
            "non-polymer",
        ),
        "MND": ("N-METHYL-D-ASPARAGINE", "peptide"),
        "424": (
            "(2S,4R)-1-[1-(4-CHLOROPHENYL)CYCLOPROPYL]CARBONYL-4-(2-CHLOROPHENYL)SULFONYL-N-[1-(IMINOMETHYL)CYCLOPROPYL]PYRROLIDINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "427": (
            "(R)-1-(4-(4-(HYDROXYMETHYL)-1,3,2-DIOXABOROLAN-2-YL)PHENYL)GUANIDINE",
            "non-polymer",
        ),
        "426": (
            "6-[(Z)-AMINO(IMINO)METHYL]-N-[3-(CYCLOPENTYLOXY)PHENYL]-2-NAPHTHAMIDE",
            "non-polymer",
        ),
        "MN8": ("2-(4-CARCOXY-5-ISOPROPYLTHIAZOLYL)BENZOPIPERIDINE", "non-polymer"),
        "PBT": (
            "[3-HYDROXY-5-(5-METHYL-2,4-DIOXOTETRAHYDRO-1(2H)-PYRIMIDINYL)TETRAHYDRO-2-FURANYL]METHYL DIHYDROGEN PHOSPHATE",
            "nucleic-acid",
        ),
        "CJH": (
            "(2Z)-2-CYANO-N-(3'-ETHOXYBIPHENYL-4-YL)-3-HYDROXYBUT-2-ENAMIDE",
            "non-polymer",
        ),
        "42M": (
            "{4-[4-(HYDROXYMETHYL)-2-METHOXYPHENOXY]BENZENE-1,3-DIYL}BIS[HYDROXY(OXO)AMMONIUM]",
            "non-polymer",
        ),
        "CJO": (
            "[(2R,4R)-2-[(1R,2R)-1-AMINO-2-HYDROXYPROPYL]-4-(4-HYDROXYBENZYL)-5-OXOIMIDAZOLIDIN-1-YL]ACETIC ACID",
            "peptide",
        ),
        "PBS": (
            "(2S,3S,4R)-N-OCTANOYL-1-[(ALPHA-D-GALACTOPYRANOSYL)OXY]-2-AMINO-OCTADECANE-3,4-DIOL",
            "non-polymer",
        ),
        "PBR": ("PENTABROMOPHENOL", "non-polymer"),
        "CJB": ("1-BETA-D-GLUCOPYRANOSYLPYRIMIDINE-2,4(1H,3H)-DIONE", "non-polymer"),
        "CJC": (
            "7-({4-[(3R)-3-AMINOPYRROLIDIN-1-YL]-5-CHLORO-6-ETHYL-7H-PYRROLO[2,3-D]PYRIMIDIN-2-YL}SULFANYL)PYRIDO[2,3-B]PYRAZIN-2(1H)-ONE",
            "non-polymer",
        ),
        "42C": ("N,N-DIMETHYL-7H-PURIN-6-AMINE", "non-polymer"),
        "42B": ("4-AMINO-2-HYDROXYBUTANOIC ACID", "non-polymer"),
        "42D": (
            "3,5-DIDEOXY-5-[(METHOXYCARBONYL)AMINO]-D-GLYCERO-ALPHA-D-GALACTO-NON-2-ULOPYRANOSONIC ACID",
            "non-polymer",
        ),
        "MN6": ("MANGANESE ION, 6 WATERS COORDINATED", "non-polymer"),
        "MN7": ("1-CARBOXY-4-ISOBUTYLAMINOMETHYLBENZENE", "non-polymer"),
        "CJZ": ("2-[(3,5-DICHLORO-4-HYDROXYPHENYL)AMINO]BENZOIC ACID", "non-polymer"),
        "PBD": (
            "1-(3,4-DIMETHOXYPHENYL)-3-[3-(1H-IMIDAZOL-1-YL)PROPYL]THIOUREA",
            "non-polymer",
        ),
        "PBG": (
            "3-[5-(AMINOMETHYL)-4-(CARBOXYMETHYL)-1H-PYRROL-3-YL]PROPANOIC ACID",
            "non-polymer",
        ),
        "42Z": (
            "N-[(5,6-DICHLORO-1H-BENZIMIDAZOL-2-YL)CARBONYL]GLYCINE",
            "non-polymer",
        ),
        "PBA": ("PHENYLETHANE BORONIC ACID", "non-polymer"),
        "SC8": (
            "5-(2,3-DICHLOROPHENYL)-N-(PYRIDIN-4-YLMETHYL)PYRAZOLO[1,5-A]PYRIMIDIN-7-AMINE",
            "non-polymer",
        ),
        "PBC": ("PHENYL BORONIC ACID", "non-polymer"),
        "PBB": ("S-(4-BROMOBENZYL)CYSTEINE", "peptide"),
        "42Q": (
            "1-{2-FLUORO-4-[(5-METHYL-5H-PYRROLO[3,2-D]PYRIMIDIN-4-YL)OXY]PHENYL}-3-[3-(TRIFLUOROMETHYL)PHENYL]UREA",
            "non-polymer",
        ),
        "PBL": ("PENTANE-2,2,4,4-TETROL", "non-polymer"),
        "PBO": ("1-PYRIDIN-3-YLBUTAN-1-ONE", "non-polymer"),
        "PBN": ("4-PHENYLBUTYLAMINE", "non-polymer"),
        "PBI": (
            "[2-AMINOMETHYL-5-OXO-4-(4-OXO-CYCLOHEXA-2,5-DIENYLMETHYL)-4,5-DIHYDRO-IMIDAZOL-1-YL] -ACETALDEHYDE",
            "non-polymer",
        ),
        "PBH": (
            "(11AS)-7-METHOXY-8-(3-{4-[6-(4-METHYLPIPERAZIN-1-YL)-1H-BENZIMIDAZOL-2-YL]PHENOXY}PROPOXY)-1,2,3,10,11,11A-HEXAHYDRO-5H-PYRROLO[2,1-C][1,4]BENZODIAZEPIN-5-ONE",
            "non-polymer",
        ),
        "PBK": (
            "5'-O-[(R)-(DODECANOYLOXY)(HYDROXY)PHOSPHORYL]ADENOSINE",
            "non-polymer",
        ),
        "SC2": ("N-ACETYL-L-CYSTEINE", "peptide"),
        "TIT": (
            "N-((3S,4S)-5-[(4-BROMOBENZYL)OXY]-3-HYDROXY-4-{[N-(PYRIDIN-2-YLCARBONYL)-L-VALYL]AMINO}PENTANOYL)-L-ALANYL-L-LEUCINAMIDE",
            "non-polymer",
        ),
        "M3S": ("2-(ACETYLOXY)-3-METHYLBENZOIC ACID", "non-polymer"),
        "M3R": ("METHYL PHOSPHONATED L-ARGININE", "peptide"),
        "M3P": ("(2E,4E)-2-HYDROXY-6-OXOHEPTA-2,4-DIENOIC ACID", "non-polymer"),
        "M3M": ("3-O-ALPHA-D-MANNOPYRANOSYL-ALPHA-D-MANNOPYRANOSE", "saccharide"),
        "M3L": ("N-TRIMETHYLLYSINE", "peptide"),
        "M3K": (
            "2-([3R-3,4-DIHYDROXYBUTYL]OXY)-4-FLUORO-6-[(2-FLUORO-4-IODOPHENYL)AMINO]BENZAMIDE",
            "non-polymer",
        ),
        "M3I": ("5-METHYL-1,2-OXAZOL-3-AMINE", "non-polymer"),
        "TKT": (
            "(3-ENDO)-8-METHYL-8-AZABICYCLO[3.2.1]OCT-3-YL 1H-INDOLE-3-CARBOXYLATE",
            "non-polymer",
        ),
        "M3F": (
            "1-METHYL-3-(TRIFLUOROMETHYL)BENZO[C][1,8]NAPHTHYRIDIN-6(5H)-ONE",
            "non-polymer",
        ),
        "M3C": (
            "(2S)-2-AMINO-4-(METHYLSULFANYL)-1-(1,3-THIAZOL-2-YL)BUTANE-1,1-DIOL",
            "non-polymer",
        ),
        "FM5": (
            "3-(N-HYDROXYACETAMIDO)-1-(3,4-DICHLOROPHENYL)PROPYLPHOSPHONIC ACID",
            "non-polymer",
        ),
        "BR6": ("(E)-4-((4-BROMOPHENYLIMINO)METHYL)BENZIMIDAMIDE", "non-polymer"),
        "BR5": ("2-ETHOXYETHYLPHOSPHINATE", "non-polymer"),
        "BR4": ("6-PHENYL-4(R)-(7-PHENYL-HEPTANOYLAMINO)-HEXANOIC ACID", "non-polymer"),
        "BR3": (
            "(6-METHYL-3,4-DIHYDRO-2H-CHROMEN-2-YL)METHYLPHOSPHINATE",
            "non-polymer",
        ),
        "BR2": (
            "2,6-DIFLUORO-N-[(5S)-3-METHOXY-5H-PYRAZOLO[3,4-B]PYRIDIN-5-YL]-3-[(PHENYLSULFONYL)AMINO]BENZAMIDE",
            "non-polymer",
        ),
        "BR0": (
            "(3AS,4R,9BR)-9-NITRO-3A,4,5,9B-TETRAHYDRO-3H-CYCLOPENTA[C]QUINOLINE-4,6-DICARBOXYLIC ACID",
            "non-polymer",
        ),
        "K9S": ("1-(1-BENZOTHIOPHEN-3-YL)METHANAMINE", "non-polymer"),
        "AD6": ("4-[(5-BROMO-1,3-THIAZOL-2-YL)AMINO]-N-METHYLBENZAMIDE", "non-polymer"),
        "AD7": (
            "OCTYL 3-AMINO-3-DEOXY-2-O-(2,6-DIDEOXY-ALPHA-L-LYXO-HEXOPYRANOSYL)-BETA-D-GALACTOPYRANOSIDE",
            "non-polymer",
        ),
        "AD4": ("((((3'-DEOXY-3'-THYMIDINYL)-AMINO)CARBONYL)METHYL)", "non-polymer"),
        "AD5": (
            "N~6~-CYCLOHEXYL-N~2~-(4-MORPHOLIN-4-YLPHENYL)-9H-PURINE-2,6-DIAMINE",
            "non-polymer",
        ),
        "AD2": ("2'-DEOXY-ADENOSINE-3'-5'-DIPHOSPHATE", "nucleic-acid"),
        "AD3": ("3-DEAZA-ADENOSINE", "non-polymer"),
        "AD0": ("DIGALACTURONIC ACID", "non-polymer"),
        "AD1": (
            "(3S)-3-HYDROXY-1-METHYL-2,3-DIHYDRO-1H-INDOLE-5,6-DIONE",
            "non-polymer",
        ),
        "QMQ": (
            "1-(4-(4-(4-((6-AMINO-9H-PURIN-9-YL)METHYL)PHENYL)BUTYL)BENZYL)-4- (DIMETHYLAMINO)PYRIDINIUM",
            "non-polymer",
        ),
        "QMP": ("CYCLOHEXYL METHYLPHOSPHONATE", "non-polymer"),
        "QMS": ("N-(QUINOLIN-8-YL)METHANESULFONAMIDE", "non-polymer"),
        "QMR": ("VARENICLINE", "non-polymer"),
        "AD8": ("1-(ADAMANTAN-1-YL)-2-(1H-IMIDAZOL-1-YL)ETHANONE", "non-polymer"),
        "AD9": ("ADP METAVANADATE", "non-polymer"),
        "QMI": (
            "4-BROMO-2,6-DICHLORO-N-(1,3,5-TRIMETHYL-1H-PYRAZOL-4-YL)BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "UIC": (
            "(3AS,5R,6AR)-HEXAHYDRO-2H-CYCLOPENTA[B]FURAN-5-YL {(1S,2R)-1-BENZYL-2-HYDROXY-3-[{[4-(HYDROXYMETHYL)PHENYL]SULFONYL}(ISOBUTYL)AMINO]PROPYL}CARBAMATE",
            "non-polymer",
        ),
        "QMM": ("N,N-DIMETHYL-L-GLUTAMINE", "peptide"),
        "GM3": (
            "N-{1-[(HEXOPYRANOSYLOXY)METHYL]-2-HYDROXYNONADECYL}TETRACOSANAMIDE",
            "non-polymer",
        ),
        "GM2": (
            "N-[(1S,2S)-2-HYDROXY-1-({[(2R,3R,4S,5S,6R)-3,4,5-TRIHYDROXY-6-(HYDROXYMETHYL)TETRAHYDRO-2H-PYRAN-2-YL]OXY}METHYL)OCTADECYL]OCTADECANAMIDE",
            "non-polymer",
        ),
        "GM1": ("AMINOMETHYLAMIDE", "non-polymer"),
        "GM0": ("4-O-PHOSPHONO-L-GLYCERO-ALPHA-D-MANNO-HEPTOPYRANOSE", "non-polymer"),
        "GM6": (
            "3-(N-HYDROXYCARBOXAMIDO)-2-ISOBUTYLPROPANOYL-TRP-METHYLAMIDE",
            "non-polymer",
        ),
        "ADV": ("ALPHA-BETA METHYLENE ADP-RIBOSE", "non-polymer"),
        "ADW": ("ADENOSINE-5'-DITUNGSTATE", "non-polymer"),
        "ADT": ("3'-DEOXY-3'-ACETAMIDO-THYMIDINE", "non-polymer"),
        "ADU": ("3'-DEOXY-3'-ACETAMIDO-URIDINE", "non-polymer"),
        "ADR": ("2,6-DIAMINO-2,3,6-TRIDEOXY-ALPHA-D-RIBO-HEXOPYRANOSYL", "saccharide"),
        "ADS": ("ADENOSINE-5'-(DITHIO)PHOSPHATE", "non-polymer"),
        "ADP": ("ADENOSINE-5'-DIPHOSPHATE", "non-polymer"),
        "ADQ": ("ADENOSINE-5'-DIPHOSPHATE-GLUCOSE", "non-polymer"),
        "GMC": (
            "(2R,3R,4S,5S)-4-AMINO-2-[6-(DIMETHYLAMINO)-9H-PURIN-9-YL]-5-(HYDROXYMETHYL)TETRAHYDRO-3-FURANOL",
            "non-polymer",
        ),
        "GMB": (
            "1,7-DI-O-PHOSPHONO-L-GLYCERO-BETA-D-MANNO-HEPTOPYRANOSE",
            "non-polymer",
        ),
        "GMA": ("4-AMIDO-4-CARBAMOYL-BUTYRIC ACID", "peptide"),
        "ADZ": ("9-METHYL-9H-PURIN-6-AMINE", "non-polymer"),
        "ADX": ("ADENOSINE-5'-PHOSPHOSULFATE", "nucleic-acid"),
        "ADY": ("3'-OXO-ADENOSINE", "non-polymer"),
        "ADF": (
            "4-{[(2R,5S)-5-{[(2S)-2-(AMINOMETHYL)PYRROLIDIN-1-YL]CARBONYL}PYRROLIDIN-2-YL]METHOXY}-3-TERT-BUTYLBENZOIC ACID",
            "non-polymer",
        ),
        "ADG": ("O-3-AMINO-3-DEOXY-ALPHA-D-GLUCOPYRANOSE", "saccharide"),
        "ADD": ("2,6,8-TRIMETHYL-3-AMINO-9-BENZYL-9-METHOXYNONANOIC ACID", "peptide"),
        "ADE": ("ADENINE", "non-polymer"),
        "ADB": (
            "4-[4-AMINO-6-(2,6-DICHLORO-PHENOXY)-[1,3,5]TRIAZIN-2-YLAMINO]-BENZONITRILE",
            "non-polymer",
        ),
        "ADC": (
            "(1'R,2'S)-9-(2-HYDROXY-3'-KETO-CYCLOPENTEN-1-YL)ADENINE",
            "non-polymer",
        ),
        "ADA": ("ALPHA-D-GALACTOPYRANURONIC ACID", "saccharide"),
        "ADN": ("ADENOSINE", "non-polymer"),
        "ADO": ("ADAMANTANONE", "non-polymer"),
        "ADL": (
            "(1,8-DIHYDROXY-9-OXO-9,10-DIHYDRO-ANTHRACEN-2-YL)-ACETIC ACID",
            "non-polymer",
        ),
        "ADM": ("ADAMANTANE", "non-polymer"),
        "ADJ": (
            "NICOTINAMIDE-ADENINE-DINUCLEOTIDE-ADENYLATE INTERMEDIATE",
            "non-polymer",
        ),
        "ADK": ("3-METHYL-3H-PURIN-6-YLAMINE", "non-polymer"),
        "ADH": ("1-AMINO-2,3-DIHYDROXY-5-HYDROXYMETHYL CYCLOHEX-5-ENE", "non-polymer"),
        "ADI": ("2',3'-DIDEOXYADENOSINE-5'-DIPHOSPHATE", "non-polymer"),
        "TE7": (
            "5-{[2-(DIMETHYLAMINO)ETHYL]AMINO}-8-METHOXY-6H-IMIDAZO[4,5,1-DE]ACRIDIN-6-ONE",
            "non-polymer",
        ),
        "BRZ": (
            "6-(1,1-DIMETHYLALLYL)-2-(1-HYDROXY-1-METHYLETHYL)-2,3-DIHYDRO-7H-FURO[3,2-G]CHROMEN-7-ONE",
            "non-polymer",
        ),
        "BRY": ("?", "non-polymer"),
        "BRX": ("BROMAMPHENICOL", "non-polymer"),
        "BRW": ("6-BROMOINDIRUBIN-3'-OXIME", "non-polymer"),
        "BRV": ("5-AMINO-2,4,6-TRIBROMOBENZENE-1,3-DICARBOXYLIC ACID", "non-polymer"),
        "BRU": ("5-BROMO-2'-DEOXYURIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "BRT": ("5-BROMONICOTINAMIDE", "non-polymer"),
        "BRS": ("2-[1-(4-CHLORO-PHENYL)-ETHYL]-4,6-DINITRO-PHENOL", "non-polymer"),
        "BRR": ("(2R)-2-{[FORMYL(HYDROXY)AMINO]METHYL}HEXANOIC ACID", "non-polymer"),
        "BRQ": (
            "{[4-AMINO-2-(3-CHLOROANILINO)-1,3-THIAZOL-5-YL](4-FLUOROPHENYL)METHANONE",
            "non-polymer",
        ),
        "BRP": ("2-BROMO-2-PROPENE-1-OL", "non-polymer"),
        "BRO": ("BROMO GROUP", "non-polymer"),
        "BRN": ("BERENIL", "non-polymer"),
        "BRM": ("2-BROMOACETYL GROUP", "non-polymer"),
        "BRL": (
            "2,4-THIAZOLIDIINEDIONE, 5-[[4-[2-(METHYL-2-PYRIDINYLAMINO)ETHOXY]PHENYL]METHYL]-(9CL)",
            "non-polymer",
        ),
        "BRK": (
            "{(2Z)-4-AMINO-2-[(4-METHOXYPHENYL)IMINO]-2,3-DIHYDRO-1,3-THIAZOL-5-YL}(4-METHOXYPHENYL)METHANONE",
            "non-polymer",
        ),
        "BRJ": ("2-BROMOETHANOL", "non-polymer"),
        "BRI": ("(O4)-METHYL-2-DEOXYFUCOSE", "saccharide"),
        "BRH": (
            "(S)-2-AMINO-3-(4-BROMO-3-HYDROXY-ISOXAZOL-5-YL)PROPIONIC ACID",
            "non-polymer",
        ),
        "BRG": ("8-BROMOGUANINE", "non-polymer"),
        "BRF": ("?", "non-polymer"),
        "BRE": (
            "2-BIPHENYL-4-YL-6-FLUORO-3-METHYL-QUINOLINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "BRD": ("1-BETA-RIBOFURANOSYL-1,3-DIAZEPINONE", "non-polymer"),
        "BRC": ("12-BROMODODECANOIC ACID", "non-polymer"),
        "BRB": ("PARA-BROMOBENZYL ALCOHOL", "non-polymer"),
        "BRA": (
            "9-[4-(N,N-DIMETHYLAMINO)PHENYLAMINO]-3,6-BIS(3-PYRROLIDINOPROPIONAMIDO) ACRIDINE",
            "non-polymer",
        ),
        "X9Q": (
            "{(2S)-2-[(1S)-1-AMINOETHYL]-4-BENZYL-5-OXO-2,5-DIHYDRO-1H-IMIDAZOL-1-YL}ACETIC ACID",
            "peptide",
        ),
        "1PI": (
            "(3R)-3-[(2S)-2-AMINO-3-OXOPROPYL]PIPERIDINE-1-CARBOXIMIDAMIDE",
            "peptide",
        ),
        "1PN": (
            "PROP-2-EN-1-YL (2S)-2-[(2S,3R)-3-(ACETYLOXY)-1-OXOBUTAN-2-YL]-2,3-DIHYDRO-1,3-THIAZOLE-4-CARBOXYLATE",
            "non-polymer",
        ),
        "1PL": (
            "(1S,3AS,5AR,8AS)-1,7,7-TRIMETHYL-1,2,3,3A,5A,6,7,8-OCTAHYDROCYCLOPENTA[C]PENTALENE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "1PM": (
            "METHYL (3S)-3-{2-[(1,3-BENZODIOXOL-5-YLMETHYL)AMINO]-2-OXOETHYL}-4-[2-(1H-IMIDAZOL-1-YL)PYRIMIDIN-4-YL]PIPERAZINE-1-CARBOXYLATE",
            "non-polymer",
        ),
        "1PB": ("1,4-DIPHENYL-2-BUTENE", "non-polymer"),
        "1PC": ("1-(PHENYL-1-CYCLOHEXYL)PIPERIDINE", "non-polymer"),
        "XIM": ("5,6,7,8-TETRAHYDRO-IMIDAZO[1,2-A]PYRIDINE-6,7,8-TRIOL", "non-polymer"),
        "1PA": ("4-(CARBOXYMETHYL)-L-PHENYLALANINE", "peptide"),
        "1PG": (
            "2-(2-{2-[2-(2-METHOXY-ETHOXY)-ETHOXY]-ETHOXY}-ETHOXY)-ETHANOL",
            "non-polymer",
        ),
        "1PD": (
            "2-({3-[5-HYDROXY-2-(3-METHYLBUTYL)-3-OXO-6-(1,3-THIAZOL-5-YL)-2,3-DIHYDROPYRIDAZIN-4-YL]-1,1-DIOXIDO-2H-1,2,4-BENZOTHIADIAZIN-7-YL}OXY)ACETAMIDE",
            "non-polymer",
        ),
        "1PE": ("PENTAETHYLENE GLYCOL", "non-polymer"),
        "1PY": ("2-OXO-3-PHENYLPROPIONIC ACID", "non-polymer"),
        "P2G": (
            "(2S,4R,6R,6AS)-4-(2-AMINO-6-OXO-1,6-DIHYDROPURIN-9-YL)-6-(HYDROXYMETHYL)-TETRAHYDROFURO[3,4-D][1,3]DIOXOL-2-YLPHOSPHONIC ACID",
            "non-polymer",
        ),
        "1PR": (
            "PHOSPHORIC ACID MONO-[5-({[5-CARBAMOYL-3-(5-PHOSPHONOOXY-5-DEOXY-RIBOFURANOSYL)- 3H-IMIDAZOL-4-YLAMINO]-METHYL}-AMINO)-2,3,4-TRIHYDROXY-PENTYL] ESTER",
            "nucleic-acid",
        ),
        "1PS": ("3-PYRIDINIUM-1-YLPROPANE-1-SULFONATE", "non-polymer"),
        "1PP": (
            "1-(3-TERT-BUTYL-1-PHENYL-1H-PYRAZOL-5-YL)-3-(2,3-DICHLOROPHENYL)UREA",
            "non-polymer",
        ),
        "1PQ": ("(4S)-N~4~-(6-METHOXYQUINOLIN-8-YL)PENTANE-1,4-DIAMINE", "non-polymer"),
        "1PT": ("CYCLOHEXANE-1(R),2(R)-DIAMINE-PLATINUM(II)", "non-polymer"),
        "1PU": (
            "1-(5-OXO-2,3,5,9B-TETRAHYDRO-1H-PYRROLO[2,1-A]ISOINDOL-9-YL)-3-PYRIDIN-2-YL-UREA",
            "non-polymer",
        ),
        "ES7": ("1-METHYL-1H-BENZIMIDAZOL-2-AMINE", "non-polymer"),
        "ES6": ("2-OXO-1,2-DIHYDROPYRIDINE-3-CARBOXYLIC ACID", "non-polymer"),
        "ES5": ("QUINOLIN-6-AMINE", "non-polymer"),
        "ES4": ("6-CHLORO-9H-PURINE", "non-polymer"),
        "ES3": ("4-BROMO-1H-IMIDAZOLE", "non-polymer"),
        "ES2": ("6-CHLORO-1,3-DIHYDRO-2H-INDOL-2-ONE", "non-polymer"),
        "ES1": ("QUINOLIN-4-OL", "non-polymer"),
        "ES0": ("2-AMINO-1H-BENZIMIDAZOL-7-OL", "non-polymer"),
        "DI6": ("3-[(4S)-2,5-DIOXOIMIDAZOLIDIN-4-YL]PROPANOIC ACID", "non-polymer"),
        "LTT": (
            "4-AMINO-1-[(2S,4S)-2-(HYDROXYMETHYL)-1,3-DIOXOLAN-4-YL]PYRIMIDIN-2(1H)-ONE",
            "non-polymer",
        ),
        "ES9": ("5-CHLORO-1H-BENZIMIDAZOLE", "non-polymer"),
        "XCA": (
            "(8ALPHA,10ALPHA,25S)-3-HYDROXYCHOLESTA-3,5-DIEN-26-OIC ACID",
            "non-polymer",
        ),
        "PB5": (
            "N,N'-BIS(PYRIDIN-3-YLMETHYL)PYRIMIDINE-4,6-DICARBOXAMIDE",
            "non-polymer",
        ),
        "EST": ("ESTRADIOL", "non-polymer"),
        "ESR": ("5-ALPHA-ESTRAN-3,17-DIONE", "non-polymer"),
        "ESP": ("THIENO[2,3-B]PYRIDINE-2-CARBOXAMIDINE", "non-polymer"),
        "KLN": (
            "1-ACETYL-4-(4-{[(2S,4R)-2-(2,4-DICHLOROPHENYL)-2-(1H-IMIDAZOL-1-YLMETHYL)-1,3-DIOXOLAN-4-YL]METHOXY}PHENYL)PIPERAZINE",
            "non-polymer",
        ),
        "KLH": ("N-HYDROXY-1-[(2-PHENYLETHYL)SULFONYL]-D-PROLINAMIDE", "non-polymer"),
        "6AH": ("6-AMINOHEXAN-1-OL", "non-polymer"),
        "ESY": (
            "BENZYL N-[(BENZYLOXY)CARBONYL]-L-ALANYL-N~6~-[(2R,3S,4S)-3-FORMYL-2-HYDROXY-4-METHYLHEXANOYL]-L-LYSINATE",
            "non-polymer",
        ),
        "ESX": ("BENZO[B]THIOPHENE-2-CARBOXAMIDINE", "non-polymer"),
        "ESG": (
            "L-GAMMA-GLUTAMYL-S-[(3S)-1-ETHYL-2,5-DIOXOPYRROLIDIN-3-YL]-L-CYSTEINYLGLYCINE",
            "non-polymer",
        ),
        "6AI": ("1H-INDAZOL-6-AMINE", "non-polymer"),
        "ESC": ("2-AMINO-4-ETHYL SULFANYL BUTYRIC ACID", "peptide"),
        "ESB": (
            "3-[(3E)-3-(ETHYLIMINO)-4-HYDROXY-6-OXOCYCLOHEXA-1,4-DIEN-1-YL]-L-ALANINE",
            "peptide",
        ),
        "ESA": ("ETHANESULFONIC ACID", "non-polymer"),
        "KLS": ("1-METHYL-3-(N-PHENYLCARBAMIMIDOYL)UREA", "non-polymer"),
        "ESO": ("O3-PHOSPHONOESTRONE", "non-polymer"),
        "6AN": (
            "6-[(1E)-CYCLOHEXA-2,4-DIEN-1-YLIDENEAMINO]NAPHTHALENE-2-SULFONATE",
            "non-polymer",
        ),
        "ESM": ("1,3,5(10)-ESTRATRIEN-2,3,17-BETA-TRIOL 2-METHYL ETHER", "non-polymer"),
        "ESL": ("ESTRIOL", "non-polymer"),
        "ESI": ("4-IODOBENZO[B]THIOPHENE-2-CARBOXAMIDINE", "non-polymer"),
        "ESH": (
            "4- [(1R,3AS,4R,8AS,8BR)- 2- (4-CHLOROBENZYL)- 1- ISOPROPYL- 3- OXODECAHYDROPYRROLO[3,4- A]PYRROLIZIN- 4- YL]BENZENECARBOXIMIDAMIDE",
            "non-polymer",
        ),
        "LTC": (
            "(2S)-1-({3-O-[2-(ACETYLAMINO)-4-AMINO-2,4,6-TRIDEOXY-BETA-D-GALACTOPYRANOSYL]-ALPHA-D-GLUCOPYRANOSYL}OXY)-3-(HEPTANOYLOXY)PROPAN-2-YL (7Z)-PENTADEC-7-ENOATE",
            "non-polymer",
        ),
        "X92": ("RAMIPRILAT", "non-polymer"),
        "X93": ("TRANDOLAPRILAT", "non-polymer"),
        "X94": ("PERINDOPRILAT", "non-polymer"),
        "SCK": (
            "2,2'-[(1,4-DIOXOBUTANE-1,4-DIYL)BIS(OXY)]BIS(N,N,N-TRIMETHYLETHANAMINIUM)",
            "non-polymer",
        ),
        "X96": ("2-{[4-(AMINOMETHYL)BENZYL]AMINO}-5-NITROBENZAMIDE", "non-polymer"),
        "X97": (
            "3-{4-OXO-4-[(2S)-2-(PYRROLIDIN-1-YLCARBONYL)PYRROLIDIN-1-YL]BUTYL}-5,5-DIPHENYLIMIDAZOLIDINE-2,4-DIONE",
            "non-polymer",
        ),
        "X98": (
            "1-{3-OXO-3-[(2S)-2-(PYRROLIDIN-1-YLCARBONYL)PYRROLIDIN-1-YL]PROPYL}-3-PHENYLQUINOXALIN-2(1H)-ONE",
            "non-polymer",
        ),
        "X99": (
            "2-{3-[(2S)-4,4-DIFLUORO-2-(PYRROLIDIN-1-YLCARBONYL)PYRROLIDIN-1-YL]-3-OXOPROPYL}-ISOINDOLE-1,3(2H)-DIONE",
            "non-polymer",
        ),
        "SCJ": (
            "3-BROMO-5-PHENYL-N-(PYRIDIN-3-YLMETHYL)PYRAZOLO[1,5-A]PYRIMIDIN-7-AMINE",
            "non-polymer",
        ),
        "LTA": ("4,8-DIAMINO-3-HYDROXY-OCTANOIC ACID ETHYL ESTER", "peptide"),
        "SCE": (
            "3-METHYL-N-(PYRIDIN-4-YLMETHYL)IMIDAZO[1,2-A]PYRAZIN-8-AMINE",
            "non-polymer",
        ),
        "XCG": ("2-(2-BENZOFURANYL)-2-IMIDAZOLINE", "non-polymer"),
        "SCD": ("SUCCINYL(CARBADETHIA)-COENZYME A", "non-polymer"),
        "1P2": ("?", "non-polymer"),
        "1P3": ("(3R)-1-ACETYL-3-METHYLPIPERIDINE", "non-polymer"),
        "1P1": ("CYCLIC POLYAMIDE", "non-polymer"),
        "1P6": (
            "3-(6-{[(1S)-1-(4-FLUOROPHENYL)ETHYL]AMINO}PYRIMIDIN-4-YL)PYRAZOLO[1,5-A]PYRIMIDIN-2-AMINE",
            "non-polymer",
        ),
        "SCG": (
            "METHYL (2S,3R,4S)-2-(BETA-D-GLUCOPYRANOSYLOXY)-4-(2-OXOETHYL)-3-VINYL-3,4-DIHYDRO-2H-PYRAN-5-CARBOXYLATE",
            "non-polymer",
        ),
        "1P5": (
            "(3S)-1-[6-(2-AMINOPYRAZOLO[1,5-A]PYRIMIDIN-3-YL)PYRIMIDIN-4-YL]-N,N-DIETHYLPIPERIDINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "P2O": ("3-PROP-2-YN-1-YL-1,3-OXAZOLIDIN-2-ONE", "non-polymer"),
        "SCF": (
            "5-(2-FLUOROPHENYL)-N-(PYRIDIN-4-YLMETHYL)PYRAZOLO[1,5-A]PYRIMIDIN-7-AMINE",
            "non-polymer",
        ),
        "PB9": (
            "METHYL 4-{[{(3S)-6-CYANO-1-[(1-METHYL-1H-IMIDAZOL-5-YL)METHYL]-1,2,3,4-TETRAHYDROQUINOLIN-3-YL}(PYRIDIN-2-YLSULFONYL)AMINO]METHYL}PIPERIDINE-1-CARBOXYLATE",
            "non-polymer",
        ),
        "5GF": ("5-FLUORO-BETA-D-GLUCOPYRANOSE", "saccharide"),
        "K1R": (
            "(2S)-2-AMINO-4-[({[(2R)-2-AMINO-2-CARBOXYETHYL]THIO}AMINO)SULFINYL]BUTANOIC ACID",
            "peptide",
        ),
        "LFR": ("BETA-L-FRUCTO-FURANOSE", "non-polymer"),
        "V7O": ("META VANADATE", "non-polymer"),
        "PYZ": ("4-IODOPYRAZOLE", "non-polymer"),
        "5GP": ("GUANOSINE-5'-MONOPHOSPHATE", "non-polymer"),
        "5GW": ("?", "non-polymer"),
        "LTI": (
            "(5Z)-5-(CARBAMOYLIMINO)-3-[(5R)-6,7,8,9-TETRAHYDRO-5H-BENZO[7]ANNULEN-5-YLSULFANYL]-2,5-DIHYDROISOTHIAZOLE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "MNH": ("MANGANESE PROTOPORPHYRIN IX", "non-polymer"),
        "G5A": ("5'-O-(GLYCYLSULFAMOYL)ADENOSINE", "non-polymer"),
        "HJ3": (
            "(2R,4S)-2-[(1R)-1-{[(6S)-6-AMINO-6-CARBOXYHEXANOYL]AMINO}-2-OXOETHYL]-5,5-DIMETHYL-1,3-THIAZOLIDINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "MNI": (
            "N-{[(2R,3R,4R,5R)-3,4-DIHYDROXY-5-(HYDROXYMETHYL)PYRROLIDIN-2-YL]METHYL}-4-(DIMETHYLAMINO)BENZAMIDE",
            "non-polymer",
        ),
        "HJ2": (
            "(2R)-2-[(R)-{[(6S)-6-AMINO-6-CARBOXYHEXANOYL]AMINO}(CARBOXY)METHYL]-5-METHYL-3,6-DIHYDRO-2H-1,3-THIAZINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "GLR": ("2,3-DIHYDROXY-5-OXO-HEXANEDIOATE", "non-polymer"),
        "LTN": ("L-TRYPTOPHANAMIDE", "non-polymer"),
        "TIO": ("(2-MERCAPTOMETHYL-3-PHENYL-PROPIONYL)-GLYCINE", "non-polymer"),
        "G5P": ("?", "non-polymer"),
        "SCY": ("S-ACETYL-CYSTEINE", "peptide"),
        "LTL": ("6-HYDROXY-6-METHYL-HEPTAN-3-ONE", "non-polymer"),
        "SCX": (
            "N-((2-AMINOPYRIMIDIN-5-YL)METHYL)-5-(2,6-DIFLUOROPHENYL)-3-ETHYLPYRAZOLO[1,5-A]PYRIMIDIN-7-AMINE",
            "non-polymer",
        ),
        "XXB": ("1-BUTYL-1H-PYRAZOLE-5-CARBOXYLIC ACID", "non-polymer"),
        "OPS": ("3-OXO-3-[(3-OXOPROPYL)SULFANYL]PROPANE-1-THIOLATE", "non-polymer"),
        "SCZ": (
            "3-CYCLOPROPYL-5-PHENYL-N-(PYRIDIN-3-YLMETHYL)PYRAZOLO[1,5-A]PYRIMIDIN-7-AMINE",
            "non-polymer",
        ),
        "TIH": ("BETA(2-THIENYL)ALANINE", "peptide"),
        "G52": ("?", "non-polymer"),
        "G50": (
            "3-(4-HYDROXYPHENYL)-1-(2,4,6-TRIHYDROXYPHENYL)PROPAN-1-ONE",
            "non-polymer",
        ),
        "G55": (
            "(5R)-1,3-DIOXEPAN-5-YL [(1S,2R)-1-BENZYL-2-HYDROXY-3-{[(4-METHOXYPHENYL)SULFONYL](2-METHYLPROPYL)AMINO}PROPYL]CARBAMATE",
            "non-polymer",
        ),
        "MNA": ("2-O-METHYL-5-N-ACETYL-ALPHA-D- NEURAMINIC ACID", "non-polymer"),
        "G58": (
            "N-(2-CHLORO-6-FLUOROBENZYL)-1-METHYL-N-{[3'-(METHYLSULFONYL)BIPHENYL-4-YL]METHYL}-1H-IMIDAZOLE-4-SULFONAMIDE",
            "non-polymer",
        ),
        "TL5": (
            "(5R)-4-HYDROXY-3,5-DIMETHYL-5-[(1E,3E)-2-METHYLPENTA-1,3-DIENYL]THIOPHEN-2(5H)-ONE",
            "non-polymer",
        ),
        "SCW": (
            "3-BROMO-6-PHENYL-N-(PYRIMIDIN-5-YLMETHYL)IMIDAZO[1,2-A]PYRIDIN-8-AMINE",
            "non-polymer",
        ),
        "SOT": ("MORPHOLINE-4-SULFONIC ACID", "non-polymer"),
        "MNC": ("METHYL ISOCYANIDE", "non-polymer"),
        "4WF": (
            "5-AZANYL-6-OXIDANYL-1-(PHENYLMETHYL)PYRIMIDINE-2,4-DIONE",
            "non-polymer",
        ),
        "SCQ": (
            "3-BROMO-5-PHENYL-N-(PYRIMIDIN-5-YLMETHYL)PYRAZOLO[1,5-A]PYRIDIN-7-AMINE",
            "non-polymer",
        ),
        "NI8": (
            "(4AS,7AS)-1,4-BIS[3-(HYDROXYMETHYL)BENZYL]HEXAHYDRO-1H-PYRROLO[3,4-B]PYRAZINE-2,3-DIONE",
            "non-polymer",
        ),
        "SCP": (
            "[3-HYDROXY-2-METHYL-5-PHOSPHONOOXYMETHYL-PYRIDIN-4-YLMETHYL]-N,O-CYCLOSERYLAMIDE",
            "non-polymer",
        ),
        "NI9": ("3-FLUORO-1-(2-HYDROXY-2,2-DIPHOSPHONOETHYL)PYRIDINIUM", "non-polymer"),
        "SCS": ("3-(ETHYLDISULFANYL)-L-ALANINE", "peptide"),
        "XAX": (
            "{[(5AR,8R,9AR)-2-AMINO-4-OXO-6,7-DI(SULFANYL-KAPPAS)-3,5,5A,8,9A,10-HEXAHYDRO-4H-PYRANO[3,2-G]PTERIDIN-8-YL]METHYL DIHYDROGENATO(2-) PHOSPHATE}(HYDROXY)OXO(THIOXO)MOLYBDENUM",
            "non-polymer",
        ),
        "MNG": ("MENOGARIL", "non-polymer"),
        "3EV": (
            "(1S,3S,5Z,7E,14BETA,17ALPHA)-9,10-SECOCHOLESTA-5,7,10-TRIENE-1,3,25-TRIOL",
            "non-polymer",
        ),
        "3ET": (
            "O-[(2R)-2-AMINO-3-(D-SERYLOXY)PROPANOYL]-N-[(2,3-DIHYDROXYPHENYL)CARBONYL]-L-SERINE",
            "non-polymer",
        ),
        "3ES": (
            "[(2S)-2-({3-[HYDROXYL(2-PHENYL-(1R)-1-{[(BENZYLOXY)[(2S)-2-({3-[HYDROXYL(2-PHENYL-(1R)-1-CARBONYL]-AMINO}ETHYL)PHOSPHINYL]-2-[(3-PHENYLISOXAZOL-5-YL)METHYL]-1-OXO-PROPYL}AMINO)-3-(4-HYDROXY-PHENYL)",
            "non-polymer",
        ),
        "3EQ": ("?", "non-polymer"),
        "3EP": ("TRIETHYLPHOSPHANE", "non-polymer"),
        "ZDU": ("5-(3-AMINOPROPYL)-2'-DEOXYURIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "ZDV": ("1-(3-FLUOROPHENYL)METHANAMINE", "non-polymer"),
        "ZDP": ("(2S,6S)-2,6-DIAMINO-2-METHYLHEPTANEDIOIC ACID", "non-polymer"),
        "ZDR": ("(2R,6S)-2,6-DIAMINO-2-METHYLHEPTANEDIOIC ACID", "non-polymer"),
        "3EG": ("(2S)-2-AMINO-4,4,4-TRIFLUOROBUTANOIC ACID", "peptide"),
        "NI2": ("NICKEL (II) ION, 2 WATERS COORDINATED", "non-polymer"),
        "8H1": (
            "2-(5-{[(2S)-2-AMINO-3-PHENYLPROPYL]OXY}PYRIDIN-3-YL)-8,9-DIMETHOXYBENZO[C][2,7]NAPHTHYRIDIN-4-AMINE",
            "non-polymer",
        ),
        "3EA": (
            "2-{5-[3-(7-PROPYL-3-TRIFLUOROMETHYLBENZO[D]ISOXAZOL-6-YLOXY)PROPOXY]INDOL-1-YL}ETHANOIC ACID",
            "non-polymer",
        ),
        "YLO": (
            "1-[2-(4-METHYLPHENOXY)ETHYL]-2-(2-PHENOXYETHYLSULFANYL)BENZIMIDAZOLE",
            "non-polymer",
        ),
        "PBQ": ("PENTABROMOPSEUDILIN", "non-polymer"),
        "2Y4": (
            "N-[(2S)-1-[[(2S)-1-[[(2S,3R)-1-CYCLOHEXYL-4,4-DIFLUORO-3-HYDROXY-5-(METHYLAMINO)-5-OXO-PENTAN-2-YL]AMINO]-1-OXO-HEXAN-2-YL]AMINO]-1-OXO-3-PHENYL-PROPAN-2-YL]MORPHOLINE-4-CARBOXAMIDE",
            "peptide",
        ),
        "2Y2": (
            "N-[(2R)-1-{[(2S,3R,5R)-1-CYCLOHEXYL-3-HYDROXY-5-{[2-(MORPHOLIN-4-YL)ETHYL]CARBAMOYL}OCT-7-YN-2-YL]AMINO}-3-(METHYLSULFANYL)-1-OXOPROPAN-2-YL]-1H-BENZIMIDAZOLE-2-CARBOXAMIDE",
            "peptide",
        ),
        "3EJ": (
            "4-[(5-{2-[(3-FLUOROBENZYL)CARBAMOYL]PYRIDIN-4-YL}-2H-TETRAZOL-2-YL)METHYL]BENZOIC ACID",
            "non-polymer",
        ),
        "PBP": ("1-(4-BROMO-PHENYL)-ETHANONE", "non-polymer"),
        "P2N": (
            "(5Z)-13-CHLORO-14,16-DIHYDROXY-3,4,7,8,9,10-HEXAHYDRO-1H-2-BENZOXACYCLOTETRADECINE-1,11(12H)-DIONE",
            "non-polymer",
        ),
        "IFS": (
            "BIS(1-METHYLETHYL) [2-(SULFANYLMETHYL)PHENYL]PHOSPHONATE",
            "non-polymer",
        ),
        "524": (
            "(2S,3S)-3-AMINO-4-[(3S)-3-FLUOROPYRROLIDIN-1-YL]-N,N-DIMETHYL-4-OXO-2-(TRANS-4-[1,2,4]TRIAZOLO[1,5-A]PYRIDIN-5-YLCYCLOHEXYL)BUTANAMIDE",
            "non-polymer",
        ),
        "MN0": (
            "METHYL 3,5-DIDEOXY-5-[(HYDROXYACETYL)AMINO]-D-GLYCERO-ALPHA-D-GALACTO-NON-2-ULOPYRANOSIDONIC ACID",
            "saccharide",
        ),
        "527": (
            "5-(3-{[1-(BENZYLSULFONYL)PIPERIDIN-4-YL]AMINO}PHENYL)-4-BROMO-3-(CARBOXYMETHOXY)THIOPHENE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "521": (
            "5-(4-CHLORO-5-PHENYL-3-THIENYL)-1,2,5-THIADIAZOLIDIN-3-ONE 1,1-DIOXIDE",
            "non-polymer",
        ),
        "523": ("5-METHYL-2'-DEOXY-CYTIDINE-5'-TRIPHOSPHATE", "non-polymer"),
        "X22": (
            "(3S,6S,7S,9AS)-6-{[(2S)-2-AMINOBUTANOYL]AMINO}-N-(DIPHENYLMETHYL)-7-(HYDROXYMETHYL)-5-OXOOCTAHYDRO-1H-PYRROLO[1,2-A]AZEPINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "528": (
            "(2S)-2-[3-(AMINOMETHYL)PHENYL]-3-{(R)-HYDROXY[(1R)-2-METHYL-1-{[(3-PHENYLPROPYL)SULFONYL]AMINO}PROPYL]PHOSPHORYL}PROPANOIC ACID",
            "non-polymer",
        ),
        "529": (
            "(2S)-N-[(3Z)-5-CYCLOPROPYL-3H-PYRAZOL-3-YLIDENE]-2-[4-(2-OXOIMIDAZOLIDIN-1-YL)PHENYL]PROPANAMIDE",
            "non-polymer",
        ),
        "52D": (
            "PYRIDIN-4-YL[4-(2-PYRROLIDIN-1-YLETHOXY)PHENYL]METHANONE",
            "non-polymer",
        ),
        "P2I": ("1,2-DIACETOYL-SN-GLYCERO-3-PHOSPHOINOSITOL", "non-polymer"),
        "MN3": ("MANGANESE (III) ION", "non-polymer"),
        "52A": ("(2R,4R)-4-AMINOPYRROLIDINE-2,4-DICARBOXYLIC ACID", "non-polymer"),
        "52B": (
            "(2S)-6,8-DICHLORO-2-(TRIFLUOROMETHYL)-2H-CHROMENE-3-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "52C": ("5-(2-CHLOROPHENYL)-1H-TETRAZOLE", "non-polymer"),
        "52H": (
            "5'-O-{[(2R)-2-HYDROXY-3,3-DIMETHYLBUTANOYL]SULFAMOYL}ADENOSINE",
            "non-polymer",
        ),
        "52I": (
            "(2E)-3-{5-[(2,4-DIAMINOPYRIMIDIN-5-YL)METHYL]-2,3-DIMETHOXYPHENYL}-1-[(1R)-1-(2-METHYLPROPYL)PHTHALAZIN-2(1H)-YL]PROP-2-EN-1-ONE",
            "non-polymer",
        ),
        "52J": (
            "(2E)-3-{5-[(2,4-DIAMINOPYRIMIDIN-5-YL)METHYL]-2,3-DIMETHOXYPHENYL}-1-[(1S)-1-(2-METHYLPROPYL)PHTHALAZIN-2(1H)-YL]PROP-2-EN-1-ONE",
            "non-polymer",
        ),
        "MN5": ("MANGANESE ION, 5 WATERS COORDINATED", "non-polymer"),
        "52V": (
            "5-[(3S)-3-(5-METHOXY-2',6'-DIMETHYLBIPHENYL-3-YL)BUT-1-YN-1-YL]-6-METHYLPYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "52P": (
            "5-(2,6-DICHLOROPHENYL)-2-[(2,4-DIFLUOROPHENYL)SULFANYL]-6H-PYRIMIDO[1,6-B]PYRIDAZIN-6-ONE",
            "non-polymer",
        ),
        "UIQ": (
            "(1R,3AS,4R,8AS,8BR)-4-(2-BENZO[1,3]DIOXOL-5-YLMETHYL-1-ISOPROPYL-3-OXO-DECAHYDRO-PYRROLO[3,4-A]PYRROLIZIN-4-YL)-BENZAMIDINE",
            "non-polymer",
        ),
        "PBZ": ("P-AMINO BENZAMIDINE", "non-polymer"),
        "52X": (
            "METHYL {3-[2-(ACETYLAMINO)ETHYL]-2-IODO-1H-INDOL-5-YL}CARBAMATE",
            "non-polymer",
        ),
        "WGZ": (
            "5-{6-[(1-METHYLPIPERIDIN-4-YL)OXY]-1H-BENZIMIDAZOL-1-YL}-3-{(1R)-1-[2-(TRIFLUOROMETHYL)PHENYL]ETHOXY}THIOPHENE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "PBE": ("1,1-DIMETHYL-PROLINIUM", "non-polymer"),
        "8HG": ("2'-DEOXY-8-OXOGUANOSINE", "non-polymer"),
        "12F": ("unknown", "non-polymer"),
        "PYQ": ("PYROQUILON", "non-polymer"),
        "8HI": (
            "(3R,5R)-7-{3-(4-FLUOROPHENYL)-1-(1-METHYLETHYL)-4-PHENYL-5-[(4-SULFAMOYLPHENYL)CARBAMOYL]-1H-PYRROL-2-YL}-3,5-DIHYDROXYHEPTANOIC ACID",
            "non-polymer",
        ),
        "TH8": ("THIAMPHENICOL", "non-polymer"),
        "TH7": ("2,4-DIHYDROXYBENZENESULFENIC ACID", "non-polymer"),
        "N9H": (
            "(5S)-5-(3-AMINOPROPYL)-3-(2,5-DIFLUOROPHENYL)-N-ETHYL-5-PHENYL-4,5-DIHYDRO-1H-PYRAZOLE-1-CARBOXAMIDE",
            "non-polymer",
        ),
        "PBF": ("PARA-(BENZOYL)-PHENYLALANINE", "peptide"),
        "UIP": (
            "(1R,3AS,4R,8AS,8BR)-4-(2-BENZO[1,3]DIOXOL-5-YLMETHYL-1-ETHYL-3-OXO-DECAHYDRO-PYRROLO[3,4-A]PYRROLIZIN-4-YL)-BENZAMIDINE",
            "non-polymer",
        ),
        "12I": ("4-(4-METHYLPIPERAZIN-1-YL)QUINAZOLINE", "non-polymer"),
        "SC9": (
            "6-(2-FLUOROPHENYL)-N-(PYRIDIN-3-YLMETHYL)IMIDAZO[1,2-A]PYRAZIN-8-AMINE",
            "non-polymer",
        ),
        "TH5": ("O-ACETYL-L-THREONINE", "peptide"),
        "GB": ("METHYLPHOSPHONIC ACID ESTER GROUP", "non-polymer"),
        "7TP": (
            "5-{(3R)-3-(4-HYDROXYBENZYL)-4-[(4-METHOXYPHENYL)SULFONYL]-1-[(1-METHYL-1H-IMIDAZOL-5-YL)METHYL]-2,3,4,5-TETRAHYDRO-1H-1,4-BENZODIAZEPIN-7-YL}FURAN-2-CARBALDEHYDE",
            "non-polymer",
        ),
        "7TQ": (
            "4-({(3R)-7-CYANO-4-[(4-METHOXYPHENYL)SULFONYL]-1-[(1-METHYL-1H-IMIDAZOL-5-YL)METHYL]-2,3,4,5-TETRAHYDRO-1H-1,4-BENZODIAZEPIN-3-YL}METHYL)PHENYL BENZYLCARBAMATE",
            "non-polymer",
        ),
        "0C8": (
            "4-[(4-{[2-(TRIFLUOROMETHOXY)PHENYL]AMINO}PYRIMIDIN-2-YL)AMINO]BENZOIC ACID",
            "non-polymer",
        ),
        "0C9": (
            "2-({2-[(4-CARBOXYPHENYL)AMINO]PYRIMIDIN-4-YL}AMINO)BENZOIC ACID",
            "non-polymer",
        ),
        "L3A": (
            "(2R)-3-[(R)-[(1R)-1-AMINO-3-METHYLBUTYL](HYDROXY)PHOSPHORYL]-2-METHYLPROPANOIC ACID",
            "peptide",
        ),
        "0C2": (
            "(3S)-6-(2,5-DIFLUOROBENZYL)-3-METHYL-N-(METHYLSULFONYL)-8-(2-OXO-1,2-DIHYDROPYRIDIN-3-YL)-3,6-DIHYDRO-2H-FURO[2,3-E]INDOLE-7-CARBOXAMIDE",
            "non-polymer",
        ),
        "0C3": (
            "4-[(4-{[2-(TRIFLUOROMETHYL)PHENYL]AMINO}PYRIMIDIN-2-YL)AMINO]BENZOIC ACID",
            "non-polymer",
        ),
        "0C0": (
            "4-{[4-(BIPHENYL-2-YLAMINO)PYRIMIDIN-2-YL]AMINO}BENZOIC ACID",
            "non-polymer",
        ),
        "0C1": (
            "6-(2,5-DIFLUOROBENZYL)-N-(METHYLSULFONYL)-8-(2-OXO-1,2-DIHYDROPYRIDIN-3-YL)-3,6-DIHYDRO-2H-FURO[2,3-E]INDOLE-7-CARBOXAMIDE",
            "non-polymer",
        ),
        "0C6": (
            "4-({4-[(2-CHLOROPHENYL)AMINO]-5-FLUOROPYRIMIDIN-2-YL}AMINO)BENZOIC ACID",
            "non-polymer",
        ),
        "0C7": (
            "N~4~-(2-CHLOROPHENYL)-N~2~-[4-(1H-TETRAZOL-5-YL)PHENYL]PYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "0C4": (
            "4-({4-[(2-BROMOPHENYL)AMINO]PYRIMIDIN-2-YL}AMINO)BENZOIC ACID",
            "non-polymer",
        ),
        "0C5": (
            "4-({4-[(2-CYANOPHENYL)AMINO]PYRIMIDIN-2-YL}AMINO)BENZOIC ACID",
            "non-polymer",
        ),
        "PBM": ("TRIMETHYL LEAD ION", "non-polymer"),
        "L3T": (
            "N-(2,2-DIMETHYLPROPYL)-N~2~-[1H-INDOL-3-YL(OXO)ACETYL]-L-ASPARAGINYL-N-(2-METHYLBENZYL)-3-PYRIDIN-4-YL-L-ALANINAMIDE",
            "non-polymer",
        ),
        "TH1": ("2-(BETA-D-GLUCOPYRANOSYL)-5-METHYL-BENZOTHIAZOLE", "non-polymer"),
        "L3P": (
            "2,3-DI-O-PHYTANLY-3-SN-GLYCERO-1-PHOSPHORYL-3'-SN-GLYCEROL-1'-PHOSPHATE",
            "non-polymer",
        ),
        "SC4": (
            "1[2-CHLORO-4-METHOXY-PHENYL-OXYMETHYL]-4-[2,6-DICHLORO-PHENYL-OXYMETHYL]-BENZENE",
            "non-polymer",
        ),
        "A70": (
            "N-ETHYL-N-[(4-METHYLPIPERAZIN-1-YL)CARBONYL]-D-PHENYLALANYL-N-[(1S,2S,4R)-4-(BUTYLCARBAMOYL)-1-(CYCLOHEXYLMETHYL)-2-HYDROXY-5-METHYLHEXYL]-L-NORLEUCINAMIDE",
            "non-polymer",
        ),
        "7TM": (
            "4-({(3R)-7-CYANO-4-[(4-METHOXYPHENYL)SULFONYL]-1-[(1-METHYL-1H-IMIDAZOL-5-YL)METHYL]-2,3,4,5-TETRAHYDRO-1H-1,4-BENZODIAZEPIN-3-YL}METHYL)PHENYL HEXYLCARBAMATE",
            "non-polymer",
        ),
        "7TO": (
            "4-({(3R)-7-CYANO-4-[(4-METHOXYPHENYL)SULFONYL]-1-[(1-METHYL-1H-IMIDAZOL-5-YL)METHYL]-2,3,4,5-TETRAHYDRO-1H-1,4-BENZODIAZEPIN-3-YL}METHYL)PHENYL TRIFLUOROMETHANESULFONATE",
            "non-polymer",
        ),
        "TIG": ("N-(2-AMINOETHYL)-L-TRYPTOPHAN", "peptide"),
        "XFH": (
            "6-(2-{3-[2-(2-AMINO-6-METHYLPYRIDIN-4-YL)ETHYL]PHENYL}ETHYL)-4-METHYLPYRIDIN-2-AMINE",
            "non-polymer",
        ),
        "XFK": (
            "6-(2-{5-[2-(2-AMINO-6-METHYLPYRIDIN-4-YL)ETHYL]PYRIDIN-3-YL}ETHYL)-4-METHYLPYRIDIN-2-AMINE",
            "non-polymer",
        ),
        "6A1": (
            "HYDROXY(2-{[(5S)-6-HYDROXY-8-METHOXY-4,5-DIHYDRO-3H-IMIDAZO[4,5,1-DE]ACRIDIN-5-YL]AMINO}ETHYL)DIMETHYLAMMONIUM",
            "non-polymer",
        ),
        "XFM": (
            "6-(2-{6-[2-(2-AMINO-6-METHYLPYRIDIN-4-YL)ETHYL]PYRIDIN-2-YL}ETHYL)-4-METHYLPYRIDIN-2-AMINE",
            "non-polymer",
        ),
        "U94": ("?", "non-polymer"),
        "NLT": ("N-DODECANOYL-L-TYROSINE", "non-polymer"),
        "12Q": ("1-METHYLQUINOLIN-2(1H)-ONE", "non-polymer"),
        "SC1": (
            "[3R-[3A,4A,5B(S*)]]-5-(1-CARBOXY-1-PHOSPHONOETHOXY)-4-HYDROXY-3-(PHOSPHONOOXY)-1-CYCLOHEXENE-1-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "XFE": (
            "(3R)-1-(5-METHYL-7H-PYRROLO[2,3-D]PYRIMIDIN-4-YL)PYRROLIDIN-3-AMINE",
            "non-polymer",
        ),
        "SC0": (
            "(6E,11E)-HEPTADECA-6,11-DIENE-9,9-DIYLBIS(PHOSPHONIC ACID)",
            "non-polymer",
        ),
        "UIR": (
            "(1R,3AS,4R,8AS,8BR)-4-(2-BENZO[1,3]DIOXOL-5-YL-1-CYCLOPROPYL-3-OXO-DECAHYDRO-PYRROLO[3,4-A]PYRROLIZIN-4-YL)-BENZAMIDINE",
            "non-polymer",
        ),
        "SBU": ("SEC-BUTYL GROUP", "non-polymer"),
        "SC3": (
            "7-BENZYL-1,3-DIMETHYL-8-PIPERAZIN-1-YL-3,7-DIHYDRO-PURINE-2,6-DIONE",
            "non-polymer",
        ),
        "12T": ("2-AMINOBENZENE-1,4-DICARBOXYLIC ACID", "non-polymer"),
        "6A5": (
            "SA-(+)-(6-(AMINOMETHYL)-5-(2,4-DICHLOROPHENYL)-7-METHYLIMIDAZO[1,2-A]PYRIMIDIN-2-YL)(MORPHOLINO)METHANONE",
            "non-polymer",
        ),
        "PYT": ("2-AMINOPROP-2-ENAMIDE", "non-polymer"),
        "12V": ("URIDINE DIPHOSPHO-5-THIO-N-ACETYLGLUCOSAMINE", "non-polymer"),
        "12X": (
            "(4R)-4-[(8-BROMO-7-METHOXY-2-{2-[(2-METHYLPROPANOYL)AMINO]-1,3-THIAZOL-4-YL}QUINOLIN-4-YL)OXY]-L-PROLINE",
            "peptide",
        ),
        "12Y": (
            "(4R)-4-[(7-METHOXY-2-{2-[(2-METHYLPROPANOYL)AMINO]-1,3-THIAZOL-4-YL}QUINOLIN-4-YL)OXY]-L-PROLINE",
            "peptide",
        ),
        "MRS": ("(S)-2-METHYLMYRISTOYL-COENZYME A", "non-polymer"),
        "TIA": ("2-(1,5-DIDEOXYRIBOSE)-4-AMIDO-THIAZOLE", "non-polymer"),
        "PYU": ("2-(1H-PYRROL-1-YLCARBONYL)BENZENE-1,3,5-TRIOL", "non-polymer"),
        "0CZ": ("4,4'-(1,1,1,3,3,3-HEXAFLUOROPROPANE-2,2-DIYL)DIPHENOL", "non-polymer"),
        "0CX": (
            "(2R,3AS,5AR,5BS,9S,13S,14R,16AS,16BR)-9-ETHYL-2,13-DIHYDROXY-14-METHYL-2,3,3A,5A,5B,6,9,10,11,12,13,14,16A,16B-TETRADECAHYDRO-1H-AS-INDACENO[3,2-D]OXACYCLODODECINE-7,15-DIONE",
            "non-polymer",
        ),
        "6FL": ("5,5,5,5',5',5'-HEXAFLUORO-L-LEUCINE", "peptide"),
        "0CR": (
            "MOMO-2-[4-(2-(4-(METHOXY)-1H-1,2,3-TRIAZOL-1-YL)ETHYL)BENZENESULFONAMIDE]-7,12-BIS-[3-(4-(METHOXY)-1H-1,2,3-TRIAZOL-1-YL)PROPANOIC ACID]-CRYPTOPHANE-A",
            "non-polymer",
        ),
        "0CS": ("3-[(S)-HYDROPEROXYSULFINYL]-L-ALANINE", "peptide"),
        "0CP": (
            "5-{3-[(1S,2S,4R)-BICYCLO[2.2.1]HEPT-2-YLOXY]-4-METHOXYPHENYL}TETRAHYDROPYRIMIDIN-2(1H)-ONE",
            "non-polymer",
        ),
        "0CV": (
            "6-METHOXY-3,8-DIMETHYL-4-(PIPERAZIN-1-YL)-1H-PYRAZOLO[3,4-B]QUINOLINE",
            "non-polymer",
        ),
        "0CT": ("1H-BENZOTRIAZOLE", "non-polymer"),
        "0CU": ("5,7-DICHLORO-1H-BENZOTRIAZOLE", "non-polymer"),
        "0CJ": ("4-OXOPENTYL TRIHYDROGEN DIPHOSPHATE", "non-polymer"),
        "0CK": (
            "4-[1-CYCLOHEXYL-4-(4-FLUOROPHENYL)-1H-IMIDAZOL-5-YL]PYRIMIDIN-2-AMINE",
            "non-polymer",
        ),
        "0CH": ("4-HYDROXYBUTYL TRIHYDROGEN DIPHOSPHATE", "non-polymer"),
        "L35": (
            "2-[4-({[(3,5-DICHLOROPHENYL)AMINO]CARBONYL}AMINO)PHENOXY]-2-METHYLPROPANOIC ACID",
            "non-polymer",
        ),
        "0CN": ("BUTA-2,3-DIEN-1-YL TRIHYDROGEN DIPHOSPHATE", "non-polymer"),
        "0CO": (
            "(1S,3R,5Z,7E,14BETA,17ALPHA,23R)-23-(2-HYDROXY-2-METHYLPROPYL)-20,24-EPOXY-9,10-SECOCHOLA-5,7,10-TRIENE-1,3-DIOL",
            "non-polymer",
        ),
        "0CL": ("1,3-DICHLOROPROPANE", "non-polymer"),
        "0CM": ("BUT-2-YN-1-YL TRIHYDROGEN DIPHOSPHATE", "non-polymer"),
        "0CB": (
            "{[(7-CARBAMIMIDOYLNAPHTHALEN-2-YL)METHYL][4-({1-[(1E)-ETHANIMIDOYL]PIPERIDIN-4-YL}OXY)PHENYL]SULFAMOYL}ACETIC ACID",
            "non-polymer",
        ),
        "RS8": (
            "N-[(2-AMINO-4-OXO-1,4-DIHYDROPTERIDIN-7-YL)CARBONYL]GLYCYL-L-PHENYLALANINE",
            "peptide",
        ),
        "TIB": (
            "4-CHLORO-8-METHYL-7-(3-METHYL-BUT-2-ENYL)-6,7,8,9-TETRAHYDRO-2H-2,7,9A-TRIAZA-BENZO[CD]AZULENE-1-THIONE",
            "non-polymer",
        ),
        "0CA": (
            "1-[(7-CARBAMIMIDOYLNAPHTHALEN-2-YL)METHYL]-6-({1-[(1Z)-ETHANIMIDOYL]PIPERIDIN-4-YL}OXY)-2-(PROPAN-2-YL)-1H-INDOLE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "0CG": ("PROP-2-YN-1-YL TRIHYDROGEN DIPHOSPHATE", "non-polymer"),
        "0CE": (
            "2-METHYL-11-(1-METHYLETHYL)-8-[(2S)-TETRAHYDRO-2H-PYRAN-2-YL]-2,11,12,13-TETRAHYDRO-4H-INDAZOLO[5,4-A]PYRROLO[3,4-C]CARBAZOL-4-ONE",
            "non-polymer",
        ),
        "GPL": ("?", "peptide"),
        "GPM": ("GLUCOPYRANOSYL-1-METHYL-PHOSPHONIC ACID", "saccharide"),
        "SUL": ("SULFATE ANION", "non-polymer"),
        "GPN": ("2-AMINOETHYLGLYCINE-CARBONYLMETHYLENE-GUANINE", "peptide"),
        "GPO": ("4-O-PHOSPHONO-D-GLYCERO-BETA-D-MANNO-HEPTOPYRANOSE", "non-polymer"),
        "MRM": ("4-AMINOPIPERIDINE-1-CARBOXYLIC ACID", "non-polymer"),
        "LRB": ("?", "non-polymer"),
        "TIC": ("TICLOPIDINE", "non-polymer"),
        "GPH": ("4-O-PHOSPHONO-D-GLYCERO-BETA-D-MANNO-HEPTOPYRANOSE", "saccharide"),
        "THZ": (
            "1-[(4S)-4-AMINO-5-(1,3-BENZOTHIAZOL-2-YL)-5-OXOPENTYL]GUANIDINE",
            "peptide",
        ),
        "VPU": (
            "[[(3R,4R,5S,6R)-4,5-DIHYDROXY-6-(HYDROXYMETHYL)-3-(PENTANOYLAMINO)OXAN-2-YLIDENE]AMINO] N-PHENYLCARBAMATE",
            "non-polymer",
        ),
        "KLT": (
            "2-CHLORO-5-[(1S)-1-HYDROXY-3-OXO-2H-ISOINDOL-1-YL]BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "GPJ": ("GLYPHOSATE", "non-polymer"),
        "VPR": ("2-PROPYLPENTANAMIDE", "non-polymer"),
        "S35": (
            "(3S)-3-({[1-(2-FLUOROPHENYL)-5-{[(2R)-2-HYDROXY-3,3-DIMETHYLBUTYL]OXY}-1H-PYRAZOL-3-YL]CARBONYL}AMINO)-3-(2-METHYLPHENYL)PROPANOIC ACID",
            "non-polymer",
        ),
        "THX": ("CDPI3", "nucleic-acid"),
        "S30": (
            "2-{[4-AMINO-3-(3-HYDROXYPROP-1-YN-1-YL)-1H-PYRAZOLO[3,4-D]PYRIMIDIN-1-YL]METHYL}-5-METHYL-3-(2-METHYLPHENYL)QUINAZOLIN-4(3H)-ONE",
            "non-polymer",
        ),
        "S33": (
            "(2S)-N-[[2-(AMINOMETHYL)-5-CHLORANYL-PHENYL]METHYL]-1-[(2S)-2-[(3-CHLORANYL-4-METHOXY-PHENYL)SULFONYLAMINO]-4-[(4-CYANOPHENYL)METHYLAMINO]-4-OXIDANYLIDENE-BUTANOYL]PYRROLIDINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "RS1": (
            "4-[4-(4-CHLORO-PHENOXY)-BENZENESULFONYLMETHYL]-TETRAHYDRO-PYRAN-4-CARBOXYLIC ACID HYDROXYAMIDE",
            "non-polymer",
        ),
        "THW": ("C2-1-HYDROXYPHENYL-THIAMIN DIPHOSPHATE", "non-polymer"),
        "VPF": (
            "N-(3-CARBOXYPROPANOYL)-L-VALYL-N-[(1S)-2-PHENYL-1-PHOSPHONOETHYL]-L-PROLINAMIDE",
            "peptide",
        ),
        "THV": ("C2-1-HYDROXY-3-METHYL-PROPYL-THIAMIN DIPHOSPHATE", "non-polymer"),
        "THU": ("TETRAHYDRODEOXYURIDINE", "non-polymer"),
        "M4M": (
            "1-AMINO-9,10-DIOXO-4-[(4-SULFAMOYLPHENYL)AMINO]-9,10-DIHYDROANTHRACENE-2-SULFONIC ACID",
            "non-polymer",
        ),
        "YPP": ("HYDROLYZED PIPERACILLIN", "non-polymer"),
        "GPG": ("GUANYLYL-2',5'-PHOSPHOGUANOSINE", "non-polymer"),
        "VXA": ("METHYLPHOSPHONIC ACID ESTER GROUP", "non-polymer"),
        "S3P": ("SHIKIMATE-3-PHOSPHATE", "non-polymer"),
        "GPA": ("2-AMINO-3-GUANIDINO-PROPIONIC ACID", "non-polymer"),
        "PU2": ("8-(2,5-DIMETHOXY-BENZYL)-2-FLUORO-9H-PURIN-6-YLAMINE", "non-polymer"),
        "YPA": (
            "(S)-3-(4-(2-CARBAZOL-9-YL-ETHOXY)-PHENYL)-2-ETHOXY-PROPIONIC ACID",
            "non-polymer",
        ),
        "S3H": ("TRISULFANE", "non-polymer"),
        "D09": (
            "5-[(1E)-2-(2-METHOXYPHENYL)PROP-1-EN-1-YL]FURO[2,3-D]PYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "THP": ("THYMIDINE-3',5'-DIPHOSPHATE", "non-polymer"),
        "S3B": ("3-METHYL-2-(PROPANOYLOXY)BENZOIC ACID", "non-polymer"),
        "THN": (
            "2-[CARBOXY-(2-THIOPHEN-2-YL-ACETYLAMINO)-METHYL]-5-METHYLENE-5,6-DIHYDRO-2H-[1,3]THIAZINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "THK": (
            "4-(4-CHLORO-PHENYL)-1-{3-[2-(4-FLUORO-PHENYL)-[1,3]DITHIOLAN-2-YL]-PROPYL}-PIPERIDIN-4-OL",
            "non-polymer",
        ),
        "U89": (
            "N-[4-[[3-(2,4-DIAMINO-1,6-DIHYDRO-6-OXO-4-PYRIMIDINYL)-PROPYL]-[2-((2-OXO-2-((4-PHOSPHORIBOXY)-BUTYL)-AMINO)-ETHYL)-THIO-ACETYL]-AMINO]BENZOYL]-1-GLUTAMIC ACID",
            "non-polymer",
        ),
        "THJ": ("THIOSULFATE", "non-polymer"),
        "THI": ("THIAZOLO-3-PHENYLISOINDOL-1-ONE", "non-polymer"),
        "3PP": ("3-PHOSPHONOPROPANOIC ACID", "non-polymer"),
        "THH": (
            "N-[4-({[(6S)-2-AMINO-4-HYDROXY-5-METHYL-5,6,7,8-TETRAHYDROPTERIDIN-6-YL]METHYL}AMINO)BENZOYL]-L-GLUTAMIC ACID",
            "non-polymer",
        ),
        "THG": ("(6S)-5,6,7,8-TETRAHYDROFOLATE", "non-polymer"),
        "K9H": (
            "{2-[5-(DECYLOXY)PYRIDIN-3-YL]ETHANE-1,1-DIYL}BIS(PHOSPHONIC ACID)",
            "non-polymer",
        ),
        "THF": ("5-HYDROXYMETHYLENE-6-HYDROFOLIC ACID", "non-polymer"),
        "THE": ("THREONATE ION", "non-polymer"),
        "MPY": (
            "(3E)-N-(2,6-DIETHYLPHENYL)-3-{[4-(4-METHYLPIPERAZIN-1-YL)BENZOYL]IMINO}PYRROLO[3,4-C]PYRAZOLE-5(3H)-CARBOXAMIDE",
            "non-polymer",
        ),
        "NKR": (
            "(2R)-2-HYDROXY-3-(PHOSPHONOOXY)PROPYL (4Z,7E,10E,13Z,16Z,19Z)-DOCOSA-4,7,10,13,16,19-HEXAENOATE",
            "non-polymer",
        ),
        "QPH": ("N-FORMYL-L-PHENYLALANINE", "peptide"),
        "GPW": (
            "1-(BETA-D-GLUCOPYRANOSYL)-5-(TRIFLUOROMETHYL)PYRIMIDINE-2,4(1H,3H)-DIONE",
            "non-polymer",
        ),
        "THC": ("N-METHYLCARBONYLTHREONINE", "peptide"),
        "HR7": ("(3S)-3-AMINO-6-[(DIAMINOMETHYLIDENE)AMINO]HEXANOIC ACID", "peptide"),
        "QPA": ("S-[(1S)-1-CARBOXY-1-(PHOSPHONOOXY)ETHYL]-L-CYSTEINE", "peptide"),
        "HR1": (
            "(3R,5R)-7-[1-(4-FLUOROPHENYL)-3-ISOPROPYL-4-OXO-5-PHENYL-4,5-DIHYDRO-3H-PYRROLO[2,3-C]QUINOLIN-2-YL]-3,5-DIHYDROXYHEPTANOIC ACID",
            "non-polymer",
        ),
        "WZ3": (
            "METHYL ALPHA-D-MANNOPYRANOSYL-(1->3)-[6-THIO-ALPHA-D-MANNOPYRANOSYL-(1->6)]-BETA-D-MANNOPYRANOSIDE",
            "non-polymer",
        ),
        "GPQ": (
            "5-CHLORO-1-(BETA-D-GLUCOPYRANOSYL)PYRIMIDINE-2,4(1H,3H)-DIONE",
            "non-polymer",
        ),
        "TP7": (
            "PHOSPHORIC ACID N-(1-HYDROXY-7-MERCAPTO-HEPTYL)THREONINE ESTER",
            "non-polymer",
        ),
        "TP6": (
            "3-(1,3,7-TRIHYDRO-9-D-RIBITYL-2,6,8-PURINETRIONE-7-YL) 1-PHOSPHATE",
            "non-polymer",
        ),
        "TP5": ("THIOPHENE-3-CARBOXIMIDAMIDE", "non-polymer"),
        "TP4": (
            "N-[4-[[GLUTAMIC ACID]-CARBONYL]-BENZENE-SULFONYL-D-PROLINYL]-3-AMINO-PROPANOIC ACID",
            "non-polymer",
        ),
        "TP3": (
            "4-[[GLUTAMIC ACID]-CARBONYL]-BENZENE-SULFONYL-D-PROLINE",
            "non-polymer",
        ),
        "THA": ("TACRINE", "non-polymer"),
        "TP1": (
            "2-(METHYLAMINO)-ETHYLGLYCINE-CARBONYLMETHYLENE-THYMINE",
            "nucleic-acid",
        ),
        "TP0": ("AMITRIPTYLINE", "non-polymer"),
        "QPS": ("BETA-ACARBOSE", "saccharide"),
        "QPP": (
            "N-(5-METHYL-1H-PYRAZOL-3-YL)-2-PHENYLQUINAZOLIN-4-AMINE",
            "non-polymer",
        ),
        "VXR": (
            "(5Z)-5-[(5-ETHYL-2-FURYL)METHYLENE]-2-{[(S)-(4-FLUOROPHENYL)(1H-TETRAZOL-5-YL)METHYL]AMINO}-1,3-THIAZOL-4(5H)-ONE",
            "non-polymer",
        ),
        "RFB": ("?", "non-polymer"),
        "QPT": ("CARBOPLATIN", "non-polymer"),
        "QPU": (
            "1,5-ANHYDRO-4-O-(4,6-DIDEOXY-4-{[(1S,2S,3S,4R,5S,6R)-2,3,4,6-TETRAHYDROXY-5-METHYLCYCLOHEXYL]AMINO}-ALPHA-D-GLUCOPYRANOSYL)-D-GLUCITOL",
            "non-polymer",
        ),
        "TPG": (
            "2,2,7-TRIMETHYL-GUANOSINE-5'-TRIPHOSPHATE-5'-GUANOSINE",
            "nucleic-acid",
        ),
        "TPF": (
            "2-(2,4-DIFLUOROPHENYL)-1,3-DI(1H-1,2,4-TRIAZOL-1-YL)PROPAN-2-OL",
            "non-polymer",
        ),
        "TPE": ("?", "non-polymer"),
        "TPD": ("N-(2-THIENYLMETHYL)-2,5-THIOPHENEDISULFONAMIDE", "non-polymer"),
        "TPC": ("?", "nucleic-acid"),
        "TPB": (
            "4-[4-(2,4,6-TRIMETHYL-PHENYLAMINO)-PYRIMIDIN-2-YLAMINO]-BENZONITRILE",
            "non-polymer",
        ),
        "TPA": ("TRANS-2-PHENYLCYCLOPROPYLAMINE", "non-polymer"),
        "X2L": (
            "(3Z)-5-AMINO-3-[(4-METHOXYPHENYL)METHYLIDENE]-1,3-DIHYDRO-2H-INDOL-2-ONE",
            "non-polymer",
        ),
        "TPO": ("PHOSPHOTHREONINE", "peptide"),
        "TPN": ("2-AMINOETHYLGLYCINE-CARBONYLMETHYLENE-THYMINE", "peptide"),
        "MRK": (
            "2-AMINO-4-FLUORO-5-[(1-METHYL-1H-IMIDAZOL-2-YL)SULFANYL]-N-(1,3-THIAZOL-2-YL)BENZAMIDE",
            "non-polymer",
        ),
        "TPL": ("TRYPTOPHANOL", "peptide"),
        "TPK": ("(1R)-1-[(2S)-PYRROLIDIN-2-YL]ETHANE-1,2-DIOL", "peptide"),
        "HRP": ("5-HYDROXY-L-TRYPTOPHAN", "peptide"),
        "TPI": (
            "4-CARBAMOYL-4-{[6-(DIFLUORO-PHOSPHONO-METHYL)-NAPHTHALENE-2-CARBONYL]-AMINO}-BUTYRIC ACID",
            "non-polymer",
        ),
        "TPH": ("2-AMINO-3-PHENYL-PROPANE-1,1-DIOL GROUP", "peptide"),
        "HRM": ("7-METHOXY-1-METHYL-9H-BETA-CARBOLINE", "non-polymer"),
        "TPV": (
            "N-(3-{(1R)-1-[(6R)-4-HYDROXY-2-OXO-6-PHENETHYL-6-PROPYL-5,6-DIHYDRO-2H-PYRAN-3-YL]PROPYL}PHENYL)-5-(TRIFLUOROMETHYL)-2-PYRIDINESULFONAMIDE",
            "non-polymer",
        ),
        "TPU": (
            "2-{1-[(4-AMINO-2-METHYLPYRIMIDIN-5-YL)METHYL]-5-METHYL-1H-1,2,3-TRIAZOL-4-YL}ETHYL TRIHYDROGEN DIPHOSPHATE",
            "non-polymer",
        ),
        "NII": (
            "2,7-BIS[3-(DIMETHYLAMINO)PROPYL]-4,9-BIS[(3-HYDROXYPROPYL)AMINO]BENZO[LMN][3,8]PHENANTHROLINE-1,3,6,8(2H,7H)-TETRONE",
            "non-polymer",
        ),
        "HRI": ("ALPHA-CHLOROISOCAPROIC ACID", "non-polymer"),
        "TPR": ("TOSYL-D-PROLINE", "non-polymer"),
        "TPQ": ("5-(2-CARBOXY-2-AMINOETHYL)-2-HYDROXY-1,4-BENZOQUINONE", "peptide"),
        "MT4": (
            "5-{4-[(6,7-DIMETHOXYQUINOLIN-4-YL)OXY]-3-FLUOROPHENYL}-2-[(4-FLUOROPHENYL)AMINO]-3-METHYLPYRIMIDIN-4(3H)-ONE",
            "non-polymer",
        ),
        "HRD": (
            "2,6-BIS(METHYLAMINO)-1,7-DIHYDRO-8H-IMIDAZO[4,5-G]QUINAZOLIN-8-ONE",
            "non-polymer",
        ),
        "HRG": ("L-HOMOARGININE", "peptide"),
        "WZB": (
            "(9BETA,13ALPHA,14BETA,17ALPHA)-2-ETHYLESTRA-1(10),2,4-TRIENE-3,17-DIYL DISULFAMATE",
            "non-polymer",
        ),
        "TPZ": ("O-PHOSPHONO-N-(5-SULFANYLPENTANOYL)-L-THREONINE", "non-polymer"),
        "TPY": (
            "(2S,3S)-3-FORMYL-2-({[(4-NITROPHENYL)SULFONYL]AMINO}METHYL)PENTANOIC ACID",
            "non-polymer",
        ),
        "HRB": (
            "N-ACETYL-1-[(2R,3S,5R)-5-HYDROXY-2-PENTYLTETRAHYDROFURAN-3-YL]-L-HISTIDINE",
            "non-polymer",
        ),
        "XE4": (
            "1,3-CYCLOHEXANEDIOL, 4-METHYLENE-5-[(2E)-[(1S,3AS,7AS)-OCTAHYDRO-1-(5-HYDROXY-5-METHYL-1,3-HEXADIYNYL)-7A-METHYL-4H-INDEN-4-YLIDENE]ETHYLIDENE]-, (1R,3S,5Z)",
            "non-polymer",
        ),
        "8UP": (
            "8-(3-BROMOPROPOXY)-7-METHOXYNAPHTHALENE-2-CARBOXIMIDAMIDE",
            "non-polymer",
        ),
        "MCX": (
            "(1R,2S,5S)-N-[(2S,3R)-4-AMINO-1-CYCLOBUTYL-3-HYDROXY-4-OXOBUTAN-2-YL]-6,6-DIMETHYL-3-{3-METHYL-N-[(1-METHYLCYCLOHEXYL)CARBAMOYL]-L-VALYL}-3-AZABICYCLO[3.1.0]HEXANE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "W19": (
            "4-{[4-(1-BENZOTHIOPHEN-4-YLOXY)-3-CHLOROPHENYL]AMINO}-N-(2-HYDROXYETHYL)-8,9-DIHYDRO-7H-PYRIMIDO[4,5-B]AZEPINE-6-CARBOXAMIDE",
            "non-polymer",
        ),
        "W16": ("5-(2-CHLOROETHYL)-1,3,4-THIADIAZOL-2-AMINE", "non-polymer"),
        "W14": (
            "2-(TRIFLUOROACETYL)-1,2,3,4-TETRAHYDROISOQUINOLIN-7-OL",
            "non-polymer",
        ),
        "SSF": (
            "2-AMINO-5-[(1-METHOXY-2-METHYLINDOLIZIN-3-YL)CARBONYL]BENZOIC ACID",
            "non-polymer",
        ),
        "W11": (
            "3-{3,5-DIMETHYL-4-[3-(3-METHYL-ISOXAZOL-5-YL)-PROPOXY]-PHENYL}-5-TRIFLUOROMETHYL-[1,2,4]OXADIAZOLE",
            "non-polymer",
        ),
        "P53": (
            "2-(HYDROXYMETHYL)PHENYL 6-O-PHOSPHONO-BETA-D-GLUCOPYRANOSIDE",
            "saccharide",
        ),
        "NIE": ("NITROETHANE", "non-polymer"),
        "F1J": (
            "(2S)-4-(4-FLUOROBENZYL)-N-(2-SULFANYLETHYL)PIPERAZINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "F1K": (
            "(2S)-4-(4-FLUOROBENZYL)-N-(3-SULFANYLPROPYL)PIPERAZINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "F1H": ("N-(1-BENZYLPIPERIDIN-4-YL)-4-SULFANYLBUTANAMIDE", "non-polymer"),
        "F1I": (
            "N-[1-(2,6-DIMETHOXYBENZYL)PIPERIDIN-4-YL]-4-SULFANYLBUTANAMIDE",
            "non-polymer",
        ),
        "F1N": (
            "N-{1-[5-CHLORO-3-METHOXY-2-(1-METHYLETHOXY)BENZYL]PIPERIDIN-4-YL}-2-(2-METHYL-4-SULFAMOYLPHENOXY)ACETAMIDE",
            "non-polymer",
        ),
        "F1L": (
            "N-[1-(5-BROMO-2,3-DIMETHOXYBENZYL)PIPERIDIN-4-YL]-4-SULFANYLBUTANAMIDE",
            "non-polymer",
        ),
        "F1M": (
            "N-{1-[5-CHLORO-3-METHOXY-2-(1-METHYLETHOXY)BENZYL]PIPERIDIN-4-YL}-2-(4-SULFAMOYLPHENOXY)ACETAMIDE",
            "non-polymer",
        ),
        "UGB": (
            "(2S,3R,4S,5R,6R)-6-[[[(2R,3S,4R,5R)-5-(2,4-DIOXOPYRIMIDIN-1-YL)-3,4-DIHYDROXY-OXOLAN-2-YL]METHOXY-HYDROXY-PHOSPHORYL]OXY-HYDROXY-PHOSPHORYL]OXY-3,4,5-TRIHYDROXY-OXANE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "UGA": ("UDP-GLUCURONIC ACID", "non-polymer"),
        "F1G": (
            "1-METHYL-3-TRIFLUOROMETHYL-1H-THIENO[2,3-C]PYRAZOLE-5-CARBOXYLIC ACID (2-MERCAPTO-ETHYL)-AMIDE",
            "non-polymer",
        ),
        "EN1": ("(2E)-2-HYDROXY-3-(4-HYDROXYPHENYL)PROP-2-ENOIC ACID", "non-polymer"),
        "F1X": ("1-O-PHOSPHONO-BETA-D-FRUCTOFURANOSE", "saccharide"),
        "NIY": ("META-NITRO-TYROSINE", "peptide"),
        "TK4": ("2-ACETYLAMINO-NAPTHALENE-1,5-DISULFONIC ACID", "non-polymer"),
        "F1S": (
            "N-(5-{[3-(4-AMINO-6-METHYL-1,3,5-TRIAZIN-2-YL)-5-(TETRAHYDRO-2H-PYRAN-4-YL)PYRIDIN-2-YL]AMINO}-2-CHLOROPYRIDIN-3-YL)METHANESULFONAMIDE",
            "non-polymer",
        ),
        "F1P": ("1-O-PHOSPHONO-BETA-D-FRUCTOPYRANOSE", "saccharide"),
        "ENA": ("ETHENO-NAD", "nucleic-acid"),
        "ENB": (
            "ETHYL (2E,4S)-4-[((2R)-2-{[N-(TERT-BUTOXYCARBONYL)-L-VALYL]AMINO}-2-PHENYLETHANOYL)AMINO]-5-[(3S)-2-OXOPYRROLIDIN-3-YL]PENT-2-ENOATE",
            "non-polymer",
        ),
        "ENC": ("ETHYL ISOCYANIDE", "non-polymer"),
        "END": (
            "1,6:5,9:8,12:11,16-TETRAANHYDRO-2,3,4,10,13,14-HEXADEOXY-D-GLYCERO-D-ALLO-D-GULO-HEPTADECA-2,13-DIENITOL",
            "non-polymer",
        ),
        "ENE": (
            "(4Z)-2,8:7,12:11,15:14,18:17,22-PENTAANHYDRO-4,5,6,9,10,13,19,20,21-NONADEOXY-D-ARABINO-D-ALLO-D-ALLO-DOCOSA-4,9,20-TRIENITOL",
            "non-polymer",
        ),
        "PYA": ("3-(1,10-PHENANTHROL-2-YL)-L-ALANINE", "peptide"),
        "ENH": (
            "1,7,8,9,10,10-HEXACHLORO-4-METHYL-4-AZA-TRICYCLO[5.2.1.0(2,6)]DEC-8-ENE-3,5-DIONE",
            "non-polymer",
        ),
        "RFZ": ("5,6-DICHLORO-1-BETA-D-RIBOFURANOSYL-1H-BENZIMIDAZOLE", "non-polymer"),
        "RFX": (
            "(3R)-N-METHYL-3-PHENYL-3-[4-(TRIFLUOROMETHYL)PHENOXY]PROPAN-1-AMINE",
            "non-polymer",
        ),
        "ENL": (
            "(1R,2S,3R,4S)-7-OXABICYCLO[2.2.1]HEPTANE-2,3-DICARBOXYLIC ACID",
            "non-polymer",
        ),
        "ENM": (
            "(5S,8R,9S,10S,13R,14S,17S)-13-{2-[(3,5-DIFLUOROBENZYL)OXY]ETHYL}-17-HYDROXY-10-METHYLHEXADECAHYDRO-3H-CYCLOPENTA[A]PHENANTHREN-3-ONE",
            "non-polymer",
        ),
        "OJ8": (
            "(S)-[(2S)-6-PHENOXY-1,2,3,4-TETRAHYDRONAPHTHALEN-2-YL](5-PYRIDIN-2-YL-1,3-OXAZOL-2-YL)METHANOL",
            "non-polymer",
        ),
        "ENO": ("3-(4-HYDROXY-PHENYL)PYRUVIC ACID", "non-polymer"),
        "ENP": ("ETHENO-NADP", "nucleic-acid"),
        "ENQ": (
            "[[(2R,3S,4R,5R)-3,4-DIHYDROXY-5-(9H-IMIDAZO[2,1-F]PURIN-6-IUM-3-YL)OXOLAN-2-YL]METHOXY-OXIDANIDYL-PHOSPHORYL] [(2R,3S,4R,5R)-3,4-DIHYDROXY-5-OXIDANIDYL-OXOLAN-2-YL]METHYL PHOSPHATE",
            "nucleic-acid",
        ),
        "F18": ("4-[(E)-(3,5-DIAMINO-1H-PYRAZOL-4-YL)DIAZENYL]PHENOL", "non-polymer"),
        "ENS": (
            "METHYL (2Z)-3-[(3-CHLORO-1H-INDOL-7-YL)AMINO]-2-CYANO-3-{[(3S)-2-OXO-1-(2-OXO-2-PYRROLIDIN-1-YLETHYL)AZEPAN-3-YL]AMINO}ACRYLATE",
            "non-polymer",
        ),
        "ENT": (
            "3(R)-METHYLCARBAMOYL-7-SULFOAMINO-3,4-DIHYDRO-1H-ISOQUINOLINE-2-CARBOXYLIC ACID TERT-BUTYL ESTER",
            "non-polymer",
        ),
        "NKX": (
            "(2S)-2,3-BIS(NITROOXY)PROPYL ETHYL[(4S,6S)-6-METHYL-7,7-DIOXIDO-2-SULFAMOYL-5,6-DIHYDRO-4H-THIENO[2,3-B]THIOPYRAN-4-YL]CARBAMATE",
            "non-polymer",
        ),
        "ENV": (
            "(2R)-N-{3-[(5-FLUORO-4-OXOPENTYL)AMINO]-3-OXOPROPYL}-2,4-DIHYDROXY-3,3-DIMETHYLBUTANAMIDE",
            "non-polymer",
        ),
        "ENW": (
            "2-({N-[(2S)-2,4-DIHYDROXY-3,3-DIMETHYLBUTANOYL]-BETA-ALANYL}AMINO)ETHYL FLUOROACETATE",
            "non-polymer",
        ),
        "ENX": ("ENACYLOXIN IIA", "non-polymer"),
        "F13": ("3-FLUORO-N-[3-(1H-TETRAZOL-5-YL)PHENYL]BENZAMIDE", "non-polymer"),
        "F10": (
            "2-[2-(2-FLUOROPHENYL)PYRIDIN-4-YL]-1,5,6,7-TETRAHYDRO-4H-PYRROLO[3,2-C]PYRIDIN-4-ONE",
            "non-polymer",
        ),
        "F11": (
            "N,N'-DI-1,2,3,4-TETRAHYDROACRIDIN-9-YLPENTANE-1,5-DIAMINE",
            "non-polymer",
        ),
        "F16": ("5-(3-HYDROXYPHENYL)ISOTHIAZOL-3(2H)-ONE 1,1-DIOXIDE", "non-polymer"),
        "F17": (
            "N-{[4-(1,1-DIOXIDO-3-OXO-2,3-DIHYDROISOTHIAZOL-5-YL)PHENYL]ACETYL}-L-PHENYLALANYL-4-(1,1-DIOXIDO-3-OXO-2,3-DIHYDROISOTHIAZOL-5-YL)-L-PHENYLALANINAMIDE",
            "non-polymer",
        ),
        "F15": ("PENTADECANOIC ACID", "non-polymer"),
        "ZRG": (
            "(2S)-2-ACETAMIDO-N-METHYL-5-[[N-(METHYLCARBAMOYL)CARBAMIMIDOYL]AMINO]PENTANAMIDE",
            "peptide",
        ),
        "OMD": ("2-(3,6-DIHYDROXYPHENYL)ACETIC ACID", "non-polymer"),
        "M1A": ("6-AMINO-1-METHYLPURINE", "non-polymer"),
        "M1C": ("(3S)-3-AMINO-1-(CYCLOPROPYLAMINO)HEPTANE-2,2-DIOL", "non-polymer"),
        "M1B": (
            "2-{3'-[AMINO(IMINO)METHYL]BIPHENYL-4-YL}-1H-BENZIMIDAZOLE-5-CARBOXIMIDAMIDE",
            "non-polymer",
        ),
        "M1G": (
            "3-(2-DEOXY-BETA-D-RIBOFURANOSYL)-PYRIDO[5,6-A]-PURINE-10-ONE-5'-MONOPHOSPHATE",
            "nucleic-acid",
        ),
        "T3E": (
            "3-[[5-CYANO-4-[4-METHYL-2-(METHYLAMINO)-1,3-THIAZOL-5-YL]PYRIMIDIN-2-YL]AMINO]BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "TNR": ("O-(2-ACETAMIDO-2-DEOXY-ALPHA-D-GALACTOPYRANOSYL)-L-SERINE", "peptide"),
        "LC4": ("12-(1H-IMIDAZOL-1-YL)DODECANOIC ACID", "non-polymer"),
        "M1N": (
            "(1R)-3-METHYL-1-{[N-(MORPHOLIN-4-YLCARBONYL)-3-(1-NAPHTHYL)-D-ALANYL]AMINO}BUTYLBORONIC ACID",
            "non-polymer",
        ),
        "M1P": ("ALPHA-D-MANNOSE 1-PHOSPHATE", "non-polymer"),
        "M1S": (
            "(5E)-14-CHLORO-15,17-DIHYDROXY-4,7,8,9,10,11-HEXAHYDRO-2-BENZOXACYCLOPENTADECINE-1,12(3H,13H)-DIONE",
            "non-polymer",
        ),
        "M1T": ("6,6-DIHYDROXY-1-METHOXYHEXAN-2-ONE", "non-polymer"),
        "LC2": (
            "N-[(1S,2R,3E,5E,7S,9E,11E,13S,15R,19R)-7,13-DIHYDROXY-1,4,10,19-TETRAMETHYL-17,18-DIOXO-16-OXABICYCLO[13.2.2]NONADECA-3,5,9,11-TETRAEN-2-YL]-2-OXOPROPANAMIDE",
            "non-polymer",
        ),
        "ZRM": (
            "7-(4-HYDROXYPHENYL)-2-PYRIDIN-4-YL-5H-THIENO[3,2-C]PYRIDIN-4-ONE",
            "non-polymer",
        ),
        "L21": ("3-(3-METHYLBUT-2-EN-1-YL)-3H-PURIN-6-AMINE", "non-polymer"),
        "SEU": ("S-ETHYLISOTHIOUREA", "non-polymer"),
        "SET": ("AMINOSERINE", "peptide"),
        "SES": ("4-BUTYL-5-PROPYL-1,3-SELENAZOL-2-AMINE", "non-polymer"),
        "LC0": (
            "3-({4-METHOXY-5-[(4-METHOXYBENZYL)OXY]PYRIDIN-2-YL}METHOXY)-5-(1-METHYL-1H-PYRAZOL-4-YL)PYRAZIN-2-AMINE",
            "non-polymer",
        ),
        "PD8": ("PHOSPHORYLATED DIHYDROPTEROATE", "non-polymer"),
        "PD7": ("(2R)-3-(PHOSPHONOOXY)PROPANE-1,2-DIYL DIHEPTANOATE", "non-polymer"),
        "PD6": ("5-AMINO-1-BUTYL-7-PHENYL-1,6-NAPHTHYRIDIN-4(1H)-ONE", "non-polymer"),
        "PD5": (
            "1-{4-[4-AMINO-1-(1-METHYLETHYL)-1H-PYRAZOLO[3,4-D]PYRIMIDIN-3-YL]PHENYL}-3-[3-(TRIFLUOROMETHYL)PHENYL]UREA",
            "non-polymer",
        ),
        "LC1": ("2-(1,3-THIAZOL-4-YL)-1H-BENZIMIDAZOLE-5-SULFONAMIDE", "non-polymer"),
        "PD3": (
            "1-{3-[(4-AMINO-1-CYCLOPENTYL-1H-PYRAZOLO[3,4-D]PYRIMIDIN-3-YL)METHYL]PHENYL}-3-[3-(TRIFLUOROMETHYL)PHENYL]UREA",
            "non-polymer",
        ),
        "PD2": ("PYRIDINE-2,4-DICARBOXYLIC ACID", "non-polymer"),
        "PD1": (
            "1-TERT-BUTYL-3-[6-(3,5-DIMETHOXY-PHENYL)-2-(4-DIETHYLAMINO-BUTYLAMINO)-PYRIDO[2,3-D]PYRIMIDIN-7-YL]-UREA",
            "non-polymer",
        ),
        "IM1": ("?", "non-polymer"),
        "IM3": ("(2E)-3-(3,4-DIHYDROXYPHENYL)-2-IMINOPROPANOIC ACID", "non-polymer"),
        "IM2": (
            "(5R)-5-[(1S,2R)-1-FORMYL-2-HYDROXYPROPYL]-3-[(2-{[(E)-IMINOMETHYL]AMINO}ETHYL)SULFANYL]-4,5-DIHYDRO-1H-PYRROLE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "CL8": ("CHLORAMPHENICOL SUCCINATE", "non-polymer"),
        "CL9": ("2-CHLORO-2'-DEOXYADENOSINE", "non-polymer"),
        "SEA": ("THIOETHYLAMINE", "non-polymer"),
        "IM6": (
            "4-TERT-BUTYL-N-[6-(1H-IMIDAZOL-1-YL)IMIDAZO[1,2-A]PYRIDIN-2-YL]BENZAMIDE",
            "non-polymer",
        ),
        "CL4": (
            "N-{3-[5-(6-AMINO-PURIN-9-YL)-3,4-DIHYDROXY-TETRAHYDRO-FURAN-2-YL]-ALLYL}-2,3-DIHYDROXY-5-NITRO-BENZAMIDE",
            "non-polymer",
        ),
        "CL5": ("N-(3-OXOPROPYL)GLYCINE", "non-polymer"),
        "CL6": ("1-[(2-CHLOROPHENYL)(DIPHENYL)METHYL]-1H-IMIDAZOLE", "non-polymer"),
        "CL7": ("CHLOROPHYLL D", "non-polymer"),
        "SEK": ("SELENOCYANATE ION", "non-polymer"),
        "CL1": ("ALPHA CHLOROPHYLL A", "non-polymer"),
        "CL2": ("BETA CHLOROPHYLL A", "non-polymer"),
        "CL3": (
            "N-METHYL-N-[3-(6-PHENYL[1,2,4]TRIAZOLO[4,3-B]PYRIDAZIN-3-YL)PHENYL]ACETAMIDE",
            "non-polymer",
        ),
        "CLL": ("CHOLESTERYL LINOLEATE", "non-polymer"),
        "CLM": ("CHLORAMPHENICOL", "non-polymer"),
        "CLN": ("SULFUR SUBSTITUTED PROTOPORPHYRIN IX", "non-polymer"),
        "CLO": ("CHLORO GROUP", "non-polymer"),
        "CLH": (
            "2-AMINO-6-[2-(2-OXO-ACETYLAMINO)-ACETYLAMINO]-HEXANOIC ACID",
            "peptide",
        ),
        "CLI": (
            "3-(2-CARBOXY-ETHYL)-4,6-DICHLORO-1H-INDOLE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "CLJ": ("CALICHEAMICIN ALPHA3I", "non-polymer"),
        "CLK": ("ALPHA-N-DICHLOROACETYL-P-AMINOPHENYLSERINOL", "non-polymer"),
        "CLD": ("D-PARA-CHLOROPHENYL-1-ACTEAMIDOBORONIC ACID ALANINE", "peptide"),
        "CLE": ("LEUCINE AMIDE", "peptide"),
        "CLF": ("FE(8)-S(7) CLUSTER", "non-polymer"),
        "CLG": (
            "2-AMINO-6-[2-(2-AMINOOXY-ACETYLAMINO)-ACETYLAMINO]-HEXANOIC ACID",
            "peptide",
        ),
        "IMM": (
            "1-(4-IODOBENZOYL)-5-METHOXY-2-METHYL INDOLE-3-ACETIC ACID",
            "non-polymer",
        ),
        "CLA": ("CHLOROPHYLL A", "non-polymer"),
        "CLB": ("D-PARA-CHLOROPHENYL-1-ACETAMIDOBORONIC ACID ALANINE", "peptide"),
        "CLC": ("N-ACETYL-P-NITROPHENYLSERINOL", "non-polymer"),
        "IMQ": (
            "(2R,3R,4S)-2-(HYDROXYMETHYL)-1-(QUINOLIN-8-YLMETHYL)PYRROLIDINE-3,4-DIOL",
            "non-polymer",
        ),
        "IMP": ("INOSINIC ACID", "non-polymer"),
        "IMS": (
            "2-[1-(4-CHLOROBENZOYL)-5-METHOXY-2-METHYL-1H-INDOL-3-YL]-N-[(1S)-1-(HYDROXYMETHYL)PROPYL]ACETAMIDE",
            "non-polymer",
        ),
        "IMR": ("IMINORIBITOL", "non-polymer"),
        "CLX": ("TRI-CHLORO-ACETALDEHYDE", "non-polymer"),
        "CLY": ("CLINDAMYCIN", "non-polymer"),
        "CLZ": ("5-CHLORYL-2,4,6-QUINAZOLINETRIAMINE", "non-polymer"),
        "IMV": (
            "5-[(2-OXO-2,3-DIHYDRO-1H-INDOL-1-YL)METHYL]-1,3-BENZODIOXOLE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "CLT": ("4-PHENYL-BUTANOIC ACID", "non-polymer"),
        "CLU": ("2,6-DICHLORO-N-IMIDAZOLIDIN-2-YLIDENEANILINE", "non-polymer"),
        "CLV": (
            "{(2S)-2-[(1S)-1-AMINOETHYL]-5-OXO-2,5-DIHYDRO-1H-IMIDAZOL-1-YL}ACETIC ACID",
            "peptide",
        ),
        "CLW": ("CHLORZOXAZONE", "non-polymer"),
        "CLP": ("FE-S CLUSTER", "non-polymer"),
        "CLQ": (
            "N4-(7-CHLORO-QUINOLIN-4-YL)-N1,N1-DIETHYL-PENTANE-1,4-DIAMINE",
            "non-polymer",
        ),
        "CLR": ("CHOLESTEROL", "non-polymer"),
        "CLS": ("CEPHALOTHIN", "non-polymer"),
        "XAA": ("(1S,2S)-1-AMINO-1-(1,3-THIAZOL-2-YL)PROPAN-2-OL", "non-polymer"),
        "NTM": ("QUINOLINIC ACID", "non-polymer"),
        "OMH": ("O-[(S)-HYDROXY(METHOXY)PHOSPHORYL]-L-SERINE", "peptide"),
        "FDS": ("FLUORESCIN", "non-polymer"),
        "M11": (
            "N^6^-[(1R,2S)-1-({[(1R)-1-CARBOXY-2-METHYLPROPYL]OXY}CARBONYL)-2-SULFANYLPROPYL]-6-OXO-L-LYSINE",
            "non-polymer",
        ),
        "M13": (
            "METHYL 3-O-ALPHA-D-MANNOPYRANOSYL-BETA-D-ALTROPYRANOSIDE",
            "non-polymer",
        ),
        "FDR": (
            "2-METHYL-3,5,6,7-TETRAHYDRO-4H-CYCLOPENTA[4,5]THIENO[2,3-D]PYRIMIDIN-4-ONE",
            "non-polymer",
        ),
        "M14": (
            "2-[3-CHLORO-5-(3-CHLORO-5-CYANOPHENOXY)PHENOXY]-N-(2-CHLORO-4-SULFAMOYLPHENYL)ACETAMIDE",
            "non-polymer",
        ),
        "M18": (
            "{(2S)-1-[N-(TERT-BUTOXYCARBONYL)GLYCYL]PYRROLIDIN-2-YL}METHYL (3-CHLOROPHENYL)ACETATE",
            "non-polymer",
        ),
        "NLX": (
            "(5A,17R)-4,5-EPOXY-3,14-DIHYDROXY-17-METHYL-6-OXO-17-(2-PROPENYL)-MORPHINANIUM",
            "non-polymer",
        ),
        "BPA": (
            "7S,8R,9R-TRIHYDROXY-7,8,9,10-TETRAHYDRO BENZO[A]PYRENE",
            "non-polymer",
        ),
        "3XA": (
            "6-{[(3S,4S)-4-(2-{[2,2-DIFLUORO-2-(3-FLUOROPHENYL)ETHYL]AMINO}ETHOXY)PYRROLIDIN-3-YL]METHYL}-4-METHYLPYRIDIN-2-AMINE",
            "non-polymer",
        ),
        "3XB": (
            "6-{[(3R,4R)-4-(2-{[2,2-DIFLUORO-2-(3-FLUOROPHENYL)ETHYL]AMINO}ETHOXY)PYRROLIDIN-3-YL]METHYL}-4-METHYLPYRIDIN-2-AMINE",
            "non-polymer",
        ),
        "BPB": ("BACTERIOPHEOPHYTIN B", "non-polymer"),
        "BPE": ("(2S)-2-AMINO-3-[(3-AMINOPROPYL)SULFANYL]PROPAN-1-OL", "peptide"),
        "3XE": (
            "(4S,6S)-6-{[(3R,4R)-4-{2-[(2,2-DIFLUORO-2-PHENYLETHYL)AMINO]ETHOXY}PYRROLIDIN-3-YL]METHYL}-4-METHYL-3,4,5,6-TETRAHYDROPYRIDIN-2-AMINE",
            "non-polymer",
        ),
        "BPG": ("9-(4-HYDROXYBUTYL)-N2-PHENYLGUANINE", "non-polymer"),
        "BPF": (
            "N,N'-(FURAN-2,5-DIYLBIS{BENZENE-4,1-DIYL[(Z)-AMINOMETHYLYLIDENE]})DIETHANAMINIUM",
            "non-polymer",
        ),
        "BPI": (
            "(1S)-1,2,3,4-TETRAHYDRO-BENZO[C]PHENANTHRENE-2,3,4-TRIOL",
            "non-polymer",
        ),
        "BPH": ("BACTERIOPHEOPHYTIN A", "non-polymer"),
        "BPK": (
            "{3-[(5-METHYL-2-PHENYL-1,3-OXAZOL-4-YL)METHOXY]PHENYL}METHANOL",
            "non-polymer",
        ),
        "BPJ": (
            "(1R)-1,2,3,4-TETRAHYDRO-BENZO[C]PHENANTHRENE-2,3,4-TRIOL",
            "non-polymer",
        ),
        "BPM": ("4-PHOSPHONOOXY-PHENYL-METHYL-[4-PHOSPHONOOXY]BENZEN", "non-polymer"),
        "BPL": ("1-BENZO[A]PYREN-6-YL-3-METHYL-UREA", "non-polymer"),
        "BPO": (
            "3-[(Z)-AMINO(IMINO)METHYL]-N-[2-(4-BENZOYL-1-PIPERIDINYL)-2-OXO-1-PHENYLETHYL]BENZAMIDE",
            "non-polymer",
        ),
        "BPN": ("PARANITROBENZYL ALCOHOL", "non-polymer"),
        "BPQ": (
            "TERT-BUTYL [(2-AMINO-4-OXO-4,7-DIHYDRO-3H-PYRROLO[2,3-D]PYRIMIDIN-5-YL)METHYL]CARBAMATE",
            "non-polymer",
        ),
        "BPP": (
            "(4-BENZYL-PIPERIDIN-1-YL)-(5-AMIDINOMETHYL-3AH-INDOL-2-YL-METHANONE",
            "non-polymer",
        ),
        "BPS": ("BIPHENYL-2-SULFINIC ACID", "non-polymer"),
        "BPR": (
            "(2R)-N-[(2R)-2-(DIHYDROXYBORYL)-1-L-PROLYLPYRROLIDIN-2-YL]-N-[(5R)-5-(DIHYDROXYBORYL)-1-L-PROLYLPYRROLIDIN-2-YL]-L-PROLINAMIDE",
            "non-polymer",
        ),
        "BPU": ("7-BROMOPYRROLO[1,2-A]QUINOXALIN-4(5H)-ONE", "non-polymer"),
        "BPT": (
            "BIS(TRANS-PLATINUM ETHYLENEDIAMINE DIAMINE CHLORO)COMPLEX",
            "non-polymer",
        ),
        "BPV": ("BROMOPYRUVATE", "non-polymer"),
        "BPY": ("BIPHENYL-2,3-DIOL", "non-polymer"),
        "BPX": (
            "(2R)-4,4-DIHYDROXY-5-NITRO-2-(PHENYLMETHYL)PENTANOIC ACID",
            "non-polymer",
        ),
        "BPZ": ("4,4'-CYCLOHEXANE-1,1-DIYLDIPHENOL", "non-polymer"),
        "GMN": (
            "2,2',2-[1,2,3-BENZENE-TRIYLTRIS(OXY)]TRIS[N,N,N-TRIETHYLETHANAMINIUM]",
            "non-polymer",
        ),
        "GMM": ("GLUCOSE MONOMYCOLATE", "non-polymer"),
        "OMN": (
            "4-BROMO-3-(5'-CARBOXY-4'-CHLORO-2'-FLUOROPHENYL)-1-METHYL-5-TRIFLUOROMETHYL-PYRAZOL",
            "non-polymer",
        ),
        "T50": (
            "1-(2,6-DIFLUOROBENZYL)-2-(2,6-DIFLUOROPHENYL)-4-METHYL-1H-BENZIMIDAZOLE",
            "non-polymer",
        ),
        "AJ1": ("N-(2-METHYLAMINO)-1,3-BENZOTHIAZOL-6-YL)ACETAMIDE", "non-polymer"),
        "AJ3": ("3-(PROP-2-ENE-1-SULFINYL)-PROPENE-1-THIOL", "non-polymer"),
        "T55": ("8-METHYLNONANOIC ACID", "non-polymer"),
        "ML2": ("N-[2-(2-IODO-5-METHOXY-1H-INDOL-3-YL)ETHYL]ACETAMIDE", "non-polymer"),
        "ML3": (
            "2-{[(2R)-2-AMINO-2-CARBOXYETHYL]SULFANYL}-N,N,N-TRIMETHYLETHANAMINIUM",
            "peptide",
        ),
        "ML0": (
            "4-{2-AMINO-6-[(1R)-2,2,2-TRIFLUORO-1-(3'-FLUOROBIPHENYL-4-YL)ETHOXY]PYRIMIDIN-4-YL}-L-PHENYLALANINE",
            "non-polymer",
        ),
        "ML1": ("N-[2-(5-METHOXY-1H-INDOL-3-YL)ETHYL]ACETAMIDE", "non-polymer"),
        "ML4": (
            "4-(5-{[(2'-METHYLBIPHENYL-2-YL)METHYL]AMINO}PYRAZIN-2-YL)-L-PHENYLALANINE",
            "non-polymer",
        ),
        "ML5": (
            "{(3R)-3-AMINO-4-[(3-HEXYLPHENYL)AMINO]-4-OXOBUTYL}PHOSPHONIC ACID",
            "non-polymer",
        ),
        "ML8": (
            "8-CYCLOPENTYL-6-[3-(HYDROXYMETHYL)PHENYL]-4-METHYL-2-(METHYLAMINO)PYRIDO[2,3-D]PYRIMIDIN-7(8H)-ONE",
            "non-polymer",
        ),
        "ML9": (
            "2-AMINO-8-[TRANS-4-(2-HYDROXYETHOXY)CYCLOHEXYL]-6-(6-METHOXYPYRIDIN-3-YL)-4-METHYLPYRIDO[2,3-D]PYRIMIDIN-7(8H)-ONE",
            "non-polymer",
        ),
        "GO1": (
            "3-(CARBOXYMETHYL)-5,7-DICHLORO-1H-INDOLE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "GO0": (
            "3-(CARBOXYMETHYL)-4,7-DICHLORO-1H-INDOLE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "GOM": ("GLUTAMOL-AMP", "nucleic-acid"),
        "GOL": ("GLYCEROL", "non-polymer"),
        "AJZ": ("3-(4-MORPHOLIN-4-YLTHIENO[3,2-D]PYRIMIDIN-2-YL)PHENOL", "non-polymer"),
        "MLA": ("MALONIC ACID", "non-polymer"),
        "GOI": (
            "2-{4-[3-(5H-DIBENZO[B,F]AZEPIN-5-YL)PROPYL]PIPERAZIN-1-YL}ETHYL BETA-D-GLUCOPYRANOSIDE",
            "non-polymer",
        ),
        "MLG": (
            "N-[3-(2,4-DICHLOROPHENOXY)PROPYL]-N-METHYL-N-PROP-2-YNYLAMINE",
            "non-polymer",
        ),
        "MLD": (
            "GLCNAC(BETA1-4)-MURNAC(1,6-ANHYDRO)-L-ALA-GAMMA-D-GLU-MESO-A2PM-D-ALA",
            "non-polymer",
        ),
        "MLE": ("N-METHYLLEUCINE", "peptide"),
        "T5P": (
            "3-{2,6,8-TRIOXO-9-[(2S,3S,4R)-2,3,4,5-TETRAHYDROXYPENTYL]-1,2,3,6,8,9-HEXAHYDRO-7H-PURIN-7-YL}PROPYL DIHYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "GMG": (
            "[4-(5-NAPHTHALEN-2-YL-1H-PYRROLO[2,3-B]PYRIDIN-3-YL)PHENYL]ACETIC ACID",
            "non-polymer",
        ),
        "MLI": ("MALONATE ION", "non-polymer"),
        "GOA": ("GLYCOLIC ACID", "non-polymer"),
        "MLO": ("5,5'-DI(PROP-2-EN-1-YL)BIPHENYL-2,2'-DIOL", "non-polymer"),
        "MLL": ("METHYL L-LEUCINATE", "peptide"),
        "GOB": (
            "4-({5-[(4-AMINOPHENYL)(PHENYL)SULFAMOYL]-2,4-DIMETHOXYPHENYL}AMINO)-4-OXOBUTANOIC ACID",
            "non-polymer",
        ),
        "AJH": (
            "4-({4-[(3R)-3-AMINO-4-(2,4,5-TRIFLUOROPHENYL)BUTANOYL]-1,4-DIAZEPAN-1-YL}CARBONYL)BENZOIC ACID",
            "non-polymer",
        ),
        "AJI": (
            "3-[(4-FLUOROPHENYL)SULFANYL]-N-(4-METHYL-1,3-THIAZOL-2-YL)-6-[(4-METHYL-4H-1,2,4-TRIAZOL-3-YL)SULFANYL]PYRIDINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "MLP": ("1-AMINOCYCLOPROPYLPHOSPHONATE", "non-polymer"),
        "AJM": ("AJMALINE", "non-polymer"),
        "MLT": ("D-MALATE", "non-polymer"),
        "MLU": ("N-METHYL-D-LEUCINE", "peptide"),
        "MLZ": ("N-METHYL-LYSINE", "peptide"),
        "AJA": (
            "(6AR,10AR)-3-(1,1-DIMETHYLHEPTYL)-1-HYDROXY-6,6-DIMETHYL-6A,7,10,10A-TETRAHYDRO-6H-BENZO[C]CHROMENE-9-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "AJB": (
            "2-AMINO-N-(4-METHYL-1,3-THIAZOL-2-YL)-5-[(4-METHYL-4H-1,2,4-TRIAZOL-3-YL)SULFANYL]BENZAMIDE",
            "non-polymer",
        ),
        "MLY": ("N-DIMETHYL-LYSINE", "peptide"),
        "AJD": ("5-BROMO-2-[(3R)-PIPERIDIN-3-YLAMINO]BENZOIC ACID", "non-polymer"),
        "T5E": (
            "2,7-BIS[4-(4-METHYLPIPERAZIN-1-YL)BUTYL]-4,9-BIS{[4-(4-METHYLPIPERAZIN-1-YL)BUTYL]AMINO}BENZO[LMN][3,8]PHENANTHROLINE-1,3,6,8(2H,7H)-TETRONE",
            "non-polymer",
        ),
        "LPR": ("[N2-[(S)-1-CARBOXY-3-PHENYLPROPYL]-L-LYSYL-L-PROLINE", "non-polymer"),
        "PEY": ("PHENANTHRENE", "non-polymer"),
        "OPR": ("(5S)-5-AMINO-8-CARBAMIMIDAMIDO-4-OXOOCTANOIC ACID", "peptide"),
        "NRI": (
            "PHOSPHORIC ACID MONO-(4-HYDROXY-PYRROLIDIN-3-YLMETHYL) ESTER",
            "nucleic-acid",
        ),
        "RSW": (
            "[(1,2,3,4,5,6-ETA)-(1S,2R,3R,4R,5S,6S)-1-CARBOXYCYCLOHEXANE-1,2,3,4,5,6-HEXAYL](CHLORO)(3-METHYL-5,7-DIOXO-6,7-DIHYDRO-5H-PYRIDO[2,3-A]PYRROLO[3,4-C]CARBAZOL-12-IDE-KAPPA~2~N~1~,N~12~)RUTHENIUM(1+)",
            "non-polymer",
        ),
        "NRO": ("3-[5-(2-NITROPENT-1-EN-1-YL)FURAN-2-YL]BENZOIC ACID", "non-polymer"),
        "NRM": (
            "N-(3-METHYLISOTHIAZOL-5-YL)-3-(1H-PYRAZOL-4-YL)IMIDAZO[1,2-A]PYRAZIN-8-AMINE",
            "non-polymer",
        ),
        "GMZ": (
            "1,7-DI-O-PHOSPHONO-D-GLYCERO-BETA-D-MANNO-HEPTOPYRANOSE",
            "non-polymer",
        ),
        "NRB": ("NARBOMYCIN", "non-polymer"),
        "NRA": ("NORATHYRIOL", "non-polymer"),
        "NRG": ("N-OMEGA-NITRO-L-ARGININE", "peptide"),
        "GMY": ("GELDANAMYCIN", "non-polymer"),
        "BP1": ("6-BROMO-7H-PURINE", "non-polymer"),
        "BP0": (
            "5,5'-(SULFANEDIYLDIETHANE-2,1-DIYL)BIS(1,3,4-THIADIAZOL-2-AMINE)",
            "non-polymer",
        ),
        "BP3": ("2'-CHLORO-BIPHENYL-2,3-DIOL", "non-polymer"),
        "BP2": ("(+)-BORNYL DIPHOSPHATE", "non-polymer"),
        "BP5": ("3-(2,2'-BIPYRIDIN-5-YL)-L-ALANINE", "peptide"),
        "BP4": ("BIPHENYL-4-YLACETIC ACID", "non-polymer"),
        "BP7": ("1,1'-BIPHENYL-3,4-DIOL", "non-polymer"),
        "BP6": ("2',6'-DICHLORO-BIPHENYL-2,6-DIOL", "non-polymer"),
        "BP9": ("5,5'-BUTANE-1,4-DIYLBIS(1,3,4-THIADIAZOL-2-AMINE)", "non-polymer"),
        "BP8": ("5,5'-PENTANE-1,5-DIYLBIS(1,3,4-THIADIAZOL-2-AMINE)", "non-polymer"),
        "NRQ": (
            "{(4Z)-4-(4-HYDROXYBENZYLIDENE)-2-[3-(METHYLTHIO)PROPANIMIDOYL]-5-OXO-4,5-DIHYDRO-1H-IMIDAZOL-1-YL}ACETIC ACID",
            "peptide",
        ),
        "NRP": (
            "{4-[(4-HYDROXYPHENYL)METHYLIDENE]-2-[(1E)-3-METHYLBUTANIMIDOYL]-5-OXO-4,5-DIHYDRO-1H-IMIDAZOL-1-YL}ACETIC ACID",
            "peptide",
        ),
        "NRU": ("RUTHENIUM (III) HEXAAMINE ION", "non-polymer"),
        "NRT": ("6-(TETRAHYDRO-2-FURANYLMETHOXY)-9H-PURIN-2-AMINE", "non-polymer"),
        "LX2": (
            "[4-(3-{[2-CHLORO-3-(TRIFLUOROMETHYL)BENZYL](2,2-DIPHENYLETHYL)AMINO}PROPOXY)-1H-INDOL-1-YL]ACETIC ACID",
            "non-polymer",
        ),
        "LX1": ("L-XYLULOSE 5-PHOSPHATE", "non-polymer"),
        "LX0": (
            "4-(4-AMINO-6-{[(1R)-1-NAPHTHALEN-2-YLETHYL]AMINO}-1,3,5-TRIAZIN-2-YL)-L-PHENYLALANINE",
            "non-polymer",
        ),
        "PDQ": (
            "3-AMINO-7-{(3R)-3-[(1S)-1-AMINOETHYL]PYRROLIDIN-1-YL}-1-CYCLOPROPYL-6-FLUORO-8-METHYLQUINAZOLINE-2,4(1H,3H)-DIONE",
            "non-polymer",
        ),
        "YYG": (
            "4-(3-[5-O-PHOSPHONORIBOFURANOSYL]-4,6-DIMETHYL-8-OXO-4,8-DIHYDRO-3H-1,3,4,5,7A-PENTAAZA-S-INDACEN-YLAMINO-BUTYRIC ACID METHYL ESTER",
            "nucleic-acid",
        ),
        "LX9": (
            "6-{3-AMINO-1-[3-(1H-INDOL-6-YL)PHENYL]-1H-PYRAZOL-4-YL}-3,4-DIHYDROISOQUINOLIN-1(2H)-ONE",
            "non-polymer",
        ),
        "LX8": (
            "2-[(2S)-6-(2-METHYLBUT-3-EN-2-YL)-7-OXO-2,3-DIHYDRO-7H-FURO[3,2-G]CHROMEN-2-YL]PROPAN-2-YL ACETATE",
            "non-polymer",
        ),
        "GMS": ("2'-DEOXYGUANOSINE-5'-MONOSELENOPHOSPHATE", "nucleic-acid"),
        "PXC": ("PHENAZINE-1,6-DICARBOXYLIC ACID", "non-polymer"),
        "NTC": ("2-HYDROXY-2-NITROMETHYL SUCCINIC ACID", "non-polymer"),
        "GMP": ("GUANOSINE", "non-polymer"),
        "GMU": ("?", "nucleic-acid"),
        "KN2": ("4-(6-HYDROXY-1H-INDAZOL-3-YL)BENZENE-1,3-DIOL", "non-polymer"),
        "KN3": (
            "4-[1-(3-METHYLBUT-2-EN-1-YL)-7-(TRIFLUOROMETHYL)-1H-INDAZOL-3-YL]BENZENE-1,3-DIOL",
            "non-polymer",
        ),
        "KN0": (
            "4-[1-BENZYL-7-(TRIFLUOROMETHYL)-1H-INDAZOL-3-YL]BENZENE-1,3-DIOL",
            "non-polymer",
        ),
        "KN1": (
            "4-[1-ALLYL-7-(TRIFLUOROMETHYL)-1H-INDAZOL-3-YL]BENZENE-1,3-DIOL",
            "non-polymer",
        ),
        "KNN": ("1-(2-METHYLPHENYL)METHANAMINE", "non-polymer"),
        "KNL": ("(2,6-DICHLOROPHENOXY)ACETIC ACID", "non-polymer"),
        "KNJ": (
            "(4R)-3-[(2S,3S)-3-{[(4-AMINO-2,6-DIMETHYLPHENOXY)ACETYL]AMINO}-2-HYDROXY-4-PHENYLBUTANOYL]-5,5-DIMETHYL-N-(2-METHYLPROP-2-EN-1-YL)-1,3-THIAZOLIDINE-4-CARBOXAMIDE",
            "peptide",
        ),
        "KNI": (
            "(4R)-N-TERT-BUTYL-3-[(2S,3S)-2-HYDROXY-3-({N-[(ISOQUINOLIN-5-YLOXY)ACETYL]-S-METHYL-L-CYSTEINYL}AMINO)-4-PHENYLBUTANOYL]-1,3-THIAZOLIDINE-4-CARBOXAMIDE",
            "peptide",
        ),
        "EQP": (
            "(4-ACETAMIDO-2,4-DIDEOXY-D-GLYCERO-ALPHA-D-GALACTO-1-OCTOPYRANOSYL)PHOSPHONIC ACID",
            "non-polymer",
        ),
        "EQU": ("EQUILENIN", "non-polymer"),
        "KNC": ("KANAMYCIN C", "non-polymer"),
        "KNA": ("NONANOIC ACID", "non-polymer"),
        "EQI": ("EQUILIN", "non-polymer"),
        "EQH": (
            "CIS-4-[3-AMINO-6-(3-CYCLOPROPYLTHIOPHEN-2-YL)PYRAZIN-2-YL]CYCLOHEXANECARBOXYLIC ACID",
            "non-polymer",
        ),
        "XEG": (
            "(2R,3S)-2-(3,4-DIHYDROXYPHENYL)-5,7-DIHYDROXY-3,4-DIHYDRO-2H-CHROMEN-3-YL 3,4,5-TRIHYDROXYBENZOATE",
            "non-polymer",
        ),
        "PY6": ("2-[O-PHOSPHONOPYRIDOXYL]-AMINO-HEXANOIC ACID", "non-polymer"),
        "EQO": (
            "4-{(E)-2-[5,5-DIMETHYL-8-(PHENYLETHYNYL)-5,6-DIHYDRONAPHTHALEN-2-YL]ETHENYL}BENZOIC ACID",
            "non-polymer",
        ),
        "EQN": (
            "4-{[(5,5,8,8-TETRAMETHYL-5,6,7,8-TETRAHYDRONAPHTHALEN-2-YL)CARBONYL]AMINO}BENZOIC ACID",
            "non-polymer",
        ),
        "JFD": (
            "6-AMINO-5-BROMO-1,2,3,4-TETRAHYDROPYRIMIDINE-2,4-DIONE",
            "non-polymer",
        ),
        "KNP": ("(2,6-DIFLUOROPHENOXY)ACETIC ACID", "non-polymer"),
        "LXV": (
            "2,5-DIBROMO-N-{(3R,5S)-1-[(Z)-IMINOMETHYL]-5-METHYLPYRROLIDIN-3-YL}BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "LXT": (
            "(2S)-2-AMINO-N-[(1S)-1-({(1S)-1-[(4-METHYLBENZYL)CARBAMOYL]-3-PHENYLPROPYL}CARBAMOYL)-3-PHENYLPROPYL]-4-PHENYLBUTANAMIDE",
            "non-polymer",
        ),
        "LXP": ("L-XYLITOL 5-PHOSPHATE", "non-polymer"),
        "IP0": ("2-(1-METHYLETHYL)PHENOL", "non-polymer"),
        "LXX": (
            "6-(1H-PYRAZOLO[3,4-B]PYRIDIN-5-YL)-4-PYRIDIN-4-YLQUINOLINE",
            "non-polymer",
        ),
        "LXG": (
            "8-[(E)-2-CYCLOPROPYLETHENYL]-2-[(DIMETHYLAMINO)METHYL][1]BENZOTHIENO[3,2-D]PYRIMIDIN-4(3H)-ONE",
            "non-polymer",
        ),
        "P0A": ("(2R)-2-AMINOHEPTANEDIOIC ACID", "peptide"),
        "LXC": ("BETA-L-XYLOPYRANOSE", "non-polymer"),
        "LXB": ("2-(ACETYLAMINO)-2-DEOXY-BETA-D-GULOPYRANOSE", "saccharide"),
        "HBM": (
            "2-(2,4-DICHLORO-PHENYL)-7-HYDROXY-1H-BENZOIMIDAZOLE-4-CARBOXYLIC ACID [2-(4-METHANESULFONYLAMINO-PHENYL)-ETHYL]-AMIDE",
            "non-polymer",
        ),
        "C39": (
            "N-[(2S,4S,5S,7R)-4-AMINO-8-(BUTYLAMINO)-5-HYDROXY-7-METHYL-2-(1-METHYLETHYL)-8-OXOOCTYL]-2-(3-METHOXYPROPOXY)BENZAMIDE",
            "non-polymer",
        ),
        "IP7": (
            "N~3~-[5-(1H-INDOL-6-YL)-2-(PYRIDIN-2-YLMETHOXY)BENZYL]PYRIDINE-2,3-DIAMINE",
            "non-polymer",
        ),
        "5EL": (
            "4-(4-{[(3S)-3-HYDROXY-1-AZABICYCLO[2.2.2]OCT-3-YL]ETHYNYL}PHENOXY)-N-(PYRIDIN-4-YLMETHYL)BENZAMIDE",
            "non-polymer",
        ),
        "5EO": (
            "6-ETHYL-2-{5-[(4-METHYLPIPERAZIN-1-YL)SULFONYL]-2-PROPOXYPHENYL}PYRIMIDIN-4(3H)-ONE",
            "non-polymer",
        ),
        "ED7": (
            "N-BENZYL-N-(2-{(4-CYANOPHENYL)[(1-METHYL-1H-IMIDAZOL-5-YL)METHYL]AMINO}ETHYL)-2-METHYLBENZENESULFONAMIDE",
            "non-polymer",
        ),
        "5EH": (
            "(3E)-3-(DIBENZO[B,E]OXEPIN-11(6H)-YLIDENE)-N,N-DIMETHYLPROPAN-1-AMINE",
            "non-polymer",
        ),
        "5EE": (
            "5R-(3,4-DICHLOROPHENYLMETHYL)-3-(2-THIOPHENESULFONYLAMINO)-4-OXO-2-THIONOTHIAZOLIDINE",
            "non-polymer",
        ),
        "IP5": (
            "(1R,2S,3R,4S,5S,6R)-6-HYDROXYCYCLOHEXANE-1,2,3,4,5-PENTAYL PENTAKIS[DIHYDROGEN (PHOSPHATE)]",
            "non-polymer",
        ),
        "5EA": (
            "5-{3,4-DIFLUORO-2-[(2-FLUORO-4-IODOPHENYL)AMINO]PHENYL}-N-(2-MORPHOLIN-4-YLETHYL)-1,3,4-OXADIAZOL-2-AMINE",
            "non-polymer",
        ),
        "X67": ("2-(4,6-DIAMINO-1,3,5-TRIAZIN-2-YL)-4-FLUOROPHENOL", "non-polymer"),
        "494": (
            "(2E)-3-(3-CHLORO-5-{4-CHLORO-2-[2-(2,4-DIOXO-3,4-DIHYDROPYRIMIDIN-1(2H)-YL)ETHOXY]PHENOXY}PHENYL)PROP-2-ENENITRILE",
            "non-polymer",
        ),
        "C37": ("5-FLUORO-2'-DEOXY-CYTIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "C36": ("5-METHYL-5-FLUORO-2'-DEOXY-CYTIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "YRR": ("3-HYDROXY-PYRROLIDIN-2-YLMETHYL-MONOPHOSPHATE GROUP", "nucleic-acid"),
        "G7M": ("N7-METHYL-GUANOSINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "K99": (
            "(2R,3R,4R,5R,6S)-2,3-DIFLUORO-4,5-DIHYDROXY-6-[(1R,2R)-1,2,3-TRIHYDROXYPROPYL]OXANE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "HBO": (
            "2,4-DIHYDROXY-7-(METHYLOXY)-2H-1,4-BENZOXAZIN-3(4H)-ONE",
            "non-polymer",
        ),
        "OPY": ("(3S)-4-OXO-4-PIPERIDIN-1-YLBUTANE-1,3-DIAMINE", "non-polymer"),
        "C33": ("2ALPHA-PROPYL-1ALPHA,25-DIHYDROXYVITAMIN D3", "non-polymer"),
        "YK9": ("4-(1H-IMIDAZO[4,5-C]PYRIDIN-2-YL)FLUOREN-9-ONE", "non-polymer"),
        "493": (
            "{4-[2-ACETYLAMINO-2-(1-BIPHENYL-4-YLMETHYL-2-OXO-AZEPAN-3-YLCARBAMOYL)-ETHYL]-2-PHOSPHONO-PHENOXY}-ACETIC ACID",
            "non-polymer",
        ),
        "G7X": ("ISPINESIB MESILATE", "non-polymer"),
        "V5X": (
            "(2R)-N~8~-HYDROXY-2-{[(5-METHOXY-2-METHYL-1H-INDOL-3-YL)ACETYL]AMINO}-N~1~-[2-(2-PHENYL-1H-INDOL-3-YL)ETHYL]OCTANEDIAMIDE",
            "non-polymer",
        ),
        "YKC": (
            "N-[4-(3H-IMIDAZO[4,5-C]PYRIDIN-2-YL)-9H-FLUOREN-9-YL-ISONICOTINAMIDE",
            "non-polymer",
        ),
        "YKB": (
            "N-[4-(3H-IMIDAZO[4,5-C]PYRIDIN-2-YL)-9H-FLUOREN-9-YL]-SUCCINAMIDE",
            "non-polymer",
        ),
        "YKE": (
            "N-[(4R)-4-(3H-IMIDAZO[4,5-C]PYRIDIN-2-YL)-4H-FLUOREN-9-YL]QUINOLINE-5-CARBOXAMIDE",
            "non-polymer",
        ),
        "YKJ": (
            "2-AMINO-N-[4-(3H-IMIDAZO[4,5-C]PYRIDIN-2-YL)--9H-FLUOREN-9-YL]-ISONICOTINAMIDE",
            "non-polymer",
        ),
        "YKI": (
            "1-H-PYRROLO[2,3-B]PYRIDINE-4-CARBOXYLIC ACID [4-(3H-IMIDAZO[4,5-C]PYRIDIN-2-YL)-9H-FLUOREN-9-YL]-AMIDE",
            "non-polymer",
        ),
        "YKN": (
            "2-[DIFLUORO-[(4-METHYL-PYRIMIDINYL)-THIO]METHYL]-BENZOXAZOLE",
            "non-polymer",
        ),
        "G75": (
            "ETHYL (4R)-4-({N-[(BENZYLOXY)CARBONYL]-L-PHENYLALANYL}AMINO)-5-[(3S)-2-OXOPYRROLIDIN-3-YL]PENTANOATE",
            "non-polymer",
        ),
        "G74": (
            "ETHYL (4R)-4-[(TERT-BUTOXYCARBONYL)AMINO]-5-[(3S)-2-OXOPYRROLIDIN-3-YL]PENTANOATE",
            "non-polymer",
        ),
        "V55": ("4-HYDROXY-3-METHOXYBENZALDEHYDE", "non-polymer"),
        "G73": (
            "(2E)-3-[4-({6-[(4-METHOXYPHENYL)AMINO]-7H-PURIN-2-YL}AMINO)-3,5-DIMETHYLPHENYL]PROP-2-ENENITRILE",
            "non-polymer",
        ),
        "G72": (
            "(1R,3R,7E,17BETA)-17-[(1S)-6,6,6-TRIFLUORO-5-HYDROXY-1-(4-HYDROXY-4-METHYLPENTYL)-5-(TRIFLUOROMETHYL)HEX-3-YN-1-YL]-9,10-SECOESTRA-5,7-DIENE-1,3-DIOL",
            "non-polymer",
        ),
        "K91": ("4-CHLORO-2-[(5-CHLOROQUINOLIN-8-YL)OXY]PHENOL", "non-polymer"),
        "4QB": (
            "3-{2-ETHYL-6-[4-(QUINOXALIN-2-YLAMINO)-BENZOYLAMINO]-HEXANOYLAMINO}-4-OXO-BUTYRIC ACID",
            "non-polymer",
        ),
        "4QC": (
            "3-[6-{2'-[(DIMETHYLAMINO)METHYL]BIPHENYL-4-YL}-7-OXO-3-(TRIFLUOROMETHYL)-4,5,6,7-TETRAHYDRO-1H-PYRAZOLO[3,4-C]PYRIDIN-1-YL]BENZAMIDE",
            "non-polymer",
        ),
        "P0E": ("PHOSPHATIDYL ETHANOL", "non-polymer"),
        "TOH": ("(3E)-4-(2-HYDROXYPHENYL)-2-OXOBUT-3-ENOIC ACID", "non-polymer"),
        "K95": (
            "(4R)-N-[(1S,2R)-2-HYDROXY-2,3-DIHYDRO-1H-INDEN-1-YL]-3-[(2S,3S)-2-HYDROXY-3-{[S-METHYL-N-(PHENYLACETYL)-L-CYSTEINYL]AMINO}-4-PHENYLBUTANOYL]-5,5-DIMETHYL-1,3-THIAZOLIDINE-4-CARBOXAMIDE",
            "peptide",
        ),
        "4QS": (
            "[(3R,5S,6R,7R,10R,11S,12E)-5,11,21-TRIMETHOXY-3,7,9,15,19-PENTAMETHYL-6-OXIDANYL-16,20,22-TRIS(OXIDANYLIDENE)-17-AZABICYCLO[16.3.1]DOCOSA-1(21),8,12,14,18-PENTAEN-10-YL] CARBAMATE",
            "non-polymer",
        ),
        "NAV": (
            "6,7,8,9-TETRAHYDROXY-5-METHYLCARBOXAMIDO-4-OXONONANOIC ACID",
            "non-polymer",
        ),
        "2G0": (
            "(2S)-1-[(2S)-6-AMINO-2-({[(2S,3S,4R,5S,6S)-3,4,5-TRIHYDROXY-6-METHYLTETRAHYDRO-2H-PYRAN-2-YL]ACETYL}AMINO)HEXANOYL]-N-[(1S)-1-CARBAMOYL-3-METHYLBUTYL]PYRROLIDINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "LI3": ("3-FLUORO-N-1H-INDOL-5-YL-5-MORPHOLIN-4-YLBENZAMIDE", "non-polymer"),
        "9DI": ("9-DEAZAINOSINE", "non-polymer"),
        "3GP": ("GUANOSINE-3'-MONOPHOSPHATE", "non-polymer"),
        "3GR": ("GLYCERALDEHYDE", "non-polymer"),
        "3GU": ("N6-CYCLOPENTYLADENOSINE", "non-polymer"),
        "3GW": (
            "5-(5-FORMYLTHIOPHEN-2-YL)URIDINE-5'(2-(ACETYLAMINO)-2-DEOXY-ALPHA-D-GALACTOSYL)-DIPHOSPHATE",
            "non-polymer",
        ),
        "3GV": (
            "(3S)-1-(2-HYDROXYPHENYL)-5-OXOPYRROLIDINE-3-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "3GH": (
            "N-{(2S,3R)-1-[(3-DEOXY-ALPHA-D-XYLO-HEXOPYRANOSYL)OXY]-3-HYDROXYOCTADECAN-2-YL}HEXACOSANAMIDE",
            "non-polymer",
        ),
        "3GL": ("(2S,4S)-2-AMINO-4-HYDROXY-PENTANEDIOIC ACID", "peptide"),
        "3GO": (
            "4-[[(2R,3S,4R,5R)-5-[6-AMINO-8-(QUINOLIN-6-YLMETHYLAMINO)PURIN-9-YL]-3,4-DIHYDROXY-OXOLAN-2-YL]METHOXYMETHYL]BENZONITRILE",
            "non-polymer",
        ),
        "3GN": (
            "2-[(2S)-2-METHYLPYRROLIDIN-2-YL]-1H-BENZIMIDAZOLE-7-CARBOXAMIDE",
            "non-polymer",
        ),
        "3GA": (
            "(3S)-3-[(2S)-2-AMINO-3-OXOPROPYL]PIPERIDINE-1-CARBOXIMIDAMIDE",
            "peptide",
        ),
        "3GC": ("GAMMA-GLUTAMYLCYSTEINE", "non-polymer"),
        "3GB": (
            "N-[(2S,3R,4E)-1-{[ALPHA-D-GALACTOPYRANOSYL-(1->4)-BETA-D-GALACTOPYRANOSYL-(1->4)-BETA-D-GLUCOPYRANOSYL]OXY}-3-HYDROXYOCTADEC-4-EN-2-YL]HEXACOSANAMIDE",
            "non-polymer",
        ),
        "9DR": (
            "6-ETHYL-5-{3-[3-METHOXY-5-(PYRIDIN-4-YL)PHENYL]PROP-1-YN-1-YL}PYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "3GD": (
            "N-[(2S,3R,4E)-1-{[4-O-(BETA-D-GALACTOPYRANOSYL)-BETA-D-GLUCOPYRANOSYL]OXY}-3-HYDROXYOCTADEC-4-EN-2-YL]DOCOSANAMIDE",
            "non-polymer",
        ),
        "9DP": (
            "(3S)-6-(3-TERT-BUTYL-5-{[(2,3-DICHLOROPHENYL)CARBAMOYL]AMINO}-1H-PYRAZOL-1-YL)-1,2,3,4-TETRAHYDROISOQUINOLINE-3-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "ZBC": (
            "(2S)-3-(4-AMINO-5-BROMO-2-OXOPYRIMIDIN-1(2H)-YL)-2-HYDROXYPROPYL DIHYDROGEN PHOSPHATE",
            "nucleic-acid",
        ),
        "ZBA": (
            "12,13-EPOXYTRICHOTHEC-9-ENE-3,4,8,15-TETROL-4,15-DIACETATE-8-ISOVALERATE",
            "non-polymer",
        ),
        "OMY": ("(BETAR)-3-CHLORO-BETA-HYDROXY-L-TYROSINE", "peptide"),
        "ZBU": ("(S)-1'-(2',3'-DIHYDROXYPROPYL)-5-BROMO-URACIL", "nucleic-acid"),
        "63C": (
            "(3R,4R)-1-METHYLCARBAMOYLMETHYL-PYRROLIDINE-3,4-DICARBOXYLIC ACID 3-[(4-CHLORO-PHENYL)-AMIDE] 4-{[2-FLUORO-4-(2-OXO-2H-PYRIDIN-1-YL)-PHENYL]-AMIDE}",
            "non-polymer",
        ),
        "ZBR": ("1,3,5-TRIS(BROMOMETHYL)BENZENE", "non-polymer"),
        "508": (
            "(1R,4'S)-3,4-DIHYDRO-2H-SPIRO[NAPHTHALENE-1,3'-PYRROLIDIN]-4'-YL[(2S,4R)-2,4-DIPHENYLPIPERIDIN-1-YL]METHANONE",
            "non-polymer",
        ),
        "509": (
            "4-BROMO-3-(CARBOXYMETHOXY)-5-PHENYLTHIOPHENE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "506": (
            "(2E)-3-(3-{2-[2-(2,4-DIOXO-3,4-DIHYDROPYRIMIDIN-1(2H)-YL)ETHOXY]-4-FLUOROPHENOXY}-5-FLUOROPHENYL)PROP-2-ENENITRILE",
            "non-polymer",
        ),
        "507": (
            "1-{3-[6-(TETRAHYDRO-2H-PYRAN-4-YLAMINO)IMIDAZO[1,2-B]PYRIDAZIN-3-YL]PHENYL}ETHANONE",
            "non-polymer",
        ),
        "505": ("(2R)-1-(2,6-DIMETHYLPHENOXY)PROPAN-2-AMINE", "non-polymer"),
        "WA2": ("(2R)-2-METHYL-4-(7H-PURIN-6-YLAMINO)BUTAN-1-OL", "non-polymer"),
        "WA1": ("1-PHENYL-3-(1,2,3-THIADIAZOL-5-YL)UREA", "non-polymer"),
        "501": (
            "N7-BUTYL-N2-(5-CHLORO-2-METHYLPHENYL)-5-METHYL[1,2,4]TRIAZOLO[1,5-A]PYRIMIDINE-2,7-DIAMINE",
            "non-polymer",
        ),
        "WAN": (
            "5-ETHOXY-4-(1-METHYL-7-OXO-3-PROPYL-6,7-DIHYDRO-1H-PYRAZOLO[4,3-D]PYRIMIDIN-5-YL)THIOPHENE-2-SULFONAMIDE",
            "non-polymer",
        ),
        "632": (
            "2-(5-BENZYLAMINO-2-METHYLSULFANYL-6-OXO-6H-PYRIMIDIN-1-YL)-N-[4-GUANIDINO-1-(THIAZOLE-2-CARBONYL)-BUTYL]-ACETAMIDE",
            "non-polymer",
        ),
        "635": (
            "1-METHYLETHYL 8-FLUORO-1,1-DIMETHYL-3-{[4-(3-MORPHOLIN-4-YLPROPOXY)PHENYL]CARBONYL}-1,2,3,6-TETRAHYDROAZEPINO[4,5-B]INDOLE-5-CARBOXYLATE",
            "non-polymer",
        ),
        "PS8": (
            "2,6-DICHLORO-4-(6-PIPERAZIN-1-YLPYRIDIN-3-YL)-N-(1,3,5-TRIMETHYL-1H-PYRAZOL-4-YL)BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "637": (
            "N-[(E)-3-[(2R,3S,4R,5R)-3,4-DIHYDROXY-5-(6-METHYLPURIN-9-YL)OXOLAN-2-YL]PROP-2-ENYL]-5-(4-FLUOROPHENYL)-2,3-DIHYDROXY-BENZAMIDE",
            "non-polymer",
        ),
        "638": (
            "[4R-(4ALPHA,5ALPHA,6BETA,7BETA)]-3,3'-[[TETRAHYDRO-5,6-DIHYDROXY-2-OXO-4,7-BIS(PHENYLMETHYL)-1H-1,3-DIAZEPINE-1,3(2H)-DIYL]BIS(METHYLENE)]BIS[N-2-THIAZOLYLBENZAMIDE]",
            "non-polymer",
        ),
        "OMX": ("(BETAR)-BETA-HYDROXY-L-TYROSINE", "peptide"),
        "WAC": (
            "N-{3-[(12BS)-7-OXO-1,3,4,6,7,12B-HEXAHYDROINDOLO[2,3-A]QUINOLIZIN-12(2H)-YL]PROPYL}PROPANE-2-SULFONAMIDE",
            "non-polymer",
        ),
        "50A": ("2-DEOXY-2-FLUORO-5-O-PHOSPHONO-ALPHA-D-RIBOFURANOSE", "saccharide"),
        "RNH": (
            "(2R)-N-[3-(HEPTYLAMINO)-3-OXOPROPYL]-2,4-DIHYDROXY-3,3-DIMETHYLBUTANAMIDE",
            "non-polymer",
        ),
        "50Z": (
            "[4-AMINO-2-(PROP-2-EN-1-YLAMINO)-1,3-THIAZOL-5-YL](PYRIDIN-2-YL)METHANONE",
            "non-polymer",
        ),
        "50X": (
            "(2R,3R)-4-[(2R)-2-(3-CHLOROPHENYL)PYRROLIDIN-1-YL]-2,3-DIHYDROXY-4-OXO-N-[(5-{[2-(TRIFLUOROMETHYL)-1H-BENZIMIDAZOL-1-YL]METHYL}THIOPHEN-2-YL)METHYL]BUTANAMIDE",
            "non-polymer",
        ),
        "RNO": ("(R)-PARA-NITROSTYRENE OXIDE", "non-polymer"),
        "50U": (
            "(S)-N-(4-CARBAMIMIDOYLBENZYL)-1-(3-CYCLOHEXYLPROPANOYL)PYRROLIDINE-2-CARBOXAMIDE",
            "peptide",
        ),
        "LCS": (
            "[5-HYDROXY-6-METHYL-4-({[(4E)-3-OXO-1,2-OXAZOLIDIN-4-YLIDENE]AMINO}METHYL)PYRIDIN-3-YL]METHYL DIHYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "2GP": ("GUANOSINE-2'-MONOPHOSPHATE", "non-polymer"),
        "2GS": ("2-O-METHYL-ALPHA-D-GALACTOPYRANOSE", "non-polymer"),
        "2GT": ("?", "nucleic-acid"),
        "LCP": ("PERCHLORATE ION", "non-polymer"),
        "2GW": ("5-PHENYL-URIDINE-5'-ALPHA-D-GALACTOSYL-DIPHOSPHATE", "non-polymer"),
        "3G3": (
            "2-[2-(1H-TETRAZOL-5-YL)ETHYL]-1H-ISOINDOLE-1,3(2H)-DIONE",
            "non-polymer",
        ),
        "3G4": (
            "2-(6-{[(3-CHLORO-2-METHYLPHENYL)SULFONYL]AMINO}PYRIDIN-2-YL)-N,N-DIETHYLACETAMIDE",
            "non-polymer",
        ),
        "2GA": (
            "2-[(3,5-DIMETHYL-1H-PYRAZOL-4-YL)SULFANYL]-5-METHYLSULFANYL-1,3,4-THIADIAZOLE",
            "non-polymer",
        ),
        "2GG": (
            "5-(5-CHLORO-2,4-DIHYDROXYPHENYL)-N-ETHYL-4-(4-METHOXYPHENYL)ISOXAZOLE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "2GJ": (
            "5-[2,4-DIHYDROXY-5-(1-METHYLETHYL)PHENYL]-N-ETHYL-4-[4-(MORPHOLIN-4-YLMETHYL)PHENYL]ISOXAZOLE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "2GL": ("4-O-ACETYL-2,6-DIDEOXY-ALPHA-D-GALACTO-HEXOPYRANOSE", "saccharide"),
        "2GM": ("DIHYDROERGOTAMINE", "non-polymer"),
        "8JZ": ("BUTYL ACETATE", "non-polymer"),
        "URD": (
            "1-((2R,3R,4S,5R)-TETRAHYDRO-3,4-DIHYDROXY-5-(HYDROXYMETHYL)FURAN-2-YL)PYRIDINE-2,4(1H,3H)-DIONE",
            "nucleic-acid",
        ),
        "O30": (
            "N-(3-AMINOPROPYL)-2-({[3-(TRIFLUOROMETHYL)-4,5,6,7-TETRAHYDRO-1H-INDAZOL-1-YL]ACETYL}AMINO)-4,5,6,7-TETRAHYDRO-1-BENZOTHIOPHENE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "0E8": (
            "DI-TERT-BUTYL {IMINOBIS[(2S,3S)-3-HYDROXY-1-PHENYLBUTANE-4,2-DIYL]}BISCARBAMATE",
            "peptide",
        ),
        "0E9": (
            "BENZYL [(1R,4S,6S,9R)-4,6-DIBENZYL-5-HYDROXY-1,9-BIS(1-METHYLETHYL)-2,8,11-TRIOXO-13-PHENYL-12-OXA-3,7,10-TRIAZATRIDEC-1-YL]CARBAMATE",
            "peptide",
        ),
        "2SM": (
            "METHYL 2-{[(4-METHYLPYRIMIDIN-2-YL)CARBAMOYL]SULFAMOYL}BENZOATE",
            "non-polymer",
        ),
        "0E5": ("(2S,3R)-2-AZANYL-3-METHYL-3-OXIDANYL-PENTANOIC ACID", "peptide"),
        "0E6": (
            "AMINO{[(4S)-4-{[(BENZYLOXY)CARBONYL]AMINO}-5-{[(2S)-1-(BENZYLOXY)-4-CHLORO-3-OXOBUTAN-2-YL]AMINO}-5-OXOPENTYL]AMINO}METHANIMINIUM",
            "peptide",
        ),
        "0E7": (
            "METHYL N-(4-CARBAMIMIDAMIDOBUTANOYL)-L-PHENYLALANYL-L-ALLOTHREONYL-L-PHENYLALANINATE",
            "peptide",
        ),
        "LCX": ("CARBOXYLATED LYSINE", "peptide"),
        "X9I": (
            "4-{[(2-CARBAMOYL-4-NITROPHENYL)AMINO]METHYL}BENZOIC ACID",
            "non-polymer",
        ),
        "LCY": ("1-METHYLPYRROLIDINE-2,5-DIONE", "non-polymer"),
        "AYH": (
            "(1R,3S)-3-[1-(ACETYLAMINO)-1-METHYLETHYL]-N-[(1S,2S,4R)-1-BENZYL-5-(BUTYLAMINO)-2-HYDROXY-4-METHYL-5-OXOPENTYL]CYCLOHEXANECARBOXAMIDE",
            "non-polymer",
        ),
        "FD1": (
            "N-ALPHA-(2-NAPHTHYLSULFONYL)-N-(3-AMIDINO-L-PHENYLALANINYL)-D-PIPECOLINIC ACID",
            "non-polymer",
        ),
        "PET": ("1,5-DI(4-AMIDINOPHENOXY)-3-OXA-PENTANE", "non-polymer"),
        "XXP": ("2-KETO,5-NITRO,6-HYDROXY-3,5-HEXADIENOIC ACID", "non-polymer"),
        "U73": ("N-(5-ISOPROPYL-THIAZOL-2-YL)-2-PYRIDIN-3-YL-ACETAMIDE", "non-polymer"),
        "PC": ("PHOSPHOCHOLINE", "non-polymer"),
        "XXY": (
            "2-[(1R,2R)-1-AMINO-2-HYDROXYPROPYL]-1-(CARBOXYMETHYL)-4-(1H-IMIDAZOL-5-YLMETHYL)-1H-IMIDAZOL-5-OLATE",
            "peptide",
        ),
        "XXX": (
            "(2R,3S,4R,5R,6R)-6-((1R,2R,3S,4R,6S)-4,6-DIAMINO-2,3-DIHYDROXYCYCLOHEXYLOXY)-5-AMINO-2-(AMINOMETHYL)-TETRAHYDRO-2H-PYRAN-3,4-DIOL",
            "non-polymer",
        ),
        "FD3": (
            "N-ALPHA-(2-NAPHTHYLSULFONYL)-3-AMIDINO-L-PHENYLALANINE PIPERAZIDE",
            "non-polymer",
        ),
        "2SD": (
            "(2S,3R,4R,5R,6R)-3,4-DIHYDROXY-6-(HYDROXYMETHYL)-5-[(2R,3R,4S,5R,6R)-3,4,5-TRIHYDROXY-6-(HYDROXYMETHYL)OXAN-2-YL]OXY-OXANE-2-SULFONAMIDE",
            "non-polymer",
        ),
        "XXA": (
            "2,4,5-TRIDEOXY-2-(METHYLAMINO)-4-[(2E)-PENTA-2,4-DIEN-1-YL]-L-XYLONIC ACID",
            "peptide",
        ),
        "FD2": (
            "N-ALPHA-(2-NAPHTHYLSULFONYL)-N(3-AMIDINO-L-PHENYLALANINYL)ISOPIPECOLINIC ACID METHYL ESTER",
            "non-polymer",
        ),
        "XXG": (
            "3,3-DIHYDROXY-4-OXOCYCLOHEXA-1,5-DIENE-1-SULFONIC ACID",
            "non-polymer",
        ),
        "XXD": ("(2S)-PENTANE-1,2,5-TRIOL", "non-polymer"),
        "O7U": (
            "(2R)-2-AMINO-2,5-DIHYDRO-1,5,2-DIAZAPHOSPHININ-6(1H)-ONE 2-OXIDE",
            "non-polymer",
        ),
        "XXH": ("4-NITROBENZALDEHYDE", "non-polymer"),
        "LCC": (
            "[(1R,3R,4R,7S)-7-HYDROXY-3-(5-METHYLCYTOSIN-1-YL)-2,5-DIOXABICYCLO[2.2.1]HEPT-1-YL]METHYL DIHYDROGEN PHOSPHATE",
            "nucleic-acid",
        ),
        "XX3": (
            "(1S)-1-HYDROPEROXY-1-HYDROXY-2-KETO-5-NITROCYCLOHEXA-3,5-DIENE",
            "non-polymer",
        ),
        "XX2": ("1-KETO,2-HYDROXY,4-NITROBENZENE, 1 ELECTRON OXIDIZED", "non-polymer"),
        "XX1": ("N~6~-7H-PURIN-6-YL-L-LYSINE", "peptide"),
        "URO": ("(2E)-3-(1H-IMIDAZOL-4-YL)ACRYLIC ACID", "non-polymer"),
        "XX7": (
            "2-{[(2S)-2-HYDROXY-3-(9-PHENANTHRYLOXY)PROPYL]AMINO}PROPANE-1,3-DIOL",
            "non-polymer",
        ),
        "XX6": (
            "2-{[(2R)-2-HYDROXY-3-(9-PHENANTHRYLOXY)PROPYL]AMINO}PROPANE-1,3-DIOL",
            "non-polymer",
        ),
        "XX5": (
            "(S,S)-2-{1-CARBOXY-2-[3-(3,5-DICHLORO-BENZYL)-3H-IMIDAZOL-4-YL]-ETHYLAMINO}-4-METHYL-PENTANOIC ACID",
            "non-polymer",
        ),
        "XX4": (
            "3-(2-AMINO-6-BENZOYLQUINAZOLIN-3(4H)-YL)-N-CYCLOHEXYL-N-METHYLPROPANAMIDE",
            "non-polymer",
        ),
        "IR": ("IRIDIUM ION", "non-polymer"),
        "IU": ("5-IODOURIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "LCN": ("1,5-ANHYDRO-D-ARABINO-HEX-1-ENITOL", "saccharide"),
        "XTG": (
            "4-NITROPHENYL 6-THIO-6-S-ALPHA-D-XYLOPYRANOSYL-BETA-D-GLUCOPYRANOSIDE",
            "non-polymer",
        ),
        "AYC": ("ALAREMYCIN 2", "non-polymer"),
        "FD8": (
            "N-[3-(PENTAFLUOROPHENOXY)PHENYL]-2-(TRIFLUOROMETHYL)BENZAMIDE",
            "non-polymer",
        ),
        "IA": (
            "2'-(3-IODOPHENYL)-5-(4-METHYL-1-PIPERAZINYL)-2,5'-BI-BENZIMIDAZOLE",
            "non-polymer",
        ),
        "IC": ("ISOCYTIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "AYB": (
            "5'-O-[(S)-({(2S)-2-AMINO-6-[(PROPOXYCARBONYL)AMINO]HEXANOYL}OXY)(HYDROXY)PHOSPHORYL]ADENOSINE",
            "non-polymer",
        ),
        "HBA": ("P-HYDROXYBENZALDEHYDE", "non-polymer"),
        "TE6": ("ORTHOTELLURIC ACID", "non-polymer"),
        "6DR": (
            "6-ETHYL-5-{3-[2-METHOXY-5-(PYRIDIN-4-YL)PHENYL]PROP-1-YN-1-YL}PYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "0EY": (
            "1-{[1-(ALPHA-L-ARABINOFURANOSYL)-1H-1,2,3-TRIAZOL-4-YL]METHYL}-2,4-DIOXO-1,2,3,4-TETRAHYDROPYRIMIDINE",
            "non-polymer",
        ),
        "0EZ": (
            "TERT-BUTYL [(1S,2S)-1-BENZYL-2-HYDROXY-3-{[(8S,11R)-8-[(1R)-1-METHYLPROPYL]-7,10-DIOXO-2-OXA-6,9-DIAZABICYCLO[11.2.2]HEPTADECA-1(15),13,16-TRIEN-11-YL]AMINO}PROPYL]CARBAMATE",
            "peptide",
        ),
        "0ET": ("UNDECAN-2-ONE-COA", "non-polymer"),
        "0EU": (
            "1-[3-({4-[(5S)-3,3-DIMETHYL-1-OXO-2-OXA-7-AZASPIRO[4.5]DEC-7-YL]PIPERIDIN-1-YL}CARBONYL)-1-BENZOTHIOPHEN-2-YL]-3-ETHYLUREA",
            "non-polymer",
        ),
        "LCK": ("(Z)-N~6~-(2-CARBOXY-1-METHYLETHYLIDENE)-L-LYSINE", "peptide"),
        "0EL": (
            "6-AMMONIO-N-[(2R,4R,5R)-5-{[N-(TERT-BUTOXYCARBONYL)-L-PHENYLALANYL-3-(1H-IMIDAZOL-3-IUM-4-YL)-L-ALANYL]AMINO}-6-CYCLOHEXYL-4-HYDROXY-2-(2-METHYLPROPYL)HEXANOYL]-L-NORLEUCYLPHENYLALANINE",
            "peptide",
        ),
        "0EM": (
            "N-(TERT-BUTOXYCARBONYL)-L-PHENYLALANYL-N-{(1S)-1-[(R)-HYDROXY(2-{[(2S)-2-METHYLBUTYL]AMINO}-2-OXOETHYL)PHOSPHORYL]-3-METHYLBUTYL}-3-(1H-IMIDAZOL-3-IUM-4-YL)-L-ALANINAMIDE",
            "peptide",
        ),
        "0EN": (
            "N-[(1R)-2-(TERT-BUTYLAMINO)-2-OXO-1-(PYRIDIN-3-YL)ETHYL]-N-(4-TERT-BUTYLPHENYL)FURAN-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "0EO": (
            "(2S)-2-[[(3S,4S)-5-CYCLOHEXYL-4-[[(4S,5S)-5-[(2-METHYLPROPAN-2-YL)OXYCARBONYLAMINO]-4-OXIDANYL-6-PHENYL-HEXANOYL]AMINO]-3-OXIDANYL-PENTANOYL]AMINO]-4-METHYL-PENTANOIC ACID",
            "peptide",
        ),
        "0EH": ("(2R)-2-AMINO-2-METHYLNONANOIC ACID", "peptide"),
        "LCH": (
            "[(1R,3R,4R,7S)-7-HYDROXY-3-(5-METHYLCYTOSIN-1-YL)-2,5-DIOXABICYCLO[2.2.1]HEPT-1-YL]METHYL DIHYDROGEN PHOSPHATE",
            "nucleic-acid",
        ),
        "0EJ": (
            "(2-OXO-2-{[(1S,2R,3S,4S,5R,6S)-2,3,4,5,6-PENTAKIS(PHOSPHONOOXY)CYCLOHEXYL]OXY}ETHYL)PHOSPHONIC ACID",
            "non-polymer",
        ),
        "0EK": (
            "6-AMMONIO-N-{[(2R,3R)-3-{[N-(TERT-BUTOXYCARBONYL)-L-PHENYLALANYL-3-(1H-IMIDAZOL-3-IUM-4-YL)-L-ALANYL]AMINO}-4-CYCLOHEXYL-2-HYDROXYBUTYL](2-METHYLPROPYL)CARBAMOYL}-L-NORLEUCYL-L-PHENYLALANINE",
            "peptide",
        ),
        "0ED": (
            "N-[(1S)-3-{[(BENZYLOXY)CARBONYL]AMINO}-1-CARBOXYPROPYL]-L-LEUCYL-N-(2-MORPHOLIN-4-YLETHYL)-L-PHENYLALANINAMIDE",
            "peptide",
        ),
        "0EF": (
            "N-(TERT-BUTOXYCARBONYL)-L-ALANYL-N-[(1R)-1-(CARBOXYAMINO)-2-PHENYLETHYL]-L-PROLINAMIDE",
            "peptide",
        ),
        "0EG": (
            "N-(TERT-BUTOXYCARBONYL)-L-ALANYL-N-[(1R)-1-(DIHYDROXYBORANYL)-2-METHYLPROPYL]-L-PROLINAMIDE",
            "peptide",
        ),
        "NEN": ("1-ETHYL-PYRROLIDINE-2,5-DIONE", "non-polymer"),
        "0EA": ("(BETAR)-BETA,3-DIHYDROXY-N-METHYL-L-TYROSINE", "peptide"),
        "0EC": ("6,7-DIMETHOXY-2-(PIPERAZIN-1-YL)QUINAZOLIN-4-AMINE", "non-polymer"),
        "SC7": (
            "N'-{(1S,2R)-1-(3,5-DIFLUOROBENZYL)-2-[(2R,4S)-4-ETHOXYPIPERIDIN-2-YL]-2-HYDROXYETHYL}-5-METHYL-N,N-DIPROPYLISOPHTHALAMIDE",
            "non-polymer",
        ),
        "PL6": (
            "(E)-N-({3-HYDROXY-2-METHYL-5-[(PHOSPHONOOXY)METHYL]PYRIDIN-4-YL}METHYLIDENE)-L-GLUTAMIC ACID",
            "non-polymer",
        ),
        "TDV": (
            "1-[(2S,4S)-4-[4-(4-FLUOROPHENYL)-1,2,3-TRIAZOL-1-YL]-2-PYRROLIDIN-1-YLCARBONYL-PYRROLIDIN-1-YL]-4-PHENYL-BUTAN-1-ONE",
            "non-polymer",
        ),
        "S57": ("1-[6-(2-CHLORO-4-METHYXYPHENOXY)-HEXYL]-IMIDAZOLE", "non-polymer"),
        "S55": (
            "METHYL (2S,3R,4S)-3-ETHYL-2-(BETA-D-GLUCOPYRANOSYLOXY)-4-[(1S)-2,3,4,9-TETRAHYDRO-1H-BETA-CARBOLIN-1-YLMETHYL]-3,4-DIHYDRO-2H-PYRAN-5-CARBOXYLATE",
            "non-polymer",
        ),
        "PL4": (
            "(2S,6S)-2-AMINO-6-[({3-HYDROXY-2-METHYL-5-[(PHOSPHONOOXY)METHYL]PYRIDIN-4-YL}METHYL)AMINO]HEPTANEDIOIC ACID",
            "non-polymer",
        ),
        "S53": (
            "[7-BENZYL-2-(5-FLUORO-2-METHYLPHENOXY)-1-PHENYL-1H-PYRROLO[2,3-C]PYRIDIN-3-YL](PIPERAZIN-1-YL)METHANONE",
            "non-polymer",
        ),
        "S52": (
            "2-PHENOXY-1-PHENYL-3-(PIPERAZIN-1-YLCARBONYL)-1H-INDOLE",
            "non-polymer",
        ),
        "S51": (
            "2-BENZYL-1-PHENYL-3-(PIPERAZIN-1-YLCARBONYL)-1H-INDOLE",
            "non-polymer",
        ),
        "5DP": (
            "5-(2,5-DIOXOPYRROLIDIN-1-YL)-N-[2-(2-{2-[(4-O-ALPHA-D-IDOPYRANOSYL-ALPHA-D-MANNOPYRANOSYL)OXY]ETHOXY}ETHOXY)ETHYL]PENTANAMIDE",
            "non-polymer",
        ),
        "GMH": ("L-GLYCERO-D-MANNO-HEPTOPYRANOSE", "saccharide"),
        "S58": (
            "1-PHENYLSULFONAMIDE-3-TRIFLUOROMETHYL-5-PARABROMOPHENYLPYRAZOLE",
            "non-polymer",
        ),
        "SXA": (
            "THIOACETIC ACID S-{2-[3-(2-HYDROXY-3,3-DIMETHYL-4-PHOSPHONOOXY-BUTYRYLAMINO)-PROPIONYLAMINO]-ETHYL} ESTER",
            "non-polymer",
        ),
        "SXC": (
            "BROMO(4-{3-[(R)-ETHOXY(4-NITROPHENOXY)PHOSPHORYL]PROPYL}-2,6-BIS[(METHYLSULFANYL-KAPPAS)METHYL]PHENYL-KAPPAC~1~)PALLADIUM(2+)",
            "non-polymer",
        ),
        "SXD": (
            "N-{2-[(3,5-DIOXOHEXYL)SULFANYL]ETHYL}-N~3~-[(2S)-2-HYDROXY-3,3-DIMETHYL-4-(PHOSPHONOOXY)BUTANOYL]-BETA-ALANINAMIDE",
            "non-polymer",
        ),
        "SXE": ("O-{(S)-ETHOXY[(1-METHYLETHYL)AMINO]PHOSPHORYL}-L-SERINE", "peptide"),
        "SXH": (
            "S-[2-({N-[(2S)-2-HYDROXY-3,3-DIMETHYL-4-(PHOSPHONOOXY)BUTANOYL]-BETA-ALANYL}AMINO)ETHYL] HEXANETHIOATE",
            "non-polymer",
        ),
        "2SP": (
            "(1S)-2-(PHOSPHONOOXY)-1-[(STEAROYLOXY)METHYL]ETHYL STEARATE",
            "non-polymer",
        ),
        "P0L": (
            "(2S)-2-{[(S)-[(1R)-1-AMINO-3-METHYLBUTYL](HYDROXY)PHOSPHORYL]OXY}-3-PHENYLPROPANOIC ACID",
            "peptide",
        ),
        "SXM": (
            "3-{[2-({N-[(2S)-2-HYDROXY-3,3-DIMETHYL-4-(PHOSPHONOOXY)BUTANOYL]-BETA-ALANYL}AMINO)ETHYL]SULFANYL}-3-OXOPROPANOIC ACID",
            "non-polymer",
        ),
        "SXN": ("SALINIXANTHIN", "non-polymer"),
        "SXO": (
            "S-[2-({N-[(2S)-2-HYDROXY-3,3-DIMETHYL-4-(PHOSPHONOOXY)BUTANOYL]-BETA-ALANYL}AMINO)ETHYL] OCTANETHIOATE",
            "non-polymer",
        ),
        "PS0": ("3(S)-AMINO-4-PHENYL-BUTAN-2(S)-OL", "peptide"),
        "HBB": (
            "N-{(2R)-2-HYDROXY-2-[(8S,11S)-8-ISOPROPYL-6,9-DIOXO-2-OXA-7,10-DIAZABICYCLO[11.2.2]HEPTADECA-1(15),13,16-TRIEN-11-YL]ETHYL}-N-ISOPENTYLBENZENESULFONAMIDE",
            "non-polymer",
        ),
        "IPB": ("5-METHYL-2-(1-METHYLETHYL)PHENOL", "non-polymer"),
        "YVQ": (
            "3-(1H-INDOL-3-YL)-6-(1H-PYRAZOL-4-YL)IMIDAZO[1,2-A]PYRAZINE",
            "non-polymer",
        ),
        "PYY": ("D-RIBOFURANOSYL-BENZENE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "DK4": (
            "1-(3-DEOXY-3-FLUORO-BETA-D-GLUCOPYRANOSYL)-5-FLUOROPYRIMIDINE-2,4(1H,3H)-DIONE",
            "non-polymer",
        ),
        "SX0": (
            "(2S)-2-AMINO-4-({[(2S,3S,4R,5R)-5-(4-AMINO-5-BROMO-7H-PYRROLO[2,3-D]PYRIMIDIN-7-YL)-3,4-DIHYDROXYTETRAHYDROFURAN-2-YL]METHYL}SULFANYL)BUTANOIC ACID (NON-PREFERRED NAME)",
            "non-polymer",
        ),
        "SX1": ("2-AMINO-5-BROMOBENZOIC ACID", "non-polymer"),
        "SX2": ("4-[(5-BROMOPYRIDIN-2-YL)AMINO]-4-OXOBUTANOIC ACID", "non-polymer"),
        "IPA": ("ISOPROPYL ALCOHOL", "non-polymer"),
        "SX4": (
            "4-BROMO-2-{[(2R)-2-(2-CHLOROBENZYL)PYRROLIDIN-1-YL]CARBONYL}ANILINE",
            "non-polymer",
        ),
        "PL9": (
            "2,3-DIMETHYL-5-(3,7,11,15,19,23,27,31,35-NONAMETHYL-2,6,10,14,18,22,26,30,34-HEXATRIACONTANONAENYL-2,5-CYCLOHEXADIENE-1,4-DIONE-2,3-DIMETHYL-5-SOLANESYL-1,4-BENZOQUINONE",
            "non-polymer",
        ),
        "SX6": (
            "N-(4-BROMO-2-{[(3R,5S)-3,5-DIMETHYLPIPERIDIN-1-YL]CARBONYL}PHENYL)-4-MORPHOLIN-4-YL-4-OXOBUTANAMIDE",
            "non-polymer",
        ),
        "SX7": (
            "2-AMINO-5-[3-(1-ETHYL-1H-PYRAZOL-5-YL)-1H-PYRROLO[2,3-B]PYRIDIN-5-YL]-N,N-DIMETHYLBENZAMIDE",
            "non-polymer",
        ),
        "SX8": (
            "6-{[6-(1-METHYL-1H-PYRAZOL-4-YL)[1,2,4]TRIAZOLO[4,3-B]PYRIDAZIN-3-YL]SULFANYL}QUINOLINE",
            "non-polymer",
        ),
        "IPF": (
            "N-ACETYLTRYPTOPHYL-N~1~-{3-[1-(N-ACETYLTRYPTOPHYLVALYL)PYRROLIDIN-2-YL]-1-BENZYL-2,3-DIHYDROXYPROPYL}VALINAMIDE",
            "non-polymer",
        ),
        "YVN": (
            "N-({(1R,2R,3S)-2-(METHYLCARBAMOYL)-3-[4-(PHOSPHONOOXY)PHENYL]CYCLOPROPYL}CARBONYL)-L-VALYL-L-ASPARTAMIDE",
            "non-polymer",
        ),
        "3KP": ("1,2,3-TRICHLOROPROPANE", "non-polymer"),
        "S5B": (
            "4-[(2-TERT-BUTYLPHENYL)AMINO]-7-(3,5-DIMETHYL-1,2-OXAZOL-4-YL)QUINOLINE-3-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "IPE": ("3-METHYLBUT-3-ENYL TRIHYDROGEN DIPHOSPHATE", "non-polymer"),
        "S5H": ("5-ALPHA-PREGNANE-3-BETA-OL-HEMISUCCINATE", "non-polymer"),
        "P0P": (
            "[4-(DIHYDROXYMETHYL)-5-HYDROXY-6-METHYLPYRIDIN-3-YL]METHYL DIHYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "S5S": ("2-SULFANYL-2LAMBDA~4~-TETRASULFANE", "non-polymer"),
        "IPK": (
            "2-({4-[(4-BENZYLPIPERIDIN-1-YL)CARBONYL]BENZYL}SULFANYL)-3H-IMIDAZO[4,5-C]PYRIDINE",
            "non-polymer",
        ),
        "MPO": ("3[N-MORPHOLINO]PROPANE SULFONIC ACID", "non-polymer"),
        "PYR": ("PYRUVIC ACID", "non-polymer"),
        "Y2F": (
            "6-{[(3R,4R)-4-(2-{[(2S)-1-(3-FLUOROPHENYL)PROPAN-2-YL]AMINO}ETHOXY)PYRROLIDIN-3-YL]METHYL}-4-METHYLPYRIDIN-2-AMINE",
            "non-polymer",
        ),
        "PYS": ("2-PYRIDINETHIOL", "non-polymer"),
        "WCX": (
            "4-{5-[(1-METHYLPIPERIDIN-4-YL)OXY]-1H-BENZIMIDAZOL-1-YL}-2-{(1R)-1-[2-(TRIFLUOROMETHYL)PHENYL]ETHOXY}BENZAMIDE",
            "non-polymer",
        ),
        "IPN": ("2-AMINOETHYLGLYCINE-CARBONYLMETHYLENE-5-IODOURACIL", "nucleic-acid"),
        "RI5": (
            "(1AR,2AR,3S,6R,6AS,8AS,8BR,9R)-2A-HYDROXY-8B-METHYL-9-(PROP-1-EN-2-YL)HEXAHYDRO-3,6-METHANO-1,5,7-TRIOXACYCLOPENTA[IJ]CYCLOPROPA[A]AZULENE-4,8(3H)-DIONE",
            "non-polymer",
        ),
        "IPO": ("PARA-IODO-D-PHENYLALANINE HYDROXAMIC ACID", "non-polymer"),
        "HT7": ("(3S)-3-AMINO-4-(1H-INDOL-3-YL)BUTANOIC ACID", "peptide"),
        "HT5": (
            "(4S)-4-(DECANOYLAMINO)-5-HYDROXY-3,4-DIHYDRO-2H-THIOPHENIUM",
            "non-polymer",
        ),
        "KLG": ("N-HYDROXY-N~2~-[(2-PHENYLETHYL)SULFONYL]GLYCINAMIDE", "non-polymer"),
        "HT3": ("(2R,3S)-HEPTANE-1,2,3-TRIOL", "non-polymer"),
        "HT2": (
            "5-AMIDINO-2-[2-(4-AMIDINOPHENYL)-5-BENZIMIDAZOLYL]BENZIMIDAZOLE",
            "non-polymer",
        ),
        "HT1": (
            "2'-(4-ETHOXYPHENYL)-5-(4-METHYL-1-PIPERAZINYL)-2,5'-BI-BENZIMIDAZOLE",
            "non-polymer",
        ),
        "LEA": ("PENTANOIC ACID", "non-polymer"),
        "SC6": (
            "N-{(1S,2R)-1-(3,5-DIFLUOROBENZYL)-2-HYDROXY-2-[(2R,4R)-4-PHENOXYPYRROLIDIN-2-YL]ETHYL}-3-{[(2R)-2-(METHOXYMETHYL)PYRROLIDIN-1-YL]CARBONYL}-5-METHYLBENZAMIDE",
            "non-polymer",
        ),
        "A47": ("?", "nucleic-acid"),
        "IPM": ("3-ISOPROPYLMALIC ACID", "non-polymer"),
        "RHS": (
            "3-(CARBOXYAMIDE(2-CARBOXYAMIDE-2-TERTBUTYLETHYL))PENTAN",
            "non-polymer",
        ),
        "PYH": (
            "N~6~-{[(2R,3R)-3-METHYL-3,4-DIHYDRO-2H-PYRROL-2-YL]CARBONYL}-L-LYSINE",
            "peptide",
        ),
        "PYI": (
            "1-[(4-AMINO-2-METHYLPYRIMIDIN-5-YL)METHYL]-3-(2-{[HYDROXY(PHOSPHONOOXY)PHOSPHORYL]OXY}ETHYL)-2-METHYLPYRIDINIUM",
            "non-polymer",
        ),
        "H9K": (
            "3-(4'-HYDROXYBIPHENYL-4-YL)-2,4-DIHYDROINDENO[1,2-C]PYRAZOL-6-OL",
            "non-polymer",
        ),
        "DVP": (
            "METHYL 1-(4-{[(2,4-DIAMINOPTERIDIN-6-YL)METHYL]AMINO}BENZOYL)PIPERIDINE-4-CARBOXYLATE",
            "non-polymer",
        ),
        "IBI": (
            "N-{TRANS-4-[4-(CYCLOPROPYLMETHYL)PIPERAZIN-1-YL]CYCLOHEXYL}-4-{[(7R)-7-ETHYL-5-METHYL-8-(1-METHYLETHYL)-6-OXO-5,6,7,8-TETRAHYDROPTERIDIN-2-YL]AMINO}-3-METHOXYBENZAMIDE",
            "non-polymer",
        ),
        "TR8": ("?", "non-polymer"),
        "TR1": (
            "2-(2-{2-[(BIPHENYL-4-YLMETHYL)-AMINO]-3-MERCAPTO-PENTANOYLAMINO}-ACETYLAMINO)-3-METHYL-BUTYRIC ACID METHYL ESTER",
            "non-polymer",
        ),
        "PYL": ("PHENYLETHANE", "non-polymer"),
        "TR3": (
            "N-[3-(METHYLSULFANYL)-1-PHENYL-1H-INDAZOL-6-YL]PIPERIDINE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "M9F": (
            "(2S)-2-AMINO-6-[[(2R)-1-[[(2S)-1-HYDROXY-3-METHYLSULFANYL-1-OXO-BUTAN-2-YL]AMINO]-1-OXO-3-SULFANYL-PROPAN-2-YL]AMINO]-6-OXO-HEXANOIC ACID",
            "non-polymer",
        ),
        "TR5": (
            "5-METHYL-1-(5-O-PHOSPHONO-BETA-D-RIBOFURANOSYL)-1,5-DIHYDRO-1,4,5,6,8-PENTAAZAACENAPHTHYLEN-3-AMINE",
            "non-polymer",
        ),
        "TR4": (
            "5-AMINO-3-(METHYLSULFANYL)-1-PHENYL-1H-PYRAZOLE-4-CARBONITRILE",
            "non-polymer",
        ),
        "TR7": ("METHYL L-TRYPTOPHANATE", "non-polymer"),
        "PYM": ("PYRUVAMIDE", "non-polymer"),
        "TRI": ("1,2,4-TRIAZOLE", "non-polymer"),
        "TRH": ("2'-DEOXY-THYMIDINE-BETA-L-RHAMNOSE", "non-polymer"),
        "TRK": ("TIRANDAMYCIN D", "non-polymer"),
        "TRJ": ("META-DI(AMINOMETHYL)BENZENE", "non-polymer"),
        "TRM": ("1H-BENOXIMIDAZOLE-2-CARBOXYLIC ACID", "non-polymer"),
        "PYN": (
            "3-PYRROLIDIN-1-YL-N-[6-(3-PYRROLIDIN-1-YL-PROPIONYLAMINO)-ACRIDIN-3-YL]-PROPIONAMIDE",
            "non-polymer",
        ),
        "TRO": ("2-HYDROXY-TRYPTOPHAN", "peptide"),
        "TRN": ("NZ2-TRYPTOPHAN", "peptide"),
        "TRA": ("ACONITATE ION", "non-polymer"),
        "TRC": ("TRICARBALLYLIC ACID", "non-polymer"),
        "PYO": (
            "1-(BETA-D-RIBOFURANOSYL)-PYRIMIDIN-2-ONE-5'-PHOSPHATE",
            "nucleic-acid",
        ),
        "TRE": ("TREHALOSE", "saccharide"),
        "TRD": ("TRIDECANE", "non-polymer"),
        "TRG": ("L-(N,N DIMETHYL)LYSINE", "peptide"),
        "KLI": (
            "3-(5-BENZYL-1,2,4-OXADIAZOL-3-YL)-6-FLUORO-1-[2-FLUORO-4-(TRIFLUOROMETHYL)BENZYL]-7-(4-METHYLPIPERAZIN-1-YL)QUINOLIN-4(1H)-ONE",
            "non-polymer",
        ),
        "H90": (
            "6-(4-BENZOYLPHENYL)-1-(PYRIDIN-4-YLMETHYL)-1H-PYRAZOLO[3,4-B]PYRIDINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "KBG": ("2-KETO-BETA-D-GLUCOSE", "saccharide"),
        "W33": (
            "5-(5-(6-CHLORO-4-(4,5-DIHYDRO-2-OXAZOLYL)PHENOXY)PENTYL)-3-METHYL ISOXAZOLE",
            "non-polymer",
        ),
        "TRZ": ("TRAZEOLIDE", "non-polymer"),
        "QV8": (
            "9,10-DIOXO-4-(PHENYLAMINO)-9,10-DIHYDROANTHRACENE-2-SULFONIC ACID",
            "non-polymer",
        ),
        "KLJ": (
            "N-HYDROXY-N~2~-{[2-(4-METHOXYPHENYL)ETHYL]SULFONYL}GLYCINAMIDE",
            "non-polymer",
        ),
        "W37": (
            "N-(6-{[1-(4-BROMOPHENYL)ISOQUINOLIN-6-YL]OXY}HEXYL)-N-METHYLPROP-2-EN-1-AMINE",
            "non-polymer",
        ),
        "H97": (
            "(1R,3R,7E,17BETA)-17-[(1R)-6,6,6-TRIFLUORO-5-HYDROXY-1-(4-HYDROXY-4-METHYLPENTYL)-5-(TRIFLUOROMETHYL)HEX-3-YN-1-YL]-9,10-SECOESTRA-5,7-DIENE-1,3-DIOL",
            "non-polymer",
        ),
        "QV4": (
            "4,6-DIDEOXY-4-{[(1S,4R,5R,6S)-4-{[ALPHA-D-GLUCOPYRANOSYL-(1->4)-ALPHA-D-GLUCOPYRANOSYL-(1->4)-ALPHA-D-GLUCOPYRANOSYL]OXY}-5,6-DIHYDROXY-3-(HYDROXYMETHYL)CYCLOHEX-2-EN-1-YL]AMINO}-ALPHA-D-GLUCOPYRANOSE",
            "saccharide",
        ),
        "TRP": ("TRYPTOPHAN", "peptide"),
        "QV6": (
            "3-({4-[(2,4-DIMETHYL-5-SULFOPHENYL)AMINO]-9,10-DIOXO-9,10-DIHYDROANTHRACEN-1-YL}AMINO)-2,4,6-TRIMETHYLBENZENESULFONIC ACID",
            "non-polymer",
        ),
        "QV7": (
            "3-HYDROXY-4-[(E)-{2-SULFO-4-[(E)-(4-SULFOPHENYL)DIAZENYL]PHENYL}DIAZENYL]NAPHTHALENE-2,7-DISULFONIC ACID",
            "non-polymer",
        ),
        "GD1": (
            "2-AMINO-5-[(Z)-IMINOMETHYL]-3,7-DIHYDRO-4H-PYRROLO[2,3-D]PYRIMIDIN-4-ONE",
            "non-polymer",
        ),
        "PYB": ("4-AMINO-(1-METHYLPYRROLE)-2-CARBOXYLIC ACID", "non-polymer"),
        "TRW": (
            "TRW3-(2-AMINO-3-HYDROXY-PROPYL)-6-(N'-CYCLOHEXYL-HYDRAZINO)OCTAHYDRO-INDOL-7-OL",
            "peptide",
        ),
        "HTG": ("HEPTYL 1-THIOHEXOPYRANOSIDE", "non-polymer"),
        "HTF": ("N-(2-OXOTETRAHYDROFURAN-3-YL)OCTANAMIDE", "non-polymer"),
        "HTD": (
            "2-[(9AR)-2,7-DIMETHYL-9A,10-DIHYDRO-5H-PYRIMIDO[4,5-D][1,3]THIAZOLO[3,2-A]PYRIMIDIN-8-YL]ETHYL TRIHYDROGEN DIPHOSPHATE",
            "non-polymer",
        ),
        "HTC": (
            "3-{[(4-CARBOXY-2-HYDROXYANILINE]SULFONYL}THIOPHENE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "HTA": (
            "N-[3-(N'-HYDROXYCARBOXAMIDO)-2-(2-METHYLPROPYL)-PROPANOYL]-O-TYROSINE-N-METHYLAMIDE",
            "non-polymer",
        ),
        "HTO": ("HEPTANE-1,2,3-TRIOL", "non-polymer"),
        "HTN": ("(3R)-3-HYDROXY-N-METHYL-D-ASPARAGINE", "peptide"),
        "HTM": (
            "6-(3-HYDROXYPROPOXY)-5-ACETAMIDO-5,6-DIHYDRO-4-HYDROXY-4H-PYRAN-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "HTL": ("2-ACETYL-THIAMINE DIPHOSPHATE", "non-polymer"),
        "HBX": ("BENZALDEHYDE", "non-polymer"),
        "HTJ": (
            "6-PHENYL-1-(PYRIDIN-4-YLMETHYL)-1H-PYRAZOLO[3,4-B]PYRIDINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "HTI": ("(4S)-4-{[(2S)-2-AMINO-3-OXOPROPYL]SULFANYL}-L-HOMOSERINE", "peptide"),
        "HTH": ("(2S,3R)-HEPTANE-1,2,3-TRIOL", "non-polymer"),
        "ESD": ("(2-AMINO-ETHYLSULFANYL)-ACETIC ACID", "non-polymer"),
        "HTS": ("2-MERCAPTOPHENOL", "non-polymer"),
        "HTR": ("BETA-HYDROXYTRYPTOPHANE", "peptide"),
        "HTQ": ("HOMOTROPINE", "non-polymer"),
        "HTP": (
            "4,5,6-TRIHYDROXY-7-HYDROXYMETHYL-4,5,6,7-TETRAHYDRO-1H-[1,2,3]TRIAZOLO[1,5-A]PYRIDIN-8-YLIUM",
            "non-polymer",
        ),
        "HTY": ("(BETAR)-BETA-HYDROXY-D-TYROSINE", "peptide"),
        "HTX": ("HEPTAN-2-ONE", "non-polymer"),
        "FB2": ("BENZENESULFONAMIDE", "non-polymer"),
        "FB1": ("2-CHLOROBENZENESULFONAMIDE", "non-polymer"),
        "FB0": ("?", "non-polymer"),
        "FB6": ("6-(DIHYDROXYBORANYL)-2-METHYL-L-NORLEUCINE", "peptide"),
        "FB5": ("2-(DIFLUOROMETHYL)-6-(DIHYDROXYBORANYL)-L-NORLEUCINE", "peptide"),
        "FB4": ("2-(2-AMINO-4-CHLOROPHENOXY)-5-CHLOROPHENOL", "non-polymer"),
        "O96": (
            "N-[(1S)-3-AMINO-1-ETHYL-2,3-DIOXOPROPYL]-2-CHLORO-4-(PYRIDIN-2-YLMETHOXY)-3-(TRIFLUOROMETHYL)BENZAMIDE",
            "non-polymer",
        ),
        "O92": (
            "3-AMINO-6-{4-CHLORO-3-[(2,3-DIFLUOROPHENYL)SULFAMOYL]PHENYL}-N-METHYLPYRAZINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "MPL": ("N-METHYL-PYRIDOXAL-5'-PHOSPHATE", "non-polymer"),
        "LE1": ("3-SULFANYL-L-VALINE", "peptide"),
        "LE2": ("N-(2-PHENYLETHYL)THIOFORMAMIDE", "non-polymer"),
        "LE3": ("N-{4-[(R)-METHYLSULFINYL]BUTYL}THIOFORMAMIDE", "non-polymer"),
        "N68": ("6-AMINOHEXYL DIHYDROGEN PHOSPHATE", "non-polymer"),
        "F7L": (
            "(2S)-1-[(2-FLUOROBENZYL)OXY]-3-(PYRROLIDIN-1-YL)PROPAN-2-OL",
            "non-polymer",
        ),
        "75V": (
            "N-{[(2S,3S)-3-(ETHOXYCARBONYL)OXIRAN-2-YL]CARBONYL}-L-ISOLEUCYL-L-ALANINE",
            "non-polymer",
        ),
        "IBM": ("3-ISOBUTYL-1-METHYLXANTHINE", "non-polymer"),
        "ODR": (
            "(1R)-1-[3-(3,5-DIMETHYL-1,2-OXAZOL-4-YL)-5-ETHOXYPHENYL]ETHANOL",
            "non-polymer",
        ),
        "NGC": (
            "3,5-DIDEOXY-5-[(HYDROXYACETYL)AMINO]-D-GLYCERO-ALPHA-D-GALACTO-NON-2-ULOPYRANOSONIC ACID",
            "saccharide",
        ),
        "ODT": ("(11E,13E,15Z)-OCTADECA-11,13,15-TRIENOIC ACID", "non-polymer"),
        "F7A": ("N-(3-PHENOXYPHENYL)-2-(TRIFLUOROMETHYL)BENZAMIDE", "non-polymer"),
        "ODI": ("OCTANE-1,8-DIOL", "non-polymer"),
        "75A": (
            "N-ACETYL-4-[DIFLUORO(PHOSPHONO)METHYL]-L-PHENYLALANYL-N~5~-(3-IODOBENZOYL)-L-ORNITHINAMIDE",
            "peptide",
        ),
        "F91": (
            "N-(PYRIDIN-4-YLMETHYL)-2,3-DIHYDRO-1,4-BENZODIOXIN-6-AMINE",
            "non-polymer",
        ),
        "TB0": ("TERT-BUTYL ISOCYANIDE", "non-polymer"),
        "EL3": (
            "(3S,3AR,6S)-3,7,7,8-TETRAMETHYL-2,3,4,5,6,7-HEXAHYDRO-1H-3A,6-METHANOAZULENE",
            "non-polymer",
        ),
        "ODD": ("(10E,12Z)-OCTADECA-10,12-DIENOIC ACID", "non-polymer"),
        "ODE": (
            "DIETHYL (1R,2S,3R,4S)-5,6-BIS(4-HYDROXYPHENYL)-7-OXABICYCLO[2.2.1]HEPT-5-ENE-2,3-DICARBOXYLATE",
            "non-polymer",
        ),
        "RDT": (
            "(1R)-6,7-DIMETHOXY-1-METHYL-3,4-DIHYDROISOQUINOLINE-2(1H)-SULFONAMIDE",
            "non-polymer",
        ),
        "751": (
            "2-{4-[(4'-METHOXYBIPHENYL-3-YL)SULFONYL]PIPERAZIN-1-YL}-3-(4-METHOXYPHENYL)PYRAZINE",
            "non-polymer",
        ),
        "750": (
            "5-(2-MORPHOLIN-4-YLETHOXY)BENZOFURAN-2-CARBOXYLIC ACID ((S)-3-METHYL-1-{(S)-3-OXO-1-[2-(3-PYRIDIN-2-YLPHENYL)ACETYL]AZEPAN-4-YLCARBAMOYL}BUTYL)AMIDE",
            "non-polymer",
        ),
        "757": ("?", "non-polymer"),
        "F94": ("7-HYDROXY-2-(3-HYDROXYPHENYL)-4H-CHROMEN-4-ONE", "non-polymer"),
        "IBN": ("2-METHYLPROPAN-1-AMINE", "non-polymer"),
        "754": (
            "(2E,4E,6Z)-3-METHYL-7-(5,5,8,8-TETRAMETHYL-3-PROPOXY-5,6,7,8-TETRAHYDRONAPHTHALEN-2-YL)OCTA-2,4,6-TRIENOIC ACID",
            "non-polymer",
        ),
        "I4B": ("ISOBUTYLBENZENE", "non-polymer"),
        "ELA": ("9-OCTADECENOIC ACID", "non-polymer"),
        "A41": ("5-METHYL-2-[(PHENYLSULFONYL)AMINO]BENZOIC ACID", "non-polymer"),
        "I4C": ("D-MYO-INOSITOL-4-PHOSPHATE", "non-polymer"),
        "RDE": (
            "2-(3-AMINO-2,5,6-TRIMETHOXYPHENYL)ETHYL 5-CHLORO-2,4-DIHYDROXYBENZOATE",
            "non-polymer",
        ),
        "RDD": ("1-BETA-D-RIBOFURANOSYL-1,3,5-TRIAZINANE-2,4,6-TRIONE", "non-polymer"),
        "ELY": ("N~6~,N~6~-DIETHYL-L-LYSINE", "peptide"),
        "F78": (
            "3-HYDROXY-N-[(2S,3R)-3-HYDROXY-4-{[(4-METHOXYPHENYL)SULFONYL](3-PHENYLPROPYL)AMINO}-1-PHENYLBUTAN-2-YL]BENZAMIDE",
            "non-polymer",
        ),
        "F79": (
            "2-(2-AMINOETHOXY)-3-ETHYL-6-{[(4-FLUOROPHENYL)SULFONYL]AMINO}BENZOIC ACID",
            "non-polymer",
        ),
        "RDC": ("RADICICOL", "non-polymer"),
        "RDL": ("6,7-DIOXO-5H-8-RIBITYLAMINOLUMAZINE", "non-polymer"),
        "ELP": (
            "(1R)-1-[((1E)-{3-HYDROXY-2-METHYL-5-[(PHOSPHONOOXY)METHYL]PYRIDIN-4-YL}METHYLENE)AMINO]-1-PHENYLETHYLPHOSPHONIC ACID",
            "non-polymer",
        ),
        "F77": (
            "3-ETHYL-6-{[(4-FLUOROPHENYL)SULFONYL]AMINO}-2-METHYLBENZOIC ACID",
            "non-polymer",
        ),
        "ELV": (
            "6-(3-CHLORO-2-FLUOROBENZYL)-1-[(1S)-1-(HYDROXYMETHYL)-2-METHYLPROPYL]-7-METHOXY-4-OXO-1,4-DIHYDROQUINOLINE-3-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "F71": (
            "3-HYDROXY-N-[(2S,3R)-3-HYDROXY-4-([(2S)-2-METHYLBUTYL]{[5-(1,2-OXAZOL-5-YL)THIOPHEN-2-YL]SULFONYL}AMINO)-1-PHENYLBUTAN-2-YL]BENZAMIDE",
            "non-polymer",
        ),
        "F72": (
            "N-{(2S,3R)-4-[(1,3-BENZOTHIAZOL-6-YLSULFONYL)(CYCLOHEXYLMETHYL)AMINO]-3-HYDROXY-1-PHENYLBUTAN-2-YL}-3-HYDROXYBENZAMIDE",
            "non-polymer",
        ),
        "X95": (
            "(S)-1-N2-(1-CARBOXY-3-PHENYLPROPYL)-L-LYSYL-L-TRYPTOPHAN",
            "non-polymer",
        ),
        "FBS": ("4-FLOUROBENZENESULFONAMIDE", "non-polymer"),
        "FBR": (
            "(9AS)-4-BROMO-9A-BUTYL-7-HYDROXY-1,2,9,9A-TETRAHYDRO-3H-FLUOREN-3-ONE",
            "non-polymer",
        ),
        "FBQ": (
            "1-[3-({[(4-AMINO-5-FLUORO-2-METHYLQUINOLIN-3-YL)METHYL]THIO}METHYL)PHENYL]-2,2,2-TRIFLUOROETHANE-1,1-DIOL",
            "non-polymer",
        ),
        "FBP": ("BETA-FRUCTOSE-1,6-DIPHOSPHATE", "saccharide"),
        "FBW": ("3-FLUOROBENZENESULFONAMIDE", "non-polymer"),
        "FBV": ("2-FLUOROBENZENESULFONAMIDE", "non-polymer"),
        "FBU": ("3,5-DIFLUOROBENZENESULFONAMIDE", "non-polymer"),
        "FBT": ("2,6-DIFLUOROBENZENESULFONAMIDE", "non-polymer"),
        "LEH": ("N-[12-(1H-IMIDAZOL-1-YL)DODECANOYL]-L-LEUCINE", "peptide"),
        "FBZ": ("(3S)-3-PHENYLBUTANOIC ACID", "non-polymer"),
        "LEM": ("(3S)-3-AMINO-5-METHYLHEXAN-2-ONE", "peptide"),
        "LEN": ("3-METHYLBUTAN-1-AMINE", "peptide"),
        "LEO": ("D-LIMONENE 1,2-EPOXIDE", "non-polymer"),
        "FBC": ("?", "non-polymer"),
        "FBB": ("6-FLUORO-1,3-BENZOTHIAZOL-2-AMINE", "non-polymer"),
        "FBA": ("4-FLUOROBENZYLAMINE", "non-polymer"),
        "FBG": ("6-FLUORO-4H-1,3-BENZODIOXINE-8-CARBOXYLIC ACID", "non-polymer"),
        "FBF": ("4-(TRIFLUOROMETHYL)BENZENECARBOXIMIDAMIDE", "non-polymer"),
        "FBE": ("(3E)-5-AMINOPENT-3-ENOIC ACID", "peptide"),
        "FBD": (
            "N-({(1R)-1-CARBOXY-2-[(4-FLUOROBENZYL)SULFANYL]ETHYL}CARBAMOYL)-L-GLUTAMIC ACID",
            "non-polymer",
        ),
        "LEX": ("(3S)-3-AMINO-1-HYDROXY-5-METHYLHEXAN-2-ONE", "peptide"),
        "MG7": ("7-METHYLGUANOSINE", "non-polymer"),
        "FBI": (
            "7-[4-(4-FLUORO-PHENYL)-6-ISOPROPYL-2-(METHANESULFONYL-METHYL-AMINO)-PYRIMIDIN-5-YL] -3,5-DIHYDROXY-HEPTANOIC ACID",
            "non-polymer",
        ),
        "FBO": (
            "L-PHENYLALANYL-N6-[(BENZYLOXY)CARBONYL]-N1-HYDROXY-L-LYSINAMIDE",
            "non-polymer",
        ),
        "FBN": (
            "5'-DEOXY-5'-[4-({[(2,3-DIHYDROXY-5-NITROPHENYL)CARBONYL]AMINO}METHYL)-1H-1,2,3-TRIAZOL-1-YL]ADENOSINE",
            "non-polymer",
        ),
        "FBL": (
            "(2S)-1-[4-({4-[(2,5-DICHLOROPHENYL)AMINO]PYRIMIDIN-2-YL}AMINO)PHENOXY]-3-(DIMETHYLAMINO)PROPAN-2-OL",
            "non-polymer",
        ),
        "V11": (
            "[(1S,5R)-8-[(2R)-2-HYDROXY-2-PHENYL-ETHYL]-8-AZABICYCLO[3.2.1]OCTAN-3-YL] BENZOATE",
            "non-polymer",
        ),
        "OBN": ("OUABAIN", "non-polymer"),
        "MG5": (
            "2,3,4-TRI-O-PROPANOYL-6-O-SULFAMOYL-BETA-D-GLUCOPYRANOSE",
            "saccharide",
        ),
        "DX9": (
            "(2S)-3-(7-CARBAMIMIDOYLNAPHTHALEN-2-YL)-2-[4-({(3R)-1-[(1Z)-ETHANIMIDOYL]PYRROLIDIN-3-YL}OXY)PHENYL]PROPANOIC ACID",
            "non-polymer",
        ),
        "DX8": (
            "2-AMINO-6-(4-METHOXYPHENYL)-4-OXO-4,7-DIHYDRO-3H-PYRROLO[2,3-D]PYRIMIDINE-5-CARBONITRILE",
            "non-polymer",
        ),
        "DX7": (
            "2-AMINO-4-OXO-6-PHENYL-4,7-DIHYDRO-3H-PYRROLO[2,3-D]PYRIMIDINE-5-CARBONITRILE",
            "non-polymer",
        ),
        "DX6": (
            "2-AMINO-5-(2-PHENYLETHYL)-3,7-DIHYDRO-4H-PYRROLO[2,3-D]PYRIMIDIN-4-ONE",
            "non-polymer",
        ),
        "DX5": ("D-XYLITOL-5-PHOSPHATE", "non-polymer"),
        "DX4": ("2-AMINO-1,9-DIHYDRO-6H-PURINE-6-THIONE", "non-polymer"),
        "DX3": ("6,7,7-TRIMETHYL-7,8-DIHYDROPTERIDINE-2,4-DIAMINE", "non-polymer"),
        "DX2": ("6-PHENYLPTERIDINE-2,4,7-TRIAMINE", "non-polymer"),
        "DX1": ("6,7-BIS(1-METHYLETHYL)PTERIDINE-2,4-DIAMINE", "non-polymer"),
        "PF1": (
            "N-[(4-AMINO-2-METHYLPYRIMIDIN-5-YL)METHYL]-N-(2-HYDROXYETHYL)FORMAMIDE",
            "non-polymer",
        ),
        "PF0": ("3-HYDROXY-2-METHYLBENZOIC ACID", "non-polymer"),
        "PF3": (
            "(2R)-3-({[4-[(PYRIDIN-4-YL)PHENYL]-THIEN-2-YL}CARBOXAMIDO)(PHENYL)PROPANOIC ACID",
            "non-polymer",
        ),
        "PF2": (
            "2-(4-{(3S,5S)-5-[(3,3-DIFLUOROPYRROLIDIN-1-YL)CARBONYL]PYRROLIDIN-3-YL}PIPERAZIN-1-YL)PYRIMIDINE",
            "non-polymer",
        ),
        "PF5": ("2,3,4,5,6-PENTAFLUORO-L-PHENYLALANINE", "peptide"),
        "PF4": ("4,5-BIS(4-METHOXYPHENYL)-2-THIOPHEN-2-YL-1H-IMIDAZOLE", "non-polymer"),
        "PF7": ("4-(QUINOLIN-3-YLMETHYL)PIPERIDINE-1-CARBOXYLIC ACID", "non-polymer"),
        "PF6": (
            "2-({4-[4-(PYRIDIN-4-YLMETHYL)-1H-PYRAZOL-3-YL]PHENOXY}METHYL)QUINOLINE",
            "non-polymer",
        ),
        "NLA": ("NAPHTHALEN-1-YL-ACETIC ACID", "non-polymer"),
        "N6S": (
            "(2S,4AS,6R,7R,7AR)-6-(6-{[(1S)-1-METHYL-2-PHENYLETHYL]AMINO}-9H-PURIN-9-YL)TETRAHYDRO-4H-FURO[3,2-D][1,3,2]DIOXAPHOSPHININE-2,7-DIOL 2-OXIDE",
            "non-polymer",
        ),
        "CN6": (
            "(2R,5R,11S,14R)-2-(BUTANOYLOXY)-5,8,11-TRIHYDROXY-5,11-DIOXIDO-16-OXO-14-[(PROPANOYLOXY)METHYL]-4,6,10,12,15-PENTAOXA-5,11-DIPHOSPHANONADEC-1-YL UNDECANOATE",
            "non-polymer",
        ),
        "SGX": ("O-[(S)-AMINO(METHOXY)PHOSPHORYL]-L-SERINE", "non-polymer"),
        "CN5": (
            "(5S,11R)-5,8,11-TRIHYDROXY-5,11-DIOXIDO-17-OXO-4,6,10,12,16-PENTAOXA-5,11-DIPHOSPHAOCTADEC-1-YL PENTADECANOATE",
            "non-polymer",
        ),
        "CN2": (
            "2-MERCAPTO-N-[1,2,3,10-TETRAMETHOXY-9-OXO-5,6,7,9-TETRAHYDRO-BENZO[A]HEPTALEN-7-YL]ACETAMIDE",
            "non-polymer",
        ),
        "CN3": (
            "(2R,5S,11R,14R)-5,8,11-TRIHYDROXY-2-(NONANOYLOXY)-5,11-DIOXIDO-16-OXO-14-[(PROPANOYLOXY)METHYL]-4,6,10,12,15-PENTAOXA-5,11-DIPHOSPHANONADEC-1-YL UNDECANOATE",
            "non-polymer",
        ),
        "CN1": ("OXO-IRON CLUSTER 2", "non-polymer"),
        "SGP": ("GUANOSINE-2',3'-CYCLOPHOSPHOROTHIOATE", "non-polymer"),
        "SGS": (
            "(E)-(2S,3R,4S,5S,6R)-TETRAHYDRO-3,4,5-TRIHYDROXY-6-(HYDROXYMETHYL)-2H-PYRAN-2-YL 3-(4-HYDROXY-3,5-DIMETHOXYPHENYL)ACRYLATE",
            "saccharide",
        ),
        "SGR": ("O-[(R)-AMINO(METHOXY)PHOSPHORYL]-L-SERINE", "non-polymer"),
        "TMP": ("THYMIDINE-5'-PHOSPHATE", "non-polymer"),
        "SGV": ("SANGIVAMYCIN", "non-polymer"),
        "SGI": ("3-(4-HYDROXYPHENYL)PROPANAMIDE", "non-polymer"),
        "IC8": (
            "2-[(6-AMINO-9H-PURIN-9-YL)METHYL]-5-METHYL-3-(2-METHYLPHENYL)QUINAZOLIN-4(3H)-ONE",
            "non-polymer",
        ),
        "SGM": ("MONOTHIOGLYCEROL", "non-polymer"),
        "NGZ": ("2-(ACETYLAMINO)-2-DEOXY-ALPHA-L-GLUCOPYRANOSE", "non-polymer"),
        "OBG": ("6-(BENZYLOXY)-9H-PURIN-2-AMINE", "non-polymer"),
        "IC3": (
            "[2-(5-AMINO-4-CYANO-1H-PYRAZOL-1-YL)-5-(TRIFLUOROMETHYL)PHENYL](HYDROXY)OXOAMMONIUM",
            "non-polymer",
        ),
        "IC1": ("3-[(2,4,6-TRIMETHOXY-PHENYL)-METHYLENE]-INDOLIN-2-ONE", "non-polymer"),
        "SGB": ("O-[(S)-METHYL(1-METHYLETHOXY)PHOSPHORYL]-L-SERINE", "peptide"),
        "SGE": (
            "13-METHYL-13,14-DIHYDRO[1,3]BENZODIOXOLO[5,6-C][1,3]DIOXOLO[4,5-I]PHENANTHRIDINE",
            "non-polymer",
        ),
        "IBB": (
            "5-(2-IMIDAZOLINYL)-2-[2-(4-HYDROXYPHENYL)-5-BENZIMIDAZOLYL]BENZIMIDAZOLE",
            "non-polymer",
        ),
        "SGF": (
            "(2S,3R,4E)-2-AMINO-3-HYDROXYOCTADEC-4-EN-1-YL 3-O-SULFO-BETA-D-GALACTOPYRANOSIDE",
            "non-polymer",
        ),
        "CNV": ("PROPANENITRILE", "non-polymer"),
        "CNW": ("7-DIETHYLAMINO-2-OXO-CHROMENE-3-CARBOXAMIDE", "non-polymer"),
        "CNT": ("N-METHYL-META-CHLORO-TYROSINE", "non-polymer"),
        "CNU": ("5-CYANO-URIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "CNR": ("(2S)-2-CHLOROPROPANOIC ACID", "non-polymer"),
        "CNS": ("HEXADECYL OCTANOATE", "non-polymer"),
        "CNP": ("2-PROPENYL-N-ACETYL-NEURAMIC ACID", "non-polymer"),
        "CNQ": ("3-(4-CHLOROPHENYL)QUINOXALINE-5-CARBOXAMIDE", "non-polymer"),
        "ICC": ("AZEPAN-2-ONE", "non-polymer"),
        "ICA": ("ISOCITRATE CALCIUM COMPLEX", "non-polymer"),
        "SG2": (
            "1-METHYL-3-OXO-1,3-DIHYDRO-BENZO[C]ISOTHIAZOLE-5-SULFONIC ACID AMIDE",
            "non-polymer",
        ),
        "SG5": (
            "2,3,4-TRI-O-PROPANOYL-6-O-SULFAMOYL-ALPHA-D-GLUCOPYRANOSE",
            "saccharide",
        ),
        "ICF": ("1-CHLORO-2,2,2-TRIFLUOROETHYL DIFLUOROMETHYL ETHER", "non-polymer"),
        "CNX": ("HEPTANENITRILE", "non-polymer"),
        "CNY": (
            "13,15-DIAMINO-2-(AMINOMETHYL)-3,4,9,12-TETRAHYDROXYHEXADECAHYDRO-2H-7,10-EPOXYPYRANO[2,3-B][1,10,4]BENZODIOXAZACYCLODODECIN-8-YL 2,6-DIAMINO-2,6-DIDEOXYHEXOPYRANOSIDE",
            "non-polymer",
        ),
        "CNF": ("OXO-IRON CLUSTER 3", "non-polymer"),
        "CND": (
            "5-BETA-D-RIBOFURANOSYLNICOTINAMIDE ADENINE DINUCLEOTIDE",
            "non-polymer",
        ),
        "CNE": (
            "(3S)-5-[(2-CHLORO-6-FLUOROBENZYL)SULFANYL]-3-{[N-({2-ETHOXY-5-[(1E)-3-METHOXY-3-OXOPROP-1-ENYL]PHENYL}ACETYL)-D-VALYL]AMINO}-4-OXOPENTANOIC ACID",
            "non-polymer",
        ),
        "CNB": ("OXO-IRON CLUSTER 1", "non-polymer"),
        "CNC": ("CO-CYANOCOBALAMIN", "non-polymer"),
        "CNA": ("CARBA-NICOTINAMIDE-ADENINE-DINUCLEOTIDE", "non-polymer"),
        "CNN": ("CYANAMIDE", "non-polymer"),
        "CNO": (
            "2-{4-[(3{2-[4-(1-CARBOXY-1-METHYL-ETHOXY)-PHENYL]-ACETYLAMINO}-PHENYLCARBAMOYL)-METHYL]-PHENOXY}-2-METHYL-PROPIONIC ACID",
            "non-polymer",
        ),
        "CNL": ("1,3,3-TRIMETHYL-2-OXABICYCLO[2.2.2]OCTANE", "non-polymer"),
        "CNM": ("ACETAMIDE", "non-polymer"),
        "HBP": ("1,7-HEPTYLENE-BIS-N,N'-SYN-2-PYRIDINIUMALDOXIME", "non-polymer"),
        "CNH": ("2-HYDROXY-2-METHYLPROPANENITRILE", "non-polymer"),
        "CNI": (
            "7-NITRO-2,3-DIOXO-2,3-DIHYDROQUINOXALINE-6-CARBONITRILE",
            "non-polymer",
        ),
        "PFA": (
            "[4-(4-HYDROXY-3-ISOPROPYL-PHENOXY)-3,5-DIMETHYL-PHENYL]-6-AZAURACIL",
            "non-polymer",
        ),
        "PFC": ("PHENYLFERRICROCIN-IRON", "non-polymer"),
        "PFB": ("2,3,4,5,6-PENTAFLUOROBENZYL ALCOHOL", "non-polymer"),
        "PFE": (
            "{4-[3-(6,7-DIETHOXY-QUINAZOLIN-4-YLAMINO)-PHENYL]-THIAZOL-2-YL}-METHANOL",
            "non-polymer",
        ),
        "PFD": (
            "5-(2-ETHOXYETHYL)-5-[4-(4-FLUOROPHENOXY)PHENOXY]PYRIMIDINE-2,4,6(1H,3H,5H)-TRIONE",
            "non-polymer",
        ),
        "PFG": ("10-PARPARGYL-5,8-DIDEAZAFOLATE-4-GLUTAMIC ACID", "non-polymer"),
        "DXX": ("METHYLMALONIC ACID", "non-polymer"),
        "DXW": (
            "6-{[(3S,4S)-4-{2-[(FURAN-2-YLMETHYL)AMINO]ETHOXY}PYRROLIDIN-3-YL]METHYL}-4-METHYLPYRIDIN-2-AMINE",
            "non-polymer",
        ),
        "PFH": (
            "6,7-DIMETHOXY-4-[(3R)-3-(2-NAPHTHYLOXY)PYRROLIDIN-1-YL]QUINAZOLINE",
            "non-polymer",
        ),
        "PFK": ("6-METHOXY-7-[2-(QUINOLIN-2-YL)ETHOXY]QUINAZOLINE", "non-polymer"),
        "DXT": (
            "(4S,4AR,5S,5AR,6R,12AS)-4-(DIMETHYLAMINO)-3,5,10,12,12A-PENTAHYDROXY-6-METHYL-1,11-DIOXO-1,4,4A,5,5A,6,11,12A-OCTAHYDROTETRACENE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "PFM": (
            "1-AMINO-7-(2-METHYL-3-OXIDO-5-((PHOSPHONOXY)METHYL)-4-PYRIDOXAL-5-OXO-6-HEPTENATE",
            "non-polymer",
        ),
        "DXR": (
            "3-(3-BROMOBENZYL)-1-TERT-BUTYL-1H-PYRAZOLO[3,4-D]PYRIMIDIN-4-AMINE",
            "non-polymer",
        ),
        "DXQ": (
            "7-[3-(DIMETHYLAMINO)PROPOXY]-6-METHOXY-2-(4-METHYL-1,4-DIAZEPAN-1-YL)-N-(1-METHYLPIPERIDIN-4-YL)QUINAZOLIN-4-AMINE",
            "non-polymer",
        ),
        "DXP": ("1-DEOXY-D-XYLULOSE-5-PHOSPHATE", "non-polymer"),
        "PFQ": (
            "2-[(5,6-DIPHENYLFURO[2,3-D]PYRIMIDIN-4-YL)AMINO]ETHANOL",
            "non-polymer",
        ),
        "DXN": (
            "(1R,3S,4R)-4-(PHOSPHOOXYMETHYL)-CYCLOPENTANE-1,3-DIOL",
            "nucleic-acid",
        ),
        "PFS": ("1-O-OCTADECYL-2-ACETYL-SN-GLYCEROL-3-PHOSPHOCHOLINE", "non-polymer"),
        "DXL": ("D-XYLARIC ACID", "non-polymer"),
        "PFU": (
            "(1S)-1,4-ANHYDRO-1-(5-CARBAMOYL-4-HYDROXY-1H-PYRAZOL-3-YL)-5-O-PHOSPHONO-D-RIBITOL",
            "non-polymer",
        ),
        "PFT": (
            "4-(1,3-DIHYDRO-2H-ISOINDOL-2-YLCARBONYL)BENZENE-1,3-DIOL",
            "non-polymer",
        ),
        "DXI": (
            "BETA-D-GLUCOPYRANOSYL-(1->4)-BETA-D-GALACTOPYRANOSYL-(1->4)-ALPHA-D-GLUCOPYRANOSE",
            "non-polymer",
        ),
        "RKE": ("(R)-KETAMINE", "non-polymer"),
        "DXG": ("4-DEOXYGLUCARATE", "non-polymer"),
        "DXF": (
            "(2R)-5-[(CARBAMOYLOXY)METHYL]-2-[(1R)-1-{[(2Z)-2-(FURAN-2-YL)-2-(METHOXYIMINO)ACETYL]AMINO}-2-OXOETHYL]-3,6-DIHYDRO-2H-1,3-THIAZINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "DXE": ("1,2-DIMETHOXYETHANE", "non-polymer"),
        "DXD": (
            "(1S,3S,4R)-4-(PHOSPHOOXYMETHYL)-CYCLOPENTANE-1,3-DIOL",
            "nucleic-acid",
        ),
        "DXC": (
            "(3ALPHA,5BETA,12ALPHA)-3,12-DIHYDROXYCHOLAN-24-OIC ACID",
            "non-polymer",
        ),
        "DXB": (
            "1,2-HYDRO-1-OXY-3,4-HYDRO-3-(1-METHOXY-2-OXY-3,4-DIHYDROXYPENTYL)-8,9-DIHYDROXY-7-METHYL-ANTHRACENE",
            "non-polymer",
        ),
        "DXA": (
            "1,2-HYDRO-1-OXY-3,4-HYDRO-3-(1-METHOXY-2-OXY-3,4-DIHYDROXYPENTYL)-8,9-DIHYDROXY-7-(SEC-BUTYL)-ANTHRACENE",
            "non-polymer",
        ),
        "3ZA": (
            "(4R)-4-{3-[HYDROXY(METHYL)AMINO]-3-OXOPROPYL}-L-GLUTAMIC ACID",
            "non-polymer",
        ),
        "UDH": ("6-AMINOHEXYL-URIDINE-C1,5'-DIPHOSPHATE", "non-polymer"),
        "AH0": (
            "2-(2-ACETYLAMINO-4-HYDROXY-6,8-DIOXA-BICYCLO[3.2.1]OCT-3-YLOXY)-PROPIONIC ACID",
            "non-polymer",
        ),
        "AH1": ("AHA001", "non-polymer"),
        "T76": (
            "(6R,21AS)-17-CHLORO-6-CYCLOHEXYL-2,3,6,7,10,11,19,20-OCTAHYDRO-1H,5H-PYRROLO[1,2-K][1,4,8,11,14]BENZOXATETRAAZA-CYCLOHEPTADECINE-5,8,12,21(9H,13H,21AH)-TETRONE",
            "non-polymer",
        ),
        "T77": (
            "6-{4-[2-(DIMETHYLAMINO)ETHOXY]PHENYL}-N-(1,3-DITHIOLAN-2-YLMETHYL)-5-PHENYL-7H-PYRROLO[2,3-D]PYRIMIDIN-4-AMINE",
            "non-polymer",
        ),
        "T74": (
            "N~3~-(2,6-DIMETHYLPHENYL)-1-(3-METHOXY-3-METHYLBUTYL)-N~6~-(4-PIPERAZIN-1-YLPHENYL)-1H-PYRAZOLO[3,4-D]PYRIMIDINE-3,6-DIAMINE",
            "non-polymer",
        ),
        "T75": ("TAK-715", "non-polymer"),
        "NP5": (
            "(5E)-12-CHLORO-13,15-DIHYDROXY-4,7,8,9-TETRAHYDRO-2-BENZOXACYCLOTRIDECINE-1,10(3H,11H)-DIONE",
            "non-polymer",
        ),
        "NP4": (
            "(5Z)-12-CHLORO-13,15-DIHYDROXY-4,7,8,9-TETRAHYDRO-2-BENZOXACYCLOTRIDECINE-1,10(3H,11H)-DIONE",
            "non-polymer",
        ),
        "AH8": (
            "(2R,3R,4R,5S)-2-AZIDO-5-(HYDROXYMETHYL)TETRAHYDROFURAN-3,4-DIOL",
            "non-polymer",
        ),
        "NP6": (
            "[[(3R,4R,5S,6R)-3-(BUTANOYLAMINO)-4,5-DIHYDROXY-6-(HYDROXYMETHYL)OXAN-2-YLIDENE]AMINO] N-PHENYLCARBAMATE",
            "non-polymer",
        ),
        "NP1": ("N-(3-AMINOPROPYL)-4-METHYL-2-NITROBENZENAMINE", "non-polymer"),
        "N6M": ("N-METHYL-9H-PURIN-6-AMINE", "non-polymer"),
        "NP3": (
            "1-[2-DEOXY-RIBOFURANOSYL]-1H-[3-NITRO-PYRROL]-5'-PHOSPHATE",
            "nucleic-acid",
        ),
        "NP2": ("N-(3-AMINOPROPYL)-2-NITROBENZENAMINE", "non-polymer"),
        "GI4": ("METHYL ACETIC ACID-SUBSTITUTED GLUCOIMIDAZOLE", "non-polymer"),
        "GI3": ("CARBOXYLATE-SUBSTITUTED GLUCOIMIDAZOLE", "non-polymer"),
        "GI2": ("METHYL ACETATE-SUBSTITUTED GLUCOIMIDAZOLE", "non-polymer"),
        "GI1": ("METHOXYCARBONYL-SUBSTITUTED GLUCOIMIDAZOLE", "non-polymer"),
        "J35": (
            "(2S)-2-[4-METHOXY-3-({[4-(TRIFLUOROMETHYL)BENZOYL]AMINO}METHYL)BENZYL]PENTANOIC ACID",
            "non-polymer",
        ),
        "J30": (
            "4-[4-(NAPHTHALEN-1-YLMETHYL)PIPERAZIN-1-YL]-4-OXOBUTANE-1-THIOL",
            "non-polymer",
        ),
        "U01": (
            "3-[1-(4-BROMO-PHENYL)-2-METHYL-PROPYL]-4-HYDROXY-CHROMEN-2-ONE",
            "non-polymer",
        ),
        "GIV": ("BETA-L-GALACTOPYRANOSE", "saccharide"),
        "MJF": (
            "[(1R)-3-(4-CHLOROPHENYL)-3-OXO-1-PHENYLPROPYL]PROPANEDIOIC ACID",
            "non-polymer",
        ),
        "MJG": (
            "N-[5-({5-[(4-ACETYLPIPERAZIN-1-YL)CARBONYL]-4-METHOXY-2-METHYLPHENYL}SULFANYL)-1,3-THIAZOL-2-YL]-4-({[(1S)-1,2,2-TRIMETHYLPROPYL]AMINO}METHYL)BENZAMIDE",
            "non-polymer",
        ),
        "GIS": ("ETHYL-(2-CARBOXY-4-GUANIDINIUM-PHENYL)-CHLOROACETATE", "non-polymer"),
        "GIR": ("GIRODAZOLE", "non-polymer"),
        "GIQ": ("3-METHYL-3H-IMIDAZO[4,5-F]QUINOLIN-2-AMINE", "non-polymer"),
        "GIP": ("S-(N-HYDROXY-N-IODOPHENYLCARBAMOYL)GLUTATHIONE", "non-polymer"),
        "MJI": (
            "1-HEXADECYL-3-TRIFLUOROETHYL-SN-GLYCERO-2-PHOSPHATE METHANE",
            "non-polymer",
        ),
        "GIX": ("PROTOPORPHYRIN IX CONTAINING GA", "non-polymer"),
        "GIG": (
            "METHYL (5-{4-[({[2-FLUORO-5-(TRIFLUOROMETHYL)PHENYL]AMINO}CARBONYL)AMINO]PHENOXY}-1H-BENZIMIDAZOL-2-YL)CARBAMATE",
            "non-polymer",
        ),
        "TNE": ("8-METHYL-8-AZABICYCLO[3,2,1]OCTAN-3-ONE", "non-polymer"),
        "PLT": (
            "[3-HYDROXY-2-METHYL-5-PHOSPHONOOXYMETHYL-PYRIDIN-4-YLMETHYL]-L-TRYPTOPHANE",
            "non-polymer",
        ),
        "GIC": ("N-(2-AMINOETHYL)-N-(1H-INDOL-3-YLACETYL)GLYCINE", "peptide"),
        "GIA": ("3-FLUORO-N-(NAPHTHALEN-2-YLCARBONYL)-D-PHENYLALANINE", "non-polymer"),
        "MJS": ("(5R)-5-AMINO-5-[HYDROXY(METHOXY)BORANYL]PENTAN-1-AMINIUM", "peptide"),
        "GIO": ("CYCLO-(GLYCINE-L-PROLINE) INHIBITOR", "non-polymer"),
        "GIN": (
            "2-{[(6-OXO-1,6-DIHYDROPYRIDIN-3-YL)METHYL]AMINO}-N-[4-PROPYL-3-(TRIFLUOROMETHYL)PHENYL]BENZAMIDE",
            "non-polymer",
        ),
        "GIM": ("GLUCOIMIDAZOLE", "non-polymer"),
        "GIL": (
            "2-{[6-{3-[AMINO(IMINO)METHYL]PHENOXY}-4-(DIISOPROPYLAMINO)-3,5-DIFLUOROPYRIDIN-2-YL]OXY}-5-[(ISOBUTYLAMINO)CARBONYL]BENZOIC ACID",
            "non-polymer",
        ),
        "MJZ": ("?", "non-polymer"),
        "PLR": (
            "(5-HYDROXY-4,6-DIMETHYLPYRIDIN-3-YL)METHYL DIHYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "AHB": ("BETA-HYDROXYASPARAGINE", "peptide"),
        "AHC": ("4-AMINOHYDROCINNAMIC ACID", "non-polymer"),
        "NPO": ("P-NITROPHENOL", "non-polymer"),
        "AHA": ("6-AMINO HEXANOIC ACID", "non-polymer"),
        "AHF": (
            "2-[4-(HYDROXY-METHOXY-METHYL)-BENZYL]-7-(4-HYDROXYMETHYL-BENZYL)-1,1-DIOXO-3,6-BIS-PHENOXYMETHYL-1LAMBDA6-[1,2,7]THIADIAZEPANE-4,5-DIOL",
            "non-polymer",
        ),
        "AHG": ("2,5-ANHYDROGLUCITOL-1,6-BIPHOSPHATE", "non-polymer"),
        "AHD": ("4-AMINO-1-HYDROXYBUTANE-1,1-DIYLDIPHOSPHONATE", "non-polymer"),
        "AHE": (
            "2-AMINO-4-[1-CARBOXYMETHYL-CARBAMOYL)-2-HYDROXYMETHYLSULFANYL-ETHYLCARBAMOYL]-BUTYRIC ACID",
            "non-polymer",
        ),
        "NPE": ("5-(PARA-NITROPHENYL PHOSPHONATE)-PENTANOIC ACID", "non-polymer"),
        "NPD": ("2-METHYL-3-NAPHTHALEN-1-YL-PROPIONALDEHYDE", "non-polymer"),
        "AHH": ("AMINO-HYDROXYHEPTANOIC ACID", "peptide"),
        "AHI": (
            "3-{[(E)-AMINO(HYDROXYIMINO)METHYL]AMINO}PROPAN-1-AMINIUM",
            "non-polymer",
        ),
        "AHN": ("N-[2-(1H-IMIDAZOL-4-YL)ETHYL]ACETAMIDE", "non-polymer"),
        "AHO": ("N-ACETYL-N-HYDROXY-L-ORNITHINE", "peptide"),
        "AHL": ("N-HYDROXY-L-ARGININAMIDE", "non-polymer"),
        "AHM": ("2,5-ANHYDROMANNITOL-1,6-DIPHOSPHATE", "non-polymer"),
        "AHR": ("ALPHA-L-ARABINOFURANOSE", "saccharide"),
        "AHS": (
            "(3-AMINO-4-CYCLOHEXYL-2-HYDROXY-BUTYL)-ISOBUTYL-CARBAMIC ACID",
            "peptide",
        ),
        "AHP": ("2-AMINO-HEPTANOIC ACID", "peptide"),
        "AHQ": (
            "(2S,7R,9R)-4,5-DIHYDROXY-2,3,6,7,8,9-HEXAHYDRO-1H-PYRROLO[2,3-F]QUINOLINE-2,7,9-TRICARBOXYLIC ACID",
            "non-polymer",
        ),
        "NPY": ("NAPHTHALENE", "non-polymer"),
        "NPX": ("(2R)-2-(6-METHOXYNAPHTHALEN-2-YL)PROPANOIC ACID", "non-polymer"),
        "AHT": ("4-(2,5-DIAMINO-5-HYDROXY-PENTYL)-PHENOL", "peptide"),
        "AHU": (
            "1',5'-ANHYDRO-2',3'-DIDEOXY-2'-(5-IODOURACIL-1-YL)-D-ABABINO-HEXITOL",
            "non-polymer",
        ),
        "AHZ": (
            "ADENOSINE DIPHOSPHATE 5-(BETA-ETHYL)-4-METHYL-THIAZOLE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "NPT": ("NEOPENTYLAMINE", "non-polymer"),
        "AHX": ("SERYL-HYDROXAMATE-ADENOSINE MONOPHOSPHATE", "non-polymer"),
        "AHY": ("(2S,3R)-3-AMINO-2-HYDROXYDECANOIC ACID", "non-polymer"),
        "NPP": (
            "N-(2-AMINO-ETHYL)-4,6-DINITRO-N'-(2,2,6,6-TETRAMETHYL-1-OXY-PIPERIDIN-4-YL)-BENZENE-1,3-DIAMINE",
            "non-polymer",
        ),
        "NPS": ("(2S)-2-(6-METHOXYNAPHTHALEN-2-YL)PROPANOIC ACID", "non-polymer"),
        "NPR": (
            "2-AMINO-7,8-DIHYDRO-6-(1,2,3-TRIHYDROXYPROPYL)-4(1H)-PTERIDINONE",
            "non-polymer",
        ),
        "LZ9": (
            "{[(2,6-DIFLUOROPHENYL)CARBONYL]AMINO}-N-(4-FLUOROPHENYL)-1H-PYRAZOLE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "LZ8": (
            "N-(4-FLUOROPHENYL)-4-[(PHENYLCARBONYL)AMINO]-1H-PYRAZOLE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "1TS": (
            "(2S)-N-[(4-CARBAMIMIDOYLPHENYL)METHYL]-1-[(2S)-2-[(3-CHLORO-4-METHOXYBENZENE)SULFONAMIDO]-3-{[(4-CYANOPHENYL)METHYL]CARBAMOYL}PROPANOYL]PYRROLIDINE-2-CARBOXAMIDE",
            "peptide",
        ),
        "1TP": (
            "1-(2-{(2S,4R,5R)-3-[(4-AMINO-2-METHYLPYRIMIDIN-5-YL)METHYL]-2-[(1S)-1-CARBOXY-1-HYDROXYETHYL]-4-METHYL-1,3-THIAZOLIDIN-5-YL}ETHOXY)-1,1,3,3-TETRAHYDROXY-1LAMBDA~5~-DIPHOSPHOX-1-EN-2-IUM 3-OXIDE",
            "non-polymer",
        ),
        "1TQ": ("6-(FORMYLAMINO)-7-HYDROXY-L-TRYPTOPHAN", "peptide"),
        "LZ1": ("1H-INDAZOLE", "non-polymer"),
        "LZ0": (
            "[1-(2-OXOETHYL)-1H-1,2,3-TRIAZOL-5-YL]METHYL 6-DEOXY-ALPHA-L-GALACTOPYRANOSIDE",
            "non-polymer",
        ),
        "LZ3": ("N-(4-SULFAMOYLPHENYL)-1H-INDAZOLE-3-CARBOXAMIDE", "non-polymer"),
        "LZ2": ("5-HYDROXYNAPHTHALENE-1-SULFONAMIDE", "non-polymer"),
        "LZ5": ("N-PHENYL-1H-PYRAZOLE-3-CARBOXAMIDE", "non-polymer"),
        "LZ4": ("4-[(6-CHLOROPYRAZIN-2-YL)AMINO]BENZENESULFONAMIDE", "non-polymer"),
        "LZ7": (
            "4-(ACETYLAMINO)-N-(4-FLUOROPHENYL)-1H-PYRAZOLE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "1TY": (
            "3-{(3E)-4-HYDROXY-6-OXO-3-[(2-PHENYLETHYL)IMINO]CYCLOHEXA-1,4-DIEN-1-YL}ALANINE",
            "peptide",
        ),
        "1TB": (
            "METHYL 2-[4-METHOXY-6-METHYL-1,3,5-TRAZIN-2-YL(METHYL)CARBAMOYLSULFAMOYL]BENZOATE",
            "non-polymer",
        ),
        "PLY": ("PALMITOYL", "non-polymer"),
        "R99": (
            "N-[3-CARBOXY-2-HYDROXY-PROPIONYL]-D-HOMOPHENYLALANYL-AMINO-2-METHYLBUTANE",
            "peptide",
        ),
        "R96": (
            "N-[4-({4-[5-(3,3-DIMETHYLPIPERIDIN-1-YL)-2-HYDROXYBENZOYL]BENZOYL}AMINO)AZEPAN-3-YL]ISONICOTINAMIDE",
            "non-polymer",
        ),
        "T2D": ("1,2,5-THIADIAZOLIDIN-3-ONE-1,1-DIOXIDE", "non-polymer"),
        "1TL": (
            "1-{2,5-ANHYDRO-6-DEOXY-4-[(PHOSPHONOOXY)METHYL]-ALPHA-L-MANNOFURANOSYL}PYRIMIDINE-2,4(1H,3H)-DIONE",
            "nucleic-acid",
        ),
        "PLX": (
            "(9R,11S)-9-({[(1S)-1-HYDROXYHEXADECYL]OXY}METHYL)-2,2-DIMETHYL-5,7,10-TRIOXA-2LAMBDA~5~-AZA-6LAMBDA~5~-PHOSPHAOCTACOSANE-6,6,11-TRIOL",
            "non-polymer",
        ),
        "PLG": (
            "N-GLYCINE-[3-HYDROXY-2-METHYL-5-PHOSPHONOOXYMETHYL-PYRIDIN-4-YL-METHANE]",
            "non-polymer",
        ),
        "RYU": (
            "(2E,3S)-3-HYDROXY-5'-[(4-HYDROXYPIPERIDIN-1-YL)SULFONYL]-3-METHYL-1,3-DIHYDRO-2,3'-BIINDOL-2'(1'H)-ONE",
            "non-polymer",
        ),
        "PLF": (
            "2-[5-(1-AMINO-2-METHYL-BUTYL)-[1,2,3]TRIAZOL-1-YL]-PROPIONIC ACID",
            "peptide",
        ),
        "SNG": (
            "METHYL 2-ACETAMIDO-1,2-DIDEOXY-1-SELENO-BETA-D-GLUCOPYRANOSIDE",
            "non-polymer",
        ),
        "PLE": ("[(1R)-1-AMINO-3-METHYLBUTYL]PHOSPHONIC ACID", "peptide"),
        "IBZ": ("2-IODOBENZYLTHIO GROUP", "non-polymer"),
        "PLD": ("DIUNDECYL PHOSPHATIDYL CHOLINE", "non-polymer"),
        "YZ9": (
            "7-HYDROXY-2-OXO-CHROMENE-3-CARBOXYLIC ACID ETHYL ESTER",
            "non-polymer",
        ),
        "PLC": ("DIUNDECYL PHOSPHATIDYL CHOLINE", "non-polymer"),
        "RMN": ("(R)-MANDELIC ACID", "non-polymer"),
        "TMY": ("ETHYL (2,3,4-TRIMETHOXY-6-OCTANOYLPHENYL)ACETATE", "non-polymer"),
        "PLB": (
            "2-[(2,4-DICHLOROBENZOYL)AMINO]-5-(PYRIMIDIN-2-YLOXY)BENZOIC ACID",
            "non-polymer",
        ),
        "KPS": (
            "1-DEOXY-1-(2,4-DIOXO-1,2,3,4-TETRAHYDROPYRIMIDIN-5-YL)-5-O-PHOSPHONO-D-RIBITOL",
            "non-polymer",
        ),
        "KPT": (
            "6,7-DIHYDRO[1,3]DIOXOLO[4,5-G][1,3]DIOXOLO[7,8]ISOQUINO[3,2-A]ISOQUINOLIN-5-IUM",
            "non-polymer",
        ),
        "PLA": (
            "2-[(3-HYDROXY-2-METHYL-5-PHOSPHONOOXYMETHYL-PYRIDIN-4-YLMETHYL)-AMINO]-2-METHYL-SUCCINIC ACID",
            "non-polymer",
        ),
        "KPV": ("5-PHENYL-2-KETO-VALERIC ACID", "non-polymer"),
        "KPY": (
            "(2R)-2-AZANYL-3-[2-[(E)-(1-OXIDANYL-1-OXIDANYLIDENE-PROPAN-2-YLIDENE)AMINO]ETHYLSULFANYL]PROPANOIC ACID",
            "peptide",
        ),
        "KPA": ("2-OXO-5-PHOSPHONOPENTANOIC ACID", "non-polymer"),
        "KPC": ("(2-[2-KETOPROPYLTHIO]ETHANESULFONATE", "non-polymer"),
        "KPG": (
            "[(3S,4R,5S)-3,4,5,6,6-PENTAHYDROXY-2-OXO-HEXYL] DIHYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "KPH": ("[(1R)-1,5-DIAMINOPENTYL]PHOSPHONIC ACID", "non-polymer"),
        "KPI": (
            "(2S)-2-AMINO-6-[(1-HYDROXY-1-OXO-PROPAN-2-YLIDENE)AMINO]HEXANOIC ACID",
            "peptide",
        ),
        "TMZ": ("3,4,5-TRIMETHYL-1,3-THIAZOLE", "non-polymer"),
        "KPL": ("KETOPANTOATE", "non-polymer"),
        "KPN": ("N-(2-AMINOETHYL)-N-(2,2'-BIPYRIDIN-5-YLACETYL)GLYCINE", "peptide"),
        "SHA": ("SALICYLHYDROXAMIC ACID", "non-polymer"),
        "LZQ": ("4-[3-(DIBENZYLAMINO)PHENYL]-2,4-DIOXOBUTANOIC ACID", "non-polymer"),
        "LZP": (
            "4,4'-PIPERAZINE-1,4-DIYLBIS{1-[3-(BENZYLOXY)PHENYL]-4-OXOBUTANE-1,3-DIONE}",
            "non-polymer",
        ),
        "LZS": (
            "N-{6-(CYCLOPROPYLMETHOXY)-7-METHOXY-2-[6-(2-METHYLPROPYL)-5-OXO-3,4,5,6-TETRAHYDRO-2,6-NAPHTHYRIDIN-2(1H)-YL]QUINAZOLIN-4-YL}-2,2,2-TRIFLUOROETHANESULFONAMIDE",
            "non-polymer",
        ),
        "LZR": (
            "6-(CYCLOPROP-2-EN-1-YLMETHOXY)-2-[6-(CYCLOPROPYLMETHYL)-5-OXO-3,4,5,6-TETRAHYDRO-2,6-NAPHTHYRIDIN-2(1H)-YL]-7-METHOXYQUINAZOLIN-4(3H)-ONE",
            "non-polymer",
        ),
        "LNT": ("N-[(2S)-2-AMINO-1,1-DIHYDROXY-4-METHYLPENTYL]-L-THREONINE", "peptide"),
        "PLJ": ("METHYL L-PROLINATE", "peptide"),
        "LZI": (
            "(4R)-4-{[(5-CHLOROTHIOPHEN-2-YL)CARBONYL]AMINO}-N-(CYCLOPROPYLMETHYL)-1-(2-{[2-FLUORO-4-(2-OXOPYRIDIN-1(2H)-YL)PHENYL]AMINO}-2-OXOETHYL)-L-PROLINAMIDE",
            "non-polymer",
        ),
        "LZH": (
            "5-CHLORO-THIOPHENE-2-CARBOXYLIC ACID ((3R,5S)-1-{[2-FLUORO-4-(2-OXO-PYRIDIN-1-YL)-PHENYLCARBAMOYL]-METHYL}-5-HYDROXYMETHYL-PYRROLIDIN-3-YL)-AMIDE",
            "non-polymer",
        ),
        "LZK": (
            "6-(2,6-DIMETHOXYPHENYL)PYRIDO[2,3-D]PYRIMIDINE-2,7-DIAMINE",
            "non-polymer",
        ),
        "LZJ": (
            "6-(2,6-DIBROMOPHENYL)PYRIDO[2,3-D]PYRIMIDINE-2,7-DIAMINE",
            "non-polymer",
        ),
        "LZM": ("PYRAZOLO[1,5-A]PYRIMIDINE-3-CARBONITRILE", "non-polymer"),
        "LZL": (
            "7-(2,5-DIHYDROPYRROL-1-YL)-6-PHENYL-PYRIDO[6,5-D]PYRIMIDIN-2-AMINE",
            "non-polymer",
        ),
        "LZA": (
            "{[(2,6-DIFLUOROPHENYL)CARBONYL]AMINO}-N-PIPERIDIN-4-YL-1H-PYRAZOLE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "SHD": ("ALPHA-D-ALTROPYRANOSE", "non-polymer"),
        "LZC": (
            "5-[(4-AMINOCYCLOHEXYL)AMINO]-7-(PROPAN-2-YLAMINO)PYRAZOLO[1,5-A]PYRIMIDINE-3-CARBONITRILE",
            "non-polymer",
        ),
        "LZB": (
            "5-CHLORO-7-[(1-METHYLETHYL)AMINO]PYRAZOLO[1,5-A]PYRIMIDINE-3-CARBONITRILE",
            "non-polymer",
        ),
        "LZE": (
            "4-{[(2,6-DICHLOROPHENYL)CARBONYL]AMINO}-N-PIPERIDIN-4-YL-1H-PYRAZOLE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "LZD": (
            "4-{[(2,6-DIFLUOROPHENYL)CARBONYL]AMINO}-N-[(3S)-PIPERIDIN-3-YL]-1H-PYRAZOLE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "LZG": (
            "5-CHLORO-THIOPHENE-2-CARBOXYLIC ACID ((3S,4S)-4-FLUORO- 1-{[2-FLUORO-4-(2-OXO-2H-PYRIDIN-1-YL)-PHENYLCARBAMOYL]-METHYL}-PYRROLIDIN-3-YL)-AMIDE",
            "non-polymer",
        ),
        "LZF": (
            "5-CHLORO-N-[(3S,4S)-1-(2-{[2-FLUORO-4-(2-OXOPYRIDIN-1(2H)-YL)PHENYL]AMINO}-2-OXOETHYL)-4-METHOXYPYRROLIDIN-3-YL]THIOPHENE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "3PX": ("(3S)-3-(PROPAN-2-YLOXY)-L-PROLINE", "peptide"),
        "5CS": ("2-AMINO-3-(CYSTEIN-S-YL)-ISOXAZOLIDIN-5-YL-ACETIC ACID", "peptide"),
        "SHF": ("LAEVULINIC ACID", "non-polymer"),
        "5CP": (
            "[4-({4-[(5-CYCLOPROPYL-1H-PYRAZOL-3-YL)AMINO]-6-(METHYLAMINO)PYRIMIDIN-2-YL}AMINO)PHENYL]ACETONITRILE",
            "non-polymer",
        ),
        "IYX": (
            "1-[5-(3,4-DIMETHOXYPHENYL)-1-BENZOTHIOPHEN-2-YL]METHANEDIAMINE",
            "non-polymer",
        ),
        "SHG": ("2-DEOXY-2-FLUORO-BETA-D-GLUCOPYRANOSE", "non-polymer"),
        "3PZ": (
            "(3R)-3-BENZYL-4-[(4-METHOXYPHENYL)SULFONYL]-1-[(1-METHYL-1H-IMIDAZOL-5-YL)METHYL]-2,3,4,5-TETRAHYDRO-1H-1,4-BENZODIAZEPINE-7-CARBONITRILE",
            "non-polymer",
        ),
        "5CY": ("N,N'-(DIPROPYL)-TETRAMETHYLINDODICARBOCYANINE", "non-polymer"),
        "5CG": ("5'-CHLORO-5'-DEOXY-GUANOSINE", "nucleic-acid"),
        "5CF": ("?", "nucleic-acid"),
        "5CD": ("5'-CHLORO-5'-DEOXYADENOSINE", "non-polymer"),
        "5CB": (
            "N-(5-CHLORO-BENZO[B]THIOPHEN-3-YLMETHYL)-2-[6-CHLORO-OXO-3-(2-PYRIDIN-2-YL-ETHYLAMINO)-2H-PYRAZIN-1-YL]-ACETAMIDE",
            "non-polymer",
        ),
        "5CA": ("5'-O-(N-(L-CYSTEINYL)-SULFAMOYL)ADENOSINE", "non-polymer"),
        "ET": ("ETHIDIUM", "non-polymer"),
        "5CO": (
            "5-CHLORO-6-ETHYL-2-{5-[(4-METHYLPIPERAZIN-1-YL)SULFONYL]-2-PROPOXYPHENYL}PYRIMIDIN-4(3H)-ONE",
            "non-polymer",
        ),
        "5CN": (
            "5-CYANO-FURAN-2-CARBOXYLIC ACID [5-HYDROXYMETHYL-2-(4-METHYL-PIPERIDIN-1-YL)-PHENYL]-AMIDE",
            "non-polymer",
        ),
        "5CM": ("5-METHYL-2'-DEOXY-CYTIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "5CL": ("1,2,3,4,5-PENTACHLOROBENZENE", "non-polymer"),
        "5CI": (
            "8-BROMO-5'-{[3-CARBOXY-2-(CARBOXYMETHYL)-2-HYDROXYPROPANOYL]AMINO}-5'-DEOXYADENOSINE",
            "non-polymer",
        ),
        "5CH": (
            "5-CHLORO-6'-METHYL-3-[4-(METHYLSULFONYL)PHENYL]-2,3'-BIPYRIDINE",
            "non-polymer",
        ),
        "INB": (
            "1-OCTADECYL-2-ACETAMIDO-2-DEOXY-SN-GLYCEROL-3-PHOSPHOETHYLMETHYL SULFIDE",
            "non-polymer",
        ),
        "G1M": (
            "2'-DEOXY-5'-O-[(R)-HYDROXY({(S)-HYDROXY[(1R)-1-PHOSPHONOETHYL]PHOSPHORYL}OXY)PHOSPHORYL]GUANOSINE",
            "non-polymer",
        ),
        "G1L": (
            "3-CHLORO-2,2-DIMETHYL-N-[4-(TRIFLUOROMETHYL)PHENYL]PROPANAMIDE",
            "non-polymer",
        ),
        "G1G": (
            "7-METHYL-GUANOSINE-5'-TRIPHOSPHATE-5'-(2'-O-METHYL)-GUANOSINE",
            "non-polymer",
        ),
        "LNV": (
            "5-(ACETYLAMINO)-2,6-ANHYDRO-4-CARBAMIMIDAMIDO-3,4,5-TRIDEOXY-7-O-METHYL-D-GLYCERO-D-GALACTO-NON-2-ENONIC ACID",
            "non-polymer",
        ),
        "G1C": (
            "5'-O-[(R)-{[(S)-[(S)-CHLORO(PHOSPHONO)METHYL](HYDROXY)PHOSPHORYL]OXY}(HYDROXY)PHOSPHORYL]-2'-DEOXYGUANOSINE",
            "non-polymer",
        ),
        "EP": ("EPOTHILONE A", "non-polymer"),
        "YI5": (
            "4-CHLORO-6-[5-(3,4-DIMETHOXYPHENYL)-1,2,3-THIADIAZOL-4-YL]BENZENE-1,3-DIOL",
            "non-polymer",
        ),
        "BXQ": (
            "(3S,4S,5R)-3-{4-AMINO-3-FLUORO-5-[(2S)-3,3,3-TRIFLUORO-2-HYDROXYPROPYL]BENZYL}-5-[(3-TERT-BUTYLBENZYL)AMINO]TETRAHYDRO-2H-THIOPYRAN-4-OL 1,1-DIOXIDE",
            "non-polymer",
        ),
        "YI1": (
            "[4-[3-(DIFLUORO-PHOSPHONO-METHYL)PHENYL]PHENYL]METHOXYAZANIUM",
            "non-polymer",
        ),
        "YI0": (
            "4-CHLORO-6-[5-(4-METHOXYPHENYL)-1,2,3-THIADIAZOL-4-YL]BENZENE-1,3-DIOL",
            "non-polymer",
        ),
        "AR9": (
            "(2R,4S)-N-BUTYL-4-[(2S,5S,7R)-2,7-DIMETHYL-3,15-DIOXO-1,4-DIAZACYCLOPENTADECAN-5-YL]-4-HYDROXY-2-METHYLBUTANAMIDE",
            "non-polymer",
        ),
        "EL": ("ELLIPTICINE", "non-polymer"),
        "G1R": (
            "[(2R,3R,4R,5R)-5-(2-AMINO-6-OXO-1,6-DIHYDRO-9H-PURIN-9-YL)-3,4-DIHYDROXYTETRAHYDROFURAN-2-YL]METHYL [(2R,3S,4R,5S)-3,4,5-TRIHYDROXYTETRAHYDROFURAN-2-YL]METHYL DIHYDROGEN DIPHOSPHATE",
            "non-polymer",
        ),
        "G1P": ("ALPHA-D-GLUCOSE-1-PHOSPHATE", "saccharide"),
        "TBY": ("TRIBUTYLSTANNANYL", "non-polymer"),
        "3PS": (
            "4-AMINO-N-(6-METHOXYPYRIDIN-3-YL)-2-METHYLQUINAZOLINE-8-CARBOXAMIDE",
            "non-polymer",
        ),
        "YIO": (
            "(2R,3R,4S,5R,6S)-2-(HYDROXYMETHYL)-6-SULFANYL-OXANE-3,4,5-TRIOL",
            "non-polymer",
        ),
        "YIN": (
            "(6R,7S)-6,7-BIS(4-BROMOPHENYL)-7,11-DIHYDRO-6H-CHROMENO[4,3-D][1,2,4]TRIAZOLO[1,5-A]PYRIMIDINE",
            "non-polymer",
        ),
        "Y8L": (
            "3,4-DIMETHYL-5-(2-{[(1Z)-4-PIPERAZIN-1-YLCYCLOHEXA-2,4-DIEN-1-YLIDENE]AMINO}PYRIMIDIN-4-YL)-1,3-THIAZOL-2(3H)-ONE",
            "non-polymer",
        ),
        "MGA": ("BETA-METHYL-D-GALACTOSIDE", "saccharide"),
        "YIT": (
            "N-{4-[(1E)-N-CARBAMIMIDOYLBUTANEHYDRAZONOYL]PHENYL}-5-METHOXY-1H-INDOLE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "YIW": (
            "1-(3-TERT-BUTYL-1-PHENYL-1H-PYRAZOL-5-YL)-3-(2-{[3-(1-METHYLETHYL)[1,2,4]TRIAZOLO[4,3-A]PYRIDIN-6-YL]SULFANYL}BENZYL)UREA",
            "non-polymer",
        ),
        "TBZ": (
            "2''-(4-METHOXYPHENYL)-5-(3-AMINO-1-PYRROLIDINYL)-2,5',2',5''-TRI-BENZIMIDAZOLE",
            "non-polymer",
        ),
        "YIQ": (
            "(E)-5-(1-(2-CARBAMIMIDOYLHYDRAZONO)ETHYL)-N-(1H-INDOL-6-YL)-1H-INDOLE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "G19": (
            "(2S,3AR,4R,5S,6S,8R,9R,9AR,10R)-2,5-DIHYDROXY-4,6,9,10-TETRAMETHYL-1-OXO-6-VINYLDECAHYDRO-3A,9-PROP[1]ENOCYCLOPENTA[8]ANNULEN-8-YL [(6-AMINOPYRIDAZIN-3-YL)CARBONYL]CARBAMATE",
            "non-polymer",
        ),
        "G18": ("4H-THIENO[2,3-C]ISOQUINOLIN-5-ONE", "non-polymer"),
        "G16": ("ALPHA-D-GLUCOSE 1,6-BISPHOSPHATE", "saccharide"),
        "G15": (
            "5-(5-CHLORO-2-THIENYL)-N-{(3S)-1-[(1S)-1-METHYL-2-MORPHOLIN-4-YL-2-OXOETHYL]-2-OXOPYRROLIDIN-3-YL}-1H-1,2,4-TRIAZOLE-3-SULFONAMIDE",
            "non-polymer",
        ),
        "G14": ("2-PHENYL-1H-IMIDAZOLE-4-CARBOXYLIC ACID", "non-polymer"),
        "G13": (
            "N-[(3AR,6S,6AS)-1-(N-METHYL-L-ALANYL-3-METHYL-L-VALYL)OCTAHYDROCYCLOPENTA[B]PYRROL-6-YL]NAPHTHALENE-1-CARBOXAMIDE",
            "non-polymer",
        ),
        "G12": ("3,8-DIBROMO-7-HYDROXY-4-METHYL-2H-CHROMEN-2-ONE", "non-polymer"),
        "G11": (
            "6-[5-(CYCLOPROPYLCARBAMOYL)-2-METHYLPHENYL]-N-(CYCLOPROPYLMETHYL)PYRIDINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "HD2": (
            "4-CYCLOPENTYL-N-[(1S,3R)-5-OXIDANYL-2-ADAMANTYL]-2-[[(3S)-OXOLAN-3-YL]AMINO]PYRIMIDINE-5-CARBOXAMIDE",
            "non-polymer",
        ),
        "4SO": ("4-SULFAMOYLBENZOIC ACID", "non-polymer"),
        "IBP": ("IBUPROFEN", "non-polymer"),
        "4SM": ("TETRATHIOMOLYBDATE", "non-polymer"),
        "MGE": (
            "(1S)-2-(ALPHA-L-ALLOPYRANOSYLOXY)-1-[(TRIDECANOYLOXY)METHYL]ETHYL PALMITATE",
            "non-polymer",
        ),
        "5C1": (
            "5-(2-CHLOROPHENYL)-N-HYDROXY-1,3-OXAZOLE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "MPH": ("(1-AMINO-3-METHYLSULFANYL-PROPYL)-PHOSPHONIC ACID", "peptide"),
        "4SB": (
            "(2R)-4-AMINO-N'-[(1E)-(3-BROMO-4-HYDROXYPHENYL)METHYLIDENE]-2-PHENYLBUTANEHYDRAZIDE",
            "non-polymer",
        ),
        "4SC": ("?", "nucleic-acid"),
        "ZD9": (
            "2,7-DIMETHYL-6-[(PROP-1-YN-1-YLAMINO)METHYL]QUINAZOLIN-4(3H)-ONE",
            "non-polymer",
        ),
        "IFB": (
            "N-[(2',4'-DIFLUORO-4-HYDROXY-5-IODOBIPHENYL-3-YL)CARBONYL]-BETA-ALANINE",
            "non-polymer",
        ),
        "4ST": ("1,2,3,4-TETRAHYDROGEN-STAUROSPORINE", "non-polymer"),
        "4SU": ("4-THIOURIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "4SR": (
            "(4S)-4-[3-(CYCLOPENTYLOXY)-4-METHOXYPHENYL]PYRROLIDIN-2-ONE",
            "non-polymer",
        ),
        "TBS": ("4,5,6,7-TETRABROMOBENZOTRIAZOLE", "non-polymer"),
        "4SP": ("?", "non-polymer"),
        "3PM": ("O-TERT-BUTYL-D-SERINE", "peptide"),
        "TBR": ("HEXATANTALUM DODECABROMIDE", "non-polymer"),
        "9ZT": (
            "(1R,2S,4S,5R)-2-[(4-METHOXYPHENYL)METHYL]-1,4,5-TRIS(OXIDANYL)-3-OXIDANYLIDENE-CYCLOHEXANE-1-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "2E3": (
            "(2E)-3-[4-HYDROXY-3-(3,5,5,8,8-PENTAMETHYL-5,6,7,8-TETRAHYDRONAPHTHALEN-2-YL)PHENYL]PROP-2-ENOIC ACID",
            "non-polymer",
        ),
        "2E1": (
            "5-(5-CHLORO-2,4-DIHYDROXYPHENYL)-N-ETHYL-4-PIPERAZIN-1-YL-1H-PYRAZOLE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "O15": (
            "(E)-2-(4-CARBAMIMIDOYLBENZYLIDENEAMINOOXY)-2-METHYLPROPANOIC ACID",
            "non-polymer",
        ),
        "MGN": ("2-METHYL-GLUTAMINE", "peptide"),
        "Y1": ("YTTRIUM ION", "non-polymer"),
        "MGM": (
            "2-[METHYL-(5-GERANYL-4-METHYL-PENT-3-ENYL)-AMINO]-ETHYL-DIPHOSPHATE",
            "non-polymer",
        ),
        "TBV": ("BILIVERDIN", "non-polymer"),
        "3PB": ("(3R)-3-(PHOSPHONOOXY)BUTANOIC ACID", "non-polymer"),
        "IBR": ("?", "non-polymer"),
        "LNR": ("L-NOREPINEPHRINE", "non-polymer"),
        "61X": (
            "4-{3-[3-(METHYLSULFONYL)PHENOXY]PHENYL}-8-(TRIFLUOROMETHYL)QUINOLINE",
            "non-polymer",
        ),
        "3PC": ("(3S)-3,4-DI-N-HEXANOYLOXYBUTYL-1-PHOSPHOCHOLINE", "non-polymer"),
        "C47": (
            "METHYL (3R)-1-[(5S,6S,8R)-5-AMINO-9-BUTYLAMINO-6-HYDROXY-3,3,8-TRIMETHYL-9-OXO-NONANOYL]-3,4-DIHYDRO-2H-QUINOLINE-3-CARBOXYLATE",
            "non-polymer",
        ),
        "MGR": ("MALACHITE GREEN", "non-polymer"),
        "1E3": ("(2S)-2,9-DIAMINONONANOIC ACID", "peptide"),
        "3PE": ("1,2-DIACYL-SN-GLYCERO-3-PHOSPHOETHANOLAMINE", "non-polymer"),
        "SW3": ("1-(1,5-DIMETHYL-1H-PYRAZOL-3-YL)METHANAMINE", "non-polymer"),
        "NG7": (
            "3-{2-[(CYCLOPROPYLCARBONYL)AMINO][1,3]THIAZOLO[5,4-B]PYRIDIN-5-YL}-N-{4-[(4-ETHYLPIPERAZIN-1-YL)METHYL]-3-(TRIFLUOROMETHYL)PHENYL}BENZAMIDE",
            "non-polymer",
        ),
        "1E5": (
            "6-[3-(CARBOXYMETHOXY)-5-FLUOROPHENYL]PYRIDINE-3-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "3PG": ("3-PHOSPHOGLYCERIC ACID", "non-polymer"),
        "61E": (
            "4-(2-CHLOROPHENYL)-8-[3-(DIMETHYLAMINO)PROPOXY]-9-HYDROXY-6-METHYLPYRROLO[3,4-C]CARBAZOLE-1,3(2H,6H)-DIONE",
            "non-polymer",
        ),
        "PPC": ("5-PHOSPHORIBOSYL-1-(BETA-METHYLENE) PYROPHOSPHATE", "non-polymer"),
        "618": (
            "(3R,6R,9AR)-2,2-DIMETHYL-6-[(N-METHYL-L-ALANYL)AMINO]-N-(3-METHYL-1-PHENYL-1H-PYRAZOL-5-YL)-5-OXO-2,3,5,6,9,9A-HEXAHYDRO[1,3]THIAZOLO[3,2-A]AZEPINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "619": (
            "N-[(E)-3-[(2R,3S,4R,5R)-5-(6-ETHYLAMINOPURIN-9-YL)-3,4-DIHYDROXY-OXOLAN-2-YL]PROP-2-ENYL]-5-(4-FLUOROPHENYL)-2,3-DIHYDROXY-BENZAMIDE",
            "non-polymer",
        ),
        "I48": (
            "N-{3-METHYL-5-[2-(PYRIDIN-4-YLAMINO)-ETHOXY]-PHENYL}-BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "T87": (
            "[(1-{2[(4-CARBAMIMIDOYL-PHENYLAMINO)-METHYL]-1-METHYL-1H-BENZOIMIDAZOL-5-YL}-CYCLOPROPYL)-PYRIDIN-2-YL-METHYLENEAMINOOXY]-ACETIC ACID ETHYL ESTER",
            "non-polymer",
        ),
        "PPB": (
            "1-{[1-(2-AMINO-3-PHENYL-PROPIONYL)-PYRROLIDINE-2-CARBONYL]-AMINO}-2-(3-CYANO-PHENYL)-ETHANEBORONIC ACID",
            "non-polymer",
        ),
        "612": (
            "6-(CYCLOHEXYLSULFANYL)-1-(ETHOXYMETHYL)-5-(1-METHYLETHYL)PYRIMIDINE-2,4(1H,3H)-DIONE",
            "non-polymer",
        ),
        "613": ("4-(4-CHLOROPHENYL)PIPERAZINE-1-CARBOXIMIDAMIDE", "non-polymer"),
        "610": (
            "N-[(E)-3-[(2R,3S,4R,5R)-3,4-DIHYDROXY-5-PURIN-9-YL-OXOLAN-2-YL]PROP-2-ENYL]-2,3-DIHYDROXY-5-NITRO-BENZAMIDE",
            "non-polymer",
        ),
        "611": (
            "N-[(E)-3-[(2R,3S,4R,5S)-3,4-DIHYDROXY-5-PYRIDIN-4-YLSULFANYL-OXOLAN-2-YL]PROP-2-ENYL]-2,3-DIHYDROXY-5-NITRO-BENZAMIDE",
            "non-polymer",
        ),
        "616": (
            "N-[2-(1-FORMYL-2-METHYL-PROPYL)-1-(4-PIPERIDIN-1-YL-BUT-2-ENOYL)-PYRROLIDIN-3-YL]-METHANESULFONAMIDE",
            "non-polymer",
        ),
        "617": ("2-{[(4-CHLOROPHENOXY)ACETYL]AMINO}BENZOIC ACID", "non-polymer"),
        "614": (
            "7-[(4-METHYLPIPERAZIN-1-YL)METHYL]-4-[(3-METHYL-1H-PYRAZOL-5-YL)AMINO]-2-(TETRAHYDRO-2H-PYRAN-4-YL)PHTHALAZIN-1(2H)-ONE",
            "non-polymer",
        ),
        "615": (
            "(1R,2S)-N~2~-HYDROXY-1-{4-[(2-PHENYLQUINOLIN-4-YL)METHOXY]BENZYL}CYCLOPROPANE-1,2-DICARBOXAMIDE",
            "non-polymer",
        ),
        "WCC": ("FE(3)-NI(1)-S(4) CLUSTER", "non-polymer"),
        "XM1": (
            "(2S)-2-(4-CHLOROBENZYL)-3-OXO-3-[4-(7H-PYRROLO[2,3-D]PYRIMIDIN-4-YL)PIPERAZIN-1-YL]PROPAN-1-AMINE",
            "non-polymer",
        ),
        "PPG": (
            "(2E,3E)-4-(2-AMINOETHOXY)-2-[({3-HYDROXY-2-METHYL-5-[(PHOSPHONOOXY)METHYL]PYRIDIN-4-YL}METHYL)IMINO]BUT-3-ENOIC ACID",
            "non-polymer",
        ),
        "TBC": ("(9S,10R)-9-HYDROXY-7,8,9,10-TETRAHYDROBENZO[A]PYRENE", "non-polymer"),
        "2EZ": ("2-ETHYLIMIDAZOLE", "non-polymer"),
        "PPE": (
            "4-[(1,3-DICARBOXY-PROPYLAMINO)-METHYL]-3-HYDROXY-2-METHYL-5-PHOSPHONOOXYMETHYL-PYRIDINIUM",
            "non-polymer",
        ),
        "8LI": (
            "(4Z)-4-[HYDROXY(SULFANYL)METHYLIDENE]-2-(3-METHYLBUTANOYL)-1,3-OXAZOL-5(4H)-ONE",
            "non-polymer",
        ),
        "2ER": (
            "[(2R,3S,4R,5R)-5-[4-AMINOCARBONYL-5-[[(Z)-[(3R,4R)-3,4-DIHYDROXY-2-OXO-5-PHOSPHONOOXY-PENTYL]IMINOMETHYL]AMINO]IMIDAZOL-1-YL]-3,4-DIHYDROXY-OXOLAN-2-YL]METHYL DIHYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "TBE": ("TAZOBACTAM INTERMEDIATE", "non-polymer"),
        "2EP": ("2-ETHYLPIPERIDINE", "non-polymer"),
        "2EQ": (
            "5-(5-CHLORO-2,4-DIHYDROXYPHENYL)-N-ETHYL-4-[4-(MORPHOLIN-4-YLMETHYL)PHENYL]ISOXAZOLE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "TFB": ("TETRAHYDROFURAN-2-CARBOXYLIC ACID", "non-polymer"),
        "2EH": ("(2S)-2-ETHYLHEXAN-1-OL", "non-polymer"),
        "XL1": ("ETHANETHIOL", "non-polymer"),
        "2EC": ("9-ETHYL-9H-PURIN-6-YLAMINE", "non-polymer"),
        "2EA": (
            "(1S)-2-(1H-INDOL-3-YL)-1-{[(5-ISOQUINOLIN-6-YLPYRIDIN-3-YL)OXY]METHYL}ETHYLAMINE",
            "non-polymer",
        ),
        "2EG": ("?", "nucleic-acid"),
        "2ED": (
            "N-[(1R,2R,3E)-2-HYDROXY-1-(HYDROXYMETHYL)HEPTADEC-3-EN-1-YL]ACETAMIDE",
            "non-polymer",
        ),
        "0GF": ("4-BROMO-3-METHYLBENZOIC ACID", "non-polymer"),
        "0GG": ("3-{[(4-HYDROXY-3-METHOXYPHENYL)ACETYL]AMINO}-D-ALANINE", "peptide"),
        "0GE": (
            "N-{[5-(DIMETHYLAMINO)NAPHTHALEN-1-YL]SULFONYL}-L-ALPHA-GLUTAMYL-N-[(2S,3S)-6-CARBAMIMIDAMIDO-1-CHLORO-2-HYDROXYHEXAN-3-YL]GLYCINAMIDE",
            "peptide",
        ),
        "7PY": (
            "7-PYRIDIN-2-YL-N-(3,4,5-TRIMETHOXYPHENYL)-7H-PYRROLO[2,3-D]PYRIMIDIN-2-AMINE",
            "non-polymer",
        ),
        "0GA": (
            "3-{[(1Z)-1-CARBOXYPROP-1-EN-1-YL]OXY}-2-HYDROXYBENZOIC ACID",
            "non-polymer",
        ),
        "8GD": ("8-OXO-7,8-DIHYDRO-2'-DEOXY-GUANOSINE-5'-DIPHOSPHATE", "non-polymer"),
        "0GO": (
            "(2S)-N-[(2S,3R)-3-HYDROXY-1-PHENYL-4-{[3-(PROPAN-2-YL)BENZYL]AMINO}BUTAN-2-YL]-2-[(5S)-6-OXO-1-PROPYL-1,7-DIAZASPIRO[4.4]NON-7-YL]PROPANAMIDE",
            "non-polymer",
        ),
        "0GM": (
            "N-[(5S,9S,10S,13S)-9-HYDROXY-5,10-BIS(2-METHYLPROPYL)-4,7,12,16-TETRAOXO-3,6,11,17-TETRAAZABICYCLO[17.3.1]TRICOSA-1(23),19,21-TRIEN-13-YL]-3-(NAPHTHALEN-1-YL)-2-(NAPHTHALEN-1-YLMETHYL)PROPANAMIDE",
            "peptide",
        ),
        "7PP": (
            "5-{2-AMINO-4-CHLORO-7-[(4-METHOXY-3,5-DIMETHYLPYRIDIN-2-YL)METHYL]-7H-PYRROLO[2,3-D]PYRIMIDIN-5-YL}-2-METHYLPENT-4-YN-2-OL",
            "non-polymer",
        ),
        "7PQ": (
            "4-[2-(AMINOMETHYL)NAPHTHALEN-1-YL]-N-[2-(MORPHOLIN-4-YL)ETHYL]BENZAMIDE",
            "non-polymer",
        ),
        "0GH": (
            "N-[(2S)-1-({(2S,3R)-3-HYDROXY-1-[(2-METHYLPROPYL)AMINO]-1-OXOBUTAN-2-YL}AMINO)-3-PHENYLPROPAN-2-YL]-5-[METHYL(METHYLSULFONYL)AMINO]-N'-[(1R)-1-PHENYLETHYL]BENZENE-1,3-DICARBOXAMIDE",
            "non-polymer",
        ),
        "0GV": ("(3BETA,8ALPHA,9BETA)-3-HYDROXYCHOLEST-5-EN-7-ONE", "non-polymer"),
        "0GT": (
            "(3R,4S,5S)-3-[(3-TERT-BUTYLBENZYL)AMINO]-5-{[3-(2,2-DIFLUOROETHYL)-1H-INDOL-5-YL]METHYL}TETRAHYDRO-2H-THIOPYRAN-4-OL 1,1-DIOXIDE",
            "non-polymer",
        ),
        "0GU": ("?", "non-polymer"),
        "0GR": ("N-(FURAN-2-YLCARBONYL)-L-LEUCYL-L-TRYPTOPHAN", "peptide"),
        "0GS": (
            "(3S,4S,5R)-3-(4-AMINO-3-BROMO-5-FLUOROBENZYL)-5-{[3-(1,1-DIFLUOROETHYL)BENZYL]AMINO}TETRAHYDRO-2H-THIOPYRAN-4-OL 1,1-DIOXIDE",
            "non-polymer",
        ),
        "0GQ": (
            "4-AMINO-N-{(1R,8R,9R,13R)-16-(4-AMINO-2-METHYLPYRIMIDIN-5-YL)-1-BENZYL-8-(CYCLOHEXYLMETHYL)-9-HYDROXY-13-[(1S)-1-METHYLPROPYL]-2,6,11,14-TETRAOXO-3,7,12,15-TETRAAZAHEXADEC-1-YL}PIPERIDINE-1-CARBOXAMIDE",
            "peptide",
        ),
        "7PE": (
            "2-(2-(2-(2-(2-(2-ETHOXYETHOXY)ETHOXY)ETHOXY)ETHOXY)ETHOXY)ETHANOL",
            "non-polymer",
        ),
        "7PG": ("2,5,8,11,14,17,20,23-OCTAOXAPENTACOSAN-25-OL", "non-polymer"),
        "0GZ": ("2-OXO-2H-PYRAN-4,6-DICARBOXYLIC ACID", "non-polymer"),
        "7PA": ("PROPANE-1,3-DIYLBIS(PHOSPHONIC ACID)", "non-polymer"),
        "0GX": (
            "{(3-EXO)-3-[5-(AMINOMETHYL)-2-FLUOROPHENYL]-8-AZABICYCLO[3.2.1]OCT-8-YL}(4-BROMO-3-METHYL-5-PROPOXYTHIOPHEN-2-YL)METHANONE",
            "non-polymer",
        ),
        "0GY": (
            "(1E,3Z)-4-HYDROXYBUTA-1,3-DIENE-1,2,4-TRICARBOXYLIC ACID",
            "non-polymer",
        ),
        "WZ5": (
            "METHYL 2-(ACETYLAMINO)-2-DEOXY-BETA-D-GLUCOPYRANOSYL-(1->2)-ALPHA-D-MANNOPYRANOSYL-(1->3)-[ALPHA-D-MANNOPYRANOSYL-(1->6)-ALPHA-D-MANNOPYRANOSYL-(1->6)]-BETA-D-MANNOPYRANOSIDE",
            "non-polymer",
        ),
        "TME": ("PROPANE", "non-polymer"),
        "U55": ("4-[(6-AMINO-4-PYRIMIDINYL)AMINO]BENZENESULFONAMIDE", "non-polymer"),
        "U51": (
            "(5Z)-7-{(1R,4S,5R,6R)-6-[(1E)-OCT-1-EN-1-YL]-2,3-DIAZABICYCLO[2.2.1]HEPT-2-EN-5-YL}HEPT-5-ENOIC ACID",
            "non-polymer",
        ),
        "XZZ": ("6-CHLORO-7-HYDROXY-4-METHYL-2H-CHROMEN-2-ONE", "non-polymer"),
        "QUL": (
            "N,N-DIETHYL-N'-(10H-INDOLO[3,2-B]QUINOLIN-11-YL)ETHANE-1,2-DIAMINE",
            "non-polymer",
        ),
        "XZN": (
            "4-[(3,5-DICHLORO-4-METHOXYPHENYL)AMINO]-6-METHOXY-7-[3-(4-METHYLPIPERAZIN-1-YL)PROPOXY]QUINOLINE-3-CARBONITRILE",
            "non-polymer",
        ),
        "XZ2": (
            "2-(3-CHLORO-4-FLUOROBENZYL)-6,7-DIHYDROXY-N,N-DIMETHYL-1-OXO-2,3-DIHYDRO-1H-ISOINDOLE-4-SULFONAMIDE",
            "non-polymer",
        ),
        "U5P": ("URIDINE-5'-MONOPHOSPHATE", "non-polymer"),
        "XMM": (
            "(2R,3S,4S,5S,6R)-2-(5-BROMO-4-CHLORO-1H-INDOL-3-YLOXY)-TETRAHYDRO-6-(HYDROXYMETHYL)-2H-PYRAN-3,4,5-TRIOL",
            "saccharide",
        ),
        "RKA": (
            "N,N'-(1H-PYRROLE-2,5-DIYLDI-4,1-PHENYLENE)DIBENZENECARBOXIMIDAMIDE",
            "non-polymer",
        ),
        "ZAB": (
            "(3-{(Z)-[3-(AMINOMETHYL)PHENYL]DIAZENYL}PHENYL)ACETIC ACID",
            "non-polymer",
        ),
        "PPZ": ("2-(PYRIDO[1,2-E]PURIN-4-YL)AMINO-ETHANOL", "non-polymer"),
        "PPY": ("3-PHENYLPYRUVIC ACID", "non-polymer"),
        "6JZ": ("3,6,9,12,15-PENTAOXAHEPTADECANE", "non-polymer"),
        "VIB": (
            "3-(4-AMINO-2-METHYL-PYRIMIDIN-5-YLMETHYL)-5-(2-HYDROXY-ETHYL)-4-METHYL-THIAZOL-3-IUM",
            "non-polymer",
        ),
        "0G6": (
            "D-PHENYLALANYL-N-[(2S,3S)-6-{[AMINO(IMINIO)METHYL]AMINO}-1-CHLORO-2-HYDROXYHEXAN-3-YL]-L-PROLINAMIDE",
            "peptide",
        ),
        "0G7": (
            "D-PHENYLALANYL-N-[(3S)-6-CARBAMIMIDAMIDO-1-CHLORO-2-OXOHEXAN-3-YL]-L-PROLINAMIDE",
            "peptide",
        ),
        "0G5": ("N~5~-(3-IODOBENZOYL)-L-ORNITHINE", "peptide"),
        "0G2": (
            "3-[2-(5-PHENYL-2H-THIENO[3,2-C]PYRAZOL-3-YL)-1H-INDOL-6-YL]PENTAN-3-OL",
            "non-polymer",
        ),
        "0G0": (
            "1-{[1-(ALPHA-L-ARABINOFURANOSYL)-1H-1,2,3-TRIAZOL-4-YL]METHYL}-5-FLUORO-2,4-DIOXO-1,2,3,4-TETRAHYDROPYRIMIDINE",
            "non-polymer",
        ),
        "0G1": (
            "3-[3-(4-METHOXYPHENYL)-2-(1H-THIENO[3,2-C]PYRAZOL-3-YL)-1H-INDOL-6-YL]PENTAN-3-OL",
            "non-polymer",
        ),
        "NLD": ("NICOTINE-11-YL-METHYL-(4-ETHYLAMINO-4-OXO)-BUTANOATE", "non-polymer"),
        "S79": (
            "6-AMINO-2-[(1-NAPHTHYLMETHYL)AMINO]-3,7-DIHYDRO-8H-IMIDAZO[4,5-G]QUINAZOLIN-8-ONE",
            "non-polymer",
        ),
        "0LT": (
            "N-{[(3S,4S)-4-BENZYLPYRROLIDIN-3-YL]METHYL}-4-METHOXY-3-(3-METHOXYPROPOXY)-N-(PROPAN-2-YL)BENZAMIDE",
            "non-polymer",
        ),
        "LUI": (
            "2-{[3-(AMINOMETHYL)-2-(2-METHYLPROPYL)-1-OXO-4-PHENYL-1,2-DIHYDROISOQUINOLIN-6-YL]OXY}ACETAMIDE",
            "non-polymer",
        ),
        "S70": ("3-HYDROXY-2-OXO-4-PHOPHONOXY- BUTYL)-PHOSPHONIC ACID", "non-polymer"),
        "SZH": (
            "(2S,6E)-8-{[(R)-HYDROXY(PHOSPHONOOXY)PHOSPHORYL]OXY}-2,6-DIMETHYLOCT-6-EN-1-YL (2S)-3,3,3-TRIFLUORO-2-HYDRAZINOPROPANOATE",
            "non-polymer",
        ),
        "JT5": (
            "N~2~-(BIPHENYL-4-YLSULFONYL)-N-HYDROXY-N~2~-(2-HYDROXYETHYL)GLYCINAMIDE",
            "non-polymer",
        ),
        "0LP": (
            "5-(ACETYLAMINO)-2,6-ANHYDRO-3,5-DIDEOXY-3-[(2E)-3-PHENYLPROP-2-EN-1-YL]-D-GLYCERO-L-ALTRO-NON-2-ENONIC ACID",
            "non-polymer",
        ),
        "JT1": (
            "(2R)-4-(5-CYCLOPROPYL[1,3]THIAZOLO[4,5-D]PYRIMIDIN-2-YL)-N-[3-FLUORO-4-(TRIFLUOROMETHOXY)BENZYL]-1-{[4-(TRIFLUOROMETHYL)PHENYL]SULFONYL}PIPERAZINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "YT3": ("YTTRIUM (III) ION", "non-polymer"),
        "X76": ("2-[(PYRIDIN-3-YLMETHYL)AMINO]BENZAMIDE", "non-polymer"),
        "MTR": (
            "(5-METHYL-6-OXO-1,6-DIHYDRO-PYRIDIN-3-YL)-1,2-DIDEOXY-RIBOFURANOSE-5-MONOPHOSPHATE",
            "nucleic-acid",
        ),
        "YTZ": ("4-AMINO-N-(1,3-THIAZOL-2-YL)BENZENESULFONAMIDE", "non-polymer"),
        "XMI": (
            "4-[(6-CHLORO-2-NAPHTHALENYL)SULFONYL]-1-[[1-(4-PYRIDINYL)-4-PIPERIDINYL]METHYL]PIPERAZINONE",
            "non-polymer",
        ),
        "JTZ": (
            "(2S)-1-[(1-METHYLETHYL)AMINO]-3-(2-PROP-2-EN-1-YLPHENOXY)PROPAN-2-OL",
            "non-polymer",
        ),
        "YTT": ("(3S,6S)-3,6-BIS(4-HYDROXYBENZYL)PIPERAZINE-2,5-DIONE", "non-polymer"),
        "YTR": ("7,8-DIMETHOXY-1,4-DIMETHYLQUINOLIN-2(1H)-ONE", "non-polymer"),
        "JTP": (
            "N-[(13-CYCLOHEXYL-6,7-DIHYDROINDOLO[1,2-D][1,4]BENZOXAZEPIN-10-YL)CARBONYL]-2-METHYL-L-ALANINE",
            "non-polymer",
        ),
        "YTP": ("1-(4-HYDROXY-3-METHYLPHENYL)ETHANONE", "non-polymer"),
        "VX6": (
            "CYCLOPROPANECARBOXYLIC ACID {4-[4-(4-METHYL-PIPERAZIN-1-YL)-6-(5-METHYL-2H-PYRAZOL-3-YLAMINO)-PYRIMIDIN-2-YLSULFANYL]-PHENYL}-AMIDE",
            "non-polymer",
        ),
        "SZ8": ("METHYL (4R)-4-AMINO-1-METHYL-L-PROLINATE", "non-polymer"),
        "SZ9": ("1-(2,5-DIMETHYLFURAN-3-YL)METHANAMINE", "non-polymer"),
        "S61": (
            "(S)-3-{[1-(2-FLUORO-PHENYL)-5-HYDROXY-1H-PYRAZOLE-3-CARBONYL]-AMINO}-3-O-TOLYL-PROPIONIC ACID",
            "non-polymer",
        ),
        "0LY": (
            "N-(4-BROMOPHENYL)-N'-(2,2,6,6-TETRAMETHYLPIPERIDIN-4-YL)ETHANEDIAMIDE",
            "non-polymer",
        ),
        "SZ3": ("TRANS-4-AMINOCYCLOHEXANOL", "non-polymer"),
        "SZ0": (
            "(2S,4R)-1-[(2R,3S)-3-AMINO-2-HYDROXY-4-PHENYLBUTYL]-4-(PYRIDIN-4-YLSULFONYL)PIPERIDINE-2-CARBOXYLIC ACID",
            "peptide",
        ),
        "SZ1": ("CYCLOHEPTANAMINE", "non-polymer"),
        "YTB": ("(1R,2R,3S,4R,5R)-5-AMINOCYCLOPENTANE-1,2,3,4-TETROL", "non-polymer"),
        "VTQ": ("RRR-ALPHA-TOCOPHERYLQUINONE", "non-polymer"),
        "SZ4": ("1-(3-METHOXYPHENYL)METHANAMINE", "non-polymer"),
        "PP3": ("ALANYL-PYRIDOXAL-5'-PHOSPHATE", "non-polymer"),
        "NPL": ("N-METHYL-4-DEOXY-4-AMINO-PYRIDOXAL-5-PHOSPHATE", "non-polymer"),
        "S7B": ("2-[(CYCLOPROPYLCARBONYL)OXY]-3-METHYLBENZOIC ACID", "non-polymer"),
        "Z94": (
            "N-{4-[(4S)-2,5-DIOXOIMIDAZOLIDIN-4-YL]PHENYL}ACETAMIDE",
            "non-polymer",
        ),
        "VX1": (
            "4-[3-(4-CHLOROPHENYL)-2,1-BENZISOXAZOL-5-YL]PYRIMIDIN-2-AMINE",
            "non-polymer",
        ),
        "PP0": (
            "ETHYL 4-[(4-BENZYLPIPERAZIN-1-YL)CARBONYL]-1-ETHYL-3,5-DIMETHYL-1H-PYRROLE-2-CARBOXYLATE",
            "non-polymer",
        ),
        "TML": ("METHYL PART OF N-TRIMETHYLLYSINE", "non-polymer"),
        "PP7": (
            "METHYL CYCLO[(2S)-2-[[(1R)-1-(N-(L-N-(3-METHYLBUTANOYL)VALYL-L-ASPARTYL)AMINO)-3-METHYLBUTYL]HYDROXYPHOSPHINYLOXY]-3-(3-AMINOMETHYL)PHENYLPROPANOATE",
            "non-polymer",
        ),
        "PP6": (
            "METHYL[CYCLO-7[(2R)-((N-VALYL)AMINO)-2-(HYDROXYL-(1S)-1-METHYLOXYCARBONYL-2-PHENYLETHOXY)PHOSPHINYLOXY-ETHYL]-1-NAPHTHALENEACETAMIDE]",
            "non-polymer",
        ),
        "PCZ": (
            "(2R)-2-[(1R)-1-{[(2Z)-2-(2-AMINO-1,3-THIAZOL-4-YL)-2-(METHOXYIMINO)ACETYL]AMINO}-2-OXOETHYL]-5-METHYLIDENE-5,6-DIHYDRO-2H-1,3-THIAZINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "XMK": (
            "1-[[(1E)-2-(4-CHLOROPHENYL)ETHENYL]SULFONYL]-4-[[1-(4-PYRIDINYL)-4-PIPERIDINYL]METHYL]PIPERAZINE",
            "non-polymer",
        ),
        "PP5": (
            "METHYL (2S)-[1-((N-(NAPHTHALENEACETYL))-L-VALYL)AMINOMETHYL)HYDROXYPHOSPHINYLOXY]-3-PHENYL PROPANOATE",
            "non-polymer",
        ),
        "PCX": ("DEOXYGUANIDINOPROCLAVAMINIC ACID", "non-polymer"),
        "HV8": ("BENZYL-2-AMINO-PARAMETHOXY-BENZYLSTATINE", "peptide"),
        "CML": (
            "(2S)-2-{[(2R)-2-AMINO-2-CARBOXYETHYL]SULFANYL}BUTANEDIOIC ACID",
            "peptide",
        ),
        "TMM": ("1,3,5-BENZENETRICARBOXYLIC ACID", "non-polymer"),
        "HV1": (
            "METHYL [(1S)-1-({2-[(4R)-4-BENZYL-4-HYDROXY-5-{[(1S,2R)-2-HYDROXY-2,3-DIHYDRO-1H-INDEN-1-YL]AMINO}-5-OXOPENTANOYL]-2-(4-BROMOBENZYL)HYDRAZINO}CARBONYL)-2,2-DIMETHYLPROPYL]CARBAMATE ",
            "non-polymer",
        ),
        "IN8": (
            "[2-(5-MERCAPTO-[1,3,4]THIADIAZOL-2-YLCARBAMOYL)-1-PHENYL-ETHYL]-CARBAMIC ACID BENZYL ESTER",
            "non-polymer",
        ),
        "HV5": ("TERT-BUTYLALANINE", "peptide"),
        "PP9": ("PROTOPORPHYRIN IX", "non-polymer"),
        "HV7": ("1-METHYLAMINE-2-HYDROXY-4-METHOXY-BENZENE", "non-polymer"),
        "IN9": (
            "2-[3-(5-MERCAPTO-[1,3,4]THIADIAZOL-2YL)-UREIDO]-N-METHYL-3-PENTAFLUOROPHENYL-PROPIONAMIDE",
            "non-polymer",
        ),
        "A33": ("2-(CARBOXYMETHYL)-D-ASPARTIC ACID", "non-polymer"),
        "A32": (
            "(3-NITRO-5-(3-MORPHOLIN-4-YL-PROPYLAMINOCARBONYL)PHENYL)-GALACTOPYRANOSIDE",
            "non-polymer",
        ),
        "A31": (
            "N-{15-OXO-19-[(3AS,4S,6AR)-2-OXOHEXAHYDRO-1H-THIENO[3,4-D]IMIDAZOL-4-YL]-4,7,10-TRIOXA-14-AZANONADEC-1-YL}-N'-(8-{[(3S,5S,7S)-TRICYCLO[3.3.1.1~3,7~]DEC-1-YLCARBONYL]AMINO}OCTYL)PENTANEDIAMIDE",
            "non-polymer",
        ),
        "NPN": ("N-PROPYL ISOCYANIDE", "non-polymer"),
        "A37": (
            "2,5-DICHLORO-N-(5-CHLORO-1,3-BENZOXAZOL-2-YL)BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "TL6": (
            "(5R)-5-[(1E)-BUTA-1,3-DIENYL]-4-HYDROXY-3,5-DIMETHYLTHIOPHEN-2(5H)-ONE",
            "non-polymer",
        ),
        "A35": ("?", "nucleic-acid"),
        "A34": ("?", "nucleic-acid"),
        "A39": ("?", "nucleic-acid"),
        "A38": ("?", "nucleic-acid"),
        "S91": (
            "[4-({5-(AMINOCARBONYL)-4-[(3-METHYLPHENYL)AMINO]PYRIMIDIN-2-YL}AMINO)PHENYL]ACETIC ACID",
            "non-polymer",
        ),
        "GT0": (
            "5-(HYDROXYMETHYL)-4-(METHOXYMETHYL)-2-METHYLPYRIDIN-3-OL",
            "non-polymer",
        ),
        "CMG": ("6-O-CYCLOHEXYLMETHYL GUANINE", "non-polymer"),
        "TMN": ("TRIS(HYDROXYMETHYL)AMINOMETHANE", "non-polymer"),
        "Z98": (
            "(2S)-2-AMINO-4-{[(1Z)-{3-HYDROXY-2-METHYL-5-[(PHOSPHONOOXY)METHYL]PYRIDIN-4-YL}METHYLIDENE]AMINO}BUTANOIC ACID",
            "non-polymer",
        ),
        "IN5": (
            "{1-[(3-HYDROXY-METHYL-5-PHOSPHONOOXY-METHYL-PYRIDIN-4-YLMETHYL)-AMINO]-ETHYL}-PHOSPHONIC ACID",
            "non-polymer",
        ),
        "GT9": ("S-NONYL-CYSTEINE", "peptide"),
        "JQ1": (
            "(6S)-6-(2-TERT-BUTOXY-2-OXOETHYL)-4-(4-CHLOROPHENYL)-2,3,9-TRIMETHYL-6,7-DIHYDROTHIENO[3,2-F][1,2,4]TRIAZOLO[4,3-A][1,4]DIAZEPIN-10-IUM",
            "non-polymer",
        ),
        "IN6": (
            "N-[2-[1-(N-BENZYLOXYCARBONYLAMINO)-3-METHYLBUTYL]THIAZOL-4-YLCARBONYL]-N'-(BENZYLOXYCARBONYL-L-LEUCINYL)HYDRAZIDE",
            "non-polymer",
        ),
        "A3S": ("SERINE-3'-AMINOADENOSINE", "non-polymer"),
        "GTA": ("P1-7-METHYLGUANOSINE-P3-ADENOSINE-5',5'-TRIPHOSPHATE", "non-polymer"),
        "GTB": ("S-(P-NITROBENZYL)GLUTATHIONE", "non-polymer"),
        "A3P": ("ADENOSINE-3'-5'-DIPHOSPHATE", "nucleic-acid"),
        "GTD": ("1-(S-GLUTATHIONYL)-2,4,6-TRINITROCYCLOHEXA-2,5-DIENE", "non-polymer"),
        "GTE": ("GLUCOSE 4-O4 GROUP", "saccharide"),
        "GTF": ("?", "non-polymer"),
        "A3T": ("3'-DEOXY-3'-(L-THREONYLAMINO)ADENOSINE", "non-polymer"),
        "GTH": (
            "3-O-GLUCOPYRANOSYL-THREONINE-[2-DEOXY-2-ACETAMIDO-GLUCOPYRANOSIDE]",
            "saccharide",
        ),
        "CMC": ("CARBOXYMETHYL COENZYME *A", "non-polymer"),
        "TLY": ("(2S)-2-AZANYL-6-(ETHANETHIOYLAMINO)HEXANOIC ACID", "peptide"),
        "GTK": ("BETA-D-GALACTOFURANURONIC ACID", "saccharide"),
        "GTL": (
            "(2E,3R,4R,5R,6S)-3,4,5-TRIHYDROXY-6-(HYDROXYMETHYL)-2-PIPERIDINONE",
            "non-polymer",
        ),
        "GTM": ("O1-METHYL-4-DEOXY-4-THIO-BETA-D-GLUCOSE", "non-polymer"),
        "GTN": ("PHOSPHOAMINOPHOSPHONIC ACID GUANYLATE ESTER", "non-polymer"),
        "GTO": ("PHOSPHOMETHYLPHOSPHONIC ACID-GUANYLATE ESTER", "non-polymer"),
        "A3C": (
            "2-(METHYL{2-[(1S,4AS,8AS)-5,5,8A-TRIMETHYL-2-METHYLIDENEDECAHYDRONAPHTHALEN-1-YL]ETHYL}AMINO)ETHYL TRIHYDROGEN DIPHOSPHATE",
            "non-polymer",
        ),
        "A3B": ("D-VINYLGLYCINE", "non-polymer"),
        "A3A": ("2'DEOXY-ALPHA-ANOMERIC-ADENOSINE-5'-PHOSPHATE", "nucleic-acid"),
        "GTS": ("GLUTATHIONE SULFONIC ACID", "non-polymer"),
        "A3G": ("3'-DEOXY-3'-(GLYCYLAMINO)ADENOSINE", "non-polymer"),
        "A3F": (
            "3-[6-AMINO-5-(3,4,5-TRIMETHOXYPHENYL)PYRIDIN-3-YL]PHENOL",
            "non-polymer",
        ),
        "A3D": ("3-ACETYLPYRIDINE ADENINE DINUCLEOTIDE", "non-polymer"),
        "GTX": ("S-HEXYLGLUTATHIONE", "non-polymer"),
        "GTY": ("L-GAMMA-GLUTAMYL-S-OCTYL-D-CYSTEINYLGLYCINE", "non-polymer"),
        "GTZ": (
            "(5R, 6S, 7S, 8S)-5-HYDROXYMETHYL-6,7,8-TRIHYDROXY-TETRAZOLO[1,5-A]PIPERIDINE",
            "non-polymer",
        ),
        "IN3": (
            "1-(N-BENZYLOXYCARBONYL-L-LEUCINYL)-5-(3-BENZYLOXY BENZOYL)CARBOHYDRAZIDE",
            "non-polymer",
        ),
        "TLO": (
            "5'-O-[(S)-HYDROXY{[(S)-HYDROXY(PHENOXY)PHOSPHORYL]OXY}PHOSPHORYL]THYMIDINE",
            "non-polymer",
        ),
        "A3N": ("5'-DEOXY-5'-(PROP-2-YN-1-YLAMINO)ADENOSINE", "non-polymer"),
        "A3M": ("2-AMINO-3-METHYL-1-PYRROLIDIN-1-YL-BUTAN-1-ONE", "non-polymer"),
        "PCJ": (
            "(2R)-3-{[(2S)-3-HYDROXY-2-(PALMITOYLAMINO)PROPYL]THIO}PROPANE-1,2-DIYL DIHEXADECANOATE",
            "non-polymer",
        ),
        "S98": (
            "6-AMINO-2-[(THIOPHEN-2-YLMETHYL)AMINO]-1,7-DIHYDRO-8H-IMIDAZO[4,5-G]QUINAZOLIN-8-ONE",
            "non-polymer",
        ),
        "PCK": (
            "(7R,18S,19R)-18,19-DIBROMO-7-{[(9S,10S)-9,10-DIBROMOOCTADECANOYL]OXY}-4-HYDROXY-N,N,N-TRIMETHYL-10-OXO-3,5,9-TRIOXA-4-PHOSPHAHEPTACOSAN-1-AMINIUM 4-OXIDE",
            "non-polymer",
        ),
        "HVA": ("3-HYDROXY-L-VALINE", "peptide"),
        "PCH": ("3-AMINO-4-CYCLOHEXYL-1-(PYRIDIN-4-YLOXY)BUTAN-2-OL", "non-polymer"),
        "ILF": ("?", "non-polymer"),
        "ARX": (
            "(2S,3S,4R,5R)-5-(6-AMINO-9H-PURIN-9-YL)-N-(6-{[(1R)-4-CARBAMIMIDAMIDO-1-{[(1R)-4-CARBAMIMIDAMIDO-1-CARBAMOYLBUTYL]CARBAMOYL}BUTYL]AMINO}-6-OXOHEXYL)-3,4-DIHYDROXYTETRAHYDROFURAN-2-CARBOXAMIDE",
            "peptide",
        ),
        "PCI": ("PENTACHLOROPHENOL", "non-polymer"),
        "PCN": (
            "2-{1-[2-(2-AMINO-THIAZOL-4-YL)-2-METHOXYIMINO-ACETYLAMINO]-2-OXO-ETHYL}-5,5-DIMETHYL-THIAZOLIDINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "ARZ": (
            "(3S,5S,7S)-N-(5-SULFAMOYL-1,3,4-THIADIAZOL-2-YL)TRICYCLO[3.3.1.1~3,7~]DECANE-1-CARBOXAMIDE",
            "non-polymer",
        ),
        "S41": (
            "N,N-DIMETHYL-5-({2-METHYL-6-[(5-METHYLPYRAZIN-2-YL)CARBAMOYL]-1-BENZOFURAN-4-YL}OXY)PYRIMIDINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "PCO": (
            "2,4-DIHYDROXY-N-[2-(2-MERCAPTO-VINYLCARBAMOYL)-ETHYL]-3,3-DIMETHYL-BUTYRAMIDE",
            "non-polymer",
        ),
        "4BY": ("5-METHYL-1H-INDOLE-2-CARBOXYLIC ACID", "non-polymer"),
        "3P3": (
            "N-[(1S,2R)-2-HYDROXY-1-(HYDROXYCARBAMOYL)PROPYL]-4-(4-PHENYLBUTA-1,3-DIYN-1-YL)BENZAMIDE",
            "non-polymer",
        ),
        "F03": ("N-(4-OXO-3H-QUINAZOLIN-2-YL)ACETAMIDE", "non-polymer"),
        "4BX": ("3-(1H-BENZIMIDAZOL-2-YL)PROPANOIC ACID", "non-polymer"),
        "F02": ("4-(DIETHYLAMINO)BENZOHYDRAZIDE", "non-polymer"),
        "ILG": ("GLUTAMYL GROUP", "peptide"),
        "PCB": (
            "4,4'-BIS([H]METHYLSULFONYL)-2,2',5,5'-TETRACHLOROBIPHENYL",
            "non-polymer",
        ),
        "LG8": (
            "3-(8-DIMETHYLAMINOMETHYL-6,7,8,9-TETRAHYDRO-PYRIDO[1,2-A]INDOL-10-YL)-4-(1-METHYL-1H-INDOL-3-YL)-PYRROLE-2,5-DIONE",
            "non-polymer",
        ),
        "LG9": (
            "(1R,2R,3R,6R,7R,7AR)-3,7-BIS(HYDROXYMETHYL)HEXAHYDRO-1H-PYRROLIZINE-1,2,6-TRIOL",
            "non-polymer",
        ),
        "PCC": ("5-OXOPROLINE", "peptide"),
        "LG2": (
            "6-[1-(3,5,5,8,8-PENTAMETHYL-5,6,7,8-TETRAHYDRONAPHTHALEN-2-YL)CYCLOPROPYL]PYRIDINE-3-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "LG3": ("PYRIMIDINE-2,4-DIAMINE", "non-polymer"),
        "LG0": (
            "1-(4-METHOXYPHENYL)-6-(4-(1-(PYRROLIDIN-1-YLMETHYL)CYCLOPROPYL)PHENYL)-3-(TRIFLUOROMETHYL)-5,6-DIHYDRO-1H-PYRAZOLO[3,4-C]PYRIDIN-7(4H)-ONE",
            "non-polymer",
        ),
        "LG1": (
            "3,11-DIFLUORO-6,8,13-TRIMETHYL-8H-QUINO[4,3,2-KL]ACRIDIN-13-IUM",
            "non-polymer",
        ),
        "LG6": ("L-GULURONIC ACID 6-PHOSPHATE", "non-polymer"),
        "LG7": (
            "4-[(7R,7AS)-7-HYDROXY-1,3-DIOXOTETRAHYDRO-1H-PYRROLO[1,2-C]IMIDAZOL-2(3H)-YL]-1-NAPHTHONITRILE",
            "non-polymer",
        ),
        "LG4": ("5-METHYLPYRIDIN-2-AMINE", "non-polymer"),
        "LG5": ("1-METHYL-1,6-DIHYDROPYRIDIN-3-AMINE", "non-polymer"),
        "PCA": ("PYROGLUTAMIC ACID", "peptide"),
        "RB9": (
            "N-BENZYL-2-({N-[2-(1H-INDOL-3-YL)ETHYL]GLYCYL}AMINO)-4-PHENYLTHIOPHENE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "UCP": ("PHOSPHOMETHYL PHOSPHONIC ACID DEOXYURIDYLATE ESTER", "non-polymer"),
        "3EK": (
            "TRANS-4-[(3-{2-[(4-FLUOROBENZYL)CARBAMOYL]-6-METHYLPYRIDIN-4-YL}-1H-1,2,4-TRIAZOL-1-YL)METHYL]CYCLOHEXANECARBOXYLIC ACID",
            "non-polymer",
        ),
        "ILD": (
            "(4R,5S)-3-[(2R)-2-{(2S,2'R,4S,5S,5'R)-2,5'-DIETHYL-5'-[(1S)-1-HYDROXYETHYL]-4-METHYLOCTAHYDRO-2,2'-BIFURAN-5-YL}BUTANOYL]-4-METHYL-5-PHENYL-1,3-OXAZOLIDIN-2-ONE",
            "non-polymer",
        ),
        "RB4": (
            "ETHYL 2-({N-[2-(1H-INDOL-3-YL)ETHYL]GLYCYL}AMINO)-4-PHENYLTHIOPHENE-3-CARBOXYLATE",
            "non-polymer",
        ),
        "RB3": (
            "(1R)-3-{[(1R)-3-METHOXY-1-METHYL-3-OXOPROPYL]OXY}-1-METHYL-3-OXOPROPYL (3R)-3-HYDROXYBUTANOATE",
            "non-polymer",
        ),
        "RB2": (
            "2-({4-CHLORO-2-[HYDROXY(METHOXY)METHYL]CYCLOHEXYL}AMINO)ETHANE-1,1,2-TRIOL",
            "non-polymer",
        ),
        "RB1": (
            "3-[(4-AMINO-1-TERT-BUTYL-1H-PYRAZOLO[3,4-D]PYRIMIDIN-3-YL)METHYL]PHENOL",
            "non-polymer",
        ),
        "2Y3": (
            "N-(TERT-BUTOXYCARBONYL)-L-PHENYLALANYL-N-[(2S,3S,5R)-1-CYCLOHEXYL-3-HYDROXY-7-METHYL-5-(METHYLCARBAMOYL)OCTAN-2-YL]-L-HISTIDINAMIDE",
            "peptide",
        ),
        "77D": (
            "2-[(2-FLUORO-4-IODOPHENYL)AMINO]-5,5-DIMETHYL-8-OXO-N-[(3R)-PIPERIDIN-3-YL]-5,6,7,8-TETRAHYDRO-4H-THIENO[2,3-C]AZEPINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "77F": (
            "N-[(2S,3R)-4-{(CYCLOHEXYLMETHYL)[(4-METHOXYPHENYL)SULFONYL]AMINO}-3-HYDROXY-1-PHENYLBUTAN-2-YL]-3-HYDROXYBENZAMIDE",
            "non-polymer",
        ),
        "77A": (
            "1-(5-CHLORO-2-METHOXYPHENYL)-3-{6-[2-(DIMETHYLAMINO)-1-METHYLETHOXY]PYRAZIN-2-YL}UREA",
            "non-polymer",
        ),
        "CMQ": (
            "N~2~-[(BENZYLOXY)CARBONYL]-N-[(1S,2S)-2-HYDROXY-1-(4-HYDROXYBENZYL)PROPYL]-L-LEUCINAMIDE",
            "non-polymer",
        ),
        "18K": (
            "N'-[4-(2-AMINO-4-METHYL-1,3-THIAZOL-5-YL)PYRIMIDIN-2-YL]-N,N-DIMETHYLBENZENE-1,4-DIAMINE",
            "non-polymer",
        ),
        "EJ4": (
            "(4BS,8R,8AS,14BR)-7-(CYCLOPROPYLMETHYL)-5,6,7,8,14,14B-HEXAHYDRO-4,8-METHANO[1]BENZOFURO[2,3-A]PYRIDO[4,3-B]CARBAZOLE-1,8A(9H)-DIOL",
            "non-polymer",
        ),
        "EJ5": (
            "4-[3-(2-AMINO-4-HYDROXY-6-OXO-1,6-DIHYDROPYRIMIDIN-5-YL)PROPYL]BENZOIC ACID",
            "non-polymer",
        ),
        "UCM": (
            "REL-(9R,12S)-9,10,11,12-TETRAHYDRO-9,12-EPOXY-1H-DIINDOLO[1,2,3-FG:3',2',1'-KL]PYRROLO[3,4-I][1,6]BENZODIAZOCINE-1,3(2H)-DIONE",
            "non-polymer",
        ),
        "F5M": (
            "5-(2,5-DIOXO-2,5-DIHYDRO-1H-PYRROL-1-YL)-2-(6-HYDROXY-3-OXO-3H-XANTHEN-9-YL)BENZOIC ACID",
            "non-polymer",
        ),
        "UCK": (
            "(2S,4R)-1-(2-CHLOROPHENYL)CARBONYL-N-[(4-CHLOROPHENYL)METHYL]-4-OXIDANYL-PYRROLIDINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "WP6": (
            "(4S)-3-BENZYL-6-CHLORO-2-METHYL-4-PHENYL-3,4-DIHYDROQUINAZOLINE",
            "non-polymer",
        ),
        "F5H": (
            "3,7-ANHYDRO-1,5,6-TRIDEOXY-5-FLUORO-6-(5-METHYL-2,4-DIOXO-3,4-DIHYDROPYRIMIDIN-1(2H)-YL)-2-O-PHOSPHONO-D-GLYCERO-L-ALTRO-HEPTITOL",
            "nucleic-acid",
        ),
        "UC5": (
            "2'-DEOXY-5'-O-[(S)-HYDROXY(PHOSPHONOMETHYL)PHOSPHORYL]URIDINE",
            "non-polymer",
        ),
        "F55": (
            "N-{[(4-CHLOROPHENYL)CARBONYL]CARBAMOYL}-BETA-D-GLUCOPYRANOSYLAMINE",
            "non-polymer",
        ),
        "771": (
            "4-(4-BENZYLOXY-2-METHANESULFONYLAMINO-5-METHOXY-BENZYLAMINO)-BENZAMIDINE",
            "non-polymer",
        ),
        "770": (
            "N-[4-(2-CHLOROPHENYL)-1,3-DIOXO-1,2,3,6-TETRAHYDROPYRROLO[3,4-C]CARBAZOL-9-YL]FORMAMIDE",
            "non-polymer",
        ),
        "773": ("CETHROMYCIN", "non-polymer"),
        "772": ("2-CYANOQUINOLIN-8-YL DIHYDROGEN PHOSPHATE", "non-polymer"),
        "4BU": (
            "(2S)-2-AMINO-5-[[4-[(2S)-2-HYDROXY-2-PHENYL-ETHOXY]PHENYL]AMINO]-5-OXO-PENTANOIC ACID",
            "non-polymer",
        ),
        "RBU": ("?", "non-polymer"),
        "RBT": ("RIFABUTIN", "non-polymer"),
        "RBS": ("N-(4-PHENYLAMINO-QUINAZOLIN-6-YL)-ACRYLAMIDE", "non-polymer"),
        "778": (
            "4-[(5-{[4-(3-CHLOROPHENYL)-3-OXOPIPERAZIN-1-YL]METHYL}-1H-IMIDAZOL-1-YL)METHYL]BENZONITRILE",
            "non-polymer",
        ),
        "4BT": (
            "2-{2-[4-(2-CARBAMIMIDOYLSULFANYL-ETHYL)-PHENYL]-ETHYL}-ISOTHIOUREA",
            "non-polymer",
        ),
        "OME": ("METHOXY GROUP", "non-polymer"),
        "XMB": (
            "4-[(2R)-3-[[(6-CHLORO-2-NAPHTHALENYL)SULFONYL]AMINO]-1-OXO-2-[[[1-(4-PYRIDINYL)-4-PIPERIDINYL]METHYL]AMINO]PROPYL]-THIOMORPHOLINE-1,1-DIOXIDE",
            "non-polymer",
        ),
        "EJZ": (
            "5-NITRO-2,6-DIOXO-1,2,3,6-TETRAHYDROPYRIMIDINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "EJT": (
            "1,3-DI(N-PROPYLOXY-A-MANNOPYRANOSYL)-CARBOMYL 5-METHYAZIDO-BENZENE",
            "saccharide",
        ),
        "RBF": ("RIBOFLAVIN", "non-polymer"),
        "RBE": (
            "3-(BENZOYLOXY)-8-AZA-BICYCLO[3.2.1]OCTANE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "S9L": (
            "2-[2-(2-HYDROXYETHOXY)ETHOXY]ETHYL DIHYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "RBC": (
            "4-[(4R,5R)-3-(METHOXYMETHYL)-5-(2-NAPHTHYLMETHOXY)PIPERIDIN-4-YL]BENZYL 2-CHLOROBENZOATE",
            "non-polymer",
        ),
        "LGJ": (
            "N-(3-FLUORO-2'-(METHYLSULFONYL)BIPHENYL-4-YL)-1-(3-(5-OXO-4,5-DIHYDRO-1H-1,2,4-TRIAZOL-3-YL)PHENYL)-3-(TRIFLUOROMETHYL)-1H-PYRAZOLE-5-CARBOXAMIDE",
            "non-polymer",
        ),
        "LGK": (
            "6-(2'-(METHYLSULFONYL)BIPHENYL-4-YL)-1-(3-(5-OXO-4,5-DIHYDRO-1H-1,2,4-TRIAZOL-3-YL)PHENYL)-3-(TRIFLUOROMETHYL)-5,6-DIHYDRO-1H-PYRAZOLO[3,4-C]PYRIDIN-7(4H)-ONE",
            "non-polymer",
        ),
        "LGH": (
            "[4-(2-CHLORO-4-NITROPHENYL)PIPERAZIN-1-YL][3-(2-METHOXYPHENYL)-5-METHYL-1,2-OXAZOL-4-YL]METHANONE",
            "non-polymer",
        ),
        "LGN": (
            "N-[(2S,3R,4E)-1-{[ALPHA-D-GALACTOPYRANOSYL-(1->3)-BETA-D-GALACTOPYRANOSYL-(1->4)-BETA-D-GLUCOPYRANOSYL]OXY}-3-HYDROXYOCTADEC-4-EN-2-YL]HEXACOSANAMIDE",
            "non-polymer",
        ),
        "ILB": ("?", "non-polymer"),
        "4BH": (
            "4-CHLORO-6-(4-PIPERAZIN-1-YL-1H-PYRAZOL-5-YL)BENZENE-1,3-DIOL",
            "non-polymer",
        ),
        "LGB": (
            "2-CHLORO-4-{[(1R,3Z,7S,7AS)-7-HYDROXY-1-(TRIFLUOROMETHYL)TETRAHYDRO-1H-PYRROLO[1,2-C][1,3]OXAZOL-3-YLIDENE]AMINO}-3-METHYLBENZONITRILE",
            "non-polymer",
        ),
        "LGC": (
            "(3S,4R,5R,6S)-3,4,5-TRIHYDROXY-6-(HYDROXYMETHYL)TETRAHYDRO-2H-PYRAN-2-ONE",
            "saccharide",
        ),
        "TDT": (
            "1-AZEPAN-1-YL-2-PHENYL-2-(4-THIOXO-1,4-DIHYDRO-PYRAZOLO[3,4-D]PYRIMIDIN-5-YL)ETHANONE ADDUCT",
            "non-polymer",
        ),
        "LGA": ("PYRIMIDIN-2-AMINE", "non-polymer"),
        "LGF": (
            "1-[(3AR)-5-{[3-(2,4-DIFLUOROPHENYL)-5,6-DIHYDROIMIDAZO[1,5-A]PYRAZIN-7(8H)-YL]CARBONYL}-6-METHOXY-3AH-PYRROLO[2,3-B]PYRIDIN-3-YL]-2-[(3R)-3-HYDROXYPYRROLIDIN-1-YL]-2-OXOETHANONE",
            "non-polymer",
        ),
        "LGG": (
            "N-ACETYL-L-ALPHA-GLUTAMYL-L-PHENYLALANYL-L-GLUTAMINYL-N-[(1S)-4-AMINO-1-(2-CARBOXYETHYL)-4-OXOBUTYL]-L-LEUCINAMIDE",
            "non-polymer",
        ),
        "LGD": (
            "6-[BIS(2,2,2-TRIFLUOROETHYL)AMINO]-4-(TRIFLUOROMETHYL)QUINOLIN-2(1H)-ONE",
            "non-polymer",
        ),
        "LGE": (
            "METHYL [3-(AMINOMETHYL)-4-(2,4-DICHLOROPHENYL)-2-METHYL-7-OXO-5,7-DIHYDRO-6H-PYRROLO[3,4-B]PYRIDIN-6-YL]ACETATE",
            "non-polymer",
        ),
        "LGZ": (
            "CARBAMIC ACID 1-{5-BENZYL-5-[2-HYDROXY-4-PHENYL-3-(TETRAHYDRO-FURAN- 3-YLOXYCARBONYLAMINO)-BUTYL]-4-OXO-4,5-DIHYDRO-1H-PYRROL-3-YL}- INDAN-2-YL ESTER",
            "non-polymer",
        ),
        "UBX": (
            "N-[(S)-({[(BENZYLOXY)CARBONYL]AMINO}METHYL)(HYDROXY)PHOSPHORYL]-N-METHYL-L-LEUCINAMIDE",
            "peptide",
        ),
        "LGX": (
            "N-(5-CYCLOPROPYL-1H-PYRAZOL-3-YL)-2-{(2R)-1-[(6-FLUOROPYRIDIN-3-YL)CARBONYL]PYRROLIDIN-2-YL}PYRROLO[2,1-F][1,2,4]TRIAZIN-4-AMINE",
            "non-polymer",
        ),
        "LGY": ("(E)-N~6~-(4-OXOBUTYLIDENE)-L-LYSINE", "peptide"),
        "LGS": (
            "(3Z,5S,6R,7S,8R,8AR)-3-(OCTYLIMINO)HEXAHYDRO[1,3]OXAZOLO[3,4-A]PYRIDINE-5,6,7,8-TETROL",
            "non-polymer",
        ),
        "LGP": (
            "N9-1-HYDROXY-PROP-2-OXYMETHYL-GUANINE-3'-MONOPHOSPHATE",
            "nucleic-acid",
        ),
        "LGV": (
            "N-({4-[(3-CYCLOPROPYL-1H-PYRAZOL-5-YL)AMINO]PYRROLO[2,1-F][1,2,4]TRIAZIN-2-YL}METHYL)-6-FLUOROPYRIDINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "LGW": (
            "N-(5-CYCLOPROPYL-1H-PYRAZOL-3-YL)-2-{(2S)-1-[(6-FLUOROPYRIDIN-3-YL)CARBONYL]PYRROLIDIN-2-YL}PYRROLO[2,1-F][1,2,4]TRIAZIN-4-AMINE",
            "non-polymer",
        ),
        "LGT": ("L-GLUCARIC ACID", "non-polymer"),
        "4BL": ("6-METHYL-1H-INDOLE-2-CARBOXYLIC ACID", "non-polymer"),
        "DZA": ("DAMINOZIDE", "non-polymer"),
        "04S": ("(3R)-PYRROLIDIN-3-AMINE", "non-polymer"),
        "DZE": (
            "METHYL (3S)-3-[(TERT-BUTOXYCARBONYL)AMINO]-4-OXOPENTANOATE",
            "non-polymer",
        ),
        "18T": (
            "[(2R,3S,5R)-3-HYDROXY-5-(5-METHYL-2,4-DIOXO-3,4-DIHYDROPYRIMIDIN-1(2H)-YL)TETRAHYDROFURAN-2-YL]METHYL (2R,3R,4S,5S,6R)-3,4,5-TRIHYDROXY-6-METHYLTETRAHYDRO-2H-PYRAN-2-YL DIHYDROGEN DIPHOSPHATE",
            "saccharide",
        ),
        "DZG": (
            "1-(2,3-DIHYDRO-1,4-BENZODIOXIN-6-YLSULFONYL)-4-[(4-METHOXYPHENYL)SULFONYL]PIPERAZINE",
            "non-polymer",
        ),
        "DZF": ("5-DEAZAFOLIC ACID", "non-polymer"),
        "DZM": ("3-DEAZA-3-METHYLADENINE", "nucleic-acid"),
        "DZO": (
            "3-(3-AMINOPHENYL)-N-(3-CHLOROPHENYL)PYRAZOLO[1,5-A]PYRIMIDIN-5-AMINE",
            "non-polymer",
        ),
        "DZN": ("DAIDZIN", "non-polymer"),
        "4BA": ("4-[(2-CARBOXY-ETHYLAMINO)-METHYL]-BENZOIC ACID", "non-polymer"),
        "PH9": (
            "(2Z)-2-[(1-ADAMANTYLCARBONYL)AMINO]-3-[4-(2-BROMOPHENOXY)PHENYL]PROP-2-ENOIC ACID",
            "non-polymer",
        ),
        "GSR": ("?", "nucleic-acid"),
        "PEO": ("HYDROGEN PEROXIDE", "non-polymer"),
        "PH3": ("N-(CHLOROPHENYL)-N'-HYDROXYGUANIDINE", "non-polymer"),
        "PH2": ("2-AMINO-6-HYDROXYMETHYL-7,8-DIHYDRO-3H-PTERIDIN-4-ONE", "non-polymer"),
        "PH1": (
            "1,2-[DI-2,6,10,14-TETRAMETHYL-HEXADECAN-16-OXY]-PROPANE",
            "non-polymer",
        ),
        "DZZ": ("O,O-DIMETHYL HYDROGEN THIOPHOSPHATE", "non-polymer"),
        "PH7": (
            "(2Z)-2-(BENZOYLAMINO)-3-[4-(2-BROMOPHENOXY)PHENYL]-2-PROPENOIC ACID",
            "non-polymer",
        ),
        "PH5": ("2-PHENYL-PROP5AC", "non-polymer"),
        "GST": ("GERANYL S-THIOLODIPHOSPHATE", "non-polymer"),
        "04R": (
            "[3-(TRIFLUOROMETHYL)-4,5,6,7-TETRAHYDRO-1H-INDAZOL-1-YL]ACETIC ACID",
            "peptide",
        ),
        "TIZ": (
            "(1R)-1-[4-(AMINOCARBONYL)-1,3-THIAZOL-2-YL]-1,4-ANHYDRO-D-RIBITOL",
            "non-polymer",
        ),
        "GSV": (
            "2-(5-CHLOROTHIOPHEN-2-YL)-N-{(3S)-1-[3-FLUORO-2'-(METHYLSULFONYL)BIPHENYL-4-YL]-2-OXOPYRROLIDIN-3-YL}ETHANESULFONAMIDE",
            "non-polymer",
        ),
        "OF2": ("2 FERRIC ION, 1 BRIDGING OXYGEN", "non-polymer"),
        "GSY": ("7-(2-HYDROXYETHYL)-8-MERCAPTOGUANINE", "non-polymer"),
        "OF1": ("FERROUS ION, 1 WATER COORDINATED", "non-polymer"),
        "IA9": (
            "2-CHLORO-4-FLUORO-3-METHYL-N-[2-(4-METHYLPIPERAZIN-1-YL)-5-NITROPHENYL]BENZAMIDE",
            "non-polymer",
        ),
        "GSX": (
            "5'-CHLORO-N-{(3S)-1-[(1S)-1-METHYL-2-MORPHOLIN-4-YL-2-OXOETHYL]-2-OXOPYRROLIDIN-3-YL}-2,2'-BITHIOPHENE-5-SULFONAMIDE",
            "non-polymer",
        ),
        "04Q": ("N-BENZYLGLYCINE", "non-polymer"),
        "IAM": ("4-[(ISOPROPYLAMINO)METHYL]PHENYLALANINE", "peptide"),
        "IAN": (
            "METHYL HYDROGEN (R)-[(R)-METHOXY(PHENYL)METHYL]PHOSPHONATE",
            "non-polymer",
        ),
        "OFF": (
            "2-DEOXY-3,4-BIS-O-[3-(4-HYDROXYPHENYL)PROPANOYL]-L-THREO-PENTARIC ACID",
            "non-polymer",
        ),
        "LU1": (
            "4'-FLUORO-4,5-DIHYDROXY-N-{(2E)-3-[(2S,4R,5R)-4-HYDROXY-5-(6-METHYL-9H-PURIN-9-YL)TETRAHYDROFURAN-2-YL]PROP-2-EN-1-YL}BIPHENYL-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "IAD": ("N-[1H-INDOL-3-YL-ACETYL]ASPARTIC ACID", "non-polymer"),
        "IAG": ("N-[1H-INDOL-3-YL-ACETYL]GLYCINE ACID", "non-polymer"),
        "OFI": ("N-[6-(4-HYDROXYPHENYL)-1H-INDAZOL-3-YL]BUTANAMIDE", "non-polymer"),
        "IAA": (
            "2-{O-[2'-(2-OXO-(2H3H)INDOLE-3-YLIDENE)-(2'H3'H)INDOLE-3'-YLIDENE]-AMINOOXY}-ACETATE",
            "non-polymer",
        ),
        "LU2": (
            "2-(3,4-DIHYDROXYPHENYL)-5,7-DIHYDROXY-4H-CHROMEN-4-ONE",
            "non-polymer",
        ),
        "IAC": ("1H-INDOL-3-YLACETIC ACID", "non-polymer"),
        "IAB": ("ISO-ACARBOSE", "saccharide"),
        "KEL": (
            "N-[(2R)-2-BENZYL-4-(HYDROXYAMINO)-4-OXOBUTANOYL]-L-ALANINE",
            "non-polymer",
        ),
        "OFT": (
            "(3Z)-N,N-DIETHYL-3-[(3E)-3-(HYDROXYIMINO)-1,3-DIHYDRO-2H-INDOL-2-YLIDENE]-2-OXO-2,3-DIHYDRO-1H-INDOLE-5-SULFONAMIDE",
            "non-polymer",
        ),
        "IAV": ("N-[1H-INDOL-3-YL-ACETYL]VALINE ACID", "non-polymer"),
        "IAQ": (
            "(2Z)-4-(DIMETHYLAMINO)-N-{7-FLUORO-4-[(2-METHYLPHENYL)AMINO]IMIDAZO[1,5-A]QUINOXALIN-8-YL}-N-METHYLBUT-2-ENAMIDE",
            "non-polymer",
        ),
        "IAP": ("4-IODO-ACETAMIDO PHENYLBORONIC ACID", "non-polymer"),
        "IAS": ("BETA-L-ASPARTIC ACID", "peptide"),
        "IAR": ("(2E)-5-[(DIAMINOMETHYLIDENE)AMINO]-2-IMINOPENTANOIC ACID", "peptide"),
        "PHK": ("(2R,3S)-3-AMINO-1-CHLORO-4-PHENYL-BUTAN-2-OL", "non-polymer"),
        "PHJ": ("N-[(AMINOOXY)CARBONYL]-N-PHENYLAMINE", "non-polymer"),
        "PHI": ("IODO-PHENYLALANINE", "peptide"),
        "PHH": ("4,5,6,7-TETRACHLORO-PHTHALIDE", "non-polymer"),
        "PHO": ("PHEOPHYTIN A", "non-polymer"),
        "PHN": ("1,10-PHENANTHROLINE", "non-polymer"),
        "PHM": ("(3S)-3-AMINO-1-CHLORO-4-PHENYLBUTAN-2-ONE", "peptide"),
        "PHL": ("L-PHENYLALANINOL", "peptide"),
        "PHC": ("N-METHYL-N-(METHYLBENZYL)FORMAMIDE", "non-polymer"),
        "PHB": ("P-HYDROXYBENZOIC ACID", "non-polymer"),
        "PHA": ("PHENYLALANINAL", "peptide"),
        "PHG": ("PHENYLMERCURY", "non-polymer"),
        "PHF": ("HF-OXO-PHOSPHATE CLUSTER PHF", "non-polymer"),
        "PHE": ("PHENYLALANINE", "peptide"),
        "PHD": ("ASPARTYL PHOSPHATE", "peptide"),
        "DZ1": ("N-(3,5-DIBROMO-4-HYDROXYPHENYL)-2,6-DIMETHYLBENZAMIDE", "non-polymer"),
        "PHZ": ("1-PHENYLHYDRAZINE", "non-polymer"),
        "DZ3": (
            "N-(3,5-DIBROMO-4-HYDROXYPHENYL)-4-HYDROXY-3,5-DIMETHYLBENZAMIDE",
            "non-polymer",
        ),
        "DZ2": ("2,5-DICHLORO-N-(3,5-DIBROMO-4-HYDROXYPHENYL)BENZAMIDE", "non-polymer"),
        "DZ4": (
            "2'-DEOXY-5'-O-[(R)-HYDROXY{[(R)-HYDROXY(PHOSPHONOOXY)PHOSPHORYL]AMINO}PHOSPHORYL]ADENOSINE",
            "non-polymer",
        ),
        "DZ9": (
            "3-AMINO-N-(3-FLUOROPHENYL)-6-THIOPHEN-2-YLTHIENO[2,3-B]PYRIDINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "DZ8": (
            "2-FLUORO-8-[(6-IODO-1,3-BENZODIOXOL-5-YL)METHYL]-9-[3-(ISOPROPYLAMINO)PROPYL]-9H-PURIN-6-AMINE",
            "non-polymer",
        ),
        "PHQ": ("BENZYL CHLOROCARBONATE", "non-polymer"),
        "PHP": (
            "(2-PHENYL-1-CARBOBENZYL-OXYVALYL-AMINO)-ETHYL-PHOSPHINIC ACID",
            "non-polymer",
        ),
        "ILO": ("N5-IMINOETHYL-L-ORNITHINE", "non-polymer"),
        "PHV": (
            "N-[(BENZYLOXY)CARBONYL]-BETA-PHENYL-D-PHENYLALANYL-N-[(1S)-4-METHOXY-1-PHOSPHONOBUTYL]-L-PROLINAMIDE",
            "peptide",
        ),
        "PHU": ("1-PHENYLUREA", "non-polymer"),
        "PHT": ("PHTHALIC ACID", "non-polymer"),
        "GSG": (
            "1S,3AS,8AS-TRIMETHYL-1-OXIDO-1,2,3,3A,8,8A-HEXAHYDROPYRROLO[2,3-B]INDOL-5-YL 2-ETHYLPHENYLCARBAMATE",
            "non-polymer",
        ),
        "VYE": (
            "(3S,5R)-5-[4-(2-CHLOROPHENYL)-2,2-DIMETHYL-5-OXOPIPERAZIN-1-YL]-N-(3-METHYLBUTYL)PIPERIDINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "AN4": ("3,4,5,6-TETRAHYDRO-2,3'-BIPYRIDINE", "non-polymer"),
        "AN5": ("5-AMINO-1-PYRIDIN-3-YLPENTAN-1-ONE", "non-polymer"),
        "AN6": (
            "5'-{[(3S)-3-AMINO-3-CARBOXYPROPYL](ETHYL)AMINO}-5'-DEOXYADENOSINE",
            "peptide",
        ),
        "AN7": (
            "(3E)-4-{3-HYDROXY-2-METHYL-5-[(PHOSPHONOOXY)METHYL]PYRIDIN-4-YL}-2-OXOBUT-3-ENOIC ACID",
            "non-polymer",
        ),
        "AN0": ("N-ACETYL-L-NORVALINE", "non-polymer"),
        "AN1": ("3-(10-METHYL-ANTHRACEN-9-YL)-PROPIONIC ACID", "non-polymer"),
        "AN2": ("AMP PHOSPHORAMIDATE", "non-polymer"),
        "AN3": ("ANTHRACENE", "non-polymer"),
        "UCB": (
            "1-(CYCLOHEPTYLMETHYL)-4-{[(2,7-DICHLORO-9H-XANTHEN-9-YL)CARBONYL]AMINO}-1-ETHYLPIPERIDINIUM",
            "non-polymer",
        ),
        "AN8": ("(2S)-2-AMINONONANOIC ACID", "peptide"),
        "AN9": (
            "1,5-BIS[3-(DIETHYLAMINO)PROPIONAMIDO]ANTHRACENE-9,10-DIONE",
            "non-polymer",
        ),
        "GK8": (
            "{[(2S)-2-AMINO-3-(1H-IMIDAZOL-5-YL)PROPYL]OXY}(TRIHYDROXY)-LAMBDA~5~-ARSANYL",
            "non-polymer",
        ),
        "GSJ": (
            "1-PYRROLIDINEACETAMIDE, 3-[[(6-CHLORO-2-NAPHTHALENYL)SULFONYL]AMINO]-ALPHA-METHYL-N-(1-METHYLETHYL)-N-[2-[(METHYLSULFONYL)AMINO]ETHYL]-2-OXO-, (ALPHAS,3S)-",
            "non-polymer",
        ),
        "GK1": (
            "4-{4-[(5-HYDROXY-2-METHYLPHENYL)AMINO]QUINOLIN-7-YL}-1,3-THIAZOLE-2-CARBALDEHYDE",
            "non-polymer",
        ),
        "ROL": ("ROLIPRAM", "non-polymer"),
        "GK3": (
            "N-(3-CYANOPHENYL)-2'-METHYL-5'-(5-METHYL-1,3,4-OXADIAZOL-2-YL)-4-BIPHENYLCARBOXAMIDE",
            "non-polymer",
        ),
        "GK5": (
            "N~3~-CYCLOPROPYL-N~4~'-(CYCLOPROPYLMETHYL)-6-METHYLBIPHENYL-3,4'-DICARBOXAMIDE",
            "non-polymer",
        ),
        "GK4": (
            "N-(CYCLOPROPYLMETHYL)-2'-METHYL-5'-(5-METHYL-1,3,4-OXADIAZOL-2-YL)BIPHENYL-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "M34": (
            "N-(BENZYLSULFONYL)-D-LEUCYL-N-[2-(AMINOMETHYL)-5-CHLOROBENZYL]-L-PROLINAMIDE",
            "non-polymer",
        ),
        "GK6": (
            "N-{4'-[(CYCLOPROPYLMETHYL)CARBAMOYL]-6-METHYLBIPHENYL-3-YL}-2-MORPHOLIN-4-YLPYRIDINE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "GSO": (
            "L-GAMMA-GLUTAMYL-S-[(2S)-2-HYDROXY-2-PHENYLETHYL]-L-CYSTEINYLGLYCINE",
            "non-polymer",
        ),
        "MH9": ("(2E,4E)-DODECA-2,4-DIENOIC ACID", "non-polymer"),
        "MH6": ("3-HYDROXY-2-IMINOPROPANOIC ACID", "peptide"),
        "P27": (
            "{[2-(1H-1,2,3-BENZOTRIAZOL-1-YL)-2-(3,4-DIFLUOROPHENYL)PROPANE-1,3-DIYL]BIS[4,1-PHENYLENE(DIFLUOROMETHYLENE)]}BIS(PHOSPHONIC ACID)",
            "non-polymer",
        ),
        "MH4": (
            "1-METHYL-8-{[4-(4-METHYLPIPERAZIN-1-YL)PHENYL]AMINO}-N-[(2-METHYLPYRIDIN-4-YL)METHYL]-4,5-DIHYDRO-1H-PYRAZOLO[4,3-H]QUINAZOLINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "GSN": (
            "2-AMINO-5-[1-(CARBOXYLATOMETHYLCARBAMOYL)-2-NITROSOSULFANYL-ETHYL]AMINO-5-OXO-PENTANOATE",
            "non-polymer",
        ),
        "MH2": ("MANGANESE ION, 1 HYDROXYL COORDINATED", "non-polymer"),
        "MH3": ("MANGANESE ION, 1 HYDROXYL COORDINATED", "non-polymer"),
        "MH1": ("1-METHYL-L-HISTIDINE", "peptide"),
        "MHN": ("6-METHYLHEPTAN-1-OL", "non-polymer"),
        "MHO": ("S-OXYMETHIONINE", "peptide"),
        "MHL": ("N-METHYL-4-HYDROXY-LEUCINE", "peptide"),
        "MHM": (
            "[7,12-DIETHYL-3,8,13,17-TETRAMETHYL-21H,23H-PORPHINE-2,18-DIPORPANOTO-(2)-N21,N22,N23,N24,]IRON",
            "non-polymer",
        ),
        "MHH": ("METHYL HYDROGEN (R)-HEXYLPHOSPHONATE", "non-polymer"),
        "MHI": ("L-ALA-GAMMA-D-GLU-MESO-DIAMINOPIMELIC ACID", "non-polymer"),
        "MHF": ("5,10-METHENYL-6,7,8-TRIHYDROFOLIC ACID", "non-polymer"),
        "FFP": ("2,6-DIFLUOROPHENOL", "non-polymer"),
        "MHB": ("?", "non-polymer"),
        "MHC": (
            "3-MERCAPTO-1-(1,3,4,9-TETRAHYDRO-B-CARBOLIN-2-YL)-PROPAN-1-ONE",
            "non-polymer",
        ),
        "MHA": ("(CARBAMOYLMETHYL-CARBOXYMETHYL-AMINO)-ACETIC ACID", "non-polymer"),
        "GKK": (
            "2-CHLORO-4-{[(3S)-1-METHYLPYRROLIDIN-3-YL][2-(TRIFLUOROMETHYL)BENZYL]AMINO}BENZONITRILE",
            "non-polymer",
        ),
        "MHZ": ("5'-DEOXY-5'-[(3-HYDRAZINOPROPYL)METHYLAMINO]ADENOSINE", "non-polymer"),
        "GKL": ("N-[(2S,3S)-3-AMINO-2-HYDROXY-5-METHYLHEXYL]-L-VALINE", "peptide"),
        "MHV": ("4-OXO-L-PIPECOLIC ACID", "peptide"),
        "LLZ": (
            "N~6~-[(1S)-3-(6-METHOXYNAPHTHALEN-2-YL)-1-METHYL-3-OXOPROPYL]-L-LYSINE",
            "peptide",
        ),
        "MHT": ("(3R)-3-(METHYLSULFANYL)-1-AZABICYCLO[2.2.2]OCTANE", "non-polymer"),
        "MHU": ("4-N,N-(DIMETHYLAMINO)-L-PHENYLALANINE", "peptide"),
        "GKE": ("?", "non-polymer"),
        "GKD": ("?", "non-polymer"),
        "NNJ": ("3-(AMINOMETHYL)-1-METHYLPYRIDINIUM", "non-polymer"),
        "AND": ("3-BETA-HYDROXY-5-ANDROSTEN-17-ONE", "non-polymer"),
        "ANE": ("ADENINE", "non-polymer"),
        "ANF": ("ANTHRONE", "non-polymer"),
        "ANG": ("8-AMINOGUANINE", "non-polymer"),
        "ANA": (
            "4-O-ACETYL-ALPHA-2-OMETHYL-5-N-ACETYL-ALPHA-D-NEURAMINIC ACID",
            "non-polymer",
        ),
        "ANB": ("ANDROSTA-1,4-DIENE-3,17-DIONE", "non-polymer"),
        "ANC": ("ANTHRACEN-1-YLAMINE", "non-polymer"),
        "ANL": ("ANILINE", "non-polymer"),
        "ANM": ("ANISOMYCIN", "non-polymer"),
        "ANN": ("4-METHOXYBENZOIC ACID", "non-polymer"),
        "ANO": ("5-BETA-ANDROSTANE-3,17-DIONE", "non-polymer"),
        "ANH": (
            "METHYL N-[(4-METHYLPHENYL)SULFONYL]GLYCYL-3-[AMINO(IMINO)METHYL]-D-PHENYLALANINATE",
            "non-polymer",
        ),
        "ANI": ("4-(TRIFLUOROMETHYL)ANILINE", "non-polymer"),
        "ANJ": (
            "(2R,3S,6S,7R,8R)-3-{[3-(FORMYLAMINO)-2-HYDROXYBENZOYL]AMINO}-8-HEXYL-2,6-DIMETHYL-4,9-DIOXO-1,5-DIOXONAN-7-YL (2S)-2-METHYLBUTANOATE",
            "non-polymer",
        ),
        "ANK": (
            "9-{5-O-[(R)-HYDROXY{[(S)-HYDROXY(PHOSPHONOAMINO)PHOSPHORYL]OXY}PHOSPHORYL]-BETA-L-RIBOFURANOSYL}-9H-PURIN-6-AMINE",
            "non-polymer",
        ),
        "ANT": ("ANTHRAMYCIN", "non-polymer"),
        "ANU": ("2,2'-ANHYDROURIDINE", "non-polymer"),
        "NVE": ("NEVIRAPINE PHOSPHONATE ANALOGUE", "non-polymer"),
        "ANW": (
            "2-AMINO-7-(1-METHYLETHYL)-5-OXO-5H-CHROMENO[2,3-B]PYRIDINE-3-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "ANP": ("PHOSPHOAMINOPHOSPHONIC ACID-ADENYLATE ESTER", "non-polymer"),
        "ANQ": ("ACENAPHTHENEQUINONE", "non-polymer"),
        "ANR": ("ANHYDRORETINOL", "non-polymer"),
        "ANS": (
            "5-(DIMETHYLAMINO)-1-NAPHTHALENESULFONIC ACID(DANSYL ACID)",
            "non-polymer",
        ),
        "ANX": ("2,5-DINITROPHENOL", "non-polymer"),
        "ANY": (
            "2-METHYL-BUTYRIC ACID 3-(3-FORMYLAMINO-2-HYDROXY-BENZOYLAMINO)-8-HEPTYL-2,6-DIMETHYL-4,9-DIOXO-[1,5]DIOXONAN-7-YL ESTER",
            "non-polymer",
        ),
        "ANZ": ("?", "non-polymer"),
        "1ZZ": (
            "5'-O-[(S)-(DODECANOYLOXY)(HYDROXY)PHOSPHORYL]ADENOSINE",
            "non-polymer",
        ),
        "NEB": ("2-DEOXY-D-STREPTAMINE", "non-polymer"),
        "HKV": ("4,5-DIOXOPENTANOIC ACID", "non-polymer"),
        "1ZT": ("N-ACETYL-L-CYSTEINE", "peptide"),
        "1ZV": (
            "AMINO({3-[(3R,5R,14S,16S,21AR)-5,14-DIHYDROXY-1,4,17-TRIOXO-16-(2-PHENYLETHYL)ICOSAHYDRO-1H-PYRROLO[1,2-D][1,4,7,11]TETRAAZACYCLONONADECIN-3-YL]PROPYL}AMINO)METHANIMINIUM",
            "peptide",
        ),
        "1ZK": ("U75875", "peptide"),
        "1ZN": (
            "(2S,3S,4E,6E,8S,9S)-3-AMINO-9-METHOXY-2,6,8-TRIMETHYL-10-PHENYLDECA-4,6-DIENOIC ACID",
            "peptide",
        ),
        "1ZB": (
            "N-[(1S)-1-BENZYL-3-DIAZEN-1-IUMYLIDENE-2-OXOPROPYL]GLYCINAMIDE",
            "peptide",
        ),
        "1ZD": (
            "N-{(2S)-2-[(N-ACETYL-L-THREONYL-L-ISOLEUCYL)AMINO]HEXYL}-L-NORLEUCYL-L-GLUTAMINYL-N~5~-[AMINO(IMINIO)METHYL]-L-ORNITHINAMIDE",
            "peptide",
        ),
        "1ZG": (
            "N-(3-CARBOXYPROPANOYL)-L-VALYL-N-{(1R)-1-[(S)-HYDROXY(OXIDO)PHOSPHANYL]-2-PHENYLETHYL}-L-PROLINAMIDE",
            "peptide",
        ),
        "FFM": (
            "S-[(1R,2R)-2-HYDROXY-1-METHYL-2-PHOSPHONOETHYL]-L-CYSTEINE",
            "peptide",
        ),
        "KR2": (
            "(2R)-4-(2-BENZOYL-1,2-DIAZEPAN-1-YL)-4-OXO-1-(2,4,5-TRIFLUOROPHENYL)BUTAN-2-AMINE",
            "non-polymer",
        ),
        "KR1": (
            "3-({3-[(2-AMINO-6-FLUOROPYRIDIN-4-YL)METHYL]-5-(1-METHYLETHYL)-2,6-DIOXO-1,2,3,6-TETRAHYDROPYRIMIDIN-4-YL}CARBONYL)-5-METHYLBENZONITRILE",
            "non-polymer",
        ),
        "V10": (
            "N^6^-[(1R)-2-[(1R)-1-CARBOXY-2-(METHYLSULFANYL)ETHOXY]-2-OXO-1-(SULFANYLMETHYL)ETHYL]-6-OXO-L-LYSINE",
            "non-polymer",
        ),
        "HKI": (
            "N-(4-{[3-CHLORO-4-(PYRIDIN-2-YLMETHOXY)PHENYL]AMINO}-3-CYANO-7-ETHOXYQUINOLIN-6-YL)-4-(DIMETHYLAMINO)BUTANAMIDE",
            "non-polymer",
        ),
        "HKL": ("FE(III) PYROPHEOPHORBIDE-A METHYL ESTER", "non-polymer"),
        "MMT": ("5'-O-(DIMETHYLAMINO)-THYMIDINE", "nucleic-acid"),
        "KRP": (
            "3-METHYL-5-{[5-(1-METHYLETHYL)-2,6-DIOXO-3-PROPYL-1,2,3,6-TETRAHYDROPYRIMIDIN-4-YL]CARBONYL}BENZONITRILE",
            "non-polymer",
        ),
        "KRV": (
            "3-{[3-ETHYL-5-(1-METHYLETHYL)-2,6-DIOXO-1,2,3,6-TETRAHYDROPYRIMIDIN-4-YL]CARBONYL}-5-METHYLBENZONITRILE",
            "non-polymer",
        ),
        "KRW": (
            "2-{4-[1-(QUINOLIN-6-YLMETHYL)-1H-[1,2,3]TRIAZOLO[4,5-B]PYRAZIN-6-YL]-1H-PYRAZOL-1-YL}ETHANOL",
            "non-polymer",
        ),
        "P22": ("ETHYL DIHYDROGEN DIPHOSPHATE", "non-polymer"),
        "KRJ": (
            "4-CYANO-N-(2-CYCLOHEX-1-EN-1-YL-4-PIPERIDIN-4-YLPHENYL)-1H-IMIDAZOLE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "KRH": (
            "5-METHYL-3-(1-METHYL-1H-IMIDAZOL-5-YL)-7-(2-METHYLPROPYL)-2-(NAPHTHALEN-1-YLMETHYL)-2H-PYRAZOLO[3,4-D]PYRIMIDINE-4,6(5H,7H)-DIONE",
            "non-polymer",
        ),
        "I96": (
            "N-(2-ISOPROPYLPHENYL)-3-[(2-THIENYLMETHYL)THIO]-1H-1,2,4-TRIAZOL-5-AMINE",
            "non-polymer",
        ),
        "I93": (
            "4-(2-CHLORO-6-FLUORO-3-METHYLPHENYL)-2,4-DIOXOBUTANOIC ACID",
            "non-polymer",
        ),
        "HKC": (
            "6,7-DIMETHOXY-3-[4-(1H-TETRAZOL-5-YL)PHENYL]-1,4-DIHYDROINDENO[1,2-C]PYRAZOLE",
            "non-polymer",
        ),
        "KRB": ("KURASOIN B", "non-polymer"),
        "KRC": (
            "(9AS)-8-ACETYL-N-[(2-ETHYLNAPHTHALEN-1-YL)METHYL]-1,7-DIHYDROXY-3-METHOXY-9A-METHYL-9-OXO-9,9A-DIHYDRODIBENZO[B,D]FURAN-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "KRA": ("KURASOIN A", "non-polymer"),
        "KRG": ("3-(3,5-DIMETHYL-1,2-OXAZOL-4-YL)-5-ETHOXYBENZOIC ACID", "non-polymer"),
        "NNT": ("3-(AMINOMETHYL)-4-CHLORO-1-METHYLPYRIDINIUM", "non-polymer"),
        "1Z0": (
            "N-METHYL-D-PHENYLALANYL-N-{(1S)-4-CARBAMIMIDAMIDO-1-[(6-CARBOXY-1,3-BENZOTHIAZOL-2-YL)CARBONYL]BUTYL}-L-PROLINAMIDE",
            "peptide",
        ),
        "1Z1": (
            "N-(DIMETHYLSULFAMOYL)-L-PHENYLALANYL-N-[(1S,2S)-2-HYDROXY-4-{[(2S)-2-METHYLBUTYL]AMINO}-1-(2-METHYLPROPYL)-4-OXOBUTYL]-N~6~-(METHYLCARBAMOTHIOYL)-L-LYSINAMIDE",
            "peptide",
        ),
        "1Z7": (
            "N-(3-METHYLBUTANOYL)-L-VALYL-N-{(1S)-1-[(R)-[(1R)-1-BENZYL-2-METHOXY-2-OXOETHOXY](HYDROXY)PHOSPHORYL]-3-METHYLBUTYL}-L-VALINAMIDE",
            "peptide",
        ),
        "PC2": ("DI-STEAROYL-3-SN-PHOSPHATIDYLCHOLINE", "non-polymer"),
        "X64": (
            "4-{[4-AMINO-5-(2-NITROBENZOYL)-1,3-THIAZOL-2-YL]AMINO}BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "PC3": ("COPROPORPHYRIN I CONTAINING CO(III)", "non-polymer"),
        "Y16": (
            "(E,2R,3R,4S,5R)-N-(2-AZANYL-2-OXIDANYLIDENE-ETHYL)-2-METHOXY-8,8-DIMETHYL-3,4,5-TRIS(OXIDANYL)NON-6-ENAMIDE",
            "non-polymer",
        ),
        "PC0": (
            "1-[2-(4-ETHOXY-3-FLUOROPYRIDIN-2-YL)ETHYL]-3-(5-METHYLPYRIDIN-2-YL)THIOUREA",
            "non-polymer",
        ),
        "5AX": ("2-(ACETYLAMINO)-1,5-ANHYDRO-2-DEOXY-D-GLUCITOL", "non-polymer"),
        "5AZ": ("6-AMINO-1,3,5-TRIAZIN-2(1H)-ONE", "non-polymer"),
        "5AP": (
            "5-(AMINOMETHYL)-6-(2,4-DICHLOROPHENYL)-2-(3,5-DIMETHOXYPHENYL)PYRIMIDIN-4-AMINE",
            "non-polymer",
        ),
        "5AS": ("5'-O-(N-ETHYL-SULFAMOYL)ADENOSINE", "non-polymer"),
        "SOO": (
            "S-[2-[3-[[(2S)-2-HYDROXY-3,3-DIMETHYL-4-PHOSPHONOOXY-BUTANOYL]AMINO]PROPANOYLAMINO]ETHYL] (E)-OCT-2-ENETHIOATE",
            "non-polymer",
        ),
        "5AT": ("5'-AMINO-5'-DEOXYTHYMIDINE", "nucleic-acid"),
        "PC7": (
            "(7S)-4-HYDROXY-N,N,N-TRIMETHYL-9-OXO-7-[(PALMITOYLOXY)METHYL]-3,5,8-TRIOXA-4-PHOSPHAHEXACOSAN-1-AMINIUM 4-OXIDE",
            "non-polymer",
        ),
        "5AH": (
            "METHYL [(1S)-1-({2-[(3S)-3-BENZYL-3-HYDROXY-4-{[(1S,2R)-2-HYDROXY-2,3-DIHYDRO-1H-INDEN-1-YL]AMINO}-4-OXOBUTYL]-2-(4-PYRIDIN-2-YLBENZYL)HYDRAZINO}CARBONYL)-2,2-DIMETHYLPROPYL]CARBAMATE",
            "non-polymer",
        ),
        "5AM": (
            "BENZYL {(1S)-5-AMINO-1-[(S)-HYDROXY(5-{[4-(4-PHENYLBUTANOYL)PIPERAZIN-1-YL]METHYL}-1,2,4-OXADIAZOL-3-YL)METHYL]PENTYL}CARBAMATE",
            "non-polymer",
        ),
        "5AL": (
            "5'-O-[(R)-{[(2S)-2-AMINOPROPANOYL]OXY}(HYDROXY)PHOSPHORYL]ADENOSINE",
            "non-polymer",
        ),
        "5AO": ("3-(1,3-OXAZOL-5-YL)ANILINE", "non-polymer"),
        "5AN": ("3,5-DIFLUOROANILINE", "non-polymer"),
        "5AA": ("N6-DIMETHYL-3'-AMINO-ADENOSINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "NPU": ("(1R,2R)-2-AMINO-1-PHENYLPROPAN-1-OL", "non-polymer"),
        "5AC": ("5-AMINO-1H-IMIDAZOLE-4-CARBOXAMIDE", "non-polymer"),
        "5AB": ("[(5S)-5-AMINO-5-CARBOXYHEXYL](TRIHYDROXY)BORATE", "peptide"),
        "5AD": ("5'-DEOXYADENOSINE", "non-polymer"),
        "5AF": ("2,5-BIS{[4-(3-PENTYL)DIAMINOMETHYL]PHENYL}FURAN", "non-polymer"),
        "89P": (
            "6-(4-{[3-(2,6-DICHLOROPHENYL)-5-(1-METHYLETHYL)ISOXAZOL-4-YL]METHOXY}PHENYL)QUINOLINE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "G3P": ("SN-GLYCEROL-3-PHOSPHATE", "non-polymer"),
        "RHO": ("TETRAMETHYLRHODAMINE-5-MALEIMIDE", "non-polymer"),
        "YEX": (
            "5-METHYL-8-(1H-PYRROL-2-YL)[1,2,4]TRIAZOLO[4,3-A]QUINOLIN-1(2H)-ONE",
            "non-polymer",
        ),
        "G3X": (
            "(4AS,6R,8AS)-3-METHOXY-11-(3-PIPERIDIN-1-YLPROPYL)-5,6,9,10,11,12-HEXAHYDRO-4AH-[1]BENZOFURO[3A,3,2-EF][2]BENZAZEPIN-6-OL",
            "non-polymer",
        ),
        "G3A": ("GUANOSINE-P3-ADENOSINE-5',5'-TRIPHOSPHATE", "non-polymer"),
        "G3B": (
            "2-METHYL-5-(3-{4-[(S)-METHYLSULFINYL]PHENYL}-1-BENZOFURAN-5-YL)-1,3,4-OXADIAZOLE",
            "non-polymer",
        ),
        "G3E": ("3-HYDROXYQUINOLIN-2(1H)-ONE", "non-polymer"),
        "G3D": ("GUANOSINE-3'-MONOPHOSPHATE-5'-DIPHOSPHATE", "non-polymer"),
        "G3G": (
            "N,N'-(3S,4S)-PYRROLIDINE-3,4-DIYLBIS(4-AMINO-N-BENZYLBENZENESULFONAMIDE)",
            "non-polymer",
        ),
        "G3F": ("3-DEOXY-3-FLUORO-BETA-D-GLUCOPYRANOSE", "non-polymer"),
        "G3I": (
            "(3R,4R,5R)-3-HYDROXY-5-(HYDROXYMETHYL)PIPERIDIN-4-YL BETA-D-GLUCOPYRANOSYL-(1->4)-BETA-D-GLUCOPYRANOSYL-(1->4)-BETA-D-GLUCOPYRANOSIDE",
            "saccharide",
        ),
        "89I": (
            "5-[4-(DIMETHYLAMINO)PHENYL]-6-[(6-MORPHOLIN-4-YLPYRIDIN-3-YL)ETHYNYL]PYRIMIDIN-4-AMINE",
            "non-polymer",
        ),
        "OKD": (
            "N-({(1S)-1-CARBOXY-5-[4-(2,5,8,11,14-PENTAOXAPENTADEC-1-YL)-1H-1,2,3-TRIAZOL-1-YL]PENTYL}CARBAMOYL)-L-GLUTAMIC ACID",
            "non-polymer",
        ),
        "G31": ("3'-METHYL-2',3'-DEOXY-GUANOSINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "891": (
            "6-(2,4-DIFLUOROPHENOXY)-8-METHYL-2-{[(1S)-1-METHYL-2-(METHYLSULFONYL)ETHYL]AMINO}PYRIDO[2,3-D]PYRIMIDIN-7(8H)-ONE",
            "non-polymer",
        ),
        "892": (
            "3-(3,5-DIBROMO-4-HYDROXY-BENZOYL)-2-ETHYL-BENZOFURAN-6-SULFONIC ACID (4-SULFAMOYL-PHENYL)-AMIDE",
            "non-polymer",
        ),
        "893": (
            "5-CYANO-N-(2,5-DIMETHOXYBENZYL)-6-ETHOXYPYRIDINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "894": (
            "2-(5-CHLORO-2-THIENYL)-N-{(3S)-1-[(1S)-1-METHYL-2-MORPHOLIN-4-YL-2-OXOETHYL]-2-OXOPYRROLIDIN-3-YL}ETHANESULFONAMIDE",
            "non-polymer",
        ),
        "895": (
            "2-(5-CHLORO-2-THIENYL)-N-{(3S)-1-[(1S)-1-METHYL-2-MORPHOLIN-4-YL-2-OXOETHYL]-2-OXOPYRROLIDIN-3-YL}ETHENESULFONAMIDE",
            "non-polymer",
        ),
        "896": (
            "N-ETHYL-N-ISOPROPYL-3-METHYL-5-{[(2S)-2-(PYRIDIN-4-YLAMINO)PROPYL]OXY}BENZAMIDE",
            "non-polymer",
        ),
        "897": (
            "N-[3-(TERT-BUTYLAMINO)-3-OXOPROPYL]-N-ISOPROPYL-3-METHYL-5-{[(2S)-2-(PYRIDIN-4-YLAMINO)PROPYL]OXY}BENZAMIDE",
            "non-polymer",
        ),
        "898": (
            "(E)-2-(5-CHLOROTHIOPHEN-2-YL)-N-[(3S)-1-{4-[(1S)-1-(DIMETHYLAMINO)ETHYL]-2-FLUOROPHENYL}-2-OXOPYRROLIDIN-3-YL]ETHENESULFONAMIDE",
            "non-polymer",
        ),
        "G38": ("3'-AMINO-2'-DEOXY-GUANOSINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "CM8": (
            "N-[(1S,2R)-1-BENZYL-2-HYDROXY-3-{[3-(TRIFLUOROMETHYL)BENZYL]AMINO}PROPYL]-3-(1,1-DIOXIDO-1,2-THIAZINAN-2-YL)-5-(ETHYLAMINO)BENZAMIDE",
            "non-polymer",
        ),
        "HK7": (
            "2-(11-OXO-10,11-DIHYDRO-5H-DIBENZO[B,E][1,4]DIAZEPIN-3-YL)BENZAMIDE",
            "non-polymer",
        ),
        "IND": ("INDOLE", "non-polymer"),
        "RPR": ("RPR128515", "non-polymer"),
        "NPW": (
            "BETA-6-HYDROXY-1,4,5,6-TETRAHYDRONICOTINAMIDE ADENINE DINUCLEOTIDE PHOSPHATE",
            "non-polymer",
        ),
        "CM6": (
            "(2S)-2-[(2,1,3-BENZOTHIADIAZOL-4-YLSULFONYL)AMINO]-2-PHENYL-N-PYRIDIN-4-YLACETAMIDE",
            "non-polymer",
        ),
        "CM5": ("5-CYCLOHEXYL-1-PENTYL-BETA-D-MALTOSIDE", "non-polymer"),
        "VXX": ("METHYL VANILLATE", "non-polymer"),
        "4MP": ("1-(4-METHOXYBENZOYL)-2-PYRROLIDINONE", "non-polymer"),
        "4MQ": ("4-(4-METHYL-1,4-DIAZEPAN-1-YL)QUINAZOLINE", "non-polymer"),
        "4MR": (
            "5-(4-PHENOXYPHENYL)-5-(4-PYRIMIDIN-2-YLPIPERAZIN-1-YL)PYRIMIDINE-2,4,6(2H,3H)-TRIONE",
            "non-polymer",
        ),
        "4MS": (
            "N-({3-[(5S)-5-TERT-BUTYL-1-(3-CHLORO-4-FLUOROBENZYL)-4-HYDROXY-2-OXO-2,5-DIHYDRO-1H-PYRROL-3-YL]-1,1-DIOXIDO-1,2-BENZISOTHIAZOL-7-YL}METHYL)METHANESULFONAMIDE",
            "non-polymer",
        ),
        "4MU": ("7-HYDROXY-4-METHYL-2H-CHROMEN-2-ONE", "non-polymer"),
        "4MV": ("4-METHYL VALERIC ACID", "non-polymer"),
        "4MY": ("MYO INOSITOL 3,4,5,6 TETRAKISPHOSPHATE", "non-polymer"),
        "4MZ": ("4-METHYLIMIDAZOLE", "non-polymer"),
        "CM2": ("CARBOXYMYCOBACTIN T", "non-polymer"),
        "4MA": ("4-METHYLBENZOIC ACID", "non-polymer"),
        "4MB": ("4-[(METHYLSULFONYL)AMINO]BENZOIC ACID", "non-polymer"),
        "4MC": (
            "(Z)-N-(5-GUANIDINO-1-OXO-1-(5-PHENYL-1-(PHENYLSULFONYL)PENT-1-EN-3-YLAMINO)PENTAN-2-YL)-4-METHYLPIPERAZINE-1-CARBOXAMIDE",
            "non-polymer",
        ),
        "4MD": ("N-(4-CHLOROBENZYL)-N-METHYLBENZENE-1,4-DISULFONAMIDE", "non-polymer"),
        "4ME": ("METHYL 1H-INDOLE-4-CARBOXYLATE", "non-polymer"),
        "4MF": (
            "1-(2-DEOXY-5-O-PHOSPHONO-BETA-D-ERYTHRO-PENTOFURANOSYL)-4-METHYL-1H-INDOLE",
            "nucleic-acid",
        ),
        "FFF": ("TRIFLUOROFURNESYL DIPHOSPHATE", "non-polymer"),
        "CM0": ("?", "nucleic-acid"),
        "4ML": (
            "[(2S)-2-METHYL-5-OXO-2,5-DIHYDROFURAN-2-YL]ACETIC ACID",
            "non-polymer",
        ),
        "4MM": (
            "(1S)-1-CARBOXY-N,N,N-TRIMETHYL-3-(METHYLSULFANYL)PROPAN-1-AMINIUM",
            "peptide",
        ),
        "4MO": ("MOLYBDENUM(IV) ION", "non-polymer"),
        "N3H": (
            "N-{3-[5-HYDROXY-2-(3-METHYLBUTYL)-3-OXO-6-THIOPHEN-2-YL-2,3-DIHYDROPYRIDAZIN-4-YL]-1,1-DIOXIDO-2H-1,2,4-BENZOTHIADIAZIN-7-YL}METHANESULFONAMIDE",
            "non-polymer",
        ),
        "SB2": (
            "4-[5-(4-FLUORO-PHENYL)-2-(4-METHANESULFINYL-PHENYL)-3H-IMIDAZOL-4-YL]-PYRIDINE",
            "non-polymer",
        ),
        "ILP": ("N-[O-PHOSPHONO-PYRIDOXYL]-ISOLEUCINE", "non-polymer"),
        "LLP": (
            "2-LYSINE(3-HYDROXY-2-METHYL-5-PHOSPHONOOXYMETHYL-PYRIDIN-4-YLMETHANE)",
            "peptide",
        ),
        "AKZ": ("(3S)-3-AMINO-4,4-DIHYDROXYBUTANOIC ACID", "peptide"),
        "2C5": (
            "2-CHLORO-5-(3-CHLORO-PHENYL)-6-[(4-CYANO-PHENYL)-(3-METHYL-3H-IMIDAZOL-4-YL)- METHOXYMETHYL]-NICOTINONITRILE",
            "non-polymer",
        ),
        "N3B": (
            "N-[(4'-FLUORO-1,1'-BIPHENYL-4-YL)CARBONYL]-3-NITRO-4-{[2-(PHENYLSULFANYL)ETHYL]AMINO}BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "2C7": (
            "2-(CYCLOHEPTYLMETHYL)-1,1-DIOXIDO-1-BENZOTHIOPHEN-6-YL SULFAMATE",
            "non-polymer",
        ),
        "N3D": ("N-METHYLPROPANE-1,3-DIAMINE", "non-polymer"),
        "7TR": (
            "4-({(3R)-7-CYANO-4-[(4-METHOXYPHENYL)SULFONYL]-1-[(1-METHYL-1H-IMIDAZOL-5-YL)METHYL]-2,3,4,5-TETRAHYDRO-1H-1,4-BENZODIAZEPIN-3-YL}METHYL)PHENYL DIETHYLCARBAMATE",
            "non-polymer",
        ),
        "2C2": ("(2S)-2-HYDROXY-2H-CHROMENE-2-CARBOXYLIC ACID", "non-polymer"),
        "SB1": (
            "(21S)-1AZA-4,4-DIMETHYL-6,19-DIOXA-2,3,7,20-TETRAOXOBICYCLO[19.4.0] PENTACOSANE",
            "non-polymer",
        ),
        "N3P": (
            "N-(4-CHLOROPHENYL)-3-(PHOSPHONOOXY)NAPHTHALENE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "MM9": (
            "(2S)-N-[(4-CARBAMIMIDOYLPHENYL)METHYL]-1-[(2R)-2-CYCLOHEXYL-2-[[2-(DIMETHYLAMINO)-2-OXIDANYLIDENE-ETHYL]AMINO]ETHANOYL]AZETIDINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "N3T": ("?", "non-polymer"),
        "SB7": ("[HYDROXY(3-PHENYLPROPYL)AMINO]METHANOL", "non-polymer"),
        "NN4": (
            "(1S,3R,4S,5S,7S)-4-{[2-(4-METHOXYPHENOXY)-2-METHYLPROPANOYL]AMINO}ADAMANTANE-1-CARBOXAMIDE",
            "non-polymer",
        ),
        "SB4": (
            "4-(4-FLUOROPHENYL)-1-(4-PIPERIDINYL)-5-(2-AMINO-4-PYRIMIDINYL)-IMIDAZOLE",
            "non-polymer",
        ),
        "GLO": ("D-GLUCOSE IN LINEAR FORM", "saccharide"),
        "OTC": ("OXYTETRACYCLINE", "non-polymer"),
        "T4S": (
            "1-(2-DEOXY-5-O-PHOSPHONO-BETA-D-ERYTHRO-PENTOFURANOSYL)-4-HYDROSELENO-5-METHYLPYRIMIDIN-2(1H)-ONE",
            "nucleic-acid",
        ),
        "HX0": ("N,N,N-TRIOCTYLOCTAN-1-AMINIUM", "non-polymer"),
        "GLA": ("ALPHA D-GALACTOSE", "saccharide"),
        "VXU": ("3,4-DIHYDRO-2H-1,5-BENZODIOXEPIN-6-YLMETHANAMINE", "non-polymer"),
        "XAO": ("CYCLOHEXYLMETHYL-2,3-DIHYDROXY-5-METHYL-HEXYLAMIDE", "non-polymer"),
        "L3L": (
            "3-OXO-2-PIPERIDIN-4-YL-2,3-DIHYDRO-1H-ISOINDOLE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "GLC": ("ALPHA-D-GLUCOSE", "saccharide"),
        "678": (
            "(3-{5-[AMINO(IMINIO)METHYL]-1H-INDOL-2-YL}-5-BROMO-4-OXIDOPHENYL)ACETATE",
            "non-polymer",
        ),
        "679": (
            "2-CHLORO-5-[4-(3-CHLORO-PHENYL)-2,5-DIOXO-2,5-DIHYDRO-1H-PYRROL-3-YLAMINO]-BENZOIC ACID",
            "non-polymer",
        ),
        "675": ("6-[(Z)-AMINO(IMINO)METHYL]-N-PHENYL-2-NAPHTHAMIDE", "non-polymer"),
        "677": (
            "(2R,3R)-7-(METHYLSULFONYL)-3-(2,4,5-TRIFLUOROPHENYL)-1,2,3,4-TETRAHYDROPYRIDO[1,2-A]BENZIMIDAZOL-2-AMINE",
            "non-polymer",
        ),
        "670": (
            "P-(4-FERROCENYL-1H-1,2,3-TRIAZOL-1-YL)BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "672": (
            "(S)-4-(4-(3-(3-CARBAMIMIDOYLPHENYL)-2-(2,4,6-TRIISOPROPYLPHENYLSULFONAMIDO)PROPANOYL)PIPERAZINE-1-CARBONYL)PIPERIDINE-1-CARBOXIMIDAMIDE",
            "non-polymer",
        ),
        "673": (
            "N-(1-METHYLETHYL)-3-[(3-PROP-2-EN-1-YLBIPHENYL-4-YL)OXY]PROPAN-1-AMINE",
            "non-polymer",
        ),
        "SQX": ("13-DESMETHYL SPIROLIDE C", "non-polymer"),
        "MM3": (
            "N-HYDROXY-4-[(4-METHOXYLPHENYL)SULFONYL]-2,2-DIMETHYL-HEXAHYDRO-1,4-THIAZEPINE-3(S)-CARBOXAMIDE",
            "non-polymer",
        ),
        "GS5": (
            "5-CHLORO-N-{(3S)-1-[(1S)-1-METHYL-2-MORPHOLIN-4-YL-2-OXOETHYL]-2-OXOPYRROLIDIN-3-YL}-1-BENZOTHIOPHENE-2-SULFONAMIDE",
            "non-polymer",
        ),
        "GLG": ("ALPHA-D-GLUCOPYRANOSYL-2-CARBOXYLIC ACID AMIDE", "saccharide"),
        "T3A": (
            "N-(TRIS(HYDROXYMETHYL)METHYL)-3-AMINOPROPANESULFONIC ACID",
            "non-polymer",
        ),
        "GS4": ("4-THIO-BETA-D-GLUCOPYRANOSE", "saccharide"),
        "GLX": ("GLU/GLN AMBIGUOUS", "peptide"),
        "Q74": (
            "2-(3-((4,5,7-TRIFLUOROBENZO[D]THIAZOL-2-YL)METHYL)-1H-PYRROLO[2,3-B]PYRIDIN-1-YL)ACETIC ACID",
            "non-polymer",
        ),
        "2CY": ("5-[3-(3-FLUOROPHENOXY)PROPOXY]QUINAZOLINE-2,4-DIAMINE", "non-polymer"),
        "2CT": (
            "(2-{2-[(2,5-DIMETHOXYPHENYL)AMINO]-2-OXOETHOXY}PHENYL)PHOSPHONIC ACID",
            "non-polymer",
        ),
        "8NX": ("8-NITRO-3,7-DIHYDRO-1H-PURINE-2,6-DIONE", "non-polymer"),
        "2CV": ("HEGA-10", "non-polymer"),
        "2CP": ("2-CARBOXYPROPYL-COENZYME A", "non-polymer"),
        "GLZ": ("AMINO-ACETALDEHYDE", "peptide"),
        "2CS": (
            "3-[3-(TERT-BUTYLTHIO)-1-(4-CHLOROBENZYL)-5-(QUINOLIN-2-YLMETHOXY)-1H-INDOL-2-YL]-2,2-DIMETHYLPROPANOIC ACID",
            "non-polymer",
        ),
        "2CL": ("(2,6-DICHLOROPHENYL)ACETIC ACID", "non-polymer"),
        "2CM": ("2-CHLORO-6-METHYL-ANILINE", "non-polymer"),
        "2CN": ("5-BROMO-2-{[3-(OCTYLOXY)BENZYL]SULFANYL}BENZOIC ACID", "non-polymer"),
        "2CO": ("S-HYDROPEROXYCYSTEINE", "peptide"),
        "2CH": ("2-CHLOROPHENOL", "non-polymer"),
        "2CJ": (
            "(1-{2-[4-(DIPHENYLMETHYL)PIPERAZIN-1-YL]-2-OXOETHYL}CYCLOPENTYL)ACETIC ACID",
            "non-polymer",
        ),
        "2CD": (
            "3,5-DICHLORO-3'-[(DIETHYLAMINO)METHYL]-N-(1,3,5-TRIMETHYL-1H-PYRAZOL-4-YL)BIPHENYL-4-SULFONAMIDE",
            "non-polymer",
        ),
        "8NH": (
            "2-CHLORO-4-[(7R,7AS)-7-HYDROXY-1,3-DIOXOTETRAHYDRO-1H-PYRROLO[1,2-C]IMIDAZOL-2(3H)-YL]-3-METHYLBENZONITRILE",
            "non-polymer",
        ),
        "2CF": (
            "(2E,6E)-2-FLUORO-3,7,11-TRIMETHYLDODECA-2,6,10-TRIEN-1-YL TRIHYDROGEN DIPHOSPHATE",
            "non-polymer",
        ),
        "2CG": (
            "(1R,2R)-3-[(S)-AMINO(CARBOXY)METHYL]CYCLOPROPANE-1,2-DICARBOXYLIC ACID",
            "non-polymer",
        ),
        "N34": (
            "N-{3-[1-(3,3-DIMETHYL-BUTYL)-4-HYDROXY-2-OXO-1,2,4A,5,6,7-HEXAHYDRO-PYRROLO[1,2-B]PYRIDAZIN-3-YL]-1,1-DIOXO-1,2-DIHYDRO-1LAMBDA6-BENZO[1,2,4]THIADIAZIN-7-YL}-METHANESULFONAMIDE",
            "non-polymer",
        ),
        "2CA": (
            "BENZOFURAN-2-CARBOXYLIC ACID {(S)-3-METHYL-1-[3-OXO-1-(PYRIDIN-2-YLSULFONYL)AZEPAN-4-YLCARBAMOYL]BUTYL}AMIDE",
            "non-polymer",
        ),
        "2CB": (
            "6-{[2-(4-METHYLPIPERAZIN-1-YL)ETHYL]AMINO}-N-(1,3,5-TRIMETHYL-1H-PYRAZOL-4-YL)PYRIDINE-3-SULFONAMIDE",
            "non-polymer",
        ),
        "2CC": (
            "3,4-BIS(7-CHLORO-1H-INDOL-3-YL)-1H-PYRROLE-2,5-DICARBOXYLIC ACID",
            "non-polymer",
        ),
        "RP1": (
            "6-(6-AMINO-PURIN-9-YL)-2-THIOXO-TETRAHYDRO-2-FURO[3,2-D][1,3,2]DIOXAPHOSPHININE-2,7-DIOL",
            "non-polymer",
        ),
        "L9Q": (
            "(1S)-2-{[(S)-(2-AMINOETHOXY)(HYDROXY)PHOSPHORYL]OXY}-1-[(OCTADECANOYLOXY)METHYL]ETHYL (9Z)-OCTADEC-9-ENOATE",
            "non-polymer",
        ),
        "0IN": ("7-(PYRIDIN-3-YLMETHYL)QUINOLIN-8-OL", "non-polymer"),
        "L9X": (
            "L-GAMMA-GLUTAMYL-S-[(3R)-1-ETHYL-2,5-DIOXOPYRROLIDIN-3-YL]-L-CYSTEINYLGLYCINE",
            "non-polymer",
        ),
        "TI1": ("[2(R,S)-2-SULFANYLHEPTANOYL]-PHE-ALA", "non-polymer"),
        "U37": ("?", "nucleic-acid"),
        "GLP": ("GLUCOSAMINE 6-PHOSPHATE", "saccharide"),
        "L9G": (
            "3-{6-[2-(2,4-DIFLUOROPHENYL)ETHYL][1,2,4]TRIAZOLO[4,3-A]PYRIDIN-3-YL}-4-METHYLBENZAMIDE",
            "non-polymer",
        ),
        "U33": ("?", "nucleic-acid"),
        "TI2": ("(2-SULFANYL-3-PHENYLPROPANOYL)-PHE-TYR", "non-polymer"),
        "U31": ("?", "nucleic-acid"),
        "GLQ": ("4-AMINO-5-OXO-PENTANOIC ACID", "peptide"),
        "L9L": (
            "5-[2-({(2S)-2-AMINO-3-[4-(TRIFLUOROMETHYL)PHENYL]PROPYL}AMINO)-1,3-THIAZOL-5-YL]-1,3-DIHYDRO-2H-INDOL-2-ONE",
            "non-polymer",
        ),
        "L9M": (
            "(2S)-N~1~-[5-(3-METHYL-1H-INDAZOL-5-YL)-1,3,4-THIADIAZOL-2-YL]-3-(4-METHYLPHENYL)PROPANE-1,2-DIAMINE",
            "non-polymer",
        ),
        "L9N": (
            "(2S)-N~1~-[5-(1H-INDAZOL-5-YL)-1,3,4-THIADIAZOL-2-YL]-3-(4-METHYLPHENYL)PROPANE-1,2-DIAMINE",
            "non-polymer",
        ),
        "0IT": (
            "AMINO({(4S)-4-[({(3S)-3-[(BENZYLSULFONYL)AMINO]-2-OXOPIPERIDIN-1-YL}ACETYL)AMINO]-5-OXOPENTYL}AMINO)METHANIMINIUM",
            "peptide",
        ),
        "0IU": (
            "(2S)-2-[(2-AMINO-1,3-THIAZOL-4-YL)METHYL]-N~1~-[(1S,2R,3R)-1-(CYCLOHEXYLMETHYL)-2,3-DIHYDROXY-5-METHYLHEXYL]-N~4~-[2-(DIMETHYLAMINO)-2-OXOETHYL]-N~4~-[(1S)-1-PHENYLETHYL]BUTANEDIAMIDE",
            "non-polymer",
        ),
        "0IV": (
            "2-{(3S)-3-[(BENZYLSULFONYL)AMINO]-2-OXOPIPERIDIN-1-YL}-N-{(2S)-1-[(3R)-1-CARBAMIMIDOYLPIPERIDIN-3-YL]-3-OXOPROPAN-2-YL}ACETAMIDE",
            "peptide",
        ),
        "0IW": (
            "NALPHA-[(4-METHYLPIPERAZIN-1-YL)CARBONYL]-N-[(3S)-1-PHENYL-5-(PHENYLSULFONYL)PENTAN-3-YL]-L-PHENYLALANINAMIDE",
            "peptide",
        ),
        "7RG": (
            "6-[(3S,4S)-4-METHYL-1-(PYRIMIDIN-2-YLMETHYL)PYRROLIDIN-3-YL]-1-(TETRAHYDRO-2H-PYRAN-4-YL)-1,5-DIHYDRO-4H-PYRAZOLO[3,4-D]PYRIMIDIN-4-ONE",
            "non-polymer",
        ),
        "7RA": ("7-ALPHA-D-RIBOFURANOSYL-2-AMINOPURINE-5'-PHOSPHATE", "non-polymer"),
        "GLT": ("5-DEOXY-5-THIO-ALPHA-D-GLUCOSE", "saccharide"),
        "RP6": (
            "5-(ACETYLAMINO)-2,6-ANHYDRO-3,5-DIDEOXY-3-PROP-2-EN-1-YL-D-GLYCERO-D-GALACTO-NON-2-ENONIC ACID",
            "non-polymer",
        ),
        "VXQ": ("1-(2,3-DIHYDRO-1-BENZOFURAN-5-YL)METHANAMINE", "non-polymer"),
        "0P0": ("N-[(4R)-4-AMMONIO-4-CARBOXYBUTANOYL]-L-CYSTEINYLGLYCINE", "peptide"),
        "T4E": (
            "4-(3-AMINO-5-PHENYL-1,2,4-TRIAZIN-6-YL)-2-CHLOROPHENOL",
            "non-polymer",
        ),
        "7RP": ("7-ALPHA-D-RIBOFURANOSYL-PURINE-5'-PHOSPHATE", "non-polymer"),
        "GLW": ("6-DEOXYGLUCOSE", "saccharide"),
        "KVS": (
            "N~2~-[(2R,5S)-5-({(2S,3S)-2-[(N-ACETYL-L-THREONYL)AMINO]-3-METHYLPENT-4-ENOYL}AMINO)-2-BUTYL-4,4-DIHYDROXYNONANOYL]-L-GLUTAMINYL-L-ARGININAMIDE",
            "peptide",
        ),
        "XFI": (
            "N-((1S,2R)-3-(((1S)-2-(CYCLOHEXYLAMINO)-1-METHYL-2-OXOETHYL)AMINO)-2-HYDROXY-1-( PHENYLMETHYL)PROPYL)-3-((METHYLSULFONYL)(PHENYL)AMINO) BENZAMIDE",
            "non-polymer",
        ),
        "MN1": ("4-CARBOXYPIPERIDINE", "non-polymer"),
        "XFJ": (
            "6,6'-(PYRIDINE-3,5-DIYLDIETHANE-2,1-DIYL)BIS(4-METHYLPYRIDIN-2-AMINE)",
            "non-polymer",
        ),
        "XFL": (
            "6,6'-(PYRIDINE-2,6-DIYLDIETHANE-2,1-DIYL)BIS(4-METHYLPYRIDIN-2-AMINE)",
            "non-polymer",
        ),
        "L94": (
            "N'-{(Z)-AMINO[4-(AMINO{[3-(DIMETHYLAMMONIO)PROPYL]IMINIO}METHYL)PHENYL]METHYLIDENE}-N,N-DIMETHYLPROPANE-1,3-DIAMINIUM",
            "non-polymer",
        ),
        "HXS": ("4-[(1R,2S)-1-ETHYL-2-(4-HYDROXYPHENYL)BUTYL]PHENOL", "non-polymer"),
        "L90": (
            "8-BUTYL-1-METHYL-7-(5-METHYL-1H-INDAZOL-4-YL)-1H-IMIDAZO[2,1-F]PURINE-2,4(3H,8H)-DIONE",
            "non-polymer",
        ),
        "L92": (
            "(2S)-3-{4-[3-(5-METHYL-2-PHENYL-1,3-OXAZOL-4-YL)PROPYL]PHENYL}-2-(1H-PYRROL-1-YL)PROPANOIC ACID",
            "non-polymer",
        ),
        "MN2": ("1-CARBOXYETHYLAMINOMETHYL-4-AMINOMETHYLBENZENE", "non-polymer"),
        "6HT": (
            "1',5'-ANHYDRO-2',3'-DIDEOXY-2'-(THYMIN-1-YL)-6'-O-PHOSPHORYL-D-ARABINO-HEXITOL",
            "nucleic-acid",
        ),
        "XFN": (
            "6,6'-[(5-AMINOBENZENE-1,3-DIYL)DIETHANE-2,1-DIYL]BIS(4-METHYLPYRIDIN-2-AMINE)",
            "non-polymer",
        ),
        "L98": (
            "(2R)-2-(3-{[AMINO(IMINO)METHYL]AMINO}PHENYL)-3-SULFANYLPROPANOIC ACID",
            "non-polymer",
        ),
        "6HS": (
            "(1S,2S,3R,4S,6R)-4,6-DIAMINO-3-{[(2S,3R)-3-AMINO-6-(HYDROXYMETHYL)-3,4-DIHYDRO-2H-PYRAN-2-YL]OXY}-2-HYDROXYCYCLOHEXYL 3-DEOXY-4-C-METHYL-3-(METHYLAMINO)-BETA-L-ARABINOPYRANOSIDE",
            "non-polymer",
        ),
        "U3H": ("?", "non-polymer"),
        "6HN": ("6-NITRO-L-NORLEUCINE", "peptide"),
        "6HI": (
            "(3R,5R)-7-{2-[(4-FLUOROBENZYL)CARBAMOYL]-4-(4-FLUOROPHENYL)-1-(1-METHYLETHYL)-1H-IMIDAZOL-5-YL}-3,5-DIHYDROXYHEPTANOIC ACID",
            "non-polymer",
        ),
        "6HK": (
            "METHYL [3-(1-METHYL-1H-IMIDAZOL-5-YL)-11-OXO-10,11-DIHYDRO-5H-DIBENZO[B,E][1,4]DIAZEPIN-8-YL]ACETATE",
            "non-polymer",
        ),
        "U3P": ("3'-URIDINEMONOPHOSPHATE", "non-polymer"),
        "6HE": ("6-METHY-6-DEPROPIONATEHEMIN", "non-polymer"),
        "6HG": (
            "1',5'-ANHYDRO-2',3'-DIDEOXY-2'-(GUANIN-9-YL)-6'-O-PHOSPHORYL-D-ARABINO-HEXITOL",
            "nucleic-acid",
        ),
        "6HA": (
            "1',5'-ANHYDRO-2',3'-DIDEOXY-2'-(ADENIN-9-YL)-6'-O-PHOSPHORYL-D-ARABINO-HEXITOL",
            "nucleic-acid",
        ),
        "0I5": (
            "N-[(3S)-1-FLUORO-2-OXO-5-PHENYLPENTAN-3-YL]-N~2~-(MORPHOLIN-4-YLCARBONYL)-L-LEUCINAMIDE",
            "peptide",
        ),
        "6HC": (
            "1',5'-ANHYDRO-2',3'-DIDEOXY-2'-(CYTOSIN-1-YL)-6'-O-PHOSPHORYL-D-ARABINO-HEXITOL",
            "nucleic-acid",
        ),
        "PED": ("PENTANE-3,4-DIOL-5-PHOSPHATE", "non-polymer"),
        "TO2": (
            "(1R,2S)-1-(3,4-DIMETHOXYPHENYL)-2-(2-METHOXYPHENOXY)PROPANE-1,3-DIOL",
            "non-polymer",
        ),
        "YET": (
            "1-CYANO-2-(2-METHYL-1-BENZOFURAN-5-YL)-3-[(3S)-2-OXO-1-(2-OXO-2-PYRROLIDIN-1-YLETHYL)AZEPAN-3-YL]GUANIDINE",
            "non-polymer",
        ),
        "144": ("TRIS-HYDROXYMETHYL-METHYL-AMMONIUM", "non-polymer"),
        "EEU": ("ESTRADIOL-PYRIDINIUM TETRAACETIC ACID", "non-polymer"),
        "H02": ("N-METHYL PROTOPORPHYRIN IX 2,4-DISULFONIC ACID", "non-polymer"),
        "H01": ("PROTOPORPHYRIN IX 2,4-DISULFONIC ACID", "non-polymer"),
        "P61": (
            "(2E)-3-(3-FLUORO-4-HYDROXYPHENYL)-2-{[(Z)-{3-HYDROXY-2-METHYL-5-[(PHOSPHONOOXY)METHYL]PYRIDIN-4(1H)-YLIDENE}METHYL]IMINO}PROPANOIC ACID",
            "non-polymer",
        ),
        "FFO": (
            "N-[4-({[(6S)-2-AMINO-5-FORMYL-4-OXO-3,4,5,6,7,8-HEXAHYDROPTERIDIN-6-YL]METHYL}AMINO)BENZOYL]-L-GLUTAMIC ACID",
            "non-polymer",
        ),
        "NA": ("SODIUM ION", "non-polymer"),
        "OMO": ("MO(VI)(=O)(OH)2 CLUSTER", "non-polymer"),
        "EES": ("3-CHLORO-2-(4-HYDROXYPHENYL)-2H-INDAZOL-5-OL", "non-polymer"),
        "UBA": (
            "(S)-1-(2-AMINO-2-CARBOXYETHYL)-3(2-CARBOXYTHIOPHENE-3-YL-METHYL)-5-METHYLPYRIMIDINE-2,4-DIONE",
            "non-polymer",
        ),
        "TO1": (
            "4-AMINO-7-(BETA-D-RIBOFURANOSYL)-7H-PYRROLO[2,3-D]PYRIMIDINE-5-CARBONITRILE",
            "non-polymer",
        ),
        "JRG": (
            "N~2~-{[(1S)-1-CARBOXY-3-(METHYLSULFANYL)PROPYL]CARBAMOYL}-N~6~-(4-IODOBENZOYL)-L-LYSINE",
            "non-polymer",
        ),
        "JRO": ("6,7,9-TRIHYDROXY-3-METHYL-1H-BENZO[G]ISOCHROMEN-1-ONE", "non-polymer"),
        "LLL": (
            "(2R,3R,4R,5R)-2-((1S,2S,3R,4S,6R)-4,6-DIAMINO-3-((2R,3R,6S)-3-AMINO-6-(AMINOMETHYL)-TETRAHYDRO-2H-PYRAN-2-YLOXY)-2-HYDROXYCYCLOHEXYLOXY)-5-METHYL-4-(METHYLAMINO)-TETRAHYDRO-2H-PYRAN-3,5-DIOL",
            "non-polymer",
        ),
        "JRS": (
            "6-{[(3S,4R)-4-(2-{[2-(3-FLUOROPHENYL)ETHYL]AMINO}ETHOXY)PYRROLIDIN-3-YL]METHYL}-4-METHYLPYRIDIN-2-AMINE",
            "non-polymer",
        ),
        "JRR": (
            "6-{[(3R,4R)-4-(2-{[2-(3-FLUOROPHENYL)ETHYL]AMINO}ETHOXY)PYRROLIDIN-3-YL]METHYL}-4-METHYLPYRIDIN-2-AMINE",
            "non-polymer",
        ),
        "MHE": ("5,5-DIMETHYL-2-OXOHEXANOIC ACID", "peptide"),
        "JRZ": (
            "(2S,3S)-1-[(7-METHYL-2,3-DIHYDRO-1H-INDEN-4-YL)OXY]-3-[(1-METHYLETHYL)AMINO]BUTAN-2-OL",
            "non-polymer",
        ),
        "HH3": ("(2S)-HYDROXY[(PHENYLCARBONYL)AMINO]ETHANOIC ACID", "non-polymer"),
        "HH2": ("6-HYDROXYMETHYLPTERIN-DIPHOSPHATE", "non-polymer"),
        "HH1": (
            "(3S)-TETRAHYDROFURAN-3-YL (1R)-3-{(2R)-4-[(1S,3S)-3-(2-AMINO-2-OXOETHYL)-2,3-DIHYDRO-1H-INDEN-1-YL]-2-BENZYL-3-OXO-2,3-DIHYDRO-1H-PYRROL-2-YL}-1-BENZYL-2-HYDROXYPROPYLCARBAMATE",
            "non-polymer",
        ),
        "HH0": (
            "METHYL N-[(2S)-2-({(S)-HYDROXY[(1R)-3-METHYL-1-{[N-(3-METHYLBUTANOYL)-L-VALYL-L-VALYL]AMINO}BUTYL]PHOSPHORYL}OXY)-3-PHENYLPROPANOYL]-L-ALANYL-L-ALANINATE",
            "peptide",
        ),
        "HH6": ("3-(4-HYDROXYPHENYL)-1H-INDAZOL-6-OL", "non-polymer"),
        "T2S": ("THYMIDINE-5'-DITHIOPHOSPHORATE", "nucleic-acid"),
        "NK0": (
            "3-[(6-BENZOYL-5-METHYLPYRROLO[2,1-F][1,2,4]TRIAZIN-4-YL)AMINO]-N-CYCLOPROPYL-4-METHYLBENZAMIDE",
            "non-polymer",
        ),
        "NK1": (
            "(1R,5S,6S,7R,8S)-1-THIONIABICYCLO[4.3.0]NONAN-5,7,8-TRIOL",
            "non-polymer",
        ),
        "NK2": ("(1R,6S,7R,8S)-1-THIONIABICYCLO[4.3.0]NONAN-7,8-DIOL", "non-polymer"),
        "NK7": (
            "4-{[4-(TRIFLUOROMETHYL)BENZYL][(3S,4S)-4-{[4-(TRIFLUOROMETHYL)BENZYL]AMINO}PYRROLIDIN-3-YL]SULFAMOYL}BENZAMIDE",
            "non-polymer",
        ),
        "NK8": (
            "4-({(3S,4S)-4-[(4-CARBAMOYLBENZYL)AMINO]PYRROLIDIN-3-YL}[4-(TRIFLUOROMETHYL)BENZYL]SULFAMOYL)BENZAMIDE",
            "non-polymer",
        ),
        "NK9": (
            "4-({(3S,4S)-4-[(3,5-DIHYDROXYBENZYL)AMINO]PYRROLIDIN-3-YL}[4-(TRIFLUOROMETHYL)BENZYL]SULFAMOYL)BENZAMIDE",
            "non-polymer",
        ),
        "MMP": ("N-METHYLMESOPORPHYRIN", "non-polymer"),
        "A15": ("3'-3-DICHLOROPHENOL-1,8-3H-BENZO[DE]ISOCHROMEN-1-ONE", "non-polymer"),
        "A14": (
            "(1S,4S,5S,7R)-7-{[(5S)-5-AMINO-5-CARBOXYPENTANOYL]AMINO}-3,3-DIMETHYL-6-OXO-2-THIABICYCLO[3.2.0]HEPTANE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "A17": (
            "N-CYCLOPROPYL-4-METHYL-3-[1-(2-METHYLPHENYL)PHTHALAZIN-6-YL]BENZAMIDE",
            "non-polymer",
        ),
        "A16": (
            "4-O-(4,6-DIDEOXY-4-{[(1S,2S,3S,4R,5S)-2,3,4-TRIHYDROXY-5-(HYDROXYMETHYL)CYCLOHEXYL]AMINO}-ALPHA-D-GLUCOPYRANOSYL)-BETA-D-GLUCOPYRANOSE ",
            "non-polymer",
        ),
        "A11": (
            "ETHYL 4-[(4-CHLOROPYRIDIN-2-YL)AMINO]PIPERIDINE-1-CARBOXYLATE",
            "non-polymer",
        ),
        "ZCT": ("?", "non-polymer"),
        "A13": ("4-METHYL-6-(TRIFLUOROMETHYL)PYRIMIDIN-2-AMINE", "non-polymer"),
        "A12": ("PHOSPHOMETHYLPHOSPHONIC ACID ADENOSYL ESTER", "non-polymer"),
        "LLB": (
            "[6-HYDROXY-2-(4-HYDROXYPHENYL)-1-BENZOTHIEN-3-YL]{4-[2-(4-METHYLPIPERIDIN-1-YL)ETHOXY]PHENYL}METHANONE",
            "non-polymer",
        ),
        "A19": (
            "5-BROMO-2-{[(4-CHLOROPHENYL)SULFONYL]AMINO}BENZOIC ACID",
            "non-polymer",
        ),
        "A18": ("5-(2-CHLOROBENZYL)-2-FUROIC ACID", "non-polymer"),
        "GV9": ("3-(ACETYLAMINO)THIOPHENE-2-CARBOXYLIC ACID", "non-polymer"),
        "HNL": ("5-[(2S)-1-METHYLPYRROLIDIN-2-YL]PYRIDIN-2-OL", "non-polymer"),
        "VZ5": (
            "(14BETA,17ALPHA)-2-ETHYL-17-HYDROXYESTRA-1(10),2,4-TRIEN-3-YL SULFAMATE",
            "non-polymer",
        ),
        "VZ4": (
            "(9BETA)-2-ETHYL-17-OXOESTRA-1(10),2,4-TRIEN-3-YL SULFAMATE",
            "non-polymer",
        ),
        "GVJ": ("(S)-1-PHENYL-1-[4-(9H-PURIN-6-YL)PHENYL]METHANAMINE", "non-polymer"),
        "A1T": (
            "5-PENTYL-N-{[4'-(PIPERIDIN-1-YLCARBONYL)BIPHENYL-4-YL]METHYL}-N-[1-(PYRIDIN-2-YLMETHYL)PIPERIDIN-4-YL]PYRIDINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "GVH": ("1H-PYRROLO[2,3-B]PYRIDINE", "non-polymer"),
        "GVI": ("N-METHYL-1-[4-(9H-PURIN-6-YL)PHENYL]METHANAMINE", "non-polymer"),
        "A1Q": ("METHYL L-GLYCERO-ALPHA-D-MANNO-HEPTOPYRANOSIDE", "saccharide"),
        "A1P": (
            "9-{2-DEOXY-5-O-[HYDROXY(OXIDO)PHOSPHINO]-BETA-L-ERYTHRO-PENTOFURANOSYL}-9H-PURIN-2-AMINE",
            "nucleic-acid",
        ),
        "GVL": (
            "O-[(R)-{[(3R)-4-AMINO-3-HYDROXY-2,2-DIMETHYL-4-OXOBUTYL]OXY}(HYDROXY)PHOSPHORYL]-L-SERINE",
            "peptide",
        ),
        "A1R": (
            "5'-O-[(S)-{[(S)-{[(2R,3R,4S)-3,4-DIHYDROXYPYRROLIDIN-2-YL]METHOXY}(HYDROXY)PHOSPHORYL]OXY}(HYDROXY)PHOSPHORYL]ADENOSINE",
            "non-polymer",
        ),
        "GVB": (
            "(3R,4S)-1-(3,4-DIMETHOXYPHENYL)-3-(3-METHYLPHENYL)PIPERIDIN-4-AMINE",
            "non-polymer",
        ),
        "GVC": (
            "1-(3,5-DICHLOROPHENYL)-5-METHYL-1H-1,2,4-TRIAZOLE-3-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "GVA": ("5-{[AMINO(IMINO)METHYL]AMINO}PENTANOIC ACID", "non-polymer"),
        "GVF": (
            "BENZENE-1,2,3,4-TETRAYL TETRAKIS[DIHYDROGEN (PHOSPHATE)]",
            "non-polymer",
        ),
        "A1X": (
            "ETHYL (2S)-2-HYDROXY-4-[1-(2-METHYLPROPYL)HYDRAZINO]-4-OXOBUTANOATE",
            "non-polymer",
        ),
        "GVD": (
            "[4-({4-[(5-CYCLOPROPYL-1H-PYRAZOL-3-YL)AMINO]QUINAZOLIN-2-YL}IMINO)CYCLOHEXA-2,5-DIEN-1-YL]ACETONITRILE",
            "non-polymer",
        ),
        "GVE": ("METHYL 4-AMINOBUTANOATE", "peptide"),
        "A1E": (
            "(5R)-5-{[10-(1,2,3,4-TETRAHYDROACRIDIN-9-YLAMINO)DECYL]AMINO}-5,6,7,8-TETRAHYDROQUINOLIN-2(1H)-ONE",
            "non-polymer",
        ),
        "TND": (
            "N-1,2,3,4-TETRAHYDRONAPHTH-1-YL-2'-[3,5-DIMETHOXYBENZAMIDO]-2'-DEOXY-ADENOSINE",
            "non-polymer",
        ),
        "TNG": ("PROPANE-1,2,3-TRIYL TRINITRATE", "non-polymer"),
        "LLX": ("N-(4-AMINOBIPHENYL-3-YL)BENZAMIDE", "non-polymer"),
        "A1A": ("6-AMINO HEXANOIC ACID", "non-polymer"),
        "TNC": (
            "4-DIMETHYLAMINO-1,10,11,12-TETRAHYDROXY-3-OXO-3,4,4A,5-TETRAHYDRO-NAPHTHACENE-2-CARBOXYLIC ACID AMIDE",
            "non-polymer",
        ),
        "TNB": ("S-(2,3,6-TRINITROPHENYL)CYSTEINE", "peptide"),
        "GVR": (
            "(2R)-N-HYDROXY-3-NAPHTHALEN-2-YL-2-[(NAPHTHALEN-2-YLSULFONYL)AMINO]PROPANAMIDE",
            "non-polymer",
        ),
        "GVS": (
            "ETHYL {3-[4-AMINO-5-{3-[(CYCLOPROPYLSULFONYL)AMINO]PROP-1-YN-1-YL}-2-OXOPYRIMIDIN-1(2H)-YL]OXETAN-3-YL}ACETATE",
            "non-polymer",
        ),
        "GVP": (
            "4-(4-CHLOROPHENYL)-4-[4-(1H-PYRAZOL-4-YL)PHENYL]PIPERIDINE",
            "non-polymer",
        ),
        "GVQ": ("(2R)-2-(4-CHLOROPHENYL)-2-PHENYLETHANAMINE", "non-polymer"),
        "OTR": ("4-(2R-AMINO-1-HYDROXYETHYL)PHENOL", "non-polymer"),
        "GVT": ("3,6,9,12,15,18,21,24,27-NONAOXAHEPTATRIACONTAN-1-OL", "non-polymer"),
        "HHS": ("6-CARBOXYPTERIN", "non-polymer"),
        "HHR": ("6-HYDROXYMETHYLPTERIN", "non-polymer"),
        "HHP": ("6-HYDROXYMETHYL-7,8-DIHYDROPTERIN", "non-polymer"),
        "HHV": (
            "5-[(PYRIDIN-3-YLMETHYL)AMINO]-1H-PYRAZOLE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "NKK": (
            "N~2~-(NAPHTHALEN-2-YLCARBONYL)-L-LYSYL-N-[(1S)-4-CARBAMIMIDAMIDO-1-FORMYLBUTYL]-L-LYSINAMIDE",
            "non-polymer",
        ),
        "NKN": ("(2R)-2-HYDROXY-3-(PHOSPHONOOXY)PROPYL TETRADECANOATE", "non-polymer"),
        "NKO": ("(2R)-2-HYDROXY-3-(PHOSPHONOOXY)PROPYL HEXADECANOATE", "non-polymer"),
        "NKP": (
            "(2R)-2-HYDROXY-3-(PHOSPHONOOXY)PROPYL (9E)-OCTADEC-9-ENOATE",
            "non-polymer",
        ),
        "NKQ": (
            "(2R)-2-HYDROXY-3-(PHOSPHONOOXY)PROPYL (9E,12Z,15Z)-OCTADECA-9,12,15-TRIENOATE",
            "non-polymer",
        ),
        "HHA": ("(2S,3S)-TRANS-2,3-DIHYDRO-3-HYDROXYANTHRANILIC ACID", "non-polymer"),
        "NKS": (
            "N-(3-{[2-(4-CHLOROPHENYL)-5-METHYL-1,3-OXAZOL-4-YL]METHOXY}BENZYL)-N-[(4-METHYLPHENOXY)CARBONYL]GLYCINE",
            "peptide",
        ),
        "HHG": ("(2R)-2-HYDROXY-3-(PHOSPHONOOXY)PROPYL HEXANOATE", "non-polymer"),
        "ROA": (
            "(2R)-3-(3,4-DIHYDROXYPHENYL)-2-{[(2E)-3-(3,4-DIHYDROXYPHENYL)PROP-2-ENOYL]OXY}PROPANOIC ACID",
            "non-polymer",
        ),
        "HHK": ("(2S)-2,8-DIAMINOOCTANOIC ACID", "peptide"),
        "HHI": ("(2Z)-3-(1H-IMIDAZOL-5-YL)-2-IMINOPROPANOIC ACID", "peptide"),
        "HHH": ("(2S)-HYDROXY(4-HYDROXYPHENYL)ETHANOIC ACID", "non-polymer"),
        "HHO": ("1-(HYDROXYMETHYLENEAMINO)-8-HYDROXY-OCTANE", "non-polymer"),
        "ROB": ("1-DEOXY-RIBOFURANOSE-5'-PHOSPHATE", "saccharide"),
        "AXR": ("METHYL ALPHA-D-ARABINOFURANOSIDE", "saccharide"),
        "LLG": ("2-AMINOETHYL NAPHTHALEN-1-YLACETATE", "non-polymer"),
        "MZ5": (
            "(2S)-N-[(1S,2R)-3-{(1,3-BENZOTHIAZOL-6-YLSULFONYL)[(2S)-2-METHYLBUTYL]AMINO}-1-BENZYL-2-HYDROXYPROPYL]-2-HYDROXY-3-METHYLBUTANAMIDE",
            "non-polymer",
        ),
        "FF3": (
            "1-(5-AMINO-1,3,4-THIADIAZOL-2-YL)-1,1-DIFLUOROMETHANESULFONAMIDE",
            "non-polymer",
        ),
        "2PS": ("PHENYL (3S)-3-AMINO-5-PHENYLPENTANE-1-SULFONATE", "peptide"),
        "FF0": (
            "(2R,3S)-3-AMINO-4-PHENYL-1-[(3-PHENYLPROPYL)AMINO]BUTAN-2-OL",
            "non-polymer",
        ),
        "011": ("7-AMINOHEPTANOIC ACID", "peptide"),
        "012": (
            "(4S)-N-[(1S,2R)-1-BENZYL-3-{[3-(DIMETHYLAMINO)BENZYL]AMINO}-2-HYDROXYPROPYL]-1-(3-METHOXYBENZYL)-2-OXOIMIDAZOLIDINE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "013": ("1,3-THIAZOL-5-YLMETHYL HYDROGEN CARBONATE", "non-polymer"),
        "014": (
            "1-(5-CHLORO-6-FLUORO-1H-BENZIMIDAZOL-2-YL)-1H-PYRAZOLE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "015": ("(2S,3S,5S)-2,5-DIAMINO-1,6-DIPHENYLHEXAN-3-OL", "peptide"),
        "016": (
            "(4R)-3-[(2R,3S)-3-{[(2,6-DIMETHYLPHENOXY)ACETYL]AMINO}-2-HYDROXY-4-PHENYLBUTANOYL]-N-[(1S,2R)-2-HYDROXY-2,3-DIHYDRO-1H-INDEN-1-YL]-5,5-DIMETHYL-1,3-THIAZOLIDINE-4-CARBOXAMIDE",
            "peptide",
        ),
        "017": (
            "(3R,3AS,6AR)-HEXAHYDROFURO[2,3-B]FURAN-3-YL(1S,2R)-3-[[(4-AMINOPHENYL)SULFONYL](ISOBUTYL)AMINO]-1-BENZYL-2-HYDROXYPROPYLCARBAMATE",
            "non-polymer",
        ),
        "018": (
            "2-NITRO-N-(THIOPHEN-3-YLMETHYL)-4-(TRIFLUOROMETHYL)ANILINE",
            "non-polymer",
        ),
        "019": (
            "METHYL{[2-(PROPAN-2-YL)-1,3-THIAZOL-4-YL]METHYL}CARBAMIC ACID",
            "peptide",
        ),
        "LA1": (
            "(2R)-2-[3-ISOBUTYL-2,5-DIOXO-4-(QUINOLIN-3-YLMETHYL)-1,4-DIAZEPAN-1-YL]-N-METHYL-3-(2-NAPHTHYL)PROPANAMIDE",
            "non-polymer",
        ),
        "LA2": ("N~6~-[(6R)-6,8-DISULFANYLOCTANOYL]-L-LYSINE", "peptide"),
        "AXW": (
            "6-{[(3S,4S)-4-{2-[(2-FLUOROBENZYL)AMINO]ETHOXY}PYRROLIDIN-3-YL]METHYL}-4-METHYLPYRIDIN-2-AMINE",
            "non-polymer",
        ),
        "LLF": (
            "(4-{(E)-[(2,2-DIFLUOROETHYL)IMINO]METHYL}-5-HYDROXY-6-METHYLPYRIDIN-3-YL)METHYL DIHYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "EH5": (
            "N-(1-BENZYL-3-{[3-(1,3-DIOXO-1,3-DIHYDRO-ISOINDOL-2-YL)-PROPIONYL]-[2-(HEXAHYDRO-BENZO[1,3]DIOXOL-5-YL)-ETHYL]-AMINO}-2-HYDROXY-PROPYL)-4-BENZYLOXY-3,5-DIMETHOXY-BENZAMIDE",
            "non-polymer",
        ),
        "MZ2": (
            "(2E)-N-[(1S,2R)-1-BENZYL-2-HYDROXY-3-{(2-THIENYLMETHYL)[(2,4,5-TRIFLUOROPHENYL)SULFONYL]AMINO}PROPYL]-4,4,4-TRIFLUORO-3-METHYLBUT-2-ENAMIDE",
            "non-polymer",
        ),
        "LLJ": (
            "6,22-DIOXA-3,9,19,25-TETRAAZONIAPENTACYCLO[25.5.3.3~11,17~.0~14,37~.0~30,34~]OCTATRIACONTA-1(33),11(38),12,14(37),15,17(36),27,29,31,34-DECAENE",
            "non-polymer",
        ),
        "UAP": (
            "4-DEOXY-2-O-SULFO-ALPHA-L-THREO-HEX-4-ENOPYRANURONIC ACID",
            "non-polymer",
        ),
        "UAR": ("?", "nucleic-acid"),
        "EH9": ("(2S,3R)-3-(6-AMINO-9H-PURIN-9-YL)NONAN-2-OL", "non-polymer"),
        "LLH": (
            "(2R,3S,4R)-2,3,4-TRIHYDROXY-5-(HYDROXYAMINO)-5-OXOPENTANOIC ACID",
            "non-polymer",
        ),
        "UAL": ("(2Z)-2-AMINO-3-(CARBAMOYLAMINO)PROP-2-ENOIC ACID", "peptide"),
        "UAA": ("URACIL-6-ACETIC ACID", "non-polymer"),
        "LLO": ("N~6~-[(1S)-3-HYDROXY-1,3-DIMETHYLBUTYL]-L-LYSINE", "peptide"),
        "UAD": (
            "[(2R,3S,4R,5R)-5-(2,4-DIOXO-3,4-DIHYDROPYRIMIDIN-1(2H)-YL)-3,4-DIHYDROXYTETRAHYDROFURAN-2-YL]METHYL (2S,3R,4S,5S)-3,4,5-TRIHYDROXYTETRAHYDRO-2H-PYRAN-2-YL DIHYDROGEN DIPHOSPHATE",
            "non-polymer",
        ),
        "UAG": (
            "URIDINE-5'-DIPHOSPHATE-N-ACETYLMURAMOYL-L-ALANINE-D-GLUTAMATE",
            "non-polymer",
        ),
        "04Z": (
            "4-{[4-AMINO-5-(3-METHOXYBENZOYL)-1,3-THIAZOL-2-YL]AMINO}BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "WZ4": (
            "METHYL 2-(ACETYLAMINO)-2-DEOXY-BETA-D-GLUCOPYRANOSYL-(1->2)-ALPHA-D-MANNOPYRANOSYL-(1->3)-[ALPHA-D-MANNOPYRANOSYL-(1->6)-6-THIO-ALPHA-D-MANNOPYRANOSYL-(1->6)]-BETA-D-MANNOPYRANOSIDE",
            "non-polymer",
        ),
        "719": (
            "N-[(E)-3-[(2R,3S,4R,5R)-3,4-DIHYDROXY-5-[6-(2-HYDROXYETHYLAMINO)PURIN-9-YL]OXOLAN-2-YL]PROP-2-ENYL]-5-(4-FLUOROPHENYL)-2,3-DIHYDROXY-BENZAMIDE",
            "non-polymer",
        ),
        "717": ("IMIDAZO[2,1-B][1,3]THIAZOL-6-YLMETHANOL", "non-polymer"),
        "716": (
            "2-[2-(3-CHLORO-PHENYL)-2-HYDROXY-ACETYLAMINO]-N-[4-GUANIDINO-1-(THIAZOLE-2-CARBONYL)-BUTYL]-3-METHYL-BUTYRAMIDE",
            "non-polymer",
        ),
        "715": (
            "(2R)-4-OXO-4-[3-(TRIFLUOROMETHYL)-5,6-DIHYDRO[1,2,4]TRIAZOLO[4,3-A]PYRAZIN-7(8H)-YL]-1-(2,4,5-TRIFLUOROPHENYL)BUTAN-2-AMINE",
            "non-polymer",
        ),
        "714": ("1-(2,2-DIPHOSPHONOETHYL)-3-(OCTYLOXY)PYRIDINIUM", "non-polymer"),
        "UA5": ("(4-ETHYLPHENYL)SULFAMIC ACID", "non-polymer"),
        "712": (
            "3-({[(1R)-1-(4-FLUOROPHENYL)ETHYL]AMINO}CARBONYL)-5-[METHYL(METHYLSULFONYL)AMINO]BENZYL ALPHA-METHYL-D-PHENYLALANINATE",
            "non-polymer",
        ),
        "711": (
            "7-[[6-[[1-(1-IMINOETHYL)PIPERIDIN-4-YL]OXY]-2-METHYL-BENZIMIDAZOL-1-YL]METHYL]NAPHTHALENE-2-CARBOXIMIDAMID",
            "non-polymer",
        ),
        "710": (
            "(5-{3-[5-(PIPERIDIN-1-YLMETHYL)-1H-INDOL-2-YL]-1H-INDAZOL-6-YL}-2H-1,2,3-TRIAZOL-4-YL)METHANOL",
            "non-polymer",
        ),
        "EHG": (
            "9-(2-DEOXY-5-O-PHOSPHONO-BETA-D-ERYTHRO-PENTOFURANOSYL)-6-(2-HYDROXYETHOXY)-9H-PURIN-2-AMINE",
            "nucleic-acid",
        ),
        "EHD": (
            "4-ETHYL-4-HYDROXY-1,12-DIHYDRO-4H-2-OXA-6,12A-DIAZA-DIBENZO[B,H]FLUORENE-3,13-DIONE",
            "non-polymer",
        ),
        "T48": ("?", "nucleic-acid"),
        "EHB": (
            "(1Z)-1-(3-ETHYL-5-HYDROXY-1,3-BENZOTHIAZOL-2(3H)-YLIDENE)PROPAN-2-ONE",
            "non-polymer",
        ),
        "EHA": (
            "(5-{3-[(6-BENZOYL-1-PROPYL-2-NAPHTHYL)OXY]PROPOXY}-1H-INDOL-1-YL)ACETIC ACID",
            "non-polymer",
        ),
        "EHN": ("ETHANE", "non-polymer"),
        "EHM": ("(2R,3R)-N,2,3,4-TETRAHYDROXYBUTANAMIDE", "non-polymer"),
        "LLA": ("L-2-HYDROXY-3-PHENYL-PROPIONIC ACID", "non-polymer"),
        "LAL": ("N,N-DIMETHYL-L-ALANINE", "peptide"),
        "01A": ("4-CHLOROPHENACYL-COENZYME A", "non-polymer"),
        "01B": ("(2S,3R)-3-AMINO-2-HYDROXY-4-PHENYLBUTANOIC ACID", "peptide"),
        "01C": (
            "(4S)-4-AMINO-1,1,1,2,2-PENTAFLUORO-5-(1H-INDOL-3-YL)PENTAN-3-ONE",
            "peptide",
        ),
        "01E": (
            "(2S)-2-(3,3-DIMETHYLBUTANOYLAMINO)-N-[(2S)-1-[[(2S,3S)-3-HYDROXY-4-[(4-IODOPHENYL)METHYLAMINO]-4-OXO-BUTAN-2-YL]AMINO]-1,4-DIOXO-4-PYRROL-1-YL-BUTAN-2-YL]-3,3-DIMETHYL-BUTANAMIDE",
            "peptide",
        ),
        "01F": ("ETHYL HYDROGEN CARBONATE", "non-polymer"),
        "01G": ("?", "non-polymer"),
        "01H": ("4-HYDRAZINYLBUTAN-1-AMINE", "non-polymer"),
        "01I": ("?", "non-polymer"),
        "01J": (
            "N-[(S)-HYDROXY(MORPHOLIN-4-YL)METHYL]-BETA-PHENYL-D-PHENYLALANINE",
            "peptide",
        ),
        "01K": ("LYSINE-COENZYME A DERIVATIVE", "non-polymer"),
        "01L": ("[(1R)-1-AMINO-4-METHOXYBUTYL]BORONIC ACID", "peptide"),
        "01M": ("(3S)-1-CARBAMIMIDOYLPIPERIDINE-3-CARBOXYLIC ACID", "peptide"),
        "01N": ("N-(CARBOXYMETHYL)-3-CYCLOHEXYL-D-ALANINE", "peptide"),
        "01O": ("4-(AMINOMETHYL)-1-METHYL-1H-PYRROLE-2-CARBOXIMIDAMIDE", "non-polymer"),
        "01P": (
            "N~2~-[3-METHOXY-4-(MORPHOLIN-4-YL)PHENYL]-N~4~-(QUINOLIN-3-YL)PYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "01Q": ("5-(AMINOMETHYL)-1-METHYL-1H-PYRROLE-2-CARBOXIMIDAMIDE", "non-polymer"),
        "01R": ("N-(PYRIDIN-4-YL)ETHANE-1,2-DIAMINE", "non-polymer"),
        "01S": (
            "N-[(2R)-2-(HYDROXYCARBAMOYL)-4-METHYLPENTANOYL]-L-ALANYLGLYCINAMIDE",
            "peptide",
        ),
        "01T": (
            "[5-(AMINOMETHYL)-6-(2,2-DIMETHYLPROPYL)-2-ETHYL-4-(4-METHYLPHENYL)PYRIDIN-3-YL]ACETIC ACID",
            "non-polymer",
        ),
        "01U": ("1,3-BENZOTHIAZOLE-2,6-DIAMINE", "non-polymer"),
        "01V": ("1-(2,5-DICHLOROPHENYL)METHANAMINE", "non-polymer"),
        "7T9": (
            "N-[(2S)-5-(6-FLUORO-3-PYRIDINYL)-2,3-DIHYDRO-1H-INDEN-2-YL]-2-PROPANESULFONAMIDE",
            "non-polymer",
        ),
        "01X": ("[2-(AMINOMETHYL)-4-CHLOROPHENOXY]ACETIC ACID", "peptide"),
        "01Y": ("(CYCLOPENTYLOXY)ACETIC ACID", "non-polymer"),
        "01Z": ("(CYCLOHEXYLOXY)ACETIC ACID", "non-polymer"),
        "FFD": (
            "(1R)-1,4-ANHYDRO-2-DEOXY-1-(3-FLUOROPHENYL)-5-O-PHOSPHONO-D-ERYTHRO-PENTITOL",
            "nucleic-acid",
        ),
        "FFC": ("2-DEOXY-2-FLUORO-B-D-CELLOBIOSIDE", "saccharide"),
        "FFB": (
            "N-(2,3,4,5,6-PENTAFLOURO-BENZYL)-4-SULFAMOYL-BENZAMIDE",
            "non-polymer",
        ),
        "FFA": (
            "(10ALPHA,13ALPHA,14BETA,17ALPHA)-17-HYDROXYANDROST-4-EN-3-ONE",
            "non-polymer",
        ),
        "LAS": (
            "N-ACETYL-S-({(2R,3S,4R)-3-HYDROXY-2-[(1S)-1-HYDROXY-2-METHYLPROPYL]-4-METHYL-5-OXOPYRROLIDIN-2-YL}CARBONYL)CYSTEINE",
            "non-polymer",
        ),
        "DTI": ("3,4-DIMETHYLTHIAZOLIUM ION", "non-polymer"),
        "DTH": ("D-THREONINE", "peptide"),
        "DTO": ("1-HYDROXYSULFANYL-4-MERCAPTO-BUTANE-2,3-DIOL", "non-polymer"),
        "AXD": (
            "2-[4-({2-[(2S,5R)-2-(AMINOMETHYL)-5-ETHYNYLPYRROLIDIN-1-YL]-2-OXOETHYL}AMINO)-4-METHYLPIPERIDIN-1-YL]ISONICOTINIC ACID",
            "non-polymer",
        ),
        "DTM": (
            "2,4-DIAMINO-6-[N-(3',4',5'-TRIMETHOXYBENZYL)-N-METHYLAMINO]PYRIDO[2,3-D]PYRIMIDINE",
            "non-polymer",
        ),
        "DTL": ("D-TREITOL", "non-polymer"),
        "DTC": ("BISHYDROXY[2H-1-BENZOPYRAN-2-ONE,1,2-BENZOPYRONE]", "non-polymer"),
        "DTB": ("6-(5-METHYL-2-OXO-IMIDAZOLIDIN-4-YL)-HEXANOIC ACID", "non-polymer"),
        "DTA": (
            "(2S,3S,4R,5R,2'S,3'S,4'R,5'R)-2,2'-[DITHIOBIS(METHYLENE)]BIS[5-(6-AMINO-9H-PURIN-9-YL)TETRAHYDROFURAN-3,4-DIOL]",
            "non-polymer",
        ),
        "DTG": ("DES-AMINO T-BUTYL GLYCINE", "non-polymer"),
        "DTF": (
            "4-[(10S,14S,18S)-18-(2-AMINO-2-OXOETHYL)-14-(1-NAPHTHYLMETHYL)-8,17,20-TRIOXO-7,16,19-TRIAZASPIRO[5.14]ICOS-11-EN-10-YL]BENZYLPHOSPHONIC ACID",
            "non-polymer",
        ),
        "DTE": ("7-CL-D-TRYPTOPHAN", "non-polymer"),
        "DTD": ("DITHIANE DIOL", "non-polymer"),
        "CB2": (
            "PHOSPHORIC ACID MONO-[5-(4-AMINO-5-BROMO-2-OXO-2H-PYRIMIDIN-1-YL)-3,4-DIHYDROXY-TETRAHYDRO-FURAN-2-YLMETHYL] ESTER",
            "nucleic-acid",
        ),
        "CB3": ("10-PROPARGYL-5,8-DIDEAZAFOLIC ACID", "non-polymer"),
        "DTY": ("D-TYROSINE", "peptide"),
        "CB1": ("5-(AZIRIDIN-1-YL)-2,4-DINITROBENZAMIDE", "non-polymer"),
        "CB6": (
            "(1R,4S,5R)-1,4,5-TRIHYDROXY-3-[(5-METHYL-1-BENZOTHIOPHEN-2-YL)METHOXY]CYCLOHEX-2-ENE-1-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "CB7": (
            "(1R,4S,5R)-3-(BENZO[B]THIOPHEN-5-YL)METHOXY-2-(BENZO[B]THIOPHEN-5-YL)METHYL-1,4,5-TRIHYDROXYCYCLOHEX-2-ENE-1-CARBOXYLATE",
            "non-polymer",
        ),
        "CB4": (
            "PINACOL[[2-AMINO-ALPHA-(1-CARBOXY-1-METHYLETHOXYIMINO)-4-THIAZOLEACETYL]AMINO]METHANEBORONATE",
            "non-polymer",
        ),
        "CB5": ("COBALT BIS(1,2-DICARBOLLIDE)", "non-polymer"),
        "DTR": ("D-TRYPTOPHAN", "peptide"),
        "CB8": (
            "(1R,4S,5R)-3-(BENZO[B]THIOPHEN-2-YL)METHOXY-1,4,5-TRIHYDROXY-2-(THIEN-2-YL)METHYLCYCLOHEX-2-EN-1-CARBOXYLATE",
            "non-polymer",
        ),
        "CB9": (
            "(2R,4S)-2-[(1R)-1-{[(2S)-2-CARBOXY-2-PHENYLACETYL]AMINO}-2-OXOETHYL]-5,5-DIMETHYL-1,3-THIAZOLIDINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "DTV": ("(2S,3S)-1,4-DIMERCAPTOBUTANE-2,3-DIOL", "non-polymer"),
        "DTU": ("(2R,3S)-1,4-DIMERCAPTOBUTANE-2,3-DIOL", "non-polymer"),
        "DTT": ("2,3-DIHYDROXY-1,4-DITHIOBUTANE", "non-polymer"),
        "MMM": (
            "(S,E)-3-HYDROXY-2-((3-HYDROXY-2-METHYL-5-(PHOSPHONOOXYMETHYL)PYRIDIN-4-YL)METHYLENEAMINO)-2-METHYLPROPANOIC ACID",
            "non-polymer",
        ),
        "D53": ("2-(BENZYLAMINO)ETHANESULFONIC ACID", "non-polymer"),
        "FAM": (
            "ALPHA-FLUORO-AMIDOCARBOXYMETHYLDETHIA COENZYME A COMPLEX",
            "non-polymer",
        ),
        "D54": ("4-DEOXY-L-THREO-HEX-5-ULOSURONIC ACID", "non-polymer"),
        "L34": (
            "4-(7-AMINO-9-HYDROXY-1-OXO-3,3A,4,5-TETRAHYDRO-2,5,6,8,9B-PENTAAZA-CYCLOPENTA[A]NAPHTHALEN-2-YL)-PHENYLCARBONYL-GLUTAMIC ACID",
            "non-polymer",
        ),
        "D55": (
            "4-(2-AMINOETHOXY)-N-(2,5-DIETHOXYPHENYL)-3,5-DIMETHYLBENZAMIDE",
            "non-polymer",
        ),
        "LL3": (
            "2-PHENYL-1-{4-[3-(THIOPHEN-2-YL)-1,2,4-OXADIAZOL-5-YL]PIPERIDIN-1-YL}ETHANONE",
            "non-polymer",
        ),
        "BYE": (
            "1-(2-METHYLPHENYL)-6-[(2S)-3,3,3-TRIFLUORO-2-METHYLPROPYL]-1,7-DIHYDRO-4H-PYRAZOLO[3,4-D]PYRIMIDIN-4-ONE",
            "non-polymer",
        ),
        "NAZ": (
            "N~3~-{3-[(7-METHYL-1,8-NAPHTHYRIDIN-2-YL)AMINO]-3-OXOPROPYL}-N~1~-[(7-OXO-7,8-DIHYDRO-1,8-NAPHTHYRIDIN-2-YL)METHYL]-BETA-ALANINAMIDE",
            "non-polymer",
        ),
        "2PO": ("PHOSPHONATE", "non-polymer"),
        "LL2": ("5-BENZYL-1,3-THIAZOL-2-AMINE", "non-polymer"),
        "LL9": (
            "2-[(3-CHLORO-4-METHOXYBENZYL)AMINO]ETHANESULFONIC ACID",
            "non-polymer",
        ),
        "SBF": (
            "N-[8-({[5-(DIMETHYLAMINO)NAPHTHALEN-1-YL]SULFONYL}AMINO)OCTYL]-2-[(3S,5S,7S)-TRICYCLO[3.3.1.1~3,7~]DEC-1-YL]ACETAMIDE",
            "non-polymer",
        ),
        "2PN": ("IMIDODIPHOSPHORIC ACID", "non-polymer"),
        "LL1": ("3-PYRIDIN-4-YL-1H-INDAZOLE", "non-polymer"),
        "SBG": ("O-[(S)-HYDROXY(METHYL)PHOSPHORYL]-L-SERINE", "non-polymer"),
        "IGU": ("2'-DEOXYISOGUANINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "IGP": ("INDOLE-3-GLYCEROL PHOSPHATE", "non-polymer"),
        "MMR": (
            "METHYL 2-(ACETYLAMINO)-3-O-[(1R)-1-CARBOXYETHYL]-2-DEOXY-BETA-D-GLUCOPYRANOSIDE",
            "non-polymer",
        ),
        "L31": (
            "3-[(1R)-2-ACETYL-6-METHYL-2,3,4,9-TETRAHYDRO-1H-BETA-CARBOLIN-1-YL]PHENOL",
            "non-polymer",
        ),
        "MME": ("N-METHYL METHIONINE", "peptide"),
        "IGC": (
            "N-[(1S,2R,3E)-1-({[ALPHA-D-GALACTOPYRANOSYL-(1->3)-BETA-D-GALACTOPYRANOSYL-(1->4)-BETA-D-GLUCOPYRANOSYL]OXY}METHYL)-2-HYDROXYHEPTADEC-3-EN-1-YL]OCTANAMIDE",
            "non-polymer",
        ),
        "LL5": (
            "(2S)-2-AMINO-3-METHYL-1-{4-[3-(THIOPHEN-2-YL)-1,2,4-OXADIAZOL-5-YL]PIPERIDIN-1-YL}BUTAN-1-ONE",
            "non-polymer",
        ),
        "IGA": ("6-AMINO-3,7-DIHYDRO-2H-PURIN-2-ONE", "non-polymer"),
        "MMD": (
            "5-(7-METHANESULFONYL-2-MORPHOLIN-4-YL-6,7-DIHYDRO-5H-PYRROLO[2,3-D]PYRIMIDIN-4-YL)-PYRIMIDIN-2-YLAMINE",
            "non-polymer",
        ),
        "IGN": (
            "{[(1R)-2-((2S)-2-{[(3-{[AMINO(IMINO)METHYL]AMINO}PROPYL)AMINO]CARBONYL}PIPERIDINYL)-1-(CYCLOHEXYLMETHYL)-2-OXOETHYL]AMINO}ACETIC ACID",
            "non-polymer",
        ),
        "IGL": ("ALPHA-AMINO-2-INDANACETIC ACID", "peptide"),
        "MMG": (
            "4-(5-PHENYL-1H-PYRROLO[2,3-B]PYRIDIN-3-YL)BENZOIC ACID",
            "non-polymer",
        ),
        "CBB": (
            "1-(3-CARBAMIMIDOYL-BENZYL)-1H-INDOLE-2-CARBOXYLIC ACID 3-CARBAMIMIDOYL-BENZYLESTER",
            "non-polymer",
        ),
        "CBC": (
            "N'-(5-CHLOROBENZOFURAN-2-CARBONYL)-2-(TRIFLUOROMETHYL)BENZENESULFONOHYDRAZIDE",
            "non-polymer",
        ),
        "CBA": (
            "N-PYRIDOXYL-2,3-DIHYDROXYASPARTIC ACID-5-MONOPHOSPHATE",
            "non-polymer",
        ),
        "CBF": ("C-(1-HYDROGYL-BETA-D-GLUCOPYRANOSYL) FORMAMIDE", "non-polymer"),
        "CBG": ("PROPANAL", "non-polymer"),
        "CBD": ("CIBACRON BLUE", "non-polymer"),
        "CBE": (
            "2-METHYL-N-PHENYL-5,6-DIHYDRO-1,4-OXATHIINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "CBJ": (
            "6-[(5-BROMO-2,3-DIOXO-2,3-DIHYDRO-1H-INDOL-1-YL)METHYL]-2,3-DIHYDRO-1,4-BENZODIOXINE-5-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "CBK": ("4-O-BETA-D-GLUCOPYRANOSYL-ALPHA-D-GLUCOPYRANOSE", "saccharide"),
        "CBH": ("S-(D-CARBOXYBUTYL)-L-HOMOCYSTEINE", "non-polymer"),
        "CBI": ("CELLOBIOSE", "saccharide"),
        "CBN": ("CLOROBIOCIN", "non-polymer"),
        "CBO": ("CARBENOXOLONE", "non-polymer"),
        "CBL": ("CHLORAMBUCIL", "non-polymer"),
        "CBM": ("CARBOXYMETHYL GROUP", "non-polymer"),
        "CBR": ("5-BROMO-2'-DEOXY-CYTIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "CBS": ("DI(N-ACETYL-D-GLUCOSAMINE)", "saccharide"),
        "CBP": (
            "2-{4-[4-(4-CHLORO-PHENOXY)-BENZENESULFONYL]-TETRAHYDRO-PYRAN-4-YL}-N-HYDROXY-ACETAMIDE",
            "non-polymer",
        ),
        "CBQ": (
            "[(5-CHLORO-PYRIDIN-2-YLAMINO)-PHOSPHONO-METHYL]-PHOSPHONIC ACID",
            "non-polymer",
        ),
        "CBV": ("?", "nucleic-acid"),
        "CBW": (
            "(3BETA,5BETA,14BETA)-3-HYDROXY-11-OXOOLEAN-12-EN-29-OIC ACID",
            "non-polymer",
        ),
        "CBT": ("N,N-BIS(4-CHLOROBENZYL)-1H-1,2,3,4-TETRAAZOL-5-AMINE", "non-polymer"),
        "CBU": ("(1R,2R,3S,4S,5S,6S)-CYCLOHEXANE-1,2,3,4,5,6-HEXOL", "non-polymer"),
        "CBZ": ("CARBOBENZOXY GROUP", "non-polymer"),
        "DT2": (
            "4-{[5-(CYCLOHEXYLAMINO)[1,2,4]TRIAZOLO[1,5-A]PYRIMIDIN-7-YL]AMINO}BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "CBX": ("CARBOXY GROUP", "non-polymer"),
        "6FP": ("2-AMINO-4-OXO-3,4-DIHYDROPTERIDINE-6-CARBALDEHYDE", "non-polymer"),
        "DT7": (
            "6,7-DIMETHOXY-3,4-DIHYDROISOQUINOLINE-2(1H)-SULFONAMIDE",
            "non-polymer",
        ),
        "DT6": ("2,4-BIS(ACETYLAMINO)-2,4-DIDEOXY-BETA-D-GLUCOPYRANOSE", "saccharide"),
        "DT5": (
            "4-{[5-(CYCLOHEXYLOXY)[1,2,4]TRIAZOLO[1,5-A]PYRIMIDIN-7-YL]AMINO}BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "DT4": (
            "4-({5-[(4-AMINOCYCLOHEXYL)AMINO][1,2,4]TRIAZOLO[1,5-A]PYRIMIDIN-7-YL}AMINO)BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "BZU": (
            "(4R)-2-(2-ETHOXYETHYL)-4-(ETHYLAMINO)-3,4-DIHYDRO-2H-THIENO[3,2-E][1,2]THIAZINE-6-SULFONAMIDE 1,1-DIOXIDE",
            "non-polymer",
        ),
        "BZT": (
            "3-(3-BROMO-4-PYRROLIDIN-1-YLMETHYL-BENZYL)-2-[4-PYRROLIDIN-1-YL-ETHOXY)-PHENYL]-BENZO[B]THIOPHEN-6-OL",
            "non-polymer",
        ),
        "BZS": ("L-BENZYLSUCCINIC ACID", "non-polymer"),
        "BZR": ("1S,2R,3S,4R-TETRAHYDRO-BENZO[A]ANTHRACENE-2,3,4-TRIOL", "non-polymer"),
        "BZQ": ("DIPHENYLMETHANONE", "non-polymer"),
        "BZP": ("N1-CARBOXYPIPERAZINE", "non-polymer"),
        "NT1": ("HEPARIN HEPTASACCHARIDE", "non-polymer"),
        "048": (
            "N-{2-FLUORO-5-[(5-OXO-6,7,8,9-TETRAHYDRO-5H-BENZO[7]ANNULEN-2-YL)AMINO]PHENYL}BENZAMIDE",
            "non-polymer",
        ),
        "BZZ": ("BENZYLHYDRAZINE", "non-polymer"),
        "BZX": ("1,3-BENZODIOXOL-5-OL", "non-polymer"),
        "BZG": (
            "6-(BENZYLOXY)-9-(2-DEOXY-5-O-PHOSPHONO-BETA-D-ERYTHRO-PENTOFURANOSYL)-9H-PURIN-2-AMINE",
            "nucleic-acid",
        ),
        "BZF": ("BENZOFURAN", "non-polymer"),
        "BZD": ("?", "non-polymer"),
        "BZC": ("?", "non-polymer"),
        "BZB": ("BENZO[B]THIOPHENE-2-BORONIC ACID", "non-polymer"),
        "BZA": ("1R,2S,3R,4S-TETRAHYDRO-BENZO[A]ANTHRACENE-2,3,4-TRIOL", "non-polymer"),
        "AL9": ("N-[(4-METHOXYPHENYL)METHYL]2,5-THIOPHENEDESULFONAMIDE", "non-polymer"),
        "BZO": ("CARBOBENZOXY GROUP", "non-polymer"),
        "BZN": (
            "N-({(3R,4R)-4-[(BENZYLOXY)METHYL]PYRROLIDIN-3-YL}METHYL)-N-(2-METHYLPROPYL)BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "BZM": ("BENZOIC ACID PHENYLMETHYLESTER", "non-polymer"),
        "AL5": (
            "THIOPHENE-2,5-DISULFONIC ACID 2-AMIDE-5-(4-METHYL-BENZYLAMIDE)",
            "non-polymer",
        ),
        "AL2": (
            "3,4-DIHYDRO-4-HYDROXY-2-(4-METHOXYPHENYL)-2H-THIENO[3,2-E]-1,2-THIAZINE-6-SULFONAMIDE-1,1-DIOXIDE",
            "non-polymer",
        ),
        "AL3": (
            "3,4-DIHYDRO-4-HYDROXY-2-(2-THIENYMETHYL)-2H-THIENO[3,2-E]-1,2-THIAZINE-6-SULFONAMIDE-1,1-DIOXIDE",
            "non-polymer",
        ),
        "BZI": ("BENZIMIDAZOLE", "non-polymer"),
        "BZH": (
            "5-[(2-AMINO-1H-BENZIMIDAZOL-6-YL)AMINO]-5-OXOPENTANOIC ACID",
            "non-polymer",
        ),
        "MV1": (
            "(2S)-2,4-DIHYDROXY-3,3-DIMETHYL-N-[3-(NONYLAMINO)-3-OXOPROPYL]BUTANAMIDE",
            "non-polymer",
        ),
        "MV2": (
            "(2S)-2,4-DIHYDROXY-N-(3-HYDROXYPROPYL)-3,3-DIMETHYLBUTANAMIDE",
            "non-polymer",
        ),
        "RO3": (
            "N-{[(2Z)-5-BROMO-1,3-THIAZOL-2(3H)-YLIDENE]CARBAMOYL}-4-METHYLBENZENESULFONAMIDE",
            "non-polymer",
        ),
        "MV7": ("unknown", "non-polymer"),
        "RHE": ("RHENIUM (IV) HEXACHLORIDE", "non-polymer"),
        "043": ("4-(3-BROMOPHENYL)-2,4-DIOXOBUTANOIC ACID", "non-polymer"),
        "042": (
            "N-{[(2S,3S)-3-(ETHOXYCARBONYL)OXIRAN-2-YL]CARBONYL}-L-THREONYL-L-ISOLEUCINE",
            "non-polymer",
        ),
        "PZN": (
            "3-(2-{5-TERT-BUTYL-3-[(4-METHYL-FURAZAN-3-YLMETHYL)-AMINO]-2-OXO-2H-PYRAZIN-1-YL}-BUTYRYLAMINO)-5-(HEXYL-METHYL-AMINO)-4-OXO-PENTANOIC ACID ANION",
            "non-polymer",
        ),
        "RO6": (
            "3-(2-CHLOROPHENYL)-7-[(TRANS-4-HYDROXYCYCLOHEXYL)AMINO]-3,4-DIHYDROPYRIMIDO[4,5-D]PYRIMIDIN-2(1H)-ONE",
            "non-polymer",
        ),
        "PZI": ("2-(2-OXOPYRROLIDIN-1-YL)ACETAMIDE", "non-polymer"),
        "IN4": (
            "+/-METHYL 4-(AMINOIMINOMETHYL)-BETA-[3- INH (AMINOIMINO)PHENYL]BENZENE PENTANOATE",
            "non-polymer",
        ),
        "040": ("4-(2-METHYLPHENYL)-2,4-DIOXOBUTANOIC ACID", "non-polymer"),
        "PSR": (
            "THIOBUTYRIC ACID S-{2-[3-(2-HYDROXY-3,3-DIMETHYL-4-PHOSPHONOOXY-BUTYRYLAMINO)-PROPIONYLAMINO]-ETHYL} ESTER",
            "non-polymer",
        ),
        "IWR": (
            "4-[(3AR,4S,7R,7AS)-1,3-DIOXO-1,3,3A,4,7,7A-HEXAHYDRO-2H-4,7-METHANOISOINDOL-2-YL]-N-(QUINOLIN-8-YL)BENZAMIDE",
            "non-polymer",
        ),
        "HB1": ("RUTHENIUM-PYRIDOCARBAZOLE-1", "non-polymer"),
        "MVA": ("N-METHYLVALINE", "peptide"),
        "MVB": (
            "(1S,7S,8S,8AR)-1,2,3,7,8,8A-HEXAHYDRO-7-METHYL-8-[2-[(2R,4R)-TETRAHYDRO-4-HY DROXY-6-OXO-2H-PYRAN-2-YL]ETHYL]-1-NAPHTHALENOL",
            "non-polymer",
        ),
        "MVC": ("MONOVACCENIN", "non-polymer"),
        "MVD": ("2ALPHA-METHYL-1ALPHA,25-DIHYDROXY-VITAMIN D3", "non-polymer"),
        "MVF": (
            "(2S)-1-[(2R)-2-[(2-AZANYL-2-OXIDANYLIDENE-ETHYL)AMINO]-2-CYCLOHEXYL-ETHANOYL]-N-[(4-CARBAMIMIDOYLPHENYL)METHYL]AZETIDINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "MVI": ("MYCINAMICIN VI", "non-polymer"),
        "MVL": (
            "(5R,6R,7S,8R)-6,7,8-TRIHYDROXY-5-(HYDROXYMETHYL)-5,6,7,8-TETRAHYDRO-1H-IMIDAZO[1,2-A]PYRIDIN-4-IUM",
            "non-polymer",
        ),
        "ALN": ("NAPHTHALEN-2-YL-3-ALANINE", "peptide"),
        "ALO": ("ALLO-THREONINE", "peptide"),
        "BZ5": (
            "(1R,2R,4S,5R)-2-(1-BENZOTHIOPHEN-5-YLMETHYL)-1,4,5-TRIS(OXIDANYL)-3-OXIDANYLIDENE-CYCLOHEXANE-1-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "ALM": ("1-METHYL-ALANINAL", "peptide"),
        "ALJ": ("CYCLO-(L-ARGININE-L-PROLINE) INHIBITOR", "non-polymer"),
        "BZ2": ("1-BENZOFURAN-2-CARBOXYLIC ACID", "non-polymer"),
        "BZ1": (
            "(+)-4-ETHYLAMINO-3,4-DIHYDRO-2-(METHOXY)PROPYL-2H-THIENO[3,2-E]-1,2-THIAZINE-6-SULFONAMIDE-1,1-DIOXIDE",
            "non-polymer",
        ),
        "ALI": ("METHYL N-ACETYL ALLOSAMINE", "non-polymer"),
        "ALF": ("TETRAFLUOROALUMINATE ION", "non-polymer"),
        "ALG": ("4-CARBAMIMIDAMIDOBUTANOIC ACID", "peptide"),
        "ALD": (
            "N-[(BENZYLOXY)CARBONYL]-L-LEUCYL-N-[(2S)-1-HYDROXY-4-METHYLPENTAN-2-YL]-L-LEUCINAMIDE",
            "non-polymer",
        ),
        "ALE": ("L-EPINEPHRINE", "non-polymer"),
        "ALB": ("DELTA-2-ALBOMYCIN A1", "non-polymer"),
        "ALC": ("2-AMINO-3-CYCLOHEXYL-PROPIONIC ACID", "peptide"),
        "BZ9": (
            "N-[3-METHYL-4-({3-[2-(METHYLAMINO)PYRIMIDIN-4-YL]PYRIDIN-2-YL}OXY)NAPHTHALEN-1-YL]-1H-BENZIMIDAZOL-2-AMINE",
            "non-polymer",
        ),
        "ALA": ("ALANINE", "peptide"),
        "NTI": ("2-[(5-NITRO-1,3-THIAZOL-2-YL)CARBAMOYL]PHENYL ACETATE", "non-polymer"),
        "NTH": (
            "SUCCINIC ACID MONO-(13-METHYL-3-OXO-2,3,6,7,8,9,10,11,12,13,14,15,16,17-TETRADECAHYDRO-1H-CYCLOPENTA[A]PHENANTHREN-17-YL) ESTER",
            "non-polymer",
        ),
        "NTJ": ("R-ETHYL N,N-DIMETHYLPHOSPHONAMIDATE", "non-polymer"),
        "ALZ": ("?", "peptide"),
        "FV1": (
            "DIHYDROXY{[(2R,3S)-3-METHYLOXIRAN-2-YL]PHOSPHONATO-KAPPAO}OXOVANADIUM",
            "non-polymer",
        ),
        "NTO": ("TRISULFOAMINO HEPARIN PENTASACCHARIDE", "saccharide"),
        "ALY": ("N(6)-ACETYLLYSINE", "peptide"),
        "ALV": ("(2S)-2-AMINOPROPANE-1,1-DIOL", "peptide"),
        "ALW": (
            "METHYL 5-(PROPYLSULFINYL)-1H-BENZIMIDAZOL-2-YLCARBAMATE",
            "non-polymer",
        ),
        "ALT": ("THIOALANINE", "peptide"),
        "ALU": ("METHYL HYDROGEN (S)-ACETYLPHOSPHONATE", "non-polymer"),
        "ALR": ("ALRESTATIN", "non-polymer"),
        "ALS": ("2-AMINO-3-OXO-4-SULFO-BUTYRIC ACID", "peptide"),
        "ALP": (
            "2-(1-CARBOXY-2-HYDROXY-ETHYL)-5,5-DIMETHYL-THIAZOLIDINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "ALQ": ("2-METHYL-PROPIONIC ACID", "non-polymer"),
        "UQ": ("COENZYME Q10, (2Z,6E,10Z,14E,18E,22E,26Z)-ISOMER", "non-polymer"),
        "IN7": (
            "[4-(4-PHENYL-PIPERIDIN-1-YL)-BENZENESULFONYLAMINO]-ACETIC ACID",
            "non-polymer",
        ),
        "P01": (
            "2-({6-[(3-CHLOROPHENYL)AMINO]-9-ISOPROPYL-9H-PURIN-2-YL}AMINO)-3-METHYLBUTAN-1-OL",
            "non-polymer",
        ),
        "IHJ": (
            "9-CYCLOPENTYL-6-{2-[3-(4-METHYL-PIPERAZIN-1-YL)-PROPOXY]-PHENYLAMINO}-9H-PURINE-2-CARBONITRILE",
            "non-polymer",
        ),
        "KOS": (
            "17-DESMETHOXY-17-N,N-DIMETHYLAMINOETHYLAMINO-GELDANAMYCIN",
            "non-polymer",
        ),
        "SD2": ("N-(SULFANYLACETYL)TYROSYLPROLYLMETHIONINAMIDE", "peptide"),
        "B70": (
            "TRIPOTASSIUM (1R)-4-BIPHENYL-4-YL-1-PHOSPHONATOBUTANE-1-SULFONATE",
            "non-polymer",
        ),
        "B71": ("3-(DECYLOXY)-1-(2,2-DIPHOSPHONOETHYL)PYRIDINIUM", "non-polymer"),
        "B72": (
            "{4-[4-HYDROXY-3-(1-METHYLETHYL)BENZYL]-3,5-DIMETHYLPHENOXY}ACETIC ACID",
            "non-polymer",
        ),
        "B73": ("3-(2,2-DIPHOSPHONOETHYL)-1-DODECYL-1H-IMIDAZOL-3-IUM", "non-polymer"),
        "B75": ("unknown", "non-polymer"),
        "B76": ("?", "non-polymer"),
        "B77": ("2-[(2-METHOXY-5-METHYLPHENOXY)METHYL]PYRIDINE", "non-polymer"),
        "B78": ("4-[(5-METHOXY-2-METHYLPHENOXY)METHYL]PYRIDINE", "non-polymer"),
        "KT5": (
            "N-{4-4-(2,4-DIAMINO-6-OXO-1,6-DIHYDRO-PYRIMIDIN-5-YL)-1-(2,2,2-TRIFLUORO-1,1-DIHYDROXY-ETHYL)-BUT-2-YL-BENZOYL}-GAMMA-GLUTAMYL-GAMMA-GLUTAMYL-GAMMA-GLUTAMYL-GAMMA-GLUTAMYL-GLUTAMIC ACID",
            "non-polymer",
        ),
        "X1E": (
            "(2S,5R,6R)-6-AMINO-3,3-DIMETHYL-7-OXO-4-THIA-1-AZABICYCLO[3.2.0]HEPTANE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "KOT": ("1-BETA-D-GLUCOPYRANOSYL-4-PHENYL-1H-1,2,3-TRIAZOLE", "non-polymer"),
        "O73": (
            "5-[(4-AMINOBUTYL)AMINO]-1,2-DIMETHYL-3-[(4-NITROPHENOXY)METHYL]-1H-INDOLE-4,7-DIONE",
            "non-polymer",
        ),
        "IN0": (
            "N-{2-[TRANS-7-CHLORO-1-(2,2-DIMETHYL-PROPYL) -5-NAPHTHALEN-1-YL-2-OXO-1,2,3,5-TETRAHYDRO-BENZO[E] [1,4]OXAZEPIN-3-YL]-ACETYL}-ASPARTIC ACID",
            "non-polymer",
        ),
        "X1P": ("1-O-PHOSPHONO-ALPHA-D-XYLOPYRANOSE", "non-polymer"),
        "X1R": ("3-(PROPYLSULFONYL)PROPANOIC ACID", "non-polymer"),
        "X1S": ("3-SULFOPROPANOIC ACID", "non-polymer"),
        "KTU": (
            "PROP-2-EN-1-YL 3,5-DIDEOXY-ALPHA-D-THREO-OCT-5-EN-2-ULOPYRANOSIDONIC ACID",
            "saccharide",
        ),
        "11R": ("RUTHENIUM WIRE, 11 CARBON LINKER", "non-polymer"),
        "KTP": (
            "(2-NAPHTHALEN-2-YL-1-NAPHTHALEN-1-YL-2-OXO-ETHYL)-PHOSPHONIC ACID",
            "non-polymer",
        ),
        "KTR": (
            "(1R)-5-BENZOYL-2,3-DIHYDRO-1H-PYRROLIZINE-1-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "IN1": (
            "PROPIONIC ACID 2,2-DIHYDROXY-3-(1-GLUCOPYRANOSYLOXY)-METHYLPROPYL ESTER",
            "saccharide",
        ),
        "X19": (
            "(5S)-N-METHYL-5-(2-METHYLBUTAN-2-YL)-4,5,6,7-TETRAHYDRO-2H-INDAZOLE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "KTN": (
            "CIS-1-ACETYL-4-(4-((2-(2,4-DICHLOROPHENYL)-2-(1H-IMIDAZOL-1-YLMETHYL)-1,3-DIOXOLAN-4-YL)METHOXY)PHENYL)PIPERAZINE",
            "non-polymer",
        ),
        "B7C": (
            "3-(2-DEOXY-5-O-PHOSPHONO-BETA-D-ERYTHRO-PENTOFURANOSYL)-7,8-DIHYDROPYRIDO[2,3-D]PYRIMIDIN-2(3H)-ONE",
            "nucleic-acid",
        ),
        "B7D": (
            "(3R)-6-CHLORO-3-(DICHLOROMETHYL)-3,4-DIHYDRO-2H-1,2,4-BENZOTHIADIAZINE-7-SULFONAMIDE 1,1-DIOXIDE",
            "non-polymer",
        ),
        "PZ3": (
            "4-({(3R)-7-(5-FORMYLFURAN-2-YL)-4-[(4-METHOXYPHENYL)SULFONYL]-1-[(1-METHYL-1H-IMIDAZOL-5-YL)METHYL]-2,3,4,5-TETRAHYDRO-1H-1,4-BENZODIAZEPIN-3-YL}METHYL)PHENYL BENZYLCARBAMATE",
            "non-polymer",
        ),
        "B7G": ("HEPTYL-BETA-D-GLUCOPYRANOSIDE", "non-polymer"),
        "KTD": (
            "1-(1,2,3-BENZOTHIADIAZOL-6-YL)-3-(2-OXO-2-PHENYLETHYL)UREA",
            "non-polymer",
        ),
        "X11": ("2-(4,6-DIAMINO-1,3,5-TRIAZIN-2-YL)PHENOL", "non-polymer"),
        "X12": ("?", "non-polymer"),
        "PZ2": (
            "3-[2-BROMO-4-(1H-PYRAZOLO[3,4-C]PYRIDAZIN-3-YLMETHYL)PHENOXY]-5-METHYLBENZONITRILE",
            "non-polymer",
        ),
        "X14": (
            "(5R)-5-(2-METHYLBUTAN-2-YL)-N-(4-SULFAMOYLBENZYL)-4,5,6,7-TETRAHYDRO-2H-INDAZOLE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "B7M": ("(10E,12Z)-HEXADECA-10,12-DIENAL", "non-polymer"),
        "B7N": (
            "(1R)-2-{[(S)-HYDROXY{[(1S,2R,3R,4S,5S,6R)-2,3,4,5,6-PENTAHYDROXYCYCLOHEXYL]OXY}PHOSPHORYL]OXY}-1-[(OCTADECANOYLOXY)METHYL]ETHYL (9Z)-OCTADEC-9-ENOATE",
            "non-polymer",
        ),
        "S38": ("4-CYANO-N-[(1S,2R)-2-PHENYLCYCLOPROPYL]BENZAMIDE", "non-polymer"),
        "IN2": ("N,N-DIMETHYLCARBAMOYL-ALPHA-AZALYSINE", "non-polymer"),
        "PEA": ("2-PHENYLETHYLAMINE", "non-polymer"),
        "X3A": (
            "[4-AMINO-2-(PHENYLAMINO)-1,3-THIAZOL-5-YL](PYRIDIN-3-YL)METHANONE",
            "non-polymer",
        ),
        "IW2": ("2,6-DIBROMO-4-[2-(2,6-DICHLOROPHENYL)ETHYL]PHENOL", "non-polymer"),
        "D24": (
            "2,5-BIS-[4-[CYCLOPENTA-1,3-DIEN-5-YLAMINO-1-AMINOMETHYL]-PHEN-1-YL]FURAN",
            "non-polymer",
        ),
        "VPX": (
            "ETHYL [(1R,3AR,4AR,6R,8AR,9S,9AS)-9-{(E)-2-[5-(3-FLUOROPHENYL)PYRIDIN-2-YL]ETHENYL}-1-METHYL-3-OXODODECAHYDRONAPHTHO[2,3-C]FURAN-6-YL]CARBAMATE",
            "non-polymer",
        ),
        "133": (
            "6-FLUORO-2-(2-HYDROXY-3-ISOBUTOXY-PHENYL)-1H-BENZOIMIDAZOLE-5-CARBOXAMIDINE",
            "non-polymer",
        ),
        "132": (
            "6-CHLORO-2-(2-HYDROXY-BIPHENYL-3-YL)-1H-INDOLE-5-CARBOXAMIDINE",
            "non-polymer",
        ),
        "130": (
            "2-(2-HYDROXY-BIPHENYL)-1H-BENZOIMIDAZOLE-5-CARBOXAMIDINE",
            "non-polymer",
        ),
        "137": (
            "1-(O-CARBOXY-PHENYLAMINO)-1-DEOXY-D-RIBULOSE-5-PHOSPHATE",
            "non-polymer",
        ),
        "136": (
            "N-(4-CARBAMIMIDOYL-3-CHORO-PHENYL)-2-HYDROXY-3-IODO-5-METHYL-BENZAMIDE",
            "non-polymer",
        ),
        "135": ("N-(4-CARBAMIMIDOYL-PHENYL)-2-HYDROXY-BENZAMIDE", "non-polymer"),
        "134": (
            "6-FLUORO-2-[2-HYDROXY-3-(2-METHYL-CYCLOHEXYLOXY)-PHENYL]-1H-INDOLE-5-CARBOXAMIDINE",
            "non-polymer",
        ),
        "139": ("?", "non-polymer"),
        "138": ("?", "non-polymer"),
        "UGJ": (
            "3,6-DIAMINO-4-(2-CHLOROPHENYL)THIENO[2,3-B]PYRIDINE-2,5-DICARBONITRILE",
            "non-polymer",
        ),
        "SSR": (
            "2-(3-FLUORO-2-METHYLBENZYL)-4-METHYL-1-PHENYL-3-(PIPERAZIN-1-YLCARBONYL)-1H-INDOL-5-OL",
            "non-polymer",
        ),
        "YM3": (
            "ETHYL 4-(2-(AMINOMETHYL)MORPHOLINO)-1H-PYRAZOLO[3,4-B]PYRIDINE-5-CARBOXYLATE",
            "non-polymer",
        ),
        "D27": (
            "2-{[(2E)-3-(3,4-DIMETHOXYPHENYL)PROP-2-ENOYL]AMINO}BENZOIC ACID",
            "non-polymer",
        ),
        "YM4": (
            "6-BROMO-7-{4-[(5-METHYLISOXAZOL-3-YL)METHYL]PIPERAZIN-1-YL}-2-[4-(4-METHYLPIPERAZIN-1-YL)PHENYL]-1H-IMIDAZO[4,5-B]PYRIDINE",
            "non-polymer",
        ),
        "YM7": (
            "5-({6-[(PIPERIDIN-4-YLMETHYL)AMINO]PYRIMIDIN-4-YL}AMINO)PYRAZINE-2-CARBONITRILE",
            "non-polymer",
        ),
        "YM6": (
            "1-[(2R)-4-(9H-PYRIDO[4',3':4,5]PYRROLO[2,3-D]PYRIMIDIN-4-YL)MORPHOLIN-2-YL]METHANAMINE",
            "non-polymer",
        ),
        "YM8": (
            "(R)-5-(8-CHLOROISOQUINOLIN-3-YLAMINO)-3-(1-(DIMETHYLAMINO)PROPAN-2-YLOXY)PYRAZINE-2-CARBONITRILE",
            "non-polymer",
        ),
        "LL4": (
            "4-METHYL-1-{4-[3-(THIOPHEN-2-YL)-1,2,4-OXADIAZOL-5-YL]PIPERIDIN-1-YL}PENTAN-1-ONE",
            "non-polymer",
        ),
        "YMP": ("O-(ADENOSINE-5'-O-YL)-N-(L-TYROSYL)PHOSPHORAMIDATE", "non-polymer"),
        "VZZ": ("11-[(MERCAPTOCARBONYL)OXY]UNDECANOIC ACID", "non-polymer"),
        "XI2": (
            "6-(2,4-DIFLUOROPHENOXY)-N-[(1R)-1-METHYL-2-(METHYLSULFONYL)ETHYL]-1H-PYRAZOLO[3,4-D]PYRIMIDIN-3-AMINE",
            "non-polymer",
        ),
        "AX3": ("N~2~-CYCLOPROPYL-1,3,5-TRIAZINE-2,4,6-TRIAMINE", "non-polymer"),
        "YMA": (
            "7,10,13-TRI(CARBOXYMETHYL)-5,15-DIOXO-4,7,10,13,16-PENTAAZA-1,19-DITHIANONADECANE",
            "non-polymer",
        ),
        "3A1": (
            "N~2~-{(2S)-3-[(R)-HYDROXY{(1R)-2-PHENYL-1-[(PHENYLACETYL)AMINO]ETHYL}PHOSPHORYL]-2-METHYLPROPANOYL}-L-LYSYL-D-SERINE",
            "peptide",
        ),
        "2P2": (
            "1-[(2R,5S)-2-TERT-BUTYL-5-({(2S)-2-[(1R)-1,2-DIHYDROXYETHYL]PYRROLIDIN-1-YL}CARBONYL)PYRROLIDIN-1-YL]-4-PHENYLBUTAN-1-ONE",
            "non-polymer",
        ),
        "MZP": (
            "4-CARBAMOYL-1-BETA-D-RIBOFURANOSYL-IMIDAZOLIUM-5-OLATE-5'-PHOSPHATE",
            "non-polymer",
        ),
        "YML": (
            "[(3R,4S,5S,7R)-4,8-DIHYDROXY-3,5,7-TRIMETHYL-2-OXOOCTYL]PHOSPHONIC ACID",
            "non-polymer",
        ),
        "AX7": ("1H-BENZIMIDAZOL-2-AMINE", "non-polymer"),
        "13C": (
            "(5E)-13-CHLORO-14,16-DIHYDROXY-3,4,7,8,9,10-HEXAHYDRO-2-BENZAZACYCLOTETRADECINE-1,11(2H,12H)-DIONE",
            "non-polymer",
        ),
        "13A": (
            "7-(3,4-DIHYDROXY-5R-HYDROXYMETHYLTETRAHYDROFURAN-2-YL)-2,2-DIOXO-1,2R,3R,7-TETRAHYDRO-2L6-IMIDAZO[4,5-C][1,2,6]THIADIAZIN-4S-ONE",
            "non-polymer",
        ),
        "2P7": (
            "[3,4-BIS(FLUORANYL)-2-[(2-FLUORANYL-4-IODANYL-PHENYL)AMINO]PHENYL]-(3-OXIDANYLAZETIDIN-1-YL)METHANONE",
            "non-polymer",
        ),
        "13E": ("N-[(2S)-2-AMINO-3-PHENYLPROPYL]-L-METHIONINE", "non-polymer"),
        "13D": ("1,3-DIAMINOPROPANE", "non-polymer"),
        "13K": (
            "N~6~-(3,4-DIMETHOXYPHENYL)-2-(MORPHOLIN-4-YL)-4,5'-BIPYRIMIDINE-2',6-DIAMINE",
            "non-polymer",
        ),
        "4OS": ("4-METHOXYBENZENESULFONIC ACID", "non-polymer"),
        "13I": (
            "(E)-ETHYL 13-CHLORO-14,16-DIHYDROXY-1,11-DIOXO-1,2,3,4,7,8,9,10,11,12-DECAHYDROBENZO[C][1]AZACYCLOTETRADECINE-10-CARBOXYLATE",
            "non-polymer",
        ),
        "2P6": (
            "(2S)-1-[1-(4-PHENYLBUTANOYL)-L-PROLYL]PYRROLIDINE-2-CARBONITRILE",
            "non-polymer",
        ),
        "13N": (
            "(5E,10R)-N-BENZYL-13-CHLORO-14,16-DIHYDROXY-1,11-DIOXO-1,2,3,4,7,8,9,10,11,12-DECAHYDRO-2-BENZAZACYCLOTETRADECINE-10-CARBOXAMIDE",
            "non-polymer",
        ),
        "13M": (
            "(2S)-2-(4-METHOXY-3-{[(PYREN-1-YLCARBONYL)AMINO]METHYL}BENZYL)BUTANOIC ACID",
            "non-polymer",
        ),
        "13L": (
            "3-{4-AMINO-1-[(3R)-1-PROPANOYLPIPERIDIN-3-YL]-1H-PYRAZOLO[3,4-D]PYRIMIDIN-3-YL}-N-[4-(PROPAN-2-YL)PHENYL]BENZAMIDE",
            "non-polymer",
        ),
        "13S": ("13(S)-HYDROPEROXY-9(Z),11(E)-OCTADECADIENOIC ACID", "non-polymer"),
        "13R": ("13(R)-HYDROPEROXY-9(Z),11(E)-OCTADECADIENOIC ACID", "non-polymer"),
        "4OH": (
            "(14BETA,15ALPHA,16ALPHA,17ALPHA)-ESTRA-1,3,5(10)-TRIENE-3,15,16,17-TETROL",
            "non-polymer",
        ),
        "13P": ("1,3-DIHYDROXYACETONEPHOSPHATE", "non-polymer"),
        "13W": (
            "3-{5-[(2E,4S)-2-IMINO-1,4-DIMETHYL-6-OXOHEXAHYDROPYRIMIDIN-4-YL]THIOPHEN-3-YL}BENZONITRILE",
            "non-polymer",
        ),
        "13U": (
            "N-CYCLOOCTYLGLYCYL-N-(4-CARBAMIMIDOYLBENZYL)-L-PROLINAMIDE",
            "non-polymer",
        ),
        "13T": ("13-DEOXYTEDANOLIDE", "non-polymer"),
        "4OC": ("4N,O2'-METHYLCYTIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "13Y": (
            "2-(3-CHLOROPHENOXY)-4-{(1R)-3-METHYL-1-[(3S)-3-(5-METHYL-2,4-DIOXO-3,4-DIHYDROPYRIMIDIN-1(2H)-YL)PIPERIDIN-1-YL]BUTYL}BENZOIC ACID",
            "non-polymer",
        ),
        "4OA": (
            "(3BETA,5BETA,14BETA,17ALPHA)-3-HYDROXYCHOLAN-24-OIC ACID",
            "non-polymer",
        ),
        "WI2": (
            "1-[3-AMINO-6-(3-METHOXYPHENYL)PYRAZIN-2-YL]PIPERIDINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "2AF": ("2-AMINOPHENOL", "non-polymer"),
        "2AG": ("(2S)-2-AMINOPENT-4-ENOIC ACID", "peptide"),
        "2AD": ("2'-AMINO-2'-DEOXYADENOSINE", "peptide"),
        "2AE": ("2-AMINO-BENZAMIDE", "non-polymer"),
        "2AB": ("4-(2-AMINOETHYL)BENZENESULFONIC ACID", "non-polymer"),
        "2AC": ("2-AMINO-P-CRESOL", "non-polymer"),
        "N1T": ("?", "non-polymer"),
        "2AA": (
            "5'-O-{[({[2-({[5-(DIMETHYLAMINO)NAPHTHALEN-1-YL]SULFONYL}AMINO)ETHYL]OXY}PHOSPHINATO)OXY]PHOSPHINATO}CYT",
            "non-polymer",
        ),
        "2AN": ("8-ANILINO-1-NAPHTHALENE SULFONATE", "non-polymer"),
        "2AO": ("(2S)-2-AMINOHEXAN-1-OL", "peptide"),
        "2AL": ("1-(2,5-DIOXO-2,5-DIHYDRO-1H-IMIDAZOL-4-YL)UREA", "non-polymer"),
        "2AM": ("ADENOSINE-2'-MONOPHOSPHATE", "non-polymer"),
        "2AH": ("?", "non-polymer"),
        "2AI": ("1H-IMIDAZOL-2-AMINE", "non-polymer"),
        "LJG": (
            "N,N'-(IMINODIETHANE-2,1-DIYL)BIS[4-AMINO-N-(2-METHYLPROPYL)BENZENESULFONAMIDE]",
            "non-polymer",
        ),
        "N1C": ("?", "non-polymer"),
        "2AT": ("?", "nucleic-acid"),
        "2AU": ("2'-AMINOURIDINE", "nucleic-acid"),
        "2AR": ("2'-DEOXYARISTEROMYCIN-5'-PHOSPHATE", "nucleic-acid"),
        "2AS": ("(2S,3S)-3-METHYL-ASPARTIC ACID", "peptide"),
        "2AP": ("2-AMINOPYRIDINE", "non-polymer"),
        "2AQ": ("QUINOLIN-2-AMINE", "non-polymer"),
        "N1H": ("?", "non-polymer"),
        "2AZ": (
            "N-(3-AMINOPROPYL)-N-[(R)-(3-BENZYL-5-CHLORO-4-OXO-3,4-DIHYDROPYRROLO[2,1-F][1,2,4]TRIAZIN-2-YL)(CYCLOPROPYL)METHYL]-4-METHYLBENZAMIDE",
            "non-polymer",
        ),
        "N1L": ("2-AMINO-2-DEOXY-BETA-D-GLUCOPYRANURONIC ACID", "saccharide"),
        "N1M": ("2-IODO-N-METHYLBENZAMIDE", "non-polymer"),
        "N7O": (
            "(S)-[(1S)-1-(5-CHLORO-1-BENZOTHIOPHEN-3-YL)-2-{[(E)-2-(3,4-DIFLUOROPHENYL)ETHENYL]AMINO}-2-OXOETHYL]METHYLPHOSPHINIC ACID",
            "non-polymer",
        ),
        "WO5": ("TUNGSTATE(VI) ION", "non-polymer"),
        "WO4": ("TUNGSTATE(VI)ION", "non-polymer"),
        "WO6": ("TRIOXIDO(OXO)TUNGSTEN", "non-polymer"),
        "WO3": ("TRI-TUNGSTEN(VI) OXIDE COMPLEX", "non-polymer"),
        "WO2": ("OCTADECATUNGSTENYL DIPHOSPHATE", "non-polymer"),
        "PC5": ("1,2-DI-N-PENTANOYL-SN-GLYCERO-3-DITHIOPHOSPHOCHOLINE", "non-polymer"),
        "RWJ": (
            "CYCLOPENTANECARBOXYLIC ACID [1-(BENZOTHIAZOLE-2-CARBONYL)-4-GUANIDINO-BUTYL]-AMIDE",
            "non-polymer",
        ),
        "PMT": (
            "PHOSPHORIC ACID MONO-[3-(3-{[5-(4-AMINO-2-OXO-2H-PYRIMIDIN-1-YL)-3,4- DIHYDROXY-TETRAHYDRO-FURAN-2- YLMETHOXY]-HYDROXY-PHOSPHORYLOXY}-3-OXO-PROPYLCARBAMOYL)-3-HYDROXY-2,2- DIMETHYL-PROPYL] ESTER",
            "nucleic-acid",
        ),
        "PMU": (
            "(3S,4S)-1-{3-[2-AMINO-6-(PROPAN-2-YL)PYRIMIDIN-4-YL]-4-HYDROXYPHENYL}PYRROLIDINE-3,4-DIOL",
            "non-polymer",
        ),
        "WOW": (
            "2-CHLORO-4-{(2-METHYLBENZYL)[(3S)-1-(METHYLSULFONYL)PYRROLIDIN-3-YL]AMINO}BENZONITRILE",
            "non-polymer",
        ),
        "O60": ("DECANE-1,10-DIYL DISULFAMATE", "non-polymer"),
        "SMN": ("(S)-MANDELIC ACID", "non-polymer"),
        "SSH": ("3-DEOXY-D-ARABINO-HEXONIC ACID", "non-polymer"),
        "PB4": (
            "N,N'-BIS(4-FLUORO-3-METHYLBENZYL)PYRIMIDINE-4,6-DICARBOXAMIDE",
            "non-polymer",
        ),
        "WOE": ("4-CHLORO-6-(2-METHOXYPHENYL)PYRIMIDIN-2-AMINE", "non-polymer"),
        "PMQ": ("BENZYLCARBAMIC ACID", "non-polymer"),
        "N12": (
            "N-[(1R)-2-[(1-{[({6-[AMINO(IMINO)METHYL]PYRIDIN-3-YL}METHYL)AMINO]CARBONYL}CYCLOPENTYL)AMINO]-1-(CYCLOHEXYLMETHYL)-2-OXOETHYL]GLYCINE",
            "non-polymer",
        ),
        "N10": ("O-[(HEXYLAMINO)CARBONYL]-L-SERINE", "peptide"),
        "N11": (
            "6-[(7-NITRO-2,1,3-BENZOXADIAZOL-4-YL)SULFANYL]HEXAN-1-OL",
            "non-polymer",
        ),
        "N16": (
            "(3Z,5S)-5-BENZYL-3-[1-(PHENYLAMINO)ETHYLIDENE]PYRROLIDINE-2,4-DIONE",
            "non-polymer",
        ),
        "O64": (
            "2-CHLORO-N-[(1S)-1-FORMYLPROPYL]-3-(TRIFLUOROMETHYL)BENZAMIDE",
            "non-polymer",
        ),
        "PR2": (
            "THIENO[3,2-B]PYRIDINE-2-SULFONIC ACID [2-OXO-1-(1H-PYRROLO[2,3-C]PYRIDIN-2-YLMETHYL)-PYRROLIDIN-3-YL]-AMIDE",
            "non-polymer",
        ),
        "PMS": ("PHENYLMETHANESULFONIC ACID", "non-polymer"),
        "2A6": ("2-ANILINO-6-CYCLOHEXYLMETHOXYPURINE", "non-polymer"),
        "2A7": ("5-METHOXY-BENZENE-1,3-DIOL", "non-polymer"),
        "2A5": ("?", "non-polymer"),
        "2A2": (
            "4-[4-(3-CHLOROPHENYL)PIPERAZIN-1-YL]-4-OXOBUTANE-1-THIOL",
            "non-polymer",
        ),
        "2A3": ("(2R)-2-AMINOPROPAN-1-OL", "non-polymer"),
        "2A0": ("N-[(2S)-2-AMINOHEXYL]-L-NORLEUCINE", "peptide"),
        "2A1": ("(2S)-2-AMINOPROPAN-1-OL", "non-polymer"),
        "UQ6": (
            "5-(3,7,11,15,19,23-HEXAMETHYL-TETRACOSA-2,6,10,14,18,22-HEXAENYL)-2,3-DIMETHOXY-6-METHYL-BENZENE-1,4-DIOL",
            "non-polymer",
        ),
        "NNH": ("NOR-N-OMEGA-HYDROXY-L-ARGININE", "peptide"),
        "2A8": ("2-AMINO-PHENYLAMINO-DIBENZOSUBERONE", "non-polymer"),
        "2A9": ("5-(HYDROXYMETHYL)-2-METHOXYPHENOL", "non-polymer"),
        "0KR": (
            "(2E,5R)-5-CYCLOPROPYL-2-IMINO-3-METHYL-5-{3-[5-(PROP-1-YN-1-YL)PYRIDIN-3-YL]PHENYL}IMIDAZOLIDIN-4-ONE",
            "non-polymer",
        ),
        "0KS": (
            "N-[4-(1,1,1,3,3,3-HEXAFLUORO-2-HYDROXYPROPAN-2-YL)PHENYL]-N-METHYLBENZENESULFONAMIDE",
            "non-polymer",
        ),
        "0KP": (
            "(2E,5R)-2-IMINO-3-METHYL-5-PHENYL-5-[3-(PYRIDIN-3-YL)PHENYL]IMIDAZOLIDIN-4-ONE",
            "non-polymer",
        ),
        "0KQ": (
            "(2E,5R)-5-[3-(5-CHLOROPYRIDIN-3-YL)PHENYL]-5-CYCLOPROPYL-2-IMINO-3-METHYLIMIDAZOLIDIN-4-ONE",
            "non-polymer",
        ),
        "0KV": (
            "2-{(3S)-3-[(BENZYLSULFONYL)AMINO]-2-OXOPIPERIDIN-1-YL}-N-{(2S)-1-[(3S)-1-CARBAMIMIDOYLPIPERIDIN-3-YL]-3-OXOPROPAN-2-YL}ACETAMIDE",
            "peptide",
        ),
        "0KW": (
            "N-(4-CHLOROPHENYL)-N'-{(S)-[5-(2-HYDROXYETHYL)-4-METHYL-1,3-THIAZOL-2-YL][(2S)-PIPERIDIN-2-YL]METHYL}ETHANEDIAMIDE",
            "non-polymer",
        ),
        "6NA": ("HEXANOIC ACID", "non-polymer"),
        "0KU": (
            "N-(4-CHLOROPHENYL)-N'-{(S)-[5-(HYDROXYMETHYL)-4-METHYL-1,3-THIAZOL-2-YL][(2S)-PIPERIDIN-2-YL]METHYL}ETHANEDIAMIDE",
            "non-polymer",
        ),
        "0KZ": (
            "[(1R,2S,4S,5S)-2-HYDROXY-4-(5-METHYL-2,4-DIOXO-3,4-DIHYDROPYRIMIDIN-1(2H)-YL)BICYCLO[3.1.0]HEX-1-YL]METHYL DIHYDROGEN PHOSPHATE",
            "nucleic-acid",
        ),
        "PMX": (
            "3-CHLORO-N-[(2E)-4-METHOXY-4-OXOBUT-2-ENOYL]-L-TYROSINE",
            "non-polymer",
        ),
        "0KX": (
            "2'-DEOXY-5'-O-[(R)-HYDROXY{[(R)-HYDROXY(PHOSPHONOOXY)PHOSPHORYL]AMINO}PHOSPHORYL]CYTIDINE",
            "non-polymer",
        ),
        "0KY": (
            "3-[(2R)-2-ETHOXYPROPYL]-2-THIOXO-1,2,3,9-TETRAHYDRO-6H-PURIN-6-ONE",
            "non-polymer",
        ),
        "6NI": ("6-NITROINDAZOLE", "non-polymer"),
        "6NH": (
            "2-(4-HYDROXY-5-PHENYL-1H-PYRAZOL-3-YL)-1H-BENZOIMIDAZOLE-5-CARBOXAMIDINE",
            "non-polymer",
        ),
        "U19": (
            "5-IMINO-4-(2-TRIFLUOROMETHYL-PHENYLAZO)-5H-PYRAZOL-3-YLAMINE",
            "non-polymer",
        ),
        "U18": (
            "(S)-2-(5-(((1,2-DIHYDRO-3-METHYL-1-OXOBENZO(F)QUINAZOLIN-9-YL)METHYL)AMINO)1-OXO-2-ISOINDOLINYL)GLUTARIC ACID",
            "non-polymer",
        ),
        "0KA": ("SILVER/IRON/SULFUR CLUSTER", "non-polymer"),
        "0KF": (
            "N-[3-({2-[(CYCLOPROPYLCARBONYL)AMINO]IMIDAZO[1,2-B]PYRIDAZIN-6-YL}OXY)PHENYL]-1,3-DIMETHYL-1H-PYRAZOLE-5-CARBOXAMIDE",
            "non-polymer",
        ),
        "0KD": (
            "N-(PYRIDIN-3-YL)-2-(THIOPHEN-3-YL)-3H-IMIDAZO[4,5-B]PYRIDINE-7-CARBOXAMIDE",
            "non-polymer",
        ),
        "6NP": (
            "6((S)-3-BENZYLPIPERAZIN-1-YL)-3-(NAPHTHALEN-2-YL)-4-(PYRIDIN-4-YL)PYRAZINE",
            "non-polymer",
        ),
        "0KJ": (
            "[(2S)-5-OXOPYRROLIDIN-2-YL]METHYL [(2S,3R)-4-{(1,3-BENZOTHIAZOL-6-YLSULFONYL)[(2S)-2-METHYLBUTYL]AMINO}-3-HYDROXY-1-PHENYLBUTAN-2-YL]CARBAMATE",
            "non-polymer",
        ),
        "0KK": ("(2E)-2-IMINO-3-METHYL-5,5-DIPHENYLIMIDAZOLIDIN-4-ONE", "non-polymer"),
        "U13": ("4-(4-FLUORO-PHENYLAZO)-5-IMINO-5H-PYRAZOL-3-YLAMINE", "non-polymer"),
        "U12": (
            "5-IMINO-4-(3-TRIFLUOROMETHYL-PHENYLAZO)-5H-PYRAZOL-3-YLAMINE",
            "non-polymer",
        ),
        "0KN": (
            "(2R)-N-[(2S,3R)-1-(1,3-BENZODIOXOL-5-YL)-4-{[(4'S)-6'-(2,2-DIMETHYLPROPYL)-3',4'-DIHYDROSPIRO[CYCLOBUTANE-1,2'-PYRANO[2,3-B]PYRIDIN]-4'-YL]AMINO}-3-HYDROXYBUTAN-2-YL]-2-METHOXYPROPANAMIDE",
            "non-polymer",
        ),
        "0KO": (
            "4-(2-[(6-METHOXYPYRIDIN-3-YL)AMINO]-5-{[4-(METHYLSULFONYL)PIPERAZIN-1-YL]METHYL}PYRIDIN-3-YL)-6-METHYL-1,3,5-TRIAZIN-2-AMINE",
            "non-polymer",
        ),
        "0KL": ("?", "non-polymer"),
        "0KM": (
            "(2E,5R)-2-IMINO-5-(3'-METHOXYBIPHENYL-3-YL)-3-METHYL-5-PHENYLIMIDAZOLIDIN-4-ONE",
            "non-polymer",
        ),
        "PME": ("N-L-ALPHA-ASPARTYL L-PHENYLALANINE 1-METHYL ESTER", "non-polymer"),
        "TM2": ("?", "non-polymer"),
        "656": (
            "2-{5-[AMINO(IMINIO)METHYL]-1H-BENZIMIDAZOL-2-YL}-6-ISOBUTOXYBENZENOLATE",
            "non-polymer",
        ),
        "654": ("4-(2-THIENYL)-1-(4-METHYLBENZYL)-1H-IMIDAZOLE", "non-polymer"),
        "655": (
            "2-{5-[AMINO(IMINIO)METHYL]-1H-BENZIMIDAZOL-2-YL}-6-(CYCLOPENTYLOXY)BENZENOLATE",
            "non-polymer",
        ),
        "653": (
            "5-(2-AMINOETHYL)-3-{5-[AMINO(IMINIO)METHYL]-1H-BENZIMIDAZOL-2-YL}-1,1'-BIPHENYL-2-OLATE",
            "non-polymer",
        ),
        "651": ("(3R)-3-BIPHENYL-4-YL-1-AZABICYCLO[2.2.2]OCTAN-3-OL", "non-polymer"),
        "SNU": ("4-(1H-PYRROL-1-YL)ANILINE", "non-polymer"),
        "311": (
            "(3,3-DIMETHYLPIPERIDIN-1-YL)(6-(3-FLUORO-4-METHYLPHENYL)PYRIDIN-2-YL)METHANONE",
            "non-polymer",
        ),
        "659": (
            "N-[(E)-3-[(2R,3R,4S,5R)-3-FLUORO-4-HYDROXY-5-[6-(METHYLAMINO)PURIN-9-YL]OXOLAN-2-YL]PROP-2-ENYL]-5-(4-FLUOROPHENYL)-2,3-DIHYDROXY-BENZAMIDE",
            "non-polymer",
        ),
        "65B": (
            "4-({6-AMINO-5-BROMO-2-[(4-CYANOPHENYL)AMINO]PYRIMIDIN-4-YL}OXY)-3,5-DIMETHYLBENZONITRILE",
            "non-polymer",
        ),
        "P99": (
            "2-[2,6-DIFLUORO-4-({2-[(PHENYLSULFONYL)AMINO]ETHYL}SULFANYL)PHENOXY]ACETAMIDE",
            "non-polymer",
        ),
        "VRG": (
            "N~2~-ACETYL-N~5~-[N-(METHYLCARBAMOYL)CARBAMIMIDOYL]-L-ORNITHYL-N-METHYL-L-PHENYLALANYL-L-ASPARTIC ACID",
            "non-polymer",
        ),
        "T23": (
            "2'-O-METHYL-3'-METHYL-3'-DEOXY-ARABINOFURANOSYL-THYMINE-5'-PHOSPHATE",
            "nucleic-acid",
        ),
        "317": (
            "(2S,3S)-3-{3-[2-CHLORO-4-(METHYLSULFONYL)PHENYL]-1,2,4-OXADIAZOL-5-YL}-1-CYCLOPENTYLIDENE-4-CYCLOPROPYL-1-FLUOROBUTAN-2-AMINE",
            "non-polymer",
        ),
        "PMC": ("3-(PHOSPHONOMETHYL)PYRIDINE-2-CARBOXYLIC ACID", "non-polymer"),
        "316": (
            "N'-[(1S,2S)-2-[(2S)-4-BENZYL-3-OXOPIPERAZIN-2-YL]-1-(3,5-DIFLUOROBENZYL)-2-HYDROXYETHYL]-5-METHYL-N,N-DIPROPYLBENZENE-1,3-DICARBOXAMIDE",
            "non-polymer",
        ),
        "PML": ("PIMELIC ACID", "non-polymer"),
        "315": (
            "(2S,3S)-4-CYCLOPROPYL-3-{(3R,5R)-3-[2-FLUORO-4-(METHYLSULFONYL)PHENYL]-1,2,4-OXADIAZOLIDIN-5-YL}-1-[(3S)-3-FLUOROPYRROLIDIN-1-YL]-1-OXOBUTAN-2-AMINE",
            "non-polymer",
        ),
        "PMM": ("PTERIN-6-YL-METHYL-MONOPHOSPHATE", "non-polymer"),
        "IHI": (
            "9-CYCLOPENTYL-6-[2-(3-IMIDAZOL-1-YL-PROPOXY)-PHENYLAMINO]-9H-PURINE-2-CARBONITRILE",
            "non-polymer",
        ),
        "314": (
            "N'-[(1S,2R)-2-[(2R,4S)-4-BENZYLPIPERIDIN-2-YL]-1-(3,5-DIFLUOROBENZYL)-2-HYDROXYETHYL]-5-METHYL-N,N-DIPROPYLBENZENE-1,3-DICARBOXAMIDE",
            "non-polymer",
        ),
        "PMN": ("PLATENSIMYCIN", "non-polymer"),
        "0K2": ("4-OXOBUTYL TRIHYDROGEN DIPHOSPHATE", "non-polymer"),
        "0K3": (
            "(2Z,6Z,10Z,14Z,18Z,22Z,26Z)-3,7,11,15,19,23,27,31-OCTAMETHYLDOTRIACONTA-2,6,10,14,18,22,26,30-OCTAEN-1-YL DIHYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "0K0": (
            "3-AMINO-6-{3-[(METHYLSULFONYL)AMINO]PHENYL}-N-(PIPERIDIN-4-YLMETHYL)PYRAZINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "0K1": (
            "3-AMINO-6-[3-(1-METHYL-1H-PYRAZOL-4-YL)PHENYL]-N-[(1R,2R,3S,5S,7S)-5-HYDROXYADAMANTAN-2-YL]PYRAZINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "0K6": (
            "7-CYCLOPENTYL-2-({1-METHYL-5-[(4-METHYLPIPERAZIN-1-YL)CARBONYL]-1H-PYRROL-3-YL}AMINO)-7H-PYRROLO[2,3-D]PYRIMIDINE-6-CARBOXAMIDE",
            "non-polymer",
        ),
        "0K7": ("2-(4-METHYLPHENYL)-1,3-THIAZOLE-4-CARBOXYLIC ACID", "non-polymer"),
        "0K5": (
            "(5-HYDROXY-4-{[(1-HYDROXY-2-OXO-6-PHENOXY-1,2-DIHYDROQUINOLIN-3-YL)AMINO]METHYL}-6-METHYLPYRIDIN-3-YL)METHYL DIHYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "PMH": (
            "(5-HYDROXY-4-{[(3-HYDROXYISOXAZOL-4-YL)AMINO]METHYL}-6-METHYLPYRIDIN-3-YL)METHYL DIHYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "0K9": (
            "(2R)-N-{(2S,3R)-4-{[(4'S)-6'-(2,2-DIMETHYLPROPYL)-3',4'-DIHYDROSPIRO[CYCLOBUTANE-1,2'-PYRANO[2,3-B]PYRIDIN]-4'-YL]AMINO}-3-HYDROXY-1-[3-(1,3-THIAZOL-2-YL)PHENYL]BUTAN-2-YL}-2-METHOXYPROPANAMIDE",
            "non-polymer",
        ),
        "PMI": ("(2-AMINO-2,3-DIHYDRO-1H-INDEN-2-YL)PHOSPHONIC ACID", "non-polymer"),
        "KR": ("KRYPTON", "non-polymer"),
        "SKF": ("1,2,3,4-TETRAHYDRO-ISOQUINOLINE-7-SULFONIC ACID AMIDE", "non-polymer"),
        "X32": (
            "N-[(R)-({[(BENZYLOXY)CARBONYL]AMINO}METHYL)(HYDROXY)PHOSPHORYL]-L-LEUCYL-L-PHENYLALANINE",
            "peptide",
        ),
        "U1P": ("?", "non-polymer"),
        "U1S": (
            "1-{5-DEOXY-5-[4-(ETHOXYCARBONYL)PIPERIDIN-1-YL]-ALPHA-L-ARABINOFURANOSYL}PYRIMIDINE-2,4(1H,3H)-DIONE",
            "non-polymer",
        ),
        "JGZ": (
            "4-(4-{[4-(4-CYANO-2,6-DIMETHYLPHENOXY)PYRIMIDIN-2-YL]AMINO}PIPERIDIN-1-YL)BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "PM4": (
            "S-(2-{[N-(2-HYDROXY-4-{[HYDROXY(OXIDO)PHOSPHINO]OXY}-3,3-DIMETHYLBUTANOYL)-BETA-ALANYL]AMINO}ETHYL) HEXANETHIOATE",
            "non-polymer",
        ),
        "T25": (
            "(9Z)-11-[(2R,3S)-3-PENTYLOXIRAN-2-YL]UNDEC-9-ENOIC ACID",
            "non-polymer",
        ),
        "PM5": (
            "S-(2-{[N-(2-HYDROXY-4-{[HYDROXY(OXIDO)PHOSPHINO]OXY}-3,3-DIMETHYLBUTANOYL)-BETA-ALANYL]AMINO}ETHYL) HEPTANETHIOATE",
            "non-polymer",
        ),
        "TF6": (
            "4-(3-{[5-(TRIFLUOROMETHYL)PYRIDIN-2-YL]OXY}BENZYL)PIPERIDINE-1-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "JP1": (
            "N-[3-CHLORO-4-(4-CHLORO-2-HYDROXYPHENOXY)PHENYL]MORPHOLINE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "JP3": (
            "2-AMINO-N-(FURAN-2-YLMETHYL)-4-OXO-3,4-DIHYDROPTERIDINE-7-CARBOXAMIDE",
            "non-polymer",
        ),
        "JP2": ("2-AMINO-4-OXO-1,4-DIHYDROPTERIDINE-7-CARBOXYLIC ACID", "non-polymer"),
        "PM1": (
            "[2-AMINO-6-(2,6-DIFLUORO-BENZOYL)-IMIDAZO[1,2-A]PYRIDIN-3-YL]-PHENYL-METHANONE",
            "non-polymer",
        ),
        "C92": (
            "N'-(6-AMINOPYRIDIN-3-YL)-N-(2-CYCLOPENTYLETHYL)-4-METHYL-BENZENE-1,3-DICARBOXAMIDE",
            "non-polymer",
        ),
        "C90": (
            "N-{(1S,2R)-2-HYDROXY-1-[(HYDROXYAMINO)CARBONYL]PROPYL}-4-{[4-(MORPHOLIN-4-YLMETHYL)PHENYL]ETHYNYL}BENZAMIDE",
            "non-polymer",
        ),
        "C96": (
            "4-{5-[(1Z)-1-(2-IMINO-4-OXO-1,3-THIAZOLIDIN-5-YLIDENE)ETHYL]-2-FURYL}BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "C95": (
            "4-{5-[(Z)-(2-IMINO-4-OXO-1,3-THIAZOLIDIN-5-YLIDENE)METHYL]FURAN-2-YL}BENZOIC ACID",
            "non-polymer",
        ),
        "C94": (
            "4-{5-[(Z)-(2-IMINO-4-OXO-1,3-THIAZOLIDIN-5-YLIDENE)METHYL]FURAN-2-YL}-2-(TRIFLUOROMETHYL)BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "C99": (
            "{(2R)-2-[(1S,2R)-1-AMINO-2-HYDROXYPROPYL]-2-HYDROXY-4,5-DIOXOIMIDAZOLIDIN-1-YL}ACETIC ACID",
            "peptide",
        ),
        "PM3": ("2-AMINO-3-(4-PHOSPHONOMETHYL-PHENYL)-PROPIONIC ACID", "peptide"),
        "SOD": ("?", "non-polymer"),
        "IZ1": (
            "N-[(1S)-2-{4-[(5S)-1,1-DIOXIDO-3-OXOISOTHIAZOLIDIN-5-YL]PHENYL}-1-(4-PHENYL-1H-IMIDAZOL-2-YL)ETHYL]-3-(TRIFLUOROMETHYL)BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "IZ2": (
            "N-[(1S)-1-(4-BENZYL-1H-IMIDAZOL-2-YL)-2-{4-[(5S)-1,1-DIOXIDO-3-OXOISOTHIAZOLIDIN-5-YL]PHENYL}ETHYL]-3-(TRIFLUOROMETHYL)BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "IZ3": (
            "3-FLUORO-N-[(1S)-1-[4-[(2-FLUOROPHENYL)METHYL]IMIDAZOL-2-YL]-2-[4-[(5S)-1,1,3-TRIOXO-1,2-THIAZOLIDIN-5-YL]PHENYL]ETHYL]BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "IZ4": (
            "N-{(1S)-2-{4-[(5S)-1,1-DIOXIDO-3-OXOISOTHIAZOLIDIN-5-YL]PHENYL}-1-[(4R)-4-(2-PHENYLETHYL)-4,5-DIHYDRO-1H-IMIDAZOL-2-YL]ETHYL}-3-FLUOROBENZENESULFONAMIDE",
            "non-polymer",
        ),
        "IZ5": (
            "N-{(1S)-2-{4-[(5S)-1,1-DIOXIDO-3-OXOISOTHIAZOLIDIN-5-YL]PHENYL}-1-[4-(3-PHENYLPROPYL)-1H-IMIDAZOL-2-YL]ETHYL}-3-FLUOROBENZENESULFONAMIDE",
            "non-polymer",
        ),
        "TLT": ("TOLMETIN", "non-polymer"),
        "T27": (
            "4-{[4-({4-[(E)-2-CYANOETHENYL]-2,6-DIMETHYLPHENYL}AMINO)PYRIMIDIN-2-YL]AMINO}BENZONITRILE",
            "non-polymer",
        ),
        "IZA": (
            "2-TERT-BUTYL-9-FLUORO-3,6-DIHYDRO-7H-BENZ[H]-IMIDAZ[4,5-F]ISOQUINOLINE-7-ONE",
            "non-polymer",
        ),
        "IZB": (
            "N-[(1S)-1-(1H-BENZIMIDAZOL-2-YL)-2-{4-[(5S)-1,1-DIOXIDO-3-OXOISOTHIAZOLIDIN-5-YL]PHENYL}ETHYL]-4-METHYL-3,4-DIHYDRO-2H-1,4-BENZOXAZINE-7-SULFONAMIDE",
            "non-polymer",
        ),
        "C9P": (
            "4-(DIHYDROXYBORANYL)-2-({[4-(PHENYLSULFONYL)THIOPHEN-2-YL]SULFONYL}AMINO)BENZOIC ACID",
            "non-polymer",
        ),
        "IZD": ("ISOTHIAZOLIDINONE ANALOG", "non-polymer"),
        "IZE": (
            "N-{(1S)-2-{4-[(5R)-1,1-DIOXIDO-3-OXOISOTHIAZOLIDIN-5-YL]PHENYL}-1-[5-(TRIFLUOROMETHYL)-1H-BENZIMIDAZOL-2-YL]ETHYL}-2,2,2-TRIFLUOROACETAMIDE",
            "non-polymer",
        ),
        "IZF": (
            "METHYL 2-{[5-({3-CHLORO-4-[(5S)-1,1-DIOXIDO-3-OXOISOTHIAZOLIDIN-5-YL]-N-(PHENYLSULFONYL)-L-PHENYLALANYL}AMINO)PENTYL]OXY}-6-HYDROXYBENZOATE",
            "non-polymer",
        ),
        "IZG": (
            "3-[(2R)-2,3-DIHYDROXYPROPYL]-6-FLUORO-5-[(2-FLUORO-4-IODOPHENYL)AMINO]-8-METHYLPYRIDO[2,3-D]PYRIMIDINE-4,7(3H,8H)-DIONE",
            "non-polymer",
        ),
        "QLI": (
            "(5S,6S)-6-AMINO-5-[(1-CARBOXYETHENYL)OXY]CYCLOHEXA-1,3-DIENE-1-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "GEQ": (
            "5-{[4-(9H-FLUOREN-9-YL)PIPERAZIN-1-YL]CARBONYL}-1H-INDOLE",
            "non-polymer",
        ),
        "IZN": (
            "(4BS,6S)-6-METHOXY-1,4,6,7,9,10,12,13-OCTAHYDRO-3H,5H-PYRANO[4',3':3,4]PYRIDO[2,1-I]INDOL-3-ONE",
            "non-polymer",
        ),
        "IZO": ("(2S)-2-AMINOHEX-5-YNOIC ACID", "peptide"),
        "IZP": ("(2R)-2-[4-(2-METHYLPROPYL)PHENYL]PROPANOIC ACID", "non-polymer"),
        "C9A": (
            "4-CHLOROBENZYL 6-FLUORO-7-(4-METHYLPIPERAZIN-1-YL)-1-[4-(METHYLSULFONYL)BENZYL]-4-OXO-1,4-DIHYDROQUINOLINE-3-CARBOXYLATE",
            "non-polymer",
        ),
        "MFC": (
            "5-[1-(3-FLUORO-4-HYDROXY-PHENYL)-METH-(Z)-YLIDENE]-3,5-DIHYDRO-IMIDAZOL-4-ONE",
            "peptide",
        ),
        "IZX": (
            "1-{4-[(R)-PHENYL(3-PHENYL-1,2,4-OXADIAZOL-5-YL)METHOXY]-1-BENZOTHIOPHEN-2-YL}METHANEDIAMINE",
            "non-polymer",
        ),
        "TCZ": ("1,3,5-TRICHLORO-BENZENE", "non-polymer"),
        "IZZ": (
            "(2S)-2-{[3-(3-AMINOPHENYL)IMIDAZO[1,2-B]PYRIDAZIN-6-YL]AMINO}-3-METHYLBUTAN-1-OL",
            "non-polymer",
        ),
        "C9H": ("5-(PENTYLSULFANYL)-1H-1,2,4-TRIAZOLE", "non-polymer"),
        "JPJ": ("2-(2,4-DICHLOROPHENOXY)-5-(2-PHENYLETHYL)PHENOL", "non-polymer"),
        "JPM": ("5-BENZYL-2-(2,4-DICHLOROPHENOXY)PHENOL", "non-polymer"),
        "JPL": (
            "5-(CYCLOHEXA-1,5-DIEN-1-YLMETHYL)-2-(2,4-DICHLOROPHENOXY)PHENOL",
            "non-polymer",
        ),
        "JPN": ("2-(2,4-DICHLOROPHENOXY)-5-(3-PHENYLPROPYL)PHENOL", "non-polymer"),
        "JPA": ("4-(2,4-DICHLOROPHENOXY)-2'-METHYLBIPHENYL-3-OL", "non-polymer"),
        "JPC": (
            "3-[(2,4-DICHLOROBENZOYL)(ISOPROPYL)AMINO]-5-PHENYLTHIOPHENE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "MFG": ("(1Z)-4-(4-FLUOROPHENYL)-2-METHYLIDENEBUTAN-1-IMINE", "non-polymer"),
        "JPZ": ("QUINAZOLIN-4(1H)-ONE", "non-polymer"),
        "TCW": ("TOLCAPONE", "non-polymer"),
        "JPS": (
            "(4R,6R,7S)-2-(2-CYCLOPROPYLETHYL)-4,6,7-TRIHYDROXY-4,5,6,7-TETRAHYDRO-1-BENZOTHIOPHENE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "JPR": (
            "1-(3-CHLORO-4-METHYLPHENYL)-3-{2-[({5-[(DIMETHYLAMINO)METHYL]-2-FURYL}METHYL)THIO]ETHYL}UREA",
            "non-polymer",
        ),
        "EHP": ("3-HYDROXYPHENYLALANINE", "peptide"),
        "TCT": ("6-(4-CHLORO-2-HYDROXY-PHENOXY)-NAPHTHALEN-2-OL", "non-polymer"),
        "O47": (
            "3-{2-[(E)-IMINOMETHYL]-6-PROPYLPYRIMIDIN-4-YL}-N,N-DIMETHYL-5-(TRIFLUOROMETHYL)BENZAMIDE",
            "non-polymer",
        ),
        "BGC": ("BETA-D-GLUCOSE", "saccharide"),
        "BGD": (
            "N-BENZOYL-L-GLUTAMYL-[4-PHOSPHONO(DIFLUOROMETHYL)]-L-PHENYLALANINE-[4-PHOSPHONO(DIFLUORO-METHYL)]-L-PHENYLALANINEAMIDE",
            "non-polymer",
        ),
        "BGF": ("2,5-BIS(4-GUANYLPHENYL)FURAN", "non-polymer"),
        "BGG": ("CARBONIC ACID MONOBENZYL ESTER", "non-polymer"),
        "BFT": ("S-BENZOYLTHIAMINE O-MONOPHOSPHATE", "non-polymer"),
        "BGL": ("B-2-OCTYLGLUCOSIDE", "saccharide"),
        "BGM": ("8-BROMO-2'-DEOXYGUANOSINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "BGN": ("N-BUTANOYL-2-AMINO-2-DEOXY-GLUCOPYRANOSIDE", "non-polymer"),
        "BGO": ("3'-O-(N-METHYLANTHRANILOYL)GUANOSINE-5'-DIPHOSPHATE", "non-polymer"),
        "BGP": ("BETA-GALACTOSE-6-PHOSPHATE", "saccharide"),
        "BGR": (
            "[[(2R,3S,5R)-5-(2-AZANYL-6-OXO-1H-PURIN-9-YL)-3-HYDROXY-OXOLAN-2-YL]METHOXY-HYDROXY-PHOSPHORYL]BORANUIDE",
            "non-polymer",
        ),
        "BGS": ("BETA-D-GLUCOPYRANOSYLSULFONYLETHANE", "saccharide"),
        "BGT": ("TERT-BUTYL D-ALPHA-GLUTAMINATE", "non-polymer"),
        "BGU": (
            "2,6-DIAMINO-1,7-DIHYDRO-8H-IMIDAZO[4,5-G]QUINAZOLIN-8-ONE",
            "non-polymer",
        ),
        "BGV": (
            "(3AR,4S,4AR,7AR,8R,9AS)-4A,8-DIMETHYL-3-METHYLIDENE-2,5-DIOXO-2,3,3A,4,4A,5,7A,8,9,9A-DECAHYDROAZULENO[6,5-B]FURAN-4-YL ACETATE",
            "non-polymer",
        ),
        "BGX": ("4-METHYL-PYRROLINE-5-CARBOXYLIC ACID", "non-polymer"),
        "NI3": ("NICKEL (II) ION, 3 WATERS COORDINATED", "non-polymer"),
        "NI1": ("NICKEL ION, 1 WATER COORDINATED", "non-polymer"),
        "NI7": (
            "(4AS,7AS)-1,4-BIS(3-HYDROXYBENZYL)HEXAHYDRO-1H-PYRROLO[3,4-B]PYRAZINE-2,3-DIONE",
            "non-polymer",
        ),
        "99T": (
            "L-GAMMA-GLUTAMYL-S-(2-CARBOXYETHYL)-L-CYSTEINYLGLYCINE",
            "non-polymer",
        ),
        "99P": ("D-PHENYLALANYL-N-(PYRIDIN-3-YLMETHYL)-L-PROLINAMIDE", "peptide"),
        "A79": (
            "N-{1-BENZYL-(2S,3S)-2,3-DIHYDROXY-4-[3-METHYL-2-(3-METHYL-3-PYRIDIN-2-YLMETHYL-UREIDO)-BUTYRYLAMINO]-5-PHENYL-PENTYL}-3-METHYL-2-(3-METHYL-3-PYRIDIN-2-YLMETHYL-UREIDO)-BUTYRAMIDE",
            "non-polymer",
        ),
        "A78": (
            "N-{1-BENZYL-3-HYDROXY-4-[3-METHYL-2-(3-METHYL-3-PYRIDIN-2-YLMETHYL-UREIDO)-BUTYRYLAMINO]-5-PHENYL-PENTYL}-3-METHYL-2-(3-METHYL-3-PYRIDIN-2-YLMETHYL-UREIDO)-BUTYRAMIDE",
            "non-polymer",
        ),
        "A77": (
            "N-{1-BENZYL-(2R,3S)-2,3-DIHYDROXY-4-[3-METHYL-2-(3-METHYL-3-PYRIDIN-2-YLMETHYL-UREIDO)-BUTYRYLAMINO]-5-PHENYL-PENTYL}-3-METHYL-2-(3-METHYL-3-PYRIDIN-2-YLMETHYL-UREIDO)-BUTYRAMIDE",
            "non-polymer",
        ),
        "A76": (
            "N-{1-BENZYL-(2R,3R)-2,3-DIHYDROXY-4-[3-METHYL-2-(3-METHYL-3-PYRIDIN-2-YLMETHYL-UREIDO)-BUTYRYLAMINO]-5-PHENYL-PENTYL}-3-METHYL-2-(3-METHYL-3-PYRIDIN-2-YLMETHYL-UREIDO)-BUTYRAMIDE",
            "non-polymer",
        ),
        "A75": (
            "2-[(PHENYLSULFONYL)AMINO]-5,6,7,8-TETRAHYDRONAPHTHALENE-1-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "A74": (
            "2,5-DICHLORO-N-[5-METHOXY-7-(6-METHOXYPYRIDIN-3-YL)-1,3-BENZOXAZOL-2-YL]BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "TH3": ("TRANS DELTA2 PALMITENOYL-COENZYMEA", "non-polymer"),
        "A72": ("TRANS-BIS(HEXYLAMINE)-(DIAMMINE) PLATINUM(II) COMPLEX", "non-polymer"),
        "A71": ("(6-AMINOHEXYLAMINE)(TRIAMMINE) PLATINUM(II) COMPLEX", "non-polymer"),
        "99Z": (
            "4-{[4-AMINO-5-(2-METHYLBENZOYL)-1,3-THIAZOL-2-YL]AMINO}BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "TCM": ("THIOCAMPHOR", "non-polymer"),
        "99A": (
            "N-{(3-ENDO)-8-[5-(BENZYLCARBAMOYL)PYRIDIN-2-YL]-8-AZABICYCLO[3.2.1]OCT-3-YL}-2,5-DIMETHYLBENZENE-1,4-DICARBOXAMIDE",
            "non-polymer",
        ),
        "GP9": ("(2R)-2,3-DIHYDROXYPROPYL PHOSPHATE", "non-polymer"),
        "99B": (
            "5-[(2R)-BUTAN-2-YLAMINO]-N-{(3-ENDO)-8-[5-(CYCLOPROPYLCARBONYL)PYRIDIN-2-YL]-8-AZABICYCLO[3.2.1]OCT-3-YL}-2-METHYLBENZENE-1,4-DICARBOXAMIDE",
            "non-polymer",
        ),
        "GP4": ("GLUCOSAMINE 4-PHOSPHATE", "saccharide"),
        "GP5": ("DIGUANOSINE-PENTAPHOSPHATE", "non-polymer"),
        "GP6": ("1-(4-AMIDINOPHENYL)-3-(4-CHLOROPHENYL)UREA", "non-polymer"),
        "GP7": (
            "(1R)-2-{[(S)-(2-AMINOETHOXY)(HYDROXY)PHOSPHORYL]OXY}-1-[(PENTADECANOYLOXY)METHYL]ETHYL (12E)-HEXADECA-9,12-DIENOATE",
            "non-polymer",
        ),
        "GP0": (
            "5-FLUORO-1-[3-C-(HYDROXYMETHYL)-BETA-D-GLUCOPYRANOSYL]PYRIMIDINE-2,4(1H,3H)-DIONE",
            "non-polymer",
        ),
        "GP1": ("GLUCOSAMINE 1-PHOSPHATE", "saccharide"),
        "GP2": ("PHOSPHOMETHYLPHOSPHONIC ACID GUANOSYL ESTER", "non-polymer"),
        "GP3": ("DIGUANOSINE-5'-TRIPHOSPHATE", "non-polymer"),
        "995": (
            "4-METHYL-PENTANOIC ACID {1-[4-GUANIDINO-1-(THIAZOLE-2-CARBONYL)-BUTYLCARBAMOYL]-2-METHYL-PROPYL}-AMIDE",
            "non-polymer",
        ),
        "994": (
            "4-(2-ANILINOPYRIDIN-3-YL)-N-(3,4,5-TRIMETHOXYPHENYL)-1,3,5-TRIAZIN-2-AMINE",
            "non-polymer",
        ),
        "997": (
            "1-[3,3-DIMETHYL-2-(2-METHYLAMINO-PROPIONYLAMINO)-BUTYRYL]-4-PHENOXY-PYRROLIDINE-2-CARBOXYLIC ACID(1,2,3,4-TETRAHYDRO-NAPHTHALEN-1-YL)-AMIDE",
            "non-polymer",
        ),
        "996": (
            "N-[(2S,3R)-4-{[(4S)-2-(2,2-DIMETHYLPROPYL)-6,6-DIMETHYL-4,5,6,7-TETRAHYDRO-2H-INDAZOL-4-YL]AMINO}-3-HYDROXY-1-PHENYLBUTAN-2-YL]ACETAMIDE",
            "non-polymer",
        ),
        "991": (
            "2-{5-[AMINO(IMINIO)METHYL]-6-CHLORO-1H-INDOL-2-YL}-6-(CYCLOPENTYLOXY)BENZENOLATE",
            "non-polymer",
        ),
        "GPI": (
            "(2S)-[3-PYRIDYL-1-PROPYL]-1-[3,3-DIMETHYL-1,2-DIOXOPENTYL]-2-PYRROLIDINECARBOXYLATE",
            "non-polymer",
        ),
        "993": (
            "3-CHLORO-4-(2-METHYLAMINO-IMIDAZOL-1-YLMETHYL)-THIOPHENE-2-CARBOXYLIC ACID [4-CHLORO-2-(5-CHLORO-PYRIDIN-2-YLCARBAMOYL)-6-METHOXY-PHENYL]-AMIDE",
            "non-polymer",
        ),
        "GPK": (
            "5-FLUORO-1-(BETA-D-GLUCOPYRANOSYL)PYRIMIDINE-2,4(1H,3H)-DIONE",
            "non-polymer",
        ),
        "GPD": ("(2R,3S,5S,6R)-5-AMINO-3-HYDROXY-6-METHYL-OXAN-2-YL", "non-polymer"),
        "GPE": ("L-ALPHA-GLYCEROPHOSPHORYLETHANOLAMINE", "non-polymer"),
        "GPF": ("N-(PHOSPHONOMETHYL)GLYCINE", "non-polymer"),
        "A7T": (
            "(1R,5S)-N-CYCLOPROPYL-7-{4-[2-(2,6-DICHLORO-4-METHYLPHENOXY)ETHOXY]PHENYL}-N-(2,3-DIMETHYLBENZYL)-3,9-DIAZABICYCLO[3.3.1]NON-6-ENE-6-CARBOXAMIDE",
            "non-polymer",
        ),
        "THS": ("THYMIDINE-5'-(DITHIO)PHOSPHATE", "non-polymer"),
        "998": (
            "N-METHYLALANYL-3-METHYLVALYL-N-(1,2,3,4-TETRAHYDRONAPHTHALEN-1-YL)PROLINAMIDE",
            "non-polymer",
        ),
        "GPB": (
            "N-[4-[2-(2-AMINO-4-METHYL-7H-PYRROLO[2,3-D]PYRIMIDIN-5-YL)-ETHYL]-BENZOYL]GLUTAMIC ACID",
            "non-polymer",
        ),
        "GPC": ("2'-FLUOROGUANYLYL-(3'-5')-PHOSPHOCYTIDINE", "non-polymer"),
        "THO": ("REDUCED THREONINE", "peptide"),
        "MFX": (
            "1-CYCLOPROPYL-6-FLUORO-8-METHOXY-7-[(4AS,7AS)-OCTAHYDRO-6H-PYRROLO[3,4-B]PYRIDIN-6-YL]-4-OXO-1,4-DIHYDROQUINOLINE-3-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "THM": ("THYMIDINE", "non-polymer"),
        "THL": (
            "N-[4-({[(6S)-2-AMINO-4-OXO-1,4,5,6,7,8-HEXAHYDROPTERIDIN-6-YL]METHYL}AMINO)BENZOYL]-L-GLUTAMIC ACID",
            "non-polymer",
        ),
        "GPX": ("?", "non-polymer"),
        "PNY": (
            "(2R)-2,4-DIHYDROXY-3,3-DIMETHYL-N-{3-OXO-3-[(2-SULFANYLETHYL)AMINO]PROPYL}BUTANAMIDE",
            "non-polymer",
        ),
        "A7I": (
            "5-CHLORO-THIOPHENE-2-CARBOXYLIC ACID [2-(1--ISOPROPYL-PIPERIDIN-4-YLSULFAMOYL)-ETHYL]-AMIDE",
            "non-polymer",
        ),
        "MFY": ("2-[(4-CHLOROBENZYL)AMINO]ETHANESULFONIC ACID", "non-polymer"),
        "GPT": (
            "(2R,3S,4R)-2-[(S)-(1S)-CYCLOHEX-2-EN-1-YL(HYDROXY)METHYL]-4-(2-FLUOROETHYL)-3-HYDROXY-3-METHYL-5-OXOPYRROLIDINE-2-CARBALDEHYDE",
            "non-polymer",
        ),
        "GPU": (
            "5-BROMO-1-(BETA-D-GLUCOPYRANOSYL)PYRIMIDINE-2,4(1H,3H)-DIONE",
            "non-polymer",
        ),
        "GPV": (
            "1-(BETA-D-GLUCOPYRANOSYL)-5-IODOPYRIMIDINE-2,4(1H,3H)-DIONE",
            "non-polymer",
        ),
        "A7D": ("5'-S-[(3R)-3,4-DIHYDROXYBUTYL]-5'-THIOADENOSINE", "non-polymer"),
        "GPP": ("GERANYL DIPHOSPHATE", "non-polymer"),
        "A7B": ("8-METHOXY-4-METHYLQUINOLIN-2(1H)-ONE", "non-polymer"),
        "GPR": (
            "(9R,10R)-9-(S-GLUTATHIONYL)-10-HYDROXY-9,10-DIHYDROPHENANTHRENE",
            "non-polymer",
        ),
        "GPS": (
            "L-GAMMA-GLUTAMYL-S-[(9S,10S)-10-HYDROXY-9,10-DIHYDROPHENANTHREN-9-YL]-L-CYSTEINYLGLYCINE",
            "non-polymer",
        ),
        "NIK": ("PENTAAQUANICKEL", "non-polymer"),
        "JG3": ("1-[(2S,3S)-3-AMINO-2-HYDROXY-4-PHENYLBUTYL]-L-PROLINE", "peptide"),
        "NIN": ("DINITROPHENYLENE", "non-polymer"),
        "NIO": ("NICOTINIC ACID", "non-polymer"),
        "NIL": ("NILOTINIB", "non-polymer"),
        "NIM": ("4-NITRO-2-PHENOXYMETHANESULFONANILIDE", "non-polymer"),
        "NIC": ("NITROISOCITRIC ACID", "non-polymer"),
        "NIA": (
            "((2R,3S,4R,5R)-5-(5-AMINO-4-NITRO-1H-IMIDAZOL-1-YL)-3,4-DIHYDROXYTETRAHYDROFURAN-2-YL)METHYL DIHYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "NIF": (
            "2-ETHYL-7-PIPERAZIN-1-YL-5H-[1,3,4]THIADIAZOLO[3,2-A]PYRIMIDIN-5-ONE",
            "non-polymer",
        ),
        "NIG": ("N-(IMINOMETHYL)-L-GLUTAMIC ACID", "non-polymer"),
        "NID": ("4-NITRO-INDEN-1-ONE", "non-polymer"),
        "HJZ": (
            "3-[2-(TRANS-4-METHYLCYCLOHEXYL)PHENYL]-5-PHENYLTHIOPHENE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "NIZ": ("PYRIDINE-4-CARBOHYDRAZIDE", "non-polymer"),
        "BG1": (
            "O-[(2S)-2-{METHYL[(METHYLAMINO)SULFONYL]AMINO}PENTANOYL]-L-SERINE",
            "peptide",
        ),
        "BG2": ("8-BROMOGUANOSINE", "non-polymer"),
        "BG3": ("3-METHYL-5-SULFO-PYRROLIDINE-2-CARBOXYLIC ACID", "non-polymer"),
        "BG4": (
            "5-(HYDROXY-METHYL-AMINO)-3-METHYL-PYRROLIDINE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "BG5": ("5-HYDROXYAMINO-3-METHYL-PYRROLIDINE-2-CARBOXYLIC ACID", "non-polymer"),
        "BG6": ("BETA-D-GLUCOSE-6-PHOSPHATE", "non-polymer"),
        "NIR": (
            "3-(AMINOCARBONYL)-1-[(3R,4S,5R)-3,4-DIHYDROXY-5-METHYLTETRAHYDRO-2-FURANYL]PYRIDINIUM",
            "non-polymer",
        ),
        "NIS": ("2-NITROPROPANE", "non-polymer"),
        "NIP": (
            "4-HYDROXY-5-IODO-3-NITROPHENYLACETYL-EPSILON-AMINOCAPROIC ACID ANION",
            "non-polymer",
        ),
        "NIQ": (
            "6-AMINO-1-BUTYL-5-(CYCLOPENTYLAMINO)PYRIMIDINE-2,4(1H,3H)-DIONE",
            "non-polymer",
        ),
        "NIV": ("NEVIRAPINE", "non-polymer"),
        "NIT": ("4-NITROANILINE", "non-polymer"),
        "NIU": (
            "6,7,8,9-TETRAHYDRO-4-HYDROXY-3-(1-PHENYLPROPYL)CYCLOHEPTA[B]PYRAN-2-ONE",
            "non-polymer",
        ),
        "FDA": ("DIHYDROFLAVINE-ADENINE DINUCLEOTIDE", "non-polymer"),
        "FDC": ("IRON(III) DICITRATE", "non-polymer"),
        "FDE": ("FE(III) DEUTEROPORPHYRIN IX", "non-polymer"),
        "FDD": ("FE(III) 2,4-DIMETHYL DEUTEROPORPHYRIN IX", "non-polymer"),
        "FDG": (
            "2-AMINO-9-(2-DEOXY-2-FLUORO-5-O-PHOSPHONO-BETA-D-ARABINOFURANOSYL)-7,9-DIHYDRO-1H-PURINE-6,8-DIONE",
            "nucleic-acid",
        ),
        "FDF": (
            "(2E,6E)-12-FLUORO-11-(FLUOROMETHYL)-3,7-DIMETHYLDODECA-2,6,10-TRIEN-1-YL TRIHYDROGEN DIPHOSPHATE",
            "non-polymer",
        ),
        "FDI": (
            "4-(N-ACETYLAMINO)-3-[N-(2-ETHYLBUTANOYLAMINO)]BENZOIC ACID",
            "non-polymer",
        ),
        "FDM": (
            "PHOSPHORIC ACID MONO-[3-FLUORO-5-(5-METHYL-2,4-DIOXO-3,4-DIHYDRO-2H-PYRIMIDIN-1-YL)-TETRAHYRO-FURAN-2-YLMETHYL] ESTER",
            "non-polymer",
        ),
        "FDL": (
            "N~6~-ACETYL-N-(4-METHYL-2-OXO-2H-CHROMEN-7-YL)-L-LYSINAMIDE",
            "peptide",
        ),
        "FDN": (
            "(5S)-3-ANILINO-5-(2,4-DIFLUOROPHENYL)-5-METHYL-1,3-OXAZOLIDINE-2,4-DIONE",
            "non-polymer",
        ),
        "FDQ": (
            "D-GLUCO-2,5-ANHYDRO-1-DEOXY-1-PHOSPHONOHEXITOL-6-PHOSPHATE",
            "saccharide",
        ),
        "FDP": ("FRUCTOSE-2,6-DIPHOSPHATE", "saccharide"),
        "038": ("?", "peptide"),
        "039": (
            "2-((9H-PURIN-6-YLTHIO)METHYL)-5-CHLORO-3-(2-METHOXYPHENYL)QUINAZOLIN-4(3H)-ONE",
            "non-polymer",
        ),
        "FDT": (
            "(7R)-6-FORMYL-7-(1-METHYL-1H-1,2,3-TRIAZOL-4-YL)-4,7-DIHYDRO-1,4-THIAZEPINE-3-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "032": (
            "N-(3-{[5-(4-CHLOROPHENYL)-1H-PYRROLO[2,3-B]PYRIDIN-3-YL]CARBONYL}-2,4-DIFLUOROPHENYL)PROPANE-1-SULFONAMIDE",
            "non-polymer",
        ),
        "033": (
            "N-({4'-[(1-BENZOFURAN-2-YLCARBONYL)AMINO]-1,1'-BIPHENYL-4-YL}SULFONYL)-L-VALINE",
            "non-polymer",
        ),
        "030": (
            "(3R,3AR,5R,6AR)-3-HYDROXYHEXAHYDRO-2H-CYCLOPENTA[B]FURAN-5-YL HYDROGEN CARBONATE",
            "non-polymer",
        ),
        "031": (
            "(3AS,5R,6AR)-HEXAHYDRO-2H-CYCLOPENTA[B]FURAN-5-YL [(1S,2R)-1-BENZYL-2-HYDROXY-3-([(4-METHOXYPHENYL)SULFONYL]{[(2R)-5-OXOPYRROLIDIN-2-YL]METHYL}AMINO)PROPYL]CARBAMATE",
            "non-polymer",
        ),
        "036": ("(2R)-2,7-BIS(PHOSPHONOOXY)HEPTANOIC ACID", "non-polymer"),
        "037": (
            "(1S,3AR,6AS)-OCTAHYDROCYCLOPENTA[C]PYRROLE-1-CARBOXYLIC ACID",
            "peptide",
        ),
        "034": (
            "5-(4-{[3-(2,6-DICHLOROPHENYL)-5-(PROPAN-2-YL)-1,2-OXAZOL-4-YL]METHOXY}PHENYL)-1H-INDOLE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "035": ("(2S)-2,7-BIS(PHOSPHONOOXY)HEPTANOIC ACID", "non-polymer"),
        "EF3": (
            "(E)-1-(1-METHYL-6-{4-[3-(4-METHYLPIPERAZIN-1-YL)PROPOXY]-3-(TRIFLUOROMETHYL)PHENYL}-1H-IMIDAZO[4,5-C]PYRIDIN-4-YL)METHANIMINE",
            "non-polymer",
        ),
        "PNZ": ("P-NITRO-BENZYLAMINE", "peptide"),
        "TC6": ("(2E)-HEXENOYL-COA", "non-polymer"),
        "TC7": ("2,4,5-TRICHLOROPHENOL", "non-polymer"),
        "TC4": (
            "BUTYL-PHOSPHINIC ACID 2,3-BIS-BUTYLCARBAMOYLOXY-PROPYL ESTER GROUP",
            "non-polymer",
        ),
        "RT": ("RIBOSYLTHYMINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "EFZ": (
            "(-)-6-CHLORO-4-CYCLOPROPYLETHYNYL-4-TRIFLUOROMETHYL-1,4-DIHYDRO-2H-3,1-BENZOXAZIN-2-ONE",
            "non-polymer",
        ),
        "TC5": ("ETHYL HYDROGEN PROPYLAMIDOPHOSPHATE", "non-polymer"),
        "EFP": (
            "N-{1-[CIS-3-(ACETYLAMINO)CYCLOBUTYL]-1H-IMIDAZOL-4-YL}-2-(4-METHOXYPHENYL)ACETAMIDE",
            "non-polymer",
        ),
        "EFQ": (
            "N-[1-(CIS-3-HYDROXYCYCLOBUTYL)-1H-IMIDAZOL-4-YL]-2-(4-METHOXYPHENYL)ACETAMIDE",
            "non-polymer",
        ),
        "EFS": ("ETHYL DIHYDROGEN PHOSPHATE", "non-polymer"),
        "EFT": (
            "(2R)-3-CYCLOHEXYL-2-(6-HYDROXY-1H-INDAZOL-3-YL)PROPANENITRILE",
            "non-polymer",
        ),
        "EFU": (
            "N-BENZYL-6-HYDROXY-N-METHYL-3-(3-METHYLBENZYL)-1H-INDAZOLE-5-CARBOXAMIDE",
            "non-polymer",
        ),
        "TC3": ("ETHYL HYDROGEN METHYLAMIDOPHOSPHATE", "non-polymer"),
        "TC0": (
            "4'-[5-[[3-[(CYCLOPROPYLAMINO)METHYL]PHENYL]AMINO]-1H-PYRAZOL-3-YL]-[1,1'-BIPHENYL]-2,4-DIOL",
            "non-polymer",
        ),
        "EFO": ("OLIGOMYCIN A", "non-polymer"),
        "F50": ("ETHANEPEROXOIC ACID", "non-polymer"),
        "EFC": ("S,S-(2-FLUOROETHYL)THIOCYSTEINE", "peptide"),
        "EFE": ("ENANTIO-PYOCHELIN FE(III)", "non-polymer"),
        "EFG": (
            "1-(2-DEOXY-2-FLUORO-5-O-PHOSPHONO-BETA-D-ARABINOFURANOSYL)-1H-IMIDAZO[2,1-B]PURIN-4(5H)-ONE",
            "non-polymer",
        ),
        "03J": ("3-[4-(4-PHENYLTHIOPHEN-2-YL)PHENYL]PROPANOIC ACID", "non-polymer"),
        "03H": ("(2S)-2-CHLORO-4-METHYLPENTANOIC ACID", "non-polymer"),
        "03I": (
            "(2R)-N-HYDROXY-2-METHYL-2-(METHYLSULFONYL)-4-{4'-[3-(MORPHOLIN-4-YL)PROPOXY]BIPHENYL-4-YL}BUTANAMIDE",
            "non-polymer",
        ),
        "03N": ("4-(AMINOMETHYL)PIPERIDINE-1-CARBOXIMIDAMIDE", "non-polymer"),
        "03O": (
            "4-({[(6S)-2-AMINO-5-METHYL-4-OXO-3,4,5,6,7,8-HEXAHYDROPTERIDIN-6-YL]METHYL}AMINO)BENZOIC ACID",
            "non-polymer",
        ),
        "03L": (
            "2-({2-CHLORO-6-[(2,4-DICHLOROPHENYL)SULFANYL]BENZYL}CARBAMOYL)BENZOIC ACID",
            "non-polymer",
        ),
        "03M": (
            "(5Z)-5-[(6-CHLORO-7-METHYL-1H-INDOL-3-YL)METHYLIDENE]-3-(3,4-DIFLUOROBENZYL)IMIDAZOLIDINE-2,4-DIONE",
            "non-polymer",
        ),
        "03B": (
            "5-(4-CHLOROPHENYL)-4-{3-[4-(4-{[(4-{[(2R)-4-(DIMETHYLAMINO)-1-(PHENYLSULFANYL)BUTAN-2-YL]AMINO}-3-NITROPHENYL)SULFONYL]AMINO}PHENYL)PIPERAZIN-1-YL]PHENYL}-1,2-DIMETHYL-1H-PYRROLE-3-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "03C": (
            "[5-AMINO-1-(2-METHYLPHENYL)-1H-PYRAZOL-4-YL]{3-[1-(METHYLSULFONYL)PIPERIDIN-4-YL]PHENYL}METHANONE",
            "non-polymer",
        ),
        "RNA": (
            "4-METHYL-N-{(5E)-5-[(5-METHYL-2-FURYL)METHYLENE]-4-OXO-4,5-DIHYDRO-1,3-THIAZOL-2-YL}BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "03A": ("(3S)-3-AMINO-N-CYCLOPROPYL-2-OXOHEXANAMIDE", "non-polymer"),
        "03F": (
            "(9Z)-N-[(2S,3R,4E)-1-(BETA-D-GLUCOPYRANOSYLOXY)-3-HYDROXYOCTADEC-4-EN-2-YL]OCTADEC-9-ENAMIDE",
            "non-polymer",
        ),
        "03G": (
            "N-(4-CHLOROPHENYL)-N'-(2,2,6,6-TETRAMETHYLPIPERIDIN-4-YL)ETHANEDIAMIDE",
            "non-polymer",
        ),
        "03D": (
            "(2R,4S,5S)-N-[(2S,3R,4S)-1-CYCLOHEXYL-3,4-DIHYDROXY-6-METHYLHEPTAN-2-YL]-2-(CYCLOPROPYLMETHYL)-4,5-DIHYDROXY-6-PHENYLHEXANAMIDE",
            "non-polymer",
        ),
        "03E": ("1-AMINOCYCLOHEPTANECARBOXYLIC ACID", "peptide"),
        "03Z": (
            "4-[(4-AMINO-5-BENZOYL-1,3-THIAZOL-2-YL)AMINO]BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "LCG": (
            "[(1R,3R,4R,7S)-7-HYDROXY-3-(GUANIN-9-YL)-2,5-DIOXABICYCLO[2.2.1]HEPT-1-YL]METHYL DIHYDROGEN PHOSPHATE",
            "nucleic-acid",
        ),
        "03X": (
            "N-{4-[(6,7-DIMETHOXYQUINOLIN-4-YL)OXY]-3-FLUOROPHENYL}-1,5-DIMETHYL-3-OXO-2-PHENYL-2,3-DIHYDRO-1H-PYRAZOLE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "03Y": ("2-METHYL-L-CYSTEINE", "peptide"),
        "FD4": (
            "N-ALPHA-(2-NAPHTHYLSULFONYL)-N(3-AMIDINO-L-PHENYLALANINYL)-4-ACETYL-PIPERAZINE",
            "non-polymer",
        ),
        "LCA": (
            "[(1R,3R,4R,7S)-7-HYDROXY-3-(ADENIN-9-YL)-2,5-DIOXABICYCLO[2.2.1]HEPT-1-YL]METHYL DIHYDROGEN PHOSPHATE",
            "nucleic-acid",
        ),
        "03R": (
            "2-({6-{3-[AMINO(IMINO)METHYL]PHENOXY}-3,5-DIFLUORO-4-[(1-METHYL-3-PHENYLPROPYL)AMINO]-2-PYRIDINYL}OXY)BENZOIC ACID",
            "non-polymer",
        ),
        "03S": ("METHANESULFONIC ACID", "non-polymer"),
        "03P": (
            "N-{2-[4-({3-CHLORO-4-[3-(TRIFLUOROMETHYL)PHENOXY]PHENYL}AMINO)-5H-PYRROLO[3,2-D]PYRIMIDIN-5-YL]ETHYL}-3-HYDROXY-3-METHYLBUTANAMIDE",
            "non-polymer",
        ),
        "03Q": (
            "2-{2-[4-({5-CHLORO-6-[3-(TRIFLUOROMETHYL)PHENOXY]PYRIDIN-3-YL}AMINO)-5H-PYRROLO[3,2-D]PYRIMIDIN-5-YL]ETHOXY}ETHANOL",
            "non-polymer",
        ),
        "03V": ("NAPHTHALEN-2-OL", "non-polymer"),
        "03W": ("(2R)-PROPANE-1,1,2,3-TETROL", "non-polymer"),
        "03T": ("1-BENZOTHIOPHENE-2-SULFONAMIDE", "non-polymer"),
        "03U": ("?", "non-polymer"),
        "PL7": (
            "(2E)-2-({(2S)-2-CARBOXY-2-[(PHENOXYACETYL)AMINO]ETHOXY}IMINO)PENTANEDIOIC ACID",
            "non-polymer",
        ),
        "0MT": (
            "4-[(4R)-4-(4-NITROPHENYL)-6-OXIDANYLIDENE-3-PHENYL-1,4-DIHYDROPYRROLO[3,4-C]PYRAZOL-5-YL]BENZOIC ACID",
            "non-polymer",
        ),
        "PL5": (
            "(2S,6S)-2-AMINO-6-{[(1E)-{3-HYDROXY-2-METHYL-5-[(PHOSPHONOOXY)METHYL]PYRIDIN-4-YL}METHYLIDENE]AMINO}HEPTANEDIOIC ACID",
            "non-polymer",
        ),
        "DVN": (
            "[[(2R,3R,4R,5R)-5-(6-AMINOPURIN-9-YL)-3-OXIDANYL-4-PHOSPHONOOXY-OXOLAN-2-YL]METHOXY-OXIDANYL-PHOSPHORYL] [(2R,3S,4R,5S)-3,4,5-TRIS(OXIDANYL)OXOLAN-2-YL]METHYL HYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "PL3": ("HEXADECAN-1-OL", "non-polymer"),
        "PL2": (
            "(4-{[(4-{[AMINO(IMINO)METHYL]AMINO}BUTYL)AMINO]METHYL}-5-HYDROXY-6-METHYLPYRIDIN-3-YL)METHYL DIHYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "PL1": ("COBALT (III)-DEGLYCOPEPLEOMYCIN", "non-polymer"),
        "PL0": ("1-PHENYLGUANIDINE", "non-polymer"),
        "WZ2": (
            "METHYL ALPHA-D-MANNOPYRANOSYL-(1->3)-[ALPHA-D-MANNOPYRANOSYL-(1->6)]-3-THIO-ALPHA-D-MANNOPYRANOSIDE",
            "non-polymer",
        ),
        "DVA": ("D-VALINE", "peptide"),
        "5ZA": (
            "(5Z)-2-[(1S,2R)-1-AMINO-2-HYDROXYPROPYL]-5-[(4-AMINO-1H-INDOL-3-YL)METHYLENE]-3-(2-HYDROXYETHYL)-3,5-DIHYDRO-4H-IMIDAZOL-4-ONE",
            "peptide",
        ),
        "DVC": ("(2R,4S,6S)-4-AZANYL-4,6-DIMETHYL-OXANE-2,5,5-TRIOL", "saccharide"),
        "PL8": (
            "[5-HYDROXY-4-({[6-(3-HYDROXYPROPYL)-2-OXO-1,2-DIHYDROPYRIDIN-3-YL]AMINO}METHYL)-6-METHYLPYRIDIN-3-YL]METHYL DIHYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "CD4": (
            "(2R,5R,11R,14R)-5,8,11-TRIHYDROXY-5,11-DIOXIDO-17-OXO-2,14-BIS(TETRADECANOYLOXY)-4,6,10,12,16-PENTAOXA-5,11-DIPHOSPHATRIACONT-1-YL TETRADECANOATE",
            "non-polymer",
        ),
        "CD5": ("CADMIUM ION, 5 WATERS COORDINATED", "non-polymer"),
        "CD6": (
            "(2R)-5-METHYLIDENE-2-{(1R)-2-OXO-1-[(PHENYLACETYL)AMINO]ETHYL}-5,6-DIHYDRO-2H-1,3-THIAZINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "CD1": ("CADMIUM ION, 1 WATER COORDINATED", "non-polymer"),
        "CD2": (
            "1-(2-AMINOETHYL)-3-[2-(QUINOLIN-3-YL)PYRIDIN-4-YL]-1H-PYRAZOLE-5-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "CD3": ("CADMIUM ION, 3 WATERS COORDINATED", "non-polymer"),
        "DVT": ("DECAVANADATE", "non-polymer"),
        "HR2": (
            "(3R,5R)-7-[3-(4-FLUOROPHENYL)-1-ISOPROPYL-8-OXO-7-PHENYL-1,4,5,6,7,8-HEXAHYDROPYRROLO[2,3-C]AZEPIN-2-YL]-3,5-DIHYDROXYHEPTANOIC ACID",
            "non-polymer",
        ),
        "CD8": (
            "(2R)-2-[(1S)-1-(BENZOYLAMINO)-1-METHOXY-2-OXOETHYL]-5-METHYLIDENE-5,6-DIHYDRO-2H-1,3-THIAZINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "CD9": (
            "5-[(6-CHLORO-2-OXO-2,3-DIHYDRO-1H-INDOL-1-YL)METHYL]-1,3-BENZODIOXOLE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "DVR": (
            "3-(4-CARBAMOYL-1-CARBOXY-2-METHYLSULFONYL-BUTA-1,3-DIENYLAMINO)-INDOLIZINE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "O59": ("OCTANE-1,8-DIYL DISULFAMATE", "non-polymer"),
        "IE2": (
            "4-(7-METHYLPYRAZOLO[3',4':4,5]THIOPYRANO[2,3-B]PYRIDIN-1(4H)-YL)BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "730": (
            "(S)-3-(3-(4-(2-GUANIDINOETHYL)PIPERIDIN-1-YL)-2-(NAPHTHALENE-2-SULFONAMIDO)-3-OXOPROPYL)BENZIMIDAMIDE",
            "non-polymer",
        ),
        "F90": ("4-CHLOROBENZYL CARBAMIMIDOTHIOATE", "non-polymer"),
        "732": (
            "5-(4-CHLOROPHENYL)-N-[(1S)-1-CYCLOHEXYL-2-(METHYLAMINO)-2-OXOETHYL]FURAN-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "735": (
            "2-METHYL-2-(4-{[({4-METHYL-2-[4-(TRIFLUOROMETHYL)PHENYL]-1,3-THIAZOL-5-YL}CARBONYL)AMINO]METHYL}PHENOXY)PROPANOIC ACID",
            "non-polymer",
        ),
        "734": (
            "5-CHLORO-7-[(R)-FURAN-2-YL(PYRIDIN-2-YLAMINO)METHYL]QUINOLIN-8-OL",
            "non-polymer",
        ),
        "737": (
            "N-[4-GUANIDINO-1-(THIAZOLE-2-CARBONYL)-BUTYL]-2-{6-OXO-5-[(QUINOLIN-8-YLMETHYL)-AMINO]-2-M-TOLYL-6H-PYRIMIDIN-1-YL}-ACETAMIDE",
            "non-polymer",
        ),
        "736": (
            "(11S)-8-CHLORO-11-[1-(METHYLSULFONYL)PIPERIDIN-4-YL]-6-PIPERAZIN-1-YL-11H-BENZO[5,6]CYCLOHEPTA[1,2-B]PYRIDINE",
            "non-polymer",
        ),
        "739": (
            "2(S)-{2(S)-[2(R)-AMINO-3-MERCAPTO]PROPYLAMINO-3(S)-METHYL}PENTYLOXY-3-PHENYLPROPIONYLMETHIONINE SULFONE",
            "non-polymer",
        ),
        "738": (
            "N-(3-CYANO-4,5,6,7-TETRAHYDRO-1-BENZOTHIEN-2-YL)-2-FLUOROBENZAMIDE",
            "non-polymer",
        ),
        "F98": ("3-[ETHANOYL(HYDROXY)AMINO]PROPYLPHOSPHONIC ACID", "non-polymer"),
        "TP2": ("N-[TOSYL-D-PROLINYL]AMINO-ETHANETHIOL", "non-polymer"),
        "IEY": (
            "2-((1E)-2-(5-IMIDAZOLYL)ETHENYL)-4-(P-HYDROXYBENZYLIDENE)-5-IMIDAZOLINONE",
            "peptide",
        ),
        "F9A": (
            "N-{3-[(DIMETHYLAMINO)METHYL]PHENYL}-2-(TRIFLUOROMETHYL)BENZAMIDE",
            "non-polymer",
        ),
        "F9F": (
            "2-({[4-(TRIFLUOROMETHOXY)PHENYL]SULFONYL}AMINO)ETHYL DIHYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "PNQ": (
            "P-NITROPHENYL AMINOETHYLDIPHOSPHATE BERYLLIUM TRIFLUORIDE",
            "non-polymer",
        ),
        "OBI": ("1,1'-(OXYDIMETHYLENE)BIS(4-FORMYLPYRIDINIUM)DIOXIME", "non-polymer"),
        "OBF": ("(2S)-2-AMINO-4,4-DIFLUOROBUTANOIC ACID", "peptide"),
        "F9K": ("?", "non-polymer"),
        "OBE": (
            "3-(3-HYDROXY-BENZOYLOXY)-8-METHYL-8-AZA-BICYCLO[3.2.1]OCTANE-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "IET": (
            "1-(4-CYANO-PHENYL)-3-[2-(2,6-DICHLORO-PHENYL)-1-IMINO-ETHYL]-THIOUREA",
            "non-polymer",
        ),
        "OBA": ("2-(OXALYL-AMINO)-BENZOIC ACID", "non-polymer"),
        "73Q": (
            "6-CHLORO-9-HYDROXY-1,3-DIMETHYL-1,9-DIHYDRO-4H-PYRAZOLO[3,4-B]QUINOLIN-4-ONE",
            "non-polymer",
        ),
        "IEM": (
            "5-ACETAMIDO-5,6-DIHYDRO-4-HYDROXY-6-ISOBUTOXY-4H-PYRAN-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "IEL": ("N~6~-[(1Z)-ETHANIMIDOYL]-L-LYSINE", "peptide"),
        "73V": (
            "BENZYL N-({(2S,3S)-3-[(PROPYLAMINO)CARBONYL]OXIRAN-2-YL}CARBONYL)-L-ISOLEUCYL-L-PROLINATE",
            "non-polymer",
        ),
        "OBS": (
            "(Z)-N^6-[(4S,5R)-5-(2-CARBOXYETHYL)-4-(CARBOXYMETHYL)-1-HYDROXYDIHYDRO-2H-THIOPYRANIUM-3(4H)-YLIDENE]-L-LYSINE",
            "peptide",
        ),
        "OBP": ("unknown", "non-polymer"),
        "CDD": (
            "2-(6-CHLORO-3-{[2,2-DIFLUORO-2-(1-OXIDO-2-PYRIDINYL)ETHYL]AMINO}-2-OXO-1(2H)-PYRAZINYL)-N-[(2-FLUOROPHENYL)METHYL]ACETAMIDE",
            "non-polymer",
        ),
        "CDE": ("1,2-DIMETHYL-PROPYLAMINE", "peptide"),
        "CDF": (
            "4-AMINO-1-{5-O-[(R)-HYDROXY(PHOSPHONOOXY)PHOSPHORYL]-ALPHA-D-ARABINOFURANOSYL}PYRIMIDIN-2(1H)-ONE",
            "non-polymer",
        ),
        "CDG": (
            "METHYL 4,6-O-[(1R)-1-CARBOXYETHYLIDENE]-BETA-D-GALACTOPYRANOSIDE",
            "non-polymer",
        ),
        "PLS": (
            "[3-HYDROXY-2-METHYL-5-PHOSPHONOOXYMETHYL-PYRIDIN-4-YLMETHYL]-SERINE",
            "non-polymer",
        ),
        "CDA": (
            "2-(6-CHLORO-3-{[2,2-DIFLUORO-2-(2-PYRIDINYL)ETHYL]AMINO}-2-OXO-1(2H)-PYRAZINYL)-N-[(2-FLUORO-6-PYRIDINYL)METHYL]ACETAMIDE",
            "non-polymer",
        ),
        "CDB": (
            "2-(6-CHLORO-3-{[2,2-DIFLUORO-2-(2-PYRIDINYL)ETHYL]AMINO}-2-OXO-1(2H)-PYRAZINYL)-N-[(2-FLUORO-3-METHYL-6-PYRIDINYL)METHYL]ACETAMIDE",
            "non-polymer",
        ),
        "CDC": (
            "[2-CYTIDYLATE-O'-PHOSPHONYLOXYL]-ETHYL-TRIMETHYL-AMMONIUM",
            "non-polymer",
        ),
        "CDL": ("CARDIOLIPIN", "non-polymer"),
        "CDM": ("4-DIPHOSPHOCYTIDYL-2-C-METHYL-D-ERYTHRITOL", "non-polymer"),
        "CDN": ("CARDIOLIPIN", "non-polymer"),
        "CDO": (
            "2-((R)-1-((S)-2-(N-(6-CARBAMIMIDOYLPYRIDIN-3-YL)METHYLCARBAMOYL)-2H-PYRROL-1(5H)-YL)-3-CYCLOHEXYL-1-OXOPROPAN-2-YLAMINO)ACETIC ACID",
            "non-polymer",
        ),
        "CDH": ("D-(L-A-AMINOADIPOYL)-L-CYSTEINYL-D-ISODEHYDROVALINE", "non-polymer"),
        "CDI": ("2C-METHYL-D-ERYTHRITOL 2,4-CYCLODIPHOSPHATE", "non-polymer"),
        "M97": (
            "1-[(3R,4R)-4-(1H-INDOL-3-YL)-2,5-DIOXOPYRROLIDIN-3-YL]PYRROLO[3,2,1-IJ]QUINOLINIUM",
            "non-polymer",
        ),
        "CDK": (
            "4-[4-(3,4-DIMETHYL-2-OXO-2,3-DIHYDRO-THIAZOL-5-YL)-PYRIMIDIN-2-YLAMINO]-N-(2-METHOXY-ETHYL)-BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "CDT": ("4-METHYLSULFANYL-2-UREIDO-BUTYRIC ACID", "non-polymer"),
        "CDU": ("N-CYCLOHEXYL-N'-DECYLUREA", "non-polymer"),
        "CDV": ("3-METHYL-2-UREIDO-BUTYRIC ACID", "peptide"),
        "CDW": (
            "(4R)-3-(2-DEOXY-5-O-PHOSPHONO-BETA-D-ERYTHRO-PENTOFURANOSYL)-1,3-DIAZABICYCLO[2.2.0]HEX-5-EN-2-ONE",
            "nucleic-acid",
        ),
        "CDP": ("CYTIDINE-5'-DIPHOSPHATE", "non-polymer"),
        "CDQ": (
            "6-[(5-CHLORO-2-OXO-2,3-DIHYDRO-1H-INDOL-1-YL)METHYL]-2,3-DIHYDRO-1,4-BENZODIOXINE-5-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "CDR": ("2,3-DIDEOXYFUCOSE", "saccharide"),
        "CDS": (
            "(2R)-2-[(1R)-1-{[(2Z)-2-(2-AMINO-1,3-THIAZOL-4-YL)-2-(METHOXYIMINO)ACETYL]AMINO}-2-OXOETHYL]-5-[(Z)-2-(4-METHYL-1,3-THIAZOL-5-YL)VINYL]-3,6-DIHYDRO-2H-1,3-THIAZINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "PLO": ("(3BETA)-3-HYDROXYPREGN-5-EN-20-ONE", "non-polymer"),
        "5Z5": (
            "5-[(Z)-(5-CHLORO-2-OXO-1,2-DIHYDRO-3H-INDOL-3-YLIDENE)METHYL]-N-(DIETHYLAMINO)ETHYL]-2,4-DIMETHYL-1H-PYRROLE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "PLM": ("PALMITIC ACID", "non-polymer"),
        "PLL": ("PALLADIUM(II) ALLYL COMPLEX", "non-polymer"),
        "CDX": (
            "(S)-4,4'-(1-METHYL-1,2-ETHANEDIYL)BIS-2,6-PIPERAZINEDIONE",
            "non-polymer",
        ),
        "CDY": ("2-CHLORODIDEOXYADENOSINE", "non-polymer"),
        "CDZ": (
            "3,5-DIHYDROXY-4,6,6-TRIS(3-METHYLBUT-2-EN-1-YL)-2-(2-METHYLPROPANOYL)CYCLOHEXA-2,4-DIEN-1-ONE",
            "non-polymer",
        ),
        "PLH": ("METHYLAMINO-PHENYLALANYL-LEUCYL-HYDROXAMIC ACID", "non-polymer"),
        "BXY": ("ALPHA-D-ARABINOFURANOSE", "saccharide"),
        "BXX": ("BETA-D-ARABINOFURANOSE", "saccharide"),
        "AR2": ("ARGINYL-BENZOTHIAZOLE-6-CARBOXYLIC ACID", "peptide"),
        "BXZ": (
            "4-BROMO-6-(6-HYDROXY-1,2-BENZISOXAZOL-3-YL)BENZENE-1,3-DIOL",
            "non-polymer",
        ),
        "AR4": (
            "2-AZANYL-5-[3-FLUORANYL-5-(HYDROXYMETHYL)-3,4-BIS(OXIDANYL)OXOLAN-2-YL]OXY-5-OXIDANYLIDENE-PENTANOIC ACID",
            "peptide",
        ),
        "RJP": (
            "(1R,5R)-1,5-DIHYDROXY-4-OXO-3-[3-OXO-3-(PHENYLAMINO)PROPYL]CYCLOHEX-2-ENE-1-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "AR6": (
            "[(2R,3S,4R,5R)-5-(6-AMINOPURIN-9-YL)-3,4-DIHYDROXY-OXOLAN-2-YL]METHYL [HYDROXY-[[(2R,3S,4R,5S)-3,4,5-TRIHYDROXYOXOLAN-2-YL]METHOXY]PHOSPHORYL] HYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "AR7": (
            "AMINO{[(4S)-4-AMINO-5,5-DIHYDROXYPENTYL]AMINO}METHANIMINIUM",
            "peptide",
        ),
        "AR8": (
            "N-{[(1S)-5-{4-[25-({2,4-BIS[HYDROXY(OXO)AMMONIO]PHENYL}AMINO)-2,5,8,11,14,17,20,23-OCTAOXAPENTACOS-1-YL]-1H-1,2,3-TRIAZOL-1-YL}-1-CARBOXYPENTYL]CARBAMOYL}-L-GLUTAMIC ACID",
            "non-polymer",
        ),
        "BXP": ("4-O-BETA-D-XYLOPYRANOSYL-BETA-D-XYLOPYRANOSE", "saccharide"),
        "BXR": ("2-[5-CHLORO-2-(1H-TETRAZOL-1-YL)PHENYL]ETHANAMINE", "non-polymer"),
        "3PT": (
            "(2R)-3-{[(S)-HYDROXY{[(1S,2R,3R,4S,5S,6S)-2,3,6-TRIHYDROXY-4,5-BIS(PHOSPHONOOXY)CYCLOHEXYL]OXY}PHOSPHORYL]OXY}PROPANE-1,2-DIYL DIBUTANOATE",
            "non-polymer",
        ),
        "BXW": (
            "3-OXO-3,4-DIHYDRO-2H-1,4-BENZOTHIAZINE-7-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "3PH": ("1,2-DIACYL-GLYCEROL-3-SN-PHOSPHATE", "non-polymer"),
        "3PI": (
            "(1S)-2-(1-HYDROXYBUTOXY)-1-{[(HYDROXY{[(2R,3S,5R,6S)-2,4,6-TRIHYDROXY-3,5-BIS(PHOSPHONOOXY)CYCLOHEXYL]OXY}PHOSPHORYL)OXY]METHYL}ETHYL BUTYRATE",
            "non-polymer",
        ),
        "3PJ": (
            "N-[TRANS-4-(3-AMINO-3-OXOPROPYL)-4-PHENYLCYCLOHEXYL]-N-CYCLOPROPYL-4-[(1S)-2,2,2-TRIFLUORO-1-HYDROXY-1-METHYLETHYL]BENZAMIDE",
            "non-polymer",
        ),
        "3PL": ("3-PHENYLPROPANAL", "non-polymer"),
        "BXL": ("(3E)-N~8~-(2-AMINOPHENYL)-N~1~-PHENYLOCT-3-ENEDIAMIDE", "non-polymer"),
        "BXO": ("(1Z)-BUTANAL OXIME", "non-polymer"),
        "3PO": ("TRIPHOSPHATE", "non-polymer"),
        "BXA": ("BROMOACETIC ACID", "non-polymer"),
        "3PA": ("3-CYCLOPENTYL-PROPIONIC ACID", "non-polymer"),
        "BXC": (
            "(3R,7R,11R)-3,7,11,15-TETRAMETHYLHEXADECYL ALPHA-D-RIBOPYRANOSIDE",
            "non-polymer",
        ),
        "BXB": ("N-(1,3-BENZODIOXOL-5-YLMETHYL)-2,6-DICHLOROBENZAMIDE", "non-polymer"),
        "3PD": (
            "2-AMINO-9-(2-DEOXY-3-O-PHOSPHONOPENTOFURANOSYL)-1,9-DIHYDRO-6H-PURIN-6-ONE",
            "non-polymer",
        ),
        "BXD": (
            "(3S,4S,5R)-3-(4-AMINO-3-{[(2R)-3-ETHOXY-1,1,1-TRIFLUOROPROPAN-2-YL]OXY}-5-FLUOROBENZYL)-5-[(3-TERT-BUTYLBENZYL)AMINO]TETRAHYDRO-2H-THIOPYRAN-4-OL 1,1-DIOXIDE",
            "non-polymer",
        ),
        "3PF": ("3-METHYL-3,4-DIHYDROQUINAZOLIN-2(1H)-ONE", "non-polymer"),
        "BXF": ("2-DEOXY-2-FLUORO-BETA-D-XYLOPYRANOSE", "saccharide"),
        "MT9": (
            "(3R,4S,5S,7R,9E,11S,12R)-12-ETHYL-11-HYDROXY-3,5,7,11-TETRAMETHYL-2,8-DIOXOOXACYCLODODEC-9-EN-4-YL 3,4,6-TRIDEOXY-3-(DIMETHYLAMINO)-BETA-D-XYLO-HEXOPYRANOSIDE",
            "non-polymer",
        ),
        "MT2": ("[(3S)-3-AMINO-3-CARBOXYPROPYL](ETHYL)METHYLSULFONIUM", "peptide"),
        "MT3": (
            "2-BENZYL-5-{4-[(6,7-DIMETHOXYQUINOLIN-4-YL)OXY]-3-FLUOROPHENYL}-3-METHYLPYRIMIDIN-4(3H)-ONE",
            "non-polymer",
        ),
        "MT0": ("1-CHLORO-4-METHYLPHTHALAZINE", "non-polymer"),
        "MT1": (
            "N-(4-{[(2,4-DIAMINOPTERIDIN-1-IUM-6-YL)METHYL](METHYL)AMINO}BENZOYL)-L-GLUTAMIC ACID",
            "non-polymer",
        ),
        "MT7": (
            "ALPHA-D-GLUCOPYRANOSYL-(1->4)-ALPHA-D-GLUCOPYRANOSYL-(1->4)-ALPHA-D-GLUCOPYRANOSYL-(1->4)-ALPHA-D-GLUCOPYRANOSYL-(1->4)-ALPHA-D-GLUCOPYRANOSYL-(1->4)-ALPHA-D-GLUCOPYRANOSYL-(1->4)-BETA-D-GLUCOPYRANOSE",
            "non-polymer",
        ),
        "JY": (
            "3-(2-{4-[2-(5-HYDROXY-2-METHYLENE-CYCLOHEXYLIDENE)-ETHYLIDENE]-7A-METHYL-OCTAHYDRO-INDEN-1-YL}-PROPYL)-PHENOL",
            "non-polymer",
        ),
        "MT5": (
            "N-[(3E,5R,6R,7S,8R,8AR)-5,6,7,8-TETRAHYDROXYHEXAHYDROIMIDAZO[1,5-A]PYRIDIN-3(2H)-YLIDENE]OCTAN-1-AMINIUM",
            "non-polymer",
        ),
        "D04": (
            "4-(4-CYCLOHEXYL-1H-1,2,3-TRIAZOL-1-YL)-2,3,5,6-TETRAFLUOROBENZENESULFONAMIDE",
            "non-polymer",
        ),
        "TPM": ("2-(4-AMINOBENZYLAMINO)-3,4,5,6-TETRAHYDROPYRIDINIUM", "non-polymer"),
        "MTZ": (
            "4-METHYL-5-{(2E)-2-[(4-MORPHOLIN-4-YLPHENYL)IMINO]-2,5-DIHYDROPYRIMIDIN-4-YL}-1,3-THIAZOL-2-AMINE",
            "non-polymer",
        ),
        "DW2": ("RU-PYRIDOCARBAZOLE-2", "non-polymer"),
        "MTX": ("METHOTREXATE", "non-polymer"),
        "MTY": ("META-TYROSINE", "peptide"),
        "LHG": ("1,2-DIPALMITOYL-PHOSPHATIDYL-GLYCEROLE", "non-polymer"),
        "MTS": (
            "(4S-TRANS)-4-(METHYLAMINO)-5,6-DIHYDRO-6-METHYL-4H-THIENO(2,3-B)THIOPYRAN-2-SULFONAMIDE-7,7-DIOXIDE",
            "non-polymer",
        ),
        "MTP": (
            "2-HYDROXYMETHYL-5-(6-METHYLSULFANYL-PURIN-9-YL)-TETRAHYDRO-FURAN-3,4-DIOL",
            "non-polymer",
        ),
        "MTQ": ("(MOLYBDOPTERIN-S,S)-DIOXO-THIO-MOLYBDENUM(VI)", "non-polymer"),
        "MTV": ("(MOLYBDOPTERIN-S,S)-DIOXO-THIO-MOLYBDENUM(IV)", "non-polymer"),
        "MTW": (
            "HYDROXY(OXO)(3-{[(2Z)-4-[3-(1H-1,2,4-TRIAZOL-1-YLMETHYL)PHENYL]PYRIMIDIN-2(5H)-YLIDENE]AMINO}PHENYL)AMMONIUM",
            "non-polymer",
        ),
        "MTT": ("MALTOTETRAOSE", "saccharide"),
        "MTU": ("9-BETA-D-RIBOFURANOSYL-9H-PURIN-2-AMINE", "nucleic-acid"),
        "MTJ": ("N,O-DIMETHYL-L-TYROSINE", "non-polymer"),
        "MTK": ("MONTELUKAST", "non-polymer"),
        "MTH": (
            "2-(4-AMINO-PYRROLO[2,3-D]PYRIMIDIN-7-YL)-5-METHYLSULFANYLMETHYL-TETRAHYDRO-FURAN-3,4-DIOL",
            "non-polymer",
        ),
        "MTI": (
            "3,4-DIHYDROXY-2-[(METHYLSULFANYL)METHYL]-5-(4-OXO-4,5-DIHYDRO-3H-PYRROLO[3,2-D]PYRIMIDIN-7-YL)PYRROLIDINIUM",
            "non-polymer",
        ),
        "MTN": (
            "S-[(1-OXYL-2,2,5,5-TETRAMETHYL-2,5-DIHYDRO-1H-PYRROL-3-YL)METHYL] METHANESULFONOTHIOATE",
            "non-polymer",
        ),
        "MTO": ("BOUND WATER", "non-polymer"),
        "MTL": ("D-MANNITOL", "non-polymer"),
        "MTM": (
            "(3S,4R)-2-(4-AMINO-5H-PYRROLO[3,2-D]PYRIMIDIN-7-YL)-5-[(METHYLSULFANYL)METHYL]PYRROLIDINE-3,4-DIOL",
            "non-polymer",
        ),
        "MTB": ("?", "non-polymer"),
        "MTC": (
            "[1,8-DIHYDROXY-11-OXO-13-(2-METHYLTRITHIO-ETHYLIDENE)-BICYCLO[7.3.1]TRIDECA-4,9-DIENE-2,6-DIYN-10-YL]-CARBAMIC ACID METHYL ESTER",
            "non-polymer",
        ),
        "MTA": ("5'-DEOXY-5'-METHYLTHIOADENOSINE", "non-polymer"),
        "MTF": ("N-METHYLTHIOFORMAMIDE", "non-polymer"),
        "MTG": ("[METHYLTHIO]ACETATE", "non-polymer"),
        "MTD": ("[METHYLTELLURO]ACETATE", "non-polymer"),
        "MTE": (
            "PHOSPHONIC ACIDMONO-(2-AMINO-5,6-DIMERCAPTO-4-OXO-3,7,8A,9,10,10A-HEXAHYDRO-4H-8-OXA-1,3,9,10-TETRAAZA-ANTHRACEN-7-YLMETHYL)ESTER",
            "non-polymer",
        ),
        "ARP": ("9-HYDROXYPROPYLADENINE, R-ISOMER", "non-polymer"),
        "ARQ": (
            "BENZOYLAMINO-BENZYL-METHYL-[2-HYDROXY-3-[1-METHYL-ETHYL-OXY-N-FORMAMIDYL]-4-PHENYL-BUTYL]-AMMONIUM",
            "non-polymer",
        ),
        "ARR": (
            "N-(4-{2-[(3-CHLOROBENZYL)AMINO]ETHYL}PHENYL)THIOPHENE-2-CARBOXIMIDAMIDE",
            "non-polymer",
        ),
        "ARS": ("ARSENIC", "non-polymer"),
        "ART": ("ARSENATE", "non-polymer"),
        "ARU": (
            "5'-O-[(R)-{[(R)-[DIBROMO(PHOSPHONO)METHYL](HYDROXY)PHOSPHORYL]OXY}(HYDROXY)PHOSPHORYL]-N,N-DIETHYLADENOSINE",
            "non-polymer",
        ),
        "ARV": ("5-N-ALLYL-ARGININE", "peptide"),
        "ARW": ("METHYL BETA-D-ARABINOPYRANOSIDE", "non-polymer"),
        "BX1": (
            "1-[3-AMINO-6-(3,4,5-TRIMETHOXYPHENYL)PYRAZIN-2-YL]PIPERIDINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "3P1": ("8-[(QUINOLIN-2-YLMETHYL)AMINO]ADENOSINE", "non-polymer"),
        "BX3": (
            "(+)-2-[4-[(-1-ACETIMIDOYL-4-PIPERIDINYL)OXY]-3-(7-AMIDINO-2-NAPHTHYL)PROPIONIC ACID",
            "non-polymer",
        ),
        "BX2": (
            "(5S)-2-AMINO-5-(2,6-DIETHYLPYRIDIN-4-YL)-3-METHYL-5-(3-PYRIMIDIN-5-YLPHENYL)-3,5-DIHYDRO-4H-IMIDAZOL-4-ONE",
            "non-polymer",
        ),
        "3P4": (
            "5-[2-(BUT-3-EN-1-YLOXY)-5-(1-HYDROXYVINYL)PYRIDIN-3-YL]-3-ETHYL-2-(1-ETHYLAZETIDIN-3-YL)-1,2,6,7A-TETRAHYDRO-7H-PYRAZOLO[4,3-D]PYRIMIDIN-7-ONE",
            "non-polymer",
        ),
        "3P5": (
            "(3S,4S,5R)-3-(3-BROMO-4-HYDROXYBENZYL)-5-[(3-CYCLOPROPYLBENZYL)AMINO]TETRAHYDRO-2H-THIOPYRAN-4-OL 1,1-DIOXIDE",
            "non-polymer",
        ),
        "BX7": (
            "N-(3-{[5-IODO-4-({3-[(THIOPHEN-2-YLCARBONYL)AMINO]PROPYL}AMINO)PYRIMIDIN-2-YL]AMINO}PHENYL)PYRROLIDINE-1-CARBOXAMIDE",
            "non-polymer",
        ),
        "ARA": ("ALPHA-L-ARABINOSE", "saccharide"),
        "ARB": ("BETA-L-ARABINOSE", "saccharide"),
        "ARC": ("3,7,11,15-TETRAMETHYL-HEXADECAN-1-OL", "non-polymer"),
        "ARD": ("C15-(R)-METHYLTHIENYL RAPAMYCIN", "non-polymer"),
        "ARE": ("ACARBOSE DERIVED PENTASACCHARIDE", "saccharide"),
        "ARF": ("FORMAMIDE", "peptide"),
        "ARG": ("ARGININE", "peptide"),
        "ARH": (
            "3-(HYDROXYMETHYL)-1-METHYL-5-(2-METHYLAZIRIDIN-1-YL)-2-PHENYL-1H-INDOLE-4,7-DIONE",
            "non-polymer",
        ),
        "ARI": ("[O4]-ACETOXY-2,3-DIDEOXYFUCOSE", "saccharide"),
        "ARJ": (
            "(2S,3R,5R)-3-(6-AMINO-9H-PURIN-9-YL)-2-HYDROXY-5-(HYDROXYMETHYL)CYCLOPENTANONE",
            "non-polymer",
        ),
        "ARK": (
            "N-({(1S)-5-[4-({2-[2-({2,4-BIS[HYDROXY(OXO)AMMONIO]PHENYL}AMINO)ETHOXY]ETHOXY}METHYL)-1H-1,2,3-TRIAZOL-1-YL]-1-CARBOXYPENTYL}CARBAMOYL)-L-GLUTAMIC ACID",
            "non-polymer",
        ),
        "ARL": (
            "7-(3,5-DITERT-BUTYLPHENYL)-3-METHYLOCTA-2,4,6-TRIENOIC ACID",
            "non-polymer",
        ),
        "ARM": ("DEOXY-METHYL-ARGININE", "peptide"),
        "ARN": ("1-IMINO-5-PENTANONE", "non-polymer"),
        "ARO": ("C-GAMMA-HYDROXY ARGININE", "peptide"),
        "0PC": (
            "NALPHA-[(BENZYLOXY)CARBONYL]-N-[(1S)-3-CHLORO-1-METHYL-2-OXOPROPYL]-L-PHENYLALANINAMIDE",
            "peptide",
        ),
        "0PA": ("CYCLOPENTYLACETIC ACID", "non-polymer"),
        "0PF": (
            "1-ETHYL-8-(4-ETHYLPHENYL)-5-METHYL-1,5-DIHYDROPYRAZOLO[4,3-C][2,1]BENZOTHIAZINE 4,4-DIOXIDE",
            "non-polymer",
        ),
        "CUZ": ("(MU-4-SULFIDO)-TETRA-NUCLEAR COPPER ION", "non-polymer"),
        "0PK": (
            "N-[(S)-[(1R)-1-{[(BENZYLOXY)CARBONYL]AMINO}-2-PHENYLETHYL](HYDROXY)PHOSPHORYL]-L-LEUCYL-L-ALANINE",
            "peptide",
        ),
        "0PJ": (
            "N-[(S)-({[(BENZYLOXY)CARBONYL]AMINO}METHYL)(HYDROXY)PHOSPHORYL]-L-LEUCYL-L-LEUCINE",
            "peptide",
        ),
        "0PI": (
            "N-[(2R,4S)-4-HYDROXY-2-(2-METHYLPROPYL)-4-OXIDO-7-OXO-9-PHENYL-3,8-DIOXA-6-AZA-4-PHOSPHANONAN-1-OYL]-L-LEUCINE",
            "peptide",
        ),
        "TPP": ("THIAMINE DIPHOSPHATE", "non-polymer"),
        "0PO": (
            "N-[(1R)-1-{[(1S,2S)-1-BENZYL-3-{(2R,4S)-2-(TERT-BUTYLCARBAMOYL)-4-[(PYRIDIN-3-YLMETHYL)SULFANYL]PIPERIDIN-1-YL}-2-HYDROXYPROPYL]CARBAMOYL}-2-METHYLPROPYL]QUINOLINE-2-CARBOXAMIDE",
            "peptide",
        ),
        "D01": (
            "2-({4-[(5-CHLORO-1H-INDOL-2-YL)SULFONYL]PIPERAZIN-1-YL}CARBONYL)THIENO[3,2-B]PYRIDINE 4-OXIDE",
            "non-polymer",
        ),
        "0PQ": (
            "N-{(2R)-3-[(S)-[(1R)-1-AMINO-2-PHENYLETHYL](HYDROXY)PHOSPHORYL]-2-BENZYLPROPANOYL}-L-PHENYLALANINE",
            "peptide",
        ),
        "0PP": (
            "(2S)-2-TERT-BUTYL-N~4~-(1-ETHYLPROPYL)-N~1~-[(2R,3S)-2-HYDROXY-4-PHENYL-3-{[N-(QUINOLIN-2-YLCARBONYL)-L-THREONYL]AMINO}BUTYL]BUTANEDIAMIDE",
            "peptide",
        ),
        "0PY": ("PYRIDINE", "non-polymer"),
        "0PX": (
            "N~1~-{(1S,2S,4S)-1-(CYCLOHEXYLMETHYL)-4-[(2,2-DIMETHYLPROPYL)CARBAMOYL]-2-HYDROXY-5-METHYLHEXYL}-N~2~-(QUINOLIN-2-YLCARBONYL)-L-ASPARTAMIDE",
            "peptide",
        ),
        "VIV": (
            "(2R)-2,5,7,8-TETRAMETHYL-2-[(4R,8R)-4,8,12-TRIMETHYLTRIDECYL]CHROMAN-6-OL",
            "non-polymer",
        ),
        "B58": ("1H-IMIDAZO[4,5-D]PYRIDAZINE-4,7-DIAMINE", "non-polymer"),
        "B59": (
            "1,3-DIMETHYL-2,4,7-TRIOXO-1,2,3,4,7,8-HEXAHYDROPTERIDINE-6-CARBALDEHYDE",
            "non-polymer",
        ),
        "B52": (
            "(2R)-2-(7-AMINO-1-METHYL-4,5-DIOXO-1,4,5,6-TETRAHYDROPYRIMIDO[4,5-C]PYRIDAZIN-3-YL)PROPANOIC ACID",
            "non-polymer",
        ),
        "B53": ("2,6-DIAMINO-5-NITROSOPYRIMIDIN-4(3H)-ONE", "non-polymer"),
        "B50": (
            "4-(4-CHLOROPHENYL)-1-[(3S)-3,4-DIHYDROXYBUTYL]-N-[3-(4-METHYLPIPERAZIN-1-YL)PROPYL]-3-PHENYL-1H-PYRROLE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "B51": ("NICKEL/IRON/SULFUR CLUSTER", "non-polymer"),
        "B56": (
            "2-AMINO-8-METHYL-4-OXO-3,4,7,8-TETRAHYDROPTERIDINE-6-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "B57": ("2,6-DIAMINO-5-NITROPYRIMIDIN-4(3H)-ONE", "non-polymer"),
        "B54": (
            "4-{3-[(2-AMINO-5-NITROSO-6-OXO-1,6-DIHYDROPYRIMIDIN-4-YL)AMINO]PROPOXY}BENZOIC ACID",
            "non-polymer",
        ),
        "B55": ("2-AMINO-8-SULFANYL-1,9-DIHYDRO-6H-PURIN-6-ONE", "non-polymer"),
        "PHW": (
            "N-[(BENZYLOXY)CARBONYL]-BETA-PHENYL-D-PHENYLALANYL-N-{(1S,3E)-1-[DIHYDROXY(DIPHENOXY)-LAMBDA~5~-PHOSPHANYL]-4-METHOXYBUT-3-EN-1-YL}-L-PROLINAMIDE",
            "peptide",
        ),
        "X3J": (
            "(2S)-2-{2-[(2-AMINOETHYL)AMINO]-2-OXOETHYL}-2-HYDROXYBUTANEDIOIC ACID",
            "non-polymer",
        ),
        "TPX": (
            "(2S,3S)-3-FORMYL-2-({[(4-METHYLPHENYL)SULFONYL]AMINO}METHYL)PENTANOIC ACID",
            "non-polymer",
        ),
        "WZC": (
            "(13ALPHA,14BETA,17ALPHA)-3-HYDROXY-2-METHOXYESTRA-1,3,5(10)-TRIEN-17-YL SULFAMATE",
            "non-polymer",
        ),
        "LHC": (
            "(2S)-2-AMINO-4-(4-AMINO-2-OXOPYRIMIDIN-1(2H)-YL)BUTANOIC ACID",
            "peptide",
        ),
        "X36": (
            "[4-AMINO-2-(PHENYLAMINO)-1,3-THIAZOL-5-YL](PHENYL)METHANONE",
            "non-polymer",
        ),
        "X37": (
            "4-(4-TERT-BUTYLBENZYL)-1-(7H-PYRROLO[2,3-D]PYRIMIDIN-4-YL)PIPERIDIN-4-AMINIUM",
            "non-polymer",
        ),
        "X35": (
            "[4-AMINO-2-(PROP-2-EN-1-YLAMINO)-1,3-THIAZOL-5-YL](PYRIDIN-3-YL)METHANONE",
            "non-polymer",
        ),
        "B5R": (
            "(2S)-3-(4-CHLORO-3-FLUOROPHENOXY)-N-[4-CYANO-3-(TRIFLUOROMETHYL)PHENYL]-2-HYDROXY-2-METHYLPROPANAMIDE",
            "non-polymer",
        ),
        "B5P": (
            "N-{3-[5-HYDROXY-8-(3-METHYLBUTYL)-7-OXO-7,8-DIHYDROIMIDAZO[1,2-A]PYRIMIDIN-6-YL]-1,1-DIOXIDO-4H-1,2,4-BENZOTHIADIAZIN-7-YL}METHANESULFONAMIDE",
            "non-polymer",
        ),
        "X39": (
            "4-AMINO-N-(4-CHLOROBENZYL)-1-(7H-PYRROLO[2,3-D]PYRIMIDIN-4-YL)PIPERIDINE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "PCR": ("P-CRESOL", "non-polymer"),
        "B5U": (
            "6-(3,4-DIMETHYLPHENYL)-3-[[4-[3-(4-METHYLPIPERAZIN-1-YL)PROPOXY]PHENYL]AMINO]PYRAZINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "T": ("THYMIDINE-5'-MONOPHOSPHATE", "nucleic-acid"),
        "B5A": (
            "1-(PIPERIDIN-4-YLMETHYL)-1H-PYRAZOLO[3,4-D]PYRIMIDIN-4-AMINE",
            "non-polymer",
        ),
        "B5F": ("?", "non-polymer"),
        "B5D": (
            "(3S)-7-CHLORO-3-METHYL-3,4-DIHYDRO-2H-1,2,4-BENZOTHIADIAZINE 1,1-DIOXIDE",
            "non-polymer",
        ),
        "H0K": (
            "4'-[7-(HYDROXYMETHYL)-2,4-DIHYDROINDENO[1,2-C]PYRAZOL-3-YL]BIPHENYL-4-OL",
            "non-polymer",
        ),
        "0P2": (
            "N~2~-[(BENZYLOXY)CARBONYL]-N-[(1R,2S)-1-(DIHYDROXYBORANYL)-2-METHYLBUTYL]-L-ALANINAMIDE",
            "peptide",
        ),
        "0P1": (
            "N-[(BENZYLOXY)CARBONYL]-L-ALANYL-N-{(1S)-1-[(R)-[(1R)-1-BENZYL-2-METHOXY-2-OXOETHOXY](HYDROXY)PHOSPHORYL]-3-METHYLBUTYL}-L-ALANINAMIDE",
            "peptide",
        ),
        "4CM": ("(4-CARBAMIMIDOYLPHENYL)-METHYL-PHOSPHINIC ACID", "non-polymer"),
        "TEP": ("THEOPHYLLINE", "non-polymer"),
        "0P6": ("2-AMINOQUINOLIN-8-OL", "non-polymer"),
        "0P5": (
            "4-{4-[(1H-BENZIMIDAZOL-2-YLMETHYL)AMINO]-6-(2-CHLORO-4-METHOXYPHENOXY)PYRIMIDIN-2-YL}PIPERAZIN-2-ONE",
            "non-polymer",
        ),
        "0P4": (
            "N-(1H-BENZIMIDAZOL-2-YL)-N'-(3,5-DICHLOROBENZYL)PROPANE-1,3-DIAMINE",
            "non-polymer",
        ),
        "WZA": (
            "(9BETA,14BETA,17BETA)-17-HYDROXY-2-METHOXYESTRA-1,3,5(10)-TRIEN-3-YL SULFAMATE",
            "non-polymer",
        ),
        "0P8": (
            "2-({3-[(3,5-DIBROMO-2-METHOXYBENZYL)AMINO]PROPYL}AMINO)QUINOLIN-4(1H)-ONE",
            "non-polymer",
        ),
        "LHA": ("DOUBLY FUNCTIONALIZED PAROMOMYCIN PM-II-162", "non-polymer"),
        "KVY": (
            "4-AMINO-1-METHYL-N-{1-METHYL-5-[(1-METHYL-5-{[3-(METHYLAMINO)-3-OXOPROPYL]CARBAMOYL}-1H-PYRROL-3-YL)CARBAMOYL]-1H-PYRROL-3-YL}-1H-PYRROLE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "4I4": (
            "(5R,6R,8S)-8-(3-{[AMINO(IMINO)METHYL]AMINO}PHENYL)-5-CYCLOHEXYL-6-HYDROXY-3-OXO-1-PHENYL-2,7-DIOXA-4-AZA-6-PHOSPHANONAN-9-OIC ACID 6-OXIDE",
            "non-polymer",
        ),
        "4I5": (
            "(6S)-2-CHLORO-5,6,7,8,9,10-HEXAHYDROCYCLOHEPTA[B]INDOLE-6-CARBOXAMIDE",
            "non-polymer",
        ),
        "115": (
            "7-[3-(4-FLUORO-PHENYL)-1-ISOPROPYL-1H-INDOL-2-YL]-3,5-DIHYDROXY-HEPTANOIC ACID",
            "non-polymer",
        ),
        "114": ("COMPACTIN", "non-polymer"),
        "117": (
            "7-[2-(4-FLUORO-PHENYL)-5-ISOPROPYL-3-PHENYL-4-PHENYLCARBAMOYL-PYRROL-1-YL]- 3,5-DIHYDROXY-HEPTANOIC ACID",
            "non-polymer",
        ),
        "116": (
            "7-[4-(4-FLUORO-PHENYL)-5-HYDROXYMETHYL-2,6-DIISOPROPYL-PYRIDIN-3-YL]-3,5-DIHYDROXY-HEPTANOIC ACID",
            "non-polymer",
        ),
        "111": (
            "(1N)-4-N-BUTOXYPHENYLSULFONYL-(2R)-N-HYDROXYCARBOXAMIDO-(4S)-METHANESULFONYLAMINO-PYRROLIDINE",
            "non-polymer",
        ),
        "110": (
            "3-(4-AMINO-CYCLOHEXYL)-2-HYDROXY-3-[(4-OXO-2-PHENYLMETHANESULFONYL-1,2,3,4-TETRAHYDRO-PYRROLO[1,2-A]PYRAZINE-6-CARBONYL)-AMINO]-PROPIONIC ACID BUTYL ESTER",
            "non-polymer",
        ),
        "113": (
            "7,8-DIHYDROXY-1-METHOXY-3-METHYL-10-OXO-4,10-DIHYDRO-1H,3H-PYRANO[4,3-B]CHROMENE-9-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "112": (
            "THIOPHOSPHORIC ACID O-((ADENOSYL-PHOSPHO)PHOSPHO)-S-ACETAMIDYL-DIESTER",
            "non-polymer",
        ),
        "119": ("TETRAPHENYLPHOSPHONIUM ION", "non-polymer"),
        "118": ("TETRAPHENYLANTIMONIUM ION", "non-polymer"),
        "CR": ("CHROMIUM ION", "non-polymer"),
        "YC2": (
            "N-{[(1S)-1-CARBOXY-5-{[(4-IODOPHENYL)CARBONYL]AMINO}PENTYL]CARBAMOYL}-L-GLUTAMIC ACID",
            "non-polymer",
        ),
        "ZGA": ("ZARAGOZIC ACID A", "non-polymer"),
        "MUS": (
            "4-METHYL-2-OXO-2H-CHROMEN-7-YL 5-(ACETYLAMINO)-3,5-DIDEOXY-L-ERYTHRO-NON-2-ULOPYRANOSIDONIC ACID",
            "non-polymer",
        ),
        "KDM": (
            "(2R,4S,5R,6R)-2,4,5-TRIHYDROXY-6-[(1R,2R)-1,2,3-TRIHYDROXYPROPYL]OXANE-2-CARBOXYLIC-ACID",
            "saccharide",
        ),
        "ZGB": ("3-(4-METHOXY-3-METHYLPHENYL)PROPANOIC ACID", "non-polymer"),
        "IH4": ("N-(R-CARBOXY-ETHYL)-ALPHA-(S)-(2-PHENYLETHYL)", "non-polymer"),
        "YCP": ("(2S)-PIPERIDINE-2-CARBOXYLIC ACID", "peptide"),
        "KDN": (
            "3-DEOXY-D-GLYCERO-BETA-D-GALACTO-NON-2-ULOPYRANOSONIC ACID",
            "non-polymer",
        ),
        "YCU": (
            "1-[5-(2-METHOXYETHYL)-4-METHYL-THIOPHEN-2-YL]SULFONYL-3-[4-METHOXY-6-(METHYLCARBAMOYLAMINO)PYRIDIN-2-YL]UREA",
            "non-polymer",
        ),
        "KDO": ("3-DEOXY-D-MANNO-OCT-2-ULOSONIC ACID", "saccharide"),
        "YCO": (
            "3-(2-DEOXY-5-O-PHOSPHONO-BETA-D-ERYTHRO-PENTOFURANOSYL)-1H-PYRIMIDO[5,4-B][1,4]BENZOXAZIN-2(3H)-ONE",
            "nucleic-acid",
        ),
        "YCN": ("1,4,7,10-TETRAAZACYCLODODECANE", "non-polymer"),
        "YCM": ("S-(2-AMINO-2-OXOETHYL)-L-CYSTEINE", "peptide"),
        "YCL": (
            "2-{[(2S)-1-AMINO-5-{[(1Z)-2-CHLOROETHANIMIDOYL]AMINO}-1-OXOPENTAN-2-YL]CARBAMOYL}BENZOIC ACID",
            "non-polymer",
        ),
        "KDH": (
            "(2R,3R)-5,7-DIHYDROXY-2-(3,4,5-TRIHYDROXYPHENYL)-3,4-DIHYDRO-2H-CHROMEN-3-YL 3,4,5-TRIHYDROXYBENZOATE",
            "non-polymer",
        ),
        "YCF": (
            "(2E)-N-HYDROXY-2-[1-(4-{[(4-{(1E)-1-[2-(N'-HYDROXYCARBAMIMIDOYL)HYDRAZINYLIDENE]ETHYL}PHENYL)CARBAMOYL]AMINO}PHENYL)ETHYLIDENE]HYDRAZINECARBOXIMIDAMIDE",
            "non-polymer",
        ),
        "KDI": (
            "3-AMINO-N-(3-METHOXYPROPYL)-6-{4-[(4-METHYLPIPERAZIN-1-YL)SULFONYL]PHENYL}PYRAZINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "11D": (
            "11-({[5-(DIMETHYLAMINO)NAPHTHALEN-1-YL]SULFONYL}AMINO)UNDECANOIC ACID",
            "non-polymer",
        ),
        "11F": (
            "5-[(3S)-3-(2-METHOXYBIPHENYL-4-YL)BUT-1-YN-1-YL]-6-METHYLPYRIMIDINE-2,4-DIAMINE",
            "non-polymer",
        ),
        "11A": ("UNDECANOIC ACID", "non-polymer"),
        "QEG": (
            "N~2~-{3-[4-(4-PHENYLTHIOPHEN-2-YL)PHENYL]PROPANOYL}-L-GLUTAMINYL-L-ALPHA-GLUTAMINE",
            "non-polymer",
        ),
        "11C": ("(2R)-2-AMINOHEXANEDIOIC ACID", "non-polymer"),
        "11B": (
            "(5R,6S,8S)-8-(3-{[AMINO(IMINO)METHYL]AMINO}PHENYL)-6-HYDROXY-3-OXO-5-PENTYL-1-PHENYL-2,7-DIOXA-4-AZA-6-PHOSPHANONAN-9-OIC ACID 6-OXIDE",
            "non-polymer",
        ),
        "11M": (
            "1-(1'-{[3-(METHYLSULFANYL)-2-BENZOTHIOPHEN-1-YL]CARBONYL}SPIRO[1-BENZOFURAN-3,4'-PIPERIDIN]-5-YL)METHANAMINE",
            "non-polymer",
        ),
        "11O": ("(TRANS-12,13-EPOXY)-11-HYDROXY-9(Z)-OCTADECENOIC ACID", "non-polymer"),
        "11N": (
            "1-[1'-(3-PHENYLACRYLOYL)SPIRO[1-BENZOFURAN-3,4'-PIPERIDIN]-5-YL]METHANAMINE",
            "non-polymer",
        ),
        "4IP": ("INOSITOL-(1,3,4,5)-TETRAKISPHOSPHATE", "non-polymer"),
        "11H": ("9,10-DIOXO-9,10-DIHYDROANTHRACENE-2,6-DISULFONIC ACID", "non-polymer"),
        "4IR": (
            "{N-(4-{[2-(AMINO-KAPPAN)ETHYL]SULFAMOYL-KAPPAN}PHENYL)-5-[(3AS,4S,6AR)-2-OXOHEXAHYDRO-1H-THIENO[3,4-D]IMIDAZOL-4-YL]PENTANAMIDE}(CHLORO)[(1,2,3,4,5-ETA)-1,2,3,4,5-PENTAMETHYLCYCLOPENTADIENYL]IRIDIUM(III)",
            "non-polymer",
        ),
        "11J": ("N-BIPHENYL-3-YL-2-IODOBENZAMIDE", "non-polymer"),
        "11U": (
            "(S)-N-(4-CARBAMIMIDOYLBENZYL)-1-(2-(CYCLOHEXYLAMINO)ETHANOYL)PYRROLIDINE-2-CARBOXAMIDE",
            "peptide",
        ),
        "QEI": (
            "2-AMINO-5-({[(1S,4S,5R)-4,5-DIHYDROXYCYCLOPENT-2-EN-1-YL]AMINO}METHYL)-3,7-DIHYDRO-4H-PYRROLO[2,3-D]PYRIMIDIN-4-ONE",
            "non-polymer",
        ),
        "4IN": ("4-AMINO-L-TRYPTOPHAN", "peptide"),
        "M8M": (
            "5'-{[2-(AMINOOXY)ETHYL](METHYL)AMINO}-5'-DEOXY-8-METHYLADENOSINE",
            "non-polymer",
        ),
        "11Q": ("1-(CYCLOHEXYLMETHYL)-L-PROLINE", "peptide"),
        "11P": (
            "[(7R)-6,7-DIHYDRO-5H-CYCLOPENTA[C]PYRIDIN-7-YL(HYDROXY)METHYLENE]BIS(PHOSPHONIC ACID)",
            "non-polymer",
        ),
        "11S": ("5-CHLORO-1H-INDOLE", "non-polymer"),
        "4IK": (
            "(2S,5S)-2,6-DIAMINO-5-{[(2R,3S,4R,5R)-5-(6-AMINO-9H-PURIN-9-YL)-3,4-DIHYDROXYTETRAHYDROFURAN-2-YL]METHYL}HEXANOIC ACID",
            "peptide",
        ),
        "4IG": (
            "(2S)-6-(2,4-DIAMINO-6-ETHYLPYRIMIDIN-5-YL)-2-(3,5-DIFLUOROPHENYL)-4-(3-METHOXYPROPYL)-2H-1,4-BENZOXAZIN-3(4H)-ONE",
            "non-polymer",
        ),
        "11X": ("N-(PYRIDIN-3-YLMETHYL)ANILINE", "non-polymer"),
        "UZ9": (
            "(2S)-2-AMINO-6-({(4R)-4-[(10R,13S)-10,13-DIMETHYL-3-OXOHEXADECAHYDRO-1H-CYCLOPENTA[A]PHENANTHREN-17-YL]PENTANOYL}AMINO)HEXANOIC ACID",
            "non-polymer",
        ),
        "2OH": ("4,4'-PROPANE-2,2-DIYLDIPHENOL", "non-polymer"),
        "QEJ": ("(2S)-PYRROLIDINE-2-CARBONITRILE", "non-polymer"),
        "N7P": ("1-ACETYL-L-PROLINE", "peptide"),
        "2OM": ("?", "nucleic-acid"),
        "N7R": (
            "N-[(3R,6S)-6-(2,3-DIFLUOROPHENYL)-2-OXO-1-(2,2,2-TRIFLUOROETHYL)AZEPAN-3-YL]-4-(2-OXO-2,3-DIHYDRO-1H-IMIDAZO[4,5-B]PYRIDIN-1-YL)PIPERIDINE-1-CARBOXAMIDE",
            "non-polymer",
        ),
        "2OA": ("2'DEOXYADENOSINE-5'-TRIPHOSPHATE", "non-polymer"),
        "2OB": ("CHOLESTERYL OLEATE", "non-polymer"),
        "2OF": ("FERROUS ION, 2 WATERS COORDINATED", "non-polymer"),
        "2OG": ("2-OXO-GLUTARIC ACID", "non-polymer"),
        "N7F": (
            "6-[(3S)-3-AMINOPIPERIDIN-1-YL]-5-BENZYL-4-OXO-3-(QUINOLIN-4-YLMETHYL)-4,5-DIHYDRO-3H-PYRROLO[3,2-D]PYRIMIDINE-7-CARBONITRILE",
            "non-polymer",
        ),
        "VLB": ("(2ALPHA,2'BETA,3BETA,4ALPHA,5BETA)-VINCALEUKOBLASTINE", "non-polymer"),
        "2OP": ("(2S)-2-HYDROXYPROPANOIC ACID", "non-polymer"),
        "MBQ": (
            "2-HYDROXY-5-({1-[(4-METHYLPHENOXY)METHYL]-3-OXOPROP-1-ENYL}AMINO)-L-TYROSINE",
            "peptide",
        ),
        "IMD": ("IMIDAZOLE", "non-polymer"),
        "2OS": ("3-N-OCTANOYLSUCROSE", "non-polymer"),
        "2OT": ("?", "nucleic-acid"),
        "N7I": ("4-[(1E)-3-HYDROXYPROP-1-EN-1-YL]-2-METHOXYPHENOL", "non-polymer"),
        "VTP": (
            "(2Z,6Z,10Z,14Z,18Z,22E,26E)-3,7,11,15,19,23,27,31-OCTAMETHYLDOTRIACONTA-2,6,10,14,18,22,26,30-OCTAEN-1-YL TRIHYDROGEN DIPHOSPHATE",
            "non-polymer",
        ),
        "PZR": ("1H-PYRAZOLO[4,3-C]PYRIDINE", "non-polymer"),
        "ZZZ": ("6-FORMYLTETRAHYDROPTERIN", "non-polymer"),
        "ZZY": (
            "1-[(2-NITROPHENYL)SULFONYL]-1H-PYRROLO[3,2-B]PYRIDINE-6-CARBOXAMIDE",
            "non-polymer",
        ),
        "ZZX": (
            "(6S)-2-(3-CHLORO-4-FLUOROBENZYL)-8-ETHYL-10-HYDROXY-N,6-DIMETHYL-1,9-DIOXO-1,2,6,7,8,9-HEXAHYDROPYRAZINO[1',2':1,5]PYRROLO[2,3-D]PYRIDAZINE-4-CARBOXAMIDE",
            "non-polymer",
        ),
        "ZZW": (
            "9-(4-FLUOROBENZYL)-N-HYDROXY-9H-BETA-CARBOLINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "ZZV": (
            "N-[7-(4-FLUOROBENZYL)-9-HYDROXY-8-OXO-7,8-DIHYDRO-6H-PYRROLO[3,4-G]QUINOLIN-5-YL]-N-METHYLMETHANESULFONAMIDE",
            "non-polymer",
        ),
        "ZZU": ("(2S,3S)-HYDROXYARGININE", "peptide"),
        "ZZT": ("2-METHOXY-5-METHYLANILINE", "non-polymer"),
        "ZZS": ("1,3,5-TRIAZINE-2,4-DIAMINE", "non-polymer"),
        "ZZR": (
            "3,6-DIAMINO-1,5-DIHYDRO[1,2,4]TRIAZOLO[4,3-B][1,2,4]TRIAZOL-4-IUM",
            "non-polymer",
        ),
        "ZZQ": (
            "2-{[3-(2-METHOXYPHENYL)-4-OXO-3,4,5,6,7,8-HEXAHYDROQUINAZOLIN-2-YL]SULFANYL}-N-QUINOXALIN-6-YLACETAMIDE",
            "non-polymer",
        ),
        "M8C": ("METHYL ALPHA-D-GALACTOPYRANURONATE", "saccharide"),
        "ZZO": (
            "2-{[4-AMINO-3-(3-FLUORO-4-HYDROXYPHENYL)-1H-PYRAZOLO[3,4-D]PYRIMIDIN-1-YL]METHYL}-5-METHYL-3-(2-METHYLPHENYL)QUINAZOLIN-4(3H)-ONE",
            "non-polymer",
        ),
        "ZZN": (
            "2-{[4-AMINO-3-(3-FLUORO-5-HYDROXYPHENYL)-1H-PYRAZOLO[3,4-D]PYRIMIDIN-1-YL]METHYL}-5-METHYL-3-(2-METHYLPHENYL)QUINAZOLIN-4(3H)-ONE",
            "non-polymer",
        ),
        "ZZM": (
            "1-(1-METHYLETHYL)-3-(PYRIDIN-3-YLETHYNYL)-1H-PYRAZOLO[3,4-D]PYRIMIDIN-4-AMINE",
            "non-polymer",
        ),
        "ZZL": (
            "4-{[9-CHLORO-7-(2,6-DIFLUOROPHENYL)-5H-PYRIMIDO[5,4-D][2]BENZAZEPIN-2-YL]AMINO}BENZOIC ACID",
            "non-polymer",
        ),
        "ZZK": (
            "4-[2-AMINO-5-(1,3-BENZODIOXOL-4-YL)PYRIDIN-3-YL]BENZAMIDE",
            "non-polymer",
        ),
        "ZZJ": ("N^2^-METHYL-L-ALANINAMIDE", "peptide"),
        "ZZI": ("(2R,3R)-2,3,4-TRIHYDROXY-N,N-DIPROPYLBUTANAMIDE", "non-polymer"),
        "ZZH": ("(2S)-2-(4-BENZYLPHENOXY)-3-PHENYLPROPANOIC ACID", "non-polymer"),
        "ZZG": (
            "4-[(5,6-DIMETHYL-2,2'-BIPYRIDIN-3-YL)OXY]-N-(3,4,5-TRIMETHOXYPHENYL)PYRIDIN-2-AMINE",
            "non-polymer",
        ),
        "PFX": (
            "(2S)-1-({2-[2-(CARBOXYMETHOXY)ETHOXY]ETHYL}AMINO)-1-OXO-3-(4-PHOSPHONOPHENYL)PROPAN-2-AMINIUM",
            "peptide",
        ),
        "ZZE": (
            "5-{[3,5-DIETHYL-1-(2-HYDROXYETHYL)-1H-PYRAZOL-4-YL]OXY}BENZENE-1,3-DICARBONITRILE ",
            "non-polymer",
        ),
        "ZZD": ("S-TRITYL-L-CYSTEINE", "peptide"),
        "ZZC": (
            "[(2R,3S,4R,5R)-5-(6-AMINOPURIN-9-YL)-3,4-DIHYDROXY-OXOLAN-2-YL]METHYL [[(2R,3R)-2,3-DIHYDROXY-4-OXO-PENTOXY]-OXIDO-PHOSPHORYL] PHOSPHATE",
            "non-polymer",
        ),
        "ZZB": ("?", "non-polymer"),
        "ZZA": ("1-PHENYL-1H-PYRAZOLE-4-CARBOXYLIC ACID", "non-polymer"),
        "ZZ9": (
            "5'-O-[(S)-HYDROXY{[(1R)-1-HYDROXYUNDECYL]OXY}PHOSPHORYL]ADENOSINE",
            "non-polymer",
        ),
        "ZZ8": ("2-(ACETYLAMINO)BENZOIC ACID", "non-polymer"),
        "ZZ7": (
            "(2R,4S)-2-[(R)-{[(2R)-2-AMINO-2-PHENYLACETYL]AMINO}(CARBOXY)METHYL]-5,5-DIMETHYL-1,3-THIAZOLIDINE-4-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "ZZ6": (
            "2-AMINO-4-(2,4-DICHLOROPHENYL)-N-ETHYLTHIENO[2,3-D]PYRIMIDINE-6-CARBOXAMIDE",
            "non-polymer",
        ),
        "ZZ5": (
            "3,6-DIAMINO-5-CYANO-4-(4-ETHOXYPHENYL)THIENO[2,3-B]PYRIDINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "ZZ4": (
            "4-(2,4-DICHLOROPHENYL)-5-PHENYLDIAZENYL-PYRIMIDIN-2-AMINE",
            "non-polymer",
        ),
        "ZZ3": ("4-METHYL-6-(METHYLSULFANYL)-1,3,5-TRIAZIN-2-AMINE", "non-polymer"),
        "ZZ2": ("4-(2-METHOXYETHOXY)-6-METHYLPYRIMIDIN-2-AMINE", "non-polymer"),
        "ZZ1": ("4-METHYL-2H-CHROMEN-2-ONE", "non-polymer"),
        "ZZ0": ("2-(4-CHLORO-PHENYLAMINO)-NICOTINIC ACID", "non-polymer"),
        "N76": (
            "3-(6-CYCLOHEXYLMETHOXY-9H-PURIN-2-YLAMINO)-BENZENESULFONAMIDE",
            "non-polymer",
        ),
        "8BX": (
            "6-{[(3R,4R)-4-(2-{[2-(2-CHLOROPHENYL)-2,2-DIFLUOROETHYL]AMINO}ETHOXY)PYRROLIDIN-3-YL]METHYL}-4-METHYLPYRIDIN-2-AMINE",
            "non-polymer",
        ),
        "WIA": ("METHYL(6S)-1-THIO-L-MANNO-HEXODIALDO-6,2-PYRANOSIDE", "non-polymer"),
        "WIN": (
            "METHYL (5BETA,7ALPHA,9BETA,10ALPHA,11ALPHA,12ALPHA,13BETA,15ALPHA)-15-{[(2E)-3,4-DIMETHYLPENT-2-ENOYL]OXY}-3,11,12-TRIHYDROXY-2,16-DIOXO-13,20-EPOXYPICRAS-3-EN-21-OATE",
            "non-polymer",
        ),
        "N79": ("?", "non-polymer"),
        "8BR": ("8-BROMO-ADENOSINE-5'-MONOPHOSPHATE", "non-polymer"),
        "8BD": (
            "3-({[(1Z)-(2-METHOXYPHENYL)METHYLIDENE]AMINO}OXY)PROPANOIC ACID",
            "non-polymer",
        ),
        "8BB": (
            "TRANS-4-{[(E)-1-AMINO-2-SULFANYLETHENYL]CARBAMOYL}CYCLOHEXANECARBOXYLIC ACID",
            "peptide",
        ),
        "6LI": ("?", "non-polymer"),
        "0MV": (
            "(3S)-3-(1H-INDOL-3-YLMETHYL)-3,4-DIHYDRO-1H-1,4-BENZODIAZEPINE-2,5-DIONE",
            "non-polymer",
        ),
        "0MW": ("[2-(HEXYLAMINO)ETHANE-1,1-DIYL]BIS(PHOSPHONIC ACID)", "non-polymer"),
        "0MP": (
            "N-((2S,3R)-1-(4-FLUOROPHENYL)-3-HYDROXY-4-((6'-NEOPENTYL-3',4'-DIHYDROSPIRO[CYCLOBUTANE-1,2'-PYRANO[2,3-B]PYRIDIN]-4'-YL)AMINO)BUTAN-2-YL)ACETAMIDE",
            "non-polymer",
        ),
        "0MQ": (
            "[2-(CYCLOHEXYLAMINO)ETHANE-1,1-DIYL]BISPHOSPHONIC ACID",
            "non-polymer",
        ),
        "0MR": (
            "N-[4-CHLORANYL-5-[4-[[3-(2-METHOXYPHENYL)-5-METHYL-1,2-OXAZOL-4-YL]CARBONYL]PIPERAZIN-1-YL]-2-NITRO-PHENYL]PYRIDINE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "0MS": (
            "[4-(5-BROMANYL-3-METHYL-PYRIDIN-2-YL)PIPERAZIN-1-YL]-[3-(2-CHLOROPHENYL)-5-METHYL-1,2-OXAZOL-4-YL]METHANONE",
            "non-polymer",
        ),
        "0MX": (
            "N-[5-(4-{[(3-CHLOROPHENYL)SULFONYL]AMINO}PHENYL)-1H-INDAZOL-3-YL]FURAN-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "0MY": (
            "N-[4-(3-AMINO-1H-INDAZOL-5-YL)PHENYL]-3-CHLOROBENZENESULFONAMIDE",
            "non-polymer",
        ),
        "0MZ": (
            "2-{(2Z,5S)-4-HYDROXY-2-[(2E)-(2-HYDROXYBENZYLIDENE)HYDRAZINYLIDENE]-2,5-DIHYDRO-1,3-THIAZOL-5-YL}-N-[3-(TRIFLUOROMETHYL)PHENYL]ACETAMIDE",
            "non-polymer",
        ),
        "0MD": (
            "{3-[(1R)-3-(3,4-DIMETHOXYPHENYL)-1-({[(2S)-1-{[(1R,2S)-2-ETHYL-1-HYDROXYCYCLOHEXYL](OXO)ACETYL}PIPERIDIN-2-YL]CARBONYL}OXY)PROPYL]PHENOXY}ACETIC ACID",
            "non-polymer",
        ),
        "0ME": (
            "(3S)-N-(9H-XANTHEN-9-YLMETHYL)PIPERIDINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "0MF": (
            "N-[4-CHLORANYL-5-[4-[[3-(2-METHOXYPHENYL)-5-METHYL-1,2-OXAZOL-4-YL]CARBONYL]PIPERAZIN-1-YL]-2-NITRO-PHENYL]FURAN-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "0MG": (
            "AMINO{[(4S)-4-AMINO-5-CARBOXY-5-OXOPENTYL]AMINO}METHANIMINIUM",
            "peptide",
        ),
        "0MA": ("MASLINIC ACID", "non-polymer"),
        "0MB": ("1-(3-PHENYL-1,2,4-THIADIAZOL-5-YL)-1,4-DIAZEPANE", "non-polymer"),
        "0MC": (
            "{3-[(1R)-3-(3,4-DIMETHOXYPHENYL)-1-({[(2S)-1-{[(1S,2R)-2-ETHYL-1-HYDROXYCYCLOHEXYL](OXO)ACETYL}PIPERIDIN-2-YL]CARBONYL}OXY)PROPYL]PHENOXY}ACETIC ACID",
            "non-polymer",
        ),
        "6LQ": (
            "2-AMINO-5-{4-[(4-METHYLPIPERAZIN-1-YL)SULFONYL]PHENYL}-N-[4-(PYRROLIDIN-1-YLMETHYL)PYRIDIN-3-YL]PYRIDINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "0MM": (
            "[4-(2-CHLORO-4-NITROPHENYL)PIPERAZIN-1-YL][3-(2-CHLOROPYRIDIN-3-YL)-5-METHYL-1,2-OXAZOL-4-YL]METHANONE",
            "non-polymer",
        ),
        "0MN": ("?", "non-polymer"),
        "0MO": ("(4R)-4-(3-BUTOXY-4-METHOXYBENZYL)IMIDAZOLIDIN-2-ONE", "non-polymer"),
        "0MH": (
            "N-[4-CHLORANYL-5-[4-[[3-(2-METHOXYPHENYL)-5-METHYL-1,2-OXAZOL-4-YL]CARBONYL]PIPERAZIN-1-YL]-2-NITRO-PHENYL]THIOPHENE-2-CARBOXAMIDE",
            "non-polymer",
        ),
        "0MJ": (
            "(3S,5R)-5-{[(4-METHYLPHENYL)SULFONYL]AMINO}-N-(9H-XANTHEN-9-YLMETHYL)PIPERIDINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "0MK": ("L-RIBOPYRANOSE", "saccharide"),
        "TSL": ("TRANS-ENAMINE INTERMEDIATE OF SULBACTAM", "non-polymer"),
        "NDN": (
            "[(2R,3S,5R)-3-HYDROXY-5-(5-NITRO-2,4-DIOXO-PYRIMIDIN-1-YL)OXOLAN-2-YL]METHYL DIHYDROGEN PHOSPHATE",
            "nucleic-acid",
        ),
        "XME": (
            "4-[(6-CHLORO-2-NAPHTHALENYL)SULFONYL]-1-[[4-HYDROXYIMINOMETHYL-1-(4-PYRIDINYL)-4-PIPERIDINYL]METHYL]PIPERAZINONE",
            "non-polymer",
        ),
        "T3Q": (
            "[(3R,4S,5S,6R)-4-AMINO-3,5-DIHYDROXY-6-METHYLOXAN-2-YL][HYDROXY-[[(2R,3S,5R)-3-HYDROXY-5-(5-METHYL-2,4-DIOXOPYRIMIDIN-1-YL)OXOLAN-2-YL]METHOXY]PHOSPHORYL] HYDROGEN PHOSPHATE",
            "non-polymer",
        ),
        "RF3": (
            "1-(THIOPHEN-2-YLACETYL)-4-(3-THIOPHEN-2-YL-1,2,4-OXADIAZOL-5-YL)PIPERIDINE",
            "non-polymer",
        ),
        "XMG": (
            "4-[(6-CHLORO-2-NAPHTHALENYL)SULFONYL]-1-[[1-(4-PYRIDINYL)-4-PIPERIDINYL] METHYL]-2-PIPERAZINECARBOXYLIC ACID",
            "non-polymer",
        ),
        "PRH": ("6-HYDROXY-1,6-DIHYDRO PURINE NUCLEOSIDE", "non-polymer"),
        "RF2": ("3-(4-FLUOROPHENYL)-5-PHENYL-4H-1,2,4-TRIAZOLE", "non-polymer"),
        "TWO": ("3,4-DIMETHOXYBENZOIC ACID", "non-polymer"),
        "NDE": ("NAD WITH (1-HYDROXYL-2-OXO-3-PHOSPHONOOXY-PROPANE)", "non-polymer"),
        "T00": (
            "N-[(2R,3S)-3-AMINO-2-HYDROXY-4-PHENYLBUTYL]-L-PHENYLALANINE",
            "peptide",
        ),
        "RF1": (
            "TERT-BUTYL 4-(3-THIOPHEN-2-YL-1,2,4-OXADIAZOL-5-YL)PIPERIDINE-1-CARBOXYLATE",
            "non-polymer",
        ),
        "NDD": ("2,6-DICARBOXYNAPHTHALENE", "non-polymer"),
        "NDG": ("2-(ACETYLAMINO)-2-DEOXY-A-D-GLUCOPYRANOSE", "saccharide"),
        "XMC": (
            "4-[[4-[(6-CHLORO-2-NAPHTHALENYL)SULFONYL]-6-OXO-1-[[1-(4-PYRIDINYL)-4-PIPERIDINYL]METHYL]-2-PIPERAZINYL]CARBONYL]MORPHOLINE",
            "non-polymer",
        ),
        "0M4": (
            "N-(4-CHLORO-3-FLUOROPHENYresponse=urllib2.urlopL)-N'-[(3AS,6AS)-HEXAHYDROCYCLOPENTA[C]PYRROL-3A(1H)-YLMETHYL]ETHANEDIAMIDE",
            "non-polymer",
        ),
        "0M5": (
            "N-(4-CHLORO-3-FLUOROPHENYL)-N'-{[(3R)-1-CYCLOPROPYLPYRROLIDIN-3-YL]METHYL}ETHANEDIAMIDE",
            "non-polymer",
        ),
        "0M6": (
            "N-[(2S)-1,4-DIHYDROXYBUTAN-2-YL]-N~2~-(PHENYLCARBAMOTHIOYL)-L-LEUCINAMIDE",
            "non-polymer",
        ),
        "0M7": ("[2-(PENTYLAMINO)ETHANE-1,1-DIYL]BIS(PHOSPHONIC ACID)", "non-polymer"),
        "0M1": (
            "N-{[(4S,5S)-5-(AMINOMETHYL)-2,2-DIMETHYL-1,3-DIOXOLAN-4-YL]METHYL}-N'-(4-CHLORO-3-FLUOROPHENYL)ETHANEDIAMIDE",
            "non-polymer",
        ),
        "0M2": (
            "(2R)-1-(3,8-DIHYDRODIBENZO[B,F]PYRROLO[3,4-D]AZEPIN-2(1H)-YL)PROPAN-2-OL",
            "non-polymer",
        ),
        "0M3": (
            "(3S,5R)-N-(2,2-DIPHENYLETHYL)-5-{[(4-METHYLPHENYL)SULFONYL]AMINO}PIPERIDINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "2FT": ("2,2-DIFLUOROPENTANEDIOIC ACID", "non-polymer"),
        "0M8": ("[2-(HEPTYLAMINO)ETHANE-1,1-DIYL]BIS(PHOSPHONIC ACID)", "non-polymer"),
        "0M9": ("[2-(PROPYLAMINO)ETHANE-1,1-DIYL]BIS(PHOSPHONIC ACID)", "non-polymer"),
        "WXV": (
            "N,1-DIMETHYL-8-{[1-(METHYLSULFONYL)PIPERIDIN-4-YL]AMINO}-1H-PYRAZOLO[4,3-H]QUINAZOLINE-3-CARBOXAMIDE",
            "non-polymer",
        ),
        "NDZ": (
            "(2R,3AR,6R,7R,7AR)-2-[(2S)-2-AMINO-2-CARBOXYETHYL]-6,7-DIHYDROXYHEXAHYDRO-2H-FURO[3,2-B]PYRAN-2-CARBOXYLIC ACID",
            "non-polymer",
        ),
        "I05": (
            "(3R,4R)-N-{4-[4-(2-FLUORO-6-HYDROXY-3-METHOXY-BENZOYL)-BENZYLOXY]-AZEPAN-3-YL}-ISONICOTINAMIDE",
            "non-polymer",
        ),
        "2LJ": (
            "1-DEOXY-1-({2,6-DIOXO-5-[(E)-(2-OXOPROPYLIDENE)AMINO]-1,2,3,6-TETRAHYDROPYRIMIDIN-4-YL}AMINO)-D-RIBITOL",
            "non-polymer",
        ),
        "7WP": ("2,4-DIAMINOPTERIDINE-6-CARBALDEHYDE", "non-polymer"),
        "7WO": ("2-HYDROXYNAPHTHALENE-1-CARBALDEHYDE", "non-polymer"),
        "7WQ": ("2-HYDROXY-5-METHOXYBENZALDEHYDE", "non-polymer"),
        "7ZS": ("3-METHANOYL-2-OXIDANYL-BENZOIC ACID", "non-polymer"),
        "30W": ("N-(6-FORMYL-4-OXO-3,4-DIHYDROPTERIDIN-2-YL)ACETAMIDE", "non-polymer"),
        "7ZV": ("5-HYDROXYDICLOFENAC", "non-polymer"),
    }


# residue name to standard amino acid name and the expected heavy atom names
resname2standard = {
    "AGM": (
        "ARG",
        ["N", "CA", "CB", "CG", "CD", "CE2", "NE1", "CZ", "NH1", "NH2", "C", "O"],
    ),
    "LLP": (
        "LYS",
        [
            "N1",
            "C2",
            "C2'",
            "C3",
            "O3",
            "C4",
            "C4'",
            "C5",
            "C6",
            "C5'",
            "OP4",
            "P",
            "OP1",
            "OP2",
            "OP3",
            "N",
            "CA",
            "CB",
            "CG",
            "CD",
            "CE",
            "NZ",
            "C",
            "O",
        ],
    ),
    "NYS": (
        "CYS",
        [
            "N",
            "CA",
            "CB",
            "C",
            "O",
            "SG",
            "CXP",
            "CXH",
            "CXQ",
            "OXF",
            "CXO",
            "OXE",
            "CXG",
            "CXN",
            "CXJ",
            "OXG",
            "CXI",
            "NXA",
        ],
    ),
    "SEN": (
        "SER",
        ["N", "CA", "C", "O", "CB", "OG", "P", "O2", "O3", "N1", "C1", "C2"],
    ),
    "SIB": (
        "CYS",
        [
            "N",
            "CA",
            "CB",
            "CG",
            "SD",
            "C",
            "O",
            "C5'",
            "C4'",
            "O4'",
            "C3'",
            "O3'",
            "C2'",
            "O2'",
            "C1'",
            "N9",
            "C8",
            "N7",
            "C5",
            "C6",
            "O6",
            "N1",
            "C2",
            "N3",
            "C4",
        ],
    ),
    "B2I": ("ILE", ["N", "CA", "CB", "CG1", "CG2", "CD1", "B", "O1", "O2"]),
    "NYB": ("CYS", ["N", "CA", "CB", "SG", "CD", "CE1", "CE2", "CF", "CN", "C", "O"]),
    "B2F": (
        "PHE",
        ["N", "CA", "CB", "CG", "CD1", "CD2", "CE1", "CE2", "CZ", "B", "O1", "O2"],
    ),
    "AGT": (
        "CYS",
        [
            "O",
            "C",
            "CA",
            "N",
            "CB",
            "SG",
            "CZ",
            "NH2",
            "NE",
            "CD",
            "CG",
            "CK",
            "CJ",
            "NB",
        ],
    ),
    "NYG": (
        "ASN",
        [
            "OD1",
            "CG1",
            "ND1",
            "CB1",
            "CA1",
            "N",
            "C1",
            "N3",
            "CA3",
            "C",
            "O",
            "N2",
            "CA2",
            "C2",
            "O2",
            "CB2",
            "CG2",
            "CD2",
            "CE2",
            "CZ",
            "OH",
            "CE1",
            "CD1",
        ],
    ),
    "AGQ": (
        "TYR",
        [
            "OXT",
            "C",
            "N",
            "O",
            "N5",
            "N6",
            "C7",
            "N8",
            "N9",
            "CA",
            "CB",
            "CG",
            "OH",
            "CZ",
            "OZ",
            "CD1",
            "CD2",
            "CE1",
            "CE2",
        ],
    ),
    "TAV": (
        "ASP",
        [
            "C1",
            "C2",
            "C3",
            "C4",
            "C5",
            "C6",
            "C9",
            "C10",
            "C13",
            "C14",
            "S18",
            "O19",
            "N20",
            "C21",
            "C22",
            "C23",
            "C24",
            "C25",
            "C26",
            "C29",
            "S31",
            "C32",
            "C33",
            "C34",
            "C35",
            "C36",
            "C37",
            "C41",
            "C42",
            "C45",
            "C46",
            "N49",
            "O50",
            "O51",
            "O52",
            "O53",
            "O54",
            "C56",
            "C57",
            "C61",
            "O62",
            "O63",
            "C64",
            "O65",
            "O66",
            "C67",
        ],
    ),
    "KYN": (
        "TRP",
        [
            "C",
            "N",
            "C1",
            "N1",
            "O2",
            "CA",
            "CB",
            "CG",
            "CZ",
            "CD1",
            "CD2",
            "CE1",
            "CE2",
            "O",
        ],
    ),
    "R1A": (
        "CYS",
        [
            "OXT",
            "C",
            "CA",
            "N",
            "CB",
            "SG",
            "SD",
            "CE",
            "C3",
            "C2",
            "C9",
            "C8",
            "C4",
            "C5",
            "C7",
            "C6",
            "N1",
            "O1",
            "O",
        ],
    ),
    "B2V": ("VAL", ["N", "CA", "CB", "CG1", "CG2", "B", "O1", "O2"]),
    "MNL": ("LEU", ["N", "CA", "C", "O", "CB", "CG", "CD", "CE", "CM1", "CM4"]),
    "KYQ": (
        "LYS",
        [
            "N",
            "CA",
            "CB",
            "CG",
            "CD",
            "CE",
            "NZ",
            "C11",
            "C12",
            "O16",
            "O15",
            "C13",
            "O14",
            "C",
            "O",
        ],
    ),
    "TY1": (
        "TYR",
        [
            "N",
            "CG",
            "OH",
            "CL",
            "CZ",
            "CAA",
            "CAB",
            "CAC",
            "CBL",
            "C",
            "CA",
            "CD1",
            "CD2",
            "CE1",
            "CE2",
            "O",
        ],
    ),
    "ALN": (
        "ALA",
        [
            "N",
            "CA",
            "C",
            "O",
            "CB",
            "CG1",
            "CD1",
            "CE1",
            "CD2",
            "CE2",
            "CZ1",
            "CG2",
            "CD3",
            "CE3",
            "CZ2",
        ],
    ),
    "VAH": ("VAL", ["C", "N", "O", "CA", "CB", "CD1", "CG2", "OG1"]),
    "VAI": ("VAL", ["C", "N", "O", "CA", "CB", "CG1", "CG2"]),
    "VAL": ("VAL", ["N", "CA", "C", "O", "CB", "CG1", "CG2"]),
    "VAF": ("VAL", ["N", "CA", "CB", "CG1", "CG2", "CF", "CJ", "O"]),
    "0FL": ("ALA", ["C1", "O1", "N", "CA", "C", "O", "CB", "N2"]),
    "VAD": ("VAL", ["CA", "CB", "CG1", "CG2", "C", "O", "OE"]),
    "HZP": ("PRO", ["OXT", "N", "CA", "C", "O", "CB", "CG", "CD", "OD1"]),
    "XPR": (
        "PRO",
        [
            "N",
            "CD",
            "CG",
            "CB",
            "CA",
            "C14",
            "C13",
            "C12",
            "C11",
            "O2",
            "S",
            "OS1",
            "OS2",
            "OS3",
            "C",
            "O",
        ],
    ),
    "FIO": (
        "ARG",
        [
            "N10",
            "C11",
            "C12",
            "O14",
            "C15",
            "C16",
            "C17",
            "N18",
            "C19",
            "C20",
            "N21",
            "F1",
        ],
    ),
    "AHB": ("ASN", ["N", "CA", "CB", "OB", "CG", "OD1", "ND2", "C", "O"]),
    "NPI": ("ALA", ["CA", "CB", "C4", "C5", "C6", "C", "O", "N", "C7", "O71", "O72"]),
    "NPH": (
        "CYS",
        [
            "N",
            "CA",
            "CB",
            "SG",
            "CD",
            "CE",
            "OZ",
            "NZ",
            "C6",
            "C5",
            "C6A",
            "C4A",
            "C10",
            "C7",
            "C4",
            "C1A",
            "N10",
            "C8",
            "C3",
            "N1",
            "C9",
            "C2",
            "C",
            "O",
        ],
    ),
    "B27": ("THR", ["N", "CA", "CB", "CG2", "OG1", "SG", "CX"]),
    "AHO": (
        "ALA",
        ["C1", "C2", "C3", "C4", "C5", "C6", "C7", "N1", "N2", "O1", "O2", "O3"],
    ),
    "MAA": ("ALA", ["N", "CM", "CA", "CB", "C", "O"]),
    "AHP": ("ALA", ["N", "CA", "C", "O", "CB", "CG", "CD", "CE", "CZ"]),
    "2FM": ("MET", ["O", "C", "CA", "N", "CB", "CG", "SD", "CE", "FZ1", "FZ2"]),
    "LMQ": ("GLN", ["N", "CA", "CB", "CB2", "CG", "CD", "OE1", "NE2", "C", "O"]),
    "MAI": (
        "ARG",
        ["N", "CA", "CF", "CJ", "CB", "CG", "CD", "NE", "CZ", "NH1", "NH2", "OH"],
    ),
    "CML": (
        "CYS",
        [
            "O5",
            "C4",
            "O4",
            "C3",
            "C2",
            "C1",
            "O1",
            "O2",
            "SG",
            "CB",
            "CA",
            "C",
            "O",
            "N",
        ],
    ),
    "4AR": (
        "ARG",
        [
            "N1",
            "C1",
            "C2",
            "O1",
            "C3",
            "C4",
            "C5",
            "N2",
            "C6",
            "N3",
            "N4",
            "C7",
            "O2",
            "C8",
            "C9",
            "C10",
            "C11",
            "C12",
            "C13",
            "C14",
            "O3",
            "C15",
            "C16",
        ],
    ),
    "1TQ": (
        "TRP",
        [
            "N",
            "CA",
            "C",
            "O",
            "CB",
            "CG",
            "CD1",
            "CD2",
            "NE1",
            "CE2",
            "CE3",
            "CZ2",
            "CZ3",
            "CH2",
            "O2",
            "CH3",
            "N2",
            "O1",
        ],
    ),
    "PCS": (
        "PHE",
        [
            "N",
            "CA",
            "C",
            "CB",
            "O",
            "CD1",
            "CG",
            "CD2",
            "CE2",
            "CZ",
            "CE1",
            "CT",
            "CL1",
        ],
    ),
    "CME": ("CYS", ["N", "CA", "CB", "SG", "SD", "CE", "CZ", "OH", "C", "O"]),
    "HL2": ("LEU", ["C", "N", "O", "CA", "CB", "CG", "OH", "CD1", "CD2"]),
    "1TY": (
        "TYR",
        [
            "N",
            "CA",
            "CB",
            "C",
            "O",
            "CG",
            "CD1",
            "OX1",
            "CE1",
            "CD2",
            "CE2",
            "NX1",
            "CZ",
            "OH",
            "C3'",
            "C2'",
            "C1'",
            "C6'",
            "C5'",
            "C4'",
            "C2",
            "C1",
        ],
    ),
    "4AF": (
        "PHE",
        [
            "C11",
            "C8",
            "O2",
            "C7",
            "C9",
            "C10",
            "C6",
            "C5",
            "C4",
            "C3",
            "C1",
            "C2",
            "OX1",
            "O",
            "N",
        ],
    ),
    "11Q": (
        "PRO",
        [
            "N",
            "CA",
            "C",
            "CB",
            "O",
            "CX",
            "C1",
            "C6",
            "C5",
            "C4",
            "C3",
            "C2",
            "CD",
            "CG",
        ],
    ),
    "PCC": ("PRO", ["N", "CA", "CB", "CG", "CD", "C", "O", "OE1"]),
    "PCA": ("GLU", ["N", "CA", "CB", "CG", "CD", "OE", "C", "O"]),
    "C4R": (
        "CYS",
        [
            "OXT",
            "O",
            "C",
            "CA",
            "N",
            "CB",
            "SG",
            "C15",
            "C13",
            "C12",
            "C11",
            "C16",
            "C17",
            "C10",
            "RH",
        ],
    ),
    "FT6": (
        "TRP",
        [
            "N",
            "CA",
            "C",
            "CB",
            "CG",
            "CD2",
            "CD1",
            "NE1",
            "CE2",
            "CZ2",
            "CH2",
            "FAP",
            "CZ3",
            "CE3",
            "O",
        ],
    ),
    "TTS": (
        "TYR",
        [
            "N",
            "CA",
            "CB",
            "C",
            "O",
            "CG",
            "CD1",
            "OD1",
            "CE1",
            "CD2",
            "CE2",
            "NE2",
            "CZ",
            "OH",
            "C3'",
            "C2'",
            "C1'",
            "C6'",
            "C5'",
            "C4'",
            "C2",
            "C1",
            "O1",
        ],
    ),
    "TTQ": (
        "TRP",
        [
            "N",
            "CA",
            "C",
            "O",
            "CB",
            "CG",
            "CD1",
            "CD2",
            "NE1",
            "CE2",
            "CE3",
            "CZ2",
            "CZ3",
            "CH2",
            "O2",
            "N2",
        ],
    ),
    "MMO": (
        "ARG",
        ["N", "CA", "C", "O", "CB", "CG", "CD", "NE", "CZ", "NH2", "NH1", "CN"],
    ),
    "KBE": ("LYS", ["C", "N", "O", "CA", "CB", "CD", "CE", "CG", "NZ"]),
    "M2L": ("LYS", ["N", "CA", "CB", "SG", "CD", "CE", "NZ", "CM1", "CM2", "C", "O"]),
    "I58": (
        "LYS",
        ["N", "C", "O", "CA", "CB", "CG", "F", "CD", "CE", "NZ", "CX", "NX", "CT"],
    ),
    "B2A": ("ALA", ["N", "CA", "CB", "B", "O1", "O2"]),
    "CSD": ("CYS", ["N", "CA", "CB", "SG", "C", "O", "OD1", "OD2"]),
    "I2M": ("ILE", ["N", "CA", "CB", "CG3", "CG2", "CG1", "CD1", "C", "O"]),
    "FTR": (
        "TRP",
        [
            "N",
            "CA",
            "CB",
            "CG",
            "CD2",
            "CE2",
            "CE3",
            "CD1",
            "NE1",
            "CZ2",
            "CZ3",
            "F",
            "CH2",
            "C",
            "O",
        ],
    ),
    "BLY": ("LYS", ["N", "CA", "CB", "CG", "CD", "CE", "NZ", "B", "O"]),
    "NFA": (
        "PHE",
        ["N", "CA", "C", "O", "CB", "CG", "CD1", "CD2", "CE1", "CE2", "CZ", "NXT"],
    ),
    "FTY": (
        "TYR",
        [
            "N",
            "CA",
            "C",
            "O",
            "CB",
            "CG",
            "CD1",
            "CD2",
            "CE1",
            "CE2",
            "CZ",
            "C1",
            "F1",
            "F2",
            "P",
            "O1P",
            "O2P",
            "O3P",
        ],
    ),
    "KPY": (
        "LYS",
        [
            "NZ",
            "CE",
            "CD",
            "SG",
            "CB",
            "CA",
            "C",
            "O",
            "N",
            "OL1",
            "CL",
            "OL2",
            "CI",
            "CK",
        ],
    ),
    "H14": (
        "PHE",
        ["N", "CA", "C", "O", "CB", "OB", "CG", "CD1", "CD2", "CE1", "CE2", "CZ"],
    ),
    "HIC": ("HIS", ["N", "CA", "C", "O", "CB", "CG", "ND1", "CD2", "CE1", "NE2", "CZ"]),
    "ALV": ("ALA", ["C", "N", "O", "CA", "CB"]),
    "BLE": ("LEU", ["CA", "CB", "CG", "CD1", "CD2", "N", "B", "O1", "O2"]),
    "KPI": (
        "LYS",
        [
            "OXT",
            "N",
            "CA",
            "CB",
            "CG",
            "CD",
            "CE",
            "NZ",
            "CX1",
            "C1",
            "CX2",
            "O1",
            "O2",
            "C",
            "O",
        ],
    ),
    "IGL": (
        "GLY",
        [
            "N",
            "CA",
            "C",
            "O",
            "CB",
            "CG1",
            "CG2",
            "CD1",
            "CE1",
            "CH1",
            "CH2",
            "CE2",
            "CD2",
        ],
    ),
    "MA": ("ALA", ["N", "CA", "C", "O", "CB", "CH3"]),
    "HLU": ("LEU", ["N", "CA", "CB", "CG", "CD1", "CD2", "C", "O", "OH"]),
    "A5N": ("ASN", ["N", "CA", "C", "O", "CB", "CG", "OD1", "ND2", "C1", "O2", "O3"]),
    "3PX": ("PRO", ["C", "N", "O", "CA", "CB", "CD", "CG", "OB1", "CB2", "CB3", "CB4"]),
    "HAR": (
        "ARG",
        ["N", "CA", "C", "O", "CB", "CG", "CD", "NE", "CZ", "NH1", "NH2", "OH1"],
    ),
    "SBL": (
        "SER",
        [
            "N",
            "CA",
            "CB",
            "O3",
            "C",
            "O",
            "N13",
            "C12",
            "C11",
            "C1",
            "C10",
            "C5",
            "C2",
            "C3",
            "C4",
            "O1",
            "O2",
            "B",
            "C14",
            "O15",
            "C16",
            "C9",
            "C8",
            "C7",
            "C6",
        ],
    ),
    "TS9": ("ILE", ["N", "CA", "C", "CB", "OG3", "CG2", "CG1", "OD2", "CD1", "O"]),
    "NOT": ("LEU", ["N", "CA", "C", "O", "CB", "CG", "C1", "C2", "N1", "N2", "N3"]),
    "GLH": (
        "GLN",
        [
            "N",
            "CA",
            "CB",
            "C",
            "CG",
            "CD",
            "OE1",
            "N1",
            "O",
            "O1",
            "C1",
            "C2",
            "N2",
            "C3",
            "C7",
            "C4",
            "C5",
            "C6",
            "C8",
            "C9",
            "C13",
            "C10",
            "C11",
            "C12",
        ],
    ),
    "AKZ": ("ASP", ["C", "N", "O", "CA", "CB", "CG", "OD1", "OD2"]),
    "GLJ": ("GLU", ["N", "CA", "C", "O", "CB", "CG", "CD", "OE1", "OE2"]),
    "GLK": ("GLU", ["N", "CA", "C", "CB", "CG", "CD", "OE1", "OE2", "O"]),
    "5CS": (
        "CYS",
        [
            "N",
            "CA",
            "CB",
            "SG",
            "C",
            "O",
            "CI",
            "NI",
            "CJ",
            "OJ1",
            "OJ2",
            "CF",
            "OF",
            "NX",
            "CD",
            "CE",
        ],
    ),
    "GLN": ("GLN", ["N", "CA", "C", "O", "CB", "CG", "CD", "OE1", "NE2"]),
    "L3O": ("LEU", ["CZ", "OZ", "CA", "CB", "CG", "CD1", "CD2", "N", "C", "O"]),
    "CSZ": ("CYS", ["N", "CA", "C", "O", "CB", "SG", "SE"]),
    "TOX": (
        "TRP",
        [
            "N",
            "CA",
            "CB",
            "CG",
            "CD1",
            "NE1",
            "CE2",
            "CD2",
            "CE3",
            "CZ3",
            "CH2",
            "CZ2",
            "C",
            "O",
            "O2",
            "O1",
        ],
    ),
    "CTE": (
        "TRP",
        [
            "O",
            "C",
            "CA",
            "N",
            "CB",
            "CG",
            "CD1",
            "NE1",
            "CE2",
            "CD2",
            "CE3",
            "CZ3",
            "CH2",
            "CZ2",
            "CL",
        ],
    ),
    "GLZ": ("GLY", ["N", "CA", "C", "O"]),
    "AKL": ("ASP", ["N", "CA", "CB", "CG", "OD2", "OD1", "C", "O", "CM", "CL1"]),
    "GLQ": ("GLU", ["N", "CA", "C", "O", "CB", "CG", "CD", "OE1", "OE2"]),
    "SME": ("MET", ["N", "CA", "CB", "CG", "S", "OE", "CE", "C", "O"]),
    "CTH": ("THR", ["N", "CA", "C", "O", "CB", "CG2", "CL2", "OG1"]),
    "GLU": ("GLU", ["N", "CA", "C", "O", "CB", "CG", "CD", "OE1", "OE2"]),
    "CS3": (
        "CYS",
        [
            "OXT",
            "C",
            "CA",
            "N",
            "CB",
            "SG",
            "C1H",
            "C1I",
            "C1L",
            "O1C",
            "C1M",
            "C1G",
            "S1K",
            "C1F",
            "C1E",
            "O",
        ],
    ),
    "MDF": (
        "TYR",
        [
            "N",
            "CA",
            "CG",
            "CD1",
            "CD2",
            "CE1",
            "CE2",
            "CZ",
            "OH1",
            "OH2",
            "C",
            "O",
            "CH3",
        ],
    ),
    "CSP": ("CYS", ["N", "CA", "CB", "SG", "C", "O", "P", "O1P", "O2P", "O3P"]),
    "0E5": ("THR", ["C1", "C2", "C3", "C4", "N1", "C19", "C20", "O6", "O3"]),
    "BSE": ("SER", ["N", "CC", "CA", "C", "O", "CB", "OG"]),
    "0YG": (
        "TYR",
        [
            "C",
            "O",
            "C2",
            "N2",
            "O2",
            "N3",
            "OH",
            "CZ",
            "CA2",
            "CA3",
            "CB2",
            "CD1",
            "CD2",
            "CE1",
            "CE2",
            "CG2",
            "O4",
        ],
    ),
    "FCL": (
        "PHE",
        ["N", "CA", "C", "O", "CB", "CG", "CD1", "CD2", "CE1", "CE2", "CL1", "CZ"],
    ),
    "FC0": (
        "PHE",
        [
            "CD1",
            "CE1",
            "CZ",
            "CE2",
            "CD2",
            "CG",
            "CB",
            "CA",
            "C",
            "O",
            "N",
            "C1",
            "O1",
            "OX0",
        ],
    ),
    "LTR": (
        "TRP",
        [
            "N",
            "CA",
            "C",
            "O",
            "CB",
            "CG",
            "CD1",
            "CD2",
            "NE1",
            "CE2",
            "CE3",
            "CZ2",
            "CZ3",
            "CH2",
        ],
    ),
    "MVA": ("VAL", ["N", "CN", "CA", "CB", "CG1", "CG2", "C", "O"]),
    "MD3": (
        "CYS",
        [
            "C16",
            "N",
            "O05",
            "SG",
            "C01",
            "C02",
            "O",
            "CA",
            "O03",
            "C",
            "O04",
            "CB",
            "C06",
            "C07",
            "C08",
            "C09",
            "C10",
            "C11",
            "C12",
            "C13",
            "C14",
            "C15",
        ],
    ),
    "FME": ("MET", ["N", "CN", "O1", "CA", "CB", "CG", "SD", "CE", "C", "O"]),
    "MD6": (
        "GLY",
        ["C", "N", "O", "CA", "N2", "C3", "O3", "C4", "O4", "C5", "C6", "C7", "C8"],
    ),
    "MD5": (
        "CYS",
        [
            "C",
            "N",
            "CA",
            "CB",
            "CAA",
            "CAB",
            "OAH",
            "OAI",
            "OAJ",
            "SAL",
            "CAM",
            "CAN",
            "CAO",
            "CAP",
            "CAQ",
            "CAR",
            "CAS",
            "NAT",
            "OAU",
            "OAV",
            "O",
        ],
    ),
    "NTY": ("TYR", ["N", "CA", "CG", "CD1", "CD2", "CE1", "CE2", "CZ", "OH", "C", "O"]),
    "ALO": ("THR", ["N", "CA", "CB", "CG2", "OG1", "C", "O"]),
    "05N": ("PRO", ["N3", "C19", "C20", "C21", "C22", "O9", "C52", "C53", "O2", "O3"]),
    "ALM": ("ALA", ["N", "CA", "C", "O", "CB", "CM"]),
    "RE3": (
        "TRP",
        [
            "N",
            "CA",
            "CB",
            "CG",
            "OG",
            "CD1",
            "OD1",
            "NE1",
            "CE2",
            "CZ2",
            "CH2",
            "CZ3",
            "CE3",
            "CD2",
            "C",
            "O",
        ],
    ),
    "RE0": (
        "TRP",
        [
            "N",
            "CA",
            "CB",
            "CG",
            "OG",
            "CD1",
            "OD1",
            "NE1",
            "CE2",
            "CZ2",
            "CH2",
            "CZ3",
            "CE3",
            "CD2",
            "C",
            "O",
        ],
    ),
    "NTR": (
        "TYR",
        ["N", "CA", "CG", "CD1", "CD2", "CE1", "CE2", "CZ", "OH", "OE3", "C", "O"],
    ),
    "ALC": ("ALA", ["N", "CA", "C", "O", "CB", "CG", "CD2", "CE2", "CZ", "CE1", "CD1"]),
    "ORQ": ("ARG", ["N", "CA", "CB", "CG", "CD", "NE", "O1", "C", "O", "C1", "C2"]),
    "ORN": ("ALA", ["N", "CA", "CB", "CG", "CD", "NE", "C", "O"]),
    "VB1": (
        "LYS",
        [
            "C1",
            "C2",
            "C3",
            "CAV",
            "C4",
            "C7",
            "C10",
            "N11",
            "C12",
            "C13",
            "N14",
            "O15",
            "C16",
            "S17",
            "O18",
            "O19",
            "O20",
            "N29",
            "C31",
            "C32",
            "C33",
            "O42",
            "O43",
        ],
    ),
    "JJL": (
        "CYS",
        [
            "C",
            "N",
            "O",
            "N1",
            "C2",
            "C3",
            "C4",
            "C5",
            "C6",
            "C7",
            "O7",
            "CA",
            "CB",
            "SG",
            "O15",
            "C16",
        ],
    ),
    "JJK": (
        "CYS",
        [
            "C",
            "N",
            "O",
            "N1",
            "C2",
            "C3",
            "C4",
            "C5",
            "C6",
            "C7",
            "O7",
            "CA",
            "CB",
            "SG",
        ],
    ),
    "JJJ": (
        "CYS",
        [
            "C",
            "N",
            "O",
            "N1",
            "C2",
            "C3",
            "C4",
            "C5",
            "C6",
            "C7",
            "O7",
            "CA",
            "CB",
            "SG",
        ],
    ),
    "ALY": (
        "LYS",
        ["OH", "CH", "CH3", "NZ", "CE", "CD", "CG", "CB", "CA", "N", "C", "O"],
    ),
    "MME": ("MET", ["N", "CA", "C", "O", "CB", "CG", "SD", "CE", "CM"]),
    "ALT": ("ALA", ["N", "CA", "CB", "C", "S"]),
    "GL3": ("GLY", ["CA", "N", "C", "O", "S"]),
    "KOR": (
        "MET",
        [
            "OAO",
            "CAN",
            "OAQ",
            "CAJ",
            "NAK",
            "CAI",
            "C",
            "O",
            "N",
            "OAA",
            "CAH",
            "SAG",
            "NAF",
            "SAE",
            "CB",
            "CA",
        ],
    ),
    "LME": ("GLU", ["N", "CA", "CB", "C1", "CG", "CD", "OE1", "OE2", "C", "O"]),
    "LMF": (
        "LYS",
        ["N", "CA", "C", "O", "CB", "CG", "CD", "CE", "NZ", "C1", "S", "N1", "CM"],
    ),
    "C22": ("ALA", ["C", "N", "O", "CL", "CA", "CB"]),
    "CAY": ("CYS", ["N", "CA", "CB", "CG", "SD", "CE", "C", "O", "OF1", "OF2"]),
    "DPQ": (
        "TYR",
        [
            "OE2",
            "N",
            "CA",
            "C",
            "O",
            "CB",
            "CG",
            "CD1",
            "CD2",
            "CE1",
            "CE2",
            "CZ",
            "OH",
        ],
    ),
    "DPP": ("ALA", ["N", "CA", "C", "O", "CB", "NG"]),
    "CAS": ("CYS", ["N", "CA", "CB", "C", "O", "SG", "AS", "CE1", "CE2"]),
    "NC1": (
        "SER",
        [
            "S1",
            "C2",
            "N3",
            "C4",
            "C5",
            "C6",
            "C7",
            "O8",
            "C9",
            "N10",
            "C11",
            "O12",
            "C13",
            "C14",
            "S15",
            "C16",
            "C17",
            "C18",
            "C19",
            "O20",
            "O21",
            "C22",
            "C23",
            "C24",
            "C25",
            "C26",
            "C27",
            "C28",
            "C29",
            "N30",
            "O31",
            "O32",
            "N33",
            "O34",
            "O35",
            "N",
            "CA",
            "CB",
            "OG",
            "C",
            "O",
        ],
    ),
    "TOQ": (
        "TRP",
        [
            "C",
            "N",
            "O",
            "O6",
            "O7",
            "CA",
            "CB",
            "CG",
            "CD1",
            "CD2",
            "NE1",
            "CE2",
            "CE3",
            "CH2",
            "CZ2",
            "CZ3",
        ],
    ),
    "LCK": (
        "LYS",
        [
            "OXT",
            "O2",
            "C4",
            "O1",
            "C3",
            "C2",
            "C1",
            "NZ",
            "CE",
            "CD",
            "CG",
            "CB",
            "CA",
            "N",
            "C",
            "O",
        ],
    ),
    "CSO": ("CYS", ["N", "CA", "CB", "SG", "C", "O", "OD"]),
    "DPL": ("PRO", ["N", "CA", "C", "O", "CB", "CG", "CD", "OG1"]),
    "CAF": ("CYS", ["N", "CA", "CB", "C", "O", "SG", "AS", "CE1", "CE2", "O1"]),
    "HIP": (
        "HIS",
        [
            "N",
            "CA",
            "CB",
            "CG",
            "CD2",
            "NE2",
            "CE1",
            "ND1",
            "P",
            "O1P",
            "O2P",
            "O3P",
            "C",
            "O",
        ],
    ),
    "HIQ": (
        "HIS",
        [
            "N",
            "CA",
            "C",
            "O",
            "CB",
            "CG",
            "ND1",
            "CD2",
            "CE1",
            "NE2",
            "O1",
            "O2",
            "O3",
            "C1",
            "C2",
            "C3",
        ],
    ),
    "ONH": ("ALA", ["C", "N", "O", "CA", "CB", "CD", "NE", "CG", "OZ"]),
    "BH2": ("ASP", ["C", "N", "O", "CA", "CB", "OB", "CG", "OD1", "OD2"]),
    "EHP": (
        "PHE",
        ["N", "CA", "C", "O", "CB", "CG", "CD1", "CD2", "CE1", "CE2", "CZ", "OH"],
    ),
    "DYS": (
        "CYS",
        [
            "O",
            "C",
            "CA",
            "N",
            "CB",
            "SG",
            "CXP",
            "CXH",
            "CXQ",
            "OXF",
            "CXO",
            "OXE",
            "CXG",
            "CXN",
            "CXJ",
            "CXI",
            "NXA",
        ],
    ),
    "HIA": (
        "HIS",
        ["N", "CA", "CB", "CG", "ND1", "CD2", "CE1", "NE2", "C", "O", "NXT"],
    ),
    "6HN": ("LYS", ["N1", "N", "CA", "C", "O", "CB", "CG", "CD", "CE", "O2", "O1"]),
    "TBM": (
        "THR",
        [
            "N",
            "CT",
            "CN",
            "CA",
            "C",
            "O",
            "CB",
            "OG1",
            "CG2",
            "CD1",
            "CD2",
            "CE",
            "CZ",
            "CH",
        ],
    ),
    "AIB": ("ALA", ["N", "CA", "C", "O", "CB1", "CB2"]),
    "PGY": ("GLY", ["N", "CA", "CB", "CG1", "CG2", "CD1", "CD2", "CE", "C", "O"]),
    "ABA": ("ALA", ["N", "CA", "C", "O", "CB", "CG"]),
    "0UO": (
        "TRP",
        [
            "O3",
            "C11",
            "C13",
            "C14",
            "C15",
            "C16",
            "C17",
            "C18",
            "C19",
            "C",
            "CA",
            "C9",
            "C10",
            "N",
            "O",
            "N6",
        ],
    ),
    "143": (
        "CYS",
        [
            "N",
            "CA",
            "CB",
            "SG",
            "C",
            "O",
            "CI",
            "NI",
            "CJ",
            "OJ1",
            "OJ2",
            "CF",
            "OF",
            "NX",
            "CD",
            "CE",
        ],
    ),
    "HIS": ("HIS", ["N", "CA", "C", "O", "CB", "CG", "ND1", "CD2", "CE1", "NE2"]),
    "1X6": ("SER", ["N", "CA", "C", "O", "CB", "OG", "C7", "O8", "C6", "C5", "N4"]),
    "PG1": (
        "SER",
        [
            "S1",
            "C2",
            "C3",
            "N4",
            "C5",
            "C6",
            "C7",
            "O8",
            "C9",
            "C10",
            "C11",
            "O12",
            "O13",
            "N14",
            "C15",
            "O16",
            "C17",
            "C18",
            "C19",
            "C20",
            "C21",
            "C22",
            "C23",
            "N",
            "CA",
            "CB",
            "OG",
            "C",
            "O",
        ],
    ),
    "2LU": ("LEU", ["N", "CA", "C", "CB", "CG", "CD1", "CD2"]),
    "NZH": (
        "HIS",
        [
            "N",
            "CA",
            "CB",
            "CG",
            "ND1",
            "CE1",
            "NE2",
            "CD2",
            "C",
            "O",
            "N1",
            "N2",
            "N3",
            "N4",
            "C1",
        ],
    ),
    "AZK": ("LYS", ["N", "CA", "CB", "CG", "CD", "CE", "NZ", "NZ2", "NZ3", "C", "O"]),
    "STY": (
        "TYR",
        [
            "N",
            "CA",
            "CB",
            "CG",
            "CD1",
            "CE1",
            "CD2",
            "CE2",
            "CZ",
            "OH",
            "C",
            "O",
            "S",
            "O2",
            "O3",
            "O4",
        ],
    ),
    "A9D": (
        "SER",
        [
            "C",
            "N",
            "O",
            "C1",
            "O2",
            "C3",
            "C4",
            "C5",
            "C6",
            "N7",
            "C8",
            "O9",
            "CA",
            "CB",
            "OG",
            "C10",
            "C11",
            "C12",
            "C13",
            "N14",
            "C15",
            "O16",
            "O17",
        ],
    ),
    "23F": ("PHE", ["N", "CA", "C", "CB", "CG", "CD1", "CD2", "CE1", "CE2", "CZ", "O"]),
    "BHD": ("ASP", ["N", "CA", "CB", "OB", "CG", "OD1", "OD2", "C", "O"]),
    "BTK": (
        "LYS",
        [
            "C",
            "N",
            "O",
            "CA",
            "CB",
            "CD",
            "CE",
            "CG",
            "NZ",
            "CAA",
            "OAD",
            "CAF",
            "CAJ",
            "CAN",
        ],
    ),
    "1AC": ("ALA", ["CB", "CG", "CA", "C", "O", "N"]),
    "NCB": ("ALA", ["N", "CA", "C", "O", "CB", "O3", "C3", "N2"]),
    "TFQ": (
        "PHE",
        [
            "F16",
            "C14",
            "F17",
            "F15",
            "C13",
            "C1",
            "C6",
            "C5",
            "C2",
            "C3",
            "C4",
            "C7",
            "C8",
            "N9",
            "C10",
            "O12",
            "O11",
        ],
    ),
    "GHW": (
        "GLU",
        [
            "C4",
            "C5",
            "C6",
            "C13",
            "N1",
            "N2",
            "C3",
            "N3",
            "S4",
            "C2",
            "O6",
            "C1",
            "SAT",
            "CAZ",
            "CAJ",
            "CAL",
            "CAI",
            "CAK",
            "CBA",
            "CAX",
            "OAE",
            "NAS",
            "CAP",
            "CAW",
            "OAG",
            "OAD",
            "CBG",
            "CAN",
            "CAV",
            "OAC",
        ],
    ),
    "LYS": ("LYS", ["N", "CA", "C", "O", "CB", "CG", "CD", "CE", "NZ"]),
    "GHG": ("GLN", ["N", "CA", "C", "O", "CB", "CG", "OG1", "CD", "OE1", "NE2"]),
    "TY8": (
        "TYR",
        [
            "C",
            "N",
            "C1",
            "C2",
            "O2",
            "C3",
            "C4",
            "O4",
            "C5",
            "O5",
            "C6",
            "CA",
            "CB",
            "O2A",
            "O4A",
            "O",
        ],
    ),
    "TY9": (
        "TYR",
        [
            "C",
            "N",
            "O",
            "C1",
            "C2",
            "C3",
            "O3",
            "C4",
            "O4",
            "C5",
            "O5",
            "C6",
            "CA",
            "CB",
            "O3A",
            "O4A",
        ],
    ),
    "GHC": (
        "GLU",
        [
            "C4",
            "C5",
            "C6",
            "C13",
            "N1",
            "N2",
            "C3",
            "N3",
            "C29",
            "S4",
            "C2",
            "O6",
            "C1",
            "SAT",
            "CAZ",
            "CAJ",
            "CAL",
            "CAI",
            "CAK",
            "CBA",
            "CAX",
            "OAE",
            "NAS",
            "CAP",
            "CAW",
            "OAG",
            "OAD",
            "CBG",
            "CAN",
            "CAV",
            "OAF",
            "OAC",
        ],
    ),
    "TY5": (
        "TYR",
        [
            "O",
            "C",
            "CA",
            "N",
            "CB",
            "CG",
            "CD2",
            "CE2",
            "CZ",
            "CE1",
            "CD1",
            "OH",
            "C49",
            "C50",
            "C51",
            "C52",
            "C53",
            "C54",
            "C55",
        ],
    ),
    "MIS": (
        "SER",
        ["N", "CA", "CB", "OG", "P", "O1P", "O2P", "O3P", "C1", "C2", "C3", "C", "O"],
    ),
    "MIR": (
        "SER",
        ["N", "CA", "CB", "OG", "P", "O1P", "O2P", "O3P", "C1", "C2", "C", "O"],
    ),
    "HS8": (
        "HIS",
        [
            "N",
            "CA",
            "CB",
            "CG",
            "ND1",
            "CE1",
            "NE2",
            "CD2",
            "C",
            "O",
            "O3",
            "S",
            "O1",
            "O2",
        ],
    ),
    "TY2": (
        "TYR",
        [
            "N",
            "CA",
            "C",
            "O",
            "CB",
            "CG",
            "CD1",
            "CD2",
            "CE1",
            "CE2",
            "CZ",
            "OH",
            "NE2",
        ],
    ),
    "TY3": (
        "TYR",
        [
            "N",
            "CA",
            "C",
            "O",
            "CB",
            "CG",
            "CD1",
            "CD2",
            "CE1",
            "CE2",
            "CZ",
            "OH",
            "OE2",
        ],
    ),
    "QPH": (
        "PHE",
        ["C", "N", "O", "CA", "CB", "CG", "CZ", "CD1", "CD2", "CE1", "CE2", "O1", "CN"],
    ),
    "SAH": (
        "CYS",
        [
            "N",
            "CA",
            "CB",
            "CG",
            "SD",
            "C",
            "O",
            "C5'",
            "C4'",
            "O4'",
            "C3'",
            "O3'",
            "C2'",
            "O2'",
            "C1'",
            "N9",
            "C8",
            "N7",
            "C5",
            "C6",
            "N6",
            "N1",
            "C2",
            "N3",
            "C4",
        ],
    ),
    "SAC": ("SER", ["C1A", "C2A", "OAC", "N", "CA", "C", "O", "CB", "OG"]),
    "02L": (
        "ASN",
        [
            "CA",
            "C",
            "N",
            "CB",
            "CG",
            "OD1",
            "ND2",
            "C36",
            "C37",
            "C40",
            "C38",
            "C39",
            "O",
        ],
    ),
    "PRO": ("PRO", ["N", "CA", "C", "O", "CB", "CG", "CD"]),
    "QPA": (
        "CYS",
        [
            "C",
            "N",
            "O",
            "P2",
            "C8",
            "C9",
            "O9",
            "CA",
            "CB",
            "SG",
            "C10",
            "O10",
            "O11",
            "O12",
            "O13",
            "O14",
        ],
    ),
    "6CL": (
        "LYS",
        ["N", "CA", "CB", "CG", "CD", "CE", "CZ", "OH1", "OH2", "NZ", "C", "O"],
    ),
    "PRK": (
        "LYS",
        ["C", "N", "O", "CA", "CB", "CD", "CE", "CG", "NZ", "CAA", "OAD", "CAF", "CAL"],
    ),
    "PRJ": (
        "PRO",
        ["O", "N", "C", "CA", "CB", "CG", "C11", "C12", "C13", "C14", "CD", "O2"],
    ),
    "FGL": ("GLY", ["N", "CA", "C", "O", "OG1", "OG2", "CB"]),
    "PRV": (
        "GLY",
        ["C", "N", "O", "N1", "O1", "O2", "CA", "CG", "CZ", "CD1", "CD2", "CE1", "CE2"],
    ),
    "P1L": (
        "CYS",
        [
            "O",
            "N",
            "CA",
            "CB",
            "SG",
            "C7",
            "O7",
            "C8",
            "C9",
            "C10",
            "C11",
            "C12",
            "C13",
            "C14",
            "C15",
            "C16",
            "C17",
            "C18",
            "C19",
            "C20",
            "C",
            "C21",
            "C22",
        ],
    ),
    "PRS": ("PRO", ["N", "CD", "CA", "CB", "SG", "C", "O"]),
    "CH7": (
        "LYS",
        [
            "NZ",
            "CA1",
            "CB",
            "CG",
            "CD",
            "CE",
            "C1",
            "N2",
            "OH",
            "CD2",
            "CE2",
            "CZ",
            "CE1",
            "CD1",
            "CG2",
            "CB2",
            "CA2",
            "C2",
            "O2",
            "N3",
            "CA3",
            "C3",
            "O3",
        ],
    ),
    "AZY": (
        "TYR",
        [
            "OXT",
            "C",
            "O",
            "CA",
            "N",
            "CB",
            "CG",
            "CD2",
            "CD1",
            "CE1",
            "CZ",
            "OH",
            "CE2",
            "N1",
            "N2",
            "N3",
        ],
    ),
    "SAR": ("GLY", ["N", "CA", "C", "O", "CN"]),
    "02Y": ("ALA", ["N", "CA", "C2", "C3", "C4", "O", "O2", "C6", "N2", "N3", "C1"]),
    "LAL": ("ALA", ["CB", "C", "O", "N", "CA", "CH1", "CH2"]),
    "BFD": (
        "ASP",
        ["N", "CA", "C", "O", "CB", "CG", "OD1", "OD2", "BE", "F1", "F2", "F3"],
    ),
    "TPO": ("THR", ["N", "CA", "CB", "CG2", "OG1", "P", "O1P", "O2P", "O3P", "C", "O"]),
    "H5M": ("PRO", ["N", "CA", "C", "O", "CB", "CG", "CD", "OB", "CD1"]),
    "AYG": (
        "ALA",
        [
            "CA1",
            "CA2",
            "CA3",
            "CB1",
            "CB2",
            "CG2",
            "CD1",
            "CD2",
            "CE1",
            "CE2",
            "CZ",
            "OH",
            "C1",
            "N1",
            "C2",
            "N2",
            "O2",
            "C3",
            "N3",
            "O3",
        ],
    ),
    "TPL": (
        "TRP",
        [
            "N",
            "CA",
            "CB1",
            "CG",
            "CD1",
            "CD2",
            "CE3",
            "CE2",
            "NE1",
            "CZ3",
            "CZ2",
            "CH2",
            "C",
        ],
    ),
    "AYA": ("ALA", ["N", "CA", "CB", "C", "O", "CT", "OT", "CM"]),
    "HRP": (
        "TRP",
        [
            "N",
            "CA",
            "CB",
            "CG",
            "CD1",
            "NE1",
            "CE2",
            "CZ2",
            "CH2",
            "CZ3",
            "CE3",
            "CD2",
            "C",
            "O",
            "OX",
        ],
    ),
    "4FB": ("PRO", ["O", "C", "CA", "CBX", "CGX", "FGX", "CDX", "N"]),
    "CHP": (
        "GLY",
        ["N", "CA", "C", "O", "C1", "C2", "C3", "CL3", "C4", "O4", "C5", "C6"],
    ),
    "PR4": (
        "PRO",
        [
            "C",
            "N",
            "O",
            "CA",
            "CB",
            "CD",
            "CG",
            "CAB",
            "CAC",
            "CAD",
            "OAE",
            "OAF",
            "OAG",
            "O5",
        ],
    ),
    "PR7": (
        "PRO",
        [
            "C",
            "N",
            "O",
            "CA",
            "CB",
            "CD",
            "CG",
            "CAB",
            "CAC",
            "CAD",
            "OAE",
            "OAF",
            "OAG",
            "O5",
        ],
    ),
    "Z3E": (
        "THR",
        [
            "C",
            "N",
            "O",
            "C1",
            "C2",
            "C3",
            "C4",
            "C5",
            "C6",
            "C7",
            "CA",
            "CB",
            "OG1",
            "CG2",
        ],
    ),
    "PR3": ("CYS", ["N", "CA", "C", "O", "CB", "SG", "SD", "CE", "CZ", "CH"]),
    "IIL": ("ILE", ["N", "CA", "C", "O", "CB", "CG2", "CG1", "CD1"]),
    "Z70": (
        "HIS",
        ["CA", "C", "C3", "C4", "C5", "C6", "C7", "N", "N2", "N3", "N4", "O"],
    ),
    "0NC": ("ALA", ["N", "CA", "C", "O", "CB", "N1", "C1"]),
    "LPD": ("PRO", ["O", "C", "N2", "CA", "N", "CD", "CG", "CB"]),
    "PR9": ("PRO", ["O", "C", "CA", "NXT", "N", "CD", "CG", "CB"]),
    "SNC": ("CYS", ["N", "CA", "CB", "SG", "ND", "OE", "C", "O"]),
    "PYX": (
        "CYS",
        [
            "N",
            "CA",
            "C",
            "O",
            "CB",
            "SG",
            "SD",
            "CE",
            "C1",
            "C2",
            "C6",
            "N3",
            "C5",
            "CM6",
            "C4",
            "O5",
            "N6",
            "CM4",
        ],
    ),
    "TYN": (
        "TYR",
        [
            "N",
            "CA",
            "CB",
            "CG",
            "CD1",
            "CE1",
            "CD2",
            "CE2",
            "CZ",
            "OC",
            "C",
            "O",
            "O1",
            "N1",
            "N2",
            "N3",
            "C1",
            "C2",
            "C3",
            "C4",
            "C5",
            "C6",
        ],
    ),
    "TYO": (
        "TYR",
        [
            "N",
            "CA",
            "CB",
            "CG",
            "CD1",
            "CE1",
            "OE1",
            "CD2",
            "CE2",
            "CZ",
            "OH",
            "OH1",
            "C",
            "O",
        ],
    ),
    "IPG": ("GLY", ["N", "CA", "C", "O", "CN", "CG1", "CG2"]),
    "TYJ": (
        "TYR",
        [
            "N",
            "CA",
            "C",
            "O",
            "CN",
            "CB7",
            "CG5",
            "CD12",
            "CD21",
            "CE11",
            "CE21",
            "CZ1",
            "OZ",
            "OH1",
            "OG1",
        ],
    ),
    "TYB": (
        "TYR",
        ["N", "CA", "C", "O", "CB", "CG", "CD1", "CD2", "CE1", "CE2", "CZ", "OH"],
    ),
    "DBY": (
        "TYR",
        [
            "N",
            "CA",
            "C",
            "O",
            "CB",
            "CG",
            "CD1",
            "CE1",
            "CD2",
            "CE2",
            "CZ",
            "OH",
            "BR1",
            "BR2",
        ],
    ),
    "SNN": ("ASN", ["N1", "C2", "C3", "N3", "C4", "C5", "O2", "O5"]),
    "API": (
        "LYS",
        ["C1", "C2", "C3", "C4", "C5", "C6", "C7", "O1", "O2", "O3", "O4", "N2", "N6"],
    ),
    "C3Y": ("CYS", ["N", "CA", "C", "O", "CB", "SG", "C1", "O2", "N1", "O3", "C2"]),
    "TYY": (
        "TYR",
        [
            "N",
            "CA",
            "C",
            "O",
            "CB",
            "CG",
            "CD1",
            "CD2",
            "CE1",
            "CE2",
            "CZ",
            "OZ",
            "N5",
            "OH",
        ],
    ),
    "TYT": (
        "TYR",
        [
            "N1",
            "C1",
            "C2",
            "O1",
            "O2",
            "C3",
            "C4",
            "C5",
            "O3",
            "C6",
            "C7",
            "O4",
            "C8",
            "N2",
            "C9",
            "C10",
            "C11",
            "C12",
            "C13",
            "C14",
            "C15",
            "C16",
            "C17",
            "C18",
        ],
    ),
    "PYA": (
        "ALA",
        [
            "C1",
            "C2",
            "C3",
            "N4",
            "C4A",
            "C4B",
            "N5",
            "C6",
            "C7",
            "C8",
            "C8A",
            "C9",
            "C10",
            "C1A",
            "CB",
            "CA",
            "N",
            "C",
            "O",
        ],
    ),
    "DBS": (
        "SER",
        [
            "O4",
            "C4",
            "C7",
            "C1",
            "O1",
            "C16",
            "C13",
            "C10",
            "C19",
            "O7",
            "N1",
            "C22",
            "C28",
            "O13",
            "C25",
            "O10",
            "O15",
        ],
    ),
    "TYW": (
        "TYR",
        ["C", "N", "O", "CA", "CB", "CG", "OH", "CZ", "CD1", "CD2", "CE1", "CE2"],
    ),
    "DBU": ("THR", ["N", "CA", "CB", "CG", "C", "O"]),
    "TYQ": (
        "TYR",
        [
            "N",
            "CA",
            "C",
            "O",
            "CB",
            "CG",
            "CD1",
            "CD2",
            "CE1",
            "CE2",
            "CZ",
            "OZ",
            "N5",
            "OH",
        ],
    ),
    "TYR": (
        "TYR",
        ["N", "CA", "C", "O", "CB", "CG", "CD1", "CD2", "CE1", "CE2", "CZ", "OH"],
    ),
    "TYS": (
        "TYR",
        [
            "N",
            "CA",
            "CB",
            "CG",
            "CD1",
            "CD2",
            "CE1",
            "CE2",
            "CZ",
            "OH",
            "S",
            "O1",
            "O2",
            "O3",
            "C",
            "O",
        ],
    ),
    "KST": (
        "LYS",
        [
            "O2",
            "C9",
            "O1",
            "C8",
            "S",
            "C6",
            "C7",
            "C5",
            "N1",
            "C4",
            "C3",
            "C2",
            "C1",
            "CA",
            "N",
            "C",
            "O",
        ],
    ),
    "LBY": (
        "LYS",
        [
            "C1",
            "CT",
            "C2",
            "C3",
            "O2",
            "CZ",
            "O1",
            "NZ",
            "CE",
            "CD",
            "CG",
            "CB",
            "CA",
            "C",
            "O",
            "N",
        ],
    ),
    "CZ2": ("CYS", ["N", "CA", "CB", "SG", "C", "O", "AS", "O1", "O2"]),
    "SVX": (
        "SER",
        ["N", "CA", "C", "CB", "O", "OG", "P1", "C4", "C2", "O6", "O5", "C1"],
    ),
    "LAA": ("ASP", ["N", "N1", "CA", "C", "O", "CB", "CG", "OD1", "OD2", "O1"]),
    "OCY": ("CYS", ["N", "CA", "CB", "SG", "CD", "CE", "OZ", "C", "O"]),
    "P2Q": (
        "TYR",
        [
            "C",
            "N",
            "O",
            "C1",
            "C2",
            "C3",
            "C4",
            "C5",
            "C6",
            "C7",
            "C8",
            "C9",
            "CA",
            "CB",
            "CG",
            "CZ",
            "C10",
            "C11",
            "OAD",
            "NAS",
            "OAT",
            "OAZ",
            "OC5",
            "CD1",
            "CD2",
            "CE1",
            "CE2",
        ],
    ),
    "OCS": ("CYS", ["N", "CA", "CB", "SG", "C", "O", "OD1", "OD2", "OD3"]),
    "P2Y": ("PRO", ["NA", "C", "CA", "N", "CD", "CG", "CB"]),
    "0Y8": (
        "PRO",
        [
            "C1",
            "N2",
            "C3",
            "C4",
            "C5",
            "C6",
            "C7",
            "C8",
            "C9",
            "C10",
            "BR1",
            "O12",
            "C13",
            "C14",
            "C15",
            "N16",
            "C17",
            "C18",
            "O19",
        ],
    ),
    "NLP": ("LEU", ["N", "CA", "CB", "CG", "CD", "CE", "P", "O1", "O2", "O3"]),
    "CEA": ("CYS", ["N", "CA", "CB", "SG", "O1", "O", "C"]),
    "2TY": (
        "TYR",
        [
            "N",
            "CA",
            "CB",
            "C",
            "O",
            "CG",
            "CD1",
            "OX1",
            "CE1",
            "CD2",
            "CE2",
            "NX1",
            "CZ",
            "OH",
            "C1'",
            "C6'",
            "C5'",
            "C4'",
            "C3'",
            "C2'",
            "C2",
            "C1",
        ],
    ),
    "FFM": (
        "CYS",
        [
            "C",
            "N",
            "O",
            "C1",
            "O1",
            "P1",
            "C2",
            "O2",
            "C3",
            "O3",
            "O4",
            "CA",
            "CB",
            "SG",
        ],
    ),
    "HMR": (
        "ARG",
        ["N", "CA", "CB", "CG", "CD", "NE", "CZ", "NH1", "NH2", "C", "O", "CH2"],
    ),
    "THC": ("THR", ["N", "CN", "ON", "CM", "CA", "CB", "OG1", "CG2", "C", "O"]),
    "UGY": ("GLY", ["OXT", "C", "O", "CA", "N", "NB", "CG", "OE", "NE"]),
    "NLE": ("LEU", ["N", "CA", "C", "O", "CB", "CG", "CD", "CE"]),
    "BMT": (
        "THR",
        ["N", "CN", "CA", "C", "O", "CB", "OG1", "CG2", "CD1", "CD2", "CE", "CZ", "CH"],
    ),
    "CS4": (
        "CYS",
        [
            "OXT",
            "C",
            "CA",
            "N",
            "CB",
            "SG",
            "C1J",
            "C1K",
            "C1N",
            "O1C",
            "C1Q",
            "C1I",
            "C1H",
            "C1G",
            "C1O",
            "CL1D",
            "C1P",
            "CL1E",
            "O",
        ],
    ),
    "CS1": (
        "CYS",
        [
            "O",
            "C",
            "CA",
            "N",
            "CB",
            "SG",
            "SD",
            "CE",
            "C5",
            "C4",
            "C3",
            "C2",
            "C1",
            "N1",
        ],
    ),
    "CS0": ("CYS", ["N", "CA", "C", "O", "CB", "SG", "C1", "C2", "O2"]),
    "NLO": ("LEU", ["N", "CA", "C", "O", "CB", "CG", "CD", "CE", "OH", "C1"]),
    "NLN": ("LEU", ["N", "CA", "C", "O", "CB", "CG", "CD", "CE", "NH2"]),
    "SHP": (
        "GLY",
        [
            "N",
            "CA",
            "C",
            "O",
            "C1B",
            "C2B",
            "C3B",
            "C4B",
            "O4B",
            "C5B",
            "C6B",
            "C1",
            "C2",
            "C3",
            "C4",
            "C5",
            "C6",
            "O1",
            "O3",
            "O4",
            "O5",
            "O6",
            "C1'",
            "C2'",
            "C3'",
            "C4'",
            "C5'",
            "C6'",
            "O2'",
            "O3'",
            "O4'",
            "O5'",
            "O6'",
        ],
    ),
    "SHR": (
        "LYS",
        [
            "N1",
            "C10",
            "C9",
            "C8",
            "C7",
            "C6",
            "N",
            "C11",
            "O5",
            "O6",
            "C4",
            "C3",
            "C2",
            "C1",
            "O1",
            "O2",
            "C5",
            "O3",
            "O4",
        ],
    ),
    "B3X": ("ASN", ["OE1", "CD", "NE2", "CG", "CB", "N", "CA", "C", "O"]),
    "B3Y": (
        "TYR",
        ["O", "C", "CA", "CB", "N", "CG", "CD", "CE2", "CF2", "CZ", "OH", "CF1", "CE1"],
    ),
    "CSJ": (
        "CYS",
        [
            "N",
            "CA",
            "C",
            "O",
            "CB",
            "SG",
            "CO",
            "O2",
            "C1",
            "C2",
            "C3",
            "N2",
            "C4",
            "C5",
            "C6",
        ],
    ),
    "CSE": ("CYS", ["N", "CA", "CB", "SE", "C", "O"]),
    "B3U": ("HIS", ["N", "CA", "CB", "CG", "CD2", "NE2", "CE1", "ND1", "NT", "C", "O"]),
    "CSA": ("CYS", ["N", "CA", "CB", "C", "SG", "C3", "C2", "O4", "C1", "O"]),
    "AFA": (
        "ASN",
        [
            "N",
            "CA",
            "CB",
            "CG",
            "OD1",
            "ND2",
            "C",
            "O",
            "C1",
            "O1",
            "C2",
            "C3",
            "C4",
            "C5",
            "C6",
            "C7",
            "C8",
            "C9",
        ],
    ),
    "CSB": ("CYS", ["N", "CA", "CB", "SG", "PB", "C", "O"]),
    "DLS": (
        "LYS",
        [
            "C2",
            "O1",
            "C1",
            "N",
            "CA",
            "C",
            "O",
            "CB",
            "CG",
            "CD",
            "CE",
            "NZ",
            "CH3",
            "OH",
            "CH",
        ],
    ),
    "SHC": (
        "CYS",
        ["N", "CA", "CB", "SG", "C1", "C2", "C3", "C4", "C5", "C6", "C", "O"],
    ),
    "CSX": ("CYS", ["N", "CA", "CB", "SG", "C", "O", "OD"]),
    "B3K": ("LYS", ["N", "CB", "CG", "CD", "CE", "CF", "NZ", "CA", "C", "O"]),
    "B3D": ("ASP", ["OE1", "CD", "OE2", "CG", "CB", "N", "CA", "C", "O"]),
    "B3E": ("GLU", ["N", "CB", "CG", "CD", "CE", "OF2", "OF1", "CA", "C", "O"]),
    "CSW": ("CYS", ["N", "CA", "CB", "SG", "C", "O", "OD1", "OD2"]),
    "B3A": ("ALA", ["CG", "CB", "N", "CA", "C", "O"]),
    "CSS": ("CYS", ["N", "CA", "CB", "SG", "SD", "C", "O"]),
    "2HF": ("HIS", ["C", "N", "O", "F1", "CA", "CB", "CG", "ND1", "CD2", "CE1", "NE2"]),
    "CZZ": ("CYS", ["N", "CA", "CB", "SG", "AS", "O1", "C", "O"]),
    "N10": (
        "SER",
        [
            "C1",
            "O2",
            "N",
            "CA",
            "CB",
            "OG",
            "N1",
            "CA1",
            "CB2",
            "CG3",
            "CD4",
            "CG5",
            "CD6",
            "C",
            "O",
        ],
    ),
    "MGG": (
        "ARG",
        [
            "N",
            "CA",
            "CB",
            "CG",
            "CD",
            "NE",
            "CZ",
            "NH1",
            "NH2",
            "C3",
            "O3",
            "C2",
            "C1",
            "O12",
            "O1",
            "C",
            "O",
        ],
    ),
    "BTR": (
        "TRP",
        [
            "N",
            "CA",
            "C",
            "O",
            "CB",
            "CG",
            "CD1",
            "NE1",
            "CE2",
            "CD2",
            "CE3",
            "CZ3",
            "CH2",
            "BR2",
            "CZ2",
        ],
    ),
    "SEP": ("SER", ["N", "CA", "CB", "OG", "C", "O", "P", "O1P", "O2P", "O3P"]),
    "XCN": ("CYS", ["C", "N", "O", "CA", "CB", "NC", "SG", "CS"]),
    "MGN": ("GLN", ["N", "CA", "CB1", "CB2", "CG", "CD", "OE1", "NE2", "C", "O"]),
    "C1X": (
        "LYS",
        [
            "O4",
            "C11",
            "O3",
            "C10",
            "C9",
            "C8",
            "C7",
            "C5",
            "C3",
            "C2",
            "C1",
            "O1",
            "O2",
            "C4",
            "NZ",
            "CE",
            "CD",
            "CG",
            "CB",
            "CA",
            "N",
            "C",
            "O",
            "S6",
        ],
    ),
    "BTA": ("LEU", ["N", "CA", "CB", "CG", "CD", "C", "O"]),
    "BTC": ("CYS", ["N", "CA", "CB", "SG", "C", "O"]),
    "MGY": ("GLY", ["N", "CA", "C", "O", "CN"]),
    "T0I": (
        "TYR",
        [
            "C14",
            "C15",
            "N11",
            "C12",
            "N13",
            "C5",
            "C6",
            "C4",
            "O4",
            "C3",
            "C2",
            "O2",
            "C1",
            "CB",
            "CA",
            "C",
            "O",
            "N",
        ],
    ),
    "TBG": ("VAL", ["N", "CA", "CB", "CG1", "CG2", "CG3", "C", "O"]),
    "EME": ("GLU", ["C7", "N2", "CA", "C", "O", "CB", "CG", "CD", "OE2", "OE1"]),
    "R4K": (
        "TRP",
        [
            "C50",
            "C51",
            "C52",
            "C53",
            "C54",
            "C55",
            "N31",
            "C32",
            "O33",
            "N",
            "CA",
            "C",
            "O",
            "CB",
            "O34",
            "C29",
            "O30",
        ],
    ),
    "12L": (
        "PRO",
        ["N19", "C20", "C21", "C22", "C23", "C24", "C25", "C26", "C27", "O28"],
    ),
    "PEC": (
        "CYS",
        ["N", "CA", "CB", "SG", "SD", "C1", "C2", "C3", "C4", "C5", "C", "O"],
    ),
    "HV5": ("ALA", ["N", "CA", "CB", "CG1", "CG2", "CG3", "C", "O"]),
    "IT1": (
        "LYS",
        [
            "C2",
            "C3",
            "C4",
            "C5",
            "C6",
            "N",
            "CA",
            "C",
            "O",
            "CB",
            "CG",
            "CD",
            "CE",
            "NZ",
            "OP1",
            "P",
            "OP2",
            "OP3",
            "OP4",
            "C5A",
            "C4A",
            "O3",
            "N1",
            "C2A",
        ],
    ),
    "SEG": ("ALA", ["N", "OD", "CB", "CA", "C", "O"]),
    "ASQ": (
        "ASP",
        ["N", "CA", "C", "O", "CB", "CG", "OD1", "OD2", "P", "O1P", "O2P", "O3P"],
    ),
    "ASP": ("ASP", ["N", "CA", "C", "O", "CB", "CG", "OD1", "OD2"]),
    "ECC": ("GLN", ["C", "N", "O", "CA", "CB", "CD", "CG", "OE1", "NE2"]),
    "SER": ("SER", ["N", "CA", "C", "O", "CB", "OG"]),
    "ASB": (
        "ASP",
        ["N", "CA", "C", "O", "CB", "CG", "OD1", "OD2", "C2", "C1", "O1", "O2"],
    ),
    "CMH": ("CYS", ["N", "CA", "C", "O", "CB", "SG", "CM", "HG"]),
    "ASA": ("ASP", ["N", "CA", "C", "O", "CB", "CG", "OD1", "OD2"]),
    "LNM": (
        "LEU",
        ["N13", "C14", "C15", "N16", "C17", "C22", "O23", "C27", "C28", "C29"],
    ),
    "0A8": (
        "CYS",
        ["N", "CA", "C", "O", "CB", "SG", "C1", "O1", "N1", "C2", "C3", "CL1"],
    ),
    "SEE": ("SER", ["N", "CA", "CB", "OG", "B", "O1B", "O2B", "O3B", "C", "O"]),
    "SEC": ("ALA", ["N", "CA", "C", "O", "CB", "SEG", "OD1", "OD2"]),
    "HPE": (
        "PHE",
        ["N", "CA", "C", "O", "CB", "CG", "CD", "CE1", "CE2", "CZ1", "CZ2", "CH"],
    ),
    "BB9": ("CYS", ["N", "CA", "C", "O", "CB", "SG"]),
    "BB8": (
        "PHE",
        ["N", "CA", "C", "CB", "OB", "CG", "CD1", "CE1", "CZ", "CE2", "CD2", "O"],
    ),
    "LNE": (
        "LEU",
        ["N9", "C10", "C11", "N12", "C13", "C14", "C15", "C16", "C17", "C18", "O25"],
    ),
    "GT9": (
        "CYS",
        [
            "N",
            "CA",
            "CB",
            "SG",
            "C",
            "O",
            "C1",
            "C2",
            "C3",
            "C4",
            "C5",
            "C6",
            "C7",
            "C8",
            "C9",
        ],
    ),
    "SEM": (
        "SER",
        [
            "C",
            "O",
            "CA",
            "N",
            "CB",
            "OG",
            "C'",
            "C1'",
            "C2'",
            "C3'",
            "C4'",
            "C5'",
            "C6'",
        ],
    ),
    "SEL": ("SER", ["N", "CA", "CB", "OG", "C"]),
    "ASN": ("ASN", ["N", "CA", "C", "O", "CB", "CG", "OD1", "ND2"]),
    "ECX": ("CYS", ["C", "N", "O", "CA", "CB", "SC", "CD", "CE"]),
    "AS7": (
        "ASN",
        [
            "C",
            "N",
            "O",
            "C1",
            "C2",
            "C3",
            "C4",
            "C5",
            "C6",
            "C7",
            "C8",
            "N9",
            "CA",
            "CB",
            "CG",
            "OD1",
            "ND2",
        ],
    ),
    "CLE": ("LEU", ["N", "CA", "CB", "CG", "CD1", "CD2", "C", "O", "N2"]),
    "TLY": (
        "LYS",
        ["C", "N", "O", "CA", "CB", "CD", "CE", "CG", "CX", "SX", "CY", "NZ"],
    ),
    "CLG": (
        "LYS",
        [
            "N",
            "CA",
            "CB",
            "CG",
            "CD",
            "CE",
            "NZ",
            "CH",
            "OI",
            "CI",
            "NJ",
            "CK",
            "OL",
            "C",
            "O",
            "CL",
            "OM",
            "NM",
        ],
    ),
    "BBC": (
        "CYS",
        ["N", "CA", "CB", "SG", "C05", "C", "O", "C08", "C09", "C10", "N11", "O12"],
    ),
    "IML": ("ILE", ["N", "CA", "C", "O", "CB", "CN", "CG2", "CG1", "CD1"]),
    "BB7": ("CYS", ["N", "CA", "C", "O", "CB", "SG", "CB1", "OB2", "CB3"]),
    "OXX": (
        "ASP",
        ["N", "CA", "CB", "CG", "OD2", "OD1", "C2", "O3", "C1", "O1", "O2", "C", "O"],
    ),
    "PVH": (
        "HIS",
        ["N", "CA", "C", "O", "CB", "CG", "ND1", "CD2", "CE1", "NE2", "OE", "CM"],
    ),
    "ASI": ("ASP", ["N", "CA", "C1", "O1", "O2", "C2", "C", "O"]),
    "4BF": (
        "TYR",
        ["CD1", "CE1", "CZ", "BR", "CE2", "CD2", "CG", "CB", "CA", "N", "C", "O"],
    ),
    "HVA": ("VAL", ["N", "CA", "CB", "CG1", "CG2", "OG3", "C", "O"]),
    "OMH": ("SER", ["O", "C", "CA", "N", "CB", "OG", "P", "O3", "O2", "O1", "C31"]),
    "3O3": (
        "GLU",
        [
            "N2",
            "N3",
            "O3",
            "C4",
            "C5",
            "O5",
            "C6",
            "O6",
            "C7",
            "O7",
            "C8",
            "O8",
            "C9",
            "C10",
            "C11",
            "C12",
        ],
    ),
    "OMT": ("MET", ["N", "CA", "CB", "CG", "SD", "CE", "C", "O", "OD1", "OD2"]),
    "OMY": (
        "TYR",
        [
            "N",
            "CA",
            "OCZ",
            "CE2",
            "CE1",
            "CZ",
            "CG",
            "CD2",
            "CD1",
            "CB",
            "CL",
            "O",
            "C",
            "ODE",
        ],
    ),
    "ASL": (
        "ASP",
        ["N", "CA", "C", "O", "CB", "CG", "OD1", "OD2", "C3", "C2", "C1", "O1", "O2"],
    ),
    "2MR": (
        "ARG",
        ["N", "CA", "CB", "CG", "CD", "NE", "CZ", "NH1", "CQ1", "NH2", "CQ2", "C", "O"],
    ),
    "BPE": ("CYS", ["N", "CA", "CB", "SG", "C05", "C", "O", "C08", "C09", "N10"]),
    "2MT": ("PRO", ["N", "CA", "C", "O", "CB", "SG", "CD1", "CD2", "CD3"]),
    "3XH": (
        "GLY",
        [
            "O11",
            "C10",
            "O12",
            "C9",
            "N8",
            "C7",
            "O14",
            "C6",
            "C5",
            "C4",
            "C3",
            "C2",
            "O13",
            "C1",
        ],
    ),
    "AAR": (
        "ARG",
        ["N", "CA", "CB", "CG", "CD", "NE", "CZ", "NH1", "NH2", "C", "O", "NT"],
    ),
    "CLH": (
        "LYS",
        [
            "N",
            "CA",
            "CB",
            "CG",
            "CD",
            "CE",
            "NZ",
            "CH",
            "OI",
            "CI",
            "NJ",
            "CK",
            "OL",
            "C",
            "O",
            "CL",
            "OM",
        ],
    ),
    "G01": (
        "GLU",
        ["CD", "CG", "CB", "CA", "C", "C6", "P", "OPA", "OPB", "N", "O", "OE1", "OE2"],
    ),
    "2ML": ("LEU", ["N", "CA", "CB1", "CB2", "CG", "CD1", "CD2", "C", "O"]),
    "CYS": ("CYS", ["N", "CA", "C", "O", "CB", "SG"]),
    "CYR": (
        "CYS",
        [
            "O1",
            "C1",
            "O2",
            "C2",
            "N2",
            "C3",
            "C4",
            "C5",
            "N6",
            "C7",
            "N7",
            "SG",
            "CB",
            "CA",
            "C",
            "O",
            "N",
        ],
    ),
    "CYQ": ("CYS", ["N", "CA", "CB", "SG", "CD", "C", "O", "P", "O1P", "O2P", "O3P"]),
    "CMT": ("CYS", ["N", "CA", "C", "O", "CB", "SG", "C1"]),
    "CYW": ("CYS", ["C", "N", "O", "CA", "CB", "SG", "C1", "O1", "C2", "O2", "C3"]),
    "BIF": (
        "PHE",
        [
            "N",
            "CA",
            "C",
            "CB",
            "CG",
            "CD2",
            "CE2",
            "CZ",
            "CE1",
            "CD1",
            "C8",
            "C9",
            "C10",
            "C11",
            "C12",
            "C13",
            "O",
        ],
    ),
    "5PG": (
        "GLY",
        ["N", "CA", "C", "O", "CN", "CB", "CC1", "CC2", "CD1", "CD2", "CE", "OH"],
    ),
    "4U7": (
        "ALA",
        [
            "B1",
            "C1",
            "F1",
            "N1",
            "O1",
            "C2",
            "F2",
            "O2",
            "C3",
            "O3",
            "C4",
            "O4",
            "C5",
            "O5",
            "C6",
            "C9",
        ],
    ),
    "ML3": (
        "LYS",
        ["N", "CA", "CB", "SG", "CD", "CE", "NZ", "CM1", "CM2", "CM3", "C", "O"],
    ),
    "CYA": ("CYS", ["OD1", "OD2"]),
    "POM": ("PRO", ["N", "CA", "C", "O", "CB", "CG", "CD", "OG", "CD1"]),
    "CYG": (
        "CYS",
        [
            "N",
            "CA",
            "C",
            "O",
            "CB",
            "SG",
            "N1",
            "CA1",
            "CB1",
            "CG1",
            "CD1",
            "OE2",
            "C1",
            "O1",
            "O2",
        ],
    ),
    "CYF": (
        "CYS",
        [
            "C1",
            "O1",
            "C2",
            "C3",
            "O2",
            "C4",
            "C5",
            "C6",
            "O3",
            "C7",
            "C8",
            "C9",
            "C10",
            "C11",
            "C12",
            "C13",
            "C14",
            "C15",
            "C16",
            "C17",
            "C18",
            "C19",
            "C20",
            "O4",
            "O5",
            "NL",
            "CL",
            "OL",
            "CAL",
            "SG",
            "CB",
            "CA",
            "N",
            "O",
            "C",
            "NT",
        ],
    ),
    "CYD": (
        "CYS",
        [
            "N",
            "CA",
            "CB",
            "SG",
            "C",
            "O",
            "C1",
            "N1",
            "C5",
            "C4",
            "C3",
            "C2",
            "O1",
            "C6",
            "OT3",
            "OT4",
        ],
    ),
    "BIU": ("ILE", ["N", "CA", "CB", "CG2", "CG1", "CD1", "C", "O", "BR"]),
    "CYM": ("CYS", ["N", "CA", "CB", "SG", "CD", "C", "O"]),
    "CY3": ("CYS", ["N", "CA", "C", "O", "CB", "SG", "N1"]),
    "DHN": ("VAL", ["N", "CA", "CB", "CG", "CD", "OG", "C", "O"]),
    "CY1": ("CYS", ["N", "CA", "CB", "SG", "CD", "NE", "CZ", "OAC", "CM", "C", "O"]),
    "CY0": (
        "CYS",
        [
            "CAI",
            "CAF",
            "CAE",
            "CAG",
            "CAJ",
            "CAX",
            "NAT",
            "C6",
            "C5",
            "CAM",
            "N1",
            "C2",
            "N3",
            "C4",
            "CAL",
            "CAK",
            "CAY",
            "NAS",
            "CAW",
            "OAC",
            "CAO",
            "CAN",
            "SAU",
            "CB",
            "CA",
            "C",
            "O",
            "N",
        ],
    ),
    "SLZ": ("LYS", ["N", "CA", "CB", "SG", "CD", "CE", "NZ", "C", "O"]),
    "CY4": (
        "CYS",
        ["O", "C", "CA", "N", "CB", "SG", "CA1", "CA2", "OA1", "CA3", "CA4"],
    ),
    "CWR": ("SER", ["C1", "N2", "N3", "C2", "O2", "CA2", "CA3", "C3", "O3", "CB2"]),
    "MLL": ("LEU", ["N", "CA", "CB", "CG", "CD2", "CD1", "C", "O", "O9", "C10"]),
    "IZO": ("MET", ["N", "CA", "C", "O", "CB", "CG", "CD", "CE"]),
    "SLL": (
        "LYS",
        [
            "C",
            "N",
            "O",
            "CA",
            "CB",
            "CD",
            "CE",
            "CG",
            "CK",
            "CL",
            "CP",
            "CX",
            "OX",
            "NZ",
            "OP1",
            "OP2",
        ],
    ),
    "3QN": (
        "LYS",
        [
            "C1",
            "N1",
            "O1",
            "P1",
            "C2",
            "C3",
            "C4",
            "O2",
            "C5",
            "C6",
            "C7",
            "C8",
            "C9",
            "C10",
            "C11",
            "C12",
            "N2",
            "O3",
            "C13",
            "O4",
            "O5",
            "C14",
            "O6",
            "C15",
        ],
    ),
    "MLZ": ("LYS", ["N", "CA", "CB", "CG", "CD", "CE", "NZ", "CM", "C", "O"]),
    "MLY": ("LYS", ["N", "CA", "CB", "CG", "CD", "CE", "NZ", "CH1", "CH2", "C", "O"]),
    "CWD": (
        "ALA",
        [
            "N1",
            "C2",
            "O2",
            "N3",
            "C4",
            "O4",
            "C5",
            "CL5",
            "C6",
            "C7",
            "C8",
            "N8",
            "C9",
            "O91",
            "O92",
        ],
    ),
    "OTH": ("THR", ["N", "C", "CA", "O", "OG1", "CB", "C22", "CG2", "C24"]),
    "MCL": (
        "LYS",
        [
            "N",
            "CA",
            "CB",
            "CG",
            "CD",
            "CE",
            "NZ",
            "CX1",
            "C1",
            "CX2",
            "O1",
            "O2",
            "C",
            "O",
        ],
    ),
    "GFT": (
        "SER",
        [
            "N",
            "CA",
            "C",
            "O",
            "CB",
            "OG",
            "CAA",
            "OAC",
            "CAD",
            "CAE",
            "CAF",
            "CAG",
            "CAH",
            "OAI",
            "CAJ",
            "PAK",
        ],
    ),
    "ARM": (
        "ARG",
        ["N", "CA", "CB", "CG", "CD", "NE", "CZ", "NH1", "NH2", "C", "O", "CM"],
    ),
    "AA4": ("ALA", ["N", "CA", "C", "O", "CB", "CG", "CD", "OE"]),
    "LGY": (
        "LYS",
        [
            "OXT",
            "C",
            "N",
            "O",
            "N1",
            "C3",
            "C4",
            "C5",
            "O6",
            "CA",
            "CB",
            "CD",
            "CE",
            "CG",
            "C11",
        ],
    ),
    "BP5": (
        "ALA",
        [
            "C9",
            "C8",
            "C7",
            "C6",
            "C5",
            "C4",
            "C3",
            "C2",
            "C1",
            "C12",
            "C11",
            "N1",
            "N2",
            "C13",
            "C14",
            "O1",
            "O2",
            "N",
        ],
    ),
    "AA3": ("ALA", ["N", "CA", "C", "O", "CB", "CG"]),
    "NRQ": (
        "MET",
        [
            "N1",
            "CE",
            "SD",
            "CG1",
            "CB1",
            "CA1",
            "C1",
            "N2",
            "OH",
            "CD2",
            "CE2",
            "CZ",
            "CE1",
            "CD1",
            "CG2",
            "CB2",
            "CA2",
            "C2",
            "O2",
            "N3",
            "CA3",
            "C3",
            "O3",
        ],
    ),
    "NRP": (
        "LEU",
        [
            "C",
            "N",
            "O",
            "C1",
            "C2",
            "N2",
            "O2",
            "N3",
            "OH",
            "CZ",
            "CA1",
            "CA2",
            "CA3",
            "CB1",
            "CB2",
            "CD1",
            "CD2",
            "CD3",
            "CD4",
            "CE1",
            "CE2",
            "CG1",
            "CG2",
        ],
    ),
    "MCS": (
        "CYS",
        ["O", "C", "CA", "N", "CB", "SAI", "CAK", "OAC", "CAG", "CAJ", "OAE", "OAB"],
    ),
    "ZU0": ("THR", ["N", "CA", "C", "O", "CB", "O3", "C4", "C5", "C6", "C7", "C8"]),
    "PAQ": (
        "TYR",
        [
            "N",
            "CA",
            "CB",
            "CG",
            "CD1",
            "O2",
            "CD2",
            "CE1",
            "CE2",
            "CZ",
            "OH",
            "N1",
            "N2",
            "C1",
            "C2",
            "C3",
            "C4",
            "C5",
            "N3",
            "C",
            "O",
        ],
    ),
    "PAS": (
        "ASP",
        ["N", "CA", "CB", "CG", "OD1", "P1", "O1", "O2", "O3", "O4", "C", "O"],
    ),
    "PAT": (
        "TRP",
        [
            "N",
            "CA",
            "C",
            "O",
            "CB",
            "CG",
            "CD1",
            "CD2",
            "NE1",
            "CE2",
            "CE3",
            "CZ2",
            "CZ3",
            "CH2",
            "P",
            "O1P",
            "O2P",
            "O3P",
        ],
    ),
    "3CF": (
        "PHE",
        ["C", "N", "C1", "N1", "CA", "CB", "CG", "CZ", "CD1", "CD2", "CE1", "CE2", "O"],
    ),
    "TQI": (
        "TRP",
        [
            "N",
            "CA",
            "C",
            "O",
            "CB",
            "CG",
            "CD1",
            "CD2",
            "NE1",
            "CE2",
            "CE3",
            "CZ2",
            "CZ3",
            "CH2",
            "S1",
            "C2",
            "C3",
            "C7",
            "C8",
            "O14",
            "O15",
            "C16",
            "C19",
            "F20",
            "F21",
            "F22",
            "C23",
            "CL",
        ],
    ),
    "TQQ": (
        "TRP",
        [
            "N",
            "CA",
            "C",
            "O",
            "CB",
            "CG",
            "CD1",
            "CD2",
            "NE1",
            "CE2",
            "CE3",
            "CZ2",
            "CZ3",
            "CH2",
            "O2",
            "N2",
        ],
    ),
    "M2S": ("MET", ["N", "CA", "CB", "CG1", "CG2", "CG", "SD", "OE", "CE", "C", "O"]),
    "EXY": ("LEU", ["N", "CA", "C", "O", "CB", "CG", "CD", "CE", "CF", "CH", "OF"]),
    "HSE": ("SER", ["NA", "C1", "C2", "C3", "O1", "O2", "C4", "O3"]),
    "TH6": ("THR", ["N", "CA", "CB", "OB1", "CG", "OG1", "C", "O"]),
    "TH5": ("THR", ["CAA", "CAH", "OAD", "OG1", "CB", "CG2", "CA", "C", "O", "N"]),
    "4HT": (
        "TRP",
        [
            "N",
            "CA",
            "C",
            "O",
            "CB",
            "CG",
            "CD1",
            "NE1",
            "CE2",
            "CD2",
            "CE3",
            "OE3",
            "CZ3",
            "CH2",
            "CZ2",
        ],
    ),
    "CSU": ("CYS", ["N", "CA", "CB", "SG", "S", "C", "O", "OD1", "OD2", "OD3"]),
    "HSO": ("HIS", ["N", "CA", "CB", "CG", "ND1", "CD2", "CE1", "NE2", "C", "O"]),
    "HSL": ("SER", ["N", "CA", "C", "O", "CB", "CG", "OD"]),
    "12X": (
        "PRO",
        [
            "C10",
            "O11",
            "C12",
            "C13",
            "C14",
            "C15",
            "N16",
            "O33",
            "C34",
            "C35",
            "C36",
            "N37",
            "C38",
            "C39",
            "C40",
            "C41",
            "C42",
            "C43",
            "BR44",
            "O45",
            "C46",
            "C47",
            "C48",
            "S49",
            "C50",
            "N51",
            "N52",
            "C53",
            "O54",
            "C55",
            "C56",
            "C57",
        ],
    ),
    "12Y": (
        "PRO",
        [
            "C10",
            "O11",
            "C12",
            "C13",
            "C14",
            "C15",
            "N16",
            "O33",
            "C34",
            "C35",
            "C36",
            "N37",
            "C38",
            "C39",
            "C40",
            "C41",
            "C42",
            "C43",
            "O45",
            "C46",
            "C47",
            "C48",
            "S49",
            "C50",
            "N51",
            "N52",
            "C53",
            "O54",
            "C55",
            "C56",
            "C57",
        ],
    ),
    "IAM": (
        "ALA",
        [
            "N",
            "CA",
            "CB",
            "CG",
            "CD1",
            "CE1",
            "CZ",
            "CE2",
            "CD2",
            "CT",
            "NH",
            "CI",
            "CK1",
            "CK2",
            "C",
            "O",
        ],
    ),
    "THZ": (
        "ARG",
        [
            "C",
            "N",
            "O",
            "S1",
            "CA",
            "CB",
            "CD",
            "NE",
            "CG",
            "CZ",
            "N11",
            "C17",
            "C21",
            "C31",
            "C41",
            "C51",
            "C61",
            "C71",
            "NH1",
            "NH2",
        ],
    ),
    "NLQ": (
        "GLN",
        ["OXT", "C", "O", "CA", "N", "C6", "O4", "C7", "CB", "CG", "CD", "NE2", "OE1"],
    ),
    "HAC": (
        "ALA",
        ["C1", "C2", "C3", "C4", "C5", "C6", "C7", "C8", "C9", "N8", "O1", "O2"],
    ),
    "KNB": ("ALA", ["C", "O", "CA", "CB", "S", "O2", "CS", "O1", "N"]),
    "THR": ("THR", ["N", "CA", "C", "O", "CB", "OG1", "CG2"]),
    "FVA": ("VAL", ["C", "N", "O", "CA", "CB", "CG1", "CG2", "O1", "CN"]),
    "BNN": (
        "PHE",
        [
            "N",
            "CA",
            "C",
            "O",
            "O1",
            "CH",
            "C11",
            "N16",
            "N17",
            "C1",
            "C2",
            "C3",
            "C4",
            "C5",
            "C6",
            "C7",
            "C15",
        ],
    ),
    "C1T": (
        "CYS",
        [
            "C",
            "N",
            "O",
            "C1",
            "C2",
            "C3",
            "C4",
            "C5",
            "C6",
            "C7",
            "O8",
            "O9",
            "CA",
            "CB",
            "SG",
            "O10",
            "O11",
        ],
    ),
    "LED": ("LEU", ["N", "CA", "CB", "CG", "CD1", "CD2", "OE", "C", "O"]),
    "RZ4": ("SER", ["C", "N", "O", "C1", "C2", "C3", "CA", "CB", "CE", "OG"]),
    "HS9": (
        "HIS",
        ["N", "CA", "CB", "CG", "ND1", "CD2", "CE1", "NE2", "C", "O", "NXT", "CM"],
    ),
    "DC2": (
        "CYS",
        [
            "O1",
            "CL1",
            "C2",
            "CL2",
            "C4",
            "N",
            "C6",
            "O7",
            "C8",
            "C9",
            "S10",
            "C11",
            "C12",
        ],
    ),
    "IAR": ("ARG", ["C", "N", "O", "CA", "CB", "CD", "NE", "CG", "CZ", "NH1", "NH2"]),
    "2P0": (
        "PRO",
        [
            "C",
            "N",
            "O",
            "CA",
            "CB",
            "CD",
            "CG",
            "OAB",
            "OAC",
            "OAD",
            "OAF",
            "OAG",
            "OAH",
            "CAK",
            "OAM",
            "OAN",
            "CAP",
            "PAS",
            "PAT",
        ],
    ),
    "PHI": (
        "PHE",
        ["N", "CA", "CB", "CG", "CD1", "CD2", "CE1", "CE2", "CZ", "I", "C", "O"],
    ),
    "F2F": (
        "PHE",
        ["O", "C", "CA", "N", "CB", "CG", "CD2", "CE2", "F2", "CZ", "F1", "CE1", "CD1"],
    ),
    "PHM": (
        "PHE",
        [
            "C1",
            "N1",
            "C2",
            "C3",
            "C4",
            "C5",
            "C6",
            "C7",
            "C8",
            "C9",
            "C10",
            "O1",
            "CL1",
        ],
    ),
    "PHL": ("PHE", ["N", "CA", "C", "O", "CB", "CG", "CD1", "CD2", "CE1", "CE2", "CZ"]),
    "PHA": ("PHE", ["N", "CA", "C", "O", "CB", "CG", "CD1", "CD2", "CE1", "CE2", "CZ"]),
    "PHE": ("PHE", ["N", "CA", "C", "O", "CB", "CG", "CD1", "CD2", "CE1", "CE2", "CZ"]),
    "PHD": (
        "ASP",
        ["N", "CA", "C", "O", "CB", "CG", "OD1", "OD2", "P", "OP1", "OP2", "OP3"],
    ),
    "SYS": ("CYS", ["N", "CA", "CB", "SE", "C", "O", "O2", "C4", "C5", "N2", "O3"]),
    "BG1": (
        "SER",
        [
            "N",
            "CA",
            "CB",
            "OG",
            "C",
            "O",
            "C1",
            "C2",
            "C3",
            "C4",
            "C5",
            "C6",
            "C7",
            "N1",
            "N2",
            "O1",
            "O2",
            "O3",
            "S",
        ],
    ),
    "NIY": (
        "TYR",
        [
            "N",
            "CA",
            "C",
            "O",
            "CB",
            "CG",
            "CD1",
            "CD2",
            "CE1",
            "CE2",
            "CZ",
            "OH",
            "NN",
            "O1",
            "O2",
        ],
    ),
    "TIH": ("ALA", ["N", "CA", "C", "O", "CB", "CG", "CD", "CE1", "CE2", "SD"]),
    "X2W": (
        "GLU",
        [
            "CA",
            "CB",
            "CG",
            "C3",
            "C",
            "O",
            "C1",
            "O1",
            "P",
            "O2P",
            "O1P",
            "O3P",
            "N1",
            "C2",
            "O2",
            "OE2",
        ],
    ),
    "0A0": ("ASP", ["N", "CA", "C", "O", "CB", "CG", "OD1", "OD2", "CM"]),
    "0A1": (
        "TYR",
        ["N", "CA", "CB", "CG", "CD1", "CE1", "CZ", "OH", "CM", "CE2", "CD2", "C", "O"],
    ),
    "0A2": (
        "LYS",
        [
            "C01",
            "C02",
            "N03",
            "C04",
            "C05",
            "C06",
            "C07",
            "C08",
            "N09",
            "B10",
            "O11",
            "O12",
            "C18",
            "C19",
        ],
    ),
    "MHO": ("MET", ["N", "CA", "CB", "CG", "SD", "CE", "C", "O", "OD1"]),
    "DMH": ("ASN", ["N", "CA", "CB", "CG", "OD1", "ND2", "CE1", "CE2", "C", "O"]),
    "0A9": (
        "PHE",
        ["N", "CA", "C", "CB", "CG", "CD1", "CD2", "CE1", "CE2", "CZ", "O", "CM"],
    ),
    "DMK": ("ASP", ["N", "CA", "CB", "CG1", "CG2", "CG3", "OD1", "OD2", "C", "O"]),
    "FDL": (
        "LYS",
        [
            "CAL",
            "CAD",
            "CAC",
            "CAB",
            "OAK",
            "OAA",
            "CAF",
            "CAG",
            "CAE",
            "CAH",
            "CAI",
            "CAJ",
            "NAM",
            "C",
            "O",
            "CA",
            "N",
            "CB",
            "CAS",
            "CAT",
            "CAU",
            "NAV",
            "CAW",
            "OAY",
            "CAX",
        ],
    ),
    "7JA": (
        "ILE",
        [
            "C",
            "N",
            "O",
            "CA",
            "CB",
            "C01",
            "C02",
            "C03",
            "C04",
            "C05",
            "C06",
            "C07",
            "O08",
            "C09",
            "C10",
            "C11",
            "C12",
            "C13",
            "O14",
            "CD1",
            "CG1",
            "CG2",
        ],
    ),
    "AN6": (
        "LEU",
        [
            "C'",
            "N'",
            "O'",
            "C1",
            "N1",
            "C2",
            "C3",
            "N3",
            "C4",
            "C5",
            "C6",
            "N6",
            "N7",
            "C8",
            "N9",
            "C1'",
            "N10",
            "C2'",
            "O2'",
            "C3'",
            "O3'",
            "C4'",
            "O4'",
            "C5'",
            "O5'",
            "CA'",
            "CB'",
            "CG'",
        ],
    ),
    "T11": (
        "PHE",
        [
            "F17",
            "C14",
            "F15",
            "F16",
            "C13",
            "N19",
            "N18",
            "C1",
            "C6",
            "C5",
            "C4",
            "C3",
            "C2",
            "C7",
            "CA",
            "N",
            "C",
            "O",
        ],
    ),
    "AEI": (
        "ASP",
        [
            "N",
            "CA",
            "C",
            "O",
            "CB",
            "CG2",
            "OG1",
            "CD",
            "OE1",
            "CE2",
            "CZ",
            "NH1",
            "CH2",
            "OT1",
            "OT2",
        ],
    ),
    "AN8": (
        "ALA",
        ["C", "N", "O", "CA", "CB", "CAA", "CAI", "CAJ", "CAK", "CAL", "CAM"],
    ),
    "037": ("PRO", ["C", "N", "O", "CA", "CB", "CG", "CBE", "CBF", "CD", "CD2"]),
    "LDH": ("LYS", ["N", "CA", "CB", "CG", "CD", "CE", "NZ", "CH", "C1", "C", "O"]),
    "TCR": (
        "TRP",
        [
            "N",
            "CA",
            "CB",
            "CG",
            "CD2",
            "CE2",
            "CE3",
            "CD1",
            "NE1",
            "CZ2",
            "CZ3",
            "CH2",
            "C9",
            "C",
            "O1",
        ],
    ),
    "TCQ": (
        "TYR",
        [
            "N",
            "CA",
            "C",
            "O",
            "CB",
            "C1",
            "C2",
            "O2",
            "C3",
            "C4",
            "O4",
            "C5",
            "N5",
            "C7",
            "C9",
            "C8",
            "C10",
            "C11",
            "C12",
            "C13",
            "C14",
            "C15",
            "C6",
        ],
    ),
    "GEE": ("GLY", ["C", "N", "O", "C1", "O1", "C2", "CA"]),
    "MH6": ("SER", ["N", "CA", "C", "CB", "OG", "O"]),
    "BUC": ("CYS", ["N", "CA", "CB", "SG", "SD", "C1", "C2", "C3", "C4", "C", "O"]),
    "BUG": ("VAL", ["N", "CA", "CB", "CG1", "CG2", "CG3", "C", "O"]),
    "MH1": ("HIS", ["N", "C2", "C", "NE2", "CD2", "CE1", "ND1", "CG", "CB", "CA", "O"]),
    "DDZ": ("ALA", ["N", "CA", "C", "O", "OG1", "OG2", "CB"]),
    "MHL": ("LEU", ["N", "CA", "CB", "CG", "CD1", "CD2", "C", "O", "CN", "OD3"]),
    "26B": (
        "THR",
        [
            "N",
            "CA",
            "C",
            "O",
            "C1",
            "C2",
            "O3",
            "C4",
            "C5",
            "C6",
            "C7",
            "C8",
            "O9",
            "C10",
            "C11",
            "C12",
            "C13",
            "C14",
            "C15",
            "C16",
            "C17",
            "C18",
            "C19",
            "N20",
            "C21",
            "O22",
            "N23",
            "C24",
            "C25",
            "C26",
            "C27",
            "C28",
            "C29",
            "C30",
            "C31",
            "C32",
        ],
    ),
    "0QL": ("CYS", ["N", "CA", "CB", "SG", "SD", "C1", "C2", "N3", "C", "O"]),
    "FOE": (
        "CYS",
        [
            "N",
            "CA",
            "C",
            "O",
            "CB",
            "SG",
            "C1",
            "C2",
            "O3",
            "N4",
            "C5",
            "C6",
            "C7",
            "C8",
            "C9",
            "C10",
            "C11",
            "C12",
            "C13",
            "F14",
        ],
    ),
    "LYK": ("LYS", ["C", "N", "O", "CA", "CB", "CD", "CE", "CG", "NZ"]),
    "MHU": (
        "PHE",
        [
            "N",
            "CA",
            "C",
            "O",
            "CB",
            "CG",
            "CD1",
            "CE1",
            "CZ",
            "CE2",
            "CD2",
            "NZ",
            "CZ1",
            "CZ2",
            "CM",
        ],
    ),
    "MHS": ("HIS", ["N", "CA", "C", "O", "CB", "CG", "ND1", "CD2", "CE1", "NE2", "CM"]),
    "DDE": (
        "HIS",
        [
            "N",
            "CA",
            "C",
            "O",
            "CB",
            "CG",
            "ND1",
            "CD2",
            "CE1",
            "NE2",
            "NAD",
            "CBI",
            "OAG",
            "CBW",
            "NCB",
            "CAB",
            "CAC",
            "CAA",
            "CAU",
            "CAT",
        ],
    ),
    "SXE": (
        "SER",
        [
            "N",
            "CA",
            "C",
            "O",
            "CB",
            "OG",
            "O1",
            "P1",
            "N1",
            "C1",
            "C3",
            "C2",
            "O2",
            "C4",
            "C5",
        ],
    ),
    "0AR": (
        "ARG",
        [
            "C",
            "N",
            "O",
            "CA",
            "CB",
            "CD",
            "NE",
            "CG",
            "CZ",
            "C0B",
            "O0C",
            "C1B",
            "NH1",
            "NH2",
        ],
    ),
    "OMX": (
        "TYR",
        ["N", "CA", "C", "O", "CB", "OC", "CG", "CD1", "CD2", "CE1", "CE2", "CZ", "OH"],
    ),
    "CR5": (
        "GLY",
        ["N", "CA1", "C1", "N2", "N3", "C2", "O2", "CA2", "CA3", "C", "O3", "O4", "O5"],
    ),
    "PSA": (
        "PHE",
        [
            "N",
            "CA",
            "CB",
            "CG",
            "CD1",
            "CD2",
            "CE1",
            "CE2",
            "CZ",
            "CH",
            "OH",
            "CM",
            "C",
            "O",
        ],
    ),
    "LCX": (
        "LYS",
        ["N", "CA", "C", "O", "CB", "CG", "CD", "CE", "NZ", "CX", "OX1", "OX2"],
    ),
    "DM0": (
        "LYS",
        [
            "OXT",
            "N",
            "CA",
            "CB",
            "CG",
            "CD",
            "CE",
            "NZ",
            "CH1",
            "CH2",
            "C",
            "O",
            "CM1",
            "CM2",
        ],
    ),
    "0AA": ("VAL", ["N", "CA", "C", "O", "CB", "CG1", "CG2", "O1", "C1"]),
    "0AB": ("VAL", ["N", "CA", "C", "O", "CB", "CG1", "CG2", "O1", "C1"]),
    "0AC": (
        "GLY",
        [
            "N",
            "CA",
            "CB",
            "CG",
            "CD1",
            "CD2",
            "CE1",
            "CE2",
            "CZ",
            "C1",
            "C2",
            "O",
            "CA1",
            "C",
            "O1",
        ],
    ),
    "NVA": ("VAL", ["N", "CA", "CB", "CG", "CD", "C", "O"]),
    "0AG": (
        "LEU",
        ["C3", "C4", "O2", "C1", "O1", "N", "CA", "C", "O", "CB", "CG", "CD1", "CD2"],
    ),
    "0AH": ("SER", ["N", "CA", "C", "O", "CB", "C1", "O1", "O2", "C2", "BR2"]),
    "0AK": ("ASP", ["N", "CA", "C", "O", "CB", "CG", "OD1", "OD2", "C1", "C2", "CL"]),
    "02O": (
        "ALA",
        [
            "C",
            "C41",
            "C40",
            "C42",
            "C43",
            "C37",
            "N36",
            "N38",
            "O39",
            "C1",
            "CB",
            "CA",
            "N",
            "O",
        ],
    ),
    "CSR": ("CYS", ["N", "CA", "CB", "SG", "AS", "O1", "O2", "O3", "C", "O"]),
    "9DN": (
        "ASN",
        [
            "N",
            "CA",
            "C",
            "O",
            "CB",
            "CG",
            "OD1",
            "ND2",
            "S1",
            "O2",
            "O3",
            "N1",
            "C1",
            "C2",
            "C3",
            "C4",
            "C5",
            "C6",
            "C7",
            "C8",
            "C9",
            "C10",
            "C11",
            "C12",
        ],
    ),
    "ILG": ("GLU", ["N", "CA", "C", "O", "CB", "CG", "CD", "OE1"]),
    "HNC": (
        "CYS",
        [
            "C",
            "N",
            "O",
            "CA",
            "CB",
            "SG",
            "CAA",
            "OAC",
            "OAE",
            "CAF",
            "CAH",
            "CAI",
            "CAJ",
            "CAK",
            "CAL",
            "CAP",
            "CAQ",
        ],
    ),
    "ILE": ("ILE", ["N", "CA", "C", "O", "CB", "CG1", "CG2", "CD1"]),
    "4CY": ("MET", ["C", "N", "O", "CA", "CB", "CCS", "SD", "CG", "NSC"]),
    "FGP": ("SER", ["N", "CA", "CB", "OG1", "OG2", "C", "O", "P", "O1P", "O2P", "O3P"]),
    "TMB": (
        "THR",
        [
            "N",
            "CT",
            "CN",
            "CA",
            "C",
            "O",
            "CB",
            "OG1",
            "CG2",
            "CD1",
            "CD2",
            "CE",
            "CZ",
            "CH",
        ],
    ),
    "LWY": ("PRO", ["C", "N", "P", "C1", "O1", "O2", "CA", "CB", "CD", "CG", "O"]),
    "TMD": (
        "THR",
        ["N", "CN", "CA", "C", "O", "CB", "CD1", "CD2", "CE", "CZ", "CH", "OG2", "CG1"],
    ),
    "02K": ("ALA", ["O", "CD", "CG", "CE", "CB", "CH", "N", "C", "CA"]),
    "4CF": (
        "PHE",
        ["C", "N", "O", "C1", "N1", "CA", "CB", "CG", "CZ", "CD1", "CD2", "CE1", "CE2"],
    ),
    "4PH": (
        "PHE",
        ["CA", "CB", "CG", "CD2", "CD1", "CE2", "CE1", "CZ", "C", "C33", "O", "N"],
    ),
    "9DS": (
        "GLY",
        [
            "N",
            "CA",
            "C",
            "O",
            "C6",
            "S8",
            "O9",
            "O10",
            "C11",
            "C12",
            "C13",
            "C14",
            "C15",
            "C16",
            "C17",
            "C18",
            "C19",
            "C20",
            "N21",
            "C23",
            "C22",
        ],
    ),
    "ILX": ("ILE", ["N", "CA", "C", "O", "CB", "CG1", "OG1", "CG2", "CD1", "OD1"]),
    "BCS": (
        "CYS",
        ["N", "CA", "CB", "SG", "CD", "CE", "CZ1", "CZ2", "CT1", "CT2", "CH", "C", "O"],
    ),
    "EFC": ("CYS", ["N", "CA", "CB", "SG", "SD", "C1", "C2", "F2", "C", "O"]),
    "OLD": (
        "HIS",
        [
            "N",
            "CA",
            "CB",
            "CG",
            "CD2",
            "ND1",
            "CE1",
            "NE2",
            "CH",
            "CG1",
            "CD1",
            "CE3",
            "CD3",
            "CE2",
            "CZ",
            "BR",
            "C9",
            "O2",
            "C",
            "O",
        ],
    ),
    "GPL": (
        "LYS",
        [
            "P",
            "O1P",
            "O2P",
            "O5'",
            "C5'",
            "C4'",
            "O4'",
            "C3'",
            "O3'",
            "C2'",
            "O2'",
            "C1'",
            "N9",
            "C8",
            "N7",
            "C5",
            "C6",
            "O6",
            "N1",
            "C2",
            "N2",
            "N3",
            "C4",
            "N",
            "CA",
            "CB",
            "CG",
            "CD",
            "CE",
            "NZ",
            "C",
            "O",
        ],
    ),
    "S12": (
        "SER",
        [
            "C",
            "N",
            "P",
            "OG",
            "C2",
            "O2",
            "C3",
            "O3",
            "C4",
            "O4",
            "C5",
            "C6",
            "C7",
            "C8",
            "C9",
            "CA",
            "CB",
            "C10",
            "C11",
            "O11",
            "C12",
            "C13",
            "C14",
            "O51",
            "O52",
            "O",
            "C15",
            "C16",
            "C17",
            "C18",
            "C19",
            "C20",
            "C21",
            "C22",
        ],
    ),
    "LYF": (
        "LYS",
        [
            "C",
            "N",
            "O",
            "CA",
            "CB",
            "CD",
            "CE",
            "CG",
            "NZ",
            "C11",
            "C12",
            "C13",
            "C14",
            "O15",
            "O16",
            "O17",
            "C18",
            "C20",
            "O21",
            "O22",
            "C41",
        ],
    ),
    "AR2": (
        "ARG",
        [
            "N",
            "CA",
            "CB",
            "CG",
            "CD",
            "NE",
            "CZ",
            "NH1",
            "NH2",
            "N5",
            "S5",
            "C5",
            "C61",
            "C62",
            "C63",
            "C64",
            "C65",
            "C66",
            "C1",
            "O1",
            "O2",
            "O",
            "C",
        ],
    ),
    "OLT": ("THR", ["N", "CA", "C", "O", "CB", "OG1", "CG2", "CD1"]),
    "OLZ": ("SER", ["OXT", "C", "O", "CA", "N", "CB", "OG", "CD", "CE", "NZ"]),
    "C6C": (
        "CYS",
        ["N", "CA", "C", "O", "CB", "SG", "SD", "C1", "C2", "C3", "C4", "C5", "C6"],
    ),
    "VR0": (
        "ARG",
        [
            "C1",
            "NH2",
            "O1",
            "O",
            "CG",
            "CD",
            "CB",
            "NH3",
            "NE",
            "NH1",
            "N",
            "CZ",
            "C6",
            "C",
            "CA",
        ],
    ),
    "6CW": (
        "TRP",
        [
            "N",
            "CA",
            "CB",
            "CG",
            "CD2",
            "CE3",
            "CE2",
            "NE1",
            "CD1",
            "CZ2",
            "CH2",
            "CLL",
            "CZ3",
            "C",
            "O",
        ],
    ),
    "PSH": (
        "HIS",
        [
            "N",
            "CA",
            "CB",
            "CG",
            "CD2",
            "ND1",
            "CE1",
            "NE2",
            "C",
            "O",
            "P",
            "S",
            "O1",
            "O2",
        ],
    ),
    "S1H": (
        "SER",
        [
            "N",
            "CA",
            "C",
            "O",
            "CB",
            "OG",
            "C1",
            "C2",
            "C3",
            "C4",
            "C5",
            "C6",
            "C7",
            "C8",
            "C9",
            "C10",
            "C11",
            "C12",
            "C13",
            "C14",
            "C15",
            "C16",
            "S1",
            "O1S",
            "O2S",
        ],
    ),
    "IEL": (
        "LYS",
        ["O", "C", "CA", "N", "CB", "CG'", "CG", "CD", "NE", "CZ", "NH1", "CH2"],
    ),
    "MK8": ("LEU", ["C", "N", "O", "CA", "CB", "CD", "CE", "CG", "CB1", "O2"]),
    "P3Q": (
        "TYR",
        [
            "C",
            "N",
            "O",
            "C1",
            "C2",
            "C3",
            "C4",
            "C5",
            "C6",
            "C7",
            "C8",
            "C9",
            "CA",
            "CB",
            "CG",
            "CZ",
            "C10",
            "C11",
            "C12",
            "O12",
            "OAD",
            "NAE",
            "OAZ",
            "CD1",
            "CD2",
            "CE1",
            "CE2",
        ],
    ),
    "K1R": (
        "CYS",
        [
            "N1",
            "CA1",
            "C1",
            "O11",
            "O12",
            "CB1",
            "CG1",
            "OE1",
            "N2",
            "CA2",
            "C2",
            "SAG",
            "NAF",
            "O2",
            "O21",
            "CB2",
            "SG2",
        ],
    ),
    "UM1": (
        "ALA",
        [
            "NCF",
            "CCG",
            "CCR",
            "OCS",
            "NCT",
            "CCU",
            "CCV",
            "CCW",
            "OCX",
            "NCY",
            "CCZ",
            "CDA",
            "CDB",
            "ODC",
            "ODB",
        ],
    ),
    "UM2": (
        "ALA",
        ["CCH", "CCI", "CCJ", "CCK", "NCL", "CCM", "OCN", "CCO", "CCP", "NCQ"],
    ),
    "PE1": (
        "LYS",
        [
            "N1",
            "C2",
            "C2A",
            "C3",
            "O3",
            "C4",
            "C4A",
            "C5",
            "C6",
            "C5A",
            "O4P",
            "P",
            "O1P",
            "O2P",
            "O3P",
            "N",
            "CA",
            "C",
            "O",
            "CB",
            "CG",
            "CD",
            "CE",
            "NZ",
        ],
    ),
    "NMC": ("GLY", ["N", "CA", "C", "O", "CN", "CX1", "CX2", "CX3"]),
    "SET": ("SER", ["N", "CA", "CB", "OG", "NT", "C", "O"]),
    "NMM": (
        "ARG",
        ["CAA", "NH2", "CZ", "NH1", "NE", "CD", "CG", "CB", "CA", "C", "O", "N"],
    ),
    "PLJ": ("PRO", ["CM", "O", "CG", "CB", "CD", "C", "CA", "N"]),
    "SD4": ("ASN", ["C", "N", "O", "CA", "CB", "CG", "OAD", "OD1", "ND2"]),
    "FHO": ("LYS", ["N", "CA", "CB", "CG", "CD", "NE", "OZ", "CZ", "OH", "C", "O"]),
    "AR4": (
        "GLU",
        [
            "N",
            "CA",
            "CB",
            "CG",
            "CD",
            "OE1",
            "OE2",
            "C",
            "O",
            "O5'",
            "C5'",
            "C4'",
            "O4'",
            "C1'",
            "C2'",
            "F2'",
            "C3'",
            "O3'",
            "O2'",
        ],
    ),
    "AR7": ("ARG", ["N", "CA", "C", "O", "CB", "CG", "CD", "NE", "CZ", "NH1", "NH2"]),
    "Z01": (
        "ALA",
        ["N", "CA", "CB", "N1", "C1", "C", "O", "O1", "C2", "C23", "C21", "C22"],
    ),
    "G8M": ("GLU", ["CD", "OE1", "OE2", "CG", "C", "O", "CA", "N", "CB", "C6", "C7"]),
    "HYP": ("PRO", ["N", "CA", "C", "O", "CB", "CG", "CD", "OD1"]),
    "GNC": ("GLN", ["C1", "NE2", "OE1", "O", "CG", "CB", "CD", "C", "CA", "N"]),
    "IYR": (
        "TYR",
        ["N", "CA", "CB", "CC", "CD", "CE", "IE", "CF", "OF", "CG", "CH", "C", "O"],
    ),
    "5AB": (
        "ALA",
        [
            "B1",
            "C1",
            "N1",
            "O1",
            "C2",
            "O2",
            "C3",
            "O3",
            "C4",
            "O4",
            "C5",
            "O5",
            "C6",
            "C9",
        ],
    ),
    "DIR": ("ARG", ["CA", "CB", "N", "C", "NG", "CD", "O", "NH1", "NH2", "OH1"]),
    "0CS": ("ALA", ["N", "CA", "CB", "C", "O", "SG", "OD1", "OD2", "OD3"]),
    "SOC": ("CYS", ["N", "CA", "CB", "SE", "OD1", "OD2", "C", "O"]),
    "MT2": ("MET", ["O", "C", "CA", "N", "CB", "CG", "SD", "CE", "C4+", "C5+"]),
    "GAU": ("GLU", ["N", "CA", "CB", "CG", "CD", "OE2", "OE1", "C"]),
    "PPN": (
        "PHE",
        [
            "N",
            "CA",
            "C",
            "O",
            "CB",
            "CG",
            "CD1",
            "CD2",
            "CE1",
            "CE2",
            "CZ",
            "N1",
            "O1",
            "O2",
        ],
    ),
    "03Y": ("CYS", ["CB", "C", "CA", "CM", "N", "O", "SG"]),
    "MC1": (
        "SER",
        [
            "S1",
            "C2",
            "C3",
            "N4",
            "C5",
            "C6",
            "C7",
            "O8",
            "C9",
            "C10",
            "C11",
            "O12",
            "O13",
            "N14",
            "C15",
            "O16",
            "C17",
            "C18",
            "C19",
            "C20",
            "C21",
            "C22",
            "O23",
            "C24",
            "O25",
            "C26",
            "N",
            "CA",
            "CB",
            "OG",
            "C",
            "O",
        ],
    ),
    "SOY": (
        "SER",
        [
            "O",
            "C",
            "CA",
            "N",
            "CB",
            "OG",
            "C1",
            "O2",
            "C3",
            "C4",
            "S5",
            "C6",
            "C7",
            "C8",
            "C9",
            "C10",
            "O12",
            "O11",
            "N13",
            "N14",
            "C15",
            "O16",
            "C17",
            "C18",
            "C19",
            "O20",
            "N21",
            "C22",
            "C23",
            "C28",
            "C24",
            "C25",
            "C26",
            "C27",
        ],
    ),
    "MBQ": (
        "TYR",
        [
            "N",
            "CA",
            "CB",
            "CQ1",
            "CQ2",
            "OQ2",
            "CQ3",
            "CQ4",
            "OQ4",
            "CQ5",
            "CQ6",
            "NQ5",
            "CAD",
            "CAC",
            "CAB",
            "OAA",
            "CAE",
            "OAF",
            "CAG",
            "CAM",
            "CAL",
            "CAJ",
            "CAK",
            "CAI",
            "CAH",
            "C",
            "O",
        ],
    ),
    "3NF": (
        "TYR",
        [
            "O1",
            "C2",
            "C3",
            "N4",
            "C5",
            "C6",
            "O7",
            "C9",
            "C10",
            "C11",
            "C12",
            "C13",
            "C14",
            "C15",
            "O16",
        ],
    ),
    "XX1": (
        "LYS",
        [
            "O",
            "C",
            "CA",
            "N",
            "CB",
            "CG",
            "CD",
            "CE",
            "NZ",
            "C6",
            "C5",
            "N7",
            "C8",
            "N1",
            "C2",
            "N3",
            "C4",
            "N9",
        ],
    ),
    "MTY": (
        "TYR",
        ["N", "CA", "CB", "CG", "CD1", "CE1", "CD2", "CE2", "CZ", "OH", "C", "O"],
    ),
    "0AF": (
        "TRP",
        [
            "N",
            "CA",
            "C",
            "O",
            "CB",
            "CG",
            "CD1",
            "CD2",
            "NE1",
            "CE2",
            "CE3",
            "CZ2",
            "O1",
            "CZ3",
            "CH2",
        ],
    ),
    "S2C": (
        "CYS",
        ["OXT", "N", "CA", "C", "O", "CB", "SG", "CD", "CE", "B", "O1", "O2", "O3"],
    ),
    "YOF": (
        "TYR",
        ["N", "CA", "C", "O", "CB", "CG", "CD1", "CD2", "CE1", "CE2", "CZ", "OH", "F"],
    ),
    "HPQ": (
        "PHE",
        ["N", "CA", "C", "O", "CB", "CG", "CD", "CE1", "CE2", "CZ1", "CZ2", "CH", "CM"],
    ),
    "M0H": ("CYS", ["OXT", "O", "C", "CA", "N", "CB", "SG", "CD", "OE"]),
    "TPJ": ("PRO", ["C", "O", "CA", "CB", "CG", "CD", "C16", "C17", "C18", "C19", "N"]),
    "0LF": (
        "PRO",
        [
            "N",
            "CA",
            "C",
            "O",
            "CD",
            "CG",
            "CB",
            "O26",
            "C27",
            "C28",
            "C29",
            "C30",
            "C31",
            "C32",
            "C33",
            "C34",
            "C35",
            "C36",
        ],
    ),
    "LVG": ("GLY", ["N", "CA", "C", "O", "CB", "CG1"]),
    "HPH": ("PHE", ["C", "O", "CA", "N", "CB", "CG", "CD1", "CD2", "CE1", "CE2", "CZ"]),
    "HPC": (
        "PHE",
        ["N", "CA", "C", "O", "CB", "CG", "CD1", "CD2", "CE1", "CE2", "CZ", "CH"],
    ),
    "LVN": ("VAL", ["C", "N", "O", "CA", "CB", "CG1", "OG1", "CG2"]),
    "4DP": (
        "TRP",
        [
            "OAP",
            "CAJ",
            "CAL",
            "CAG",
            "CAH",
            "NAM",
            "CAD",
            "CAC",
            "CAE",
            "CAF",
            "CAK",
            "CAI",
            "OAO",
            "NAN",
            "CB",
            "CA",
            "C",
            "O",
            "N",
        ],
    ),
    "8SP": (
        "SER",
        [
            "N",
            "P",
            "OG",
            "C1A",
            "O1A",
            "C1B",
            "O1B",
            "CB",
            "C1G",
            "O1G",
            "C2A",
            "C2B",
            "CA",
            "C2G",
            "O2G",
            "O2P",
            "C3A",
            "C3B",
            "C",
            "O",
            "C3G",
            "O3G",
            "O3P",
            "C4A",
            "C4B",
            "C5A",
            "C5B",
            "C6A",
            "C6B",
            "C7A",
            "C7B",
            "C8A",
            "C8B",
        ],
    ),
    "SDP": (
        "SER",
        [
            "N",
            "CA",
            "CB",
            "OG",
            "C",
            "O",
            "C31",
            "C32",
            "C41",
            "C42",
            "O1",
            "O2",
            "O3",
            "P",
        ],
    ),
    "RC7": (
        "HIS",
        [
            "C1",
            "N2",
            "CA2",
            "C2",
            "O2",
            "N3",
            "CA3",
            "C",
            "O",
            "CA1",
            "CB2",
            "CG2",
            "CD1",
            "CD2",
            "CE1",
            "CE2",
            "CZ",
            "OH",
            "CB1",
            "CG1",
            "ND1",
            "CD3",
            "NE1",
            "CE3",
        ],
    ),
    "PSW": ("CYS", ["N", "C", "O", "CA", "CB", "SE", "SD"]),
    "LA2": (
        "LYS",
        [
            "CB",
            "C",
            "O",
            "N",
            "CA",
            "O1",
            "C1",
            "NZ",
            "CE",
            "CD",
            "CG",
            "C2",
            "C3",
            "C4",
            "C5",
            "C6",
            "S6",
            "C7",
            "C8",
            "S8",
        ],
    ),
    "HY3": ("PRO", ["C1", "O1", "N1", "C2", "C3", "C4", "C5", "O2"]),
    "ARG": ("ARG", ["N", "CA", "C", "O", "CB", "CG", "CD", "NE", "CZ", "NH1", "NH2"]),
    "07O": (
        "CYS",
        [
            "N",
            "CA",
            "CB",
            "SG",
            "CD",
            "OD",
            "C1",
            "C1A",
            "C2",
            "C3",
            "C4",
            "C5",
            "C6",
            "C7",
            "C",
            "O",
        ],
    ),
    "0EA": (
        "TYR",
        [
            "O1",
            "C5",
            "C6",
            "C7",
            "O5",
            "C8",
            "C9",
            "C10",
            "C11",
            "O8",
            "C12",
            "N6",
            "C13",
            "C14",
            "O3",
        ],
    ),
    "ARO": (
        "ARG",
        ["N", "CA", "C", "O", "CB", "CG", "OH", "CD", "NE", "CZ", "NH1", "NH2"],
    ),
    "HBN": (
        "HIS",
        [
            "C1",
            "C2",
            "C3",
            "C4",
            "C4A",
            "C8A",
            "C8",
            "C7",
            "C6",
            "C5",
            "N1",
            "N",
            "CA",
            "C",
            "O",
            "CB",
            "CG",
            "ND1",
            "CD2",
            "CE1",
            "NE2",
        ],
    ),
    "GSU": (
        "GLU",
        [
            "C2",
            "S",
            "N",
            "CA",
            "CB",
            "CG",
            "CD",
            "OE1",
            "OE2",
            "C",
            "O",
            "N10",
            "O1S",
            "O2S",
            "O5'",
            "C5'",
            "C4'",
            "O4'",
            "C1'",
            "N9",
            "C4",
            "N3",
            "N1",
            "C6",
            "N6",
            "C5",
            "N7",
            "C8",
            "C2'",
            "O2'",
            "C3'",
            "O3'",
        ],
    ),
    "TPQ": (
        "TYR",
        [
            "N",
            "CA",
            "CB",
            "C",
            "O",
            "C1",
            "C2",
            "O2",
            "C3",
            "C4",
            "O4",
            "C5",
            "O5",
            "C6",
        ],
    ),
    "BOR": (
        "ARG",
        ["O1", "B", "O2", "CA", "N", "CB", "CG", "CD", "NE", "CZ", "NH1", "NH2"],
    ),
    "ZYJ": (
        "PRO",
        [
            "C",
            "N",
            "O",
            "N1",
            "O2",
            "C4",
            "O4",
            "O5",
            "C6",
            "C7",
            "C9",
            "CA",
            "CB",
            "CD",
            "CG",
            "C10",
            "C11",
            "C12",
            "C13",
            "C14",
            "O51",
        ],
    ),
    "ZYK": (
        "PRO",
        [
            "C",
            "N",
            "O",
            "C1",
            "N1",
            "O1",
            "C2",
            "O2",
            "C3",
            "O3",
            "C4",
            "O4",
            "C5",
            "C6",
            "C7",
            "C8",
            "C9",
            "CA",
            "CB",
            "CD",
            "CG",
            "O6",
        ],
    ),
    "CCS": ("CYS", ["N", "CA", "CB", "SG", "CD", "CE", "OZ1", "OZ2", "C", "O"]),
    "TIS": ("SER", ["N", "CA", "CB", "OG", "C", "O", "C1T", "O1T", "O2T", "C2T"]),
    "LPG": ("GLY", ["CB", "C", "O", "N", "CA", "CG"]),
    "GSC": ("GLY", ["N", "CA", "C", "O", "S", "C1", "C2"]),
    "OAS": ("SER", ["N", "CA", "CB", "OG", "C", "O", "C2A", "C1A", "OAC"]),
    "PYH": (
        "LYS",
        [
            "CAA",
            "CAP",
            "CAJ",
            "CAF",
            "NAL",
            "CAR",
            "CAO",
            "OAD",
            "NZ",
            "CE",
            "CD",
            "CG",
            "CB",
            "CA",
            "C",
            "O",
            "N",
        ],
    ),
    "DON": (
        "LEU",
        ["O", "C", "CA", "N", "CB", "C1E", "C1D", "O1J", "C1C", "N1B", "N1A"],
    ),
    "NAL": (
        "ALA",
        [
            "C1",
            "C2",
            "C3",
            "C4",
            "C4A",
            "C5",
            "C6",
            "C7",
            "C8",
            "C8A",
            "C9",
            "CA",
            "C",
            "N",
            "O",
        ],
    ),
    "HRG": (
        "ARG",
        ["N", "CA", "CB", "CG", "CG'", "CD", "NE", "CZ", "NH1", "NH2", "C", "O"],
    ),
    "LYU": (
        "LYS",
        [
            "N",
            "CA",
            "CB",
            "CG",
            "CD",
            "CE",
            "NZ",
            "C1'",
            "C2'",
            "C3'",
            "C4'",
            "C5'",
            "C6'",
            "C",
            "O",
        ],
    ),
    "OHI": (
        "HIS",
        ["CE1", "ND1", "NE2", "CD2", "CG", "CB", "CA", "N", "C", "O", "O12"],
    ),
    "LYR": (
        "LYS",
        [
            "N",
            "CA",
            "C",
            "O",
            "CB",
            "CG",
            "CD",
            "CE",
            "NZ",
            "C1",
            "C2",
            "C3",
            "C4",
            "C5",
            "C6",
            "C7",
            "C80",
            "C8",
            "C9",
            "C10",
            "C11",
            "C12",
            "C13",
            "C14",
            "C15",
            "C16",
            "C17",
            "C18",
            "C19",
        ],
    ),
    "NNH": ("ARG", ["N", "CA", "C", "O", "CB", "CG", "ND", "CE", "NH1", "NH2", "OH1"]),
    "2XA": (
        "CYS",
        [
            "N",
            "CA",
            "CB",
            "C",
            "O",
            "SG",
            "CAM",
            "CAG",
            "CAD",
            "CAH",
            "CAN",
            "CAV",
            "CAF",
            "CAL",
            "CAR",
            "CAX",
            "CAQ",
            "CAK",
            "CAE",
            "CAJ",
            "CAP",
            "CAW",
            "CAO",
            "CAI",
            "CL",
            "CAZ",
        ],
    ),
    "PIA": (
        "ALA",
        [
            "N1",
            "CA1",
            "CB1",
            "C1",
            "N2",
            "N3",
            "C2",
            "O2",
            "CA2",
            "CA3",
            "C3",
            "O3",
            "CB2",
            "CG2",
            "CD1",
            "CD2",
            "CE1",
            "CE2",
            "CZ",
            "OH",
            "O31",
        ],
    ),
    "LYX": (
        "LYS",
        [
            "O1",
            "P2",
            "O3",
            "P4",
            "O5",
            "C6",
            "C7",
            "C8",
            "C9",
            "C10",
            "O11",
            "C12",
            "N13",
            "C14",
            "C15",
            "C16",
            "N17",
            "C18",
            "C19",
            "S20",
            "C21",
            "C22",
            "C23",
            "NZ",
            "CE",
            "CD",
            "CG",
            "CB",
            "CA",
            "N",
            "C",
            "O",
            "O33",
            "O34",
            "O35",
            "O36",
            "O37",
            "O38",
            "O39",
            "C40",
            "C41",
            "C42",
            "O43",
            "P44",
            "O45",
            "O46",
            "O47",
            "C48",
            "O49",
            "C50",
            "N51",
            "C52",
            "N53",
            "C54",
            "C55",
            "N56",
            "C57",
            "N58",
            "C59",
            "N60",
            "O61",
        ],
    ),
    "S2P": (
        "ALA",
        ["CD2", "CG", "ND1", "CB", "CA", "OD2", "O", "C", "NE2", "N", "SE1"],
    ),
    "LYZ": ("LYS", ["N", "CA", "C", "O", "CB", "CG", "CD", "CE", "NZ", "OH"]),
    "MYK": (
        "LYS",
        [
            "OXT",
            "C",
            "N",
            "O",
            "CA",
            "CB",
            "CD",
            "CE",
            "CG",
            "CI",
            "CK",
            "CL",
            "CM",
            "CP",
            "CQ",
            "CR",
            "CS",
            "CT",
            "CU",
            "CV",
            "CW",
            "CX",
            "OX",
            "CY",
            "NZ",
        ],
    ),
    "MYN": ("ARG", ["CA", "N", "C", "CB", "CG", "CD", "NE", "CZ", "N1", "NZ", "O"]),
    "NAM": (
        "ALA",
        [
            "N",
            "CA",
            "CB",
            "C",
            "O",
            "NXT",
            "CG",
            "CD1",
            "CE1",
            "CZ1",
            "CE2",
            "CH1",
            "CP",
            "CH2",
            "CZ2",
            "CD2",
        ],
    ),
    "LYM": ("LYS", ["N", "CA", "C", "O", "CB", "CG", "CD", "CE", "NZ", "CM"]),
    "LYN": ("LYS", ["N", "CA", "CB", "CG", "CD", "CE", "NZ", "C", "O", "NT"]),
    "RVX": (
        "SER",
        [
            "N",
            "CA",
            "C",
            "O",
            "CB",
            "OG",
            "P1",
            "O1",
            "C1",
            "O2",
            "C2",
            "C3",
            "C5",
            "C4",
        ],
    ),
    "UF0": (
        "SER",
        [
            "C",
            "N",
            "O",
            "O1",
            "CA",
            "CB",
            "O1A",
            "C1B",
            "C1C",
            "C1D",
            "C1E",
            "N1F",
            "O1G",
            "C1H",
            "C1I",
            "C1J",
            "N1K",
            "O1L",
            "C1M",
            "C1N",
            "N1O",
            "C1P",
            "C1Q",
            "P1R",
            "O1S",
            "O1T",
            "C1U",
            "C1V",
            "O1W",
            "O1X",
            "O1Y",
        ],
    ),
    "OHS": (
        "ASP",
        ["OXT", "C", "O", "CA", "N", "CB", "CG", "OD1", "OD2", "SE", "CZ", "OW", "OZ"],
    ),
    "A8E": ("VAL", ["O", "C", "N", "BR", "CA", "CB", "CG", "CD1"]),
    "0WZ": (
        "TYR",
        [
            "C4",
            "C3",
            "N2",
            "C5",
            "N1",
            "N",
            "CA",
            "CB",
            "CG",
            "CD1",
            "CE1",
            "CZ",
            "OH",
            "CE2",
            "CD2",
            "C",
            "O1",
        ],
    ),
    "SVA": ("SER", ["N", "CA", "CB", "OG", "C", "O", "V", "O1", "O2", "O3", "O4"]),
    "APK": (
        "LYS",
        [
            "O",
            "C",
            "CA",
            "N",
            "CB",
            "CG",
            "CD",
            "CE",
            "NZ",
            "P",
            "O1P",
            "O2P",
            "O5'",
            "C5'",
            "C4'",
            "O4'",
            "C3'",
            "O3'",
            "C2'",
            "O2'",
            "C1'",
            "N9",
            "C8",
            "N7",
            "C5",
            "C4",
            "N3",
            "C2",
            "N1",
            "C6",
            "N6",
        ],
    ),
    "5HP": ("GLU", ["N", "CA", "C", "O", "CB", "CG", "CD", "OD"]),
    "SVY": (
        "SER",
        [
            "N",
            "CA",
            "C",
            "O",
            "CB",
            "OG",
            "C1",
            "C2",
            "C3",
            "C1'",
            "C2'",
            "C3'",
            "P",
            "O1P",
            "O2P",
            "O3P",
        ],
    ),
    "SVV": ("SER", ["N", "CA", "C", "O", "CB", "OG", "O3", "P1", "O2", "N4"]),
    "SVW": ("SER", ["N", "CA", "C", "O", "CB", "OG", "O3", "P1", "O4", "N2"]),
    "TYI": (
        "TYR",
        [
            "N",
            "CA",
            "CB",
            "CG",
            "CD1",
            "CE1",
            "CD2",
            "CE2",
            "CZ",
            "OH",
            "C",
            "O",
            "I1",
            "I2",
        ],
    ),
    "GLY": ("GLY", ["N", "CA", "C", "O"]),
    "ASK": ("ASP", ["N", "CA", "CB", "CG", "OD1", "OD2", "C", "O", "CM"]),
    "2CO": ("CYS", ["N", "CA", "C", "O", "CB", "SG", "OD", "OE"]),
    "UMA": (
        "ALA",
        [
            "N1",
            "C2",
            "N3",
            "C4",
            "C5",
            "C6",
            "O2",
            "O4",
            "C1B",
            "C2B",
            "O2'",
            "C3B",
            "C4B",
            "O4B",
            "O3B",
            "C5B",
            "O5B",
            "PA",
            "O1A",
            "O2A",
            "O3A",
            "PB",
            "O1B",
            "O2B",
            "O1'",
            "C1'",
            "C2'",
            "N2'",
            "C7'",
            "O7'",
            "C8'",
            "C3'",
            "O3'",
            "C4'",
            "O4'",
            "C5'",
            "O5'",
            "C6'",
            "O6'",
            "C18",
            "C19",
            "O18",
            "C20",
            "N4",
            "C21",
            "C22",
            "O19",
            "O20",
            "C23",
        ],
    ),
    "NA8": (
        "ALA",
        [
            "C11",
            "N",
            "O2",
            "C1",
            "C2",
            "C3",
            "C4",
            "C4A",
            "C5",
            "C6",
            "C7",
            "C8",
            "C8A",
            "C9",
            "C10",
        ],
    ),
    "5OH": (
        "ALA",
        ["C", "N", "O", "CA", "CB", "NP", "CQ", "NQ", "CR", "NR", "CS", "OS"],
    ),
    "CXM": ("MET", ["N", "CA", "CB", "CG", "SD", "CE", "C", "O", "CN", "ON1", "ON2"]),
    "ALA": ("ALA", ["N", "CA", "C", "O", "CB"]),
    "3GL": ("GLU", ["N", "CA", "C", "CB", "CG", "OH", "CD", "OE1", "OE2", "O"]),
    "2VA": (
        "VAL",
        [
            "CG2",
            "CB",
            "CG1",
            "CA",
            "N",
            "C",
            "O",
            "NAZ",
            "C2'",
            "C3'",
            "O3'",
            "C4'",
            "C5'",
            "O5'",
            "O4'",
            "C1'",
            "N9",
            "C8",
            "N7",
            "C5",
            "C4",
            "N3",
            "C2",
            "N1",
            "C6",
            "N6",
        ],
    ),
    "200": (
        "PHE",
        ["N", "CA", "C", "O", "CB", "CG", "CD1", "CD2", "CE1", "CE2", "CZ", "CL"],
    ),
    "DOH": ("ASP", ["N", "CA", "C", "O", "CB", "CG", "OG1", "OD1", "OD2"]),
    "FLT": (
        "TYR",
        [
            "N",
            "CA",
            "CB",
            "CG",
            "CD1",
            "CE1",
            "CD2",
            "CE2",
            "CZ",
            "OH",
            "C1",
            "C2",
            "F1",
            "O1",
            "O2",
            "O3",
            "O4",
            "C3",
            "O",
            "C",
        ],
    ),
    "M30": ("GLY", ["OXT", "C", "O", "CA", "N", "C3", "C4", "C5", "C6", "C7", "C8"]),
    "HT7": (
        "TRP",
        [
            "N",
            "CB",
            "CG",
            "CD",
            "CE2",
            "CZ3",
            "CH3",
            "CT2",
            "CH2",
            "CZ2",
            "NZ1",
            "CE1",
            "CA",
            "C",
            "O",
        ],
    ),
    "3GA": (
        "ALA",
        [
            "C2",
            "C3",
            "N",
            "CA",
            "C",
            "O",
            "CB",
            "CZ",
            "NE",
            "C1",
            "CG",
            "CD",
            "NH1",
            "NH2",
        ],
    ),
    "SMF": (
        "PHE",
        [
            "N",
            "CA",
            "CB",
            "CG",
            "CD1",
            "CE1",
            "CD2",
            "CE2",
            "CZ",
            "CH",
            "S",
            "O1",
            "O2",
            "O3",
            "C",
            "O",
        ],
    ),
    "FLA": ("ALA", ["N", "CA", "C", "O", "CB", "F1", "F2", "F3"]),
    "XSN": ("ASN", ["C", "N", "O", "CA", "CB", "CG", "OD1", "OD2", "N1"]),
    "MP8": ("PRO", ["C", "N", "O", "CA", "CB", "CD", "CE", "CG"]),
    "LLZ": (
        "LYS",
        [
            "C",
            "N",
            "O",
            "C1",
            "O1",
            "C2",
            "C3",
            "O3",
            "C4",
            "C5",
            "C6",
            "C7",
            "C8",
            "C9",
            "CA",
            "CB",
            "CD",
            "CE",
            "CG",
            "NZ",
            "C10",
            "C11",
            "C12",
            "C13",
            "C14",
            "C31",
        ],
    ),
    "PTM": (
        "TYR",
        [
            "N",
            "CA",
            "CH3",
            "C",
            "O",
            "CB",
            "CG",
            "CD1",
            "CD2",
            "CE1",
            "CE2",
            "CZ",
            "OH",
            "P",
            "O1P",
            "O2P",
            "O3P",
        ],
    ),
    "SCH": ("CYS", ["N", "CA", "CB", "SG", "SD", "CE", "C", "O"]),
    "PTH": (
        "TYR",
        [
            "N",
            "CA",
            "C",
            "O",
            "CB",
            "CG",
            "CD1",
            "CD2",
            "CE1",
            "CE2",
            "CZ",
            "CF",
            "OH",
            "P",
            "O1P",
            "O2P",
            "O3P",
            "OF",
        ],
    ),
    "SMC": ("CYS", ["N", "CA", "CB", "SG", "CS", "C", "O"]),
    "MLE": ("LEU", ["N", "CN", "CA", "CB", "CG", "CD1", "CD2", "C", "O"]),
    "MNV": ("VAL", ["N", "CN", "CA", "CB", "CG1", "CG2", "C", "O", "NT"]),
    "NWD": (
        "ALA",
        [
            "N1",
            "O1",
            "C2",
            "N2",
            "O2",
            "N3",
            "O3",
            "C4",
            "O4",
            "C5",
            "C6",
            "C7",
            "C8",
            "N8",
            "C9",
            "O91",
            "O92",
        ],
    ),
    "ALS": (
        "ALA",
        ["C1", "O1", "O2", "C2", "N2", "C3", "O3", "OS1", "S", "OS2", "OS3", "OS4"],
    ),
    "DBZ": (
        "ALA",
        [
            "N",
            "CA",
            "CB",
            "NG",
            "CD2",
            "OD2",
            "C1",
            "C2",
            "C6",
            "C3",
            "C5",
            "C4",
            "C",
            "O1",
            "O",
        ],
    ),
    "SCY": ("CYS", ["N", "CA", "CB", "SG", "CD", "OCD", "CE", "C", "O"]),
    "04V": ("PRO", ["N7", "C6", "C5", "C4", "C3", "C1", "N2"]),
    "04U": (
        "PRO",
        ["O9", "C8", "C10", "C25", "C24", "N26", "N27", "N28", "C23", "N11"],
    ),
    "0BN": (
        "PHE",
        [
            "N",
            "CA",
            "C",
            "O",
            "N2",
            "N3",
            "C3",
            "C4",
            "C5",
            "C6",
            "C7",
            "C8",
            "C9",
            "C10",
        ],
    ),
    "OSE": ("SER", ["N", "CA", "CB", "OG", "C", "O", "S", "O1S", "O2S", "O3S"]),
    "PTR": (
        "TYR",
        [
            "N",
            "CA",
            "C",
            "O",
            "CB",
            "CG",
            "CD1",
            "CD2",
            "CE1",
            "CE2",
            "CZ",
            "OH",
            "P",
            "O1P",
            "O2P",
            "O3P",
        ],
    ),
    "SCS": ("CYS", ["N", "CA", "C", "O", "CB", "SG", "SD", "CE", "CZ"]),
    "TRO": (
        "TRP",
        [
            "N",
            "CA",
            "CB",
            "CG",
            "CD1",
            "OD1",
            "NE1",
            "CE2",
            "CZ2",
            "CH2",
            "CZ3",
            "CE3",
            "CD2",
            "C",
            "O",
        ],
    ),
    "TRN": (
        "TRP",
        [
            "N",
            "CA",
            "CB",
            "CG",
            "CD1",
            "CD2",
            "NE1",
            "CE2",
            "CE3",
            "NZ2",
            "CH2",
            "CZ3",
            "C",
            "O",
        ],
    ),
    "BB6": ("CYS", ["N", "CA", "C", "O", "CB", "SG", "CD"]),
    "TRG": ("LYS", ["N", "CA", "C", "O", "CB", "CG", "CD", "CE", "NZ", "CH1", "CH2"]),
    "TRF": (
        "TRP",
        [
            "N",
            "CA",
            "CB",
            "CG",
            "CD1",
            "NE1",
            "CE2",
            "CZ2",
            "CH2",
            "CZ3",
            "CE3",
            "CD2",
            "C",
            "C1",
            "O1",
            "O",
        ],
    ),
    "TRY": (
        "TRP",
        [
            "CD2",
            "CE3",
            "CZ3",
            "CH2",
            "CZ2",
            "CE2",
            "NE1",
            "CD1",
            "CG",
            "CB",
            "CA2",
            "N2",
            "C2",
            "O2",
        ],
    ),
    "TRX": (
        "TRP",
        [
            "N",
            "CA",
            "C",
            "O",
            "CB",
            "CG",
            "CD1",
            "CD2",
            "NE1",
            "CE2",
            "CE3",
            "CZ2",
            "CZ3",
            "CH2",
            "OH2",
        ],
    ),
    "PBF": (
        "PHE",
        [
            "N",
            "C",
            "O",
            "CA",
            "CB",
            "CG",
            "CD1",
            "CD2",
            "CE1",
            "CE2",
            "CZ",
            "CN1",
            "ON2",
            "CT",
            "CI1",
            "CI2",
            "CK1",
            "CK2",
            "CL",
        ],
    ),
    "DHA": ("SER", ["N", "CA", "CB", "C", "O"]),
    "YCM": ("CYS", ["N", "CA", "CB", "SG", "CD", "CE", "OZ1", "NZ2", "C", "O"]),
    "PBB": (
        "CYS",
        [
            "CA",
            "C",
            "O",
            "N",
            "CB",
            "SG",
            "BR",
            "C1",
            "C2",
            "C3",
            "C4",
            "C5",
            "C6",
            "C7",
        ],
    ),
    "TRQ": (
        "TRP",
        [
            "N",
            "CA",
            "C",
            "O",
            "CB",
            "CG",
            "CD1",
            "NE1",
            "CE2",
            "CZ2",
            "CH2",
            "CZ3",
            "CE3",
            "CD2",
            "O6",
            "O7",
        ],
    ),
    "TRP": (
        "TRP",
        [
            "N",
            "CA",
            "C",
            "O",
            "CB",
            "CG",
            "CD1",
            "CD2",
            "NE1",
            "CE2",
            "CE3",
            "CZ2",
            "CZ3",
            "CH2",
        ],
    ),
    "TRW": (
        "TRP",
        [
            "N",
            "CA",
            "CB",
            "CG",
            "CD2",
            "CE2",
            "CE3",
            "CD1",
            "NE1",
            "CZ2",
            "CZ3",
            "CH2",
            "N6",
            "O7",
            "C",
            "O",
            "N1",
            "C1",
            "C2",
            "C3",
            "C4",
            "C5",
            "C6",
        ],
    ),
    "BCX": ("CYS", ["C", "N", "CA", "CB", "CC", "SG", "O"]),
    "4FW": (
        "TRP",
        [
            "N",
            "CA",
            "CB",
            "CG",
            "CD2",
            "CE2",
            "CE3",
            "CD1",
            "NE1",
            "CZ2",
            "CZ3",
            "F",
            "CH2",
            "C",
            "O",
        ],
    ),
    "3MY": (
        "TYR",
        [
            "OBD",
            "CZ",
            "CE2",
            "CD2",
            "CL",
            "CE1",
            "CD1",
            "CG",
            "CB",
            "CA",
            "C",
            "O",
            "N",
        ],
    ),
    "9NE": (
        "GLU",
        [
            "N",
            "CA",
            "C",
            "O",
            "CB",
            "CG",
            "CD",
            "OE1",
            "OE2",
            "S1",
            "O2",
            "O3",
            "C1",
            "C2",
            "C3",
            "C4",
            "C5",
            "C6",
            "C7",
            "C8",
            "C9",
            "C10",
            "N1",
            "C11",
            "C12",
        ],
    ),
    "9NF": (
        "PHE",
        [
            "N",
            "CA",
            "C",
            "CB",
            "CG",
            "CD1",
            "CD2",
            "CE1",
            "CE2",
            "CZ",
            "O",
            "S1",
            "O2",
            "O3",
            "N1",
            "C1",
            "C2",
            "C3",
            "C4",
            "C5",
            "C6",
            "C7",
            "C8",
            "C9",
            "C10",
            "C11",
            "C12",
        ],
    ),
    "M3R": (
        "LYS",
        [
            "CP",
            "P",
            "OPA",
            "OPB",
            "N",
            "CA",
            "C",
            "CB",
            "CG",
            "CD",
            "NE",
            "CZ",
            "NH2",
            "NH1",
            "O",
        ],
    ),
    "HTI": (
        "CYS",
        [
            "N",
            "CA",
            "CB",
            "S",
            "CAH",
            "OAI",
            "CAJ",
            "CAK",
            "NAL",
            "CAM",
            "OAN",
            "OAO",
            "C",
            "O",
        ],
    ),
    "4IN": (
        "TRP",
        [
            "N",
            "CA",
            "CB",
            "CG",
            "CD1",
            "NE1",
            "CE2",
            "CZ2",
            "CH2",
            "CZ3",
            "CE3",
            "N16",
            "CD2",
            "C",
            "O",
        ],
    ),
    "M3L": (
        "LYS",
        ["N", "CA", "CB", "CG", "CD", "CE", "NZ", "C", "O", "CM1", "CM2", "CM3"],
    ),
    "C5C": (
        "CYS",
        ["N", "CA", "C", "O", "CB", "SG", "SD", "C1", "C2", "C3", "C4", "C5"],
    ),
    "HTR": (
        "TRP",
        [
            "N",
            "CA",
            "CB",
            "OH",
            "CG",
            "CD1",
            "NE1",
            "CE2",
            "CZ2",
            "CH2",
            "CZ3",
            "CE3",
            "CD2",
            "C",
            "O",
        ],
    ),
    "MPQ": ("GLY", ["N", "CM", "CA", "CG", "CD1", "CD2", "CE1", "CE2", "CZ", "C", "O"]),
    "9NV": (
        "VAL",
        [
            "N",
            "CA",
            "C",
            "O",
            "CB",
            "CG",
            "CD",
            "S8",
            "O9",
            "O10",
            "C11",
            "C12",
            "C13",
            "C14",
            "C15",
            "C16",
            "C17",
            "C18",
            "C19",
            "C20",
            "N21",
            "C23",
            "C22",
        ],
    ),
    "3MD": ("ASP", ["N", "CB", "CG", "OD1", "OD2", "CA", "CB1", "C", "O"]),
    "KCX": (
        "LYS",
        ["N", "CA", "CB", "CG", "CD", "CE", "NZ", "C", "O", "CX", "OQ1", "OQ2"],
    ),
    "9NR": (
        "ARG",
        [
            "N",
            "CA",
            "C",
            "O",
            "CB",
            "CG",
            "CD",
            "NE",
            "CZ",
            "NH1",
            "NH2",
            "S1",
            "O2",
            "O3",
            "N1",
            "C1",
            "C2",
            "C3",
            "C4",
            "C5",
            "C6",
            "C7",
            "C8",
            "C9",
            "C10",
            "C11",
            "C12",
        ],
    ),
    "NEM": ("HIS", ["N", "CA", "C", "O", "CB", "CG", "ND1", "CD2", "CE1", "NE2", "CM"]),
    "N7P": (
        "PRO",
        ["O10", "C1", "O11", "C2", "N6", "C7", "O9", "C8", "C5", "C4", "C3"],
    ),
    "FB6": (
        "ALA",
        ["B1", "C1", "N1", "O1", "C2", "O2", "C3", "O3", "C4", "O4", "C5", "C6", "C7"],
    ),
    "FB5": (
        "ALA",
        [
            "B1",
            "F1",
            "N1",
            "O1",
            "C2",
            "F2",
            "O2",
            "C3",
            "O3",
            "C4",
            "O4",
            "C5",
            "C6",
            "C7",
            "C8",
        ],
    ),
    "2AG": ("ALA", ["C1A", "C1E", "CB", "CA", "C", "O", "N"]),
    "SUN": (
        "SER",
        [
            "N",
            "CA",
            "CB",
            "OG",
            "C",
            "O",
            "O1",
            "P1",
            "N1",
            "C2",
            "C1",
            "O2",
            "C3",
            "C4",
        ],
    ),
    "0TD": ("ASP", ["N", "CA", "C", "O", "CSB", "SB", "CB", "CG", "OD2", "OD1"]),
    "ACL": (
        "ARG",
        ["N", "CA", "CB", "CG", "CD", "NE", "CZ", "NH1", "NH2", "C", "O", "CM", "CL"],
    ),
    "LE1": ("VAL", ["O", "C", "CA", "N", "CB", "C9", "C8", "SG"]),
    "LLO": (
        "LYS",
        [
            "C",
            "N",
            "O",
            "O1",
            "C8",
            "CA",
            "CB",
            "CD",
            "CE",
            "CG",
            "NZ",
            "C11",
            "C12",
            "C13",
            "C14",
            "C16",
        ],
    ),
    "ME0": ("MET", ["C", "N", "O", "CA", "CB", "SD", "CE", "OF", "CG"]),
    "NEP": (
        "HIS",
        [
            "N",
            "CA",
            "C",
            "O",
            "CB",
            "CG",
            "ND1",
            "CD2",
            "CE1",
            "NE2",
            "P",
            "O1P",
            "O2P",
            "O3P",
        ],
    ),
    "CGA": (
        "GLU",
        ["N", "CA", "C", "O", "CB", "CG", "CD", "OE1", "OE2", "C1", "C2", "O1", "O2"],
    ),
    "NBQ": (
        "TYR",
        [
            "N",
            "CA",
            "CB",
            "CQ1",
            "CQ2",
            "OQ2",
            "CQ3",
            "CQ4",
            "OQ4",
            "CQ5",
            "CQ6",
            "NQ5",
            "CAD",
            "CAC",
            "CAB",
            "OAA",
            "CAE",
            "OAF",
            "CAG",
            "CAP",
            "CAO",
            "CAN",
            "CAM",
            "CAI",
            "CAH",
            "CAJ",
            "CAK",
            "CAL",
            "C",
            "O",
        ],
    ),
    "TEF": (
        "PHE",
        [
            "CA",
            "N",
            "CB",
            "N2",
            "O",
            "CG",
            "N3",
            "CD1",
            "N4",
            "CE1",
            "N5",
            "CZ",
            "C7",
            "CE2",
            "CD2",
            "C",
        ],
    ),
    "QMM": ("GLN", ["C", "CD", "CG", "N", "OE1", "NE2", "CH1", "CH2", "CA", "CB", "O"]),
    "ZZJ": ("ALA", ["N", "CA", "C", "O", "CB", "CM", "NXT"]),
    "56A": (
        "HIS",
        [
            "C",
            "N",
            "O",
            "CA",
            "CB",
            "CG",
            "C27",
            "C28",
            "C29",
            "C30",
            "C31",
            "C32",
            "C33",
            "C34",
            "C35",
            "C36",
            "C37",
            "C38",
            "C39",
            "C40",
            "ND1",
            "CD2",
            "CE1",
            "NE2",
        ],
    ),
    "ZZD": (
        "CYS",
        [
            "N",
            "CA",
            "CB",
            "C",
            "O",
            "SG",
            "CAM",
            "CAG",
            "CAD",
            "CAH",
            "CAN",
            "CAV",
            "CAZ",
            "CAX",
            "CAQ",
            "CAK",
            "CAF",
            "CAL",
            "CAR",
            "CAW",
            "CAO",
            "CAI",
            "CAE",
            "CAJ",
            "CAP",
        ],
    ),
    "CGU": (
        "GLU",
        ["N", "CA", "C", "O", "CB", "CG", "CD1", "CD2", "OE11", "OE12", "OE21", "OE22"],
    ),
    "PM3": (
        "PHE",
        [
            "N",
            "CA",
            "CB",
            "CG",
            "CD1",
            "CE1",
            "CD2",
            "CE2",
            "CZ",
            "CH4",
            "P",
            "O1",
            "O2",
            "O3",
            "C",
            "O",
        ],
    ),
    "DNL": ("LYS", ["CA", "C", "O", "CB", "CG", "CD", "CE", "NZ"]),
    "GMA": ("GLU", ["N", "CA", "C", "O", "CB", "CG", "CD", "OE1", "OE2", "N2"]),
    "GME": ("GLU", ["C", "N", "O", "CA", "CB", "CD", "CG", "CX", "OE1", "OE2"]),
    "ELY": (
        "LYS",
        ["C", "N", "O", "CA", "CB", "CD", "CE", "CG", "NZ", "CH1", "CH2", "CT1", "CT2"],
    ),
    "DNP": ("ALA", ["N", "CA", "CB", "NG", "C", "O"]),
    "DNS": (
        "LYS",
        [
            "C1",
            "N1",
            "C2",
            "C3",
            "C4",
            "C5",
            "C6",
            "C7",
            "C8",
            "S",
            "OA",
            "OB",
            "NZ",
            "CE",
            "CD",
            "CG",
            "CB",
            "CA",
            "C",
            "O",
            "N",
            "C9",
            "C10",
            "C11",
            "C12",
        ],
    ),
    "HR7": (
        "ARG",
        ["N", "CB", "CG", "CD", "CE", "NZ", "CH", "NH2", "NH1", "C", "O", "CA"],
    ),
    "B1F": (
        "PHE",
        ["N", "CA", "CB", "CG", "CD1", "CD2", "CE1", "CE2", "CZ", "B", "O1", "O2"],
    ),
    "MEG": ("GLU", ["N", "CA", "CB", "CG", "CD", "OE1", "OE2", "C", "O", "C1"]),
    "MEA": (
        "PHE",
        ["C1", "N", "CA", "C", "O", "CB", "CG", "CD1", "CE1", "CZ", "CE2", "CD2"],
    ),
    "LEF": ("LEU", ["CD1", "CD2", "CG", "CB", "N", "CA", "C", "O", "F1"]),
    "LEH": (
        "LEU",
        [
            "O3",
            "C7",
            "N1",
            "C6",
            "C5",
            "C4",
            "C3",
            "C2",
            "C1",
            "O2",
            "O1",
            "C8",
            "C9",
            "C10",
            "C11",
            "C12",
            "C13",
            "C14",
            "C15",
            "C16",
            "C17",
            "C18",
            "N2",
            "C20",
            "C21",
            "N3",
            "C19",
        ],
    ),
    "MEN": ("ASN", ["N", "CA", "C", "O", "CB", "CG", "OD1", "ND2", "CE2"]),
    "LEM": ("LEU", ["O", "C", "CZ", "CA", "N", "CB", "CG", "CD2", "CD1"]),
    "LEN": ("LEU", ["CD2", "CG", "CD1", "CB", "CA", "N"]),
    "MEU": ("GLY", ["CB", "N", "CA", "C", "O"]),
    "MET": ("MET", ["N", "CA", "C", "O", "CB", "CG", "SD", "CE"]),
    "MEQ": ("GLN", ["N", "CA", "CB", "CG", "CD", "OE1", "NE2", "CE", "C", "O"]),
    "LEU": ("LEU", ["N", "CA", "C", "O", "CB", "CG", "CD1", "CD2"]),
    "LEX": ("LEU", ["C", "N", "O", "CA", "CB", "CG", "CAC", "CD1", "CD2"]),
    "AZH": ("ALA", ["N", "CA", "CB", "CG", "ND", "NE", "NZ", "C", "O", "OX0"]),
    "1PI": (
        "ALA",
        [
            "C",
            "N",
            "O",
            "C1",
            "C2",
            "C3",
            "CA",
            "CB",
            "CD",
            "NE",
            "CG",
            "CZ",
            "NH1",
            "NH2",
        ],
    ),
    "1PA": (
        "PHE",
        [
            "N",
            "CA",
            "CB",
            "CG",
            "CD1",
            "CE1",
            "CD2",
            "CE2",
            "CZ",
            "CH",
            "CO",
            "O1",
            "O2",
            "C",
            "O",
        ],
    ),
    "2RA": ("ALA", ["C", "N", "O", "CA", "CB", "NG"]),
    "MSO": ("MET", ["N", "CA", "CB", "C", "CG", "SE", "OE", "CE", "O"]),
    "CIR": (
        "ARG",
        ["C1", "O1", "O2", "C2", "N2", "C3", "C4", "C5", "N6", "C7", "O7", "N8"],
    ),
    "MSL": ("MET", ["OXT", "C", "O", "CA", "N", "CB", "CG", "SD", "OE", "NE", "CE"]),
    "PF5": (
        "PHE",
        [
            "N",
            "CA",
            "CB",
            "CG",
            "CD1",
            "FD1",
            "CD2",
            "FD2",
            "CE1",
            "FE1",
            "CE2",
            "FE2",
            "CZ",
            "FZ",
            "C",
            "O",
        ],
    ),
    "SEB": (
        "SER",
        [
            "C",
            "N",
            "CI2",
            "CH2",
            "CJ",
            "CI1",
            "CH1",
            "CZ",
            "CE",
            "OD2",
            "OD1",
            "SD",
            "OG",
            "CB",
            "CA",
            "O",
        ],
    ),
    "HQA": (
        "ALA",
        [
            "OXT",
            "C",
            "O",
            "CA",
            "N",
            "C03",
            "C04",
            "C05",
            "N06",
            "C07",
            "C08",
            "O09",
            "C10",
            "C11",
            "C12",
            "C13",
            "C14",
        ],
    ),
    "MSE": ("MET", ["N", "CA", "C", "O", "CB", "CG", "SE", "CE"]),
    "AZS": ("SER", ["O", "C", "CA", "N", "CB", "OG", "CH", "CI", "NJ", "NK", "OH"]),
    "WFP": (
        "PHE",
        ["C", "N", "O", "F1", "F2", "CA", "CB", "CG", "CZ", "CD1", "CD2", "CE1", "CE2"],
    ),
    "MSA": ("GLY", ["N", "CA", "C", "O", "CN", "SB", "CG"]),
    "QIL": ("ILE", ["C", "N", "O", "CA", "CB", "CD1", "CG1", "CG2", "CN", "O1"]),
    "3AH": (
        "HIS",
        [
            "N",
            "CA",
            "CB",
            "CG",
            "CD2",
            "ND1",
            "CE1",
            "NE2",
            "C",
            "O",
            "N1",
            "N2",
            "C3",
            "N4",
            "C5",
            "N3A",
        ],
    ),
    "LP6": (
        "LYS",
        [
            "N",
            "CA",
            "CB",
            "CG",
            "CD",
            "CE",
            "NZ",
            "CM",
            "C",
            "C13",
            "O",
            "CCA",
            "CCC",
            "CCD",
        ],
    ),
    "DAB": ("ALA", ["N", "CA", "C", "O", "CB", "CG", "ND"]),
    "FLE": (
        "LEU",
        [
            "N",
            "CA",
            "CB",
            "CG",
            "CD1",
            "CD2",
            "C",
            "O",
            "C2",
            "O3",
            "C4",
            "O8",
            "C5",
            "C6",
            "C7",
        ],
    ),
    "DAH": (
        "PHE",
        [
            "N",
            "CA",
            "C",
            "O",
            "CB",
            "CG",
            "CD1",
            "CD2",
            "CE1",
            "CE2",
            "CZ",
            "OE2",
            "OZ",
        ],
    ),
    "PXU": ("PRO", ["N", "CA", "OA", "CB", "CG", "CD", "C", "O"]),
    "RGL": ("ARG", ["N", "CA", "C", "O", "CB", "CG", "CD", "NE", "CZ", "NH1", "NH2"]),
    "SGB": (
        "SER",
        ["N", "CA", "C", "O", "CB", "OG", "O1", "P1", "C1", "O2", "C2", "C4", "C3"],
    ),
    "FP9": ("PRO", ["C", "N", "O", "CA", "CB", "CD", "FD", "CG"]),
    "PKR": ("PRO", ["N", "CA", "C", "O", "CB", "CG", "CD", "C6"]),
    "TNR": (
        "SER",
        [
            "OXT",
            "C",
            "O",
            "CA",
            "N",
            "CB",
            "O1",
            "C1",
            "C2",
            "N2",
            "C7",
            "C8",
            "O7",
            "C3",
            "O3",
            "C4",
            "O4",
            "C5",
            "O5",
            "C6",
            "O6",
        ],
    ),
    "LPS": (
        "SER",
        [
            "OT2",
            "C",
            "OT1",
            "C12",
            "N",
            "C11",
            "O12",
            "P",
            "O13",
            "O14",
            "O11",
            "C1",
            "C2",
            "O21",
            "C3",
            "O31",
            "C31",
            "O32",
            "C32",
            "C33",
            "C34",
            "C35",
            "C36",
            "C37",
            "C38",
            "C39",
            "C40",
            "C41",
            "C42",
            "C43",
            "C44",
            "C45",
        ],
    ),
    "TNY": (
        "THR",
        [
            "C",
            "N",
            "O",
            "C1",
            "C2",
            "N2",
            "C3",
            "O3",
            "C4",
            "O4",
            "C5",
            "O5",
            "C6",
            "O6",
            "C7",
            "O7",
            "C8",
            "CA",
            "CB",
            "OG1",
            "CG2",
        ],
    ),
    "LLY": (
        "LYS",
        [
            "N",
            "CA",
            "CB",
            "CG",
            "CD",
            "CE",
            "NZ",
            "CH",
            "C1",
            "O1",
            "O2",
            "C2",
            "O3",
            "O4",
            "C",
            "O",
        ],
    ),
    "ICY": (
        "CYS",
        [
            "C",
            "N",
            "O",
            "CA",
            "CB",
            "SG",
            "I",
            "C0",
            "C1",
            "C2",
            "C3",
            "C4",
            "C5",
            "C6",
        ],
    ),
    "ESC": ("MET", ["N", "CA", "CB", "CG", "SD", "C", "O", "CD", "CE"]),
    "ESB": (
        "TYR",
        [
            "OX",
            "C2",
            "C1",
            "N1",
            "N",
            "CA",
            "C",
            "O",
            "CB",
            "CG",
            "CD1",
            "CD2",
            "CE1",
            "CE2",
            "CZ",
            "OH",
        ],
    ),
    "TNB": (
        "CYS",
        [
            "CA",
            "C",
            "O",
            "N",
            "CB",
            "SG",
            "C1",
            "C2",
            "C3",
            "C4",
            "C5",
            "C6",
            "N2",
            "N4",
            "N6",
            "O21",
            "O22",
            "O41",
            "O42",
            "O61",
            "O62",
        ],
    ),
    "HOX": (
        "PHE",
        ["O", "C", "N", "CA", "CB", "CG", "CZ", "NZ", "CD1", "CD2", "CE1", "CE2"],
    ),
    "PFF": (
        "PHE",
        ["N", "CA", "C", "O", "CB", "CG", "CD1", "CD2", "CE1", "CE2", "CZ", "F"],
    ),
    "ZCL": (
        "PHE",
        [
            "N",
            "O",
            "CLZ",
            "CE2",
            "CE1",
            "CLE1",
            "CD2",
            "CD1",
            "CB",
            "C",
            "CZ",
            "CG",
            "CA",
        ],
    ),
    "XW1": (
        "ALA",
        [
            "C1",
            "N",
            "CA",
            "C13",
            "C14",
            "C15",
            "C17",
            "C",
            "O",
            "O91",
            "O92",
            "C93",
            "C94",
        ],
    ),
    "KGC": (
        "LYS",
        [
            "N",
            "CA",
            "CB",
            "CG",
            "CD",
            "CE",
            "NZ",
            "C11",
            "C13",
            "O15",
            "O14",
            "O30",
            "C28",
            "O29",
            "C27",
            "C12",
            "C",
            "O1",
        ],
    ),
    "HHI": ("HIS", ["C", "N", "O", "CA", "CB", "CG", "ND1", "CD2", "CE1", "NE2"]),
    "TPK": ("PRO", ["C", "O", "C3", "O4", "CA", "CB", "CG", "CD", "N"]),
    "FZN": (
        "LYS",
        [
            "N",
            "CA",
            "CB",
            "C",
            "O",
            "C4",
            "C5",
            "N2",
            "C6",
            "C7",
            "S1",
            "C8",
            "O3",
            "C11",
            "C10",
            "O2",
            "C9",
            "O1",
            "C12",
            "O4",
            "P1",
            "O5",
            "O6",
            "O7",
            "P2",
            "O8",
            "O9",
            "O10",
            "CZ1",
            "CZ2",
            "CZ4",
            "OZ5",
            "CZ6",
            "OZ7",
            "CZ8",
            "OZ3",
            "N9",
            "C17",
            "C14",
            "N10",
            "CZ9",
            "N13",
            "C16",
            "N12",
            "C15",
            "N11",
            "C13",
        ],
    ),
}


def is_common_buffer(residue):
    """
    Is the residue a common buffer?
    If it occurs in the L{common buffers<common_buffers>} list it is considered a common buffer.

    @param residue: A AbPDB residue object or residues identifier e.g. PO4

    @return: Flag if the residue is a common buffer.
    @rtype: C{bool}
    """
    if not isinstance(residue, str):
        residue = residue.get_resname()
    residue = residue.strip()
    if residue in common_buffers:
        return True
    else:
        return False


def is_aa(residue, standard=False):
    if not isinstance(residue, str):
        residue = residue.get_resname()
    residue = residue.strip()
    if standard:
        return bio_is_aa(residue, standard)

    if get_res_type(residue) == "peptide":
        return True
    else:
        return False


def is_carbohydrate(residue):
    if get_res_type(residue) == "saccharide":
        return True
    else:
        return False


def is_nucleic_acid(residue):
    if get_res_type(residue) == "nucleic-acid":
        return True
    else:
        return False


def is_polymer(residue):

    if get_res_type(residue) == "non-polymer":
        return False
    else:
        return True


def get_res_type(residue):
    if not isinstance(residue, str):
        residue = residue.get_resname()
    residue = residue.strip()
    return get_name_type(residue)[1]


def get_chemical_name(residue):
    if not isinstance(residue, str):
        residue = residue.get_resname()
    residue = residue.strip()
    return get_name_type(residue)[0]


def get_name_type(residue):
    if not isinstance(residue, str):
        residue = residue.get_resname()
    residue = residue.strip()
    if residue in resname_to_name_and_type:
        return resname_to_name_and_type[residue]
    else:
        return get_from_expo(residue)


def get_from_expo(residue):
    """
    The PDB has a habit of updating ...therefore, if we don't have the three letter code try to get it from ligand expo database online.
    """
    if not isinstance(residue, str):
        residue = residue.get_resname()
    residue = residue.strip()
    # Get the list of files in expo database
    if residue:
        url = expourl % (residue[0], residue, residue)
    try:

        response = urllib.request.urlopen(url, timeout=1)

        # parse the file for three things.
        d = str(response.read())
        ident = chem_comp_id.search(d).groups()[0].strip().strip("\"'")
        ctype = chem_comp_type.search(d).groups()[0].strip().strip("\"'").lower()
        name = chem_comp_name.search(d).groups()[0].strip().strip("\"'")
        if ident == residue and ctype and name:
            if "non-polymer" in ctype:
                ctype = "non-polymer"
            elif "peptide" in ctype:
                ctype = "peptide"
            elif "dna" in ctype or "rna" in ctype:
                ctype = "nucleic-acid"
            elif "saccharide" in ctype:
                ctype = "saccharide"
            else:
                raise
        else:
            raise
        resname_to_name_and_type[residue] = (name, ctype)
        return name, ctype
    except Exception:  # lazy exception handling. - end result is the same anyway
        sys.stdout.write("Warning: Unknown residue code %s\n" % residue)
        return "unknown", None


def is_complete(residue, quiet=True):
    """
    Check the whether a residue has all the heavy atoms we would expect for the residue type.
    Works on the residue object in biopython pdb.
    Returns a flag (True or False) and a list of missing atom names missing
    """
    resname = residue.get_resname()  # get the 3 digit code
    if not is_aa(resname):
        if not quiet:
            sys.stderr.write(
                "Warning. Residue %s is not a recognised amino acid. Assuming complete.\n"
                % resname
            )
        return True, []
    if resname not in resname2standard:
        if not quiet:
            sys.stderr.write(
                "Warning. Residue %s does not have a standard amino acid equivalent. Assuming complete.\n"
                % resname
            )
        return True, []
    else:
        _, expected_atoms = resname2standard[resname]
        found_atoms = set([a.id for a in residue])
        if set(expected_atoms) - found_atoms:
            return False, list(set(expected_atoms) - found_atoms)
        else:
            return True, []


if __name__ == "__main__":

    if is_aa("LEU", standard=True):
        print("correct")
    if is_nucleic_acid("DA"):
        print("correct")
    if not is_polymer("GOL"):
        print("correct")

    print((get_res_type("LEU")))
